#include <tree_sitter/parser.h>

#if defined(__GNUC__) || defined(__clang__)
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wmissing-field-initializers"
#endif

#define LANGUAGE_VERSION 14
#define STATE_COUNT 2565
#define LARGE_STATE_COUNT 639
#define SYMBOL_COUNT 319
#define ALIAS_COUNT 4
#define TOKEN_COUNT 139
#define EXTERNAL_TOKEN_COUNT 4
#define FIELD_COUNT 28
#define MAX_ALIAS_SEQUENCE_LENGTH 10
#define PRODUCTION_ID_COUNT 246

enum {
  sym_identifier = 1,
  anon_sym_SEMI = 2,
  anon_sym_macro_rules_BANG = 3,
  anon_sym_LPAREN = 4,
  anon_sym_RPAREN = 5,
  anon_sym_LBRACE = 6,
  anon_sym_RBRACE = 7,
  anon_sym_EQ_GT = 8,
  anon_sym_LBRACK = 9,
  anon_sym_RBRACK = 10,
  anon_sym_COLON = 11,
  anon_sym_DOLLAR = 12,
  aux_sym_token_repetition_pattern_token1 = 13,
  anon_sym_PLUS = 14,
  anon_sym_STAR = 15,
  anon_sym_QMARK = 16,
  anon_sym_block = 17,
  anon_sym_expr = 18,
  anon_sym_ident = 19,
  anon_sym_item = 20,
  anon_sym_lifetime = 21,
  anon_sym_literal = 22,
  anon_sym_meta = 23,
  anon_sym_pat = 24,
  anon_sym_path = 25,
  anon_sym_stmt = 26,
  anon_sym_tt = 27,
  anon_sym_ty = 28,
  anon_sym_vis = 29,
  anon_sym_u8 = 30,
  anon_sym_i8 = 31,
  anon_sym_u16 = 32,
  anon_sym_i16 = 33,
  anon_sym_u32 = 34,
  anon_sym_i32 = 35,
  anon_sym_u64 = 36,
  anon_sym_i64 = 37,
  anon_sym_u128 = 38,
  anon_sym_i128 = 39,
  anon_sym_isize = 40,
  anon_sym_usize = 41,
  anon_sym_f32 = 42,
  anon_sym_f64 = 43,
  anon_sym_bool = 44,
  anon_sym_str = 45,
  anon_sym_char = 46,
  aux_sym__non_special_token_token1 = 47,
  anon_sym_SQUOTE = 48,
  anon_sym_as = 49,
  anon_sym_async = 50,
  anon_sym_await = 51,
  anon_sym_break = 52,
  anon_sym_const = 53,
  anon_sym_continue = 54,
  anon_sym_default = 55,
  anon_sym_enum = 56,
  anon_sym_fn = 57,
  anon_sym_for = 58,
  anon_sym_if = 59,
  anon_sym_impl = 60,
  anon_sym_let = 61,
  anon_sym_loop = 62,
  anon_sym_match = 63,
  anon_sym_mod = 64,
  anon_sym_pub = 65,
  anon_sym_return = 66,
  anon_sym_static = 67,
  anon_sym_struct = 68,
  anon_sym_trait = 69,
  anon_sym_type = 70,
  anon_sym_union = 71,
  anon_sym_unsafe = 72,
  anon_sym_use = 73,
  anon_sym_where = 74,
  anon_sym_while = 75,
  anon_sym_POUND = 76,
  anon_sym_BANG = 77,
  anon_sym_EQ = 78,
  anon_sym_COMMA = 79,
  anon_sym_extern = 80,
  anon_sym_ref = 81,
  anon_sym_DASH_GT = 82,
  anon_sym_LT = 83,
  anon_sym_GT = 84,
  anon_sym_else = 85,
  anon_sym_COLON_COLON = 86,
  anon_sym__ = 87,
  anon_sym_AMP = 88,
  anon_sym_DOT_DOT_DOT = 89,
  anon_sym_in = 90,
  anon_sym_LT2 = 91,
  anon_sym_dyn = 92,
  sym_mutable_specifier = 93,
  anon_sym_DOT_DOT = 94,
  anon_sym_DOT_DOT_EQ = 95,
  anon_sym_DASH = 96,
  anon_sym_AMP_AMP = 97,
  anon_sym_PIPE_PIPE = 98,
  anon_sym_PIPE = 99,
  anon_sym_CARET = 100,
  anon_sym_EQ_EQ = 101,
  anon_sym_BANG_EQ = 102,
  anon_sym_LT_EQ = 103,
  anon_sym_GT_EQ = 104,
  anon_sym_LT_LT = 105,
  anon_sym_GT_GT = 106,
  anon_sym_SLASH = 107,
  anon_sym_PERCENT = 108,
  anon_sym_PLUS_EQ = 109,
  anon_sym_DASH_EQ = 110,
  anon_sym_STAR_EQ = 111,
  anon_sym_SLASH_EQ = 112,
  anon_sym_PERCENT_EQ = 113,
  anon_sym_AMP_EQ = 114,
  anon_sym_PIPE_EQ = 115,
  anon_sym_CARET_EQ = 116,
  anon_sym_LT_LT_EQ = 117,
  anon_sym_GT_GT_EQ = 118,
  anon_sym_yield = 119,
  anon_sym_move = 120,
  anon_sym_DOT = 121,
  anon_sym_AT = 122,
  sym_integer_literal = 123,
  aux_sym_string_literal_token1 = 124,
  anon_sym_DQUOTE = 125,
  sym_char_literal = 126,
  sym_escape_sequence = 127,
  anon_sym_true = 128,
  anon_sym_false = 129,
  sym_line_comment = 130,
  sym_self = 131,
  sym_super = 132,
  sym_crate = 133,
  sym_metavariable = 134,
  sym__string_content = 135,
  sym_raw_string_literal = 136,
  sym_float_literal = 137,
  sym_block_comment = 138,
  sym_source_file = 139,
  sym__statement = 140,
  sym_empty_statement = 141,
  sym_expression_statement = 142,
  sym_macro_definition = 143,
  sym_macro_rule = 144,
  sym__token_pattern = 145,
  sym_token_tree_pattern = 146,
  sym_token_binding_pattern = 147,
  sym_token_repetition_pattern = 148,
  sym_fragment_specifier = 149,
  sym_token_tree = 150,
  sym_token_repetition = 151,
  sym_attribute_item = 152,
  sym_inner_attribute_item = 153,
  sym_attribute = 154,
  sym_mod_item = 155,
  sym_foreign_mod_item = 156,
  sym_declaration_list = 157,
  sym_struct_item = 158,
  sym_union_item = 159,
  sym_enum_item = 160,
  sym_enum_variant_list = 161,
  sym_enum_variant = 162,
  sym_field_declaration_list = 163,
  sym_field_declaration = 164,
  sym_ordered_field_declaration_list = 165,
  sym_extern_crate_declaration = 166,
  sym_const_item = 167,
  sym_static_item = 168,
  sym_type_item = 169,
  sym_function_item = 170,
  sym_function_signature_item = 171,
  sym_function_modifiers = 172,
  sym_where_clause = 173,
  sym_where_predicate = 174,
  sym_impl_item = 175,
  sym_trait_item = 176,
  sym_associated_type = 177,
  sym_trait_bounds = 178,
  sym_higher_ranked_trait_bound = 179,
  sym_removed_trait_bound = 180,
  sym_type_parameters = 181,
  sym_const_parameter = 182,
  sym_constrained_type_parameter = 183,
  sym_optional_type_parameter = 184,
  sym_let_declaration = 185,
  sym_use_declaration = 186,
  sym__use_clause = 187,
  sym_scoped_use_list = 188,
  sym_use_list = 189,
  sym_use_as_clause = 190,
  sym_use_wildcard = 191,
  sym_parameters = 192,
  sym_self_parameter = 193,
  sym_variadic_parameter = 194,
  sym_parameter = 195,
  sym_extern_modifier = 196,
  sym_visibility_modifier = 197,
  sym__type = 198,
  sym_bracketed_type = 199,
  sym_qualified_type = 200,
  sym_lifetime = 201,
  sym_array_type = 202,
  sym_for_lifetimes = 203,
  sym_function_type = 204,
  sym_tuple_type = 205,
  sym_unit_type = 206,
  sym_generic_function = 207,
  sym_generic_type = 208,
  sym_generic_type_with_turbofish = 209,
  sym_bounded_type = 210,
  sym_type_arguments = 211,
  sym_type_binding = 212,
  sym_reference_type = 213,
  sym_pointer_type = 214,
  sym_empty_type = 215,
  sym_abstract_type = 216,
  sym_dynamic_type = 217,
  sym__expression_except_range = 218,
  sym__expression = 219,
  sym_macro_invocation = 220,
  sym_delim_token_tree = 221,
  sym__delim_tokens = 222,
  sym__non_delim_token = 223,
  sym_scoped_identifier = 224,
  sym_scoped_type_identifier_in_expression_position = 225,
  sym_scoped_type_identifier = 226,
  sym_range_expression = 227,
  sym_unary_expression = 228,
  sym_try_expression = 229,
  sym_reference_expression = 230,
  sym_binary_expression = 231,
  sym_assignment_expression = 232,
  sym_compound_assignment_expr = 233,
  sym_type_cast_expression = 234,
  sym_return_expression = 235,
  sym_yield_expression = 236,
  sym_call_expression = 237,
  sym_arguments = 238,
  sym_array_expression = 239,
  sym_parenthesized_expression = 240,
  sym_tuple_expression = 241,
  sym_unit_expression = 242,
  sym_struct_expression = 243,
  sym_field_initializer_list = 244,
  sym_shorthand_field_initializer = 245,
  sym_field_initializer = 246,
  sym_base_field_initializer = 247,
  sym_if_expression = 248,
  sym_let_condition = 249,
  sym__let_chain = 250,
  sym__condition = 251,
  sym_else_clause = 252,
  sym_match_expression = 253,
  sym_match_block = 254,
  sym_match_arm = 255,
  sym_last_match_arm = 256,
  sym_match_pattern = 257,
  sym_while_expression = 258,
  sym_loop_expression = 259,
  sym_for_expression = 260,
  sym_const_block = 261,
  sym_closure_expression = 262,
  sym_closure_parameters = 263,
  sym_loop_label = 264,
  sym_break_expression = 265,
  sym_continue_expression = 266,
  sym_index_expression = 267,
  sym_await_expression = 268,
  sym_field_expression = 269,
  sym_unsafe_block = 270,
  sym_async_block = 271,
  sym_block = 272,
  sym__pattern = 273,
  sym_tuple_pattern = 274,
  sym_slice_pattern = 275,
  sym_tuple_struct_pattern = 276,
  sym_struct_pattern = 277,
  sym_field_pattern = 278,
  sym_remaining_field_pattern = 279,
  sym_mut_pattern = 280,
  sym_range_pattern = 281,
  sym_ref_pattern = 282,
  sym_captured_pattern = 283,
  sym_reference_pattern = 284,
  sym_or_pattern = 285,
  sym__literal = 286,
  sym__literal_pattern = 287,
  sym_negative_literal = 288,
  sym_string_literal = 289,
  sym_boolean_literal = 290,
  aux_sym_source_file_repeat1 = 291,
  aux_sym_macro_definition_repeat1 = 292,
  aux_sym_token_tree_pattern_repeat1 = 293,
  aux_sym_token_tree_repeat1 = 294,
  aux_sym_declaration_list_repeat1 = 295,
  aux_sym_enum_variant_list_repeat1 = 296,
  aux_sym_enum_variant_list_repeat2 = 297,
  aux_sym_field_declaration_list_repeat1 = 298,
  aux_sym_ordered_field_declaration_list_repeat1 = 299,
  aux_sym_function_modifiers_repeat1 = 300,
  aux_sym_where_clause_repeat1 = 301,
  aux_sym_trait_bounds_repeat1 = 302,
  aux_sym_type_parameters_repeat1 = 303,
  aux_sym_use_list_repeat1 = 304,
  aux_sym_parameters_repeat1 = 305,
  aux_sym_for_lifetimes_repeat1 = 306,
  aux_sym_tuple_type_repeat1 = 307,
  aux_sym_type_arguments_repeat1 = 308,
  aux_sym_delim_token_tree_repeat1 = 309,
  aux_sym_arguments_repeat1 = 310,
  aux_sym_array_expression_repeat1 = 311,
  aux_sym_tuple_expression_repeat1 = 312,
  aux_sym_field_initializer_list_repeat1 = 313,
  aux_sym_match_block_repeat1 = 314,
  aux_sym_closure_parameters_repeat1 = 315,
  aux_sym_tuple_pattern_repeat1 = 316,
  aux_sym_struct_pattern_repeat1 = 317,
  aux_sym_string_literal_repeat1 = 318,
  alias_sym_field_identifier = 319,
  alias_sym_let_chain = 320,
  alias_sym_shorthand_field_identifier = 321,
  alias_sym_type_identifier = 322,
};

static const char * const ts_symbol_names[] = {
  [ts_builtin_sym_end] = "end",
  [sym_identifier] = "identifier",
  [anon_sym_SEMI] = ";",
  [anon_sym_macro_rules_BANG] = "macro_rules!",
  [anon_sym_LPAREN] = "(",
  [anon_sym_RPAREN] = ")",
  [anon_sym_LBRACE] = "{",
  [anon_sym_RBRACE] = "}",
  [anon_sym_EQ_GT] = "=>",
  [anon_sym_LBRACK] = "[",
  [anon_sym_RBRACK] = "]",
  [anon_sym_COLON] = ":",
  [anon_sym_DOLLAR] = "$",
  [aux_sym_token_repetition_pattern_token1] = "token_repetition_pattern_token1",
  [anon_sym_PLUS] = "+",
  [anon_sym_STAR] = "*",
  [anon_sym_QMARK] = "\?",
  [anon_sym_block] = "block",
  [anon_sym_expr] = "expr",
  [anon_sym_ident] = "ident",
  [anon_sym_item] = "item",
  [anon_sym_lifetime] = "lifetime",
  [anon_sym_literal] = "literal",
  [anon_sym_meta] = "meta",
  [anon_sym_pat] = "pat",
  [anon_sym_path] = "path",
  [anon_sym_stmt] = "stmt",
  [anon_sym_tt] = "tt",
  [anon_sym_ty] = "ty",
  [anon_sym_vis] = "vis",
  [anon_sym_u8] = "primitive_type",
  [anon_sym_i8] = "primitive_type",
  [anon_sym_u16] = "primitive_type",
  [anon_sym_i16] = "primitive_type",
  [anon_sym_u32] = "primitive_type",
  [anon_sym_i32] = "primitive_type",
  [anon_sym_u64] = "primitive_type",
  [anon_sym_i64] = "primitive_type",
  [anon_sym_u128] = "primitive_type",
  [anon_sym_i128] = "primitive_type",
  [anon_sym_isize] = "primitive_type",
  [anon_sym_usize] = "primitive_type",
  [anon_sym_f32] = "primitive_type",
  [anon_sym_f64] = "primitive_type",
  [anon_sym_bool] = "primitive_type",
  [anon_sym_str] = "primitive_type",
  [anon_sym_char] = "primitive_type",
  [aux_sym__non_special_token_token1] = "_non_special_token_token1",
  [anon_sym_SQUOTE] = "'",
  [anon_sym_as] = "as",
  [anon_sym_async] = "async",
  [anon_sym_await] = "await",
  [anon_sym_break] = "break",
  [anon_sym_const] = "const",
  [anon_sym_continue] = "continue",
  [anon_sym_default] = "default",
  [anon_sym_enum] = "enum",
  [anon_sym_fn] = "fn",
  [anon_sym_for] = "for",
  [anon_sym_if] = "if",
  [anon_sym_impl] = "impl",
  [anon_sym_let] = "let",
  [anon_sym_loop] = "loop",
  [anon_sym_match] = "match",
  [anon_sym_mod] = "mod",
  [anon_sym_pub] = "pub",
  [anon_sym_return] = "return",
  [anon_sym_static] = "static",
  [anon_sym_struct] = "struct",
  [anon_sym_trait] = "trait",
  [anon_sym_type] = "type",
  [anon_sym_union] = "union",
  [anon_sym_unsafe] = "unsafe",
  [anon_sym_use] = "use",
  [anon_sym_where] = "where",
  [anon_sym_while] = "while",
  [anon_sym_POUND] = "#",
  [anon_sym_BANG] = "!",
  [anon_sym_EQ] = "=",
  [anon_sym_COMMA] = ",",
  [anon_sym_extern] = "extern",
  [anon_sym_ref] = "ref",
  [anon_sym_DASH_GT] = "->",
  [anon_sym_LT] = "<",
  [anon_sym_GT] = ">",
  [anon_sym_else] = "else",
  [anon_sym_COLON_COLON] = "::",
  [anon_sym__] = "_",
  [anon_sym_AMP] = "&",
  [anon_sym_DOT_DOT_DOT] = "...",
  [anon_sym_in] = "in",
  [anon_sym_LT2] = "<",
  [anon_sym_dyn] = "dyn",
  [sym_mutable_specifier] = "mutable_specifier",
  [anon_sym_DOT_DOT] = "..",
  [anon_sym_DOT_DOT_EQ] = "..=",
  [anon_sym_DASH] = "-",
  [anon_sym_AMP_AMP] = "&&",
  [anon_sym_PIPE_PIPE] = "||",
  [anon_sym_PIPE] = "|",
  [anon_sym_CARET] = "^",
  [anon_sym_EQ_EQ] = "==",
  [anon_sym_BANG_EQ] = "!=",
  [anon_sym_LT_EQ] = "<=",
  [anon_sym_GT_EQ] = ">=",
  [anon_sym_LT_LT] = "<<",
  [anon_sym_GT_GT] = ">>",
  [anon_sym_SLASH] = "/",
  [anon_sym_PERCENT] = "%",
  [anon_sym_PLUS_EQ] = "+=",
  [anon_sym_DASH_EQ] = "-=",
  [anon_sym_STAR_EQ] = "*=",
  [anon_sym_SLASH_EQ] = "/=",
  [anon_sym_PERCENT_EQ] = "%=",
  [anon_sym_AMP_EQ] = "&=",
  [anon_sym_PIPE_EQ] = "|=",
  [anon_sym_CARET_EQ] = "^=",
  [anon_sym_LT_LT_EQ] = "<<=",
  [anon_sym_GT_GT_EQ] = ">>=",
  [anon_sym_yield] = "yield",
  [anon_sym_move] = "move",
  [anon_sym_DOT] = ".",
  [anon_sym_AT] = "@",
  [sym_integer_literal] = "integer_literal",
  [aux_sym_string_literal_token1] = "\"",
  [anon_sym_DQUOTE] = "\"",
  [sym_char_literal] = "char_literal",
  [sym_escape_sequence] = "escape_sequence",
  [anon_sym_true] = "true",
  [anon_sym_false] = "false",
  [sym_line_comment] = "line_comment",
  [sym_self] = "self",
  [sym_super] = "super",
  [sym_crate] = "crate",
  [sym_metavariable] = "metavariable",
  [sym__string_content] = "_string_content",
  [sym_raw_string_literal] = "raw_string_literal",
  [sym_float_literal] = "float_literal",
  [sym_block_comment] = "block_comment",
  [sym_source_file] = "source_file",
  [sym__statement] = "_statement",
  [sym_empty_statement] = "empty_statement",
  [sym_expression_statement] = "expression_statement",
  [sym_macro_definition] = "macro_definition",
  [sym_macro_rule] = "macro_rule",
  [sym__token_pattern] = "_token_pattern",
  [sym_token_tree_pattern] = "token_tree_pattern",
  [sym_token_binding_pattern] = "token_binding_pattern",
  [sym_token_repetition_pattern] = "token_repetition_pattern",
  [sym_fragment_specifier] = "fragment_specifier",
  [sym_token_tree] = "token_tree",
  [sym_token_repetition] = "token_repetition",
  [sym_attribute_item] = "attribute_item",
  [sym_inner_attribute_item] = "inner_attribute_item",
  [sym_attribute] = "attribute",
  [sym_mod_item] = "mod_item",
  [sym_foreign_mod_item] = "foreign_mod_item",
  [sym_declaration_list] = "declaration_list",
  [sym_struct_item] = "struct_item",
  [sym_union_item] = "union_item",
  [sym_enum_item] = "enum_item",
  [sym_enum_variant_list] = "enum_variant_list",
  [sym_enum_variant] = "enum_variant",
  [sym_field_declaration_list] = "field_declaration_list",
  [sym_field_declaration] = "field_declaration",
  [sym_ordered_field_declaration_list] = "ordered_field_declaration_list",
  [sym_extern_crate_declaration] = "extern_crate_declaration",
  [sym_const_item] = "const_item",
  [sym_static_item] = "static_item",
  [sym_type_item] = "type_item",
  [sym_function_item] = "function_item",
  [sym_function_signature_item] = "function_signature_item",
  [sym_function_modifiers] = "function_modifiers",
  [sym_where_clause] = "where_clause",
  [sym_where_predicate] = "where_predicate",
  [sym_impl_item] = "impl_item",
  [sym_trait_item] = "trait_item",
  [sym_associated_type] = "associated_type",
  [sym_trait_bounds] = "trait_bounds",
  [sym_higher_ranked_trait_bound] = "higher_ranked_trait_bound",
  [sym_removed_trait_bound] = "removed_trait_bound",
  [sym_type_parameters] = "type_parameters",
  [sym_const_parameter] = "const_parameter",
  [sym_constrained_type_parameter] = "constrained_type_parameter",
  [sym_optional_type_parameter] = "optional_type_parameter",
  [sym_let_declaration] = "let_declaration",
  [sym_use_declaration] = "use_declaration",
  [sym__use_clause] = "_use_clause",
  [sym_scoped_use_list] = "scoped_use_list",
  [sym_use_list] = "use_list",
  [sym_use_as_clause] = "use_as_clause",
  [sym_use_wildcard] = "use_wildcard",
  [sym_parameters] = "parameters",
  [sym_self_parameter] = "self_parameter",
  [sym_variadic_parameter] = "variadic_parameter",
  [sym_parameter] = "parameter",
  [sym_extern_modifier] = "extern_modifier",
  [sym_visibility_modifier] = "visibility_modifier",
  [sym__type] = "_type",
  [sym_bracketed_type] = "bracketed_type",
  [sym_qualified_type] = "qualified_type",
  [sym_lifetime] = "lifetime",
  [sym_array_type] = "array_type",
  [sym_for_lifetimes] = "for_lifetimes",
  [sym_function_type] = "function_type",
  [sym_tuple_type] = "tuple_type",
  [sym_unit_type] = "unit_type",
  [sym_generic_function] = "generic_function",
  [sym_generic_type] = "generic_type",
  [sym_generic_type_with_turbofish] = "generic_type_with_turbofish",
  [sym_bounded_type] = "bounded_type",
  [sym_type_arguments] = "type_arguments",
  [sym_type_binding] = "type_binding",
  [sym_reference_type] = "reference_type",
  [sym_pointer_type] = "pointer_type",
  [sym_empty_type] = "empty_type",
  [sym_abstract_type] = "abstract_type",
  [sym_dynamic_type] = "dynamic_type",
  [sym__expression_except_range] = "_expression_except_range",
  [sym__expression] = "_expression",
  [sym_macro_invocation] = "macro_invocation",
  [sym_delim_token_tree] = "token_tree",
  [sym__delim_tokens] = "_delim_tokens",
  [sym__non_delim_token] = "_non_delim_token",
  [sym_scoped_identifier] = "scoped_identifier",
  [sym_scoped_type_identifier_in_expression_position] = "scoped_type_identifier",
  [sym_scoped_type_identifier] = "scoped_type_identifier",
  [sym_range_expression] = "range_expression",
  [sym_unary_expression] = "unary_expression",
  [sym_try_expression] = "try_expression",
  [sym_reference_expression] = "reference_expression",
  [sym_binary_expression] = "binary_expression",
  [sym_assignment_expression] = "assignment_expression",
  [sym_compound_assignment_expr] = "compound_assignment_expr",
  [sym_type_cast_expression] = "type_cast_expression",
  [sym_return_expression] = "return_expression",
  [sym_yield_expression] = "yield_expression",
  [sym_call_expression] = "call_expression",
  [sym_arguments] = "arguments",
  [sym_array_expression] = "array_expression",
  [sym_parenthesized_expression] = "parenthesized_expression",
  [sym_tuple_expression] = "tuple_expression",
  [sym_unit_expression] = "unit_expression",
  [sym_struct_expression] = "struct_expression",
  [sym_field_initializer_list] = "field_initializer_list",
  [sym_shorthand_field_initializer] = "shorthand_field_initializer",
  [sym_field_initializer] = "field_initializer",
  [sym_base_field_initializer] = "base_field_initializer",
  [sym_if_expression] = "if_expression",
  [sym_let_condition] = "let_condition",
  [sym__let_chain] = "_let_chain",
  [sym__condition] = "_condition",
  [sym_else_clause] = "else_clause",
  [sym_match_expression] = "match_expression",
  [sym_match_block] = "match_block",
  [sym_match_arm] = "match_arm",
  [sym_last_match_arm] = "match_arm",
  [sym_match_pattern] = "match_pattern",
  [sym_while_expression] = "while_expression",
  [sym_loop_expression] = "loop_expression",
  [sym_for_expression] = "for_expression",
  [sym_const_block] = "const_block",
  [sym_closure_expression] = "closure_expression",
  [sym_closure_parameters] = "closure_parameters",
  [sym_loop_label] = "loop_label",
  [sym_break_expression] = "break_expression",
  [sym_continue_expression] = "continue_expression",
  [sym_index_expression] = "index_expression",
  [sym_await_expression] = "await_expression",
  [sym_field_expression] = "field_expression",
  [sym_unsafe_block] = "unsafe_block",
  [sym_async_block] = "async_block",
  [sym_block] = "block",
  [sym__pattern] = "_pattern",
  [sym_tuple_pattern] = "tuple_pattern",
  [sym_slice_pattern] = "slice_pattern",
  [sym_tuple_struct_pattern] = "tuple_struct_pattern",
  [sym_struct_pattern] = "struct_pattern",
  [sym_field_pattern] = "field_pattern",
  [sym_remaining_field_pattern] = "remaining_field_pattern",
  [sym_mut_pattern] = "mut_pattern",
  [sym_range_pattern] = "range_pattern",
  [sym_ref_pattern] = "ref_pattern",
  [sym_captured_pattern] = "captured_pattern",
  [sym_reference_pattern] = "reference_pattern",
  [sym_or_pattern] = "or_pattern",
  [sym__literal] = "_literal",
  [sym__literal_pattern] = "_literal_pattern",
  [sym_negative_literal] = "negative_literal",
  [sym_string_literal] = "string_literal",
  [sym_boolean_literal] = "boolean_literal",
  [aux_sym_source_file_repeat1] = "source_file_repeat1",
  [aux_sym_macro_definition_repeat1] = "macro_definition_repeat1",
  [aux_sym_token_tree_pattern_repeat1] = "token_tree_pattern_repeat1",
  [aux_sym_token_tree_repeat1] = "token_tree_repeat1",
  [aux_sym_declaration_list_repeat1] = "declaration_list_repeat1",
  [aux_sym_enum_variant_list_repeat1] = "enum_variant_list_repeat1",
  [aux_sym_enum_variant_list_repeat2] = "enum_variant_list_repeat2",
  [aux_sym_field_declaration_list_repeat1] = "field_declaration_list_repeat1",
  [aux_sym_ordered_field_declaration_list_repeat1] = "ordered_field_declaration_list_repeat1",
  [aux_sym_function_modifiers_repeat1] = "function_modifiers_repeat1",
  [aux_sym_where_clause_repeat1] = "where_clause_repeat1",
  [aux_sym_trait_bounds_repeat1] = "trait_bounds_repeat1",
  [aux_sym_type_parameters_repeat1] = "type_parameters_repeat1",
  [aux_sym_use_list_repeat1] = "use_list_repeat1",
  [aux_sym_parameters_repeat1] = "parameters_repeat1",
  [aux_sym_for_lifetimes_repeat1] = "for_lifetimes_repeat1",
  [aux_sym_tuple_type_repeat1] = "tuple_type_repeat1",
  [aux_sym_type_arguments_repeat1] = "type_arguments_repeat1",
  [aux_sym_delim_token_tree_repeat1] = "delim_token_tree_repeat1",
  [aux_sym_arguments_repeat1] = "arguments_repeat1",
  [aux_sym_array_expression_repeat1] = "array_expression_repeat1",
  [aux_sym_tuple_expression_repeat1] = "tuple_expression_repeat1",
  [aux_sym_field_initializer_list_repeat1] = "field_initializer_list_repeat1",
  [aux_sym_match_block_repeat1] = "match_block_repeat1",
  [aux_sym_closure_parameters_repeat1] = "closure_parameters_repeat1",
  [aux_sym_tuple_pattern_repeat1] = "tuple_pattern_repeat1",
  [aux_sym_struct_pattern_repeat1] = "struct_pattern_repeat1",
  [aux_sym_string_literal_repeat1] = "string_literal_repeat1",
  [alias_sym_field_identifier] = "field_identifier",
  [alias_sym_let_chain] = "let_chain",
  [alias_sym_shorthand_field_identifier] = "shorthand_field_identifier",
  [alias_sym_type_identifier] = "type_identifier",
};

static const TSSymbol ts_symbol_map[] = {
  [ts_builtin_sym_end] = ts_builtin_sym_end,
  [sym_identifier] = sym_identifier,
  [anon_sym_SEMI] = anon_sym_SEMI,
  [anon_sym_macro_rules_BANG] = anon_sym_macro_rules_BANG,
  [anon_sym_LPAREN] = anon_sym_LPAREN,
  [anon_sym_RPAREN] = anon_sym_RPAREN,
  [anon_sym_LBRACE] = anon_sym_LBRACE,
  [anon_sym_RBRACE] = anon_sym_RBRACE,
  [anon_sym_EQ_GT] = anon_sym_EQ_GT,
  [anon_sym_LBRACK] = anon_sym_LBRACK,
  [anon_sym_RBRACK] = anon_sym_RBRACK,
  [anon_sym_COLON] = anon_sym_COLON,
  [anon_sym_DOLLAR] = anon_sym_DOLLAR,
  [aux_sym_token_repetition_pattern_token1] = aux_sym_token_repetition_pattern_token1,
  [anon_sym_PLUS] = anon_sym_PLUS,
  [anon_sym_STAR] = anon_sym_STAR,
  [anon_sym_QMARK] = anon_sym_QMARK,
  [anon_sym_block] = anon_sym_block,
  [anon_sym_expr] = anon_sym_expr,
  [anon_sym_ident] = anon_sym_ident,
  [anon_sym_item] = anon_sym_item,
  [anon_sym_lifetime] = anon_sym_lifetime,
  [anon_sym_literal] = anon_sym_literal,
  [anon_sym_meta] = anon_sym_meta,
  [anon_sym_pat] = anon_sym_pat,
  [anon_sym_path] = anon_sym_path,
  [anon_sym_stmt] = anon_sym_stmt,
  [anon_sym_tt] = anon_sym_tt,
  [anon_sym_ty] = anon_sym_ty,
  [anon_sym_vis] = anon_sym_vis,
  [anon_sym_u8] = anon_sym_u8,
  [anon_sym_i8] = anon_sym_u8,
  [anon_sym_u16] = anon_sym_u8,
  [anon_sym_i16] = anon_sym_u8,
  [anon_sym_u32] = anon_sym_u8,
  [anon_sym_i32] = anon_sym_u8,
  [anon_sym_u64] = anon_sym_u8,
  [anon_sym_i64] = anon_sym_u8,
  [anon_sym_u128] = anon_sym_u8,
  [anon_sym_i128] = anon_sym_u8,
  [anon_sym_isize] = anon_sym_u8,
  [anon_sym_usize] = anon_sym_u8,
  [anon_sym_f32] = anon_sym_u8,
  [anon_sym_f64] = anon_sym_u8,
  [anon_sym_bool] = anon_sym_u8,
  [anon_sym_str] = anon_sym_u8,
  [anon_sym_char] = anon_sym_u8,
  [aux_sym__non_special_token_token1] = aux_sym__non_special_token_token1,
  [anon_sym_SQUOTE] = anon_sym_SQUOTE,
  [anon_sym_as] = anon_sym_as,
  [anon_sym_async] = anon_sym_async,
  [anon_sym_await] = anon_sym_await,
  [anon_sym_break] = anon_sym_break,
  [anon_sym_const] = anon_sym_const,
  [anon_sym_continue] = anon_sym_continue,
  [anon_sym_default] = anon_sym_default,
  [anon_sym_enum] = anon_sym_enum,
  [anon_sym_fn] = anon_sym_fn,
  [anon_sym_for] = anon_sym_for,
  [anon_sym_if] = anon_sym_if,
  [anon_sym_impl] = anon_sym_impl,
  [anon_sym_let] = anon_sym_let,
  [anon_sym_loop] = anon_sym_loop,
  [anon_sym_match] = anon_sym_match,
  [anon_sym_mod] = anon_sym_mod,
  [anon_sym_pub] = anon_sym_pub,
  [anon_sym_return] = anon_sym_return,
  [anon_sym_static] = anon_sym_static,
  [anon_sym_struct] = anon_sym_struct,
  [anon_sym_trait] = anon_sym_trait,
  [anon_sym_type] = anon_sym_type,
  [anon_sym_union] = anon_sym_union,
  [anon_sym_unsafe] = anon_sym_unsafe,
  [anon_sym_use] = anon_sym_use,
  [anon_sym_where] = anon_sym_where,
  [anon_sym_while] = anon_sym_while,
  [anon_sym_POUND] = anon_sym_POUND,
  [anon_sym_BANG] = anon_sym_BANG,
  [anon_sym_EQ] = anon_sym_EQ,
  [anon_sym_COMMA] = anon_sym_COMMA,
  [anon_sym_extern] = anon_sym_extern,
  [anon_sym_ref] = anon_sym_ref,
  [anon_sym_DASH_GT] = anon_sym_DASH_GT,
  [anon_sym_LT] = anon_sym_LT,
  [anon_sym_GT] = anon_sym_GT,
  [anon_sym_else] = anon_sym_else,
  [anon_sym_COLON_COLON] = anon_sym_COLON_COLON,
  [anon_sym__] = anon_sym__,
  [anon_sym_AMP] = anon_sym_AMP,
  [anon_sym_DOT_DOT_DOT] = anon_sym_DOT_DOT_DOT,
  [anon_sym_in] = anon_sym_in,
  [anon_sym_LT2] = anon_sym_LT,
  [anon_sym_dyn] = anon_sym_dyn,
  [sym_mutable_specifier] = sym_mutable_specifier,
  [anon_sym_DOT_DOT] = anon_sym_DOT_DOT,
  [anon_sym_DOT_DOT_EQ] = anon_sym_DOT_DOT_EQ,
  [anon_sym_DASH] = anon_sym_DASH,
  [anon_sym_AMP_AMP] = anon_sym_AMP_AMP,
  [anon_sym_PIPE_PIPE] = anon_sym_PIPE_PIPE,
  [anon_sym_PIPE] = anon_sym_PIPE,
  [anon_sym_CARET] = anon_sym_CARET,
  [anon_sym_EQ_EQ] = anon_sym_EQ_EQ,
  [anon_sym_BANG_EQ] = anon_sym_BANG_EQ,
  [anon_sym_LT_EQ] = anon_sym_LT_EQ,
  [anon_sym_GT_EQ] = anon_sym_GT_EQ,
  [anon_sym_LT_LT] = anon_sym_LT_LT,
  [anon_sym_GT_GT] = anon_sym_GT_GT,
  [anon_sym_SLASH] = anon_sym_SLASH,
  [anon_sym_PERCENT] = anon_sym_PERCENT,
  [anon_sym_PLUS_EQ] = anon_sym_PLUS_EQ,
  [anon_sym_DASH_EQ] = anon_sym_DASH_EQ,
  [anon_sym_STAR_EQ] = anon_sym_STAR_EQ,
  [anon_sym_SLASH_EQ] = anon_sym_SLASH_EQ,
  [anon_sym_PERCENT_EQ] = anon_sym_PERCENT_EQ,
  [anon_sym_AMP_EQ] = anon_sym_AMP_EQ,
  [anon_sym_PIPE_EQ] = anon_sym_PIPE_EQ,
  [anon_sym_CARET_EQ] = anon_sym_CARET_EQ,
  [anon_sym_LT_LT_EQ] = anon_sym_LT_LT_EQ,
  [anon_sym_GT_GT_EQ] = anon_sym_GT_GT_EQ,
  [anon_sym_yield] = anon_sym_yield,
  [anon_sym_move] = anon_sym_move,
  [anon_sym_DOT] = anon_sym_DOT,
  [anon_sym_AT] = anon_sym_AT,
  [sym_integer_literal] = sym_integer_literal,
  [aux_sym_string_literal_token1] = anon_sym_DQUOTE,
  [anon_sym_DQUOTE] = anon_sym_DQUOTE,
  [sym_char_literal] = sym_char_literal,
  [sym_escape_sequence] = sym_escape_sequence,
  [anon_sym_true] = anon_sym_true,
  [anon_sym_false] = anon_sym_false,
  [sym_line_comment] = sym_line_comment,
  [sym_self] = sym_self,
  [sym_super] = sym_super,
  [sym_crate] = sym_crate,
  [sym_metavariable] = sym_metavariable,
  [sym__string_content] = sym__string_content,
  [sym_raw_string_literal] = sym_raw_string_literal,
  [sym_float_literal] = sym_float_literal,
  [sym_block_comment] = sym_block_comment,
  [sym_source_file] = sym_source_file,
  [sym__statement] = sym__statement,
  [sym_empty_statement] = sym_empty_statement,
  [sym_expression_statement] = sym_expression_statement,
  [sym_macro_definition] = sym_macro_definition,
  [sym_macro_rule] = sym_macro_rule,
  [sym__token_pattern] = sym__token_pattern,
  [sym_token_tree_pattern] = sym_token_tree_pattern,
  [sym_token_binding_pattern] = sym_token_binding_pattern,
  [sym_token_repetition_pattern] = sym_token_repetition_pattern,
  [sym_fragment_specifier] = sym_fragment_specifier,
  [sym_token_tree] = sym_token_tree,
  [sym_token_repetition] = sym_token_repetition,
  [sym_attribute_item] = sym_attribute_item,
  [sym_inner_attribute_item] = sym_inner_attribute_item,
  [sym_attribute] = sym_attribute,
  [sym_mod_item] = sym_mod_item,
  [sym_foreign_mod_item] = sym_foreign_mod_item,
  [sym_declaration_list] = sym_declaration_list,
  [sym_struct_item] = sym_struct_item,
  [sym_union_item] = sym_union_item,
  [sym_enum_item] = sym_enum_item,
  [sym_enum_variant_list] = sym_enum_variant_list,
  [sym_enum_variant] = sym_enum_variant,
  [sym_field_declaration_list] = sym_field_declaration_list,
  [sym_field_declaration] = sym_field_declaration,
  [sym_ordered_field_declaration_list] = sym_ordered_field_declaration_list,
  [sym_extern_crate_declaration] = sym_extern_crate_declaration,
  [sym_const_item] = sym_const_item,
  [sym_static_item] = sym_static_item,
  [sym_type_item] = sym_type_item,
  [sym_function_item] = sym_function_item,
  [sym_function_signature_item] = sym_function_signature_item,
  [sym_function_modifiers] = sym_function_modifiers,
  [sym_where_clause] = sym_where_clause,
  [sym_where_predicate] = sym_where_predicate,
  [sym_impl_item] = sym_impl_item,
  [sym_trait_item] = sym_trait_item,
  [sym_associated_type] = sym_associated_type,
  [sym_trait_bounds] = sym_trait_bounds,
  [sym_higher_ranked_trait_bound] = sym_higher_ranked_trait_bound,
  [sym_removed_trait_bound] = sym_removed_trait_bound,
  [sym_type_parameters] = sym_type_parameters,
  [sym_const_parameter] = sym_const_parameter,
  [sym_constrained_type_parameter] = sym_constrained_type_parameter,
  [sym_optional_type_parameter] = sym_optional_type_parameter,
  [sym_let_declaration] = sym_let_declaration,
  [sym_use_declaration] = sym_use_declaration,
  [sym__use_clause] = sym__use_clause,
  [sym_scoped_use_list] = sym_scoped_use_list,
  [sym_use_list] = sym_use_list,
  [sym_use_as_clause] = sym_use_as_clause,
  [sym_use_wildcard] = sym_use_wildcard,
  [sym_parameters] = sym_parameters,
  [sym_self_parameter] = sym_self_parameter,
  [sym_variadic_parameter] = sym_variadic_parameter,
  [sym_parameter] = sym_parameter,
  [sym_extern_modifier] = sym_extern_modifier,
  [sym_visibility_modifier] = sym_visibility_modifier,
  [sym__type] = sym__type,
  [sym_bracketed_type] = sym_bracketed_type,
  [sym_qualified_type] = sym_qualified_type,
  [sym_lifetime] = sym_lifetime,
  [sym_array_type] = sym_array_type,
  [sym_for_lifetimes] = sym_for_lifetimes,
  [sym_function_type] = sym_function_type,
  [sym_tuple_type] = sym_tuple_type,
  [sym_unit_type] = sym_unit_type,
  [sym_generic_function] = sym_generic_function,
  [sym_generic_type] = sym_generic_type,
  [sym_generic_type_with_turbofish] = sym_generic_type_with_turbofish,
  [sym_bounded_type] = sym_bounded_type,
  [sym_type_arguments] = sym_type_arguments,
  [sym_type_binding] = sym_type_binding,
  [sym_reference_type] = sym_reference_type,
  [sym_pointer_type] = sym_pointer_type,
  [sym_empty_type] = sym_empty_type,
  [sym_abstract_type] = sym_abstract_type,
  [sym_dynamic_type] = sym_dynamic_type,
  [sym__expression_except_range] = sym__expression_except_range,
  [sym__expression] = sym__expression,
  [sym_macro_invocation] = sym_macro_invocation,
  [sym_delim_token_tree] = sym_token_tree,
  [sym__delim_tokens] = sym__delim_tokens,
  [sym__non_delim_token] = sym__non_delim_token,
  [sym_scoped_identifier] = sym_scoped_identifier,
  [sym_scoped_type_identifier_in_expression_position] = sym_scoped_type_identifier,
  [sym_scoped_type_identifier] = sym_scoped_type_identifier,
  [sym_range_expression] = sym_range_expression,
  [sym_unary_expression] = sym_unary_expression,
  [sym_try_expression] = sym_try_expression,
  [sym_reference_expression] = sym_reference_expression,
  [sym_binary_expression] = sym_binary_expression,
  [sym_assignment_expression] = sym_assignment_expression,
  [sym_compound_assignment_expr] = sym_compound_assignment_expr,
  [sym_type_cast_expression] = sym_type_cast_expression,
  [sym_return_expression] = sym_return_expression,
  [sym_yield_expression] = sym_yield_expression,
  [sym_call_expression] = sym_call_expression,
  [sym_arguments] = sym_arguments,
  [sym_array_expression] = sym_array_expression,
  [sym_parenthesized_expression] = sym_parenthesized_expression,
  [sym_tuple_expression] = sym_tuple_expression,
  [sym_unit_expression] = sym_unit_expression,
  [sym_struct_expression] = sym_struct_expression,
  [sym_field_initializer_list] = sym_field_initializer_list,
  [sym_shorthand_field_initializer] = sym_shorthand_field_initializer,
  [sym_field_initializer] = sym_field_initializer,
  [sym_base_field_initializer] = sym_base_field_initializer,
  [sym_if_expression] = sym_if_expression,
  [sym_let_condition] = sym_let_condition,
  [sym__let_chain] = sym__let_chain,
  [sym__condition] = sym__condition,
  [sym_else_clause] = sym_else_clause,
  [sym_match_expression] = sym_match_expression,
  [sym_match_block] = sym_match_block,
  [sym_match_arm] = sym_match_arm,
  [sym_last_match_arm] = sym_match_arm,
  [sym_match_pattern] = sym_match_pattern,
  [sym_while_expression] = sym_while_expression,
  [sym_loop_expression] = sym_loop_expression,
  [sym_for_expression] = sym_for_expression,
  [sym_const_block] = sym_const_block,
  [sym_closure_expression] = sym_closure_expression,
  [sym_closure_parameters] = sym_closure_parameters,
  [sym_loop_label] = sym_loop_label,
  [sym_break_expression] = sym_break_expression,
  [sym_continue_expression] = sym_continue_expression,
  [sym_index_expression] = sym_index_expression,
  [sym_await_expression] = sym_await_expression,
  [sym_field_expression] = sym_field_expression,
  [sym_unsafe_block] = sym_unsafe_block,
  [sym_async_block] = sym_async_block,
  [sym_block] = sym_block,
  [sym__pattern] = sym__pattern,
  [sym_tuple_pattern] = sym_tuple_pattern,
  [sym_slice_pattern] = sym_slice_pattern,
  [sym_tuple_struct_pattern] = sym_tuple_struct_pattern,
  [sym_struct_pattern] = sym_struct_pattern,
  [sym_field_pattern] = sym_field_pattern,
  [sym_remaining_field_pattern] = sym_remaining_field_pattern,
  [sym_mut_pattern] = sym_mut_pattern,
  [sym_range_pattern] = sym_range_pattern,
  [sym_ref_pattern] = sym_ref_pattern,
  [sym_captured_pattern] = sym_captured_pattern,
  [sym_reference_pattern] = sym_reference_pattern,
  [sym_or_pattern] = sym_or_pattern,
  [sym__literal] = sym__literal,
  [sym__literal_pattern] = sym__literal_pattern,
  [sym_negative_literal] = sym_negative_literal,
  [sym_string_literal] = sym_string_literal,
  [sym_boolean_literal] = sym_boolean_literal,
  [aux_sym_source_file_repeat1] = aux_sym_source_file_repeat1,
  [aux_sym_macro_definition_repeat1] = aux_sym_macro_definition_repeat1,
  [aux_sym_token_tree_pattern_repeat1] = aux_sym_token_tree_pattern_repeat1,
  [aux_sym_token_tree_repeat1] = aux_sym_token_tree_repeat1,
  [aux_sym_declaration_list_repeat1] = aux_sym_declaration_list_repeat1,
  [aux_sym_enum_variant_list_repeat1] = aux_sym_enum_variant_list_repeat1,
  [aux_sym_enum_variant_list_repeat2] = aux_sym_enum_variant_list_repeat2,
  [aux_sym_field_declaration_list_repeat1] = aux_sym_field_declaration_list_repeat1,
  [aux_sym_ordered_field_declaration_list_repeat1] = aux_sym_ordered_field_declaration_list_repeat1,
  [aux_sym_function_modifiers_repeat1] = aux_sym_function_modifiers_repeat1,
  [aux_sym_where_clause_repeat1] = aux_sym_where_clause_repeat1,
  [aux_sym_trait_bounds_repeat1] = aux_sym_trait_bounds_repeat1,
  [aux_sym_type_parameters_repeat1] = aux_sym_type_parameters_repeat1,
  [aux_sym_use_list_repeat1] = aux_sym_use_list_repeat1,
  [aux_sym_parameters_repeat1] = aux_sym_parameters_repeat1,
  [aux_sym_for_lifetimes_repeat1] = aux_sym_for_lifetimes_repeat1,
  [aux_sym_tuple_type_repeat1] = aux_sym_tuple_type_repeat1,
  [aux_sym_type_arguments_repeat1] = aux_sym_type_arguments_repeat1,
  [aux_sym_delim_token_tree_repeat1] = aux_sym_delim_token_tree_repeat1,
  [aux_sym_arguments_repeat1] = aux_sym_arguments_repeat1,
  [aux_sym_array_expression_repeat1] = aux_sym_array_expression_repeat1,
  [aux_sym_tuple_expression_repeat1] = aux_sym_tuple_expression_repeat1,
  [aux_sym_field_initializer_list_repeat1] = aux_sym_field_initializer_list_repeat1,
  [aux_sym_match_block_repeat1] = aux_sym_match_block_repeat1,
  [aux_sym_closure_parameters_repeat1] = aux_sym_closure_parameters_repeat1,
  [aux_sym_tuple_pattern_repeat1] = aux_sym_tuple_pattern_repeat1,
  [aux_sym_struct_pattern_repeat1] = aux_sym_struct_pattern_repeat1,
  [aux_sym_string_literal_repeat1] = aux_sym_string_literal_repeat1,
  [alias_sym_field_identifier] = alias_sym_field_identifier,
  [alias_sym_let_chain] = alias_sym_let_chain,
  [alias_sym_shorthand_field_identifier] = alias_sym_shorthand_field_identifier,
  [alias_sym_type_identifier] = alias_sym_type_identifier,
};

static const TSSymbolMetadata ts_symbol_metadata[] = {
  [ts_builtin_sym_end] = {
    .visible = false,
    .named = true,
  },
  [sym_identifier] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_SEMI] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_macro_rules_BANG] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RPAREN] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_RBRACK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOLLAR] = {
    .visible = true,
    .named = false,
  },
  [aux_sym_token_repetition_pattern_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_PLUS] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_QMARK] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_block] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_expr] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_ident] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_item] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_lifetime] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_literal] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_meta] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_pat] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_path] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_stmt] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_tt] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_ty] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_vis] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_u8] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_i8] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_u16] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_i16] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_u32] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_i32] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_u64] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_i64] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_u128] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_i128] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_isize] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_usize] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_f32] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_f64] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_bool] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_str] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_char] = {
    .visible = true,
    .named = true,
  },
  [aux_sym__non_special_token_token1] = {
    .visible = false,
    .named = false,
  },
  [anon_sym_SQUOTE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_as] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_async] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_await] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_break] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_const] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_continue] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_default] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_enum] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_fn] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_for] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_if] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_impl] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_let] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_loop] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_match] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_mod] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_pub] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_return] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_static] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_struct] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_trait] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_type] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_union] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_unsafe] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_use] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_where] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_while] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_POUND] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BANG] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COMMA] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_extern] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_ref] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_else] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_COLON_COLON] = {
    .visible = true,
    .named = false,
  },
  [anon_sym__] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT_DOT_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_in] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT2] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_dyn] = {
    .visible = true,
    .named = false,
  },
  [sym_mutable_specifier] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_DOT_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT_DOT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP_AMP] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE_PIPE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_CARET] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_EQ_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_BANG_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_LT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_GT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SLASH] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PERCENT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PLUS_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DASH_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_STAR_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_SLASH_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PERCENT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AMP_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_PIPE_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_CARET_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_LT_LT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_GT_GT_EQ] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_yield] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_move] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DOT] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_AT] = {
    .visible = true,
    .named = false,
  },
  [sym_integer_literal] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_string_literal_token1] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_DQUOTE] = {
    .visible = true,
    .named = false,
  },
  [sym_char_literal] = {
    .visible = true,
    .named = true,
  },
  [sym_escape_sequence] = {
    .visible = true,
    .named = true,
  },
  [anon_sym_true] = {
    .visible = true,
    .named = false,
  },
  [anon_sym_false] = {
    .visible = true,
    .named = false,
  },
  [sym_line_comment] = {
    .visible = true,
    .named = true,
  },
  [sym_self] = {
    .visible = true,
    .named = true,
  },
  [sym_super] = {
    .visible = true,
    .named = true,
  },
  [sym_crate] = {
    .visible = true,
    .named = true,
  },
  [sym_metavariable] = {
    .visible = true,
    .named = true,
  },
  [sym__string_content] = {
    .visible = false,
    .named = true,
  },
  [sym_raw_string_literal] = {
    .visible = true,
    .named = true,
  },
  [sym_float_literal] = {
    .visible = true,
    .named = true,
  },
  [sym_block_comment] = {
    .visible = true,
    .named = true,
  },
  [sym_source_file] = {
    .visible = true,
    .named = true,
  },
  [sym__statement] = {
    .visible = false,
    .named = true,
  },
  [sym_empty_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_expression_statement] = {
    .visible = true,
    .named = true,
  },
  [sym_macro_definition] = {
    .visible = true,
    .named = true,
  },
  [sym_macro_rule] = {
    .visible = true,
    .named = true,
  },
  [sym__token_pattern] = {
    .visible = false,
    .named = true,
  },
  [sym_token_tree_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_token_binding_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_token_repetition_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_fragment_specifier] = {
    .visible = true,
    .named = true,
  },
  [sym_token_tree] = {
    .visible = true,
    .named = true,
  },
  [sym_token_repetition] = {
    .visible = true,
    .named = true,
  },
  [sym_attribute_item] = {
    .visible = true,
    .named = true,
  },
  [sym_inner_attribute_item] = {
    .visible = true,
    .named = true,
  },
  [sym_attribute] = {
    .visible = true,
    .named = true,
  },
  [sym_mod_item] = {
    .visible = true,
    .named = true,
  },
  [sym_foreign_mod_item] = {
    .visible = true,
    .named = true,
  },
  [sym_declaration_list] = {
    .visible = true,
    .named = true,
  },
  [sym_struct_item] = {
    .visible = true,
    .named = true,
  },
  [sym_union_item] = {
    .visible = true,
    .named = true,
  },
  [sym_enum_item] = {
    .visible = true,
    .named = true,
  },
  [sym_enum_variant_list] = {
    .visible = true,
    .named = true,
  },
  [sym_enum_variant] = {
    .visible = true,
    .named = true,
  },
  [sym_field_declaration_list] = {
    .visible = true,
    .named = true,
  },
  [sym_field_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_ordered_field_declaration_list] = {
    .visible = true,
    .named = true,
  },
  [sym_extern_crate_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_const_item] = {
    .visible = true,
    .named = true,
  },
  [sym_static_item] = {
    .visible = true,
    .named = true,
  },
  [sym_type_item] = {
    .visible = true,
    .named = true,
  },
  [sym_function_item] = {
    .visible = true,
    .named = true,
  },
  [sym_function_signature_item] = {
    .visible = true,
    .named = true,
  },
  [sym_function_modifiers] = {
    .visible = true,
    .named = true,
  },
  [sym_where_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_where_predicate] = {
    .visible = true,
    .named = true,
  },
  [sym_impl_item] = {
    .visible = true,
    .named = true,
  },
  [sym_trait_item] = {
    .visible = true,
    .named = true,
  },
  [sym_associated_type] = {
    .visible = true,
    .named = true,
  },
  [sym_trait_bounds] = {
    .visible = true,
    .named = true,
  },
  [sym_higher_ranked_trait_bound] = {
    .visible = true,
    .named = true,
  },
  [sym_removed_trait_bound] = {
    .visible = true,
    .named = true,
  },
  [sym_type_parameters] = {
    .visible = true,
    .named = true,
  },
  [sym_const_parameter] = {
    .visible = true,
    .named = true,
  },
  [sym_constrained_type_parameter] = {
    .visible = true,
    .named = true,
  },
  [sym_optional_type_parameter] = {
    .visible = true,
    .named = true,
  },
  [sym_let_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym_use_declaration] = {
    .visible = true,
    .named = true,
  },
  [sym__use_clause] = {
    .visible = false,
    .named = true,
  },
  [sym_scoped_use_list] = {
    .visible = true,
    .named = true,
  },
  [sym_use_list] = {
    .visible = true,
    .named = true,
  },
  [sym_use_as_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_use_wildcard] = {
    .visible = true,
    .named = true,
  },
  [sym_parameters] = {
    .visible = true,
    .named = true,
  },
  [sym_self_parameter] = {
    .visible = true,
    .named = true,
  },
  [sym_variadic_parameter] = {
    .visible = true,
    .named = true,
  },
  [sym_parameter] = {
    .visible = true,
    .named = true,
  },
  [sym_extern_modifier] = {
    .visible = true,
    .named = true,
  },
  [sym_visibility_modifier] = {
    .visible = true,
    .named = true,
  },
  [sym__type] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_bracketed_type] = {
    .visible = true,
    .named = true,
  },
  [sym_qualified_type] = {
    .visible = true,
    .named = true,
  },
  [sym_lifetime] = {
    .visible = true,
    .named = true,
  },
  [sym_array_type] = {
    .visible = true,
    .named = true,
  },
  [sym_for_lifetimes] = {
    .visible = true,
    .named = true,
  },
  [sym_function_type] = {
    .visible = true,
    .named = true,
  },
  [sym_tuple_type] = {
    .visible = true,
    .named = true,
  },
  [sym_unit_type] = {
    .visible = true,
    .named = true,
  },
  [sym_generic_function] = {
    .visible = true,
    .named = true,
  },
  [sym_generic_type] = {
    .visible = true,
    .named = true,
  },
  [sym_generic_type_with_turbofish] = {
    .visible = true,
    .named = true,
  },
  [sym_bounded_type] = {
    .visible = true,
    .named = true,
  },
  [sym_type_arguments] = {
    .visible = true,
    .named = true,
  },
  [sym_type_binding] = {
    .visible = true,
    .named = true,
  },
  [sym_reference_type] = {
    .visible = true,
    .named = true,
  },
  [sym_pointer_type] = {
    .visible = true,
    .named = true,
  },
  [sym_empty_type] = {
    .visible = true,
    .named = true,
  },
  [sym_abstract_type] = {
    .visible = true,
    .named = true,
  },
  [sym_dynamic_type] = {
    .visible = true,
    .named = true,
  },
  [sym__expression_except_range] = {
    .visible = false,
    .named = true,
  },
  [sym__expression] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_macro_invocation] = {
    .visible = true,
    .named = true,
  },
  [sym_delim_token_tree] = {
    .visible = true,
    .named = true,
  },
  [sym__delim_tokens] = {
    .visible = false,
    .named = true,
  },
  [sym__non_delim_token] = {
    .visible = false,
    .named = true,
  },
  [sym_scoped_identifier] = {
    .visible = true,
    .named = true,
  },
  [sym_scoped_type_identifier_in_expression_position] = {
    .visible = true,
    .named = true,
  },
  [sym_scoped_type_identifier] = {
    .visible = true,
    .named = true,
  },
  [sym_range_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_unary_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_try_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_reference_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_binary_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_assignment_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_compound_assignment_expr] = {
    .visible = true,
    .named = true,
  },
  [sym_type_cast_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_return_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_yield_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_call_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_arguments] = {
    .visible = true,
    .named = true,
  },
  [sym_array_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_parenthesized_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_tuple_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_unit_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_struct_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_field_initializer_list] = {
    .visible = true,
    .named = true,
  },
  [sym_shorthand_field_initializer] = {
    .visible = true,
    .named = true,
  },
  [sym_field_initializer] = {
    .visible = true,
    .named = true,
  },
  [sym_base_field_initializer] = {
    .visible = true,
    .named = true,
  },
  [sym_if_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_let_condition] = {
    .visible = true,
    .named = true,
  },
  [sym__let_chain] = {
    .visible = false,
    .named = true,
  },
  [sym__condition] = {
    .visible = false,
    .named = true,
  },
  [sym_else_clause] = {
    .visible = true,
    .named = true,
  },
  [sym_match_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_match_block] = {
    .visible = true,
    .named = true,
  },
  [sym_match_arm] = {
    .visible = true,
    .named = true,
  },
  [sym_last_match_arm] = {
    .visible = true,
    .named = true,
  },
  [sym_match_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_while_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_loop_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_for_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_const_block] = {
    .visible = true,
    .named = true,
  },
  [sym_closure_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_closure_parameters] = {
    .visible = true,
    .named = true,
  },
  [sym_loop_label] = {
    .visible = true,
    .named = true,
  },
  [sym_break_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_continue_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_index_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_await_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_field_expression] = {
    .visible = true,
    .named = true,
  },
  [sym_unsafe_block] = {
    .visible = true,
    .named = true,
  },
  [sym_async_block] = {
    .visible = true,
    .named = true,
  },
  [sym_block] = {
    .visible = true,
    .named = true,
  },
  [sym__pattern] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_tuple_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_slice_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_tuple_struct_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_struct_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_field_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_remaining_field_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_mut_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_range_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_ref_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_captured_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_reference_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym_or_pattern] = {
    .visible = true,
    .named = true,
  },
  [sym__literal] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym__literal_pattern] = {
    .visible = false,
    .named = true,
    .supertype = true,
  },
  [sym_negative_literal] = {
    .visible = true,
    .named = true,
  },
  [sym_string_literal] = {
    .visible = true,
    .named = true,
  },
  [sym_boolean_literal] = {
    .visible = true,
    .named = true,
  },
  [aux_sym_source_file_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_macro_definition_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_token_tree_pattern_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_token_tree_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_declaration_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_enum_variant_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_enum_variant_list_repeat2] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_field_declaration_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_ordered_field_declaration_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_function_modifiers_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_where_clause_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_trait_bounds_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_type_parameters_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_use_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_parameters_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_for_lifetimes_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_tuple_type_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_type_arguments_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_delim_token_tree_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_arguments_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_array_expression_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_tuple_expression_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_field_initializer_list_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_match_block_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_closure_parameters_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_tuple_pattern_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_struct_pattern_repeat1] = {
    .visible = false,
    .named = false,
  },
  [aux_sym_string_literal_repeat1] = {
    .visible = false,
    .named = false,
  },
  [alias_sym_field_identifier] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_let_chain] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_shorthand_field_identifier] = {
    .visible = true,
    .named = true,
  },
  [alias_sym_type_identifier] = {
    .visible = true,
    .named = true,
  },
};

enum {
  field_alias = 1,
  field_alternative = 2,
  field_argument = 3,
  field_arguments = 4,
  field_body = 5,
  field_bounds = 6,
  field_condition = 7,
  field_consequence = 8,
  field_default_type = 9,
  field_element = 10,
  field_field = 11,
  field_function = 12,
  field_left = 13,
  field_length = 14,
  field_list = 15,
  field_macro = 16,
  field_name = 17,
  field_operator = 18,
  field_parameters = 19,
  field_path = 20,
  field_pattern = 21,
  field_return_type = 22,
  field_right = 23,
  field_trait = 24,
  field_type = 25,
  field_type_arguments = 26,
  field_type_parameters = 27,
  field_value = 28,
};

static const char * const ts_field_names[] = {
  [0] = NULL,
  [field_alias] = "alias",
  [field_alternative] = "alternative",
  [field_argument] = "argument",
  [field_arguments] = "arguments",
  [field_body] = "body",
  [field_bounds] = "bounds",
  [field_condition] = "condition",
  [field_consequence] = "consequence",
  [field_default_type] = "default_type",
  [field_element] = "element",
  [field_field] = "field",
  [field_function] = "function",
  [field_left] = "left",
  [field_length] = "length",
  [field_list] = "list",
  [field_macro] = "macro",
  [field_name] = "name",
  [field_operator] = "operator",
  [field_parameters] = "parameters",
  [field_path] = "path",
  [field_pattern] = "pattern",
  [field_return_type] = "return_type",
  [field_right] = "right",
  [field_trait] = "trait",
  [field_type] = "type",
  [field_type_arguments] = "type_arguments",
  [field_type_parameters] = "type_parameters",
  [field_value] = "value",
};

static const TSFieldMapSlice ts_field_map_slices[PRODUCTION_ID_COUNT] = {
  [2] = {.index = 0, .length = 1},
  [5] = {.index = 1, .length = 1},
  [6] = {.index = 1, .length = 1},
  [7] = {.index = 2, .length = 1},
  [8] = {.index = 3, .length = 2},
  [9] = {.index = 3, .length = 2},
  [10] = {.index = 5, .length = 2},
  [11] = {.index = 7, .length = 2},
  [12] = {.index = 9, .length = 2},
  [13] = {.index = 9, .length = 2},
  [14] = {.index = 11, .length = 1},
  [15] = {.index = 12, .length = 2},
  [16] = {.index = 14, .length = 2},
  [17] = {.index = 14, .length = 2},
  [18] = {.index = 16, .length = 2},
  [19] = {.index = 18, .length = 1},
  [20] = {.index = 19, .length = 1},
  [21] = {.index = 19, .length = 1},
  [22] = {.index = 20, .length = 1},
  [23] = {.index = 21, .length = 2},
  [24] = {.index = 23, .length = 2},
  [25] = {.index = 21, .length = 2},
  [26] = {.index = 25, .length = 1},
  [27] = {.index = 26, .length = 2},
  [28] = {.index = 12, .length = 2},
  [29] = {.index = 28, .length = 1},
  [30] = {.index = 29, .length = 1},
  [31] = {.index = 30, .length = 2},
  [32] = {.index = 32, .length = 1},
  [33] = {.index = 33, .length = 2},
  [34] = {.index = 11, .length = 1},
  [35] = {.index = 9, .length = 2},
  [36] = {.index = 9, .length = 2},
  [37] = {.index = 35, .length = 2},
  [38] = {.index = 37, .length = 2},
  [39] = {.index = 39, .length = 1},
  [40] = {.index = 9, .length = 2},
  [41] = {.index = 9, .length = 2},
  [42] = {.index = 40, .length = 3},
  [43] = {.index = 43, .length = 2},
  [44] = {.index = 45, .length = 2},
  [45] = {.index = 47, .length = 2},
  [46] = {.index = 47, .length = 2},
  [47] = {.index = 37, .length = 2},
  [48] = {.index = 1, .length = 1},
  [49] = {.index = 49, .length = 1},
  [50] = {.index = 50, .length = 2},
  [51] = {.index = 52, .length = 3},
  [52] = {.index = 55, .length = 2},
  [53] = {.index = 57, .length = 3},
  [55] = {.index = 60, .length = 1},
  [56] = {.index = 60, .length = 1},
  [57] = {.index = 49, .length = 1},
  [59] = {.index = 61, .length = 3},
  [60] = {.index = 64, .length = 1},
  [61] = {.index = 65, .length = 1},
  [63] = {.index = 66, .length = 2},
  [64] = {.index = 66, .length = 2},
  [65] = {.index = 68, .length = 1},
  [66] = {.index = 69, .length = 2},
  [67] = {.index = 71, .length = 3},
  [68] = {.index = 74, .length = 2},
  [69] = {.index = 76, .length = 2},
  [70] = {.index = 76, .length = 2},
  [71] = {.index = 78, .length = 1},
  [72] = {.index = 79, .length = 2},
  [73] = {.index = 81, .length = 3},
  [74] = {.index = 84, .length = 2},
  [75] = {.index = 86, .length = 2},
  [76] = {.index = 88, .length = 2},
  [77] = {.index = 90, .length = 2},
  [78] = {.index = 92, .length = 2},
  [79] = {.index = 90, .length = 2},
  [80] = {.index = 92, .length = 2},
  [81] = {.index = 94, .length = 1},
  [82] = {.index = 94, .length = 1},
  [83] = {.index = 95, .length = 1},
  [84] = {.index = 96, .length = 2},
  [85] = {.index = 98, .length = 2},
  [86] = {.index = 88, .length = 2},
  [87] = {.index = 95, .length = 1},
  [88] = {.index = 100, .length = 1},
  [89] = {.index = 101, .length = 3},
  [90] = {.index = 104, .length = 1},
  [91] = {.index = 105, .length = 1},
  [92] = {.index = 106, .length = 2},
  [93] = {.index = 108, .length = 3},
  [94] = {.index = 111, .length = 3},
  [95] = {.index = 114, .length = 4},
  [96] = {.index = 118, .length = 3},
  [97] = {.index = 1, .length = 1},
  [98] = {.index = 121, .length = 3},
  [99] = {.index = 124, .length = 2},
  [100] = {.index = 126, .length = 2},
  [101] = {.index = 128, .length = 2},
  [102] = {.index = 128, .length = 2},
  [103] = {.index = 130, .length = 1},
  [104] = {.index = 131, .length = 2},
  [105] = {.index = 133, .length = 3},
  [106] = {.index = 136, .length = 3},
  [107] = {.index = 139, .length = 3},
  [108] = {.index = 142, .length = 1},
  [109] = {.index = 131, .length = 2},
  [110] = {.index = 133, .length = 3},
  [111] = {.index = 136, .length = 3},
  [112] = {.index = 143, .length = 2},
  [113] = {.index = 145, .length = 2},
  [115] = {.index = 147, .length = 3},
  [116] = {.index = 150, .length = 4},
  [117] = {.index = 106, .length = 2},
  [118] = {.index = 154, .length = 3},
  [119] = {.index = 157, .length = 2},
  [120] = {.index = 159, .length = 3},
  [121] = {.index = 162, .length = 2},
  [122] = {.index = 164, .length = 2},
  [123] = {.index = 166, .length = 3},
  [124] = {.index = 169, .length = 3},
  [125] = {.index = 32, .length = 1},
  [126] = {.index = 172, .length = 3},
  [127] = {.index = 175, .length = 2},
  [128] = {.index = 177, .length = 2},
  [129] = {.index = 179, .length = 3},
  [130] = {.index = 182, .length = 2},
  [131] = {.index = 184, .length = 2},
  [132] = {.index = 186, .length = 1},
  [133] = {.index = 187, .length = 2},
  [134] = {.index = 189, .length = 1},
  [135] = {.index = 175, .length = 2},
  [136] = {.index = 190, .length = 4},
  [137] = {.index = 194, .length = 3},
  [138] = {.index = 197, .length = 4},
  [139] = {.index = 95, .length = 1},
  [140] = {.index = 201, .length = 2},
  [141] = {.index = 203, .length = 2},
  [142] = {.index = 205, .length = 2},
  [143] = {.index = 207, .length = 3},
  [144] = {.index = 210, .length = 2},
  [145] = {.index = 212, .length = 3},
  [146] = {.index = 215, .length = 4},
  [147] = {.index = 212, .length = 3},
  [148] = {.index = 215, .length = 4},
  [149] = {.index = 219, .length = 3},
  [150] = {.index = 219, .length = 3},
  [151] = {.index = 207, .length = 3},
  [152] = {.index = 222, .length = 2},
  [153] = {.index = 224, .length = 2},
  [154] = {.index = 226, .length = 2},
  [155] = {.index = 228, .length = 2},
  [156] = {.index = 230, .length = 1},
  [157] = {.index = 231, .length = 2},
  [158] = {.index = 233, .length = 2},
  [159] = {.index = 235, .length = 2},
  [160] = {.index = 203, .length = 2},
  [161] = {.index = 237, .length = 4},
  [162] = {.index = 241, .length = 3},
  [163] = {.index = 244, .length = 2},
  [164] = {.index = 246, .length = 3},
  [165] = {.index = 249, .length = 3},
  [166] = {.index = 244, .length = 2},
  [167] = {.index = 246, .length = 3},
  [168] = {.index = 252, .length = 3},
  [169] = {.index = 255, .length = 3},
  [170] = {.index = 258, .length = 4},
  [171] = {.index = 262, .length = 2},
  [172] = {.index = 264, .length = 2},
  [173] = {.index = 266, .length = 3},
  [174] = {.index = 269, .length = 4},
  [175] = {.index = 273, .length = 3},
  [176] = {.index = 231, .length = 2},
  [177] = {.index = 276, .length = 2},
  [178] = {.index = 278, .length = 3},
  [179] = {.index = 281, .length = 3},
  [180] = {.index = 284, .length = 2},
  [181] = {.index = 286, .length = 3},
  [182] = {.index = 203, .length = 2},
  [183] = {.index = 289, .length = 3},
  [184] = {.index = 292, .length = 3},
  [185] = {.index = 264, .length = 2},
  [186] = {.index = 295, .length = 4},
  [187] = {.index = 299, .length = 5},
  [188] = {.index = 304, .length = 4},
  [189] = {.index = 308, .length = 2},
  [190] = {.index = 310, .length = 3},
  [191] = {.index = 313, .length = 4},
  [192] = {.index = 313, .length = 4},
  [193] = {.index = 317, .length = 2},
  [194] = {.index = 319, .length = 3},
  [195] = {.index = 322, .length = 3},
  [196] = {.index = 325, .length = 3},
  [197] = {.index = 328, .length = 2},
  [198] = {.index = 330, .length = 2},
  [199] = {.index = 106, .length = 2},
  [200] = {.index = 332, .length = 3},
  [201] = {.index = 335, .length = 3},
  [202] = {.index = 338, .length = 4},
  [203] = {.index = 335, .length = 3},
  [204] = {.index = 338, .length = 4},
  [205] = {.index = 332, .length = 3},
  [206] = {.index = 342, .length = 4},
  [207] = {.index = 346, .length = 4},
  [208] = {.index = 350, .length = 3},
  [209] = {.index = 353, .length = 4},
  [210] = {.index = 357, .length = 3},
  [211] = {.index = 360, .length = 3},
  [212] = {.index = 363, .length = 3},
  [213] = {.index = 366, .length = 4},
  [214] = {.index = 370, .length = 2},
  [215] = {.index = 372, .length = 3},
  [216] = {.index = 375, .length = 4},
  [217] = {.index = 379, .length = 3},
  [218] = {.index = 382, .length = 3},
  [219] = {.index = 385, .length = 3},
  [220] = {.index = 388, .length = 5},
  [221] = {.index = 393, .length = 2},
  [222] = {.index = 395, .length = 3},
  [223] = {.index = 398, .length = 3},
  [224] = {.index = 401, .length = 3},
  [225] = {.index = 404, .length = 3},
  [226] = {.index = 407, .length = 2},
  [227] = {.index = 409, .length = 4},
  [228] = {.index = 409, .length = 4},
  [229] = {.index = 413, .length = 4},
  [230] = {.index = 417, .length = 5},
  [231] = {.index = 422, .length = 4},
  [232] = {.index = 426, .length = 2},
  [233] = {.index = 428, .length = 4},
  [234] = {.index = 432, .length = 4},
  [235] = {.index = 436, .length = 3},
  [236] = {.index = 439, .length = 4},
  [237] = {.index = 443, .length = 4},
  [238] = {.index = 447, .length = 3},
  [239] = {.index = 450, .length = 5},
  [240] = {.index = 455, .length = 4},
  [241] = {.index = 459, .length = 5},
  [242] = {.index = 464, .length = 4},
  [243] = {.index = 468, .length = 4},
  [244] = {.index = 472, .length = 3},
  [245] = {.index = 475, .length = 5},
};

static const TSFieldMapEntry ts_field_map_entries[] = {
  [0] =
    {field_body, 1},
  [1] =
    {field_name, 1},
  [2] =
    {field_value, 1},
  [3] =
    {field_body, 1},
    {field_name, 0},
  [5] =
    {field_arguments, 1},
    {field_function, 0},
  [7] =
    {field_body, 1},
    {field_parameters, 0},
  [9] =
    {field_name, 2},
    {field_path, 0},
  [11] =
    {field_macro, 0},
  [12] =
    {field_body, 2},
    {field_name, 1},
  [14] =
    {field_type, 0},
    {field_type_arguments, 1},
  [16] =
    {field_condition, 1},
    {field_consequence, 2},
  [18] =
    {field_parameters, 1},
  [19] =
    {field_trait, 1},
  [20] =
    {field_type, 1},
  [21] =
    {field_parameters, 1},
    {field_trait, 0},
  [23] =
    {field_body, 2},
    {field_type, 1},
  [25] =
    {field_pattern, 1},
  [26] =
    {field_body, 2},
    {field_value, 1},
  [28] =
    {field_list, 1},
  [29] =
    {field_argument, 1},
  [30] =
    {field_body, 2},
    {field_condition, 1},
  [32] =
    {field_value, 2},
  [33] =
    {field_body, 2},
    {field_parameters, 1},
  [35] =
    {field_function, 0},
    {field_type_arguments, 2},
  [37] =
    {field_type, 0},
    {field_type_arguments, 2},
  [39] =
    {field_body, 2},
  [40] =
    {field_left, 0},
    {field_operator, 1},
    {field_right, 2},
  [43] =
    {field_type, 2},
    {field_value, 0},
  [45] =
    {field_left, 0},
    {field_right, 2},
  [47] =
    {field_field, 2},
    {field_value, 0},
  [49] =
    {field_name, 0},
  [50] =
    {field_body, 3},
    {field_name, 1},
  [52] =
    {field_body, 3},
    {field_name, 1},
    {field_type_parameters, 2},
  [55] =
    {field_name, 1},
    {field_parameters, 2},
  [57] =
    {field_body, 3},
    {field_name, 1},
    {field_parameters, 2},
  [60] =
    {field_type, 0},
  [61] =
    {field_alternative, 3},
    {field_condition, 1},
    {field_consequence, 2},
  [64] =
    {field_element, 1},
  [65] =
    {field_type, 2},
  [66] =
    {field_bounds, 1},
    {field_left, 0},
  [68] =
    {field_parameters, 2},
  [69] =
    {field_type, 2},
    {field_type_parameters, 1},
  [71] =
    {field_body, 3},
    {field_type, 2},
    {field_type_parameters, 1},
  [74] =
    {field_body, 3},
    {field_type, 1},
  [76] =
    {field_parameters, 2},
    {field_trait, 1},
  [78] =
    {field_pattern, 2},
  [79] =
    {field_name, 1},
    {field_type_parameters, 2},
  [81] =
    {field_body, 3},
    {field_bounds, 2},
    {field_name, 1},
  [84] =
    {field_bounds, 2},
    {field_name, 1},
  [86] =
    {field_body, 3},
    {field_type, 2},
  [88] =
    {field_body, 3},
    {field_name, 2},
  [90] =
    {field_alias, 2},
    {field_path, 0},
  [92] =
    {field_list, 2},
    {field_path, 0},
  [94] =
    {field_arguments, 1},
  [95] =
    {field_name, 2},
  [96] =
    {field_alias, 2},
    {field_type, 0},
  [98] =
    {field_pattern, 0},
    {field_type, 2},
  [100] =
    {field_argument, 2},
  [101] =
    {field_body, 3},
    {field_parameters, 0},
    {field_return_type, 2},
  [104] =
    {field_body, 3},
  [105] =
    {field_length, 3},
  [106] =
    {field_name, 1},
    {field_type, 3},
  [108] =
    {field_body, 4},
    {field_name, 1},
    {field_type_parameters, 2},
  [111] =
    {field_name, 1},
    {field_parameters, 3},
    {field_type_parameters, 2},
  [114] =
    {field_body, 4},
    {field_name, 1},
    {field_parameters, 3},
    {field_type_parameters, 2},
  [118] =
    {field_body, 4},
    {field_name, 1},
    {field_parameters, 2},
  [121] =
    {field_body, 4},
    {field_pattern, 1},
    {field_value, 3},
  [124] =
    {field_pattern, 1},
    {field_value, 3},
  [126] =
    {field_parameters, 1},
    {field_return_type, 3},
  [128] =
    {field_default_type, 2},
    {field_name, 0},
  [130] =
    {field_type, 3},
  [131] =
    {field_trait, 1},
    {field_type, 3},
  [133] =
    {field_body, 4},
    {field_trait, 1},
    {field_type, 3},
  [136] =
    {field_parameters, 1},
    {field_return_type, 3},
    {field_trait, 0},
  [139] =
    {field_body, 4},
    {field_type, 2},
    {field_type_parameters, 1},
  [142] =
    {field_parameters, 3},
  [143] =
    {field_pattern, 1},
    {field_type, 3},
  [145] =
    {field_alternative, 3},
    {field_pattern, 1},
  [147] =
    {field_body, 4},
    {field_bounds, 2},
    {field_name, 1},
  [150] =
    {field_body, 4},
    {field_bounds, 3},
    {field_name, 1},
    {field_type_parameters, 2},
  [154] =
    {field_bounds, 3},
    {field_name, 1},
    {field_type_parameters, 2},
  [157] =
    {field_type, 3},
    {field_type_parameters, 2},
  [159] =
    {field_body, 4},
    {field_type, 3},
    {field_type_parameters, 2},
  [162] =
    {field_body, 4},
    {field_type, 2},
  [164] =
    {field_body, 4},
    {field_name, 2},
  [166] =
    {field_body, 4},
    {field_bounds, 3},
    {field_name, 2},
  [169] =
    {field_body, 4},
    {field_name, 2},
    {field_type_parameters, 3},
  [172] =
    {field_body, 4},
    {field_parameters, 1},
    {field_return_type, 3},
  [175] =
    {field_name, 0},
    {field_value, 2},
  [177] =
    {field_name, 2},
    {field_parameters, 3},
  [179] =
    {field_body, 4},
    {field_name, 2},
    {field_parameters, 3},
  [182] =
    {field_name, 2},
    {field_type_parameters, 3},
  [184] =
    {field_body, 4},
    {field_name, 3},
  [186] =
    {field_name, 3},
  [187] =
    {field_body, 4},
    {field_condition, 3},
  [189] =
    {field_length, 4},
  [190] =
    {field_body, 5},
    {field_name, 1},
    {field_parameters, 3},
    {field_type_parameters, 2},
  [194] =
    {field_name, 1},
    {field_parameters, 2},
    {field_return_type, 4},
  [197] =
    {field_body, 5},
    {field_name, 1},
    {field_parameters, 2},
    {field_return_type, 4},
  [201] =
    {field_name, 0},
    {field_pattern, 2},
  [203] =
    {field_name, 0},
    {field_type, 2},
  [205] =
    {field_element, 1},
    {field_length, 3},
  [207] =
    {field_body, 5},
    {field_trait, 1},
    {field_type, 3},
  [210] =
    {field_parameters, 2},
    {field_return_type, 4},
  [212] =
    {field_trait, 2},
    {field_type, 4},
    {field_type_parameters, 1},
  [215] =
    {field_body, 5},
    {field_trait, 2},
    {field_type, 4},
    {field_type_parameters, 1},
  [219] =
    {field_parameters, 2},
    {field_return_type, 4},
    {field_trait, 1},
  [222] =
    {field_pattern, 2},
    {field_type, 4},
  [224] =
    {field_pattern, 2},
    {field_value, 4},
  [226] =
    {field_alternative, 4},
    {field_pattern, 2},
  [228] =
    {field_pattern, 0},
    {field_value, 2},
  [230] =
    {field_condition, 2},
  [231] =
    {field_name, 2},
    {field_type, 4},
  [233] =
    {field_type, 1},
    {field_type, 2, .inherited = true},
  [235] =
    {field_type, 0, .inherited = true},
    {field_type, 1, .inherited = true},
  [237] =
    {field_body, 5},
    {field_bounds, 3},
    {field_name, 1},
    {field_type_parameters, 2},
  [241] =
    {field_name, 1},
    {field_type, 4},
    {field_type_parameters, 2},
  [244] =
    {field_trait, 2},
    {field_type, 4},
  [246] =
    {field_body, 5},
    {field_trait, 2},
    {field_type, 4},
  [249] =
    {field_body, 5},
    {field_type, 3},
    {field_type_parameters, 2},
  [252] =
    {field_body, 5},
    {field_bounds, 3},
    {field_name, 2},
  [255] =
    {field_body, 5},
    {field_name, 2},
    {field_type_parameters, 3},
  [258] =
    {field_body, 5},
    {field_bounds, 4},
    {field_name, 2},
    {field_type_parameters, 3},
  [262] =
    {field_alias, 4},
    {field_name, 2},
  [264] =
    {field_name, 1},
    {field_value, 3},
  [266] =
    {field_name, 2},
    {field_parameters, 4},
    {field_type_parameters, 3},
  [269] =
    {field_body, 5},
    {field_name, 2},
    {field_parameters, 4},
    {field_type_parameters, 3},
  [273] =
    {field_body, 5},
    {field_name, 2},
    {field_parameters, 3},
  [276] =
    {field_body, 5},
    {field_name, 3},
  [278] =
    {field_body, 5},
    {field_bounds, 4},
    {field_name, 3},
  [281] =
    {field_body, 5},
    {field_name, 3},
    {field_type_parameters, 4},
  [284] =
    {field_name, 3},
    {field_parameters, 4},
  [286] =
    {field_body, 5},
    {field_name, 3},
    {field_parameters, 4},
  [289] =
    {field_name, 1},
    {field_type, 3},
    {field_value, 5},
  [292] =
    {field_body, 1},
    {field_name, 0},
    {field_value, 3},
  [295] =
    {field_name, 1},
    {field_parameters, 3},
    {field_return_type, 5},
    {field_type_parameters, 2},
  [299] =
    {field_body, 6},
    {field_name, 1},
    {field_parameters, 3},
    {field_return_type, 5},
    {field_type_parameters, 2},
  [304] =
    {field_body, 6},
    {field_name, 1},
    {field_parameters, 2},
    {field_return_type, 4},
  [308] =
    {field_name, 1},
    {field_pattern, 3},
  [310] =
    {field_name, 0},
    {field_type, 3},
    {field_type_arguments, 1},
  [313] =
    {field_body, 6},
    {field_trait, 2},
    {field_type, 4},
    {field_type_parameters, 1},
  [317] =
    {field_parameters, 3},
    {field_return_type, 5},
  [319] =
    {field_pattern, 1},
    {field_type, 3},
    {field_value, 5},
  [322] =
    {field_alternative, 5},
    {field_pattern, 1},
    {field_type, 3},
  [325] =
    {field_alternative, 5},
    {field_pattern, 1},
    {field_value, 3},
  [328] =
    {field_name, 3},
    {field_type, 5},
  [330] =
    {field_type, 2},
    {field_type, 3, .inherited = true},
  [332] =
    {field_body, 6},
    {field_trait, 2},
    {field_type, 4},
  [335] =
    {field_trait, 3},
    {field_type, 5},
    {field_type_parameters, 2},
  [338] =
    {field_body, 6},
    {field_trait, 3},
    {field_type, 5},
    {field_type_parameters, 2},
  [342] =
    {field_body, 6},
    {field_bounds, 4},
    {field_name, 2},
    {field_type_parameters, 3},
  [346] =
    {field_body, 6},
    {field_name, 2},
    {field_parameters, 4},
    {field_type_parameters, 3},
  [350] =
    {field_name, 2},
    {field_parameters, 3},
    {field_return_type, 5},
  [353] =
    {field_body, 6},
    {field_name, 2},
    {field_parameters, 3},
    {field_return_type, 5},
  [357] =
    {field_name, 2},
    {field_type, 5},
    {field_type_parameters, 3},
  [360] =
    {field_body, 6},
    {field_bounds, 4},
    {field_name, 3},
  [363] =
    {field_body, 6},
    {field_name, 3},
    {field_type_parameters, 4},
  [366] =
    {field_body, 6},
    {field_bounds, 5},
    {field_name, 3},
    {field_type_parameters, 4},
  [370] =
    {field_alias, 5},
    {field_name, 3},
  [372] =
    {field_name, 3},
    {field_parameters, 5},
    {field_type_parameters, 4},
  [375] =
    {field_body, 6},
    {field_name, 3},
    {field_parameters, 5},
    {field_type_parameters, 4},
  [379] =
    {field_body, 6},
    {field_name, 3},
    {field_parameters, 4},
  [382] =
    {field_body, 6},
    {field_pattern, 3},
    {field_value, 5},
  [385] =
    {field_body, 2},
    {field_name, 1},
    {field_value, 4},
  [388] =
    {field_body, 7},
    {field_name, 1},
    {field_parameters, 3},
    {field_return_type, 5},
    {field_type_parameters, 2},
  [393] =
    {field_name, 2},
    {field_pattern, 4},
  [395] =
    {field_pattern, 2},
    {field_type, 4},
    {field_value, 6},
  [398] =
    {field_alternative, 6},
    {field_pattern, 2},
    {field_type, 4},
  [401] =
    {field_alternative, 6},
    {field_pattern, 2},
    {field_value, 4},
  [404] =
    {field_name, 2},
    {field_type, 4},
    {field_value, 6},
  [407] =
    {field_type, 3},
    {field_type, 4, .inherited = true},
  [409] =
    {field_body, 7},
    {field_trait, 3},
    {field_type, 5},
    {field_type_parameters, 2},
  [413] =
    {field_name, 2},
    {field_parameters, 4},
    {field_return_type, 6},
    {field_type_parameters, 3},
  [417] =
    {field_body, 7},
    {field_name, 2},
    {field_parameters, 4},
    {field_return_type, 6},
    {field_type_parameters, 3},
  [422] =
    {field_body, 7},
    {field_name, 2},
    {field_parameters, 3},
    {field_return_type, 5},
  [426] =
    {field_name, 4},
    {field_type, 6},
  [428] =
    {field_body, 7},
    {field_bounds, 5},
    {field_name, 3},
    {field_type_parameters, 4},
  [432] =
    {field_body, 7},
    {field_name, 3},
    {field_parameters, 5},
    {field_type_parameters, 4},
  [436] =
    {field_name, 3},
    {field_parameters, 4},
    {field_return_type, 6},
  [439] =
    {field_body, 7},
    {field_name, 3},
    {field_parameters, 4},
    {field_return_type, 6},
  [443] =
    {field_alternative, 7},
    {field_pattern, 1},
    {field_type, 3},
    {field_value, 5},
  [447] =
    {field_name, 3},
    {field_type, 5},
    {field_value, 7},
  [450] =
    {field_body, 8},
    {field_name, 2},
    {field_parameters, 4},
    {field_return_type, 6},
    {field_type_parameters, 3},
  [455] =
    {field_name, 3},
    {field_parameters, 5},
    {field_return_type, 7},
    {field_type_parameters, 4},
  [459] =
    {field_body, 8},
    {field_name, 3},
    {field_parameters, 5},
    {field_return_type, 7},
    {field_type_parameters, 4},
  [464] =
    {field_body, 8},
    {field_name, 3},
    {field_parameters, 4},
    {field_return_type, 6},
  [468] =
    {field_alternative, 8},
    {field_pattern, 2},
    {field_type, 4},
    {field_value, 6},
  [472] =
    {field_name, 4},
    {field_type, 6},
    {field_value, 8},
  [475] =
    {field_body, 9},
    {field_name, 3},
    {field_parameters, 5},
    {field_return_type, 7},
    {field_type_parameters, 4},
};

static const TSSymbol ts_alias_sequences[PRODUCTION_ID_COUNT][MAX_ALIAS_SEQUENCE_LENGTH] = {
  [0] = {0},
  [1] = {
    [0] = sym_identifier,
  },
  [3] = {
    [0] = alias_sym_let_chain,
  },
  [4] = {
    [0] = alias_sym_type_identifier,
  },
  [6] = {
    [1] = alias_sym_type_identifier,
  },
  [8] = {
    [0] = alias_sym_type_identifier,
  },
  [12] = {
    [0] = sym_identifier,
  },
  [13] = {
    [0] = sym_identifier,
    [2] = alias_sym_type_identifier,
  },
  [14] = {
    [0] = sym_identifier,
  },
  [15] = {
    [1] = alias_sym_type_identifier,
  },
  [16] = {
    [0] = alias_sym_type_identifier,
  },
  [20] = {
    [1] = alias_sym_type_identifier,
  },
  [23] = {
    [0] = alias_sym_type_identifier,
  },
  [36] = {
    [2] = alias_sym_type_identifier,
  },
  [38] = {
    [0] = alias_sym_type_identifier,
  },
  [40] = {
    [0] = sym_generic_type,
  },
  [41] = {
    [0] = sym_generic_type,
    [2] = alias_sym_type_identifier,
  },
  [46] = {
    [2] = alias_sym_field_identifier,
  },
  [48] = {
    [1] = sym_identifier,
  },
  [50] = {
    [1] = alias_sym_type_identifier,
  },
  [51] = {
    [1] = alias_sym_type_identifier,
  },
  [54] = {
    [0] = sym_identifier,
    [2] = sym_identifier,
  },
  [56] = {
    [0] = alias_sym_type_identifier,
  },
  [57] = {
    [0] = alias_sym_shorthand_field_identifier,
  },
  [58] = {
    [2] = sym_identifier,
  },
  [62] = {
    [1] = alias_sym_type_identifier,
  },
  [63] = {
    [0] = alias_sym_type_identifier,
  },
  [69] = {
    [1] = alias_sym_type_identifier,
  },
  [72] = {
    [1] = alias_sym_type_identifier,
  },
  [73] = {
    [1] = alias_sym_type_identifier,
  },
  [74] = {
    [1] = alias_sym_type_identifier,
  },
  [76] = {
    [2] = alias_sym_type_identifier,
  },
  [77] = {
    [0] = sym_identifier,
  },
  [78] = {
    [0] = sym_identifier,
  },
  [81] = {
    [0] = sym_identifier,
  },
  [87] = {
    [2] = alias_sym_type_identifier,
  },
  [93] = {
    [1] = alias_sym_type_identifier,
  },
  [97] = {
    [1] = alias_sym_shorthand_field_identifier,
  },
  [101] = {
    [0] = alias_sym_type_identifier,
  },
  [104] = {
    [1] = alias_sym_type_identifier,
  },
  [105] = {
    [1] = alias_sym_type_identifier,
  },
  [106] = {
    [0] = alias_sym_type_identifier,
  },
  [114] = {
    [3] = sym_identifier,
  },
  [115] = {
    [1] = alias_sym_type_identifier,
  },
  [116] = {
    [1] = alias_sym_type_identifier,
  },
  [117] = {
    [1] = alias_sym_type_identifier,
  },
  [118] = {
    [1] = alias_sym_type_identifier,
  },
  [122] = {
    [2] = alias_sym_type_identifier,
  },
  [123] = {
    [2] = alias_sym_type_identifier,
  },
  [124] = {
    [2] = alias_sym_type_identifier,
  },
  [125] = {
    [0] = sym_identifier,
  },
  [127] = {
    [0] = alias_sym_field_identifier,
  },
  [130] = {
    [2] = alias_sym_type_identifier,
  },
  [131] = {
    [3] = alias_sym_type_identifier,
  },
  [139] = {
    [2] = alias_sym_shorthand_field_identifier,
  },
  [140] = {
    [0] = alias_sym_field_identifier,
  },
  [141] = {
    [0] = alias_sym_type_identifier,
  },
  [143] = {
    [1] = alias_sym_type_identifier,
  },
  [145] = {
    [2] = alias_sym_type_identifier,
  },
  [146] = {
    [2] = alias_sym_type_identifier,
  },
  [149] = {
    [1] = alias_sym_type_identifier,
  },
  [160] = {
    [0] = alias_sym_field_identifier,
  },
  [161] = {
    [1] = alias_sym_type_identifier,
  },
  [162] = {
    [1] = alias_sym_type_identifier,
  },
  [163] = {
    [2] = alias_sym_type_identifier,
  },
  [164] = {
    [2] = alias_sym_type_identifier,
  },
  [168] = {
    [2] = alias_sym_type_identifier,
  },
  [169] = {
    [2] = alias_sym_type_identifier,
  },
  [170] = {
    [2] = alias_sym_type_identifier,
  },
  [172] = {
    [1] = alias_sym_field_identifier,
  },
  [176] = {
    [2] = alias_sym_type_identifier,
  },
  [177] = {
    [3] = alias_sym_type_identifier,
  },
  [178] = {
    [3] = alias_sym_type_identifier,
  },
  [179] = {
    [3] = alias_sym_type_identifier,
  },
  [189] = {
    [1] = alias_sym_field_identifier,
  },
  [190] = {
    [0] = alias_sym_type_identifier,
  },
  [191] = {
    [2] = alias_sym_type_identifier,
  },
  [199] = {
    [1] = alias_sym_field_identifier,
  },
  [200] = {
    [2] = alias_sym_type_identifier,
  },
  [201] = {
    [3] = alias_sym_type_identifier,
  },
  [202] = {
    [3] = alias_sym_type_identifier,
  },
  [206] = {
    [2] = alias_sym_type_identifier,
  },
  [210] = {
    [2] = alias_sym_type_identifier,
  },
  [211] = {
    [3] = alias_sym_type_identifier,
  },
  [212] = {
    [3] = alias_sym_type_identifier,
  },
  [213] = {
    [3] = alias_sym_type_identifier,
  },
  [221] = {
    [2] = alias_sym_field_identifier,
  },
  [227] = {
    [3] = alias_sym_type_identifier,
  },
  [233] = {
    [3] = alias_sym_type_identifier,
  },
};

static const uint16_t ts_non_terminal_alias_map[] = {
  sym_generic_type_with_turbofish, 2,
    sym_generic_type_with_turbofish,
    sym_generic_type,
  sym__let_chain, 2,
    sym__let_chain,
    alias_sym_let_chain,
  0,
};

static const TSStateId ts_primary_state_ids[STATE_COUNT] = {
  [0] = 0,
  [1] = 1,
  [2] = 2,
  [3] = 3,
  [4] = 3,
  [5] = 2,
  [6] = 2,
  [7] = 2,
  [8] = 8,
  [9] = 9,
  [10] = 2,
  [11] = 3,
  [12] = 8,
  [13] = 3,
  [14] = 3,
  [15] = 3,
  [16] = 2,
  [17] = 17,
  [18] = 18,
  [19] = 19,
  [20] = 19,
  [21] = 18,
  [22] = 22,
  [23] = 23,
  [24] = 24,
  [25] = 17,
  [26] = 18,
  [27] = 23,
  [28] = 19,
  [29] = 24,
  [30] = 22,
  [31] = 18,
  [32] = 19,
  [33] = 33,
  [34] = 34,
  [35] = 35,
  [36] = 36,
  [37] = 37,
  [38] = 38,
  [39] = 39,
  [40] = 40,
  [41] = 37,
  [42] = 42,
  [43] = 39,
  [44] = 40,
  [45] = 45,
  [46] = 40,
  [47] = 47,
  [48] = 48,
  [49] = 39,
  [50] = 37,
  [51] = 51,
  [52] = 52,
  [53] = 53,
  [54] = 54,
  [55] = 55,
  [56] = 56,
  [57] = 57,
  [58] = 58,
  [59] = 58,
  [60] = 60,
  [61] = 61,
  [62] = 62,
  [63] = 63,
  [64] = 64,
  [65] = 65,
  [66] = 66,
  [67] = 67,
  [68] = 68,
  [69] = 69,
  [70] = 70,
  [71] = 71,
  [72] = 72,
  [73] = 73,
  [74] = 74,
  [75] = 75,
  [76] = 76,
  [77] = 77,
  [78] = 78,
  [79] = 79,
  [80] = 80,
  [81] = 81,
  [82] = 80,
  [83] = 83,
  [84] = 84,
  [85] = 85,
  [86] = 86,
  [87] = 87,
  [88] = 88,
  [89] = 89,
  [90] = 90,
  [91] = 91,
  [92] = 92,
  [93] = 93,
  [94] = 94,
  [95] = 91,
  [96] = 68,
  [97] = 94,
  [98] = 90,
  [99] = 99,
  [100] = 100,
  [101] = 101,
  [102] = 102,
  [103] = 103,
  [104] = 104,
  [105] = 105,
  [106] = 106,
  [107] = 107,
  [108] = 108,
  [109] = 109,
  [110] = 110,
  [111] = 111,
  [112] = 112,
  [113] = 113,
  [114] = 114,
  [115] = 115,
  [116] = 111,
  [117] = 117,
  [118] = 118,
  [119] = 112,
  [120] = 120,
  [121] = 121,
  [122] = 117,
  [123] = 123,
  [124] = 124,
  [125] = 111,
  [126] = 126,
  [127] = 127,
  [128] = 128,
  [129] = 105,
  [130] = 113,
  [131] = 131,
  [132] = 106,
  [133] = 133,
  [134] = 134,
  [135] = 135,
  [136] = 136,
  [137] = 137,
  [138] = 138,
  [139] = 109,
  [140] = 124,
  [141] = 110,
  [142] = 120,
  [143] = 128,
  [144] = 103,
  [145] = 145,
  [146] = 136,
  [147] = 147,
  [148] = 148,
  [149] = 149,
  [150] = 149,
  [151] = 103,
  [152] = 152,
  [153] = 153,
  [154] = 154,
  [155] = 120,
  [156] = 156,
  [157] = 108,
  [158] = 100,
  [159] = 126,
  [160] = 160,
  [161] = 148,
  [162] = 100,
  [163] = 100,
  [164] = 133,
  [165] = 165,
  [166] = 166,
  [167] = 165,
  [168] = 121,
  [169] = 169,
  [170] = 115,
  [171] = 169,
  [172] = 166,
  [173] = 127,
  [174] = 107,
  [175] = 153,
  [176] = 135,
  [177] = 152,
  [178] = 178,
  [179] = 179,
  [180] = 179,
  [181] = 179,
  [182] = 182,
  [183] = 182,
  [184] = 184,
  [185] = 185,
  [186] = 186,
  [187] = 185,
  [188] = 186,
  [189] = 184,
  [190] = 190,
  [191] = 191,
  [192] = 192,
  [193] = 192,
  [194] = 194,
  [195] = 195,
  [196] = 195,
  [197] = 195,
  [198] = 198,
  [199] = 199,
  [200] = 200,
  [201] = 200,
  [202] = 202,
  [203] = 202,
  [204] = 204,
  [205] = 202,
  [206] = 199,
  [207] = 200,
  [208] = 204,
  [209] = 209,
  [210] = 48,
  [211] = 53,
  [212] = 56,
  [213] = 54,
  [214] = 214,
  [215] = 60,
  [216] = 85,
  [217] = 72,
  [218] = 218,
  [219] = 75,
  [220] = 62,
  [221] = 74,
  [222] = 77,
  [223] = 76,
  [224] = 69,
  [225] = 73,
  [226] = 63,
  [227] = 67,
  [228] = 61,
  [229] = 65,
  [230] = 230,
  [231] = 230,
  [232] = 232,
  [233] = 66,
  [234] = 84,
  [235] = 232,
  [236] = 236,
  [237] = 237,
  [238] = 237,
  [239] = 239,
  [240] = 240,
  [241] = 241,
  [242] = 242,
  [243] = 242,
  [244] = 241,
  [245] = 56,
  [246] = 246,
  [247] = 247,
  [248] = 248,
  [249] = 249,
  [250] = 250,
  [251] = 251,
  [252] = 252,
  [253] = 253,
  [254] = 254,
  [255] = 255,
  [256] = 256,
  [257] = 257,
  [258] = 258,
  [259] = 259,
  [260] = 260,
  [261] = 261,
  [262] = 262,
  [263] = 263,
  [264] = 264,
  [265] = 265,
  [266] = 266,
  [267] = 267,
  [268] = 268,
  [269] = 269,
  [270] = 270,
  [271] = 271,
  [272] = 272,
  [273] = 273,
  [274] = 274,
  [275] = 275,
  [276] = 276,
  [277] = 277,
  [278] = 278,
  [279] = 279,
  [280] = 280,
  [281] = 281,
  [282] = 282,
  [283] = 283,
  [284] = 284,
  [285] = 285,
  [286] = 286,
  [287] = 287,
  [288] = 288,
  [289] = 289,
  [290] = 290,
  [291] = 291,
  [292] = 292,
  [293] = 293,
  [294] = 294,
  [295] = 295,
  [296] = 296,
  [297] = 297,
  [298] = 298,
  [299] = 299,
  [300] = 300,
  [301] = 301,
  [302] = 302,
  [303] = 303,
  [304] = 304,
  [305] = 305,
  [306] = 306,
  [307] = 307,
  [308] = 308,
  [309] = 309,
  [310] = 310,
  [311] = 311,
  [312] = 312,
  [313] = 313,
  [314] = 314,
  [315] = 315,
  [316] = 316,
  [317] = 317,
  [318] = 318,
  [319] = 319,
  [320] = 320,
  [321] = 321,
  [322] = 322,
  [323] = 323,
  [324] = 324,
  [325] = 325,
  [326] = 326,
  [327] = 327,
  [328] = 328,
  [329] = 329,
  [330] = 53,
  [331] = 331,
  [332] = 332,
  [333] = 333,
  [334] = 334,
  [335] = 335,
  [336] = 336,
  [337] = 54,
  [338] = 338,
  [339] = 339,
  [340] = 340,
  [341] = 341,
  [342] = 342,
  [343] = 343,
  [344] = 344,
  [345] = 345,
  [346] = 346,
  [347] = 347,
  [348] = 348,
  [349] = 349,
  [350] = 350,
  [351] = 351,
  [352] = 352,
  [353] = 353,
  [354] = 354,
  [355] = 355,
  [356] = 356,
  [357] = 357,
  [358] = 358,
  [359] = 359,
  [360] = 360,
  [361] = 361,
  [362] = 362,
  [363] = 363,
  [364] = 364,
  [365] = 365,
  [366] = 366,
  [367] = 367,
  [368] = 368,
  [369] = 369,
  [370] = 370,
  [371] = 371,
  [372] = 372,
  [373] = 373,
  [374] = 374,
  [375] = 375,
  [376] = 376,
  [377] = 377,
  [378] = 378,
  [379] = 379,
  [380] = 380,
  [381] = 381,
  [382] = 382,
  [383] = 383,
  [384] = 384,
  [385] = 385,
  [386] = 386,
  [387] = 387,
  [388] = 388,
  [389] = 389,
  [390] = 390,
  [391] = 365,
  [392] = 392,
  [393] = 393,
  [394] = 394,
  [395] = 395,
  [396] = 396,
  [397] = 397,
  [398] = 398,
  [399] = 399,
  [400] = 400,
  [401] = 401,
  [402] = 402,
  [403] = 403,
  [404] = 404,
  [405] = 405,
  [406] = 406,
  [407] = 407,
  [408] = 408,
  [409] = 409,
  [410] = 410,
  [411] = 411,
  [412] = 412,
  [413] = 413,
  [414] = 414,
  [415] = 415,
  [416] = 416,
  [417] = 417,
  [418] = 418,
  [419] = 419,
  [420] = 420,
  [421] = 421,
  [422] = 422,
  [423] = 423,
  [424] = 424,
  [425] = 425,
  [426] = 426,
  [427] = 427,
  [428] = 428,
  [429] = 429,
  [430] = 430,
  [431] = 431,
  [432] = 432,
  [433] = 433,
  [434] = 434,
  [435] = 435,
  [436] = 436,
  [437] = 437,
  [438] = 438,
  [439] = 439,
  [440] = 440,
  [441] = 441,
  [442] = 442,
  [443] = 443,
  [444] = 444,
  [445] = 445,
  [446] = 446,
  [447] = 447,
  [448] = 448,
  [449] = 449,
  [450] = 450,
  [451] = 451,
  [452] = 452,
  [453] = 453,
  [454] = 454,
  [455] = 455,
  [456] = 456,
  [457] = 457,
  [458] = 458,
  [459] = 459,
  [460] = 460,
  [461] = 461,
  [462] = 462,
  [463] = 463,
  [464] = 464,
  [465] = 465,
  [466] = 466,
  [467] = 467,
  [468] = 468,
  [469] = 365,
  [470] = 470,
  [471] = 471,
  [472] = 472,
  [473] = 473,
  [474] = 474,
  [475] = 475,
  [476] = 476,
  [477] = 477,
  [478] = 478,
  [479] = 479,
  [480] = 480,
  [481] = 481,
  [482] = 482,
  [483] = 483,
  [484] = 484,
  [485] = 485,
  [486] = 484,
  [487] = 487,
  [488] = 488,
  [489] = 489,
  [490] = 484,
  [491] = 491,
  [492] = 492,
  [493] = 489,
  [494] = 494,
  [495] = 487,
  [496] = 492,
  [497] = 488,
  [498] = 481,
  [499] = 494,
  [500] = 500,
  [501] = 501,
  [502] = 502,
  [503] = 503,
  [504] = 504,
  [505] = 505,
  [506] = 506,
  [507] = 505,
  [508] = 508,
  [509] = 509,
  [510] = 504,
  [511] = 505,
  [512] = 501,
  [513] = 502,
  [514] = 514,
  [515] = 515,
  [516] = 516,
  [517] = 508,
  [518] = 506,
  [519] = 502,
  [520] = 508,
  [521] = 505,
  [522] = 509,
  [523] = 509,
  [524] = 504,
  [525] = 525,
  [526] = 508,
  [527] = 504,
  [528] = 501,
  [529] = 502,
  [530] = 508,
  [531] = 509,
  [532] = 509,
  [533] = 514,
  [534] = 505,
  [535] = 516,
  [536] = 502,
  [537] = 537,
  [538] = 525,
  [539] = 539,
  [540] = 504,
  [541] = 503,
  [542] = 501,
  [543] = 501,
  [544] = 515,
  [545] = 545,
  [546] = 546,
  [547] = 547,
  [548] = 548,
  [549] = 549,
  [550] = 550,
  [551] = 551,
  [552] = 552,
  [553] = 553,
  [554] = 554,
  [555] = 555,
  [556] = 556,
  [557] = 557,
  [558] = 558,
  [559] = 559,
  [560] = 560,
  [561] = 561,
  [562] = 562,
  [563] = 563,
  [564] = 564,
  [565] = 565,
  [566] = 566,
  [567] = 567,
  [568] = 568,
  [569] = 569,
  [570] = 570,
  [571] = 571,
  [572] = 572,
  [573] = 573,
  [574] = 574,
  [575] = 575,
  [576] = 576,
  [577] = 577,
  [578] = 578,
  [579] = 579,
  [580] = 580,
  [581] = 581,
  [582] = 582,
  [583] = 583,
  [584] = 584,
  [585] = 585,
  [586] = 586,
  [587] = 587,
  [588] = 588,
  [589] = 584,
  [590] = 590,
  [591] = 587,
  [592] = 585,
  [593] = 593,
  [594] = 594,
  [595] = 595,
  [596] = 596,
  [597] = 597,
  [598] = 598,
  [599] = 599,
  [600] = 600,
  [601] = 567,
  [602] = 602,
  [603] = 603,
  [604] = 604,
  [605] = 64,
  [606] = 606,
  [607] = 607,
  [608] = 608,
  [609] = 609,
  [610] = 610,
  [611] = 611,
  [612] = 597,
  [613] = 600,
  [614] = 575,
  [615] = 88,
  [616] = 598,
  [617] = 597,
  [618] = 587,
  [619] = 611,
  [620] = 620,
  [621] = 621,
  [622] = 622,
  [623] = 622,
  [624] = 622,
  [625] = 625,
  [626] = 626,
  [627] = 626,
  [628] = 625,
  [629] = 629,
  [630] = 630,
  [631] = 631,
  [632] = 629,
  [633] = 633,
  [634] = 634,
  [635] = 630,
  [636] = 626,
  [637] = 633,
  [638] = 631,
  [639] = 639,
  [640] = 640,
  [641] = 641,
  [642] = 642,
  [643] = 643,
  [644] = 644,
  [645] = 645,
  [646] = 646,
  [647] = 647,
  [648] = 648,
  [649] = 640,
  [650] = 650,
  [651] = 651,
  [652] = 652,
  [653] = 653,
  [654] = 654,
  [655] = 655,
  [656] = 656,
  [657] = 657,
  [658] = 658,
  [659] = 659,
  [660] = 650,
  [661] = 661,
  [662] = 643,
  [663] = 663,
  [664] = 664,
  [665] = 665,
  [666] = 666,
  [667] = 667,
  [668] = 668,
  [669] = 669,
  [670] = 670,
  [671] = 671,
  [672] = 639,
  [673] = 673,
  [674] = 644,
  [675] = 675,
  [676] = 676,
  [677] = 677,
  [678] = 678,
  [679] = 648,
  [680] = 652,
  [681] = 681,
  [682] = 663,
  [683] = 681,
  [684] = 684,
  [685] = 654,
  [686] = 686,
  [687] = 676,
  [688] = 642,
  [689] = 689,
  [690] = 690,
  [691] = 691,
  [692] = 651,
  [693] = 693,
  [694] = 694,
  [695] = 695,
  [696] = 696,
  [697] = 697,
  [698] = 698,
  [699] = 665,
  [700] = 700,
  [701] = 701,
  [702] = 702,
  [703] = 703,
  [704] = 704,
  [705] = 705,
  [706] = 655,
  [707] = 698,
  [708] = 708,
  [709] = 709,
  [710] = 710,
  [711] = 711,
  [712] = 704,
  [713] = 658,
  [714] = 711,
  [715] = 667,
  [716] = 689,
  [717] = 708,
  [718] = 661,
  [719] = 670,
  [720] = 675,
  [721] = 694,
  [722] = 657,
  [723] = 659,
  [724] = 646,
  [725] = 644,
  [726] = 664,
  [727] = 653,
  [728] = 696,
  [729] = 695,
  [730] = 666,
  [731] = 701,
  [732] = 318,
  [733] = 733,
  [734] = 710,
  [735] = 735,
  [736] = 703,
  [737] = 705,
  [738] = 738,
  [739] = 709,
  [740] = 640,
  [741] = 741,
  [742] = 742,
  [743] = 690,
  [744] = 678,
  [745] = 653,
  [746] = 746,
  [747] = 747,
  [748] = 748,
  [749] = 749,
  [750] = 750,
  [751] = 318,
  [752] = 752,
  [753] = 753,
  [754] = 754,
  [755] = 755,
  [756] = 756,
  [757] = 757,
  [758] = 758,
  [759] = 759,
  [760] = 760,
  [761] = 761,
  [762] = 762,
  [763] = 461,
  [764] = 764,
  [765] = 765,
  [766] = 766,
  [767] = 464,
  [768] = 320,
  [769] = 769,
  [770] = 770,
  [771] = 771,
  [772] = 772,
  [773] = 256,
  [774] = 326,
  [775] = 775,
  [776] = 776,
  [777] = 777,
  [778] = 778,
  [779] = 779,
  [780] = 780,
  [781] = 781,
  [782] = 782,
  [783] = 783,
  [784] = 198,
  [785] = 785,
  [786] = 786,
  [787] = 787,
  [788] = 788,
  [789] = 789,
  [790] = 790,
  [791] = 791,
  [792] = 792,
  [793] = 793,
  [794] = 794,
  [795] = 795,
  [796] = 796,
  [797] = 797,
  [798] = 798,
  [799] = 799,
  [800] = 800,
  [801] = 48,
  [802] = 802,
  [803] = 803,
  [804] = 804,
  [805] = 805,
  [806] = 806,
  [807] = 807,
  [808] = 290,
  [809] = 306,
  [810] = 369,
  [811] = 384,
  [812] = 387,
  [813] = 813,
  [814] = 392,
  [815] = 585,
  [816] = 816,
  [817] = 817,
  [818] = 62,
  [819] = 401,
  [820] = 402,
  [821] = 403,
  [822] = 404,
  [823] = 405,
  [824] = 406,
  [825] = 407,
  [826] = 408,
  [827] = 411,
  [828] = 413,
  [829] = 414,
  [830] = 415,
  [831] = 419,
  [832] = 423,
  [833] = 425,
  [834] = 426,
  [835] = 428,
  [836] = 431,
  [837] = 433,
  [838] = 838,
  [839] = 434,
  [840] = 435,
  [841] = 436,
  [842] = 842,
  [843] = 843,
  [844] = 440,
  [845] = 246,
  [846] = 443,
  [847] = 444,
  [848] = 450,
  [849] = 452,
  [850] = 456,
  [851] = 468,
  [852] = 77,
  [853] = 476,
  [854] = 477,
  [855] = 479,
  [856] = 856,
  [857] = 480,
  [858] = 412,
  [859] = 475,
  [860] = 471,
  [861] = 470,
  [862] = 862,
  [863] = 465,
  [864] = 463,
  [865] = 865,
  [866] = 866,
  [867] = 867,
  [868] = 868,
  [869] = 66,
  [870] = 69,
  [871] = 871,
  [872] = 872,
  [873] = 453,
  [874] = 451,
  [875] = 447,
  [876] = 446,
  [877] = 420,
  [878] = 416,
  [879] = 88,
  [880] = 398,
  [881] = 375,
  [882] = 373,
  [883] = 75,
  [884] = 357,
  [885] = 338,
  [886] = 335,
  [887] = 327,
  [888] = 325,
  [889] = 324,
  [890] = 323,
  [891] = 891,
  [892] = 321,
  [893] = 319,
  [894] = 388,
  [895] = 383,
  [896] = 317,
  [897] = 314,
  [898] = 898,
  [899] = 311,
  [900] = 304,
  [901] = 302,
  [902] = 301,
  [903] = 297,
  [904] = 293,
  [905] = 905,
  [906] = 417,
  [907] = 418,
  [908] = 273,
  [909] = 272,
  [910] = 271,
  [911] = 270,
  [912] = 257,
  [913] = 362,
  [914] = 422,
  [915] = 73,
  [916] = 254,
  [917] = 360,
  [918] = 83,
  [919] = 424,
  [920] = 432,
  [921] = 342,
  [922] = 922,
  [923] = 349,
  [924] = 924,
  [925] = 437,
  [926] = 252,
  [927] = 389,
  [928] = 251,
  [929] = 348,
  [930] = 81,
  [931] = 931,
  [932] = 932,
  [933] = 250,
  [934] = 249,
  [935] = 935,
  [936] = 347,
  [937] = 438,
  [938] = 439,
  [939] = 346,
  [940] = 454,
  [941] = 941,
  [942] = 942,
  [943] = 457,
  [944] = 458,
  [945] = 345,
  [946] = 946,
  [947] = 344,
  [948] = 948,
  [949] = 949,
  [950] = 950,
  [951] = 951,
  [952] = 74,
  [953] = 953,
  [954] = 343,
  [955] = 247,
  [956] = 331,
  [957] = 356,
  [958] = 341,
  [959] = 339,
  [960] = 336,
  [961] = 85,
  [962] = 329,
  [963] = 394,
  [964] = 395,
  [965] = 427,
  [966] = 64,
  [967] = 54,
  [968] = 448,
  [969] = 455,
  [970] = 467,
  [971] = 466,
  [972] = 445,
  [973] = 81,
  [974] = 72,
  [975] = 409,
  [976] = 390,
  [977] = 386,
  [978] = 382,
  [979] = 979,
  [980] = 377,
  [981] = 309,
  [982] = 374,
  [983] = 368,
  [984] = 367,
  [985] = 60,
  [986] = 308,
  [987] = 987,
  [988] = 361,
  [989] = 567,
  [990] = 307,
  [991] = 460,
  [992] = 364,
  [993] = 305,
  [994] = 303,
  [995] = 300,
  [996] = 298,
  [997] = 352,
  [998] = 350,
  [999] = 296,
  [1000] = 294,
  [1001] = 292,
  [1002] = 1002,
  [1003] = 462,
  [1004] = 291,
  [1005] = 1005,
  [1006] = 289,
  [1007] = 288,
  [1008] = 287,
  [1009] = 1009,
  [1010] = 286,
  [1011] = 1011,
  [1012] = 1012,
  [1013] = 285,
  [1014] = 284,
  [1015] = 1015,
  [1016] = 282,
  [1017] = 1017,
  [1018] = 281,
  [1019] = 279,
  [1020] = 276,
  [1021] = 275,
  [1022] = 277,
  [1023] = 83,
  [1024] = 84,
  [1025] = 53,
  [1026] = 478,
  [1027] = 474,
  [1028] = 584,
  [1029] = 1029,
  [1030] = 269,
  [1031] = 1031,
  [1032] = 280,
  [1033] = 283,
  [1034] = 268,
  [1035] = 266,
  [1036] = 264,
  [1037] = 259,
  [1038] = 61,
  [1039] = 63,
  [1040] = 1040,
  [1041] = 255,
  [1042] = 1042,
  [1043] = 248,
  [1044] = 473,
  [1045] = 67,
  [1046] = 1046,
  [1047] = 76,
  [1048] = 400,
  [1049] = 56,
  [1050] = 442,
  [1051] = 1051,
  [1052] = 1052,
  [1053] = 472,
  [1054] = 1054,
  [1055] = 253,
  [1056] = 261,
  [1057] = 262,
  [1058] = 381,
  [1059] = 263,
  [1060] = 1060,
  [1061] = 1061,
  [1062] = 265,
  [1063] = 370,
  [1064] = 359,
  [1065] = 459,
  [1066] = 267,
  [1067] = 1067,
  [1068] = 376,
  [1069] = 1069,
  [1070] = 1070,
  [1071] = 1071,
  [1072] = 358,
  [1073] = 1073,
  [1074] = 322,
  [1075] = 274,
  [1076] = 278,
  [1077] = 295,
  [1078] = 299,
  [1079] = 1079,
  [1080] = 318,
  [1081] = 1081,
  [1082] = 1082,
  [1083] = 1083,
  [1084] = 310,
  [1085] = 449,
  [1086] = 312,
  [1087] = 1087,
  [1088] = 313,
  [1089] = 315,
  [1090] = 441,
  [1091] = 316,
  [1092] = 328,
  [1093] = 332,
  [1094] = 333,
  [1095] = 334,
  [1096] = 340,
  [1097] = 351,
  [1098] = 353,
  [1099] = 54,
  [1100] = 354,
  [1101] = 355,
  [1102] = 53,
  [1103] = 363,
  [1104] = 366,
  [1105] = 430,
  [1106] = 371,
  [1107] = 56,
  [1108] = 372,
  [1109] = 429,
  [1110] = 258,
  [1111] = 1111,
  [1112] = 378,
  [1113] = 379,
  [1114] = 88,
  [1115] = 380,
  [1116] = 385,
  [1117] = 421,
  [1118] = 575,
  [1119] = 1119,
  [1120] = 393,
  [1121] = 1121,
  [1122] = 396,
  [1123] = 397,
  [1124] = 64,
  [1125] = 399,
  [1126] = 410,
  [1127] = 65,
  [1128] = 1128,
  [1129] = 1129,
  [1130] = 1130,
  [1131] = 1131,
  [1132] = 1132,
  [1133] = 1133,
  [1134] = 1134,
  [1135] = 1135,
  [1136] = 1136,
  [1137] = 1137,
  [1138] = 1138,
  [1139] = 1139,
  [1140] = 1140,
  [1141] = 1141,
  [1142] = 1142,
  [1143] = 1143,
  [1144] = 1144,
  [1145] = 1145,
  [1146] = 1146,
  [1147] = 1147,
  [1148] = 1148,
  [1149] = 1149,
  [1150] = 1150,
  [1151] = 1151,
  [1152] = 1152,
  [1153] = 621,
  [1154] = 1154,
  [1155] = 1155,
  [1156] = 1156,
  [1157] = 1157,
  [1158] = 1158,
  [1159] = 1159,
  [1160] = 1160,
  [1161] = 758,
  [1162] = 1162,
  [1163] = 318,
  [1164] = 1164,
  [1165] = 1165,
  [1166] = 1166,
  [1167] = 1165,
  [1168] = 1168,
  [1169] = 1164,
  [1170] = 1015,
  [1171] = 758,
  [1172] = 792,
  [1173] = 1173,
  [1174] = 1174,
  [1175] = 1175,
  [1176] = 792,
  [1177] = 780,
  [1178] = 1178,
  [1179] = 1179,
  [1180] = 1180,
  [1181] = 769,
  [1182] = 1182,
  [1183] = 1183,
  [1184] = 1184,
  [1185] = 1185,
  [1186] = 766,
  [1187] = 764,
  [1188] = 1188,
  [1189] = 772,
  [1190] = 793,
  [1191] = 796,
  [1192] = 1031,
  [1193] = 1193,
  [1194] = 1119,
  [1195] = 1195,
  [1196] = 1195,
  [1197] = 1197,
  [1198] = 1136,
  [1199] = 1199,
  [1200] = 1155,
  [1201] = 1201,
  [1202] = 1199,
  [1203] = 1157,
  [1204] = 1204,
  [1205] = 1154,
  [1206] = 1206,
  [1207] = 1207,
  [1208] = 1130,
  [1209] = 1209,
  [1210] = 1210,
  [1211] = 1211,
  [1212] = 1212,
  [1213] = 1213,
  [1214] = 1201,
  [1215] = 1215,
  [1216] = 1213,
  [1217] = 1210,
  [1218] = 1218,
  [1219] = 1206,
  [1220] = 1206,
  [1221] = 1148,
  [1222] = 1211,
  [1223] = 1223,
  [1224] = 1140,
  [1225] = 1225,
  [1226] = 1226,
  [1227] = 1227,
  [1228] = 1213,
  [1229] = 1229,
  [1230] = 1213,
  [1231] = 1209,
  [1232] = 1143,
  [1233] = 1204,
  [1234] = 1234,
  [1235] = 1144,
  [1236] = 1146,
  [1237] = 1211,
  [1238] = 1138,
  [1239] = 1206,
  [1240] = 1147,
  [1241] = 1129,
  [1242] = 1206,
  [1243] = 1243,
  [1244] = 1128,
  [1245] = 1245,
  [1246] = 1246,
  [1247] = 1152,
  [1248] = 1210,
  [1249] = 1158,
  [1250] = 1250,
  [1251] = 1246,
  [1252] = 1252,
  [1253] = 1142,
  [1254] = 1254,
  [1255] = 1255,
  [1256] = 1145,
  [1257] = 1206,
  [1258] = 1258,
  [1259] = 1201,
  [1260] = 1213,
  [1261] = 1213,
  [1262] = 1156,
  [1263] = 1132,
  [1264] = 1264,
  [1265] = 1265,
  [1266] = 1266,
  [1267] = 1267,
  [1268] = 1268,
  [1269] = 1269,
  [1270] = 1270,
  [1271] = 1271,
  [1272] = 1272,
  [1273] = 1273,
  [1274] = 1274,
  [1275] = 1275,
  [1276] = 1276,
  [1277] = 1277,
  [1278] = 1269,
  [1279] = 1265,
  [1280] = 1270,
  [1281] = 1275,
  [1282] = 1282,
  [1283] = 1268,
  [1284] = 1271,
  [1285] = 1274,
  [1286] = 1286,
  [1287] = 1272,
  [1288] = 1277,
  [1289] = 1289,
  [1290] = 1290,
  [1291] = 1291,
  [1292] = 1292,
  [1293] = 1293,
  [1294] = 1267,
  [1295] = 1295,
  [1296] = 1296,
  [1297] = 1297,
  [1298] = 1297,
  [1299] = 1299,
  [1300] = 1299,
  [1301] = 1301,
  [1302] = 1302,
  [1303] = 1303,
  [1304] = 1304,
  [1305] = 1305,
  [1306] = 1305,
  [1307] = 1304,
  [1308] = 1305,
  [1309] = 1305,
  [1310] = 1305,
  [1311] = 1311,
  [1312] = 1311,
  [1313] = 567,
  [1314] = 575,
  [1315] = 1315,
  [1316] = 752,
  [1317] = 762,
  [1318] = 755,
  [1319] = 753,
  [1320] = 756,
  [1321] = 755,
  [1322] = 762,
  [1323] = 756,
  [1324] = 761,
  [1325] = 761,
  [1326] = 754,
  [1327] = 759,
  [1328] = 1328,
  [1329] = 757,
  [1330] = 1330,
  [1331] = 760,
  [1332] = 1332,
  [1333] = 81,
  [1334] = 1334,
  [1335] = 88,
  [1336] = 83,
  [1337] = 1337,
  [1338] = 64,
  [1339] = 1334,
  [1340] = 789,
  [1341] = 1183,
  [1342] = 797,
  [1343] = 1015,
  [1344] = 804,
  [1345] = 782,
  [1346] = 785,
  [1347] = 799,
  [1348] = 1348,
  [1349] = 803,
  [1350] = 791,
  [1351] = 1351,
  [1352] = 779,
  [1353] = 781,
  [1354] = 776,
  [1355] = 805,
  [1356] = 783,
  [1357] = 802,
  [1358] = 584,
  [1359] = 790,
  [1360] = 1180,
  [1361] = 807,
  [1362] = 786,
  [1363] = 798,
  [1364] = 1364,
  [1365] = 795,
  [1366] = 1366,
  [1367] = 806,
  [1368] = 788,
  [1369] = 1369,
  [1370] = 775,
  [1371] = 1184,
  [1372] = 1185,
  [1373] = 1369,
  [1374] = 787,
  [1375] = 1369,
  [1376] = 950,
  [1377] = 1042,
  [1378] = 868,
  [1379] = 867,
  [1380] = 1380,
  [1381] = 932,
  [1382] = 931,
  [1383] = 865,
  [1384] = 1012,
  [1385] = 946,
  [1386] = 949,
  [1387] = 862,
  [1388] = 1017,
  [1389] = 1389,
  [1390] = 1040,
  [1391] = 771,
  [1392] = 935,
  [1393] = 1393,
  [1394] = 1083,
  [1395] = 1395,
  [1396] = 1396,
  [1397] = 1397,
  [1398] = 941,
  [1399] = 1005,
  [1400] = 951,
  [1401] = 1002,
  [1402] = 953,
  [1403] = 922,
  [1404] = 905,
  [1405] = 856,
  [1406] = 1406,
  [1407] = 56,
  [1408] = 1395,
  [1409] = 1409,
  [1410] = 1410,
  [1411] = 1411,
  [1412] = 1412,
  [1413] = 1413,
  [1414] = 54,
  [1415] = 53,
  [1416] = 1416,
  [1417] = 1417,
  [1418] = 1418,
  [1419] = 762,
  [1420] = 1420,
  [1421] = 1421,
  [1422] = 1422,
  [1423] = 756,
  [1424] = 761,
  [1425] = 1425,
  [1426] = 1426,
  [1427] = 1427,
  [1428] = 1428,
  [1429] = 1429,
  [1430] = 756,
  [1431] = 762,
  [1432] = 1432,
  [1433] = 755,
  [1434] = 1434,
  [1435] = 755,
  [1436] = 1436,
  [1437] = 1437,
  [1438] = 1438,
  [1439] = 1395,
  [1440] = 1440,
  [1441] = 1441,
  [1442] = 1442,
  [1443] = 1443,
  [1444] = 1444,
  [1445] = 761,
  [1446] = 1446,
  [1447] = 1447,
  [1448] = 1448,
  [1449] = 761,
  [1450] = 1450,
  [1451] = 1451,
  [1452] = 1452,
  [1453] = 1453,
  [1454] = 1454,
  [1455] = 67,
  [1456] = 1456,
  [1457] = 755,
  [1458] = 1458,
  [1459] = 756,
  [1460] = 1460,
  [1461] = 762,
  [1462] = 1462,
  [1463] = 1462,
  [1464] = 1464,
  [1465] = 1465,
  [1466] = 1465,
  [1467] = 1464,
  [1468] = 1468,
  [1469] = 1469,
  [1470] = 1469,
  [1471] = 1471,
  [1472] = 1471,
  [1473] = 1468,
  [1474] = 1474,
  [1475] = 1475,
  [1476] = 1476,
  [1477] = 1477,
  [1478] = 1478,
  [1479] = 1478,
  [1480] = 1480,
  [1481] = 1480,
  [1482] = 1482,
  [1483] = 1483,
  [1484] = 1484,
  [1485] = 1485,
  [1486] = 1486,
  [1487] = 1487,
  [1488] = 1488,
  [1489] = 1487,
  [1490] = 1490,
  [1491] = 1483,
  [1492] = 1492,
  [1493] = 1493,
  [1494] = 1494,
  [1495] = 1495,
  [1496] = 1486,
  [1497] = 1497,
  [1498] = 1495,
  [1499] = 1493,
  [1500] = 1500,
  [1501] = 1484,
  [1502] = 1490,
  [1503] = 1494,
  [1504] = 1504,
  [1505] = 1337,
  [1506] = 1482,
  [1507] = 1497,
  [1508] = 1508,
  [1509] = 1509,
  [1510] = 1510,
  [1511] = 1511,
  [1512] = 1512,
  [1513] = 1512,
  [1514] = 1485,
  [1515] = 1500,
  [1516] = 1516,
  [1517] = 1511,
  [1518] = 1508,
  [1519] = 1509,
  [1520] = 1492,
  [1521] = 1516,
  [1522] = 1522,
  [1523] = 1523,
  [1524] = 1524,
  [1525] = 1525,
  [1526] = 1526,
  [1527] = 1527,
  [1528] = 1337,
  [1529] = 1529,
  [1530] = 1530,
  [1531] = 1531,
  [1532] = 1532,
  [1533] = 1531,
  [1534] = 1534,
  [1535] = 1535,
  [1536] = 1536,
  [1537] = 1537,
  [1538] = 1538,
  [1539] = 1539,
  [1540] = 1540,
  [1541] = 1541,
  [1542] = 1542,
  [1543] = 1542,
  [1544] = 1544,
  [1545] = 1545,
  [1546] = 1545,
  [1547] = 1547,
  [1548] = 1548,
  [1549] = 1549,
  [1550] = 1550,
  [1551] = 1551,
  [1552] = 1545,
  [1553] = 1553,
  [1554] = 1554,
  [1555] = 1534,
  [1556] = 1556,
  [1557] = 1348,
  [1558] = 1558,
  [1559] = 1559,
  [1560] = 1559,
  [1561] = 1561,
  [1562] = 1527,
  [1563] = 1532,
  [1564] = 1538,
  [1565] = 1337,
  [1566] = 1566,
  [1567] = 1567,
  [1568] = 1568,
  [1569] = 1569,
  [1570] = 1570,
  [1571] = 1571,
  [1572] = 1572,
  [1573] = 1573,
  [1574] = 1574,
  [1575] = 1575,
  [1576] = 1576,
  [1577] = 1577,
  [1578] = 1578,
  [1579] = 1579,
  [1580] = 1580,
  [1581] = 1579,
  [1582] = 1576,
  [1583] = 1569,
  [1584] = 1584,
  [1585] = 1573,
  [1586] = 1586,
  [1587] = 1587,
  [1588] = 1588,
  [1589] = 1589,
  [1590] = 1590,
  [1591] = 1591,
  [1592] = 1580,
  [1593] = 1593,
  [1594] = 1594,
  [1595] = 1595,
  [1596] = 1348,
  [1597] = 1597,
  [1598] = 1594,
  [1599] = 1599,
  [1600] = 1586,
  [1601] = 1584,
  [1602] = 1597,
  [1603] = 1603,
  [1604] = 1604,
  [1605] = 1587,
  [1606] = 1606,
  [1607] = 1607,
  [1608] = 1608,
  [1609] = 1571,
  [1610] = 1590,
  [1611] = 1588,
  [1612] = 1612,
  [1613] = 1613,
  [1614] = 1614,
  [1615] = 1615,
  [1616] = 1616,
  [1617] = 1617,
  [1618] = 1618,
  [1619] = 1619,
  [1620] = 1620,
  [1621] = 1621,
  [1622] = 1622,
  [1623] = 1623,
  [1624] = 1624,
  [1625] = 1613,
  [1626] = 1626,
  [1627] = 1612,
  [1628] = 1628,
  [1629] = 1629,
  [1630] = 1630,
  [1631] = 1631,
  [1632] = 1632,
  [1633] = 1633,
  [1634] = 1634,
  [1635] = 1635,
  [1636] = 1624,
  [1637] = 1637,
  [1638] = 1638,
  [1639] = 1639,
  [1640] = 1640,
  [1641] = 1635,
  [1642] = 1640,
  [1643] = 1637,
  [1644] = 1620,
  [1645] = 1645,
  [1646] = 1646,
  [1647] = 1647,
  [1648] = 1626,
  [1649] = 1646,
  [1650] = 1650,
  [1651] = 1639,
  [1652] = 1623,
  [1653] = 1348,
  [1654] = 1654,
  [1655] = 1619,
  [1656] = 1656,
  [1657] = 1657,
  [1658] = 1658,
  [1659] = 1650,
  [1660] = 1647,
  [1661] = 1661,
  [1662] = 1645,
  [1663] = 1663,
  [1664] = 1616,
  [1665] = 1665,
  [1666] = 1615,
  [1667] = 1667,
  [1668] = 1656,
  [1669] = 1669,
  [1670] = 1614,
  [1671] = 1671,
  [1672] = 1672,
  [1673] = 1673,
  [1674] = 1674,
  [1675] = 1638,
  [1676] = 1632,
  [1677] = 1677,
  [1678] = 1631,
  [1679] = 1679,
  [1680] = 1661,
  [1681] = 1615,
  [1682] = 1673,
  [1683] = 1683,
  [1684] = 1663,
  [1685] = 1657,
  [1686] = 1679,
  [1687] = 1677,
  [1688] = 1669,
  [1689] = 1683,
  [1690] = 1690,
  [1691] = 1628,
  [1692] = 1622,
  [1693] = 1667,
  [1694] = 1633,
  [1695] = 1695,
  [1696] = 1673,
  [1697] = 1618,
  [1698] = 1672,
  [1699] = 1658,
  [1700] = 1628,
  [1701] = 1617,
  [1702] = 1674,
  [1703] = 1621,
  [1704] = 1671,
  [1705] = 1705,
  [1706] = 1706,
  [1707] = 1707,
  [1708] = 1707,
  [1709] = 1139,
  [1710] = 1710,
  [1711] = 1711,
  [1712] = 1712,
  [1713] = 1713,
  [1714] = 1714,
  [1715] = 1711,
  [1716] = 1716,
  [1717] = 1707,
  [1718] = 1718,
  [1719] = 1719,
  [1720] = 1720,
  [1721] = 1721,
  [1722] = 1722,
  [1723] = 1723,
  [1724] = 1724,
  [1725] = 1135,
  [1726] = 1134,
  [1727] = 1727,
  [1728] = 1710,
  [1729] = 1729,
  [1730] = 1710,
  [1731] = 1723,
  [1732] = 1732,
  [1733] = 1733,
  [1734] = 1734,
  [1735] = 1735,
  [1736] = 1736,
  [1737] = 1737,
  [1738] = 1716,
  [1739] = 1739,
  [1740] = 1740,
  [1741] = 1150,
  [1742] = 1141,
  [1743] = 1718,
  [1744] = 1740,
  [1745] = 1745,
  [1746] = 1735,
  [1747] = 1747,
  [1748] = 1748,
  [1749] = 1749,
  [1750] = 1720,
  [1751] = 1736,
  [1752] = 1752,
  [1753] = 1151,
  [1754] = 1716,
  [1755] = 1755,
  [1756] = 1756,
  [1757] = 1757,
  [1758] = 1758,
  [1759] = 1759,
  [1760] = 1760,
  [1761] = 1761,
  [1762] = 1762,
  [1763] = 1763,
  [1764] = 1764,
  [1765] = 1765,
  [1766] = 1766,
  [1767] = 1767,
  [1768] = 1764,
  [1769] = 1769,
  [1770] = 1770,
  [1771] = 1771,
  [1772] = 1772,
  [1773] = 1773,
  [1774] = 1774,
  [1775] = 1775,
  [1776] = 1761,
  [1777] = 1777,
  [1778] = 1765,
  [1779] = 1769,
  [1780] = 1780,
  [1781] = 1781,
  [1782] = 1782,
  [1783] = 1783,
  [1784] = 1784,
  [1785] = 1785,
  [1786] = 1786,
  [1787] = 1781,
  [1788] = 1788,
  [1789] = 1789,
  [1790] = 1790,
  [1791] = 1791,
  [1792] = 1792,
  [1793] = 1774,
  [1794] = 1794,
  [1795] = 1786,
  [1796] = 1780,
  [1797] = 1771,
  [1798] = 1798,
  [1799] = 1785,
  [1800] = 1800,
  [1801] = 1762,
  [1802] = 1767,
  [1803] = 1760,
  [1804] = 1804,
  [1805] = 1774,
  [1806] = 1806,
  [1807] = 1807,
  [1808] = 1769,
  [1809] = 1809,
  [1810] = 1777,
  [1811] = 1811,
  [1812] = 1812,
  [1813] = 1813,
  [1814] = 1814,
  [1815] = 1807,
  [1816] = 1812,
  [1817] = 1782,
  [1818] = 1792,
  [1819] = 1809,
  [1820] = 1820,
  [1821] = 1821,
  [1822] = 1758,
  [1823] = 1823,
  [1824] = 1824,
  [1825] = 1825,
  [1826] = 1780,
  [1827] = 1827,
  [1828] = 1783,
  [1829] = 1762,
  [1830] = 1814,
  [1831] = 1827,
  [1832] = 1832,
  [1833] = 1833,
  [1834] = 1760,
  [1835] = 1775,
  [1836] = 1836,
  [1837] = 1837,
  [1838] = 1838,
  [1839] = 1839,
  [1840] = 1811,
  [1841] = 1772,
  [1842] = 1842,
  [1843] = 1843,
  [1844] = 1800,
  [1845] = 1760,
  [1846] = 1846,
  [1847] = 1847,
  [1848] = 1769,
  [1849] = 1774,
  [1850] = 1813,
  [1851] = 1851,
  [1852] = 1833,
  [1853] = 1762,
  [1854] = 1854,
  [1855] = 1855,
  [1856] = 1856,
  [1857] = 1857,
  [1858] = 1858,
  [1859] = 1859,
  [1860] = 1860,
  [1861] = 1861,
  [1862] = 1862,
  [1863] = 1863,
  [1864] = 1864,
  [1865] = 1865,
  [1866] = 1866,
  [1867] = 1867,
  [1868] = 1868,
  [1869] = 1869,
  [1870] = 1870,
  [1871] = 574,
  [1872] = 1872,
  [1873] = 1873,
  [1874] = 1874,
  [1875] = 1875,
  [1876] = 1876,
  [1877] = 1877,
  [1878] = 1878,
  [1879] = 1879,
  [1880] = 1880,
  [1881] = 1881,
  [1882] = 1882,
  [1883] = 1883,
  [1884] = 1884,
  [1885] = 1863,
  [1886] = 1886,
  [1887] = 1887,
  [1888] = 1862,
  [1889] = 1889,
  [1890] = 1890,
  [1891] = 1891,
  [1892] = 1892,
  [1893] = 1893,
  [1894] = 1862,
  [1895] = 1895,
  [1896] = 1863,
  [1897] = 1897,
  [1898] = 1898,
  [1899] = 571,
  [1900] = 1900,
  [1901] = 1901,
  [1902] = 1902,
  [1903] = 1903,
  [1904] = 1904,
  [1905] = 1905,
  [1906] = 1863,
  [1907] = 1907,
  [1908] = 1908,
  [1909] = 1862,
  [1910] = 1910,
  [1911] = 1911,
  [1912] = 1912,
  [1913] = 1913,
  [1914] = 1914,
  [1915] = 1915,
  [1916] = 1916,
  [1917] = 1917,
  [1918] = 1918,
  [1919] = 1919,
  [1920] = 1920,
  [1921] = 1898,
  [1922] = 1922,
  [1923] = 1881,
  [1924] = 1861,
  [1925] = 1925,
  [1926] = 1926,
  [1927] = 1927,
  [1928] = 1928,
  [1929] = 1929,
  [1930] = 1930,
  [1931] = 1931,
  [1932] = 1932,
  [1933] = 1933,
  [1934] = 1934,
  [1935] = 1935,
  [1936] = 1936,
  [1937] = 1937,
  [1938] = 1865,
  [1939] = 1939,
  [1940] = 1940,
  [1941] = 1941,
  [1942] = 1942,
  [1943] = 1943,
  [1944] = 1858,
  [1945] = 1945,
  [1946] = 1946,
  [1947] = 1947,
  [1948] = 1948,
  [1949] = 1873,
  [1950] = 1875,
  [1951] = 1951,
  [1952] = 1876,
  [1953] = 1953,
  [1954] = 1954,
  [1955] = 1856,
  [1956] = 1877,
  [1957] = 1878,
  [1958] = 1958,
  [1959] = 1959,
  [1960] = 1960,
  [1961] = 1942,
  [1962] = 1962,
  [1963] = 1963,
  [1964] = 1964,
  [1965] = 1965,
  [1966] = 1966,
  [1967] = 1891,
  [1968] = 1968,
  [1969] = 1969,
  [1970] = 1970,
  [1971] = 1971,
  [1972] = 1972,
  [1973] = 1900,
  [1974] = 1974,
  [1975] = 1975,
  [1976] = 1976,
  [1977] = 1977,
  [1978] = 1907,
  [1979] = 1910,
  [1980] = 1980,
  [1981] = 1981,
  [1982] = 1982,
  [1983] = 1911,
  [1984] = 1912,
  [1985] = 1919,
  [1986] = 1986,
  [1987] = 1987,
  [1988] = 1965,
  [1989] = 1926,
  [1990] = 1927,
  [1991] = 1941,
  [1992] = 1992,
  [1993] = 1993,
  [1994] = 1994,
  [1995] = 1943,
  [1996] = 1996,
  [1997] = 1997,
  [1998] = 1998,
  [1999] = 1999,
  [2000] = 2000,
  [2001] = 1965,
  [2002] = 2002,
  [2003] = 1998,
  [2004] = 2004,
  [2005] = 1958,
  [2006] = 2006,
  [2007] = 2007,
  [2008] = 2008,
  [2009] = 2009,
  [2010] = 2010,
  [2011] = 1965,
  [2012] = 2012,
  [2013] = 2013,
  [2014] = 1863,
  [2015] = 1925,
  [2016] = 1862,
  [2017] = 1992,
  [2018] = 1862,
  [2019] = 1993,
  [2020] = 2020,
  [2021] = 2009,
  [2022] = 2022,
  [2023] = 2006,
  [2024] = 2024,
  [2025] = 2025,
  [2026] = 2007,
  [2027] = 1863,
  [2028] = 2012,
  [2029] = 1997,
  [2030] = 2022,
  [2031] = 2025,
  [2032] = 2032,
  [2033] = 2033,
  [2034] = 2034,
  [2035] = 2035,
  [2036] = 1987,
  [2037] = 2033,
  [2038] = 2038,
  [2039] = 2039,
  [2040] = 2040,
  [2041] = 2041,
  [2042] = 2042,
  [2043] = 2035,
  [2044] = 2044,
  [2045] = 2045,
  [2046] = 2046,
  [2047] = 2047,
  [2048] = 1975,
  [2049] = 2049,
  [2050] = 2050,
  [2051] = 2051,
  [2052] = 2052,
  [2053] = 1892,
  [2054] = 2054,
  [2055] = 2055,
  [2056] = 2056,
  [2057] = 2057,
  [2058] = 2047,
  [2059] = 2059,
  [2060] = 2060,
  [2061] = 2061,
  [2062] = 2062,
  [2063] = 2062,
  [2064] = 2064,
  [2065] = 2032,
  [2066] = 2066,
  [2067] = 1982,
  [2068] = 1869,
  [2069] = 2064,
  [2070] = 2066,
  [2071] = 2071,
  [2072] = 1998,
  [2073] = 1880,
  [2074] = 2074,
  [2075] = 2075,
  [2076] = 1857,
  [2077] = 2077,
  [2078] = 2078,
  [2079] = 2079,
  [2080] = 1901,
  [2081] = 2000,
  [2082] = 2075,
  [2083] = 1903,
  [2084] = 2077,
  [2085] = 2085,
  [2086] = 1948,
  [2087] = 1884,
  [2088] = 2088,
  [2089] = 2089,
  [2090] = 1963,
  [2091] = 2091,
  [2092] = 1969,
  [2093] = 2093,
  [2094] = 1855,
  [2095] = 2095,
  [2096] = 2096,
  [2097] = 2020,
  [2098] = 2098,
  [2099] = 2099,
  [2100] = 2034,
  [2101] = 2098,
  [2102] = 1994,
  [2103] = 2103,
  [2104] = 1999,
  [2105] = 2010,
  [2106] = 2008,
  [2107] = 2107,
  [2108] = 2108,
  [2109] = 2109,
  [2110] = 2110,
  [2111] = 2111,
  [2112] = 2112,
  [2113] = 2113,
  [2114] = 2114,
  [2115] = 2115,
  [2116] = 2116,
  [2117] = 2117,
  [2118] = 2107,
  [2119] = 2119,
  [2120] = 2120,
  [2121] = 2121,
  [2122] = 2122,
  [2123] = 2123,
  [2124] = 2115,
  [2125] = 2125,
  [2126] = 2112,
  [2127] = 2127,
  [2128] = 2128,
  [2129] = 2129,
  [2130] = 2116,
  [2131] = 2131,
  [2132] = 2128,
  [2133] = 2133,
  [2134] = 2134,
  [2135] = 2129,
  [2136] = 2134,
  [2137] = 2137,
  [2138] = 2138,
  [2139] = 2139,
  [2140] = 2140,
  [2141] = 2141,
  [2142] = 2142,
  [2143] = 2143,
  [2144] = 2120,
  [2145] = 2141,
  [2146] = 2146,
  [2147] = 2121,
  [2148] = 2143,
  [2149] = 2149,
  [2150] = 2150,
  [2151] = 2151,
  [2152] = 2152,
  [2153] = 2153,
  [2154] = 2154,
  [2155] = 2155,
  [2156] = 2156,
  [2157] = 2157,
  [2158] = 2158,
  [2159] = 2159,
  [2160] = 2160,
  [2161] = 2127,
  [2162] = 2162,
  [2163] = 2163,
  [2164] = 2164,
  [2165] = 2165,
  [2166] = 2166,
  [2167] = 2122,
  [2168] = 2168,
  [2169] = 2169,
  [2170] = 2170,
  [2171] = 2171,
  [2172] = 2119,
  [2173] = 2140,
  [2174] = 2146,
  [2175] = 2175,
  [2176] = 2176,
  [2177] = 2177,
  [2178] = 2178,
  [2179] = 2179,
  [2180] = 2117,
  [2181] = 2181,
  [2182] = 2182,
  [2183] = 2183,
  [2184] = 2184,
  [2185] = 2114,
  [2186] = 2186,
  [2187] = 2187,
  [2188] = 2188,
  [2189] = 2189,
  [2190] = 2190,
  [2191] = 2191,
  [2192] = 2192,
  [2193] = 2193,
  [2194] = 2194,
  [2195] = 2195,
  [2196] = 2113,
  [2197] = 2197,
  [2198] = 2187,
  [2199] = 2176,
  [2200] = 2170,
  [2201] = 2201,
  [2202] = 2175,
  [2203] = 2177,
  [2204] = 2179,
  [2205] = 2181,
  [2206] = 2182,
  [2207] = 2207,
  [2208] = 2208,
  [2209] = 2209,
  [2210] = 2184,
  [2211] = 2211,
  [2212] = 2111,
  [2213] = 2188,
  [2214] = 2214,
  [2215] = 2110,
  [2216] = 2157,
  [2217] = 2137,
  [2218] = 2218,
  [2219] = 2109,
  [2220] = 2108,
  [2221] = 2191,
  [2222] = 2192,
  [2223] = 2193,
  [2224] = 2142,
  [2225] = 2194,
  [2226] = 2226,
  [2227] = 2227,
  [2228] = 2168,
  [2229] = 2171,
  [2230] = 2230,
  [2231] = 2231,
  [2232] = 2232,
  [2233] = 2166,
  [2234] = 2153,
  [2235] = 2125,
  [2236] = 2165,
  [2237] = 2131,
  [2238] = 2214,
  [2239] = 2171,
  [2240] = 2183,
  [2241] = 2159,
  [2242] = 2163,
  [2243] = 2190,
  [2244] = 2244,
  [2245] = 2245,
  [2246] = 2197,
  [2247] = 2247,
  [2248] = 2248,
  [2249] = 2211,
  [2250] = 2186,
  [2251] = 2218,
  [2252] = 1595,
  [2253] = 2253,
  [2254] = 2123,
  [2255] = 2255,
  [2256] = 2256,
  [2257] = 2227,
  [2258] = 2245,
  [2259] = 2253,
  [2260] = 2260,
  [2261] = 2244,
  [2262] = 2262,
  [2263] = 2263,
  [2264] = 2264,
  [2265] = 2265,
  [2266] = 2160,
  [2267] = 2256,
  [2268] = 2162,
  [2269] = 2218,
  [2270] = 2270,
  [2271] = 2271,
  [2272] = 2211,
  [2273] = 2133,
  [2274] = 2274,
  [2275] = 2183,
  [2276] = 2226,
  [2277] = 2197,
  [2278] = 2137,
  [2279] = 2230,
  [2280] = 2231,
  [2281] = 2281,
  [2282] = 865,
  [2283] = 2232,
  [2284] = 2284,
  [2285] = 2133,
  [2286] = 2286,
  [2287] = 2287,
  [2288] = 2288,
  [2289] = 941,
  [2290] = 2290,
  [2291] = 946,
  [2292] = 2159,
  [2293] = 2149,
  [2294] = 2158,
  [2295] = 2295,
  [2296] = 2296,
  [2297] = 2297,
  [2298] = 2150,
  [2299] = 2256,
  [2300] = 2134,
  [2301] = 2301,
  [2302] = 2178,
  [2303] = 2271,
  [2304] = 2304,
  [2305] = 1005,
  [2306] = 2255,
  [2307] = 2189,
  [2308] = 2151,
  [2309] = 2154,
  [2310] = 2156,
  [2311] = 2169,
  [2312] = 2301,
  [2313] = 585,
  [2314] = 2314,
  [2315] = 2315,
  [2316] = 2316,
  [2317] = 2317,
  [2318] = 2318,
  [2319] = 2319,
  [2320] = 2320,
  [2321] = 2321,
  [2322] = 2322,
  [2323] = 2323,
  [2324] = 2324,
  [2325] = 2325,
  [2326] = 2326,
  [2327] = 2327,
  [2328] = 2328,
  [2329] = 2324,
  [2330] = 2330,
  [2331] = 2331,
  [2332] = 2332,
  [2333] = 2333,
  [2334] = 2317,
  [2335] = 2335,
  [2336] = 2319,
  [2337] = 2320,
  [2338] = 2338,
  [2339] = 2322,
  [2340] = 2340,
  [2341] = 2341,
  [2342] = 2322,
  [2343] = 2343,
  [2344] = 2320,
  [2345] = 2345,
  [2346] = 2324,
  [2347] = 2319,
  [2348] = 2317,
  [2349] = 2317,
  [2350] = 2319,
  [2351] = 2320,
  [2352] = 2352,
  [2353] = 2324,
  [2354] = 2354,
  [2355] = 2355,
  [2356] = 2356,
  [2357] = 2357,
  [2358] = 2358,
  [2359] = 2359,
  [2360] = 2360,
  [2361] = 2361,
  [2362] = 2362,
  [2363] = 2322,
  [2364] = 2364,
  [2365] = 2365,
  [2366] = 2366,
  [2367] = 2367,
  [2368] = 2368,
  [2369] = 2369,
  [2370] = 2370,
  [2371] = 2371,
  [2372] = 2372,
  [2373] = 2373,
  [2374] = 2374,
  [2375] = 2375,
  [2376] = 2376,
  [2377] = 2377,
  [2378] = 2378,
  [2379] = 2379,
  [2380] = 2380,
  [2381] = 2381,
  [2382] = 2382,
  [2383] = 2383,
  [2384] = 2384,
  [2385] = 2338,
  [2386] = 2386,
  [2387] = 2387,
  [2388] = 2388,
  [2389] = 2389,
  [2390] = 2390,
  [2391] = 2391,
  [2392] = 2392,
  [2393] = 2393,
  [2394] = 2394,
  [2395] = 2395,
  [2396] = 2396,
  [2397] = 2397,
  [2398] = 2398,
  [2399] = 2399,
  [2400] = 2400,
  [2401] = 2332,
  [2402] = 2402,
  [2403] = 2403,
  [2404] = 2404,
  [2405] = 2405,
  [2406] = 2406,
  [2407] = 2407,
  [2408] = 2408,
  [2409] = 2405,
  [2410] = 2410,
  [2411] = 2404,
  [2412] = 2412,
  [2413] = 2394,
  [2414] = 2320,
  [2415] = 2415,
  [2416] = 2319,
  [2417] = 2417,
  [2418] = 2418,
  [2419] = 2317,
  [2420] = 2420,
  [2421] = 2421,
  [2422] = 2422,
  [2423] = 2423,
  [2424] = 2331,
  [2425] = 2328,
  [2426] = 2426,
  [2427] = 2333,
  [2428] = 2399,
  [2429] = 2314,
  [2430] = 2430,
  [2431] = 2431,
  [2432] = 2432,
  [2433] = 2433,
  [2434] = 2434,
  [2435] = 2435,
  [2436] = 2436,
  [2437] = 2422,
  [2438] = 2438,
  [2439] = 2398,
  [2440] = 2440,
  [2441] = 2397,
  [2442] = 2442,
  [2443] = 2396,
  [2444] = 2433,
  [2445] = 2445,
  [2446] = 2446,
  [2447] = 2447,
  [2448] = 2448,
  [2449] = 2377,
  [2450] = 2412,
  [2451] = 2403,
  [2452] = 2395,
  [2453] = 2453,
  [2454] = 2338,
  [2455] = 2383,
  [2456] = 2456,
  [2457] = 2457,
  [2458] = 2458,
  [2459] = 564,
  [2460] = 2460,
  [2461] = 2442,
  [2462] = 2440,
  [2463] = 2393,
  [2464] = 2464,
  [2465] = 2465,
  [2466] = 2343,
  [2467] = 2316,
  [2468] = 2468,
  [2469] = 2469,
  [2470] = 2321,
  [2471] = 2471,
  [2472] = 2386,
  [2473] = 2323,
  [2474] = 2474,
  [2475] = 2330,
  [2476] = 2476,
  [2477] = 2322,
  [2478] = 2352,
  [2479] = 2318,
  [2480] = 2391,
  [2481] = 2481,
  [2482] = 2333,
  [2483] = 2483,
  [2484] = 2320,
  [2485] = 2415,
  [2486] = 2319,
  [2487] = 2379,
  [2488] = 2488,
  [2489] = 2378,
  [2490] = 2317,
  [2491] = 2378,
  [2492] = 2379,
  [2493] = 2423,
  [2494] = 2418,
  [2495] = 2434,
  [2496] = 2324,
  [2497] = 2497,
  [2498] = 2435,
  [2499] = 2499,
  [2500] = 2500,
  [2501] = 2501,
  [2502] = 2445,
  [2503] = 2370,
  [2504] = 2378,
  [2505] = 2379,
  [2506] = 2458,
  [2507] = 2417,
  [2508] = 2378,
  [2509] = 2379,
  [2510] = 2468,
  [2511] = 2378,
  [2512] = 2379,
  [2513] = 2471,
  [2514] = 568,
  [2515] = 2352,
  [2516] = 2408,
  [2517] = 2517,
  [2518] = 2501,
  [2519] = 2497,
  [2520] = 2324,
  [2521] = 2517,
  [2522] = 2522,
  [2523] = 2407,
  [2524] = 2464,
  [2525] = 2453,
  [2526] = 2499,
  [2527] = 2369,
  [2528] = 2528,
  [2529] = 2529,
  [2530] = 2367,
  [2531] = 2531,
  [2532] = 2388,
  [2533] = 2531,
  [2534] = 2534,
  [2535] = 2362,
  [2536] = 2341,
  [2537] = 2332,
  [2538] = 2392,
  [2539] = 2400,
  [2540] = 2361,
  [2541] = 2360,
  [2542] = 2389,
  [2543] = 2456,
  [2544] = 2544,
  [2545] = 2447,
  [2546] = 2517,
  [2547] = 2501,
  [2548] = 2548,
  [2549] = 2384,
  [2550] = 2359,
  [2551] = 2382,
  [2552] = 2528,
  [2553] = 2553,
  [2554] = 2465,
  [2555] = 2380,
  [2556] = 2325,
  [2557] = 2438,
  [2558] = 2436,
  [2559] = 2390,
  [2560] = 2560,
  [2561] = 2373,
  [2562] = 2446,
  [2563] = 2354,
  [2564] = 2534,
};

static inline bool sym_identifier_character_set_1(int32_t c) {
  return (c < 43514
    ? (c < 4193
      ? (c < 2707
        ? (c < 1994
          ? (c < 931
            ? (c < 748
              ? (c < 192
                ? (c < 170
                  ? (c < 'a'
                    ? (c >= 'A' && c <= '_')
                    : c <= 'z')
                  : (c <= 170 || (c < 186
                    ? c == 181
                    : c <= 186)))
                : (c <= 214 || (c < 710
                  ? (c < 248
                    ? (c >= 216 && c <= 246)
                    : c <= 705)
                  : (c <= 721 || (c >= 736 && c <= 740)))))
              : (c <= 748 || (c < 895
                ? (c < 886
                  ? (c < 880
                    ? c == 750
                    : c <= 884)
                  : (c <= 887 || (c >= 891 && c <= 893)))
                : (c <= 895 || (c < 908
                  ? (c < 904
                    ? c == 902
                    : c <= 906)
                  : (c <= 908 || (c >= 910 && c <= 929)))))))
            : (c <= 1013 || (c < 1649
              ? (c < 1376
                ? (c < 1329
                  ? (c < 1162
                    ? (c >= 1015 && c <= 1153)
                    : c <= 1327)
                  : (c <= 1366 || c == 1369))
                : (c <= 1416 || (c < 1568
                  ? (c < 1519
                    ? (c >= 1488 && c <= 1514)
                    : c <= 1522)
                  : (c <= 1610 || (c >= 1646 && c <= 1647)))))
              : (c <= 1747 || (c < 1791
                ? (c < 1774
                  ? (c < 1765
                    ? c == 1749
                    : c <= 1766)
                  : (c <= 1775 || (c >= 1786 && c <= 1788)))
                : (c <= 1791 || (c < 1869
                  ? (c < 1810
                    ? c == 1808
                    : c <= 1839)
                  : (c <= 1957 || c == 1969))))))))
          : (c <= 2026 || (c < 2482
            ? (c < 2208
              ? (c < 2088
                ? (c < 2048
                  ? (c < 2042
                    ? (c >= 2036 && c <= 2037)
                    : c <= 2042)
                  : (c <= 2069 || (c < 2084
                    ? c == 2074
                    : c <= 2084)))
                : (c <= 2088 || (c < 2160
                  ? (c < 2144
                    ? (c >= 2112 && c <= 2136)
                    : c <= 2154)
                  : (c <= 2183 || (c >= 2185 && c <= 2190)))))
              : (c <= 2249 || (c < 2417
                ? (c < 2384
                  ? (c < 2365
                    ? (c >= 2308 && c <= 2361)
                    : c <= 2365)
                  : (c <= 2384 || (c >= 2392 && c <= 2401)))
                : (c <= 2432 || (c < 2451
                  ? (c < 2447
                    ? (c >= 2437 && c <= 2444)
                    : c <= 2448)
                  : (c <= 2472 || (c >= 2474 && c <= 2480)))))))
            : (c <= 2482 || (c < 2579
              ? (c < 2527
                ? (c < 2510
                  ? (c < 2493
                    ? (c >= 2486 && c <= 2489)
                    : c <= 2493)
                  : (c <= 2510 || (c >= 2524 && c <= 2525)))
                : (c <= 2529 || (c < 2565
                  ? (c < 2556
                    ? (c >= 2544 && c <= 2545)
                    : c <= 2556)
                  : (c <= 2570 || (c >= 2575 && c <= 2576)))))
              : (c <= 2600 || (c < 2649
                ? (c < 2613
                  ? (c < 2610
                    ? (c >= 2602 && c <= 2608)
                    : c <= 2611)
                  : (c <= 2614 || (c >= 2616 && c <= 2617)))
                : (c <= 2652 || (c < 2693
                  ? (c < 2674
                    ? c == 2654
                    : c <= 2676)
                  : (c <= 2701 || (c >= 2703 && c <= 2705)))))))))))
        : (c <= 2728 || (c < 3242
          ? (c < 2962
            ? (c < 2858
              ? (c < 2784
                ? (c < 2741
                  ? (c < 2738
                    ? (c >= 2730 && c <= 2736)
                    : c <= 2739)
                  : (c <= 2745 || (c < 2768
                    ? c == 2749
                    : c <= 2768)))
                : (c <= 2785 || (c < 2831
                  ? (c < 2821
                    ? c == 2809
                    : c <= 2828)
                  : (c <= 2832 || (c >= 2835 && c <= 2856)))))
              : (c <= 2864 || (c < 2911
                ? (c < 2877
                  ? (c < 2869
                    ? (c >= 2866 && c <= 2867)
                    : c <= 2873)
                  : (c <= 2877 || (c >= 2908 && c <= 2909)))
                : (c <= 2913 || (c < 2949
                  ? (c < 2947
                    ? c == 2929
                    : c <= 2947)
                  : (c <= 2954 || (c >= 2958 && c <= 2960)))))))
            : (c <= 2965 || (c < 3090
              ? (c < 2984
                ? (c < 2974
                  ? (c < 2972
                    ? (c >= 2969 && c <= 2970)
                    : c <= 2972)
                  : (c <= 2975 || (c >= 2979 && c <= 2980)))
                : (c <= 2986 || (c < 3077
                  ? (c < 3024
                    ? (c >= 2990 && c <= 3001)
                    : c <= 3024)
                  : (c <= 3084 || (c >= 3086 && c <= 3088)))))
              : (c <= 3112 || (c < 3168
                ? (c < 3160
                  ? (c < 3133
                    ? (c >= 3114 && c <= 3129)
                    : c <= 3133)
                  : (c <= 3162 || c == 3165))
                : (c <= 3169 || (c < 3214
                  ? (c < 3205
                    ? c == 3200
                    : c <= 3212)
                  : (c <= 3216 || (c >= 3218 && c <= 3240)))))))))
          : (c <= 3251 || (c < 3648
            ? (c < 3412
              ? (c < 3332
                ? (c < 3293
                  ? (c < 3261
                    ? (c >= 3253 && c <= 3257)
                    : c <= 3261)
                  : (c <= 3294 || (c < 3313
                    ? (c >= 3296 && c <= 3297)
                    : c <= 3314)))
                : (c <= 3340 || (c < 3389
                  ? (c < 3346
                    ? (c >= 3342 && c <= 3344)
                    : c <= 3386)
                  : (c <= 3389 || c == 3406))))
              : (c <= 3414 || (c < 3507
                ? (c < 3461
                  ? (c < 3450
                    ? (c >= 3423 && c <= 3425)
                    : c <= 3455)
                  : (c <= 3478 || (c >= 3482 && c <= 3505)))
                : (c <= 3515 || (c < 3585
                  ? (c < 3520
                    ? c == 3517
                    : c <= 3526)
                  : (c <= 3632 || c == 3634))))))
            : (c <= 3654 || (c < 3782
              ? (c < 3749
                ? (c < 3718
                  ? (c < 3716
                    ? (c >= 3713 && c <= 3714)
                    : c <= 3716)
                  : (c <= 3722 || (c >= 3724 && c <= 3747)))
                : (c <= 3749 || (c < 3773
                  ? (c < 3762
                    ? (c >= 3751 && c <= 3760)
                    : c <= 3762)
                  : (c <= 3773 || (c >= 3776 && c <= 3780)))))
              : (c <= 3782 || (c < 3976
                ? (c < 3904
                  ? (c < 3840
                    ? (c >= 3804 && c <= 3807)
                    : c <= 3840)
                  : (c <= 3911 || (c >= 3913 && c <= 3948)))
                : (c <= 3980 || (c < 4176
                  ? (c < 4159
                    ? (c >= 4096 && c <= 4138)
                    : c <= 4159)
                  : (c <= 4181 || (c >= 4186 && c <= 4189)))))))))))))
      : (c <= 4193 || (c < 8134
        ? (c < 6176
          ? (c < 4808
            ? (c < 4688
              ? (c < 4295
                ? (c < 4213
                  ? (c < 4206
                    ? (c >= 4197 && c <= 4198)
                    : c <= 4208)
                  : (c <= 4225 || (c < 4256
                    ? c == 4238
                    : c <= 4293)))
                : (c <= 4295 || (c < 4348
                  ? (c < 4304
                    ? c == 4301
                    : c <= 4346)
                  : (c <= 4680 || (c >= 4682 && c <= 4685)))))
              : (c <= 4694 || (c < 4752
                ? (c < 4704
                  ? (c < 4698
                    ? c == 4696
                    : c <= 4701)
                  : (c <= 4744 || (c >= 4746 && c <= 4749)))
                : (c <= 4784 || (c < 4800
                  ? (c < 4792
                    ? (c >= 4786 && c <= 4789)
                    : c <= 4798)
                  : (c <= 4800 || (c >= 4802 && c <= 4805)))))))
            : (c <= 4822 || (c < 5792
              ? (c < 5024
                ? (c < 4888
                  ? (c < 4882
                    ? (c >= 4824 && c <= 4880)
                    : c <= 4885)
                  : (c <= 4954 || (c >= 4992 && c <= 5007)))
                : (c <= 5109 || (c < 5743
                  ? (c < 5121
                    ? (c >= 5112 && c <= 5117)
                    : c <= 5740)
                  : (c <= 5759 || (c >= 5761 && c <= 5786)))))
              : (c <= 5866 || (c < 5984
                ? (c < 5919
                  ? (c < 5888
                    ? (c >= 5870 && c <= 5880)
                    : c <= 5905)
                  : (c <= 5937 || (c >= 5952 && c <= 5969)))
                : (c <= 5996 || (c < 6103
                  ? (c < 6016
                    ? (c >= 5998 && c <= 6000)
                    : c <= 6067)
                  : (c <= 6103 || c == 6108))))))))
          : (c <= 6264 || (c < 7312
            ? (c < 6823
              ? (c < 6512
                ? (c < 6320
                  ? (c < 6314
                    ? (c >= 6272 && c <= 6312)
                    : c <= 6314)
                  : (c <= 6389 || (c < 6480
                    ? (c >= 6400 && c <= 6430)
                    : c <= 6509)))
                : (c <= 6516 || (c < 6656
                  ? (c < 6576
                    ? (c >= 6528 && c <= 6571)
                    : c <= 6601)
                  : (c <= 6678 || (c >= 6688 && c <= 6740)))))
              : (c <= 6823 || (c < 7098
                ? (c < 7043
                  ? (c < 6981
                    ? (c >= 6917 && c <= 6963)
                    : c <= 6988)
                  : (c <= 7072 || (c >= 7086 && c <= 7087)))
                : (c <= 7141 || (c < 7258
                  ? (c < 7245
                    ? (c >= 7168 && c <= 7203)
                    : c <= 7247)
                  : (c <= 7293 || (c >= 7296 && c <= 7304)))))))
            : (c <= 7354 || (c < 8008
              ? (c < 7418
                ? (c < 7406
                  ? (c < 7401
                    ? (c >= 7357 && c <= 7359)
                    : c <= 7404)
                  : (c <= 7411 || (c >= 7413 && c <= 7414)))
                : (c <= 7418 || (c < 7960
                  ? (c < 7680
                    ? (c >= 7424 && c <= 7615)
                    : c <= 7957)
                  : (c <= 7965 || (c >= 7968 && c <= 8005)))))
              : (c <= 8013 || (c < 8031
                ? (c < 8027
                  ? (c < 8025
                    ? (c >= 8016 && c <= 8023)
                    : c <= 8025)
                  : (c <= 8027 || c == 8029))
                : (c <= 8061 || (c < 8126
                  ? (c < 8118
                    ? (c >= 8064 && c <= 8116)
                    : c <= 8124)
                  : (c <= 8126 || (c >= 8130 && c <= 8132)))))))))))
        : (c <= 8140 || (c < 12337
          ? (c < 8544
            ? (c < 8458
              ? (c < 8305
                ? (c < 8160
                  ? (c < 8150
                    ? (c >= 8144 && c <= 8147)
                    : c <= 8155)
                  : (c <= 8172 || (c < 8182
                    ? (c >= 8178 && c <= 8180)
                    : c <= 8188)))
                : (c <= 8305 || (c < 8450
                  ? (c < 8336
                    ? c == 8319
                    : c <= 8348)
                  : (c <= 8450 || c == 8455))))
              : (c <= 8467 || (c < 8488
                ? (c < 8484
                  ? (c < 8472
                    ? c == 8469
                    : c <= 8477)
                  : (c <= 8484 || c == 8486))
                : (c <= 8488 || (c < 8517
                  ? (c < 8508
                    ? (c >= 8490 && c <= 8505)
                    : c <= 8511)
                  : (c <= 8521 || c == 8526))))))
            : (c <= 8584 || (c < 11680
              ? (c < 11559
                ? (c < 11506
                  ? (c < 11499
                    ? (c >= 11264 && c <= 11492)
                    : c <= 11502)
                  : (c <= 11507 || (c >= 11520 && c <= 11557)))
                : (c <= 11559 || (c < 11631
                  ? (c < 11568
                    ? c == 11565
                    : c <= 11623)
                  : (c <= 11631 || (c >= 11648 && c <= 11670)))))
              : (c <= 11686 || (c < 11720
                ? (c < 11704
                  ? (c < 11696
                    ? (c >= 11688 && c <= 11694)
                    : c <= 11702)
                  : (c <= 11710 || (c >= 11712 && c <= 11718)))
                : (c <= 11726 || (c < 12293
                  ? (c < 11736
                    ? (c >= 11728 && c <= 11734)
                    : c <= 11742)
                  : (c <= 12295 || (c >= 12321 && c <= 12329)))))))))
          : (c <= 12341 || (c < 42891
            ? (c < 19968
              ? (c < 12549
                ? (c < 12445
                  ? (c < 12353
                    ? (c >= 12344 && c <= 12348)
                    : c <= 12438)
                  : (c <= 12447 || (c < 12540
                    ? (c >= 12449 && c <= 12538)
                    : c <= 12543)))
                : (c <= 12591 || (c < 12784
                  ? (c < 12704
                    ? (c >= 12593 && c <= 12686)
                    : c <= 12735)
                  : (c <= 12799 || (c >= 13312 && c <= 19903)))))
              : (c <= 42124 || (c < 42560
                ? (c < 42512
                  ? (c < 42240
                    ? (c >= 42192 && c <= 42237)
                    : c <= 42508)
                  : (c <= 42527 || (c >= 42538 && c <= 42539)))
                : (c <= 42606 || (c < 42775
                  ? (c < 42656
                    ? (c >= 42623 && c <= 42653)
                    : c <= 42735)
                  : (c <= 42783 || (c >= 42786 && c <= 42888)))))))
            : (c <= 42954 || (c < 43250
              ? (c < 43011
                ? (c < 42965
                  ? (c < 42963
                    ? (c >= 42960 && c <= 42961)
                    : c <= 42963)
                  : (c <= 42969 || (c >= 42994 && c <= 43009)))
                : (c <= 43013 || (c < 43072
                  ? (c < 43020
                    ? (c >= 43015 && c <= 43018)
                    : c <= 43042)
                  : (c <= 43123 || (c >= 43138 && c <= 43187)))))
              : (c <= 43255 || (c < 43360
                ? (c < 43274
                  ? (c < 43261
                    ? c == 43259
                    : c <= 43262)
                  : (c <= 43301 || (c >= 43312 && c <= 43334)))
                : (c <= 43388 || (c < 43488
                  ? (c < 43471
                    ? (c >= 43396 && c <= 43442)
                    : c <= 43471)
                  : (c <= 43492 || (c >= 43494 && c <= 43503)))))))))))))))
    : (c <= 43518 || (c < 70727
      ? (c < 66956
        ? (c < 64914
          ? (c < 43868
            ? (c < 43714
              ? (c < 43646
                ? (c < 43588
                  ? (c < 43584
                    ? (c >= 43520 && c <= 43560)
                    : c <= 43586)
                  : (c <= 43595 || (c < 43642
                    ? (c >= 43616 && c <= 43638)
                    : c <= 43642)))
                : (c <= 43695 || (c < 43705
                  ? (c < 43701
                    ? c == 43697
                    : c <= 43702)
                  : (c <= 43709 || c == 43712))))
              : (c <= 43714 || (c < 43785
                ? (c < 43762
                  ? (c < 43744
                    ? (c >= 43739 && c <= 43741)
                    : c <= 43754)
                  : (c <= 43764 || (c >= 43777 && c <= 43782)))
                : (c <= 43790 || (c < 43816
                  ? (c < 43808
                    ? (c >= 43793 && c <= 43798)
                    : c <= 43814)
                  : (c <= 43822 || (c >= 43824 && c <= 43866)))))))
            : (c <= 43881 || (c < 64287
              ? (c < 63744
                ? (c < 55216
                  ? (c < 44032
                    ? (c >= 43888 && c <= 44002)
                    : c <= 55203)
                  : (c <= 55238 || (c >= 55243 && c <= 55291)))
                : (c <= 64109 || (c < 64275
                  ? (c < 64256
                    ? (c >= 64112 && c <= 64217)
                    : c <= 64262)
                  : (c <= 64279 || c == 64285))))
              : (c <= 64296 || (c < 64323
                ? (c < 64318
                  ? (c < 64312
                    ? (c >= 64298 && c <= 64310)
                    : c <= 64316)
                  : (c <= 64318 || (c >= 64320 && c <= 64321)))
                : (c <= 64324 || (c < 64612
                  ? (c < 64467
                    ? (c >= 64326 && c <= 64433)
                    : c <= 64605)
                  : (c <= 64829 || (c >= 64848 && c <= 64911)))))))))
          : (c <= 64967 || (c < 65599
            ? (c < 65382
              ? (c < 65147
                ? (c < 65139
                  ? (c < 65137
                    ? (c >= 65008 && c <= 65017)
                    : c <= 65137)
                  : (c <= 65139 || (c < 65145
                    ? c == 65143
                    : c <= 65145)))
                : (c <= 65147 || (c < 65313
                  ? (c < 65151
                    ? c == 65149
                    : c <= 65276)
                  : (c <= 65338 || (c >= 65345 && c <= 65370)))))
              : (c <= 65437 || (c < 65498
                ? (c < 65482
                  ? (c < 65474
                    ? (c >= 65440 && c <= 65470)
                    : c <= 65479)
                  : (c <= 65487 || (c >= 65490 && c <= 65495)))
                : (c <= 65500 || (c < 65576
                  ? (c < 65549
                    ? (c >= 65536 && c <= 65547)
                    : c <= 65574)
                  : (c <= 65594 || (c >= 65596 && c <= 65597)))))))
            : (c <= 65613 || (c < 66464
              ? (c < 66208
                ? (c < 65856
                  ? (c < 65664
                    ? (c >= 65616 && c <= 65629)
                    : c <= 65786)
                  : (c <= 65908 || (c >= 66176 && c <= 66204)))
                : (c <= 66256 || (c < 66384
                  ? (c < 66349
                    ? (c >= 66304 && c <= 66335)
                    : c <= 66378)
                  : (c <= 66421 || (c >= 66432 && c <= 66461)))))
              : (c <= 66499 || (c < 66776
                ? (c < 66560
                  ? (c < 66513
                    ? (c >= 66504 && c <= 66511)
                    : c <= 66517)
                  : (c <= 66717 || (c >= 66736 && c <= 66771)))
                : (c <= 66811 || (c < 66928
                  ? (c < 66864
                    ? (c >= 66816 && c <= 66855)
                    : c <= 66915)
                  : (c <= 66938 || (c >= 66940 && c <= 66954)))))))))))
        : (c <= 66962 || (c < 68864
          ? (c < 67828
            ? (c < 67506
              ? (c < 67072
                ? (c < 66979
                  ? (c < 66967
                    ? (c >= 66964 && c <= 66965)
                    : c <= 66977)
                  : (c <= 66993 || (c < 67003
                    ? (c >= 66995 && c <= 67001)
                    : c <= 67004)))
                : (c <= 67382 || (c < 67456
                  ? (c < 67424
                    ? (c >= 67392 && c <= 67413)
                    : c <= 67431)
                  : (c <= 67461 || (c >= 67463 && c <= 67504)))))
              : (c <= 67514 || (c < 67644
                ? (c < 67594
                  ? (c < 67592
                    ? (c >= 67584 && c <= 67589)
                    : c <= 67592)
                  : (c <= 67637 || (c >= 67639 && c <= 67640)))
                : (c <= 67644 || (c < 67712
                  ? (c < 67680
                    ? (c >= 67647 && c <= 67669)
                    : c <= 67702)
                  : (c <= 67742 || (c >= 67808 && c <= 67826)))))))
            : (c <= 67829 || (c < 68224
              ? (c < 68096
                ? (c < 67968
                  ? (c < 67872
                    ? (c >= 67840 && c <= 67861)
                    : c <= 67897)
                  : (c <= 68023 || (c >= 68030 && c <= 68031)))
                : (c <= 68096 || (c < 68121
                  ? (c < 68117
                    ? (c >= 68112 && c <= 68115)
                    : c <= 68119)
                  : (c <= 68149 || (c >= 68192 && c <= 68220)))))
              : (c <= 68252 || (c < 68448
                ? (c < 68352
                  ? (c < 68297
                    ? (c >= 68288 && c <= 68295)
                    : c <= 68324)
                  : (c <= 68405 || (c >= 68416 && c <= 68437)))
                : (c <= 68466 || (c < 68736
                  ? (c < 68608
                    ? (c >= 68480 && c <= 68497)
                    : c <= 68680)
                  : (c <= 68786 || (c >= 68800 && c <= 68850)))))))))
          : (c <= 68899 || (c < 70106
            ? (c < 69749
              ? (c < 69488
                ? (c < 69376
                  ? (c < 69296
                    ? (c >= 69248 && c <= 69289)
                    : c <= 69297)
                  : (c <= 69404 || (c < 69424
                    ? c == 69415
                    : c <= 69445)))
                : (c <= 69505 || (c < 69635
                  ? (c < 69600
                    ? (c >= 69552 && c <= 69572)
                    : c <= 69622)
                  : (c <= 69687 || (c >= 69745 && c <= 69746)))))
              : (c <= 69749 || (c < 69959
                ? (c < 69891
                  ? (c < 69840
                    ? (c >= 69763 && c <= 69807)
                    : c <= 69864)
                  : (c <= 69926 || c == 69956))
                : (c <= 69959 || (c < 70019
                  ? (c < 70006
                    ? (c >= 69968 && c <= 70002)
                    : c <= 70006)
                  : (c <= 70066 || (c >= 70081 && c <= 70084)))))))
            : (c <= 70106 || (c < 70405
              ? (c < 70280
                ? (c < 70163
                  ? (c < 70144
                    ? c == 70108
                    : c <= 70161)
                  : (c <= 70187 || (c >= 70272 && c <= 70278)))
                : (c <= 70280 || (c < 70303
                  ? (c < 70287
                    ? (c >= 70282 && c <= 70285)
                    : c <= 70301)
                  : (c <= 70312 || (c >= 70320 && c <= 70366)))))
              : (c <= 70412 || (c < 70453
                ? (c < 70442
                  ? (c < 70419
                    ? (c >= 70415 && c <= 70416)
                    : c <= 70440)
                  : (c <= 70448 || (c >= 70450 && c <= 70451)))
                : (c <= 70457 || (c < 70493
                  ? (c < 70480
                    ? c == 70461
                    : c <= 70480)
                  : (c <= 70497 || (c >= 70656 && c <= 70708)))))))))))))
      : (c <= 70730 || (c < 119894
        ? (c < 73056
          ? (c < 72001
            ? (c < 71424
              ? (c < 71128
                ? (c < 70852
                  ? (c < 70784
                    ? (c >= 70751 && c <= 70753)
                    : c <= 70831)
                  : (c <= 70853 || (c < 71040
                    ? c == 70855
                    : c <= 71086)))
                : (c <= 71131 || (c < 71296
                  ? (c < 71236
                    ? (c >= 71168 && c <= 71215)
                    : c <= 71236)
                  : (c <= 71338 || c == 71352))))
              : (c <= 71450 || (c < 71945
                ? (c < 71840
                  ? (c < 71680
                    ? (c >= 71488 && c <= 71494)
                    : c <= 71723)
                  : (c <= 71903 || (c >= 71935 && c <= 71942)))
                : (c <= 71945 || (c < 71960
                  ? (c < 71957
                    ? (c >= 71948 && c <= 71955)
                    : c <= 71958)
                  : (c <= 71983 || c == 71999))))))
            : (c <= 72001 || (c < 72349
              ? (c < 72192
                ? (c < 72161
                  ? (c < 72106
                    ? (c >= 72096 && c <= 72103)
                    : c <= 72144)
                  : (c <= 72161 || c == 72163))
                : (c <= 72192 || (c < 72272
                  ? (c < 72250
                    ? (c >= 72203 && c <= 72242)
                    : c <= 72250)
                  : (c <= 72272 || (c >= 72284 && c <= 72329)))))
              : (c <= 72349 || (c < 72818
                ? (c < 72714
                  ? (c < 72704
                    ? (c >= 72368 && c <= 72440)
                    : c <= 72712)
                  : (c <= 72750 || c == 72768))
                : (c <= 72847 || (c < 72971
                  ? (c < 72968
                    ? (c >= 72960 && c <= 72966)
                    : c <= 72969)
                  : (c <= 73008 || c == 73030))))))))
          : (c <= 73061 || (c < 93952
            ? (c < 82944
              ? (c < 73728
                ? (c < 73112
                  ? (c < 73066
                    ? (c >= 73063 && c <= 73064)
                    : c <= 73097)
                  : (c <= 73112 || (c < 73648
                    ? (c >= 73440 && c <= 73458)
                    : c <= 73648)))
                : (c <= 74649 || (c < 77712
                  ? (c < 74880
                    ? (c >= 74752 && c <= 74862)
                    : c <= 75075)
                  : (c <= 77808 || (c >= 77824 && c <= 78894)))))
              : (c <= 83526 || (c < 92928
                ? (c < 92784
                  ? (c < 92736
                    ? (c >= 92160 && c <= 92728)
                    : c <= 92766)
                  : (c <= 92862 || (c >= 92880 && c <= 92909)))
                : (c <= 92975 || (c < 93053
                  ? (c < 93027
                    ? (c >= 92992 && c <= 92995)
                    : c <= 93047)
                  : (c <= 93071 || (c >= 93760 && c <= 93823)))))))
            : (c <= 94026 || (c < 110589
              ? (c < 94208
                ? (c < 94176
                  ? (c < 94099
                    ? c == 94032
                    : c <= 94111)
                  : (c <= 94177 || c == 94179))
                : (c <= 100343 || (c < 110576
                  ? (c < 101632
                    ? (c >= 100352 && c <= 101589)
                    : c <= 101640)
                  : (c <= 110579 || (c >= 110581 && c <= 110587)))))
              : (c <= 110590 || (c < 113664
                ? (c < 110948
                  ? (c < 110928
                    ? (c >= 110592 && c <= 110882)
                    : c <= 110930)
                  : (c <= 110951 || (c >= 110960 && c <= 111355)))
                : (c <= 113770 || (c < 113808
                  ? (c < 113792
                    ? (c >= 113776 && c <= 113788)
                    : c <= 113800)
                  : (c <= 113817 || (c >= 119808 && c <= 119892)))))))))))
        : (c <= 119964 || (c < 125259
          ? (c < 120572
            ? (c < 120086
              ? (c < 119995
                ? (c < 119973
                  ? (c < 119970
                    ? (c >= 119966 && c <= 119967)
                    : c <= 119970)
                  : (c <= 119974 || (c < 119982
                    ? (c >= 119977 && c <= 119980)
                    : c <= 119993)))
                : (c <= 119995 || (c < 120071
                  ? (c < 120005
                    ? (c >= 119997 && c <= 120003)
                    : c <= 120069)
                  : (c <= 120074 || (c >= 120077 && c <= 120084)))))
              : (c <= 120092 || (c < 120138
                ? (c < 120128
                  ? (c < 120123
                    ? (c >= 120094 && c <= 120121)
                    : c <= 120126)
                  : (c <= 120132 || c == 120134))
                : (c <= 120144 || (c < 120514
                  ? (c < 120488
                    ? (c >= 120146 && c <= 120485)
                    : c <= 120512)
                  : (c <= 120538 || (c >= 120540 && c <= 120570)))))))
            : (c <= 120596 || (c < 123191
              ? (c < 120714
                ? (c < 120656
                  ? (c < 120630
                    ? (c >= 120598 && c <= 120628)
                    : c <= 120654)
                  : (c <= 120686 || (c >= 120688 && c <= 120712)))
                : (c <= 120744 || (c < 122624
                  ? (c < 120772
                    ? (c >= 120746 && c <= 120770)
                    : c <= 120779)
                  : (c <= 122654 || (c >= 123136 && c <= 123180)))))
              : (c <= 123197 || (c < 124904
                ? (c < 123584
                  ? (c < 123536
                    ? c == 123214
                    : c <= 123565)
                  : (c <= 123627 || (c >= 124896 && c <= 124902)))
                : (c <= 124907 || (c < 124928
                  ? (c < 124912
                    ? (c >= 124909 && c <= 124910)
                    : c <= 124926)
                  : (c <= 125124 || (c >= 125184 && c <= 125251)))))))))
          : (c <= 125259 || (c < 126559
            ? (c < 126535
              ? (c < 126505
                ? (c < 126497
                  ? (c < 126469
                    ? (c >= 126464 && c <= 126467)
                    : c <= 126495)
                  : (c <= 126498 || (c < 126503
                    ? c == 126500
                    : c <= 126503)))
                : (c <= 126514 || (c < 126523
                  ? (c < 126521
                    ? (c >= 126516 && c <= 126519)
                    : c <= 126521)
                  : (c <= 126523 || c == 126530))))
              : (c <= 126535 || (c < 126548
                ? (c < 126541
                  ? (c < 126539
                    ? c == 126537
                    : c <= 126539)
                  : (c <= 126543 || (c >= 126545 && c <= 126546)))
                : (c <= 126548 || (c < 126555
                  ? (c < 126553
                    ? c == 126551
                    : c <= 126553)
                  : (c <= 126555 || c == 126557))))))
            : (c <= 126559 || (c < 126625
              ? (c < 126580
                ? (c < 126567
                  ? (c < 126564
                    ? (c >= 126561 && c <= 126562)
                    : c <= 126564)
                  : (c <= 126570 || (c >= 126572 && c <= 126578)))
                : (c <= 126583 || (c < 126592
                  ? (c < 126590
                    ? (c >= 126585 && c <= 126588)
                    : c <= 126590)
                  : (c <= 126601 || (c >= 126603 && c <= 126619)))))
              : (c <= 126627 || (c < 177984
                ? (c < 131072
                  ? (c < 126635
                    ? (c >= 126629 && c <= 126633)
                    : c <= 126651)
                  : (c <= 173791 || (c >= 173824 && c <= 177976)))
                : (c <= 178205 || (c < 194560
                  ? (c < 183984
                    ? (c >= 178208 && c <= 183969)
                    : c <= 191456)
                  : (c <= 195101 || (c >= 196608 && c <= 201546)))))))))))))))));
}

static inline bool sym_identifier_character_set_2(int32_t c) {
  return (c < 43514
    ? (c < 4193
      ? (c < 2707
        ? (c < 1994
          ? (c < 910
            ? (c < 736
              ? (c < 186
                ? (c < 'a'
                  ? (c < '_'
                    ? (c >= 'A' && c <= 'Z')
                    : c <= '_')
                  : (c <= 'z' || (c < 181
                    ? c == 170
                    : c <= 181)))
                : (c <= 186 || (c < 248
                  ? (c < 216
                    ? (c >= 192 && c <= 214)
                    : c <= 246)
                  : (c <= 705 || (c >= 710 && c <= 721)))))
              : (c <= 740 || (c < 891
                ? (c < 880
                  ? (c < 750
                    ? c == 748
                    : c <= 750)
                  : (c <= 884 || (c >= 886 && c <= 887)))
                : (c <= 893 || (c < 904
                  ? (c < 902
                    ? c == 895
                    : c <= 902)
                  : (c <= 906 || c == 908))))))
            : (c <= 929 || (c < 1649
              ? (c < 1376
                ? (c < 1162
                  ? (c < 1015
                    ? (c >= 931 && c <= 1013)
                    : c <= 1153)
                  : (c <= 1327 || (c < 1369
                    ? (c >= 1329 && c <= 1366)
                    : c <= 1369)))
                : (c <= 1416 || (c < 1568
                  ? (c < 1519
                    ? (c >= 1488 && c <= 1514)
                    : c <= 1522)
                  : (c <= 1610 || (c >= 1646 && c <= 1647)))))
              : (c <= 1747 || (c < 1791
                ? (c < 1774
                  ? (c < 1765
                    ? c == 1749
                    : c <= 1766)
                  : (c <= 1775 || (c >= 1786 && c <= 1788)))
                : (c <= 1791 || (c < 1869
                  ? (c < 1810
                    ? c == 1808
                    : c <= 1839)
                  : (c <= 1957 || c == 1969))))))))
          : (c <= 2026 || (c < 2482
            ? (c < 2208
              ? (c < 2088
                ? (c < 2048
                  ? (c < 2042
                    ? (c >= 2036 && c <= 2037)
                    : c <= 2042)
                  : (c <= 2069 || (c < 2084
                    ? c == 2074
                    : c <= 2084)))
                : (c <= 2088 || (c < 2160
                  ? (c < 2144
                    ? (c >= 2112 && c <= 2136)
                    : c <= 2154)
                  : (c <= 2183 || (c >= 2185 && c <= 2190)))))
              : (c <= 2249 || (c < 2417
                ? (c < 2384
                  ? (c < 2365
                    ? (c >= 2308 && c <= 2361)
                    : c <= 2365)
                  : (c <= 2384 || (c >= 2392 && c <= 2401)))
                : (c <= 2432 || (c < 2451
                  ? (c < 2447
                    ? (c >= 2437 && c <= 2444)
                    : c <= 2448)
                  : (c <= 2472 || (c >= 2474 && c <= 2480)))))))
            : (c <= 2482 || (c < 2579
              ? (c < 2527
                ? (c < 2510
                  ? (c < 2493
                    ? (c >= 2486 && c <= 2489)
                    : c <= 2493)
                  : (c <= 2510 || (c >= 2524 && c <= 2525)))
                : (c <= 2529 || (c < 2565
                  ? (c < 2556
                    ? (c >= 2544 && c <= 2545)
                    : c <= 2556)
                  : (c <= 2570 || (c >= 2575 && c <= 2576)))))
              : (c <= 2600 || (c < 2649
                ? (c < 2613
                  ? (c < 2610
                    ? (c >= 2602 && c <= 2608)
                    : c <= 2611)
                  : (c <= 2614 || (c >= 2616 && c <= 2617)))
                : (c <= 2652 || (c < 2693
                  ? (c < 2674
                    ? c == 2654
                    : c <= 2676)
                  : (c <= 2701 || (c >= 2703 && c <= 2705)))))))))))
        : (c <= 2728 || (c < 3242
          ? (c < 2962
            ? (c < 2858
              ? (c < 2784
                ? (c < 2741
                  ? (c < 2738
                    ? (c >= 2730 && c <= 2736)
                    : c <= 2739)
                  : (c <= 2745 || (c < 2768
                    ? c == 2749
                    : c <= 2768)))
                : (c <= 2785 || (c < 2831
                  ? (c < 2821
                    ? c == 2809
                    : c <= 2828)
                  : (c <= 2832 || (c >= 2835 && c <= 2856)))))
              : (c <= 2864 || (c < 2911
                ? (c < 2877
                  ? (c < 2869
                    ? (c >= 2866 && c <= 2867)
                    : c <= 2873)
                  : (c <= 2877 || (c >= 2908 && c <= 2909)))
                : (c <= 2913 || (c < 2949
                  ? (c < 2947
                    ? c == 2929
                    : c <= 2947)
                  : (c <= 2954 || (c >= 2958 && c <= 2960)))))))
            : (c <= 2965 || (c < 3090
              ? (c < 2984
                ? (c < 2974
                  ? (c < 2972
                    ? (c >= 2969 && c <= 2970)
                    : c <= 2972)
                  : (c <= 2975 || (c >= 2979 && c <= 2980)))
                : (c <= 2986 || (c < 3077
                  ? (c < 3024
                    ? (c >= 2990 && c <= 3001)
                    : c <= 3024)
                  : (c <= 3084 || (c >= 3086 && c <= 3088)))))
              : (c <= 3112 || (c < 3168
                ? (c < 3160
                  ? (c < 3133
                    ? (c >= 3114 && c <= 3129)
                    : c <= 3133)
                  : (c <= 3162 || c == 3165))
                : (c <= 3169 || (c < 3214
                  ? (c < 3205
                    ? c == 3200
                    : c <= 3212)
                  : (c <= 3216 || (c >= 3218 && c <= 3240)))))))))
          : (c <= 3251 || (c < 3648
            ? (c < 3412
              ? (c < 3332
                ? (c < 3293
                  ? (c < 3261
                    ? (c >= 3253 && c <= 3257)
                    : c <= 3261)
                  : (c <= 3294 || (c < 3313
                    ? (c >= 3296 && c <= 3297)
                    : c <= 3314)))
                : (c <= 3340 || (c < 3389
                  ? (c < 3346
                    ? (c >= 3342 && c <= 3344)
                    : c <= 3386)
                  : (c <= 3389 || c == 3406))))
              : (c <= 3414 || (c < 3507
                ? (c < 3461
                  ? (c < 3450
                    ? (c >= 3423 && c <= 3425)
                    : c <= 3455)
                  : (c <= 3478 || (c >= 3482 && c <= 3505)))
                : (c <= 3515 || (c < 3585
                  ? (c < 3520
                    ? c == 3517
                    : c <= 3526)
                  : (c <= 3632 || c == 3634))))))
            : (c <= 3654 || (c < 3782
              ? (c < 3749
                ? (c < 3718
                  ? (c < 3716
                    ? (c >= 3713 && c <= 3714)
                    : c <= 3716)
                  : (c <= 3722 || (c >= 3724 && c <= 3747)))
                : (c <= 3749 || (c < 3773
                  ? (c < 3762
                    ? (c >= 3751 && c <= 3760)
                    : c <= 3762)
                  : (c <= 3773 || (c >= 3776 && c <= 3780)))))
              : (c <= 3782 || (c < 3976
                ? (c < 3904
                  ? (c < 3840
                    ? (c >= 3804 && c <= 3807)
                    : c <= 3840)
                  : (c <= 3911 || (c >= 3913 && c <= 3948)))
                : (c <= 3980 || (c < 4176
                  ? (c < 4159
                    ? (c >= 4096 && c <= 4138)
                    : c <= 4159)
                  : (c <= 4181 || (c >= 4186 && c <= 4189)))))))))))))
      : (c <= 4193 || (c < 8134
        ? (c < 6176
          ? (c < 4808
            ? (c < 4688
              ? (c < 4295
                ? (c < 4213
                  ? (c < 4206
                    ? (c >= 4197 && c <= 4198)
                    : c <= 4208)
                  : (c <= 4225 || (c < 4256
                    ? c == 4238
                    : c <= 4293)))
                : (c <= 4295 || (c < 4348
                  ? (c < 4304
                    ? c == 4301
                    : c <= 4346)
                  : (c <= 4680 || (c >= 4682 && c <= 4685)))))
              : (c <= 4694 || (c < 4752
                ? (c < 4704
                  ? (c < 4698
                    ? c == 4696
                    : c <= 4701)
                  : (c <= 4744 || (c >= 4746 && c <= 4749)))
                : (c <= 4784 || (c < 4800
                  ? (c < 4792
                    ? (c >= 4786 && c <= 4789)
                    : c <= 4798)
                  : (c <= 4800 || (c >= 4802 && c <= 4805)))))))
            : (c <= 4822 || (c < 5792
              ? (c < 5024
                ? (c < 4888
                  ? (c < 4882
                    ? (c >= 4824 && c <= 4880)
                    : c <= 4885)
                  : (c <= 4954 || (c >= 4992 && c <= 5007)))
                : (c <= 5109 || (c < 5743
                  ? (c < 5121
                    ? (c >= 5112 && c <= 5117)
                    : c <= 5740)
                  : (c <= 5759 || (c >= 5761 && c <= 5786)))))
              : (c <= 5866 || (c < 5984
                ? (c < 5919
                  ? (c < 5888
                    ? (c >= 5870 && c <= 5880)
                    : c <= 5905)
                  : (c <= 5937 || (c >= 5952 && c <= 5969)))
                : (c <= 5996 || (c < 6103
                  ? (c < 6016
                    ? (c >= 5998 && c <= 6000)
                    : c <= 6067)
                  : (c <= 6103 || c == 6108))))))))
          : (c <= 6264 || (c < 7312
            ? (c < 6823
              ? (c < 6512
                ? (c < 6320
                  ? (c < 6314
                    ? (c >= 6272 && c <= 6312)
                    : c <= 6314)
                  : (c <= 6389 || (c < 6480
                    ? (c >= 6400 && c <= 6430)
                    : c <= 6509)))
                : (c <= 6516 || (c < 6656
                  ? (c < 6576
                    ? (c >= 6528 && c <= 6571)
                    : c <= 6601)
                  : (c <= 6678 || (c >= 6688 && c <= 6740)))))
              : (c <= 6823 || (c < 7098
                ? (c < 7043
                  ? (c < 6981
                    ? (c >= 6917 && c <= 6963)
                    : c <= 6988)
                  : (c <= 7072 || (c >= 7086 && c <= 7087)))
                : (c <= 7141 || (c < 7258
                  ? (c < 7245
                    ? (c >= 7168 && c <= 7203)
                    : c <= 7247)
                  : (c <= 7293 || (c >= 7296 && c <= 7304)))))))
            : (c <= 7354 || (c < 8008
              ? (c < 7418
                ? (c < 7406
                  ? (c < 7401
                    ? (c >= 7357 && c <= 7359)
                    : c <= 7404)
                  : (c <= 7411 || (c >= 7413 && c <= 7414)))
                : (c <= 7418 || (c < 7960
                  ? (c < 7680
                    ? (c >= 7424 && c <= 7615)
                    : c <= 7957)
                  : (c <= 7965 || (c >= 7968 && c <= 8005)))))
              : (c <= 8013 || (c < 8031
                ? (c < 8027
                  ? (c < 8025
                    ? (c >= 8016 && c <= 8023)
                    : c <= 8025)
                  : (c <= 8027 || c == 8029))
                : (c <= 8061 || (c < 8126
                  ? (c < 8118
                    ? (c >= 8064 && c <= 8116)
                    : c <= 8124)
                  : (c <= 8126 || (c >= 8130 && c <= 8132)))))))))))
        : (c <= 8140 || (c < 12337
          ? (c < 8544
            ? (c < 8458
              ? (c < 8305
                ? (c < 8160
                  ? (c < 8150
                    ? (c >= 8144 && c <= 8147)
                    : c <= 8155)
                  : (c <= 8172 || (c < 8182
                    ? (c >= 8178 && c <= 8180)
                    : c <= 8188)))
                : (c <= 8305 || (c < 8450
                  ? (c < 8336
                    ? c == 8319
                    : c <= 8348)
                  : (c <= 8450 || c == 8455))))
              : (c <= 8467 || (c < 8488
                ? (c < 8484
                  ? (c < 8472
                    ? c == 8469
                    : c <= 8477)
                  : (c <= 8484 || c == 8486))
                : (c <= 8488 || (c < 8517
                  ? (c < 8508
                    ? (c >= 8490 && c <= 8505)
                    : c <= 8511)
                  : (c <= 8521 || c == 8526))))))
            : (c <= 8584 || (c < 11680
              ? (c < 11559
                ? (c < 11506
                  ? (c < 11499
                    ? (c >= 11264 && c <= 11492)
                    : c <= 11502)
                  : (c <= 11507 || (c >= 11520 && c <= 11557)))
                : (c <= 11559 || (c < 11631
                  ? (c < 11568
                    ? c == 11565
                    : c <= 11623)
                  : (c <= 11631 || (c >= 11648 && c <= 11670)))))
              : (c <= 11686 || (c < 11720
                ? (c < 11704
                  ? (c < 11696
                    ? (c >= 11688 && c <= 11694)
                    : c <= 11702)
                  : (c <= 11710 || (c >= 11712 && c <= 11718)))
                : (c <= 11726 || (c < 12293
                  ? (c < 11736
                    ? (c >= 11728 && c <= 11734)
                    : c <= 11742)
                  : (c <= 12295 || (c >= 12321 && c <= 12329)))))))))
          : (c <= 12341 || (c < 42891
            ? (c < 19968
              ? (c < 12549
                ? (c < 12445
                  ? (c < 12353
                    ? (c >= 12344 && c <= 12348)
                    : c <= 12438)
                  : (c <= 12447 || (c < 12540
                    ? (c >= 12449 && c <= 12538)
                    : c <= 12543)))
                : (c <= 12591 || (c < 12784
                  ? (c < 12704
                    ? (c >= 12593 && c <= 12686)
                    : c <= 12735)
                  : (c <= 12799 || (c >= 13312 && c <= 19903)))))
              : (c <= 42124 || (c < 42560
                ? (c < 42512
                  ? (c < 42240
                    ? (c >= 42192 && c <= 42237)
                    : c <= 42508)
                  : (c <= 42527 || (c >= 42538 && c <= 42539)))
                : (c <= 42606 || (c < 42775
                  ? (c < 42656
                    ? (c >= 42623 && c <= 42653)
                    : c <= 42735)
                  : (c <= 42783 || (c >= 42786 && c <= 42888)))))))
            : (c <= 42954 || (c < 43250
              ? (c < 43011
                ? (c < 42965
                  ? (c < 42963
                    ? (c >= 42960 && c <= 42961)
                    : c <= 42963)
                  : (c <= 42969 || (c >= 42994 && c <= 43009)))
                : (c <= 43013 || (c < 43072
                  ? (c < 43020
                    ? (c >= 43015 && c <= 43018)
                    : c <= 43042)
                  : (c <= 43123 || (c >= 43138 && c <= 43187)))))
              : (c <= 43255 || (c < 43360
                ? (c < 43274
                  ? (c < 43261
                    ? c == 43259
                    : c <= 43262)
                  : (c <= 43301 || (c >= 43312 && c <= 43334)))
                : (c <= 43388 || (c < 43488
                  ? (c < 43471
                    ? (c >= 43396 && c <= 43442)
                    : c <= 43471)
                  : (c <= 43492 || (c >= 43494 && c <= 43503)))))))))))))))
    : (c <= 43518 || (c < 70727
      ? (c < 66956
        ? (c < 64914
          ? (c < 43868
            ? (c < 43714
              ? (c < 43646
                ? (c < 43588
                  ? (c < 43584
                    ? (c >= 43520 && c <= 43560)
                    : c <= 43586)
                  : (c <= 43595 || (c < 43642
                    ? (c >= 43616 && c <= 43638)
                    : c <= 43642)))
                : (c <= 43695 || (c < 43705
                  ? (c < 43701
                    ? c == 43697
                    : c <= 43702)
                  : (c <= 43709 || c == 43712))))
              : (c <= 43714 || (c < 43785
                ? (c < 43762
                  ? (c < 43744
                    ? (c >= 43739 && c <= 43741)
                    : c <= 43754)
                  : (c <= 43764 || (c >= 43777 && c <= 43782)))
                : (c <= 43790 || (c < 43816
                  ? (c < 43808
                    ? (c >= 43793 && c <= 43798)
                    : c <= 43814)
                  : (c <= 43822 || (c >= 43824 && c <= 43866)))))))
            : (c <= 43881 || (c < 64287
              ? (c < 63744
                ? (c < 55216
                  ? (c < 44032
                    ? (c >= 43888 && c <= 44002)
                    : c <= 55203)
                  : (c <= 55238 || (c >= 55243 && c <= 55291)))
                : (c <= 64109 || (c < 64275
                  ? (c < 64256
                    ? (c >= 64112 && c <= 64217)
                    : c <= 64262)
                  : (c <= 64279 || c == 64285))))
              : (c <= 64296 || (c < 64323
                ? (c < 64318
                  ? (c < 64312
                    ? (c >= 64298 && c <= 64310)
                    : c <= 64316)
                  : (c <= 64318 || (c >= 64320 && c <= 64321)))
                : (c <= 64324 || (c < 64612
                  ? (c < 64467
                    ? (c >= 64326 && c <= 64433)
                    : c <= 64605)
                  : (c <= 64829 || (c >= 64848 && c <= 64911)))))))))
          : (c <= 64967 || (c < 65599
            ? (c < 65382
              ? (c < 65147
                ? (c < 65139
                  ? (c < 65137
                    ? (c >= 65008 && c <= 65017)
                    : c <= 65137)
                  : (c <= 65139 || (c < 65145
                    ? c == 65143
                    : c <= 65145)))
                : (c <= 65147 || (c < 65313
                  ? (c < 65151
                    ? c == 65149
                    : c <= 65276)
                  : (c <= 65338 || (c >= 65345 && c <= 65370)))))
              : (c <= 65437 || (c < 65498
                ? (c < 65482
                  ? (c < 65474
                    ? (c >= 65440 && c <= 65470)
                    : c <= 65479)
                  : (c <= 65487 || (c >= 65490 && c <= 65495)))
                : (c <= 65500 || (c < 65576
                  ? (c < 65549
                    ? (c >= 65536 && c <= 65547)
                    : c <= 65574)
                  : (c <= 65594 || (c >= 65596 && c <= 65597)))))))
            : (c <= 65613 || (c < 66464
              ? (c < 66208
                ? (c < 65856
                  ? (c < 65664
                    ? (c >= 65616 && c <= 65629)
                    : c <= 65786)
                  : (c <= 65908 || (c >= 66176 && c <= 66204)))
                : (c <= 66256 || (c < 66384
                  ? (c < 66349
                    ? (c >= 66304 && c <= 66335)
                    : c <= 66378)
                  : (c <= 66421 || (c >= 66432 && c <= 66461)))))
              : (c <= 66499 || (c < 66776
                ? (c < 66560
                  ? (c < 66513
                    ? (c >= 66504 && c <= 66511)
                    : c <= 66517)
                  : (c <= 66717 || (c >= 66736 && c <= 66771)))
                : (c <= 66811 || (c < 66928
                  ? (c < 66864
                    ? (c >= 66816 && c <= 66855)
                    : c <= 66915)
                  : (c <= 66938 || (c >= 66940 && c <= 66954)))))))))))
        : (c <= 66962 || (c < 68864
          ? (c < 67828
            ? (c < 67506
              ? (c < 67072
                ? (c < 66979
                  ? (c < 66967
                    ? (c >= 66964 && c <= 66965)
                    : c <= 66977)
                  : (c <= 66993 || (c < 67003
                    ? (c >= 66995 && c <= 67001)
                    : c <= 67004)))
                : (c <= 67382 || (c < 67456
                  ? (c < 67424
                    ? (c >= 67392 && c <= 67413)
                    : c <= 67431)
                  : (c <= 67461 || (c >= 67463 && c <= 67504)))))
              : (c <= 67514 || (c < 67644
                ? (c < 67594
                  ? (c < 67592
                    ? (c >= 67584 && c <= 67589)
                    : c <= 67592)
                  : (c <= 67637 || (c >= 67639 && c <= 67640)))
                : (c <= 67644 || (c < 67712
                  ? (c < 67680
                    ? (c >= 67647 && c <= 67669)
                    : c <= 67702)
                  : (c <= 67742 || (c >= 67808 && c <= 67826)))))))
            : (c <= 67829 || (c < 68224
              ? (c < 68096
                ? (c < 67968
                  ? (c < 67872
                    ? (c >= 67840 && c <= 67861)
                    : c <= 67897)
                  : (c <= 68023 || (c >= 68030 && c <= 68031)))
                : (c <= 68096 || (c < 68121
                  ? (c < 68117
                    ? (c >= 68112 && c <= 68115)
                    : c <= 68119)
                  : (c <= 68149 || (c >= 68192 && c <= 68220)))))
              : (c <= 68252 || (c < 68448
                ? (c < 68352
                  ? (c < 68297
                    ? (c >= 68288 && c <= 68295)
                    : c <= 68324)
                  : (c <= 68405 || (c >= 68416 && c <= 68437)))
                : (c <= 68466 || (c < 68736
                  ? (c < 68608
                    ? (c >= 68480 && c <= 68497)
                    : c <= 68680)
                  : (c <= 68786 || (c >= 68800 && c <= 68850)))))))))
          : (c <= 68899 || (c < 70106
            ? (c < 69749
              ? (c < 69488
                ? (c < 69376
                  ? (c < 69296
                    ? (c >= 69248 && c <= 69289)
                    : c <= 69297)
                  : (c <= 69404 || (c < 69424
                    ? c == 69415
                    : c <= 69445)))
                : (c <= 69505 || (c < 69635
                  ? (c < 69600
                    ? (c >= 69552 && c <= 69572)
                    : c <= 69622)
                  : (c <= 69687 || (c >= 69745 && c <= 69746)))))
              : (c <= 69749 || (c < 69959
                ? (c < 69891
                  ? (c < 69840
                    ? (c >= 69763 && c <= 69807)
                    : c <= 69864)
                  : (c <= 69926 || c == 69956))
                : (c <= 69959 || (c < 70019
                  ? (c < 70006
                    ? (c >= 69968 && c <= 70002)
                    : c <= 70006)
                  : (c <= 70066 || (c >= 70081 && c <= 70084)))))))
            : (c <= 70106 || (c < 70405
              ? (c < 70280
                ? (c < 70163
                  ? (c < 70144
                    ? c == 70108
                    : c <= 70161)
                  : (c <= 70187 || (c >= 70272 && c <= 70278)))
                : (c <= 70280 || (c < 70303
                  ? (c < 70287
                    ? (c >= 70282 && c <= 70285)
                    : c <= 70301)
                  : (c <= 70312 || (c >= 70320 && c <= 70366)))))
              : (c <= 70412 || (c < 70453
                ? (c < 70442
                  ? (c < 70419
                    ? (c >= 70415 && c <= 70416)
                    : c <= 70440)
                  : (c <= 70448 || (c >= 70450 && c <= 70451)))
                : (c <= 70457 || (c < 70493
                  ? (c < 70480
                    ? c == 70461
                    : c <= 70480)
                  : (c <= 70497 || (c >= 70656 && c <= 70708)))))))))))))
      : (c <= 70730 || (c < 119894
        ? (c < 73056
          ? (c < 72001
            ? (c < 71424
              ? (c < 71128
                ? (c < 70852
                  ? (c < 70784
                    ? (c >= 70751 && c <= 70753)
                    : c <= 70831)
                  : (c <= 70853 || (c < 71040
                    ? c == 70855
                    : c <= 71086)))
                : (c <= 71131 || (c < 71296
                  ? (c < 71236
                    ? (c >= 71168 && c <= 71215)
                    : c <= 71236)
                  : (c <= 71338 || c == 71352))))
              : (c <= 71450 || (c < 71945
                ? (c < 71840
                  ? (c < 71680
                    ? (c >= 71488 && c <= 71494)
                    : c <= 71723)
                  : (c <= 71903 || (c >= 71935 && c <= 71942)))
                : (c <= 71945 || (c < 71960
                  ? (c < 71957
                    ? (c >= 71948 && c <= 71955)
                    : c <= 71958)
                  : (c <= 71983 || c == 71999))))))
            : (c <= 72001 || (c < 72349
              ? (c < 72192
                ? (c < 72161
                  ? (c < 72106
                    ? (c >= 72096 && c <= 72103)
                    : c <= 72144)
                  : (c <= 72161 || c == 72163))
                : (c <= 72192 || (c < 72272
                  ? (c < 72250
                    ? (c >= 72203 && c <= 72242)
                    : c <= 72250)
                  : (c <= 72272 || (c >= 72284 && c <= 72329)))))
              : (c <= 72349 || (c < 72818
                ? (c < 72714
                  ? (c < 72704
                    ? (c >= 72368 && c <= 72440)
                    : c <= 72712)
                  : (c <= 72750 || c == 72768))
                : (c <= 72847 || (c < 72971
                  ? (c < 72968
                    ? (c >= 72960 && c <= 72966)
                    : c <= 72969)
                  : (c <= 73008 || c == 73030))))))))
          : (c <= 73061 || (c < 93952
            ? (c < 82944
              ? (c < 73728
                ? (c < 73112
                  ? (c < 73066
                    ? (c >= 73063 && c <= 73064)
                    : c <= 73097)
                  : (c <= 73112 || (c < 73648
                    ? (c >= 73440 && c <= 73458)
                    : c <= 73648)))
                : (c <= 74649 || (c < 77712
                  ? (c < 74880
                    ? (c >= 74752 && c <= 74862)
                    : c <= 75075)
                  : (c <= 77808 || (c >= 77824 && c <= 78894)))))
              : (c <= 83526 || (c < 92928
                ? (c < 92784
                  ? (c < 92736
                    ? (c >= 92160 && c <= 92728)
                    : c <= 92766)
                  : (c <= 92862 || (c >= 92880 && c <= 92909)))
                : (c <= 92975 || (c < 93053
                  ? (c < 93027
                    ? (c >= 92992 && c <= 92995)
                    : c <= 93047)
                  : (c <= 93071 || (c >= 93760 && c <= 93823)))))))
            : (c <= 94026 || (c < 110589
              ? (c < 94208
                ? (c < 94176
                  ? (c < 94099
                    ? c == 94032
                    : c <= 94111)
                  : (c <= 94177 || c == 94179))
                : (c <= 100343 || (c < 110576
                  ? (c < 101632
                    ? (c >= 100352 && c <= 101589)
                    : c <= 101640)
                  : (c <= 110579 || (c >= 110581 && c <= 110587)))))
              : (c <= 110590 || (c < 113664
                ? (c < 110948
                  ? (c < 110928
                    ? (c >= 110592 && c <= 110882)
                    : c <= 110930)
                  : (c <= 110951 || (c >= 110960 && c <= 111355)))
                : (c <= 113770 || (c < 113808
                  ? (c < 113792
                    ? (c >= 113776 && c <= 113788)
                    : c <= 113800)
                  : (c <= 113817 || (c >= 119808 && c <= 119892)))))))))))
        : (c <= 119964 || (c < 125259
          ? (c < 120572
            ? (c < 120086
              ? (c < 119995
                ? (c < 119973
                  ? (c < 119970
                    ? (c >= 119966 && c <= 119967)
                    : c <= 119970)
                  : (c <= 119974 || (c < 119982
                    ? (c >= 119977 && c <= 119980)
                    : c <= 119993)))
                : (c <= 119995 || (c < 120071
                  ? (c < 120005
                    ? (c >= 119997 && c <= 120003)
                    : c <= 120069)
                  : (c <= 120074 || (c >= 120077 && c <= 120084)))))
              : (c <= 120092 || (c < 120138
                ? (c < 120128
                  ? (c < 120123
                    ? (c >= 120094 && c <= 120121)
                    : c <= 120126)
                  : (c <= 120132 || c == 120134))
                : (c <= 120144 || (c < 120514
                  ? (c < 120488
                    ? (c >= 120146 && c <= 120485)
                    : c <= 120512)
                  : (c <= 120538 || (c >= 120540 && c <= 120570)))))))
            : (c <= 120596 || (c < 123191
              ? (c < 120714
                ? (c < 120656
                  ? (c < 120630
                    ? (c >= 120598 && c <= 120628)
                    : c <= 120654)
                  : (c <= 120686 || (c >= 120688 && c <= 120712)))
                : (c <= 120744 || (c < 122624
                  ? (c < 120772
                    ? (c >= 120746 && c <= 120770)
                    : c <= 120779)
                  : (c <= 122654 || (c >= 123136 && c <= 123180)))))
              : (c <= 123197 || (c < 124904
                ? (c < 123584
                  ? (c < 123536
                    ? c == 123214
                    : c <= 123565)
                  : (c <= 123627 || (c >= 124896 && c <= 124902)))
                : (c <= 124907 || (c < 124928
                  ? (c < 124912
                    ? (c >= 124909 && c <= 124910)
                    : c <= 124926)
                  : (c <= 125124 || (c >= 125184 && c <= 125251)))))))))
          : (c <= 125259 || (c < 126559
            ? (c < 126535
              ? (c < 126505
                ? (c < 126497
                  ? (c < 126469
                    ? (c >= 126464 && c <= 126467)
                    : c <= 126495)
                  : (c <= 126498 || (c < 126503
                    ? c == 126500
                    : c <= 126503)))
                : (c <= 126514 || (c < 126523
                  ? (c < 126521
                    ? (c >= 126516 && c <= 126519)
                    : c <= 126521)
                  : (c <= 126523 || c == 126530))))
              : (c <= 126535 || (c < 126548
                ? (c < 126541
                  ? (c < 126539
                    ? c == 126537
                    : c <= 126539)
                  : (c <= 126543 || (c >= 126545 && c <= 126546)))
                : (c <= 126548 || (c < 126555
                  ? (c < 126553
                    ? c == 126551
                    : c <= 126553)
                  : (c <= 126555 || c == 126557))))))
            : (c <= 126559 || (c < 126625
              ? (c < 126580
                ? (c < 126567
                  ? (c < 126564
                    ? (c >= 126561 && c <= 126562)
                    : c <= 126564)
                  : (c <= 126570 || (c >= 126572 && c <= 126578)))
                : (c <= 126583 || (c < 126592
                  ? (c < 126590
                    ? (c >= 126585 && c <= 126588)
                    : c <= 126590)
                  : (c <= 126601 || (c >= 126603 && c <= 126619)))))
              : (c <= 126627 || (c < 177984
                ? (c < 131072
                  ? (c < 126635
                    ? (c >= 126629 && c <= 126633)
                    : c <= 126651)
                  : (c <= 173791 || (c >= 173824 && c <= 177976)))
                : (c <= 178205 || (c < 194560
                  ? (c < 183984
                    ? (c >= 178208 && c <= 183969)
                    : c <= 191456)
                  : (c <= 195101 || (c >= 196608 && c <= 201546)))))))))))))))));
}

static inline bool sym_identifier_character_set_3(int32_t c) {
  return (c < 43514
    ? (c < 4193
      ? (c < 2707
        ? (c < 1994
          ? (c < 931
            ? (c < 748
              ? (c < 192
                ? (c < 170
                  ? (c < 'a'
                    ? (c >= 'A' && c <= 'Z')
                    : c <= 'z')
                  : (c <= 170 || (c < 186
                    ? c == 181
                    : c <= 186)))
                : (c <= 214 || (c < 710
                  ? (c < 248
                    ? (c >= 216 && c <= 246)
                    : c <= 705)
                  : (c <= 721 || (c >= 736 && c <= 740)))))
              : (c <= 748 || (c < 895
                ? (c < 886
                  ? (c < 880
                    ? c == 750
                    : c <= 884)
                  : (c <= 887 || (c >= 891 && c <= 893)))
                : (c <= 895 || (c < 908
                  ? (c < 904
                    ? c == 902
                    : c <= 906)
                  : (c <= 908 || (c >= 910 && c <= 929)))))))
            : (c <= 1013 || (c < 1649
              ? (c < 1376
                ? (c < 1329
                  ? (c < 1162
                    ? (c >= 1015 && c <= 1153)
                    : c <= 1327)
                  : (c <= 1366 || c == 1369))
                : (c <= 1416 || (c < 1568
                  ? (c < 1519
                    ? (c >= 1488 && c <= 1514)
                    : c <= 1522)
                  : (c <= 1610 || (c >= 1646 && c <= 1647)))))
              : (c <= 1747 || (c < 1791
                ? (c < 1774
                  ? (c < 1765
                    ? c == 1749
                    : c <= 1766)
                  : (c <= 1775 || (c >= 1786 && c <= 1788)))
                : (c <= 1791 || (c < 1869
                  ? (c < 1810
                    ? c == 1808
                    : c <= 1839)
                  : (c <= 1957 || c == 1969))))))))
          : (c <= 2026 || (c < 2482
            ? (c < 2208
              ? (c < 2088
                ? (c < 2048
                  ? (c < 2042
                    ? (c >= 2036 && c <= 2037)
                    : c <= 2042)
                  : (c <= 2069 || (c < 2084
                    ? c == 2074
                    : c <= 2084)))
                : (c <= 2088 || (c < 2160
                  ? (c < 2144
                    ? (c >= 2112 && c <= 2136)
                    : c <= 2154)
                  : (c <= 2183 || (c >= 2185 && c <= 2190)))))
              : (c <= 2249 || (c < 2417
                ? (c < 2384
                  ? (c < 2365
                    ? (c >= 2308 && c <= 2361)
                    : c <= 2365)
                  : (c <= 2384 || (c >= 2392 && c <= 2401)))
                : (c <= 2432 || (c < 2451
                  ? (c < 2447
                    ? (c >= 2437 && c <= 2444)
                    : c <= 2448)
                  : (c <= 2472 || (c >= 2474 && c <= 2480)))))))
            : (c <= 2482 || (c < 2579
              ? (c < 2527
                ? (c < 2510
                  ? (c < 2493
                    ? (c >= 2486 && c <= 2489)
                    : c <= 2493)
                  : (c <= 2510 || (c >= 2524 && c <= 2525)))
                : (c <= 2529 || (c < 2565
                  ? (c < 2556
                    ? (c >= 2544 && c <= 2545)
                    : c <= 2556)
                  : (c <= 2570 || (c >= 2575 && c <= 2576)))))
              : (c <= 2600 || (c < 2649
                ? (c < 2613
                  ? (c < 2610
                    ? (c >= 2602 && c <= 2608)
                    : c <= 2611)
                  : (c <= 2614 || (c >= 2616 && c <= 2617)))
                : (c <= 2652 || (c < 2693
                  ? (c < 2674
                    ? c == 2654
                    : c <= 2676)
                  : (c <= 2701 || (c >= 2703 && c <= 2705)))))))))))
        : (c <= 2728 || (c < 3242
          ? (c < 2962
            ? (c < 2858
              ? (c < 2784
                ? (c < 2741
                  ? (c < 2738
                    ? (c >= 2730 && c <= 2736)
                    : c <= 2739)
                  : (c <= 2745 || (c < 2768
                    ? c == 2749
                    : c <= 2768)))
                : (c <= 2785 || (c < 2831
                  ? (c < 2821
                    ? c == 2809
                    : c <= 2828)
                  : (c <= 2832 || (c >= 2835 && c <= 2856)))))
              : (c <= 2864 || (c < 2911
                ? (c < 2877
                  ? (c < 2869
                    ? (c >= 2866 && c <= 2867)
                    : c <= 2873)
                  : (c <= 2877 || (c >= 2908 && c <= 2909)))
                : (c <= 2913 || (c < 2949
                  ? (c < 2947
                    ? c == 2929
                    : c <= 2947)
                  : (c <= 2954 || (c >= 2958 && c <= 2960)))))))
            : (c <= 2965 || (c < 3090
              ? (c < 2984
                ? (c < 2974
                  ? (c < 2972
                    ? (c >= 2969 && c <= 2970)
                    : c <= 2972)
                  : (c <= 2975 || (c >= 2979 && c <= 2980)))
                : (c <= 2986 || (c < 3077
                  ? (c < 3024
                    ? (c >= 2990 && c <= 3001)
                    : c <= 3024)
                  : (c <= 3084 || (c >= 3086 && c <= 3088)))))
              : (c <= 3112 || (c < 3168
                ? (c < 3160
                  ? (c < 3133
                    ? (c >= 3114 && c <= 3129)
                    : c <= 3133)
                  : (c <= 3162 || c == 3165))
                : (c <= 3169 || (c < 3214
                  ? (c < 3205
                    ? c == 3200
                    : c <= 3212)
                  : (c <= 3216 || (c >= 3218 && c <= 3240)))))))))
          : (c <= 3251 || (c < 3648
            ? (c < 3412
              ? (c < 3332
                ? (c < 3293
                  ? (c < 3261
                    ? (c >= 3253 && c <= 3257)
                    : c <= 3261)
                  : (c <= 3294 || (c < 3313
                    ? (c >= 3296 && c <= 3297)
                    : c <= 3314)))
                : (c <= 3340 || (c < 3389
                  ? (c < 3346
                    ? (c >= 3342 && c <= 3344)
                    : c <= 3386)
                  : (c <= 3389 || c == 3406))))
              : (c <= 3414 || (c < 3507
                ? (c < 3461
                  ? (c < 3450
                    ? (c >= 3423 && c <= 3425)
                    : c <= 3455)
                  : (c <= 3478 || (c >= 3482 && c <= 3505)))
                : (c <= 3515 || (c < 3585
                  ? (c < 3520
                    ? c == 3517
                    : c <= 3526)
                  : (c <= 3632 || c == 3634))))))
            : (c <= 3654 || (c < 3782
              ? (c < 3749
                ? (c < 3718
                  ? (c < 3716
                    ? (c >= 3713 && c <= 3714)
                    : c <= 3716)
                  : (c <= 3722 || (c >= 3724 && c <= 3747)))
                : (c <= 3749 || (c < 3773
                  ? (c < 3762
                    ? (c >= 3751 && c <= 3760)
                    : c <= 3762)
                  : (c <= 3773 || (c >= 3776 && c <= 3780)))))
              : (c <= 3782 || (c < 3976
                ? (c < 3904
                  ? (c < 3840
                    ? (c >= 3804 && c <= 3807)
                    : c <= 3840)
                  : (c <= 3911 || (c >= 3913 && c <= 3948)))
                : (c <= 3980 || (c < 4176
                  ? (c < 4159
                    ? (c >= 4096 && c <= 4138)
                    : c <= 4159)
                  : (c <= 4181 || (c >= 4186 && c <= 4189)))))))))))))
      : (c <= 4193 || (c < 8134
        ? (c < 6176
          ? (c < 4808
            ? (c < 4688
              ? (c < 4295
                ? (c < 4213
                  ? (c < 4206
                    ? (c >= 4197 && c <= 4198)
                    : c <= 4208)
                  : (c <= 4225 || (c < 4256
                    ? c == 4238
                    : c <= 4293)))
                : (c <= 4295 || (c < 4348
                  ? (c < 4304
                    ? c == 4301
                    : c <= 4346)
                  : (c <= 4680 || (c >= 4682 && c <= 4685)))))
              : (c <= 4694 || (c < 4752
                ? (c < 4704
                  ? (c < 4698
                    ? c == 4696
                    : c <= 4701)
                  : (c <= 4744 || (c >= 4746 && c <= 4749)))
                : (c <= 4784 || (c < 4800
                  ? (c < 4792
                    ? (c >= 4786 && c <= 4789)
                    : c <= 4798)
                  : (c <= 4800 || (c >= 4802 && c <= 4805)))))))
            : (c <= 4822 || (c < 5792
              ? (c < 5024
                ? (c < 4888
                  ? (c < 4882
                    ? (c >= 4824 && c <= 4880)
                    : c <= 4885)
                  : (c <= 4954 || (c >= 4992 && c <= 5007)))
                : (c <= 5109 || (c < 5743
                  ? (c < 5121
                    ? (c >= 5112 && c <= 5117)
                    : c <= 5740)
                  : (c <= 5759 || (c >= 5761 && c <= 5786)))))
              : (c <= 5866 || (c < 5984
                ? (c < 5919
                  ? (c < 5888
                    ? (c >= 5870 && c <= 5880)
                    : c <= 5905)
                  : (c <= 5937 || (c >= 5952 && c <= 5969)))
                : (c <= 5996 || (c < 6103
                  ? (c < 6016
                    ? (c >= 5998 && c <= 6000)
                    : c <= 6067)
                  : (c <= 6103 || c == 6108))))))))
          : (c <= 6264 || (c < 7312
            ? (c < 6823
              ? (c < 6512
                ? (c < 6320
                  ? (c < 6314
                    ? (c >= 6272 && c <= 6312)
                    : c <= 6314)
                  : (c <= 6389 || (c < 6480
                    ? (c >= 6400 && c <= 6430)
                    : c <= 6509)))
                : (c <= 6516 || (c < 6656
                  ? (c < 6576
                    ? (c >= 6528 && c <= 6571)
                    : c <= 6601)
                  : (c <= 6678 || (c >= 6688 && c <= 6740)))))
              : (c <= 6823 || (c < 7098
                ? (c < 7043
                  ? (c < 6981
                    ? (c >= 6917 && c <= 6963)
                    : c <= 6988)
                  : (c <= 7072 || (c >= 7086 && c <= 7087)))
                : (c <= 7141 || (c < 7258
                  ? (c < 7245
                    ? (c >= 7168 && c <= 7203)
                    : c <= 7247)
                  : (c <= 7293 || (c >= 7296 && c <= 7304)))))))
            : (c <= 7354 || (c < 8008
              ? (c < 7418
                ? (c < 7406
                  ? (c < 7401
                    ? (c >= 7357 && c <= 7359)
                    : c <= 7404)
                  : (c <= 7411 || (c >= 7413 && c <= 7414)))
                : (c <= 7418 || (c < 7960
                  ? (c < 7680
                    ? (c >= 7424 && c <= 7615)
                    : c <= 7957)
                  : (c <= 7965 || (c >= 7968 && c <= 8005)))))
              : (c <= 8013 || (c < 8031
                ? (c < 8027
                  ? (c < 8025
                    ? (c >= 8016 && c <= 8023)
                    : c <= 8025)
                  : (c <= 8027 || c == 8029))
                : (c <= 8061 || (c < 8126
                  ? (c < 8118
                    ? (c >= 8064 && c <= 8116)
                    : c <= 8124)
                  : (c <= 8126 || (c >= 8130 && c <= 8132)))))))))))
        : (c <= 8140 || (c < 12337
          ? (c < 8544
            ? (c < 8458
              ? (c < 8305
                ? (c < 8160
                  ? (c < 8150
                    ? (c >= 8144 && c <= 8147)
                    : c <= 8155)
                  : (c <= 8172 || (c < 8182
                    ? (c >= 8178 && c <= 8180)
                    : c <= 8188)))
                : (c <= 8305 || (c < 8450
                  ? (c < 8336
                    ? c == 8319
                    : c <= 8348)
                  : (c <= 8450 || c == 8455))))
              : (c <= 8467 || (c < 8488
                ? (c < 8484
                  ? (c < 8472
                    ? c == 8469
                    : c <= 8477)
                  : (c <= 8484 || c == 8486))
                : (c <= 8488 || (c < 8517
                  ? (c < 8508
                    ? (c >= 8490 && c <= 8505)
                    : c <= 8511)
                  : (c <= 8521 || c == 8526))))))
            : (c <= 8584 || (c < 11680
              ? (c < 11559
                ? (c < 11506
                  ? (c < 11499
                    ? (c >= 11264 && c <= 11492)
                    : c <= 11502)
                  : (c <= 11507 || (c >= 11520 && c <= 11557)))
                : (c <= 11559 || (c < 11631
                  ? (c < 11568
                    ? c == 11565
                    : c <= 11623)
                  : (c <= 11631 || (c >= 11648 && c <= 11670)))))
              : (c <= 11686 || (c < 11720
                ? (c < 11704
                  ? (c < 11696
                    ? (c >= 11688 && c <= 11694)
                    : c <= 11702)
                  : (c <= 11710 || (c >= 11712 && c <= 11718)))
                : (c <= 11726 || (c < 12293
                  ? (c < 11736
                    ? (c >= 11728 && c <= 11734)
                    : c <= 11742)
                  : (c <= 12295 || (c >= 12321 && c <= 12329)))))))))
          : (c <= 12341 || (c < 42891
            ? (c < 19968
              ? (c < 12549
                ? (c < 12445
                  ? (c < 12353
                    ? (c >= 12344 && c <= 12348)
                    : c <= 12438)
                  : (c <= 12447 || (c < 12540
                    ? (c >= 12449 && c <= 12538)
                    : c <= 12543)))
                : (c <= 12591 || (c < 12784
                  ? (c < 12704
                    ? (c >= 12593 && c <= 12686)
                    : c <= 12735)
                  : (c <= 12799 || (c >= 13312 && c <= 19903)))))
              : (c <= 42124 || (c < 42560
                ? (c < 42512
                  ? (c < 42240
                    ? (c >= 42192 && c <= 42237)
                    : c <= 42508)
                  : (c <= 42527 || (c >= 42538 && c <= 42539)))
                : (c <= 42606 || (c < 42775
                  ? (c < 42656
                    ? (c >= 42623 && c <= 42653)
                    : c <= 42735)
                  : (c <= 42783 || (c >= 42786 && c <= 42888)))))))
            : (c <= 42954 || (c < 43250
              ? (c < 43011
                ? (c < 42965
                  ? (c < 42963
                    ? (c >= 42960 && c <= 42961)
                    : c <= 42963)
                  : (c <= 42969 || (c >= 42994 && c <= 43009)))
                : (c <= 43013 || (c < 43072
                  ? (c < 43020
                    ? (c >= 43015 && c <= 43018)
                    : c <= 43042)
                  : (c <= 43123 || (c >= 43138 && c <= 43187)))))
              : (c <= 43255 || (c < 43360
                ? (c < 43274
                  ? (c < 43261
                    ? c == 43259
                    : c <= 43262)
                  : (c <= 43301 || (c >= 43312 && c <= 43334)))
                : (c <= 43388 || (c < 43488
                  ? (c < 43471
                    ? (c >= 43396 && c <= 43442)
                    : c <= 43471)
                  : (c <= 43492 || (c >= 43494 && c <= 43503)))))))))))))))
    : (c <= 43518 || (c < 70727
      ? (c < 66956
        ? (c < 64914
          ? (c < 43868
            ? (c < 43714
              ? (c < 43646
                ? (c < 43588
                  ? (c < 43584
                    ? (c >= 43520 && c <= 43560)
                    : c <= 43586)
                  : (c <= 43595 || (c < 43642
                    ? (c >= 43616 && c <= 43638)
                    : c <= 43642)))
                : (c <= 43695 || (c < 43705
                  ? (c < 43701
                    ? c == 43697
                    : c <= 43702)
                  : (c <= 43709 || c == 43712))))
              : (c <= 43714 || (c < 43785
                ? (c < 43762
                  ? (c < 43744
                    ? (c >= 43739 && c <= 43741)
                    : c <= 43754)
                  : (c <= 43764 || (c >= 43777 && c <= 43782)))
                : (c <= 43790 || (c < 43816
                  ? (c < 43808
                    ? (c >= 43793 && c <= 43798)
                    : c <= 43814)
                  : (c <= 43822 || (c >= 43824 && c <= 43866)))))))
            : (c <= 43881 || (c < 64287
              ? (c < 63744
                ? (c < 55216
                  ? (c < 44032
                    ? (c >= 43888 && c <= 44002)
                    : c <= 55203)
                  : (c <= 55238 || (c >= 55243 && c <= 55291)))
                : (c <= 64109 || (c < 64275
                  ? (c < 64256
                    ? (c >= 64112 && c <= 64217)
                    : c <= 64262)
                  : (c <= 64279 || c == 64285))))
              : (c <= 64296 || (c < 64323
                ? (c < 64318
                  ? (c < 64312
                    ? (c >= 64298 && c <= 64310)
                    : c <= 64316)
                  : (c <= 64318 || (c >= 64320 && c <= 64321)))
                : (c <= 64324 || (c < 64612
                  ? (c < 64467
                    ? (c >= 64326 && c <= 64433)
                    : c <= 64605)
                  : (c <= 64829 || (c >= 64848 && c <= 64911)))))))))
          : (c <= 64967 || (c < 65599
            ? (c < 65382
              ? (c < 65147
                ? (c < 65139
                  ? (c < 65137
                    ? (c >= 65008 && c <= 65017)
                    : c <= 65137)
                  : (c <= 65139 || (c < 65145
                    ? c == 65143
                    : c <= 65145)))
                : (c <= 65147 || (c < 65313
                  ? (c < 65151
                    ? c == 65149
                    : c <= 65276)
                  : (c <= 65338 || (c >= 65345 && c <= 65370)))))
              : (c <= 65437 || (c < 65498
                ? (c < 65482
                  ? (c < 65474
                    ? (c >= 65440 && c <= 65470)
                    : c <= 65479)
                  : (c <= 65487 || (c >= 65490 && c <= 65495)))
                : (c <= 65500 || (c < 65576
                  ? (c < 65549
                    ? (c >= 65536 && c <= 65547)
                    : c <= 65574)
                  : (c <= 65594 || (c >= 65596 && c <= 65597)))))))
            : (c <= 65613 || (c < 66464
              ? (c < 66208
                ? (c < 65856
                  ? (c < 65664
                    ? (c >= 65616 && c <= 65629)
                    : c <= 65786)
                  : (c <= 65908 || (c >= 66176 && c <= 66204)))
                : (c <= 66256 || (c < 66384
                  ? (c < 66349
                    ? (c >= 66304 && c <= 66335)
                    : c <= 66378)
                  : (c <= 66421 || (c >= 66432 && c <= 66461)))))
              : (c <= 66499 || (c < 66776
                ? (c < 66560
                  ? (c < 66513
                    ? (c >= 66504 && c <= 66511)
                    : c <= 66517)
                  : (c <= 66717 || (c >= 66736 && c <= 66771)))
                : (c <= 66811 || (c < 66928
                  ? (c < 66864
                    ? (c >= 66816 && c <= 66855)
                    : c <= 66915)
                  : (c <= 66938 || (c >= 66940 && c <= 66954)))))))))))
        : (c <= 66962 || (c < 68864
          ? (c < 67828
            ? (c < 67506
              ? (c < 67072
                ? (c < 66979
                  ? (c < 66967
                    ? (c >= 66964 && c <= 66965)
                    : c <= 66977)
                  : (c <= 66993 || (c < 67003
                    ? (c >= 66995 && c <= 67001)
                    : c <= 67004)))
                : (c <= 67382 || (c < 67456
                  ? (c < 67424
                    ? (c >= 67392 && c <= 67413)
                    : c <= 67431)
                  : (c <= 67461 || (c >= 67463 && c <= 67504)))))
              : (c <= 67514 || (c < 67644
                ? (c < 67594
                  ? (c < 67592
                    ? (c >= 67584 && c <= 67589)
                    : c <= 67592)
                  : (c <= 67637 || (c >= 67639 && c <= 67640)))
                : (c <= 67644 || (c < 67712
                  ? (c < 67680
                    ? (c >= 67647 && c <= 67669)
                    : c <= 67702)
                  : (c <= 67742 || (c >= 67808 && c <= 67826)))))))
            : (c <= 67829 || (c < 68224
              ? (c < 68096
                ? (c < 67968
                  ? (c < 67872
                    ? (c >= 67840 && c <= 67861)
                    : c <= 67897)
                  : (c <= 68023 || (c >= 68030 && c <= 68031)))
                : (c <= 68096 || (c < 68121
                  ? (c < 68117
                    ? (c >= 68112 && c <= 68115)
                    : c <= 68119)
                  : (c <= 68149 || (c >= 68192 && c <= 68220)))))
              : (c <= 68252 || (c < 68448
                ? (c < 68352
                  ? (c < 68297
                    ? (c >= 68288 && c <= 68295)
                    : c <= 68324)
                  : (c <= 68405 || (c >= 68416 && c <= 68437)))
                : (c <= 68466 || (c < 68736
                  ? (c < 68608
                    ? (c >= 68480 && c <= 68497)
                    : c <= 68680)
                  : (c <= 68786 || (c >= 68800 && c <= 68850)))))))))
          : (c <= 68899 || (c < 70106
            ? (c < 69749
              ? (c < 69488
                ? (c < 69376
                  ? (c < 69296
                    ? (c >= 69248 && c <= 69289)
                    : c <= 69297)
                  : (c <= 69404 || (c < 69424
                    ? c == 69415
                    : c <= 69445)))
                : (c <= 69505 || (c < 69635
                  ? (c < 69600
                    ? (c >= 69552 && c <= 69572)
                    : c <= 69622)
                  : (c <= 69687 || (c >= 69745 && c <= 69746)))))
              : (c <= 69749 || (c < 69959
                ? (c < 69891
                  ? (c < 69840
                    ? (c >= 69763 && c <= 69807)
                    : c <= 69864)
                  : (c <= 69926 || c == 69956))
                : (c <= 69959 || (c < 70019
                  ? (c < 70006
                    ? (c >= 69968 && c <= 70002)
                    : c <= 70006)
                  : (c <= 70066 || (c >= 70081 && c <= 70084)))))))
            : (c <= 70106 || (c < 70405
              ? (c < 70280
                ? (c < 70163
                  ? (c < 70144
                    ? c == 70108
                    : c <= 70161)
                  : (c <= 70187 || (c >= 70272 && c <= 70278)))
                : (c <= 70280 || (c < 70303
                  ? (c < 70287
                    ? (c >= 70282 && c <= 70285)
                    : c <= 70301)
                  : (c <= 70312 || (c >= 70320 && c <= 70366)))))
              : (c <= 70412 || (c < 70453
                ? (c < 70442
                  ? (c < 70419
                    ? (c >= 70415 && c <= 70416)
                    : c <= 70440)
                  : (c <= 70448 || (c >= 70450 && c <= 70451)))
                : (c <= 70457 || (c < 70493
                  ? (c < 70480
                    ? c == 70461
                    : c <= 70480)
                  : (c <= 70497 || (c >= 70656 && c <= 70708)))))))))))))
      : (c <= 70730 || (c < 119894
        ? (c < 73056
          ? (c < 72001
            ? (c < 71424
              ? (c < 71128
                ? (c < 70852
                  ? (c < 70784
                    ? (c >= 70751 && c <= 70753)
                    : c <= 70831)
                  : (c <= 70853 || (c < 71040
                    ? c == 70855
                    : c <= 71086)))
                : (c <= 71131 || (c < 71296
                  ? (c < 71236
                    ? (c >= 71168 && c <= 71215)
                    : c <= 71236)
                  : (c <= 71338 || c == 71352))))
              : (c <= 71450 || (c < 71945
                ? (c < 71840
                  ? (c < 71680
                    ? (c >= 71488 && c <= 71494)
                    : c <= 71723)
                  : (c <= 71903 || (c >= 71935 && c <= 71942)))
                : (c <= 71945 || (c < 71960
                  ? (c < 71957
                    ? (c >= 71948 && c <= 71955)
                    : c <= 71958)
                  : (c <= 71983 || c == 71999))))))
            : (c <= 72001 || (c < 72349
              ? (c < 72192
                ? (c < 72161
                  ? (c < 72106
                    ? (c >= 72096 && c <= 72103)
                    : c <= 72144)
                  : (c <= 72161 || c == 72163))
                : (c <= 72192 || (c < 72272
                  ? (c < 72250
                    ? (c >= 72203 && c <= 72242)
                    : c <= 72250)
                  : (c <= 72272 || (c >= 72284 && c <= 72329)))))
              : (c <= 72349 || (c < 72818
                ? (c < 72714
                  ? (c < 72704
                    ? (c >= 72368 && c <= 72440)
                    : c <= 72712)
                  : (c <= 72750 || c == 72768))
                : (c <= 72847 || (c < 72971
                  ? (c < 72968
                    ? (c >= 72960 && c <= 72966)
                    : c <= 72969)
                  : (c <= 73008 || c == 73030))))))))
          : (c <= 73061 || (c < 93952
            ? (c < 82944
              ? (c < 73728
                ? (c < 73112
                  ? (c < 73066
                    ? (c >= 73063 && c <= 73064)
                    : c <= 73097)
                  : (c <= 73112 || (c < 73648
                    ? (c >= 73440 && c <= 73458)
                    : c <= 73648)))
                : (c <= 74649 || (c < 77712
                  ? (c < 74880
                    ? (c >= 74752 && c <= 74862)
                    : c <= 75075)
                  : (c <= 77808 || (c >= 77824 && c <= 78894)))))
              : (c <= 83526 || (c < 92928
                ? (c < 92784
                  ? (c < 92736
                    ? (c >= 92160 && c <= 92728)
                    : c <= 92766)
                  : (c <= 92862 || (c >= 92880 && c <= 92909)))
                : (c <= 92975 || (c < 93053
                  ? (c < 93027
                    ? (c >= 92992 && c <= 92995)
                    : c <= 93047)
                  : (c <= 93071 || (c >= 93760 && c <= 93823)))))))
            : (c <= 94026 || (c < 110589
              ? (c < 94208
                ? (c < 94176
                  ? (c < 94099
                    ? c == 94032
                    : c <= 94111)
                  : (c <= 94177 || c == 94179))
                : (c <= 100343 || (c < 110576
                  ? (c < 101632
                    ? (c >= 100352 && c <= 101589)
                    : c <= 101640)
                  : (c <= 110579 || (c >= 110581 && c <= 110587)))))
              : (c <= 110590 || (c < 113664
                ? (c < 110948
                  ? (c < 110928
                    ? (c >= 110592 && c <= 110882)
                    : c <= 110930)
                  : (c <= 110951 || (c >= 110960 && c <= 111355)))
                : (c <= 113770 || (c < 113808
                  ? (c < 113792
                    ? (c >= 113776 && c <= 113788)
                    : c <= 113800)
                  : (c <= 113817 || (c >= 119808 && c <= 119892)))))))))))
        : (c <= 119964 || (c < 125259
          ? (c < 120572
            ? (c < 120086
              ? (c < 119995
                ? (c < 119973
                  ? (c < 119970
                    ? (c >= 119966 && c <= 119967)
                    : c <= 119970)
                  : (c <= 119974 || (c < 119982
                    ? (c >= 119977 && c <= 119980)
                    : c <= 119993)))
                : (c <= 119995 || (c < 120071
                  ? (c < 120005
                    ? (c >= 119997 && c <= 120003)
                    : c <= 120069)
                  : (c <= 120074 || (c >= 120077 && c <= 120084)))))
              : (c <= 120092 || (c < 120138
                ? (c < 120128
                  ? (c < 120123
                    ? (c >= 120094 && c <= 120121)
                    : c <= 120126)
                  : (c <= 120132 || c == 120134))
                : (c <= 120144 || (c < 120514
                  ? (c < 120488
                    ? (c >= 120146 && c <= 120485)
                    : c <= 120512)
                  : (c <= 120538 || (c >= 120540 && c <= 120570)))))))
            : (c <= 120596 || (c < 123191
              ? (c < 120714
                ? (c < 120656
                  ? (c < 120630
                    ? (c >= 120598 && c <= 120628)
                    : c <= 120654)
                  : (c <= 120686 || (c >= 120688 && c <= 120712)))
                : (c <= 120744 || (c < 122624
                  ? (c < 120772
                    ? (c >= 120746 && c <= 120770)
                    : c <= 120779)
                  : (c <= 122654 || (c >= 123136 && c <= 123180)))))
              : (c <= 123197 || (c < 124904
                ? (c < 123584
                  ? (c < 123536
                    ? c == 123214
                    : c <= 123565)
                  : (c <= 123627 || (c >= 124896 && c <= 124902)))
                : (c <= 124907 || (c < 124928
                  ? (c < 124912
                    ? (c >= 124909 && c <= 124910)
                    : c <= 124926)
                  : (c <= 125124 || (c >= 125184 && c <= 125251)))))))))
          : (c <= 125259 || (c < 126559
            ? (c < 126535
              ? (c < 126505
                ? (c < 126497
                  ? (c < 126469
                    ? (c >= 126464 && c <= 126467)
                    : c <= 126495)
                  : (c <= 126498 || (c < 126503
                    ? c == 126500
                    : c <= 126503)))
                : (c <= 126514 || (c < 126523
                  ? (c < 126521
                    ? (c >= 126516 && c <= 126519)
                    : c <= 126521)
                  : (c <= 126523 || c == 126530))))
              : (c <= 126535 || (c < 126548
                ? (c < 126541
                  ? (c < 126539
                    ? c == 126537
                    : c <= 126539)
                  : (c <= 126543 || (c >= 126545 && c <= 126546)))
                : (c <= 126548 || (c < 126555
                  ? (c < 126553
                    ? c == 126551
                    : c <= 126553)
                  : (c <= 126555 || c == 126557))))))
            : (c <= 126559 || (c < 126625
              ? (c < 126580
                ? (c < 126567
                  ? (c < 126564
                    ? (c >= 126561 && c <= 126562)
                    : c <= 126564)
                  : (c <= 126570 || (c >= 126572 && c <= 126578)))
                : (c <= 126583 || (c < 126592
                  ? (c < 126590
                    ? (c >= 126585 && c <= 126588)
                    : c <= 126590)
                  : (c <= 126601 || (c >= 126603 && c <= 126619)))))
              : (c <= 126627 || (c < 177984
                ? (c < 131072
                  ? (c < 126635
                    ? (c >= 126629 && c <= 126633)
                    : c <= 126651)
                  : (c <= 173791 || (c >= 173824 && c <= 177976)))
                : (c <= 178205 || (c < 194560
                  ? (c < 183984
                    ? (c >= 178208 && c <= 183969)
                    : c <= 191456)
                  : (c <= 195101 || (c >= 196608 && c <= 201546)))))))))))))))));
}

static inline bool sym_identifier_character_set_4(int32_t c) {
  return (c < 43642
    ? (c < 3792
      ? (c < 2763
        ? (c < 2112
          ? (c < 1162
            ? (c < 748
              ? (c < 186
                ? (c < 170
                  ? (c < 'a'
                    ? (c >= '0' && c <= 'Z')
                    : c <= 'z')
                  : (c <= 170 || (c < 183
                    ? c == 181
                    : c <= 183)))
                : (c <= 186 || (c < 248
                  ? (c < 216
                    ? (c >= 192 && c <= 214)
                    : c <= 246)
                  : (c <= 705 || (c < 736
                    ? (c >= 710 && c <= 721)
                    : c <= 740)))))
              : (c <= 748 || (c < 902
                ? (c < 886
                  ? (c < 768
                    ? c == 750
                    : c <= 884)
                  : (c <= 887 || (c < 895
                    ? (c >= 891 && c <= 893)
                    : c <= 895)))
                : (c <= 906 || (c < 931
                  ? (c < 910
                    ? c == 908
                    : c <= 929)
                  : (c <= 1013 || (c < 1155
                    ? (c >= 1015 && c <= 1153)
                    : c <= 1159)))))))
            : (c <= 1327 || (c < 1568
              ? (c < 1473
                ? (c < 1376
                  ? (c < 1369
                    ? (c >= 1329 && c <= 1366)
                    : c <= 1369)
                  : (c <= 1416 || (c < 1471
                    ? (c >= 1425 && c <= 1469)
                    : c <= 1471)))
                : (c <= 1474 || (c < 1488
                  ? (c < 1479
                    ? (c >= 1476 && c <= 1477)
                    : c <= 1479)
                  : (c <= 1514 || (c < 1552
                    ? (c >= 1519 && c <= 1522)
                    : c <= 1562)))))
              : (c <= 1641 || (c < 1808
                ? (c < 1759
                  ? (c < 1749
                    ? (c >= 1646 && c <= 1747)
                    : c <= 1756)
                  : (c <= 1768 || (c < 1791
                    ? (c >= 1770 && c <= 1788)
                    : c <= 1791)))
                : (c <= 1866 || (c < 2042
                  ? (c < 1984
                    ? (c >= 1869 && c <= 1969)
                    : c <= 2037)
                  : (c <= 2042 || (c < 2048
                    ? c == 2045
                    : c <= 2093)))))))))
          : (c <= 2139 || (c < 2565
            ? (c < 2482
              ? (c < 2406
                ? (c < 2185
                  ? (c < 2160
                    ? (c >= 2144 && c <= 2154)
                    : c <= 2183)
                  : (c <= 2190 || (c < 2275
                    ? (c >= 2200 && c <= 2273)
                    : c <= 2403)))
                : (c <= 2415 || (c < 2447
                  ? (c < 2437
                    ? (c >= 2417 && c <= 2435)
                    : c <= 2444)
                  : (c <= 2448 || (c < 2474
                    ? (c >= 2451 && c <= 2472)
                    : c <= 2480)))))
              : (c <= 2482 || (c < 2524
                ? (c < 2503
                  ? (c < 2492
                    ? (c >= 2486 && c <= 2489)
                    : c <= 2500)
                  : (c <= 2504 || (c < 2519
                    ? (c >= 2507 && c <= 2510)
                    : c <= 2519)))
                : (c <= 2525 || (c < 2556
                  ? (c < 2534
                    ? (c >= 2527 && c <= 2531)
                    : c <= 2545)
                  : (c <= 2556 || (c < 2561
                    ? c == 2558
                    : c <= 2563)))))))
            : (c <= 2570 || (c < 2649
              ? (c < 2616
                ? (c < 2602
                  ? (c < 2579
                    ? (c >= 2575 && c <= 2576)
                    : c <= 2600)
                  : (c <= 2608 || (c < 2613
                    ? (c >= 2610 && c <= 2611)
                    : c <= 2614)))
                : (c <= 2617 || (c < 2631
                  ? (c < 2622
                    ? c == 2620
                    : c <= 2626)
                  : (c <= 2632 || (c < 2641
                    ? (c >= 2635 && c <= 2637)
                    : c <= 2641)))))
              : (c <= 2652 || (c < 2707
                ? (c < 2689
                  ? (c < 2662
                    ? c == 2654
                    : c <= 2677)
                  : (c <= 2691 || (c < 2703
                    ? (c >= 2693 && c <= 2701)
                    : c <= 2705)))
                : (c <= 2728 || (c < 2741
                  ? (c < 2738
                    ? (c >= 2730 && c <= 2736)
                    : c <= 2739)
                  : (c <= 2745 || (c < 2759
                    ? (c >= 2748 && c <= 2757)
                    : c <= 2761)))))))))))
        : (c <= 2765 || (c < 3200
          ? (c < 2969
            ? (c < 2876
              ? (c < 2821
                ? (c < 2790
                  ? (c < 2784
                    ? c == 2768
                    : c <= 2787)
                  : (c <= 2799 || (c < 2817
                    ? (c >= 2809 && c <= 2815)
                    : c <= 2819)))
                : (c <= 2828 || (c < 2858
                  ? (c < 2835
                    ? (c >= 2831 && c <= 2832)
                    : c <= 2856)
                  : (c <= 2864 || (c < 2869
                    ? (c >= 2866 && c <= 2867)
                    : c <= 2873)))))
              : (c <= 2884 || (c < 2918
                ? (c < 2901
                  ? (c < 2891
                    ? (c >= 2887 && c <= 2888)
                    : c <= 2893)
                  : (c <= 2903 || (c < 2911
                    ? (c >= 2908 && c <= 2909)
                    : c <= 2915)))
                : (c <= 2927 || (c < 2949
                  ? (c < 2946
                    ? c == 2929
                    : c <= 2947)
                  : (c <= 2954 || (c < 2962
                    ? (c >= 2958 && c <= 2960)
                    : c <= 2965)))))))
            : (c <= 2970 || (c < 3072
              ? (c < 3006
                ? (c < 2979
                  ? (c < 2974
                    ? c == 2972
                    : c <= 2975)
                  : (c <= 2980 || (c < 2990
                    ? (c >= 2984 && c <= 2986)
                    : c <= 3001)))
                : (c <= 3010 || (c < 3024
                  ? (c < 3018
                    ? (c >= 3014 && c <= 3016)
                    : c <= 3021)
                  : (c <= 3024 || (c < 3046
                    ? c == 3031
                    : c <= 3055)))))
              : (c <= 3084 || (c < 3146
                ? (c < 3114
                  ? (c < 3090
                    ? (c >= 3086 && c <= 3088)
                    : c <= 3112)
                  : (c <= 3129 || (c < 3142
                    ? (c >= 3132 && c <= 3140)
                    : c <= 3144)))
                : (c <= 3149 || (c < 3165
                  ? (c < 3160
                    ? (c >= 3157 && c <= 3158)
                    : c <= 3162)
                  : (c <= 3165 || (c < 3174
                    ? (c >= 3168 && c <= 3171)
                    : c <= 3183)))))))))
          : (c <= 3203 || (c < 3461
            ? (c < 3302
              ? (c < 3260
                ? (c < 3218
                  ? (c < 3214
                    ? (c >= 3205 && c <= 3212)
                    : c <= 3216)
                  : (c <= 3240 || (c < 3253
                    ? (c >= 3242 && c <= 3251)
                    : c <= 3257)))
                : (c <= 3268 || (c < 3285
                  ? (c < 3274
                    ? (c >= 3270 && c <= 3272)
                    : c <= 3277)
                  : (c <= 3286 || (c < 3296
                    ? (c >= 3293 && c <= 3294)
                    : c <= 3299)))))
              : (c <= 3311 || (c < 3402
                ? (c < 3342
                  ? (c < 3328
                    ? (c >= 3313 && c <= 3314)
                    : c <= 3340)
                  : (c <= 3344 || (c < 3398
                    ? (c >= 3346 && c <= 3396)
                    : c <= 3400)))
                : (c <= 3406 || (c < 3430
                  ? (c < 3423
                    ? (c >= 3412 && c <= 3415)
                    : c <= 3427)
                  : (c <= 3439 || (c < 3457
                    ? (c >= 3450 && c <= 3455)
                    : c <= 3459)))))))
            : (c <= 3478 || (c < 3648
              ? (c < 3535
                ? (c < 3517
                  ? (c < 3507
                    ? (c >= 3482 && c <= 3505)
                    : c <= 3515)
                  : (c <= 3517 || (c < 3530
                    ? (c >= 3520 && c <= 3526)
                    : c <= 3530)))
                : (c <= 3540 || (c < 3558
                  ? (c < 3544
                    ? c == 3542
                    : c <= 3551)
                  : (c <= 3567 || (c < 3585
                    ? (c >= 3570 && c <= 3571)
                    : c <= 3642)))))
              : (c <= 3662 || (c < 3749
                ? (c < 3716
                  ? (c < 3713
                    ? (c >= 3664 && c <= 3673)
                    : c <= 3714)
                  : (c <= 3716 || (c < 3724
                    ? (c >= 3718 && c <= 3722)
                    : c <= 3747)))
                : (c <= 3749 || (c < 3782
                  ? (c < 3776
                    ? (c >= 3751 && c <= 3773)
                    : c <= 3780)
                  : (c <= 3782 || (c >= 3784 && c <= 3789)))))))))))))
      : (c <= 3801 || (c < 8027
        ? (c < 5952
          ? (c < 4698
            ? (c < 3993
              ? (c < 3895
                ? (c < 3864
                  ? (c < 3840
                    ? (c >= 3804 && c <= 3807)
                    : c <= 3840)
                  : (c <= 3865 || (c < 3893
                    ? (c >= 3872 && c <= 3881)
                    : c <= 3893)))
                : (c <= 3895 || (c < 3913
                  ? (c < 3902
                    ? c == 3897
                    : c <= 3911)
                  : (c <= 3948 || (c < 3974
                    ? (c >= 3953 && c <= 3972)
                    : c <= 3991)))))
              : (c <= 4028 || (c < 4301
                ? (c < 4176
                  ? (c < 4096
                    ? c == 4038
                    : c <= 4169)
                  : (c <= 4253 || (c < 4295
                    ? (c >= 4256 && c <= 4293)
                    : c <= 4295)))
                : (c <= 4301 || (c < 4682
                  ? (c < 4348
                    ? (c >= 4304 && c <= 4346)
                    : c <= 4680)
                  : (c <= 4685 || (c < 4696
                    ? (c >= 4688 && c <= 4694)
                    : c <= 4696)))))))
            : (c <= 4701 || (c < 4957
              ? (c < 4800
                ? (c < 4752
                  ? (c < 4746
                    ? (c >= 4704 && c <= 4744)
                    : c <= 4749)
                  : (c <= 4784 || (c < 4792
                    ? (c >= 4786 && c <= 4789)
                    : c <= 4798)))
                : (c <= 4800 || (c < 4824
                  ? (c < 4808
                    ? (c >= 4802 && c <= 4805)
                    : c <= 4822)
                  : (c <= 4880 || (c < 4888
                    ? (c >= 4882 && c <= 4885)
                    : c <= 4954)))))
              : (c <= 4959 || (c < 5743
                ? (c < 5024
                  ? (c < 4992
                    ? (c >= 4969 && c <= 4977)
                    : c <= 5007)
                  : (c <= 5109 || (c < 5121
                    ? (c >= 5112 && c <= 5117)
                    : c <= 5740)))
                : (c <= 5759 || (c < 5870
                  ? (c < 5792
                    ? (c >= 5761 && c <= 5786)
                    : c <= 5866)
                  : (c <= 5880 || (c < 5919
                    ? (c >= 5888 && c <= 5909)
                    : c <= 5940)))))))))
          : (c <= 5971 || (c < 6783
            ? (c < 6320
              ? (c < 6108
                ? (c < 6002
                  ? (c < 5998
                    ? (c >= 5984 && c <= 5996)
                    : c <= 6000)
                  : (c <= 6003 || (c < 6103
                    ? (c >= 6016 && c <= 6099)
                    : c <= 6103)))
                : (c <= 6109 || (c < 6159
                  ? (c < 6155
                    ? (c >= 6112 && c <= 6121)
                    : c <= 6157)
                  : (c <= 6169 || (c < 6272
                    ? (c >= 6176 && c <= 6264)
                    : c <= 6314)))))
              : (c <= 6389 || (c < 6528
                ? (c < 6448
                  ? (c < 6432
                    ? (c >= 6400 && c <= 6430)
                    : c <= 6443)
                  : (c <= 6459 || (c < 6512
                    ? (c >= 6470 && c <= 6509)
                    : c <= 6516)))
                : (c <= 6571 || (c < 6656
                  ? (c < 6608
                    ? (c >= 6576 && c <= 6601)
                    : c <= 6618)
                  : (c <= 6683 || (c < 6752
                    ? (c >= 6688 && c <= 6750)
                    : c <= 6780)))))))
            : (c <= 6793 || (c < 7296
              ? (c < 6992
                ? (c < 6832
                  ? (c < 6823
                    ? (c >= 6800 && c <= 6809)
                    : c <= 6823)
                  : (c <= 6845 || (c < 6912
                    ? (c >= 6847 && c <= 6862)
                    : c <= 6988)))
                : (c <= 7001 || (c < 7168
                  ? (c < 7040
                    ? (c >= 7019 && c <= 7027)
                    : c <= 7155)
                  : (c <= 7223 || (c < 7245
                    ? (c >= 7232 && c <= 7241)
                    : c <= 7293)))))
              : (c <= 7304 || (c < 7960
                ? (c < 7376
                  ? (c < 7357
                    ? (c >= 7312 && c <= 7354)
                    : c <= 7359)
                  : (c <= 7378 || (c < 7424
                    ? (c >= 7380 && c <= 7418)
                    : c <= 7957)))
                : (c <= 7965 || (c < 8016
                  ? (c < 8008
                    ? (c >= 7968 && c <= 8005)
                    : c <= 8013)
                  : (c <= 8023 || c == 8025))))))))))
        : (c <= 8027 || (c < 11728
          ? (c < 8469
            ? (c < 8182
              ? (c < 8130
                ? (c < 8064
                  ? (c < 8031
                    ? c == 8029
                    : c <= 8061)
                  : (c <= 8116 || (c < 8126
                    ? (c >= 8118 && c <= 8124)
                    : c <= 8126)))
                : (c <= 8132 || (c < 8150
                  ? (c < 8144
                    ? (c >= 8134 && c <= 8140)
                    : c <= 8147)
                  : (c <= 8155 || (c < 8178
                    ? (c >= 8160 && c <= 8172)
                    : c <= 8180)))))
              : (c <= 8188 || (c < 8400
                ? (c < 8305
                  ? (c < 8276
                    ? (c >= 8255 && c <= 8256)
                    : c <= 8276)
                  : (c <= 8305 || (c < 8336
                    ? c == 8319
                    : c <= 8348)))
                : (c <= 8412 || (c < 8450
                  ? (c < 8421
                    ? c == 8417
                    : c <= 8432)
                  : (c <= 8450 || (c < 8458
                    ? c == 8455
                    : c <= 8467)))))))
            : (c <= 8469 || (c < 11520
              ? (c < 8508
                ? (c < 8486
                  ? (c < 8484
                    ? (c >= 8472 && c <= 8477)
                    : c <= 8484)
                  : (c <= 8486 || (c < 8490
                    ? c == 8488
                    : c <= 8505)))
                : (c <= 8511 || (c < 8544
                  ? (c < 8526
                    ? (c >= 8517 && c <= 8521)
                    : c <= 8526)
                  : (c <= 8584 || (c < 11499
                    ? (c >= 11264 && c <= 11492)
                    : c <= 11507)))))
              : (c <= 11557 || (c < 11680
                ? (c < 11568
                  ? (c < 11565
                    ? c == 11559
                    : c <= 11565)
                  : (c <= 11623 || (c < 11647
                    ? c == 11631
                    : c <= 11670)))
                : (c <= 11686 || (c < 11704
                  ? (c < 11696
                    ? (c >= 11688 && c <= 11694)
                    : c <= 11702)
                  : (c <= 11710 || (c < 11720
                    ? (c >= 11712 && c <= 11718)
                    : c <= 11726)))))))))
          : (c <= 11734 || (c < 42775
            ? (c < 12549
              ? (c < 12344
                ? (c < 12293
                  ? (c < 11744
                    ? (c >= 11736 && c <= 11742)
                    : c <= 11775)
                  : (c <= 12295 || (c < 12337
                    ? (c >= 12321 && c <= 12335)
                    : c <= 12341)))
                : (c <= 12348 || (c < 12445
                  ? (c < 12441
                    ? (c >= 12353 && c <= 12438)
                    : c <= 12442)
                  : (c <= 12447 || (c < 12540
                    ? (c >= 12449 && c <= 12538)
                    : c <= 12543)))))
              : (c <= 12591 || (c < 42192
                ? (c < 12784
                  ? (c < 12704
                    ? (c >= 12593 && c <= 12686)
                    : c <= 12735)
                  : (c <= 12799 || (c < 19968
                    ? (c >= 13312 && c <= 19903)
                    : c <= 42124)))
                : (c <= 42237 || (c < 42560
                  ? (c < 42512
                    ? (c >= 42240 && c <= 42508)
                    : c <= 42539)
                  : (c <= 42607 || (c < 42623
                    ? (c >= 42612 && c <= 42621)
                    : c <= 42737)))))))
            : (c <= 42783 || (c < 43259
              ? (c < 42994
                ? (c < 42960
                  ? (c < 42891
                    ? (c >= 42786 && c <= 42888)
                    : c <= 42954)
                  : (c <= 42961 || (c < 42965
                    ? c == 42963
                    : c <= 42969)))
                : (c <= 43047 || (c < 43136
                  ? (c < 43072
                    ? c == 43052
                    : c <= 43123)
                  : (c <= 43205 || (c < 43232
                    ? (c >= 43216 && c <= 43225)
                    : c <= 43255)))))
              : (c <= 43259 || (c < 43488
                ? (c < 43360
                  ? (c < 43312
                    ? (c >= 43261 && c <= 43309)
                    : c <= 43347)
                  : (c <= 43388 || (c < 43471
                    ? (c >= 43392 && c <= 43456)
                    : c <= 43481)))
                : (c <= 43518 || (c < 43600
                  ? (c < 43584
                    ? (c >= 43520 && c <= 43574)
                    : c <= 43597)
                  : (c <= 43609 || (c >= 43616 && c <= 43638)))))))))))))))
    : (c <= 43714 || (c < 71472
      ? (c < 67644
        ? (c < 65382
          ? (c < 64318
            ? (c < 44012
              ? (c < 43793
                ? (c < 43762
                  ? (c < 43744
                    ? (c >= 43739 && c <= 43741)
                    : c <= 43759)
                  : (c <= 43766 || (c < 43785
                    ? (c >= 43777 && c <= 43782)
                    : c <= 43790)))
                : (c <= 43798 || (c < 43824
                  ? (c < 43816
                    ? (c >= 43808 && c <= 43814)
                    : c <= 43822)
                  : (c <= 43866 || (c < 43888
                    ? (c >= 43868 && c <= 43881)
                    : c <= 44010)))))
              : (c <= 44013 || (c < 64112
                ? (c < 55216
                  ? (c < 44032
                    ? (c >= 44016 && c <= 44025)
                    : c <= 55203)
                  : (c <= 55238 || (c < 63744
                    ? (c >= 55243 && c <= 55291)
                    : c <= 64109)))
                : (c <= 64217 || (c < 64285
                  ? (c < 64275
                    ? (c >= 64256 && c <= 64262)
                    : c <= 64279)
                  : (c <= 64296 || (c < 64312
                    ? (c >= 64298 && c <= 64310)
                    : c <= 64316)))))))
            : (c <= 64318 || (c < 65101
              ? (c < 64848
                ? (c < 64326
                  ? (c < 64323
                    ? (c >= 64320 && c <= 64321)
                    : c <= 64324)
                  : (c <= 64433 || (c < 64612
                    ? (c >= 64467 && c <= 64605)
                    : c <= 64829)))
                : (c <= 64911 || (c < 65024
                  ? (c < 65008
                    ? (c >= 64914 && c <= 64967)
                    : c <= 65017)
                  : (c <= 65039 || (c < 65075
                    ? (c >= 65056 && c <= 65071)
                    : c <= 65076)))))
              : (c <= 65103 || (c < 65149
                ? (c < 65143
                  ? (c < 65139
                    ? c == 65137
                    : c <= 65139)
                  : (c <= 65143 || (c < 65147
                    ? c == 65145
                    : c <= 65147)))
                : (c <= 65149 || (c < 65313
                  ? (c < 65296
                    ? (c >= 65151 && c <= 65276)
                    : c <= 65305)
                  : (c <= 65338 || (c < 65345
                    ? c == 65343
                    : c <= 65370)))))))))
          : (c <= 65470 || (c < 66560
            ? (c < 65856
              ? (c < 65549
                ? (c < 65490
                  ? (c < 65482
                    ? (c >= 65474 && c <= 65479)
                    : c <= 65487)
                  : (c <= 65495 || (c < 65536
                    ? (c >= 65498 && c <= 65500)
                    : c <= 65547)))
                : (c <= 65574 || (c < 65599
                  ? (c < 65596
                    ? (c >= 65576 && c <= 65594)
                    : c <= 65597)
                  : (c <= 65613 || (c < 65664
                    ? (c >= 65616 && c <= 65629)
                    : c <= 65786)))))
              : (c <= 65908 || (c < 66349
                ? (c < 66208
                  ? (c < 66176
                    ? c == 66045
                    : c <= 66204)
                  : (c <= 66256 || (c < 66304
                    ? c == 66272
                    : c <= 66335)))
                : (c <= 66378 || (c < 66464
                  ? (c < 66432
                    ? (c >= 66384 && c <= 66426)
                    : c <= 66461)
                  : (c <= 66499 || (c < 66513
                    ? (c >= 66504 && c <= 66511)
                    : c <= 66517)))))))
            : (c <= 66717 || (c < 66995
              ? (c < 66928
                ? (c < 66776
                  ? (c < 66736
                    ? (c >= 66720 && c <= 66729)
                    : c <= 66771)
                  : (c <= 66811 || (c < 66864
                    ? (c >= 66816 && c <= 66855)
                    : c <= 66915)))
                : (c <= 66938 || (c < 66964
                  ? (c < 66956
                    ? (c >= 66940 && c <= 66954)
                    : c <= 66962)
                  : (c <= 66965 || (c < 66979
                    ? (c >= 66967 && c <= 66977)
                    : c <= 66993)))))
              : (c <= 67001 || (c < 67463
                ? (c < 67392
                  ? (c < 67072
                    ? (c >= 67003 && c <= 67004)
                    : c <= 67382)
                  : (c <= 67413 || (c < 67456
                    ? (c >= 67424 && c <= 67431)
                    : c <= 67461)))
                : (c <= 67504 || (c < 67592
                  ? (c < 67584
                    ? (c >= 67506 && c <= 67514)
                    : c <= 67589)
                  : (c <= 67592 || (c < 67639
                    ? (c >= 67594 && c <= 67637)
                    : c <= 67640)))))))))))
        : (c <= 67644 || (c < 69968
          ? (c < 68480
            ? (c < 68108
              ? (c < 67840
                ? (c < 67712
                  ? (c < 67680
                    ? (c >= 67647 && c <= 67669)
                    : c <= 67702)
                  : (c <= 67742 || (c < 67828
                    ? (c >= 67808 && c <= 67826)
                    : c <= 67829)))
                : (c <= 67861 || (c < 68030
                  ? (c < 67968
                    ? (c >= 67872 && c <= 67897)
                    : c <= 68023)
                  : (c <= 68031 || (c < 68101
                    ? (c >= 68096 && c <= 68099)
                    : c <= 68102)))))
              : (c <= 68115 || (c < 68224
                ? (c < 68152
                  ? (c < 68121
                    ? (c >= 68117 && c <= 68119)
                    : c <= 68149)
                  : (c <= 68154 || (c < 68192
                    ? c == 68159
                    : c <= 68220)))
                : (c <= 68252 || (c < 68352
                  ? (c < 68297
                    ? (c >= 68288 && c <= 68295)
                    : c <= 68326)
                  : (c <= 68405 || (c < 68448
                    ? (c >= 68416 && c <= 68437)
                    : c <= 68466)))))))
            : (c <= 68497 || (c < 69488
              ? (c < 69248
                ? (c < 68800
                  ? (c < 68736
                    ? (c >= 68608 && c <= 68680)
                    : c <= 68786)
                  : (c <= 68850 || (c < 68912
                    ? (c >= 68864 && c <= 68903)
                    : c <= 68921)))
                : (c <= 69289 || (c < 69376
                  ? (c < 69296
                    ? (c >= 69291 && c <= 69292)
                    : c <= 69297)
                  : (c <= 69404 || (c < 69424
                    ? c == 69415
                    : c <= 69456)))))
              : (c <= 69509 || (c < 69826
                ? (c < 69632
                  ? (c < 69600
                    ? (c >= 69552 && c <= 69572)
                    : c <= 69622)
                  : (c <= 69702 || (c < 69759
                    ? (c >= 69734 && c <= 69749)
                    : c <= 69818)))
                : (c <= 69826 || (c < 69888
                  ? (c < 69872
                    ? (c >= 69840 && c <= 69864)
                    : c <= 69881)
                  : (c <= 69940 || (c < 69956
                    ? (c >= 69942 && c <= 69951)
                    : c <= 69959)))))))))
          : (c <= 70003 || (c < 70471
            ? (c < 70287
              ? (c < 70144
                ? (c < 70089
                  ? (c < 70016
                    ? c == 70006
                    : c <= 70084)
                  : (c <= 70092 || (c < 70108
                    ? (c >= 70094 && c <= 70106)
                    : c <= 70108)))
                : (c <= 70161 || (c < 70272
                  ? (c < 70206
                    ? (c >= 70163 && c <= 70199)
                    : c <= 70206)
                  : (c <= 70278 || (c < 70282
                    ? c == 70280
                    : c <= 70285)))))
              : (c <= 70301 || (c < 70415
                ? (c < 70384
                  ? (c < 70320
                    ? (c >= 70303 && c <= 70312)
                    : c <= 70378)
                  : (c <= 70393 || (c < 70405
                    ? (c >= 70400 && c <= 70403)
                    : c <= 70412)))
                : (c <= 70416 || (c < 70450
                  ? (c < 70442
                    ? (c >= 70419 && c <= 70440)
                    : c <= 70448)
                  : (c <= 70451 || (c < 70459
                    ? (c >= 70453 && c <= 70457)
                    : c <= 70468)))))))
            : (c <= 70472 || (c < 70864
              ? (c < 70512
                ? (c < 70487
                  ? (c < 70480
                    ? (c >= 70475 && c <= 70477)
                    : c <= 70480)
                  : (c <= 70487 || (c < 70502
                    ? (c >= 70493 && c <= 70499)
                    : c <= 70508)))
                : (c <= 70516 || (c < 70750
                  ? (c < 70736
                    ? (c >= 70656 && c <= 70730)
                    : c <= 70745)
                  : (c <= 70753 || (c < 70855
                    ? (c >= 70784 && c <= 70853)
                    : c <= 70855)))))
              : (c <= 70873 || (c < 71248
                ? (c < 71128
                  ? (c < 71096
                    ? (c >= 71040 && c <= 71093)
                    : c <= 71104)
                  : (c <= 71133 || (c < 71236
                    ? (c >= 71168 && c <= 71232)
                    : c <= 71236)))
                : (c <= 71257 || (c < 71424
                  ? (c < 71360
                    ? (c >= 71296 && c <= 71352)
                    : c <= 71369)
                  : (c <= 71450 || (c >= 71453 && c <= 71467)))))))))))))
      : (c <= 71481 || (c < 119973
        ? (c < 82944
          ? (c < 72784
            ? (c < 72096
              ? (c < 71948
                ? (c < 71840
                  ? (c < 71680
                    ? (c >= 71488 && c <= 71494)
                    : c <= 71738)
                  : (c <= 71913 || (c < 71945
                    ? (c >= 71935 && c <= 71942)
                    : c <= 71945)))
                : (c <= 71955 || (c < 71991
                  ? (c < 71960
                    ? (c >= 71957 && c <= 71958)
                    : c <= 71989)
                  : (c <= 71992 || (c < 72016
                    ? (c >= 71995 && c <= 72003)
                    : c <= 72025)))))
              : (c <= 72103 || (c < 72272
                ? (c < 72163
                  ? (c < 72154
                    ? (c >= 72106 && c <= 72151)
                    : c <= 72161)
                  : (c <= 72164 || (c < 72263
                    ? (c >= 72192 && c <= 72254)
                    : c <= 72263)))
                : (c <= 72345 || (c < 72704
                  ? (c < 72368
                    ? c == 72349
                    : c <= 72440)
                  : (c <= 72712 || (c < 72760
                    ? (c >= 72714 && c <= 72758)
                    : c <= 72768)))))))
            : (c <= 72793 || (c < 73063
              ? (c < 72971
                ? (c < 72873
                  ? (c < 72850
                    ? (c >= 72818 && c <= 72847)
                    : c <= 72871)
                  : (c <= 72886 || (c < 72968
                    ? (c >= 72960 && c <= 72966)
                    : c <= 72969)))
                : (c <= 73014 || (c < 73023
                  ? (c < 73020
                    ? c == 73018
                    : c <= 73021)
                  : (c <= 73031 || (c < 73056
                    ? (c >= 73040 && c <= 73049)
                    : c <= 73061)))))
              : (c <= 73064 || (c < 73648
                ? (c < 73107
                  ? (c < 73104
                    ? (c >= 73066 && c <= 73102)
                    : c <= 73105)
                  : (c <= 73112 || (c < 73440
                    ? (c >= 73120 && c <= 73129)
                    : c <= 73462)))
                : (c <= 73648 || (c < 74880
                  ? (c < 74752
                    ? (c >= 73728 && c <= 74649)
                    : c <= 74862)
                  : (c <= 75075 || (c < 77824
                    ? (c >= 77712 && c <= 77808)
                    : c <= 78894)))))))))
          : (c <= 83526 || (c < 110581
            ? (c < 93053
              ? (c < 92880
                ? (c < 92768
                  ? (c < 92736
                    ? (c >= 92160 && c <= 92728)
                    : c <= 92766)
                  : (c <= 92777 || (c < 92864
                    ? (c >= 92784 && c <= 92862)
                    : c <= 92873)))
                : (c <= 92909 || (c < 92992
                  ? (c < 92928
                    ? (c >= 92912 && c <= 92916)
                    : c <= 92982)
                  : (c <= 92995 || (c < 93027
                    ? (c >= 93008 && c <= 93017)
                    : c <= 93047)))))
              : (c <= 93071 || (c < 94179
                ? (c < 94031
                  ? (c < 93952
                    ? (c >= 93760 && c <= 93823)
                    : c <= 94026)
                  : (c <= 94087 || (c < 94176
                    ? (c >= 94095 && c <= 94111)
                    : c <= 94177)))
                : (c <= 94180 || (c < 100352
                  ? (c < 94208
                    ? (c >= 94192 && c <= 94193)
                    : c <= 100343)
                  : (c <= 101589 || (c < 110576
                    ? (c >= 101632 && c <= 101640)
                    : c <= 110579)))))))
            : (c <= 110587 || (c < 118576
              ? (c < 113664
                ? (c < 110928
                  ? (c < 110592
                    ? (c >= 110589 && c <= 110590)
                    : c <= 110882)
                  : (c <= 110930 || (c < 110960
                    ? (c >= 110948 && c <= 110951)
                    : c <= 111355)))
                : (c <= 113770 || (c < 113808
                  ? (c < 113792
                    ? (c >= 113776 && c <= 113788)
                    : c <= 113800)
                  : (c <= 113817 || (c < 118528
                    ? (c >= 113821 && c <= 113822)
                    : c <= 118573)))))
              : (c <= 118598 || (c < 119362
                ? (c < 119163
                  ? (c < 119149
                    ? (c >= 119141 && c <= 119145)
                    : c <= 119154)
                  : (c <= 119170 || (c < 119210
                    ? (c >= 119173 && c <= 119179)
                    : c <= 119213)))
                : (c <= 119364 || (c < 119966
                  ? (c < 119894
                    ? (c >= 119808 && c <= 119892)
                    : c <= 119964)
                  : (c <= 119967 || c == 119970))))))))))
        : (c <= 119974 || (c < 124912
          ? (c < 120746
            ? (c < 120134
              ? (c < 120071
                ? (c < 119995
                  ? (c < 119982
                    ? (c >= 119977 && c <= 119980)
                    : c <= 119993)
                  : (c <= 119995 || (c < 120005
                    ? (c >= 119997 && c <= 120003)
                    : c <= 120069)))
                : (c <= 120074 || (c < 120094
                  ? (c < 120086
                    ? (c >= 120077 && c <= 120084)
                    : c <= 120092)
                  : (c <= 120121 || (c < 120128
                    ? (c >= 120123 && c <= 120126)
                    : c <= 120132)))))
              : (c <= 120134 || (c < 120572
                ? (c < 120488
                  ? (c < 120146
                    ? (c >= 120138 && c <= 120144)
                    : c <= 120485)
                  : (c <= 120512 || (c < 120540
                    ? (c >= 120514 && c <= 120538)
                    : c <= 120570)))
                : (c <= 120596 || (c < 120656
                  ? (c < 120630
                    ? (c >= 120598 && c <= 120628)
                    : c <= 120654)
                  : (c <= 120686 || (c < 120714
                    ? (c >= 120688 && c <= 120712)
                    : c <= 120744)))))))
            : (c <= 120770 || (c < 122907
              ? (c < 121476
                ? (c < 121344
                  ? (c < 120782
                    ? (c >= 120772 && c <= 120779)
                    : c <= 120831)
                  : (c <= 121398 || (c < 121461
                    ? (c >= 121403 && c <= 121452)
                    : c <= 121461)))
                : (c <= 121476 || (c < 122624
                  ? (c < 121505
                    ? (c >= 121499 && c <= 121503)
                    : c <= 121519)
                  : (c <= 122654 || (c < 122888
                    ? (c >= 122880 && c <= 122886)
                    : c <= 122904)))))
              : (c <= 122913 || (c < 123214
                ? (c < 123136
                  ? (c < 122918
                    ? (c >= 122915 && c <= 122916)
                    : c <= 122922)
                  : (c <= 123180 || (c < 123200
                    ? (c >= 123184 && c <= 123197)
                    : c <= 123209)))
                : (c <= 123214 || (c < 124896
                  ? (c < 123584
                    ? (c >= 123536 && c <= 123566)
                    : c <= 123641)
                  : (c <= 124902 || (c < 124909
                    ? (c >= 124904 && c <= 124907)
                    : c <= 124910)))))))))
          : (c <= 124926 || (c < 126557
            ? (c < 126521
              ? (c < 126469
                ? (c < 125184
                  ? (c < 125136
                    ? (c >= 124928 && c <= 125124)
                    : c <= 125142)
                  : (c <= 125259 || (c < 126464
                    ? (c >= 125264 && c <= 125273)
                    : c <= 126467)))
                : (c <= 126495 || (c < 126503
                  ? (c < 126500
                    ? (c >= 126497 && c <= 126498)
                    : c <= 126500)
                  : (c <= 126503 || (c < 126516
                    ? (c >= 126505 && c <= 126514)
                    : c <= 126519)))))
              : (c <= 126521 || (c < 126541
                ? (c < 126535
                  ? (c < 126530
                    ? c == 126523
                    : c <= 126530)
                  : (c <= 126535 || (c < 126539
                    ? c == 126537
                    : c <= 126539)))
                : (c <= 126543 || (c < 126551
                  ? (c < 126548
                    ? (c >= 126545 && c <= 126546)
                    : c <= 126548)
                  : (c <= 126551 || (c < 126555
                    ? c == 126553
                    : c <= 126555)))))))
            : (c <= 126557 || (c < 126629
              ? (c < 126580
                ? (c < 126564
                  ? (c < 126561
                    ? c == 126559
                    : c <= 126562)
                  : (c <= 126564 || (c < 126572
                    ? (c >= 126567 && c <= 126570)
                    : c <= 126578)))
                : (c <= 126583 || (c < 126592
                  ? (c < 126590
                    ? (c >= 126585 && c <= 126588)
                    : c <= 126590)
                  : (c <= 126601 || (c < 126625
                    ? (c >= 126603 && c <= 126619)
                    : c <= 126627)))))
              : (c <= 126633 || (c < 178208
                ? (c < 131072
                  ? (c < 130032
                    ? (c >= 126635 && c <= 126651)
                    : c <= 130041)
                  : (c <= 173791 || (c < 177984
                    ? (c >= 173824 && c <= 177976)
                    : c <= 178205)))
                : (c <= 183969 || (c < 196608
                  ? (c < 194560
                    ? (c >= 183984 && c <= 191456)
                    : c <= 195101)
                  : (c <= 201546 || (c >= 917760 && c <= 917999)))))))))))))))));
}

static inline bool sym_identifier_character_set_5(int32_t c) {
  return (c < 43616
    ? (c < 3782
      ? (c < 2748
        ? (c < 2045
          ? (c < 1015
            ? (c < 710
              ? (c < 181
                ? (c < '_'
                  ? (c < 'A'
                    ? (c >= '0' && c <= '9')
                    : c <= 'Z')
                  : (c <= '_' || (c < 170
                    ? (c >= 'a' && c <= 'z')
                    : c <= 170)))
                : (c <= 181 || (c < 192
                  ? (c < 186
                    ? c == 183
                    : c <= 186)
                  : (c <= 214 || (c < 248
                    ? (c >= 216 && c <= 246)
                    : c <= 705)))))
              : (c <= 721 || (c < 891
                ? (c < 750
                  ? (c < 748
                    ? (c >= 736 && c <= 740)
                    : c <= 748)
                  : (c <= 750 || (c < 886
                    ? (c >= 768 && c <= 884)
                    : c <= 887)))
                : (c <= 893 || (c < 908
                  ? (c < 902
                    ? c == 895
                    : c <= 906)
                  : (c <= 908 || (c < 931
                    ? (c >= 910 && c <= 929)
                    : c <= 1013)))))))
            : (c <= 1153 || (c < 1519
              ? (c < 1425
                ? (c < 1329
                  ? (c < 1162
                    ? (c >= 1155 && c <= 1159)
                    : c <= 1327)
                  : (c <= 1366 || (c < 1376
                    ? c == 1369
                    : c <= 1416)))
                : (c <= 1469 || (c < 1476
                  ? (c < 1473
                    ? c == 1471
                    : c <= 1474)
                  : (c <= 1477 || (c < 1488
                    ? c == 1479
                    : c <= 1514)))))
              : (c <= 1522 || (c < 1770
                ? (c < 1646
                  ? (c < 1568
                    ? (c >= 1552 && c <= 1562)
                    : c <= 1641)
                  : (c <= 1747 || (c < 1759
                    ? (c >= 1749 && c <= 1756)
                    : c <= 1768)))
                : (c <= 1788 || (c < 1869
                  ? (c < 1808
                    ? c == 1791
                    : c <= 1866)
                  : (c <= 1969 || (c < 2042
                    ? (c >= 1984 && c <= 2037)
                    : c <= 2042)))))))))
          : (c <= 2045 || (c < 2558
            ? (c < 2451
              ? (c < 2200
                ? (c < 2144
                  ? (c < 2112
                    ? (c >= 2048 && c <= 2093)
                    : c <= 2139)
                  : (c <= 2154 || (c < 2185
                    ? (c >= 2160 && c <= 2183)
                    : c <= 2190)))
                : (c <= 2273 || (c < 2417
                  ? (c < 2406
                    ? (c >= 2275 && c <= 2403)
                    : c <= 2415)
                  : (c <= 2435 || (c < 2447
                    ? (c >= 2437 && c <= 2444)
                    : c <= 2448)))))
              : (c <= 2472 || (c < 2507
                ? (c < 2486
                  ? (c < 2482
                    ? (c >= 2474 && c <= 2480)
                    : c <= 2482)
                  : (c <= 2489 || (c < 2503
                    ? (c >= 2492 && c <= 2500)
                    : c <= 2504)))
                : (c <= 2510 || (c < 2527
                  ? (c < 2524
                    ? c == 2519
                    : c <= 2525)
                  : (c <= 2531 || (c < 2556
                    ? (c >= 2534 && c <= 2545)
                    : c <= 2556)))))))
            : (c <= 2558 || (c < 2635
              ? (c < 2610
                ? (c < 2575
                  ? (c < 2565
                    ? (c >= 2561 && c <= 2563)
                    : c <= 2570)
                  : (c <= 2576 || (c < 2602
                    ? (c >= 2579 && c <= 2600)
                    : c <= 2608)))
                : (c <= 2611 || (c < 2620
                  ? (c < 2616
                    ? (c >= 2613 && c <= 2614)
                    : c <= 2617)
                  : (c <= 2620 || (c < 2631
                    ? (c >= 2622 && c <= 2626)
                    : c <= 2632)))))
              : (c <= 2637 || (c < 2693
                ? (c < 2654
                  ? (c < 2649
                    ? c == 2641
                    : c <= 2652)
                  : (c <= 2654 || (c < 2689
                    ? (c >= 2662 && c <= 2677)
                    : c <= 2691)))
                : (c <= 2701 || (c < 2730
                  ? (c < 2707
                    ? (c >= 2703 && c <= 2705)
                    : c <= 2728)
                  : (c <= 2736 || (c < 2741
                    ? (c >= 2738 && c <= 2739)
                    : c <= 2745)))))))))))
        : (c <= 2757 || (c < 3168
          ? (c < 2958
            ? (c < 2866
              ? (c < 2809
                ? (c < 2768
                  ? (c < 2763
                    ? (c >= 2759 && c <= 2761)
                    : c <= 2765)
                  : (c <= 2768 || (c < 2790
                    ? (c >= 2784 && c <= 2787)
                    : c <= 2799)))
                : (c <= 2815 || (c < 2831
                  ? (c < 2821
                    ? (c >= 2817 && c <= 2819)
                    : c <= 2828)
                  : (c <= 2832 || (c < 2858
                    ? (c >= 2835 && c <= 2856)
                    : c <= 2864)))))
              : (c <= 2867 || (c < 2908
                ? (c < 2887
                  ? (c < 2876
                    ? (c >= 2869 && c <= 2873)
                    : c <= 2884)
                  : (c <= 2888 || (c < 2901
                    ? (c >= 2891 && c <= 2893)
                    : c <= 2903)))
                : (c <= 2909 || (c < 2929
                  ? (c < 2918
                    ? (c >= 2911 && c <= 2915)
                    : c <= 2927)
                  : (c <= 2929 || (c < 2949
                    ? (c >= 2946 && c <= 2947)
                    : c <= 2954)))))))
            : (c <= 2960 || (c < 3031
              ? (c < 2984
                ? (c < 2972
                  ? (c < 2969
                    ? (c >= 2962 && c <= 2965)
                    : c <= 2970)
                  : (c <= 2972 || (c < 2979
                    ? (c >= 2974 && c <= 2975)
                    : c <= 2980)))
                : (c <= 2986 || (c < 3014
                  ? (c < 3006
                    ? (c >= 2990 && c <= 3001)
                    : c <= 3010)
                  : (c <= 3016 || (c < 3024
                    ? (c >= 3018 && c <= 3021)
                    : c <= 3024)))))
              : (c <= 3031 || (c < 3132
                ? (c < 3086
                  ? (c < 3072
                    ? (c >= 3046 && c <= 3055)
                    : c <= 3084)
                  : (c <= 3088 || (c < 3114
                    ? (c >= 3090 && c <= 3112)
                    : c <= 3129)))
                : (c <= 3140 || (c < 3157
                  ? (c < 3146
                    ? (c >= 3142 && c <= 3144)
                    : c <= 3149)
                  : (c <= 3158 || (c < 3165
                    ? (c >= 3160 && c <= 3162)
                    : c <= 3165)))))))))
          : (c <= 3171 || (c < 3450
            ? (c < 3293
              ? (c < 3242
                ? (c < 3205
                  ? (c < 3200
                    ? (c >= 3174 && c <= 3183)
                    : c <= 3203)
                  : (c <= 3212 || (c < 3218
                    ? (c >= 3214 && c <= 3216)
                    : c <= 3240)))
                : (c <= 3251 || (c < 3270
                  ? (c < 3260
                    ? (c >= 3253 && c <= 3257)
                    : c <= 3268)
                  : (c <= 3272 || (c < 3285
                    ? (c >= 3274 && c <= 3277)
                    : c <= 3286)))))
              : (c <= 3294 || (c < 3346
                ? (c < 3313
                  ? (c < 3302
                    ? (c >= 3296 && c <= 3299)
                    : c <= 3311)
                  : (c <= 3314 || (c < 3342
                    ? (c >= 3328 && c <= 3340)
                    : c <= 3344)))
                : (c <= 3396 || (c < 3412
                  ? (c < 3402
                    ? (c >= 3398 && c <= 3400)
                    : c <= 3406)
                  : (c <= 3415 || (c < 3430
                    ? (c >= 3423 && c <= 3427)
                    : c <= 3439)))))))
            : (c <= 3455 || (c < 3570
              ? (c < 3520
                ? (c < 3482
                  ? (c < 3461
                    ? (c >= 3457 && c <= 3459)
                    : c <= 3478)
                  : (c <= 3505 || (c < 3517
                    ? (c >= 3507 && c <= 3515)
                    : c <= 3517)))
                : (c <= 3526 || (c < 3542
                  ? (c < 3535
                    ? c == 3530
                    : c <= 3540)
                  : (c <= 3542 || (c < 3558
                    ? (c >= 3544 && c <= 3551)
                    : c <= 3567)))))
              : (c <= 3571 || (c < 3718
                ? (c < 3664
                  ? (c < 3648
                    ? (c >= 3585 && c <= 3642)
                    : c <= 3662)
                  : (c <= 3673 || (c < 3716
                    ? (c >= 3713 && c <= 3714)
                    : c <= 3716)))
                : (c <= 3722 || (c < 3751
                  ? (c < 3749
                    ? (c >= 3724 && c <= 3747)
                    : c <= 3749)
                  : (c <= 3773 || (c >= 3776 && c <= 3780)))))))))))))
      : (c <= 3782 || (c < 8025
        ? (c < 5888
          ? (c < 4688
            ? (c < 3953
              ? (c < 3872
                ? (c < 3804
                  ? (c < 3792
                    ? (c >= 3784 && c <= 3789)
                    : c <= 3801)
                  : (c <= 3807 || (c < 3864
                    ? c == 3840
                    : c <= 3865)))
                : (c <= 3881 || (c < 3897
                  ? (c < 3895
                    ? c == 3893
                    : c <= 3895)
                  : (c <= 3897 || (c < 3913
                    ? (c >= 3902 && c <= 3911)
                    : c <= 3948)))))
              : (c <= 3972 || (c < 4256
                ? (c < 4038
                  ? (c < 3993
                    ? (c >= 3974 && c <= 3991)
                    : c <= 4028)
                  : (c <= 4038 || (c < 4176
                    ? (c >= 4096 && c <= 4169)
                    : c <= 4253)))
                : (c <= 4293 || (c < 4304
                  ? (c < 4301
                    ? c == 4295
                    : c <= 4301)
                  : (c <= 4346 || (c < 4682
                    ? (c >= 4348 && c <= 4680)
                    : c <= 4685)))))))
            : (c <= 4694 || (c < 4882
              ? (c < 4786
                ? (c < 4704
                  ? (c < 4698
                    ? c == 4696
                    : c <= 4701)
                  : (c <= 4744 || (c < 4752
                    ? (c >= 4746 && c <= 4749)
                    : c <= 4784)))
                : (c <= 4789 || (c < 4802
                  ? (c < 4800
                    ? (c >= 4792 && c <= 4798)
                    : c <= 4800)
                  : (c <= 4805 || (c < 4824
                    ? (c >= 4808 && c <= 4822)
                    : c <= 4880)))))
              : (c <= 4885 || (c < 5112
                ? (c < 4969
                  ? (c < 4957
                    ? (c >= 4888 && c <= 4954)
                    : c <= 4959)
                  : (c <= 4977 || (c < 5024
                    ? (c >= 4992 && c <= 5007)
                    : c <= 5109)))
                : (c <= 5117 || (c < 5761
                  ? (c < 5743
                    ? (c >= 5121 && c <= 5740)
                    : c <= 5759)
                  : (c <= 5786 || (c < 5870
                    ? (c >= 5792 && c <= 5866)
                    : c <= 5880)))))))))
          : (c <= 5909 || (c < 6688
            ? (c < 6176
              ? (c < 6016
                ? (c < 5984
                  ? (c < 5952
                    ? (c >= 5919 && c <= 5940)
                    : c <= 5971)
                  : (c <= 5996 || (c < 6002
                    ? (c >= 5998 && c <= 6000)
                    : c <= 6003)))
                : (c <= 6099 || (c < 6112
                  ? (c < 6108
                    ? c == 6103
                    : c <= 6109)
                  : (c <= 6121 || (c < 6159
                    ? (c >= 6155 && c <= 6157)
                    : c <= 6169)))))
              : (c <= 6264 || (c < 6470
                ? (c < 6400
                  ? (c < 6320
                    ? (c >= 6272 && c <= 6314)
                    : c <= 6389)
                  : (c <= 6430 || (c < 6448
                    ? (c >= 6432 && c <= 6443)
                    : c <= 6459)))
                : (c <= 6509 || (c < 6576
                  ? (c < 6528
                    ? (c >= 6512 && c <= 6516)
                    : c <= 6571)
                  : (c <= 6601 || (c < 6656
                    ? (c >= 6608 && c <= 6618)
                    : c <= 6683)))))))
            : (c <= 6750 || (c < 7232
              ? (c < 6847
                ? (c < 6800
                  ? (c < 6783
                    ? (c >= 6752 && c <= 6780)
                    : c <= 6793)
                  : (c <= 6809 || (c < 6832
                    ? c == 6823
                    : c <= 6845)))
                : (c <= 6862 || (c < 7019
                  ? (c < 6992
                    ? (c >= 6912 && c <= 6988)
                    : c <= 7001)
                  : (c <= 7027 || (c < 7168
                    ? (c >= 7040 && c <= 7155)
                    : c <= 7223)))))
              : (c <= 7241 || (c < 7380
                ? (c < 7312
                  ? (c < 7296
                    ? (c >= 7245 && c <= 7293)
                    : c <= 7304)
                  : (c <= 7354 || (c < 7376
                    ? (c >= 7357 && c <= 7359)
                    : c <= 7378)))
                : (c <= 7418 || (c < 7968
                  ? (c < 7960
                    ? (c >= 7424 && c <= 7957)
                    : c <= 7965)
                  : (c <= 8005 || (c < 8016
                    ? (c >= 8008 && c <= 8013)
                    : c <= 8023)))))))))))
        : (c <= 8025 || (c < 11720
          ? (c < 8458
            ? (c < 8178
              ? (c < 8126
                ? (c < 8031
                  ? (c < 8029
                    ? c == 8027
                    : c <= 8029)
                  : (c <= 8061 || (c < 8118
                    ? (c >= 8064 && c <= 8116)
                    : c <= 8124)))
                : (c <= 8126 || (c < 8144
                  ? (c < 8134
                    ? (c >= 8130 && c <= 8132)
                    : c <= 8140)
                  : (c <= 8147 || (c < 8160
                    ? (c >= 8150 && c <= 8155)
                    : c <= 8172)))))
              : (c <= 8180 || (c < 8336
                ? (c < 8276
                  ? (c < 8255
                    ? (c >= 8182 && c <= 8188)
                    : c <= 8256)
                  : (c <= 8276 || (c < 8319
                    ? c == 8305
                    : c <= 8319)))
                : (c <= 8348 || (c < 8421
                  ? (c < 8417
                    ? (c >= 8400 && c <= 8412)
                    : c <= 8417)
                  : (c <= 8432 || (c < 8455
                    ? c == 8450
                    : c <= 8455)))))))
            : (c <= 8467 || (c < 11499
              ? (c < 8490
                ? (c < 8484
                  ? (c < 8472
                    ? c == 8469
                    : c <= 8477)
                  : (c <= 8484 || (c < 8488
                    ? c == 8486
                    : c <= 8488)))
                : (c <= 8505 || (c < 8526
                  ? (c < 8517
                    ? (c >= 8508 && c <= 8511)
                    : c <= 8521)
                  : (c <= 8526 || (c < 11264
                    ? (c >= 8544 && c <= 8584)
                    : c <= 11492)))))
              : (c <= 11507 || (c < 11647
                ? (c < 11565
                  ? (c < 11559
                    ? (c >= 11520 && c <= 11557)
                    : c <= 11559)
                  : (c <= 11565 || (c < 11631
                    ? (c >= 11568 && c <= 11623)
                    : c <= 11631)))
                : (c <= 11670 || (c < 11696
                  ? (c < 11688
                    ? (c >= 11680 && c <= 11686)
                    : c <= 11694)
                  : (c <= 11702 || (c < 11712
                    ? (c >= 11704 && c <= 11710)
                    : c <= 11718)))))))))
          : (c <= 11726 || (c < 42623
            ? (c < 12540
              ? (c < 12337
                ? (c < 11744
                  ? (c < 11736
                    ? (c >= 11728 && c <= 11734)
                    : c <= 11742)
                  : (c <= 11775 || (c < 12321
                    ? (c >= 12293 && c <= 12295)
                    : c <= 12335)))
                : (c <= 12341 || (c < 12441
                  ? (c < 12353
                    ? (c >= 12344 && c <= 12348)
                    : c <= 12438)
                  : (c <= 12442 || (c < 12449
                    ? (c >= 12445 && c <= 12447)
                    : c <= 12538)))))
              : (c <= 12543 || (c < 19968
                ? (c < 12704
                  ? (c < 12593
                    ? (c >= 12549 && c <= 12591)
                    : c <= 12686)
                  : (c <= 12735 || (c < 13312
                    ? (c >= 12784 && c <= 12799)
                    : c <= 19903)))
                : (c <= 42124 || (c < 42512
                  ? (c < 42240
                    ? (c >= 42192 && c <= 42237)
                    : c <= 42508)
                  : (c <= 42539 || (c < 42612
                    ? (c >= 42560 && c <= 42607)
                    : c <= 42621)))))))
            : (c <= 42737 || (c < 43232
              ? (c < 42965
                ? (c < 42891
                  ? (c < 42786
                    ? (c >= 42775 && c <= 42783)
                    : c <= 42888)
                  : (c <= 42954 || (c < 42963
                    ? (c >= 42960 && c <= 42961)
                    : c <= 42963)))
                : (c <= 42969 || (c < 43072
                  ? (c < 43052
                    ? (c >= 42994 && c <= 43047)
                    : c <= 43052)
                  : (c <= 43123 || (c < 43216
                    ? (c >= 43136 && c <= 43205)
                    : c <= 43225)))))
              : (c <= 43255 || (c < 43471
                ? (c < 43312
                  ? (c < 43261
                    ? c == 43259
                    : c <= 43309)
                  : (c <= 43347 || (c < 43392
                    ? (c >= 43360 && c <= 43388)
                    : c <= 43456)))
                : (c <= 43481 || (c < 43584
                  ? (c < 43520
                    ? (c >= 43488 && c <= 43518)
                    : c <= 43574)
                  : (c <= 43597 || (c >= 43600 && c <= 43609)))))))))))))))
    : (c <= 43638 || (c < 71453
      ? (c < 67639
        ? (c < 65345
          ? (c < 64312
            ? (c < 43888
              ? (c < 43785
                ? (c < 43744
                  ? (c < 43739
                    ? (c >= 43642 && c <= 43714)
                    : c <= 43741)
                  : (c <= 43759 || (c < 43777
                    ? (c >= 43762 && c <= 43766)
                    : c <= 43782)))
                : (c <= 43790 || (c < 43816
                  ? (c < 43808
                    ? (c >= 43793 && c <= 43798)
                    : c <= 43814)
                  : (c <= 43822 || (c < 43868
                    ? (c >= 43824 && c <= 43866)
                    : c <= 43881)))))
              : (c <= 44010 || (c < 63744
                ? (c < 44032
                  ? (c < 44016
                    ? (c >= 44012 && c <= 44013)
                    : c <= 44025)
                  : (c <= 55203 || (c < 55243
                    ? (c >= 55216 && c <= 55238)
                    : c <= 55291)))
                : (c <= 64109 || (c < 64275
                  ? (c < 64256
                    ? (c >= 64112 && c <= 64217)
                    : c <= 64262)
                  : (c <= 64279 || (c < 64298
                    ? (c >= 64285 && c <= 64296)
                    : c <= 64310)))))))
            : (c <= 64316 || (c < 65075
              ? (c < 64612
                ? (c < 64323
                  ? (c < 64320
                    ? c == 64318
                    : c <= 64321)
                  : (c <= 64324 || (c < 64467
                    ? (c >= 64326 && c <= 64433)
                    : c <= 64605)))
                : (c <= 64829 || (c < 65008
                  ? (c < 64914
                    ? (c >= 64848 && c <= 64911)
                    : c <= 64967)
                  : (c <= 65017 || (c < 65056
                    ? (c >= 65024 && c <= 65039)
                    : c <= 65071)))))
              : (c <= 65076 || (c < 65147
                ? (c < 65139
                  ? (c < 65137
                    ? (c >= 65101 && c <= 65103)
                    : c <= 65137)
                  : (c <= 65139 || (c < 65145
                    ? c == 65143
                    : c <= 65145)))
                : (c <= 65147 || (c < 65296
                  ? (c < 65151
                    ? c == 65149
                    : c <= 65276)
                  : (c <= 65305 || (c < 65343
                    ? (c >= 65313 && c <= 65338)
                    : c <= 65343)))))))))
          : (c <= 65370 || (c < 66513
            ? (c < 65664
              ? (c < 65536
                ? (c < 65482
                  ? (c < 65474
                    ? (c >= 65382 && c <= 65470)
                    : c <= 65479)
                  : (c <= 65487 || (c < 65498
                    ? (c >= 65490 && c <= 65495)
                    : c <= 65500)))
                : (c <= 65547 || (c < 65596
                  ? (c < 65576
                    ? (c >= 65549 && c <= 65574)
                    : c <= 65594)
                  : (c <= 65597 || (c < 65616
                    ? (c >= 65599 && c <= 65613)
                    : c <= 65629)))))
              : (c <= 65786 || (c < 66304
                ? (c < 66176
                  ? (c < 66045
                    ? (c >= 65856 && c <= 65908)
                    : c <= 66045)
                  : (c <= 66204 || (c < 66272
                    ? (c >= 66208 && c <= 66256)
                    : c <= 66272)))
                : (c <= 66335 || (c < 66432
                  ? (c < 66384
                    ? (c >= 66349 && c <= 66378)
                    : c <= 66426)
                  : (c <= 66461 || (c < 66504
                    ? (c >= 66464 && c <= 66499)
                    : c <= 66511)))))))
            : (c <= 66517 || (c < 66979
              ? (c < 66864
                ? (c < 66736
                  ? (c < 66720
                    ? (c >= 66560 && c <= 66717)
                    : c <= 66729)
                  : (c <= 66771 || (c < 66816
                    ? (c >= 66776 && c <= 66811)
                    : c <= 66855)))
                : (c <= 66915 || (c < 66956
                  ? (c < 66940
                    ? (c >= 66928 && c <= 66938)
                    : c <= 66954)
                  : (c <= 66962 || (c < 66967
                    ? (c >= 66964 && c <= 66965)
                    : c <= 66977)))))
              : (c <= 66993 || (c < 67456
                ? (c < 67072
                  ? (c < 67003
                    ? (c >= 66995 && c <= 67001)
                    : c <= 67004)
                  : (c <= 67382 || (c < 67424
                    ? (c >= 67392 && c <= 67413)
                    : c <= 67431)))
                : (c <= 67461 || (c < 67584
                  ? (c < 67506
                    ? (c >= 67463 && c <= 67504)
                    : c <= 67514)
                  : (c <= 67589 || (c < 67594
                    ? c == 67592
                    : c <= 67637)))))))))))
        : (c <= 67640 || (c < 69956
          ? (c < 68448
            ? (c < 68101
              ? (c < 67828
                ? (c < 67680
                  ? (c < 67647
                    ? c == 67644
                    : c <= 67669)
                  : (c <= 67702 || (c < 67808
                    ? (c >= 67712 && c <= 67742)
                    : c <= 67826)))
                : (c <= 67829 || (c < 67968
                  ? (c < 67872
                    ? (c >= 67840 && c <= 67861)
                    : c <= 67897)
                  : (c <= 68023 || (c < 68096
                    ? (c >= 68030 && c <= 68031)
                    : c <= 68099)))))
              : (c <= 68102 || (c < 68192
                ? (c < 68121
                  ? (c < 68117
                    ? (c >= 68108 && c <= 68115)
                    : c <= 68119)
                  : (c <= 68149 || (c < 68159
                    ? (c >= 68152 && c <= 68154)
                    : c <= 68159)))
                : (c <= 68220 || (c < 68297
                  ? (c < 68288
                    ? (c >= 68224 && c <= 68252)
                    : c <= 68295)
                  : (c <= 68326 || (c < 68416
                    ? (c >= 68352 && c <= 68405)
                    : c <= 68437)))))))
            : (c <= 68466 || (c < 69424
              ? (c < 68912
                ? (c < 68736
                  ? (c < 68608
                    ? (c >= 68480 && c <= 68497)
                    : c <= 68680)
                  : (c <= 68786 || (c < 68864
                    ? (c >= 68800 && c <= 68850)
                    : c <= 68903)))
                : (c <= 68921 || (c < 69296
                  ? (c < 69291
                    ? (c >= 69248 && c <= 69289)
                    : c <= 69292)
                  : (c <= 69297 || (c < 69415
                    ? (c >= 69376 && c <= 69404)
                    : c <= 69415)))))
              : (c <= 69456 || (c < 69759
                ? (c < 69600
                  ? (c < 69552
                    ? (c >= 69488 && c <= 69509)
                    : c <= 69572)
                  : (c <= 69622 || (c < 69734
                    ? (c >= 69632 && c <= 69702)
                    : c <= 69749)))
                : (c <= 69818 || (c < 69872
                  ? (c < 69840
                    ? c == 69826
                    : c <= 69864)
                  : (c <= 69881 || (c < 69942
                    ? (c >= 69888 && c <= 69940)
                    : c <= 69951)))))))))
          : (c <= 69959 || (c < 70459
            ? (c < 70282
              ? (c < 70108
                ? (c < 70016
                  ? (c < 70006
                    ? (c >= 69968 && c <= 70003)
                    : c <= 70006)
                  : (c <= 70084 || (c < 70094
                    ? (c >= 70089 && c <= 70092)
                    : c <= 70106)))
                : (c <= 70108 || (c < 70206
                  ? (c < 70163
                    ? (c >= 70144 && c <= 70161)
                    : c <= 70199)
                  : (c <= 70206 || (c < 70280
                    ? (c >= 70272 && c <= 70278)
                    : c <= 70280)))))
              : (c <= 70285 || (c < 70405
                ? (c < 70320
                  ? (c < 70303
                    ? (c >= 70287 && c <= 70301)
                    : c <= 70312)
                  : (c <= 70378 || (c < 70400
                    ? (c >= 70384 && c <= 70393)
                    : c <= 70403)))
                : (c <= 70412 || (c < 70442
                  ? (c < 70419
                    ? (c >= 70415 && c <= 70416)
                    : c <= 70440)
                  : (c <= 70448 || (c < 70453
                    ? (c >= 70450 && c <= 70451)
                    : c <= 70457)))))))
            : (c <= 70468 || (c < 70855
              ? (c < 70502
                ? (c < 70480
                  ? (c < 70475
                    ? (c >= 70471 && c <= 70472)
                    : c <= 70477)
                  : (c <= 70480 || (c < 70493
                    ? c == 70487
                    : c <= 70499)))
                : (c <= 70508 || (c < 70736
                  ? (c < 70656
                    ? (c >= 70512 && c <= 70516)
                    : c <= 70730)
                  : (c <= 70745 || (c < 70784
                    ? (c >= 70750 && c <= 70753)
                    : c <= 70853)))))
              : (c <= 70855 || (c < 71236
                ? (c < 71096
                  ? (c < 71040
                    ? (c >= 70864 && c <= 70873)
                    : c <= 71093)
                  : (c <= 71104 || (c < 71168
                    ? (c >= 71128 && c <= 71133)
                    : c <= 71232)))
                : (c <= 71236 || (c < 71360
                  ? (c < 71296
                    ? (c >= 71248 && c <= 71257)
                    : c <= 71352)
                  : (c <= 71369 || (c >= 71424 && c <= 71450)))))))))))))
      : (c <= 71467 || (c < 119973
        ? (c < 77824
          ? (c < 72760
            ? (c < 72016
              ? (c < 71945
                ? (c < 71680
                  ? (c < 71488
                    ? (c >= 71472 && c <= 71481)
                    : c <= 71494)
                  : (c <= 71738 || (c < 71935
                    ? (c >= 71840 && c <= 71913)
                    : c <= 71942)))
                : (c <= 71945 || (c < 71960
                  ? (c < 71957
                    ? (c >= 71948 && c <= 71955)
                    : c <= 71958)
                  : (c <= 71989 || (c < 71995
                    ? (c >= 71991 && c <= 71992)
                    : c <= 72003)))))
              : (c <= 72025 || (c < 72263
                ? (c < 72154
                  ? (c < 72106
                    ? (c >= 72096 && c <= 72103)
                    : c <= 72151)
                  : (c <= 72161 || (c < 72192
                    ? (c >= 72163 && c <= 72164)
                    : c <= 72254)))
                : (c <= 72263 || (c < 72368
                  ? (c < 72349
                    ? (c >= 72272 && c <= 72345)
                    : c <= 72349)
                  : (c <= 72440 || (c < 72714
                    ? (c >= 72704 && c <= 72712)
                    : c <= 72758)))))))
            : (c <= 72768 || (c < 73056
              ? (c < 72968
                ? (c < 72850
                  ? (c < 72818
                    ? (c >= 72784 && c <= 72793)
                    : c <= 72847)
                  : (c <= 72871 || (c < 72960
                    ? (c >= 72873 && c <= 72886)
                    : c <= 72966)))
                : (c <= 72969 || (c < 73020
                  ? (c < 73018
                    ? (c >= 72971 && c <= 73014)
                    : c <= 73018)
                  : (c <= 73021 || (c < 73040
                    ? (c >= 73023 && c <= 73031)
                    : c <= 73049)))))
              : (c <= 73061 || (c < 73440
                ? (c < 73104
                  ? (c < 73066
                    ? (c >= 73063 && c <= 73064)
                    : c <= 73102)
                  : (c <= 73105 || (c < 73120
                    ? (c >= 73107 && c <= 73112)
                    : c <= 73129)))
                : (c <= 73462 || (c < 74752
                  ? (c < 73728
                    ? c == 73648
                    : c <= 74649)
                  : (c <= 74862 || (c < 77712
                    ? (c >= 74880 && c <= 75075)
                    : c <= 77808)))))))))
          : (c <= 78894 || (c < 110576
            ? (c < 93027
              ? (c < 92864
                ? (c < 92736
                  ? (c < 92160
                    ? (c >= 82944 && c <= 83526)
                    : c <= 92728)
                  : (c <= 92766 || (c < 92784
                    ? (c >= 92768 && c <= 92777)
                    : c <= 92862)))
                : (c <= 92873 || (c < 92928
                  ? (c < 92912
                    ? (c >= 92880 && c <= 92909)
                    : c <= 92916)
                  : (c <= 92982 || (c < 93008
                    ? (c >= 92992 && c <= 92995)
                    : c <= 93017)))))
              : (c <= 93047 || (c < 94176
                ? (c < 93952
                  ? (c < 93760
                    ? (c >= 93053 && c <= 93071)
                    : c <= 93823)
                  : (c <= 94026 || (c < 94095
                    ? (c >= 94031 && c <= 94087)
                    : c <= 94111)))
                : (c <= 94177 || (c < 94208
                  ? (c < 94192
                    ? (c >= 94179 && c <= 94180)
                    : c <= 94193)
                  : (c <= 100343 || (c < 101632
                    ? (c >= 100352 && c <= 101589)
                    : c <= 101640)))))))
            : (c <= 110579 || (c < 118528
              ? (c < 110960
                ? (c < 110592
                  ? (c < 110589
                    ? (c >= 110581 && c <= 110587)
                    : c <= 110590)
                  : (c <= 110882 || (c < 110948
                    ? (c >= 110928 && c <= 110930)
                    : c <= 110951)))
                : (c <= 111355 || (c < 113792
                  ? (c < 113776
                    ? (c >= 113664 && c <= 113770)
                    : c <= 113788)
                  : (c <= 113800 || (c < 113821
                    ? (c >= 113808 && c <= 113817)
                    : c <= 113822)))))
              : (c <= 118573 || (c < 119210
                ? (c < 119149
                  ? (c < 119141
                    ? (c >= 118576 && c <= 118598)
                    : c <= 119145)
                  : (c <= 119154 || (c < 119173
                    ? (c >= 119163 && c <= 119170)
                    : c <= 119179)))
                : (c <= 119213 || (c < 119894
                  ? (c < 119808
                    ? (c >= 119362 && c <= 119364)
                    : c <= 119892)
                  : (c <= 119964 || (c < 119970
                    ? (c >= 119966 && c <= 119967)
                    : c <= 119970)))))))))))
        : (c <= 119974 || (c < 124912
          ? (c < 120746
            ? (c < 120134
              ? (c < 120071
                ? (c < 119995
                  ? (c < 119982
                    ? (c >= 119977 && c <= 119980)
                    : c <= 119993)
                  : (c <= 119995 || (c < 120005
                    ? (c >= 119997 && c <= 120003)
                    : c <= 120069)))
                : (c <= 120074 || (c < 120094
                  ? (c < 120086
                    ? (c >= 120077 && c <= 120084)
                    : c <= 120092)
                  : (c <= 120121 || (c < 120128
                    ? (c >= 120123 && c <= 120126)
                    : c <= 120132)))))
              : (c <= 120134 || (c < 120572
                ? (c < 120488
                  ? (c < 120146
                    ? (c >= 120138 && c <= 120144)
                    : c <= 120485)
                  : (c <= 120512 || (c < 120540
                    ? (c >= 120514 && c <= 120538)
                    : c <= 120570)))
                : (c <= 120596 || (c < 120656
                  ? (c < 120630
                    ? (c >= 120598 && c <= 120628)
                    : c <= 120654)
                  : (c <= 120686 || (c < 120714
                    ? (c >= 120688 && c <= 120712)
                    : c <= 120744)))))))
            : (c <= 120770 || (c < 122907
              ? (c < 121476
                ? (c < 121344
                  ? (c < 120782
                    ? (c >= 120772 && c <= 120779)
                    : c <= 120831)
                  : (c <= 121398 || (c < 121461
                    ? (c >= 121403 && c <= 121452)
                    : c <= 121461)))
                : (c <= 121476 || (c < 122624
                  ? (c < 121505
                    ? (c >= 121499 && c <= 121503)
                    : c <= 121519)
                  : (c <= 122654 || (c < 122888
                    ? (c >= 122880 && c <= 122886)
                    : c <= 122904)))))
              : (c <= 122913 || (c < 123214
                ? (c < 123136
                  ? (c < 122918
                    ? (c >= 122915 && c <= 122916)
                    : c <= 122922)
                  : (c <= 123180 || (c < 123200
                    ? (c >= 123184 && c <= 123197)
                    : c <= 123209)))
                : (c <= 123214 || (c < 124896
                  ? (c < 123584
                    ? (c >= 123536 && c <= 123566)
                    : c <= 123641)
                  : (c <= 124902 || (c < 124909
                    ? (c >= 124904 && c <= 124907)
                    : c <= 124910)))))))))
          : (c <= 124926 || (c < 126557
            ? (c < 126521
              ? (c < 126469
                ? (c < 125184
                  ? (c < 125136
                    ? (c >= 124928 && c <= 125124)
                    : c <= 125142)
                  : (c <= 125259 || (c < 126464
                    ? (c >= 125264 && c <= 125273)
                    : c <= 126467)))
                : (c <= 126495 || (c < 126503
                  ? (c < 126500
                    ? (c >= 126497 && c <= 126498)
                    : c <= 126500)
                  : (c <= 126503 || (c < 126516
                    ? (c >= 126505 && c <= 126514)
                    : c <= 126519)))))
              : (c <= 126521 || (c < 126541
                ? (c < 126535
                  ? (c < 126530
                    ? c == 126523
                    : c <= 126530)
                  : (c <= 126535 || (c < 126539
                    ? c == 126537
                    : c <= 126539)))
                : (c <= 126543 || (c < 126551
                  ? (c < 126548
                    ? (c >= 126545 && c <= 126546)
                    : c <= 126548)
                  : (c <= 126551 || (c < 126555
                    ? c == 126553
                    : c <= 126555)))))))
            : (c <= 126557 || (c < 126629
              ? (c < 126580
                ? (c < 126564
                  ? (c < 126561
                    ? c == 126559
                    : c <= 126562)
                  : (c <= 126564 || (c < 126572
                    ? (c >= 126567 && c <= 126570)
                    : c <= 126578)))
                : (c <= 126583 || (c < 126592
                  ? (c < 126590
                    ? (c >= 126585 && c <= 126588)
                    : c <= 126590)
                  : (c <= 126601 || (c < 126625
                    ? (c >= 126603 && c <= 126619)
                    : c <= 126627)))))
              : (c <= 126633 || (c < 178208
                ? (c < 131072
                  ? (c < 130032
                    ? (c >= 126635 && c <= 126651)
                    : c <= 130041)
                  : (c <= 173791 || (c < 177984
                    ? (c >= 173824 && c <= 177976)
                    : c <= 178205)))
                : (c <= 183969 || (c < 196608
                  ? (c < 194560
                    ? (c >= 183984 && c <= 191456)
                    : c <= 195101)
                  : (c <= 201546 || (c >= 917760 && c <= 917999)))))))))))))))));
}

static inline bool sym_identifier_character_set_6(int32_t c) {
  return (c < 43642
    ? (c < 3784
      ? (c < 2759
        ? (c < 2048
          ? (c < 1155
            ? (c < 736
              ? (c < 183
                ? (c < 'a'
                  ? (c < 'A'
                    ? (c >= '0' && c <= '9')
                    : c <= 'Z')
                  : (c <= 'z' || (c < 181
                    ? c == 170
                    : c <= 181)))
                : (c <= 183 || (c < 216
                  ? (c < 192
                    ? c == 186
                    : c <= 214)
                  : (c <= 246 || (c < 710
                    ? (c >= 248 && c <= 705)
                    : c <= 721)))))
              : (c <= 740 || (c < 895
                ? (c < 768
                  ? (c < 750
                    ? c == 748
                    : c <= 750)
                  : (c <= 884 || (c < 891
                    ? (c >= 886 && c <= 887)
                    : c <= 893)))
                : (c <= 895 || (c < 910
                  ? (c < 908
                    ? (c >= 902 && c <= 906)
                    : c <= 908)
                  : (c <= 929 || (c < 1015
                    ? (c >= 931 && c <= 1013)
                    : c <= 1153)))))))
            : (c <= 1159 || (c < 1552
              ? (c < 1471
                ? (c < 1369
                  ? (c < 1329
                    ? (c >= 1162 && c <= 1327)
                    : c <= 1366)
                  : (c <= 1369 || (c < 1425
                    ? (c >= 1376 && c <= 1416)
                    : c <= 1469)))
                : (c <= 1471 || (c < 1479
                  ? (c < 1476
                    ? (c >= 1473 && c <= 1474)
                    : c <= 1477)
                  : (c <= 1479 || (c < 1519
                    ? (c >= 1488 && c <= 1514)
                    : c <= 1522)))))
              : (c <= 1562 || (c < 1791
                ? (c < 1749
                  ? (c < 1646
                    ? (c >= 1568 && c <= 1641)
                    : c <= 1747)
                  : (c <= 1756 || (c < 1770
                    ? (c >= 1759 && c <= 1768)
                    : c <= 1788)))
                : (c <= 1791 || (c < 1984
                  ? (c < 1869
                    ? (c >= 1808 && c <= 1866)
                    : c <= 1969)
                  : (c <= 2037 || (c < 2045
                    ? c == 2042
                    : c <= 2045)))))))))
          : (c <= 2093 || (c < 2561
            ? (c < 2474
              ? (c < 2275
                ? (c < 2160
                  ? (c < 2144
                    ? (c >= 2112 && c <= 2139)
                    : c <= 2154)
                  : (c <= 2183 || (c < 2200
                    ? (c >= 2185 && c <= 2190)
                    : c <= 2273)))
                : (c <= 2403 || (c < 2437
                  ? (c < 2417
                    ? (c >= 2406 && c <= 2415)
                    : c <= 2435)
                  : (c <= 2444 || (c < 2451
                    ? (c >= 2447 && c <= 2448)
                    : c <= 2472)))))
              : (c <= 2480 || (c < 2519
                ? (c < 2492
                  ? (c < 2486
                    ? c == 2482
                    : c <= 2489)
                  : (c <= 2500 || (c < 2507
                    ? (c >= 2503 && c <= 2504)
                    : c <= 2510)))
                : (c <= 2519 || (c < 2534
                  ? (c < 2527
                    ? (c >= 2524 && c <= 2525)
                    : c <= 2531)
                  : (c <= 2545 || (c < 2558
                    ? c == 2556
                    : c <= 2558)))))))
            : (c <= 2563 || (c < 2641
              ? (c < 2613
                ? (c < 2579
                  ? (c < 2575
                    ? (c >= 2565 && c <= 2570)
                    : c <= 2576)
                  : (c <= 2600 || (c < 2610
                    ? (c >= 2602 && c <= 2608)
                    : c <= 2611)))
                : (c <= 2614 || (c < 2622
                  ? (c < 2620
                    ? (c >= 2616 && c <= 2617)
                    : c <= 2620)
                  : (c <= 2626 || (c < 2635
                    ? (c >= 2631 && c <= 2632)
                    : c <= 2637)))))
              : (c <= 2641 || (c < 2703
                ? (c < 2662
                  ? (c < 2654
                    ? (c >= 2649 && c <= 2652)
                    : c <= 2654)
                  : (c <= 2677 || (c < 2693
                    ? (c >= 2689 && c <= 2691)
                    : c <= 2701)))
                : (c <= 2705 || (c < 2738
                  ? (c < 2730
                    ? (c >= 2707 && c <= 2728)
                    : c <= 2736)
                  : (c <= 2739 || (c < 2748
                    ? (c >= 2741 && c <= 2745)
                    : c <= 2757)))))))))))
        : (c <= 2761 || (c < 3174
          ? (c < 2962
            ? (c < 2869
              ? (c < 2817
                ? (c < 2784
                  ? (c < 2768
                    ? (c >= 2763 && c <= 2765)
                    : c <= 2768)
                  : (c <= 2787 || (c < 2809
                    ? (c >= 2790 && c <= 2799)
                    : c <= 2815)))
                : (c <= 2819 || (c < 2835
                  ? (c < 2831
                    ? (c >= 2821 && c <= 2828)
                    : c <= 2832)
                  : (c <= 2856 || (c < 2866
                    ? (c >= 2858 && c <= 2864)
                    : c <= 2867)))))
              : (c <= 2873 || (c < 2911
                ? (c < 2891
                  ? (c < 2887
                    ? (c >= 2876 && c <= 2884)
                    : c <= 2888)
                  : (c <= 2893 || (c < 2908
                    ? (c >= 2901 && c <= 2903)
                    : c <= 2909)))
                : (c <= 2915 || (c < 2946
                  ? (c < 2929
                    ? (c >= 2918 && c <= 2927)
                    : c <= 2929)
                  : (c <= 2947 || (c < 2958
                    ? (c >= 2949 && c <= 2954)
                    : c <= 2960)))))))
            : (c <= 2965 || (c < 3046
              ? (c < 2990
                ? (c < 2974
                  ? (c < 2972
                    ? (c >= 2969 && c <= 2970)
                    : c <= 2972)
                  : (c <= 2975 || (c < 2984
                    ? (c >= 2979 && c <= 2980)
                    : c <= 2986)))
                : (c <= 3001 || (c < 3018
                  ? (c < 3014
                    ? (c >= 3006 && c <= 3010)
                    : c <= 3016)
                  : (c <= 3021 || (c < 3031
                    ? c == 3024
                    : c <= 3031)))))
              : (c <= 3055 || (c < 3142
                ? (c < 3090
                  ? (c < 3086
                    ? (c >= 3072 && c <= 3084)
                    : c <= 3088)
                  : (c <= 3112 || (c < 3132
                    ? (c >= 3114 && c <= 3129)
                    : c <= 3140)))
                : (c <= 3144 || (c < 3160
                  ? (c < 3157
                    ? (c >= 3146 && c <= 3149)
                    : c <= 3158)
                  : (c <= 3162 || (c < 3168
                    ? c == 3165
                    : c <= 3171)))))))))
          : (c <= 3183 || (c < 3457
            ? (c < 3296
              ? (c < 3253
                ? (c < 3214
                  ? (c < 3205
                    ? (c >= 3200 && c <= 3203)
                    : c <= 3212)
                  : (c <= 3216 || (c < 3242
                    ? (c >= 3218 && c <= 3240)
                    : c <= 3251)))
                : (c <= 3257 || (c < 3274
                  ? (c < 3270
                    ? (c >= 3260 && c <= 3268)
                    : c <= 3272)
                  : (c <= 3277 || (c < 3293
                    ? (c >= 3285 && c <= 3286)
                    : c <= 3294)))))
              : (c <= 3299 || (c < 3398
                ? (c < 3328
                  ? (c < 3313
                    ? (c >= 3302 && c <= 3311)
                    : c <= 3314)
                  : (c <= 3340 || (c < 3346
                    ? (c >= 3342 && c <= 3344)
                    : c <= 3396)))
                : (c <= 3400 || (c < 3423
                  ? (c < 3412
                    ? (c >= 3402 && c <= 3406)
                    : c <= 3415)
                  : (c <= 3427 || (c < 3450
                    ? (c >= 3430 && c <= 3439)
                    : c <= 3455)))))))
            : (c <= 3459 || (c < 3585
              ? (c < 3530
                ? (c < 3507
                  ? (c < 3482
                    ? (c >= 3461 && c <= 3478)
                    : c <= 3505)
                  : (c <= 3515 || (c < 3520
                    ? c == 3517
                    : c <= 3526)))
                : (c <= 3530 || (c < 3544
                  ? (c < 3542
                    ? (c >= 3535 && c <= 3540)
                    : c <= 3542)
                  : (c <= 3551 || (c < 3570
                    ? (c >= 3558 && c <= 3567)
                    : c <= 3571)))))
              : (c <= 3642 || (c < 3724
                ? (c < 3713
                  ? (c < 3664
                    ? (c >= 3648 && c <= 3662)
                    : c <= 3673)
                  : (c <= 3714 || (c < 3718
                    ? c == 3716
                    : c <= 3722)))
                : (c <= 3747 || (c < 3776
                  ? (c < 3751
                    ? c == 3749
                    : c <= 3773)
                  : (c <= 3780 || c == 3782))))))))))))
      : (c <= 3789 || (c < 8027
        ? (c < 5919
          ? (c < 4696
            ? (c < 3974
              ? (c < 3893
                ? (c < 3840
                  ? (c < 3804
                    ? (c >= 3792 && c <= 3801)
                    : c <= 3807)
                  : (c <= 3840 || (c < 3872
                    ? (c >= 3864 && c <= 3865)
                    : c <= 3881)))
                : (c <= 3893 || (c < 3902
                  ? (c < 3897
                    ? c == 3895
                    : c <= 3897)
                  : (c <= 3911 || (c < 3953
                    ? (c >= 3913 && c <= 3948)
                    : c <= 3972)))))
              : (c <= 3991 || (c < 4295
                ? (c < 4096
                  ? (c < 4038
                    ? (c >= 3993 && c <= 4028)
                    : c <= 4038)
                  : (c <= 4169 || (c < 4256
                    ? (c >= 4176 && c <= 4253)
                    : c <= 4293)))
                : (c <= 4295 || (c < 4348
                  ? (c < 4304
                    ? c == 4301
                    : c <= 4346)
                  : (c <= 4680 || (c < 4688
                    ? (c >= 4682 && c <= 4685)
                    : c <= 4694)))))))
            : (c <= 4696 || (c < 4888
              ? (c < 4792
                ? (c < 4746
                  ? (c < 4704
                    ? (c >= 4698 && c <= 4701)
                    : c <= 4744)
                  : (c <= 4749 || (c < 4786
                    ? (c >= 4752 && c <= 4784)
                    : c <= 4789)))
                : (c <= 4798 || (c < 4808
                  ? (c < 4802
                    ? c == 4800
                    : c <= 4805)
                  : (c <= 4822 || (c < 4882
                    ? (c >= 4824 && c <= 4880)
                    : c <= 4885)))))
              : (c <= 4954 || (c < 5121
                ? (c < 4992
                  ? (c < 4969
                    ? (c >= 4957 && c <= 4959)
                    : c <= 4977)
                  : (c <= 5007 || (c < 5112
                    ? (c >= 5024 && c <= 5109)
                    : c <= 5117)))
                : (c <= 5740 || (c < 5792
                  ? (c < 5761
                    ? (c >= 5743 && c <= 5759)
                    : c <= 5786)
                  : (c <= 5866 || (c < 5888
                    ? (c >= 5870 && c <= 5880)
                    : c <= 5909)))))))))
          : (c <= 5940 || (c < 6752
            ? (c < 6272
              ? (c < 6103
                ? (c < 5998
                  ? (c < 5984
                    ? (c >= 5952 && c <= 5971)
                    : c <= 5996)
                  : (c <= 6000 || (c < 6016
                    ? (c >= 6002 && c <= 6003)
                    : c <= 6099)))
                : (c <= 6103 || (c < 6155
                  ? (c < 6112
                    ? (c >= 6108 && c <= 6109)
                    : c <= 6121)
                  : (c <= 6157 || (c < 6176
                    ? (c >= 6159 && c <= 6169)
                    : c <= 6264)))))
              : (c <= 6314 || (c < 6512
                ? (c < 6432
                  ? (c < 6400
                    ? (c >= 6320 && c <= 6389)
                    : c <= 6430)
                  : (c <= 6443 || (c < 6470
                    ? (c >= 6448 && c <= 6459)
                    : c <= 6509)))
                : (c <= 6516 || (c < 6608
                  ? (c < 6576
                    ? (c >= 6528 && c <= 6571)
                    : c <= 6601)
                  : (c <= 6618 || (c < 6688
                    ? (c >= 6656 && c <= 6683)
                    : c <= 6750)))))))
            : (c <= 6780 || (c < 7245
              ? (c < 6912
                ? (c < 6823
                  ? (c < 6800
                    ? (c >= 6783 && c <= 6793)
                    : c <= 6809)
                  : (c <= 6823 || (c < 6847
                    ? (c >= 6832 && c <= 6845)
                    : c <= 6862)))
                : (c <= 6988 || (c < 7040
                  ? (c < 7019
                    ? (c >= 6992 && c <= 7001)
                    : c <= 7027)
                  : (c <= 7155 || (c < 7232
                    ? (c >= 7168 && c <= 7223)
                    : c <= 7241)))))
              : (c <= 7293 || (c < 7424
                ? (c < 7357
                  ? (c < 7312
                    ? (c >= 7296 && c <= 7304)
                    : c <= 7354)
                  : (c <= 7359 || (c < 7380
                    ? (c >= 7376 && c <= 7378)
                    : c <= 7418)))
                : (c <= 7957 || (c < 8008
                  ? (c < 7968
                    ? (c >= 7960 && c <= 7965)
                    : c <= 8005)
                  : (c <= 8013 || (c < 8025
                    ? (c >= 8016 && c <= 8023)
                    : c <= 8025)))))))))))
        : (c <= 8027 || (c < 11728
          ? (c < 8469
            ? (c < 8182
              ? (c < 8130
                ? (c < 8064
                  ? (c < 8031
                    ? c == 8029
                    : c <= 8061)
                  : (c <= 8116 || (c < 8126
                    ? (c >= 8118 && c <= 8124)
                    : c <= 8126)))
                : (c <= 8132 || (c < 8150
                  ? (c < 8144
                    ? (c >= 8134 && c <= 8140)
                    : c <= 8147)
                  : (c <= 8155 || (c < 8178
                    ? (c >= 8160 && c <= 8172)
                    : c <= 8180)))))
              : (c <= 8188 || (c < 8400
                ? (c < 8305
                  ? (c < 8276
                    ? (c >= 8255 && c <= 8256)
                    : c <= 8276)
                  : (c <= 8305 || (c < 8336
                    ? c == 8319
                    : c <= 8348)))
                : (c <= 8412 || (c < 8450
                  ? (c < 8421
                    ? c == 8417
                    : c <= 8432)
                  : (c <= 8450 || (c < 8458
                    ? c == 8455
                    : c <= 8467)))))))
            : (c <= 8469 || (c < 11520
              ? (c < 8508
                ? (c < 8486
                  ? (c < 8484
                    ? (c >= 8472 && c <= 8477)
                    : c <= 8484)
                  : (c <= 8486 || (c < 8490
                    ? c == 8488
                    : c <= 8505)))
                : (c <= 8511 || (c < 8544
                  ? (c < 8526
                    ? (c >= 8517 && c <= 8521)
                    : c <= 8526)
                  : (c <= 8584 || (c < 11499
                    ? (c >= 11264 && c <= 11492)
                    : c <= 11507)))))
              : (c <= 11557 || (c < 11680
                ? (c < 11568
                  ? (c < 11565
                    ? c == 11559
                    : c <= 11565)
                  : (c <= 11623 || (c < 11647
                    ? c == 11631
                    : c <= 11670)))
                : (c <= 11686 || (c < 11704
                  ? (c < 11696
                    ? (c >= 11688 && c <= 11694)
                    : c <= 11702)
                  : (c <= 11710 || (c < 11720
                    ? (c >= 11712 && c <= 11718)
                    : c <= 11726)))))))))
          : (c <= 11734 || (c < 42775
            ? (c < 12549
              ? (c < 12344
                ? (c < 12293
                  ? (c < 11744
                    ? (c >= 11736 && c <= 11742)
                    : c <= 11775)
                  : (c <= 12295 || (c < 12337
                    ? (c >= 12321 && c <= 12335)
                    : c <= 12341)))
                : (c <= 12348 || (c < 12445
                  ? (c < 12441
                    ? (c >= 12353 && c <= 12438)
                    : c <= 12442)
                  : (c <= 12447 || (c < 12540
                    ? (c >= 12449 && c <= 12538)
                    : c <= 12543)))))
              : (c <= 12591 || (c < 42192
                ? (c < 12784
                  ? (c < 12704
                    ? (c >= 12593 && c <= 12686)
                    : c <= 12735)
                  : (c <= 12799 || (c < 19968
                    ? (c >= 13312 && c <= 19903)
                    : c <= 42124)))
                : (c <= 42237 || (c < 42560
                  ? (c < 42512
                    ? (c >= 42240 && c <= 42508)
                    : c <= 42539)
                  : (c <= 42607 || (c < 42623
                    ? (c >= 42612 && c <= 42621)
                    : c <= 42737)))))))
            : (c <= 42783 || (c < 43259
              ? (c < 42994
                ? (c < 42960
                  ? (c < 42891
                    ? (c >= 42786 && c <= 42888)
                    : c <= 42954)
                  : (c <= 42961 || (c < 42965
                    ? c == 42963
                    : c <= 42969)))
                : (c <= 43047 || (c < 43136
                  ? (c < 43072
                    ? c == 43052
                    : c <= 43123)
                  : (c <= 43205 || (c < 43232
                    ? (c >= 43216 && c <= 43225)
                    : c <= 43255)))))
              : (c <= 43259 || (c < 43488
                ? (c < 43360
                  ? (c < 43312
                    ? (c >= 43261 && c <= 43309)
                    : c <= 43347)
                  : (c <= 43388 || (c < 43471
                    ? (c >= 43392 && c <= 43456)
                    : c <= 43481)))
                : (c <= 43518 || (c < 43600
                  ? (c < 43584
                    ? (c >= 43520 && c <= 43574)
                    : c <= 43597)
                  : (c <= 43609 || (c >= 43616 && c <= 43638)))))))))))))))
    : (c <= 43714 || (c < 71472
      ? (c < 67644
        ? (c < 65382
          ? (c < 64318
            ? (c < 44012
              ? (c < 43793
                ? (c < 43762
                  ? (c < 43744
                    ? (c >= 43739 && c <= 43741)
                    : c <= 43759)
                  : (c <= 43766 || (c < 43785
                    ? (c >= 43777 && c <= 43782)
                    : c <= 43790)))
                : (c <= 43798 || (c < 43824
                  ? (c < 43816
                    ? (c >= 43808 && c <= 43814)
                    : c <= 43822)
                  : (c <= 43866 || (c < 43888
                    ? (c >= 43868 && c <= 43881)
                    : c <= 44010)))))
              : (c <= 44013 || (c < 64112
                ? (c < 55216
                  ? (c < 44032
                    ? (c >= 44016 && c <= 44025)
                    : c <= 55203)
                  : (c <= 55238 || (c < 63744
                    ? (c >= 55243 && c <= 55291)
                    : c <= 64109)))
                : (c <= 64217 || (c < 64285
                  ? (c < 64275
                    ? (c >= 64256 && c <= 64262)
                    : c <= 64279)
                  : (c <= 64296 || (c < 64312
                    ? (c >= 64298 && c <= 64310)
                    : c <= 64316)))))))
            : (c <= 64318 || (c < 65101
              ? (c < 64848
                ? (c < 64326
                  ? (c < 64323
                    ? (c >= 64320 && c <= 64321)
                    : c <= 64324)
                  : (c <= 64433 || (c < 64612
                    ? (c >= 64467 && c <= 64605)
                    : c <= 64829)))
                : (c <= 64911 || (c < 65024
                  ? (c < 65008
                    ? (c >= 64914 && c <= 64967)
                    : c <= 65017)
                  : (c <= 65039 || (c < 65075
                    ? (c >= 65056 && c <= 65071)
                    : c <= 65076)))))
              : (c <= 65103 || (c < 65149
                ? (c < 65143
                  ? (c < 65139
                    ? c == 65137
                    : c <= 65139)
                  : (c <= 65143 || (c < 65147
                    ? c == 65145
                    : c <= 65147)))
                : (c <= 65149 || (c < 65313
                  ? (c < 65296
                    ? (c >= 65151 && c <= 65276)
                    : c <= 65305)
                  : (c <= 65338 || (c < 65345
                    ? c == 65343
                    : c <= 65370)))))))))
          : (c <= 65470 || (c < 66560
            ? (c < 65856
              ? (c < 65549
                ? (c < 65490
                  ? (c < 65482
                    ? (c >= 65474 && c <= 65479)
                    : c <= 65487)
                  : (c <= 65495 || (c < 65536
                    ? (c >= 65498 && c <= 65500)
                    : c <= 65547)))
                : (c <= 65574 || (c < 65599
                  ? (c < 65596
                    ? (c >= 65576 && c <= 65594)
                    : c <= 65597)
                  : (c <= 65613 || (c < 65664
                    ? (c >= 65616 && c <= 65629)
                    : c <= 65786)))))
              : (c <= 65908 || (c < 66349
                ? (c < 66208
                  ? (c < 66176
                    ? c == 66045
                    : c <= 66204)
                  : (c <= 66256 || (c < 66304
                    ? c == 66272
                    : c <= 66335)))
                : (c <= 66378 || (c < 66464
                  ? (c < 66432
                    ? (c >= 66384 && c <= 66426)
                    : c <= 66461)
                  : (c <= 66499 || (c < 66513
                    ? (c >= 66504 && c <= 66511)
                    : c <= 66517)))))))
            : (c <= 66717 || (c < 66995
              ? (c < 66928
                ? (c < 66776
                  ? (c < 66736
                    ? (c >= 66720 && c <= 66729)
                    : c <= 66771)
                  : (c <= 66811 || (c < 66864
                    ? (c >= 66816 && c <= 66855)
                    : c <= 66915)))
                : (c <= 66938 || (c < 66964
                  ? (c < 66956
                    ? (c >= 66940 && c <= 66954)
                    : c <= 66962)
                  : (c <= 66965 || (c < 66979
                    ? (c >= 66967 && c <= 66977)
                    : c <= 66993)))))
              : (c <= 67001 || (c < 67463
                ? (c < 67392
                  ? (c < 67072
                    ? (c >= 67003 && c <= 67004)
                    : c <= 67382)
                  : (c <= 67413 || (c < 67456
                    ? (c >= 67424 && c <= 67431)
                    : c <= 67461)))
                : (c <= 67504 || (c < 67592
                  ? (c < 67584
                    ? (c >= 67506 && c <= 67514)
                    : c <= 67589)
                  : (c <= 67592 || (c < 67639
                    ? (c >= 67594 && c <= 67637)
                    : c <= 67640)))))))))))
        : (c <= 67644 || (c < 69968
          ? (c < 68480
            ? (c < 68108
              ? (c < 67840
                ? (c < 67712
                  ? (c < 67680
                    ? (c >= 67647 && c <= 67669)
                    : c <= 67702)
                  : (c <= 67742 || (c < 67828
                    ? (c >= 67808 && c <= 67826)
                    : c <= 67829)))
                : (c <= 67861 || (c < 68030
                  ? (c < 67968
                    ? (c >= 67872 && c <= 67897)
                    : c <= 68023)
                  : (c <= 68031 || (c < 68101
                    ? (c >= 68096 && c <= 68099)
                    : c <= 68102)))))
              : (c <= 68115 || (c < 68224
                ? (c < 68152
                  ? (c < 68121
                    ? (c >= 68117 && c <= 68119)
                    : c <= 68149)
                  : (c <= 68154 || (c < 68192
                    ? c == 68159
                    : c <= 68220)))
                : (c <= 68252 || (c < 68352
                  ? (c < 68297
                    ? (c >= 68288 && c <= 68295)
                    : c <= 68326)
                  : (c <= 68405 || (c < 68448
                    ? (c >= 68416 && c <= 68437)
                    : c <= 68466)))))))
            : (c <= 68497 || (c < 69488
              ? (c < 69248
                ? (c < 68800
                  ? (c < 68736
                    ? (c >= 68608 && c <= 68680)
                    : c <= 68786)
                  : (c <= 68850 || (c < 68912
                    ? (c >= 68864 && c <= 68903)
                    : c <= 68921)))
                : (c <= 69289 || (c < 69376
                  ? (c < 69296
                    ? (c >= 69291 && c <= 69292)
                    : c <= 69297)
                  : (c <= 69404 || (c < 69424
                    ? c == 69415
                    : c <= 69456)))))
              : (c <= 69509 || (c < 69826
                ? (c < 69632
                  ? (c < 69600
                    ? (c >= 69552 && c <= 69572)
                    : c <= 69622)
                  : (c <= 69702 || (c < 69759
                    ? (c >= 69734 && c <= 69749)
                    : c <= 69818)))
                : (c <= 69826 || (c < 69888
                  ? (c < 69872
                    ? (c >= 69840 && c <= 69864)
                    : c <= 69881)
                  : (c <= 69940 || (c < 69956
                    ? (c >= 69942 && c <= 69951)
                    : c <= 69959)))))))))
          : (c <= 70003 || (c < 70471
            ? (c < 70287
              ? (c < 70144
                ? (c < 70089
                  ? (c < 70016
                    ? c == 70006
                    : c <= 70084)
                  : (c <= 70092 || (c < 70108
                    ? (c >= 70094 && c <= 70106)
                    : c <= 70108)))
                : (c <= 70161 || (c < 70272
                  ? (c < 70206
                    ? (c >= 70163 && c <= 70199)
                    : c <= 70206)
                  : (c <= 70278 || (c < 70282
                    ? c == 70280
                    : c <= 70285)))))
              : (c <= 70301 || (c < 70415
                ? (c < 70384
                  ? (c < 70320
                    ? (c >= 70303 && c <= 70312)
                    : c <= 70378)
                  : (c <= 70393 || (c < 70405
                    ? (c >= 70400 && c <= 70403)
                    : c <= 70412)))
                : (c <= 70416 || (c < 70450
                  ? (c < 70442
                    ? (c >= 70419 && c <= 70440)
                    : c <= 70448)
                  : (c <= 70451 || (c < 70459
                    ? (c >= 70453 && c <= 70457)
                    : c <= 70468)))))))
            : (c <= 70472 || (c < 70864
              ? (c < 70512
                ? (c < 70487
                  ? (c < 70480
                    ? (c >= 70475 && c <= 70477)
                    : c <= 70480)
                  : (c <= 70487 || (c < 70502
                    ? (c >= 70493 && c <= 70499)
                    : c <= 70508)))
                : (c <= 70516 || (c < 70750
                  ? (c < 70736
                    ? (c >= 70656 && c <= 70730)
                    : c <= 70745)
                  : (c <= 70753 || (c < 70855
                    ? (c >= 70784 && c <= 70853)
                    : c <= 70855)))))
              : (c <= 70873 || (c < 71248
                ? (c < 71128
                  ? (c < 71096
                    ? (c >= 71040 && c <= 71093)
                    : c <= 71104)
                  : (c <= 71133 || (c < 71236
                    ? (c >= 71168 && c <= 71232)
                    : c <= 71236)))
                : (c <= 71257 || (c < 71424
                  ? (c < 71360
                    ? (c >= 71296 && c <= 71352)
                    : c <= 71369)
                  : (c <= 71450 || (c >= 71453 && c <= 71467)))))))))))))
      : (c <= 71481 || (c < 119973
        ? (c < 82944
          ? (c < 72784
            ? (c < 72096
              ? (c < 71948
                ? (c < 71840
                  ? (c < 71680
                    ? (c >= 71488 && c <= 71494)
                    : c <= 71738)
                  : (c <= 71913 || (c < 71945
                    ? (c >= 71935 && c <= 71942)
                    : c <= 71945)))
                : (c <= 71955 || (c < 71991
                  ? (c < 71960
                    ? (c >= 71957 && c <= 71958)
                    : c <= 71989)
                  : (c <= 71992 || (c < 72016
                    ? (c >= 71995 && c <= 72003)
                    : c <= 72025)))))
              : (c <= 72103 || (c < 72272
                ? (c < 72163
                  ? (c < 72154
                    ? (c >= 72106 && c <= 72151)
                    : c <= 72161)
                  : (c <= 72164 || (c < 72263
                    ? (c >= 72192 && c <= 72254)
                    : c <= 72263)))
                : (c <= 72345 || (c < 72704
                  ? (c < 72368
                    ? c == 72349
                    : c <= 72440)
                  : (c <= 72712 || (c < 72760
                    ? (c >= 72714 && c <= 72758)
                    : c <= 72768)))))))
            : (c <= 72793 || (c < 73063
              ? (c < 72971
                ? (c < 72873
                  ? (c < 72850
                    ? (c >= 72818 && c <= 72847)
                    : c <= 72871)
                  : (c <= 72886 || (c < 72968
                    ? (c >= 72960 && c <= 72966)
                    : c <= 72969)))
                : (c <= 73014 || (c < 73023
                  ? (c < 73020
                    ? c == 73018
                    : c <= 73021)
                  : (c <= 73031 || (c < 73056
                    ? (c >= 73040 && c <= 73049)
                    : c <= 73061)))))
              : (c <= 73064 || (c < 73648
                ? (c < 73107
                  ? (c < 73104
                    ? (c >= 73066 && c <= 73102)
                    : c <= 73105)
                  : (c <= 73112 || (c < 73440
                    ? (c >= 73120 && c <= 73129)
                    : c <= 73462)))
                : (c <= 73648 || (c < 74880
                  ? (c < 74752
                    ? (c >= 73728 && c <= 74649)
                    : c <= 74862)
                  : (c <= 75075 || (c < 77824
                    ? (c >= 77712 && c <= 77808)
                    : c <= 78894)))))))))
          : (c <= 83526 || (c < 110581
            ? (c < 93053
              ? (c < 92880
                ? (c < 92768
                  ? (c < 92736
                    ? (c >= 92160 && c <= 92728)
                    : c <= 92766)
                  : (c <= 92777 || (c < 92864
                    ? (c >= 92784 && c <= 92862)
                    : c <= 92873)))
                : (c <= 92909 || (c < 92992
                  ? (c < 92928
                    ? (c >= 92912 && c <= 92916)
                    : c <= 92982)
                  : (c <= 92995 || (c < 93027
                    ? (c >= 93008 && c <= 93017)
                    : c <= 93047)))))
              : (c <= 93071 || (c < 94179
                ? (c < 94031
                  ? (c < 93952
                    ? (c >= 93760 && c <= 93823)
                    : c <= 94026)
                  : (c <= 94087 || (c < 94176
                    ? (c >= 94095 && c <= 94111)
                    : c <= 94177)))
                : (c <= 94180 || (c < 100352
                  ? (c < 94208
                    ? (c >= 94192 && c <= 94193)
                    : c <= 100343)
                  : (c <= 101589 || (c < 110576
                    ? (c >= 101632 && c <= 101640)
                    : c <= 110579)))))))
            : (c <= 110587 || (c < 118576
              ? (c < 113664
                ? (c < 110928
                  ? (c < 110592
                    ? (c >= 110589 && c <= 110590)
                    : c <= 110882)
                  : (c <= 110930 || (c < 110960
                    ? (c >= 110948 && c <= 110951)
                    : c <= 111355)))
                : (c <= 113770 || (c < 113808
                  ? (c < 113792
                    ? (c >= 113776 && c <= 113788)
                    : c <= 113800)
                  : (c <= 113817 || (c < 118528
                    ? (c >= 113821 && c <= 113822)
                    : c <= 118573)))))
              : (c <= 118598 || (c < 119362
                ? (c < 119163
                  ? (c < 119149
                    ? (c >= 119141 && c <= 119145)
                    : c <= 119154)
                  : (c <= 119170 || (c < 119210
                    ? (c >= 119173 && c <= 119179)
                    : c <= 119213)))
                : (c <= 119364 || (c < 119966
                  ? (c < 119894
                    ? (c >= 119808 && c <= 119892)
                    : c <= 119964)
                  : (c <= 119967 || c == 119970))))))))))
        : (c <= 119974 || (c < 124912
          ? (c < 120746
            ? (c < 120134
              ? (c < 120071
                ? (c < 119995
                  ? (c < 119982
                    ? (c >= 119977 && c <= 119980)
                    : c <= 119993)
                  : (c <= 119995 || (c < 120005
                    ? (c >= 119997 && c <= 120003)
                    : c <= 120069)))
                : (c <= 120074 || (c < 120094
                  ? (c < 120086
                    ? (c >= 120077 && c <= 120084)
                    : c <= 120092)
                  : (c <= 120121 || (c < 120128
                    ? (c >= 120123 && c <= 120126)
                    : c <= 120132)))))
              : (c <= 120134 || (c < 120572
                ? (c < 120488
                  ? (c < 120146
                    ? (c >= 120138 && c <= 120144)
                    : c <= 120485)
                  : (c <= 120512 || (c < 120540
                    ? (c >= 120514 && c <= 120538)
                    : c <= 120570)))
                : (c <= 120596 || (c < 120656
                  ? (c < 120630
                    ? (c >= 120598 && c <= 120628)
                    : c <= 120654)
                  : (c <= 120686 || (c < 120714
                    ? (c >= 120688 && c <= 120712)
                    : c <= 120744)))))))
            : (c <= 120770 || (c < 122907
              ? (c < 121476
                ? (c < 121344
                  ? (c < 120782
                    ? (c >= 120772 && c <= 120779)
                    : c <= 120831)
                  : (c <= 121398 || (c < 121461
                    ? (c >= 121403 && c <= 121452)
                    : c <= 121461)))
                : (c <= 121476 || (c < 122624
                  ? (c < 121505
                    ? (c >= 121499 && c <= 121503)
                    : c <= 121519)
                  : (c <= 122654 || (c < 122888
                    ? (c >= 122880 && c <= 122886)
                    : c <= 122904)))))
              : (c <= 122913 || (c < 123214
                ? (c < 123136
                  ? (c < 122918
                    ? (c >= 122915 && c <= 122916)
                    : c <= 122922)
                  : (c <= 123180 || (c < 123200
                    ? (c >= 123184 && c <= 123197)
                    : c <= 123209)))
                : (c <= 123214 || (c < 124896
                  ? (c < 123584
                    ? (c >= 123536 && c <= 123566)
                    : c <= 123641)
                  : (c <= 124902 || (c < 124909
                    ? (c >= 124904 && c <= 124907)
                    : c <= 124910)))))))))
          : (c <= 124926 || (c < 126557
            ? (c < 126521
              ? (c < 126469
                ? (c < 125184
                  ? (c < 125136
                    ? (c >= 124928 && c <= 125124)
                    : c <= 125142)
                  : (c <= 125259 || (c < 126464
                    ? (c >= 125264 && c <= 125273)
                    : c <= 126467)))
                : (c <= 126495 || (c < 126503
                  ? (c < 126500
                    ? (c >= 126497 && c <= 126498)
                    : c <= 126500)
                  : (c <= 126503 || (c < 126516
                    ? (c >= 126505 && c <= 126514)
                    : c <= 126519)))))
              : (c <= 126521 || (c < 126541
                ? (c < 126535
                  ? (c < 126530
                    ? c == 126523
                    : c <= 126530)
                  : (c <= 126535 || (c < 126539
                    ? c == 126537
                    : c <= 126539)))
                : (c <= 126543 || (c < 126551
                  ? (c < 126548
                    ? (c >= 126545 && c <= 126546)
                    : c <= 126548)
                  : (c <= 126551 || (c < 126555
                    ? c == 126553
                    : c <= 126555)))))))
            : (c <= 126557 || (c < 126629
              ? (c < 126580
                ? (c < 126564
                  ? (c < 126561
                    ? c == 126559
                    : c <= 126562)
                  : (c <= 126564 || (c < 126572
                    ? (c >= 126567 && c <= 126570)
                    : c <= 126578)))
                : (c <= 126583 || (c < 126592
                  ? (c < 126590
                    ? (c >= 126585 && c <= 126588)
                    : c <= 126590)
                  : (c <= 126601 || (c < 126625
                    ? (c >= 126603 && c <= 126619)
                    : c <= 126627)))))
              : (c <= 126633 || (c < 178208
                ? (c < 131072
                  ? (c < 130032
                    ? (c >= 126635 && c <= 126651)
                    : c <= 130041)
                  : (c <= 173791 || (c < 177984
                    ? (c >= 173824 && c <= 177976)
                    : c <= 178205)))
                : (c <= 183969 || (c < 196608
                  ? (c < 194560
                    ? (c >= 183984 && c <= 191456)
                    : c <= 195101)
                  : (c <= 201546 || (c >= 917760 && c <= 917999)))))))))))))))));
}

static inline bool sym_identifier_character_set_7(int32_t c) {
  return (c < 43616
    ? (c < 3782
      ? (c < 2748
        ? (c < 2045
          ? (c < 1015
            ? (c < 710
              ? (c < 181
                ? (c < '_'
                  ? (c < 'A'
                    ? (c >= '0' && c <= '9')
                    : c <= 'Z')
                  : (c <= '_' || (c < 170
                    ? (c >= 'b' && c <= 'z')
                    : c <= 170)))
                : (c <= 181 || (c < 192
                  ? (c < 186
                    ? c == 183
                    : c <= 186)
                  : (c <= 214 || (c < 248
                    ? (c >= 216 && c <= 246)
                    : c <= 705)))))
              : (c <= 721 || (c < 891
                ? (c < 750
                  ? (c < 748
                    ? (c >= 736 && c <= 740)
                    : c <= 748)
                  : (c <= 750 || (c < 886
                    ? (c >= 768 && c <= 884)
                    : c <= 887)))
                : (c <= 893 || (c < 908
                  ? (c < 902
                    ? c == 895
                    : c <= 906)
                  : (c <= 908 || (c < 931
                    ? (c >= 910 && c <= 929)
                    : c <= 1013)))))))
            : (c <= 1153 || (c < 1519
              ? (c < 1425
                ? (c < 1329
                  ? (c < 1162
                    ? (c >= 1155 && c <= 1159)
                    : c <= 1327)
                  : (c <= 1366 || (c < 1376
                    ? c == 1369
                    : c <= 1416)))
                : (c <= 1469 || (c < 1476
                  ? (c < 1473
                    ? c == 1471
                    : c <= 1474)
                  : (c <= 1477 || (c < 1488
                    ? c == 1479
                    : c <= 1514)))))
              : (c <= 1522 || (c < 1770
                ? (c < 1646
                  ? (c < 1568
                    ? (c >= 1552 && c <= 1562)
                    : c <= 1641)
                  : (c <= 1747 || (c < 1759
                    ? (c >= 1749 && c <= 1756)
                    : c <= 1768)))
                : (c <= 1788 || (c < 1869
                  ? (c < 1808
                    ? c == 1791
                    : c <= 1866)
                  : (c <= 1969 || (c < 2042
                    ? (c >= 1984 && c <= 2037)
                    : c <= 2042)))))))))
          : (c <= 2045 || (c < 2558
            ? (c < 2451
              ? (c < 2200
                ? (c < 2144
                  ? (c < 2112
                    ? (c >= 2048 && c <= 2093)
                    : c <= 2139)
                  : (c <= 2154 || (c < 2185
                    ? (c >= 2160 && c <= 2183)
                    : c <= 2190)))
                : (c <= 2273 || (c < 2417
                  ? (c < 2406
                    ? (c >= 2275 && c <= 2403)
                    : c <= 2415)
                  : (c <= 2435 || (c < 2447
                    ? (c >= 2437 && c <= 2444)
                    : c <= 2448)))))
              : (c <= 2472 || (c < 2507
                ? (c < 2486
                  ? (c < 2482
                    ? (c >= 2474 && c <= 2480)
                    : c <= 2482)
                  : (c <= 2489 || (c < 2503
                    ? (c >= 2492 && c <= 2500)
                    : c <= 2504)))
                : (c <= 2510 || (c < 2527
                  ? (c < 2524
                    ? c == 2519
                    : c <= 2525)
                  : (c <= 2531 || (c < 2556
                    ? (c >= 2534 && c <= 2545)
                    : c <= 2556)))))))
            : (c <= 2558 || (c < 2635
              ? (c < 2610
                ? (c < 2575
                  ? (c < 2565
                    ? (c >= 2561 && c <= 2563)
                    : c <= 2570)
                  : (c <= 2576 || (c < 2602
                    ? (c >= 2579 && c <= 2600)
                    : c <= 2608)))
                : (c <= 2611 || (c < 2620
                  ? (c < 2616
                    ? (c >= 2613 && c <= 2614)
                    : c <= 2617)
                  : (c <= 2620 || (c < 2631
                    ? (c >= 2622 && c <= 2626)
                    : c <= 2632)))))
              : (c <= 2637 || (c < 2693
                ? (c < 2654
                  ? (c < 2649
                    ? c == 2641
                    : c <= 2652)
                  : (c <= 2654 || (c < 2689
                    ? (c >= 2662 && c <= 2677)
                    : c <= 2691)))
                : (c <= 2701 || (c < 2730
                  ? (c < 2707
                    ? (c >= 2703 && c <= 2705)
                    : c <= 2728)
                  : (c <= 2736 || (c < 2741
                    ? (c >= 2738 && c <= 2739)
                    : c <= 2745)))))))))))
        : (c <= 2757 || (c < 3168
          ? (c < 2958
            ? (c < 2866
              ? (c < 2809
                ? (c < 2768
                  ? (c < 2763
                    ? (c >= 2759 && c <= 2761)
                    : c <= 2765)
                  : (c <= 2768 || (c < 2790
                    ? (c >= 2784 && c <= 2787)
                    : c <= 2799)))
                : (c <= 2815 || (c < 2831
                  ? (c < 2821
                    ? (c >= 2817 && c <= 2819)
                    : c <= 2828)
                  : (c <= 2832 || (c < 2858
                    ? (c >= 2835 && c <= 2856)
                    : c <= 2864)))))
              : (c <= 2867 || (c < 2908
                ? (c < 2887
                  ? (c < 2876
                    ? (c >= 2869 && c <= 2873)
                    : c <= 2884)
                  : (c <= 2888 || (c < 2901
                    ? (c >= 2891 && c <= 2893)
                    : c <= 2903)))
                : (c <= 2909 || (c < 2929
                  ? (c < 2918
                    ? (c >= 2911 && c <= 2915)
                    : c <= 2927)
                  : (c <= 2929 || (c < 2949
                    ? (c >= 2946 && c <= 2947)
                    : c <= 2954)))))))
            : (c <= 2960 || (c < 3031
              ? (c < 2984
                ? (c < 2972
                  ? (c < 2969
                    ? (c >= 2962 && c <= 2965)
                    : c <= 2970)
                  : (c <= 2972 || (c < 2979
                    ? (c >= 2974 && c <= 2975)
                    : c <= 2980)))
                : (c <= 2986 || (c < 3014
                  ? (c < 3006
                    ? (c >= 2990 && c <= 3001)
                    : c <= 3010)
                  : (c <= 3016 || (c < 3024
                    ? (c >= 3018 && c <= 3021)
                    : c <= 3024)))))
              : (c <= 3031 || (c < 3132
                ? (c < 3086
                  ? (c < 3072
                    ? (c >= 3046 && c <= 3055)
                    : c <= 3084)
                  : (c <= 3088 || (c < 3114
                    ? (c >= 3090 && c <= 3112)
                    : c <= 3129)))
                : (c <= 3140 || (c < 3157
                  ? (c < 3146
                    ? (c >= 3142 && c <= 3144)
                    : c <= 3149)
                  : (c <= 3158 || (c < 3165
                    ? (c >= 3160 && c <= 3162)
                    : c <= 3165)))))))))
          : (c <= 3171 || (c < 3450
            ? (c < 3293
              ? (c < 3242
                ? (c < 3205
                  ? (c < 3200
                    ? (c >= 3174 && c <= 3183)
                    : c <= 3203)
                  : (c <= 3212 || (c < 3218
                    ? (c >= 3214 && c <= 3216)
                    : c <= 3240)))
                : (c <= 3251 || (c < 3270
                  ? (c < 3260
                    ? (c >= 3253 && c <= 3257)
                    : c <= 3268)
                  : (c <= 3272 || (c < 3285
                    ? (c >= 3274 && c <= 3277)
                    : c <= 3286)))))
              : (c <= 3294 || (c < 3346
                ? (c < 3313
                  ? (c < 3302
                    ? (c >= 3296 && c <= 3299)
                    : c <= 3311)
                  : (c <= 3314 || (c < 3342
                    ? (c >= 3328 && c <= 3340)
                    : c <= 3344)))
                : (c <= 3396 || (c < 3412
                  ? (c < 3402
                    ? (c >= 3398 && c <= 3400)
                    : c <= 3406)
                  : (c <= 3415 || (c < 3430
                    ? (c >= 3423 && c <= 3427)
                    : c <= 3439)))))))
            : (c <= 3455 || (c < 3570
              ? (c < 3520
                ? (c < 3482
                  ? (c < 3461
                    ? (c >= 3457 && c <= 3459)
                    : c <= 3478)
                  : (c <= 3505 || (c < 3517
                    ? (c >= 3507 && c <= 3515)
                    : c <= 3517)))
                : (c <= 3526 || (c < 3542
                  ? (c < 3535
                    ? c == 3530
                    : c <= 3540)
                  : (c <= 3542 || (c < 3558
                    ? (c >= 3544 && c <= 3551)
                    : c <= 3567)))))
              : (c <= 3571 || (c < 3718
                ? (c < 3664
                  ? (c < 3648
                    ? (c >= 3585 && c <= 3642)
                    : c <= 3662)
                  : (c <= 3673 || (c < 3716
                    ? (c >= 3713 && c <= 3714)
                    : c <= 3716)))
                : (c <= 3722 || (c < 3751
                  ? (c < 3749
                    ? (c >= 3724 && c <= 3747)
                    : c <= 3749)
                  : (c <= 3773 || (c >= 3776 && c <= 3780)))))))))))))
      : (c <= 3782 || (c < 8025
        ? (c < 5888
          ? (c < 4688
            ? (c < 3953
              ? (c < 3872
                ? (c < 3804
                  ? (c < 3792
                    ? (c >= 3784 && c <= 3789)
                    : c <= 3801)
                  : (c <= 3807 || (c < 3864
                    ? c == 3840
                    : c <= 3865)))
                : (c <= 3881 || (c < 3897
                  ? (c < 3895
                    ? c == 3893
                    : c <= 3895)
                  : (c <= 3897 || (c < 3913
                    ? (c >= 3902 && c <= 3911)
                    : c <= 3948)))))
              : (c <= 3972 || (c < 4256
                ? (c < 4038
                  ? (c < 3993
                    ? (c >= 3974 && c <= 3991)
                    : c <= 4028)
                  : (c <= 4038 || (c < 4176
                    ? (c >= 4096 && c <= 4169)
                    : c <= 4253)))
                : (c <= 4293 || (c < 4304
                  ? (c < 4301
                    ? c == 4295
                    : c <= 4301)
                  : (c <= 4346 || (c < 4682
                    ? (c >= 4348 && c <= 4680)
                    : c <= 4685)))))))
            : (c <= 4694 || (c < 4882
              ? (c < 4786
                ? (c < 4704
                  ? (c < 4698
                    ? c == 4696
                    : c <= 4701)
                  : (c <= 4744 || (c < 4752
                    ? (c >= 4746 && c <= 4749)
                    : c <= 4784)))
                : (c <= 4789 || (c < 4802
                  ? (c < 4800
                    ? (c >= 4792 && c <= 4798)
                    : c <= 4800)
                  : (c <= 4805 || (c < 4824
                    ? (c >= 4808 && c <= 4822)
                    : c <= 4880)))))
              : (c <= 4885 || (c < 5112
                ? (c < 4969
                  ? (c < 4957
                    ? (c >= 4888 && c <= 4954)
                    : c <= 4959)
                  : (c <= 4977 || (c < 5024
                    ? (c >= 4992 && c <= 5007)
                    : c <= 5109)))
                : (c <= 5117 || (c < 5761
                  ? (c < 5743
                    ? (c >= 5121 && c <= 5740)
                    : c <= 5759)
                  : (c <= 5786 || (c < 5870
                    ? (c >= 5792 && c <= 5866)
                    : c <= 5880)))))))))
          : (c <= 5909 || (c < 6688
            ? (c < 6176
              ? (c < 6016
                ? (c < 5984
                  ? (c < 5952
                    ? (c >= 5919 && c <= 5940)
                    : c <= 5971)
                  : (c <= 5996 || (c < 6002
                    ? (c >= 5998 && c <= 6000)
                    : c <= 6003)))
                : (c <= 6099 || (c < 6112
                  ? (c < 6108
                    ? c == 6103
                    : c <= 6109)
                  : (c <= 6121 || (c < 6159
                    ? (c >= 6155 && c <= 6157)
                    : c <= 6169)))))
              : (c <= 6264 || (c < 6470
                ? (c < 6400
                  ? (c < 6320
                    ? (c >= 6272 && c <= 6314)
                    : c <= 6389)
                  : (c <= 6430 || (c < 6448
                    ? (c >= 6432 && c <= 6443)
                    : c <= 6459)))
                : (c <= 6509 || (c < 6576
                  ? (c < 6528
                    ? (c >= 6512 && c <= 6516)
                    : c <= 6571)
                  : (c <= 6601 || (c < 6656
                    ? (c >= 6608 && c <= 6618)
                    : c <= 6683)))))))
            : (c <= 6750 || (c < 7232
              ? (c < 6847
                ? (c < 6800
                  ? (c < 6783
                    ? (c >= 6752 && c <= 6780)
                    : c <= 6793)
                  : (c <= 6809 || (c < 6832
                    ? c == 6823
                    : c <= 6845)))
                : (c <= 6862 || (c < 7019
                  ? (c < 6992
                    ? (c >= 6912 && c <= 6988)
                    : c <= 7001)
                  : (c <= 7027 || (c < 7168
                    ? (c >= 7040 && c <= 7155)
                    : c <= 7223)))))
              : (c <= 7241 || (c < 7380
                ? (c < 7312
                  ? (c < 7296
                    ? (c >= 7245 && c <= 7293)
                    : c <= 7304)
                  : (c <= 7354 || (c < 7376
                    ? (c >= 7357 && c <= 7359)
                    : c <= 7378)))
                : (c <= 7418 || (c < 7968
                  ? (c < 7960
                    ? (c >= 7424 && c <= 7957)
                    : c <= 7965)
                  : (c <= 8005 || (c < 8016
                    ? (c >= 8008 && c <= 8013)
                    : c <= 8023)))))))))))
        : (c <= 8025 || (c < 11720
          ? (c < 8458
            ? (c < 8178
              ? (c < 8126
                ? (c < 8031
                  ? (c < 8029
                    ? c == 8027
                    : c <= 8029)
                  : (c <= 8061 || (c < 8118
                    ? (c >= 8064 && c <= 8116)
                    : c <= 8124)))
                : (c <= 8126 || (c < 8144
                  ? (c < 8134
                    ? (c >= 8130 && c <= 8132)
                    : c <= 8140)
                  : (c <= 8147 || (c < 8160
                    ? (c >= 8150 && c <= 8155)
                    : c <= 8172)))))
              : (c <= 8180 || (c < 8336
                ? (c < 8276
                  ? (c < 8255
                    ? (c >= 8182 && c <= 8188)
                    : c <= 8256)
                  : (c <= 8276 || (c < 8319
                    ? c == 8305
                    : c <= 8319)))
                : (c <= 8348 || (c < 8421
                  ? (c < 8417
                    ? (c >= 8400 && c <= 8412)
                    : c <= 8417)
                  : (c <= 8432 || (c < 8455
                    ? c == 8450
                    : c <= 8455)))))))
            : (c <= 8467 || (c < 11499
              ? (c < 8490
                ? (c < 8484
                  ? (c < 8472
                    ? c == 8469
                    : c <= 8477)
                  : (c <= 8484 || (c < 8488
                    ? c == 8486
                    : c <= 8488)))
                : (c <= 8505 || (c < 8526
                  ? (c < 8517
                    ? (c >= 8508 && c <= 8511)
                    : c <= 8521)
                  : (c <= 8526 || (c < 11264
                    ? (c >= 8544 && c <= 8584)
                    : c <= 11492)))))
              : (c <= 11507 || (c < 11647
                ? (c < 11565
                  ? (c < 11559
                    ? (c >= 11520 && c <= 11557)
                    : c <= 11559)
                  : (c <= 11565 || (c < 11631
                    ? (c >= 11568 && c <= 11623)
                    : c <= 11631)))
                : (c <= 11670 || (c < 11696
                  ? (c < 11688
                    ? (c >= 11680 && c <= 11686)
                    : c <= 11694)
                  : (c <= 11702 || (c < 11712
                    ? (c >= 11704 && c <= 11710)
                    : c <= 11718)))))))))
          : (c <= 11726 || (c < 42623
            ? (c < 12540
              ? (c < 12337
                ? (c < 11744
                  ? (c < 11736
                    ? (c >= 11728 && c <= 11734)
                    : c <= 11742)
                  : (c <= 11775 || (c < 12321
                    ? (c >= 12293 && c <= 12295)
                    : c <= 12335)))
                : (c <= 12341 || (c < 12441
                  ? (c < 12353
                    ? (c >= 12344 && c <= 12348)
                    : c <= 12438)
                  : (c <= 12442 || (c < 12449
                    ? (c >= 12445 && c <= 12447)
                    : c <= 12538)))))
              : (c <= 12543 || (c < 19968
                ? (c < 12704
                  ? (c < 12593
                    ? (c >= 12549 && c <= 12591)
                    : c <= 12686)
                  : (c <= 12735 || (c < 13312
                    ? (c >= 12784 && c <= 12799)
                    : c <= 19903)))
                : (c <= 42124 || (c < 42512
                  ? (c < 42240
                    ? (c >= 42192 && c <= 42237)
                    : c <= 42508)
                  : (c <= 42539 || (c < 42612
                    ? (c >= 42560 && c <= 42607)
                    : c <= 42621)))))))
            : (c <= 42737 || (c < 43232
              ? (c < 42965
                ? (c < 42891
                  ? (c < 42786
                    ? (c >= 42775 && c <= 42783)
                    : c <= 42888)
                  : (c <= 42954 || (c < 42963
                    ? (c >= 42960 && c <= 42961)
                    : c <= 42963)))
                : (c <= 42969 || (c < 43072
                  ? (c < 43052
                    ? (c >= 42994 && c <= 43047)
                    : c <= 43052)
                  : (c <= 43123 || (c < 43216
                    ? (c >= 43136 && c <= 43205)
                    : c <= 43225)))))
              : (c <= 43255 || (c < 43471
                ? (c < 43312
                  ? (c < 43261
                    ? c == 43259
                    : c <= 43309)
                  : (c <= 43347 || (c < 43392
                    ? (c >= 43360 && c <= 43388)
                    : c <= 43456)))
                : (c <= 43481 || (c < 43584
                  ? (c < 43520
                    ? (c >= 43488 && c <= 43518)
                    : c <= 43574)
                  : (c <= 43597 || (c >= 43600 && c <= 43609)))))))))))))))
    : (c <= 43638 || (c < 71453
      ? (c < 67639
        ? (c < 65345
          ? (c < 64312
            ? (c < 43888
              ? (c < 43785
                ? (c < 43744
                  ? (c < 43739
                    ? (c >= 43642 && c <= 43714)
                    : c <= 43741)
                  : (c <= 43759 || (c < 43777
                    ? (c >= 43762 && c <= 43766)
                    : c <= 43782)))
                : (c <= 43790 || (c < 43816
                  ? (c < 43808
                    ? (c >= 43793 && c <= 43798)
                    : c <= 43814)
                  : (c <= 43822 || (c < 43868
                    ? (c >= 43824 && c <= 43866)
                    : c <= 43881)))))
              : (c <= 44010 || (c < 63744
                ? (c < 44032
                  ? (c < 44016
                    ? (c >= 44012 && c <= 44013)
                    : c <= 44025)
                  : (c <= 55203 || (c < 55243
                    ? (c >= 55216 && c <= 55238)
                    : c <= 55291)))
                : (c <= 64109 || (c < 64275
                  ? (c < 64256
                    ? (c >= 64112 && c <= 64217)
                    : c <= 64262)
                  : (c <= 64279 || (c < 64298
                    ? (c >= 64285 && c <= 64296)
                    : c <= 64310)))))))
            : (c <= 64316 || (c < 65075
              ? (c < 64612
                ? (c < 64323
                  ? (c < 64320
                    ? c == 64318
                    : c <= 64321)
                  : (c <= 64324 || (c < 64467
                    ? (c >= 64326 && c <= 64433)
                    : c <= 64605)))
                : (c <= 64829 || (c < 65008
                  ? (c < 64914
                    ? (c >= 64848 && c <= 64911)
                    : c <= 64967)
                  : (c <= 65017 || (c < 65056
                    ? (c >= 65024 && c <= 65039)
                    : c <= 65071)))))
              : (c <= 65076 || (c < 65147
                ? (c < 65139
                  ? (c < 65137
                    ? (c >= 65101 && c <= 65103)
                    : c <= 65137)
                  : (c <= 65139 || (c < 65145
                    ? c == 65143
                    : c <= 65145)))
                : (c <= 65147 || (c < 65296
                  ? (c < 65151
                    ? c == 65149
                    : c <= 65276)
                  : (c <= 65305 || (c < 65343
                    ? (c >= 65313 && c <= 65338)
                    : c <= 65343)))))))))
          : (c <= 65370 || (c < 66513
            ? (c < 65664
              ? (c < 65536
                ? (c < 65482
                  ? (c < 65474
                    ? (c >= 65382 && c <= 65470)
                    : c <= 65479)
                  : (c <= 65487 || (c < 65498
                    ? (c >= 65490 && c <= 65495)
                    : c <= 65500)))
                : (c <= 65547 || (c < 65596
                  ? (c < 65576
                    ? (c >= 65549 && c <= 65574)
                    : c <= 65594)
                  : (c <= 65597 || (c < 65616
                    ? (c >= 65599 && c <= 65613)
                    : c <= 65629)))))
              : (c <= 65786 || (c < 66304
                ? (c < 66176
                  ? (c < 66045
                    ? (c >= 65856 && c <= 65908)
                    : c <= 66045)
                  : (c <= 66204 || (c < 66272
                    ? (c >= 66208 && c <= 66256)
                    : c <= 66272)))
                : (c <= 66335 || (c < 66432
                  ? (c < 66384
                    ? (c >= 66349 && c <= 66378)
                    : c <= 66426)
                  : (c <= 66461 || (c < 66504
                    ? (c >= 66464 && c <= 66499)
                    : c <= 66511)))))))
            : (c <= 66517 || (c < 66979
              ? (c < 66864
                ? (c < 66736
                  ? (c < 66720
                    ? (c >= 66560 && c <= 66717)
                    : c <= 66729)
                  : (c <= 66771 || (c < 66816
                    ? (c >= 66776 && c <= 66811)
                    : c <= 66855)))
                : (c <= 66915 || (c < 66956
                  ? (c < 66940
                    ? (c >= 66928 && c <= 66938)
                    : c <= 66954)
                  : (c <= 66962 || (c < 66967
                    ? (c >= 66964 && c <= 66965)
                    : c <= 66977)))))
              : (c <= 66993 || (c < 67456
                ? (c < 67072
                  ? (c < 67003
                    ? (c >= 66995 && c <= 67001)
                    : c <= 67004)
                  : (c <= 67382 || (c < 67424
                    ? (c >= 67392 && c <= 67413)
                    : c <= 67431)))
                : (c <= 67461 || (c < 67584
                  ? (c < 67506
                    ? (c >= 67463 && c <= 67504)
                    : c <= 67514)
                  : (c <= 67589 || (c < 67594
                    ? c == 67592
                    : c <= 67637)))))))))))
        : (c <= 67640 || (c < 69956
          ? (c < 68448
            ? (c < 68101
              ? (c < 67828
                ? (c < 67680
                  ? (c < 67647
                    ? c == 67644
                    : c <= 67669)
                  : (c <= 67702 || (c < 67808
                    ? (c >= 67712 && c <= 67742)
                    : c <= 67826)))
                : (c <= 67829 || (c < 67968
                  ? (c < 67872
                    ? (c >= 67840 && c <= 67861)
                    : c <= 67897)
                  : (c <= 68023 || (c < 68096
                    ? (c >= 68030 && c <= 68031)
                    : c <= 68099)))))
              : (c <= 68102 || (c < 68192
                ? (c < 68121
                  ? (c < 68117
                    ? (c >= 68108 && c <= 68115)
                    : c <= 68119)
                  : (c <= 68149 || (c < 68159
                    ? (c >= 68152 && c <= 68154)
                    : c <= 68159)))
                : (c <= 68220 || (c < 68297
                  ? (c < 68288
                    ? (c >= 68224 && c <= 68252)
                    : c <= 68295)
                  : (c <= 68326 || (c < 68416
                    ? (c >= 68352 && c <= 68405)
                    : c <= 68437)))))))
            : (c <= 68466 || (c < 69424
              ? (c < 68912
                ? (c < 68736
                  ? (c < 68608
                    ? (c >= 68480 && c <= 68497)
                    : c <= 68680)
                  : (c <= 68786 || (c < 68864
                    ? (c >= 68800 && c <= 68850)
                    : c <= 68903)))
                : (c <= 68921 || (c < 69296
                  ? (c < 69291
                    ? (c >= 69248 && c <= 69289)
                    : c <= 69292)
                  : (c <= 69297 || (c < 69415
                    ? (c >= 69376 && c <= 69404)
                    : c <= 69415)))))
              : (c <= 69456 || (c < 69759
                ? (c < 69600
                  ? (c < 69552
                    ? (c >= 69488 && c <= 69509)
                    : c <= 69572)
                  : (c <= 69622 || (c < 69734
                    ? (c >= 69632 && c <= 69702)
                    : c <= 69749)))
                : (c <= 69818 || (c < 69872
                  ? (c < 69840
                    ? c == 69826
                    : c <= 69864)
                  : (c <= 69881 || (c < 69942
                    ? (c >= 69888 && c <= 69940)
                    : c <= 69951)))))))))
          : (c <= 69959 || (c < 70459
            ? (c < 70282
              ? (c < 70108
                ? (c < 70016
                  ? (c < 70006
                    ? (c >= 69968 && c <= 70003)
                    : c <= 70006)
                  : (c <= 70084 || (c < 70094
                    ? (c >= 70089 && c <= 70092)
                    : c <= 70106)))
                : (c <= 70108 || (c < 70206
                  ? (c < 70163
                    ? (c >= 70144 && c <= 70161)
                    : c <= 70199)
                  : (c <= 70206 || (c < 70280
                    ? (c >= 70272 && c <= 70278)
                    : c <= 70280)))))
              : (c <= 70285 || (c < 70405
                ? (c < 70320
                  ? (c < 70303
                    ? (c >= 70287 && c <= 70301)
                    : c <= 70312)
                  : (c <= 70378 || (c < 70400
                    ? (c >= 70384 && c <= 70393)
                    : c <= 70403)))
                : (c <= 70412 || (c < 70442
                  ? (c < 70419
                    ? (c >= 70415 && c <= 70416)
                    : c <= 70440)
                  : (c <= 70448 || (c < 70453
                    ? (c >= 70450 && c <= 70451)
                    : c <= 70457)))))))
            : (c <= 70468 || (c < 70855
              ? (c < 70502
                ? (c < 70480
                  ? (c < 70475
                    ? (c >= 70471 && c <= 70472)
                    : c <= 70477)
                  : (c <= 70480 || (c < 70493
                    ? c == 70487
                    : c <= 70499)))
                : (c <= 70508 || (c < 70736
                  ? (c < 70656
                    ? (c >= 70512 && c <= 70516)
                    : c <= 70730)
                  : (c <= 70745 || (c < 70784
                    ? (c >= 70750 && c <= 70753)
                    : c <= 70853)))))
              : (c <= 70855 || (c < 71236
                ? (c < 71096
                  ? (c < 71040
                    ? (c >= 70864 && c <= 70873)
                    : c <= 71093)
                  : (c <= 71104 || (c < 71168
                    ? (c >= 71128 && c <= 71133)
                    : c <= 71232)))
                : (c <= 71236 || (c < 71360
                  ? (c < 71296
                    ? (c >= 71248 && c <= 71257)
                    : c <= 71352)
                  : (c <= 71369 || (c >= 71424 && c <= 71450)))))))))))))
      : (c <= 71467 || (c < 119973
        ? (c < 77824
          ? (c < 72760
            ? (c < 72016
              ? (c < 71945
                ? (c < 71680
                  ? (c < 71488
                    ? (c >= 71472 && c <= 71481)
                    : c <= 71494)
                  : (c <= 71738 || (c < 71935
                    ? (c >= 71840 && c <= 71913)
                    : c <= 71942)))
                : (c <= 71945 || (c < 71960
                  ? (c < 71957
                    ? (c >= 71948 && c <= 71955)
                    : c <= 71958)
                  : (c <= 71989 || (c < 71995
                    ? (c >= 71991 && c <= 71992)
                    : c <= 72003)))))
              : (c <= 72025 || (c < 72263
                ? (c < 72154
                  ? (c < 72106
                    ? (c >= 72096 && c <= 72103)
                    : c <= 72151)
                  : (c <= 72161 || (c < 72192
                    ? (c >= 72163 && c <= 72164)
                    : c <= 72254)))
                : (c <= 72263 || (c < 72368
                  ? (c < 72349
                    ? (c >= 72272 && c <= 72345)
                    : c <= 72349)
                  : (c <= 72440 || (c < 72714
                    ? (c >= 72704 && c <= 72712)
                    : c <= 72758)))))))
            : (c <= 72768 || (c < 73056
              ? (c < 72968
                ? (c < 72850
                  ? (c < 72818
                    ? (c >= 72784 && c <= 72793)
                    : c <= 72847)
                  : (c <= 72871 || (c < 72960
                    ? (c >= 72873 && c <= 72886)
                    : c <= 72966)))
                : (c <= 72969 || (c < 73020
                  ? (c < 73018
                    ? (c >= 72971 && c <= 73014)
                    : c <= 73018)
                  : (c <= 73021 || (c < 73040
                    ? (c >= 73023 && c <= 73031)
                    : c <= 73049)))))
              : (c <= 73061 || (c < 73440
                ? (c < 73104
                  ? (c < 73066
                    ? (c >= 73063 && c <= 73064)
                    : c <= 73102)
                  : (c <= 73105 || (c < 73120
                    ? (c >= 73107 && c <= 73112)
                    : c <= 73129)))
                : (c <= 73462 || (c < 74752
                  ? (c < 73728
                    ? c == 73648
                    : c <= 74649)
                  : (c <= 74862 || (c < 77712
                    ? (c >= 74880 && c <= 75075)
                    : c <= 77808)))))))))
          : (c <= 78894 || (c < 110576
            ? (c < 93027
              ? (c < 92864
                ? (c < 92736
                  ? (c < 92160
                    ? (c >= 82944 && c <= 83526)
                    : c <= 92728)
                  : (c <= 92766 || (c < 92784
                    ? (c >= 92768 && c <= 92777)
                    : c <= 92862)))
                : (c <= 92873 || (c < 92928
                  ? (c < 92912
                    ? (c >= 92880 && c <= 92909)
                    : c <= 92916)
                  : (c <= 92982 || (c < 93008
                    ? (c >= 92992 && c <= 92995)
                    : c <= 93017)))))
              : (c <= 93047 || (c < 94176
                ? (c < 93952
                  ? (c < 93760
                    ? (c >= 93053 && c <= 93071)
                    : c <= 93823)
                  : (c <= 94026 || (c < 94095
                    ? (c >= 94031 && c <= 94087)
                    : c <= 94111)))
                : (c <= 94177 || (c < 94208
                  ? (c < 94192
                    ? (c >= 94179 && c <= 94180)
                    : c <= 94193)
                  : (c <= 100343 || (c < 101632
                    ? (c >= 100352 && c <= 101589)
                    : c <= 101640)))))))
            : (c <= 110579 || (c < 118528
              ? (c < 110960
                ? (c < 110592
                  ? (c < 110589
                    ? (c >= 110581 && c <= 110587)
                    : c <= 110590)
                  : (c <= 110882 || (c < 110948
                    ? (c >= 110928 && c <= 110930)
                    : c <= 110951)))
                : (c <= 111355 || (c < 113792
                  ? (c < 113776
                    ? (c >= 113664 && c <= 113770)
                    : c <= 113788)
                  : (c <= 113800 || (c < 113821
                    ? (c >= 113808 && c <= 113817)
                    : c <= 113822)))))
              : (c <= 118573 || (c < 119210
                ? (c < 119149
                  ? (c < 119141
                    ? (c >= 118576 && c <= 118598)
                    : c <= 119145)
                  : (c <= 119154 || (c < 119173
                    ? (c >= 119163 && c <= 119170)
                    : c <= 119179)))
                : (c <= 119213 || (c < 119894
                  ? (c < 119808
                    ? (c >= 119362 && c <= 119364)
                    : c <= 119892)
                  : (c <= 119964 || (c < 119970
                    ? (c >= 119966 && c <= 119967)
                    : c <= 119970)))))))))))
        : (c <= 119974 || (c < 124912
          ? (c < 120746
            ? (c < 120134
              ? (c < 120071
                ? (c < 119995
                  ? (c < 119982
                    ? (c >= 119977 && c <= 119980)
                    : c <= 119993)
                  : (c <= 119995 || (c < 120005
                    ? (c >= 119997 && c <= 120003)
                    : c <= 120069)))
                : (c <= 120074 || (c < 120094
                  ? (c < 120086
                    ? (c >= 120077 && c <= 120084)
                    : c <= 120092)
                  : (c <= 120121 || (c < 120128
                    ? (c >= 120123 && c <= 120126)
                    : c <= 120132)))))
              : (c <= 120134 || (c < 120572
                ? (c < 120488
                  ? (c < 120146
                    ? (c >= 120138 && c <= 120144)
                    : c <= 120485)
                  : (c <= 120512 || (c < 120540
                    ? (c >= 120514 && c <= 120538)
                    : c <= 120570)))
                : (c <= 120596 || (c < 120656
                  ? (c < 120630
                    ? (c >= 120598 && c <= 120628)
                    : c <= 120654)
                  : (c <= 120686 || (c < 120714
                    ? (c >= 120688 && c <= 120712)
                    : c <= 120744)))))))
            : (c <= 120770 || (c < 122907
              ? (c < 121476
                ? (c < 121344
                  ? (c < 120782
                    ? (c >= 120772 && c <= 120779)
                    : c <= 120831)
                  : (c <= 121398 || (c < 121461
                    ? (c >= 121403 && c <= 121452)
                    : c <= 121461)))
                : (c <= 121476 || (c < 122624
                  ? (c < 121505
                    ? (c >= 121499 && c <= 121503)
                    : c <= 121519)
                  : (c <= 122654 || (c < 122888
                    ? (c >= 122880 && c <= 122886)
                    : c <= 122904)))))
              : (c <= 122913 || (c < 123214
                ? (c < 123136
                  ? (c < 122918
                    ? (c >= 122915 && c <= 122916)
                    : c <= 122922)
                  : (c <= 123180 || (c < 123200
                    ? (c >= 123184 && c <= 123197)
                    : c <= 123209)))
                : (c <= 123214 || (c < 124896
                  ? (c < 123584
                    ? (c >= 123536 && c <= 123566)
                    : c <= 123641)
                  : (c <= 124902 || (c < 124909
                    ? (c >= 124904 && c <= 124907)
                    : c <= 124910)))))))))
          : (c <= 124926 || (c < 126557
            ? (c < 126521
              ? (c < 126469
                ? (c < 125184
                  ? (c < 125136
                    ? (c >= 124928 && c <= 125124)
                    : c <= 125142)
                  : (c <= 125259 || (c < 126464
                    ? (c >= 125264 && c <= 125273)
                    : c <= 126467)))
                : (c <= 126495 || (c < 126503
                  ? (c < 126500
                    ? (c >= 126497 && c <= 126498)
                    : c <= 126500)
                  : (c <= 126503 || (c < 126516
                    ? (c >= 126505 && c <= 126514)
                    : c <= 126519)))))
              : (c <= 126521 || (c < 126541
                ? (c < 126535
                  ? (c < 126530
                    ? c == 126523
                    : c <= 126530)
                  : (c <= 126535 || (c < 126539
                    ? c == 126537
                    : c <= 126539)))
                : (c <= 126543 || (c < 126551
                  ? (c < 126548
                    ? (c >= 126545 && c <= 126546)
                    : c <= 126548)
                  : (c <= 126551 || (c < 126555
                    ? c == 126553
                    : c <= 126555)))))))
            : (c <= 126557 || (c < 126629
              ? (c < 126580
                ? (c < 126564
                  ? (c < 126561
                    ? c == 126559
                    : c <= 126562)
                  : (c <= 126564 || (c < 126572
                    ? (c >= 126567 && c <= 126570)
                    : c <= 126578)))
                : (c <= 126583 || (c < 126592
                  ? (c < 126590
                    ? (c >= 126585 && c <= 126588)
                    : c <= 126590)
                  : (c <= 126601 || (c < 126625
                    ? (c >= 126603 && c <= 126619)
                    : c <= 126627)))))
              : (c <= 126633 || (c < 178208
                ? (c < 131072
                  ? (c < 130032
                    ? (c >= 126635 && c <= 126651)
                    : c <= 130041)
                  : (c <= 173791 || (c < 177984
                    ? (c >= 173824 && c <= 177976)
                    : c <= 178205)))
                : (c <= 183969 || (c < 196608
                  ? (c < 194560
                    ? (c >= 183984 && c <= 191456)
                    : c <= 195101)
                  : (c <= 201546 || (c >= 917760 && c <= 917999)))))))))))))))));
}

static bool ts_lex(TSLexer *lexer, TSStateId state) {
  START_LEXER();
  eof = lexer->eof(lexer);
  switch (state) {
    case 0:
      if (eof) ADVANCE(59);
      if (lookahead == '!') ADVANCE(91);
      if (lookahead == '"') ADVANCE(150);
      if (lookahead == '#') ADVANCE(89);
      if (lookahead == '$') ADVANCE(73);
      if (lookahead == '%') ADVANCE(127);
      if (lookahead == '&') ADVANCE(104);
      if (lookahead == '\'') ADVANCE(88);
      if (lookahead == '(') ADVANCE(62);
      if (lookahead == ')') ADVANCE(63);
      if (lookahead == '*') ADVANCE(81);
      if (lookahead == '+') ADVANCE(79);
      if (lookahead == ',') ADVANCE(96);
      if (lookahead == '-') ADVANCE(113);
      if (lookahead == '.') ADVANCE(138);
      if (lookahead == '/') ADVANCE(126);
      if (lookahead == '0') ADVANCE(144);
      if (lookahead == ':') ADVANCE(70);
      if (lookahead == ';') ADVANCE(60);
      if (lookahead == '<') ADVANCE(106);
      if (lookahead == '=') ADVANCE(94);
      if (lookahead == '>') ADVANCE(101);
      if (lookahead == '?') ADVANCE(82);
      if (lookahead == '@') ADVANCE(139);
      if (lookahead == '[') ADVANCE(67);
      if (lookahead == '\\') ADVANCE(36);
      if (lookahead == ']') ADVANCE(68);
      if (lookahead == '^') ADVANCE(119);
      if (lookahead == 'b') ADVANCE(155);
      if (lookahead == 'm') ADVANCE(159);
      if (lookahead == 'r') ADVANCE(157);
      if (lookahead == '{') ADVANCE(64);
      if (lookahead == '|') ADVANCE(118);
      if (lookahead == '}') ADVANCE(65);
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(56)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(147);
      if (sym_identifier_character_set_1(lookahead)) ADVANCE(168);
      END_STATE();
    case 1:
      if (lookahead == '!') ADVANCE(91);
      if (lookahead == '"') ADVANCE(149);
      if (lookahead == '#') ADVANCE(89);
      if (lookahead == '$') ADVANCE(54);
      if (lookahead == '%') ADVANCE(127);
      if (lookahead == '&') ADVANCE(104);
      if (lookahead == '\'') ADVANCE(88);
      if (lookahead == '(') ADVANCE(62);
      if (lookahead == ')') ADVANCE(63);
      if (lookahead == '*') ADVANCE(81);
      if (lookahead == '+') ADVANCE(79);
      if (lookahead == ',') ADVANCE(96);
      if (lookahead == '-') ADVANCE(112);
      if (lookahead == '.') ADVANCE(138);
      if (lookahead == '/') ADVANCE(126);
      if (lookahead == '0') ADVANCE(144);
      if (lookahead == ':') ADVANCE(70);
      if (lookahead == ';') ADVANCE(60);
      if (lookahead == '<') ADVANCE(99);
      if (lookahead == '=') ADVANCE(94);
      if (lookahead == '>') ADVANCE(101);
      if (lookahead == '?') ADVANCE(82);
      if (lookahead == '[') ADVANCE(67);
      if (lookahead == ']') ADVANCE(68);
      if (lookahead == '^') ADVANCE(119);
      if (lookahead == 'b') ADVANCE(155);
      if (lookahead == 'r') ADVANCE(157);
      if (lookahead == '{') ADVANCE(64);
      if (lookahead == '|') ADVANCE(118);
      if (lookahead == '}') ADVANCE(65);
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(1)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(147);
      if (sym_identifier_character_set_2(lookahead)) ADVANCE(168);
      END_STATE();
    case 2:
      if (lookahead == '!') ADVANCE(91);
      if (lookahead == '%') ADVANCE(127);
      if (lookahead == '&') ADVANCE(104);
      if (lookahead == '\'') ADVANCE(87);
      if (lookahead == '(') ADVANCE(62);
      if (lookahead == ')') ADVANCE(63);
      if (lookahead == '*') ADVANCE(81);
      if (lookahead == '+') ADVANCE(79);
      if (lookahead == ',') ADVANCE(96);
      if (lookahead == '-') ADVANCE(113);
      if (lookahead == '.') ADVANCE(138);
      if (lookahead == '/') ADVANCE(126);
      if (lookahead == ':') ADVANCE(31);
      if (lookahead == ';') ADVANCE(60);
      if (lookahead == '<') ADVANCE(99);
      if (lookahead == '=') ADVANCE(94);
      if (lookahead == '>') ADVANCE(101);
      if (lookahead == '?') ADVANCE(82);
      if (lookahead == '[') ADVANCE(67);
      if (lookahead == ']') ADVANCE(68);
      if (lookahead == '^') ADVANCE(119);
      if (lookahead == 'r') ADVANCE(157);
      if (lookahead == '{') ADVANCE(64);
      if (lookahead == '|') ADVANCE(118);
      if (lookahead == '}') ADVANCE(65);
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(2)
      if (sym_identifier_character_set_2(lookahead)) ADVANCE(168);
      END_STATE();
    case 3:
      if (lookahead == '!') ADVANCE(91);
      if (lookahead == '%') ADVANCE(127);
      if (lookahead == '&') ADVANCE(104);
      if (lookahead == '(') ADVANCE(62);
      if (lookahead == ')') ADVANCE(63);
      if (lookahead == '*') ADVANCE(81);
      if (lookahead == '+') ADVANCE(79);
      if (lookahead == ',') ADVANCE(96);
      if (lookahead == '-') ADVANCE(112);
      if (lookahead == '.') ADVANCE(138);
      if (lookahead == '/') ADVANCE(126);
      if (lookahead == ':') ADVANCE(31);
      if (lookahead == ';') ADVANCE(60);
      if (lookahead == '<') ADVANCE(106);
      if (lookahead == '=') ADVANCE(94);
      if (lookahead == '>') ADVANCE(101);
      if (lookahead == '?') ADVANCE(82);
      if (lookahead == '[') ADVANCE(67);
      if (lookahead == ']') ADVANCE(68);
      if (lookahead == '^') ADVANCE(119);
      if (lookahead == 'r') ADVANCE(157);
      if (lookahead == '{') ADVANCE(64);
      if (lookahead == '|') ADVANCE(118);
      if (lookahead == '}') ADVANCE(65);
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(3)
      if (sym_identifier_character_set_2(lookahead)) ADVANCE(168);
      END_STATE();
    case 4:
      if (lookahead == '!') ADVANCE(90);
      if (lookahead == '"') ADVANCE(150);
      if (lookahead == '#') ADVANCE(89);
      if (lookahead == '$') ADVANCE(54);
      if (lookahead == '\'') ADVANCE(87);
      if (lookahead == '(') ADVANCE(62);
      if (lookahead == ')') ADVANCE(63);
      if (lookahead == '+') ADVANCE(78);
      if (lookahead == ',') ADVANCE(96);
      if (lookahead == '.') ADVANCE(23);
      if (lookahead == '/') ADVANCE(24);
      if (lookahead == ':') ADVANCE(70);
      if (lookahead == ';') ADVANCE(60);
      if (lookahead == '<') ADVANCE(98);
      if (lookahead == '=') ADVANCE(92);
      if (lookahead == '>') ADVANCE(100);
      if (lookahead == '\\') ADVANCE(36);
      if (lookahead == ']') ADVANCE(68);
      if (lookahead == 'm') ADVANCE(159);
      if (lookahead == 'r') ADVANCE(157);
      if (lookahead == '|') ADVANCE(117);
      if (lookahead == '}') ADVANCE(65);
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(8)
      if (sym_identifier_character_set_2(lookahead)) ADVANCE(168);
      END_STATE();
    case 5:
      if (lookahead == '!') ADVANCE(90);
      if (lookahead == '"') ADVANCE(149);
      if (lookahead == '#') ADVANCE(89);
      if (lookahead == '$') ADVANCE(54);
      if (lookahead == '&') ADVANCE(103);
      if (lookahead == '\'') ADVANCE(88);
      if (lookahead == '(') ADVANCE(62);
      if (lookahead == ')') ADVANCE(63);
      if (lookahead == '*') ADVANCE(80);
      if (lookahead == ',') ADVANCE(96);
      if (lookahead == '-') ADVANCE(114);
      if (lookahead == '.') ADVANCE(21);
      if (lookahead == '/') ADVANCE(24);
      if (lookahead == '0') ADVANCE(144);
      if (lookahead == ':') ADVANCE(31);
      if (lookahead == '<') ADVANCE(98);
      if (lookahead == '>') ADVANCE(100);
      if (lookahead == '[') ADVANCE(67);
      if (lookahead == ']') ADVANCE(68);
      if (lookahead == 'b') ADVANCE(155);
      if (lookahead == 'r') ADVANCE(157);
      if (lookahead == '{') ADVANCE(64);
      if (lookahead == '|') ADVANCE(117);
      if (lookahead == '}') ADVANCE(65);
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(5)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(147);
      if (sym_identifier_character_set_2(lookahead)) ADVANCE(168);
      END_STATE();
    case 6:
      if (lookahead == '!') ADVANCE(90);
      if (lookahead == '"') ADVANCE(149);
      if (lookahead == '#') ADVANCE(89);
      if (lookahead == '$') ADVANCE(54);
      if (lookahead == '&') ADVANCE(103);
      if (lookahead == '\'') ADVANCE(88);
      if (lookahead == '(') ADVANCE(62);
      if (lookahead == ')') ADVANCE(63);
      if (lookahead == '*') ADVANCE(80);
      if (lookahead == ',') ADVANCE(96);
      if (lookahead == '-') ADVANCE(111);
      if (lookahead == '.') ADVANCE(22);
      if (lookahead == '/') ADVANCE(24);
      if (lookahead == '0') ADVANCE(144);
      if (lookahead == ':') ADVANCE(31);
      if (lookahead == '<') ADVANCE(98);
      if (lookahead == '[') ADVANCE(67);
      if (lookahead == 'b') ADVANCE(155);
      if (lookahead == 'r') ADVANCE(157);
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(6)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(147);
      if (sym_identifier_character_set_2(lookahead)) ADVANCE(168);
      END_STATE();
    case 7:
      if (lookahead == '!') ADVANCE(90);
      if (lookahead == '#') ADVANCE(89);
      if (lookahead == '$') ADVANCE(54);
      if (lookahead == '&') ADVANCE(103);
      if (lookahead == '\'') ADVANCE(87);
      if (lookahead == '(') ADVANCE(62);
      if (lookahead == ')') ADVANCE(63);
      if (lookahead == '*') ADVANCE(80);
      if (lookahead == '+') ADVANCE(78);
      if (lookahead == ',') ADVANCE(96);
      if (lookahead == '-') ADVANCE(33);
      if (lookahead == '.') ADVANCE(23);
      if (lookahead == '/') ADVANCE(24);
      if (lookahead == '0') ADVANCE(144);
      if (lookahead == ':') ADVANCE(70);
      if (lookahead == ';') ADVANCE(60);
      if (lookahead == '<') ADVANCE(98);
      if (lookahead == '=') ADVANCE(95);
      if (lookahead == '>') ADVANCE(100);
      if (lookahead == '?') ADVANCE(82);
      if (lookahead == '[') ADVANCE(67);
      if (lookahead == ']') ADVANCE(68);
      if (lookahead == 'r') ADVANCE(157);
      if (lookahead == '{') ADVANCE(64);
      if (lookahead == '|') ADVANCE(117);
      if (lookahead == '}') ADVANCE(65);
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(7)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(147);
      if (sym_identifier_character_set_2(lookahead)) ADVANCE(168);
      END_STATE();
    case 8:
      if (lookahead == '!') ADVANCE(90);
      if (lookahead == '#') ADVANCE(89);
      if (lookahead == '$') ADVANCE(54);
      if (lookahead == '\'') ADVANCE(87);
      if (lookahead == '(') ADVANCE(62);
      if (lookahead == ')') ADVANCE(63);
      if (lookahead == '+') ADVANCE(78);
      if (lookahead == ',') ADVANCE(96);
      if (lookahead == '.') ADVANCE(23);
      if (lookahead == '/') ADVANCE(24);
      if (lookahead == ':') ADVANCE(70);
      if (lookahead == ';') ADVANCE(60);
      if (lookahead == '<') ADVANCE(98);
      if (lookahead == '=') ADVANCE(92);
      if (lookahead == '>') ADVANCE(100);
      if (lookahead == ']') ADVANCE(68);
      if (lookahead == 'm') ADVANCE(159);
      if (lookahead == 'r') ADVANCE(157);
      if (lookahead == '|') ADVANCE(117);
      if (lookahead == '}') ADVANCE(65);
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(8)
      if (sym_identifier_character_set_2(lookahead)) ADVANCE(168);
      END_STATE();
    case 9:
      if (lookahead == '!') ADVANCE(90);
      if (lookahead == '(') ADVANCE(62);
      if (lookahead == ')') ADVANCE(63);
      if (lookahead == '+') ADVANCE(78);
      if (lookahead == ',') ADVANCE(96);
      if (lookahead == '.') ADVANCE(23);
      if (lookahead == '/') ADVANCE(24);
      if (lookahead == ':') ADVANCE(70);
      if (lookahead == ';') ADVANCE(60);
      if (lookahead == '<') ADVANCE(106);
      if (lookahead == '=') ADVANCE(95);
      if (lookahead == '>') ADVANCE(100);
      if (lookahead == '@') ADVANCE(139);
      if (lookahead == ']') ADVANCE(68);
      if (lookahead == 'r') ADVANCE(157);
      if (lookahead == '{') ADVANCE(64);
      if (lookahead == '|') ADVANCE(117);
      if (lookahead == '}') ADVANCE(65);
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(9)
      if (sym_identifier_character_set_2(lookahead)) ADVANCE(168);
      END_STATE();
    case 10:
      if (lookahead == '!') ADVANCE(32);
      if (lookahead == '%') ADVANCE(127);
      if (lookahead == '&') ADVANCE(104);
      if (lookahead == '(') ADVANCE(62);
      if (lookahead == ')') ADVANCE(63);
      if (lookahead == '*') ADVANCE(81);
      if (lookahead == '+') ADVANCE(79);
      if (lookahead == ',') ADVANCE(96);
      if (lookahead == '-') ADVANCE(112);
      if (lookahead == '.') ADVANCE(138);
      if (lookahead == '/') ADVANCE(126);
      if (lookahead == ':') ADVANCE(69);
      if (lookahead == ';') ADVANCE(60);
      if (lookahead == '<') ADVANCE(99);
      if (lookahead == '=') ADVANCE(94);
      if (lookahead == '>') ADVANCE(101);
      if (lookahead == '?') ADVANCE(82);
      if (lookahead == '[') ADVANCE(67);
      if (lookahead == ']') ADVANCE(68);
      if (lookahead == '^') ADVANCE(119);
      if (lookahead == 'r') ADVANCE(157);
      if (lookahead == '{') ADVANCE(64);
      if (lookahead == '|') ADVANCE(118);
      if (lookahead == '}') ADVANCE(65);
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(10)
      if (sym_identifier_character_set_2(lookahead)) ADVANCE(168);
      END_STATE();
    case 11:
      if (lookahead == '"') ADVANCE(149);
      if (lookahead == '$') ADVANCE(73);
      if (lookahead == '\'') ADVANCE(88);
      if (lookahead == '(') ADVANCE(62);
      if (lookahead == ')') ADVANCE(63);
      if (lookahead == '/') ADVANCE(83);
      if (lookahead == '0') ADVANCE(144);
      if (lookahead == ':') ADVANCE(71);
      if (lookahead == '[') ADVANCE(67);
      if (lookahead == ']') ADVANCE(68);
      if (lookahead == '_') ADVANCE(84);
      if (lookahead == 'b') ADVANCE(155);
      if (lookahead == 'r') ADVANCE(157);
      if (lookahead == '{') ADVANCE(64);
      if (lookahead == '}') ADVANCE(65);
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(11)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(147);
      if (('!' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          ('|' <= lookahead && lookahead <= '~')) ADVANCE(85);
      if (sym_identifier_character_set_3(lookahead)) ADVANCE(168);
      END_STATE();
    case 12:
      if (lookahead == '"') ADVANCE(149);
      if (lookahead == '$') ADVANCE(73);
      if (lookahead == '\'') ADVANCE(88);
      if (lookahead == '(') ADVANCE(62);
      if (lookahead == ')') ADVANCE(63);
      if (lookahead == '/') ADVANCE(83);
      if (lookahead == '0') ADVANCE(144);
      if (lookahead == '[') ADVANCE(67);
      if (lookahead == ']') ADVANCE(68);
      if (lookahead == '_') ADVANCE(84);
      if (lookahead == 'b') ADVANCE(155);
      if (lookahead == 'r') ADVANCE(157);
      if (lookahead == '{') ADVANCE(64);
      if (lookahead == '}') ADVANCE(65);
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(12)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(147);
      if (('!' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          ('|' <= lookahead && lookahead <= '~')) ADVANCE(85);
      if (sym_identifier_character_set_3(lookahead)) ADVANCE(168);
      END_STATE();
    case 13:
      if (lookahead == '"') ADVANCE(149);
      if (lookahead == '$') ADVANCE(72);
      if (lookahead == '\'') ADVANCE(88);
      if (lookahead == '(') ADVANCE(62);
      if (lookahead == ')') ADVANCE(63);
      if (lookahead == '/') ADVANCE(83);
      if (lookahead == '0') ADVANCE(144);
      if (lookahead == '[') ADVANCE(67);
      if (lookahead == ']') ADVANCE(68);
      if (lookahead == '_') ADVANCE(84);
      if (lookahead == 'b') ADVANCE(155);
      if (lookahead == 'r') ADVANCE(157);
      if (lookahead == '{') ADVANCE(64);
      if (lookahead == '}') ADVANCE(65);
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(13)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(147);
      if (('!' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          ('|' <= lookahead && lookahead <= '~')) ADVANCE(85);
      if (sym_identifier_character_set_3(lookahead)) ADVANCE(168);
      END_STATE();
    case 14:
      if (lookahead == '"') ADVANCE(149);
      if (lookahead == '/') ADVANCE(24);
      if (lookahead == ':') ADVANCE(69);
      if (lookahead == ';') ADVANCE(60);
      if (lookahead == '<') ADVANCE(98);
      if (lookahead == '=') ADVANCE(92);
      if (lookahead == 'b') ADVANCE(156);
      if (lookahead == 'r') ADVANCE(157);
      if (lookahead == '{') ADVANCE(64);
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(14)
      if (sym_identifier_character_set_2(lookahead)) ADVANCE(168);
      END_STATE();
    case 15:
      if (lookahead == '#') ADVANCE(89);
      if (lookahead == ',') ADVANCE(96);
      if (lookahead == '.') ADVANCE(21);
      if (lookahead == '/') ADVANCE(24);
      if (lookahead == ':') ADVANCE(69);
      if (lookahead == '<') ADVANCE(98);
      if (lookahead == 'r') ADVANCE(157);
      if (lookahead == '{') ADVANCE(64);
      if (lookahead == '}') ADVANCE(65);
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(15)
      if (sym_identifier_character_set_2(lookahead)) ADVANCE(168);
      END_STATE();
    case 16:
      if (lookahead == '\'') ADVANCE(151);
      END_STATE();
    case 17:
      if (lookahead == '\'') ADVANCE(151);
      if (lookahead == '\\') ADVANCE(37);
      if (lookahead != 0) ADVANCE(16);
      END_STATE();
    case 18:
      if (lookahead == '(') ADVANCE(62);
      if (lookahead == ')') ADVANCE(63);
      if (lookahead == '*') ADVANCE(80);
      if (lookahead == '+') ADVANCE(78);
      if (lookahead == ',') ADVANCE(96);
      if (lookahead == '-') ADVANCE(33);
      if (lookahead == '.') ADVANCE(23);
      if (lookahead == '/') ADVANCE(24);
      if (lookahead == ':') ADVANCE(69);
      if (lookahead == ';') ADVANCE(60);
      if (lookahead == '<') ADVANCE(106);
      if (lookahead == '=') ADVANCE(95);
      if (lookahead == '>') ADVANCE(100);
      if (lookahead == ']') ADVANCE(68);
      if (lookahead == 'r') ADVANCE(157);
      if (lookahead == '{') ADVANCE(64);
      if (lookahead == '|') ADVANCE(117);
      if (lookahead == '}') ADVANCE(65);
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(18)
      if (sym_identifier_character_set_2(lookahead)) ADVANCE(168);
      END_STATE();
    case 19:
      if (lookahead == '*') ADVANCE(80);
      if (lookahead == '+') ADVANCE(78);
      if (lookahead == '/') ADVANCE(76);
      if (lookahead == '?') ADVANCE(82);
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == '\r' ||
          lookahead == ' ') ADVANCE(75);
      if (lookahead != 0) ADVANCE(77);
      END_STATE();
    case 20:
      if (lookahead == '.') ADVANCE(105);
      if (lookahead == '=') ADVANCE(110);
      END_STATE();
    case 21:
      if (lookahead == '.') ADVANCE(107);
      END_STATE();
    case 22:
      if (lookahead == '.') ADVANCE(108);
      END_STATE();
    case 23:
      if (lookahead == '.') ADVANCE(20);
      END_STATE();
    case 24:
      if (lookahead == '/') ADVANCE(153);
      END_STATE();
    case 25:
      if (lookahead == '1') ADVANCE(27);
      if (lookahead == '3') ADVANCE(26);
      if (lookahead == '6') ADVANCE(29);
      if (lookahead == '8') ADVANCE(140);
      if (lookahead == 's') ADVANCE(35);
      END_STATE();
    case 26:
      if (lookahead == '2') ADVANCE(140);
      END_STATE();
    case 27:
      if (lookahead == '2') ADVANCE(30);
      if (lookahead == '6') ADVANCE(140);
      END_STATE();
    case 28:
      if (lookahead == '3') ADVANCE(26);
      if (lookahead == '6') ADVANCE(29);
      END_STATE();
    case 29:
      if (lookahead == '4') ADVANCE(140);
      END_STATE();
    case 30:
      if (lookahead == '8') ADVANCE(140);
      END_STATE();
    case 31:
      if (lookahead == ':') ADVANCE(102);
      END_STATE();
    case 32:
      if (lookahead == '=') ADVANCE(121);
      END_STATE();
    case 33:
      if (lookahead == '>') ADVANCE(97);
      END_STATE();
    case 34:
      if (lookahead == 'e') ADVANCE(140);
      END_STATE();
    case 35:
      if (lookahead == 'i') ADVANCE(38);
      END_STATE();
    case 36:
      if (lookahead == 'u') ADVANCE(39);
      if (lookahead == 'x') ADVANCE(50);
      if (lookahead != 0) ADVANCE(152);
      END_STATE();
    case 37:
      if (lookahead == 'u') ADVANCE(40);
      if (lookahead == 'x') ADVANCE(51);
      if (lookahead != 0) ADVANCE(16);
      END_STATE();
    case 38:
      if (lookahead == 'z') ADVANCE(34);
      END_STATE();
    case 39:
      if (lookahead == '{') ADVANCE(48);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(46);
      END_STATE();
    case 40:
      if (lookahead == '{') ADVANCE(49);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(52);
      END_STATE();
    case 41:
      if (lookahead == '}') ADVANCE(16);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(41);
      END_STATE();
    case 42:
      if (lookahead == '}') ADVANCE(152);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(42);
      END_STATE();
    case 43:
      if (lookahead == '0' ||
          lookahead == '1' ||
          lookahead == '_') ADVANCE(145);
      END_STATE();
    case 44:
      if (('0' <= lookahead && lookahead <= '7') ||
          lookahead == '_') ADVANCE(146);
      END_STATE();
    case 45:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(16);
      END_STATE();
    case 46:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(50);
      END_STATE();
    case 47:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(152);
      END_STATE();
    case 48:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(42);
      END_STATE();
    case 49:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(41);
      END_STATE();
    case 50:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(47);
      END_STATE();
    case 51:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(45);
      END_STATE();
    case 52:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(51);
      END_STATE();
    case 53:
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'f')) ADVANCE(148);
      END_STATE();
    case 54:
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(169);
      END_STATE();
    case 55:
      if (sym_identifier_character_set_2(lookahead)) ADVANCE(168);
      END_STATE();
    case 56:
      if (eof) ADVANCE(59);
      if (lookahead == '!') ADVANCE(91);
      if (lookahead == '"') ADVANCE(149);
      if (lookahead == '#') ADVANCE(89);
      if (lookahead == '$') ADVANCE(73);
      if (lookahead == '%') ADVANCE(127);
      if (lookahead == '&') ADVANCE(104);
      if (lookahead == '\'') ADVANCE(88);
      if (lookahead == '(') ADVANCE(62);
      if (lookahead == ')') ADVANCE(63);
      if (lookahead == '*') ADVANCE(81);
      if (lookahead == '+') ADVANCE(79);
      if (lookahead == ',') ADVANCE(96);
      if (lookahead == '-') ADVANCE(113);
      if (lookahead == '.') ADVANCE(138);
      if (lookahead == '/') ADVANCE(126);
      if (lookahead == '0') ADVANCE(144);
      if (lookahead == ':') ADVANCE(70);
      if (lookahead == ';') ADVANCE(60);
      if (lookahead == '<') ADVANCE(106);
      if (lookahead == '=') ADVANCE(94);
      if (lookahead == '>') ADVANCE(101);
      if (lookahead == '?') ADVANCE(82);
      if (lookahead == '@') ADVANCE(139);
      if (lookahead == '[') ADVANCE(67);
      if (lookahead == ']') ADVANCE(68);
      if (lookahead == '^') ADVANCE(119);
      if (lookahead == 'b') ADVANCE(155);
      if (lookahead == 'm') ADVANCE(159);
      if (lookahead == 'r') ADVANCE(157);
      if (lookahead == '{') ADVANCE(64);
      if (lookahead == '|') ADVANCE(118);
      if (lookahead == '}') ADVANCE(65);
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(56)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(147);
      if (sym_identifier_character_set_2(lookahead)) ADVANCE(168);
      END_STATE();
    case 57:
      if (eof) ADVANCE(59);
      if (lookahead == '!') ADVANCE(91);
      if (lookahead == '"') ADVANCE(149);
      if (lookahead == '#') ADVANCE(89);
      if (lookahead == '$') ADVANCE(54);
      if (lookahead == '%') ADVANCE(127);
      if (lookahead == '&') ADVANCE(104);
      if (lookahead == '\'') ADVANCE(88);
      if (lookahead == '(') ADVANCE(62);
      if (lookahead == '*') ADVANCE(81);
      if (lookahead == '+') ADVANCE(79);
      if (lookahead == '-') ADVANCE(112);
      if (lookahead == '.') ADVANCE(138);
      if (lookahead == '/') ADVANCE(126);
      if (lookahead == '0') ADVANCE(144);
      if (lookahead == ':') ADVANCE(31);
      if (lookahead == ';') ADVANCE(60);
      if (lookahead == '<') ADVANCE(99);
      if (lookahead == '=') ADVANCE(93);
      if (lookahead == '>') ADVANCE(101);
      if (lookahead == '?') ADVANCE(82);
      if (lookahead == '[') ADVANCE(67);
      if (lookahead == '^') ADVANCE(119);
      if (lookahead == 'b') ADVANCE(155);
      if (lookahead == 'm') ADVANCE(159);
      if (lookahead == 'r') ADVANCE(157);
      if (lookahead == '{') ADVANCE(64);
      if (lookahead == '|') ADVANCE(118);
      if (lookahead == '}') ADVANCE(65);
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(57)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(147);
      if (sym_identifier_character_set_2(lookahead)) ADVANCE(168);
      END_STATE();
    case 58:
      if (eof) ADVANCE(59);
      if (lookahead == '!') ADVANCE(90);
      if (lookahead == '"') ADVANCE(149);
      if (lookahead == '#') ADVANCE(89);
      if (lookahead == '$') ADVANCE(54);
      if (lookahead == '&') ADVANCE(103);
      if (lookahead == '\'') ADVANCE(88);
      if (lookahead == '(') ADVANCE(62);
      if (lookahead == ')') ADVANCE(63);
      if (lookahead == '*') ADVANCE(80);
      if (lookahead == '+') ADVANCE(78);
      if (lookahead == ',') ADVANCE(96);
      if (lookahead == '-') ADVANCE(114);
      if (lookahead == '.') ADVANCE(21);
      if (lookahead == '/') ADVANCE(24);
      if (lookahead == '0') ADVANCE(144);
      if (lookahead == ':') ADVANCE(31);
      if (lookahead == ';') ADVANCE(60);
      if (lookahead == '<') ADVANCE(98);
      if (lookahead == '=') ADVANCE(92);
      if (lookahead == '>') ADVANCE(100);
      if (lookahead == '?') ADVANCE(82);
      if (lookahead == '[') ADVANCE(67);
      if (lookahead == ']') ADVANCE(68);
      if (lookahead == 'b') ADVANCE(155);
      if (lookahead == 'm') ADVANCE(159);
      if (lookahead == 'r') ADVANCE(157);
      if (lookahead == '{') ADVANCE(64);
      if (lookahead == '|') ADVANCE(117);
      if (lookahead == '}') ADVANCE(65);
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(58)
      if (('1' <= lookahead && lookahead <= '9')) ADVANCE(147);
      if (sym_identifier_character_set_2(lookahead)) ADVANCE(168);
      END_STATE();
    case 59:
      ACCEPT_TOKEN(ts_builtin_sym_end);
      END_STATE();
    case 60:
      ACCEPT_TOKEN(anon_sym_SEMI);
      END_STATE();
    case 61:
      ACCEPT_TOKEN(anon_sym_macro_rules_BANG);
      END_STATE();
    case 62:
      ACCEPT_TOKEN(anon_sym_LPAREN);
      END_STATE();
    case 63:
      ACCEPT_TOKEN(anon_sym_RPAREN);
      END_STATE();
    case 64:
      ACCEPT_TOKEN(anon_sym_LBRACE);
      END_STATE();
    case 65:
      ACCEPT_TOKEN(anon_sym_RBRACE);
      END_STATE();
    case 66:
      ACCEPT_TOKEN(anon_sym_EQ_GT);
      END_STATE();
    case 67:
      ACCEPT_TOKEN(anon_sym_LBRACK);
      END_STATE();
    case 68:
      ACCEPT_TOKEN(anon_sym_RBRACK);
      END_STATE();
    case 69:
      ACCEPT_TOKEN(anon_sym_COLON);
      END_STATE();
    case 70:
      ACCEPT_TOKEN(anon_sym_COLON);
      if (lookahead == ':') ADVANCE(102);
      END_STATE();
    case 71:
      ACCEPT_TOKEN(anon_sym_COLON);
      if (lookahead == '!' ||
          lookahead == '#' ||
          lookahead == '%' ||
          lookahead == '&' ||
          ('*' <= lookahead && lookahead <= '/') ||
          (':' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '_' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(85);
      END_STATE();
    case 72:
      ACCEPT_TOKEN(anon_sym_DOLLAR);
      END_STATE();
    case 73:
      ACCEPT_TOKEN(anon_sym_DOLLAR);
      if (('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(169);
      END_STATE();
    case 74:
      ACCEPT_TOKEN(aux_sym_token_repetition_pattern_token1);
      if (lookahead == '\n') ADVANCE(77);
      if (lookahead == '*' ||
          lookahead == '+' ||
          lookahead == '?') ADVANCE(153);
      if (lookahead != 0) ADVANCE(74);
      END_STATE();
    case 75:
      ACCEPT_TOKEN(aux_sym_token_repetition_pattern_token1);
      if (lookahead == '/') ADVANCE(76);
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == '\r' ||
          lookahead == ' ') ADVANCE(75);
      if (lookahead != 0 &&
          lookahead != '*' &&
          lookahead != '+' &&
          lookahead != '?') ADVANCE(77);
      END_STATE();
    case 76:
      ACCEPT_TOKEN(aux_sym_token_repetition_pattern_token1);
      if (lookahead == '/') ADVANCE(74);
      if (lookahead != 0 &&
          lookahead != '*' &&
          lookahead != '+' &&
          lookahead != '?') ADVANCE(77);
      END_STATE();
    case 77:
      ACCEPT_TOKEN(aux_sym_token_repetition_pattern_token1);
      if (lookahead != 0 &&
          lookahead != '*' &&
          lookahead != '+' &&
          lookahead != '?') ADVANCE(77);
      END_STATE();
    case 78:
      ACCEPT_TOKEN(anon_sym_PLUS);
      END_STATE();
    case 79:
      ACCEPT_TOKEN(anon_sym_PLUS);
      if (lookahead == '=') ADVANCE(128);
      END_STATE();
    case 80:
      ACCEPT_TOKEN(anon_sym_STAR);
      END_STATE();
    case 81:
      ACCEPT_TOKEN(anon_sym_STAR);
      if (lookahead == '=') ADVANCE(130);
      END_STATE();
    case 82:
      ACCEPT_TOKEN(anon_sym_QMARK);
      END_STATE();
    case 83:
      ACCEPT_TOKEN(aux_sym__non_special_token_token1);
      if (lookahead == '/') ADVANCE(86);
      if (lookahead == '!' ||
          lookahead == '#' ||
          lookahead == '%' ||
          lookahead == '&' ||
          ('*' <= lookahead && lookahead <= '.') ||
          (':' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '_' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(85);
      END_STATE();
    case 84:
      ACCEPT_TOKEN(aux_sym__non_special_token_token1);
      if (lookahead == '_') ADVANCE(84);
      if (lookahead == '!' ||
          lookahead == '#' ||
          lookahead == '%' ||
          lookahead == '&' ||
          ('*' <= lookahead && lookahead <= '/') ||
          (':' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(85);
      if (sym_identifier_character_set_4(lookahead)) ADVANCE(168);
      END_STATE();
    case 85:
      ACCEPT_TOKEN(aux_sym__non_special_token_token1);
      if (lookahead == '!' ||
          lookahead == '#' ||
          lookahead == '%' ||
          lookahead == '&' ||
          ('*' <= lookahead && lookahead <= '/') ||
          (':' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '_' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(85);
      END_STATE();
    case 86:
      ACCEPT_TOKEN(aux_sym__non_special_token_token1);
      if (lookahead == '!' ||
          lookahead == '#' ||
          lookahead == '%' ||
          lookahead == '&' ||
          ('*' <= lookahead && lookahead <= '/') ||
          (':' <= lookahead && lookahead <= '@') ||
          lookahead == '^' ||
          lookahead == '_' ||
          lookahead == '|' ||
          lookahead == '~') ADVANCE(86);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(153);
      END_STATE();
    case 87:
      ACCEPT_TOKEN(anon_sym_SQUOTE);
      END_STATE();
    case 88:
      ACCEPT_TOKEN(anon_sym_SQUOTE);
      if (lookahead == '\'') ADVANCE(151);
      if (lookahead == '\\') ADVANCE(37);
      if (lookahead != 0) ADVANCE(16);
      END_STATE();
    case 89:
      ACCEPT_TOKEN(anon_sym_POUND);
      END_STATE();
    case 90:
      ACCEPT_TOKEN(anon_sym_BANG);
      END_STATE();
    case 91:
      ACCEPT_TOKEN(anon_sym_BANG);
      if (lookahead == '=') ADVANCE(121);
      END_STATE();
    case 92:
      ACCEPT_TOKEN(anon_sym_EQ);
      END_STATE();
    case 93:
      ACCEPT_TOKEN(anon_sym_EQ);
      if (lookahead == '=') ADVANCE(120);
      END_STATE();
    case 94:
      ACCEPT_TOKEN(anon_sym_EQ);
      if (lookahead == '=') ADVANCE(120);
      if (lookahead == '>') ADVANCE(66);
      END_STATE();
    case 95:
      ACCEPT_TOKEN(anon_sym_EQ);
      if (lookahead == '>') ADVANCE(66);
      END_STATE();
    case 96:
      ACCEPT_TOKEN(anon_sym_COMMA);
      END_STATE();
    case 97:
      ACCEPT_TOKEN(anon_sym_DASH_GT);
      END_STATE();
    case 98:
      ACCEPT_TOKEN(anon_sym_LT);
      END_STATE();
    case 99:
      ACCEPT_TOKEN(anon_sym_LT);
      if (lookahead == '<') ADVANCE(124);
      if (lookahead == '=') ADVANCE(122);
      END_STATE();
    case 100:
      ACCEPT_TOKEN(anon_sym_GT);
      END_STATE();
    case 101:
      ACCEPT_TOKEN(anon_sym_GT);
      if (lookahead == '=') ADVANCE(123);
      if (lookahead == '>') ADVANCE(125);
      END_STATE();
    case 102:
      ACCEPT_TOKEN(anon_sym_COLON_COLON);
      END_STATE();
    case 103:
      ACCEPT_TOKEN(anon_sym_AMP);
      END_STATE();
    case 104:
      ACCEPT_TOKEN(anon_sym_AMP);
      if (lookahead == '&') ADVANCE(115);
      if (lookahead == '=') ADVANCE(133);
      END_STATE();
    case 105:
      ACCEPT_TOKEN(anon_sym_DOT_DOT_DOT);
      END_STATE();
    case 106:
      ACCEPT_TOKEN(anon_sym_LT2);
      END_STATE();
    case 107:
      ACCEPT_TOKEN(anon_sym_DOT_DOT);
      END_STATE();
    case 108:
      ACCEPT_TOKEN(anon_sym_DOT_DOT);
      if (lookahead == '.') ADVANCE(105);
      END_STATE();
    case 109:
      ACCEPT_TOKEN(anon_sym_DOT_DOT);
      if (lookahead == '.') ADVANCE(105);
      if (lookahead == '=') ADVANCE(110);
      END_STATE();
    case 110:
      ACCEPT_TOKEN(anon_sym_DOT_DOT_EQ);
      END_STATE();
    case 111:
      ACCEPT_TOKEN(anon_sym_DASH);
      END_STATE();
    case 112:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '=') ADVANCE(129);
      END_STATE();
    case 113:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '=') ADVANCE(129);
      if (lookahead == '>') ADVANCE(97);
      END_STATE();
    case 114:
      ACCEPT_TOKEN(anon_sym_DASH);
      if (lookahead == '>') ADVANCE(97);
      END_STATE();
    case 115:
      ACCEPT_TOKEN(anon_sym_AMP_AMP);
      END_STATE();
    case 116:
      ACCEPT_TOKEN(anon_sym_PIPE_PIPE);
      END_STATE();
    case 117:
      ACCEPT_TOKEN(anon_sym_PIPE);
      END_STATE();
    case 118:
      ACCEPT_TOKEN(anon_sym_PIPE);
      if (lookahead == '=') ADVANCE(134);
      if (lookahead == '|') ADVANCE(116);
      END_STATE();
    case 119:
      ACCEPT_TOKEN(anon_sym_CARET);
      if (lookahead == '=') ADVANCE(135);
      END_STATE();
    case 120:
      ACCEPT_TOKEN(anon_sym_EQ_EQ);
      END_STATE();
    case 121:
      ACCEPT_TOKEN(anon_sym_BANG_EQ);
      END_STATE();
    case 122:
      ACCEPT_TOKEN(anon_sym_LT_EQ);
      END_STATE();
    case 123:
      ACCEPT_TOKEN(anon_sym_GT_EQ);
      END_STATE();
    case 124:
      ACCEPT_TOKEN(anon_sym_LT_LT);
      if (lookahead == '=') ADVANCE(136);
      END_STATE();
    case 125:
      ACCEPT_TOKEN(anon_sym_GT_GT);
      if (lookahead == '=') ADVANCE(137);
      END_STATE();
    case 126:
      ACCEPT_TOKEN(anon_sym_SLASH);
      if (lookahead == '/') ADVANCE(153);
      if (lookahead == '=') ADVANCE(131);
      END_STATE();
    case 127:
      ACCEPT_TOKEN(anon_sym_PERCENT);
      if (lookahead == '=') ADVANCE(132);
      END_STATE();
    case 128:
      ACCEPT_TOKEN(anon_sym_PLUS_EQ);
      END_STATE();
    case 129:
      ACCEPT_TOKEN(anon_sym_DASH_EQ);
      END_STATE();
    case 130:
      ACCEPT_TOKEN(anon_sym_STAR_EQ);
      END_STATE();
    case 131:
      ACCEPT_TOKEN(anon_sym_SLASH_EQ);
      END_STATE();
    case 132:
      ACCEPT_TOKEN(anon_sym_PERCENT_EQ);
      END_STATE();
    case 133:
      ACCEPT_TOKEN(anon_sym_AMP_EQ);
      END_STATE();
    case 134:
      ACCEPT_TOKEN(anon_sym_PIPE_EQ);
      END_STATE();
    case 135:
      ACCEPT_TOKEN(anon_sym_CARET_EQ);
      END_STATE();
    case 136:
      ACCEPT_TOKEN(anon_sym_LT_LT_EQ);
      END_STATE();
    case 137:
      ACCEPT_TOKEN(anon_sym_GT_GT_EQ);
      END_STATE();
    case 138:
      ACCEPT_TOKEN(anon_sym_DOT);
      if (lookahead == '.') ADVANCE(109);
      END_STATE();
    case 139:
      ACCEPT_TOKEN(anon_sym_AT);
      END_STATE();
    case 140:
      ACCEPT_TOKEN(sym_integer_literal);
      END_STATE();
    case 141:
      ACCEPT_TOKEN(sym_integer_literal);
      if (lookahead == '2') ADVANCE(148);
      if (lookahead == 'f') ADVANCE(142);
      if (lookahead == 'i') ADVANCE(25);
      if (lookahead == 'u') ADVANCE(25);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'e')) ADVANCE(148);
      END_STATE();
    case 142:
      ACCEPT_TOKEN(sym_integer_literal);
      if (lookahead == '3') ADVANCE(141);
      if (lookahead == '6') ADVANCE(143);
      if (lookahead == 'f') ADVANCE(142);
      if (lookahead == 'i') ADVANCE(25);
      if (lookahead == 'u') ADVANCE(25);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'e')) ADVANCE(148);
      END_STATE();
    case 143:
      ACCEPT_TOKEN(sym_integer_literal);
      if (lookahead == '4') ADVANCE(148);
      if (lookahead == 'f') ADVANCE(142);
      if (lookahead == 'i') ADVANCE(25);
      if (lookahead == 'u') ADVANCE(25);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'e')) ADVANCE(148);
      END_STATE();
    case 144:
      ACCEPT_TOKEN(sym_integer_literal);
      if (lookahead == 'b') ADVANCE(43);
      if (lookahead == 'f') ADVANCE(28);
      if (lookahead == 'i') ADVANCE(25);
      if (lookahead == 'o') ADVANCE(44);
      if (lookahead == 'u') ADVANCE(25);
      if (lookahead == 'x') ADVANCE(53);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_') ADVANCE(147);
      END_STATE();
    case 145:
      ACCEPT_TOKEN(sym_integer_literal);
      if (lookahead == 'f') ADVANCE(28);
      if (lookahead == 'i') ADVANCE(25);
      if (lookahead == 'u') ADVANCE(25);
      if (lookahead == '0' ||
          lookahead == '1' ||
          lookahead == '_') ADVANCE(145);
      END_STATE();
    case 146:
      ACCEPT_TOKEN(sym_integer_literal);
      if (lookahead == 'f') ADVANCE(28);
      if (lookahead == 'i') ADVANCE(25);
      if (lookahead == 'u') ADVANCE(25);
      if (('0' <= lookahead && lookahead <= '7') ||
          lookahead == '_') ADVANCE(146);
      END_STATE();
    case 147:
      ACCEPT_TOKEN(sym_integer_literal);
      if (lookahead == 'f') ADVANCE(28);
      if (lookahead == 'i') ADVANCE(25);
      if (lookahead == 'u') ADVANCE(25);
      if (('0' <= lookahead && lookahead <= '9') ||
          lookahead == '_') ADVANCE(147);
      END_STATE();
    case 148:
      ACCEPT_TOKEN(sym_integer_literal);
      if (lookahead == 'f') ADVANCE(142);
      if (lookahead == 'i') ADVANCE(25);
      if (lookahead == 'u') ADVANCE(25);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'F') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'e')) ADVANCE(148);
      END_STATE();
    case 149:
      ACCEPT_TOKEN(aux_sym_string_literal_token1);
      END_STATE();
    case 150:
      ACCEPT_TOKEN(anon_sym_DQUOTE);
      END_STATE();
    case 151:
      ACCEPT_TOKEN(sym_char_literal);
      END_STATE();
    case 152:
      ACCEPT_TOKEN(sym_escape_sequence);
      END_STATE();
    case 153:
      ACCEPT_TOKEN(sym_line_comment);
      if (lookahead != 0 &&
          lookahead != '\n') ADVANCE(153);
      END_STATE();
    case 154:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '!') ADVANCE(61);
      if (sym_identifier_character_set_5(lookahead)) ADVANCE(168);
      END_STATE();
    case 155:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '"') ADVANCE(149);
      if (lookahead == '\'') ADVANCE(17);
      if (sym_identifier_character_set_5(lookahead)) ADVANCE(168);
      END_STATE();
    case 156:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '"') ADVANCE(149);
      if (sym_identifier_character_set_5(lookahead)) ADVANCE(168);
      END_STATE();
    case 157:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '#') ADVANCE(55);
      if (sym_identifier_character_set_5(lookahead)) ADVANCE(168);
      END_STATE();
    case 158:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == '_') ADVANCE(165);
      if (sym_identifier_character_set_6(lookahead)) ADVANCE(168);
      END_STATE();
    case 159:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'a') ADVANCE(160);
      if (sym_identifier_character_set_7(lookahead)) ADVANCE(168);
      END_STATE();
    case 160:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'c') ADVANCE(164);
      if (sym_identifier_character_set_5(lookahead)) ADVANCE(168);
      END_STATE();
    case 161:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'e') ADVANCE(166);
      if (sym_identifier_character_set_5(lookahead)) ADVANCE(168);
      END_STATE();
    case 162:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'l') ADVANCE(161);
      if (sym_identifier_character_set_5(lookahead)) ADVANCE(168);
      END_STATE();
    case 163:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'o') ADVANCE(158);
      if (sym_identifier_character_set_5(lookahead)) ADVANCE(168);
      END_STATE();
    case 164:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'r') ADVANCE(163);
      if (sym_identifier_character_set_5(lookahead)) ADVANCE(168);
      END_STATE();
    case 165:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'r') ADVANCE(167);
      if (sym_identifier_character_set_5(lookahead)) ADVANCE(168);
      END_STATE();
    case 166:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 's') ADVANCE(154);
      if (sym_identifier_character_set_5(lookahead)) ADVANCE(168);
      END_STATE();
    case 167:
      ACCEPT_TOKEN(sym_identifier);
      if (lookahead == 'u') ADVANCE(162);
      if (sym_identifier_character_set_5(lookahead)) ADVANCE(168);
      END_STATE();
    case 168:
      ACCEPT_TOKEN(sym_identifier);
      if (sym_identifier_character_set_5(lookahead)) ADVANCE(168);
      END_STATE();
    case 169:
      ACCEPT_TOKEN(sym_metavariable);
      if (('0' <= lookahead && lookahead <= '9') ||
          ('A' <= lookahead && lookahead <= 'Z') ||
          lookahead == '_' ||
          ('a' <= lookahead && lookahead <= 'z')) ADVANCE(169);
      END_STATE();
    default:
      return false;
  }
}

static bool ts_lex_keywords(TSLexer *lexer, TSStateId state) {
  START_LEXER();
  eof = lexer->eof(lexer);
  switch (state) {
    case 0:
      if (lookahead == '_') ADVANCE(1);
      if (lookahead == 'a') ADVANCE(2);
      if (lookahead == 'b') ADVANCE(3);
      if (lookahead == 'c') ADVANCE(4);
      if (lookahead == 'd') ADVANCE(5);
      if (lookahead == 'e') ADVANCE(6);
      if (lookahead == 'f') ADVANCE(7);
      if (lookahead == 'i') ADVANCE(8);
      if (lookahead == 'l') ADVANCE(9);
      if (lookahead == 'm') ADVANCE(10);
      if (lookahead == 'p') ADVANCE(11);
      if (lookahead == 'r') ADVANCE(12);
      if (lookahead == 's') ADVANCE(13);
      if (lookahead == 't') ADVANCE(14);
      if (lookahead == 'u') ADVANCE(15);
      if (lookahead == 'v') ADVANCE(16);
      if (lookahead == 'w') ADVANCE(17);
      if (lookahead == 'y') ADVANCE(18);
      if (lookahead == '\t' ||
          lookahead == '\n' ||
          lookahead == '\r' ||
          lookahead == ' ') SKIP(0)
      END_STATE();
    case 1:
      ACCEPT_TOKEN(anon_sym__);
      END_STATE();
    case 2:
      if (lookahead == 's') ADVANCE(19);
      if (lookahead == 'w') ADVANCE(20);
      END_STATE();
    case 3:
      if (lookahead == 'l') ADVANCE(21);
      if (lookahead == 'o') ADVANCE(22);
      if (lookahead == 'r') ADVANCE(23);
      END_STATE();
    case 4:
      if (lookahead == 'h') ADVANCE(24);
      if (lookahead == 'o') ADVANCE(25);
      if (lookahead == 'r') ADVANCE(26);
      END_STATE();
    case 5:
      if (lookahead == 'e') ADVANCE(27);
      if (lookahead == 'y') ADVANCE(28);
      END_STATE();
    case 6:
      if (lookahead == 'l') ADVANCE(29);
      if (lookahead == 'n') ADVANCE(30);
      if (lookahead == 'x') ADVANCE(31);
      END_STATE();
    case 7:
      if (lookahead == '3') ADVANCE(32);
      if (lookahead == '6') ADVANCE(33);
      if (lookahead == 'a') ADVANCE(34);
      if (lookahead == 'n') ADVANCE(35);
      if (lookahead == 'o') ADVANCE(36);
      END_STATE();
    case 8:
      if (lookahead == '1') ADVANCE(37);
      if (lookahead == '3') ADVANCE(38);
      if (lookahead == '6') ADVANCE(39);
      if (lookahead == '8') ADVANCE(40);
      if (lookahead == 'd') ADVANCE(41);
      if (lookahead == 'f') ADVANCE(42);
      if (lookahead == 'm') ADVANCE(43);
      if (lookahead == 'n') ADVANCE(44);
      if (lookahead == 's') ADVANCE(45);
      if (lookahead == 't') ADVANCE(46);
      END_STATE();
    case 9:
      if (lookahead == 'e') ADVANCE(47);
      if (lookahead == 'i') ADVANCE(48);
      if (lookahead == 'o') ADVANCE(49);
      END_STATE();
    case 10:
      if (lookahead == 'a') ADVANCE(50);
      if (lookahead == 'e') ADVANCE(51);
      if (lookahead == 'o') ADVANCE(52);
      if (lookahead == 'u') ADVANCE(53);
      END_STATE();
    case 11:
      if (lookahead == 'a') ADVANCE(54);
      if (lookahead == 'u') ADVANCE(55);
      END_STATE();
    case 12:
      if (lookahead == 'e') ADVANCE(56);
      END_STATE();
    case 13:
      if (lookahead == 'e') ADVANCE(57);
      if (lookahead == 't') ADVANCE(58);
      if (lookahead == 'u') ADVANCE(59);
      END_STATE();
    case 14:
      if (lookahead == 'r') ADVANCE(60);
      if (lookahead == 't') ADVANCE(61);
      if (lookahead == 'y') ADVANCE(62);
      END_STATE();
    case 15:
      if (lookahead == '1') ADVANCE(63);
      if (lookahead == '3') ADVANCE(64);
      if (lookahead == '6') ADVANCE(65);
      if (lookahead == '8') ADVANCE(66);
      if (lookahead == 'n') ADVANCE(67);
      if (lookahead == 's') ADVANCE(68);
      END_STATE();
    case 16:
      if (lookahead == 'i') ADVANCE(69);
      END_STATE();
    case 17:
      if (lookahead == 'h') ADVANCE(70);
      END_STATE();
    case 18:
      if (lookahead == 'i') ADVANCE(71);
      END_STATE();
    case 19:
      ACCEPT_TOKEN(anon_sym_as);
      if (lookahead == 'y') ADVANCE(72);
      END_STATE();
    case 20:
      if (lookahead == 'a') ADVANCE(73);
      END_STATE();
    case 21:
      if (lookahead == 'o') ADVANCE(74);
      END_STATE();
    case 22:
      if (lookahead == 'o') ADVANCE(75);
      END_STATE();
    case 23:
      if (lookahead == 'e') ADVANCE(76);
      END_STATE();
    case 24:
      if (lookahead == 'a') ADVANCE(77);
      END_STATE();
    case 25:
      if (lookahead == 'n') ADVANCE(78);
      END_STATE();
    case 26:
      if (lookahead == 'a') ADVANCE(79);
      END_STATE();
    case 27:
      if (lookahead == 'f') ADVANCE(80);
      END_STATE();
    case 28:
      if (lookahead == 'n') ADVANCE(81);
      END_STATE();
    case 29:
      if (lookahead == 's') ADVANCE(82);
      END_STATE();
    case 30:
      if (lookahead == 'u') ADVANCE(83);
      END_STATE();
    case 31:
      if (lookahead == 'p') ADVANCE(84);
      if (lookahead == 't') ADVANCE(85);
      END_STATE();
    case 32:
      if (lookahead == '2') ADVANCE(86);
      END_STATE();
    case 33:
      if (lookahead == '4') ADVANCE(87);
      END_STATE();
    case 34:
      if (lookahead == 'l') ADVANCE(88);
      END_STATE();
    case 35:
      ACCEPT_TOKEN(anon_sym_fn);
      END_STATE();
    case 36:
      if (lookahead == 'r') ADVANCE(89);
      END_STATE();
    case 37:
      if (lookahead == '2') ADVANCE(90);
      if (lookahead == '6') ADVANCE(91);
      END_STATE();
    case 38:
      if (lookahead == '2') ADVANCE(92);
      END_STATE();
    case 39:
      if (lookahead == '4') ADVANCE(93);
      END_STATE();
    case 40:
      ACCEPT_TOKEN(anon_sym_i8);
      END_STATE();
    case 41:
      if (lookahead == 'e') ADVANCE(94);
      END_STATE();
    case 42:
      ACCEPT_TOKEN(anon_sym_if);
      END_STATE();
    case 43:
      if (lookahead == 'p') ADVANCE(95);
      END_STATE();
    case 44:
      ACCEPT_TOKEN(anon_sym_in);
      END_STATE();
    case 45:
      if (lookahead == 'i') ADVANCE(96);
      END_STATE();
    case 46:
      if (lookahead == 'e') ADVANCE(97);
      END_STATE();
    case 47:
      if (lookahead == 't') ADVANCE(98);
      END_STATE();
    case 48:
      if (lookahead == 'f') ADVANCE(99);
      if (lookahead == 't') ADVANCE(100);
      END_STATE();
    case 49:
      if (lookahead == 'o') ADVANCE(101);
      END_STATE();
    case 50:
      if (lookahead == 't') ADVANCE(102);
      END_STATE();
    case 51:
      if (lookahead == 't') ADVANCE(103);
      END_STATE();
    case 52:
      if (lookahead == 'd') ADVANCE(104);
      if (lookahead == 'v') ADVANCE(105);
      END_STATE();
    case 53:
      if (lookahead == 't') ADVANCE(106);
      END_STATE();
    case 54:
      if (lookahead == 't') ADVANCE(107);
      END_STATE();
    case 55:
      if (lookahead == 'b') ADVANCE(108);
      END_STATE();
    case 56:
      if (lookahead == 'f') ADVANCE(109);
      if (lookahead == 't') ADVANCE(110);
      END_STATE();
    case 57:
      if (lookahead == 'l') ADVANCE(111);
      END_STATE();
    case 58:
      if (lookahead == 'a') ADVANCE(112);
      if (lookahead == 'm') ADVANCE(113);
      if (lookahead == 'r') ADVANCE(114);
      END_STATE();
    case 59:
      if (lookahead == 'p') ADVANCE(115);
      END_STATE();
    case 60:
      if (lookahead == 'a') ADVANCE(116);
      if (lookahead == 'u') ADVANCE(117);
      END_STATE();
    case 61:
      ACCEPT_TOKEN(anon_sym_tt);
      END_STATE();
    case 62:
      ACCEPT_TOKEN(anon_sym_ty);
      if (lookahead == 'p') ADVANCE(118);
      END_STATE();
    case 63:
      if (lookahead == '2') ADVANCE(119);
      if (lookahead == '6') ADVANCE(120);
      END_STATE();
    case 64:
      if (lookahead == '2') ADVANCE(121);
      END_STATE();
    case 65:
      if (lookahead == '4') ADVANCE(122);
      END_STATE();
    case 66:
      ACCEPT_TOKEN(anon_sym_u8);
      END_STATE();
    case 67:
      if (lookahead == 'i') ADVANCE(123);
      if (lookahead == 's') ADVANCE(124);
      END_STATE();
    case 68:
      if (lookahead == 'e') ADVANCE(125);
      if (lookahead == 'i') ADVANCE(126);
      END_STATE();
    case 69:
      if (lookahead == 's') ADVANCE(127);
      END_STATE();
    case 70:
      if (lookahead == 'e') ADVANCE(128);
      if (lookahead == 'i') ADVANCE(129);
      END_STATE();
    case 71:
      if (lookahead == 'e') ADVANCE(130);
      END_STATE();
    case 72:
      if (lookahead == 'n') ADVANCE(131);
      END_STATE();
    case 73:
      if (lookahead == 'i') ADVANCE(132);
      END_STATE();
    case 74:
      if (lookahead == 'c') ADVANCE(133);
      END_STATE();
    case 75:
      if (lookahead == 'l') ADVANCE(134);
      END_STATE();
    case 76:
      if (lookahead == 'a') ADVANCE(135);
      END_STATE();
    case 77:
      if (lookahead == 'r') ADVANCE(136);
      END_STATE();
    case 78:
      if (lookahead == 's') ADVANCE(137);
      if (lookahead == 't') ADVANCE(138);
      END_STATE();
    case 79:
      if (lookahead == 't') ADVANCE(139);
      END_STATE();
    case 80:
      if (lookahead == 'a') ADVANCE(140);
      END_STATE();
    case 81:
      ACCEPT_TOKEN(anon_sym_dyn);
      END_STATE();
    case 82:
      if (lookahead == 'e') ADVANCE(141);
      END_STATE();
    case 83:
      if (lookahead == 'm') ADVANCE(142);
      END_STATE();
    case 84:
      if (lookahead == 'r') ADVANCE(143);
      END_STATE();
    case 85:
      if (lookahead == 'e') ADVANCE(144);
      END_STATE();
    case 86:
      ACCEPT_TOKEN(anon_sym_f32);
      END_STATE();
    case 87:
      ACCEPT_TOKEN(anon_sym_f64);
      END_STATE();
    case 88:
      if (lookahead == 's') ADVANCE(145);
      END_STATE();
    case 89:
      ACCEPT_TOKEN(anon_sym_for);
      END_STATE();
    case 90:
      if (lookahead == '8') ADVANCE(146);
      END_STATE();
    case 91:
      ACCEPT_TOKEN(anon_sym_i16);
      END_STATE();
    case 92:
      ACCEPT_TOKEN(anon_sym_i32);
      END_STATE();
    case 93:
      ACCEPT_TOKEN(anon_sym_i64);
      END_STATE();
    case 94:
      if (lookahead == 'n') ADVANCE(147);
      END_STATE();
    case 95:
      if (lookahead == 'l') ADVANCE(148);
      END_STATE();
    case 96:
      if (lookahead == 'z') ADVANCE(149);
      END_STATE();
    case 97:
      if (lookahead == 'm') ADVANCE(150);
      END_STATE();
    case 98:
      ACCEPT_TOKEN(anon_sym_let);
      END_STATE();
    case 99:
      if (lookahead == 'e') ADVANCE(151);
      END_STATE();
    case 100:
      if (lookahead == 'e') ADVANCE(152);
      END_STATE();
    case 101:
      if (lookahead == 'p') ADVANCE(153);
      END_STATE();
    case 102:
      if (lookahead == 'c') ADVANCE(154);
      END_STATE();
    case 103:
      if (lookahead == 'a') ADVANCE(155);
      END_STATE();
    case 104:
      ACCEPT_TOKEN(anon_sym_mod);
      END_STATE();
    case 105:
      if (lookahead == 'e') ADVANCE(156);
      END_STATE();
    case 106:
      ACCEPT_TOKEN(sym_mutable_specifier);
      END_STATE();
    case 107:
      ACCEPT_TOKEN(anon_sym_pat);
      if (lookahead == 'h') ADVANCE(157);
      END_STATE();
    case 108:
      ACCEPT_TOKEN(anon_sym_pub);
      END_STATE();
    case 109:
      ACCEPT_TOKEN(anon_sym_ref);
      END_STATE();
    case 110:
      if (lookahead == 'u') ADVANCE(158);
      END_STATE();
    case 111:
      if (lookahead == 'f') ADVANCE(159);
      END_STATE();
    case 112:
      if (lookahead == 't') ADVANCE(160);
      END_STATE();
    case 113:
      if (lookahead == 't') ADVANCE(161);
      END_STATE();
    case 114:
      ACCEPT_TOKEN(anon_sym_str);
      if (lookahead == 'u') ADVANCE(162);
      END_STATE();
    case 115:
      if (lookahead == 'e') ADVANCE(163);
      END_STATE();
    case 116:
      if (lookahead == 'i') ADVANCE(164);
      END_STATE();
    case 117:
      if (lookahead == 'e') ADVANCE(165);
      END_STATE();
    case 118:
      if (lookahead == 'e') ADVANCE(166);
      END_STATE();
    case 119:
      if (lookahead == '8') ADVANCE(167);
      END_STATE();
    case 120:
      ACCEPT_TOKEN(anon_sym_u16);
      END_STATE();
    case 121:
      ACCEPT_TOKEN(anon_sym_u32);
      END_STATE();
    case 122:
      ACCEPT_TOKEN(anon_sym_u64);
      END_STATE();
    case 123:
      if (lookahead == 'o') ADVANCE(168);
      END_STATE();
    case 124:
      if (lookahead == 'a') ADVANCE(169);
      END_STATE();
    case 125:
      ACCEPT_TOKEN(anon_sym_use);
      END_STATE();
    case 126:
      if (lookahead == 'z') ADVANCE(170);
      END_STATE();
    case 127:
      ACCEPT_TOKEN(anon_sym_vis);
      END_STATE();
    case 128:
      if (lookahead == 'r') ADVANCE(171);
      END_STATE();
    case 129:
      if (lookahead == 'l') ADVANCE(172);
      END_STATE();
    case 130:
      if (lookahead == 'l') ADVANCE(173);
      END_STATE();
    case 131:
      if (lookahead == 'c') ADVANCE(174);
      END_STATE();
    case 132:
      if (lookahead == 't') ADVANCE(175);
      END_STATE();
    case 133:
      if (lookahead == 'k') ADVANCE(176);
      END_STATE();
    case 134:
      ACCEPT_TOKEN(anon_sym_bool);
      END_STATE();
    case 135:
      if (lookahead == 'k') ADVANCE(177);
      END_STATE();
    case 136:
      ACCEPT_TOKEN(anon_sym_char);
      END_STATE();
    case 137:
      if (lookahead == 't') ADVANCE(178);
      END_STATE();
    case 138:
      if (lookahead == 'i') ADVANCE(179);
      END_STATE();
    case 139:
      if (lookahead == 'e') ADVANCE(180);
      END_STATE();
    case 140:
      if (lookahead == 'u') ADVANCE(181);
      END_STATE();
    case 141:
      ACCEPT_TOKEN(anon_sym_else);
      END_STATE();
    case 142:
      ACCEPT_TOKEN(anon_sym_enum);
      END_STATE();
    case 143:
      ACCEPT_TOKEN(anon_sym_expr);
      END_STATE();
    case 144:
      if (lookahead == 'r') ADVANCE(182);
      END_STATE();
    case 145:
      if (lookahead == 'e') ADVANCE(183);
      END_STATE();
    case 146:
      ACCEPT_TOKEN(anon_sym_i128);
      END_STATE();
    case 147:
      if (lookahead == 't') ADVANCE(184);
      END_STATE();
    case 148:
      ACCEPT_TOKEN(anon_sym_impl);
      END_STATE();
    case 149:
      if (lookahead == 'e') ADVANCE(185);
      END_STATE();
    case 150:
      ACCEPT_TOKEN(anon_sym_item);
      END_STATE();
    case 151:
      if (lookahead == 't') ADVANCE(186);
      END_STATE();
    case 152:
      if (lookahead == 'r') ADVANCE(187);
      END_STATE();
    case 153:
      ACCEPT_TOKEN(anon_sym_loop);
      END_STATE();
    case 154:
      if (lookahead == 'h') ADVANCE(188);
      END_STATE();
    case 155:
      ACCEPT_TOKEN(anon_sym_meta);
      END_STATE();
    case 156:
      ACCEPT_TOKEN(anon_sym_move);
      END_STATE();
    case 157:
      ACCEPT_TOKEN(anon_sym_path);
      END_STATE();
    case 158:
      if (lookahead == 'r') ADVANCE(189);
      END_STATE();
    case 159:
      ACCEPT_TOKEN(sym_self);
      END_STATE();
    case 160:
      if (lookahead == 'i') ADVANCE(190);
      END_STATE();
    case 161:
      ACCEPT_TOKEN(anon_sym_stmt);
      END_STATE();
    case 162:
      if (lookahead == 'c') ADVANCE(191);
      END_STATE();
    case 163:
      if (lookahead == 'r') ADVANCE(192);
      END_STATE();
    case 164:
      if (lookahead == 't') ADVANCE(193);
      END_STATE();
    case 165:
      ACCEPT_TOKEN(anon_sym_true);
      END_STATE();
    case 166:
      ACCEPT_TOKEN(anon_sym_type);
      END_STATE();
    case 167:
      ACCEPT_TOKEN(anon_sym_u128);
      END_STATE();
    case 168:
      if (lookahead == 'n') ADVANCE(194);
      END_STATE();
    case 169:
      if (lookahead == 'f') ADVANCE(195);
      END_STATE();
    case 170:
      if (lookahead == 'e') ADVANCE(196);
      END_STATE();
    case 171:
      if (lookahead == 'e') ADVANCE(197);
      END_STATE();
    case 172:
      if (lookahead == 'e') ADVANCE(198);
      END_STATE();
    case 173:
      if (lookahead == 'd') ADVANCE(199);
      END_STATE();
    case 174:
      ACCEPT_TOKEN(anon_sym_async);
      END_STATE();
    case 175:
      ACCEPT_TOKEN(anon_sym_await);
      END_STATE();
    case 176:
      ACCEPT_TOKEN(anon_sym_block);
      END_STATE();
    case 177:
      ACCEPT_TOKEN(anon_sym_break);
      END_STATE();
    case 178:
      ACCEPT_TOKEN(anon_sym_const);
      END_STATE();
    case 179:
      if (lookahead == 'n') ADVANCE(200);
      END_STATE();
    case 180:
      ACCEPT_TOKEN(sym_crate);
      END_STATE();
    case 181:
      if (lookahead == 'l') ADVANCE(201);
      END_STATE();
    case 182:
      if (lookahead == 'n') ADVANCE(202);
      END_STATE();
    case 183:
      ACCEPT_TOKEN(anon_sym_false);
      END_STATE();
    case 184:
      ACCEPT_TOKEN(anon_sym_ident);
      END_STATE();
    case 185:
      ACCEPT_TOKEN(anon_sym_isize);
      END_STATE();
    case 186:
      if (lookahead == 'i') ADVANCE(203);
      END_STATE();
    case 187:
      if (lookahead == 'a') ADVANCE(204);
      END_STATE();
    case 188:
      ACCEPT_TOKEN(anon_sym_match);
      END_STATE();
    case 189:
      if (lookahead == 'n') ADVANCE(205);
      END_STATE();
    case 190:
      if (lookahead == 'c') ADVANCE(206);
      END_STATE();
    case 191:
      if (lookahead == 't') ADVANCE(207);
      END_STATE();
    case 192:
      ACCEPT_TOKEN(sym_super);
      END_STATE();
    case 193:
      ACCEPT_TOKEN(anon_sym_trait);
      END_STATE();
    case 194:
      ACCEPT_TOKEN(anon_sym_union);
      END_STATE();
    case 195:
      if (lookahead == 'e') ADVANCE(208);
      END_STATE();
    case 196:
      ACCEPT_TOKEN(anon_sym_usize);
      END_STATE();
    case 197:
      ACCEPT_TOKEN(anon_sym_where);
      END_STATE();
    case 198:
      ACCEPT_TOKEN(anon_sym_while);
      END_STATE();
    case 199:
      ACCEPT_TOKEN(anon_sym_yield);
      END_STATE();
    case 200:
      if (lookahead == 'u') ADVANCE(209);
      END_STATE();
    case 201:
      if (lookahead == 't') ADVANCE(210);
      END_STATE();
    case 202:
      ACCEPT_TOKEN(anon_sym_extern);
      END_STATE();
    case 203:
      if (lookahead == 'm') ADVANCE(211);
      END_STATE();
    case 204:
      if (lookahead == 'l') ADVANCE(212);
      END_STATE();
    case 205:
      ACCEPT_TOKEN(anon_sym_return);
      END_STATE();
    case 206:
      ACCEPT_TOKEN(anon_sym_static);
      END_STATE();
    case 207:
      ACCEPT_TOKEN(anon_sym_struct);
      END_STATE();
    case 208:
      ACCEPT_TOKEN(anon_sym_unsafe);
      END_STATE();
    case 209:
      if (lookahead == 'e') ADVANCE(213);
      END_STATE();
    case 210:
      ACCEPT_TOKEN(anon_sym_default);
      END_STATE();
    case 211:
      if (lookahead == 'e') ADVANCE(214);
      END_STATE();
    case 212:
      ACCEPT_TOKEN(anon_sym_literal);
      END_STATE();
    case 213:
      ACCEPT_TOKEN(anon_sym_continue);
      END_STATE();
    case 214:
      ACCEPT_TOKEN(anon_sym_lifetime);
      END_STATE();
    default:
      return false;
  }
}

static const TSLexMode ts_lex_modes[STATE_COUNT] = {
  [0] = {.lex_state = 0, .external_lex_state = 1},
  [1] = {.lex_state = 58, .external_lex_state = 2},
  [2] = {.lex_state = 58, .external_lex_state = 2},
  [3] = {.lex_state = 58, .external_lex_state = 2},
  [4] = {.lex_state = 58, .external_lex_state = 2},
  [5] = {.lex_state = 58, .external_lex_state = 2},
  [6] = {.lex_state = 58, .external_lex_state = 2},
  [7] = {.lex_state = 58, .external_lex_state = 2},
  [8] = {.lex_state = 58, .external_lex_state = 2},
  [9] = {.lex_state = 58, .external_lex_state = 2},
  [10] = {.lex_state = 58, .external_lex_state = 2},
  [11] = {.lex_state = 58, .external_lex_state = 2},
  [12] = {.lex_state = 58, .external_lex_state = 2},
  [13] = {.lex_state = 58, .external_lex_state = 2},
  [14] = {.lex_state = 58, .external_lex_state = 2},
  [15] = {.lex_state = 58, .external_lex_state = 2},
  [16] = {.lex_state = 58, .external_lex_state = 2},
  [17] = {.lex_state = 1, .external_lex_state = 2},
  [18] = {.lex_state = 1, .external_lex_state = 2},
  [19] = {.lex_state = 1, .external_lex_state = 2},
  [20] = {.lex_state = 1, .external_lex_state = 2},
  [21] = {.lex_state = 1, .external_lex_state = 2},
  [22] = {.lex_state = 1, .external_lex_state = 2},
  [23] = {.lex_state = 1, .external_lex_state = 2},
  [24] = {.lex_state = 1, .external_lex_state = 2},
  [25] = {.lex_state = 1, .external_lex_state = 2},
  [26] = {.lex_state = 1, .external_lex_state = 2},
  [27] = {.lex_state = 1, .external_lex_state = 2},
  [28] = {.lex_state = 1, .external_lex_state = 2},
  [29] = {.lex_state = 1, .external_lex_state = 2},
  [30] = {.lex_state = 1, .external_lex_state = 2},
  [31] = {.lex_state = 1, .external_lex_state = 2},
  [32] = {.lex_state = 1, .external_lex_state = 2},
  [33] = {.lex_state = 5, .external_lex_state = 2},
  [34] = {.lex_state = 5, .external_lex_state = 2},
  [35] = {.lex_state = 5, .external_lex_state = 2},
  [36] = {.lex_state = 5, .external_lex_state = 2},
  [37] = {.lex_state = 5, .external_lex_state = 2},
  [38] = {.lex_state = 5, .external_lex_state = 2},
  [39] = {.lex_state = 5, .external_lex_state = 2},
  [40] = {.lex_state = 5, .external_lex_state = 2},
  [41] = {.lex_state = 5, .external_lex_state = 2},
  [42] = {.lex_state = 5, .external_lex_state = 2},
  [43] = {.lex_state = 5, .external_lex_state = 2},
  [44] = {.lex_state = 5, .external_lex_state = 2},
  [45] = {.lex_state = 5, .external_lex_state = 2},
  [46] = {.lex_state = 5, .external_lex_state = 2},
  [47] = {.lex_state = 5, .external_lex_state = 2},
  [48] = {.lex_state = 57, .external_lex_state = 2},
  [49] = {.lex_state = 5, .external_lex_state = 2},
  [50] = {.lex_state = 5, .external_lex_state = 2},
  [51] = {.lex_state = 5, .external_lex_state = 2},
  [52] = {.lex_state = 5, .external_lex_state = 2},
  [53] = {.lex_state = 57, .external_lex_state = 2},
  [54] = {.lex_state = 57, .external_lex_state = 2},
  [55] = {.lex_state = 5, .external_lex_state = 2},
  [56] = {.lex_state = 57, .external_lex_state = 2},
  [57] = {.lex_state = 5, .external_lex_state = 2},
  [58] = {.lex_state = 5, .external_lex_state = 2},
  [59] = {.lex_state = 5, .external_lex_state = 2},
  [60] = {.lex_state = 57, .external_lex_state = 2},
  [61] = {.lex_state = 57, .external_lex_state = 2},
  [62] = {.lex_state = 57, .external_lex_state = 2},
  [63] = {.lex_state = 57, .external_lex_state = 2},
  [64] = {.lex_state = 57, .external_lex_state = 2},
  [65] = {.lex_state = 57, .external_lex_state = 2},
  [66] = {.lex_state = 57, .external_lex_state = 2},
  [67] = {.lex_state = 57, .external_lex_state = 2},
  [68] = {.lex_state = 57, .external_lex_state = 2},
  [69] = {.lex_state = 57, .external_lex_state = 2},
  [70] = {.lex_state = 57, .external_lex_state = 2},
  [71] = {.lex_state = 5, .external_lex_state = 2},
  [72] = {.lex_state = 57, .external_lex_state = 2},
  [73] = {.lex_state = 57, .external_lex_state = 2},
  [74] = {.lex_state = 57, .external_lex_state = 2},
  [75] = {.lex_state = 57, .external_lex_state = 2},
  [76] = {.lex_state = 57, .external_lex_state = 2},
  [77] = {.lex_state = 57, .external_lex_state = 2},
  [78] = {.lex_state = 5, .external_lex_state = 2},
  [79] = {.lex_state = 5, .external_lex_state = 2},
  [80] = {.lex_state = 5, .external_lex_state = 2},
  [81] = {.lex_state = 57, .external_lex_state = 2},
  [82] = {.lex_state = 5, .external_lex_state = 2},
  [83] = {.lex_state = 57, .external_lex_state = 2},
  [84] = {.lex_state = 57, .external_lex_state = 2},
  [85] = {.lex_state = 57, .external_lex_state = 2},
  [86] = {.lex_state = 5, .external_lex_state = 2},
  [87] = {.lex_state = 5, .external_lex_state = 2},
  [88] = {.lex_state = 57, .external_lex_state = 2},
  [89] = {.lex_state = 5, .external_lex_state = 2},
  [90] = {.lex_state = 5, .external_lex_state = 2},
  [91] = {.lex_state = 5, .external_lex_state = 2},
  [92] = {.lex_state = 5, .external_lex_state = 2},
  [93] = {.lex_state = 5, .external_lex_state = 2},
  [94] = {.lex_state = 5, .external_lex_state = 2},
  [95] = {.lex_state = 5, .external_lex_state = 2},
  [96] = {.lex_state = 57, .external_lex_state = 2},
  [97] = {.lex_state = 5, .external_lex_state = 2},
  [98] = {.lex_state = 5, .external_lex_state = 2},
  [99] = {.lex_state = 5, .external_lex_state = 2},
  [100] = {.lex_state = 5, .external_lex_state = 2},
  [101] = {.lex_state = 5, .external_lex_state = 2},
  [102] = {.lex_state = 5, .external_lex_state = 2},
  [103] = {.lex_state = 5, .external_lex_state = 2},
  [104] = {.lex_state = 5, .external_lex_state = 2},
  [105] = {.lex_state = 5, .external_lex_state = 2},
  [106] = {.lex_state = 5, .external_lex_state = 2},
  [107] = {.lex_state = 5, .external_lex_state = 2},
  [108] = {.lex_state = 5, .external_lex_state = 2},
  [109] = {.lex_state = 5, .external_lex_state = 2},
  [110] = {.lex_state = 5, .external_lex_state = 2},
  [111] = {.lex_state = 5, .external_lex_state = 2},
  [112] = {.lex_state = 5, .external_lex_state = 2},
  [113] = {.lex_state = 5, .external_lex_state = 2},
  [114] = {.lex_state = 5, .external_lex_state = 2},
  [115] = {.lex_state = 5, .external_lex_state = 2},
  [116] = {.lex_state = 5, .external_lex_state = 2},
  [117] = {.lex_state = 5, .external_lex_state = 2},
  [118] = {.lex_state = 5, .external_lex_state = 2},
  [119] = {.lex_state = 5, .external_lex_state = 2},
  [120] = {.lex_state = 5, .external_lex_state = 2},
  [121] = {.lex_state = 5, .external_lex_state = 2},
  [122] = {.lex_state = 5, .external_lex_state = 2},
  [123] = {.lex_state = 5, .external_lex_state = 2},
  [124] = {.lex_state = 5, .external_lex_state = 2},
  [125] = {.lex_state = 5, .external_lex_state = 2},
  [126] = {.lex_state = 5, .external_lex_state = 2},
  [127] = {.lex_state = 5, .external_lex_state = 2},
  [128] = {.lex_state = 5, .external_lex_state = 2},
  [129] = {.lex_state = 5, .external_lex_state = 2},
  [130] = {.lex_state = 5, .external_lex_state = 2},
  [131] = {.lex_state = 5, .external_lex_state = 2},
  [132] = {.lex_state = 5, .external_lex_state = 2},
  [133] = {.lex_state = 5, .external_lex_state = 2},
  [134] = {.lex_state = 5, .external_lex_state = 2},
  [135] = {.lex_state = 5, .external_lex_state = 2},
  [136] = {.lex_state = 5, .external_lex_state = 2},
  [137] = {.lex_state = 5, .external_lex_state = 2},
  [138] = {.lex_state = 5, .external_lex_state = 2},
  [139] = {.lex_state = 5, .external_lex_state = 2},
  [140] = {.lex_state = 5, .external_lex_state = 2},
  [141] = {.lex_state = 5, .external_lex_state = 2},
  [142] = {.lex_state = 5, .external_lex_state = 2},
  [143] = {.lex_state = 5, .external_lex_state = 2},
  [144] = {.lex_state = 5, .external_lex_state = 2},
  [145] = {.lex_state = 5, .external_lex_state = 2},
  [146] = {.lex_state = 5, .external_lex_state = 2},
  [147] = {.lex_state = 5, .external_lex_state = 2},
  [148] = {.lex_state = 5, .external_lex_state = 2},
  [149] = {.lex_state = 5, .external_lex_state = 2},
  [150] = {.lex_state = 5, .external_lex_state = 2},
  [151] = {.lex_state = 5, .external_lex_state = 2},
  [152] = {.lex_state = 5, .external_lex_state = 2},
  [153] = {.lex_state = 5, .external_lex_state = 2},
  [154] = {.lex_state = 5, .external_lex_state = 2},
  [155] = {.lex_state = 5, .external_lex_state = 2},
  [156] = {.lex_state = 5, .external_lex_state = 2},
  [157] = {.lex_state = 5, .external_lex_state = 2},
  [158] = {.lex_state = 5, .external_lex_state = 2},
  [159] = {.lex_state = 5, .external_lex_state = 2},
  [160] = {.lex_state = 5, .external_lex_state = 2},
  [161] = {.lex_state = 5, .external_lex_state = 2},
  [162] = {.lex_state = 5, .external_lex_state = 2},
  [163] = {.lex_state = 5, .external_lex_state = 2},
  [164] = {.lex_state = 5, .external_lex_state = 2},
  [165] = {.lex_state = 5, .external_lex_state = 2},
  [166] = {.lex_state = 5, .external_lex_state = 2},
  [167] = {.lex_state = 5, .external_lex_state = 2},
  [168] = {.lex_state = 5, .external_lex_state = 2},
  [169] = {.lex_state = 5, .external_lex_state = 2},
  [170] = {.lex_state = 5, .external_lex_state = 2},
  [171] = {.lex_state = 5, .external_lex_state = 2},
  [172] = {.lex_state = 5, .external_lex_state = 2},
  [173] = {.lex_state = 5, .external_lex_state = 2},
  [174] = {.lex_state = 5, .external_lex_state = 2},
  [175] = {.lex_state = 5, .external_lex_state = 2},
  [176] = {.lex_state = 5, .external_lex_state = 2},
  [177] = {.lex_state = 5, .external_lex_state = 2},
  [178] = {.lex_state = 5, .external_lex_state = 2},
  [179] = {.lex_state = 6, .external_lex_state = 2},
  [180] = {.lex_state = 6, .external_lex_state = 2},
  [181] = {.lex_state = 6, .external_lex_state = 2},
  [182] = {.lex_state = 6, .external_lex_state = 2},
  [183] = {.lex_state = 6, .external_lex_state = 2},
  [184] = {.lex_state = 6, .external_lex_state = 2},
  [185] = {.lex_state = 6, .external_lex_state = 2},
  [186] = {.lex_state = 6, .external_lex_state = 2},
  [187] = {.lex_state = 6, .external_lex_state = 2},
  [188] = {.lex_state = 6, .external_lex_state = 2},
  [189] = {.lex_state = 6, .external_lex_state = 2},
  [190] = {.lex_state = 6, .external_lex_state = 2},
  [191] = {.lex_state = 6, .external_lex_state = 2},
  [192] = {.lex_state = 6, .external_lex_state = 2},
  [193] = {.lex_state = 6, .external_lex_state = 2},
  [194] = {.lex_state = 5, .external_lex_state = 2},
  [195] = {.lex_state = 5, .external_lex_state = 2},
  [196] = {.lex_state = 5, .external_lex_state = 2},
  [197] = {.lex_state = 5, .external_lex_state = 2},
  [198] = {.lex_state = 1, .external_lex_state = 2},
  [199] = {.lex_state = 5, .external_lex_state = 2},
  [200] = {.lex_state = 5, .external_lex_state = 2},
  [201] = {.lex_state = 5, .external_lex_state = 2},
  [202] = {.lex_state = 5, .external_lex_state = 2},
  [203] = {.lex_state = 5, .external_lex_state = 2},
  [204] = {.lex_state = 5, .external_lex_state = 2},
  [205] = {.lex_state = 5, .external_lex_state = 2},
  [206] = {.lex_state = 5, .external_lex_state = 2},
  [207] = {.lex_state = 5, .external_lex_state = 2},
  [208] = {.lex_state = 5, .external_lex_state = 2},
  [209] = {.lex_state = 5, .external_lex_state = 2},
  [210] = {.lex_state = 1, .external_lex_state = 2},
  [211] = {.lex_state = 1, .external_lex_state = 2},
  [212] = {.lex_state = 1, .external_lex_state = 2},
  [213] = {.lex_state = 1, .external_lex_state = 2},
  [214] = {.lex_state = 1, .external_lex_state = 2},
  [215] = {.lex_state = 1, .external_lex_state = 2},
  [216] = {.lex_state = 1, .external_lex_state = 2},
  [217] = {.lex_state = 1, .external_lex_state = 2},
  [218] = {.lex_state = 1, .external_lex_state = 2},
  [219] = {.lex_state = 1, .external_lex_state = 2},
  [220] = {.lex_state = 1, .external_lex_state = 2},
  [221] = {.lex_state = 1, .external_lex_state = 2},
  [222] = {.lex_state = 1, .external_lex_state = 2},
  [223] = {.lex_state = 1, .external_lex_state = 2},
  [224] = {.lex_state = 1, .external_lex_state = 2},
  [225] = {.lex_state = 1, .external_lex_state = 2},
  [226] = {.lex_state = 1, .external_lex_state = 2},
  [227] = {.lex_state = 1, .external_lex_state = 2},
  [228] = {.lex_state = 1, .external_lex_state = 2},
  [229] = {.lex_state = 1, .external_lex_state = 2},
  [230] = {.lex_state = 5, .external_lex_state = 2},
  [231] = {.lex_state = 5, .external_lex_state = 2},
  [232] = {.lex_state = 5, .external_lex_state = 2},
  [233] = {.lex_state = 1, .external_lex_state = 2},
  [234] = {.lex_state = 1, .external_lex_state = 2},
  [235] = {.lex_state = 5, .external_lex_state = 2},
  [236] = {.lex_state = 5, .external_lex_state = 2},
  [237] = {.lex_state = 5, .external_lex_state = 2},
  [238] = {.lex_state = 5, .external_lex_state = 2},
  [239] = {.lex_state = 12, .external_lex_state = 2},
  [240] = {.lex_state = 4, .external_lex_state = 3},
  [241] = {.lex_state = 4, .external_lex_state = 3},
  [242] = {.lex_state = 4, .external_lex_state = 3},
  [243] = {.lex_state = 4, .external_lex_state = 3},
  [244] = {.lex_state = 4, .external_lex_state = 3},
  [245] = {.lex_state = 58, .external_lex_state = 2},
  [246] = {.lex_state = 58, .external_lex_state = 2},
  [247] = {.lex_state = 58, .external_lex_state = 2},
  [248] = {.lex_state = 58, .external_lex_state = 2},
  [249] = {.lex_state = 58, .external_lex_state = 2},
  [250] = {.lex_state = 58, .external_lex_state = 2},
  [251] = {.lex_state = 58, .external_lex_state = 2},
  [252] = {.lex_state = 58, .external_lex_state = 2},
  [253] = {.lex_state = 58, .external_lex_state = 2},
  [254] = {.lex_state = 58, .external_lex_state = 2},
  [255] = {.lex_state = 58, .external_lex_state = 2},
  [256] = {.lex_state = 58, .external_lex_state = 2},
  [257] = {.lex_state = 58, .external_lex_state = 2},
  [258] = {.lex_state = 58, .external_lex_state = 2},
  [259] = {.lex_state = 58, .external_lex_state = 2},
  [260] = {.lex_state = 58, .external_lex_state = 2},
  [261] = {.lex_state = 58, .external_lex_state = 2},
  [262] = {.lex_state = 58, .external_lex_state = 2},
  [263] = {.lex_state = 58, .external_lex_state = 2},
  [264] = {.lex_state = 58, .external_lex_state = 2},
  [265] = {.lex_state = 58, .external_lex_state = 2},
  [266] = {.lex_state = 58, .external_lex_state = 2},
  [267] = {.lex_state = 58, .external_lex_state = 2},
  [268] = {.lex_state = 58, .external_lex_state = 2},
  [269] = {.lex_state = 58, .external_lex_state = 2},
  [270] = {.lex_state = 58, .external_lex_state = 2},
  [271] = {.lex_state = 58, .external_lex_state = 2},
  [272] = {.lex_state = 58, .external_lex_state = 2},
  [273] = {.lex_state = 58, .external_lex_state = 2},
  [274] = {.lex_state = 58, .external_lex_state = 2},
  [275] = {.lex_state = 58, .external_lex_state = 2},
  [276] = {.lex_state = 58, .external_lex_state = 2},
  [277] = {.lex_state = 58, .external_lex_state = 2},
  [278] = {.lex_state = 58, .external_lex_state = 2},
  [279] = {.lex_state = 58, .external_lex_state = 2},
  [280] = {.lex_state = 58, .external_lex_state = 2},
  [281] = {.lex_state = 58, .external_lex_state = 2},
  [282] = {.lex_state = 58, .external_lex_state = 2},
  [283] = {.lex_state = 58, .external_lex_state = 2},
  [284] = {.lex_state = 58, .external_lex_state = 2},
  [285] = {.lex_state = 58, .external_lex_state = 2},
  [286] = {.lex_state = 58, .external_lex_state = 2},
  [287] = {.lex_state = 58, .external_lex_state = 2},
  [288] = {.lex_state = 58, .external_lex_state = 2},
  [289] = {.lex_state = 58, .external_lex_state = 2},
  [290] = {.lex_state = 58, .external_lex_state = 2},
  [291] = {.lex_state = 58, .external_lex_state = 2},
  [292] = {.lex_state = 58, .external_lex_state = 2},
  [293] = {.lex_state = 58, .external_lex_state = 2},
  [294] = {.lex_state = 58, .external_lex_state = 2},
  [295] = {.lex_state = 58, .external_lex_state = 2},
  [296] = {.lex_state = 58, .external_lex_state = 2},
  [297] = {.lex_state = 58, .external_lex_state = 2},
  [298] = {.lex_state = 58, .external_lex_state = 2},
  [299] = {.lex_state = 58, .external_lex_state = 2},
  [300] = {.lex_state = 58, .external_lex_state = 2},
  [301] = {.lex_state = 58, .external_lex_state = 2},
  [302] = {.lex_state = 58, .external_lex_state = 2},
  [303] = {.lex_state = 58, .external_lex_state = 2},
  [304] = {.lex_state = 58, .external_lex_state = 2},
  [305] = {.lex_state = 58, .external_lex_state = 2},
  [306] = {.lex_state = 58, .external_lex_state = 2},
  [307] = {.lex_state = 58, .external_lex_state = 2},
  [308] = {.lex_state = 58, .external_lex_state = 2},
  [309] = {.lex_state = 58, .external_lex_state = 2},
  [310] = {.lex_state = 58, .external_lex_state = 2},
  [311] = {.lex_state = 58, .external_lex_state = 2},
  [312] = {.lex_state = 58, .external_lex_state = 2},
  [313] = {.lex_state = 58, .external_lex_state = 2},
  [314] = {.lex_state = 58, .external_lex_state = 2},
  [315] = {.lex_state = 58, .external_lex_state = 2},
  [316] = {.lex_state = 58, .external_lex_state = 2},
  [317] = {.lex_state = 58, .external_lex_state = 2},
  [318] = {.lex_state = 58, .external_lex_state = 2},
  [319] = {.lex_state = 58, .external_lex_state = 2},
  [320] = {.lex_state = 58, .external_lex_state = 2},
  [321] = {.lex_state = 58, .external_lex_state = 2},
  [322] = {.lex_state = 58, .external_lex_state = 2},
  [323] = {.lex_state = 58, .external_lex_state = 2},
  [324] = {.lex_state = 58, .external_lex_state = 2},
  [325] = {.lex_state = 58, .external_lex_state = 2},
  [326] = {.lex_state = 58, .external_lex_state = 2},
  [327] = {.lex_state = 58, .external_lex_state = 2},
  [328] = {.lex_state = 58, .external_lex_state = 2},
  [329] = {.lex_state = 58, .external_lex_state = 2},
  [330] = {.lex_state = 58, .external_lex_state = 2},
  [331] = {.lex_state = 58, .external_lex_state = 2},
  [332] = {.lex_state = 58, .external_lex_state = 2},
  [333] = {.lex_state = 58, .external_lex_state = 2},
  [334] = {.lex_state = 58, .external_lex_state = 2},
  [335] = {.lex_state = 58, .external_lex_state = 2},
  [336] = {.lex_state = 58, .external_lex_state = 2},
  [337] = {.lex_state = 58, .external_lex_state = 2},
  [338] = {.lex_state = 58, .external_lex_state = 2},
  [339] = {.lex_state = 58, .external_lex_state = 2},
  [340] = {.lex_state = 58, .external_lex_state = 2},
  [341] = {.lex_state = 58, .external_lex_state = 2},
  [342] = {.lex_state = 58, .external_lex_state = 2},
  [343] = {.lex_state = 58, .external_lex_state = 2},
  [344] = {.lex_state = 58, .external_lex_state = 2},
  [345] = {.lex_state = 58, .external_lex_state = 2},
  [346] = {.lex_state = 58, .external_lex_state = 2},
  [347] = {.lex_state = 58, .external_lex_state = 2},
  [348] = {.lex_state = 58, .external_lex_state = 2},
  [349] = {.lex_state = 58, .external_lex_state = 2},
  [350] = {.lex_state = 58, .external_lex_state = 2},
  [351] = {.lex_state = 58, .external_lex_state = 2},
  [352] = {.lex_state = 58, .external_lex_state = 2},
  [353] = {.lex_state = 58, .external_lex_state = 2},
  [354] = {.lex_state = 58, .external_lex_state = 2},
  [355] = {.lex_state = 58, .external_lex_state = 2},
  [356] = {.lex_state = 58, .external_lex_state = 2},
  [357] = {.lex_state = 58, .external_lex_state = 2},
  [358] = {.lex_state = 58, .external_lex_state = 2},
  [359] = {.lex_state = 58, .external_lex_state = 2},
  [360] = {.lex_state = 58, .external_lex_state = 2},
  [361] = {.lex_state = 58, .external_lex_state = 2},
  [362] = {.lex_state = 58, .external_lex_state = 2},
  [363] = {.lex_state = 58, .external_lex_state = 2},
  [364] = {.lex_state = 58, .external_lex_state = 2},
  [365] = {.lex_state = 5, .external_lex_state = 2},
  [366] = {.lex_state = 58, .external_lex_state = 2},
  [367] = {.lex_state = 58, .external_lex_state = 2},
  [368] = {.lex_state = 58, .external_lex_state = 2},
  [369] = {.lex_state = 58, .external_lex_state = 2},
  [370] = {.lex_state = 58, .external_lex_state = 2},
  [371] = {.lex_state = 58, .external_lex_state = 2},
  [372] = {.lex_state = 58, .external_lex_state = 2},
  [373] = {.lex_state = 58, .external_lex_state = 2},
  [374] = {.lex_state = 58, .external_lex_state = 2},
  [375] = {.lex_state = 58, .external_lex_state = 2},
  [376] = {.lex_state = 58, .external_lex_state = 2},
  [377] = {.lex_state = 58, .external_lex_state = 2},
  [378] = {.lex_state = 58, .external_lex_state = 2},
  [379] = {.lex_state = 58, .external_lex_state = 2},
  [380] = {.lex_state = 58, .external_lex_state = 2},
  [381] = {.lex_state = 58, .external_lex_state = 2},
  [382] = {.lex_state = 58, .external_lex_state = 2},
  [383] = {.lex_state = 58, .external_lex_state = 2},
  [384] = {.lex_state = 58, .external_lex_state = 2},
  [385] = {.lex_state = 58, .external_lex_state = 2},
  [386] = {.lex_state = 58, .external_lex_state = 2},
  [387] = {.lex_state = 58, .external_lex_state = 2},
  [388] = {.lex_state = 58, .external_lex_state = 2},
  [389] = {.lex_state = 58, .external_lex_state = 2},
  [390] = {.lex_state = 58, .external_lex_state = 2},
  [391] = {.lex_state = 5, .external_lex_state = 2},
  [392] = {.lex_state = 58, .external_lex_state = 2},
  [393] = {.lex_state = 58, .external_lex_state = 2},
  [394] = {.lex_state = 58, .external_lex_state = 2},
  [395] = {.lex_state = 58, .external_lex_state = 2},
  [396] = {.lex_state = 58, .external_lex_state = 2},
  [397] = {.lex_state = 58, .external_lex_state = 2},
  [398] = {.lex_state = 58, .external_lex_state = 2},
  [399] = {.lex_state = 58, .external_lex_state = 2},
  [400] = {.lex_state = 58, .external_lex_state = 2},
  [401] = {.lex_state = 58, .external_lex_state = 2},
  [402] = {.lex_state = 58, .external_lex_state = 2},
  [403] = {.lex_state = 58, .external_lex_state = 2},
  [404] = {.lex_state = 58, .external_lex_state = 2},
  [405] = {.lex_state = 58, .external_lex_state = 2},
  [406] = {.lex_state = 58, .external_lex_state = 2},
  [407] = {.lex_state = 58, .external_lex_state = 2},
  [408] = {.lex_state = 58, .external_lex_state = 2},
  [409] = {.lex_state = 58, .external_lex_state = 2},
  [410] = {.lex_state = 58, .external_lex_state = 2},
  [411] = {.lex_state = 58, .external_lex_state = 2},
  [412] = {.lex_state = 58, .external_lex_state = 2},
  [413] = {.lex_state = 58, .external_lex_state = 2},
  [414] = {.lex_state = 58, .external_lex_state = 2},
  [415] = {.lex_state = 58, .external_lex_state = 2},
  [416] = {.lex_state = 58, .external_lex_state = 2},
  [417] = {.lex_state = 58, .external_lex_state = 2},
  [418] = {.lex_state = 58, .external_lex_state = 2},
  [419] = {.lex_state = 58, .external_lex_state = 2},
  [420] = {.lex_state = 58, .external_lex_state = 2},
  [421] = {.lex_state = 58, .external_lex_state = 2},
  [422] = {.lex_state = 58, .external_lex_state = 2},
  [423] = {.lex_state = 58, .external_lex_state = 2},
  [424] = {.lex_state = 58, .external_lex_state = 2},
  [425] = {.lex_state = 58, .external_lex_state = 2},
  [426] = {.lex_state = 58, .external_lex_state = 2},
  [427] = {.lex_state = 58, .external_lex_state = 2},
  [428] = {.lex_state = 58, .external_lex_state = 2},
  [429] = {.lex_state = 58, .external_lex_state = 2},
  [430] = {.lex_state = 58, .external_lex_state = 2},
  [431] = {.lex_state = 58, .external_lex_state = 2},
  [432] = {.lex_state = 58, .external_lex_state = 2},
  [433] = {.lex_state = 58, .external_lex_state = 2},
  [434] = {.lex_state = 58, .external_lex_state = 2},
  [435] = {.lex_state = 58, .external_lex_state = 2},
  [436] = {.lex_state = 58, .external_lex_state = 2},
  [437] = {.lex_state = 58, .external_lex_state = 2},
  [438] = {.lex_state = 58, .external_lex_state = 2},
  [439] = {.lex_state = 58, .external_lex_state = 2},
  [440] = {.lex_state = 58, .external_lex_state = 2},
  [441] = {.lex_state = 58, .external_lex_state = 2},
  [442] = {.lex_state = 58, .external_lex_state = 2},
  [443] = {.lex_state = 58, .external_lex_state = 2},
  [444] = {.lex_state = 58, .external_lex_state = 2},
  [445] = {.lex_state = 58, .external_lex_state = 2},
  [446] = {.lex_state = 58, .external_lex_state = 2},
  [447] = {.lex_state = 58, .external_lex_state = 2},
  [448] = {.lex_state = 58, .external_lex_state = 2},
  [449] = {.lex_state = 58, .external_lex_state = 2},
  [450] = {.lex_state = 58, .external_lex_state = 2},
  [451] = {.lex_state = 58, .external_lex_state = 2},
  [452] = {.lex_state = 58, .external_lex_state = 2},
  [453] = {.lex_state = 58, .external_lex_state = 2},
  [454] = {.lex_state = 58, .external_lex_state = 2},
  [455] = {.lex_state = 58, .external_lex_state = 2},
  [456] = {.lex_state = 58, .external_lex_state = 2},
  [457] = {.lex_state = 58, .external_lex_state = 2},
  [458] = {.lex_state = 58, .external_lex_state = 2},
  [459] = {.lex_state = 58, .external_lex_state = 2},
  [460] = {.lex_state = 58, .external_lex_state = 2},
  [461] = {.lex_state = 58, .external_lex_state = 2},
  [462] = {.lex_state = 58, .external_lex_state = 2},
  [463] = {.lex_state = 58, .external_lex_state = 2},
  [464] = {.lex_state = 58, .external_lex_state = 2},
  [465] = {.lex_state = 58, .external_lex_state = 2},
  [466] = {.lex_state = 58, .external_lex_state = 2},
  [467] = {.lex_state = 58, .external_lex_state = 2},
  [468] = {.lex_state = 58, .external_lex_state = 2},
  [469] = {.lex_state = 5, .external_lex_state = 2},
  [470] = {.lex_state = 58, .external_lex_state = 2},
  [471] = {.lex_state = 58, .external_lex_state = 2},
  [472] = {.lex_state = 58, .external_lex_state = 2},
  [473] = {.lex_state = 58, .external_lex_state = 2},
  [474] = {.lex_state = 58, .external_lex_state = 2},
  [475] = {.lex_state = 58, .external_lex_state = 2},
  [476] = {.lex_state = 58, .external_lex_state = 2},
  [477] = {.lex_state = 58, .external_lex_state = 2},
  [478] = {.lex_state = 58, .external_lex_state = 2},
  [479] = {.lex_state = 58, .external_lex_state = 2},
  [480] = {.lex_state = 58, .external_lex_state = 2},
  [481] = {.lex_state = 12, .external_lex_state = 2},
  [482] = {.lex_state = 12, .external_lex_state = 2},
  [483] = {.lex_state = 13, .external_lex_state = 2},
  [484] = {.lex_state = 5, .external_lex_state = 2},
  [485] = {.lex_state = 12, .external_lex_state = 2},
  [486] = {.lex_state = 5, .external_lex_state = 2},
  [487] = {.lex_state = 12, .external_lex_state = 2},
  [488] = {.lex_state = 12, .external_lex_state = 2},
  [489] = {.lex_state = 12, .external_lex_state = 2},
  [490] = {.lex_state = 5, .external_lex_state = 2},
  [491] = {.lex_state = 12, .external_lex_state = 2},
  [492] = {.lex_state = 12, .external_lex_state = 2},
  [493] = {.lex_state = 12, .external_lex_state = 2},
  [494] = {.lex_state = 12, .external_lex_state = 2},
  [495] = {.lex_state = 12, .external_lex_state = 2},
  [496] = {.lex_state = 12, .external_lex_state = 2},
  [497] = {.lex_state = 12, .external_lex_state = 2},
  [498] = {.lex_state = 12, .external_lex_state = 2},
  [499] = {.lex_state = 12, .external_lex_state = 2},
  [500] = {.lex_state = 5, .external_lex_state = 2},
  [501] = {.lex_state = 13, .external_lex_state = 2},
  [502] = {.lex_state = 13, .external_lex_state = 2},
  [503] = {.lex_state = 12, .external_lex_state = 2},
  [504] = {.lex_state = 13, .external_lex_state = 2},
  [505] = {.lex_state = 13, .external_lex_state = 2},
  [506] = {.lex_state = 12, .external_lex_state = 2},
  [507] = {.lex_state = 13, .external_lex_state = 2},
  [508] = {.lex_state = 13, .external_lex_state = 2},
  [509] = {.lex_state = 13, .external_lex_state = 2},
  [510] = {.lex_state = 13, .external_lex_state = 2},
  [511] = {.lex_state = 13, .external_lex_state = 2},
  [512] = {.lex_state = 13, .external_lex_state = 2},
  [513] = {.lex_state = 13, .external_lex_state = 2},
  [514] = {.lex_state = 12, .external_lex_state = 2},
  [515] = {.lex_state = 12, .external_lex_state = 2},
  [516] = {.lex_state = 12, .external_lex_state = 2},
  [517] = {.lex_state = 13, .external_lex_state = 2},
  [518] = {.lex_state = 12, .external_lex_state = 2},
  [519] = {.lex_state = 13, .external_lex_state = 2},
  [520] = {.lex_state = 13, .external_lex_state = 2},
  [521] = {.lex_state = 13, .external_lex_state = 2},
  [522] = {.lex_state = 13, .external_lex_state = 2},
  [523] = {.lex_state = 13, .external_lex_state = 2},
  [524] = {.lex_state = 13, .external_lex_state = 2},
  [525] = {.lex_state = 12, .external_lex_state = 2},
  [526] = {.lex_state = 13, .external_lex_state = 2},
  [527] = {.lex_state = 13, .external_lex_state = 2},
  [528] = {.lex_state = 13, .external_lex_state = 2},
  [529] = {.lex_state = 13, .external_lex_state = 2},
  [530] = {.lex_state = 13, .external_lex_state = 2},
  [531] = {.lex_state = 13, .external_lex_state = 2},
  [532] = {.lex_state = 13, .external_lex_state = 2},
  [533] = {.lex_state = 12, .external_lex_state = 2},
  [534] = {.lex_state = 13, .external_lex_state = 2},
  [535] = {.lex_state = 12, .external_lex_state = 2},
  [536] = {.lex_state = 13, .external_lex_state = 2},
  [537] = {.lex_state = 12, .external_lex_state = 2},
  [538] = {.lex_state = 12, .external_lex_state = 2},
  [539] = {.lex_state = 12, .external_lex_state = 2},
  [540] = {.lex_state = 13, .external_lex_state = 2},
  [541] = {.lex_state = 12, .external_lex_state = 2},
  [542] = {.lex_state = 13, .external_lex_state = 2},
  [543] = {.lex_state = 13, .external_lex_state = 2},
  [544] = {.lex_state = 12, .external_lex_state = 2},
  [545] = {.lex_state = 7, .external_lex_state = 3},
  [546] = {.lex_state = 5, .external_lex_state = 2},
  [547] = {.lex_state = 5, .external_lex_state = 2},
  [548] = {.lex_state = 7, .external_lex_state = 3},
  [549] = {.lex_state = 7, .external_lex_state = 3},
  [550] = {.lex_state = 7, .external_lex_state = 3},
  [551] = {.lex_state = 7, .external_lex_state = 3},
  [552] = {.lex_state = 7, .external_lex_state = 3},
  [553] = {.lex_state = 7, .external_lex_state = 3},
  [554] = {.lex_state = 7, .external_lex_state = 3},
  [555] = {.lex_state = 7, .external_lex_state = 3},
  [556] = {.lex_state = 5, .external_lex_state = 2},
  [557] = {.lex_state = 7, .external_lex_state = 3},
  [558] = {.lex_state = 11, .external_lex_state = 2},
  [559] = {.lex_state = 5, .external_lex_state = 2},
  [560] = {.lex_state = 5, .external_lex_state = 2},
  [561] = {.lex_state = 5, .external_lex_state = 2},
  [562] = {.lex_state = 5, .external_lex_state = 2},
  [563] = {.lex_state = 12, .external_lex_state = 2},
  [564] = {.lex_state = 12, .external_lex_state = 2},
  [565] = {.lex_state = 12, .external_lex_state = 2},
  [566] = {.lex_state = 5, .external_lex_state = 2},
  [567] = {.lex_state = 12, .external_lex_state = 2},
  [568] = {.lex_state = 12, .external_lex_state = 2},
  [569] = {.lex_state = 7, .external_lex_state = 3},
  [570] = {.lex_state = 5, .external_lex_state = 2},
  [571] = {.lex_state = 12, .external_lex_state = 2},
  [572] = {.lex_state = 12, .external_lex_state = 2},
  [573] = {.lex_state = 5, .external_lex_state = 2},
  [574] = {.lex_state = 12, .external_lex_state = 2},
  [575] = {.lex_state = 12, .external_lex_state = 2},
  [576] = {.lex_state = 5, .external_lex_state = 2},
  [577] = {.lex_state = 12, .external_lex_state = 2},
  [578] = {.lex_state = 12, .external_lex_state = 2},
  [579] = {.lex_state = 12, .external_lex_state = 2},
  [580] = {.lex_state = 5, .external_lex_state = 2},
  [581] = {.lex_state = 5, .external_lex_state = 2},
  [582] = {.lex_state = 12, .external_lex_state = 2},
  [583] = {.lex_state = 12, .external_lex_state = 2},
  [584] = {.lex_state = 12, .external_lex_state = 2},
  [585] = {.lex_state = 12, .external_lex_state = 2},
  [586] = {.lex_state = 5, .external_lex_state = 2},
  [587] = {.lex_state = 5, .external_lex_state = 2},
  [588] = {.lex_state = 5, .external_lex_state = 2},
  [589] = {.lex_state = 13, .external_lex_state = 2},
  [590] = {.lex_state = 5, .external_lex_state = 2},
  [591] = {.lex_state = 5, .external_lex_state = 2},
  [592] = {.lex_state = 13, .external_lex_state = 2},
  [593] = {.lex_state = 5, .external_lex_state = 2},
  [594] = {.lex_state = 7, .external_lex_state = 3},
  [595] = {.lex_state = 7, .external_lex_state = 3},
  [596] = {.lex_state = 7, .external_lex_state = 3},
  [597] = {.lex_state = 5, .external_lex_state = 2},
  [598] = {.lex_state = 5, .external_lex_state = 2},
  [599] = {.lex_state = 5, .external_lex_state = 2},
  [600] = {.lex_state = 5, .external_lex_state = 2},
  [601] = {.lex_state = 13, .external_lex_state = 2},
  [602] = {.lex_state = 5, .external_lex_state = 2},
  [603] = {.lex_state = 5, .external_lex_state = 2},
  [604] = {.lex_state = 5, .external_lex_state = 2},
  [605] = {.lex_state = 13, .external_lex_state = 2},
  [606] = {.lex_state = 5, .external_lex_state = 2},
  [607] = {.lex_state = 5, .external_lex_state = 2},
  [608] = {.lex_state = 5, .external_lex_state = 2},
  [609] = {.lex_state = 5, .external_lex_state = 2},
  [610] = {.lex_state = 7, .external_lex_state = 3},
  [611] = {.lex_state = 5, .external_lex_state = 2},
  [612] = {.lex_state = 5, .external_lex_state = 2},
  [613] = {.lex_state = 5, .external_lex_state = 2},
  [614] = {.lex_state = 13, .external_lex_state = 2},
  [615] = {.lex_state = 13, .external_lex_state = 2},
  [616] = {.lex_state = 5, .external_lex_state = 2},
  [617] = {.lex_state = 5, .external_lex_state = 2},
  [618] = {.lex_state = 5, .external_lex_state = 2},
  [619] = {.lex_state = 5, .external_lex_state = 2},
  [620] = {.lex_state = 7, .external_lex_state = 3},
  [621] = {.lex_state = 5, .external_lex_state = 2},
  [622] = {.lex_state = 7, .external_lex_state = 3},
  [623] = {.lex_state = 7, .external_lex_state = 3},
  [624] = {.lex_state = 7, .external_lex_state = 3},
  [625] = {.lex_state = 7, .external_lex_state = 3},
  [626] = {.lex_state = 7, .external_lex_state = 3},
  [627] = {.lex_state = 7, .external_lex_state = 3},
  [628] = {.lex_state = 7, .external_lex_state = 3},
  [629] = {.lex_state = 7, .external_lex_state = 3},
  [630] = {.lex_state = 7, .external_lex_state = 3},
  [631] = {.lex_state = 7, .external_lex_state = 3},
  [632] = {.lex_state = 7, .external_lex_state = 3},
  [633] = {.lex_state = 7, .external_lex_state = 3},
  [634] = {.lex_state = 7, .external_lex_state = 3},
  [635] = {.lex_state = 7, .external_lex_state = 3},
  [636] = {.lex_state = 7, .external_lex_state = 3},
  [637] = {.lex_state = 7, .external_lex_state = 3},
  [638] = {.lex_state = 7, .external_lex_state = 3},
  [639] = {.lex_state = 7, .external_lex_state = 3},
  [640] = {.lex_state = 7, .external_lex_state = 3},
  [641] = {.lex_state = 7, .external_lex_state = 3},
  [642] = {.lex_state = 7, .external_lex_state = 3},
  [643] = {.lex_state = 7, .external_lex_state = 3},
  [644] = {.lex_state = 7, .external_lex_state = 3},
  [645] = {.lex_state = 7, .external_lex_state = 3},
  [646] = {.lex_state = 7, .external_lex_state = 3},
  [647] = {.lex_state = 7, .external_lex_state = 3},
  [648] = {.lex_state = 7, .external_lex_state = 3},
  [649] = {.lex_state = 7, .external_lex_state = 3},
  [650] = {.lex_state = 7, .external_lex_state = 3},
  [651] = {.lex_state = 7, .external_lex_state = 3},
  [652] = {.lex_state = 7, .external_lex_state = 3},
  [653] = {.lex_state = 7, .external_lex_state = 3},
  [654] = {.lex_state = 7, .external_lex_state = 3},
  [655] = {.lex_state = 7, .external_lex_state = 3},
  [656] = {.lex_state = 7, .external_lex_state = 3},
  [657] = {.lex_state = 7, .external_lex_state = 3},
  [658] = {.lex_state = 7, .external_lex_state = 3},
  [659] = {.lex_state = 7, .external_lex_state = 3},
  [660] = {.lex_state = 7, .external_lex_state = 3},
  [661] = {.lex_state = 7, .external_lex_state = 3},
  [662] = {.lex_state = 7, .external_lex_state = 3},
  [663] = {.lex_state = 7, .external_lex_state = 3},
  [664] = {.lex_state = 7, .external_lex_state = 3},
  [665] = {.lex_state = 7, .external_lex_state = 3},
  [666] = {.lex_state = 7, .external_lex_state = 3},
  [667] = {.lex_state = 7, .external_lex_state = 3},
  [668] = {.lex_state = 7, .external_lex_state = 3},
  [669] = {.lex_state = 7, .external_lex_state = 3},
  [670] = {.lex_state = 7, .external_lex_state = 3},
  [671] = {.lex_state = 7, .external_lex_state = 3},
  [672] = {.lex_state = 7, .external_lex_state = 3},
  [673] = {.lex_state = 7, .external_lex_state = 3},
  [674] = {.lex_state = 7, .external_lex_state = 3},
  [675] = {.lex_state = 7, .external_lex_state = 3},
  [676] = {.lex_state = 7, .external_lex_state = 3},
  [677] = {.lex_state = 7, .external_lex_state = 3},
  [678] = {.lex_state = 7, .external_lex_state = 3},
  [679] = {.lex_state = 7, .external_lex_state = 3},
  [680] = {.lex_state = 7, .external_lex_state = 3},
  [681] = {.lex_state = 7, .external_lex_state = 3},
  [682] = {.lex_state = 7, .external_lex_state = 3},
  [683] = {.lex_state = 7, .external_lex_state = 3},
  [684] = {.lex_state = 7, .external_lex_state = 3},
  [685] = {.lex_state = 7, .external_lex_state = 3},
  [686] = {.lex_state = 7, .external_lex_state = 3},
  [687] = {.lex_state = 7, .external_lex_state = 3},
  [688] = {.lex_state = 7, .external_lex_state = 3},
  [689] = {.lex_state = 7, .external_lex_state = 3},
  [690] = {.lex_state = 7, .external_lex_state = 3},
  [691] = {.lex_state = 7, .external_lex_state = 3},
  [692] = {.lex_state = 7, .external_lex_state = 3},
  [693] = {.lex_state = 7, .external_lex_state = 3},
  [694] = {.lex_state = 7, .external_lex_state = 3},
  [695] = {.lex_state = 7, .external_lex_state = 3},
  [696] = {.lex_state = 7, .external_lex_state = 3},
  [697] = {.lex_state = 7, .external_lex_state = 3},
  [698] = {.lex_state = 7, .external_lex_state = 3},
  [699] = {.lex_state = 7, .external_lex_state = 3},
  [700] = {.lex_state = 7, .external_lex_state = 3},
  [701] = {.lex_state = 7, .external_lex_state = 3},
  [702] = {.lex_state = 7, .external_lex_state = 3},
  [703] = {.lex_state = 7, .external_lex_state = 3},
  [704] = {.lex_state = 7, .external_lex_state = 3},
  [705] = {.lex_state = 7, .external_lex_state = 3},
  [706] = {.lex_state = 7, .external_lex_state = 3},
  [707] = {.lex_state = 7, .external_lex_state = 3},
  [708] = {.lex_state = 7, .external_lex_state = 3},
  [709] = {.lex_state = 7, .external_lex_state = 3},
  [710] = {.lex_state = 7, .external_lex_state = 3},
  [711] = {.lex_state = 7, .external_lex_state = 3},
  [712] = {.lex_state = 7, .external_lex_state = 3},
  [713] = {.lex_state = 7, .external_lex_state = 3},
  [714] = {.lex_state = 7, .external_lex_state = 3},
  [715] = {.lex_state = 7, .external_lex_state = 3},
  [716] = {.lex_state = 7, .external_lex_state = 3},
  [717] = {.lex_state = 7, .external_lex_state = 3},
  [718] = {.lex_state = 7, .external_lex_state = 3},
  [719] = {.lex_state = 7, .external_lex_state = 3},
  [720] = {.lex_state = 7, .external_lex_state = 3},
  [721] = {.lex_state = 7, .external_lex_state = 3},
  [722] = {.lex_state = 7, .external_lex_state = 3},
  [723] = {.lex_state = 7, .external_lex_state = 3},
  [724] = {.lex_state = 7, .external_lex_state = 3},
  [725] = {.lex_state = 7, .external_lex_state = 3},
  [726] = {.lex_state = 7, .external_lex_state = 3},
  [727] = {.lex_state = 7, .external_lex_state = 3},
  [728] = {.lex_state = 7, .external_lex_state = 3},
  [729] = {.lex_state = 7, .external_lex_state = 3},
  [730] = {.lex_state = 7, .external_lex_state = 3},
  [731] = {.lex_state = 7, .external_lex_state = 3},
  [732] = {.lex_state = 5, .external_lex_state = 2},
  [733] = {.lex_state = 7, .external_lex_state = 3},
  [734] = {.lex_state = 7, .external_lex_state = 3},
  [735] = {.lex_state = 7, .external_lex_state = 3},
  [736] = {.lex_state = 7, .external_lex_state = 3},
  [737] = {.lex_state = 7, .external_lex_state = 3},
  [738] = {.lex_state = 7, .external_lex_state = 3},
  [739] = {.lex_state = 7, .external_lex_state = 3},
  [740] = {.lex_state = 7, .external_lex_state = 3},
  [741] = {.lex_state = 7, .external_lex_state = 3},
  [742] = {.lex_state = 7, .external_lex_state = 3},
  [743] = {.lex_state = 7, .external_lex_state = 3},
  [744] = {.lex_state = 7, .external_lex_state = 3},
  [745] = {.lex_state = 7, .external_lex_state = 3},
  [746] = {.lex_state = 7, .external_lex_state = 3},
  [747] = {.lex_state = 5, .external_lex_state = 2},
  [748] = {.lex_state = 5, .external_lex_state = 2},
  [749] = {.lex_state = 5, .external_lex_state = 2},
  [750] = {.lex_state = 5, .external_lex_state = 2},
  [751] = {.lex_state = 6, .external_lex_state = 2},
  [752] = {.lex_state = 3, .external_lex_state = 3},
  [753] = {.lex_state = 3, .external_lex_state = 3},
  [754] = {.lex_state = 3, .external_lex_state = 3},
  [755] = {.lex_state = 3, .external_lex_state = 3},
  [756] = {.lex_state = 3, .external_lex_state = 3},
  [757] = {.lex_state = 3, .external_lex_state = 3},
  [758] = {.lex_state = 2, .external_lex_state = 3},
  [759] = {.lex_state = 3, .external_lex_state = 3},
  [760] = {.lex_state = 3, .external_lex_state = 3},
  [761] = {.lex_state = 3, .external_lex_state = 3},
  [762] = {.lex_state = 3, .external_lex_state = 3},
  [763] = {.lex_state = 4, .external_lex_state = 3},
  [764] = {.lex_state = 2, .external_lex_state = 3},
  [765] = {.lex_state = 2, .external_lex_state = 3},
  [766] = {.lex_state = 2, .external_lex_state = 3},
  [767] = {.lex_state = 4, .external_lex_state = 3},
  [768] = {.lex_state = 4, .external_lex_state = 3},
  [769] = {.lex_state = 2, .external_lex_state = 3},
  [770] = {.lex_state = 2, .external_lex_state = 3},
  [771] = {.lex_state = 2, .external_lex_state = 3},
  [772] = {.lex_state = 2, .external_lex_state = 3},
  [773] = {.lex_state = 4, .external_lex_state = 3},
  [774] = {.lex_state = 4, .external_lex_state = 3},
  [775] = {.lex_state = 2, .external_lex_state = 3},
  [776] = {.lex_state = 2, .external_lex_state = 3},
  [777] = {.lex_state = 2, .external_lex_state = 3},
  [778] = {.lex_state = 2, .external_lex_state = 3},
  [779] = {.lex_state = 2, .external_lex_state = 3},
  [780] = {.lex_state = 2, .external_lex_state = 3},
  [781] = {.lex_state = 2, .external_lex_state = 3},
  [782] = {.lex_state = 2, .external_lex_state = 3},
  [783] = {.lex_state = 2, .external_lex_state = 3},
  [784] = {.lex_state = 10, .external_lex_state = 3},
  [785] = {.lex_state = 2, .external_lex_state = 3},
  [786] = {.lex_state = 2, .external_lex_state = 3},
  [787] = {.lex_state = 2, .external_lex_state = 3},
  [788] = {.lex_state = 2, .external_lex_state = 3},
  [789] = {.lex_state = 3, .external_lex_state = 3},
  [790] = {.lex_state = 2, .external_lex_state = 3},
  [791] = {.lex_state = 2, .external_lex_state = 3},
  [792] = {.lex_state = 2, .external_lex_state = 3},
  [793] = {.lex_state = 2, .external_lex_state = 3},
  [794] = {.lex_state = 2, .external_lex_state = 3},
  [795] = {.lex_state = 2, .external_lex_state = 3},
  [796] = {.lex_state = 2, .external_lex_state = 3},
  [797] = {.lex_state = 2, .external_lex_state = 3},
  [798] = {.lex_state = 2, .external_lex_state = 3},
  [799] = {.lex_state = 2, .external_lex_state = 3},
  [800] = {.lex_state = 2, .external_lex_state = 3},
  [801] = {.lex_state = 2, .external_lex_state = 3},
  [802] = {.lex_state = 2, .external_lex_state = 3},
  [803] = {.lex_state = 2, .external_lex_state = 3},
  [804] = {.lex_state = 2, .external_lex_state = 3},
  [805] = {.lex_state = 2, .external_lex_state = 3},
  [806] = {.lex_state = 2, .external_lex_state = 3},
  [807] = {.lex_state = 2, .external_lex_state = 3},
  [808] = {.lex_state = 4, .external_lex_state = 3},
  [809] = {.lex_state = 4, .external_lex_state = 3},
  [810] = {.lex_state = 4, .external_lex_state = 3},
  [811] = {.lex_state = 4, .external_lex_state = 3},
  [812] = {.lex_state = 4, .external_lex_state = 3},
  [813] = {.lex_state = 2, .external_lex_state = 3},
  [814] = {.lex_state = 4, .external_lex_state = 3},
  [815] = {.lex_state = 2, .external_lex_state = 3},
  [816] = {.lex_state = 2, .external_lex_state = 3},
  [817] = {.lex_state = 2, .external_lex_state = 3},
  [818] = {.lex_state = 2, .external_lex_state = 3},
  [819] = {.lex_state = 4, .external_lex_state = 3},
  [820] = {.lex_state = 4, .external_lex_state = 3},
  [821] = {.lex_state = 4, .external_lex_state = 3},
  [822] = {.lex_state = 4, .external_lex_state = 3},
  [823] = {.lex_state = 4, .external_lex_state = 3},
  [824] = {.lex_state = 4, .external_lex_state = 3},
  [825] = {.lex_state = 4, .external_lex_state = 3},
  [826] = {.lex_state = 4, .external_lex_state = 3},
  [827] = {.lex_state = 4, .external_lex_state = 3},
  [828] = {.lex_state = 4, .external_lex_state = 3},
  [829] = {.lex_state = 4, .external_lex_state = 3},
  [830] = {.lex_state = 4, .external_lex_state = 3},
  [831] = {.lex_state = 4, .external_lex_state = 3},
  [832] = {.lex_state = 4, .external_lex_state = 3},
  [833] = {.lex_state = 4, .external_lex_state = 3},
  [834] = {.lex_state = 4, .external_lex_state = 3},
  [835] = {.lex_state = 4, .external_lex_state = 3},
  [836] = {.lex_state = 4, .external_lex_state = 3},
  [837] = {.lex_state = 4, .external_lex_state = 3},
  [838] = {.lex_state = 2, .external_lex_state = 3},
  [839] = {.lex_state = 4, .external_lex_state = 3},
  [840] = {.lex_state = 4, .external_lex_state = 3},
  [841] = {.lex_state = 4, .external_lex_state = 3},
  [842] = {.lex_state = 2, .external_lex_state = 3},
  [843] = {.lex_state = 2, .external_lex_state = 3},
  [844] = {.lex_state = 4, .external_lex_state = 3},
  [845] = {.lex_state = 4, .external_lex_state = 3},
  [846] = {.lex_state = 4, .external_lex_state = 3},
  [847] = {.lex_state = 4, .external_lex_state = 3},
  [848] = {.lex_state = 4, .external_lex_state = 3},
  [849] = {.lex_state = 4, .external_lex_state = 3},
  [850] = {.lex_state = 4, .external_lex_state = 3},
  [851] = {.lex_state = 4, .external_lex_state = 3},
  [852] = {.lex_state = 2, .external_lex_state = 3},
  [853] = {.lex_state = 4, .external_lex_state = 3},
  [854] = {.lex_state = 4, .external_lex_state = 3},
  [855] = {.lex_state = 4, .external_lex_state = 3},
  [856] = {.lex_state = 2, .external_lex_state = 3},
  [857] = {.lex_state = 4, .external_lex_state = 3},
  [858] = {.lex_state = 4, .external_lex_state = 3},
  [859] = {.lex_state = 4, .external_lex_state = 3},
  [860] = {.lex_state = 4, .external_lex_state = 3},
  [861] = {.lex_state = 4, .external_lex_state = 3},
  [862] = {.lex_state = 2, .external_lex_state = 3},
  [863] = {.lex_state = 4, .external_lex_state = 3},
  [864] = {.lex_state = 4, .external_lex_state = 3},
  [865] = {.lex_state = 2, .external_lex_state = 3},
  [866] = {.lex_state = 2, .external_lex_state = 3},
  [867] = {.lex_state = 2, .external_lex_state = 3},
  [868] = {.lex_state = 2, .external_lex_state = 3},
  [869] = {.lex_state = 2, .external_lex_state = 3},
  [870] = {.lex_state = 2, .external_lex_state = 3},
  [871] = {.lex_state = 2, .external_lex_state = 3},
  [872] = {.lex_state = 2, .external_lex_state = 3},
  [873] = {.lex_state = 4, .external_lex_state = 3},
  [874] = {.lex_state = 4, .external_lex_state = 3},
  [875] = {.lex_state = 4, .external_lex_state = 3},
  [876] = {.lex_state = 4, .external_lex_state = 3},
  [877] = {.lex_state = 4, .external_lex_state = 3},
  [878] = {.lex_state = 4, .external_lex_state = 3},
  [879] = {.lex_state = 2, .external_lex_state = 3},
  [880] = {.lex_state = 4, .external_lex_state = 3},
  [881] = {.lex_state = 4, .external_lex_state = 3},
  [882] = {.lex_state = 4, .external_lex_state = 3},
  [883] = {.lex_state = 2, .external_lex_state = 3},
  [884] = {.lex_state = 4, .external_lex_state = 3},
  [885] = {.lex_state = 4, .external_lex_state = 3},
  [886] = {.lex_state = 4, .external_lex_state = 3},
  [887] = {.lex_state = 4, .external_lex_state = 3},
  [888] = {.lex_state = 4, .external_lex_state = 3},
  [889] = {.lex_state = 4, .external_lex_state = 3},
  [890] = {.lex_state = 4, .external_lex_state = 3},
  [891] = {.lex_state = 2, .external_lex_state = 3},
  [892] = {.lex_state = 4, .external_lex_state = 3},
  [893] = {.lex_state = 4, .external_lex_state = 3},
  [894] = {.lex_state = 4, .external_lex_state = 3},
  [895] = {.lex_state = 4, .external_lex_state = 3},
  [896] = {.lex_state = 4, .external_lex_state = 3},
  [897] = {.lex_state = 4, .external_lex_state = 3},
  [898] = {.lex_state = 2, .external_lex_state = 3},
  [899] = {.lex_state = 4, .external_lex_state = 3},
  [900] = {.lex_state = 4, .external_lex_state = 3},
  [901] = {.lex_state = 4, .external_lex_state = 3},
  [902] = {.lex_state = 4, .external_lex_state = 3},
  [903] = {.lex_state = 4, .external_lex_state = 3},
  [904] = {.lex_state = 4, .external_lex_state = 3},
  [905] = {.lex_state = 2, .external_lex_state = 3},
  [906] = {.lex_state = 4, .external_lex_state = 3},
  [907] = {.lex_state = 4, .external_lex_state = 3},
  [908] = {.lex_state = 4, .external_lex_state = 3},
  [909] = {.lex_state = 4, .external_lex_state = 3},
  [910] = {.lex_state = 4, .external_lex_state = 3},
  [911] = {.lex_state = 4, .external_lex_state = 3},
  [912] = {.lex_state = 4, .external_lex_state = 3},
  [913] = {.lex_state = 4, .external_lex_state = 3},
  [914] = {.lex_state = 4, .external_lex_state = 3},
  [915] = {.lex_state = 2, .external_lex_state = 3},
  [916] = {.lex_state = 4, .external_lex_state = 3},
  [917] = {.lex_state = 4, .external_lex_state = 3},
  [918] = {.lex_state = 4, .external_lex_state = 3},
  [919] = {.lex_state = 4, .external_lex_state = 3},
  [920] = {.lex_state = 4, .external_lex_state = 3},
  [921] = {.lex_state = 4, .external_lex_state = 3},
  [922] = {.lex_state = 2, .external_lex_state = 3},
  [923] = {.lex_state = 4, .external_lex_state = 3},
  [924] = {.lex_state = 2, .external_lex_state = 3},
  [925] = {.lex_state = 4, .external_lex_state = 3},
  [926] = {.lex_state = 4, .external_lex_state = 3},
  [927] = {.lex_state = 4, .external_lex_state = 3},
  [928] = {.lex_state = 4, .external_lex_state = 3},
  [929] = {.lex_state = 4, .external_lex_state = 3},
  [930] = {.lex_state = 4, .external_lex_state = 3},
  [931] = {.lex_state = 2, .external_lex_state = 3},
  [932] = {.lex_state = 2, .external_lex_state = 3},
  [933] = {.lex_state = 4, .external_lex_state = 3},
  [934] = {.lex_state = 4, .external_lex_state = 3},
  [935] = {.lex_state = 2, .external_lex_state = 3},
  [936] = {.lex_state = 4, .external_lex_state = 3},
  [937] = {.lex_state = 4, .external_lex_state = 3},
  [938] = {.lex_state = 4, .external_lex_state = 3},
  [939] = {.lex_state = 4, .external_lex_state = 3},
  [940] = {.lex_state = 4, .external_lex_state = 3},
  [941] = {.lex_state = 2, .external_lex_state = 3},
  [942] = {.lex_state = 2, .external_lex_state = 3},
  [943] = {.lex_state = 4, .external_lex_state = 3},
  [944] = {.lex_state = 4, .external_lex_state = 3},
  [945] = {.lex_state = 4, .external_lex_state = 3},
  [946] = {.lex_state = 2, .external_lex_state = 3},
  [947] = {.lex_state = 4, .external_lex_state = 3},
  [948] = {.lex_state = 2, .external_lex_state = 3},
  [949] = {.lex_state = 2, .external_lex_state = 3},
  [950] = {.lex_state = 2, .external_lex_state = 3},
  [951] = {.lex_state = 2, .external_lex_state = 3},
  [952] = {.lex_state = 2, .external_lex_state = 3},
  [953] = {.lex_state = 2, .external_lex_state = 3},
  [954] = {.lex_state = 4, .external_lex_state = 3},
  [955] = {.lex_state = 4, .external_lex_state = 3},
  [956] = {.lex_state = 4, .external_lex_state = 3},
  [957] = {.lex_state = 4, .external_lex_state = 3},
  [958] = {.lex_state = 4, .external_lex_state = 3},
  [959] = {.lex_state = 4, .external_lex_state = 3},
  [960] = {.lex_state = 4, .external_lex_state = 3},
  [961] = {.lex_state = 2, .external_lex_state = 3},
  [962] = {.lex_state = 4, .external_lex_state = 3},
  [963] = {.lex_state = 4, .external_lex_state = 3},
  [964] = {.lex_state = 4, .external_lex_state = 3},
  [965] = {.lex_state = 4, .external_lex_state = 3},
  [966] = {.lex_state = 2, .external_lex_state = 3},
  [967] = {.lex_state = 2, .external_lex_state = 3},
  [968] = {.lex_state = 4, .external_lex_state = 3},
  [969] = {.lex_state = 4, .external_lex_state = 3},
  [970] = {.lex_state = 4, .external_lex_state = 3},
  [971] = {.lex_state = 4, .external_lex_state = 3},
  [972] = {.lex_state = 4, .external_lex_state = 3},
  [973] = {.lex_state = 2, .external_lex_state = 3},
  [974] = {.lex_state = 2, .external_lex_state = 3},
  [975] = {.lex_state = 4, .external_lex_state = 3},
  [976] = {.lex_state = 4, .external_lex_state = 3},
  [977] = {.lex_state = 4, .external_lex_state = 3},
  [978] = {.lex_state = 4, .external_lex_state = 3},
  [979] = {.lex_state = 2, .external_lex_state = 3},
  [980] = {.lex_state = 4, .external_lex_state = 3},
  [981] = {.lex_state = 4, .external_lex_state = 3},
  [982] = {.lex_state = 4, .external_lex_state = 3},
  [983] = {.lex_state = 4, .external_lex_state = 3},
  [984] = {.lex_state = 4, .external_lex_state = 3},
  [985] = {.lex_state = 2, .external_lex_state = 3},
  [986] = {.lex_state = 4, .external_lex_state = 3},
  [987] = {.lex_state = 2, .external_lex_state = 3},
  [988] = {.lex_state = 4, .external_lex_state = 3},
  [989] = {.lex_state = 2, .external_lex_state = 3},
  [990] = {.lex_state = 4, .external_lex_state = 3},
  [991] = {.lex_state = 4, .external_lex_state = 3},
  [992] = {.lex_state = 4, .external_lex_state = 3},
  [993] = {.lex_state = 4, .external_lex_state = 3},
  [994] = {.lex_state = 4, .external_lex_state = 3},
  [995] = {.lex_state = 4, .external_lex_state = 3},
  [996] = {.lex_state = 4, .external_lex_state = 3},
  [997] = {.lex_state = 4, .external_lex_state = 3},
  [998] = {.lex_state = 4, .external_lex_state = 3},
  [999] = {.lex_state = 4, .external_lex_state = 3},
  [1000] = {.lex_state = 4, .external_lex_state = 3},
  [1001] = {.lex_state = 4, .external_lex_state = 3},
  [1002] = {.lex_state = 2, .external_lex_state = 3},
  [1003] = {.lex_state = 4, .external_lex_state = 3},
  [1004] = {.lex_state = 4, .external_lex_state = 3},
  [1005] = {.lex_state = 2, .external_lex_state = 3},
  [1006] = {.lex_state = 4, .external_lex_state = 3},
  [1007] = {.lex_state = 4, .external_lex_state = 3},
  [1008] = {.lex_state = 4, .external_lex_state = 3},
  [1009] = {.lex_state = 2, .external_lex_state = 3},
  [1010] = {.lex_state = 4, .external_lex_state = 3},
  [1011] = {.lex_state = 2, .external_lex_state = 3},
  [1012] = {.lex_state = 2, .external_lex_state = 3},
  [1013] = {.lex_state = 4, .external_lex_state = 3},
  [1014] = {.lex_state = 4, .external_lex_state = 3},
  [1015] = {.lex_state = 2, .external_lex_state = 3},
  [1016] = {.lex_state = 4, .external_lex_state = 3},
  [1017] = {.lex_state = 2, .external_lex_state = 3},
  [1018] = {.lex_state = 4, .external_lex_state = 3},
  [1019] = {.lex_state = 4, .external_lex_state = 3},
  [1020] = {.lex_state = 4, .external_lex_state = 3},
  [1021] = {.lex_state = 4, .external_lex_state = 3},
  [1022] = {.lex_state = 4, .external_lex_state = 3},
  [1023] = {.lex_state = 2, .external_lex_state = 3},
  [1024] = {.lex_state = 2, .external_lex_state = 3},
  [1025] = {.lex_state = 2, .external_lex_state = 3},
  [1026] = {.lex_state = 4, .external_lex_state = 3},
  [1027] = {.lex_state = 4, .external_lex_state = 3},
  [1028] = {.lex_state = 2, .external_lex_state = 3},
  [1029] = {.lex_state = 2, .external_lex_state = 3},
  [1030] = {.lex_state = 4, .external_lex_state = 3},
  [1031] = {.lex_state = 2, .external_lex_state = 3},
  [1032] = {.lex_state = 4, .external_lex_state = 3},
  [1033] = {.lex_state = 4, .external_lex_state = 3},
  [1034] = {.lex_state = 4, .external_lex_state = 3},
  [1035] = {.lex_state = 4, .external_lex_state = 3},
  [1036] = {.lex_state = 4, .external_lex_state = 3},
  [1037] = {.lex_state = 4, .external_lex_state = 3},
  [1038] = {.lex_state = 2, .external_lex_state = 3},
  [1039] = {.lex_state = 2, .external_lex_state = 3},
  [1040] = {.lex_state = 2, .external_lex_state = 3},
  [1041] = {.lex_state = 4, .external_lex_state = 3},
  [1042] = {.lex_state = 2, .external_lex_state = 3},
  [1043] = {.lex_state = 4, .external_lex_state = 3},
  [1044] = {.lex_state = 4, .external_lex_state = 3},
  [1045] = {.lex_state = 2, .external_lex_state = 3},
  [1046] = {.lex_state = 2, .external_lex_state = 3},
  [1047] = {.lex_state = 2, .external_lex_state = 3},
  [1048] = {.lex_state = 4, .external_lex_state = 3},
  [1049] = {.lex_state = 2, .external_lex_state = 3},
  [1050] = {.lex_state = 4, .external_lex_state = 3},
  [1051] = {.lex_state = 2, .external_lex_state = 3},
  [1052] = {.lex_state = 2, .external_lex_state = 3},
  [1053] = {.lex_state = 4, .external_lex_state = 3},
  [1054] = {.lex_state = 2, .external_lex_state = 3},
  [1055] = {.lex_state = 4, .external_lex_state = 3},
  [1056] = {.lex_state = 4, .external_lex_state = 3},
  [1057] = {.lex_state = 4, .external_lex_state = 3},
  [1058] = {.lex_state = 4, .external_lex_state = 3},
  [1059] = {.lex_state = 4, .external_lex_state = 3},
  [1060] = {.lex_state = 2, .external_lex_state = 3},
  [1061] = {.lex_state = 2, .external_lex_state = 3},
  [1062] = {.lex_state = 4, .external_lex_state = 3},
  [1063] = {.lex_state = 4, .external_lex_state = 3},
  [1064] = {.lex_state = 4, .external_lex_state = 3},
  [1065] = {.lex_state = 4, .external_lex_state = 3},
  [1066] = {.lex_state = 4, .external_lex_state = 3},
  [1067] = {.lex_state = 2, .external_lex_state = 3},
  [1068] = {.lex_state = 4, .external_lex_state = 3},
  [1069] = {.lex_state = 2, .external_lex_state = 3},
  [1070] = {.lex_state = 2, .external_lex_state = 3},
  [1071] = {.lex_state = 2, .external_lex_state = 3},
  [1072] = {.lex_state = 4, .external_lex_state = 3},
  [1073] = {.lex_state = 2, .external_lex_state = 3},
  [1074] = {.lex_state = 4, .external_lex_state = 3},
  [1075] = {.lex_state = 4, .external_lex_state = 3},
  [1076] = {.lex_state = 4, .external_lex_state = 3},
  [1077] = {.lex_state = 4, .external_lex_state = 3},
  [1078] = {.lex_state = 4, .external_lex_state = 3},
  [1079] = {.lex_state = 2, .external_lex_state = 3},
  [1080] = {.lex_state = 4, .external_lex_state = 3},
  [1081] = {.lex_state = 2, .external_lex_state = 3},
  [1082] = {.lex_state = 2, .external_lex_state = 3},
  [1083] = {.lex_state = 2, .external_lex_state = 3},
  [1084] = {.lex_state = 4, .external_lex_state = 3},
  [1085] = {.lex_state = 4, .external_lex_state = 3},
  [1086] = {.lex_state = 4, .external_lex_state = 3},
  [1087] = {.lex_state = 2, .external_lex_state = 3},
  [1088] = {.lex_state = 4, .external_lex_state = 3},
  [1089] = {.lex_state = 4, .external_lex_state = 3},
  [1090] = {.lex_state = 4, .external_lex_state = 3},
  [1091] = {.lex_state = 4, .external_lex_state = 3},
  [1092] = {.lex_state = 4, .external_lex_state = 3},
  [1093] = {.lex_state = 4, .external_lex_state = 3},
  [1094] = {.lex_state = 4, .external_lex_state = 3},
  [1095] = {.lex_state = 4, .external_lex_state = 3},
  [1096] = {.lex_state = 4, .external_lex_state = 3},
  [1097] = {.lex_state = 4, .external_lex_state = 3},
  [1098] = {.lex_state = 4, .external_lex_state = 3},
  [1099] = {.lex_state = 4, .external_lex_state = 3},
  [1100] = {.lex_state = 4, .external_lex_state = 3},
  [1101] = {.lex_state = 4, .external_lex_state = 3},
  [1102] = {.lex_state = 4, .external_lex_state = 3},
  [1103] = {.lex_state = 4, .external_lex_state = 3},
  [1104] = {.lex_state = 4, .external_lex_state = 3},
  [1105] = {.lex_state = 4, .external_lex_state = 3},
  [1106] = {.lex_state = 4, .external_lex_state = 3},
  [1107] = {.lex_state = 4, .external_lex_state = 3},
  [1108] = {.lex_state = 4, .external_lex_state = 3},
  [1109] = {.lex_state = 4, .external_lex_state = 3},
  [1110] = {.lex_state = 4, .external_lex_state = 3},
  [1111] = {.lex_state = 2, .external_lex_state = 3},
  [1112] = {.lex_state = 4, .external_lex_state = 3},
  [1113] = {.lex_state = 4, .external_lex_state = 3},
  [1114] = {.lex_state = 4, .external_lex_state = 3},
  [1115] = {.lex_state = 4, .external_lex_state = 3},
  [1116] = {.lex_state = 4, .external_lex_state = 3},
  [1117] = {.lex_state = 4, .external_lex_state = 3},
  [1118] = {.lex_state = 2, .external_lex_state = 3},
  [1119] = {.lex_state = 2, .external_lex_state = 3},
  [1120] = {.lex_state = 4, .external_lex_state = 3},
  [1121] = {.lex_state = 2, .external_lex_state = 3},
  [1122] = {.lex_state = 4, .external_lex_state = 3},
  [1123] = {.lex_state = 4, .external_lex_state = 3},
  [1124] = {.lex_state = 4, .external_lex_state = 3},
  [1125] = {.lex_state = 4, .external_lex_state = 3},
  [1126] = {.lex_state = 4, .external_lex_state = 3},
  [1127] = {.lex_state = 2, .external_lex_state = 3},
  [1128] = {.lex_state = 2, .external_lex_state = 3},
  [1129] = {.lex_state = 2, .external_lex_state = 3},
  [1130] = {.lex_state = 2, .external_lex_state = 3},
  [1131] = {.lex_state = 7, .external_lex_state = 3},
  [1132] = {.lex_state = 2, .external_lex_state = 3},
  [1133] = {.lex_state = 7, .external_lex_state = 3},
  [1134] = {.lex_state = 7, .external_lex_state = 3},
  [1135] = {.lex_state = 7, .external_lex_state = 3},
  [1136] = {.lex_state = 2, .external_lex_state = 3},
  [1137] = {.lex_state = 5, .external_lex_state = 2},
  [1138] = {.lex_state = 2, .external_lex_state = 3},
  [1139] = {.lex_state = 7, .external_lex_state = 3},
  [1140] = {.lex_state = 2, .external_lex_state = 3},
  [1141] = {.lex_state = 7, .external_lex_state = 3},
  [1142] = {.lex_state = 2, .external_lex_state = 3},
  [1143] = {.lex_state = 2, .external_lex_state = 3},
  [1144] = {.lex_state = 2, .external_lex_state = 3},
  [1145] = {.lex_state = 2, .external_lex_state = 3},
  [1146] = {.lex_state = 2, .external_lex_state = 3},
  [1147] = {.lex_state = 2, .external_lex_state = 3},
  [1148] = {.lex_state = 2, .external_lex_state = 3},
  [1149] = {.lex_state = 5, .external_lex_state = 2},
  [1150] = {.lex_state = 7, .external_lex_state = 3},
  [1151] = {.lex_state = 7, .external_lex_state = 3},
  [1152] = {.lex_state = 2, .external_lex_state = 3},
  [1153] = {.lex_state = 7, .external_lex_state = 3},
  [1154] = {.lex_state = 2, .external_lex_state = 3},
  [1155] = {.lex_state = 2, .external_lex_state = 3},
  [1156] = {.lex_state = 2, .external_lex_state = 3},
  [1157] = {.lex_state = 2, .external_lex_state = 3},
  [1158] = {.lex_state = 2, .external_lex_state = 3},
  [1159] = {.lex_state = 5, .external_lex_state = 2},
  [1160] = {.lex_state = 5, .external_lex_state = 2},
  [1161] = {.lex_state = 2, .external_lex_state = 3},
  [1162] = {.lex_state = 7, .external_lex_state = 3},
  [1163] = {.lex_state = 7, .external_lex_state = 3},
  [1164] = {.lex_state = 7, .external_lex_state = 3},
  [1165] = {.lex_state = 7, .external_lex_state = 3},
  [1166] = {.lex_state = 2, .external_lex_state = 3},
  [1167] = {.lex_state = 7, .external_lex_state = 3},
  [1168] = {.lex_state = 7, .external_lex_state = 3},
  [1169] = {.lex_state = 7, .external_lex_state = 3},
  [1170] = {.lex_state = 7, .external_lex_state = 3},
  [1171] = {.lex_state = 2, .external_lex_state = 3},
  [1172] = {.lex_state = 2, .external_lex_state = 3},
  [1173] = {.lex_state = 5, .external_lex_state = 2},
  [1174] = {.lex_state = 5, .external_lex_state = 2},
  [1175] = {.lex_state = 2, .external_lex_state = 3},
  [1176] = {.lex_state = 2, .external_lex_state = 3},
  [1177] = {.lex_state = 2, .external_lex_state = 3},
  [1178] = {.lex_state = 7, .external_lex_state = 3},
  [1179] = {.lex_state = 7, .external_lex_state = 3},
  [1180] = {.lex_state = 7, .external_lex_state = 3},
  [1181] = {.lex_state = 2, .external_lex_state = 3},
  [1182] = {.lex_state = 2, .external_lex_state = 3},
  [1183] = {.lex_state = 7, .external_lex_state = 3},
  [1184] = {.lex_state = 7, .external_lex_state = 3},
  [1185] = {.lex_state = 7, .external_lex_state = 3},
  [1186] = {.lex_state = 2, .external_lex_state = 3},
  [1187] = {.lex_state = 2, .external_lex_state = 3},
  [1188] = {.lex_state = 7, .external_lex_state = 3},
  [1189] = {.lex_state = 2, .external_lex_state = 3},
  [1190] = {.lex_state = 2, .external_lex_state = 3},
  [1191] = {.lex_state = 2, .external_lex_state = 3},
  [1192] = {.lex_state = 2, .external_lex_state = 3},
  [1193] = {.lex_state = 2, .external_lex_state = 3},
  [1194] = {.lex_state = 2, .external_lex_state = 3},
  [1195] = {.lex_state = 7, .external_lex_state = 3},
  [1196] = {.lex_state = 7, .external_lex_state = 3},
  [1197] = {.lex_state = 2, .external_lex_state = 3},
  [1198] = {.lex_state = 2, .external_lex_state = 3},
  [1199] = {.lex_state = 2, .external_lex_state = 3},
  [1200] = {.lex_state = 2, .external_lex_state = 3},
  [1201] = {.lex_state = 2, .external_lex_state = 3},
  [1202] = {.lex_state = 2, .external_lex_state = 3},
  [1203] = {.lex_state = 2, .external_lex_state = 3},
  [1204] = {.lex_state = 2, .external_lex_state = 3},
  [1205] = {.lex_state = 2, .external_lex_state = 3},
  [1206] = {.lex_state = 2, .external_lex_state = 3},
  [1207] = {.lex_state = 7, .external_lex_state = 3},
  [1208] = {.lex_state = 2, .external_lex_state = 3},
  [1209] = {.lex_state = 2, .external_lex_state = 3},
  [1210] = {.lex_state = 2, .external_lex_state = 3},
  [1211] = {.lex_state = 2, .external_lex_state = 3},
  [1212] = {.lex_state = 2, .external_lex_state = 3},
  [1213] = {.lex_state = 2, .external_lex_state = 3},
  [1214] = {.lex_state = 2, .external_lex_state = 3},
  [1215] = {.lex_state = 2, .external_lex_state = 3},
  [1216] = {.lex_state = 2, .external_lex_state = 3},
  [1217] = {.lex_state = 2, .external_lex_state = 3},
  [1218] = {.lex_state = 2, .external_lex_state = 3},
  [1219] = {.lex_state = 2, .external_lex_state = 3},
  [1220] = {.lex_state = 2, .external_lex_state = 3},
  [1221] = {.lex_state = 2, .external_lex_state = 3},
  [1222] = {.lex_state = 2, .external_lex_state = 3},
  [1223] = {.lex_state = 2, .external_lex_state = 3},
  [1224] = {.lex_state = 2, .external_lex_state = 3},
  [1225] = {.lex_state = 2, .external_lex_state = 3},
  [1226] = {.lex_state = 2, .external_lex_state = 3},
  [1227] = {.lex_state = 2, .external_lex_state = 3},
  [1228] = {.lex_state = 2, .external_lex_state = 3},
  [1229] = {.lex_state = 2, .external_lex_state = 3},
  [1230] = {.lex_state = 2, .external_lex_state = 3},
  [1231] = {.lex_state = 2, .external_lex_state = 3},
  [1232] = {.lex_state = 2, .external_lex_state = 3},
  [1233] = {.lex_state = 2, .external_lex_state = 3},
  [1234] = {.lex_state = 2, .external_lex_state = 3},
  [1235] = {.lex_state = 2, .external_lex_state = 3},
  [1236] = {.lex_state = 2, .external_lex_state = 3},
  [1237] = {.lex_state = 2, .external_lex_state = 3},
  [1238] = {.lex_state = 2, .external_lex_state = 3},
  [1239] = {.lex_state = 2, .external_lex_state = 3},
  [1240] = {.lex_state = 2, .external_lex_state = 3},
  [1241] = {.lex_state = 2, .external_lex_state = 3},
  [1242] = {.lex_state = 2, .external_lex_state = 3},
  [1243] = {.lex_state = 2, .external_lex_state = 3},
  [1244] = {.lex_state = 2, .external_lex_state = 3},
  [1245] = {.lex_state = 2, .external_lex_state = 3},
  [1246] = {.lex_state = 2, .external_lex_state = 3},
  [1247] = {.lex_state = 2, .external_lex_state = 3},
  [1248] = {.lex_state = 2, .external_lex_state = 3},
  [1249] = {.lex_state = 2, .external_lex_state = 3},
  [1250] = {.lex_state = 2, .external_lex_state = 3},
  [1251] = {.lex_state = 2, .external_lex_state = 3},
  [1252] = {.lex_state = 2, .external_lex_state = 3},
  [1253] = {.lex_state = 2, .external_lex_state = 3},
  [1254] = {.lex_state = 2, .external_lex_state = 3},
  [1255] = {.lex_state = 2, .external_lex_state = 3},
  [1256] = {.lex_state = 2, .external_lex_state = 3},
  [1257] = {.lex_state = 2, .external_lex_state = 3},
  [1258] = {.lex_state = 2, .external_lex_state = 3},
  [1259] = {.lex_state = 2, .external_lex_state = 3},
  [1260] = {.lex_state = 2, .external_lex_state = 3},
  [1261] = {.lex_state = 2, .external_lex_state = 3},
  [1262] = {.lex_state = 2, .external_lex_state = 3},
  [1263] = {.lex_state = 2, .external_lex_state = 3},
  [1264] = {.lex_state = 2, .external_lex_state = 3},
  [1265] = {.lex_state = 2, .external_lex_state = 3},
  [1266] = {.lex_state = 2, .external_lex_state = 3},
  [1267] = {.lex_state = 2, .external_lex_state = 3},
  [1268] = {.lex_state = 2, .external_lex_state = 3},
  [1269] = {.lex_state = 2, .external_lex_state = 3},
  [1270] = {.lex_state = 2, .external_lex_state = 3},
  [1271] = {.lex_state = 2, .external_lex_state = 3},
  [1272] = {.lex_state = 2, .external_lex_state = 3},
  [1273] = {.lex_state = 2, .external_lex_state = 3},
  [1274] = {.lex_state = 2, .external_lex_state = 3},
  [1275] = {.lex_state = 2, .external_lex_state = 3},
  [1276] = {.lex_state = 2, .external_lex_state = 3},
  [1277] = {.lex_state = 2, .external_lex_state = 3},
  [1278] = {.lex_state = 2, .external_lex_state = 3},
  [1279] = {.lex_state = 2, .external_lex_state = 3},
  [1280] = {.lex_state = 2, .external_lex_state = 3},
  [1281] = {.lex_state = 2, .external_lex_state = 3},
  [1282] = {.lex_state = 2, .external_lex_state = 3},
  [1283] = {.lex_state = 2, .external_lex_state = 3},
  [1284] = {.lex_state = 2, .external_lex_state = 3},
  [1285] = {.lex_state = 2, .external_lex_state = 3},
  [1286] = {.lex_state = 2, .external_lex_state = 3},
  [1287] = {.lex_state = 2, .external_lex_state = 3},
  [1288] = {.lex_state = 2, .external_lex_state = 3},
  [1289] = {.lex_state = 7, .external_lex_state = 3},
  [1290] = {.lex_state = 2, .external_lex_state = 3},
  [1291] = {.lex_state = 2, .external_lex_state = 3},
  [1292] = {.lex_state = 2, .external_lex_state = 3},
  [1293] = {.lex_state = 2, .external_lex_state = 3},
  [1294] = {.lex_state = 2, .external_lex_state = 3},
  [1295] = {.lex_state = 7, .external_lex_state = 3},
  [1296] = {.lex_state = 7, .external_lex_state = 3},
  [1297] = {.lex_state = 7, .external_lex_state = 3},
  [1298] = {.lex_state = 7, .external_lex_state = 3},
  [1299] = {.lex_state = 7, .external_lex_state = 3},
  [1300] = {.lex_state = 7, .external_lex_state = 3},
  [1301] = {.lex_state = 7, .external_lex_state = 3},
  [1302] = {.lex_state = 7, .external_lex_state = 3},
  [1303] = {.lex_state = 7, .external_lex_state = 3},
  [1304] = {.lex_state = 7, .external_lex_state = 3},
  [1305] = {.lex_state = 7, .external_lex_state = 3},
  [1306] = {.lex_state = 7, .external_lex_state = 3},
  [1307] = {.lex_state = 7, .external_lex_state = 3},
  [1308] = {.lex_state = 7, .external_lex_state = 3},
  [1309] = {.lex_state = 7, .external_lex_state = 3},
  [1310] = {.lex_state = 7, .external_lex_state = 3},
  [1311] = {.lex_state = 7, .external_lex_state = 3},
  [1312] = {.lex_state = 7, .external_lex_state = 3},
  [1313] = {.lex_state = 18, .external_lex_state = 3},
  [1314] = {.lex_state = 18, .external_lex_state = 3},
  [1315] = {.lex_state = 9, .external_lex_state = 3},
  [1316] = {.lex_state = 9, .external_lex_state = 3},
  [1317] = {.lex_state = 9, .external_lex_state = 3},
  [1318] = {.lex_state = 9, .external_lex_state = 3},
  [1319] = {.lex_state = 9, .external_lex_state = 3},
  [1320] = {.lex_state = 9, .external_lex_state = 3},
  [1321] = {.lex_state = 9, .external_lex_state = 3},
  [1322] = {.lex_state = 9, .external_lex_state = 3},
  [1323] = {.lex_state = 9, .external_lex_state = 3},
  [1324] = {.lex_state = 9, .external_lex_state = 3},
  [1325] = {.lex_state = 9, .external_lex_state = 3},
  [1326] = {.lex_state = 9, .external_lex_state = 3},
  [1327] = {.lex_state = 18, .external_lex_state = 3},
  [1328] = {.lex_state = 7, .external_lex_state = 3},
  [1329] = {.lex_state = 18, .external_lex_state = 3},
  [1330] = {.lex_state = 7, .external_lex_state = 3},
  [1331] = {.lex_state = 18, .external_lex_state = 3},
  [1332] = {.lex_state = 7, .external_lex_state = 3},
  [1333] = {.lex_state = 18, .external_lex_state = 3},
  [1334] = {.lex_state = 7, .external_lex_state = 3},
  [1335] = {.lex_state = 18, .external_lex_state = 3},
  [1336] = {.lex_state = 18, .external_lex_state = 3},
  [1337] = {.lex_state = 7, .external_lex_state = 3},
  [1338] = {.lex_state = 18, .external_lex_state = 3},
  [1339] = {.lex_state = 7, .external_lex_state = 3},
  [1340] = {.lex_state = 18, .external_lex_state = 3},
  [1341] = {.lex_state = 7, .external_lex_state = 3},
  [1342] = {.lex_state = 7, .external_lex_state = 3},
  [1343] = {.lex_state = 18, .external_lex_state = 3},
  [1344] = {.lex_state = 7, .external_lex_state = 3},
  [1345] = {.lex_state = 7, .external_lex_state = 3},
  [1346] = {.lex_state = 7, .external_lex_state = 3},
  [1347] = {.lex_state = 7, .external_lex_state = 3},
  [1348] = {.lex_state = 7, .external_lex_state = 3},
  [1349] = {.lex_state = 18, .external_lex_state = 3},
  [1350] = {.lex_state = 18, .external_lex_state = 3},
  [1351] = {.lex_state = 7, .external_lex_state = 3},
  [1352] = {.lex_state = 18, .external_lex_state = 3},
  [1353] = {.lex_state = 18, .external_lex_state = 3},
  [1354] = {.lex_state = 18, .external_lex_state = 3},
  [1355] = {.lex_state = 7, .external_lex_state = 3},
  [1356] = {.lex_state = 18, .external_lex_state = 3},
  [1357] = {.lex_state = 18, .external_lex_state = 3},
  [1358] = {.lex_state = 18, .external_lex_state = 3},
  [1359] = {.lex_state = 7, .external_lex_state = 3},
  [1360] = {.lex_state = 7, .external_lex_state = 3},
  [1361] = {.lex_state = 18, .external_lex_state = 3},
  [1362] = {.lex_state = 7, .external_lex_state = 3},
  [1363] = {.lex_state = 18, .external_lex_state = 3},
  [1364] = {.lex_state = 7, .external_lex_state = 3},
  [1365] = {.lex_state = 18, .external_lex_state = 3},
  [1366] = {.lex_state = 7, .external_lex_state = 3},
  [1367] = {.lex_state = 18, .external_lex_state = 3},
  [1368] = {.lex_state = 7, .external_lex_state = 3},
  [1369] = {.lex_state = 9, .external_lex_state = 3},
  [1370] = {.lex_state = 18, .external_lex_state = 3},
  [1371] = {.lex_state = 7, .external_lex_state = 3},
  [1372] = {.lex_state = 7, .external_lex_state = 3},
  [1373] = {.lex_state = 9, .external_lex_state = 3},
  [1374] = {.lex_state = 7, .external_lex_state = 3},
  [1375] = {.lex_state = 9, .external_lex_state = 3},
  [1376] = {.lex_state = 18, .external_lex_state = 3},
  [1377] = {.lex_state = 18, .external_lex_state = 3},
  [1378] = {.lex_state = 18, .external_lex_state = 3},
  [1379] = {.lex_state = 18, .external_lex_state = 3},
  [1380] = {.lex_state = 9, .external_lex_state = 3},
  [1381] = {.lex_state = 18, .external_lex_state = 3},
  [1382] = {.lex_state = 18, .external_lex_state = 3},
  [1383] = {.lex_state = 18, .external_lex_state = 3},
  [1384] = {.lex_state = 18, .external_lex_state = 3},
  [1385] = {.lex_state = 18, .external_lex_state = 3},
  [1386] = {.lex_state = 18, .external_lex_state = 3},
  [1387] = {.lex_state = 18, .external_lex_state = 3},
  [1388] = {.lex_state = 18, .external_lex_state = 3},
  [1389] = {.lex_state = 7, .external_lex_state = 3},
  [1390] = {.lex_state = 18, .external_lex_state = 3},
  [1391] = {.lex_state = 7, .external_lex_state = 3},
  [1392] = {.lex_state = 18, .external_lex_state = 3},
  [1393] = {.lex_state = 18, .external_lex_state = 3},
  [1394] = {.lex_state = 18, .external_lex_state = 3},
  [1395] = {.lex_state = 7, .external_lex_state = 3},
  [1396] = {.lex_state = 18, .external_lex_state = 3},
  [1397] = {.lex_state = 18, .external_lex_state = 3},
  [1398] = {.lex_state = 18, .external_lex_state = 3},
  [1399] = {.lex_state = 18, .external_lex_state = 3},
  [1400] = {.lex_state = 18, .external_lex_state = 3},
  [1401] = {.lex_state = 18, .external_lex_state = 3},
  [1402] = {.lex_state = 18, .external_lex_state = 3},
  [1403] = {.lex_state = 18, .external_lex_state = 3},
  [1404] = {.lex_state = 18, .external_lex_state = 3},
  [1405] = {.lex_state = 18, .external_lex_state = 3},
  [1406] = {.lex_state = 7, .external_lex_state = 3},
  [1407] = {.lex_state = 18, .external_lex_state = 3},
  [1408] = {.lex_state = 7, .external_lex_state = 3},
  [1409] = {.lex_state = 7, .external_lex_state = 3},
  [1410] = {.lex_state = 7, .external_lex_state = 3},
  [1411] = {.lex_state = 7, .external_lex_state = 3},
  [1412] = {.lex_state = 7, .external_lex_state = 3},
  [1413] = {.lex_state = 7, .external_lex_state = 3},
  [1414] = {.lex_state = 18, .external_lex_state = 3},
  [1415] = {.lex_state = 18, .external_lex_state = 3},
  [1416] = {.lex_state = 7, .external_lex_state = 3},
  [1417] = {.lex_state = 7, .external_lex_state = 3},
  [1418] = {.lex_state = 18, .external_lex_state = 3},
  [1419] = {.lex_state = 9, .external_lex_state = 3},
  [1420] = {.lex_state = 18, .external_lex_state = 3},
  [1421] = {.lex_state = 18, .external_lex_state = 3},
  [1422] = {.lex_state = 18, .external_lex_state = 3},
  [1423] = {.lex_state = 9, .external_lex_state = 3},
  [1424] = {.lex_state = 9, .external_lex_state = 3},
  [1425] = {.lex_state = 18, .external_lex_state = 3},
  [1426] = {.lex_state = 18, .external_lex_state = 3},
  [1427] = {.lex_state = 18, .external_lex_state = 3},
  [1428] = {.lex_state = 18, .external_lex_state = 3},
  [1429] = {.lex_state = 18, .external_lex_state = 3},
  [1430] = {.lex_state = 9, .external_lex_state = 3},
  [1431] = {.lex_state = 9, .external_lex_state = 3},
  [1432] = {.lex_state = 18, .external_lex_state = 3},
  [1433] = {.lex_state = 9, .external_lex_state = 3},
  [1434] = {.lex_state = 18, .external_lex_state = 3},
  [1435] = {.lex_state = 9, .external_lex_state = 3},
  [1436] = {.lex_state = 18, .external_lex_state = 3},
  [1437] = {.lex_state = 18, .external_lex_state = 3},
  [1438] = {.lex_state = 18, .external_lex_state = 3},
  [1439] = {.lex_state = 7, .external_lex_state = 3},
  [1440] = {.lex_state = 18, .external_lex_state = 3},
  [1441] = {.lex_state = 18, .external_lex_state = 3},
  [1442] = {.lex_state = 18, .external_lex_state = 3},
  [1443] = {.lex_state = 18, .external_lex_state = 3},
  [1444] = {.lex_state = 18, .external_lex_state = 3},
  [1445] = {.lex_state = 9, .external_lex_state = 3},
  [1446] = {.lex_state = 18, .external_lex_state = 3},
  [1447] = {.lex_state = 18, .external_lex_state = 3},
  [1448] = {.lex_state = 18, .external_lex_state = 3},
  [1449] = {.lex_state = 9, .external_lex_state = 3},
  [1450] = {.lex_state = 18, .external_lex_state = 3},
  [1451] = {.lex_state = 18, .external_lex_state = 3},
  [1452] = {.lex_state = 18, .external_lex_state = 3},
  [1453] = {.lex_state = 18, .external_lex_state = 3},
  [1454] = {.lex_state = 18, .external_lex_state = 3},
  [1455] = {.lex_state = 18, .external_lex_state = 3},
  [1456] = {.lex_state = 18, .external_lex_state = 3},
  [1457] = {.lex_state = 9, .external_lex_state = 3},
  [1458] = {.lex_state = 18, .external_lex_state = 3},
  [1459] = {.lex_state = 9, .external_lex_state = 3},
  [1460] = {.lex_state = 18, .external_lex_state = 3},
  [1461] = {.lex_state = 9, .external_lex_state = 3},
  [1462] = {.lex_state = 9, .external_lex_state = 3},
  [1463] = {.lex_state = 9, .external_lex_state = 3},
  [1464] = {.lex_state = 14, .external_lex_state = 3},
  [1465] = {.lex_state = 9, .external_lex_state = 3},
  [1466] = {.lex_state = 9, .external_lex_state = 3},
  [1467] = {.lex_state = 14, .external_lex_state = 3},
  [1468] = {.lex_state = 9, .external_lex_state = 3},
  [1469] = {.lex_state = 9, .external_lex_state = 3},
  [1470] = {.lex_state = 9, .external_lex_state = 3},
  [1471] = {.lex_state = 14, .external_lex_state = 3},
  [1472] = {.lex_state = 14, .external_lex_state = 3},
  [1473] = {.lex_state = 9, .external_lex_state = 3},
  [1474] = {.lex_state = 18, .external_lex_state = 3},
  [1475] = {.lex_state = 15, .external_lex_state = 3},
  [1476] = {.lex_state = 7, .external_lex_state = 3},
  [1477] = {.lex_state = 9, .external_lex_state = 3},
  [1478] = {.lex_state = 7, .external_lex_state = 3},
  [1479] = {.lex_state = 7, .external_lex_state = 3},
  [1480] = {.lex_state = 7, .external_lex_state = 3},
  [1481] = {.lex_state = 7, .external_lex_state = 3},
  [1482] = {.lex_state = 18, .external_lex_state = 3},
  [1483] = {.lex_state = 18, .external_lex_state = 3},
  [1484] = {.lex_state = 7, .external_lex_state = 3},
  [1485] = {.lex_state = 7, .external_lex_state = 3},
  [1486] = {.lex_state = 7, .external_lex_state = 3},
  [1487] = {.lex_state = 7, .external_lex_state = 3},
  [1488] = {.lex_state = 7, .external_lex_state = 3},
  [1489] = {.lex_state = 7, .external_lex_state = 3},
  [1490] = {.lex_state = 7, .external_lex_state = 3},
  [1491] = {.lex_state = 18, .external_lex_state = 3},
  [1492] = {.lex_state = 7, .external_lex_state = 3},
  [1493] = {.lex_state = 18, .external_lex_state = 3},
  [1494] = {.lex_state = 7, .external_lex_state = 3},
  [1495] = {.lex_state = 7, .external_lex_state = 3},
  [1496] = {.lex_state = 7, .external_lex_state = 3},
  [1497] = {.lex_state = 18, .external_lex_state = 3},
  [1498] = {.lex_state = 7, .external_lex_state = 3},
  [1499] = {.lex_state = 18, .external_lex_state = 3},
  [1500] = {.lex_state = 7, .external_lex_state = 3},
  [1501] = {.lex_state = 7, .external_lex_state = 3},
  [1502] = {.lex_state = 7, .external_lex_state = 3},
  [1503] = {.lex_state = 7, .external_lex_state = 3},
  [1504] = {.lex_state = 15, .external_lex_state = 3},
  [1505] = {.lex_state = 4, .external_lex_state = 3},
  [1506] = {.lex_state = 18, .external_lex_state = 3},
  [1507] = {.lex_state = 18, .external_lex_state = 3},
  [1508] = {.lex_state = 7, .external_lex_state = 3},
  [1509] = {.lex_state = 7, .external_lex_state = 3},
  [1510] = {.lex_state = 7, .external_lex_state = 3},
  [1511] = {.lex_state = 7, .external_lex_state = 3},
  [1512] = {.lex_state = 7, .external_lex_state = 3},
  [1513] = {.lex_state = 7, .external_lex_state = 3},
  [1514] = {.lex_state = 7, .external_lex_state = 3},
  [1515] = {.lex_state = 7, .external_lex_state = 3},
  [1516] = {.lex_state = 7, .external_lex_state = 3},
  [1517] = {.lex_state = 7, .external_lex_state = 3},
  [1518] = {.lex_state = 7, .external_lex_state = 3},
  [1519] = {.lex_state = 7, .external_lex_state = 3},
  [1520] = {.lex_state = 7, .external_lex_state = 3},
  [1521] = {.lex_state = 7, .external_lex_state = 3},
  [1522] = {.lex_state = 15, .external_lex_state = 3},
  [1523] = {.lex_state = 7, .external_lex_state = 3},
  [1524] = {.lex_state = 7, .external_lex_state = 3},
  [1525] = {.lex_state = 7, .external_lex_state = 3},
  [1526] = {.lex_state = 7, .external_lex_state = 3},
  [1527] = {.lex_state = 7, .external_lex_state = 3},
  [1528] = {.lex_state = 4, .external_lex_state = 3},
  [1529] = {.lex_state = 15, .external_lex_state = 3},
  [1530] = {.lex_state = 7, .external_lex_state = 3},
  [1531] = {.lex_state = 15, .external_lex_state = 3},
  [1532] = {.lex_state = 15, .external_lex_state = 3},
  [1533] = {.lex_state = 15, .external_lex_state = 3},
  [1534] = {.lex_state = 15, .external_lex_state = 3},
  [1535] = {.lex_state = 7, .external_lex_state = 3},
  [1536] = {.lex_state = 15, .external_lex_state = 3},
  [1537] = {.lex_state = 7, .external_lex_state = 3},
  [1538] = {.lex_state = 7, .external_lex_state = 3},
  [1539] = {.lex_state = 7, .external_lex_state = 3},
  [1540] = {.lex_state = 7, .external_lex_state = 3},
  [1541] = {.lex_state = 7, .external_lex_state = 3},
  [1542] = {.lex_state = 7, .external_lex_state = 3},
  [1543] = {.lex_state = 7, .external_lex_state = 3},
  [1544] = {.lex_state = 0, .external_lex_state = 3},
  [1545] = {.lex_state = 4, .external_lex_state = 3},
  [1546] = {.lex_state = 4, .external_lex_state = 3},
  [1547] = {.lex_state = 7, .external_lex_state = 3},
  [1548] = {.lex_state = 15, .external_lex_state = 3},
  [1549] = {.lex_state = 7, .external_lex_state = 3},
  [1550] = {.lex_state = 7, .external_lex_state = 3},
  [1551] = {.lex_state = 7, .external_lex_state = 3},
  [1552] = {.lex_state = 4, .external_lex_state = 3},
  [1553] = {.lex_state = 7, .external_lex_state = 3},
  [1554] = {.lex_state = 7, .external_lex_state = 3},
  [1555] = {.lex_state = 15, .external_lex_state = 3},
  [1556] = {.lex_state = 7, .external_lex_state = 3},
  [1557] = {.lex_state = 4, .external_lex_state = 3},
  [1558] = {.lex_state = 7, .external_lex_state = 3},
  [1559] = {.lex_state = 7, .external_lex_state = 3},
  [1560] = {.lex_state = 7, .external_lex_state = 3},
  [1561] = {.lex_state = 14, .external_lex_state = 3},
  [1562] = {.lex_state = 7, .external_lex_state = 3},
  [1563] = {.lex_state = 15, .external_lex_state = 3},
  [1564] = {.lex_state = 7, .external_lex_state = 3},
  [1565] = {.lex_state = 4, .external_lex_state = 3},
  [1566] = {.lex_state = 0, .external_lex_state = 3},
  [1567] = {.lex_state = 7, .external_lex_state = 3},
  [1568] = {.lex_state = 7, .external_lex_state = 3},
  [1569] = {.lex_state = 7, .external_lex_state = 3},
  [1570] = {.lex_state = 7, .external_lex_state = 3},
  [1571] = {.lex_state = 7, .external_lex_state = 3},
  [1572] = {.lex_state = 7, .external_lex_state = 3},
  [1573] = {.lex_state = 7, .external_lex_state = 3},
  [1574] = {.lex_state = 15, .external_lex_state = 3},
  [1575] = {.lex_state = 7, .external_lex_state = 3},
  [1576] = {.lex_state = 0, .external_lex_state = 3},
  [1577] = {.lex_state = 7, .external_lex_state = 3},
  [1578] = {.lex_state = 0, .external_lex_state = 3},
  [1579] = {.lex_state = 0, .external_lex_state = 3},
  [1580] = {.lex_state = 0, .external_lex_state = 3},
  [1581] = {.lex_state = 0, .external_lex_state = 3},
  [1582] = {.lex_state = 0, .external_lex_state = 3},
  [1583] = {.lex_state = 7, .external_lex_state = 3},
  [1584] = {.lex_state = 0, .external_lex_state = 3},
  [1585] = {.lex_state = 7, .external_lex_state = 3},
  [1586] = {.lex_state = 7, .external_lex_state = 3},
  [1587] = {.lex_state = 0, .external_lex_state = 3},
  [1588] = {.lex_state = 7, .external_lex_state = 3},
  [1589] = {.lex_state = 15, .external_lex_state = 3},
  [1590] = {.lex_state = 0, .external_lex_state = 3},
  [1591] = {.lex_state = 7, .external_lex_state = 3},
  [1592] = {.lex_state = 0, .external_lex_state = 3},
  [1593] = {.lex_state = 7, .external_lex_state = 3},
  [1594] = {.lex_state = 0, .external_lex_state = 3},
  [1595] = {.lex_state = 7, .external_lex_state = 3},
  [1596] = {.lex_state = 4, .external_lex_state = 3},
  [1597] = {.lex_state = 0, .external_lex_state = 3},
  [1598] = {.lex_state = 0, .external_lex_state = 3},
  [1599] = {.lex_state = 0, .external_lex_state = 3},
  [1600] = {.lex_state = 7, .external_lex_state = 3},
  [1601] = {.lex_state = 0, .external_lex_state = 3},
  [1602] = {.lex_state = 0, .external_lex_state = 3},
  [1603] = {.lex_state = 0, .external_lex_state = 3},
  [1604] = {.lex_state = 15, .external_lex_state = 3},
  [1605] = {.lex_state = 0, .external_lex_state = 3},
  [1606] = {.lex_state = 0, .external_lex_state = 3},
  [1607] = {.lex_state = 15, .external_lex_state = 3},
  [1608] = {.lex_state = 0, .external_lex_state = 3},
  [1609] = {.lex_state = 7, .external_lex_state = 3},
  [1610] = {.lex_state = 0, .external_lex_state = 3},
  [1611] = {.lex_state = 7, .external_lex_state = 3},
  [1612] = {.lex_state = 7, .external_lex_state = 3},
  [1613] = {.lex_state = 7, .external_lex_state = 3},
  [1614] = {.lex_state = 7, .external_lex_state = 3},
  [1615] = {.lex_state = 18, .external_lex_state = 3},
  [1616] = {.lex_state = 7, .external_lex_state = 3},
  [1617] = {.lex_state = 7, .external_lex_state = 3},
  [1618] = {.lex_state = 10, .external_lex_state = 3},
  [1619] = {.lex_state = 7, .external_lex_state = 3},
  [1620] = {.lex_state = 10, .external_lex_state = 3},
  [1621] = {.lex_state = 7, .external_lex_state = 3},
  [1622] = {.lex_state = 7, .external_lex_state = 3},
  [1623] = {.lex_state = 7, .external_lex_state = 3},
  [1624] = {.lex_state = 7, .external_lex_state = 3},
  [1625] = {.lex_state = 7, .external_lex_state = 3},
  [1626] = {.lex_state = 7, .external_lex_state = 3},
  [1627] = {.lex_state = 7, .external_lex_state = 3},
  [1628] = {.lex_state = 58, .external_lex_state = 3},
  [1629] = {.lex_state = 18, .external_lex_state = 3},
  [1630] = {.lex_state = 18, .external_lex_state = 3},
  [1631] = {.lex_state = 7, .external_lex_state = 3},
  [1632] = {.lex_state = 7, .external_lex_state = 3},
  [1633] = {.lex_state = 7, .external_lex_state = 3},
  [1634] = {.lex_state = 7, .external_lex_state = 3},
  [1635] = {.lex_state = 14, .external_lex_state = 3},
  [1636] = {.lex_state = 7, .external_lex_state = 3},
  [1637] = {.lex_state = 7, .external_lex_state = 3},
  [1638] = {.lex_state = 7, .external_lex_state = 3},
  [1639] = {.lex_state = 7, .external_lex_state = 3},
  [1640] = {.lex_state = 7, .external_lex_state = 3},
  [1641] = {.lex_state = 14, .external_lex_state = 3},
  [1642] = {.lex_state = 7, .external_lex_state = 3},
  [1643] = {.lex_state = 7, .external_lex_state = 3},
  [1644] = {.lex_state = 10, .external_lex_state = 3},
  [1645] = {.lex_state = 7, .external_lex_state = 3},
  [1646] = {.lex_state = 7, .external_lex_state = 3},
  [1647] = {.lex_state = 7, .external_lex_state = 3},
  [1648] = {.lex_state = 7, .external_lex_state = 3},
  [1649] = {.lex_state = 7, .external_lex_state = 3},
  [1650] = {.lex_state = 7, .external_lex_state = 3},
  [1651] = {.lex_state = 7, .external_lex_state = 3},
  [1652] = {.lex_state = 7, .external_lex_state = 3},
  [1653] = {.lex_state = 4, .external_lex_state = 3},
  [1654] = {.lex_state = 18, .external_lex_state = 3},
  [1655] = {.lex_state = 7, .external_lex_state = 3},
  [1656] = {.lex_state = 7, .external_lex_state = 3},
  [1657] = {.lex_state = 7, .external_lex_state = 3},
  [1658] = {.lex_state = 0, .external_lex_state = 3},
  [1659] = {.lex_state = 7, .external_lex_state = 3},
  [1660] = {.lex_state = 7, .external_lex_state = 3},
  [1661] = {.lex_state = 10, .external_lex_state = 3},
  [1662] = {.lex_state = 7, .external_lex_state = 3},
  [1663] = {.lex_state = 7, .external_lex_state = 3},
  [1664] = {.lex_state = 7, .external_lex_state = 3},
  [1665] = {.lex_state = 0, .external_lex_state = 3},
  [1666] = {.lex_state = 18, .external_lex_state = 3},
  [1667] = {.lex_state = 7, .external_lex_state = 3},
  [1668] = {.lex_state = 7, .external_lex_state = 3},
  [1669] = {.lex_state = 7, .external_lex_state = 3},
  [1670] = {.lex_state = 7, .external_lex_state = 3},
  [1671] = {.lex_state = 7, .external_lex_state = 3},
  [1672] = {.lex_state = 7, .external_lex_state = 3},
  [1673] = {.lex_state = 4, .external_lex_state = 3},
  [1674] = {.lex_state = 7, .external_lex_state = 3},
  [1675] = {.lex_state = 7, .external_lex_state = 3},
  [1676] = {.lex_state = 7, .external_lex_state = 3},
  [1677] = {.lex_state = 18, .external_lex_state = 3},
  [1678] = {.lex_state = 7, .external_lex_state = 3},
  [1679] = {.lex_state = 7, .external_lex_state = 3},
  [1680] = {.lex_state = 10, .external_lex_state = 3},
  [1681] = {.lex_state = 58, .external_lex_state = 3},
  [1682] = {.lex_state = 4, .external_lex_state = 3},
  [1683] = {.lex_state = 7, .external_lex_state = 3},
  [1684] = {.lex_state = 7, .external_lex_state = 3},
  [1685] = {.lex_state = 7, .external_lex_state = 3},
  [1686] = {.lex_state = 7, .external_lex_state = 3},
  [1687] = {.lex_state = 18, .external_lex_state = 3},
  [1688] = {.lex_state = 7, .external_lex_state = 3},
  [1689] = {.lex_state = 7, .external_lex_state = 3},
  [1690] = {.lex_state = 18, .external_lex_state = 3},
  [1691] = {.lex_state = 18, .external_lex_state = 3},
  [1692] = {.lex_state = 7, .external_lex_state = 3},
  [1693] = {.lex_state = 7, .external_lex_state = 3},
  [1694] = {.lex_state = 7, .external_lex_state = 3},
  [1695] = {.lex_state = 15, .external_lex_state = 3},
  [1696] = {.lex_state = 4, .external_lex_state = 3},
  [1697] = {.lex_state = 10, .external_lex_state = 3},
  [1698] = {.lex_state = 7, .external_lex_state = 3},
  [1699] = {.lex_state = 0, .external_lex_state = 3},
  [1700] = {.lex_state = 18, .external_lex_state = 3},
  [1701] = {.lex_state = 7, .external_lex_state = 3},
  [1702] = {.lex_state = 7, .external_lex_state = 3},
  [1703] = {.lex_state = 7, .external_lex_state = 3},
  [1704] = {.lex_state = 7, .external_lex_state = 3},
  [1705] = {.lex_state = 7, .external_lex_state = 3},
  [1706] = {.lex_state = 7, .external_lex_state = 3},
  [1707] = {.lex_state = 18, .external_lex_state = 3},
  [1708] = {.lex_state = 18, .external_lex_state = 3},
  [1709] = {.lex_state = 10, .external_lex_state = 3},
  [1710] = {.lex_state = 18, .external_lex_state = 3},
  [1711] = {.lex_state = 18, .external_lex_state = 3},
  [1712] = {.lex_state = 0, .external_lex_state = 3},
  [1713] = {.lex_state = 7, .external_lex_state = 3},
  [1714] = {.lex_state = 7, .external_lex_state = 3},
  [1715] = {.lex_state = 18, .external_lex_state = 3},
  [1716] = {.lex_state = 18, .external_lex_state = 3},
  [1717] = {.lex_state = 18, .external_lex_state = 3},
  [1718] = {.lex_state = 0, .external_lex_state = 3},
  [1719] = {.lex_state = 7, .external_lex_state = 3},
  [1720] = {.lex_state = 0, .external_lex_state = 3},
  [1721] = {.lex_state = 7, .external_lex_state = 3},
  [1722] = {.lex_state = 7, .external_lex_state = 3},
  [1723] = {.lex_state = 18, .external_lex_state = 3},
  [1724] = {.lex_state = 7, .external_lex_state = 3},
  [1725] = {.lex_state = 10, .external_lex_state = 3},
  [1726] = {.lex_state = 10, .external_lex_state = 3},
  [1727] = {.lex_state = 7, .external_lex_state = 3},
  [1728] = {.lex_state = 18, .external_lex_state = 3},
  [1729] = {.lex_state = 7, .external_lex_state = 3},
  [1730] = {.lex_state = 58, .external_lex_state = 3},
  [1731] = {.lex_state = 18, .external_lex_state = 3},
  [1732] = {.lex_state = 7, .external_lex_state = 3},
  [1733] = {.lex_state = 7, .external_lex_state = 3},
  [1734] = {.lex_state = 7, .external_lex_state = 3},
  [1735] = {.lex_state = 18, .external_lex_state = 3},
  [1736] = {.lex_state = 18, .external_lex_state = 3},
  [1737] = {.lex_state = 7, .external_lex_state = 3},
  [1738] = {.lex_state = 18, .external_lex_state = 3},
  [1739] = {.lex_state = 7, .external_lex_state = 3},
  [1740] = {.lex_state = 0, .external_lex_state = 3},
  [1741] = {.lex_state = 10, .external_lex_state = 3},
  [1742] = {.lex_state = 10, .external_lex_state = 3},
  [1743] = {.lex_state = 0, .external_lex_state = 3},
  [1744] = {.lex_state = 0, .external_lex_state = 3},
  [1745] = {.lex_state = 7, .external_lex_state = 3},
  [1746] = {.lex_state = 18, .external_lex_state = 3},
  [1747] = {.lex_state = 58, .external_lex_state = 3},
  [1748] = {.lex_state = 18, .external_lex_state = 3},
  [1749] = {.lex_state = 7, .external_lex_state = 3},
  [1750] = {.lex_state = 0, .external_lex_state = 3},
  [1751] = {.lex_state = 18, .external_lex_state = 3},
  [1752] = {.lex_state = 0, .external_lex_state = 3},
  [1753] = {.lex_state = 10, .external_lex_state = 3},
  [1754] = {.lex_state = 58, .external_lex_state = 3},
  [1755] = {.lex_state = 18, .external_lex_state = 3},
  [1756] = {.lex_state = 7, .external_lex_state = 3},
  [1757] = {.lex_state = 0, .external_lex_state = 3},
  [1758] = {.lex_state = 58, .external_lex_state = 3},
  [1759] = {.lex_state = 58, .external_lex_state = 3},
  [1760] = {.lex_state = 0, .external_lex_state = 3},
  [1761] = {.lex_state = 7, .external_lex_state = 3},
  [1762] = {.lex_state = 4, .external_lex_state = 4},
  [1763] = {.lex_state = 19, .external_lex_state = 3},
  [1764] = {.lex_state = 7, .external_lex_state = 3},
  [1765] = {.lex_state = 7, .external_lex_state = 3},
  [1766] = {.lex_state = 58, .external_lex_state = 3},
  [1767] = {.lex_state = 58, .external_lex_state = 3},
  [1768] = {.lex_state = 7, .external_lex_state = 3},
  [1769] = {.lex_state = 0, .external_lex_state = 3},
  [1770] = {.lex_state = 18, .external_lex_state = 3},
  [1771] = {.lex_state = 7, .external_lex_state = 3},
  [1772] = {.lex_state = 7, .external_lex_state = 3},
  [1773] = {.lex_state = 18, .external_lex_state = 3},
  [1774] = {.lex_state = 4, .external_lex_state = 4},
  [1775] = {.lex_state = 58, .external_lex_state = 3},
  [1776] = {.lex_state = 7, .external_lex_state = 3},
  [1777] = {.lex_state = 58, .external_lex_state = 3},
  [1778] = {.lex_state = 7, .external_lex_state = 3},
  [1779] = {.lex_state = 0, .external_lex_state = 3},
  [1780] = {.lex_state = 7, .external_lex_state = 3},
  [1781] = {.lex_state = 58, .external_lex_state = 3},
  [1782] = {.lex_state = 7, .external_lex_state = 3},
  [1783] = {.lex_state = 7, .external_lex_state = 3},
  [1784] = {.lex_state = 0, .external_lex_state = 3},
  [1785] = {.lex_state = 58, .external_lex_state = 3},
  [1786] = {.lex_state = 58, .external_lex_state = 3},
  [1787] = {.lex_state = 58, .external_lex_state = 3},
  [1788] = {.lex_state = 18, .external_lex_state = 3},
  [1789] = {.lex_state = 58, .external_lex_state = 3},
  [1790] = {.lex_state = 0, .external_lex_state = 3},
  [1791] = {.lex_state = 58, .external_lex_state = 3},
  [1792] = {.lex_state = 0, .external_lex_state = 3},
  [1793] = {.lex_state = 4, .external_lex_state = 4},
  [1794] = {.lex_state = 0, .external_lex_state = 3},
  [1795] = {.lex_state = 58, .external_lex_state = 3},
  [1796] = {.lex_state = 7, .external_lex_state = 3},
  [1797] = {.lex_state = 7, .external_lex_state = 3},
  [1798] = {.lex_state = 18, .external_lex_state = 3},
  [1799] = {.lex_state = 58, .external_lex_state = 3},
  [1800] = {.lex_state = 7, .external_lex_state = 3},
  [1801] = {.lex_state = 4, .external_lex_state = 4},
  [1802] = {.lex_state = 58, .external_lex_state = 3},
  [1803] = {.lex_state = 0, .external_lex_state = 3},
  [1804] = {.lex_state = 58, .external_lex_state = 3},
  [1805] = {.lex_state = 4, .external_lex_state = 4},
  [1806] = {.lex_state = 0, .external_lex_state = 3},
  [1807] = {.lex_state = 7, .external_lex_state = 3},
  [1808] = {.lex_state = 0, .external_lex_state = 3},
  [1809] = {.lex_state = 7, .external_lex_state = 3},
  [1810] = {.lex_state = 58, .external_lex_state = 3},
  [1811] = {.lex_state = 7, .external_lex_state = 3},
  [1812] = {.lex_state = 7, .external_lex_state = 3},
  [1813] = {.lex_state = 10, .external_lex_state = 3},
  [1814] = {.lex_state = 7, .external_lex_state = 3},
  [1815] = {.lex_state = 7, .external_lex_state = 3},
  [1816] = {.lex_state = 7, .external_lex_state = 3},
  [1817] = {.lex_state = 7, .external_lex_state = 3},
  [1818] = {.lex_state = 0, .external_lex_state = 3},
  [1819] = {.lex_state = 7, .external_lex_state = 3},
  [1820] = {.lex_state = 18, .external_lex_state = 3},
  [1821] = {.lex_state = 58, .external_lex_state = 3},
  [1822] = {.lex_state = 58, .external_lex_state = 3},
  [1823] = {.lex_state = 58, .external_lex_state = 3},
  [1824] = {.lex_state = 19, .external_lex_state = 3},
  [1825] = {.lex_state = 0, .external_lex_state = 3},
  [1826] = {.lex_state = 7, .external_lex_state = 3},
  [1827] = {.lex_state = 58, .external_lex_state = 3},
  [1828] = {.lex_state = 7, .external_lex_state = 3},
  [1829] = {.lex_state = 4, .external_lex_state = 4},
  [1830] = {.lex_state = 7, .external_lex_state = 3},
  [1831] = {.lex_state = 58, .external_lex_state = 3},
  [1832] = {.lex_state = 18, .external_lex_state = 3},
  [1833] = {.lex_state = 58, .external_lex_state = 3},
  [1834] = {.lex_state = 0, .external_lex_state = 3},
  [1835] = {.lex_state = 58, .external_lex_state = 3},
  [1836] = {.lex_state = 19, .external_lex_state = 3},
  [1837] = {.lex_state = 58, .external_lex_state = 3},
  [1838] = {.lex_state = 7, .external_lex_state = 3},
  [1839] = {.lex_state = 4, .external_lex_state = 4},
  [1840] = {.lex_state = 7, .external_lex_state = 3},
  [1841] = {.lex_state = 7, .external_lex_state = 3},
  [1842] = {.lex_state = 18, .external_lex_state = 3},
  [1843] = {.lex_state = 0, .external_lex_state = 3},
  [1844] = {.lex_state = 7, .external_lex_state = 3},
  [1845] = {.lex_state = 0, .external_lex_state = 3},
  [1846] = {.lex_state = 58, .external_lex_state = 3},
  [1847] = {.lex_state = 19, .external_lex_state = 3},
  [1848] = {.lex_state = 0, .external_lex_state = 3},
  [1849] = {.lex_state = 4, .external_lex_state = 4},
  [1850] = {.lex_state = 10, .external_lex_state = 3},
  [1851] = {.lex_state = 7, .external_lex_state = 3},
  [1852] = {.lex_state = 58, .external_lex_state = 3},
  [1853] = {.lex_state = 4, .external_lex_state = 4},
  [1854] = {.lex_state = 7, .external_lex_state = 3},
  [1855] = {.lex_state = 58, .external_lex_state = 3},
  [1856] = {.lex_state = 7, .external_lex_state = 3},
  [1857] = {.lex_state = 3, .external_lex_state = 3},
  [1858] = {.lex_state = 0, .external_lex_state = 3},
  [1859] = {.lex_state = 58, .external_lex_state = 3},
  [1860] = {.lex_state = 58, .external_lex_state = 3},
  [1861] = {.lex_state = 0, .external_lex_state = 3},
  [1862] = {.lex_state = 3, .external_lex_state = 3},
  [1863] = {.lex_state = 3, .external_lex_state = 3},
  [1864] = {.lex_state = 58, .external_lex_state = 3},
  [1865] = {.lex_state = 0, .external_lex_state = 3},
  [1866] = {.lex_state = 0, .external_lex_state = 3},
  [1867] = {.lex_state = 58, .external_lex_state = 3},
  [1868] = {.lex_state = 0, .external_lex_state = 3},
  [1869] = {.lex_state = 58, .external_lex_state = 3},
  [1870] = {.lex_state = 58, .external_lex_state = 3},
  [1871] = {.lex_state = 0, .external_lex_state = 3},
  [1872] = {.lex_state = 58, .external_lex_state = 3},
  [1873] = {.lex_state = 0, .external_lex_state = 3},
  [1874] = {.lex_state = 58, .external_lex_state = 3},
  [1875] = {.lex_state = 58, .external_lex_state = 3},
  [1876] = {.lex_state = 0, .external_lex_state = 3},
  [1877] = {.lex_state = 0, .external_lex_state = 3},
  [1878] = {.lex_state = 0, .external_lex_state = 3},
  [1879] = {.lex_state = 58, .external_lex_state = 3},
  [1880] = {.lex_state = 0, .external_lex_state = 3},
  [1881] = {.lex_state = 58, .external_lex_state = 3},
  [1882] = {.lex_state = 0, .external_lex_state = 3},
  [1883] = {.lex_state = 58, .external_lex_state = 3},
  [1884] = {.lex_state = 7, .external_lex_state = 3},
  [1885] = {.lex_state = 3, .external_lex_state = 3},
  [1886] = {.lex_state = 0, .external_lex_state = 3},
  [1887] = {.lex_state = 0, .external_lex_state = 3},
  [1888] = {.lex_state = 3, .external_lex_state = 3},
  [1889] = {.lex_state = 58, .external_lex_state = 3},
  [1890] = {.lex_state = 58, .external_lex_state = 3},
  [1891] = {.lex_state = 0, .external_lex_state = 3},
  [1892] = {.lex_state = 0, .external_lex_state = 3},
  [1893] = {.lex_state = 58, .external_lex_state = 3},
  [1894] = {.lex_state = 3, .external_lex_state = 3},
  [1895] = {.lex_state = 7, .external_lex_state = 3},
  [1896] = {.lex_state = 3, .external_lex_state = 3},
  [1897] = {.lex_state = 0, .external_lex_state = 3},
  [1898] = {.lex_state = 7, .external_lex_state = 3},
  [1899] = {.lex_state = 0, .external_lex_state = 3},
  [1900] = {.lex_state = 0, .external_lex_state = 3},
  [1901] = {.lex_state = 0, .external_lex_state = 3},
  [1902] = {.lex_state = 7, .external_lex_state = 3},
  [1903] = {.lex_state = 0, .external_lex_state = 3},
  [1904] = {.lex_state = 0, .external_lex_state = 3},
  [1905] = {.lex_state = 0, .external_lex_state = 3},
  [1906] = {.lex_state = 3, .external_lex_state = 3},
  [1907] = {.lex_state = 0, .external_lex_state = 3},
  [1908] = {.lex_state = 0, .external_lex_state = 3},
  [1909] = {.lex_state = 3, .external_lex_state = 3},
  [1910] = {.lex_state = 0, .external_lex_state = 3},
  [1911] = {.lex_state = 0, .external_lex_state = 3},
  [1912] = {.lex_state = 0, .external_lex_state = 3},
  [1913] = {.lex_state = 0, .external_lex_state = 3},
  [1914] = {.lex_state = 10, .external_lex_state = 3},
  [1915] = {.lex_state = 58, .external_lex_state = 3},
  [1916] = {.lex_state = 0, .external_lex_state = 3},
  [1917] = {.lex_state = 58, .external_lex_state = 3},
  [1918] = {.lex_state = 0, .external_lex_state = 3},
  [1919] = {.lex_state = 58, .external_lex_state = 3},
  [1920] = {.lex_state = 58, .external_lex_state = 3},
  [1921] = {.lex_state = 7, .external_lex_state = 3},
  [1922] = {.lex_state = 10, .external_lex_state = 3},
  [1923] = {.lex_state = 58, .external_lex_state = 3},
  [1924] = {.lex_state = 0, .external_lex_state = 3},
  [1925] = {.lex_state = 7, .external_lex_state = 3},
  [1926] = {.lex_state = 0, .external_lex_state = 3},
  [1927] = {.lex_state = 0, .external_lex_state = 3},
  [1928] = {.lex_state = 58, .external_lex_state = 3},
  [1929] = {.lex_state = 58, .external_lex_state = 3},
  [1930] = {.lex_state = 4, .external_lex_state = 3},
  [1931] = {.lex_state = 0, .external_lex_state = 3},
  [1932] = {.lex_state = 58, .external_lex_state = 3},
  [1933] = {.lex_state = 0, .external_lex_state = 3},
  [1934] = {.lex_state = 58, .external_lex_state = 3},
  [1935] = {.lex_state = 58, .external_lex_state = 3},
  [1936] = {.lex_state = 58, .external_lex_state = 3},
  [1937] = {.lex_state = 0, .external_lex_state = 3},
  [1938] = {.lex_state = 0, .external_lex_state = 3},
  [1939] = {.lex_state = 58, .external_lex_state = 3},
  [1940] = {.lex_state = 10, .external_lex_state = 3},
  [1941] = {.lex_state = 0, .external_lex_state = 3},
  [1942] = {.lex_state = 0, .external_lex_state = 3},
  [1943] = {.lex_state = 0, .external_lex_state = 3},
  [1944] = {.lex_state = 0, .external_lex_state = 3},
  [1945] = {.lex_state = 0, .external_lex_state = 3},
  [1946] = {.lex_state = 0, .external_lex_state = 3},
  [1947] = {.lex_state = 58, .external_lex_state = 3},
  [1948] = {.lex_state = 58, .external_lex_state = 3},
  [1949] = {.lex_state = 0, .external_lex_state = 3},
  [1950] = {.lex_state = 58, .external_lex_state = 3},
  [1951] = {.lex_state = 0, .external_lex_state = 3},
  [1952] = {.lex_state = 0, .external_lex_state = 3},
  [1953] = {.lex_state = 0, .external_lex_state = 3},
  [1954] = {.lex_state = 0, .external_lex_state = 3},
  [1955] = {.lex_state = 7, .external_lex_state = 3},
  [1956] = {.lex_state = 0, .external_lex_state = 3},
  [1957] = {.lex_state = 0, .external_lex_state = 3},
  [1958] = {.lex_state = 0, .external_lex_state = 3},
  [1959] = {.lex_state = 0, .external_lex_state = 3},
  [1960] = {.lex_state = 0, .external_lex_state = 3},
  [1961] = {.lex_state = 0, .external_lex_state = 3},
  [1962] = {.lex_state = 0, .external_lex_state = 3},
  [1963] = {.lex_state = 0, .external_lex_state = 3},
  [1964] = {.lex_state = 0, .external_lex_state = 3},
  [1965] = {.lex_state = 0, .external_lex_state = 3},
  [1966] = {.lex_state = 58, .external_lex_state = 3},
  [1967] = {.lex_state = 0, .external_lex_state = 3},
  [1968] = {.lex_state = 58, .external_lex_state = 3},
  [1969] = {.lex_state = 58, .external_lex_state = 3},
  [1970] = {.lex_state = 58, .external_lex_state = 3},
  [1971] = {.lex_state = 0, .external_lex_state = 3},
  [1972] = {.lex_state = 0, .external_lex_state = 3},
  [1973] = {.lex_state = 0, .external_lex_state = 3},
  [1974] = {.lex_state = 0, .external_lex_state = 3},
  [1975] = {.lex_state = 7, .external_lex_state = 3},
  [1976] = {.lex_state = 0, .external_lex_state = 3},
  [1977] = {.lex_state = 0, .external_lex_state = 3},
  [1978] = {.lex_state = 0, .external_lex_state = 3},
  [1979] = {.lex_state = 0, .external_lex_state = 3},
  [1980] = {.lex_state = 58, .external_lex_state = 3},
  [1981] = {.lex_state = 7, .external_lex_state = 3},
  [1982] = {.lex_state = 0, .external_lex_state = 3},
  [1983] = {.lex_state = 0, .external_lex_state = 3},
  [1984] = {.lex_state = 0, .external_lex_state = 3},
  [1985] = {.lex_state = 58, .external_lex_state = 3},
  [1986] = {.lex_state = 4, .external_lex_state = 3},
  [1987] = {.lex_state = 7, .external_lex_state = 3},
  [1988] = {.lex_state = 0, .external_lex_state = 3},
  [1989] = {.lex_state = 0, .external_lex_state = 3},
  [1990] = {.lex_state = 0, .external_lex_state = 3},
  [1991] = {.lex_state = 0, .external_lex_state = 3},
  [1992] = {.lex_state = 58, .external_lex_state = 3},
  [1993] = {.lex_state = 0, .external_lex_state = 3},
  [1994] = {.lex_state = 3, .external_lex_state = 3},
  [1995] = {.lex_state = 0, .external_lex_state = 3},
  [1996] = {.lex_state = 0, .external_lex_state = 3},
  [1997] = {.lex_state = 0, .external_lex_state = 3},
  [1998] = {.lex_state = 7, .external_lex_state = 3},
  [1999] = {.lex_state = 58, .external_lex_state = 3},
  [2000] = {.lex_state = 3, .external_lex_state = 3},
  [2001] = {.lex_state = 0, .external_lex_state = 3},
  [2002] = {.lex_state = 0, .external_lex_state = 3},
  [2003] = {.lex_state = 7, .external_lex_state = 3},
  [2004] = {.lex_state = 0, .external_lex_state = 3},
  [2005] = {.lex_state = 0, .external_lex_state = 3},
  [2006] = {.lex_state = 58, .external_lex_state = 3},
  [2007] = {.lex_state = 0, .external_lex_state = 3},
  [2008] = {.lex_state = 0, .external_lex_state = 3},
  [2009] = {.lex_state = 58, .external_lex_state = 3},
  [2010] = {.lex_state = 7, .external_lex_state = 3},
  [2011] = {.lex_state = 0, .external_lex_state = 3},
  [2012] = {.lex_state = 0, .external_lex_state = 3},
  [2013] = {.lex_state = 0, .external_lex_state = 3},
  [2014] = {.lex_state = 3, .external_lex_state = 3},
  [2015] = {.lex_state = 7, .external_lex_state = 3},
  [2016] = {.lex_state = 3, .external_lex_state = 3},
  [2017] = {.lex_state = 58, .external_lex_state = 3},
  [2018] = {.lex_state = 3, .external_lex_state = 3},
  [2019] = {.lex_state = 0, .external_lex_state = 3},
  [2020] = {.lex_state = 0, .external_lex_state = 3},
  [2021] = {.lex_state = 58, .external_lex_state = 3},
  [2022] = {.lex_state = 0, .external_lex_state = 3},
  [2023] = {.lex_state = 58, .external_lex_state = 3},
  [2024] = {.lex_state = 0, .external_lex_state = 3},
  [2025] = {.lex_state = 58, .external_lex_state = 3},
  [2026] = {.lex_state = 0, .external_lex_state = 3},
  [2027] = {.lex_state = 3, .external_lex_state = 3},
  [2028] = {.lex_state = 0, .external_lex_state = 3},
  [2029] = {.lex_state = 0, .external_lex_state = 3},
  [2030] = {.lex_state = 0, .external_lex_state = 3},
  [2031] = {.lex_state = 58, .external_lex_state = 3},
  [2032] = {.lex_state = 3, .external_lex_state = 3},
  [2033] = {.lex_state = 0, .external_lex_state = 3},
  [2034] = {.lex_state = 0, .external_lex_state = 3},
  [2035] = {.lex_state = 0, .external_lex_state = 3},
  [2036] = {.lex_state = 7, .external_lex_state = 3},
  [2037] = {.lex_state = 0, .external_lex_state = 3},
  [2038] = {.lex_state = 58, .external_lex_state = 3},
  [2039] = {.lex_state = 0, .external_lex_state = 3},
  [2040] = {.lex_state = 58, .external_lex_state = 3},
  [2041] = {.lex_state = 58, .external_lex_state = 3},
  [2042] = {.lex_state = 0, .external_lex_state = 3},
  [2043] = {.lex_state = 0, .external_lex_state = 3},
  [2044] = {.lex_state = 0, .external_lex_state = 3},
  [2045] = {.lex_state = 58, .external_lex_state = 3},
  [2046] = {.lex_state = 58, .external_lex_state = 3},
  [2047] = {.lex_state = 58, .external_lex_state = 3},
  [2048] = {.lex_state = 7, .external_lex_state = 3},
  [2049] = {.lex_state = 0, .external_lex_state = 3},
  [2050] = {.lex_state = 0, .external_lex_state = 3},
  [2051] = {.lex_state = 58, .external_lex_state = 3},
  [2052] = {.lex_state = 0, .external_lex_state = 3},
  [2053] = {.lex_state = 0, .external_lex_state = 3},
  [2054] = {.lex_state = 0, .external_lex_state = 3},
  [2055] = {.lex_state = 0, .external_lex_state = 3},
  [2056] = {.lex_state = 0, .external_lex_state = 3},
  [2057] = {.lex_state = 10, .external_lex_state = 3},
  [2058] = {.lex_state = 58, .external_lex_state = 3},
  [2059] = {.lex_state = 0, .external_lex_state = 3},
  [2060] = {.lex_state = 0, .external_lex_state = 3},
  [2061] = {.lex_state = 0, .external_lex_state = 3},
  [2062] = {.lex_state = 0, .external_lex_state = 3},
  [2063] = {.lex_state = 0, .external_lex_state = 3},
  [2064] = {.lex_state = 0, .external_lex_state = 3},
  [2065] = {.lex_state = 3, .external_lex_state = 3},
  [2066] = {.lex_state = 0, .external_lex_state = 3},
  [2067] = {.lex_state = 0, .external_lex_state = 3},
  [2068] = {.lex_state = 58, .external_lex_state = 3},
  [2069] = {.lex_state = 0, .external_lex_state = 3},
  [2070] = {.lex_state = 0, .external_lex_state = 3},
  [2071] = {.lex_state = 0, .external_lex_state = 3},
  [2072] = {.lex_state = 7, .external_lex_state = 3},
  [2073] = {.lex_state = 0, .external_lex_state = 3},
  [2074] = {.lex_state = 0, .external_lex_state = 3},
  [2075] = {.lex_state = 0, .external_lex_state = 3},
  [2076] = {.lex_state = 3, .external_lex_state = 3},
  [2077] = {.lex_state = 0, .external_lex_state = 3},
  [2078] = {.lex_state = 0, .external_lex_state = 3},
  [2079] = {.lex_state = 0, .external_lex_state = 3},
  [2080] = {.lex_state = 0, .external_lex_state = 3},
  [2081] = {.lex_state = 3, .external_lex_state = 3},
  [2082] = {.lex_state = 0, .external_lex_state = 3},
  [2083] = {.lex_state = 0, .external_lex_state = 3},
  [2084] = {.lex_state = 0, .external_lex_state = 3},
  [2085] = {.lex_state = 7, .external_lex_state = 3},
  [2086] = {.lex_state = 58, .external_lex_state = 3},
  [2087] = {.lex_state = 7, .external_lex_state = 3},
  [2088] = {.lex_state = 18, .external_lex_state = 3},
  [2089] = {.lex_state = 7, .external_lex_state = 3},
  [2090] = {.lex_state = 0, .external_lex_state = 3},
  [2091] = {.lex_state = 58, .external_lex_state = 3},
  [2092] = {.lex_state = 58, .external_lex_state = 3},
  [2093] = {.lex_state = 58, .external_lex_state = 3},
  [2094] = {.lex_state = 58, .external_lex_state = 3},
  [2095] = {.lex_state = 0, .external_lex_state = 3},
  [2096] = {.lex_state = 0, .external_lex_state = 3},
  [2097] = {.lex_state = 0, .external_lex_state = 3},
  [2098] = {.lex_state = 7, .external_lex_state = 3},
  [2099] = {.lex_state = 10, .external_lex_state = 3},
  [2100] = {.lex_state = 0, .external_lex_state = 3},
  [2101] = {.lex_state = 7, .external_lex_state = 3},
  [2102] = {.lex_state = 3, .external_lex_state = 3},
  [2103] = {.lex_state = 7, .external_lex_state = 3},
  [2104] = {.lex_state = 58, .external_lex_state = 3},
  [2105] = {.lex_state = 7, .external_lex_state = 3},
  [2106] = {.lex_state = 0, .external_lex_state = 3},
  [2107] = {.lex_state = 0, .external_lex_state = 3},
  [2108] = {.lex_state = 0, .external_lex_state = 3},
  [2109] = {.lex_state = 0, .external_lex_state = 3},
  [2110] = {.lex_state = 0, .external_lex_state = 3},
  [2111] = {.lex_state = 0, .external_lex_state = 3},
  [2112] = {.lex_state = 0, .external_lex_state = 3},
  [2113] = {.lex_state = 0, .external_lex_state = 3},
  [2114] = {.lex_state = 0, .external_lex_state = 3},
  [2115] = {.lex_state = 0, .external_lex_state = 3},
  [2116] = {.lex_state = 0, .external_lex_state = 3},
  [2117] = {.lex_state = 58, .external_lex_state = 3},
  [2118] = {.lex_state = 0, .external_lex_state = 3},
  [2119] = {.lex_state = 0, .external_lex_state = 3},
  [2120] = {.lex_state = 0, .external_lex_state = 3},
  [2121] = {.lex_state = 0, .external_lex_state = 3},
  [2122] = {.lex_state = 0, .external_lex_state = 3},
  [2123] = {.lex_state = 7, .external_lex_state = 3},
  [2124] = {.lex_state = 0, .external_lex_state = 3},
  [2125] = {.lex_state = 0, .external_lex_state = 3},
  [2126] = {.lex_state = 0, .external_lex_state = 3},
  [2127] = {.lex_state = 0, .external_lex_state = 3},
  [2128] = {.lex_state = 0, .external_lex_state = 3},
  [2129] = {.lex_state = 7, .external_lex_state = 3},
  [2130] = {.lex_state = 0, .external_lex_state = 3},
  [2131] = {.lex_state = 0, .external_lex_state = 3},
  [2132] = {.lex_state = 0, .external_lex_state = 3},
  [2133] = {.lex_state = 7, .external_lex_state = 3},
  [2134] = {.lex_state = 0, .external_lex_state = 3},
  [2135] = {.lex_state = 7, .external_lex_state = 3},
  [2136] = {.lex_state = 0, .external_lex_state = 3},
  [2137] = {.lex_state = 0, .external_lex_state = 3},
  [2138] = {.lex_state = 58, .external_lex_state = 3},
  [2139] = {.lex_state = 58, .external_lex_state = 3},
  [2140] = {.lex_state = 0, .external_lex_state = 3},
  [2141] = {.lex_state = 0, .external_lex_state = 3},
  [2142] = {.lex_state = 58, .external_lex_state = 3},
  [2143] = {.lex_state = 0, .external_lex_state = 3},
  [2144] = {.lex_state = 0, .external_lex_state = 3},
  [2145] = {.lex_state = 0, .external_lex_state = 3},
  [2146] = {.lex_state = 0, .external_lex_state = 3},
  [2147] = {.lex_state = 0, .external_lex_state = 3},
  [2148] = {.lex_state = 0, .external_lex_state = 3},
  [2149] = {.lex_state = 0, .external_lex_state = 3},
  [2150] = {.lex_state = 7, .external_lex_state = 3},
  [2151] = {.lex_state = 0, .external_lex_state = 3},
  [2152] = {.lex_state = 0, .external_lex_state = 3},
  [2153] = {.lex_state = 0, .external_lex_state = 3},
  [2154] = {.lex_state = 0, .external_lex_state = 3},
  [2155] = {.lex_state = 0, .external_lex_state = 3},
  [2156] = {.lex_state = 0, .external_lex_state = 3},
  [2157] = {.lex_state = 58, .external_lex_state = 3},
  [2158] = {.lex_state = 0, .external_lex_state = 3},
  [2159] = {.lex_state = 0, .external_lex_state = 3},
  [2160] = {.lex_state = 0, .external_lex_state = 3},
  [2161] = {.lex_state = 0, .external_lex_state = 3},
  [2162] = {.lex_state = 7, .external_lex_state = 3},
  [2163] = {.lex_state = 0, .external_lex_state = 3},
  [2164] = {.lex_state = 58, .external_lex_state = 3},
  [2165] = {.lex_state = 0, .external_lex_state = 3},
  [2166] = {.lex_state = 0, .external_lex_state = 3},
  [2167] = {.lex_state = 0, .external_lex_state = 3},
  [2168] = {.lex_state = 0, .external_lex_state = 3},
  [2169] = {.lex_state = 0, .external_lex_state = 3},
  [2170] = {.lex_state = 0, .external_lex_state = 3},
  [2171] = {.lex_state = 7, .external_lex_state = 3},
  [2172] = {.lex_state = 0, .external_lex_state = 3},
  [2173] = {.lex_state = 0, .external_lex_state = 3},
  [2174] = {.lex_state = 0, .external_lex_state = 3},
  [2175] = {.lex_state = 0, .external_lex_state = 3},
  [2176] = {.lex_state = 0, .external_lex_state = 3},
  [2177] = {.lex_state = 58, .external_lex_state = 3},
  [2178] = {.lex_state = 0, .external_lex_state = 3},
  [2179] = {.lex_state = 0, .external_lex_state = 3},
  [2180] = {.lex_state = 58, .external_lex_state = 3},
  [2181] = {.lex_state = 0, .external_lex_state = 3},
  [2182] = {.lex_state = 0, .external_lex_state = 3},
  [2183] = {.lex_state = 0, .external_lex_state = 3},
  [2184] = {.lex_state = 0, .external_lex_state = 3},
  [2185] = {.lex_state = 0, .external_lex_state = 3},
  [2186] = {.lex_state = 0, .external_lex_state = 3},
  [2187] = {.lex_state = 58, .external_lex_state = 3},
  [2188] = {.lex_state = 0, .external_lex_state = 3},
  [2189] = {.lex_state = 0, .external_lex_state = 3},
  [2190] = {.lex_state = 0, .external_lex_state = 3},
  [2191] = {.lex_state = 0, .external_lex_state = 3},
  [2192] = {.lex_state = 58, .external_lex_state = 3},
  [2193] = {.lex_state = 0, .external_lex_state = 3},
  [2194] = {.lex_state = 0, .external_lex_state = 3},
  [2195] = {.lex_state = 58, .external_lex_state = 3},
  [2196] = {.lex_state = 0, .external_lex_state = 3},
  [2197] = {.lex_state = 7, .external_lex_state = 3},
  [2198] = {.lex_state = 58, .external_lex_state = 3},
  [2199] = {.lex_state = 0, .external_lex_state = 3},
  [2200] = {.lex_state = 0, .external_lex_state = 3},
  [2201] = {.lex_state = 4, .external_lex_state = 3},
  [2202] = {.lex_state = 0, .external_lex_state = 3},
  [2203] = {.lex_state = 58, .external_lex_state = 3},
  [2204] = {.lex_state = 0, .external_lex_state = 3},
  [2205] = {.lex_state = 0, .external_lex_state = 3},
  [2206] = {.lex_state = 0, .external_lex_state = 3},
  [2207] = {.lex_state = 0, .external_lex_state = 3},
  [2208] = {.lex_state = 0, .external_lex_state = 5},
  [2209] = {.lex_state = 0, .external_lex_state = 3},
  [2210] = {.lex_state = 0, .external_lex_state = 3},
  [2211] = {.lex_state = 0, .external_lex_state = 3},
  [2212] = {.lex_state = 0, .external_lex_state = 3},
  [2213] = {.lex_state = 0, .external_lex_state = 3},
  [2214] = {.lex_state = 7, .external_lex_state = 3},
  [2215] = {.lex_state = 0, .external_lex_state = 3},
  [2216] = {.lex_state = 58, .external_lex_state = 3},
  [2217] = {.lex_state = 0, .external_lex_state = 3},
  [2218] = {.lex_state = 0, .external_lex_state = 3},
  [2219] = {.lex_state = 0, .external_lex_state = 3},
  [2220] = {.lex_state = 0, .external_lex_state = 3},
  [2221] = {.lex_state = 0, .external_lex_state = 3},
  [2222] = {.lex_state = 58, .external_lex_state = 3},
  [2223] = {.lex_state = 0, .external_lex_state = 3},
  [2224] = {.lex_state = 58, .external_lex_state = 3},
  [2225] = {.lex_state = 0, .external_lex_state = 3},
  [2226] = {.lex_state = 0, .external_lex_state = 3},
  [2227] = {.lex_state = 0, .external_lex_state = 3},
  [2228] = {.lex_state = 0, .external_lex_state = 3},
  [2229] = {.lex_state = 7, .external_lex_state = 3},
  [2230] = {.lex_state = 58, .external_lex_state = 3},
  [2231] = {.lex_state = 0, .external_lex_state = 3},
  [2232] = {.lex_state = 0, .external_lex_state = 3},
  [2233] = {.lex_state = 0, .external_lex_state = 3},
  [2234] = {.lex_state = 0, .external_lex_state = 3},
  [2235] = {.lex_state = 0, .external_lex_state = 3},
  [2236] = {.lex_state = 0, .external_lex_state = 3},
  [2237] = {.lex_state = 0, .external_lex_state = 3},
  [2238] = {.lex_state = 7, .external_lex_state = 3},
  [2239] = {.lex_state = 7, .external_lex_state = 3},
  [2240] = {.lex_state = 0, .external_lex_state = 3},
  [2241] = {.lex_state = 0, .external_lex_state = 3},
  [2242] = {.lex_state = 0, .external_lex_state = 3},
  [2243] = {.lex_state = 0, .external_lex_state = 3},
  [2244] = {.lex_state = 58, .external_lex_state = 3},
  [2245] = {.lex_state = 0, .external_lex_state = 3},
  [2246] = {.lex_state = 7, .external_lex_state = 3},
  [2247] = {.lex_state = 0, .external_lex_state = 3},
  [2248] = {.lex_state = 0, .external_lex_state = 3},
  [2249] = {.lex_state = 0, .external_lex_state = 3},
  [2250] = {.lex_state = 0, .external_lex_state = 3},
  [2251] = {.lex_state = 0, .external_lex_state = 3},
  [2252] = {.lex_state = 18, .external_lex_state = 3},
  [2253] = {.lex_state = 58, .external_lex_state = 3},
  [2254] = {.lex_state = 7, .external_lex_state = 3},
  [2255] = {.lex_state = 0, .external_lex_state = 3},
  [2256] = {.lex_state = 0, .external_lex_state = 3},
  [2257] = {.lex_state = 0, .external_lex_state = 3},
  [2258] = {.lex_state = 0, .external_lex_state = 3},
  [2259] = {.lex_state = 58, .external_lex_state = 3},
  [2260] = {.lex_state = 0, .external_lex_state = 3},
  [2261] = {.lex_state = 58, .external_lex_state = 3},
  [2262] = {.lex_state = 0, .external_lex_state = 3},
  [2263] = {.lex_state = 0, .external_lex_state = 3},
  [2264] = {.lex_state = 58, .external_lex_state = 3},
  [2265] = {.lex_state = 58, .external_lex_state = 3},
  [2266] = {.lex_state = 0, .external_lex_state = 3},
  [2267] = {.lex_state = 0, .external_lex_state = 3},
  [2268] = {.lex_state = 7, .external_lex_state = 3},
  [2269] = {.lex_state = 0, .external_lex_state = 3},
  [2270] = {.lex_state = 7, .external_lex_state = 3},
  [2271] = {.lex_state = 7, .external_lex_state = 3},
  [2272] = {.lex_state = 0, .external_lex_state = 3},
  [2273] = {.lex_state = 7, .external_lex_state = 3},
  [2274] = {.lex_state = 0, .external_lex_state = 3},
  [2275] = {.lex_state = 0, .external_lex_state = 3},
  [2276] = {.lex_state = 0, .external_lex_state = 3},
  [2277] = {.lex_state = 7, .external_lex_state = 3},
  [2278] = {.lex_state = 0, .external_lex_state = 3},
  [2279] = {.lex_state = 58, .external_lex_state = 3},
  [2280] = {.lex_state = 0, .external_lex_state = 3},
  [2281] = {.lex_state = 18, .external_lex_state = 3},
  [2282] = {.lex_state = 18, .external_lex_state = 3},
  [2283] = {.lex_state = 0, .external_lex_state = 3},
  [2284] = {.lex_state = 0, .external_lex_state = 3},
  [2285] = {.lex_state = 7, .external_lex_state = 3},
  [2286] = {.lex_state = 0, .external_lex_state = 3},
  [2287] = {.lex_state = 0, .external_lex_state = 3},
  [2288] = {.lex_state = 58, .external_lex_state = 3},
  [2289] = {.lex_state = 18, .external_lex_state = 3},
  [2290] = {.lex_state = 58, .external_lex_state = 3},
  [2291] = {.lex_state = 18, .external_lex_state = 3},
  [2292] = {.lex_state = 0, .external_lex_state = 3},
  [2293] = {.lex_state = 0, .external_lex_state = 3},
  [2294] = {.lex_state = 0, .external_lex_state = 3},
  [2295] = {.lex_state = 4, .external_lex_state = 3},
  [2296] = {.lex_state = 10, .external_lex_state = 3},
  [2297] = {.lex_state = 58, .external_lex_state = 3},
  [2298] = {.lex_state = 7, .external_lex_state = 3},
  [2299] = {.lex_state = 0, .external_lex_state = 3},
  [2300] = {.lex_state = 0, .external_lex_state = 3},
  [2301] = {.lex_state = 7, .external_lex_state = 3},
  [2302] = {.lex_state = 0, .external_lex_state = 3},
  [2303] = {.lex_state = 7, .external_lex_state = 3},
  [2304] = {.lex_state = 0, .external_lex_state = 3},
  [2305] = {.lex_state = 18, .external_lex_state = 3},
  [2306] = {.lex_state = 0, .external_lex_state = 3},
  [2307] = {.lex_state = 0, .external_lex_state = 3},
  [2308] = {.lex_state = 0, .external_lex_state = 3},
  [2309] = {.lex_state = 0, .external_lex_state = 3},
  [2310] = {.lex_state = 0, .external_lex_state = 3},
  [2311] = {.lex_state = 0, .external_lex_state = 3},
  [2312] = {.lex_state = 7, .external_lex_state = 3},
  [2313] = {.lex_state = 58, .external_lex_state = 3},
  [2314] = {.lex_state = 0, .external_lex_state = 3},
  [2315] = {.lex_state = 0, .external_lex_state = 3},
  [2316] = {.lex_state = 0, .external_lex_state = 3},
  [2317] = {.lex_state = 7, .external_lex_state = 3},
  [2318] = {.lex_state = 0, .external_lex_state = 3},
  [2319] = {.lex_state = 7, .external_lex_state = 3},
  [2320] = {.lex_state = 7, .external_lex_state = 3},
  [2321] = {.lex_state = 0, .external_lex_state = 3},
  [2322] = {.lex_state = 0, .external_lex_state = 3},
  [2323] = {.lex_state = 0, .external_lex_state = 3},
  [2324] = {.lex_state = 7, .external_lex_state = 3},
  [2325] = {.lex_state = 0, .external_lex_state = 3},
  [2326] = {.lex_state = 0, .external_lex_state = 3},
  [2327] = {.lex_state = 7, .external_lex_state = 3},
  [2328] = {.lex_state = 7, .external_lex_state = 3},
  [2329] = {.lex_state = 7, .external_lex_state = 3},
  [2330] = {.lex_state = 0, .external_lex_state = 3},
  [2331] = {.lex_state = 7, .external_lex_state = 3},
  [2332] = {.lex_state = 0, .external_lex_state = 3},
  [2333] = {.lex_state = 0, .external_lex_state = 3},
  [2334] = {.lex_state = 7, .external_lex_state = 3},
  [2335] = {.lex_state = 7, .external_lex_state = 3},
  [2336] = {.lex_state = 7, .external_lex_state = 3},
  [2337] = {.lex_state = 7, .external_lex_state = 3},
  [2338] = {.lex_state = 0, .external_lex_state = 3},
  [2339] = {.lex_state = 0, .external_lex_state = 3},
  [2340] = {.lex_state = 7, .external_lex_state = 3},
  [2341] = {.lex_state = 10, .external_lex_state = 3},
  [2342] = {.lex_state = 0, .external_lex_state = 3},
  [2343] = {.lex_state = 0, .external_lex_state = 3},
  [2344] = {.lex_state = 7, .external_lex_state = 3},
  [2345] = {.lex_state = 0, .external_lex_state = 3},
  [2346] = {.lex_state = 7, .external_lex_state = 3},
  [2347] = {.lex_state = 7, .external_lex_state = 3},
  [2348] = {.lex_state = 7, .external_lex_state = 3},
  [2349] = {.lex_state = 7, .external_lex_state = 3},
  [2350] = {.lex_state = 7, .external_lex_state = 3},
  [2351] = {.lex_state = 7, .external_lex_state = 3},
  [2352] = {.lex_state = 7, .external_lex_state = 3},
  [2353] = {.lex_state = 7, .external_lex_state = 3},
  [2354] = {.lex_state = 7, .external_lex_state = 3},
  [2355] = {.lex_state = 0, .external_lex_state = 3},
  [2356] = {.lex_state = 0, .external_lex_state = 3},
  [2357] = {.lex_state = 0, .external_lex_state = 3},
  [2358] = {.lex_state = 0, .external_lex_state = 3},
  [2359] = {.lex_state = 7, .external_lex_state = 3},
  [2360] = {.lex_state = 7, .external_lex_state = 3},
  [2361] = {.lex_state = 7, .external_lex_state = 3},
  [2362] = {.lex_state = 7, .external_lex_state = 3},
  [2363] = {.lex_state = 0, .external_lex_state = 3},
  [2364] = {.lex_state = 7, .external_lex_state = 3},
  [2365] = {.lex_state = 7, .external_lex_state = 3},
  [2366] = {.lex_state = 7, .external_lex_state = 3},
  [2367] = {.lex_state = 0, .external_lex_state = 3},
  [2368] = {.lex_state = 7, .external_lex_state = 3},
  [2369] = {.lex_state = 0, .external_lex_state = 3},
  [2370] = {.lex_state = 0, .external_lex_state = 3},
  [2371] = {.lex_state = 7, .external_lex_state = 3},
  [2372] = {.lex_state = 0, .external_lex_state = 3},
  [2373] = {.lex_state = 0, .external_lex_state = 3},
  [2374] = {.lex_state = 0, .external_lex_state = 3},
  [2375] = {.lex_state = 0, .external_lex_state = 3},
  [2376] = {.lex_state = 0, .external_lex_state = 3},
  [2377] = {.lex_state = 7, .external_lex_state = 3},
  [2378] = {.lex_state = 0, .external_lex_state = 3},
  [2379] = {.lex_state = 0, .external_lex_state = 3},
  [2380] = {.lex_state = 0, .external_lex_state = 3},
  [2381] = {.lex_state = 10, .external_lex_state = 3},
  [2382] = {.lex_state = 0, .external_lex_state = 3},
  [2383] = {.lex_state = 0, .external_lex_state = 3},
  [2384] = {.lex_state = 0, .external_lex_state = 3},
  [2385] = {.lex_state = 0, .external_lex_state = 3},
  [2386] = {.lex_state = 7, .external_lex_state = 3},
  [2387] = {.lex_state = 0, .external_lex_state = 3},
  [2388] = {.lex_state = 10, .external_lex_state = 3},
  [2389] = {.lex_state = 0, .external_lex_state = 3},
  [2390] = {.lex_state = 7, .external_lex_state = 3},
  [2391] = {.lex_state = 0, .external_lex_state = 3},
  [2392] = {.lex_state = 0, .external_lex_state = 3},
  [2393] = {.lex_state = 7, .external_lex_state = 3},
  [2394] = {.lex_state = 7, .external_lex_state = 3},
  [2395] = {.lex_state = 7, .external_lex_state = 3},
  [2396] = {.lex_state = 7, .external_lex_state = 3},
  [2397] = {.lex_state = 7, .external_lex_state = 3},
  [2398] = {.lex_state = 7, .external_lex_state = 3},
  [2399] = {.lex_state = 7, .external_lex_state = 3},
  [2400] = {.lex_state = 10, .external_lex_state = 3},
  [2401] = {.lex_state = 0, .external_lex_state = 3},
  [2402] = {.lex_state = 0, .external_lex_state = 3},
  [2403] = {.lex_state = 0, .external_lex_state = 3},
  [2404] = {.lex_state = 0, .external_lex_state = 3},
  [2405] = {.lex_state = 0, .external_lex_state = 3},
  [2406] = {.lex_state = 10, .external_lex_state = 3},
  [2407] = {.lex_state = 0, .external_lex_state = 3},
  [2408] = {.lex_state = 0, .external_lex_state = 3},
  [2409] = {.lex_state = 0, .external_lex_state = 3},
  [2410] = {.lex_state = 7, .external_lex_state = 3},
  [2411] = {.lex_state = 0, .external_lex_state = 3},
  [2412] = {.lex_state = 0, .external_lex_state = 3},
  [2413] = {.lex_state = 7, .external_lex_state = 3},
  [2414] = {.lex_state = 7, .external_lex_state = 3},
  [2415] = {.lex_state = 7, .external_lex_state = 3},
  [2416] = {.lex_state = 7, .external_lex_state = 3},
  [2417] = {.lex_state = 0, .external_lex_state = 3},
  [2418] = {.lex_state = 0, .external_lex_state = 3},
  [2419] = {.lex_state = 7, .external_lex_state = 3},
  [2420] = {.lex_state = 0, .external_lex_state = 3},
  [2421] = {.lex_state = 0, .external_lex_state = 3},
  [2422] = {.lex_state = 0, .external_lex_state = 3},
  [2423] = {.lex_state = 0, .external_lex_state = 3},
  [2424] = {.lex_state = 7, .external_lex_state = 3},
  [2425] = {.lex_state = 7, .external_lex_state = 3},
  [2426] = {.lex_state = 0, .external_lex_state = 3},
  [2427] = {.lex_state = 0, .external_lex_state = 3},
  [2428] = {.lex_state = 7, .external_lex_state = 3},
  [2429] = {.lex_state = 0, .external_lex_state = 3},
  [2430] = {.lex_state = 7, .external_lex_state = 3},
  [2431] = {.lex_state = 0, .external_lex_state = 3},
  [2432] = {.lex_state = 0, .external_lex_state = 3},
  [2433] = {.lex_state = 0, .external_lex_state = 3},
  [2434] = {.lex_state = 7, .external_lex_state = 3},
  [2435] = {.lex_state = 7, .external_lex_state = 3},
  [2436] = {.lex_state = 7, .external_lex_state = 3},
  [2437] = {.lex_state = 0, .external_lex_state = 3},
  [2438] = {.lex_state = 7, .external_lex_state = 3},
  [2439] = {.lex_state = 7, .external_lex_state = 3},
  [2440] = {.lex_state = 0, .external_lex_state = 3},
  [2441] = {.lex_state = 7, .external_lex_state = 3},
  [2442] = {.lex_state = 0, .external_lex_state = 3},
  [2443] = {.lex_state = 7, .external_lex_state = 3},
  [2444] = {.lex_state = 0, .external_lex_state = 3},
  [2445] = {.lex_state = 7, .external_lex_state = 3},
  [2446] = {.lex_state = 7, .external_lex_state = 3},
  [2447] = {.lex_state = 10, .external_lex_state = 3},
  [2448] = {.lex_state = 58, .external_lex_state = 3},
  [2449] = {.lex_state = 7, .external_lex_state = 3},
  [2450] = {.lex_state = 0, .external_lex_state = 3},
  [2451] = {.lex_state = 0, .external_lex_state = 3},
  [2452] = {.lex_state = 7, .external_lex_state = 3},
  [2453] = {.lex_state = 0, .external_lex_state = 3},
  [2454] = {.lex_state = 0, .external_lex_state = 3},
  [2455] = {.lex_state = 0, .external_lex_state = 3},
  [2456] = {.lex_state = 0, .external_lex_state = 3},
  [2457] = {.lex_state = 0, .external_lex_state = 3},
  [2458] = {.lex_state = 0, .external_lex_state = 3},
  [2459] = {.lex_state = 0, .external_lex_state = 3},
  [2460] = {.lex_state = 0, .external_lex_state = 3},
  [2461] = {.lex_state = 0, .external_lex_state = 3},
  [2462] = {.lex_state = 0, .external_lex_state = 3},
  [2463] = {.lex_state = 7, .external_lex_state = 3},
  [2464] = {.lex_state = 10, .external_lex_state = 3},
  [2465] = {.lex_state = 7, .external_lex_state = 3},
  [2466] = {.lex_state = 0, .external_lex_state = 3},
  [2467] = {.lex_state = 0, .external_lex_state = 3},
  [2468] = {.lex_state = 7, .external_lex_state = 3},
  [2469] = {.lex_state = 7, .external_lex_state = 3},
  [2470] = {.lex_state = 0, .external_lex_state = 3},
  [2471] = {.lex_state = 58, .external_lex_state = 3},
  [2472] = {.lex_state = 7, .external_lex_state = 3},
  [2473] = {.lex_state = 0, .external_lex_state = 3},
  [2474] = {.lex_state = 7, .external_lex_state = 3},
  [2475] = {.lex_state = 0, .external_lex_state = 3},
  [2476] = {.lex_state = 58, .external_lex_state = 3},
  [2477] = {.lex_state = 0, .external_lex_state = 3},
  [2478] = {.lex_state = 7, .external_lex_state = 3},
  [2479] = {.lex_state = 0, .external_lex_state = 3},
  [2480] = {.lex_state = 0, .external_lex_state = 3},
  [2481] = {.lex_state = 0, .external_lex_state = 3},
  [2482] = {.lex_state = 0, .external_lex_state = 3},
  [2483] = {.lex_state = 10, .external_lex_state = 3},
  [2484] = {.lex_state = 7, .external_lex_state = 3},
  [2485] = {.lex_state = 7, .external_lex_state = 3},
  [2486] = {.lex_state = 7, .external_lex_state = 3},
  [2487] = {.lex_state = 0, .external_lex_state = 3},
  [2488] = {.lex_state = 0, .external_lex_state = 3},
  [2489] = {.lex_state = 0, .external_lex_state = 3},
  [2490] = {.lex_state = 7, .external_lex_state = 3},
  [2491] = {.lex_state = 0, .external_lex_state = 3},
  [2492] = {.lex_state = 0, .external_lex_state = 3},
  [2493] = {.lex_state = 0, .external_lex_state = 3},
  [2494] = {.lex_state = 0, .external_lex_state = 3},
  [2495] = {.lex_state = 7, .external_lex_state = 3},
  [2496] = {.lex_state = 7, .external_lex_state = 3},
  [2497] = {.lex_state = 10, .external_lex_state = 3},
  [2498] = {.lex_state = 7, .external_lex_state = 3},
  [2499] = {.lex_state = 7, .external_lex_state = 3},
  [2500] = {.lex_state = 0, .external_lex_state = 3},
  [2501] = {.lex_state = 0, .external_lex_state = 3},
  [2502] = {.lex_state = 7, .external_lex_state = 3},
  [2503] = {.lex_state = 0, .external_lex_state = 3},
  [2504] = {.lex_state = 0, .external_lex_state = 3},
  [2505] = {.lex_state = 0, .external_lex_state = 3},
  [2506] = {.lex_state = 0, .external_lex_state = 3},
  [2507] = {.lex_state = 0, .external_lex_state = 3},
  [2508] = {.lex_state = 0, .external_lex_state = 3},
  [2509] = {.lex_state = 0, .external_lex_state = 3},
  [2510] = {.lex_state = 7, .external_lex_state = 3},
  [2511] = {.lex_state = 0, .external_lex_state = 3},
  [2512] = {.lex_state = 0, .external_lex_state = 3},
  [2513] = {.lex_state = 58, .external_lex_state = 3},
  [2514] = {.lex_state = 0, .external_lex_state = 3},
  [2515] = {.lex_state = 7, .external_lex_state = 3},
  [2516] = {.lex_state = 0, .external_lex_state = 3},
  [2517] = {.lex_state = 10, .external_lex_state = 3},
  [2518] = {.lex_state = 0, .external_lex_state = 3},
  [2519] = {.lex_state = 10, .external_lex_state = 3},
  [2520] = {.lex_state = 7, .external_lex_state = 3},
  [2521] = {.lex_state = 10, .external_lex_state = 3},
  [2522] = {.lex_state = 0, .external_lex_state = 3},
  [2523] = {.lex_state = 0, .external_lex_state = 3},
  [2524] = {.lex_state = 10, .external_lex_state = 3},
  [2525] = {.lex_state = 0, .external_lex_state = 3},
  [2526] = {.lex_state = 7, .external_lex_state = 3},
  [2527] = {.lex_state = 0, .external_lex_state = 3},
  [2528] = {.lex_state = 7, .external_lex_state = 3},
  [2529] = {.lex_state = 0, .external_lex_state = 3},
  [2530] = {.lex_state = 0, .external_lex_state = 3},
  [2531] = {.lex_state = 7, .external_lex_state = 3},
  [2532] = {.lex_state = 10, .external_lex_state = 3},
  [2533] = {.lex_state = 7, .external_lex_state = 3},
  [2534] = {.lex_state = 7, .external_lex_state = 3},
  [2535] = {.lex_state = 7, .external_lex_state = 3},
  [2536] = {.lex_state = 10, .external_lex_state = 3},
  [2537] = {.lex_state = 0, .external_lex_state = 3},
  [2538] = {.lex_state = 0, .external_lex_state = 3},
  [2539] = {.lex_state = 10, .external_lex_state = 3},
  [2540] = {.lex_state = 7, .external_lex_state = 3},
  [2541] = {.lex_state = 7, .external_lex_state = 3},
  [2542] = {.lex_state = 0, .external_lex_state = 3},
  [2543] = {.lex_state = 0, .external_lex_state = 3},
  [2544] = {.lex_state = 0, .external_lex_state = 3},
  [2545] = {.lex_state = 10, .external_lex_state = 3},
  [2546] = {.lex_state = 10, .external_lex_state = 3},
  [2547] = {.lex_state = 0, .external_lex_state = 3},
  [2548] = {.lex_state = 0, .external_lex_state = 3},
  [2549] = {.lex_state = 0, .external_lex_state = 3},
  [2550] = {.lex_state = 7, .external_lex_state = 3},
  [2551] = {.lex_state = 0, .external_lex_state = 3},
  [2552] = {.lex_state = 7, .external_lex_state = 3},
  [2553] = {.lex_state = 0, .external_lex_state = 3},
  [2554] = {.lex_state = 7, .external_lex_state = 3},
  [2555] = {.lex_state = 0, .external_lex_state = 3},
  [2556] = {.lex_state = 0, .external_lex_state = 3},
  [2557] = {.lex_state = 7, .external_lex_state = 3},
  [2558] = {.lex_state = 7, .external_lex_state = 3},
  [2559] = {.lex_state = 7, .external_lex_state = 3},
  [2560] = {.lex_state = 0, .external_lex_state = 3},
  [2561] = {.lex_state = 0, .external_lex_state = 3},
  [2562] = {.lex_state = 7, .external_lex_state = 3},
  [2563] = {.lex_state = 7, .external_lex_state = 3},
  [2564] = {.lex_state = 7, .external_lex_state = 3},
};

enum {
  ts_external_token__string_content = 0,
  ts_external_token_raw_string_literal = 1,
  ts_external_token_float_literal = 2,
  ts_external_token_block_comment = 3,
};

static const TSSymbol ts_external_scanner_symbol_map[EXTERNAL_TOKEN_COUNT] = {
  [ts_external_token__string_content] = sym__string_content,
  [ts_external_token_raw_string_literal] = sym_raw_string_literal,
  [ts_external_token_float_literal] = sym_float_literal,
  [ts_external_token_block_comment] = sym_block_comment,
};

static const bool ts_external_scanner_states[6][EXTERNAL_TOKEN_COUNT] = {
  [1] = {
    [ts_external_token__string_content] = true,
    [ts_external_token_raw_string_literal] = true,
    [ts_external_token_float_literal] = true,
    [ts_external_token_block_comment] = true,
  },
  [2] = {
    [ts_external_token_raw_string_literal] = true,
    [ts_external_token_float_literal] = true,
    [ts_external_token_block_comment] = true,
  },
  [3] = {
    [ts_external_token_block_comment] = true,
  },
  [4] = {
    [ts_external_token__string_content] = true,
    [ts_external_token_block_comment] = true,
  },
  [5] = {
    [ts_external_token_float_literal] = true,
    [ts_external_token_block_comment] = true,
  },
};

static const uint16_t ts_parse_table[LARGE_STATE_COUNT][SYMBOL_COUNT] = {
  [0] = {
    [ts_builtin_sym_end] = ACTIONS(1),
    [sym_identifier] = ACTIONS(1),
    [anon_sym_SEMI] = ACTIONS(1),
    [anon_sym_macro_rules_BANG] = ACTIONS(1),
    [anon_sym_LPAREN] = ACTIONS(1),
    [anon_sym_RPAREN] = ACTIONS(1),
    [anon_sym_LBRACE] = ACTIONS(1),
    [anon_sym_RBRACE] = ACTIONS(1),
    [anon_sym_EQ_GT] = ACTIONS(1),
    [anon_sym_LBRACK] = ACTIONS(1),
    [anon_sym_RBRACK] = ACTIONS(1),
    [anon_sym_COLON] = ACTIONS(1),
    [anon_sym_DOLLAR] = ACTIONS(1),
    [anon_sym_PLUS] = ACTIONS(1),
    [anon_sym_STAR] = ACTIONS(1),
    [anon_sym_QMARK] = ACTIONS(1),
    [anon_sym_block] = ACTIONS(1),
    [anon_sym_expr] = ACTIONS(1),
    [anon_sym_ident] = ACTIONS(1),
    [anon_sym_item] = ACTIONS(1),
    [anon_sym_lifetime] = ACTIONS(1),
    [anon_sym_literal] = ACTIONS(1),
    [anon_sym_meta] = ACTIONS(1),
    [anon_sym_pat] = ACTIONS(1),
    [anon_sym_path] = ACTIONS(1),
    [anon_sym_stmt] = ACTIONS(1),
    [anon_sym_tt] = ACTIONS(1),
    [anon_sym_ty] = ACTIONS(1),
    [anon_sym_vis] = ACTIONS(1),
    [anon_sym_u8] = ACTIONS(1),
    [anon_sym_i8] = ACTIONS(1),
    [anon_sym_u16] = ACTIONS(1),
    [anon_sym_i16] = ACTIONS(1),
    [anon_sym_u32] = ACTIONS(1),
    [anon_sym_i32] = ACTIONS(1),
    [anon_sym_u64] = ACTIONS(1),
    [anon_sym_i64] = ACTIONS(1),
    [anon_sym_u128] = ACTIONS(1),
    [anon_sym_i128] = ACTIONS(1),
    [anon_sym_isize] = ACTIONS(1),
    [anon_sym_usize] = ACTIONS(1),
    [anon_sym_f32] = ACTIONS(1),
    [anon_sym_f64] = ACTIONS(1),
    [anon_sym_bool] = ACTIONS(1),
    [anon_sym_str] = ACTIONS(1),
    [anon_sym_char] = ACTIONS(1),
    [anon_sym_SQUOTE] = ACTIONS(1),
    [anon_sym_as] = ACTIONS(1),
    [anon_sym_async] = ACTIONS(1),
    [anon_sym_await] = ACTIONS(1),
    [anon_sym_break] = ACTIONS(1),
    [anon_sym_const] = ACTIONS(1),
    [anon_sym_continue] = ACTIONS(1),
    [anon_sym_default] = ACTIONS(1),
    [anon_sym_enum] = ACTIONS(1),
    [anon_sym_fn] = ACTIONS(1),
    [anon_sym_for] = ACTIONS(1),
    [anon_sym_if] = ACTIONS(1),
    [anon_sym_impl] = ACTIONS(1),
    [anon_sym_let] = ACTIONS(1),
    [anon_sym_loop] = ACTIONS(1),
    [anon_sym_match] = ACTIONS(1),
    [anon_sym_mod] = ACTIONS(1),
    [anon_sym_pub] = ACTIONS(1),
    [anon_sym_return] = ACTIONS(1),
    [anon_sym_static] = ACTIONS(1),
    [anon_sym_struct] = ACTIONS(1),
    [anon_sym_trait] = ACTIONS(1),
    [anon_sym_type] = ACTIONS(1),
    [anon_sym_union] = ACTIONS(1),
    [anon_sym_unsafe] = ACTIONS(1),
    [anon_sym_use] = ACTIONS(1),
    [anon_sym_where] = ACTIONS(1),
    [anon_sym_while] = ACTIONS(1),
    [anon_sym_POUND] = ACTIONS(1),
    [anon_sym_BANG] = ACTIONS(1),
    [anon_sym_EQ] = ACTIONS(1),
    [anon_sym_COMMA] = ACTIONS(1),
    [anon_sym_extern] = ACTIONS(1),
    [anon_sym_ref] = ACTIONS(1),
    [anon_sym_DASH_GT] = ACTIONS(1),
    [anon_sym_LT] = ACTIONS(1),
    [anon_sym_GT] = ACTIONS(1),
    [anon_sym_else] = ACTIONS(1),
    [anon_sym_COLON_COLON] = ACTIONS(1),
    [anon_sym__] = ACTIONS(1),
    [anon_sym_AMP] = ACTIONS(1),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(1),
    [anon_sym_in] = ACTIONS(1),
    [anon_sym_LT2] = ACTIONS(1),
    [anon_sym_dyn] = ACTIONS(1),
    [sym_mutable_specifier] = ACTIONS(1),
    [anon_sym_DOT_DOT] = ACTIONS(1),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(1),
    [anon_sym_DASH] = ACTIONS(1),
    [anon_sym_AMP_AMP] = ACTIONS(1),
    [anon_sym_PIPE_PIPE] = ACTIONS(1),
    [anon_sym_PIPE] = ACTIONS(1),
    [anon_sym_CARET] = ACTIONS(1),
    [anon_sym_EQ_EQ] = ACTIONS(1),
    [anon_sym_BANG_EQ] = ACTIONS(1),
    [anon_sym_LT_EQ] = ACTIONS(1),
    [anon_sym_GT_EQ] = ACTIONS(1),
    [anon_sym_LT_LT] = ACTIONS(1),
    [anon_sym_GT_GT] = ACTIONS(1),
    [anon_sym_SLASH] = ACTIONS(1),
    [anon_sym_PERCENT] = ACTIONS(1),
    [anon_sym_PLUS_EQ] = ACTIONS(1),
    [anon_sym_DASH_EQ] = ACTIONS(1),
    [anon_sym_STAR_EQ] = ACTIONS(1),
    [anon_sym_SLASH_EQ] = ACTIONS(1),
    [anon_sym_PERCENT_EQ] = ACTIONS(1),
    [anon_sym_AMP_EQ] = ACTIONS(1),
    [anon_sym_PIPE_EQ] = ACTIONS(1),
    [anon_sym_CARET_EQ] = ACTIONS(1),
    [anon_sym_LT_LT_EQ] = ACTIONS(1),
    [anon_sym_GT_GT_EQ] = ACTIONS(1),
    [anon_sym_yield] = ACTIONS(1),
    [anon_sym_move] = ACTIONS(1),
    [anon_sym_DOT] = ACTIONS(1),
    [anon_sym_AT] = ACTIONS(1),
    [sym_integer_literal] = ACTIONS(1),
    [aux_sym_string_literal_token1] = ACTIONS(1),
    [anon_sym_DQUOTE] = ACTIONS(1),
    [sym_char_literal] = ACTIONS(1),
    [sym_escape_sequence] = ACTIONS(1),
    [anon_sym_true] = ACTIONS(1),
    [anon_sym_false] = ACTIONS(1),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1),
    [sym_super] = ACTIONS(1),
    [sym_crate] = ACTIONS(1),
    [sym_metavariable] = ACTIONS(1),
    [sym__string_content] = ACTIONS(1),
    [sym_raw_string_literal] = ACTIONS(1),
    [sym_float_literal] = ACTIONS(1),
    [sym_block_comment] = ACTIONS(3),
  },
  [1] = {
    [sym_source_file] = STATE(2529),
    [sym__statement] = STATE(9),
    [sym_empty_statement] = STATE(9),
    [sym_expression_statement] = STATE(9),
    [sym_macro_definition] = STATE(9),
    [sym_attribute_item] = STATE(9),
    [sym_inner_attribute_item] = STATE(9),
    [sym_mod_item] = STATE(9),
    [sym_foreign_mod_item] = STATE(9),
    [sym_struct_item] = STATE(9),
    [sym_union_item] = STATE(9),
    [sym_enum_item] = STATE(9),
    [sym_extern_crate_declaration] = STATE(9),
    [sym_const_item] = STATE(9),
    [sym_static_item] = STATE(9),
    [sym_type_item] = STATE(9),
    [sym_function_item] = STATE(9),
    [sym_function_signature_item] = STATE(9),
    [sym_function_modifiers] = STATE(2528),
    [sym_impl_item] = STATE(9),
    [sym_trait_item] = STATE(9),
    [sym_associated_type] = STATE(9),
    [sym_let_declaration] = STATE(9),
    [sym_use_declaration] = STATE(9),
    [sym_extern_modifier] = STATE(1484),
    [sym_visibility_modifier] = STATE(1339),
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1266),
    [sym_macro_invocation] = STATE(68),
    [sym_scoped_identifier] = STATE(1172),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(70),
    [sym_match_expression] = STATE(70),
    [sym_while_expression] = STATE(70),
    [sym_loop_expression] = STATE(70),
    [sym_for_expression] = STATE(70),
    [sym_const_block] = STATE(70),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2521),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(70),
    [sym_async_block] = STATE(70),
    [sym_block] = STATE(70),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [aux_sym_source_file_repeat1] = STATE(9),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [ts_builtin_sym_end] = ACTIONS(5),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(9),
    [anon_sym_macro_rules_BANG] = ACTIONS(11),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(25),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(29),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(33),
    [anon_sym_enum] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(37),
    [anon_sym_for] = ACTIONS(39),
    [anon_sym_if] = ACTIONS(41),
    [anon_sym_impl] = ACTIONS(43),
    [anon_sym_let] = ACTIONS(45),
    [anon_sym_loop] = ACTIONS(47),
    [anon_sym_match] = ACTIONS(49),
    [anon_sym_mod] = ACTIONS(51),
    [anon_sym_pub] = ACTIONS(53),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_static] = ACTIONS(57),
    [anon_sym_struct] = ACTIONS(59),
    [anon_sym_trait] = ACTIONS(61),
    [anon_sym_type] = ACTIONS(63),
    [anon_sym_union] = ACTIONS(65),
    [anon_sym_unsafe] = ACTIONS(67),
    [anon_sym_use] = ACTIONS(69),
    [anon_sym_while] = ACTIONS(71),
    [anon_sym_POUND] = ACTIONS(73),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_extern] = ACTIONS(75),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(101),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [2] = {
    [sym__statement] = STATE(13),
    [sym_empty_statement] = STATE(13),
    [sym_expression_statement] = STATE(13),
    [sym_macro_definition] = STATE(13),
    [sym_attribute_item] = STATE(13),
    [sym_inner_attribute_item] = STATE(13),
    [sym_mod_item] = STATE(13),
    [sym_foreign_mod_item] = STATE(13),
    [sym_struct_item] = STATE(13),
    [sym_union_item] = STATE(13),
    [sym_enum_item] = STATE(13),
    [sym_extern_crate_declaration] = STATE(13),
    [sym_const_item] = STATE(13),
    [sym_static_item] = STATE(13),
    [sym_type_item] = STATE(13),
    [sym_function_item] = STATE(13),
    [sym_function_signature_item] = STATE(13),
    [sym_function_modifiers] = STATE(2528),
    [sym_impl_item] = STATE(13),
    [sym_trait_item] = STATE(13),
    [sym_associated_type] = STATE(13),
    [sym_let_declaration] = STATE(13),
    [sym_use_declaration] = STATE(13),
    [sym_extern_modifier] = STATE(1484),
    [sym_visibility_modifier] = STATE(1339),
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1216),
    [sym_macro_invocation] = STATE(68),
    [sym_scoped_identifier] = STATE(1172),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(70),
    [sym_match_expression] = STATE(70),
    [sym_while_expression] = STATE(70),
    [sym_loop_expression] = STATE(70),
    [sym_for_expression] = STATE(70),
    [sym_const_block] = STATE(70),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2521),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(70),
    [sym_async_block] = STATE(70),
    [sym_block] = STATE(70),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [aux_sym_source_file_repeat1] = STATE(13),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(9),
    [anon_sym_macro_rules_BANG] = ACTIONS(11),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_RBRACE] = ACTIONS(105),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(25),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(29),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(33),
    [anon_sym_enum] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(37),
    [anon_sym_for] = ACTIONS(39),
    [anon_sym_if] = ACTIONS(41),
    [anon_sym_impl] = ACTIONS(43),
    [anon_sym_let] = ACTIONS(45),
    [anon_sym_loop] = ACTIONS(47),
    [anon_sym_match] = ACTIONS(49),
    [anon_sym_mod] = ACTIONS(51),
    [anon_sym_pub] = ACTIONS(53),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_static] = ACTIONS(57),
    [anon_sym_struct] = ACTIONS(59),
    [anon_sym_trait] = ACTIONS(61),
    [anon_sym_type] = ACTIONS(63),
    [anon_sym_union] = ACTIONS(65),
    [anon_sym_unsafe] = ACTIONS(67),
    [anon_sym_use] = ACTIONS(69),
    [anon_sym_while] = ACTIONS(71),
    [anon_sym_POUND] = ACTIONS(73),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_extern] = ACTIONS(75),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(101),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [3] = {
    [sym__statement] = STATE(12),
    [sym_empty_statement] = STATE(12),
    [sym_expression_statement] = STATE(12),
    [sym_macro_definition] = STATE(12),
    [sym_attribute_item] = STATE(12),
    [sym_inner_attribute_item] = STATE(12),
    [sym_mod_item] = STATE(12),
    [sym_foreign_mod_item] = STATE(12),
    [sym_struct_item] = STATE(12),
    [sym_union_item] = STATE(12),
    [sym_enum_item] = STATE(12),
    [sym_extern_crate_declaration] = STATE(12),
    [sym_const_item] = STATE(12),
    [sym_static_item] = STATE(12),
    [sym_type_item] = STATE(12),
    [sym_function_item] = STATE(12),
    [sym_function_signature_item] = STATE(12),
    [sym_function_modifiers] = STATE(2528),
    [sym_impl_item] = STATE(12),
    [sym_trait_item] = STATE(12),
    [sym_associated_type] = STATE(12),
    [sym_let_declaration] = STATE(12),
    [sym_use_declaration] = STATE(12),
    [sym_extern_modifier] = STATE(1484),
    [sym_visibility_modifier] = STATE(1339),
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1239),
    [sym_macro_invocation] = STATE(68),
    [sym_scoped_identifier] = STATE(1172),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(70),
    [sym_match_expression] = STATE(70),
    [sym_while_expression] = STATE(70),
    [sym_loop_expression] = STATE(70),
    [sym_for_expression] = STATE(70),
    [sym_const_block] = STATE(70),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2521),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(70),
    [sym_async_block] = STATE(70),
    [sym_block] = STATE(70),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [aux_sym_source_file_repeat1] = STATE(12),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(9),
    [anon_sym_macro_rules_BANG] = ACTIONS(11),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_RBRACE] = ACTIONS(107),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(25),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(29),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(33),
    [anon_sym_enum] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(37),
    [anon_sym_for] = ACTIONS(39),
    [anon_sym_if] = ACTIONS(41),
    [anon_sym_impl] = ACTIONS(43),
    [anon_sym_let] = ACTIONS(45),
    [anon_sym_loop] = ACTIONS(47),
    [anon_sym_match] = ACTIONS(49),
    [anon_sym_mod] = ACTIONS(51),
    [anon_sym_pub] = ACTIONS(53),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_static] = ACTIONS(57),
    [anon_sym_struct] = ACTIONS(59),
    [anon_sym_trait] = ACTIONS(61),
    [anon_sym_type] = ACTIONS(63),
    [anon_sym_union] = ACTIONS(65),
    [anon_sym_unsafe] = ACTIONS(67),
    [anon_sym_use] = ACTIONS(69),
    [anon_sym_while] = ACTIONS(71),
    [anon_sym_POUND] = ACTIONS(73),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_extern] = ACTIONS(75),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(101),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [4] = {
    [sym__statement] = STATE(12),
    [sym_empty_statement] = STATE(12),
    [sym_expression_statement] = STATE(12),
    [sym_macro_definition] = STATE(12),
    [sym_attribute_item] = STATE(12),
    [sym_inner_attribute_item] = STATE(12),
    [sym_mod_item] = STATE(12),
    [sym_foreign_mod_item] = STATE(12),
    [sym_struct_item] = STATE(12),
    [sym_union_item] = STATE(12),
    [sym_enum_item] = STATE(12),
    [sym_extern_crate_declaration] = STATE(12),
    [sym_const_item] = STATE(12),
    [sym_static_item] = STATE(12),
    [sym_type_item] = STATE(12),
    [sym_function_item] = STATE(12),
    [sym_function_signature_item] = STATE(12),
    [sym_function_modifiers] = STATE(2528),
    [sym_impl_item] = STATE(12),
    [sym_trait_item] = STATE(12),
    [sym_associated_type] = STATE(12),
    [sym_let_declaration] = STATE(12),
    [sym_use_declaration] = STATE(12),
    [sym_extern_modifier] = STATE(1484),
    [sym_visibility_modifier] = STATE(1339),
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1242),
    [sym_macro_invocation] = STATE(68),
    [sym_scoped_identifier] = STATE(1172),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(70),
    [sym_match_expression] = STATE(70),
    [sym_while_expression] = STATE(70),
    [sym_loop_expression] = STATE(70),
    [sym_for_expression] = STATE(70),
    [sym_const_block] = STATE(70),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2521),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(70),
    [sym_async_block] = STATE(70),
    [sym_block] = STATE(70),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [aux_sym_source_file_repeat1] = STATE(12),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(9),
    [anon_sym_macro_rules_BANG] = ACTIONS(11),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_RBRACE] = ACTIONS(109),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(25),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(29),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(33),
    [anon_sym_enum] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(37),
    [anon_sym_for] = ACTIONS(39),
    [anon_sym_if] = ACTIONS(41),
    [anon_sym_impl] = ACTIONS(43),
    [anon_sym_let] = ACTIONS(45),
    [anon_sym_loop] = ACTIONS(47),
    [anon_sym_match] = ACTIONS(49),
    [anon_sym_mod] = ACTIONS(51),
    [anon_sym_pub] = ACTIONS(53),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_static] = ACTIONS(57),
    [anon_sym_struct] = ACTIONS(59),
    [anon_sym_trait] = ACTIONS(61),
    [anon_sym_type] = ACTIONS(63),
    [anon_sym_union] = ACTIONS(65),
    [anon_sym_unsafe] = ACTIONS(67),
    [anon_sym_use] = ACTIONS(69),
    [anon_sym_while] = ACTIONS(71),
    [anon_sym_POUND] = ACTIONS(73),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_extern] = ACTIONS(75),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(101),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [5] = {
    [sym__statement] = STATE(11),
    [sym_empty_statement] = STATE(11),
    [sym_expression_statement] = STATE(11),
    [sym_macro_definition] = STATE(11),
    [sym_attribute_item] = STATE(11),
    [sym_inner_attribute_item] = STATE(11),
    [sym_mod_item] = STATE(11),
    [sym_foreign_mod_item] = STATE(11),
    [sym_struct_item] = STATE(11),
    [sym_union_item] = STATE(11),
    [sym_enum_item] = STATE(11),
    [sym_extern_crate_declaration] = STATE(11),
    [sym_const_item] = STATE(11),
    [sym_static_item] = STATE(11),
    [sym_type_item] = STATE(11),
    [sym_function_item] = STATE(11),
    [sym_function_signature_item] = STATE(11),
    [sym_function_modifiers] = STATE(2528),
    [sym_impl_item] = STATE(11),
    [sym_trait_item] = STATE(11),
    [sym_associated_type] = STATE(11),
    [sym_let_declaration] = STATE(11),
    [sym_use_declaration] = STATE(11),
    [sym_extern_modifier] = STATE(1484),
    [sym_visibility_modifier] = STATE(1339),
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1213),
    [sym_macro_invocation] = STATE(68),
    [sym_scoped_identifier] = STATE(1172),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(70),
    [sym_match_expression] = STATE(70),
    [sym_while_expression] = STATE(70),
    [sym_loop_expression] = STATE(70),
    [sym_for_expression] = STATE(70),
    [sym_const_block] = STATE(70),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2521),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(70),
    [sym_async_block] = STATE(70),
    [sym_block] = STATE(70),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [aux_sym_source_file_repeat1] = STATE(11),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(9),
    [anon_sym_macro_rules_BANG] = ACTIONS(11),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_RBRACE] = ACTIONS(111),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(25),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(29),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(33),
    [anon_sym_enum] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(37),
    [anon_sym_for] = ACTIONS(39),
    [anon_sym_if] = ACTIONS(41),
    [anon_sym_impl] = ACTIONS(43),
    [anon_sym_let] = ACTIONS(45),
    [anon_sym_loop] = ACTIONS(47),
    [anon_sym_match] = ACTIONS(49),
    [anon_sym_mod] = ACTIONS(51),
    [anon_sym_pub] = ACTIONS(53),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_static] = ACTIONS(57),
    [anon_sym_struct] = ACTIONS(59),
    [anon_sym_trait] = ACTIONS(61),
    [anon_sym_type] = ACTIONS(63),
    [anon_sym_union] = ACTIONS(65),
    [anon_sym_unsafe] = ACTIONS(67),
    [anon_sym_use] = ACTIONS(69),
    [anon_sym_while] = ACTIONS(71),
    [anon_sym_POUND] = ACTIONS(73),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_extern] = ACTIONS(75),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(101),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [6] = {
    [sym__statement] = STATE(14),
    [sym_empty_statement] = STATE(14),
    [sym_expression_statement] = STATE(14),
    [sym_macro_definition] = STATE(14),
    [sym_attribute_item] = STATE(14),
    [sym_inner_attribute_item] = STATE(14),
    [sym_mod_item] = STATE(14),
    [sym_foreign_mod_item] = STATE(14),
    [sym_struct_item] = STATE(14),
    [sym_union_item] = STATE(14),
    [sym_enum_item] = STATE(14),
    [sym_extern_crate_declaration] = STATE(14),
    [sym_const_item] = STATE(14),
    [sym_static_item] = STATE(14),
    [sym_type_item] = STATE(14),
    [sym_function_item] = STATE(14),
    [sym_function_signature_item] = STATE(14),
    [sym_function_modifiers] = STATE(2528),
    [sym_impl_item] = STATE(14),
    [sym_trait_item] = STATE(14),
    [sym_associated_type] = STATE(14),
    [sym_let_declaration] = STATE(14),
    [sym_use_declaration] = STATE(14),
    [sym_extern_modifier] = STATE(1484),
    [sym_visibility_modifier] = STATE(1339),
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1230),
    [sym_macro_invocation] = STATE(68),
    [sym_scoped_identifier] = STATE(1172),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(70),
    [sym_match_expression] = STATE(70),
    [sym_while_expression] = STATE(70),
    [sym_loop_expression] = STATE(70),
    [sym_for_expression] = STATE(70),
    [sym_const_block] = STATE(70),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2521),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(70),
    [sym_async_block] = STATE(70),
    [sym_block] = STATE(70),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [aux_sym_source_file_repeat1] = STATE(14),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(9),
    [anon_sym_macro_rules_BANG] = ACTIONS(11),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_RBRACE] = ACTIONS(113),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(25),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(29),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(33),
    [anon_sym_enum] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(37),
    [anon_sym_for] = ACTIONS(39),
    [anon_sym_if] = ACTIONS(41),
    [anon_sym_impl] = ACTIONS(43),
    [anon_sym_let] = ACTIONS(45),
    [anon_sym_loop] = ACTIONS(47),
    [anon_sym_match] = ACTIONS(49),
    [anon_sym_mod] = ACTIONS(51),
    [anon_sym_pub] = ACTIONS(53),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_static] = ACTIONS(57),
    [anon_sym_struct] = ACTIONS(59),
    [anon_sym_trait] = ACTIONS(61),
    [anon_sym_type] = ACTIONS(63),
    [anon_sym_union] = ACTIONS(65),
    [anon_sym_unsafe] = ACTIONS(67),
    [anon_sym_use] = ACTIONS(69),
    [anon_sym_while] = ACTIONS(71),
    [anon_sym_POUND] = ACTIONS(73),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_extern] = ACTIONS(75),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(101),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [7] = {
    [sym__statement] = STATE(4),
    [sym_empty_statement] = STATE(4),
    [sym_expression_statement] = STATE(4),
    [sym_macro_definition] = STATE(4),
    [sym_attribute_item] = STATE(4),
    [sym_inner_attribute_item] = STATE(4),
    [sym_mod_item] = STATE(4),
    [sym_foreign_mod_item] = STATE(4),
    [sym_struct_item] = STATE(4),
    [sym_union_item] = STATE(4),
    [sym_enum_item] = STATE(4),
    [sym_extern_crate_declaration] = STATE(4),
    [sym_const_item] = STATE(4),
    [sym_static_item] = STATE(4),
    [sym_type_item] = STATE(4),
    [sym_function_item] = STATE(4),
    [sym_function_signature_item] = STATE(4),
    [sym_function_modifiers] = STATE(2528),
    [sym_impl_item] = STATE(4),
    [sym_trait_item] = STATE(4),
    [sym_associated_type] = STATE(4),
    [sym_let_declaration] = STATE(4),
    [sym_use_declaration] = STATE(4),
    [sym_extern_modifier] = STATE(1484),
    [sym_visibility_modifier] = STATE(1339),
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1260),
    [sym_macro_invocation] = STATE(68),
    [sym_scoped_identifier] = STATE(1172),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(70),
    [sym_match_expression] = STATE(70),
    [sym_while_expression] = STATE(70),
    [sym_loop_expression] = STATE(70),
    [sym_for_expression] = STATE(70),
    [sym_const_block] = STATE(70),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2521),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(70),
    [sym_async_block] = STATE(70),
    [sym_block] = STATE(70),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [aux_sym_source_file_repeat1] = STATE(4),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(9),
    [anon_sym_macro_rules_BANG] = ACTIONS(11),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_RBRACE] = ACTIONS(115),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(25),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(29),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(33),
    [anon_sym_enum] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(37),
    [anon_sym_for] = ACTIONS(39),
    [anon_sym_if] = ACTIONS(41),
    [anon_sym_impl] = ACTIONS(43),
    [anon_sym_let] = ACTIONS(45),
    [anon_sym_loop] = ACTIONS(47),
    [anon_sym_match] = ACTIONS(49),
    [anon_sym_mod] = ACTIONS(51),
    [anon_sym_pub] = ACTIONS(53),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_static] = ACTIONS(57),
    [anon_sym_struct] = ACTIONS(59),
    [anon_sym_trait] = ACTIONS(61),
    [anon_sym_type] = ACTIONS(63),
    [anon_sym_union] = ACTIONS(65),
    [anon_sym_unsafe] = ACTIONS(67),
    [anon_sym_use] = ACTIONS(69),
    [anon_sym_while] = ACTIONS(71),
    [anon_sym_POUND] = ACTIONS(73),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_extern] = ACTIONS(75),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(101),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [8] = {
    [sym__statement] = STATE(8),
    [sym_empty_statement] = STATE(8),
    [sym_expression_statement] = STATE(8),
    [sym_macro_definition] = STATE(8),
    [sym_attribute_item] = STATE(8),
    [sym_inner_attribute_item] = STATE(8),
    [sym_mod_item] = STATE(8),
    [sym_foreign_mod_item] = STATE(8),
    [sym_struct_item] = STATE(8),
    [sym_union_item] = STATE(8),
    [sym_enum_item] = STATE(8),
    [sym_extern_crate_declaration] = STATE(8),
    [sym_const_item] = STATE(8),
    [sym_static_item] = STATE(8),
    [sym_type_item] = STATE(8),
    [sym_function_item] = STATE(8),
    [sym_function_signature_item] = STATE(8),
    [sym_function_modifiers] = STATE(2528),
    [sym_impl_item] = STATE(8),
    [sym_trait_item] = STATE(8),
    [sym_associated_type] = STATE(8),
    [sym_let_declaration] = STATE(8),
    [sym_use_declaration] = STATE(8),
    [sym_extern_modifier] = STATE(1484),
    [sym_visibility_modifier] = STATE(1339),
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1266),
    [sym_macro_invocation] = STATE(68),
    [sym_scoped_identifier] = STATE(1172),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(70),
    [sym_match_expression] = STATE(70),
    [sym_while_expression] = STATE(70),
    [sym_loop_expression] = STATE(70),
    [sym_for_expression] = STATE(70),
    [sym_const_block] = STATE(70),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2521),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(70),
    [sym_async_block] = STATE(70),
    [sym_block] = STATE(70),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [aux_sym_source_file_repeat1] = STATE(8),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [ts_builtin_sym_end] = ACTIONS(117),
    [sym_identifier] = ACTIONS(119),
    [anon_sym_SEMI] = ACTIONS(122),
    [anon_sym_macro_rules_BANG] = ACTIONS(125),
    [anon_sym_LPAREN] = ACTIONS(128),
    [anon_sym_LBRACE] = ACTIONS(131),
    [anon_sym_LBRACK] = ACTIONS(134),
    [anon_sym_STAR] = ACTIONS(137),
    [anon_sym_u8] = ACTIONS(140),
    [anon_sym_i8] = ACTIONS(140),
    [anon_sym_u16] = ACTIONS(140),
    [anon_sym_i16] = ACTIONS(140),
    [anon_sym_u32] = ACTIONS(140),
    [anon_sym_i32] = ACTIONS(140),
    [anon_sym_u64] = ACTIONS(140),
    [anon_sym_i64] = ACTIONS(140),
    [anon_sym_u128] = ACTIONS(140),
    [anon_sym_i128] = ACTIONS(140),
    [anon_sym_isize] = ACTIONS(140),
    [anon_sym_usize] = ACTIONS(140),
    [anon_sym_f32] = ACTIONS(140),
    [anon_sym_f64] = ACTIONS(140),
    [anon_sym_bool] = ACTIONS(140),
    [anon_sym_str] = ACTIONS(140),
    [anon_sym_char] = ACTIONS(140),
    [anon_sym_SQUOTE] = ACTIONS(143),
    [anon_sym_async] = ACTIONS(146),
    [anon_sym_break] = ACTIONS(149),
    [anon_sym_const] = ACTIONS(152),
    [anon_sym_continue] = ACTIONS(155),
    [anon_sym_default] = ACTIONS(158),
    [anon_sym_enum] = ACTIONS(161),
    [anon_sym_fn] = ACTIONS(164),
    [anon_sym_for] = ACTIONS(167),
    [anon_sym_if] = ACTIONS(170),
    [anon_sym_impl] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(176),
    [anon_sym_loop] = ACTIONS(179),
    [anon_sym_match] = ACTIONS(182),
    [anon_sym_mod] = ACTIONS(185),
    [anon_sym_pub] = ACTIONS(188),
    [anon_sym_return] = ACTIONS(191),
    [anon_sym_static] = ACTIONS(194),
    [anon_sym_struct] = ACTIONS(197),
    [anon_sym_trait] = ACTIONS(200),
    [anon_sym_type] = ACTIONS(203),
    [anon_sym_union] = ACTIONS(206),
    [anon_sym_unsafe] = ACTIONS(209),
    [anon_sym_use] = ACTIONS(212),
    [anon_sym_while] = ACTIONS(215),
    [anon_sym_POUND] = ACTIONS(218),
    [anon_sym_BANG] = ACTIONS(137),
    [anon_sym_extern] = ACTIONS(221),
    [anon_sym_LT] = ACTIONS(224),
    [anon_sym_COLON_COLON] = ACTIONS(227),
    [anon_sym_AMP] = ACTIONS(230),
    [anon_sym_DOT_DOT] = ACTIONS(233),
    [anon_sym_DASH] = ACTIONS(137),
    [anon_sym_PIPE] = ACTIONS(236),
    [anon_sym_yield] = ACTIONS(239),
    [anon_sym_move] = ACTIONS(242),
    [sym_integer_literal] = ACTIONS(245),
    [aux_sym_string_literal_token1] = ACTIONS(248),
    [sym_char_literal] = ACTIONS(245),
    [anon_sym_true] = ACTIONS(251),
    [anon_sym_false] = ACTIONS(251),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(254),
    [sym_super] = ACTIONS(257),
    [sym_crate] = ACTIONS(260),
    [sym_metavariable] = ACTIONS(263),
    [sym_raw_string_literal] = ACTIONS(245),
    [sym_float_literal] = ACTIONS(245),
    [sym_block_comment] = ACTIONS(3),
  },
  [9] = {
    [sym__statement] = STATE(8),
    [sym_empty_statement] = STATE(8),
    [sym_expression_statement] = STATE(8),
    [sym_macro_definition] = STATE(8),
    [sym_attribute_item] = STATE(8),
    [sym_inner_attribute_item] = STATE(8),
    [sym_mod_item] = STATE(8),
    [sym_foreign_mod_item] = STATE(8),
    [sym_struct_item] = STATE(8),
    [sym_union_item] = STATE(8),
    [sym_enum_item] = STATE(8),
    [sym_extern_crate_declaration] = STATE(8),
    [sym_const_item] = STATE(8),
    [sym_static_item] = STATE(8),
    [sym_type_item] = STATE(8),
    [sym_function_item] = STATE(8),
    [sym_function_signature_item] = STATE(8),
    [sym_function_modifiers] = STATE(2528),
    [sym_impl_item] = STATE(8),
    [sym_trait_item] = STATE(8),
    [sym_associated_type] = STATE(8),
    [sym_let_declaration] = STATE(8),
    [sym_use_declaration] = STATE(8),
    [sym_extern_modifier] = STATE(1484),
    [sym_visibility_modifier] = STATE(1339),
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1266),
    [sym_macro_invocation] = STATE(68),
    [sym_scoped_identifier] = STATE(1172),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(70),
    [sym_match_expression] = STATE(70),
    [sym_while_expression] = STATE(70),
    [sym_loop_expression] = STATE(70),
    [sym_for_expression] = STATE(70),
    [sym_const_block] = STATE(70),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2521),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(70),
    [sym_async_block] = STATE(70),
    [sym_block] = STATE(70),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [aux_sym_source_file_repeat1] = STATE(8),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [ts_builtin_sym_end] = ACTIONS(266),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(9),
    [anon_sym_macro_rules_BANG] = ACTIONS(11),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(25),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(29),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(33),
    [anon_sym_enum] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(37),
    [anon_sym_for] = ACTIONS(39),
    [anon_sym_if] = ACTIONS(41),
    [anon_sym_impl] = ACTIONS(43),
    [anon_sym_let] = ACTIONS(45),
    [anon_sym_loop] = ACTIONS(47),
    [anon_sym_match] = ACTIONS(49),
    [anon_sym_mod] = ACTIONS(51),
    [anon_sym_pub] = ACTIONS(53),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_static] = ACTIONS(57),
    [anon_sym_struct] = ACTIONS(59),
    [anon_sym_trait] = ACTIONS(61),
    [anon_sym_type] = ACTIONS(63),
    [anon_sym_union] = ACTIONS(65),
    [anon_sym_unsafe] = ACTIONS(67),
    [anon_sym_use] = ACTIONS(69),
    [anon_sym_while] = ACTIONS(71),
    [anon_sym_POUND] = ACTIONS(73),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_extern] = ACTIONS(75),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(101),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [10] = {
    [sym__statement] = STATE(3),
    [sym_empty_statement] = STATE(3),
    [sym_expression_statement] = STATE(3),
    [sym_macro_definition] = STATE(3),
    [sym_attribute_item] = STATE(3),
    [sym_inner_attribute_item] = STATE(3),
    [sym_mod_item] = STATE(3),
    [sym_foreign_mod_item] = STATE(3),
    [sym_struct_item] = STATE(3),
    [sym_union_item] = STATE(3),
    [sym_enum_item] = STATE(3),
    [sym_extern_crate_declaration] = STATE(3),
    [sym_const_item] = STATE(3),
    [sym_static_item] = STATE(3),
    [sym_type_item] = STATE(3),
    [sym_function_item] = STATE(3),
    [sym_function_signature_item] = STATE(3),
    [sym_function_modifiers] = STATE(2528),
    [sym_impl_item] = STATE(3),
    [sym_trait_item] = STATE(3),
    [sym_associated_type] = STATE(3),
    [sym_let_declaration] = STATE(3),
    [sym_use_declaration] = STATE(3),
    [sym_extern_modifier] = STATE(1484),
    [sym_visibility_modifier] = STATE(1339),
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1228),
    [sym_macro_invocation] = STATE(68),
    [sym_scoped_identifier] = STATE(1172),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(70),
    [sym_match_expression] = STATE(70),
    [sym_while_expression] = STATE(70),
    [sym_loop_expression] = STATE(70),
    [sym_for_expression] = STATE(70),
    [sym_const_block] = STATE(70),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2521),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(70),
    [sym_async_block] = STATE(70),
    [sym_block] = STATE(70),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [aux_sym_source_file_repeat1] = STATE(3),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(9),
    [anon_sym_macro_rules_BANG] = ACTIONS(11),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_RBRACE] = ACTIONS(268),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(25),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(29),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(33),
    [anon_sym_enum] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(37),
    [anon_sym_for] = ACTIONS(39),
    [anon_sym_if] = ACTIONS(41),
    [anon_sym_impl] = ACTIONS(43),
    [anon_sym_let] = ACTIONS(45),
    [anon_sym_loop] = ACTIONS(47),
    [anon_sym_match] = ACTIONS(49),
    [anon_sym_mod] = ACTIONS(51),
    [anon_sym_pub] = ACTIONS(53),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_static] = ACTIONS(57),
    [anon_sym_struct] = ACTIONS(59),
    [anon_sym_trait] = ACTIONS(61),
    [anon_sym_type] = ACTIONS(63),
    [anon_sym_union] = ACTIONS(65),
    [anon_sym_unsafe] = ACTIONS(67),
    [anon_sym_use] = ACTIONS(69),
    [anon_sym_while] = ACTIONS(71),
    [anon_sym_POUND] = ACTIONS(73),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_extern] = ACTIONS(75),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(101),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [11] = {
    [sym__statement] = STATE(12),
    [sym_empty_statement] = STATE(12),
    [sym_expression_statement] = STATE(12),
    [sym_macro_definition] = STATE(12),
    [sym_attribute_item] = STATE(12),
    [sym_inner_attribute_item] = STATE(12),
    [sym_mod_item] = STATE(12),
    [sym_foreign_mod_item] = STATE(12),
    [sym_struct_item] = STATE(12),
    [sym_union_item] = STATE(12),
    [sym_enum_item] = STATE(12),
    [sym_extern_crate_declaration] = STATE(12),
    [sym_const_item] = STATE(12),
    [sym_static_item] = STATE(12),
    [sym_type_item] = STATE(12),
    [sym_function_item] = STATE(12),
    [sym_function_signature_item] = STATE(12),
    [sym_function_modifiers] = STATE(2528),
    [sym_impl_item] = STATE(12),
    [sym_trait_item] = STATE(12),
    [sym_associated_type] = STATE(12),
    [sym_let_declaration] = STATE(12),
    [sym_use_declaration] = STATE(12),
    [sym_extern_modifier] = STATE(1484),
    [sym_visibility_modifier] = STATE(1339),
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1206),
    [sym_macro_invocation] = STATE(68),
    [sym_scoped_identifier] = STATE(1172),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(70),
    [sym_match_expression] = STATE(70),
    [sym_while_expression] = STATE(70),
    [sym_loop_expression] = STATE(70),
    [sym_for_expression] = STATE(70),
    [sym_const_block] = STATE(70),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2521),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(70),
    [sym_async_block] = STATE(70),
    [sym_block] = STATE(70),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [aux_sym_source_file_repeat1] = STATE(12),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(9),
    [anon_sym_macro_rules_BANG] = ACTIONS(11),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_RBRACE] = ACTIONS(270),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(25),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(29),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(33),
    [anon_sym_enum] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(37),
    [anon_sym_for] = ACTIONS(39),
    [anon_sym_if] = ACTIONS(41),
    [anon_sym_impl] = ACTIONS(43),
    [anon_sym_let] = ACTIONS(45),
    [anon_sym_loop] = ACTIONS(47),
    [anon_sym_match] = ACTIONS(49),
    [anon_sym_mod] = ACTIONS(51),
    [anon_sym_pub] = ACTIONS(53),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_static] = ACTIONS(57),
    [anon_sym_struct] = ACTIONS(59),
    [anon_sym_trait] = ACTIONS(61),
    [anon_sym_type] = ACTIONS(63),
    [anon_sym_union] = ACTIONS(65),
    [anon_sym_unsafe] = ACTIONS(67),
    [anon_sym_use] = ACTIONS(69),
    [anon_sym_while] = ACTIONS(71),
    [anon_sym_POUND] = ACTIONS(73),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_extern] = ACTIONS(75),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(101),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [12] = {
    [sym__statement] = STATE(12),
    [sym_empty_statement] = STATE(12),
    [sym_expression_statement] = STATE(12),
    [sym_macro_definition] = STATE(12),
    [sym_attribute_item] = STATE(12),
    [sym_inner_attribute_item] = STATE(12),
    [sym_mod_item] = STATE(12),
    [sym_foreign_mod_item] = STATE(12),
    [sym_struct_item] = STATE(12),
    [sym_union_item] = STATE(12),
    [sym_enum_item] = STATE(12),
    [sym_extern_crate_declaration] = STATE(12),
    [sym_const_item] = STATE(12),
    [sym_static_item] = STATE(12),
    [sym_type_item] = STATE(12),
    [sym_function_item] = STATE(12),
    [sym_function_signature_item] = STATE(12),
    [sym_function_modifiers] = STATE(2528),
    [sym_impl_item] = STATE(12),
    [sym_trait_item] = STATE(12),
    [sym_associated_type] = STATE(12),
    [sym_let_declaration] = STATE(12),
    [sym_use_declaration] = STATE(12),
    [sym_extern_modifier] = STATE(1484),
    [sym_visibility_modifier] = STATE(1339),
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1266),
    [sym_macro_invocation] = STATE(96),
    [sym_scoped_identifier] = STATE(1172),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(70),
    [sym_match_expression] = STATE(70),
    [sym_while_expression] = STATE(70),
    [sym_loop_expression] = STATE(70),
    [sym_for_expression] = STATE(70),
    [sym_const_block] = STATE(70),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2521),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(70),
    [sym_async_block] = STATE(70),
    [sym_block] = STATE(70),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [aux_sym_source_file_repeat1] = STATE(12),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(119),
    [anon_sym_SEMI] = ACTIONS(122),
    [anon_sym_macro_rules_BANG] = ACTIONS(125),
    [anon_sym_LPAREN] = ACTIONS(128),
    [anon_sym_LBRACE] = ACTIONS(131),
    [anon_sym_RBRACE] = ACTIONS(117),
    [anon_sym_LBRACK] = ACTIONS(134),
    [anon_sym_STAR] = ACTIONS(137),
    [anon_sym_u8] = ACTIONS(140),
    [anon_sym_i8] = ACTIONS(140),
    [anon_sym_u16] = ACTIONS(140),
    [anon_sym_i16] = ACTIONS(140),
    [anon_sym_u32] = ACTIONS(140),
    [anon_sym_i32] = ACTIONS(140),
    [anon_sym_u64] = ACTIONS(140),
    [anon_sym_i64] = ACTIONS(140),
    [anon_sym_u128] = ACTIONS(140),
    [anon_sym_i128] = ACTIONS(140),
    [anon_sym_isize] = ACTIONS(140),
    [anon_sym_usize] = ACTIONS(140),
    [anon_sym_f32] = ACTIONS(140),
    [anon_sym_f64] = ACTIONS(140),
    [anon_sym_bool] = ACTIONS(140),
    [anon_sym_str] = ACTIONS(140),
    [anon_sym_char] = ACTIONS(140),
    [anon_sym_SQUOTE] = ACTIONS(143),
    [anon_sym_async] = ACTIONS(146),
    [anon_sym_break] = ACTIONS(149),
    [anon_sym_const] = ACTIONS(152),
    [anon_sym_continue] = ACTIONS(155),
    [anon_sym_default] = ACTIONS(158),
    [anon_sym_enum] = ACTIONS(161),
    [anon_sym_fn] = ACTIONS(164),
    [anon_sym_for] = ACTIONS(167),
    [anon_sym_if] = ACTIONS(170),
    [anon_sym_impl] = ACTIONS(173),
    [anon_sym_let] = ACTIONS(176),
    [anon_sym_loop] = ACTIONS(179),
    [anon_sym_match] = ACTIONS(182),
    [anon_sym_mod] = ACTIONS(185),
    [anon_sym_pub] = ACTIONS(188),
    [anon_sym_return] = ACTIONS(191),
    [anon_sym_static] = ACTIONS(194),
    [anon_sym_struct] = ACTIONS(197),
    [anon_sym_trait] = ACTIONS(200),
    [anon_sym_type] = ACTIONS(203),
    [anon_sym_union] = ACTIONS(206),
    [anon_sym_unsafe] = ACTIONS(209),
    [anon_sym_use] = ACTIONS(212),
    [anon_sym_while] = ACTIONS(215),
    [anon_sym_POUND] = ACTIONS(218),
    [anon_sym_BANG] = ACTIONS(137),
    [anon_sym_extern] = ACTIONS(221),
    [anon_sym_LT] = ACTIONS(224),
    [anon_sym_COLON_COLON] = ACTIONS(227),
    [anon_sym_AMP] = ACTIONS(230),
    [anon_sym_DOT_DOT] = ACTIONS(233),
    [anon_sym_DASH] = ACTIONS(137),
    [anon_sym_PIPE] = ACTIONS(236),
    [anon_sym_yield] = ACTIONS(239),
    [anon_sym_move] = ACTIONS(242),
    [sym_integer_literal] = ACTIONS(245),
    [aux_sym_string_literal_token1] = ACTIONS(248),
    [sym_char_literal] = ACTIONS(245),
    [anon_sym_true] = ACTIONS(251),
    [anon_sym_false] = ACTIONS(251),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(254),
    [sym_super] = ACTIONS(257),
    [sym_crate] = ACTIONS(260),
    [sym_metavariable] = ACTIONS(263),
    [sym_raw_string_literal] = ACTIONS(245),
    [sym_float_literal] = ACTIONS(245),
    [sym_block_comment] = ACTIONS(3),
  },
  [13] = {
    [sym__statement] = STATE(12),
    [sym_empty_statement] = STATE(12),
    [sym_expression_statement] = STATE(12),
    [sym_macro_definition] = STATE(12),
    [sym_attribute_item] = STATE(12),
    [sym_inner_attribute_item] = STATE(12),
    [sym_mod_item] = STATE(12),
    [sym_foreign_mod_item] = STATE(12),
    [sym_struct_item] = STATE(12),
    [sym_union_item] = STATE(12),
    [sym_enum_item] = STATE(12),
    [sym_extern_crate_declaration] = STATE(12),
    [sym_const_item] = STATE(12),
    [sym_static_item] = STATE(12),
    [sym_type_item] = STATE(12),
    [sym_function_item] = STATE(12),
    [sym_function_signature_item] = STATE(12),
    [sym_function_modifiers] = STATE(2528),
    [sym_impl_item] = STATE(12),
    [sym_trait_item] = STATE(12),
    [sym_associated_type] = STATE(12),
    [sym_let_declaration] = STATE(12),
    [sym_use_declaration] = STATE(12),
    [sym_extern_modifier] = STATE(1484),
    [sym_visibility_modifier] = STATE(1339),
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1220),
    [sym_macro_invocation] = STATE(68),
    [sym_scoped_identifier] = STATE(1172),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(70),
    [sym_match_expression] = STATE(70),
    [sym_while_expression] = STATE(70),
    [sym_loop_expression] = STATE(70),
    [sym_for_expression] = STATE(70),
    [sym_const_block] = STATE(70),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2521),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(70),
    [sym_async_block] = STATE(70),
    [sym_block] = STATE(70),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [aux_sym_source_file_repeat1] = STATE(12),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(9),
    [anon_sym_macro_rules_BANG] = ACTIONS(11),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_RBRACE] = ACTIONS(272),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(25),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(29),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(33),
    [anon_sym_enum] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(37),
    [anon_sym_for] = ACTIONS(39),
    [anon_sym_if] = ACTIONS(41),
    [anon_sym_impl] = ACTIONS(43),
    [anon_sym_let] = ACTIONS(45),
    [anon_sym_loop] = ACTIONS(47),
    [anon_sym_match] = ACTIONS(49),
    [anon_sym_mod] = ACTIONS(51),
    [anon_sym_pub] = ACTIONS(53),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_static] = ACTIONS(57),
    [anon_sym_struct] = ACTIONS(59),
    [anon_sym_trait] = ACTIONS(61),
    [anon_sym_type] = ACTIONS(63),
    [anon_sym_union] = ACTIONS(65),
    [anon_sym_unsafe] = ACTIONS(67),
    [anon_sym_use] = ACTIONS(69),
    [anon_sym_while] = ACTIONS(71),
    [anon_sym_POUND] = ACTIONS(73),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_extern] = ACTIONS(75),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(101),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [14] = {
    [sym__statement] = STATE(12),
    [sym_empty_statement] = STATE(12),
    [sym_expression_statement] = STATE(12),
    [sym_macro_definition] = STATE(12),
    [sym_attribute_item] = STATE(12),
    [sym_inner_attribute_item] = STATE(12),
    [sym_mod_item] = STATE(12),
    [sym_foreign_mod_item] = STATE(12),
    [sym_struct_item] = STATE(12),
    [sym_union_item] = STATE(12),
    [sym_enum_item] = STATE(12),
    [sym_extern_crate_declaration] = STATE(12),
    [sym_const_item] = STATE(12),
    [sym_static_item] = STATE(12),
    [sym_type_item] = STATE(12),
    [sym_function_item] = STATE(12),
    [sym_function_signature_item] = STATE(12),
    [sym_function_modifiers] = STATE(2528),
    [sym_impl_item] = STATE(12),
    [sym_trait_item] = STATE(12),
    [sym_associated_type] = STATE(12),
    [sym_let_declaration] = STATE(12),
    [sym_use_declaration] = STATE(12),
    [sym_extern_modifier] = STATE(1484),
    [sym_visibility_modifier] = STATE(1339),
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1219),
    [sym_macro_invocation] = STATE(68),
    [sym_scoped_identifier] = STATE(1172),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(70),
    [sym_match_expression] = STATE(70),
    [sym_while_expression] = STATE(70),
    [sym_loop_expression] = STATE(70),
    [sym_for_expression] = STATE(70),
    [sym_const_block] = STATE(70),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2521),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(70),
    [sym_async_block] = STATE(70),
    [sym_block] = STATE(70),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [aux_sym_source_file_repeat1] = STATE(12),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(9),
    [anon_sym_macro_rules_BANG] = ACTIONS(11),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_RBRACE] = ACTIONS(274),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(25),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(29),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(33),
    [anon_sym_enum] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(37),
    [anon_sym_for] = ACTIONS(39),
    [anon_sym_if] = ACTIONS(41),
    [anon_sym_impl] = ACTIONS(43),
    [anon_sym_let] = ACTIONS(45),
    [anon_sym_loop] = ACTIONS(47),
    [anon_sym_match] = ACTIONS(49),
    [anon_sym_mod] = ACTIONS(51),
    [anon_sym_pub] = ACTIONS(53),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_static] = ACTIONS(57),
    [anon_sym_struct] = ACTIONS(59),
    [anon_sym_trait] = ACTIONS(61),
    [anon_sym_type] = ACTIONS(63),
    [anon_sym_union] = ACTIONS(65),
    [anon_sym_unsafe] = ACTIONS(67),
    [anon_sym_use] = ACTIONS(69),
    [anon_sym_while] = ACTIONS(71),
    [anon_sym_POUND] = ACTIONS(73),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_extern] = ACTIONS(75),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(101),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [15] = {
    [sym__statement] = STATE(12),
    [sym_empty_statement] = STATE(12),
    [sym_expression_statement] = STATE(12),
    [sym_macro_definition] = STATE(12),
    [sym_attribute_item] = STATE(12),
    [sym_inner_attribute_item] = STATE(12),
    [sym_mod_item] = STATE(12),
    [sym_foreign_mod_item] = STATE(12),
    [sym_struct_item] = STATE(12),
    [sym_union_item] = STATE(12),
    [sym_enum_item] = STATE(12),
    [sym_extern_crate_declaration] = STATE(12),
    [sym_const_item] = STATE(12),
    [sym_static_item] = STATE(12),
    [sym_type_item] = STATE(12),
    [sym_function_item] = STATE(12),
    [sym_function_signature_item] = STATE(12),
    [sym_function_modifiers] = STATE(2528),
    [sym_impl_item] = STATE(12),
    [sym_trait_item] = STATE(12),
    [sym_associated_type] = STATE(12),
    [sym_let_declaration] = STATE(12),
    [sym_use_declaration] = STATE(12),
    [sym_extern_modifier] = STATE(1484),
    [sym_visibility_modifier] = STATE(1339),
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1257),
    [sym_macro_invocation] = STATE(68),
    [sym_scoped_identifier] = STATE(1172),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(70),
    [sym_match_expression] = STATE(70),
    [sym_while_expression] = STATE(70),
    [sym_loop_expression] = STATE(70),
    [sym_for_expression] = STATE(70),
    [sym_const_block] = STATE(70),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2521),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(70),
    [sym_async_block] = STATE(70),
    [sym_block] = STATE(70),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [aux_sym_source_file_repeat1] = STATE(12),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(9),
    [anon_sym_macro_rules_BANG] = ACTIONS(11),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_RBRACE] = ACTIONS(276),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(25),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(29),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(33),
    [anon_sym_enum] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(37),
    [anon_sym_for] = ACTIONS(39),
    [anon_sym_if] = ACTIONS(41),
    [anon_sym_impl] = ACTIONS(43),
    [anon_sym_let] = ACTIONS(45),
    [anon_sym_loop] = ACTIONS(47),
    [anon_sym_match] = ACTIONS(49),
    [anon_sym_mod] = ACTIONS(51),
    [anon_sym_pub] = ACTIONS(53),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_static] = ACTIONS(57),
    [anon_sym_struct] = ACTIONS(59),
    [anon_sym_trait] = ACTIONS(61),
    [anon_sym_type] = ACTIONS(63),
    [anon_sym_union] = ACTIONS(65),
    [anon_sym_unsafe] = ACTIONS(67),
    [anon_sym_use] = ACTIONS(69),
    [anon_sym_while] = ACTIONS(71),
    [anon_sym_POUND] = ACTIONS(73),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_extern] = ACTIONS(75),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(101),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [16] = {
    [sym__statement] = STATE(15),
    [sym_empty_statement] = STATE(15),
    [sym_expression_statement] = STATE(15),
    [sym_macro_definition] = STATE(15),
    [sym_attribute_item] = STATE(15),
    [sym_inner_attribute_item] = STATE(15),
    [sym_mod_item] = STATE(15),
    [sym_foreign_mod_item] = STATE(15),
    [sym_struct_item] = STATE(15),
    [sym_union_item] = STATE(15),
    [sym_enum_item] = STATE(15),
    [sym_extern_crate_declaration] = STATE(15),
    [sym_const_item] = STATE(15),
    [sym_static_item] = STATE(15),
    [sym_type_item] = STATE(15),
    [sym_function_item] = STATE(15),
    [sym_function_signature_item] = STATE(15),
    [sym_function_modifiers] = STATE(2528),
    [sym_impl_item] = STATE(15),
    [sym_trait_item] = STATE(15),
    [sym_associated_type] = STATE(15),
    [sym_let_declaration] = STATE(15),
    [sym_use_declaration] = STATE(15),
    [sym_extern_modifier] = STATE(1484),
    [sym_visibility_modifier] = STATE(1339),
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1261),
    [sym_macro_invocation] = STATE(68),
    [sym_scoped_identifier] = STATE(1172),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(70),
    [sym_match_expression] = STATE(70),
    [sym_while_expression] = STATE(70),
    [sym_loop_expression] = STATE(70),
    [sym_for_expression] = STATE(70),
    [sym_const_block] = STATE(70),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2521),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(70),
    [sym_async_block] = STATE(70),
    [sym_block] = STATE(70),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [aux_sym_source_file_repeat1] = STATE(15),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(7),
    [anon_sym_SEMI] = ACTIONS(9),
    [anon_sym_macro_rules_BANG] = ACTIONS(11),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(15),
    [anon_sym_RBRACE] = ACTIONS(278),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(25),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(29),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(33),
    [anon_sym_enum] = ACTIONS(35),
    [anon_sym_fn] = ACTIONS(37),
    [anon_sym_for] = ACTIONS(39),
    [anon_sym_if] = ACTIONS(41),
    [anon_sym_impl] = ACTIONS(43),
    [anon_sym_let] = ACTIONS(45),
    [anon_sym_loop] = ACTIONS(47),
    [anon_sym_match] = ACTIONS(49),
    [anon_sym_mod] = ACTIONS(51),
    [anon_sym_pub] = ACTIONS(53),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_static] = ACTIONS(57),
    [anon_sym_struct] = ACTIONS(59),
    [anon_sym_trait] = ACTIONS(61),
    [anon_sym_type] = ACTIONS(63),
    [anon_sym_union] = ACTIONS(65),
    [anon_sym_unsafe] = ACTIONS(67),
    [anon_sym_use] = ACTIONS(69),
    [anon_sym_while] = ACTIONS(71),
    [anon_sym_POUND] = ACTIONS(73),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_extern] = ACTIONS(75),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(101),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [17] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1154),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_SEMI] = ACTIONS(282),
    [anon_sym_LPAREN] = ACTIONS(282),
    [anon_sym_RPAREN] = ACTIONS(282),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_RBRACE] = ACTIONS(282),
    [anon_sym_EQ_GT] = ACTIONS(282),
    [anon_sym_LBRACK] = ACTIONS(282),
    [anon_sym_RBRACK] = ACTIONS(282),
    [anon_sym_COLON] = ACTIONS(286),
    [anon_sym_PLUS] = ACTIONS(288),
    [anon_sym_STAR] = ACTIONS(288),
    [anon_sym_QMARK] = ACTIONS(282),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_as] = ACTIONS(288),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(308),
    [anon_sym_EQ] = ACTIONS(288),
    [anon_sym_COMMA] = ACTIONS(282),
    [anon_sym_LT] = ACTIONS(288),
    [anon_sym_GT] = ACTIONS(288),
    [anon_sym_else] = ACTIONS(288),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(288),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(282),
    [anon_sym_DOT_DOT] = ACTIONS(288),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(282),
    [anon_sym_DASH] = ACTIONS(288),
    [anon_sym_AMP_AMP] = ACTIONS(282),
    [anon_sym_PIPE_PIPE] = ACTIONS(282),
    [anon_sym_PIPE] = ACTIONS(288),
    [anon_sym_CARET] = ACTIONS(288),
    [anon_sym_EQ_EQ] = ACTIONS(282),
    [anon_sym_BANG_EQ] = ACTIONS(282),
    [anon_sym_LT_EQ] = ACTIONS(282),
    [anon_sym_GT_EQ] = ACTIONS(282),
    [anon_sym_LT_LT] = ACTIONS(288),
    [anon_sym_GT_GT] = ACTIONS(288),
    [anon_sym_SLASH] = ACTIONS(288),
    [anon_sym_PERCENT] = ACTIONS(288),
    [anon_sym_PLUS_EQ] = ACTIONS(282),
    [anon_sym_DASH_EQ] = ACTIONS(282),
    [anon_sym_STAR_EQ] = ACTIONS(282),
    [anon_sym_SLASH_EQ] = ACTIONS(282),
    [anon_sym_PERCENT_EQ] = ACTIONS(282),
    [anon_sym_AMP_EQ] = ACTIONS(282),
    [anon_sym_PIPE_EQ] = ACTIONS(282),
    [anon_sym_CARET_EQ] = ACTIONS(282),
    [anon_sym_LT_LT_EQ] = ACTIONS(282),
    [anon_sym_GT_GT_EQ] = ACTIONS(282),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [anon_sym_DOT] = ACTIONS(288),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [18] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(871),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_SEMI] = ACTIONS(310),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_RPAREN] = ACTIONS(310),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_RBRACE] = ACTIONS(310),
    [anon_sym_EQ_GT] = ACTIONS(310),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_RBRACK] = ACTIONS(310),
    [anon_sym_PLUS] = ACTIONS(312),
    [anon_sym_STAR] = ACTIONS(312),
    [anon_sym_QMARK] = ACTIONS(310),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_as] = ACTIONS(312),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(308),
    [anon_sym_EQ] = ACTIONS(312),
    [anon_sym_COMMA] = ACTIONS(310),
    [anon_sym_LT] = ACTIONS(312),
    [anon_sym_GT] = ACTIONS(312),
    [anon_sym_else] = ACTIONS(312),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(312),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(310),
    [anon_sym_DOT_DOT] = ACTIONS(312),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(310),
    [anon_sym_DASH] = ACTIONS(312),
    [anon_sym_AMP_AMP] = ACTIONS(310),
    [anon_sym_PIPE_PIPE] = ACTIONS(310),
    [anon_sym_PIPE] = ACTIONS(312),
    [anon_sym_CARET] = ACTIONS(312),
    [anon_sym_EQ_EQ] = ACTIONS(310),
    [anon_sym_BANG_EQ] = ACTIONS(310),
    [anon_sym_LT_EQ] = ACTIONS(310),
    [anon_sym_GT_EQ] = ACTIONS(310),
    [anon_sym_LT_LT] = ACTIONS(312),
    [anon_sym_GT_GT] = ACTIONS(312),
    [anon_sym_SLASH] = ACTIONS(312),
    [anon_sym_PERCENT] = ACTIONS(312),
    [anon_sym_PLUS_EQ] = ACTIONS(310),
    [anon_sym_DASH_EQ] = ACTIONS(310),
    [anon_sym_STAR_EQ] = ACTIONS(310),
    [anon_sym_SLASH_EQ] = ACTIONS(310),
    [anon_sym_PERCENT_EQ] = ACTIONS(310),
    [anon_sym_AMP_EQ] = ACTIONS(310),
    [anon_sym_PIPE_EQ] = ACTIONS(310),
    [anon_sym_CARET_EQ] = ACTIONS(310),
    [anon_sym_LT_LT_EQ] = ACTIONS(310),
    [anon_sym_GT_GT_EQ] = ACTIONS(310),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [anon_sym_DOT] = ACTIONS(312),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [19] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1111),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_SEMI] = ACTIONS(314),
    [anon_sym_LPAREN] = ACTIONS(314),
    [anon_sym_RPAREN] = ACTIONS(314),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_RBRACE] = ACTIONS(314),
    [anon_sym_EQ_GT] = ACTIONS(314),
    [anon_sym_LBRACK] = ACTIONS(314),
    [anon_sym_RBRACK] = ACTIONS(314),
    [anon_sym_PLUS] = ACTIONS(316),
    [anon_sym_STAR] = ACTIONS(316),
    [anon_sym_QMARK] = ACTIONS(314),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_as] = ACTIONS(316),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(308),
    [anon_sym_EQ] = ACTIONS(316),
    [anon_sym_COMMA] = ACTIONS(314),
    [anon_sym_LT] = ACTIONS(316),
    [anon_sym_GT] = ACTIONS(316),
    [anon_sym_else] = ACTIONS(316),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(316),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(314),
    [anon_sym_DOT_DOT] = ACTIONS(316),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(314),
    [anon_sym_DASH] = ACTIONS(316),
    [anon_sym_AMP_AMP] = ACTIONS(314),
    [anon_sym_PIPE_PIPE] = ACTIONS(314),
    [anon_sym_PIPE] = ACTIONS(316),
    [anon_sym_CARET] = ACTIONS(316),
    [anon_sym_EQ_EQ] = ACTIONS(314),
    [anon_sym_BANG_EQ] = ACTIONS(314),
    [anon_sym_LT_EQ] = ACTIONS(314),
    [anon_sym_GT_EQ] = ACTIONS(314),
    [anon_sym_LT_LT] = ACTIONS(316),
    [anon_sym_GT_GT] = ACTIONS(316),
    [anon_sym_SLASH] = ACTIONS(316),
    [anon_sym_PERCENT] = ACTIONS(316),
    [anon_sym_PLUS_EQ] = ACTIONS(314),
    [anon_sym_DASH_EQ] = ACTIONS(314),
    [anon_sym_STAR_EQ] = ACTIONS(314),
    [anon_sym_SLASH_EQ] = ACTIONS(314),
    [anon_sym_PERCENT_EQ] = ACTIONS(314),
    [anon_sym_AMP_EQ] = ACTIONS(314),
    [anon_sym_PIPE_EQ] = ACTIONS(314),
    [anon_sym_CARET_EQ] = ACTIONS(314),
    [anon_sym_LT_LT_EQ] = ACTIONS(314),
    [anon_sym_GT_GT_EQ] = ACTIONS(314),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [anon_sym_DOT] = ACTIONS(316),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [20] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1111),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_SEMI] = ACTIONS(314),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_RPAREN] = ACTIONS(314),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_RBRACE] = ACTIONS(314),
    [anon_sym_EQ_GT] = ACTIONS(314),
    [anon_sym_LBRACK] = ACTIONS(314),
    [anon_sym_RBRACK] = ACTIONS(314),
    [anon_sym_PLUS] = ACTIONS(316),
    [anon_sym_STAR] = ACTIONS(316),
    [anon_sym_QMARK] = ACTIONS(314),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_as] = ACTIONS(316),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(308),
    [anon_sym_EQ] = ACTIONS(316),
    [anon_sym_COMMA] = ACTIONS(314),
    [anon_sym_LT] = ACTIONS(316),
    [anon_sym_GT] = ACTIONS(316),
    [anon_sym_else] = ACTIONS(316),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(316),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(314),
    [anon_sym_DOT_DOT] = ACTIONS(316),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(314),
    [anon_sym_DASH] = ACTIONS(316),
    [anon_sym_AMP_AMP] = ACTIONS(314),
    [anon_sym_PIPE_PIPE] = ACTIONS(314),
    [anon_sym_PIPE] = ACTIONS(316),
    [anon_sym_CARET] = ACTIONS(316),
    [anon_sym_EQ_EQ] = ACTIONS(314),
    [anon_sym_BANG_EQ] = ACTIONS(314),
    [anon_sym_LT_EQ] = ACTIONS(314),
    [anon_sym_GT_EQ] = ACTIONS(314),
    [anon_sym_LT_LT] = ACTIONS(316),
    [anon_sym_GT_GT] = ACTIONS(316),
    [anon_sym_SLASH] = ACTIONS(316),
    [anon_sym_PERCENT] = ACTIONS(316),
    [anon_sym_PLUS_EQ] = ACTIONS(314),
    [anon_sym_DASH_EQ] = ACTIONS(314),
    [anon_sym_STAR_EQ] = ACTIONS(314),
    [anon_sym_SLASH_EQ] = ACTIONS(314),
    [anon_sym_PERCENT_EQ] = ACTIONS(314),
    [anon_sym_AMP_EQ] = ACTIONS(314),
    [anon_sym_PIPE_EQ] = ACTIONS(314),
    [anon_sym_CARET_EQ] = ACTIONS(314),
    [anon_sym_LT_LT_EQ] = ACTIONS(314),
    [anon_sym_GT_GT_EQ] = ACTIONS(314),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [anon_sym_DOT] = ACTIONS(316),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [21] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(871),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_SEMI] = ACTIONS(310),
    [anon_sym_LPAREN] = ACTIONS(310),
    [anon_sym_RPAREN] = ACTIONS(310),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_RBRACE] = ACTIONS(310),
    [anon_sym_EQ_GT] = ACTIONS(310),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_RBRACK] = ACTIONS(310),
    [anon_sym_PLUS] = ACTIONS(312),
    [anon_sym_STAR] = ACTIONS(312),
    [anon_sym_QMARK] = ACTIONS(310),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_as] = ACTIONS(312),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(308),
    [anon_sym_EQ] = ACTIONS(312),
    [anon_sym_COMMA] = ACTIONS(310),
    [anon_sym_LT] = ACTIONS(312),
    [anon_sym_GT] = ACTIONS(312),
    [anon_sym_else] = ACTIONS(312),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(312),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(310),
    [anon_sym_DOT_DOT] = ACTIONS(312),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(310),
    [anon_sym_DASH] = ACTIONS(312),
    [anon_sym_AMP_AMP] = ACTIONS(310),
    [anon_sym_PIPE_PIPE] = ACTIONS(310),
    [anon_sym_PIPE] = ACTIONS(312),
    [anon_sym_CARET] = ACTIONS(312),
    [anon_sym_EQ_EQ] = ACTIONS(310),
    [anon_sym_BANG_EQ] = ACTIONS(310),
    [anon_sym_LT_EQ] = ACTIONS(310),
    [anon_sym_GT_EQ] = ACTIONS(310),
    [anon_sym_LT_LT] = ACTIONS(312),
    [anon_sym_GT_GT] = ACTIONS(312),
    [anon_sym_SLASH] = ACTIONS(312),
    [anon_sym_PERCENT] = ACTIONS(312),
    [anon_sym_PLUS_EQ] = ACTIONS(310),
    [anon_sym_DASH_EQ] = ACTIONS(310),
    [anon_sym_STAR_EQ] = ACTIONS(310),
    [anon_sym_SLASH_EQ] = ACTIONS(310),
    [anon_sym_PERCENT_EQ] = ACTIONS(310),
    [anon_sym_AMP_EQ] = ACTIONS(310),
    [anon_sym_PIPE_EQ] = ACTIONS(310),
    [anon_sym_CARET_EQ] = ACTIONS(310),
    [anon_sym_LT_LT_EQ] = ACTIONS(310),
    [anon_sym_GT_GT_EQ] = ACTIONS(310),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [anon_sym_DOT] = ACTIONS(312),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [22] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1138),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(17),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_SEMI] = ACTIONS(318),
    [anon_sym_LPAREN] = ACTIONS(318),
    [anon_sym_RPAREN] = ACTIONS(318),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_RBRACE] = ACTIONS(318),
    [anon_sym_EQ_GT] = ACTIONS(318),
    [anon_sym_LBRACK] = ACTIONS(318),
    [anon_sym_RBRACK] = ACTIONS(318),
    [anon_sym_PLUS] = ACTIONS(320),
    [anon_sym_STAR] = ACTIONS(320),
    [anon_sym_QMARK] = ACTIONS(318),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_as] = ACTIONS(320),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(308),
    [anon_sym_EQ] = ACTIONS(320),
    [anon_sym_COMMA] = ACTIONS(318),
    [anon_sym_LT] = ACTIONS(320),
    [anon_sym_GT] = ACTIONS(320),
    [anon_sym_else] = ACTIONS(320),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(320),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(318),
    [anon_sym_DOT_DOT] = ACTIONS(320),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(318),
    [anon_sym_DASH] = ACTIONS(320),
    [anon_sym_AMP_AMP] = ACTIONS(318),
    [anon_sym_PIPE_PIPE] = ACTIONS(318),
    [anon_sym_PIPE] = ACTIONS(320),
    [anon_sym_CARET] = ACTIONS(320),
    [anon_sym_EQ_EQ] = ACTIONS(318),
    [anon_sym_BANG_EQ] = ACTIONS(318),
    [anon_sym_LT_EQ] = ACTIONS(318),
    [anon_sym_GT_EQ] = ACTIONS(318),
    [anon_sym_LT_LT] = ACTIONS(320),
    [anon_sym_GT_GT] = ACTIONS(320),
    [anon_sym_SLASH] = ACTIONS(320),
    [anon_sym_PERCENT] = ACTIONS(320),
    [anon_sym_PLUS_EQ] = ACTIONS(318),
    [anon_sym_DASH_EQ] = ACTIONS(318),
    [anon_sym_STAR_EQ] = ACTIONS(318),
    [anon_sym_SLASH_EQ] = ACTIONS(318),
    [anon_sym_PERCENT_EQ] = ACTIONS(318),
    [anon_sym_AMP_EQ] = ACTIONS(318),
    [anon_sym_PIPE_EQ] = ACTIONS(318),
    [anon_sym_CARET_EQ] = ACTIONS(318),
    [anon_sym_LT_LT_EQ] = ACTIONS(318),
    [anon_sym_GT_GT_EQ] = ACTIONS(318),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [anon_sym_DOT] = ACTIONS(320),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [23] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1148),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_SEMI] = ACTIONS(324),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_RPAREN] = ACTIONS(324),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_RBRACE] = ACTIONS(324),
    [anon_sym_EQ_GT] = ACTIONS(324),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_RBRACK] = ACTIONS(324),
    [anon_sym_PLUS] = ACTIONS(326),
    [anon_sym_STAR] = ACTIONS(308),
    [anon_sym_QMARK] = ACTIONS(324),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_as] = ACTIONS(326),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(308),
    [anon_sym_EQ] = ACTIONS(326),
    [anon_sym_COMMA] = ACTIONS(324),
    [anon_sym_LT] = ACTIONS(328),
    [anon_sym_GT] = ACTIONS(326),
    [anon_sym_else] = ACTIONS(326),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(330),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(324),
    [anon_sym_DOT_DOT] = ACTIONS(332),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(324),
    [anon_sym_DASH] = ACTIONS(308),
    [anon_sym_AMP_AMP] = ACTIONS(324),
    [anon_sym_PIPE_PIPE] = ACTIONS(324),
    [anon_sym_PIPE] = ACTIONS(334),
    [anon_sym_CARET] = ACTIONS(326),
    [anon_sym_EQ_EQ] = ACTIONS(324),
    [anon_sym_BANG_EQ] = ACTIONS(324),
    [anon_sym_LT_EQ] = ACTIONS(324),
    [anon_sym_GT_EQ] = ACTIONS(324),
    [anon_sym_LT_LT] = ACTIONS(326),
    [anon_sym_GT_GT] = ACTIONS(326),
    [anon_sym_SLASH] = ACTIONS(326),
    [anon_sym_PERCENT] = ACTIONS(326),
    [anon_sym_PLUS_EQ] = ACTIONS(324),
    [anon_sym_DASH_EQ] = ACTIONS(324),
    [anon_sym_STAR_EQ] = ACTIONS(324),
    [anon_sym_SLASH_EQ] = ACTIONS(324),
    [anon_sym_PERCENT_EQ] = ACTIONS(324),
    [anon_sym_AMP_EQ] = ACTIONS(324),
    [anon_sym_PIPE_EQ] = ACTIONS(324),
    [anon_sym_CARET_EQ] = ACTIONS(324),
    [anon_sym_LT_LT_EQ] = ACTIONS(324),
    [anon_sym_GT_GT_EQ] = ACTIONS(324),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [anon_sym_DOT] = ACTIONS(326),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [24] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1132),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_SEMI] = ACTIONS(336),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_RPAREN] = ACTIONS(336),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_RBRACE] = ACTIONS(336),
    [anon_sym_EQ_GT] = ACTIONS(336),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_RBRACK] = ACTIONS(336),
    [anon_sym_PLUS] = ACTIONS(338),
    [anon_sym_STAR] = ACTIONS(308),
    [anon_sym_QMARK] = ACTIONS(336),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_as] = ACTIONS(338),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(308),
    [anon_sym_EQ] = ACTIONS(338),
    [anon_sym_COMMA] = ACTIONS(336),
    [anon_sym_LT] = ACTIONS(328),
    [anon_sym_GT] = ACTIONS(338),
    [anon_sym_else] = ACTIONS(338),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(330),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(336),
    [anon_sym_DOT_DOT] = ACTIONS(332),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(336),
    [anon_sym_DASH] = ACTIONS(308),
    [anon_sym_AMP_AMP] = ACTIONS(336),
    [anon_sym_PIPE_PIPE] = ACTIONS(336),
    [anon_sym_PIPE] = ACTIONS(334),
    [anon_sym_CARET] = ACTIONS(338),
    [anon_sym_EQ_EQ] = ACTIONS(336),
    [anon_sym_BANG_EQ] = ACTIONS(336),
    [anon_sym_LT_EQ] = ACTIONS(336),
    [anon_sym_GT_EQ] = ACTIONS(336),
    [anon_sym_LT_LT] = ACTIONS(338),
    [anon_sym_GT_GT] = ACTIONS(338),
    [anon_sym_SLASH] = ACTIONS(338),
    [anon_sym_PERCENT] = ACTIONS(338),
    [anon_sym_PLUS_EQ] = ACTIONS(336),
    [anon_sym_DASH_EQ] = ACTIONS(336),
    [anon_sym_STAR_EQ] = ACTIONS(336),
    [anon_sym_SLASH_EQ] = ACTIONS(336),
    [anon_sym_PERCENT_EQ] = ACTIONS(336),
    [anon_sym_AMP_EQ] = ACTIONS(336),
    [anon_sym_PIPE_EQ] = ACTIONS(336),
    [anon_sym_CARET_EQ] = ACTIONS(336),
    [anon_sym_LT_LT_EQ] = ACTIONS(336),
    [anon_sym_GT_GT_EQ] = ACTIONS(336),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [anon_sym_DOT] = ACTIONS(338),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [25] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1942),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1205),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(1176),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(97),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(340),
    [anon_sym_LPAREN] = ACTIONS(282),
    [anon_sym_LBRACE] = ACTIONS(282),
    [anon_sym_LBRACK] = ACTIONS(282),
    [anon_sym_COLON] = ACTIONS(286),
    [anon_sym_PLUS] = ACTIONS(288),
    [anon_sym_STAR] = ACTIONS(288),
    [anon_sym_QMARK] = ACTIONS(282),
    [anon_sym_u8] = ACTIONS(342),
    [anon_sym_i8] = ACTIONS(342),
    [anon_sym_u16] = ACTIONS(342),
    [anon_sym_i16] = ACTIONS(342),
    [anon_sym_u32] = ACTIONS(342),
    [anon_sym_i32] = ACTIONS(342),
    [anon_sym_u64] = ACTIONS(342),
    [anon_sym_i64] = ACTIONS(342),
    [anon_sym_u128] = ACTIONS(342),
    [anon_sym_i128] = ACTIONS(342),
    [anon_sym_isize] = ACTIONS(342),
    [anon_sym_usize] = ACTIONS(342),
    [anon_sym_f32] = ACTIONS(342),
    [anon_sym_f64] = ACTIONS(342),
    [anon_sym_bool] = ACTIONS(342),
    [anon_sym_str] = ACTIONS(342),
    [anon_sym_char] = ACTIONS(342),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_as] = ACTIONS(288),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(344),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(346),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(348),
    [anon_sym_union] = ACTIONS(346),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(350),
    [anon_sym_EQ] = ACTIONS(288),
    [anon_sym_LT] = ACTIONS(288),
    [anon_sym_GT] = ACTIONS(288),
    [anon_sym_COLON_COLON] = ACTIONS(352),
    [anon_sym_AMP] = ACTIONS(288),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(282),
    [anon_sym_DOT_DOT] = ACTIONS(288),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(282),
    [anon_sym_DASH] = ACTIONS(288),
    [anon_sym_AMP_AMP] = ACTIONS(282),
    [anon_sym_PIPE_PIPE] = ACTIONS(282),
    [anon_sym_PIPE] = ACTIONS(288),
    [anon_sym_CARET] = ACTIONS(288),
    [anon_sym_EQ_EQ] = ACTIONS(282),
    [anon_sym_BANG_EQ] = ACTIONS(282),
    [anon_sym_LT_EQ] = ACTIONS(282),
    [anon_sym_GT_EQ] = ACTIONS(282),
    [anon_sym_LT_LT] = ACTIONS(288),
    [anon_sym_GT_GT] = ACTIONS(288),
    [anon_sym_SLASH] = ACTIONS(288),
    [anon_sym_PERCENT] = ACTIONS(288),
    [anon_sym_PLUS_EQ] = ACTIONS(282),
    [anon_sym_DASH_EQ] = ACTIONS(282),
    [anon_sym_STAR_EQ] = ACTIONS(282),
    [anon_sym_SLASH_EQ] = ACTIONS(282),
    [anon_sym_PERCENT_EQ] = ACTIONS(282),
    [anon_sym_AMP_EQ] = ACTIONS(282),
    [anon_sym_PIPE_EQ] = ACTIONS(282),
    [anon_sym_CARET_EQ] = ACTIONS(282),
    [anon_sym_LT_LT_EQ] = ACTIONS(282),
    [anon_sym_GT_GT_EQ] = ACTIONS(282),
    [anon_sym_yield] = ACTIONS(354),
    [anon_sym_move] = ACTIONS(356),
    [anon_sym_DOT] = ACTIONS(288),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(358),
    [sym_super] = ACTIONS(360),
    [sym_crate] = ACTIONS(360),
    [sym_metavariable] = ACTIONS(362),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [26] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1942),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(871),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(1176),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(97),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(340),
    [anon_sym_LPAREN] = ACTIONS(310),
    [anon_sym_LBRACE] = ACTIONS(310),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_PLUS] = ACTIONS(312),
    [anon_sym_STAR] = ACTIONS(312),
    [anon_sym_QMARK] = ACTIONS(310),
    [anon_sym_u8] = ACTIONS(342),
    [anon_sym_i8] = ACTIONS(342),
    [anon_sym_u16] = ACTIONS(342),
    [anon_sym_i16] = ACTIONS(342),
    [anon_sym_u32] = ACTIONS(342),
    [anon_sym_i32] = ACTIONS(342),
    [anon_sym_u64] = ACTIONS(342),
    [anon_sym_i64] = ACTIONS(342),
    [anon_sym_u128] = ACTIONS(342),
    [anon_sym_i128] = ACTIONS(342),
    [anon_sym_isize] = ACTIONS(342),
    [anon_sym_usize] = ACTIONS(342),
    [anon_sym_f32] = ACTIONS(342),
    [anon_sym_f64] = ACTIONS(342),
    [anon_sym_bool] = ACTIONS(342),
    [anon_sym_str] = ACTIONS(342),
    [anon_sym_char] = ACTIONS(342),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_as] = ACTIONS(312),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(344),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(346),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(348),
    [anon_sym_union] = ACTIONS(346),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(350),
    [anon_sym_EQ] = ACTIONS(312),
    [anon_sym_LT] = ACTIONS(312),
    [anon_sym_GT] = ACTIONS(312),
    [anon_sym_COLON_COLON] = ACTIONS(352),
    [anon_sym_AMP] = ACTIONS(312),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(310),
    [anon_sym_DOT_DOT] = ACTIONS(312),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(310),
    [anon_sym_DASH] = ACTIONS(312),
    [anon_sym_AMP_AMP] = ACTIONS(310),
    [anon_sym_PIPE_PIPE] = ACTIONS(310),
    [anon_sym_PIPE] = ACTIONS(312),
    [anon_sym_CARET] = ACTIONS(312),
    [anon_sym_EQ_EQ] = ACTIONS(310),
    [anon_sym_BANG_EQ] = ACTIONS(310),
    [anon_sym_LT_EQ] = ACTIONS(310),
    [anon_sym_GT_EQ] = ACTIONS(310),
    [anon_sym_LT_LT] = ACTIONS(312),
    [anon_sym_GT_GT] = ACTIONS(312),
    [anon_sym_SLASH] = ACTIONS(312),
    [anon_sym_PERCENT] = ACTIONS(312),
    [anon_sym_PLUS_EQ] = ACTIONS(310),
    [anon_sym_DASH_EQ] = ACTIONS(310),
    [anon_sym_STAR_EQ] = ACTIONS(310),
    [anon_sym_SLASH_EQ] = ACTIONS(310),
    [anon_sym_PERCENT_EQ] = ACTIONS(310),
    [anon_sym_AMP_EQ] = ACTIONS(310),
    [anon_sym_PIPE_EQ] = ACTIONS(310),
    [anon_sym_CARET_EQ] = ACTIONS(310),
    [anon_sym_LT_LT_EQ] = ACTIONS(310),
    [anon_sym_GT_GT_EQ] = ACTIONS(310),
    [anon_sym_yield] = ACTIONS(354),
    [anon_sym_move] = ACTIONS(356),
    [anon_sym_DOT] = ACTIONS(312),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(358),
    [sym_super] = ACTIONS(360),
    [sym_crate] = ACTIONS(360),
    [sym_metavariable] = ACTIONS(362),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [27] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1942),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1221),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(1176),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(97),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(340),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_PLUS] = ACTIONS(326),
    [anon_sym_STAR] = ACTIONS(350),
    [anon_sym_QMARK] = ACTIONS(324),
    [anon_sym_u8] = ACTIONS(342),
    [anon_sym_i8] = ACTIONS(342),
    [anon_sym_u16] = ACTIONS(342),
    [anon_sym_i16] = ACTIONS(342),
    [anon_sym_u32] = ACTIONS(342),
    [anon_sym_i32] = ACTIONS(342),
    [anon_sym_u64] = ACTIONS(342),
    [anon_sym_i64] = ACTIONS(342),
    [anon_sym_u128] = ACTIONS(342),
    [anon_sym_i128] = ACTIONS(342),
    [anon_sym_isize] = ACTIONS(342),
    [anon_sym_usize] = ACTIONS(342),
    [anon_sym_f32] = ACTIONS(342),
    [anon_sym_f64] = ACTIONS(342),
    [anon_sym_bool] = ACTIONS(342),
    [anon_sym_str] = ACTIONS(342),
    [anon_sym_char] = ACTIONS(342),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_as] = ACTIONS(326),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(344),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(346),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(348),
    [anon_sym_union] = ACTIONS(346),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(350),
    [anon_sym_EQ] = ACTIONS(326),
    [anon_sym_LT] = ACTIONS(328),
    [anon_sym_GT] = ACTIONS(326),
    [anon_sym_COLON_COLON] = ACTIONS(352),
    [anon_sym_AMP] = ACTIONS(364),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(324),
    [anon_sym_DOT_DOT] = ACTIONS(366),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(324),
    [anon_sym_DASH] = ACTIONS(350),
    [anon_sym_AMP_AMP] = ACTIONS(324),
    [anon_sym_PIPE_PIPE] = ACTIONS(324),
    [anon_sym_PIPE] = ACTIONS(334),
    [anon_sym_CARET] = ACTIONS(326),
    [anon_sym_EQ_EQ] = ACTIONS(324),
    [anon_sym_BANG_EQ] = ACTIONS(324),
    [anon_sym_LT_EQ] = ACTIONS(324),
    [anon_sym_GT_EQ] = ACTIONS(324),
    [anon_sym_LT_LT] = ACTIONS(326),
    [anon_sym_GT_GT] = ACTIONS(326),
    [anon_sym_SLASH] = ACTIONS(326),
    [anon_sym_PERCENT] = ACTIONS(326),
    [anon_sym_PLUS_EQ] = ACTIONS(324),
    [anon_sym_DASH_EQ] = ACTIONS(324),
    [anon_sym_STAR_EQ] = ACTIONS(324),
    [anon_sym_SLASH_EQ] = ACTIONS(324),
    [anon_sym_PERCENT_EQ] = ACTIONS(324),
    [anon_sym_AMP_EQ] = ACTIONS(324),
    [anon_sym_PIPE_EQ] = ACTIONS(324),
    [anon_sym_CARET_EQ] = ACTIONS(324),
    [anon_sym_LT_LT_EQ] = ACTIONS(324),
    [anon_sym_GT_GT_EQ] = ACTIONS(324),
    [anon_sym_yield] = ACTIONS(354),
    [anon_sym_move] = ACTIONS(356),
    [anon_sym_DOT] = ACTIONS(326),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(358),
    [sym_super] = ACTIONS(360),
    [sym_crate] = ACTIONS(360),
    [sym_metavariable] = ACTIONS(362),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [28] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1942),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1111),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(1176),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(97),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(340),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(314),
    [anon_sym_LBRACK] = ACTIONS(314),
    [anon_sym_PLUS] = ACTIONS(316),
    [anon_sym_STAR] = ACTIONS(316),
    [anon_sym_QMARK] = ACTIONS(314),
    [anon_sym_u8] = ACTIONS(342),
    [anon_sym_i8] = ACTIONS(342),
    [anon_sym_u16] = ACTIONS(342),
    [anon_sym_i16] = ACTIONS(342),
    [anon_sym_u32] = ACTIONS(342),
    [anon_sym_i32] = ACTIONS(342),
    [anon_sym_u64] = ACTIONS(342),
    [anon_sym_i64] = ACTIONS(342),
    [anon_sym_u128] = ACTIONS(342),
    [anon_sym_i128] = ACTIONS(342),
    [anon_sym_isize] = ACTIONS(342),
    [anon_sym_usize] = ACTIONS(342),
    [anon_sym_f32] = ACTIONS(342),
    [anon_sym_f64] = ACTIONS(342),
    [anon_sym_bool] = ACTIONS(342),
    [anon_sym_str] = ACTIONS(342),
    [anon_sym_char] = ACTIONS(342),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_as] = ACTIONS(316),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(344),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(346),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(348),
    [anon_sym_union] = ACTIONS(346),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(350),
    [anon_sym_EQ] = ACTIONS(316),
    [anon_sym_LT] = ACTIONS(316),
    [anon_sym_GT] = ACTIONS(316),
    [anon_sym_COLON_COLON] = ACTIONS(352),
    [anon_sym_AMP] = ACTIONS(316),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(314),
    [anon_sym_DOT_DOT] = ACTIONS(316),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(314),
    [anon_sym_DASH] = ACTIONS(316),
    [anon_sym_AMP_AMP] = ACTIONS(314),
    [anon_sym_PIPE_PIPE] = ACTIONS(314),
    [anon_sym_PIPE] = ACTIONS(316),
    [anon_sym_CARET] = ACTIONS(316),
    [anon_sym_EQ_EQ] = ACTIONS(314),
    [anon_sym_BANG_EQ] = ACTIONS(314),
    [anon_sym_LT_EQ] = ACTIONS(314),
    [anon_sym_GT_EQ] = ACTIONS(314),
    [anon_sym_LT_LT] = ACTIONS(316),
    [anon_sym_GT_GT] = ACTIONS(316),
    [anon_sym_SLASH] = ACTIONS(316),
    [anon_sym_PERCENT] = ACTIONS(316),
    [anon_sym_PLUS_EQ] = ACTIONS(314),
    [anon_sym_DASH_EQ] = ACTIONS(314),
    [anon_sym_STAR_EQ] = ACTIONS(314),
    [anon_sym_SLASH_EQ] = ACTIONS(314),
    [anon_sym_PERCENT_EQ] = ACTIONS(314),
    [anon_sym_AMP_EQ] = ACTIONS(314),
    [anon_sym_PIPE_EQ] = ACTIONS(314),
    [anon_sym_CARET_EQ] = ACTIONS(314),
    [anon_sym_LT_LT_EQ] = ACTIONS(314),
    [anon_sym_GT_GT_EQ] = ACTIONS(314),
    [anon_sym_yield] = ACTIONS(354),
    [anon_sym_move] = ACTIONS(356),
    [anon_sym_DOT] = ACTIONS(316),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(358),
    [sym_super] = ACTIONS(360),
    [sym_crate] = ACTIONS(360),
    [sym_metavariable] = ACTIONS(362),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [29] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1942),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1263),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(1176),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(97),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(340),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_PLUS] = ACTIONS(338),
    [anon_sym_STAR] = ACTIONS(350),
    [anon_sym_QMARK] = ACTIONS(336),
    [anon_sym_u8] = ACTIONS(342),
    [anon_sym_i8] = ACTIONS(342),
    [anon_sym_u16] = ACTIONS(342),
    [anon_sym_i16] = ACTIONS(342),
    [anon_sym_u32] = ACTIONS(342),
    [anon_sym_i32] = ACTIONS(342),
    [anon_sym_u64] = ACTIONS(342),
    [anon_sym_i64] = ACTIONS(342),
    [anon_sym_u128] = ACTIONS(342),
    [anon_sym_i128] = ACTIONS(342),
    [anon_sym_isize] = ACTIONS(342),
    [anon_sym_usize] = ACTIONS(342),
    [anon_sym_f32] = ACTIONS(342),
    [anon_sym_f64] = ACTIONS(342),
    [anon_sym_bool] = ACTIONS(342),
    [anon_sym_str] = ACTIONS(342),
    [anon_sym_char] = ACTIONS(342),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_as] = ACTIONS(338),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(344),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(346),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(348),
    [anon_sym_union] = ACTIONS(346),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(350),
    [anon_sym_EQ] = ACTIONS(338),
    [anon_sym_LT] = ACTIONS(328),
    [anon_sym_GT] = ACTIONS(338),
    [anon_sym_COLON_COLON] = ACTIONS(352),
    [anon_sym_AMP] = ACTIONS(364),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(336),
    [anon_sym_DOT_DOT] = ACTIONS(366),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(336),
    [anon_sym_DASH] = ACTIONS(350),
    [anon_sym_AMP_AMP] = ACTIONS(336),
    [anon_sym_PIPE_PIPE] = ACTIONS(336),
    [anon_sym_PIPE] = ACTIONS(334),
    [anon_sym_CARET] = ACTIONS(338),
    [anon_sym_EQ_EQ] = ACTIONS(336),
    [anon_sym_BANG_EQ] = ACTIONS(336),
    [anon_sym_LT_EQ] = ACTIONS(336),
    [anon_sym_GT_EQ] = ACTIONS(336),
    [anon_sym_LT_LT] = ACTIONS(338),
    [anon_sym_GT_GT] = ACTIONS(338),
    [anon_sym_SLASH] = ACTIONS(338),
    [anon_sym_PERCENT] = ACTIONS(338),
    [anon_sym_PLUS_EQ] = ACTIONS(336),
    [anon_sym_DASH_EQ] = ACTIONS(336),
    [anon_sym_STAR_EQ] = ACTIONS(336),
    [anon_sym_SLASH_EQ] = ACTIONS(336),
    [anon_sym_PERCENT_EQ] = ACTIONS(336),
    [anon_sym_AMP_EQ] = ACTIONS(336),
    [anon_sym_PIPE_EQ] = ACTIONS(336),
    [anon_sym_CARET_EQ] = ACTIONS(336),
    [anon_sym_LT_LT_EQ] = ACTIONS(336),
    [anon_sym_GT_GT_EQ] = ACTIONS(336),
    [anon_sym_yield] = ACTIONS(354),
    [anon_sym_move] = ACTIONS(356),
    [anon_sym_DOT] = ACTIONS(338),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(358),
    [sym_super] = ACTIONS(360),
    [sym_crate] = ACTIONS(360),
    [sym_metavariable] = ACTIONS(362),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [30] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1942),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1238),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(1176),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(97),
    [sym_loop_label] = STATE(25),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(340),
    [anon_sym_LPAREN] = ACTIONS(318),
    [anon_sym_LBRACE] = ACTIONS(318),
    [anon_sym_LBRACK] = ACTIONS(318),
    [anon_sym_PLUS] = ACTIONS(320),
    [anon_sym_STAR] = ACTIONS(320),
    [anon_sym_QMARK] = ACTIONS(318),
    [anon_sym_u8] = ACTIONS(342),
    [anon_sym_i8] = ACTIONS(342),
    [anon_sym_u16] = ACTIONS(342),
    [anon_sym_i16] = ACTIONS(342),
    [anon_sym_u32] = ACTIONS(342),
    [anon_sym_i32] = ACTIONS(342),
    [anon_sym_u64] = ACTIONS(342),
    [anon_sym_i64] = ACTIONS(342),
    [anon_sym_u128] = ACTIONS(342),
    [anon_sym_i128] = ACTIONS(342),
    [anon_sym_isize] = ACTIONS(342),
    [anon_sym_usize] = ACTIONS(342),
    [anon_sym_f32] = ACTIONS(342),
    [anon_sym_f64] = ACTIONS(342),
    [anon_sym_bool] = ACTIONS(342),
    [anon_sym_str] = ACTIONS(342),
    [anon_sym_char] = ACTIONS(342),
    [anon_sym_SQUOTE] = ACTIONS(322),
    [anon_sym_as] = ACTIONS(320),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(344),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(346),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(348),
    [anon_sym_union] = ACTIONS(346),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(350),
    [anon_sym_EQ] = ACTIONS(320),
    [anon_sym_LT] = ACTIONS(320),
    [anon_sym_GT] = ACTIONS(320),
    [anon_sym_COLON_COLON] = ACTIONS(352),
    [anon_sym_AMP] = ACTIONS(320),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(318),
    [anon_sym_DOT_DOT] = ACTIONS(320),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(318),
    [anon_sym_DASH] = ACTIONS(320),
    [anon_sym_AMP_AMP] = ACTIONS(318),
    [anon_sym_PIPE_PIPE] = ACTIONS(318),
    [anon_sym_PIPE] = ACTIONS(320),
    [anon_sym_CARET] = ACTIONS(320),
    [anon_sym_EQ_EQ] = ACTIONS(318),
    [anon_sym_BANG_EQ] = ACTIONS(318),
    [anon_sym_LT_EQ] = ACTIONS(318),
    [anon_sym_GT_EQ] = ACTIONS(318),
    [anon_sym_LT_LT] = ACTIONS(320),
    [anon_sym_GT_GT] = ACTIONS(320),
    [anon_sym_SLASH] = ACTIONS(320),
    [anon_sym_PERCENT] = ACTIONS(320),
    [anon_sym_PLUS_EQ] = ACTIONS(318),
    [anon_sym_DASH_EQ] = ACTIONS(318),
    [anon_sym_STAR_EQ] = ACTIONS(318),
    [anon_sym_SLASH_EQ] = ACTIONS(318),
    [anon_sym_PERCENT_EQ] = ACTIONS(318),
    [anon_sym_AMP_EQ] = ACTIONS(318),
    [anon_sym_PIPE_EQ] = ACTIONS(318),
    [anon_sym_CARET_EQ] = ACTIONS(318),
    [anon_sym_LT_LT_EQ] = ACTIONS(318),
    [anon_sym_GT_GT_EQ] = ACTIONS(318),
    [anon_sym_yield] = ACTIONS(354),
    [anon_sym_move] = ACTIONS(356),
    [anon_sym_DOT] = ACTIONS(320),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(358),
    [sym_super] = ACTIONS(360),
    [sym_crate] = ACTIONS(360),
    [sym_metavariable] = ACTIONS(362),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [31] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1942),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(871),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(1176),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(97),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(340),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(310),
    [anon_sym_LBRACK] = ACTIONS(310),
    [anon_sym_PLUS] = ACTIONS(312),
    [anon_sym_STAR] = ACTIONS(312),
    [anon_sym_QMARK] = ACTIONS(310),
    [anon_sym_u8] = ACTIONS(342),
    [anon_sym_i8] = ACTIONS(342),
    [anon_sym_u16] = ACTIONS(342),
    [anon_sym_i16] = ACTIONS(342),
    [anon_sym_u32] = ACTIONS(342),
    [anon_sym_i32] = ACTIONS(342),
    [anon_sym_u64] = ACTIONS(342),
    [anon_sym_i64] = ACTIONS(342),
    [anon_sym_u128] = ACTIONS(342),
    [anon_sym_i128] = ACTIONS(342),
    [anon_sym_isize] = ACTIONS(342),
    [anon_sym_usize] = ACTIONS(342),
    [anon_sym_f32] = ACTIONS(342),
    [anon_sym_f64] = ACTIONS(342),
    [anon_sym_bool] = ACTIONS(342),
    [anon_sym_str] = ACTIONS(342),
    [anon_sym_char] = ACTIONS(342),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_as] = ACTIONS(312),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(344),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(346),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(348),
    [anon_sym_union] = ACTIONS(346),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(350),
    [anon_sym_EQ] = ACTIONS(312),
    [anon_sym_LT] = ACTIONS(312),
    [anon_sym_GT] = ACTIONS(312),
    [anon_sym_COLON_COLON] = ACTIONS(352),
    [anon_sym_AMP] = ACTIONS(312),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(310),
    [anon_sym_DOT_DOT] = ACTIONS(312),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(310),
    [anon_sym_DASH] = ACTIONS(312),
    [anon_sym_AMP_AMP] = ACTIONS(310),
    [anon_sym_PIPE_PIPE] = ACTIONS(310),
    [anon_sym_PIPE] = ACTIONS(312),
    [anon_sym_CARET] = ACTIONS(312),
    [anon_sym_EQ_EQ] = ACTIONS(310),
    [anon_sym_BANG_EQ] = ACTIONS(310),
    [anon_sym_LT_EQ] = ACTIONS(310),
    [anon_sym_GT_EQ] = ACTIONS(310),
    [anon_sym_LT_LT] = ACTIONS(312),
    [anon_sym_GT_GT] = ACTIONS(312),
    [anon_sym_SLASH] = ACTIONS(312),
    [anon_sym_PERCENT] = ACTIONS(312),
    [anon_sym_PLUS_EQ] = ACTIONS(310),
    [anon_sym_DASH_EQ] = ACTIONS(310),
    [anon_sym_STAR_EQ] = ACTIONS(310),
    [anon_sym_SLASH_EQ] = ACTIONS(310),
    [anon_sym_PERCENT_EQ] = ACTIONS(310),
    [anon_sym_AMP_EQ] = ACTIONS(310),
    [anon_sym_PIPE_EQ] = ACTIONS(310),
    [anon_sym_CARET_EQ] = ACTIONS(310),
    [anon_sym_LT_LT_EQ] = ACTIONS(310),
    [anon_sym_GT_GT_EQ] = ACTIONS(310),
    [anon_sym_yield] = ACTIONS(354),
    [anon_sym_move] = ACTIONS(356),
    [anon_sym_DOT] = ACTIONS(312),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(358),
    [sym_super] = ACTIONS(360),
    [sym_crate] = ACTIONS(360),
    [sym_metavariable] = ACTIONS(362),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [32] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1942),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1111),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(1176),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(97),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(340),
    [anon_sym_LPAREN] = ACTIONS(314),
    [anon_sym_LBRACE] = ACTIONS(314),
    [anon_sym_LBRACK] = ACTIONS(314),
    [anon_sym_PLUS] = ACTIONS(316),
    [anon_sym_STAR] = ACTIONS(316),
    [anon_sym_QMARK] = ACTIONS(314),
    [anon_sym_u8] = ACTIONS(342),
    [anon_sym_i8] = ACTIONS(342),
    [anon_sym_u16] = ACTIONS(342),
    [anon_sym_i16] = ACTIONS(342),
    [anon_sym_u32] = ACTIONS(342),
    [anon_sym_i32] = ACTIONS(342),
    [anon_sym_u64] = ACTIONS(342),
    [anon_sym_i64] = ACTIONS(342),
    [anon_sym_u128] = ACTIONS(342),
    [anon_sym_i128] = ACTIONS(342),
    [anon_sym_isize] = ACTIONS(342),
    [anon_sym_usize] = ACTIONS(342),
    [anon_sym_f32] = ACTIONS(342),
    [anon_sym_f64] = ACTIONS(342),
    [anon_sym_bool] = ACTIONS(342),
    [anon_sym_str] = ACTIONS(342),
    [anon_sym_char] = ACTIONS(342),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_as] = ACTIONS(316),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(344),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(346),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(348),
    [anon_sym_union] = ACTIONS(346),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(350),
    [anon_sym_EQ] = ACTIONS(316),
    [anon_sym_LT] = ACTIONS(316),
    [anon_sym_GT] = ACTIONS(316),
    [anon_sym_COLON_COLON] = ACTIONS(352),
    [anon_sym_AMP] = ACTIONS(316),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(314),
    [anon_sym_DOT_DOT] = ACTIONS(316),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(314),
    [anon_sym_DASH] = ACTIONS(316),
    [anon_sym_AMP_AMP] = ACTIONS(314),
    [anon_sym_PIPE_PIPE] = ACTIONS(314),
    [anon_sym_PIPE] = ACTIONS(316),
    [anon_sym_CARET] = ACTIONS(316),
    [anon_sym_EQ_EQ] = ACTIONS(314),
    [anon_sym_BANG_EQ] = ACTIONS(314),
    [anon_sym_LT_EQ] = ACTIONS(314),
    [anon_sym_GT_EQ] = ACTIONS(314),
    [anon_sym_LT_LT] = ACTIONS(316),
    [anon_sym_GT_GT] = ACTIONS(316),
    [anon_sym_SLASH] = ACTIONS(316),
    [anon_sym_PERCENT] = ACTIONS(316),
    [anon_sym_PLUS_EQ] = ACTIONS(314),
    [anon_sym_DASH_EQ] = ACTIONS(314),
    [anon_sym_STAR_EQ] = ACTIONS(314),
    [anon_sym_SLASH_EQ] = ACTIONS(314),
    [anon_sym_PERCENT_EQ] = ACTIONS(314),
    [anon_sym_AMP_EQ] = ACTIONS(314),
    [anon_sym_PIPE_EQ] = ACTIONS(314),
    [anon_sym_CARET_EQ] = ACTIONS(314),
    [anon_sym_LT_LT_EQ] = ACTIONS(314),
    [anon_sym_GT_GT_EQ] = ACTIONS(314),
    [anon_sym_yield] = ACTIONS(354),
    [anon_sym_move] = ACTIONS(356),
    [anon_sym_DOT] = ACTIONS(316),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(358),
    [sym_super] = ACTIONS(360),
    [sym_crate] = ACTIONS(360),
    [sym_metavariable] = ACTIONS(362),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [33] = {
    [sym_attribute_item] = STATE(55),
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1197),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [aux_sym_enum_variant_list_repeat1] = STATE(55),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_RPAREN] = ACTIONS(368),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_POUND] = ACTIONS(370),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_COMMA] = ACTIONS(372),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [34] = {
    [sym_attribute_item] = STATE(621),
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1175),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [aux_sym_enum_variant_list_repeat1] = STATE(621),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_RBRACK] = ACTIONS(374),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_POUND] = ACTIONS(370),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_COMMA] = ACTIONS(376),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [35] = {
    [sym_attribute_item] = STATE(34),
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1182),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [aux_sym_enum_variant_list_repeat1] = STATE(34),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_RBRACK] = ACTIONS(378),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_POUND] = ACTIONS(370),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_COMMA] = ACTIONS(380),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [36] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1287),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_let_condition] = STATE(2243),
    [sym__let_chain] = STATE(2107),
    [sym__condition] = STATE(2544),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_let] = ACTIONS(382),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [37] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1942),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1272),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(1176),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_let_condition] = STATE(2190),
    [sym__let_chain] = STATE(2118),
    [sym__condition] = STATE(2278),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(97),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(340),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(384),
    [anon_sym_u8] = ACTIONS(342),
    [anon_sym_i8] = ACTIONS(342),
    [anon_sym_u16] = ACTIONS(342),
    [anon_sym_i16] = ACTIONS(342),
    [anon_sym_u32] = ACTIONS(342),
    [anon_sym_i32] = ACTIONS(342),
    [anon_sym_u64] = ACTIONS(342),
    [anon_sym_i64] = ACTIONS(342),
    [anon_sym_u128] = ACTIONS(342),
    [anon_sym_i128] = ACTIONS(342),
    [anon_sym_isize] = ACTIONS(342),
    [anon_sym_usize] = ACTIONS(342),
    [anon_sym_f32] = ACTIONS(342),
    [anon_sym_f64] = ACTIONS(342),
    [anon_sym_bool] = ACTIONS(342),
    [anon_sym_str] = ACTIONS(342),
    [anon_sym_char] = ACTIONS(342),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(344),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(346),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_let] = ACTIONS(386),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(348),
    [anon_sym_union] = ACTIONS(346),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(384),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(352),
    [anon_sym_AMP] = ACTIONS(388),
    [anon_sym_DOT_DOT] = ACTIONS(390),
    [anon_sym_DASH] = ACTIONS(384),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(354),
    [anon_sym_move] = ACTIONS(356),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(358),
    [sym_super] = ACTIONS(360),
    [sym_crate] = ACTIONS(360),
    [sym_metavariable] = ACTIONS(362),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [38] = {
    [sym_attribute_item] = STATE(57),
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1223),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [aux_sym_enum_variant_list_repeat1] = STATE(57),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_RPAREN] = ACTIONS(392),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_POUND] = ACTIONS(370),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [39] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1942),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1272),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(1176),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_let_condition] = STATE(2190),
    [sym__let_chain] = STATE(2118),
    [sym__condition] = STATE(2240),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(97),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(340),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(384),
    [anon_sym_u8] = ACTIONS(342),
    [anon_sym_i8] = ACTIONS(342),
    [anon_sym_u16] = ACTIONS(342),
    [anon_sym_i16] = ACTIONS(342),
    [anon_sym_u32] = ACTIONS(342),
    [anon_sym_i32] = ACTIONS(342),
    [anon_sym_u64] = ACTIONS(342),
    [anon_sym_i64] = ACTIONS(342),
    [anon_sym_u128] = ACTIONS(342),
    [anon_sym_i128] = ACTIONS(342),
    [anon_sym_isize] = ACTIONS(342),
    [anon_sym_usize] = ACTIONS(342),
    [anon_sym_f32] = ACTIONS(342),
    [anon_sym_f64] = ACTIONS(342),
    [anon_sym_bool] = ACTIONS(342),
    [anon_sym_str] = ACTIONS(342),
    [anon_sym_char] = ACTIONS(342),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(344),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(346),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_let] = ACTIONS(386),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(348),
    [anon_sym_union] = ACTIONS(346),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(384),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(352),
    [anon_sym_AMP] = ACTIONS(388),
    [anon_sym_DOT_DOT] = ACTIONS(390),
    [anon_sym_DASH] = ACTIONS(384),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(354),
    [anon_sym_move] = ACTIONS(356),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(358),
    [sym_super] = ACTIONS(360),
    [sym_crate] = ACTIONS(360),
    [sym_metavariable] = ACTIONS(362),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [40] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1942),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1272),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(1176),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_let_condition] = STATE(2190),
    [sym__let_chain] = STATE(2118),
    [sym__condition] = STATE(2300),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(97),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(340),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(384),
    [anon_sym_u8] = ACTIONS(342),
    [anon_sym_i8] = ACTIONS(342),
    [anon_sym_u16] = ACTIONS(342),
    [anon_sym_i16] = ACTIONS(342),
    [anon_sym_u32] = ACTIONS(342),
    [anon_sym_i32] = ACTIONS(342),
    [anon_sym_u64] = ACTIONS(342),
    [anon_sym_i64] = ACTIONS(342),
    [anon_sym_u128] = ACTIONS(342),
    [anon_sym_i128] = ACTIONS(342),
    [anon_sym_isize] = ACTIONS(342),
    [anon_sym_usize] = ACTIONS(342),
    [anon_sym_f32] = ACTIONS(342),
    [anon_sym_f64] = ACTIONS(342),
    [anon_sym_bool] = ACTIONS(342),
    [anon_sym_str] = ACTIONS(342),
    [anon_sym_char] = ACTIONS(342),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(344),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(346),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_let] = ACTIONS(386),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(348),
    [anon_sym_union] = ACTIONS(346),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(384),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(352),
    [anon_sym_AMP] = ACTIONS(388),
    [anon_sym_DOT_DOT] = ACTIONS(390),
    [anon_sym_DASH] = ACTIONS(384),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(354),
    [anon_sym_move] = ACTIONS(356),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(358),
    [sym_super] = ACTIONS(360),
    [sym_crate] = ACTIONS(360),
    [sym_metavariable] = ACTIONS(362),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [41] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1942),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1272),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(1176),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_let_condition] = STATE(2190),
    [sym__let_chain] = STATE(2118),
    [sym__condition] = STATE(2217),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(97),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(340),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(384),
    [anon_sym_u8] = ACTIONS(342),
    [anon_sym_i8] = ACTIONS(342),
    [anon_sym_u16] = ACTIONS(342),
    [anon_sym_i16] = ACTIONS(342),
    [anon_sym_u32] = ACTIONS(342),
    [anon_sym_i32] = ACTIONS(342),
    [anon_sym_u64] = ACTIONS(342),
    [anon_sym_i64] = ACTIONS(342),
    [anon_sym_u128] = ACTIONS(342),
    [anon_sym_i128] = ACTIONS(342),
    [anon_sym_isize] = ACTIONS(342),
    [anon_sym_usize] = ACTIONS(342),
    [anon_sym_f32] = ACTIONS(342),
    [anon_sym_f64] = ACTIONS(342),
    [anon_sym_bool] = ACTIONS(342),
    [anon_sym_str] = ACTIONS(342),
    [anon_sym_char] = ACTIONS(342),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(344),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(346),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_let] = ACTIONS(386),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(348),
    [anon_sym_union] = ACTIONS(346),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(384),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(352),
    [anon_sym_AMP] = ACTIONS(388),
    [anon_sym_DOT_DOT] = ACTIONS(390),
    [anon_sym_DASH] = ACTIONS(384),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(354),
    [anon_sym_move] = ACTIONS(356),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(358),
    [sym_super] = ACTIONS(360),
    [sym_crate] = ACTIONS(360),
    [sym_metavariable] = ACTIONS(362),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [42] = {
    [sym_attribute_item] = STATE(51),
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1226),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [aux_sym_enum_variant_list_repeat1] = STATE(51),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_RPAREN] = ACTIONS(394),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_POUND] = ACTIONS(370),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [43] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1942),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1272),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(1176),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_let_condition] = STATE(2190),
    [sym__let_chain] = STATE(2118),
    [sym__condition] = STATE(2183),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(97),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(340),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(384),
    [anon_sym_u8] = ACTIONS(342),
    [anon_sym_i8] = ACTIONS(342),
    [anon_sym_u16] = ACTIONS(342),
    [anon_sym_i16] = ACTIONS(342),
    [anon_sym_u32] = ACTIONS(342),
    [anon_sym_i32] = ACTIONS(342),
    [anon_sym_u64] = ACTIONS(342),
    [anon_sym_i64] = ACTIONS(342),
    [anon_sym_u128] = ACTIONS(342),
    [anon_sym_i128] = ACTIONS(342),
    [anon_sym_isize] = ACTIONS(342),
    [anon_sym_usize] = ACTIONS(342),
    [anon_sym_f32] = ACTIONS(342),
    [anon_sym_f64] = ACTIONS(342),
    [anon_sym_bool] = ACTIONS(342),
    [anon_sym_str] = ACTIONS(342),
    [anon_sym_char] = ACTIONS(342),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(344),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(346),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_let] = ACTIONS(386),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(348),
    [anon_sym_union] = ACTIONS(346),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(384),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(352),
    [anon_sym_AMP] = ACTIONS(388),
    [anon_sym_DOT_DOT] = ACTIONS(390),
    [anon_sym_DASH] = ACTIONS(384),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(354),
    [anon_sym_move] = ACTIONS(356),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(358),
    [sym_super] = ACTIONS(360),
    [sym_crate] = ACTIONS(360),
    [sym_metavariable] = ACTIONS(362),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [44] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1942),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1272),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(1176),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_let_condition] = STATE(2190),
    [sym__let_chain] = STATE(2118),
    [sym__condition] = STATE(2134),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(97),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(340),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(384),
    [anon_sym_u8] = ACTIONS(342),
    [anon_sym_i8] = ACTIONS(342),
    [anon_sym_u16] = ACTIONS(342),
    [anon_sym_i16] = ACTIONS(342),
    [anon_sym_u32] = ACTIONS(342),
    [anon_sym_i32] = ACTIONS(342),
    [anon_sym_u64] = ACTIONS(342),
    [anon_sym_i64] = ACTIONS(342),
    [anon_sym_u128] = ACTIONS(342),
    [anon_sym_i128] = ACTIONS(342),
    [anon_sym_isize] = ACTIONS(342),
    [anon_sym_usize] = ACTIONS(342),
    [anon_sym_f32] = ACTIONS(342),
    [anon_sym_f64] = ACTIONS(342),
    [anon_sym_bool] = ACTIONS(342),
    [anon_sym_str] = ACTIONS(342),
    [anon_sym_char] = ACTIONS(342),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(344),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(346),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_let] = ACTIONS(386),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(348),
    [anon_sym_union] = ACTIONS(346),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(384),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(352),
    [anon_sym_AMP] = ACTIONS(388),
    [anon_sym_DOT_DOT] = ACTIONS(390),
    [anon_sym_DASH] = ACTIONS(384),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(354),
    [anon_sym_move] = ACTIONS(356),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(358),
    [sym_super] = ACTIONS(360),
    [sym_crate] = ACTIONS(360),
    [sym_metavariable] = ACTIONS(362),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [45] = {
    [sym_attribute_item] = STATE(57),
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1223),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [aux_sym_enum_variant_list_repeat1] = STATE(57),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_RPAREN] = ACTIONS(396),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_POUND] = ACTIONS(370),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [46] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1942),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1272),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(1176),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_let_condition] = STATE(2190),
    [sym__let_chain] = STATE(2118),
    [sym__condition] = STATE(2136),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(97),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(340),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(384),
    [anon_sym_u8] = ACTIONS(342),
    [anon_sym_i8] = ACTIONS(342),
    [anon_sym_u16] = ACTIONS(342),
    [anon_sym_i16] = ACTIONS(342),
    [anon_sym_u32] = ACTIONS(342),
    [anon_sym_i32] = ACTIONS(342),
    [anon_sym_u64] = ACTIONS(342),
    [anon_sym_i64] = ACTIONS(342),
    [anon_sym_u128] = ACTIONS(342),
    [anon_sym_i128] = ACTIONS(342),
    [anon_sym_isize] = ACTIONS(342),
    [anon_sym_usize] = ACTIONS(342),
    [anon_sym_f32] = ACTIONS(342),
    [anon_sym_f64] = ACTIONS(342),
    [anon_sym_bool] = ACTIONS(342),
    [anon_sym_str] = ACTIONS(342),
    [anon_sym_char] = ACTIONS(342),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(344),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(346),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_let] = ACTIONS(386),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(348),
    [anon_sym_union] = ACTIONS(346),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(384),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(352),
    [anon_sym_AMP] = ACTIONS(388),
    [anon_sym_DOT_DOT] = ACTIONS(390),
    [anon_sym_DASH] = ACTIONS(384),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(354),
    [anon_sym_move] = ACTIONS(356),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(358),
    [sym_super] = ACTIONS(360),
    [sym_crate] = ACTIONS(360),
    [sym_metavariable] = ACTIONS(362),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [47] = {
    [sym_attribute_item] = STATE(57),
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1223),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [aux_sym_enum_variant_list_repeat1] = STATE(57),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_RPAREN] = ACTIONS(398),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_POUND] = ACTIONS(370),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [48] = {
    [sym_else_clause] = STATE(74),
    [ts_builtin_sym_end] = ACTIONS(400),
    [sym_identifier] = ACTIONS(402),
    [anon_sym_SEMI] = ACTIONS(400),
    [anon_sym_macro_rules_BANG] = ACTIONS(400),
    [anon_sym_LPAREN] = ACTIONS(400),
    [anon_sym_LBRACE] = ACTIONS(400),
    [anon_sym_RBRACE] = ACTIONS(400),
    [anon_sym_LBRACK] = ACTIONS(400),
    [anon_sym_PLUS] = ACTIONS(402),
    [anon_sym_STAR] = ACTIONS(402),
    [anon_sym_QMARK] = ACTIONS(400),
    [anon_sym_u8] = ACTIONS(402),
    [anon_sym_i8] = ACTIONS(402),
    [anon_sym_u16] = ACTIONS(402),
    [anon_sym_i16] = ACTIONS(402),
    [anon_sym_u32] = ACTIONS(402),
    [anon_sym_i32] = ACTIONS(402),
    [anon_sym_u64] = ACTIONS(402),
    [anon_sym_i64] = ACTIONS(402),
    [anon_sym_u128] = ACTIONS(402),
    [anon_sym_i128] = ACTIONS(402),
    [anon_sym_isize] = ACTIONS(402),
    [anon_sym_usize] = ACTIONS(402),
    [anon_sym_f32] = ACTIONS(402),
    [anon_sym_f64] = ACTIONS(402),
    [anon_sym_bool] = ACTIONS(402),
    [anon_sym_str] = ACTIONS(402),
    [anon_sym_char] = ACTIONS(402),
    [anon_sym_SQUOTE] = ACTIONS(402),
    [anon_sym_as] = ACTIONS(402),
    [anon_sym_async] = ACTIONS(402),
    [anon_sym_break] = ACTIONS(402),
    [anon_sym_const] = ACTIONS(402),
    [anon_sym_continue] = ACTIONS(402),
    [anon_sym_default] = ACTIONS(402),
    [anon_sym_enum] = ACTIONS(402),
    [anon_sym_fn] = ACTIONS(402),
    [anon_sym_for] = ACTIONS(402),
    [anon_sym_if] = ACTIONS(402),
    [anon_sym_impl] = ACTIONS(402),
    [anon_sym_let] = ACTIONS(402),
    [anon_sym_loop] = ACTIONS(402),
    [anon_sym_match] = ACTIONS(402),
    [anon_sym_mod] = ACTIONS(402),
    [anon_sym_pub] = ACTIONS(402),
    [anon_sym_return] = ACTIONS(402),
    [anon_sym_static] = ACTIONS(402),
    [anon_sym_struct] = ACTIONS(402),
    [anon_sym_trait] = ACTIONS(402),
    [anon_sym_type] = ACTIONS(402),
    [anon_sym_union] = ACTIONS(402),
    [anon_sym_unsafe] = ACTIONS(402),
    [anon_sym_use] = ACTIONS(402),
    [anon_sym_while] = ACTIONS(402),
    [anon_sym_POUND] = ACTIONS(400),
    [anon_sym_BANG] = ACTIONS(402),
    [anon_sym_EQ] = ACTIONS(402),
    [anon_sym_extern] = ACTIONS(402),
    [anon_sym_LT] = ACTIONS(402),
    [anon_sym_GT] = ACTIONS(402),
    [anon_sym_else] = ACTIONS(404),
    [anon_sym_COLON_COLON] = ACTIONS(400),
    [anon_sym_AMP] = ACTIONS(402),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(400),
    [anon_sym_DOT_DOT] = ACTIONS(402),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(400),
    [anon_sym_DASH] = ACTIONS(402),
    [anon_sym_AMP_AMP] = ACTIONS(400),
    [anon_sym_PIPE_PIPE] = ACTIONS(400),
    [anon_sym_PIPE] = ACTIONS(402),
    [anon_sym_CARET] = ACTIONS(402),
    [anon_sym_EQ_EQ] = ACTIONS(400),
    [anon_sym_BANG_EQ] = ACTIONS(400),
    [anon_sym_LT_EQ] = ACTIONS(400),
    [anon_sym_GT_EQ] = ACTIONS(400),
    [anon_sym_LT_LT] = ACTIONS(402),
    [anon_sym_GT_GT] = ACTIONS(402),
    [anon_sym_SLASH] = ACTIONS(402),
    [anon_sym_PERCENT] = ACTIONS(402),
    [anon_sym_PLUS_EQ] = ACTIONS(400),
    [anon_sym_DASH_EQ] = ACTIONS(400),
    [anon_sym_STAR_EQ] = ACTIONS(400),
    [anon_sym_SLASH_EQ] = ACTIONS(400),
    [anon_sym_PERCENT_EQ] = ACTIONS(400),
    [anon_sym_AMP_EQ] = ACTIONS(400),
    [anon_sym_PIPE_EQ] = ACTIONS(400),
    [anon_sym_CARET_EQ] = ACTIONS(400),
    [anon_sym_LT_LT_EQ] = ACTIONS(400),
    [anon_sym_GT_GT_EQ] = ACTIONS(400),
    [anon_sym_yield] = ACTIONS(402),
    [anon_sym_move] = ACTIONS(402),
    [anon_sym_DOT] = ACTIONS(402),
    [sym_integer_literal] = ACTIONS(400),
    [aux_sym_string_literal_token1] = ACTIONS(400),
    [sym_char_literal] = ACTIONS(400),
    [anon_sym_true] = ACTIONS(402),
    [anon_sym_false] = ACTIONS(402),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(402),
    [sym_super] = ACTIONS(402),
    [sym_crate] = ACTIONS(402),
    [sym_metavariable] = ACTIONS(400),
    [sym_raw_string_literal] = ACTIONS(400),
    [sym_float_literal] = ACTIONS(400),
    [sym_block_comment] = ACTIONS(3),
  },
  [49] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1942),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1272),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(1176),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_let_condition] = STATE(2190),
    [sym__let_chain] = STATE(2118),
    [sym__condition] = STATE(2275),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(97),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(340),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(384),
    [anon_sym_u8] = ACTIONS(342),
    [anon_sym_i8] = ACTIONS(342),
    [anon_sym_u16] = ACTIONS(342),
    [anon_sym_i16] = ACTIONS(342),
    [anon_sym_u32] = ACTIONS(342),
    [anon_sym_i32] = ACTIONS(342),
    [anon_sym_u64] = ACTIONS(342),
    [anon_sym_i64] = ACTIONS(342),
    [anon_sym_u128] = ACTIONS(342),
    [anon_sym_i128] = ACTIONS(342),
    [anon_sym_isize] = ACTIONS(342),
    [anon_sym_usize] = ACTIONS(342),
    [anon_sym_f32] = ACTIONS(342),
    [anon_sym_f64] = ACTIONS(342),
    [anon_sym_bool] = ACTIONS(342),
    [anon_sym_str] = ACTIONS(342),
    [anon_sym_char] = ACTIONS(342),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(344),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(346),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_let] = ACTIONS(386),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(348),
    [anon_sym_union] = ACTIONS(346),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(384),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(352),
    [anon_sym_AMP] = ACTIONS(388),
    [anon_sym_DOT_DOT] = ACTIONS(390),
    [anon_sym_DASH] = ACTIONS(384),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(354),
    [anon_sym_move] = ACTIONS(356),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(358),
    [sym_super] = ACTIONS(360),
    [sym_crate] = ACTIONS(360),
    [sym_metavariable] = ACTIONS(362),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [50] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1942),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1272),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(1176),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_let_condition] = STATE(2190),
    [sym__let_chain] = STATE(2118),
    [sym__condition] = STATE(2137),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(97),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(340),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(384),
    [anon_sym_u8] = ACTIONS(342),
    [anon_sym_i8] = ACTIONS(342),
    [anon_sym_u16] = ACTIONS(342),
    [anon_sym_i16] = ACTIONS(342),
    [anon_sym_u32] = ACTIONS(342),
    [anon_sym_i32] = ACTIONS(342),
    [anon_sym_u64] = ACTIONS(342),
    [anon_sym_i64] = ACTIONS(342),
    [anon_sym_u128] = ACTIONS(342),
    [anon_sym_i128] = ACTIONS(342),
    [anon_sym_isize] = ACTIONS(342),
    [anon_sym_usize] = ACTIONS(342),
    [anon_sym_f32] = ACTIONS(342),
    [anon_sym_f64] = ACTIONS(342),
    [anon_sym_bool] = ACTIONS(342),
    [anon_sym_str] = ACTIONS(342),
    [anon_sym_char] = ACTIONS(342),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(344),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(346),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_let] = ACTIONS(386),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(348),
    [anon_sym_union] = ACTIONS(346),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(384),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(352),
    [anon_sym_AMP] = ACTIONS(388),
    [anon_sym_DOT_DOT] = ACTIONS(390),
    [anon_sym_DASH] = ACTIONS(384),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(354),
    [anon_sym_move] = ACTIONS(356),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(358),
    [sym_super] = ACTIONS(360),
    [sym_crate] = ACTIONS(360),
    [sym_metavariable] = ACTIONS(362),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [51] = {
    [sym_attribute_item] = STATE(621),
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1273),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [aux_sym_enum_variant_list_repeat1] = STATE(621),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_POUND] = ACTIONS(370),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [52] = {
    [sym_attribute_item] = STATE(57),
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1223),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [aux_sym_enum_variant_list_repeat1] = STATE(57),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_POUND] = ACTIONS(370),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [53] = {
    [ts_builtin_sym_end] = ACTIONS(406),
    [sym_identifier] = ACTIONS(408),
    [anon_sym_SEMI] = ACTIONS(406),
    [anon_sym_macro_rules_BANG] = ACTIONS(406),
    [anon_sym_LPAREN] = ACTIONS(406),
    [anon_sym_LBRACE] = ACTIONS(406),
    [anon_sym_RBRACE] = ACTIONS(406),
    [anon_sym_LBRACK] = ACTIONS(406),
    [anon_sym_PLUS] = ACTIONS(408),
    [anon_sym_STAR] = ACTIONS(408),
    [anon_sym_QMARK] = ACTIONS(406),
    [anon_sym_u8] = ACTIONS(408),
    [anon_sym_i8] = ACTIONS(408),
    [anon_sym_u16] = ACTIONS(408),
    [anon_sym_i16] = ACTIONS(408),
    [anon_sym_u32] = ACTIONS(408),
    [anon_sym_i32] = ACTIONS(408),
    [anon_sym_u64] = ACTIONS(408),
    [anon_sym_i64] = ACTIONS(408),
    [anon_sym_u128] = ACTIONS(408),
    [anon_sym_i128] = ACTIONS(408),
    [anon_sym_isize] = ACTIONS(408),
    [anon_sym_usize] = ACTIONS(408),
    [anon_sym_f32] = ACTIONS(408),
    [anon_sym_f64] = ACTIONS(408),
    [anon_sym_bool] = ACTIONS(408),
    [anon_sym_str] = ACTIONS(408),
    [anon_sym_char] = ACTIONS(408),
    [anon_sym_SQUOTE] = ACTIONS(408),
    [anon_sym_as] = ACTIONS(408),
    [anon_sym_async] = ACTIONS(408),
    [anon_sym_break] = ACTIONS(408),
    [anon_sym_const] = ACTIONS(408),
    [anon_sym_continue] = ACTIONS(408),
    [anon_sym_default] = ACTIONS(408),
    [anon_sym_enum] = ACTIONS(408),
    [anon_sym_fn] = ACTIONS(408),
    [anon_sym_for] = ACTIONS(408),
    [anon_sym_if] = ACTIONS(408),
    [anon_sym_impl] = ACTIONS(408),
    [anon_sym_let] = ACTIONS(408),
    [anon_sym_loop] = ACTIONS(408),
    [anon_sym_match] = ACTIONS(408),
    [anon_sym_mod] = ACTIONS(408),
    [anon_sym_pub] = ACTIONS(408),
    [anon_sym_return] = ACTIONS(408),
    [anon_sym_static] = ACTIONS(408),
    [anon_sym_struct] = ACTIONS(408),
    [anon_sym_trait] = ACTIONS(408),
    [anon_sym_type] = ACTIONS(408),
    [anon_sym_union] = ACTIONS(408),
    [anon_sym_unsafe] = ACTIONS(408),
    [anon_sym_use] = ACTIONS(408),
    [anon_sym_while] = ACTIONS(408),
    [anon_sym_POUND] = ACTIONS(406),
    [anon_sym_BANG] = ACTIONS(408),
    [anon_sym_EQ] = ACTIONS(408),
    [anon_sym_extern] = ACTIONS(408),
    [anon_sym_LT] = ACTIONS(408),
    [anon_sym_GT] = ACTIONS(408),
    [anon_sym_else] = ACTIONS(408),
    [anon_sym_COLON_COLON] = ACTIONS(406),
    [anon_sym_AMP] = ACTIONS(408),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(406),
    [anon_sym_DOT_DOT] = ACTIONS(408),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(406),
    [anon_sym_DASH] = ACTIONS(408),
    [anon_sym_AMP_AMP] = ACTIONS(406),
    [anon_sym_PIPE_PIPE] = ACTIONS(406),
    [anon_sym_PIPE] = ACTIONS(408),
    [anon_sym_CARET] = ACTIONS(408),
    [anon_sym_EQ_EQ] = ACTIONS(406),
    [anon_sym_BANG_EQ] = ACTIONS(406),
    [anon_sym_LT_EQ] = ACTIONS(406),
    [anon_sym_GT_EQ] = ACTIONS(406),
    [anon_sym_LT_LT] = ACTIONS(408),
    [anon_sym_GT_GT] = ACTIONS(408),
    [anon_sym_SLASH] = ACTIONS(408),
    [anon_sym_PERCENT] = ACTIONS(408),
    [anon_sym_PLUS_EQ] = ACTIONS(406),
    [anon_sym_DASH_EQ] = ACTIONS(406),
    [anon_sym_STAR_EQ] = ACTIONS(406),
    [anon_sym_SLASH_EQ] = ACTIONS(406),
    [anon_sym_PERCENT_EQ] = ACTIONS(406),
    [anon_sym_AMP_EQ] = ACTIONS(406),
    [anon_sym_PIPE_EQ] = ACTIONS(406),
    [anon_sym_CARET_EQ] = ACTIONS(406),
    [anon_sym_LT_LT_EQ] = ACTIONS(406),
    [anon_sym_GT_GT_EQ] = ACTIONS(406),
    [anon_sym_yield] = ACTIONS(408),
    [anon_sym_move] = ACTIONS(408),
    [anon_sym_DOT] = ACTIONS(408),
    [sym_integer_literal] = ACTIONS(406),
    [aux_sym_string_literal_token1] = ACTIONS(406),
    [sym_char_literal] = ACTIONS(406),
    [anon_sym_true] = ACTIONS(408),
    [anon_sym_false] = ACTIONS(408),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(408),
    [sym_super] = ACTIONS(408),
    [sym_crate] = ACTIONS(408),
    [sym_metavariable] = ACTIONS(406),
    [sym_raw_string_literal] = ACTIONS(406),
    [sym_float_literal] = ACTIONS(406),
    [sym_block_comment] = ACTIONS(3),
  },
  [54] = {
    [ts_builtin_sym_end] = ACTIONS(410),
    [sym_identifier] = ACTIONS(412),
    [anon_sym_SEMI] = ACTIONS(410),
    [anon_sym_macro_rules_BANG] = ACTIONS(410),
    [anon_sym_LPAREN] = ACTIONS(410),
    [anon_sym_LBRACE] = ACTIONS(410),
    [anon_sym_RBRACE] = ACTIONS(410),
    [anon_sym_LBRACK] = ACTIONS(410),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_STAR] = ACTIONS(412),
    [anon_sym_QMARK] = ACTIONS(410),
    [anon_sym_u8] = ACTIONS(412),
    [anon_sym_i8] = ACTIONS(412),
    [anon_sym_u16] = ACTIONS(412),
    [anon_sym_i16] = ACTIONS(412),
    [anon_sym_u32] = ACTIONS(412),
    [anon_sym_i32] = ACTIONS(412),
    [anon_sym_u64] = ACTIONS(412),
    [anon_sym_i64] = ACTIONS(412),
    [anon_sym_u128] = ACTIONS(412),
    [anon_sym_i128] = ACTIONS(412),
    [anon_sym_isize] = ACTIONS(412),
    [anon_sym_usize] = ACTIONS(412),
    [anon_sym_f32] = ACTIONS(412),
    [anon_sym_f64] = ACTIONS(412),
    [anon_sym_bool] = ACTIONS(412),
    [anon_sym_str] = ACTIONS(412),
    [anon_sym_char] = ACTIONS(412),
    [anon_sym_SQUOTE] = ACTIONS(412),
    [anon_sym_as] = ACTIONS(412),
    [anon_sym_async] = ACTIONS(412),
    [anon_sym_break] = ACTIONS(412),
    [anon_sym_const] = ACTIONS(412),
    [anon_sym_continue] = ACTIONS(412),
    [anon_sym_default] = ACTIONS(412),
    [anon_sym_enum] = ACTIONS(412),
    [anon_sym_fn] = ACTIONS(412),
    [anon_sym_for] = ACTIONS(412),
    [anon_sym_if] = ACTIONS(412),
    [anon_sym_impl] = ACTIONS(412),
    [anon_sym_let] = ACTIONS(412),
    [anon_sym_loop] = ACTIONS(412),
    [anon_sym_match] = ACTIONS(412),
    [anon_sym_mod] = ACTIONS(412),
    [anon_sym_pub] = ACTIONS(412),
    [anon_sym_return] = ACTIONS(412),
    [anon_sym_static] = ACTIONS(412),
    [anon_sym_struct] = ACTIONS(412),
    [anon_sym_trait] = ACTIONS(412),
    [anon_sym_type] = ACTIONS(412),
    [anon_sym_union] = ACTIONS(412),
    [anon_sym_unsafe] = ACTIONS(412),
    [anon_sym_use] = ACTIONS(412),
    [anon_sym_while] = ACTIONS(412),
    [anon_sym_POUND] = ACTIONS(410),
    [anon_sym_BANG] = ACTIONS(412),
    [anon_sym_EQ] = ACTIONS(412),
    [anon_sym_extern] = ACTIONS(412),
    [anon_sym_LT] = ACTIONS(412),
    [anon_sym_GT] = ACTIONS(412),
    [anon_sym_else] = ACTIONS(412),
    [anon_sym_COLON_COLON] = ACTIONS(410),
    [anon_sym_AMP] = ACTIONS(412),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(410),
    [anon_sym_DOT_DOT] = ACTIONS(412),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(410),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_AMP_AMP] = ACTIONS(410),
    [anon_sym_PIPE_PIPE] = ACTIONS(410),
    [anon_sym_PIPE] = ACTIONS(412),
    [anon_sym_CARET] = ACTIONS(412),
    [anon_sym_EQ_EQ] = ACTIONS(410),
    [anon_sym_BANG_EQ] = ACTIONS(410),
    [anon_sym_LT_EQ] = ACTIONS(410),
    [anon_sym_GT_EQ] = ACTIONS(410),
    [anon_sym_LT_LT] = ACTIONS(412),
    [anon_sym_GT_GT] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(412),
    [anon_sym_PERCENT] = ACTIONS(412),
    [anon_sym_PLUS_EQ] = ACTIONS(410),
    [anon_sym_DASH_EQ] = ACTIONS(410),
    [anon_sym_STAR_EQ] = ACTIONS(410),
    [anon_sym_SLASH_EQ] = ACTIONS(410),
    [anon_sym_PERCENT_EQ] = ACTIONS(410),
    [anon_sym_AMP_EQ] = ACTIONS(410),
    [anon_sym_PIPE_EQ] = ACTIONS(410),
    [anon_sym_CARET_EQ] = ACTIONS(410),
    [anon_sym_LT_LT_EQ] = ACTIONS(410),
    [anon_sym_GT_GT_EQ] = ACTIONS(410),
    [anon_sym_yield] = ACTIONS(412),
    [anon_sym_move] = ACTIONS(412),
    [anon_sym_DOT] = ACTIONS(412),
    [sym_integer_literal] = ACTIONS(410),
    [aux_sym_string_literal_token1] = ACTIONS(410),
    [sym_char_literal] = ACTIONS(410),
    [anon_sym_true] = ACTIONS(412),
    [anon_sym_false] = ACTIONS(412),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(412),
    [sym_super] = ACTIONS(412),
    [sym_crate] = ACTIONS(412),
    [sym_metavariable] = ACTIONS(410),
    [sym_raw_string_literal] = ACTIONS(410),
    [sym_float_literal] = ACTIONS(410),
    [sym_block_comment] = ACTIONS(3),
  },
  [55] = {
    [sym_attribute_item] = STATE(621),
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1193),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [aux_sym_enum_variant_list_repeat1] = STATE(621),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_POUND] = ACTIONS(370),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [56] = {
    [ts_builtin_sym_end] = ACTIONS(414),
    [sym_identifier] = ACTIONS(416),
    [anon_sym_SEMI] = ACTIONS(414),
    [anon_sym_macro_rules_BANG] = ACTIONS(414),
    [anon_sym_LPAREN] = ACTIONS(414),
    [anon_sym_LBRACE] = ACTIONS(414),
    [anon_sym_RBRACE] = ACTIONS(414),
    [anon_sym_LBRACK] = ACTIONS(414),
    [anon_sym_PLUS] = ACTIONS(416),
    [anon_sym_STAR] = ACTIONS(416),
    [anon_sym_QMARK] = ACTIONS(414),
    [anon_sym_u8] = ACTIONS(416),
    [anon_sym_i8] = ACTIONS(416),
    [anon_sym_u16] = ACTIONS(416),
    [anon_sym_i16] = ACTIONS(416),
    [anon_sym_u32] = ACTIONS(416),
    [anon_sym_i32] = ACTIONS(416),
    [anon_sym_u64] = ACTIONS(416),
    [anon_sym_i64] = ACTIONS(416),
    [anon_sym_u128] = ACTIONS(416),
    [anon_sym_i128] = ACTIONS(416),
    [anon_sym_isize] = ACTIONS(416),
    [anon_sym_usize] = ACTIONS(416),
    [anon_sym_f32] = ACTIONS(416),
    [anon_sym_f64] = ACTIONS(416),
    [anon_sym_bool] = ACTIONS(416),
    [anon_sym_str] = ACTIONS(416),
    [anon_sym_char] = ACTIONS(416),
    [anon_sym_SQUOTE] = ACTIONS(416),
    [anon_sym_as] = ACTIONS(416),
    [anon_sym_async] = ACTIONS(416),
    [anon_sym_break] = ACTIONS(416),
    [anon_sym_const] = ACTIONS(416),
    [anon_sym_continue] = ACTIONS(416),
    [anon_sym_default] = ACTIONS(416),
    [anon_sym_enum] = ACTIONS(416),
    [anon_sym_fn] = ACTIONS(416),
    [anon_sym_for] = ACTIONS(416),
    [anon_sym_if] = ACTIONS(416),
    [anon_sym_impl] = ACTIONS(416),
    [anon_sym_let] = ACTIONS(416),
    [anon_sym_loop] = ACTIONS(416),
    [anon_sym_match] = ACTIONS(416),
    [anon_sym_mod] = ACTIONS(416),
    [anon_sym_pub] = ACTIONS(416),
    [anon_sym_return] = ACTIONS(416),
    [anon_sym_static] = ACTIONS(416),
    [anon_sym_struct] = ACTIONS(416),
    [anon_sym_trait] = ACTIONS(416),
    [anon_sym_type] = ACTIONS(416),
    [anon_sym_union] = ACTIONS(416),
    [anon_sym_unsafe] = ACTIONS(416),
    [anon_sym_use] = ACTIONS(416),
    [anon_sym_while] = ACTIONS(416),
    [anon_sym_POUND] = ACTIONS(414),
    [anon_sym_BANG] = ACTIONS(416),
    [anon_sym_EQ] = ACTIONS(416),
    [anon_sym_extern] = ACTIONS(416),
    [anon_sym_LT] = ACTIONS(416),
    [anon_sym_GT] = ACTIONS(416),
    [anon_sym_else] = ACTIONS(416),
    [anon_sym_COLON_COLON] = ACTIONS(414),
    [anon_sym_AMP] = ACTIONS(416),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(414),
    [anon_sym_DOT_DOT] = ACTIONS(416),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(414),
    [anon_sym_DASH] = ACTIONS(416),
    [anon_sym_AMP_AMP] = ACTIONS(414),
    [anon_sym_PIPE_PIPE] = ACTIONS(414),
    [anon_sym_PIPE] = ACTIONS(416),
    [anon_sym_CARET] = ACTIONS(416),
    [anon_sym_EQ_EQ] = ACTIONS(414),
    [anon_sym_BANG_EQ] = ACTIONS(414),
    [anon_sym_LT_EQ] = ACTIONS(414),
    [anon_sym_GT_EQ] = ACTIONS(414),
    [anon_sym_LT_LT] = ACTIONS(416),
    [anon_sym_GT_GT] = ACTIONS(416),
    [anon_sym_SLASH] = ACTIONS(416),
    [anon_sym_PERCENT] = ACTIONS(416),
    [anon_sym_PLUS_EQ] = ACTIONS(414),
    [anon_sym_DASH_EQ] = ACTIONS(414),
    [anon_sym_STAR_EQ] = ACTIONS(414),
    [anon_sym_SLASH_EQ] = ACTIONS(414),
    [anon_sym_PERCENT_EQ] = ACTIONS(414),
    [anon_sym_AMP_EQ] = ACTIONS(414),
    [anon_sym_PIPE_EQ] = ACTIONS(414),
    [anon_sym_CARET_EQ] = ACTIONS(414),
    [anon_sym_LT_LT_EQ] = ACTIONS(414),
    [anon_sym_GT_GT_EQ] = ACTIONS(414),
    [anon_sym_yield] = ACTIONS(416),
    [anon_sym_move] = ACTIONS(416),
    [anon_sym_DOT] = ACTIONS(416),
    [sym_integer_literal] = ACTIONS(414),
    [aux_sym_string_literal_token1] = ACTIONS(414),
    [sym_char_literal] = ACTIONS(414),
    [anon_sym_true] = ACTIONS(416),
    [anon_sym_false] = ACTIONS(416),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(416),
    [sym_super] = ACTIONS(416),
    [sym_crate] = ACTIONS(416),
    [sym_metavariable] = ACTIONS(414),
    [sym_raw_string_literal] = ACTIONS(414),
    [sym_float_literal] = ACTIONS(414),
    [sym_block_comment] = ACTIONS(3),
  },
  [57] = {
    [sym_attribute_item] = STATE(621),
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1255),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [aux_sym_enum_variant_list_repeat1] = STATE(621),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_POUND] = ACTIONS(370),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [58] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1942),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1240),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(1176),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_let_condition] = STATE(1962),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(97),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(340),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(384),
    [anon_sym_u8] = ACTIONS(342),
    [anon_sym_i8] = ACTIONS(342),
    [anon_sym_u16] = ACTIONS(342),
    [anon_sym_i16] = ACTIONS(342),
    [anon_sym_u32] = ACTIONS(342),
    [anon_sym_i32] = ACTIONS(342),
    [anon_sym_u64] = ACTIONS(342),
    [anon_sym_i64] = ACTIONS(342),
    [anon_sym_u128] = ACTIONS(342),
    [anon_sym_i128] = ACTIONS(342),
    [anon_sym_isize] = ACTIONS(342),
    [anon_sym_usize] = ACTIONS(342),
    [anon_sym_f32] = ACTIONS(342),
    [anon_sym_f64] = ACTIONS(342),
    [anon_sym_bool] = ACTIONS(342),
    [anon_sym_str] = ACTIONS(342),
    [anon_sym_char] = ACTIONS(342),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(344),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(346),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_let] = ACTIONS(386),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(348),
    [anon_sym_union] = ACTIONS(346),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(384),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(352),
    [anon_sym_AMP] = ACTIONS(388),
    [anon_sym_DOT_DOT] = ACTIONS(418),
    [anon_sym_DASH] = ACTIONS(384),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(354),
    [anon_sym_move] = ACTIONS(356),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(358),
    [sym_super] = ACTIONS(360),
    [sym_crate] = ACTIONS(360),
    [sym_metavariable] = ACTIONS(362),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [59] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1147),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_let_condition] = STATE(1962),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_let] = ACTIONS(382),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(420),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [60] = {
    [ts_builtin_sym_end] = ACTIONS(422),
    [sym_identifier] = ACTIONS(424),
    [anon_sym_SEMI] = ACTIONS(422),
    [anon_sym_macro_rules_BANG] = ACTIONS(422),
    [anon_sym_LPAREN] = ACTIONS(422),
    [anon_sym_LBRACE] = ACTIONS(422),
    [anon_sym_RBRACE] = ACTIONS(422),
    [anon_sym_LBRACK] = ACTIONS(422),
    [anon_sym_PLUS] = ACTIONS(424),
    [anon_sym_STAR] = ACTIONS(424),
    [anon_sym_QMARK] = ACTIONS(422),
    [anon_sym_u8] = ACTIONS(424),
    [anon_sym_i8] = ACTIONS(424),
    [anon_sym_u16] = ACTIONS(424),
    [anon_sym_i16] = ACTIONS(424),
    [anon_sym_u32] = ACTIONS(424),
    [anon_sym_i32] = ACTIONS(424),
    [anon_sym_u64] = ACTIONS(424),
    [anon_sym_i64] = ACTIONS(424),
    [anon_sym_u128] = ACTIONS(424),
    [anon_sym_i128] = ACTIONS(424),
    [anon_sym_isize] = ACTIONS(424),
    [anon_sym_usize] = ACTIONS(424),
    [anon_sym_f32] = ACTIONS(424),
    [anon_sym_f64] = ACTIONS(424),
    [anon_sym_bool] = ACTIONS(424),
    [anon_sym_str] = ACTIONS(424),
    [anon_sym_char] = ACTIONS(424),
    [anon_sym_SQUOTE] = ACTIONS(424),
    [anon_sym_as] = ACTIONS(424),
    [anon_sym_async] = ACTIONS(424),
    [anon_sym_break] = ACTIONS(424),
    [anon_sym_const] = ACTIONS(424),
    [anon_sym_continue] = ACTIONS(424),
    [anon_sym_default] = ACTIONS(424),
    [anon_sym_enum] = ACTIONS(424),
    [anon_sym_fn] = ACTIONS(424),
    [anon_sym_for] = ACTIONS(424),
    [anon_sym_if] = ACTIONS(424),
    [anon_sym_impl] = ACTIONS(424),
    [anon_sym_let] = ACTIONS(424),
    [anon_sym_loop] = ACTIONS(424),
    [anon_sym_match] = ACTIONS(424),
    [anon_sym_mod] = ACTIONS(424),
    [anon_sym_pub] = ACTIONS(424),
    [anon_sym_return] = ACTIONS(424),
    [anon_sym_static] = ACTIONS(424),
    [anon_sym_struct] = ACTIONS(424),
    [anon_sym_trait] = ACTIONS(424),
    [anon_sym_type] = ACTIONS(424),
    [anon_sym_union] = ACTIONS(424),
    [anon_sym_unsafe] = ACTIONS(424),
    [anon_sym_use] = ACTIONS(424),
    [anon_sym_while] = ACTIONS(424),
    [anon_sym_POUND] = ACTIONS(422),
    [anon_sym_BANG] = ACTIONS(424),
    [anon_sym_EQ] = ACTIONS(424),
    [anon_sym_extern] = ACTIONS(424),
    [anon_sym_LT] = ACTIONS(424),
    [anon_sym_GT] = ACTIONS(424),
    [anon_sym_COLON_COLON] = ACTIONS(422),
    [anon_sym_AMP] = ACTIONS(424),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(422),
    [anon_sym_DOT_DOT] = ACTIONS(424),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(422),
    [anon_sym_DASH] = ACTIONS(424),
    [anon_sym_AMP_AMP] = ACTIONS(422),
    [anon_sym_PIPE_PIPE] = ACTIONS(422),
    [anon_sym_PIPE] = ACTIONS(424),
    [anon_sym_CARET] = ACTIONS(424),
    [anon_sym_EQ_EQ] = ACTIONS(422),
    [anon_sym_BANG_EQ] = ACTIONS(422),
    [anon_sym_LT_EQ] = ACTIONS(422),
    [anon_sym_GT_EQ] = ACTIONS(422),
    [anon_sym_LT_LT] = ACTIONS(424),
    [anon_sym_GT_GT] = ACTIONS(424),
    [anon_sym_SLASH] = ACTIONS(424),
    [anon_sym_PERCENT] = ACTIONS(424),
    [anon_sym_PLUS_EQ] = ACTIONS(422),
    [anon_sym_DASH_EQ] = ACTIONS(422),
    [anon_sym_STAR_EQ] = ACTIONS(422),
    [anon_sym_SLASH_EQ] = ACTIONS(422),
    [anon_sym_PERCENT_EQ] = ACTIONS(422),
    [anon_sym_AMP_EQ] = ACTIONS(422),
    [anon_sym_PIPE_EQ] = ACTIONS(422),
    [anon_sym_CARET_EQ] = ACTIONS(422),
    [anon_sym_LT_LT_EQ] = ACTIONS(422),
    [anon_sym_GT_GT_EQ] = ACTIONS(422),
    [anon_sym_yield] = ACTIONS(424),
    [anon_sym_move] = ACTIONS(424),
    [anon_sym_DOT] = ACTIONS(424),
    [sym_integer_literal] = ACTIONS(422),
    [aux_sym_string_literal_token1] = ACTIONS(422),
    [sym_char_literal] = ACTIONS(422),
    [anon_sym_true] = ACTIONS(424),
    [anon_sym_false] = ACTIONS(424),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(424),
    [sym_super] = ACTIONS(424),
    [sym_crate] = ACTIONS(424),
    [sym_metavariable] = ACTIONS(422),
    [sym_raw_string_literal] = ACTIONS(422),
    [sym_float_literal] = ACTIONS(422),
    [sym_block_comment] = ACTIONS(3),
  },
  [61] = {
    [ts_builtin_sym_end] = ACTIONS(426),
    [sym_identifier] = ACTIONS(428),
    [anon_sym_SEMI] = ACTIONS(426),
    [anon_sym_macro_rules_BANG] = ACTIONS(426),
    [anon_sym_LPAREN] = ACTIONS(426),
    [anon_sym_LBRACE] = ACTIONS(426),
    [anon_sym_RBRACE] = ACTIONS(426),
    [anon_sym_LBRACK] = ACTIONS(426),
    [anon_sym_PLUS] = ACTIONS(428),
    [anon_sym_STAR] = ACTIONS(428),
    [anon_sym_QMARK] = ACTIONS(426),
    [anon_sym_u8] = ACTIONS(428),
    [anon_sym_i8] = ACTIONS(428),
    [anon_sym_u16] = ACTIONS(428),
    [anon_sym_i16] = ACTIONS(428),
    [anon_sym_u32] = ACTIONS(428),
    [anon_sym_i32] = ACTIONS(428),
    [anon_sym_u64] = ACTIONS(428),
    [anon_sym_i64] = ACTIONS(428),
    [anon_sym_u128] = ACTIONS(428),
    [anon_sym_i128] = ACTIONS(428),
    [anon_sym_isize] = ACTIONS(428),
    [anon_sym_usize] = ACTIONS(428),
    [anon_sym_f32] = ACTIONS(428),
    [anon_sym_f64] = ACTIONS(428),
    [anon_sym_bool] = ACTIONS(428),
    [anon_sym_str] = ACTIONS(428),
    [anon_sym_char] = ACTIONS(428),
    [anon_sym_SQUOTE] = ACTIONS(428),
    [anon_sym_as] = ACTIONS(428),
    [anon_sym_async] = ACTIONS(428),
    [anon_sym_break] = ACTIONS(428),
    [anon_sym_const] = ACTIONS(428),
    [anon_sym_continue] = ACTIONS(428),
    [anon_sym_default] = ACTIONS(428),
    [anon_sym_enum] = ACTIONS(428),
    [anon_sym_fn] = ACTIONS(428),
    [anon_sym_for] = ACTIONS(428),
    [anon_sym_if] = ACTIONS(428),
    [anon_sym_impl] = ACTIONS(428),
    [anon_sym_let] = ACTIONS(428),
    [anon_sym_loop] = ACTIONS(428),
    [anon_sym_match] = ACTIONS(428),
    [anon_sym_mod] = ACTIONS(428),
    [anon_sym_pub] = ACTIONS(428),
    [anon_sym_return] = ACTIONS(428),
    [anon_sym_static] = ACTIONS(428),
    [anon_sym_struct] = ACTIONS(428),
    [anon_sym_trait] = ACTIONS(428),
    [anon_sym_type] = ACTIONS(428),
    [anon_sym_union] = ACTIONS(428),
    [anon_sym_unsafe] = ACTIONS(428),
    [anon_sym_use] = ACTIONS(428),
    [anon_sym_while] = ACTIONS(428),
    [anon_sym_POUND] = ACTIONS(426),
    [anon_sym_BANG] = ACTIONS(428),
    [anon_sym_EQ] = ACTIONS(428),
    [anon_sym_extern] = ACTIONS(428),
    [anon_sym_LT] = ACTIONS(428),
    [anon_sym_GT] = ACTIONS(428),
    [anon_sym_COLON_COLON] = ACTIONS(426),
    [anon_sym_AMP] = ACTIONS(428),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(426),
    [anon_sym_DOT_DOT] = ACTIONS(428),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(426),
    [anon_sym_DASH] = ACTIONS(428),
    [anon_sym_AMP_AMP] = ACTIONS(426),
    [anon_sym_PIPE_PIPE] = ACTIONS(426),
    [anon_sym_PIPE] = ACTIONS(428),
    [anon_sym_CARET] = ACTIONS(428),
    [anon_sym_EQ_EQ] = ACTIONS(426),
    [anon_sym_BANG_EQ] = ACTIONS(426),
    [anon_sym_LT_EQ] = ACTIONS(426),
    [anon_sym_GT_EQ] = ACTIONS(426),
    [anon_sym_LT_LT] = ACTIONS(428),
    [anon_sym_GT_GT] = ACTIONS(428),
    [anon_sym_SLASH] = ACTIONS(428),
    [anon_sym_PERCENT] = ACTIONS(428),
    [anon_sym_PLUS_EQ] = ACTIONS(426),
    [anon_sym_DASH_EQ] = ACTIONS(426),
    [anon_sym_STAR_EQ] = ACTIONS(426),
    [anon_sym_SLASH_EQ] = ACTIONS(426),
    [anon_sym_PERCENT_EQ] = ACTIONS(426),
    [anon_sym_AMP_EQ] = ACTIONS(426),
    [anon_sym_PIPE_EQ] = ACTIONS(426),
    [anon_sym_CARET_EQ] = ACTIONS(426),
    [anon_sym_LT_LT_EQ] = ACTIONS(426),
    [anon_sym_GT_GT_EQ] = ACTIONS(426),
    [anon_sym_yield] = ACTIONS(428),
    [anon_sym_move] = ACTIONS(428),
    [anon_sym_DOT] = ACTIONS(428),
    [sym_integer_literal] = ACTIONS(426),
    [aux_sym_string_literal_token1] = ACTIONS(426),
    [sym_char_literal] = ACTIONS(426),
    [anon_sym_true] = ACTIONS(428),
    [anon_sym_false] = ACTIONS(428),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(428),
    [sym_super] = ACTIONS(428),
    [sym_crate] = ACTIONS(428),
    [sym_metavariable] = ACTIONS(426),
    [sym_raw_string_literal] = ACTIONS(426),
    [sym_float_literal] = ACTIONS(426),
    [sym_block_comment] = ACTIONS(3),
  },
  [62] = {
    [ts_builtin_sym_end] = ACTIONS(430),
    [sym_identifier] = ACTIONS(432),
    [anon_sym_SEMI] = ACTIONS(430),
    [anon_sym_macro_rules_BANG] = ACTIONS(430),
    [anon_sym_LPAREN] = ACTIONS(430),
    [anon_sym_LBRACE] = ACTIONS(430),
    [anon_sym_RBRACE] = ACTIONS(430),
    [anon_sym_LBRACK] = ACTIONS(430),
    [anon_sym_PLUS] = ACTIONS(432),
    [anon_sym_STAR] = ACTIONS(432),
    [anon_sym_QMARK] = ACTIONS(430),
    [anon_sym_u8] = ACTIONS(432),
    [anon_sym_i8] = ACTIONS(432),
    [anon_sym_u16] = ACTIONS(432),
    [anon_sym_i16] = ACTIONS(432),
    [anon_sym_u32] = ACTIONS(432),
    [anon_sym_i32] = ACTIONS(432),
    [anon_sym_u64] = ACTIONS(432),
    [anon_sym_i64] = ACTIONS(432),
    [anon_sym_u128] = ACTIONS(432),
    [anon_sym_i128] = ACTIONS(432),
    [anon_sym_isize] = ACTIONS(432),
    [anon_sym_usize] = ACTIONS(432),
    [anon_sym_f32] = ACTIONS(432),
    [anon_sym_f64] = ACTIONS(432),
    [anon_sym_bool] = ACTIONS(432),
    [anon_sym_str] = ACTIONS(432),
    [anon_sym_char] = ACTIONS(432),
    [anon_sym_SQUOTE] = ACTIONS(432),
    [anon_sym_as] = ACTIONS(432),
    [anon_sym_async] = ACTIONS(432),
    [anon_sym_break] = ACTIONS(432),
    [anon_sym_const] = ACTIONS(432),
    [anon_sym_continue] = ACTIONS(432),
    [anon_sym_default] = ACTIONS(432),
    [anon_sym_enum] = ACTIONS(432),
    [anon_sym_fn] = ACTIONS(432),
    [anon_sym_for] = ACTIONS(432),
    [anon_sym_if] = ACTIONS(432),
    [anon_sym_impl] = ACTIONS(432),
    [anon_sym_let] = ACTIONS(432),
    [anon_sym_loop] = ACTIONS(432),
    [anon_sym_match] = ACTIONS(432),
    [anon_sym_mod] = ACTIONS(432),
    [anon_sym_pub] = ACTIONS(432),
    [anon_sym_return] = ACTIONS(432),
    [anon_sym_static] = ACTIONS(432),
    [anon_sym_struct] = ACTIONS(432),
    [anon_sym_trait] = ACTIONS(432),
    [anon_sym_type] = ACTIONS(432),
    [anon_sym_union] = ACTIONS(432),
    [anon_sym_unsafe] = ACTIONS(432),
    [anon_sym_use] = ACTIONS(432),
    [anon_sym_while] = ACTIONS(432),
    [anon_sym_POUND] = ACTIONS(430),
    [anon_sym_BANG] = ACTIONS(432),
    [anon_sym_EQ] = ACTIONS(432),
    [anon_sym_extern] = ACTIONS(432),
    [anon_sym_LT] = ACTIONS(432),
    [anon_sym_GT] = ACTIONS(432),
    [anon_sym_COLON_COLON] = ACTIONS(430),
    [anon_sym_AMP] = ACTIONS(432),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(430),
    [anon_sym_DOT_DOT] = ACTIONS(432),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(430),
    [anon_sym_DASH] = ACTIONS(432),
    [anon_sym_AMP_AMP] = ACTIONS(430),
    [anon_sym_PIPE_PIPE] = ACTIONS(430),
    [anon_sym_PIPE] = ACTIONS(432),
    [anon_sym_CARET] = ACTIONS(432),
    [anon_sym_EQ_EQ] = ACTIONS(430),
    [anon_sym_BANG_EQ] = ACTIONS(430),
    [anon_sym_LT_EQ] = ACTIONS(430),
    [anon_sym_GT_EQ] = ACTIONS(430),
    [anon_sym_LT_LT] = ACTIONS(432),
    [anon_sym_GT_GT] = ACTIONS(432),
    [anon_sym_SLASH] = ACTIONS(432),
    [anon_sym_PERCENT] = ACTIONS(432),
    [anon_sym_PLUS_EQ] = ACTIONS(430),
    [anon_sym_DASH_EQ] = ACTIONS(430),
    [anon_sym_STAR_EQ] = ACTIONS(430),
    [anon_sym_SLASH_EQ] = ACTIONS(430),
    [anon_sym_PERCENT_EQ] = ACTIONS(430),
    [anon_sym_AMP_EQ] = ACTIONS(430),
    [anon_sym_PIPE_EQ] = ACTIONS(430),
    [anon_sym_CARET_EQ] = ACTIONS(430),
    [anon_sym_LT_LT_EQ] = ACTIONS(430),
    [anon_sym_GT_GT_EQ] = ACTIONS(430),
    [anon_sym_yield] = ACTIONS(432),
    [anon_sym_move] = ACTIONS(432),
    [anon_sym_DOT] = ACTIONS(432),
    [sym_integer_literal] = ACTIONS(430),
    [aux_sym_string_literal_token1] = ACTIONS(430),
    [sym_char_literal] = ACTIONS(430),
    [anon_sym_true] = ACTIONS(432),
    [anon_sym_false] = ACTIONS(432),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(432),
    [sym_super] = ACTIONS(432),
    [sym_crate] = ACTIONS(432),
    [sym_metavariable] = ACTIONS(430),
    [sym_raw_string_literal] = ACTIONS(430),
    [sym_float_literal] = ACTIONS(430),
    [sym_block_comment] = ACTIONS(3),
  },
  [63] = {
    [ts_builtin_sym_end] = ACTIONS(434),
    [sym_identifier] = ACTIONS(436),
    [anon_sym_SEMI] = ACTIONS(434),
    [anon_sym_macro_rules_BANG] = ACTIONS(434),
    [anon_sym_LPAREN] = ACTIONS(434),
    [anon_sym_LBRACE] = ACTIONS(434),
    [anon_sym_RBRACE] = ACTIONS(434),
    [anon_sym_LBRACK] = ACTIONS(434),
    [anon_sym_PLUS] = ACTIONS(436),
    [anon_sym_STAR] = ACTIONS(436),
    [anon_sym_QMARK] = ACTIONS(434),
    [anon_sym_u8] = ACTIONS(436),
    [anon_sym_i8] = ACTIONS(436),
    [anon_sym_u16] = ACTIONS(436),
    [anon_sym_i16] = ACTIONS(436),
    [anon_sym_u32] = ACTIONS(436),
    [anon_sym_i32] = ACTIONS(436),
    [anon_sym_u64] = ACTIONS(436),
    [anon_sym_i64] = ACTIONS(436),
    [anon_sym_u128] = ACTIONS(436),
    [anon_sym_i128] = ACTIONS(436),
    [anon_sym_isize] = ACTIONS(436),
    [anon_sym_usize] = ACTIONS(436),
    [anon_sym_f32] = ACTIONS(436),
    [anon_sym_f64] = ACTIONS(436),
    [anon_sym_bool] = ACTIONS(436),
    [anon_sym_str] = ACTIONS(436),
    [anon_sym_char] = ACTIONS(436),
    [anon_sym_SQUOTE] = ACTIONS(436),
    [anon_sym_as] = ACTIONS(436),
    [anon_sym_async] = ACTIONS(436),
    [anon_sym_break] = ACTIONS(436),
    [anon_sym_const] = ACTIONS(436),
    [anon_sym_continue] = ACTIONS(436),
    [anon_sym_default] = ACTIONS(436),
    [anon_sym_enum] = ACTIONS(436),
    [anon_sym_fn] = ACTIONS(436),
    [anon_sym_for] = ACTIONS(436),
    [anon_sym_if] = ACTIONS(436),
    [anon_sym_impl] = ACTIONS(436),
    [anon_sym_let] = ACTIONS(436),
    [anon_sym_loop] = ACTIONS(436),
    [anon_sym_match] = ACTIONS(436),
    [anon_sym_mod] = ACTIONS(436),
    [anon_sym_pub] = ACTIONS(436),
    [anon_sym_return] = ACTIONS(436),
    [anon_sym_static] = ACTIONS(436),
    [anon_sym_struct] = ACTIONS(436),
    [anon_sym_trait] = ACTIONS(436),
    [anon_sym_type] = ACTIONS(436),
    [anon_sym_union] = ACTIONS(436),
    [anon_sym_unsafe] = ACTIONS(436),
    [anon_sym_use] = ACTIONS(436),
    [anon_sym_while] = ACTIONS(436),
    [anon_sym_POUND] = ACTIONS(434),
    [anon_sym_BANG] = ACTIONS(436),
    [anon_sym_EQ] = ACTIONS(436),
    [anon_sym_extern] = ACTIONS(436),
    [anon_sym_LT] = ACTIONS(436),
    [anon_sym_GT] = ACTIONS(436),
    [anon_sym_COLON_COLON] = ACTIONS(434),
    [anon_sym_AMP] = ACTIONS(436),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(434),
    [anon_sym_DOT_DOT] = ACTIONS(436),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(434),
    [anon_sym_DASH] = ACTIONS(436),
    [anon_sym_AMP_AMP] = ACTIONS(434),
    [anon_sym_PIPE_PIPE] = ACTIONS(434),
    [anon_sym_PIPE] = ACTIONS(436),
    [anon_sym_CARET] = ACTIONS(436),
    [anon_sym_EQ_EQ] = ACTIONS(434),
    [anon_sym_BANG_EQ] = ACTIONS(434),
    [anon_sym_LT_EQ] = ACTIONS(434),
    [anon_sym_GT_EQ] = ACTIONS(434),
    [anon_sym_LT_LT] = ACTIONS(436),
    [anon_sym_GT_GT] = ACTIONS(436),
    [anon_sym_SLASH] = ACTIONS(436),
    [anon_sym_PERCENT] = ACTIONS(436),
    [anon_sym_PLUS_EQ] = ACTIONS(434),
    [anon_sym_DASH_EQ] = ACTIONS(434),
    [anon_sym_STAR_EQ] = ACTIONS(434),
    [anon_sym_SLASH_EQ] = ACTIONS(434),
    [anon_sym_PERCENT_EQ] = ACTIONS(434),
    [anon_sym_AMP_EQ] = ACTIONS(434),
    [anon_sym_PIPE_EQ] = ACTIONS(434),
    [anon_sym_CARET_EQ] = ACTIONS(434),
    [anon_sym_LT_LT_EQ] = ACTIONS(434),
    [anon_sym_GT_GT_EQ] = ACTIONS(434),
    [anon_sym_yield] = ACTIONS(436),
    [anon_sym_move] = ACTIONS(436),
    [anon_sym_DOT] = ACTIONS(436),
    [sym_integer_literal] = ACTIONS(434),
    [aux_sym_string_literal_token1] = ACTIONS(434),
    [sym_char_literal] = ACTIONS(434),
    [anon_sym_true] = ACTIONS(436),
    [anon_sym_false] = ACTIONS(436),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(436),
    [sym_super] = ACTIONS(436),
    [sym_crate] = ACTIONS(436),
    [sym_metavariable] = ACTIONS(434),
    [sym_raw_string_literal] = ACTIONS(434),
    [sym_float_literal] = ACTIONS(434),
    [sym_block_comment] = ACTIONS(3),
  },
  [64] = {
    [ts_builtin_sym_end] = ACTIONS(438),
    [sym_identifier] = ACTIONS(440),
    [anon_sym_SEMI] = ACTIONS(438),
    [anon_sym_macro_rules_BANG] = ACTIONS(438),
    [anon_sym_LPAREN] = ACTIONS(438),
    [anon_sym_LBRACE] = ACTIONS(438),
    [anon_sym_RBRACE] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(438),
    [anon_sym_PLUS] = ACTIONS(440),
    [anon_sym_STAR] = ACTIONS(440),
    [anon_sym_QMARK] = ACTIONS(438),
    [anon_sym_u8] = ACTIONS(440),
    [anon_sym_i8] = ACTIONS(440),
    [anon_sym_u16] = ACTIONS(440),
    [anon_sym_i16] = ACTIONS(440),
    [anon_sym_u32] = ACTIONS(440),
    [anon_sym_i32] = ACTIONS(440),
    [anon_sym_u64] = ACTIONS(440),
    [anon_sym_i64] = ACTIONS(440),
    [anon_sym_u128] = ACTIONS(440),
    [anon_sym_i128] = ACTIONS(440),
    [anon_sym_isize] = ACTIONS(440),
    [anon_sym_usize] = ACTIONS(440),
    [anon_sym_f32] = ACTIONS(440),
    [anon_sym_f64] = ACTIONS(440),
    [anon_sym_bool] = ACTIONS(440),
    [anon_sym_str] = ACTIONS(440),
    [anon_sym_char] = ACTIONS(440),
    [anon_sym_SQUOTE] = ACTIONS(440),
    [anon_sym_as] = ACTIONS(440),
    [anon_sym_async] = ACTIONS(440),
    [anon_sym_break] = ACTIONS(440),
    [anon_sym_const] = ACTIONS(440),
    [anon_sym_continue] = ACTIONS(440),
    [anon_sym_default] = ACTIONS(440),
    [anon_sym_enum] = ACTIONS(440),
    [anon_sym_fn] = ACTIONS(440),
    [anon_sym_for] = ACTIONS(440),
    [anon_sym_if] = ACTIONS(440),
    [anon_sym_impl] = ACTIONS(440),
    [anon_sym_let] = ACTIONS(440),
    [anon_sym_loop] = ACTIONS(440),
    [anon_sym_match] = ACTIONS(440),
    [anon_sym_mod] = ACTIONS(440),
    [anon_sym_pub] = ACTIONS(440),
    [anon_sym_return] = ACTIONS(440),
    [anon_sym_static] = ACTIONS(440),
    [anon_sym_struct] = ACTIONS(440),
    [anon_sym_trait] = ACTIONS(440),
    [anon_sym_type] = ACTIONS(440),
    [anon_sym_union] = ACTIONS(440),
    [anon_sym_unsafe] = ACTIONS(440),
    [anon_sym_use] = ACTIONS(440),
    [anon_sym_while] = ACTIONS(440),
    [anon_sym_POUND] = ACTIONS(438),
    [anon_sym_BANG] = ACTIONS(440),
    [anon_sym_EQ] = ACTIONS(440),
    [anon_sym_extern] = ACTIONS(440),
    [anon_sym_LT] = ACTIONS(440),
    [anon_sym_GT] = ACTIONS(440),
    [anon_sym_COLON_COLON] = ACTIONS(438),
    [anon_sym_AMP] = ACTIONS(440),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(438),
    [anon_sym_DOT_DOT] = ACTIONS(440),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(438),
    [anon_sym_DASH] = ACTIONS(440),
    [anon_sym_AMP_AMP] = ACTIONS(438),
    [anon_sym_PIPE_PIPE] = ACTIONS(438),
    [anon_sym_PIPE] = ACTIONS(440),
    [anon_sym_CARET] = ACTIONS(440),
    [anon_sym_EQ_EQ] = ACTIONS(438),
    [anon_sym_BANG_EQ] = ACTIONS(438),
    [anon_sym_LT_EQ] = ACTIONS(438),
    [anon_sym_GT_EQ] = ACTIONS(438),
    [anon_sym_LT_LT] = ACTIONS(440),
    [anon_sym_GT_GT] = ACTIONS(440),
    [anon_sym_SLASH] = ACTIONS(440),
    [anon_sym_PERCENT] = ACTIONS(440),
    [anon_sym_PLUS_EQ] = ACTIONS(438),
    [anon_sym_DASH_EQ] = ACTIONS(438),
    [anon_sym_STAR_EQ] = ACTIONS(438),
    [anon_sym_SLASH_EQ] = ACTIONS(438),
    [anon_sym_PERCENT_EQ] = ACTIONS(438),
    [anon_sym_AMP_EQ] = ACTIONS(438),
    [anon_sym_PIPE_EQ] = ACTIONS(438),
    [anon_sym_CARET_EQ] = ACTIONS(438),
    [anon_sym_LT_LT_EQ] = ACTIONS(438),
    [anon_sym_GT_GT_EQ] = ACTIONS(438),
    [anon_sym_yield] = ACTIONS(440),
    [anon_sym_move] = ACTIONS(440),
    [anon_sym_DOT] = ACTIONS(440),
    [sym_integer_literal] = ACTIONS(438),
    [aux_sym_string_literal_token1] = ACTIONS(438),
    [sym_char_literal] = ACTIONS(438),
    [anon_sym_true] = ACTIONS(440),
    [anon_sym_false] = ACTIONS(440),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(440),
    [sym_super] = ACTIONS(440),
    [sym_crate] = ACTIONS(440),
    [sym_metavariable] = ACTIONS(438),
    [sym_raw_string_literal] = ACTIONS(438),
    [sym_float_literal] = ACTIONS(438),
    [sym_block_comment] = ACTIONS(3),
  },
  [65] = {
    [ts_builtin_sym_end] = ACTIONS(442),
    [sym_identifier] = ACTIONS(444),
    [anon_sym_SEMI] = ACTIONS(442),
    [anon_sym_macro_rules_BANG] = ACTIONS(442),
    [anon_sym_LPAREN] = ACTIONS(442),
    [anon_sym_LBRACE] = ACTIONS(442),
    [anon_sym_RBRACE] = ACTIONS(442),
    [anon_sym_LBRACK] = ACTIONS(442),
    [anon_sym_PLUS] = ACTIONS(444),
    [anon_sym_STAR] = ACTIONS(444),
    [anon_sym_QMARK] = ACTIONS(442),
    [anon_sym_u8] = ACTIONS(444),
    [anon_sym_i8] = ACTIONS(444),
    [anon_sym_u16] = ACTIONS(444),
    [anon_sym_i16] = ACTIONS(444),
    [anon_sym_u32] = ACTIONS(444),
    [anon_sym_i32] = ACTIONS(444),
    [anon_sym_u64] = ACTIONS(444),
    [anon_sym_i64] = ACTIONS(444),
    [anon_sym_u128] = ACTIONS(444),
    [anon_sym_i128] = ACTIONS(444),
    [anon_sym_isize] = ACTIONS(444),
    [anon_sym_usize] = ACTIONS(444),
    [anon_sym_f32] = ACTIONS(444),
    [anon_sym_f64] = ACTIONS(444),
    [anon_sym_bool] = ACTIONS(444),
    [anon_sym_str] = ACTIONS(444),
    [anon_sym_char] = ACTIONS(444),
    [anon_sym_SQUOTE] = ACTIONS(444),
    [anon_sym_as] = ACTIONS(444),
    [anon_sym_async] = ACTIONS(444),
    [anon_sym_break] = ACTIONS(444),
    [anon_sym_const] = ACTIONS(444),
    [anon_sym_continue] = ACTIONS(444),
    [anon_sym_default] = ACTIONS(444),
    [anon_sym_enum] = ACTIONS(444),
    [anon_sym_fn] = ACTIONS(444),
    [anon_sym_for] = ACTIONS(444),
    [anon_sym_if] = ACTIONS(444),
    [anon_sym_impl] = ACTIONS(444),
    [anon_sym_let] = ACTIONS(444),
    [anon_sym_loop] = ACTIONS(444),
    [anon_sym_match] = ACTIONS(444),
    [anon_sym_mod] = ACTIONS(444),
    [anon_sym_pub] = ACTIONS(444),
    [anon_sym_return] = ACTIONS(444),
    [anon_sym_static] = ACTIONS(444),
    [anon_sym_struct] = ACTIONS(444),
    [anon_sym_trait] = ACTIONS(444),
    [anon_sym_type] = ACTIONS(444),
    [anon_sym_union] = ACTIONS(444),
    [anon_sym_unsafe] = ACTIONS(444),
    [anon_sym_use] = ACTIONS(444),
    [anon_sym_while] = ACTIONS(444),
    [anon_sym_POUND] = ACTIONS(442),
    [anon_sym_BANG] = ACTIONS(444),
    [anon_sym_EQ] = ACTIONS(444),
    [anon_sym_extern] = ACTIONS(444),
    [anon_sym_LT] = ACTIONS(444),
    [anon_sym_GT] = ACTIONS(444),
    [anon_sym_COLON_COLON] = ACTIONS(442),
    [anon_sym_AMP] = ACTIONS(444),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(442),
    [anon_sym_DOT_DOT] = ACTIONS(444),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(442),
    [anon_sym_DASH] = ACTIONS(444),
    [anon_sym_AMP_AMP] = ACTIONS(442),
    [anon_sym_PIPE_PIPE] = ACTIONS(442),
    [anon_sym_PIPE] = ACTIONS(444),
    [anon_sym_CARET] = ACTIONS(444),
    [anon_sym_EQ_EQ] = ACTIONS(442),
    [anon_sym_BANG_EQ] = ACTIONS(442),
    [anon_sym_LT_EQ] = ACTIONS(442),
    [anon_sym_GT_EQ] = ACTIONS(442),
    [anon_sym_LT_LT] = ACTIONS(444),
    [anon_sym_GT_GT] = ACTIONS(444),
    [anon_sym_SLASH] = ACTIONS(444),
    [anon_sym_PERCENT] = ACTIONS(444),
    [anon_sym_PLUS_EQ] = ACTIONS(442),
    [anon_sym_DASH_EQ] = ACTIONS(442),
    [anon_sym_STAR_EQ] = ACTIONS(442),
    [anon_sym_SLASH_EQ] = ACTIONS(442),
    [anon_sym_PERCENT_EQ] = ACTIONS(442),
    [anon_sym_AMP_EQ] = ACTIONS(442),
    [anon_sym_PIPE_EQ] = ACTIONS(442),
    [anon_sym_CARET_EQ] = ACTIONS(442),
    [anon_sym_LT_LT_EQ] = ACTIONS(442),
    [anon_sym_GT_GT_EQ] = ACTIONS(442),
    [anon_sym_yield] = ACTIONS(444),
    [anon_sym_move] = ACTIONS(444),
    [anon_sym_DOT] = ACTIONS(444),
    [sym_integer_literal] = ACTIONS(442),
    [aux_sym_string_literal_token1] = ACTIONS(442),
    [sym_char_literal] = ACTIONS(442),
    [anon_sym_true] = ACTIONS(444),
    [anon_sym_false] = ACTIONS(444),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(444),
    [sym_super] = ACTIONS(444),
    [sym_crate] = ACTIONS(444),
    [sym_metavariable] = ACTIONS(442),
    [sym_raw_string_literal] = ACTIONS(442),
    [sym_float_literal] = ACTIONS(442),
    [sym_block_comment] = ACTIONS(3),
  },
  [66] = {
    [ts_builtin_sym_end] = ACTIONS(446),
    [sym_identifier] = ACTIONS(448),
    [anon_sym_SEMI] = ACTIONS(446),
    [anon_sym_macro_rules_BANG] = ACTIONS(446),
    [anon_sym_LPAREN] = ACTIONS(446),
    [anon_sym_LBRACE] = ACTIONS(446),
    [anon_sym_RBRACE] = ACTIONS(446),
    [anon_sym_LBRACK] = ACTIONS(446),
    [anon_sym_PLUS] = ACTIONS(448),
    [anon_sym_STAR] = ACTIONS(448),
    [anon_sym_QMARK] = ACTIONS(446),
    [anon_sym_u8] = ACTIONS(448),
    [anon_sym_i8] = ACTIONS(448),
    [anon_sym_u16] = ACTIONS(448),
    [anon_sym_i16] = ACTIONS(448),
    [anon_sym_u32] = ACTIONS(448),
    [anon_sym_i32] = ACTIONS(448),
    [anon_sym_u64] = ACTIONS(448),
    [anon_sym_i64] = ACTIONS(448),
    [anon_sym_u128] = ACTIONS(448),
    [anon_sym_i128] = ACTIONS(448),
    [anon_sym_isize] = ACTIONS(448),
    [anon_sym_usize] = ACTIONS(448),
    [anon_sym_f32] = ACTIONS(448),
    [anon_sym_f64] = ACTIONS(448),
    [anon_sym_bool] = ACTIONS(448),
    [anon_sym_str] = ACTIONS(448),
    [anon_sym_char] = ACTIONS(448),
    [anon_sym_SQUOTE] = ACTIONS(448),
    [anon_sym_as] = ACTIONS(448),
    [anon_sym_async] = ACTIONS(448),
    [anon_sym_break] = ACTIONS(448),
    [anon_sym_const] = ACTIONS(448),
    [anon_sym_continue] = ACTIONS(448),
    [anon_sym_default] = ACTIONS(448),
    [anon_sym_enum] = ACTIONS(448),
    [anon_sym_fn] = ACTIONS(448),
    [anon_sym_for] = ACTIONS(448),
    [anon_sym_if] = ACTIONS(448),
    [anon_sym_impl] = ACTIONS(448),
    [anon_sym_let] = ACTIONS(448),
    [anon_sym_loop] = ACTIONS(448),
    [anon_sym_match] = ACTIONS(448),
    [anon_sym_mod] = ACTIONS(448),
    [anon_sym_pub] = ACTIONS(448),
    [anon_sym_return] = ACTIONS(448),
    [anon_sym_static] = ACTIONS(448),
    [anon_sym_struct] = ACTIONS(448),
    [anon_sym_trait] = ACTIONS(448),
    [anon_sym_type] = ACTIONS(448),
    [anon_sym_union] = ACTIONS(448),
    [anon_sym_unsafe] = ACTIONS(448),
    [anon_sym_use] = ACTIONS(448),
    [anon_sym_while] = ACTIONS(448),
    [anon_sym_POUND] = ACTIONS(446),
    [anon_sym_BANG] = ACTIONS(448),
    [anon_sym_EQ] = ACTIONS(448),
    [anon_sym_extern] = ACTIONS(448),
    [anon_sym_LT] = ACTIONS(448),
    [anon_sym_GT] = ACTIONS(448),
    [anon_sym_COLON_COLON] = ACTIONS(446),
    [anon_sym_AMP] = ACTIONS(448),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(446),
    [anon_sym_DOT_DOT] = ACTIONS(448),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(446),
    [anon_sym_DASH] = ACTIONS(448),
    [anon_sym_AMP_AMP] = ACTIONS(446),
    [anon_sym_PIPE_PIPE] = ACTIONS(446),
    [anon_sym_PIPE] = ACTIONS(448),
    [anon_sym_CARET] = ACTIONS(448),
    [anon_sym_EQ_EQ] = ACTIONS(446),
    [anon_sym_BANG_EQ] = ACTIONS(446),
    [anon_sym_LT_EQ] = ACTIONS(446),
    [anon_sym_GT_EQ] = ACTIONS(446),
    [anon_sym_LT_LT] = ACTIONS(448),
    [anon_sym_GT_GT] = ACTIONS(448),
    [anon_sym_SLASH] = ACTIONS(448),
    [anon_sym_PERCENT] = ACTIONS(448),
    [anon_sym_PLUS_EQ] = ACTIONS(446),
    [anon_sym_DASH_EQ] = ACTIONS(446),
    [anon_sym_STAR_EQ] = ACTIONS(446),
    [anon_sym_SLASH_EQ] = ACTIONS(446),
    [anon_sym_PERCENT_EQ] = ACTIONS(446),
    [anon_sym_AMP_EQ] = ACTIONS(446),
    [anon_sym_PIPE_EQ] = ACTIONS(446),
    [anon_sym_CARET_EQ] = ACTIONS(446),
    [anon_sym_LT_LT_EQ] = ACTIONS(446),
    [anon_sym_GT_GT_EQ] = ACTIONS(446),
    [anon_sym_yield] = ACTIONS(448),
    [anon_sym_move] = ACTIONS(448),
    [anon_sym_DOT] = ACTIONS(448),
    [sym_integer_literal] = ACTIONS(446),
    [aux_sym_string_literal_token1] = ACTIONS(446),
    [sym_char_literal] = ACTIONS(446),
    [anon_sym_true] = ACTIONS(448),
    [anon_sym_false] = ACTIONS(448),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(448),
    [sym_super] = ACTIONS(448),
    [sym_crate] = ACTIONS(448),
    [sym_metavariable] = ACTIONS(446),
    [sym_raw_string_literal] = ACTIONS(446),
    [sym_float_literal] = ACTIONS(446),
    [sym_block_comment] = ACTIONS(3),
  },
  [67] = {
    [ts_builtin_sym_end] = ACTIONS(450),
    [sym_identifier] = ACTIONS(452),
    [anon_sym_SEMI] = ACTIONS(450),
    [anon_sym_macro_rules_BANG] = ACTIONS(450),
    [anon_sym_LPAREN] = ACTIONS(450),
    [anon_sym_LBRACE] = ACTIONS(450),
    [anon_sym_RBRACE] = ACTIONS(450),
    [anon_sym_LBRACK] = ACTIONS(450),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_STAR] = ACTIONS(452),
    [anon_sym_QMARK] = ACTIONS(450),
    [anon_sym_u8] = ACTIONS(452),
    [anon_sym_i8] = ACTIONS(452),
    [anon_sym_u16] = ACTIONS(452),
    [anon_sym_i16] = ACTIONS(452),
    [anon_sym_u32] = ACTIONS(452),
    [anon_sym_i32] = ACTIONS(452),
    [anon_sym_u64] = ACTIONS(452),
    [anon_sym_i64] = ACTIONS(452),
    [anon_sym_u128] = ACTIONS(452),
    [anon_sym_i128] = ACTIONS(452),
    [anon_sym_isize] = ACTIONS(452),
    [anon_sym_usize] = ACTIONS(452),
    [anon_sym_f32] = ACTIONS(452),
    [anon_sym_f64] = ACTIONS(452),
    [anon_sym_bool] = ACTIONS(452),
    [anon_sym_str] = ACTIONS(452),
    [anon_sym_char] = ACTIONS(452),
    [anon_sym_SQUOTE] = ACTIONS(452),
    [anon_sym_as] = ACTIONS(452),
    [anon_sym_async] = ACTIONS(452),
    [anon_sym_break] = ACTIONS(452),
    [anon_sym_const] = ACTIONS(452),
    [anon_sym_continue] = ACTIONS(452),
    [anon_sym_default] = ACTIONS(452),
    [anon_sym_enum] = ACTIONS(452),
    [anon_sym_fn] = ACTIONS(452),
    [anon_sym_for] = ACTIONS(452),
    [anon_sym_if] = ACTIONS(452),
    [anon_sym_impl] = ACTIONS(452),
    [anon_sym_let] = ACTIONS(452),
    [anon_sym_loop] = ACTIONS(452),
    [anon_sym_match] = ACTIONS(452),
    [anon_sym_mod] = ACTIONS(452),
    [anon_sym_pub] = ACTIONS(452),
    [anon_sym_return] = ACTIONS(452),
    [anon_sym_static] = ACTIONS(452),
    [anon_sym_struct] = ACTIONS(452),
    [anon_sym_trait] = ACTIONS(452),
    [anon_sym_type] = ACTIONS(452),
    [anon_sym_union] = ACTIONS(452),
    [anon_sym_unsafe] = ACTIONS(452),
    [anon_sym_use] = ACTIONS(452),
    [anon_sym_while] = ACTIONS(452),
    [anon_sym_POUND] = ACTIONS(450),
    [anon_sym_BANG] = ACTIONS(452),
    [anon_sym_EQ] = ACTIONS(452),
    [anon_sym_extern] = ACTIONS(452),
    [anon_sym_LT] = ACTIONS(452),
    [anon_sym_GT] = ACTIONS(452),
    [anon_sym_COLON_COLON] = ACTIONS(450),
    [anon_sym_AMP] = ACTIONS(452),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(450),
    [anon_sym_DOT_DOT] = ACTIONS(452),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(450),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_AMP_AMP] = ACTIONS(450),
    [anon_sym_PIPE_PIPE] = ACTIONS(450),
    [anon_sym_PIPE] = ACTIONS(452),
    [anon_sym_CARET] = ACTIONS(452),
    [anon_sym_EQ_EQ] = ACTIONS(450),
    [anon_sym_BANG_EQ] = ACTIONS(450),
    [anon_sym_LT_EQ] = ACTIONS(450),
    [anon_sym_GT_EQ] = ACTIONS(450),
    [anon_sym_LT_LT] = ACTIONS(452),
    [anon_sym_GT_GT] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(452),
    [anon_sym_PERCENT] = ACTIONS(452),
    [anon_sym_PLUS_EQ] = ACTIONS(450),
    [anon_sym_DASH_EQ] = ACTIONS(450),
    [anon_sym_STAR_EQ] = ACTIONS(450),
    [anon_sym_SLASH_EQ] = ACTIONS(450),
    [anon_sym_PERCENT_EQ] = ACTIONS(450),
    [anon_sym_AMP_EQ] = ACTIONS(450),
    [anon_sym_PIPE_EQ] = ACTIONS(450),
    [anon_sym_CARET_EQ] = ACTIONS(450),
    [anon_sym_LT_LT_EQ] = ACTIONS(450),
    [anon_sym_GT_GT_EQ] = ACTIONS(450),
    [anon_sym_yield] = ACTIONS(452),
    [anon_sym_move] = ACTIONS(452),
    [anon_sym_DOT] = ACTIONS(452),
    [sym_integer_literal] = ACTIONS(450),
    [aux_sym_string_literal_token1] = ACTIONS(450),
    [sym_char_literal] = ACTIONS(450),
    [anon_sym_true] = ACTIONS(452),
    [anon_sym_false] = ACTIONS(452),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(452),
    [sym_super] = ACTIONS(452),
    [sym_crate] = ACTIONS(452),
    [sym_metavariable] = ACTIONS(450),
    [sym_raw_string_literal] = ACTIONS(450),
    [sym_float_literal] = ACTIONS(450),
    [sym_block_comment] = ACTIONS(3),
  },
  [68] = {
    [ts_builtin_sym_end] = ACTIONS(454),
    [sym_identifier] = ACTIONS(456),
    [anon_sym_SEMI] = ACTIONS(458),
    [anon_sym_macro_rules_BANG] = ACTIONS(454),
    [anon_sym_LPAREN] = ACTIONS(458),
    [anon_sym_LBRACE] = ACTIONS(454),
    [anon_sym_RBRACE] = ACTIONS(458),
    [anon_sym_LBRACK] = ACTIONS(458),
    [anon_sym_PLUS] = ACTIONS(460),
    [anon_sym_STAR] = ACTIONS(460),
    [anon_sym_QMARK] = ACTIONS(458),
    [anon_sym_u8] = ACTIONS(456),
    [anon_sym_i8] = ACTIONS(456),
    [anon_sym_u16] = ACTIONS(456),
    [anon_sym_i16] = ACTIONS(456),
    [anon_sym_u32] = ACTIONS(456),
    [anon_sym_i32] = ACTIONS(456),
    [anon_sym_u64] = ACTIONS(456),
    [anon_sym_i64] = ACTIONS(456),
    [anon_sym_u128] = ACTIONS(456),
    [anon_sym_i128] = ACTIONS(456),
    [anon_sym_isize] = ACTIONS(456),
    [anon_sym_usize] = ACTIONS(456),
    [anon_sym_f32] = ACTIONS(456),
    [anon_sym_f64] = ACTIONS(456),
    [anon_sym_bool] = ACTIONS(456),
    [anon_sym_str] = ACTIONS(456),
    [anon_sym_char] = ACTIONS(456),
    [anon_sym_SQUOTE] = ACTIONS(456),
    [anon_sym_as] = ACTIONS(460),
    [anon_sym_async] = ACTIONS(456),
    [anon_sym_break] = ACTIONS(456),
    [anon_sym_const] = ACTIONS(456),
    [anon_sym_continue] = ACTIONS(456),
    [anon_sym_default] = ACTIONS(456),
    [anon_sym_enum] = ACTIONS(456),
    [anon_sym_fn] = ACTIONS(456),
    [anon_sym_for] = ACTIONS(456),
    [anon_sym_if] = ACTIONS(456),
    [anon_sym_impl] = ACTIONS(456),
    [anon_sym_let] = ACTIONS(456),
    [anon_sym_loop] = ACTIONS(456),
    [anon_sym_match] = ACTIONS(456),
    [anon_sym_mod] = ACTIONS(456),
    [anon_sym_pub] = ACTIONS(456),
    [anon_sym_return] = ACTIONS(456),
    [anon_sym_static] = ACTIONS(456),
    [anon_sym_struct] = ACTIONS(456),
    [anon_sym_trait] = ACTIONS(456),
    [anon_sym_type] = ACTIONS(456),
    [anon_sym_union] = ACTIONS(456),
    [anon_sym_unsafe] = ACTIONS(456),
    [anon_sym_use] = ACTIONS(456),
    [anon_sym_while] = ACTIONS(456),
    [anon_sym_POUND] = ACTIONS(454),
    [anon_sym_BANG] = ACTIONS(456),
    [anon_sym_EQ] = ACTIONS(460),
    [anon_sym_extern] = ACTIONS(456),
    [anon_sym_LT] = ACTIONS(460),
    [anon_sym_GT] = ACTIONS(460),
    [anon_sym_COLON_COLON] = ACTIONS(454),
    [anon_sym_AMP] = ACTIONS(460),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(458),
    [anon_sym_DOT_DOT] = ACTIONS(460),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(458),
    [anon_sym_DASH] = ACTIONS(460),
    [anon_sym_AMP_AMP] = ACTIONS(458),
    [anon_sym_PIPE_PIPE] = ACTIONS(458),
    [anon_sym_PIPE] = ACTIONS(460),
    [anon_sym_CARET] = ACTIONS(460),
    [anon_sym_EQ_EQ] = ACTIONS(458),
    [anon_sym_BANG_EQ] = ACTIONS(458),
    [anon_sym_LT_EQ] = ACTIONS(458),
    [anon_sym_GT_EQ] = ACTIONS(458),
    [anon_sym_LT_LT] = ACTIONS(460),
    [anon_sym_GT_GT] = ACTIONS(460),
    [anon_sym_SLASH] = ACTIONS(460),
    [anon_sym_PERCENT] = ACTIONS(460),
    [anon_sym_PLUS_EQ] = ACTIONS(458),
    [anon_sym_DASH_EQ] = ACTIONS(458),
    [anon_sym_STAR_EQ] = ACTIONS(458),
    [anon_sym_SLASH_EQ] = ACTIONS(458),
    [anon_sym_PERCENT_EQ] = ACTIONS(458),
    [anon_sym_AMP_EQ] = ACTIONS(458),
    [anon_sym_PIPE_EQ] = ACTIONS(458),
    [anon_sym_CARET_EQ] = ACTIONS(458),
    [anon_sym_LT_LT_EQ] = ACTIONS(458),
    [anon_sym_GT_GT_EQ] = ACTIONS(458),
    [anon_sym_yield] = ACTIONS(456),
    [anon_sym_move] = ACTIONS(456),
    [anon_sym_DOT] = ACTIONS(460),
    [sym_integer_literal] = ACTIONS(454),
    [aux_sym_string_literal_token1] = ACTIONS(454),
    [sym_char_literal] = ACTIONS(454),
    [anon_sym_true] = ACTIONS(456),
    [anon_sym_false] = ACTIONS(456),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(456),
    [sym_super] = ACTIONS(456),
    [sym_crate] = ACTIONS(456),
    [sym_metavariable] = ACTIONS(454),
    [sym_raw_string_literal] = ACTIONS(454),
    [sym_float_literal] = ACTIONS(454),
    [sym_block_comment] = ACTIONS(3),
  },
  [69] = {
    [ts_builtin_sym_end] = ACTIONS(462),
    [sym_identifier] = ACTIONS(464),
    [anon_sym_SEMI] = ACTIONS(462),
    [anon_sym_macro_rules_BANG] = ACTIONS(462),
    [anon_sym_LPAREN] = ACTIONS(462),
    [anon_sym_LBRACE] = ACTIONS(462),
    [anon_sym_RBRACE] = ACTIONS(462),
    [anon_sym_LBRACK] = ACTIONS(462),
    [anon_sym_PLUS] = ACTIONS(464),
    [anon_sym_STAR] = ACTIONS(464),
    [anon_sym_QMARK] = ACTIONS(462),
    [anon_sym_u8] = ACTIONS(464),
    [anon_sym_i8] = ACTIONS(464),
    [anon_sym_u16] = ACTIONS(464),
    [anon_sym_i16] = ACTIONS(464),
    [anon_sym_u32] = ACTIONS(464),
    [anon_sym_i32] = ACTIONS(464),
    [anon_sym_u64] = ACTIONS(464),
    [anon_sym_i64] = ACTIONS(464),
    [anon_sym_u128] = ACTIONS(464),
    [anon_sym_i128] = ACTIONS(464),
    [anon_sym_isize] = ACTIONS(464),
    [anon_sym_usize] = ACTIONS(464),
    [anon_sym_f32] = ACTIONS(464),
    [anon_sym_f64] = ACTIONS(464),
    [anon_sym_bool] = ACTIONS(464),
    [anon_sym_str] = ACTIONS(464),
    [anon_sym_char] = ACTIONS(464),
    [anon_sym_SQUOTE] = ACTIONS(464),
    [anon_sym_as] = ACTIONS(464),
    [anon_sym_async] = ACTIONS(464),
    [anon_sym_break] = ACTIONS(464),
    [anon_sym_const] = ACTIONS(464),
    [anon_sym_continue] = ACTIONS(464),
    [anon_sym_default] = ACTIONS(464),
    [anon_sym_enum] = ACTIONS(464),
    [anon_sym_fn] = ACTIONS(464),
    [anon_sym_for] = ACTIONS(464),
    [anon_sym_if] = ACTIONS(464),
    [anon_sym_impl] = ACTIONS(464),
    [anon_sym_let] = ACTIONS(464),
    [anon_sym_loop] = ACTIONS(464),
    [anon_sym_match] = ACTIONS(464),
    [anon_sym_mod] = ACTIONS(464),
    [anon_sym_pub] = ACTIONS(464),
    [anon_sym_return] = ACTIONS(464),
    [anon_sym_static] = ACTIONS(464),
    [anon_sym_struct] = ACTIONS(464),
    [anon_sym_trait] = ACTIONS(464),
    [anon_sym_type] = ACTIONS(464),
    [anon_sym_union] = ACTIONS(464),
    [anon_sym_unsafe] = ACTIONS(464),
    [anon_sym_use] = ACTIONS(464),
    [anon_sym_while] = ACTIONS(464),
    [anon_sym_POUND] = ACTIONS(462),
    [anon_sym_BANG] = ACTIONS(464),
    [anon_sym_EQ] = ACTIONS(464),
    [anon_sym_extern] = ACTIONS(464),
    [anon_sym_LT] = ACTIONS(464),
    [anon_sym_GT] = ACTIONS(464),
    [anon_sym_COLON_COLON] = ACTIONS(462),
    [anon_sym_AMP] = ACTIONS(464),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(462),
    [anon_sym_DOT_DOT] = ACTIONS(464),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(462),
    [anon_sym_DASH] = ACTIONS(464),
    [anon_sym_AMP_AMP] = ACTIONS(462),
    [anon_sym_PIPE_PIPE] = ACTIONS(462),
    [anon_sym_PIPE] = ACTIONS(464),
    [anon_sym_CARET] = ACTIONS(464),
    [anon_sym_EQ_EQ] = ACTIONS(462),
    [anon_sym_BANG_EQ] = ACTIONS(462),
    [anon_sym_LT_EQ] = ACTIONS(462),
    [anon_sym_GT_EQ] = ACTIONS(462),
    [anon_sym_LT_LT] = ACTIONS(464),
    [anon_sym_GT_GT] = ACTIONS(464),
    [anon_sym_SLASH] = ACTIONS(464),
    [anon_sym_PERCENT] = ACTIONS(464),
    [anon_sym_PLUS_EQ] = ACTIONS(462),
    [anon_sym_DASH_EQ] = ACTIONS(462),
    [anon_sym_STAR_EQ] = ACTIONS(462),
    [anon_sym_SLASH_EQ] = ACTIONS(462),
    [anon_sym_PERCENT_EQ] = ACTIONS(462),
    [anon_sym_AMP_EQ] = ACTIONS(462),
    [anon_sym_PIPE_EQ] = ACTIONS(462),
    [anon_sym_CARET_EQ] = ACTIONS(462),
    [anon_sym_LT_LT_EQ] = ACTIONS(462),
    [anon_sym_GT_GT_EQ] = ACTIONS(462),
    [anon_sym_yield] = ACTIONS(464),
    [anon_sym_move] = ACTIONS(464),
    [anon_sym_DOT] = ACTIONS(464),
    [sym_integer_literal] = ACTIONS(462),
    [aux_sym_string_literal_token1] = ACTIONS(462),
    [sym_char_literal] = ACTIONS(462),
    [anon_sym_true] = ACTIONS(464),
    [anon_sym_false] = ACTIONS(464),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(464),
    [sym_super] = ACTIONS(464),
    [sym_crate] = ACTIONS(464),
    [sym_metavariable] = ACTIONS(462),
    [sym_raw_string_literal] = ACTIONS(462),
    [sym_float_literal] = ACTIONS(462),
    [sym_block_comment] = ACTIONS(3),
  },
  [70] = {
    [ts_builtin_sym_end] = ACTIONS(466),
    [sym_identifier] = ACTIONS(468),
    [anon_sym_SEMI] = ACTIONS(466),
    [anon_sym_macro_rules_BANG] = ACTIONS(466),
    [anon_sym_LPAREN] = ACTIONS(466),
    [anon_sym_LBRACE] = ACTIONS(466),
    [anon_sym_RBRACE] = ACTIONS(466),
    [anon_sym_LBRACK] = ACTIONS(466),
    [anon_sym_PLUS] = ACTIONS(460),
    [anon_sym_STAR] = ACTIONS(468),
    [anon_sym_QMARK] = ACTIONS(458),
    [anon_sym_u8] = ACTIONS(468),
    [anon_sym_i8] = ACTIONS(468),
    [anon_sym_u16] = ACTIONS(468),
    [anon_sym_i16] = ACTIONS(468),
    [anon_sym_u32] = ACTIONS(468),
    [anon_sym_i32] = ACTIONS(468),
    [anon_sym_u64] = ACTIONS(468),
    [anon_sym_i64] = ACTIONS(468),
    [anon_sym_u128] = ACTIONS(468),
    [anon_sym_i128] = ACTIONS(468),
    [anon_sym_isize] = ACTIONS(468),
    [anon_sym_usize] = ACTIONS(468),
    [anon_sym_f32] = ACTIONS(468),
    [anon_sym_f64] = ACTIONS(468),
    [anon_sym_bool] = ACTIONS(468),
    [anon_sym_str] = ACTIONS(468),
    [anon_sym_char] = ACTIONS(468),
    [anon_sym_SQUOTE] = ACTIONS(468),
    [anon_sym_as] = ACTIONS(460),
    [anon_sym_async] = ACTIONS(468),
    [anon_sym_break] = ACTIONS(468),
    [anon_sym_const] = ACTIONS(468),
    [anon_sym_continue] = ACTIONS(468),
    [anon_sym_default] = ACTIONS(468),
    [anon_sym_enum] = ACTIONS(468),
    [anon_sym_fn] = ACTIONS(468),
    [anon_sym_for] = ACTIONS(468),
    [anon_sym_if] = ACTIONS(468),
    [anon_sym_impl] = ACTIONS(468),
    [anon_sym_let] = ACTIONS(468),
    [anon_sym_loop] = ACTIONS(468),
    [anon_sym_match] = ACTIONS(468),
    [anon_sym_mod] = ACTIONS(468),
    [anon_sym_pub] = ACTIONS(468),
    [anon_sym_return] = ACTIONS(468),
    [anon_sym_static] = ACTIONS(468),
    [anon_sym_struct] = ACTIONS(468),
    [anon_sym_trait] = ACTIONS(468),
    [anon_sym_type] = ACTIONS(468),
    [anon_sym_union] = ACTIONS(468),
    [anon_sym_unsafe] = ACTIONS(468),
    [anon_sym_use] = ACTIONS(468),
    [anon_sym_while] = ACTIONS(468),
    [anon_sym_POUND] = ACTIONS(466),
    [anon_sym_BANG] = ACTIONS(468),
    [anon_sym_EQ] = ACTIONS(460),
    [anon_sym_extern] = ACTIONS(468),
    [anon_sym_LT] = ACTIONS(468),
    [anon_sym_GT] = ACTIONS(460),
    [anon_sym_COLON_COLON] = ACTIONS(466),
    [anon_sym_AMP] = ACTIONS(468),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(458),
    [anon_sym_DOT_DOT] = ACTIONS(468),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(458),
    [anon_sym_DASH] = ACTIONS(468),
    [anon_sym_AMP_AMP] = ACTIONS(458),
    [anon_sym_PIPE_PIPE] = ACTIONS(458),
    [anon_sym_PIPE] = ACTIONS(468),
    [anon_sym_CARET] = ACTIONS(460),
    [anon_sym_EQ_EQ] = ACTIONS(458),
    [anon_sym_BANG_EQ] = ACTIONS(458),
    [anon_sym_LT_EQ] = ACTIONS(458),
    [anon_sym_GT_EQ] = ACTIONS(458),
    [anon_sym_LT_LT] = ACTIONS(460),
    [anon_sym_GT_GT] = ACTIONS(460),
    [anon_sym_SLASH] = ACTIONS(460),
    [anon_sym_PERCENT] = ACTIONS(460),
    [anon_sym_PLUS_EQ] = ACTIONS(458),
    [anon_sym_DASH_EQ] = ACTIONS(458),
    [anon_sym_STAR_EQ] = ACTIONS(458),
    [anon_sym_SLASH_EQ] = ACTIONS(458),
    [anon_sym_PERCENT_EQ] = ACTIONS(458),
    [anon_sym_AMP_EQ] = ACTIONS(458),
    [anon_sym_PIPE_EQ] = ACTIONS(458),
    [anon_sym_CARET_EQ] = ACTIONS(458),
    [anon_sym_LT_LT_EQ] = ACTIONS(458),
    [anon_sym_GT_GT_EQ] = ACTIONS(458),
    [anon_sym_yield] = ACTIONS(468),
    [anon_sym_move] = ACTIONS(468),
    [anon_sym_DOT] = ACTIONS(460),
    [sym_integer_literal] = ACTIONS(466),
    [aux_sym_string_literal_token1] = ACTIONS(466),
    [sym_char_literal] = ACTIONS(466),
    [anon_sym_true] = ACTIONS(468),
    [anon_sym_false] = ACTIONS(468),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(468),
    [sym_super] = ACTIONS(468),
    [sym_crate] = ACTIONS(468),
    [sym_metavariable] = ACTIONS(466),
    [sym_raw_string_literal] = ACTIONS(466),
    [sym_float_literal] = ACTIONS(466),
    [sym_block_comment] = ACTIONS(3),
  },
  [71] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1243),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [aux_sym_tuple_expression_repeat1] = STATE(87),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_RPAREN] = ACTIONS(470),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [72] = {
    [ts_builtin_sym_end] = ACTIONS(472),
    [sym_identifier] = ACTIONS(474),
    [anon_sym_SEMI] = ACTIONS(472),
    [anon_sym_macro_rules_BANG] = ACTIONS(472),
    [anon_sym_LPAREN] = ACTIONS(472),
    [anon_sym_LBRACE] = ACTIONS(472),
    [anon_sym_RBRACE] = ACTIONS(472),
    [anon_sym_LBRACK] = ACTIONS(472),
    [anon_sym_PLUS] = ACTIONS(474),
    [anon_sym_STAR] = ACTIONS(474),
    [anon_sym_QMARK] = ACTIONS(472),
    [anon_sym_u8] = ACTIONS(474),
    [anon_sym_i8] = ACTIONS(474),
    [anon_sym_u16] = ACTIONS(474),
    [anon_sym_i16] = ACTIONS(474),
    [anon_sym_u32] = ACTIONS(474),
    [anon_sym_i32] = ACTIONS(474),
    [anon_sym_u64] = ACTIONS(474),
    [anon_sym_i64] = ACTIONS(474),
    [anon_sym_u128] = ACTIONS(474),
    [anon_sym_i128] = ACTIONS(474),
    [anon_sym_isize] = ACTIONS(474),
    [anon_sym_usize] = ACTIONS(474),
    [anon_sym_f32] = ACTIONS(474),
    [anon_sym_f64] = ACTIONS(474),
    [anon_sym_bool] = ACTIONS(474),
    [anon_sym_str] = ACTIONS(474),
    [anon_sym_char] = ACTIONS(474),
    [anon_sym_SQUOTE] = ACTIONS(474),
    [anon_sym_as] = ACTIONS(474),
    [anon_sym_async] = ACTIONS(474),
    [anon_sym_break] = ACTIONS(474),
    [anon_sym_const] = ACTIONS(474),
    [anon_sym_continue] = ACTIONS(474),
    [anon_sym_default] = ACTIONS(474),
    [anon_sym_enum] = ACTIONS(474),
    [anon_sym_fn] = ACTIONS(474),
    [anon_sym_for] = ACTIONS(474),
    [anon_sym_if] = ACTIONS(474),
    [anon_sym_impl] = ACTIONS(474),
    [anon_sym_let] = ACTIONS(474),
    [anon_sym_loop] = ACTIONS(474),
    [anon_sym_match] = ACTIONS(474),
    [anon_sym_mod] = ACTIONS(474),
    [anon_sym_pub] = ACTIONS(474),
    [anon_sym_return] = ACTIONS(474),
    [anon_sym_static] = ACTIONS(474),
    [anon_sym_struct] = ACTIONS(474),
    [anon_sym_trait] = ACTIONS(474),
    [anon_sym_type] = ACTIONS(474),
    [anon_sym_union] = ACTIONS(474),
    [anon_sym_unsafe] = ACTIONS(474),
    [anon_sym_use] = ACTIONS(474),
    [anon_sym_while] = ACTIONS(474),
    [anon_sym_POUND] = ACTIONS(472),
    [anon_sym_BANG] = ACTIONS(474),
    [anon_sym_EQ] = ACTIONS(474),
    [anon_sym_extern] = ACTIONS(474),
    [anon_sym_LT] = ACTIONS(474),
    [anon_sym_GT] = ACTIONS(474),
    [anon_sym_COLON_COLON] = ACTIONS(472),
    [anon_sym_AMP] = ACTIONS(474),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(472),
    [anon_sym_DOT_DOT] = ACTIONS(474),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(472),
    [anon_sym_DASH] = ACTIONS(474),
    [anon_sym_AMP_AMP] = ACTIONS(472),
    [anon_sym_PIPE_PIPE] = ACTIONS(472),
    [anon_sym_PIPE] = ACTIONS(474),
    [anon_sym_CARET] = ACTIONS(474),
    [anon_sym_EQ_EQ] = ACTIONS(472),
    [anon_sym_BANG_EQ] = ACTIONS(472),
    [anon_sym_LT_EQ] = ACTIONS(472),
    [anon_sym_GT_EQ] = ACTIONS(472),
    [anon_sym_LT_LT] = ACTIONS(474),
    [anon_sym_GT_GT] = ACTIONS(474),
    [anon_sym_SLASH] = ACTIONS(474),
    [anon_sym_PERCENT] = ACTIONS(474),
    [anon_sym_PLUS_EQ] = ACTIONS(472),
    [anon_sym_DASH_EQ] = ACTIONS(472),
    [anon_sym_STAR_EQ] = ACTIONS(472),
    [anon_sym_SLASH_EQ] = ACTIONS(472),
    [anon_sym_PERCENT_EQ] = ACTIONS(472),
    [anon_sym_AMP_EQ] = ACTIONS(472),
    [anon_sym_PIPE_EQ] = ACTIONS(472),
    [anon_sym_CARET_EQ] = ACTIONS(472),
    [anon_sym_LT_LT_EQ] = ACTIONS(472),
    [anon_sym_GT_GT_EQ] = ACTIONS(472),
    [anon_sym_yield] = ACTIONS(474),
    [anon_sym_move] = ACTIONS(474),
    [anon_sym_DOT] = ACTIONS(474),
    [sym_integer_literal] = ACTIONS(472),
    [aux_sym_string_literal_token1] = ACTIONS(472),
    [sym_char_literal] = ACTIONS(472),
    [anon_sym_true] = ACTIONS(474),
    [anon_sym_false] = ACTIONS(474),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(474),
    [sym_super] = ACTIONS(474),
    [sym_crate] = ACTIONS(474),
    [sym_metavariable] = ACTIONS(472),
    [sym_raw_string_literal] = ACTIONS(472),
    [sym_float_literal] = ACTIONS(472),
    [sym_block_comment] = ACTIONS(3),
  },
  [73] = {
    [ts_builtin_sym_end] = ACTIONS(476),
    [sym_identifier] = ACTIONS(478),
    [anon_sym_SEMI] = ACTIONS(476),
    [anon_sym_macro_rules_BANG] = ACTIONS(476),
    [anon_sym_LPAREN] = ACTIONS(476),
    [anon_sym_LBRACE] = ACTIONS(476),
    [anon_sym_RBRACE] = ACTIONS(476),
    [anon_sym_LBRACK] = ACTIONS(476),
    [anon_sym_PLUS] = ACTIONS(478),
    [anon_sym_STAR] = ACTIONS(478),
    [anon_sym_QMARK] = ACTIONS(476),
    [anon_sym_u8] = ACTIONS(478),
    [anon_sym_i8] = ACTIONS(478),
    [anon_sym_u16] = ACTIONS(478),
    [anon_sym_i16] = ACTIONS(478),
    [anon_sym_u32] = ACTIONS(478),
    [anon_sym_i32] = ACTIONS(478),
    [anon_sym_u64] = ACTIONS(478),
    [anon_sym_i64] = ACTIONS(478),
    [anon_sym_u128] = ACTIONS(478),
    [anon_sym_i128] = ACTIONS(478),
    [anon_sym_isize] = ACTIONS(478),
    [anon_sym_usize] = ACTIONS(478),
    [anon_sym_f32] = ACTIONS(478),
    [anon_sym_f64] = ACTIONS(478),
    [anon_sym_bool] = ACTIONS(478),
    [anon_sym_str] = ACTIONS(478),
    [anon_sym_char] = ACTIONS(478),
    [anon_sym_SQUOTE] = ACTIONS(478),
    [anon_sym_as] = ACTIONS(478),
    [anon_sym_async] = ACTIONS(478),
    [anon_sym_break] = ACTIONS(478),
    [anon_sym_const] = ACTIONS(478),
    [anon_sym_continue] = ACTIONS(478),
    [anon_sym_default] = ACTIONS(478),
    [anon_sym_enum] = ACTIONS(478),
    [anon_sym_fn] = ACTIONS(478),
    [anon_sym_for] = ACTIONS(478),
    [anon_sym_if] = ACTIONS(478),
    [anon_sym_impl] = ACTIONS(478),
    [anon_sym_let] = ACTIONS(478),
    [anon_sym_loop] = ACTIONS(478),
    [anon_sym_match] = ACTIONS(478),
    [anon_sym_mod] = ACTIONS(478),
    [anon_sym_pub] = ACTIONS(478),
    [anon_sym_return] = ACTIONS(478),
    [anon_sym_static] = ACTIONS(478),
    [anon_sym_struct] = ACTIONS(478),
    [anon_sym_trait] = ACTIONS(478),
    [anon_sym_type] = ACTIONS(478),
    [anon_sym_union] = ACTIONS(478),
    [anon_sym_unsafe] = ACTIONS(478),
    [anon_sym_use] = ACTIONS(478),
    [anon_sym_while] = ACTIONS(478),
    [anon_sym_POUND] = ACTIONS(476),
    [anon_sym_BANG] = ACTIONS(478),
    [anon_sym_EQ] = ACTIONS(478),
    [anon_sym_extern] = ACTIONS(478),
    [anon_sym_LT] = ACTIONS(478),
    [anon_sym_GT] = ACTIONS(478),
    [anon_sym_COLON_COLON] = ACTIONS(476),
    [anon_sym_AMP] = ACTIONS(478),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(476),
    [anon_sym_DOT_DOT] = ACTIONS(478),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(476),
    [anon_sym_DASH] = ACTIONS(478),
    [anon_sym_AMP_AMP] = ACTIONS(476),
    [anon_sym_PIPE_PIPE] = ACTIONS(476),
    [anon_sym_PIPE] = ACTIONS(478),
    [anon_sym_CARET] = ACTIONS(478),
    [anon_sym_EQ_EQ] = ACTIONS(476),
    [anon_sym_BANG_EQ] = ACTIONS(476),
    [anon_sym_LT_EQ] = ACTIONS(476),
    [anon_sym_GT_EQ] = ACTIONS(476),
    [anon_sym_LT_LT] = ACTIONS(478),
    [anon_sym_GT_GT] = ACTIONS(478),
    [anon_sym_SLASH] = ACTIONS(478),
    [anon_sym_PERCENT] = ACTIONS(478),
    [anon_sym_PLUS_EQ] = ACTIONS(476),
    [anon_sym_DASH_EQ] = ACTIONS(476),
    [anon_sym_STAR_EQ] = ACTIONS(476),
    [anon_sym_SLASH_EQ] = ACTIONS(476),
    [anon_sym_PERCENT_EQ] = ACTIONS(476),
    [anon_sym_AMP_EQ] = ACTIONS(476),
    [anon_sym_PIPE_EQ] = ACTIONS(476),
    [anon_sym_CARET_EQ] = ACTIONS(476),
    [anon_sym_LT_LT_EQ] = ACTIONS(476),
    [anon_sym_GT_GT_EQ] = ACTIONS(476),
    [anon_sym_yield] = ACTIONS(478),
    [anon_sym_move] = ACTIONS(478),
    [anon_sym_DOT] = ACTIONS(478),
    [sym_integer_literal] = ACTIONS(476),
    [aux_sym_string_literal_token1] = ACTIONS(476),
    [sym_char_literal] = ACTIONS(476),
    [anon_sym_true] = ACTIONS(478),
    [anon_sym_false] = ACTIONS(478),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(478),
    [sym_super] = ACTIONS(478),
    [sym_crate] = ACTIONS(478),
    [sym_metavariable] = ACTIONS(476),
    [sym_raw_string_literal] = ACTIONS(476),
    [sym_float_literal] = ACTIONS(476),
    [sym_block_comment] = ACTIONS(3),
  },
  [74] = {
    [ts_builtin_sym_end] = ACTIONS(480),
    [sym_identifier] = ACTIONS(482),
    [anon_sym_SEMI] = ACTIONS(480),
    [anon_sym_macro_rules_BANG] = ACTIONS(480),
    [anon_sym_LPAREN] = ACTIONS(480),
    [anon_sym_LBRACE] = ACTIONS(480),
    [anon_sym_RBRACE] = ACTIONS(480),
    [anon_sym_LBRACK] = ACTIONS(480),
    [anon_sym_PLUS] = ACTIONS(482),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_QMARK] = ACTIONS(480),
    [anon_sym_u8] = ACTIONS(482),
    [anon_sym_i8] = ACTIONS(482),
    [anon_sym_u16] = ACTIONS(482),
    [anon_sym_i16] = ACTIONS(482),
    [anon_sym_u32] = ACTIONS(482),
    [anon_sym_i32] = ACTIONS(482),
    [anon_sym_u64] = ACTIONS(482),
    [anon_sym_i64] = ACTIONS(482),
    [anon_sym_u128] = ACTIONS(482),
    [anon_sym_i128] = ACTIONS(482),
    [anon_sym_isize] = ACTIONS(482),
    [anon_sym_usize] = ACTIONS(482),
    [anon_sym_f32] = ACTIONS(482),
    [anon_sym_f64] = ACTIONS(482),
    [anon_sym_bool] = ACTIONS(482),
    [anon_sym_str] = ACTIONS(482),
    [anon_sym_char] = ACTIONS(482),
    [anon_sym_SQUOTE] = ACTIONS(482),
    [anon_sym_as] = ACTIONS(482),
    [anon_sym_async] = ACTIONS(482),
    [anon_sym_break] = ACTIONS(482),
    [anon_sym_const] = ACTIONS(482),
    [anon_sym_continue] = ACTIONS(482),
    [anon_sym_default] = ACTIONS(482),
    [anon_sym_enum] = ACTIONS(482),
    [anon_sym_fn] = ACTIONS(482),
    [anon_sym_for] = ACTIONS(482),
    [anon_sym_if] = ACTIONS(482),
    [anon_sym_impl] = ACTIONS(482),
    [anon_sym_let] = ACTIONS(482),
    [anon_sym_loop] = ACTIONS(482),
    [anon_sym_match] = ACTIONS(482),
    [anon_sym_mod] = ACTIONS(482),
    [anon_sym_pub] = ACTIONS(482),
    [anon_sym_return] = ACTIONS(482),
    [anon_sym_static] = ACTIONS(482),
    [anon_sym_struct] = ACTIONS(482),
    [anon_sym_trait] = ACTIONS(482),
    [anon_sym_type] = ACTIONS(482),
    [anon_sym_union] = ACTIONS(482),
    [anon_sym_unsafe] = ACTIONS(482),
    [anon_sym_use] = ACTIONS(482),
    [anon_sym_while] = ACTIONS(482),
    [anon_sym_POUND] = ACTIONS(480),
    [anon_sym_BANG] = ACTIONS(482),
    [anon_sym_EQ] = ACTIONS(482),
    [anon_sym_extern] = ACTIONS(482),
    [anon_sym_LT] = ACTIONS(482),
    [anon_sym_GT] = ACTIONS(482),
    [anon_sym_COLON_COLON] = ACTIONS(480),
    [anon_sym_AMP] = ACTIONS(482),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(480),
    [anon_sym_DOT_DOT] = ACTIONS(482),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(480),
    [anon_sym_DASH] = ACTIONS(482),
    [anon_sym_AMP_AMP] = ACTIONS(480),
    [anon_sym_PIPE_PIPE] = ACTIONS(480),
    [anon_sym_PIPE] = ACTIONS(482),
    [anon_sym_CARET] = ACTIONS(482),
    [anon_sym_EQ_EQ] = ACTIONS(480),
    [anon_sym_BANG_EQ] = ACTIONS(480),
    [anon_sym_LT_EQ] = ACTIONS(480),
    [anon_sym_GT_EQ] = ACTIONS(480),
    [anon_sym_LT_LT] = ACTIONS(482),
    [anon_sym_GT_GT] = ACTIONS(482),
    [anon_sym_SLASH] = ACTIONS(482),
    [anon_sym_PERCENT] = ACTIONS(482),
    [anon_sym_PLUS_EQ] = ACTIONS(480),
    [anon_sym_DASH_EQ] = ACTIONS(480),
    [anon_sym_STAR_EQ] = ACTIONS(480),
    [anon_sym_SLASH_EQ] = ACTIONS(480),
    [anon_sym_PERCENT_EQ] = ACTIONS(480),
    [anon_sym_AMP_EQ] = ACTIONS(480),
    [anon_sym_PIPE_EQ] = ACTIONS(480),
    [anon_sym_CARET_EQ] = ACTIONS(480),
    [anon_sym_LT_LT_EQ] = ACTIONS(480),
    [anon_sym_GT_GT_EQ] = ACTIONS(480),
    [anon_sym_yield] = ACTIONS(482),
    [anon_sym_move] = ACTIONS(482),
    [anon_sym_DOT] = ACTIONS(482),
    [sym_integer_literal] = ACTIONS(480),
    [aux_sym_string_literal_token1] = ACTIONS(480),
    [sym_char_literal] = ACTIONS(480),
    [anon_sym_true] = ACTIONS(482),
    [anon_sym_false] = ACTIONS(482),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(482),
    [sym_super] = ACTIONS(482),
    [sym_crate] = ACTIONS(482),
    [sym_metavariable] = ACTIONS(480),
    [sym_raw_string_literal] = ACTIONS(480),
    [sym_float_literal] = ACTIONS(480),
    [sym_block_comment] = ACTIONS(3),
  },
  [75] = {
    [ts_builtin_sym_end] = ACTIONS(484),
    [sym_identifier] = ACTIONS(486),
    [anon_sym_SEMI] = ACTIONS(484),
    [anon_sym_macro_rules_BANG] = ACTIONS(484),
    [anon_sym_LPAREN] = ACTIONS(484),
    [anon_sym_LBRACE] = ACTIONS(484),
    [anon_sym_RBRACE] = ACTIONS(484),
    [anon_sym_LBRACK] = ACTIONS(484),
    [anon_sym_PLUS] = ACTIONS(486),
    [anon_sym_STAR] = ACTIONS(486),
    [anon_sym_QMARK] = ACTIONS(484),
    [anon_sym_u8] = ACTIONS(486),
    [anon_sym_i8] = ACTIONS(486),
    [anon_sym_u16] = ACTIONS(486),
    [anon_sym_i16] = ACTIONS(486),
    [anon_sym_u32] = ACTIONS(486),
    [anon_sym_i32] = ACTIONS(486),
    [anon_sym_u64] = ACTIONS(486),
    [anon_sym_i64] = ACTIONS(486),
    [anon_sym_u128] = ACTIONS(486),
    [anon_sym_i128] = ACTIONS(486),
    [anon_sym_isize] = ACTIONS(486),
    [anon_sym_usize] = ACTIONS(486),
    [anon_sym_f32] = ACTIONS(486),
    [anon_sym_f64] = ACTIONS(486),
    [anon_sym_bool] = ACTIONS(486),
    [anon_sym_str] = ACTIONS(486),
    [anon_sym_char] = ACTIONS(486),
    [anon_sym_SQUOTE] = ACTIONS(486),
    [anon_sym_as] = ACTIONS(486),
    [anon_sym_async] = ACTIONS(486),
    [anon_sym_break] = ACTIONS(486),
    [anon_sym_const] = ACTIONS(486),
    [anon_sym_continue] = ACTIONS(486),
    [anon_sym_default] = ACTIONS(486),
    [anon_sym_enum] = ACTIONS(486),
    [anon_sym_fn] = ACTIONS(486),
    [anon_sym_for] = ACTIONS(486),
    [anon_sym_if] = ACTIONS(486),
    [anon_sym_impl] = ACTIONS(486),
    [anon_sym_let] = ACTIONS(486),
    [anon_sym_loop] = ACTIONS(486),
    [anon_sym_match] = ACTIONS(486),
    [anon_sym_mod] = ACTIONS(486),
    [anon_sym_pub] = ACTIONS(486),
    [anon_sym_return] = ACTIONS(486),
    [anon_sym_static] = ACTIONS(486),
    [anon_sym_struct] = ACTIONS(486),
    [anon_sym_trait] = ACTIONS(486),
    [anon_sym_type] = ACTIONS(486),
    [anon_sym_union] = ACTIONS(486),
    [anon_sym_unsafe] = ACTIONS(486),
    [anon_sym_use] = ACTIONS(486),
    [anon_sym_while] = ACTIONS(486),
    [anon_sym_POUND] = ACTIONS(484),
    [anon_sym_BANG] = ACTIONS(486),
    [anon_sym_EQ] = ACTIONS(486),
    [anon_sym_extern] = ACTIONS(486),
    [anon_sym_LT] = ACTIONS(486),
    [anon_sym_GT] = ACTIONS(486),
    [anon_sym_COLON_COLON] = ACTIONS(484),
    [anon_sym_AMP] = ACTIONS(486),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(484),
    [anon_sym_DOT_DOT] = ACTIONS(486),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(484),
    [anon_sym_DASH] = ACTIONS(486),
    [anon_sym_AMP_AMP] = ACTIONS(484),
    [anon_sym_PIPE_PIPE] = ACTIONS(484),
    [anon_sym_PIPE] = ACTIONS(486),
    [anon_sym_CARET] = ACTIONS(486),
    [anon_sym_EQ_EQ] = ACTIONS(484),
    [anon_sym_BANG_EQ] = ACTIONS(484),
    [anon_sym_LT_EQ] = ACTIONS(484),
    [anon_sym_GT_EQ] = ACTIONS(484),
    [anon_sym_LT_LT] = ACTIONS(486),
    [anon_sym_GT_GT] = ACTIONS(486),
    [anon_sym_SLASH] = ACTIONS(486),
    [anon_sym_PERCENT] = ACTIONS(486),
    [anon_sym_PLUS_EQ] = ACTIONS(484),
    [anon_sym_DASH_EQ] = ACTIONS(484),
    [anon_sym_STAR_EQ] = ACTIONS(484),
    [anon_sym_SLASH_EQ] = ACTIONS(484),
    [anon_sym_PERCENT_EQ] = ACTIONS(484),
    [anon_sym_AMP_EQ] = ACTIONS(484),
    [anon_sym_PIPE_EQ] = ACTIONS(484),
    [anon_sym_CARET_EQ] = ACTIONS(484),
    [anon_sym_LT_LT_EQ] = ACTIONS(484),
    [anon_sym_GT_GT_EQ] = ACTIONS(484),
    [anon_sym_yield] = ACTIONS(486),
    [anon_sym_move] = ACTIONS(486),
    [anon_sym_DOT] = ACTIONS(486),
    [sym_integer_literal] = ACTIONS(484),
    [aux_sym_string_literal_token1] = ACTIONS(484),
    [sym_char_literal] = ACTIONS(484),
    [anon_sym_true] = ACTIONS(486),
    [anon_sym_false] = ACTIONS(486),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(486),
    [sym_super] = ACTIONS(486),
    [sym_crate] = ACTIONS(486),
    [sym_metavariable] = ACTIONS(484),
    [sym_raw_string_literal] = ACTIONS(484),
    [sym_float_literal] = ACTIONS(484),
    [sym_block_comment] = ACTIONS(3),
  },
  [76] = {
    [ts_builtin_sym_end] = ACTIONS(488),
    [sym_identifier] = ACTIONS(490),
    [anon_sym_SEMI] = ACTIONS(488),
    [anon_sym_macro_rules_BANG] = ACTIONS(488),
    [anon_sym_LPAREN] = ACTIONS(488),
    [anon_sym_LBRACE] = ACTIONS(488),
    [anon_sym_RBRACE] = ACTIONS(488),
    [anon_sym_LBRACK] = ACTIONS(488),
    [anon_sym_PLUS] = ACTIONS(490),
    [anon_sym_STAR] = ACTIONS(490),
    [anon_sym_QMARK] = ACTIONS(488),
    [anon_sym_u8] = ACTIONS(490),
    [anon_sym_i8] = ACTIONS(490),
    [anon_sym_u16] = ACTIONS(490),
    [anon_sym_i16] = ACTIONS(490),
    [anon_sym_u32] = ACTIONS(490),
    [anon_sym_i32] = ACTIONS(490),
    [anon_sym_u64] = ACTIONS(490),
    [anon_sym_i64] = ACTIONS(490),
    [anon_sym_u128] = ACTIONS(490),
    [anon_sym_i128] = ACTIONS(490),
    [anon_sym_isize] = ACTIONS(490),
    [anon_sym_usize] = ACTIONS(490),
    [anon_sym_f32] = ACTIONS(490),
    [anon_sym_f64] = ACTIONS(490),
    [anon_sym_bool] = ACTIONS(490),
    [anon_sym_str] = ACTIONS(490),
    [anon_sym_char] = ACTIONS(490),
    [anon_sym_SQUOTE] = ACTIONS(490),
    [anon_sym_as] = ACTIONS(490),
    [anon_sym_async] = ACTIONS(490),
    [anon_sym_break] = ACTIONS(490),
    [anon_sym_const] = ACTIONS(490),
    [anon_sym_continue] = ACTIONS(490),
    [anon_sym_default] = ACTIONS(490),
    [anon_sym_enum] = ACTIONS(490),
    [anon_sym_fn] = ACTIONS(490),
    [anon_sym_for] = ACTIONS(490),
    [anon_sym_if] = ACTIONS(490),
    [anon_sym_impl] = ACTIONS(490),
    [anon_sym_let] = ACTIONS(490),
    [anon_sym_loop] = ACTIONS(490),
    [anon_sym_match] = ACTIONS(490),
    [anon_sym_mod] = ACTIONS(490),
    [anon_sym_pub] = ACTIONS(490),
    [anon_sym_return] = ACTIONS(490),
    [anon_sym_static] = ACTIONS(490),
    [anon_sym_struct] = ACTIONS(490),
    [anon_sym_trait] = ACTIONS(490),
    [anon_sym_type] = ACTIONS(490),
    [anon_sym_union] = ACTIONS(490),
    [anon_sym_unsafe] = ACTIONS(490),
    [anon_sym_use] = ACTIONS(490),
    [anon_sym_while] = ACTIONS(490),
    [anon_sym_POUND] = ACTIONS(488),
    [anon_sym_BANG] = ACTIONS(490),
    [anon_sym_EQ] = ACTIONS(490),
    [anon_sym_extern] = ACTIONS(490),
    [anon_sym_LT] = ACTIONS(490),
    [anon_sym_GT] = ACTIONS(490),
    [anon_sym_COLON_COLON] = ACTIONS(488),
    [anon_sym_AMP] = ACTIONS(490),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(488),
    [anon_sym_DOT_DOT] = ACTIONS(490),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(488),
    [anon_sym_DASH] = ACTIONS(490),
    [anon_sym_AMP_AMP] = ACTIONS(488),
    [anon_sym_PIPE_PIPE] = ACTIONS(488),
    [anon_sym_PIPE] = ACTIONS(490),
    [anon_sym_CARET] = ACTIONS(490),
    [anon_sym_EQ_EQ] = ACTIONS(488),
    [anon_sym_BANG_EQ] = ACTIONS(488),
    [anon_sym_LT_EQ] = ACTIONS(488),
    [anon_sym_GT_EQ] = ACTIONS(488),
    [anon_sym_LT_LT] = ACTIONS(490),
    [anon_sym_GT_GT] = ACTIONS(490),
    [anon_sym_SLASH] = ACTIONS(490),
    [anon_sym_PERCENT] = ACTIONS(490),
    [anon_sym_PLUS_EQ] = ACTIONS(488),
    [anon_sym_DASH_EQ] = ACTIONS(488),
    [anon_sym_STAR_EQ] = ACTIONS(488),
    [anon_sym_SLASH_EQ] = ACTIONS(488),
    [anon_sym_PERCENT_EQ] = ACTIONS(488),
    [anon_sym_AMP_EQ] = ACTIONS(488),
    [anon_sym_PIPE_EQ] = ACTIONS(488),
    [anon_sym_CARET_EQ] = ACTIONS(488),
    [anon_sym_LT_LT_EQ] = ACTIONS(488),
    [anon_sym_GT_GT_EQ] = ACTIONS(488),
    [anon_sym_yield] = ACTIONS(490),
    [anon_sym_move] = ACTIONS(490),
    [anon_sym_DOT] = ACTIONS(490),
    [sym_integer_literal] = ACTIONS(488),
    [aux_sym_string_literal_token1] = ACTIONS(488),
    [sym_char_literal] = ACTIONS(488),
    [anon_sym_true] = ACTIONS(490),
    [anon_sym_false] = ACTIONS(490),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(490),
    [sym_super] = ACTIONS(490),
    [sym_crate] = ACTIONS(490),
    [sym_metavariable] = ACTIONS(488),
    [sym_raw_string_literal] = ACTIONS(488),
    [sym_float_literal] = ACTIONS(488),
    [sym_block_comment] = ACTIONS(3),
  },
  [77] = {
    [ts_builtin_sym_end] = ACTIONS(492),
    [sym_identifier] = ACTIONS(494),
    [anon_sym_SEMI] = ACTIONS(492),
    [anon_sym_macro_rules_BANG] = ACTIONS(492),
    [anon_sym_LPAREN] = ACTIONS(492),
    [anon_sym_LBRACE] = ACTIONS(492),
    [anon_sym_RBRACE] = ACTIONS(492),
    [anon_sym_LBRACK] = ACTIONS(492),
    [anon_sym_PLUS] = ACTIONS(494),
    [anon_sym_STAR] = ACTIONS(494),
    [anon_sym_QMARK] = ACTIONS(492),
    [anon_sym_u8] = ACTIONS(494),
    [anon_sym_i8] = ACTIONS(494),
    [anon_sym_u16] = ACTIONS(494),
    [anon_sym_i16] = ACTIONS(494),
    [anon_sym_u32] = ACTIONS(494),
    [anon_sym_i32] = ACTIONS(494),
    [anon_sym_u64] = ACTIONS(494),
    [anon_sym_i64] = ACTIONS(494),
    [anon_sym_u128] = ACTIONS(494),
    [anon_sym_i128] = ACTIONS(494),
    [anon_sym_isize] = ACTIONS(494),
    [anon_sym_usize] = ACTIONS(494),
    [anon_sym_f32] = ACTIONS(494),
    [anon_sym_f64] = ACTIONS(494),
    [anon_sym_bool] = ACTIONS(494),
    [anon_sym_str] = ACTIONS(494),
    [anon_sym_char] = ACTIONS(494),
    [anon_sym_SQUOTE] = ACTIONS(494),
    [anon_sym_as] = ACTIONS(494),
    [anon_sym_async] = ACTIONS(494),
    [anon_sym_break] = ACTIONS(494),
    [anon_sym_const] = ACTIONS(494),
    [anon_sym_continue] = ACTIONS(494),
    [anon_sym_default] = ACTIONS(494),
    [anon_sym_enum] = ACTIONS(494),
    [anon_sym_fn] = ACTIONS(494),
    [anon_sym_for] = ACTIONS(494),
    [anon_sym_if] = ACTIONS(494),
    [anon_sym_impl] = ACTIONS(494),
    [anon_sym_let] = ACTIONS(494),
    [anon_sym_loop] = ACTIONS(494),
    [anon_sym_match] = ACTIONS(494),
    [anon_sym_mod] = ACTIONS(494),
    [anon_sym_pub] = ACTIONS(494),
    [anon_sym_return] = ACTIONS(494),
    [anon_sym_static] = ACTIONS(494),
    [anon_sym_struct] = ACTIONS(494),
    [anon_sym_trait] = ACTIONS(494),
    [anon_sym_type] = ACTIONS(494),
    [anon_sym_union] = ACTIONS(494),
    [anon_sym_unsafe] = ACTIONS(494),
    [anon_sym_use] = ACTIONS(494),
    [anon_sym_while] = ACTIONS(494),
    [anon_sym_POUND] = ACTIONS(492),
    [anon_sym_BANG] = ACTIONS(494),
    [anon_sym_EQ] = ACTIONS(494),
    [anon_sym_extern] = ACTIONS(494),
    [anon_sym_LT] = ACTIONS(494),
    [anon_sym_GT] = ACTIONS(494),
    [anon_sym_COLON_COLON] = ACTIONS(492),
    [anon_sym_AMP] = ACTIONS(494),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(492),
    [anon_sym_DOT_DOT] = ACTIONS(494),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(492),
    [anon_sym_DASH] = ACTIONS(494),
    [anon_sym_AMP_AMP] = ACTIONS(492),
    [anon_sym_PIPE_PIPE] = ACTIONS(492),
    [anon_sym_PIPE] = ACTIONS(494),
    [anon_sym_CARET] = ACTIONS(494),
    [anon_sym_EQ_EQ] = ACTIONS(492),
    [anon_sym_BANG_EQ] = ACTIONS(492),
    [anon_sym_LT_EQ] = ACTIONS(492),
    [anon_sym_GT_EQ] = ACTIONS(492),
    [anon_sym_LT_LT] = ACTIONS(494),
    [anon_sym_GT_GT] = ACTIONS(494),
    [anon_sym_SLASH] = ACTIONS(494),
    [anon_sym_PERCENT] = ACTIONS(494),
    [anon_sym_PLUS_EQ] = ACTIONS(492),
    [anon_sym_DASH_EQ] = ACTIONS(492),
    [anon_sym_STAR_EQ] = ACTIONS(492),
    [anon_sym_SLASH_EQ] = ACTIONS(492),
    [anon_sym_PERCENT_EQ] = ACTIONS(492),
    [anon_sym_AMP_EQ] = ACTIONS(492),
    [anon_sym_PIPE_EQ] = ACTIONS(492),
    [anon_sym_CARET_EQ] = ACTIONS(492),
    [anon_sym_LT_LT_EQ] = ACTIONS(492),
    [anon_sym_GT_GT_EQ] = ACTIONS(492),
    [anon_sym_yield] = ACTIONS(494),
    [anon_sym_move] = ACTIONS(494),
    [anon_sym_DOT] = ACTIONS(494),
    [sym_integer_literal] = ACTIONS(492),
    [aux_sym_string_literal_token1] = ACTIONS(492),
    [sym_char_literal] = ACTIONS(492),
    [anon_sym_true] = ACTIONS(494),
    [anon_sym_false] = ACTIONS(494),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(494),
    [sym_super] = ACTIONS(494),
    [sym_crate] = ACTIONS(494),
    [sym_metavariable] = ACTIONS(492),
    [sym_raw_string_literal] = ACTIONS(492),
    [sym_float_literal] = ACTIONS(492),
    [sym_block_comment] = ACTIONS(3),
  },
  [78] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1291),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [aux_sym_tuple_expression_repeat1] = STATE(78),
    [sym_identifier] = ACTIONS(496),
    [anon_sym_LPAREN] = ACTIONS(499),
    [anon_sym_RPAREN] = ACTIONS(502),
    [anon_sym_LBRACE] = ACTIONS(504),
    [anon_sym_LBRACK] = ACTIONS(507),
    [anon_sym_STAR] = ACTIONS(510),
    [anon_sym_u8] = ACTIONS(513),
    [anon_sym_i8] = ACTIONS(513),
    [anon_sym_u16] = ACTIONS(513),
    [anon_sym_i16] = ACTIONS(513),
    [anon_sym_u32] = ACTIONS(513),
    [anon_sym_i32] = ACTIONS(513),
    [anon_sym_u64] = ACTIONS(513),
    [anon_sym_i64] = ACTIONS(513),
    [anon_sym_u128] = ACTIONS(513),
    [anon_sym_i128] = ACTIONS(513),
    [anon_sym_isize] = ACTIONS(513),
    [anon_sym_usize] = ACTIONS(513),
    [anon_sym_f32] = ACTIONS(513),
    [anon_sym_f64] = ACTIONS(513),
    [anon_sym_bool] = ACTIONS(513),
    [anon_sym_str] = ACTIONS(513),
    [anon_sym_char] = ACTIONS(513),
    [anon_sym_SQUOTE] = ACTIONS(516),
    [anon_sym_async] = ACTIONS(519),
    [anon_sym_break] = ACTIONS(522),
    [anon_sym_const] = ACTIONS(525),
    [anon_sym_continue] = ACTIONS(528),
    [anon_sym_default] = ACTIONS(531),
    [anon_sym_for] = ACTIONS(534),
    [anon_sym_if] = ACTIONS(537),
    [anon_sym_loop] = ACTIONS(540),
    [anon_sym_match] = ACTIONS(543),
    [anon_sym_return] = ACTIONS(546),
    [anon_sym_union] = ACTIONS(531),
    [anon_sym_unsafe] = ACTIONS(549),
    [anon_sym_while] = ACTIONS(552),
    [anon_sym_BANG] = ACTIONS(510),
    [anon_sym_LT] = ACTIONS(555),
    [anon_sym_COLON_COLON] = ACTIONS(558),
    [anon_sym_AMP] = ACTIONS(561),
    [anon_sym_DOT_DOT] = ACTIONS(564),
    [anon_sym_DASH] = ACTIONS(510),
    [anon_sym_PIPE] = ACTIONS(567),
    [anon_sym_yield] = ACTIONS(570),
    [anon_sym_move] = ACTIONS(573),
    [sym_integer_literal] = ACTIONS(576),
    [aux_sym_string_literal_token1] = ACTIONS(579),
    [sym_char_literal] = ACTIONS(576),
    [anon_sym_true] = ACTIONS(582),
    [anon_sym_false] = ACTIONS(582),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(585),
    [sym_super] = ACTIONS(588),
    [sym_crate] = ACTIONS(588),
    [sym_metavariable] = ACTIONS(591),
    [sym_raw_string_literal] = ACTIONS(576),
    [sym_float_literal] = ACTIONS(576),
    [sym_block_comment] = ACTIONS(3),
  },
  [79] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1245),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [aux_sym_tuple_expression_repeat1] = STATE(78),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_RPAREN] = ACTIONS(594),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [80] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1942),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1270),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(1176),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_let_condition] = STATE(1962),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(97),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(340),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(384),
    [anon_sym_u8] = ACTIONS(342),
    [anon_sym_i8] = ACTIONS(342),
    [anon_sym_u16] = ACTIONS(342),
    [anon_sym_i16] = ACTIONS(342),
    [anon_sym_u32] = ACTIONS(342),
    [anon_sym_i32] = ACTIONS(342),
    [anon_sym_u64] = ACTIONS(342),
    [anon_sym_i64] = ACTIONS(342),
    [anon_sym_u128] = ACTIONS(342),
    [anon_sym_i128] = ACTIONS(342),
    [anon_sym_isize] = ACTIONS(342),
    [anon_sym_usize] = ACTIONS(342),
    [anon_sym_f32] = ACTIONS(342),
    [anon_sym_f64] = ACTIONS(342),
    [anon_sym_bool] = ACTIONS(342),
    [anon_sym_str] = ACTIONS(342),
    [anon_sym_char] = ACTIONS(342),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(344),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(346),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_let] = ACTIONS(386),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(348),
    [anon_sym_union] = ACTIONS(346),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(384),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(352),
    [anon_sym_AMP] = ACTIONS(388),
    [anon_sym_DOT_DOT] = ACTIONS(390),
    [anon_sym_DASH] = ACTIONS(384),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(354),
    [anon_sym_move] = ACTIONS(356),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(358),
    [sym_super] = ACTIONS(360),
    [sym_crate] = ACTIONS(360),
    [sym_metavariable] = ACTIONS(362),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [81] = {
    [ts_builtin_sym_end] = ACTIONS(596),
    [sym_identifier] = ACTIONS(598),
    [anon_sym_SEMI] = ACTIONS(596),
    [anon_sym_macro_rules_BANG] = ACTIONS(596),
    [anon_sym_LPAREN] = ACTIONS(596),
    [anon_sym_LBRACE] = ACTIONS(596),
    [anon_sym_RBRACE] = ACTIONS(596),
    [anon_sym_LBRACK] = ACTIONS(596),
    [anon_sym_PLUS] = ACTIONS(598),
    [anon_sym_STAR] = ACTIONS(598),
    [anon_sym_QMARK] = ACTIONS(596),
    [anon_sym_u8] = ACTIONS(598),
    [anon_sym_i8] = ACTIONS(598),
    [anon_sym_u16] = ACTIONS(598),
    [anon_sym_i16] = ACTIONS(598),
    [anon_sym_u32] = ACTIONS(598),
    [anon_sym_i32] = ACTIONS(598),
    [anon_sym_u64] = ACTIONS(598),
    [anon_sym_i64] = ACTIONS(598),
    [anon_sym_u128] = ACTIONS(598),
    [anon_sym_i128] = ACTIONS(598),
    [anon_sym_isize] = ACTIONS(598),
    [anon_sym_usize] = ACTIONS(598),
    [anon_sym_f32] = ACTIONS(598),
    [anon_sym_f64] = ACTIONS(598),
    [anon_sym_bool] = ACTIONS(598),
    [anon_sym_str] = ACTIONS(598),
    [anon_sym_char] = ACTIONS(598),
    [anon_sym_SQUOTE] = ACTIONS(598),
    [anon_sym_as] = ACTIONS(598),
    [anon_sym_async] = ACTIONS(598),
    [anon_sym_break] = ACTIONS(598),
    [anon_sym_const] = ACTIONS(598),
    [anon_sym_continue] = ACTIONS(598),
    [anon_sym_default] = ACTIONS(598),
    [anon_sym_enum] = ACTIONS(598),
    [anon_sym_fn] = ACTIONS(598),
    [anon_sym_for] = ACTIONS(598),
    [anon_sym_if] = ACTIONS(598),
    [anon_sym_impl] = ACTIONS(598),
    [anon_sym_let] = ACTIONS(598),
    [anon_sym_loop] = ACTIONS(598),
    [anon_sym_match] = ACTIONS(598),
    [anon_sym_mod] = ACTIONS(598),
    [anon_sym_pub] = ACTIONS(598),
    [anon_sym_return] = ACTIONS(598),
    [anon_sym_static] = ACTIONS(598),
    [anon_sym_struct] = ACTIONS(598),
    [anon_sym_trait] = ACTIONS(598),
    [anon_sym_type] = ACTIONS(598),
    [anon_sym_union] = ACTIONS(598),
    [anon_sym_unsafe] = ACTIONS(598),
    [anon_sym_use] = ACTIONS(598),
    [anon_sym_while] = ACTIONS(598),
    [anon_sym_POUND] = ACTIONS(596),
    [anon_sym_BANG] = ACTIONS(598),
    [anon_sym_EQ] = ACTIONS(598),
    [anon_sym_extern] = ACTIONS(598),
    [anon_sym_LT] = ACTIONS(598),
    [anon_sym_GT] = ACTIONS(598),
    [anon_sym_COLON_COLON] = ACTIONS(596),
    [anon_sym_AMP] = ACTIONS(598),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(596),
    [anon_sym_DOT_DOT] = ACTIONS(598),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(596),
    [anon_sym_DASH] = ACTIONS(598),
    [anon_sym_AMP_AMP] = ACTIONS(596),
    [anon_sym_PIPE_PIPE] = ACTIONS(596),
    [anon_sym_PIPE] = ACTIONS(598),
    [anon_sym_CARET] = ACTIONS(598),
    [anon_sym_EQ_EQ] = ACTIONS(596),
    [anon_sym_BANG_EQ] = ACTIONS(596),
    [anon_sym_LT_EQ] = ACTIONS(596),
    [anon_sym_GT_EQ] = ACTIONS(596),
    [anon_sym_LT_LT] = ACTIONS(598),
    [anon_sym_GT_GT] = ACTIONS(598),
    [anon_sym_SLASH] = ACTIONS(598),
    [anon_sym_PERCENT] = ACTIONS(598),
    [anon_sym_PLUS_EQ] = ACTIONS(596),
    [anon_sym_DASH_EQ] = ACTIONS(596),
    [anon_sym_STAR_EQ] = ACTIONS(596),
    [anon_sym_SLASH_EQ] = ACTIONS(596),
    [anon_sym_PERCENT_EQ] = ACTIONS(596),
    [anon_sym_AMP_EQ] = ACTIONS(596),
    [anon_sym_PIPE_EQ] = ACTIONS(596),
    [anon_sym_CARET_EQ] = ACTIONS(596),
    [anon_sym_LT_LT_EQ] = ACTIONS(596),
    [anon_sym_GT_GT_EQ] = ACTIONS(596),
    [anon_sym_yield] = ACTIONS(598),
    [anon_sym_move] = ACTIONS(598),
    [anon_sym_DOT] = ACTIONS(598),
    [sym_integer_literal] = ACTIONS(596),
    [aux_sym_string_literal_token1] = ACTIONS(596),
    [sym_char_literal] = ACTIONS(596),
    [anon_sym_true] = ACTIONS(598),
    [anon_sym_false] = ACTIONS(598),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(598),
    [sym_super] = ACTIONS(598),
    [sym_crate] = ACTIONS(598),
    [sym_metavariable] = ACTIONS(596),
    [sym_raw_string_literal] = ACTIONS(596),
    [sym_float_literal] = ACTIONS(596),
    [sym_block_comment] = ACTIONS(3),
  },
  [82] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1280),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_let_condition] = STATE(1962),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_let] = ACTIONS(382),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [83] = {
    [ts_builtin_sym_end] = ACTIONS(600),
    [sym_identifier] = ACTIONS(602),
    [anon_sym_SEMI] = ACTIONS(600),
    [anon_sym_macro_rules_BANG] = ACTIONS(600),
    [anon_sym_LPAREN] = ACTIONS(600),
    [anon_sym_LBRACE] = ACTIONS(600),
    [anon_sym_RBRACE] = ACTIONS(600),
    [anon_sym_LBRACK] = ACTIONS(600),
    [anon_sym_PLUS] = ACTIONS(602),
    [anon_sym_STAR] = ACTIONS(602),
    [anon_sym_QMARK] = ACTIONS(600),
    [anon_sym_u8] = ACTIONS(602),
    [anon_sym_i8] = ACTIONS(602),
    [anon_sym_u16] = ACTIONS(602),
    [anon_sym_i16] = ACTIONS(602),
    [anon_sym_u32] = ACTIONS(602),
    [anon_sym_i32] = ACTIONS(602),
    [anon_sym_u64] = ACTIONS(602),
    [anon_sym_i64] = ACTIONS(602),
    [anon_sym_u128] = ACTIONS(602),
    [anon_sym_i128] = ACTIONS(602),
    [anon_sym_isize] = ACTIONS(602),
    [anon_sym_usize] = ACTIONS(602),
    [anon_sym_f32] = ACTIONS(602),
    [anon_sym_f64] = ACTIONS(602),
    [anon_sym_bool] = ACTIONS(602),
    [anon_sym_str] = ACTIONS(602),
    [anon_sym_char] = ACTIONS(602),
    [anon_sym_SQUOTE] = ACTIONS(602),
    [anon_sym_as] = ACTIONS(602),
    [anon_sym_async] = ACTIONS(602),
    [anon_sym_break] = ACTIONS(602),
    [anon_sym_const] = ACTIONS(602),
    [anon_sym_continue] = ACTIONS(602),
    [anon_sym_default] = ACTIONS(602),
    [anon_sym_enum] = ACTIONS(602),
    [anon_sym_fn] = ACTIONS(602),
    [anon_sym_for] = ACTIONS(602),
    [anon_sym_if] = ACTIONS(602),
    [anon_sym_impl] = ACTIONS(602),
    [anon_sym_let] = ACTIONS(602),
    [anon_sym_loop] = ACTIONS(602),
    [anon_sym_match] = ACTIONS(602),
    [anon_sym_mod] = ACTIONS(602),
    [anon_sym_pub] = ACTIONS(602),
    [anon_sym_return] = ACTIONS(602),
    [anon_sym_static] = ACTIONS(602),
    [anon_sym_struct] = ACTIONS(602),
    [anon_sym_trait] = ACTIONS(602),
    [anon_sym_type] = ACTIONS(602),
    [anon_sym_union] = ACTIONS(602),
    [anon_sym_unsafe] = ACTIONS(602),
    [anon_sym_use] = ACTIONS(602),
    [anon_sym_while] = ACTIONS(602),
    [anon_sym_POUND] = ACTIONS(600),
    [anon_sym_BANG] = ACTIONS(602),
    [anon_sym_EQ] = ACTIONS(602),
    [anon_sym_extern] = ACTIONS(602),
    [anon_sym_LT] = ACTIONS(602),
    [anon_sym_GT] = ACTIONS(602),
    [anon_sym_COLON_COLON] = ACTIONS(600),
    [anon_sym_AMP] = ACTIONS(602),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(600),
    [anon_sym_DOT_DOT] = ACTIONS(602),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(600),
    [anon_sym_DASH] = ACTIONS(602),
    [anon_sym_AMP_AMP] = ACTIONS(600),
    [anon_sym_PIPE_PIPE] = ACTIONS(600),
    [anon_sym_PIPE] = ACTIONS(602),
    [anon_sym_CARET] = ACTIONS(602),
    [anon_sym_EQ_EQ] = ACTIONS(600),
    [anon_sym_BANG_EQ] = ACTIONS(600),
    [anon_sym_LT_EQ] = ACTIONS(600),
    [anon_sym_GT_EQ] = ACTIONS(600),
    [anon_sym_LT_LT] = ACTIONS(602),
    [anon_sym_GT_GT] = ACTIONS(602),
    [anon_sym_SLASH] = ACTIONS(602),
    [anon_sym_PERCENT] = ACTIONS(602),
    [anon_sym_PLUS_EQ] = ACTIONS(600),
    [anon_sym_DASH_EQ] = ACTIONS(600),
    [anon_sym_STAR_EQ] = ACTIONS(600),
    [anon_sym_SLASH_EQ] = ACTIONS(600),
    [anon_sym_PERCENT_EQ] = ACTIONS(600),
    [anon_sym_AMP_EQ] = ACTIONS(600),
    [anon_sym_PIPE_EQ] = ACTIONS(600),
    [anon_sym_CARET_EQ] = ACTIONS(600),
    [anon_sym_LT_LT_EQ] = ACTIONS(600),
    [anon_sym_GT_GT_EQ] = ACTIONS(600),
    [anon_sym_yield] = ACTIONS(602),
    [anon_sym_move] = ACTIONS(602),
    [anon_sym_DOT] = ACTIONS(602),
    [sym_integer_literal] = ACTIONS(600),
    [aux_sym_string_literal_token1] = ACTIONS(600),
    [sym_char_literal] = ACTIONS(600),
    [anon_sym_true] = ACTIONS(602),
    [anon_sym_false] = ACTIONS(602),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(602),
    [sym_super] = ACTIONS(602),
    [sym_crate] = ACTIONS(602),
    [sym_metavariable] = ACTIONS(600),
    [sym_raw_string_literal] = ACTIONS(600),
    [sym_float_literal] = ACTIONS(600),
    [sym_block_comment] = ACTIONS(3),
  },
  [84] = {
    [ts_builtin_sym_end] = ACTIONS(604),
    [sym_identifier] = ACTIONS(606),
    [anon_sym_SEMI] = ACTIONS(604),
    [anon_sym_macro_rules_BANG] = ACTIONS(604),
    [anon_sym_LPAREN] = ACTIONS(604),
    [anon_sym_LBRACE] = ACTIONS(604),
    [anon_sym_RBRACE] = ACTIONS(604),
    [anon_sym_LBRACK] = ACTIONS(604),
    [anon_sym_PLUS] = ACTIONS(606),
    [anon_sym_STAR] = ACTIONS(606),
    [anon_sym_QMARK] = ACTIONS(604),
    [anon_sym_u8] = ACTIONS(606),
    [anon_sym_i8] = ACTIONS(606),
    [anon_sym_u16] = ACTIONS(606),
    [anon_sym_i16] = ACTIONS(606),
    [anon_sym_u32] = ACTIONS(606),
    [anon_sym_i32] = ACTIONS(606),
    [anon_sym_u64] = ACTIONS(606),
    [anon_sym_i64] = ACTIONS(606),
    [anon_sym_u128] = ACTIONS(606),
    [anon_sym_i128] = ACTIONS(606),
    [anon_sym_isize] = ACTIONS(606),
    [anon_sym_usize] = ACTIONS(606),
    [anon_sym_f32] = ACTIONS(606),
    [anon_sym_f64] = ACTIONS(606),
    [anon_sym_bool] = ACTIONS(606),
    [anon_sym_str] = ACTIONS(606),
    [anon_sym_char] = ACTIONS(606),
    [anon_sym_SQUOTE] = ACTIONS(606),
    [anon_sym_as] = ACTIONS(606),
    [anon_sym_async] = ACTIONS(606),
    [anon_sym_break] = ACTIONS(606),
    [anon_sym_const] = ACTIONS(606),
    [anon_sym_continue] = ACTIONS(606),
    [anon_sym_default] = ACTIONS(606),
    [anon_sym_enum] = ACTIONS(606),
    [anon_sym_fn] = ACTIONS(606),
    [anon_sym_for] = ACTIONS(606),
    [anon_sym_if] = ACTIONS(606),
    [anon_sym_impl] = ACTIONS(606),
    [anon_sym_let] = ACTIONS(606),
    [anon_sym_loop] = ACTIONS(606),
    [anon_sym_match] = ACTIONS(606),
    [anon_sym_mod] = ACTIONS(606),
    [anon_sym_pub] = ACTIONS(606),
    [anon_sym_return] = ACTIONS(606),
    [anon_sym_static] = ACTIONS(606),
    [anon_sym_struct] = ACTIONS(606),
    [anon_sym_trait] = ACTIONS(606),
    [anon_sym_type] = ACTIONS(606),
    [anon_sym_union] = ACTIONS(606),
    [anon_sym_unsafe] = ACTIONS(606),
    [anon_sym_use] = ACTIONS(606),
    [anon_sym_while] = ACTIONS(606),
    [anon_sym_POUND] = ACTIONS(604),
    [anon_sym_BANG] = ACTIONS(606),
    [anon_sym_EQ] = ACTIONS(606),
    [anon_sym_extern] = ACTIONS(606),
    [anon_sym_LT] = ACTIONS(606),
    [anon_sym_GT] = ACTIONS(606),
    [anon_sym_COLON_COLON] = ACTIONS(604),
    [anon_sym_AMP] = ACTIONS(606),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(604),
    [anon_sym_DOT_DOT] = ACTIONS(606),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(604),
    [anon_sym_DASH] = ACTIONS(606),
    [anon_sym_AMP_AMP] = ACTIONS(604),
    [anon_sym_PIPE_PIPE] = ACTIONS(604),
    [anon_sym_PIPE] = ACTIONS(606),
    [anon_sym_CARET] = ACTIONS(606),
    [anon_sym_EQ_EQ] = ACTIONS(604),
    [anon_sym_BANG_EQ] = ACTIONS(604),
    [anon_sym_LT_EQ] = ACTIONS(604),
    [anon_sym_GT_EQ] = ACTIONS(604),
    [anon_sym_LT_LT] = ACTIONS(606),
    [anon_sym_GT_GT] = ACTIONS(606),
    [anon_sym_SLASH] = ACTIONS(606),
    [anon_sym_PERCENT] = ACTIONS(606),
    [anon_sym_PLUS_EQ] = ACTIONS(604),
    [anon_sym_DASH_EQ] = ACTIONS(604),
    [anon_sym_STAR_EQ] = ACTIONS(604),
    [anon_sym_SLASH_EQ] = ACTIONS(604),
    [anon_sym_PERCENT_EQ] = ACTIONS(604),
    [anon_sym_AMP_EQ] = ACTIONS(604),
    [anon_sym_PIPE_EQ] = ACTIONS(604),
    [anon_sym_CARET_EQ] = ACTIONS(604),
    [anon_sym_LT_LT_EQ] = ACTIONS(604),
    [anon_sym_GT_GT_EQ] = ACTIONS(604),
    [anon_sym_yield] = ACTIONS(606),
    [anon_sym_move] = ACTIONS(606),
    [anon_sym_DOT] = ACTIONS(606),
    [sym_integer_literal] = ACTIONS(604),
    [aux_sym_string_literal_token1] = ACTIONS(604),
    [sym_char_literal] = ACTIONS(604),
    [anon_sym_true] = ACTIONS(606),
    [anon_sym_false] = ACTIONS(606),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(606),
    [sym_super] = ACTIONS(606),
    [sym_crate] = ACTIONS(606),
    [sym_metavariable] = ACTIONS(604),
    [sym_raw_string_literal] = ACTIONS(604),
    [sym_float_literal] = ACTIONS(604),
    [sym_block_comment] = ACTIONS(3),
  },
  [85] = {
    [ts_builtin_sym_end] = ACTIONS(608),
    [sym_identifier] = ACTIONS(610),
    [anon_sym_SEMI] = ACTIONS(608),
    [anon_sym_macro_rules_BANG] = ACTIONS(608),
    [anon_sym_LPAREN] = ACTIONS(608),
    [anon_sym_LBRACE] = ACTIONS(608),
    [anon_sym_RBRACE] = ACTIONS(608),
    [anon_sym_LBRACK] = ACTIONS(608),
    [anon_sym_PLUS] = ACTIONS(610),
    [anon_sym_STAR] = ACTIONS(610),
    [anon_sym_QMARK] = ACTIONS(608),
    [anon_sym_u8] = ACTIONS(610),
    [anon_sym_i8] = ACTIONS(610),
    [anon_sym_u16] = ACTIONS(610),
    [anon_sym_i16] = ACTIONS(610),
    [anon_sym_u32] = ACTIONS(610),
    [anon_sym_i32] = ACTIONS(610),
    [anon_sym_u64] = ACTIONS(610),
    [anon_sym_i64] = ACTIONS(610),
    [anon_sym_u128] = ACTIONS(610),
    [anon_sym_i128] = ACTIONS(610),
    [anon_sym_isize] = ACTIONS(610),
    [anon_sym_usize] = ACTIONS(610),
    [anon_sym_f32] = ACTIONS(610),
    [anon_sym_f64] = ACTIONS(610),
    [anon_sym_bool] = ACTIONS(610),
    [anon_sym_str] = ACTIONS(610),
    [anon_sym_char] = ACTIONS(610),
    [anon_sym_SQUOTE] = ACTIONS(610),
    [anon_sym_as] = ACTIONS(610),
    [anon_sym_async] = ACTIONS(610),
    [anon_sym_break] = ACTIONS(610),
    [anon_sym_const] = ACTIONS(610),
    [anon_sym_continue] = ACTIONS(610),
    [anon_sym_default] = ACTIONS(610),
    [anon_sym_enum] = ACTIONS(610),
    [anon_sym_fn] = ACTIONS(610),
    [anon_sym_for] = ACTIONS(610),
    [anon_sym_if] = ACTIONS(610),
    [anon_sym_impl] = ACTIONS(610),
    [anon_sym_let] = ACTIONS(610),
    [anon_sym_loop] = ACTIONS(610),
    [anon_sym_match] = ACTIONS(610),
    [anon_sym_mod] = ACTIONS(610),
    [anon_sym_pub] = ACTIONS(610),
    [anon_sym_return] = ACTIONS(610),
    [anon_sym_static] = ACTIONS(610),
    [anon_sym_struct] = ACTIONS(610),
    [anon_sym_trait] = ACTIONS(610),
    [anon_sym_type] = ACTIONS(610),
    [anon_sym_union] = ACTIONS(610),
    [anon_sym_unsafe] = ACTIONS(610),
    [anon_sym_use] = ACTIONS(610),
    [anon_sym_while] = ACTIONS(610),
    [anon_sym_POUND] = ACTIONS(608),
    [anon_sym_BANG] = ACTIONS(610),
    [anon_sym_EQ] = ACTIONS(610),
    [anon_sym_extern] = ACTIONS(610),
    [anon_sym_LT] = ACTIONS(610),
    [anon_sym_GT] = ACTIONS(610),
    [anon_sym_COLON_COLON] = ACTIONS(608),
    [anon_sym_AMP] = ACTIONS(610),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(608),
    [anon_sym_DOT_DOT] = ACTIONS(610),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(608),
    [anon_sym_DASH] = ACTIONS(610),
    [anon_sym_AMP_AMP] = ACTIONS(608),
    [anon_sym_PIPE_PIPE] = ACTIONS(608),
    [anon_sym_PIPE] = ACTIONS(610),
    [anon_sym_CARET] = ACTIONS(610),
    [anon_sym_EQ_EQ] = ACTIONS(608),
    [anon_sym_BANG_EQ] = ACTIONS(608),
    [anon_sym_LT_EQ] = ACTIONS(608),
    [anon_sym_GT_EQ] = ACTIONS(608),
    [anon_sym_LT_LT] = ACTIONS(610),
    [anon_sym_GT_GT] = ACTIONS(610),
    [anon_sym_SLASH] = ACTIONS(610),
    [anon_sym_PERCENT] = ACTIONS(610),
    [anon_sym_PLUS_EQ] = ACTIONS(608),
    [anon_sym_DASH_EQ] = ACTIONS(608),
    [anon_sym_STAR_EQ] = ACTIONS(608),
    [anon_sym_SLASH_EQ] = ACTIONS(608),
    [anon_sym_PERCENT_EQ] = ACTIONS(608),
    [anon_sym_AMP_EQ] = ACTIONS(608),
    [anon_sym_PIPE_EQ] = ACTIONS(608),
    [anon_sym_CARET_EQ] = ACTIONS(608),
    [anon_sym_LT_LT_EQ] = ACTIONS(608),
    [anon_sym_GT_GT_EQ] = ACTIONS(608),
    [anon_sym_yield] = ACTIONS(610),
    [anon_sym_move] = ACTIONS(610),
    [anon_sym_DOT] = ACTIONS(610),
    [sym_integer_literal] = ACTIONS(608),
    [aux_sym_string_literal_token1] = ACTIONS(608),
    [sym_char_literal] = ACTIONS(608),
    [anon_sym_true] = ACTIONS(610),
    [anon_sym_false] = ACTIONS(610),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(610),
    [sym_super] = ACTIONS(610),
    [sym_crate] = ACTIONS(610),
    [sym_metavariable] = ACTIONS(608),
    [sym_raw_string_literal] = ACTIONS(608),
    [sym_float_literal] = ACTIONS(608),
    [sym_block_comment] = ACTIONS(3),
  },
  [86] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1234),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [aux_sym_tuple_expression_repeat1] = STATE(79),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_RPAREN] = ACTIONS(612),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [87] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1234),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [aux_sym_tuple_expression_repeat1] = STATE(78),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_RPAREN] = ACTIONS(612),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [88] = {
    [ts_builtin_sym_end] = ACTIONS(614),
    [sym_identifier] = ACTIONS(616),
    [anon_sym_SEMI] = ACTIONS(614),
    [anon_sym_macro_rules_BANG] = ACTIONS(614),
    [anon_sym_LPAREN] = ACTIONS(614),
    [anon_sym_LBRACE] = ACTIONS(614),
    [anon_sym_RBRACE] = ACTIONS(614),
    [anon_sym_LBRACK] = ACTIONS(614),
    [anon_sym_PLUS] = ACTIONS(616),
    [anon_sym_STAR] = ACTIONS(616),
    [anon_sym_QMARK] = ACTIONS(614),
    [anon_sym_u8] = ACTIONS(616),
    [anon_sym_i8] = ACTIONS(616),
    [anon_sym_u16] = ACTIONS(616),
    [anon_sym_i16] = ACTIONS(616),
    [anon_sym_u32] = ACTIONS(616),
    [anon_sym_i32] = ACTIONS(616),
    [anon_sym_u64] = ACTIONS(616),
    [anon_sym_i64] = ACTIONS(616),
    [anon_sym_u128] = ACTIONS(616),
    [anon_sym_i128] = ACTIONS(616),
    [anon_sym_isize] = ACTIONS(616),
    [anon_sym_usize] = ACTIONS(616),
    [anon_sym_f32] = ACTIONS(616),
    [anon_sym_f64] = ACTIONS(616),
    [anon_sym_bool] = ACTIONS(616),
    [anon_sym_str] = ACTIONS(616),
    [anon_sym_char] = ACTIONS(616),
    [anon_sym_SQUOTE] = ACTIONS(616),
    [anon_sym_as] = ACTIONS(616),
    [anon_sym_async] = ACTIONS(616),
    [anon_sym_break] = ACTIONS(616),
    [anon_sym_const] = ACTIONS(616),
    [anon_sym_continue] = ACTIONS(616),
    [anon_sym_default] = ACTIONS(616),
    [anon_sym_enum] = ACTIONS(616),
    [anon_sym_fn] = ACTIONS(616),
    [anon_sym_for] = ACTIONS(616),
    [anon_sym_if] = ACTIONS(616),
    [anon_sym_impl] = ACTIONS(616),
    [anon_sym_let] = ACTIONS(616),
    [anon_sym_loop] = ACTIONS(616),
    [anon_sym_match] = ACTIONS(616),
    [anon_sym_mod] = ACTIONS(616),
    [anon_sym_pub] = ACTIONS(616),
    [anon_sym_return] = ACTIONS(616),
    [anon_sym_static] = ACTIONS(616),
    [anon_sym_struct] = ACTIONS(616),
    [anon_sym_trait] = ACTIONS(616),
    [anon_sym_type] = ACTIONS(616),
    [anon_sym_union] = ACTIONS(616),
    [anon_sym_unsafe] = ACTIONS(616),
    [anon_sym_use] = ACTIONS(616),
    [anon_sym_while] = ACTIONS(616),
    [anon_sym_POUND] = ACTIONS(614),
    [anon_sym_BANG] = ACTIONS(616),
    [anon_sym_EQ] = ACTIONS(616),
    [anon_sym_extern] = ACTIONS(616),
    [anon_sym_LT] = ACTIONS(616),
    [anon_sym_GT] = ACTIONS(616),
    [anon_sym_COLON_COLON] = ACTIONS(614),
    [anon_sym_AMP] = ACTIONS(616),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(614),
    [anon_sym_DOT_DOT] = ACTIONS(616),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(614),
    [anon_sym_DASH] = ACTIONS(616),
    [anon_sym_AMP_AMP] = ACTIONS(614),
    [anon_sym_PIPE_PIPE] = ACTIONS(614),
    [anon_sym_PIPE] = ACTIONS(616),
    [anon_sym_CARET] = ACTIONS(616),
    [anon_sym_EQ_EQ] = ACTIONS(614),
    [anon_sym_BANG_EQ] = ACTIONS(614),
    [anon_sym_LT_EQ] = ACTIONS(614),
    [anon_sym_GT_EQ] = ACTIONS(614),
    [anon_sym_LT_LT] = ACTIONS(616),
    [anon_sym_GT_GT] = ACTIONS(616),
    [anon_sym_SLASH] = ACTIONS(616),
    [anon_sym_PERCENT] = ACTIONS(616),
    [anon_sym_PLUS_EQ] = ACTIONS(614),
    [anon_sym_DASH_EQ] = ACTIONS(614),
    [anon_sym_STAR_EQ] = ACTIONS(614),
    [anon_sym_SLASH_EQ] = ACTIONS(614),
    [anon_sym_PERCENT_EQ] = ACTIONS(614),
    [anon_sym_AMP_EQ] = ACTIONS(614),
    [anon_sym_PIPE_EQ] = ACTIONS(614),
    [anon_sym_CARET_EQ] = ACTIONS(614),
    [anon_sym_LT_LT_EQ] = ACTIONS(614),
    [anon_sym_GT_GT_EQ] = ACTIONS(614),
    [anon_sym_yield] = ACTIONS(616),
    [anon_sym_move] = ACTIONS(616),
    [anon_sym_DOT] = ACTIONS(616),
    [sym_integer_literal] = ACTIONS(614),
    [aux_sym_string_literal_token1] = ACTIONS(614),
    [sym_char_literal] = ACTIONS(614),
    [anon_sym_true] = ACTIONS(616),
    [anon_sym_false] = ACTIONS(616),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(616),
    [sym_super] = ACTIONS(616),
    [sym_crate] = ACTIONS(616),
    [sym_metavariable] = ACTIONS(614),
    [sym_raw_string_literal] = ACTIONS(614),
    [sym_float_literal] = ACTIONS(614),
    [sym_block_comment] = ACTIONS(3),
  },
  [89] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1212),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_RBRACK] = ACTIONS(618),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [90] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1156),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_DASH_GT] = ACTIONS(620),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(420),
    [anon_sym_DASH] = ACTIONS(308),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [91] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1130),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [sym_mutable_specifier] = ACTIONS(622),
    [anon_sym_DOT_DOT] = ACTIONS(420),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [92] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1212),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_RBRACK] = ACTIONS(624),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [93] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1212),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_RBRACK] = ACTIONS(626),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [94] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1145),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_DASH_GT] = ACTIONS(628),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(420),
    [anon_sym_DASH] = ACTIONS(308),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [95] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1942),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1208),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(1176),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(97),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(340),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(384),
    [anon_sym_u8] = ACTIONS(342),
    [anon_sym_i8] = ACTIONS(342),
    [anon_sym_u16] = ACTIONS(342),
    [anon_sym_i16] = ACTIONS(342),
    [anon_sym_u32] = ACTIONS(342),
    [anon_sym_i32] = ACTIONS(342),
    [anon_sym_u64] = ACTIONS(342),
    [anon_sym_i64] = ACTIONS(342),
    [anon_sym_u128] = ACTIONS(342),
    [anon_sym_i128] = ACTIONS(342),
    [anon_sym_isize] = ACTIONS(342),
    [anon_sym_usize] = ACTIONS(342),
    [anon_sym_f32] = ACTIONS(342),
    [anon_sym_f64] = ACTIONS(342),
    [anon_sym_bool] = ACTIONS(342),
    [anon_sym_str] = ACTIONS(342),
    [anon_sym_char] = ACTIONS(342),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(344),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(346),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(348),
    [anon_sym_union] = ACTIONS(346),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(384),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(352),
    [anon_sym_AMP] = ACTIONS(388),
    [sym_mutable_specifier] = ACTIONS(630),
    [anon_sym_DOT_DOT] = ACTIONS(418),
    [anon_sym_DASH] = ACTIONS(384),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(354),
    [anon_sym_move] = ACTIONS(356),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(358),
    [sym_super] = ACTIONS(360),
    [sym_crate] = ACTIONS(360),
    [sym_metavariable] = ACTIONS(362),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [96] = {
    [sym_identifier] = ACTIONS(456),
    [anon_sym_SEMI] = ACTIONS(458),
    [anon_sym_macro_rules_BANG] = ACTIONS(454),
    [anon_sym_LPAREN] = ACTIONS(458),
    [anon_sym_LBRACE] = ACTIONS(454),
    [anon_sym_RBRACE] = ACTIONS(454),
    [anon_sym_LBRACK] = ACTIONS(458),
    [anon_sym_PLUS] = ACTIONS(460),
    [anon_sym_STAR] = ACTIONS(460),
    [anon_sym_QMARK] = ACTIONS(458),
    [anon_sym_u8] = ACTIONS(456),
    [anon_sym_i8] = ACTIONS(456),
    [anon_sym_u16] = ACTIONS(456),
    [anon_sym_i16] = ACTIONS(456),
    [anon_sym_u32] = ACTIONS(456),
    [anon_sym_i32] = ACTIONS(456),
    [anon_sym_u64] = ACTIONS(456),
    [anon_sym_i64] = ACTIONS(456),
    [anon_sym_u128] = ACTIONS(456),
    [anon_sym_i128] = ACTIONS(456),
    [anon_sym_isize] = ACTIONS(456),
    [anon_sym_usize] = ACTIONS(456),
    [anon_sym_f32] = ACTIONS(456),
    [anon_sym_f64] = ACTIONS(456),
    [anon_sym_bool] = ACTIONS(456),
    [anon_sym_str] = ACTIONS(456),
    [anon_sym_char] = ACTIONS(456),
    [anon_sym_SQUOTE] = ACTIONS(456),
    [anon_sym_as] = ACTIONS(460),
    [anon_sym_async] = ACTIONS(456),
    [anon_sym_break] = ACTIONS(456),
    [anon_sym_const] = ACTIONS(456),
    [anon_sym_continue] = ACTIONS(456),
    [anon_sym_default] = ACTIONS(456),
    [anon_sym_enum] = ACTIONS(456),
    [anon_sym_fn] = ACTIONS(456),
    [anon_sym_for] = ACTIONS(456),
    [anon_sym_if] = ACTIONS(456),
    [anon_sym_impl] = ACTIONS(456),
    [anon_sym_let] = ACTIONS(456),
    [anon_sym_loop] = ACTIONS(456),
    [anon_sym_match] = ACTIONS(456),
    [anon_sym_mod] = ACTIONS(456),
    [anon_sym_pub] = ACTIONS(456),
    [anon_sym_return] = ACTIONS(456),
    [anon_sym_static] = ACTIONS(456),
    [anon_sym_struct] = ACTIONS(456),
    [anon_sym_trait] = ACTIONS(456),
    [anon_sym_type] = ACTIONS(456),
    [anon_sym_union] = ACTIONS(456),
    [anon_sym_unsafe] = ACTIONS(456),
    [anon_sym_use] = ACTIONS(456),
    [anon_sym_while] = ACTIONS(456),
    [anon_sym_POUND] = ACTIONS(454),
    [anon_sym_BANG] = ACTIONS(456),
    [anon_sym_EQ] = ACTIONS(460),
    [anon_sym_extern] = ACTIONS(456),
    [anon_sym_LT] = ACTIONS(460),
    [anon_sym_GT] = ACTIONS(460),
    [anon_sym_COLON_COLON] = ACTIONS(454),
    [anon_sym_AMP] = ACTIONS(460),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(458),
    [anon_sym_DOT_DOT] = ACTIONS(460),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(458),
    [anon_sym_DASH] = ACTIONS(460),
    [anon_sym_AMP_AMP] = ACTIONS(458),
    [anon_sym_PIPE_PIPE] = ACTIONS(458),
    [anon_sym_PIPE] = ACTIONS(460),
    [anon_sym_CARET] = ACTIONS(460),
    [anon_sym_EQ_EQ] = ACTIONS(458),
    [anon_sym_BANG_EQ] = ACTIONS(458),
    [anon_sym_LT_EQ] = ACTIONS(458),
    [anon_sym_GT_EQ] = ACTIONS(458),
    [anon_sym_LT_LT] = ACTIONS(460),
    [anon_sym_GT_GT] = ACTIONS(460),
    [anon_sym_SLASH] = ACTIONS(460),
    [anon_sym_PERCENT] = ACTIONS(460),
    [anon_sym_PLUS_EQ] = ACTIONS(458),
    [anon_sym_DASH_EQ] = ACTIONS(458),
    [anon_sym_STAR_EQ] = ACTIONS(458),
    [anon_sym_SLASH_EQ] = ACTIONS(458),
    [anon_sym_PERCENT_EQ] = ACTIONS(458),
    [anon_sym_AMP_EQ] = ACTIONS(458),
    [anon_sym_PIPE_EQ] = ACTIONS(458),
    [anon_sym_CARET_EQ] = ACTIONS(458),
    [anon_sym_LT_LT_EQ] = ACTIONS(458),
    [anon_sym_GT_GT_EQ] = ACTIONS(458),
    [anon_sym_yield] = ACTIONS(456),
    [anon_sym_move] = ACTIONS(456),
    [anon_sym_DOT] = ACTIONS(460),
    [sym_integer_literal] = ACTIONS(454),
    [aux_sym_string_literal_token1] = ACTIONS(454),
    [sym_char_literal] = ACTIONS(454),
    [anon_sym_true] = ACTIONS(456),
    [anon_sym_false] = ACTIONS(456),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(456),
    [sym_super] = ACTIONS(456),
    [sym_crate] = ACTIONS(456),
    [sym_metavariable] = ACTIONS(454),
    [sym_raw_string_literal] = ACTIONS(454),
    [sym_float_literal] = ACTIONS(454),
    [sym_block_comment] = ACTIONS(3),
  },
  [97] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1942),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1256),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(1176),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(97),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(340),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(384),
    [anon_sym_u8] = ACTIONS(342),
    [anon_sym_i8] = ACTIONS(342),
    [anon_sym_u16] = ACTIONS(342),
    [anon_sym_i16] = ACTIONS(342),
    [anon_sym_u32] = ACTIONS(342),
    [anon_sym_i32] = ACTIONS(342),
    [anon_sym_u64] = ACTIONS(342),
    [anon_sym_i64] = ACTIONS(342),
    [anon_sym_u128] = ACTIONS(342),
    [anon_sym_i128] = ACTIONS(342),
    [anon_sym_isize] = ACTIONS(342),
    [anon_sym_usize] = ACTIONS(342),
    [anon_sym_f32] = ACTIONS(342),
    [anon_sym_f64] = ACTIONS(342),
    [anon_sym_bool] = ACTIONS(342),
    [anon_sym_str] = ACTIONS(342),
    [anon_sym_char] = ACTIONS(342),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(344),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(346),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(348),
    [anon_sym_union] = ACTIONS(346),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(384),
    [anon_sym_DASH_GT] = ACTIONS(628),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(352),
    [anon_sym_AMP] = ACTIONS(388),
    [anon_sym_DOT_DOT] = ACTIONS(418),
    [anon_sym_DASH] = ACTIONS(350),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(354),
    [anon_sym_move] = ACTIONS(356),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(358),
    [sym_super] = ACTIONS(360),
    [sym_crate] = ACTIONS(360),
    [sym_metavariable] = ACTIONS(362),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [98] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1942),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1262),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(1176),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(97),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(340),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(384),
    [anon_sym_u8] = ACTIONS(342),
    [anon_sym_i8] = ACTIONS(342),
    [anon_sym_u16] = ACTIONS(342),
    [anon_sym_i16] = ACTIONS(342),
    [anon_sym_u32] = ACTIONS(342),
    [anon_sym_i32] = ACTIONS(342),
    [anon_sym_u64] = ACTIONS(342),
    [anon_sym_i64] = ACTIONS(342),
    [anon_sym_u128] = ACTIONS(342),
    [anon_sym_i128] = ACTIONS(342),
    [anon_sym_isize] = ACTIONS(342),
    [anon_sym_usize] = ACTIONS(342),
    [anon_sym_f32] = ACTIONS(342),
    [anon_sym_f64] = ACTIONS(342),
    [anon_sym_bool] = ACTIONS(342),
    [anon_sym_str] = ACTIONS(342),
    [anon_sym_char] = ACTIONS(342),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(344),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(346),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(348),
    [anon_sym_union] = ACTIONS(346),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(384),
    [anon_sym_DASH_GT] = ACTIONS(620),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(352),
    [anon_sym_AMP] = ACTIONS(388),
    [anon_sym_DOT_DOT] = ACTIONS(418),
    [anon_sym_DASH] = ACTIONS(350),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(354),
    [anon_sym_move] = ACTIONS(356),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(358),
    [sym_super] = ACTIONS(360),
    [sym_crate] = ACTIONS(360),
    [sym_metavariable] = ACTIONS(362),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [99] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1252),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [100] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1942),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(871),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(1176),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(97),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(340),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(384),
    [anon_sym_u8] = ACTIONS(342),
    [anon_sym_i8] = ACTIONS(342),
    [anon_sym_u16] = ACTIONS(342),
    [anon_sym_i16] = ACTIONS(342),
    [anon_sym_u32] = ACTIONS(342),
    [anon_sym_i32] = ACTIONS(342),
    [anon_sym_u64] = ACTIONS(342),
    [anon_sym_i64] = ACTIONS(342),
    [anon_sym_u128] = ACTIONS(342),
    [anon_sym_i128] = ACTIONS(342),
    [anon_sym_isize] = ACTIONS(342),
    [anon_sym_usize] = ACTIONS(342),
    [anon_sym_f32] = ACTIONS(342),
    [anon_sym_f64] = ACTIONS(342),
    [anon_sym_bool] = ACTIONS(342),
    [anon_sym_str] = ACTIONS(342),
    [anon_sym_char] = ACTIONS(342),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(344),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(346),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(348),
    [anon_sym_union] = ACTIONS(346),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(384),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(352),
    [anon_sym_AMP] = ACTIONS(388),
    [anon_sym_DOT_DOT] = ACTIONS(390),
    [anon_sym_DASH] = ACTIONS(384),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(354),
    [anon_sym_move] = ACTIONS(356),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(358),
    [sym_super] = ACTIONS(360),
    [sym_crate] = ACTIONS(360),
    [sym_metavariable] = ACTIONS(362),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [101] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1218),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [102] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1282),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [103] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1942),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1214),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(1176),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(97),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(340),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(384),
    [anon_sym_u8] = ACTIONS(342),
    [anon_sym_i8] = ACTIONS(342),
    [anon_sym_u16] = ACTIONS(342),
    [anon_sym_i16] = ACTIONS(342),
    [anon_sym_u32] = ACTIONS(342),
    [anon_sym_i32] = ACTIONS(342),
    [anon_sym_u64] = ACTIONS(342),
    [anon_sym_i64] = ACTIONS(342),
    [anon_sym_u128] = ACTIONS(342),
    [anon_sym_i128] = ACTIONS(342),
    [anon_sym_isize] = ACTIONS(342),
    [anon_sym_usize] = ACTIONS(342),
    [anon_sym_f32] = ACTIONS(342),
    [anon_sym_f64] = ACTIONS(342),
    [anon_sym_bool] = ACTIONS(342),
    [anon_sym_str] = ACTIONS(342),
    [anon_sym_char] = ACTIONS(342),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(344),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(346),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(348),
    [anon_sym_union] = ACTIONS(346),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(384),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(352),
    [anon_sym_AMP] = ACTIONS(388),
    [anon_sym_DOT_DOT] = ACTIONS(390),
    [anon_sym_DASH] = ACTIONS(384),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(354),
    [anon_sym_move] = ACTIONS(356),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(358),
    [sym_super] = ACTIONS(360),
    [sym_crate] = ACTIONS(360),
    [sym_metavariable] = ACTIONS(362),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [104] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1212),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [105] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1942),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1198),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(1176),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(97),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(340),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(384),
    [anon_sym_u8] = ACTIONS(342),
    [anon_sym_i8] = ACTIONS(342),
    [anon_sym_u16] = ACTIONS(342),
    [anon_sym_i16] = ACTIONS(342),
    [anon_sym_u32] = ACTIONS(342),
    [anon_sym_i32] = ACTIONS(342),
    [anon_sym_u64] = ACTIONS(342),
    [anon_sym_i64] = ACTIONS(342),
    [anon_sym_u128] = ACTIONS(342),
    [anon_sym_i128] = ACTIONS(342),
    [anon_sym_isize] = ACTIONS(342),
    [anon_sym_usize] = ACTIONS(342),
    [anon_sym_f32] = ACTIONS(342),
    [anon_sym_f64] = ACTIONS(342),
    [anon_sym_bool] = ACTIONS(342),
    [anon_sym_str] = ACTIONS(342),
    [anon_sym_char] = ACTIONS(342),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(344),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(346),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(348),
    [anon_sym_union] = ACTIONS(346),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(384),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(352),
    [anon_sym_AMP] = ACTIONS(388),
    [anon_sym_DOT_DOT] = ACTIONS(418),
    [anon_sym_DASH] = ACTIONS(384),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(354),
    [anon_sym_move] = ACTIONS(356),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(358),
    [sym_super] = ACTIONS(360),
    [sym_crate] = ACTIONS(360),
    [sym_metavariable] = ACTIONS(362),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [106] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1942),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1224),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(1176),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(97),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(340),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(384),
    [anon_sym_u8] = ACTIONS(342),
    [anon_sym_i8] = ACTIONS(342),
    [anon_sym_u16] = ACTIONS(342),
    [anon_sym_i16] = ACTIONS(342),
    [anon_sym_u32] = ACTIONS(342),
    [anon_sym_i32] = ACTIONS(342),
    [anon_sym_u64] = ACTIONS(342),
    [anon_sym_i64] = ACTIONS(342),
    [anon_sym_u128] = ACTIONS(342),
    [anon_sym_i128] = ACTIONS(342),
    [anon_sym_isize] = ACTIONS(342),
    [anon_sym_usize] = ACTIONS(342),
    [anon_sym_f32] = ACTIONS(342),
    [anon_sym_f64] = ACTIONS(342),
    [anon_sym_bool] = ACTIONS(342),
    [anon_sym_str] = ACTIONS(342),
    [anon_sym_char] = ACTIONS(342),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(344),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(346),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(348),
    [anon_sym_union] = ACTIONS(346),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(384),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(352),
    [anon_sym_AMP] = ACTIONS(388),
    [anon_sym_DOT_DOT] = ACTIONS(418),
    [anon_sym_DASH] = ACTIONS(384),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(354),
    [anon_sym_move] = ACTIONS(356),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(358),
    [sym_super] = ACTIONS(360),
    [sym_crate] = ACTIONS(360),
    [sym_metavariable] = ACTIONS(362),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [107] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1281),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [108] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1279),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [109] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1942),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1232),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(1176),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(97),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(340),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(384),
    [anon_sym_u8] = ACTIONS(342),
    [anon_sym_i8] = ACTIONS(342),
    [anon_sym_u16] = ACTIONS(342),
    [anon_sym_i16] = ACTIONS(342),
    [anon_sym_u32] = ACTIONS(342),
    [anon_sym_i32] = ACTIONS(342),
    [anon_sym_u64] = ACTIONS(342),
    [anon_sym_i64] = ACTIONS(342),
    [anon_sym_u128] = ACTIONS(342),
    [anon_sym_i128] = ACTIONS(342),
    [anon_sym_isize] = ACTIONS(342),
    [anon_sym_usize] = ACTIONS(342),
    [anon_sym_f32] = ACTIONS(342),
    [anon_sym_f64] = ACTIONS(342),
    [anon_sym_bool] = ACTIONS(342),
    [anon_sym_str] = ACTIONS(342),
    [anon_sym_char] = ACTIONS(342),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(344),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(346),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(348),
    [anon_sym_union] = ACTIONS(346),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(384),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(352),
    [anon_sym_AMP] = ACTIONS(388),
    [anon_sym_DOT_DOT] = ACTIONS(418),
    [anon_sym_DASH] = ACTIONS(384),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(354),
    [anon_sym_move] = ACTIONS(356),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(358),
    [sym_super] = ACTIONS(360),
    [sym_crate] = ACTIONS(360),
    [sym_metavariable] = ACTIONS(362),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [110] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1942),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1235),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(1176),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(97),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(340),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(384),
    [anon_sym_u8] = ACTIONS(342),
    [anon_sym_i8] = ACTIONS(342),
    [anon_sym_u16] = ACTIONS(342),
    [anon_sym_i16] = ACTIONS(342),
    [anon_sym_u32] = ACTIONS(342),
    [anon_sym_i32] = ACTIONS(342),
    [anon_sym_u64] = ACTIONS(342),
    [anon_sym_i64] = ACTIONS(342),
    [anon_sym_u128] = ACTIONS(342),
    [anon_sym_i128] = ACTIONS(342),
    [anon_sym_isize] = ACTIONS(342),
    [anon_sym_usize] = ACTIONS(342),
    [anon_sym_f32] = ACTIONS(342),
    [anon_sym_f64] = ACTIONS(342),
    [anon_sym_bool] = ACTIONS(342),
    [anon_sym_str] = ACTIONS(342),
    [anon_sym_char] = ACTIONS(342),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(344),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(346),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(348),
    [anon_sym_union] = ACTIONS(346),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(384),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(352),
    [anon_sym_AMP] = ACTIONS(388),
    [anon_sym_DOT_DOT] = ACTIONS(418),
    [anon_sym_DASH] = ACTIONS(384),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(354),
    [anon_sym_move] = ACTIONS(356),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(358),
    [sym_super] = ACTIONS(360),
    [sym_crate] = ACTIONS(360),
    [sym_metavariable] = ACTIONS(362),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [111] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1942),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1222),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(1176),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(97),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(340),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(384),
    [anon_sym_u8] = ACTIONS(342),
    [anon_sym_i8] = ACTIONS(342),
    [anon_sym_u16] = ACTIONS(342),
    [anon_sym_i16] = ACTIONS(342),
    [anon_sym_u32] = ACTIONS(342),
    [anon_sym_i32] = ACTIONS(342),
    [anon_sym_u64] = ACTIONS(342),
    [anon_sym_i64] = ACTIONS(342),
    [anon_sym_u128] = ACTIONS(342),
    [anon_sym_i128] = ACTIONS(342),
    [anon_sym_isize] = ACTIONS(342),
    [anon_sym_usize] = ACTIONS(342),
    [anon_sym_f32] = ACTIONS(342),
    [anon_sym_f64] = ACTIONS(342),
    [anon_sym_bool] = ACTIONS(342),
    [anon_sym_str] = ACTIONS(342),
    [anon_sym_char] = ACTIONS(342),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(344),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(346),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(348),
    [anon_sym_union] = ACTIONS(346),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(384),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(352),
    [anon_sym_AMP] = ACTIONS(388),
    [anon_sym_DOT_DOT] = ACTIONS(390),
    [anon_sym_DASH] = ACTIONS(384),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(354),
    [anon_sym_move] = ACTIONS(356),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(358),
    [sym_super] = ACTIONS(360),
    [sym_crate] = ACTIONS(360),
    [sym_metavariable] = ACTIONS(362),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [112] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1231),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [113] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1274),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [114] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1293),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [115] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1246),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [116] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1942),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1211),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(1176),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(97),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(340),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(384),
    [anon_sym_u8] = ACTIONS(342),
    [anon_sym_i8] = ACTIONS(342),
    [anon_sym_u16] = ACTIONS(342),
    [anon_sym_i16] = ACTIONS(342),
    [anon_sym_u32] = ACTIONS(342),
    [anon_sym_i32] = ACTIONS(342),
    [anon_sym_u64] = ACTIONS(342),
    [anon_sym_i64] = ACTIONS(342),
    [anon_sym_u128] = ACTIONS(342),
    [anon_sym_i128] = ACTIONS(342),
    [anon_sym_isize] = ACTIONS(342),
    [anon_sym_usize] = ACTIONS(342),
    [anon_sym_f32] = ACTIONS(342),
    [anon_sym_f64] = ACTIONS(342),
    [anon_sym_bool] = ACTIONS(342),
    [anon_sym_str] = ACTIONS(342),
    [anon_sym_char] = ACTIONS(342),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(344),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(346),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(348),
    [anon_sym_union] = ACTIONS(346),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(384),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(352),
    [anon_sym_AMP] = ACTIONS(388),
    [anon_sym_DOT_DOT] = ACTIONS(390),
    [anon_sym_DASH] = ACTIONS(384),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(354),
    [anon_sym_move] = ACTIONS(356),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(358),
    [sym_super] = ACTIONS(360),
    [sym_crate] = ACTIONS(360),
    [sym_metavariable] = ACTIONS(362),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [117] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1294),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [118] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1264),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [119] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1209),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [120] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1942),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1248),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(1176),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(97),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(340),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(384),
    [anon_sym_u8] = ACTIONS(342),
    [anon_sym_i8] = ACTIONS(342),
    [anon_sym_u16] = ACTIONS(342),
    [anon_sym_i16] = ACTIONS(342),
    [anon_sym_u32] = ACTIONS(342),
    [anon_sym_i32] = ACTIONS(342),
    [anon_sym_u64] = ACTIONS(342),
    [anon_sym_i64] = ACTIONS(342),
    [anon_sym_u128] = ACTIONS(342),
    [anon_sym_i128] = ACTIONS(342),
    [anon_sym_isize] = ACTIONS(342),
    [anon_sym_usize] = ACTIONS(342),
    [anon_sym_f32] = ACTIONS(342),
    [anon_sym_f64] = ACTIONS(342),
    [anon_sym_bool] = ACTIONS(342),
    [anon_sym_str] = ACTIONS(342),
    [anon_sym_char] = ACTIONS(342),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(344),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(346),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(348),
    [anon_sym_union] = ACTIONS(346),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(384),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(352),
    [anon_sym_AMP] = ACTIONS(388),
    [anon_sym_DOT_DOT] = ACTIONS(390),
    [anon_sym_DASH] = ACTIONS(384),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(354),
    [anon_sym_move] = ACTIONS(356),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(358),
    [sym_super] = ACTIONS(360),
    [sym_crate] = ACTIONS(360),
    [sym_metavariable] = ACTIONS(362),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [121] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1233),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [122] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1267),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [123] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1229),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(214),
    [sym_match_expression] = STATE(214),
    [sym_while_expression] = STATE(214),
    [sym_loop_expression] = STATE(214),
    [sym_for_expression] = STATE(214),
    [sym_const_block] = STATE(214),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2546),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(214),
    [sym_async_block] = STATE(214),
    [sym_block] = STATE(214),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(632),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(634),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(636),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(638),
    [anon_sym_if] = ACTIONS(640),
    [anon_sym_loop] = ACTIONS(642),
    [anon_sym_match] = ACTIONS(644),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(646),
    [anon_sym_while] = ACTIONS(648),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [124] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1277),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [125] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1942),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1237),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(1176),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(97),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(340),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(384),
    [anon_sym_u8] = ACTIONS(342),
    [anon_sym_i8] = ACTIONS(342),
    [anon_sym_u16] = ACTIONS(342),
    [anon_sym_i16] = ACTIONS(342),
    [anon_sym_u32] = ACTIONS(342),
    [anon_sym_i32] = ACTIONS(342),
    [anon_sym_u64] = ACTIONS(342),
    [anon_sym_i64] = ACTIONS(342),
    [anon_sym_u128] = ACTIONS(342),
    [anon_sym_i128] = ACTIONS(342),
    [anon_sym_isize] = ACTIONS(342),
    [anon_sym_usize] = ACTIONS(342),
    [anon_sym_f32] = ACTIONS(342),
    [anon_sym_f64] = ACTIONS(342),
    [anon_sym_bool] = ACTIONS(342),
    [anon_sym_str] = ACTIONS(342),
    [anon_sym_char] = ACTIONS(342),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(344),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(346),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(348),
    [anon_sym_union] = ACTIONS(346),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(384),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(352),
    [anon_sym_AMP] = ACTIONS(388),
    [anon_sym_DOT_DOT] = ACTIONS(390),
    [anon_sym_DASH] = ACTIONS(384),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(354),
    [anon_sym_move] = ACTIONS(356),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(358),
    [sym_super] = ACTIONS(360),
    [sym_crate] = ACTIONS(360),
    [sym_metavariable] = ACTIONS(362),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [126] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1278),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [127] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1942),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1268),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(1176),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(97),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(340),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(384),
    [anon_sym_u8] = ACTIONS(342),
    [anon_sym_i8] = ACTIONS(342),
    [anon_sym_u16] = ACTIONS(342),
    [anon_sym_i16] = ACTIONS(342),
    [anon_sym_u32] = ACTIONS(342),
    [anon_sym_i32] = ACTIONS(342),
    [anon_sym_u64] = ACTIONS(342),
    [anon_sym_i64] = ACTIONS(342),
    [anon_sym_u128] = ACTIONS(342),
    [anon_sym_i128] = ACTIONS(342),
    [anon_sym_isize] = ACTIONS(342),
    [anon_sym_usize] = ACTIONS(342),
    [anon_sym_f32] = ACTIONS(342),
    [anon_sym_f64] = ACTIONS(342),
    [anon_sym_bool] = ACTIONS(342),
    [anon_sym_str] = ACTIONS(342),
    [anon_sym_char] = ACTIONS(342),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(344),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(346),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(348),
    [anon_sym_union] = ACTIONS(346),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(384),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(352),
    [anon_sym_AMP] = ACTIONS(388),
    [anon_sym_DOT_DOT] = ACTIONS(390),
    [anon_sym_DASH] = ACTIONS(384),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(354),
    [anon_sym_move] = ACTIONS(356),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(358),
    [sym_super] = ACTIONS(360),
    [sym_crate] = ACTIONS(360),
    [sym_metavariable] = ACTIONS(362),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [128] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1942),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1236),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(1176),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(97),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(340),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(384),
    [anon_sym_u8] = ACTIONS(342),
    [anon_sym_i8] = ACTIONS(342),
    [anon_sym_u16] = ACTIONS(342),
    [anon_sym_i16] = ACTIONS(342),
    [anon_sym_u32] = ACTIONS(342),
    [anon_sym_i32] = ACTIONS(342),
    [anon_sym_u64] = ACTIONS(342),
    [anon_sym_i64] = ACTIONS(342),
    [anon_sym_u128] = ACTIONS(342),
    [anon_sym_i128] = ACTIONS(342),
    [anon_sym_isize] = ACTIONS(342),
    [anon_sym_usize] = ACTIONS(342),
    [anon_sym_f32] = ACTIONS(342),
    [anon_sym_f64] = ACTIONS(342),
    [anon_sym_bool] = ACTIONS(342),
    [anon_sym_str] = ACTIONS(342),
    [anon_sym_char] = ACTIONS(342),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(344),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(346),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(348),
    [anon_sym_union] = ACTIONS(346),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(384),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(352),
    [anon_sym_AMP] = ACTIONS(388),
    [anon_sym_DOT_DOT] = ACTIONS(418),
    [anon_sym_DASH] = ACTIONS(384),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(354),
    [anon_sym_move] = ACTIONS(356),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(358),
    [sym_super] = ACTIONS(360),
    [sym_crate] = ACTIONS(360),
    [sym_metavariable] = ACTIONS(362),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [129] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1136),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(420),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [130] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1285),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [131] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1286),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(218),
    [sym_match_expression] = STATE(218),
    [sym_while_expression] = STATE(218),
    [sym_loop_expression] = STATE(218),
    [sym_for_expression] = STATE(218),
    [sym_const_block] = STATE(218),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2546),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(218),
    [sym_async_block] = STATE(218),
    [sym_block] = STATE(218),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(632),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(634),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(636),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(638),
    [anon_sym_if] = ACTIONS(640),
    [anon_sym_loop] = ACTIONS(642),
    [anon_sym_match] = ACTIONS(644),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(646),
    [anon_sym_while] = ACTIONS(648),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [132] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1140),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(420),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [133] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1271),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [134] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1292),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(214),
    [sym_match_expression] = STATE(214),
    [sym_while_expression] = STATE(214),
    [sym_loop_expression] = STATE(214),
    [sym_for_expression] = STATE(214),
    [sym_const_block] = STATE(214),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2546),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(214),
    [sym_async_block] = STATE(214),
    [sym_block] = STATE(214),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(632),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(634),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(636),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(638),
    [anon_sym_if] = ACTIONS(640),
    [anon_sym_loop] = ACTIONS(642),
    [anon_sym_match] = ACTIONS(644),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(646),
    [anon_sym_while] = ACTIONS(648),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [135] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1202),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [136] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1942),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1240),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(1176),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(97),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(340),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(384),
    [anon_sym_u8] = ACTIONS(342),
    [anon_sym_i8] = ACTIONS(342),
    [anon_sym_u16] = ACTIONS(342),
    [anon_sym_i16] = ACTIONS(342),
    [anon_sym_u32] = ACTIONS(342),
    [anon_sym_i32] = ACTIONS(342),
    [anon_sym_u64] = ACTIONS(342),
    [anon_sym_i64] = ACTIONS(342),
    [anon_sym_u128] = ACTIONS(342),
    [anon_sym_i128] = ACTIONS(342),
    [anon_sym_isize] = ACTIONS(342),
    [anon_sym_usize] = ACTIONS(342),
    [anon_sym_f32] = ACTIONS(342),
    [anon_sym_f64] = ACTIONS(342),
    [anon_sym_bool] = ACTIONS(342),
    [anon_sym_str] = ACTIONS(342),
    [anon_sym_char] = ACTIONS(342),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(344),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(346),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(348),
    [anon_sym_union] = ACTIONS(346),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(384),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(352),
    [anon_sym_AMP] = ACTIONS(388),
    [anon_sym_DOT_DOT] = ACTIONS(418),
    [anon_sym_DASH] = ACTIONS(384),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(354),
    [anon_sym_move] = ACTIONS(356),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(358),
    [sym_super] = ACTIONS(360),
    [sym_crate] = ACTIONS(360),
    [sym_metavariable] = ACTIONS(362),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [137] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1258),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [138] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1290),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [139] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1143),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(420),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [140] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1288),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [141] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1144),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(420),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [142] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1942),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1217),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(1176),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(97),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(340),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(384),
    [anon_sym_u8] = ACTIONS(342),
    [anon_sym_i8] = ACTIONS(342),
    [anon_sym_u16] = ACTIONS(342),
    [anon_sym_i16] = ACTIONS(342),
    [anon_sym_u32] = ACTIONS(342),
    [anon_sym_i32] = ACTIONS(342),
    [anon_sym_u64] = ACTIONS(342),
    [anon_sym_i64] = ACTIONS(342),
    [anon_sym_u128] = ACTIONS(342),
    [anon_sym_i128] = ACTIONS(342),
    [anon_sym_isize] = ACTIONS(342),
    [anon_sym_usize] = ACTIONS(342),
    [anon_sym_f32] = ACTIONS(342),
    [anon_sym_f64] = ACTIONS(342),
    [anon_sym_bool] = ACTIONS(342),
    [anon_sym_str] = ACTIONS(342),
    [anon_sym_char] = ACTIONS(342),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(344),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(346),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(348),
    [anon_sym_union] = ACTIONS(346),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(384),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(352),
    [anon_sym_AMP] = ACTIONS(388),
    [anon_sym_DOT_DOT] = ACTIONS(390),
    [anon_sym_DASH] = ACTIONS(384),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(354),
    [anon_sym_move] = ACTIONS(356),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(358),
    [sym_super] = ACTIONS(360),
    [sym_crate] = ACTIONS(360),
    [sym_metavariable] = ACTIONS(362),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [143] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1146),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(420),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [144] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1942),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1201),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(1176),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(97),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(340),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(384),
    [anon_sym_u8] = ACTIONS(342),
    [anon_sym_i8] = ACTIONS(342),
    [anon_sym_u16] = ACTIONS(342),
    [anon_sym_i16] = ACTIONS(342),
    [anon_sym_u32] = ACTIONS(342),
    [anon_sym_i32] = ACTIONS(342),
    [anon_sym_u64] = ACTIONS(342),
    [anon_sym_i64] = ACTIONS(342),
    [anon_sym_u128] = ACTIONS(342),
    [anon_sym_i128] = ACTIONS(342),
    [anon_sym_isize] = ACTIONS(342),
    [anon_sym_usize] = ACTIONS(342),
    [anon_sym_f32] = ACTIONS(342),
    [anon_sym_f64] = ACTIONS(342),
    [anon_sym_bool] = ACTIONS(342),
    [anon_sym_str] = ACTIONS(342),
    [anon_sym_char] = ACTIONS(342),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(344),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(346),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(348),
    [anon_sym_union] = ACTIONS(346),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(384),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(352),
    [anon_sym_AMP] = ACTIONS(388),
    [anon_sym_DOT_DOT] = ACTIONS(390),
    [anon_sym_DASH] = ACTIONS(384),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(354),
    [anon_sym_move] = ACTIONS(356),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(358),
    [sym_super] = ACTIONS(360),
    [sym_crate] = ACTIONS(360),
    [sym_metavariable] = ACTIONS(362),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [145] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1227),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [146] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1147),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(420),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [147] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1225),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [148] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1142),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(420),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [149] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1155),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(420),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [150] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1942),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1200),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(1176),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(97),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(340),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(384),
    [anon_sym_u8] = ACTIONS(342),
    [anon_sym_i8] = ACTIONS(342),
    [anon_sym_u16] = ACTIONS(342),
    [anon_sym_i16] = ACTIONS(342),
    [anon_sym_u32] = ACTIONS(342),
    [anon_sym_i32] = ACTIONS(342),
    [anon_sym_u64] = ACTIONS(342),
    [anon_sym_i64] = ACTIONS(342),
    [anon_sym_u128] = ACTIONS(342),
    [anon_sym_i128] = ACTIONS(342),
    [anon_sym_isize] = ACTIONS(342),
    [anon_sym_usize] = ACTIONS(342),
    [anon_sym_f32] = ACTIONS(342),
    [anon_sym_f64] = ACTIONS(342),
    [anon_sym_bool] = ACTIONS(342),
    [anon_sym_str] = ACTIONS(342),
    [anon_sym_char] = ACTIONS(342),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(344),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(346),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(348),
    [anon_sym_union] = ACTIONS(346),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(384),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(352),
    [anon_sym_AMP] = ACTIONS(388),
    [anon_sym_DOT_DOT] = ACTIONS(418),
    [anon_sym_DASH] = ACTIONS(384),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(354),
    [anon_sym_move] = ACTIONS(356),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(358),
    [sym_super] = ACTIONS(360),
    [sym_crate] = ACTIONS(360),
    [sym_metavariable] = ACTIONS(362),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [151] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1942),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1259),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(1176),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(97),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(340),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(384),
    [anon_sym_u8] = ACTIONS(342),
    [anon_sym_i8] = ACTIONS(342),
    [anon_sym_u16] = ACTIONS(342),
    [anon_sym_i16] = ACTIONS(342),
    [anon_sym_u32] = ACTIONS(342),
    [anon_sym_i32] = ACTIONS(342),
    [anon_sym_u64] = ACTIONS(342),
    [anon_sym_i64] = ACTIONS(342),
    [anon_sym_u128] = ACTIONS(342),
    [anon_sym_i128] = ACTIONS(342),
    [anon_sym_isize] = ACTIONS(342),
    [anon_sym_usize] = ACTIONS(342),
    [anon_sym_f32] = ACTIONS(342),
    [anon_sym_f64] = ACTIONS(342),
    [anon_sym_bool] = ACTIONS(342),
    [anon_sym_str] = ACTIONS(342),
    [anon_sym_char] = ACTIONS(342),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(344),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(346),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(348),
    [anon_sym_union] = ACTIONS(346),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(384),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(352),
    [anon_sym_AMP] = ACTIONS(388),
    [anon_sym_DOT_DOT] = ACTIONS(390),
    [anon_sym_DASH] = ACTIONS(384),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(354),
    [anon_sym_move] = ACTIONS(356),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(358),
    [sym_super] = ACTIONS(360),
    [sym_crate] = ACTIONS(360),
    [sym_metavariable] = ACTIONS(362),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [152] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1942),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1249),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(1176),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(97),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(340),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(384),
    [anon_sym_u8] = ACTIONS(342),
    [anon_sym_i8] = ACTIONS(342),
    [anon_sym_u16] = ACTIONS(342),
    [anon_sym_i16] = ACTIONS(342),
    [anon_sym_u32] = ACTIONS(342),
    [anon_sym_i32] = ACTIONS(342),
    [anon_sym_u64] = ACTIONS(342),
    [anon_sym_i64] = ACTIONS(342),
    [anon_sym_u128] = ACTIONS(342),
    [anon_sym_i128] = ACTIONS(342),
    [anon_sym_isize] = ACTIONS(342),
    [anon_sym_usize] = ACTIONS(342),
    [anon_sym_f32] = ACTIONS(342),
    [anon_sym_f64] = ACTIONS(342),
    [anon_sym_bool] = ACTIONS(342),
    [anon_sym_str] = ACTIONS(342),
    [anon_sym_char] = ACTIONS(342),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(344),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(346),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(348),
    [anon_sym_union] = ACTIONS(346),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(384),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(352),
    [anon_sym_AMP] = ACTIONS(388),
    [anon_sym_DOT_DOT] = ACTIONS(418),
    [anon_sym_DASH] = ACTIONS(384),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(354),
    [anon_sym_move] = ACTIONS(356),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(358),
    [sym_super] = ACTIONS(360),
    [sym_crate] = ACTIONS(360),
    [sym_metavariable] = ACTIONS(362),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [153] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1942),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1203),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(1176),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(97),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(340),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(384),
    [anon_sym_u8] = ACTIONS(342),
    [anon_sym_i8] = ACTIONS(342),
    [anon_sym_u16] = ACTIONS(342),
    [anon_sym_i16] = ACTIONS(342),
    [anon_sym_u32] = ACTIONS(342),
    [anon_sym_i32] = ACTIONS(342),
    [anon_sym_u64] = ACTIONS(342),
    [anon_sym_i64] = ACTIONS(342),
    [anon_sym_u128] = ACTIONS(342),
    [anon_sym_i128] = ACTIONS(342),
    [anon_sym_isize] = ACTIONS(342),
    [anon_sym_usize] = ACTIONS(342),
    [anon_sym_f32] = ACTIONS(342),
    [anon_sym_f64] = ACTIONS(342),
    [anon_sym_bool] = ACTIONS(342),
    [anon_sym_str] = ACTIONS(342),
    [anon_sym_char] = ACTIONS(342),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(344),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(346),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(348),
    [anon_sym_union] = ACTIONS(346),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(384),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(352),
    [anon_sym_AMP] = ACTIONS(388),
    [anon_sym_DOT_DOT] = ACTIONS(418),
    [anon_sym_DASH] = ACTIONS(384),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(354),
    [anon_sym_move] = ACTIONS(356),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(358),
    [sym_super] = ACTIONS(360),
    [sym_crate] = ACTIONS(360),
    [sym_metavariable] = ACTIONS(362),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [154] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1276),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [155] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1942),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1210),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(1176),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(97),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(340),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(384),
    [anon_sym_u8] = ACTIONS(342),
    [anon_sym_i8] = ACTIONS(342),
    [anon_sym_u16] = ACTIONS(342),
    [anon_sym_i16] = ACTIONS(342),
    [anon_sym_u32] = ACTIONS(342),
    [anon_sym_i32] = ACTIONS(342),
    [anon_sym_u64] = ACTIONS(342),
    [anon_sym_i64] = ACTIONS(342),
    [anon_sym_u128] = ACTIONS(342),
    [anon_sym_i128] = ACTIONS(342),
    [anon_sym_isize] = ACTIONS(342),
    [anon_sym_usize] = ACTIONS(342),
    [anon_sym_f32] = ACTIONS(342),
    [anon_sym_f64] = ACTIONS(342),
    [anon_sym_bool] = ACTIONS(342),
    [anon_sym_str] = ACTIONS(342),
    [anon_sym_char] = ACTIONS(342),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(344),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(346),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(348),
    [anon_sym_union] = ACTIONS(346),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(384),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(352),
    [anon_sym_AMP] = ACTIONS(388),
    [anon_sym_DOT_DOT] = ACTIONS(390),
    [anon_sym_DASH] = ACTIONS(384),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(354),
    [anon_sym_move] = ACTIONS(356),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(358),
    [sym_super] = ACTIONS(360),
    [sym_crate] = ACTIONS(360),
    [sym_metavariable] = ACTIONS(362),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [156] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1254),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [157] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1265),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [158] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(871),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [159] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1269),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [160] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1250),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [161] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1942),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1253),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(1176),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(97),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(340),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(384),
    [anon_sym_u8] = ACTIONS(342),
    [anon_sym_i8] = ACTIONS(342),
    [anon_sym_u16] = ACTIONS(342),
    [anon_sym_i16] = ACTIONS(342),
    [anon_sym_u32] = ACTIONS(342),
    [anon_sym_i32] = ACTIONS(342),
    [anon_sym_u64] = ACTIONS(342),
    [anon_sym_i64] = ACTIONS(342),
    [anon_sym_u128] = ACTIONS(342),
    [anon_sym_i128] = ACTIONS(342),
    [anon_sym_isize] = ACTIONS(342),
    [anon_sym_usize] = ACTIONS(342),
    [anon_sym_f32] = ACTIONS(342),
    [anon_sym_f64] = ACTIONS(342),
    [anon_sym_bool] = ACTIONS(342),
    [anon_sym_str] = ACTIONS(342),
    [anon_sym_char] = ACTIONS(342),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(344),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(346),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(348),
    [anon_sym_union] = ACTIONS(346),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(384),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(352),
    [anon_sym_AMP] = ACTIONS(388),
    [anon_sym_DOT_DOT] = ACTIONS(418),
    [anon_sym_DASH] = ACTIONS(384),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(354),
    [anon_sym_move] = ACTIONS(356),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(358),
    [sym_super] = ACTIONS(360),
    [sym_crate] = ACTIONS(360),
    [sym_metavariable] = ACTIONS(362),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [162] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1942),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(871),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(1176),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(97),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(340),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(384),
    [anon_sym_u8] = ACTIONS(342),
    [anon_sym_i8] = ACTIONS(342),
    [anon_sym_u16] = ACTIONS(342),
    [anon_sym_i16] = ACTIONS(342),
    [anon_sym_u32] = ACTIONS(342),
    [anon_sym_i32] = ACTIONS(342),
    [anon_sym_u64] = ACTIONS(342),
    [anon_sym_i64] = ACTIONS(342),
    [anon_sym_u128] = ACTIONS(342),
    [anon_sym_i128] = ACTIONS(342),
    [anon_sym_isize] = ACTIONS(342),
    [anon_sym_usize] = ACTIONS(342),
    [anon_sym_f32] = ACTIONS(342),
    [anon_sym_f64] = ACTIONS(342),
    [anon_sym_bool] = ACTIONS(342),
    [anon_sym_str] = ACTIONS(342),
    [anon_sym_char] = ACTIONS(342),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(344),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(346),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(348),
    [anon_sym_union] = ACTIONS(346),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(384),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(352),
    [anon_sym_AMP] = ACTIONS(388),
    [anon_sym_DOT_DOT] = ACTIONS(418),
    [anon_sym_DASH] = ACTIONS(384),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(354),
    [anon_sym_move] = ACTIONS(356),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(358),
    [sym_super] = ACTIONS(360),
    [sym_crate] = ACTIONS(360),
    [sym_metavariable] = ACTIONS(362),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [163] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(871),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(420),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [164] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1284),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [165] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1942),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1241),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(1176),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(97),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(340),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(384),
    [anon_sym_u8] = ACTIONS(342),
    [anon_sym_i8] = ACTIONS(342),
    [anon_sym_u16] = ACTIONS(342),
    [anon_sym_i16] = ACTIONS(342),
    [anon_sym_u32] = ACTIONS(342),
    [anon_sym_i32] = ACTIONS(342),
    [anon_sym_u64] = ACTIONS(342),
    [anon_sym_i64] = ACTIONS(342),
    [anon_sym_u128] = ACTIONS(342),
    [anon_sym_i128] = ACTIONS(342),
    [anon_sym_isize] = ACTIONS(342),
    [anon_sym_usize] = ACTIONS(342),
    [anon_sym_f32] = ACTIONS(342),
    [anon_sym_f64] = ACTIONS(342),
    [anon_sym_bool] = ACTIONS(342),
    [anon_sym_str] = ACTIONS(342),
    [anon_sym_char] = ACTIONS(342),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(344),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(346),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(348),
    [anon_sym_union] = ACTIONS(346),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(384),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(352),
    [anon_sym_AMP] = ACTIONS(388),
    [anon_sym_DOT_DOT] = ACTIONS(418),
    [anon_sym_DASH] = ACTIONS(384),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(354),
    [anon_sym_move] = ACTIONS(356),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(358),
    [sym_super] = ACTIONS(360),
    [sym_crate] = ACTIONS(360),
    [sym_metavariable] = ACTIONS(362),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [166] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1942),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1247),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(1176),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(97),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(340),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(384),
    [anon_sym_u8] = ACTIONS(342),
    [anon_sym_i8] = ACTIONS(342),
    [anon_sym_u16] = ACTIONS(342),
    [anon_sym_i16] = ACTIONS(342),
    [anon_sym_u32] = ACTIONS(342),
    [anon_sym_i32] = ACTIONS(342),
    [anon_sym_u64] = ACTIONS(342),
    [anon_sym_i64] = ACTIONS(342),
    [anon_sym_u128] = ACTIONS(342),
    [anon_sym_i128] = ACTIONS(342),
    [anon_sym_isize] = ACTIONS(342),
    [anon_sym_usize] = ACTIONS(342),
    [anon_sym_f32] = ACTIONS(342),
    [anon_sym_f64] = ACTIONS(342),
    [anon_sym_bool] = ACTIONS(342),
    [anon_sym_str] = ACTIONS(342),
    [anon_sym_char] = ACTIONS(342),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(344),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(346),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(348),
    [anon_sym_union] = ACTIONS(346),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(384),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(352),
    [anon_sym_AMP] = ACTIONS(388),
    [anon_sym_DOT_DOT] = ACTIONS(418),
    [anon_sym_DASH] = ACTIONS(384),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(354),
    [anon_sym_move] = ACTIONS(356),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(358),
    [sym_super] = ACTIONS(360),
    [sym_crate] = ACTIONS(360),
    [sym_metavariable] = ACTIONS(362),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [167] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1129),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(420),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [168] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1204),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [169] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1942),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1244),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(1176),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(97),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(340),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(384),
    [anon_sym_u8] = ACTIONS(342),
    [anon_sym_i8] = ACTIONS(342),
    [anon_sym_u16] = ACTIONS(342),
    [anon_sym_i16] = ACTIONS(342),
    [anon_sym_u32] = ACTIONS(342),
    [anon_sym_i32] = ACTIONS(342),
    [anon_sym_u64] = ACTIONS(342),
    [anon_sym_i64] = ACTIONS(342),
    [anon_sym_u128] = ACTIONS(342),
    [anon_sym_i128] = ACTIONS(342),
    [anon_sym_isize] = ACTIONS(342),
    [anon_sym_usize] = ACTIONS(342),
    [anon_sym_f32] = ACTIONS(342),
    [anon_sym_f64] = ACTIONS(342),
    [anon_sym_bool] = ACTIONS(342),
    [anon_sym_str] = ACTIONS(342),
    [anon_sym_char] = ACTIONS(342),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(344),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(346),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(348),
    [anon_sym_union] = ACTIONS(346),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(384),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(352),
    [anon_sym_AMP] = ACTIONS(388),
    [anon_sym_DOT_DOT] = ACTIONS(418),
    [anon_sym_DASH] = ACTIONS(384),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(354),
    [anon_sym_move] = ACTIONS(356),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(358),
    [sym_super] = ACTIONS(360),
    [sym_crate] = ACTIONS(360),
    [sym_metavariable] = ACTIONS(362),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [170] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1251),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [171] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1128),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(420),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [172] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1152),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(420),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [173] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1283),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [174] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1275),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [175] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1157),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(420),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [176] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1199),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [177] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1158),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(777),
    [sym_match_expression] = STATE(777),
    [sym_while_expression] = STATE(777),
    [sym_loop_expression] = STATE(777),
    [sym_for_expression] = STATE(777),
    [sym_const_block] = STATE(777),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2517),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(777),
    [sym_async_block] = STATE(777),
    [sym_block] = STATE(777),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(284),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(290),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(292),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(296),
    [anon_sym_if] = ACTIONS(298),
    [anon_sym_loop] = ACTIONS(300),
    [anon_sym_match] = ACTIONS(302),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(304),
    [anon_sym_while] = ACTIONS(306),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(420),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [178] = {
    [sym_bracketed_type] = STATE(2527),
    [sym_generic_function] = STATE(777),
    [sym_generic_type_with_turbofish] = STATE(1961),
    [sym__expression_except_range] = STATE(777),
    [sym__expression] = STATE(1215),
    [sym_macro_invocation] = STATE(777),
    [sym_scoped_identifier] = STATE(792),
    [sym_scoped_type_identifier_in_expression_position] = STATE(2247),
    [sym_range_expression] = STATE(1054),
    [sym_unary_expression] = STATE(777),
    [sym_try_expression] = STATE(777),
    [sym_reference_expression] = STATE(777),
    [sym_binary_expression] = STATE(777),
    [sym_assignment_expression] = STATE(777),
    [sym_compound_assignment_expr] = STATE(777),
    [sym_type_cast_expression] = STATE(777),
    [sym_return_expression] = STATE(777),
    [sym_yield_expression] = STATE(777),
    [sym_call_expression] = STATE(777),
    [sym_array_expression] = STATE(777),
    [sym_parenthesized_expression] = STATE(777),
    [sym_tuple_expression] = STATE(777),
    [sym_unit_expression] = STATE(777),
    [sym_struct_expression] = STATE(777),
    [sym_if_expression] = STATE(218),
    [sym_match_expression] = STATE(218),
    [sym_while_expression] = STATE(218),
    [sym_loop_expression] = STATE(218),
    [sym_for_expression] = STATE(218),
    [sym_const_block] = STATE(218),
    [sym_closure_expression] = STATE(777),
    [sym_closure_parameters] = STATE(94),
    [sym_loop_label] = STATE(2546),
    [sym_break_expression] = STATE(777),
    [sym_continue_expression] = STATE(777),
    [sym_index_expression] = STATE(777),
    [sym_await_expression] = STATE(777),
    [sym_field_expression] = STATE(778),
    [sym_unsafe_block] = STATE(218),
    [sym_async_block] = STATE(218),
    [sym_block] = STATE(218),
    [sym__literal] = STATE(777),
    [sym_string_literal] = STATE(815),
    [sym_boolean_literal] = STATE(815),
    [sym_identifier] = ACTIONS(280),
    [anon_sym_LPAREN] = ACTIONS(13),
    [anon_sym_LBRACE] = ACTIONS(632),
    [anon_sym_LBRACK] = ACTIONS(17),
    [anon_sym_STAR] = ACTIONS(19),
    [anon_sym_u8] = ACTIONS(21),
    [anon_sym_i8] = ACTIONS(21),
    [anon_sym_u16] = ACTIONS(21),
    [anon_sym_i16] = ACTIONS(21),
    [anon_sym_u32] = ACTIONS(21),
    [anon_sym_i32] = ACTIONS(21),
    [anon_sym_u64] = ACTIONS(21),
    [anon_sym_i64] = ACTIONS(21),
    [anon_sym_u128] = ACTIONS(21),
    [anon_sym_i128] = ACTIONS(21),
    [anon_sym_isize] = ACTIONS(21),
    [anon_sym_usize] = ACTIONS(21),
    [anon_sym_f32] = ACTIONS(21),
    [anon_sym_f64] = ACTIONS(21),
    [anon_sym_bool] = ACTIONS(21),
    [anon_sym_str] = ACTIONS(21),
    [anon_sym_char] = ACTIONS(21),
    [anon_sym_SQUOTE] = ACTIONS(23),
    [anon_sym_async] = ACTIONS(634),
    [anon_sym_break] = ACTIONS(27),
    [anon_sym_const] = ACTIONS(636),
    [anon_sym_continue] = ACTIONS(31),
    [anon_sym_default] = ACTIONS(294),
    [anon_sym_for] = ACTIONS(638),
    [anon_sym_if] = ACTIONS(640),
    [anon_sym_loop] = ACTIONS(642),
    [anon_sym_match] = ACTIONS(644),
    [anon_sym_return] = ACTIONS(55),
    [anon_sym_union] = ACTIONS(294),
    [anon_sym_unsafe] = ACTIONS(646),
    [anon_sym_while] = ACTIONS(648),
    [anon_sym_BANG] = ACTIONS(19),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(79),
    [anon_sym_AMP] = ACTIONS(81),
    [anon_sym_DOT_DOT] = ACTIONS(83),
    [anon_sym_DASH] = ACTIONS(19),
    [anon_sym_PIPE] = ACTIONS(85),
    [anon_sym_yield] = ACTIONS(87),
    [anon_sym_move] = ACTIONS(89),
    [sym_integer_literal] = ACTIONS(91),
    [aux_sym_string_literal_token1] = ACTIONS(93),
    [sym_char_literal] = ACTIONS(91),
    [anon_sym_true] = ACTIONS(95),
    [anon_sym_false] = ACTIONS(95),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(97),
    [sym_super] = ACTIONS(99),
    [sym_crate] = ACTIONS(99),
    [sym_metavariable] = ACTIONS(103),
    [sym_raw_string_literal] = ACTIONS(91),
    [sym_float_literal] = ACTIONS(91),
    [sym_block_comment] = ACTIONS(3),
  },
  [179] = {
    [sym_attribute_item] = STATE(192),
    [sym_function_modifiers] = STATE(2472),
    [sym_self_parameter] = STATE(1901),
    [sym_variadic_parameter] = STATE(1901),
    [sym_parameter] = STATE(1901),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(1781),
    [sym_bracketed_type] = STATE(2508),
    [sym_lifetime] = STATE(1902),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2509),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1728),
    [sym_scoped_identifier] = STATE(1505),
    [sym_scoped_type_identifier] = STATE(1474),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1748),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(650),
    [anon_sym_LPAREN] = ACTIONS(652),
    [anon_sym_RPAREN] = ACTIONS(654),
    [anon_sym_LBRACK] = ACTIONS(656),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(660),
    [anon_sym_i8] = ACTIONS(660),
    [anon_sym_u16] = ACTIONS(660),
    [anon_sym_i16] = ACTIONS(660),
    [anon_sym_u32] = ACTIONS(660),
    [anon_sym_i32] = ACTIONS(660),
    [anon_sym_u64] = ACTIONS(660),
    [anon_sym_i64] = ACTIONS(660),
    [anon_sym_u128] = ACTIONS(660),
    [anon_sym_i128] = ACTIONS(660),
    [anon_sym_isize] = ACTIONS(660),
    [anon_sym_usize] = ACTIONS(660),
    [anon_sym_f32] = ACTIONS(660),
    [anon_sym_f64] = ACTIONS(660),
    [anon_sym_bool] = ACTIONS(660),
    [anon_sym_str] = ACTIONS(660),
    [anon_sym_char] = ACTIONS(660),
    [anon_sym_SQUOTE] = ACTIONS(662),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(666),
    [anon_sym_default] = ACTIONS(668),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(676),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_POUND] = ACTIONS(678),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_COMMA] = ACTIONS(682),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(688),
    [anon_sym__] = ACTIONS(690),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(694),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_mutable_specifier] = ACTIONS(698),
    [anon_sym_DOT_DOT] = ACTIONS(700),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(710),
    [sym_super] = ACTIONS(712),
    [sym_crate] = ACTIONS(712),
    [sym_metavariable] = ACTIONS(714),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [180] = {
    [sym_attribute_item] = STATE(192),
    [sym_function_modifiers] = STATE(2472),
    [sym_self_parameter] = STATE(1901),
    [sym_variadic_parameter] = STATE(1901),
    [sym_parameter] = STATE(1901),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(1781),
    [sym_bracketed_type] = STATE(2508),
    [sym_lifetime] = STATE(1902),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2509),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1728),
    [sym_scoped_identifier] = STATE(1505),
    [sym_scoped_type_identifier] = STATE(1474),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1748),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(650),
    [anon_sym_LPAREN] = ACTIONS(652),
    [anon_sym_RPAREN] = ACTIONS(716),
    [anon_sym_LBRACK] = ACTIONS(656),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(660),
    [anon_sym_i8] = ACTIONS(660),
    [anon_sym_u16] = ACTIONS(660),
    [anon_sym_i16] = ACTIONS(660),
    [anon_sym_u32] = ACTIONS(660),
    [anon_sym_i32] = ACTIONS(660),
    [anon_sym_u64] = ACTIONS(660),
    [anon_sym_i64] = ACTIONS(660),
    [anon_sym_u128] = ACTIONS(660),
    [anon_sym_i128] = ACTIONS(660),
    [anon_sym_isize] = ACTIONS(660),
    [anon_sym_usize] = ACTIONS(660),
    [anon_sym_f32] = ACTIONS(660),
    [anon_sym_f64] = ACTIONS(660),
    [anon_sym_bool] = ACTIONS(660),
    [anon_sym_str] = ACTIONS(660),
    [anon_sym_char] = ACTIONS(660),
    [anon_sym_SQUOTE] = ACTIONS(662),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(666),
    [anon_sym_default] = ACTIONS(668),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(676),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_POUND] = ACTIONS(678),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_COMMA] = ACTIONS(718),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(688),
    [anon_sym__] = ACTIONS(690),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(694),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_mutable_specifier] = ACTIONS(698),
    [anon_sym_DOT_DOT] = ACTIONS(700),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(710),
    [sym_super] = ACTIONS(712),
    [sym_crate] = ACTIONS(712),
    [sym_metavariable] = ACTIONS(714),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [181] = {
    [sym_attribute_item] = STATE(192),
    [sym_function_modifiers] = STATE(2472),
    [sym_self_parameter] = STATE(1901),
    [sym_variadic_parameter] = STATE(1901),
    [sym_parameter] = STATE(1901),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(1781),
    [sym_bracketed_type] = STATE(2508),
    [sym_lifetime] = STATE(1902),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2509),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1728),
    [sym_scoped_identifier] = STATE(1505),
    [sym_scoped_type_identifier] = STATE(1474),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1748),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(650),
    [anon_sym_LPAREN] = ACTIONS(652),
    [anon_sym_RPAREN] = ACTIONS(720),
    [anon_sym_LBRACK] = ACTIONS(656),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(660),
    [anon_sym_i8] = ACTIONS(660),
    [anon_sym_u16] = ACTIONS(660),
    [anon_sym_i16] = ACTIONS(660),
    [anon_sym_u32] = ACTIONS(660),
    [anon_sym_i32] = ACTIONS(660),
    [anon_sym_u64] = ACTIONS(660),
    [anon_sym_i64] = ACTIONS(660),
    [anon_sym_u128] = ACTIONS(660),
    [anon_sym_i128] = ACTIONS(660),
    [anon_sym_isize] = ACTIONS(660),
    [anon_sym_usize] = ACTIONS(660),
    [anon_sym_f32] = ACTIONS(660),
    [anon_sym_f64] = ACTIONS(660),
    [anon_sym_bool] = ACTIONS(660),
    [anon_sym_str] = ACTIONS(660),
    [anon_sym_char] = ACTIONS(660),
    [anon_sym_SQUOTE] = ACTIONS(662),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(666),
    [anon_sym_default] = ACTIONS(668),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(676),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_POUND] = ACTIONS(678),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_COMMA] = ACTIONS(722),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(688),
    [anon_sym__] = ACTIONS(690),
    [anon_sym_AMP] = ACTIONS(692),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(694),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_mutable_specifier] = ACTIONS(698),
    [anon_sym_DOT_DOT] = ACTIONS(700),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(710),
    [sym_super] = ACTIONS(712),
    [sym_crate] = ACTIONS(712),
    [sym_metavariable] = ACTIONS(714),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [182] = {
    [sym_attribute_item] = STATE(192),
    [sym_function_modifiers] = STATE(2472),
    [sym_self_parameter] = STATE(1901),
    [sym_variadic_parameter] = STATE(1901),
    [sym_parameter] = STATE(1901),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(1781),
    [sym_bracketed_type] = STATE(2504),
    [sym_lifetime] = STATE(1902),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2505),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1710),
    [sym_scoped_identifier] = STATE(1565),
    [sym_scoped_type_identifier] = STATE(1474),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(2281),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(724),
    [anon_sym_LPAREN] = ACTIONS(726),
    [anon_sym_RPAREN] = ACTIONS(728),
    [anon_sym_LBRACK] = ACTIONS(656),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(730),
    [anon_sym_i8] = ACTIONS(730),
    [anon_sym_u16] = ACTIONS(730),
    [anon_sym_i16] = ACTIONS(730),
    [anon_sym_u32] = ACTIONS(730),
    [anon_sym_i32] = ACTIONS(730),
    [anon_sym_u64] = ACTIONS(730),
    [anon_sym_i64] = ACTIONS(730),
    [anon_sym_u128] = ACTIONS(730),
    [anon_sym_i128] = ACTIONS(730),
    [anon_sym_isize] = ACTIONS(730),
    [anon_sym_usize] = ACTIONS(730),
    [anon_sym_f32] = ACTIONS(730),
    [anon_sym_f64] = ACTIONS(730),
    [anon_sym_bool] = ACTIONS(730),
    [anon_sym_str] = ACTIONS(730),
    [anon_sym_char] = ACTIONS(730),
    [anon_sym_SQUOTE] = ACTIONS(662),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(666),
    [anon_sym_default] = ACTIONS(732),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(734),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_POUND] = ACTIONS(678),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_COMMA] = ACTIONS(736),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(738),
    [anon_sym__] = ACTIONS(740),
    [anon_sym_AMP] = ACTIONS(742),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(694),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_mutable_specifier] = ACTIONS(698),
    [anon_sym_DOT_DOT] = ACTIONS(700),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(744),
    [sym_super] = ACTIONS(746),
    [sym_crate] = ACTIONS(746),
    [sym_metavariable] = ACTIONS(748),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [183] = {
    [sym_attribute_item] = STATE(193),
    [sym_function_modifiers] = STATE(2472),
    [sym_self_parameter] = STATE(2080),
    [sym_variadic_parameter] = STATE(2080),
    [sym_parameter] = STATE(2080),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(1787),
    [sym_bracketed_type] = STATE(2504),
    [sym_lifetime] = STATE(1902),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2505),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1710),
    [sym_scoped_identifier] = STATE(1565),
    [sym_scoped_type_identifier] = STATE(1474),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(2281),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(724),
    [anon_sym_LPAREN] = ACTIONS(726),
    [anon_sym_RPAREN] = ACTIONS(750),
    [anon_sym_LBRACK] = ACTIONS(656),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(730),
    [anon_sym_i8] = ACTIONS(730),
    [anon_sym_u16] = ACTIONS(730),
    [anon_sym_i16] = ACTIONS(730),
    [anon_sym_u32] = ACTIONS(730),
    [anon_sym_i32] = ACTIONS(730),
    [anon_sym_u64] = ACTIONS(730),
    [anon_sym_i64] = ACTIONS(730),
    [anon_sym_u128] = ACTIONS(730),
    [anon_sym_i128] = ACTIONS(730),
    [anon_sym_isize] = ACTIONS(730),
    [anon_sym_usize] = ACTIONS(730),
    [anon_sym_f32] = ACTIONS(730),
    [anon_sym_f64] = ACTIONS(730),
    [anon_sym_bool] = ACTIONS(730),
    [anon_sym_str] = ACTIONS(730),
    [anon_sym_char] = ACTIONS(730),
    [anon_sym_SQUOTE] = ACTIONS(662),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(666),
    [anon_sym_default] = ACTIONS(732),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(734),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_POUND] = ACTIONS(678),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_COMMA] = ACTIONS(752),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(738),
    [anon_sym__] = ACTIONS(754),
    [anon_sym_AMP] = ACTIONS(742),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(694),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_mutable_specifier] = ACTIONS(698),
    [anon_sym_DOT_DOT] = ACTIONS(700),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(744),
    [sym_super] = ACTIONS(746),
    [sym_crate] = ACTIONS(746),
    [sym_metavariable] = ACTIONS(748),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [184] = {
    [sym_attribute_item] = STATE(191),
    [sym_function_modifiers] = STATE(2472),
    [sym_self_parameter] = STATE(2260),
    [sym_variadic_parameter] = STATE(2260),
    [sym_parameter] = STATE(2260),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(1947),
    [sym_bracketed_type] = STATE(2504),
    [sym_lifetime] = STATE(1902),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2505),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1710),
    [sym_scoped_identifier] = STATE(1565),
    [sym_scoped_type_identifier] = STATE(1474),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(2281),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(724),
    [anon_sym_LPAREN] = ACTIONS(726),
    [anon_sym_RPAREN] = ACTIONS(756),
    [anon_sym_LBRACK] = ACTIONS(656),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(730),
    [anon_sym_i8] = ACTIONS(730),
    [anon_sym_u16] = ACTIONS(730),
    [anon_sym_i16] = ACTIONS(730),
    [anon_sym_u32] = ACTIONS(730),
    [anon_sym_i32] = ACTIONS(730),
    [anon_sym_u64] = ACTIONS(730),
    [anon_sym_i64] = ACTIONS(730),
    [anon_sym_u128] = ACTIONS(730),
    [anon_sym_i128] = ACTIONS(730),
    [anon_sym_isize] = ACTIONS(730),
    [anon_sym_usize] = ACTIONS(730),
    [anon_sym_f32] = ACTIONS(730),
    [anon_sym_f64] = ACTIONS(730),
    [anon_sym_bool] = ACTIONS(730),
    [anon_sym_str] = ACTIONS(730),
    [anon_sym_char] = ACTIONS(730),
    [anon_sym_SQUOTE] = ACTIONS(662),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(666),
    [anon_sym_default] = ACTIONS(732),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(734),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_POUND] = ACTIONS(678),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(738),
    [anon_sym__] = ACTIONS(758),
    [anon_sym_AMP] = ACTIONS(742),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(694),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_mutable_specifier] = ACTIONS(698),
    [anon_sym_DOT_DOT] = ACTIONS(700),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(744),
    [sym_super] = ACTIONS(746),
    [sym_crate] = ACTIONS(746),
    [sym_metavariable] = ACTIONS(748),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [185] = {
    [sym_attribute_item] = STATE(191),
    [sym_function_modifiers] = STATE(2472),
    [sym_self_parameter] = STATE(2260),
    [sym_variadic_parameter] = STATE(2260),
    [sym_parameter] = STATE(2260),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(1947),
    [sym_bracketed_type] = STATE(2504),
    [sym_lifetime] = STATE(1902),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2505),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1710),
    [sym_scoped_identifier] = STATE(1565),
    [sym_scoped_type_identifier] = STATE(1474),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(2281),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(724),
    [anon_sym_LPAREN] = ACTIONS(726),
    [anon_sym_RPAREN] = ACTIONS(760),
    [anon_sym_LBRACK] = ACTIONS(656),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(730),
    [anon_sym_i8] = ACTIONS(730),
    [anon_sym_u16] = ACTIONS(730),
    [anon_sym_i16] = ACTIONS(730),
    [anon_sym_u32] = ACTIONS(730),
    [anon_sym_i32] = ACTIONS(730),
    [anon_sym_u64] = ACTIONS(730),
    [anon_sym_i64] = ACTIONS(730),
    [anon_sym_u128] = ACTIONS(730),
    [anon_sym_i128] = ACTIONS(730),
    [anon_sym_isize] = ACTIONS(730),
    [anon_sym_usize] = ACTIONS(730),
    [anon_sym_f32] = ACTIONS(730),
    [anon_sym_f64] = ACTIONS(730),
    [anon_sym_bool] = ACTIONS(730),
    [anon_sym_str] = ACTIONS(730),
    [anon_sym_char] = ACTIONS(730),
    [anon_sym_SQUOTE] = ACTIONS(662),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(666),
    [anon_sym_default] = ACTIONS(732),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(734),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_POUND] = ACTIONS(678),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(738),
    [anon_sym__] = ACTIONS(758),
    [anon_sym_AMP] = ACTIONS(742),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(694),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_mutable_specifier] = ACTIONS(698),
    [anon_sym_DOT_DOT] = ACTIONS(700),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(744),
    [sym_super] = ACTIONS(746),
    [sym_crate] = ACTIONS(746),
    [sym_metavariable] = ACTIONS(748),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [186] = {
    [sym_attribute_item] = STATE(191),
    [sym_function_modifiers] = STATE(2472),
    [sym_self_parameter] = STATE(2260),
    [sym_variadic_parameter] = STATE(2260),
    [sym_parameter] = STATE(2260),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(1947),
    [sym_bracketed_type] = STATE(2504),
    [sym_lifetime] = STATE(1902),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2505),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1710),
    [sym_scoped_identifier] = STATE(1565),
    [sym_scoped_type_identifier] = STATE(1474),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(2281),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(724),
    [anon_sym_LPAREN] = ACTIONS(726),
    [anon_sym_RPAREN] = ACTIONS(762),
    [anon_sym_LBRACK] = ACTIONS(656),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(730),
    [anon_sym_i8] = ACTIONS(730),
    [anon_sym_u16] = ACTIONS(730),
    [anon_sym_i16] = ACTIONS(730),
    [anon_sym_u32] = ACTIONS(730),
    [anon_sym_i32] = ACTIONS(730),
    [anon_sym_u64] = ACTIONS(730),
    [anon_sym_i64] = ACTIONS(730),
    [anon_sym_u128] = ACTIONS(730),
    [anon_sym_i128] = ACTIONS(730),
    [anon_sym_isize] = ACTIONS(730),
    [anon_sym_usize] = ACTIONS(730),
    [anon_sym_f32] = ACTIONS(730),
    [anon_sym_f64] = ACTIONS(730),
    [anon_sym_bool] = ACTIONS(730),
    [anon_sym_str] = ACTIONS(730),
    [anon_sym_char] = ACTIONS(730),
    [anon_sym_SQUOTE] = ACTIONS(662),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(666),
    [anon_sym_default] = ACTIONS(732),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(734),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_POUND] = ACTIONS(678),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(738),
    [anon_sym__] = ACTIONS(758),
    [anon_sym_AMP] = ACTIONS(742),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(694),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_mutable_specifier] = ACTIONS(698),
    [anon_sym_DOT_DOT] = ACTIONS(700),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(744),
    [sym_super] = ACTIONS(746),
    [sym_crate] = ACTIONS(746),
    [sym_metavariable] = ACTIONS(748),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [187] = {
    [sym_attribute_item] = STATE(191),
    [sym_function_modifiers] = STATE(2472),
    [sym_self_parameter] = STATE(2260),
    [sym_variadic_parameter] = STATE(2260),
    [sym_parameter] = STATE(2260),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(1947),
    [sym_bracketed_type] = STATE(2504),
    [sym_lifetime] = STATE(1902),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2505),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1710),
    [sym_scoped_identifier] = STATE(1565),
    [sym_scoped_type_identifier] = STATE(1474),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(2281),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(724),
    [anon_sym_LPAREN] = ACTIONS(726),
    [anon_sym_RPAREN] = ACTIONS(764),
    [anon_sym_LBRACK] = ACTIONS(656),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(730),
    [anon_sym_i8] = ACTIONS(730),
    [anon_sym_u16] = ACTIONS(730),
    [anon_sym_i16] = ACTIONS(730),
    [anon_sym_u32] = ACTIONS(730),
    [anon_sym_i32] = ACTIONS(730),
    [anon_sym_u64] = ACTIONS(730),
    [anon_sym_i64] = ACTIONS(730),
    [anon_sym_u128] = ACTIONS(730),
    [anon_sym_i128] = ACTIONS(730),
    [anon_sym_isize] = ACTIONS(730),
    [anon_sym_usize] = ACTIONS(730),
    [anon_sym_f32] = ACTIONS(730),
    [anon_sym_f64] = ACTIONS(730),
    [anon_sym_bool] = ACTIONS(730),
    [anon_sym_str] = ACTIONS(730),
    [anon_sym_char] = ACTIONS(730),
    [anon_sym_SQUOTE] = ACTIONS(662),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(666),
    [anon_sym_default] = ACTIONS(732),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(734),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_POUND] = ACTIONS(678),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(738),
    [anon_sym__] = ACTIONS(758),
    [anon_sym_AMP] = ACTIONS(742),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(694),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_mutable_specifier] = ACTIONS(698),
    [anon_sym_DOT_DOT] = ACTIONS(700),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(744),
    [sym_super] = ACTIONS(746),
    [sym_crate] = ACTIONS(746),
    [sym_metavariable] = ACTIONS(748),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [188] = {
    [sym_attribute_item] = STATE(191),
    [sym_function_modifiers] = STATE(2472),
    [sym_self_parameter] = STATE(2260),
    [sym_variadic_parameter] = STATE(2260),
    [sym_parameter] = STATE(2260),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(1947),
    [sym_bracketed_type] = STATE(2504),
    [sym_lifetime] = STATE(1902),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2505),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1710),
    [sym_scoped_identifier] = STATE(1565),
    [sym_scoped_type_identifier] = STATE(1474),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(2281),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(724),
    [anon_sym_LPAREN] = ACTIONS(726),
    [anon_sym_RPAREN] = ACTIONS(766),
    [anon_sym_LBRACK] = ACTIONS(656),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(730),
    [anon_sym_i8] = ACTIONS(730),
    [anon_sym_u16] = ACTIONS(730),
    [anon_sym_i16] = ACTIONS(730),
    [anon_sym_u32] = ACTIONS(730),
    [anon_sym_i32] = ACTIONS(730),
    [anon_sym_u64] = ACTIONS(730),
    [anon_sym_i64] = ACTIONS(730),
    [anon_sym_u128] = ACTIONS(730),
    [anon_sym_i128] = ACTIONS(730),
    [anon_sym_isize] = ACTIONS(730),
    [anon_sym_usize] = ACTIONS(730),
    [anon_sym_f32] = ACTIONS(730),
    [anon_sym_f64] = ACTIONS(730),
    [anon_sym_bool] = ACTIONS(730),
    [anon_sym_str] = ACTIONS(730),
    [anon_sym_char] = ACTIONS(730),
    [anon_sym_SQUOTE] = ACTIONS(662),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(666),
    [anon_sym_default] = ACTIONS(732),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(734),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_POUND] = ACTIONS(678),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(738),
    [anon_sym__] = ACTIONS(758),
    [anon_sym_AMP] = ACTIONS(742),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(694),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_mutable_specifier] = ACTIONS(698),
    [anon_sym_DOT_DOT] = ACTIONS(700),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(744),
    [sym_super] = ACTIONS(746),
    [sym_crate] = ACTIONS(746),
    [sym_metavariable] = ACTIONS(748),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [189] = {
    [sym_attribute_item] = STATE(191),
    [sym_function_modifiers] = STATE(2472),
    [sym_self_parameter] = STATE(2260),
    [sym_variadic_parameter] = STATE(2260),
    [sym_parameter] = STATE(2260),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(1947),
    [sym_bracketed_type] = STATE(2504),
    [sym_lifetime] = STATE(1902),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2505),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1710),
    [sym_scoped_identifier] = STATE(1565),
    [sym_scoped_type_identifier] = STATE(1474),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(2281),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(724),
    [anon_sym_LPAREN] = ACTIONS(726),
    [anon_sym_RPAREN] = ACTIONS(768),
    [anon_sym_LBRACK] = ACTIONS(656),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(730),
    [anon_sym_i8] = ACTIONS(730),
    [anon_sym_u16] = ACTIONS(730),
    [anon_sym_i16] = ACTIONS(730),
    [anon_sym_u32] = ACTIONS(730),
    [anon_sym_i32] = ACTIONS(730),
    [anon_sym_u64] = ACTIONS(730),
    [anon_sym_i64] = ACTIONS(730),
    [anon_sym_u128] = ACTIONS(730),
    [anon_sym_i128] = ACTIONS(730),
    [anon_sym_isize] = ACTIONS(730),
    [anon_sym_usize] = ACTIONS(730),
    [anon_sym_f32] = ACTIONS(730),
    [anon_sym_f64] = ACTIONS(730),
    [anon_sym_bool] = ACTIONS(730),
    [anon_sym_str] = ACTIONS(730),
    [anon_sym_char] = ACTIONS(730),
    [anon_sym_SQUOTE] = ACTIONS(662),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(666),
    [anon_sym_default] = ACTIONS(732),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(734),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_POUND] = ACTIONS(678),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(738),
    [anon_sym__] = ACTIONS(758),
    [anon_sym_AMP] = ACTIONS(742),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(694),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_mutable_specifier] = ACTIONS(698),
    [anon_sym_DOT_DOT] = ACTIONS(700),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(744),
    [sym_super] = ACTIONS(746),
    [sym_crate] = ACTIONS(746),
    [sym_metavariable] = ACTIONS(748),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [190] = {
    [sym_attribute_item] = STATE(191),
    [sym_function_modifiers] = STATE(2472),
    [sym_self_parameter] = STATE(2260),
    [sym_variadic_parameter] = STATE(2260),
    [sym_parameter] = STATE(2260),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(1947),
    [sym_bracketed_type] = STATE(2504),
    [sym_lifetime] = STATE(1902),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2505),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1710),
    [sym_scoped_identifier] = STATE(1565),
    [sym_scoped_type_identifier] = STATE(1474),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(2281),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(724),
    [anon_sym_LPAREN] = ACTIONS(726),
    [anon_sym_LBRACK] = ACTIONS(656),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(730),
    [anon_sym_i8] = ACTIONS(730),
    [anon_sym_u16] = ACTIONS(730),
    [anon_sym_i16] = ACTIONS(730),
    [anon_sym_u32] = ACTIONS(730),
    [anon_sym_i32] = ACTIONS(730),
    [anon_sym_u64] = ACTIONS(730),
    [anon_sym_i64] = ACTIONS(730),
    [anon_sym_u128] = ACTIONS(730),
    [anon_sym_i128] = ACTIONS(730),
    [anon_sym_isize] = ACTIONS(730),
    [anon_sym_usize] = ACTIONS(730),
    [anon_sym_f32] = ACTIONS(730),
    [anon_sym_f64] = ACTIONS(730),
    [anon_sym_bool] = ACTIONS(730),
    [anon_sym_str] = ACTIONS(730),
    [anon_sym_char] = ACTIONS(730),
    [anon_sym_SQUOTE] = ACTIONS(662),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(666),
    [anon_sym_default] = ACTIONS(732),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(734),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_POUND] = ACTIONS(678),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(738),
    [anon_sym__] = ACTIONS(758),
    [anon_sym_AMP] = ACTIONS(742),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(694),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_mutable_specifier] = ACTIONS(698),
    [anon_sym_DOT_DOT] = ACTIONS(700),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(744),
    [sym_super] = ACTIONS(746),
    [sym_crate] = ACTIONS(746),
    [sym_metavariable] = ACTIONS(748),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [191] = {
    [sym_function_modifiers] = STATE(2472),
    [sym_self_parameter] = STATE(2287),
    [sym_variadic_parameter] = STATE(2287),
    [sym_parameter] = STATE(2287),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(1893),
    [sym_bracketed_type] = STATE(2504),
    [sym_lifetime] = STATE(1902),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2505),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1710),
    [sym_scoped_identifier] = STATE(1565),
    [sym_scoped_type_identifier] = STATE(1474),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(2281),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(724),
    [anon_sym_LPAREN] = ACTIONS(726),
    [anon_sym_LBRACK] = ACTIONS(656),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(730),
    [anon_sym_i8] = ACTIONS(730),
    [anon_sym_u16] = ACTIONS(730),
    [anon_sym_i16] = ACTIONS(730),
    [anon_sym_u32] = ACTIONS(730),
    [anon_sym_i32] = ACTIONS(730),
    [anon_sym_u64] = ACTIONS(730),
    [anon_sym_i64] = ACTIONS(730),
    [anon_sym_u128] = ACTIONS(730),
    [anon_sym_i128] = ACTIONS(730),
    [anon_sym_isize] = ACTIONS(730),
    [anon_sym_usize] = ACTIONS(730),
    [anon_sym_f32] = ACTIONS(730),
    [anon_sym_f64] = ACTIONS(730),
    [anon_sym_bool] = ACTIONS(730),
    [anon_sym_str] = ACTIONS(730),
    [anon_sym_char] = ACTIONS(730),
    [anon_sym_SQUOTE] = ACTIONS(662),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(666),
    [anon_sym_default] = ACTIONS(732),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(734),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(738),
    [anon_sym__] = ACTIONS(770),
    [anon_sym_AMP] = ACTIONS(742),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(694),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_mutable_specifier] = ACTIONS(698),
    [anon_sym_DOT_DOT] = ACTIONS(700),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(744),
    [sym_super] = ACTIONS(746),
    [sym_crate] = ACTIONS(746),
    [sym_metavariable] = ACTIONS(748),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [192] = {
    [sym_function_modifiers] = STATE(2472),
    [sym_self_parameter] = STATE(2069),
    [sym_variadic_parameter] = STATE(2069),
    [sym_parameter] = STATE(2069),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(1802),
    [sym_bracketed_type] = STATE(2504),
    [sym_lifetime] = STATE(1902),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2505),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1710),
    [sym_scoped_identifier] = STATE(1565),
    [sym_scoped_type_identifier] = STATE(1474),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(2281),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(724),
    [anon_sym_LPAREN] = ACTIONS(726),
    [anon_sym_LBRACK] = ACTIONS(656),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(730),
    [anon_sym_i8] = ACTIONS(730),
    [anon_sym_u16] = ACTIONS(730),
    [anon_sym_i16] = ACTIONS(730),
    [anon_sym_u32] = ACTIONS(730),
    [anon_sym_i32] = ACTIONS(730),
    [anon_sym_u64] = ACTIONS(730),
    [anon_sym_i64] = ACTIONS(730),
    [anon_sym_u128] = ACTIONS(730),
    [anon_sym_i128] = ACTIONS(730),
    [anon_sym_isize] = ACTIONS(730),
    [anon_sym_usize] = ACTIONS(730),
    [anon_sym_f32] = ACTIONS(730),
    [anon_sym_f64] = ACTIONS(730),
    [anon_sym_bool] = ACTIONS(730),
    [anon_sym_str] = ACTIONS(730),
    [anon_sym_char] = ACTIONS(730),
    [anon_sym_SQUOTE] = ACTIONS(662),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(666),
    [anon_sym_default] = ACTIONS(732),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(734),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(738),
    [anon_sym__] = ACTIONS(772),
    [anon_sym_AMP] = ACTIONS(742),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(694),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_mutable_specifier] = ACTIONS(698),
    [anon_sym_DOT_DOT] = ACTIONS(700),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(744),
    [sym_super] = ACTIONS(746),
    [sym_crate] = ACTIONS(746),
    [sym_metavariable] = ACTIONS(748),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [193] = {
    [sym_function_modifiers] = STATE(2472),
    [sym_self_parameter] = STATE(2064),
    [sym_variadic_parameter] = STATE(2064),
    [sym_parameter] = STATE(2064),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(1767),
    [sym_bracketed_type] = STATE(2504),
    [sym_lifetime] = STATE(1902),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2505),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1710),
    [sym_scoped_identifier] = STATE(1565),
    [sym_scoped_type_identifier] = STATE(1474),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(2281),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(724),
    [anon_sym_LPAREN] = ACTIONS(726),
    [anon_sym_LBRACK] = ACTIONS(656),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(730),
    [anon_sym_i8] = ACTIONS(730),
    [anon_sym_u16] = ACTIONS(730),
    [anon_sym_i16] = ACTIONS(730),
    [anon_sym_u32] = ACTIONS(730),
    [anon_sym_i32] = ACTIONS(730),
    [anon_sym_u64] = ACTIONS(730),
    [anon_sym_i64] = ACTIONS(730),
    [anon_sym_u128] = ACTIONS(730),
    [anon_sym_i128] = ACTIONS(730),
    [anon_sym_isize] = ACTIONS(730),
    [anon_sym_usize] = ACTIONS(730),
    [anon_sym_f32] = ACTIONS(730),
    [anon_sym_f64] = ACTIONS(730),
    [anon_sym_bool] = ACTIONS(730),
    [anon_sym_str] = ACTIONS(730),
    [anon_sym_char] = ACTIONS(730),
    [anon_sym_SQUOTE] = ACTIONS(662),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(666),
    [anon_sym_default] = ACTIONS(732),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(734),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(738),
    [anon_sym__] = ACTIONS(774),
    [anon_sym_AMP] = ACTIONS(742),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(694),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_mutable_specifier] = ACTIONS(698),
    [anon_sym_DOT_DOT] = ACTIONS(700),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(744),
    [sym_super] = ACTIONS(746),
    [sym_crate] = ACTIONS(746),
    [sym_metavariable] = ACTIONS(748),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [194] = {
    [sym_function_modifiers] = STATE(2472),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(2009),
    [sym_bracketed_type] = STATE(2511),
    [sym_lifetime] = STATE(2471),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2512),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1730),
    [sym_scoped_identifier] = STATE(1528),
    [sym_scoped_type_identifier] = STATE(1474),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1791),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(776),
    [anon_sym_LPAREN] = ACTIONS(778),
    [anon_sym_LBRACK] = ACTIONS(656),
    [anon_sym_RBRACK] = ACTIONS(780),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(782),
    [anon_sym_i8] = ACTIONS(782),
    [anon_sym_u16] = ACTIONS(782),
    [anon_sym_i16] = ACTIONS(782),
    [anon_sym_u32] = ACTIONS(782),
    [anon_sym_i32] = ACTIONS(782),
    [anon_sym_u64] = ACTIONS(782),
    [anon_sym_i64] = ACTIONS(782),
    [anon_sym_u128] = ACTIONS(782),
    [anon_sym_i128] = ACTIONS(782),
    [anon_sym_isize] = ACTIONS(782),
    [anon_sym_usize] = ACTIONS(782),
    [anon_sym_f32] = ACTIONS(782),
    [anon_sym_f64] = ACTIONS(782),
    [anon_sym_bool] = ACTIONS(782),
    [anon_sym_str] = ACTIONS(782),
    [anon_sym_char] = ACTIONS(782),
    [anon_sym_SQUOTE] = ACTIONS(662),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(666),
    [anon_sym_default] = ACTIONS(784),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(786),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_COMMA] = ACTIONS(788),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(790),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(794),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_mutable_specifier] = ACTIONS(796),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(800),
    [sym_super] = ACTIONS(800),
    [sym_crate] = ACTIONS(800),
    [sym_metavariable] = ACTIONS(802),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [195] = {
    [sym_function_modifiers] = STATE(2472),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(1831),
    [sym_bracketed_type] = STATE(2508),
    [sym_lifetime] = STATE(2471),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2509),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1728),
    [sym_scoped_identifier] = STATE(1505),
    [sym_scoped_type_identifier] = STATE(1474),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1789),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(650),
    [anon_sym_LPAREN] = ACTIONS(652),
    [anon_sym_RPAREN] = ACTIONS(804),
    [anon_sym_LBRACK] = ACTIONS(656),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(660),
    [anon_sym_i8] = ACTIONS(660),
    [anon_sym_u16] = ACTIONS(660),
    [anon_sym_i16] = ACTIONS(660),
    [anon_sym_u32] = ACTIONS(660),
    [anon_sym_i32] = ACTIONS(660),
    [anon_sym_u64] = ACTIONS(660),
    [anon_sym_i64] = ACTIONS(660),
    [anon_sym_u128] = ACTIONS(660),
    [anon_sym_i128] = ACTIONS(660),
    [anon_sym_isize] = ACTIONS(660),
    [anon_sym_usize] = ACTIONS(660),
    [anon_sym_f32] = ACTIONS(660),
    [anon_sym_f64] = ACTIONS(660),
    [anon_sym_bool] = ACTIONS(660),
    [anon_sym_str] = ACTIONS(660),
    [anon_sym_char] = ACTIONS(660),
    [anon_sym_SQUOTE] = ACTIONS(662),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(666),
    [anon_sym_default] = ACTIONS(668),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(676),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_COMMA] = ACTIONS(806),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(688),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(808),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_mutable_specifier] = ACTIONS(796),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(712),
    [sym_super] = ACTIONS(712),
    [sym_crate] = ACTIONS(712),
    [sym_metavariable] = ACTIONS(714),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [196] = {
    [sym_function_modifiers] = STATE(2472),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(1831),
    [sym_bracketed_type] = STATE(2508),
    [sym_lifetime] = STATE(2471),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2509),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1728),
    [sym_scoped_identifier] = STATE(1505),
    [sym_scoped_type_identifier] = STATE(1474),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1789),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(650),
    [anon_sym_LPAREN] = ACTIONS(652),
    [anon_sym_RPAREN] = ACTIONS(810),
    [anon_sym_LBRACK] = ACTIONS(656),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(660),
    [anon_sym_i8] = ACTIONS(660),
    [anon_sym_u16] = ACTIONS(660),
    [anon_sym_i16] = ACTIONS(660),
    [anon_sym_u32] = ACTIONS(660),
    [anon_sym_i32] = ACTIONS(660),
    [anon_sym_u64] = ACTIONS(660),
    [anon_sym_i64] = ACTIONS(660),
    [anon_sym_u128] = ACTIONS(660),
    [anon_sym_i128] = ACTIONS(660),
    [anon_sym_isize] = ACTIONS(660),
    [anon_sym_usize] = ACTIONS(660),
    [anon_sym_f32] = ACTIONS(660),
    [anon_sym_f64] = ACTIONS(660),
    [anon_sym_bool] = ACTIONS(660),
    [anon_sym_str] = ACTIONS(660),
    [anon_sym_char] = ACTIONS(660),
    [anon_sym_SQUOTE] = ACTIONS(662),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(666),
    [anon_sym_default] = ACTIONS(668),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(676),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_COMMA] = ACTIONS(806),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(688),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(808),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_mutable_specifier] = ACTIONS(796),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(712),
    [sym_super] = ACTIONS(712),
    [sym_crate] = ACTIONS(712),
    [sym_metavariable] = ACTIONS(714),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [197] = {
    [sym_function_modifiers] = STATE(2472),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(1831),
    [sym_bracketed_type] = STATE(2508),
    [sym_lifetime] = STATE(2471),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2509),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1728),
    [sym_scoped_identifier] = STATE(1505),
    [sym_scoped_type_identifier] = STATE(1474),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1789),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(650),
    [anon_sym_LPAREN] = ACTIONS(652),
    [anon_sym_RPAREN] = ACTIONS(812),
    [anon_sym_LBRACK] = ACTIONS(656),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(660),
    [anon_sym_i8] = ACTIONS(660),
    [anon_sym_u16] = ACTIONS(660),
    [anon_sym_i16] = ACTIONS(660),
    [anon_sym_u32] = ACTIONS(660),
    [anon_sym_i32] = ACTIONS(660),
    [anon_sym_u64] = ACTIONS(660),
    [anon_sym_i64] = ACTIONS(660),
    [anon_sym_u128] = ACTIONS(660),
    [anon_sym_i128] = ACTIONS(660),
    [anon_sym_isize] = ACTIONS(660),
    [anon_sym_usize] = ACTIONS(660),
    [anon_sym_f32] = ACTIONS(660),
    [anon_sym_f64] = ACTIONS(660),
    [anon_sym_bool] = ACTIONS(660),
    [anon_sym_str] = ACTIONS(660),
    [anon_sym_char] = ACTIONS(660),
    [anon_sym_SQUOTE] = ACTIONS(662),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(666),
    [anon_sym_default] = ACTIONS(668),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(676),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_COMMA] = ACTIONS(806),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(688),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(808),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_mutable_specifier] = ACTIONS(796),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(712),
    [sym_super] = ACTIONS(712),
    [sym_crate] = ACTIONS(712),
    [sym_metavariable] = ACTIONS(714),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [198] = {
    [sym_identifier] = ACTIONS(814),
    [anon_sym_SEMI] = ACTIONS(816),
    [anon_sym_LPAREN] = ACTIONS(816),
    [anon_sym_RPAREN] = ACTIONS(816),
    [anon_sym_LBRACE] = ACTIONS(816),
    [anon_sym_RBRACE] = ACTIONS(816),
    [anon_sym_EQ_GT] = ACTIONS(816),
    [anon_sym_LBRACK] = ACTIONS(816),
    [anon_sym_RBRACK] = ACTIONS(816),
    [anon_sym_COLON] = ACTIONS(814),
    [anon_sym_PLUS] = ACTIONS(814),
    [anon_sym_STAR] = ACTIONS(814),
    [anon_sym_QMARK] = ACTIONS(816),
    [anon_sym_u8] = ACTIONS(814),
    [anon_sym_i8] = ACTIONS(814),
    [anon_sym_u16] = ACTIONS(814),
    [anon_sym_i16] = ACTIONS(814),
    [anon_sym_u32] = ACTIONS(814),
    [anon_sym_i32] = ACTIONS(814),
    [anon_sym_u64] = ACTIONS(814),
    [anon_sym_i64] = ACTIONS(814),
    [anon_sym_u128] = ACTIONS(814),
    [anon_sym_i128] = ACTIONS(814),
    [anon_sym_isize] = ACTIONS(814),
    [anon_sym_usize] = ACTIONS(814),
    [anon_sym_f32] = ACTIONS(814),
    [anon_sym_f64] = ACTIONS(814),
    [anon_sym_bool] = ACTIONS(814),
    [anon_sym_str] = ACTIONS(814),
    [anon_sym_char] = ACTIONS(814),
    [anon_sym_SQUOTE] = ACTIONS(814),
    [anon_sym_as] = ACTIONS(814),
    [anon_sym_async] = ACTIONS(814),
    [anon_sym_break] = ACTIONS(814),
    [anon_sym_const] = ACTIONS(814),
    [anon_sym_continue] = ACTIONS(814),
    [anon_sym_default] = ACTIONS(814),
    [anon_sym_for] = ACTIONS(814),
    [anon_sym_if] = ACTIONS(814),
    [anon_sym_loop] = ACTIONS(814),
    [anon_sym_match] = ACTIONS(814),
    [anon_sym_return] = ACTIONS(814),
    [anon_sym_union] = ACTIONS(814),
    [anon_sym_unsafe] = ACTIONS(814),
    [anon_sym_while] = ACTIONS(814),
    [anon_sym_BANG] = ACTIONS(814),
    [anon_sym_EQ] = ACTIONS(814),
    [anon_sym_COMMA] = ACTIONS(816),
    [anon_sym_LT] = ACTIONS(814),
    [anon_sym_GT] = ACTIONS(814),
    [anon_sym_else] = ACTIONS(814),
    [anon_sym_COLON_COLON] = ACTIONS(816),
    [anon_sym_AMP] = ACTIONS(814),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(816),
    [anon_sym_DOT_DOT] = ACTIONS(814),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(816),
    [anon_sym_DASH] = ACTIONS(814),
    [anon_sym_AMP_AMP] = ACTIONS(816),
    [anon_sym_PIPE_PIPE] = ACTIONS(816),
    [anon_sym_PIPE] = ACTIONS(814),
    [anon_sym_CARET] = ACTIONS(814),
    [anon_sym_EQ_EQ] = ACTIONS(816),
    [anon_sym_BANG_EQ] = ACTIONS(816),
    [anon_sym_LT_EQ] = ACTIONS(816),
    [anon_sym_GT_EQ] = ACTIONS(816),
    [anon_sym_LT_LT] = ACTIONS(814),
    [anon_sym_GT_GT] = ACTIONS(814),
    [anon_sym_SLASH] = ACTIONS(814),
    [anon_sym_PERCENT] = ACTIONS(814),
    [anon_sym_PLUS_EQ] = ACTIONS(816),
    [anon_sym_DASH_EQ] = ACTIONS(816),
    [anon_sym_STAR_EQ] = ACTIONS(816),
    [anon_sym_SLASH_EQ] = ACTIONS(816),
    [anon_sym_PERCENT_EQ] = ACTIONS(816),
    [anon_sym_AMP_EQ] = ACTIONS(816),
    [anon_sym_PIPE_EQ] = ACTIONS(816),
    [anon_sym_CARET_EQ] = ACTIONS(816),
    [anon_sym_LT_LT_EQ] = ACTIONS(816),
    [anon_sym_GT_GT_EQ] = ACTIONS(816),
    [anon_sym_yield] = ACTIONS(814),
    [anon_sym_move] = ACTIONS(814),
    [anon_sym_DOT] = ACTIONS(814),
    [sym_integer_literal] = ACTIONS(816),
    [aux_sym_string_literal_token1] = ACTIONS(816),
    [sym_char_literal] = ACTIONS(816),
    [anon_sym_true] = ACTIONS(814),
    [anon_sym_false] = ACTIONS(814),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(814),
    [sym_super] = ACTIONS(814),
    [sym_crate] = ACTIONS(814),
    [sym_metavariable] = ACTIONS(816),
    [sym_raw_string_literal] = ACTIONS(816),
    [sym_float_literal] = ACTIONS(816),
    [sym_block_comment] = ACTIONS(3),
  },
  [199] = {
    [sym_function_modifiers] = STATE(2472),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(1398),
    [sym_bracketed_type] = STATE(2504),
    [sym_lifetime] = STATE(2471),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2505),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1710),
    [sym_scoped_identifier] = STATE(1565),
    [sym_scoped_type_identifier] = STATE(1474),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1426),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(724),
    [anon_sym_LPAREN] = ACTIONS(726),
    [anon_sym_LBRACK] = ACTIONS(656),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(730),
    [anon_sym_i8] = ACTIONS(730),
    [anon_sym_u16] = ACTIONS(730),
    [anon_sym_i16] = ACTIONS(730),
    [anon_sym_u32] = ACTIONS(730),
    [anon_sym_i32] = ACTIONS(730),
    [anon_sym_u64] = ACTIONS(730),
    [anon_sym_i64] = ACTIONS(730),
    [anon_sym_u128] = ACTIONS(730),
    [anon_sym_i128] = ACTIONS(730),
    [anon_sym_isize] = ACTIONS(730),
    [anon_sym_usize] = ACTIONS(730),
    [anon_sym_f32] = ACTIONS(730),
    [anon_sym_f64] = ACTIONS(730),
    [anon_sym_bool] = ACTIONS(730),
    [anon_sym_str] = ACTIONS(730),
    [anon_sym_char] = ACTIONS(730),
    [anon_sym_SQUOTE] = ACTIONS(662),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(666),
    [anon_sym_default] = ACTIONS(732),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(734),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(738),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(818),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_mutable_specifier] = ACTIONS(796),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(820),
    [sym_super] = ACTIONS(746),
    [sym_crate] = ACTIONS(746),
    [sym_metavariable] = ACTIONS(748),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [200] = {
    [sym_function_modifiers] = STATE(2472),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(1398),
    [sym_bracketed_type] = STATE(2504),
    [sym_lifetime] = STATE(2471),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2505),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1710),
    [sym_scoped_identifier] = STATE(1565),
    [sym_scoped_type_identifier] = STATE(1474),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1426),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(724),
    [anon_sym_LPAREN] = ACTIONS(726),
    [anon_sym_LBRACK] = ACTIONS(656),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(730),
    [anon_sym_i8] = ACTIONS(730),
    [anon_sym_u16] = ACTIONS(730),
    [anon_sym_i16] = ACTIONS(730),
    [anon_sym_u32] = ACTIONS(730),
    [anon_sym_i32] = ACTIONS(730),
    [anon_sym_u64] = ACTIONS(730),
    [anon_sym_i64] = ACTIONS(730),
    [anon_sym_u128] = ACTIONS(730),
    [anon_sym_i128] = ACTIONS(730),
    [anon_sym_isize] = ACTIONS(730),
    [anon_sym_usize] = ACTIONS(730),
    [anon_sym_f32] = ACTIONS(730),
    [anon_sym_f64] = ACTIONS(730),
    [anon_sym_bool] = ACTIONS(730),
    [anon_sym_str] = ACTIONS(730),
    [anon_sym_char] = ACTIONS(730),
    [anon_sym_SQUOTE] = ACTIONS(662),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(666),
    [anon_sym_default] = ACTIONS(732),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(734),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(738),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(818),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_mutable_specifier] = ACTIONS(796),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(746),
    [sym_super] = ACTIONS(746),
    [sym_crate] = ACTIONS(746),
    [sym_metavariable] = ACTIONS(748),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [201] = {
    [sym_function_modifiers] = STATE(2472),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(1398),
    [sym_bracketed_type] = STATE(2511),
    [sym_lifetime] = STATE(2471),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2512),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1730),
    [sym_scoped_identifier] = STATE(1528),
    [sym_scoped_type_identifier] = STATE(1474),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1426),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(776),
    [anon_sym_LPAREN] = ACTIONS(778),
    [anon_sym_LBRACK] = ACTIONS(656),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(782),
    [anon_sym_i8] = ACTIONS(782),
    [anon_sym_u16] = ACTIONS(782),
    [anon_sym_i16] = ACTIONS(782),
    [anon_sym_u32] = ACTIONS(782),
    [anon_sym_i32] = ACTIONS(782),
    [anon_sym_u64] = ACTIONS(782),
    [anon_sym_i64] = ACTIONS(782),
    [anon_sym_u128] = ACTIONS(782),
    [anon_sym_i128] = ACTIONS(782),
    [anon_sym_isize] = ACTIONS(782),
    [anon_sym_usize] = ACTIONS(782),
    [anon_sym_f32] = ACTIONS(782),
    [anon_sym_f64] = ACTIONS(782),
    [anon_sym_bool] = ACTIONS(782),
    [anon_sym_str] = ACTIONS(782),
    [anon_sym_char] = ACTIONS(782),
    [anon_sym_SQUOTE] = ACTIONS(662),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(666),
    [anon_sym_default] = ACTIONS(784),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(786),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(790),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(794),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_mutable_specifier] = ACTIONS(796),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(800),
    [sym_super] = ACTIONS(800),
    [sym_crate] = ACTIONS(800),
    [sym_metavariable] = ACTIONS(802),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [202] = {
    [sym_function_modifiers] = STATE(2472),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(1399),
    [sym_bracketed_type] = STATE(2508),
    [sym_lifetime] = STATE(624),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2509),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1728),
    [sym_scoped_identifier] = STATE(1505),
    [sym_scoped_type_identifier] = STATE(1474),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1453),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(650),
    [anon_sym_LPAREN] = ACTIONS(652),
    [anon_sym_LBRACK] = ACTIONS(656),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(660),
    [anon_sym_i8] = ACTIONS(660),
    [anon_sym_u16] = ACTIONS(660),
    [anon_sym_i16] = ACTIONS(660),
    [anon_sym_u32] = ACTIONS(660),
    [anon_sym_i32] = ACTIONS(660),
    [anon_sym_u64] = ACTIONS(660),
    [anon_sym_i64] = ACTIONS(660),
    [anon_sym_u128] = ACTIONS(660),
    [anon_sym_i128] = ACTIONS(660),
    [anon_sym_isize] = ACTIONS(660),
    [anon_sym_usize] = ACTIONS(660),
    [anon_sym_f32] = ACTIONS(660),
    [anon_sym_f64] = ACTIONS(660),
    [anon_sym_bool] = ACTIONS(660),
    [anon_sym_str] = ACTIONS(660),
    [anon_sym_char] = ACTIONS(660),
    [anon_sym_SQUOTE] = ACTIONS(822),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(666),
    [anon_sym_default] = ACTIONS(668),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(676),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(688),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(808),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_mutable_specifier] = ACTIONS(824),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(712),
    [sym_super] = ACTIONS(712),
    [sym_crate] = ACTIONS(712),
    [sym_metavariable] = ACTIONS(714),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [203] = {
    [sym_function_modifiers] = STATE(2472),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(1399),
    [sym_bracketed_type] = STATE(2504),
    [sym_lifetime] = STATE(624),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2505),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1710),
    [sym_scoped_identifier] = STATE(1565),
    [sym_scoped_type_identifier] = STATE(1474),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1453),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(724),
    [anon_sym_LPAREN] = ACTIONS(726),
    [anon_sym_LBRACK] = ACTIONS(656),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(730),
    [anon_sym_i8] = ACTIONS(730),
    [anon_sym_u16] = ACTIONS(730),
    [anon_sym_i16] = ACTIONS(730),
    [anon_sym_u32] = ACTIONS(730),
    [anon_sym_i32] = ACTIONS(730),
    [anon_sym_u64] = ACTIONS(730),
    [anon_sym_i64] = ACTIONS(730),
    [anon_sym_u128] = ACTIONS(730),
    [anon_sym_i128] = ACTIONS(730),
    [anon_sym_isize] = ACTIONS(730),
    [anon_sym_usize] = ACTIONS(730),
    [anon_sym_f32] = ACTIONS(730),
    [anon_sym_f64] = ACTIONS(730),
    [anon_sym_bool] = ACTIONS(730),
    [anon_sym_str] = ACTIONS(730),
    [anon_sym_char] = ACTIONS(730),
    [anon_sym_SQUOTE] = ACTIONS(822),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(666),
    [anon_sym_default] = ACTIONS(732),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(734),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(738),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(818),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_mutable_specifier] = ACTIONS(826),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(746),
    [sym_super] = ACTIONS(746),
    [sym_crate] = ACTIONS(746),
    [sym_metavariable] = ACTIONS(748),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [204] = {
    [sym_function_modifiers] = STATE(2472),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(1399),
    [sym_bracketed_type] = STATE(2504),
    [sym_lifetime] = STATE(620),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2505),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1710),
    [sym_scoped_identifier] = STATE(1565),
    [sym_scoped_type_identifier] = STATE(1474),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1453),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(724),
    [anon_sym_LPAREN] = ACTIONS(726),
    [anon_sym_LBRACK] = ACTIONS(656),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(730),
    [anon_sym_i8] = ACTIONS(730),
    [anon_sym_u16] = ACTIONS(730),
    [anon_sym_i16] = ACTIONS(730),
    [anon_sym_u32] = ACTIONS(730),
    [anon_sym_i32] = ACTIONS(730),
    [anon_sym_u64] = ACTIONS(730),
    [anon_sym_i64] = ACTIONS(730),
    [anon_sym_u128] = ACTIONS(730),
    [anon_sym_i128] = ACTIONS(730),
    [anon_sym_isize] = ACTIONS(730),
    [anon_sym_usize] = ACTIONS(730),
    [anon_sym_f32] = ACTIONS(730),
    [anon_sym_f64] = ACTIONS(730),
    [anon_sym_bool] = ACTIONS(730),
    [anon_sym_str] = ACTIONS(730),
    [anon_sym_char] = ACTIONS(730),
    [anon_sym_SQUOTE] = ACTIONS(822),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(666),
    [anon_sym_default] = ACTIONS(732),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(734),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(738),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(818),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_mutable_specifier] = ACTIONS(828),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(830),
    [sym_super] = ACTIONS(746),
    [sym_crate] = ACTIONS(746),
    [sym_metavariable] = ACTIONS(748),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [205] = {
    [sym_function_modifiers] = STATE(2472),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(1399),
    [sym_bracketed_type] = STATE(2511),
    [sym_lifetime] = STATE(624),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2512),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1730),
    [sym_scoped_identifier] = STATE(1528),
    [sym_scoped_type_identifier] = STATE(1474),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1453),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(776),
    [anon_sym_LPAREN] = ACTIONS(778),
    [anon_sym_LBRACK] = ACTIONS(656),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(782),
    [anon_sym_i8] = ACTIONS(782),
    [anon_sym_u16] = ACTIONS(782),
    [anon_sym_i16] = ACTIONS(782),
    [anon_sym_u32] = ACTIONS(782),
    [anon_sym_i32] = ACTIONS(782),
    [anon_sym_u64] = ACTIONS(782),
    [anon_sym_i64] = ACTIONS(782),
    [anon_sym_u128] = ACTIONS(782),
    [anon_sym_i128] = ACTIONS(782),
    [anon_sym_isize] = ACTIONS(782),
    [anon_sym_usize] = ACTIONS(782),
    [anon_sym_f32] = ACTIONS(782),
    [anon_sym_f64] = ACTIONS(782),
    [anon_sym_bool] = ACTIONS(782),
    [anon_sym_str] = ACTIONS(782),
    [anon_sym_char] = ACTIONS(782),
    [anon_sym_SQUOTE] = ACTIONS(822),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(666),
    [anon_sym_default] = ACTIONS(784),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(786),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(790),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(794),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_mutable_specifier] = ACTIONS(832),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(800),
    [sym_super] = ACTIONS(800),
    [sym_crate] = ACTIONS(800),
    [sym_metavariable] = ACTIONS(802),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [206] = {
    [sym_function_modifiers] = STATE(2472),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(1398),
    [sym_bracketed_type] = STATE(2508),
    [sym_lifetime] = STATE(2471),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2509),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1728),
    [sym_scoped_identifier] = STATE(1505),
    [sym_scoped_type_identifier] = STATE(1474),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1426),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(650),
    [anon_sym_LPAREN] = ACTIONS(652),
    [anon_sym_LBRACK] = ACTIONS(656),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(660),
    [anon_sym_i8] = ACTIONS(660),
    [anon_sym_u16] = ACTIONS(660),
    [anon_sym_i16] = ACTIONS(660),
    [anon_sym_u32] = ACTIONS(660),
    [anon_sym_i32] = ACTIONS(660),
    [anon_sym_u64] = ACTIONS(660),
    [anon_sym_i64] = ACTIONS(660),
    [anon_sym_u128] = ACTIONS(660),
    [anon_sym_i128] = ACTIONS(660),
    [anon_sym_isize] = ACTIONS(660),
    [anon_sym_usize] = ACTIONS(660),
    [anon_sym_f32] = ACTIONS(660),
    [anon_sym_f64] = ACTIONS(660),
    [anon_sym_bool] = ACTIONS(660),
    [anon_sym_str] = ACTIONS(660),
    [anon_sym_char] = ACTIONS(660),
    [anon_sym_SQUOTE] = ACTIONS(662),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(666),
    [anon_sym_default] = ACTIONS(668),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(676),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(688),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(808),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_mutable_specifier] = ACTIONS(796),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(834),
    [sym_super] = ACTIONS(712),
    [sym_crate] = ACTIONS(712),
    [sym_metavariable] = ACTIONS(714),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [207] = {
    [sym_function_modifiers] = STATE(2472),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(1398),
    [sym_bracketed_type] = STATE(2508),
    [sym_lifetime] = STATE(2471),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2509),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1728),
    [sym_scoped_identifier] = STATE(1505),
    [sym_scoped_type_identifier] = STATE(1474),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1426),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(650),
    [anon_sym_LPAREN] = ACTIONS(652),
    [anon_sym_LBRACK] = ACTIONS(656),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(660),
    [anon_sym_i8] = ACTIONS(660),
    [anon_sym_u16] = ACTIONS(660),
    [anon_sym_i16] = ACTIONS(660),
    [anon_sym_u32] = ACTIONS(660),
    [anon_sym_i32] = ACTIONS(660),
    [anon_sym_u64] = ACTIONS(660),
    [anon_sym_i64] = ACTIONS(660),
    [anon_sym_u128] = ACTIONS(660),
    [anon_sym_i128] = ACTIONS(660),
    [anon_sym_isize] = ACTIONS(660),
    [anon_sym_usize] = ACTIONS(660),
    [anon_sym_f32] = ACTIONS(660),
    [anon_sym_f64] = ACTIONS(660),
    [anon_sym_bool] = ACTIONS(660),
    [anon_sym_str] = ACTIONS(660),
    [anon_sym_char] = ACTIONS(660),
    [anon_sym_SQUOTE] = ACTIONS(662),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(666),
    [anon_sym_default] = ACTIONS(668),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(676),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(688),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(808),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_mutable_specifier] = ACTIONS(796),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(712),
    [sym_super] = ACTIONS(712),
    [sym_crate] = ACTIONS(712),
    [sym_metavariable] = ACTIONS(714),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [208] = {
    [sym_function_modifiers] = STATE(2472),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(1399),
    [sym_bracketed_type] = STATE(2508),
    [sym_lifetime] = STATE(620),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2509),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1728),
    [sym_scoped_identifier] = STATE(1505),
    [sym_scoped_type_identifier] = STATE(1474),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1453),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(650),
    [anon_sym_LPAREN] = ACTIONS(652),
    [anon_sym_LBRACK] = ACTIONS(656),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(660),
    [anon_sym_i8] = ACTIONS(660),
    [anon_sym_u16] = ACTIONS(660),
    [anon_sym_i16] = ACTIONS(660),
    [anon_sym_u32] = ACTIONS(660),
    [anon_sym_i32] = ACTIONS(660),
    [anon_sym_u64] = ACTIONS(660),
    [anon_sym_i64] = ACTIONS(660),
    [anon_sym_u128] = ACTIONS(660),
    [anon_sym_i128] = ACTIONS(660),
    [anon_sym_isize] = ACTIONS(660),
    [anon_sym_usize] = ACTIONS(660),
    [anon_sym_f32] = ACTIONS(660),
    [anon_sym_f64] = ACTIONS(660),
    [anon_sym_bool] = ACTIONS(660),
    [anon_sym_str] = ACTIONS(660),
    [anon_sym_char] = ACTIONS(660),
    [anon_sym_SQUOTE] = ACTIONS(822),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(666),
    [anon_sym_default] = ACTIONS(668),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(676),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(688),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(808),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_mutable_specifier] = ACTIONS(836),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(838),
    [sym_super] = ACTIONS(712),
    [sym_crate] = ACTIONS(712),
    [sym_metavariable] = ACTIONS(714),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [209] = {
    [sym_bracketed_type] = STATE(2489),
    [sym_generic_type] = STATE(2488),
    [sym_generic_type_with_turbofish] = STATE(2487),
    [sym_macro_invocation] = STATE(1422),
    [sym_scoped_identifier] = STATE(1337),
    [sym_scoped_type_identifier] = STATE(2013),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1458),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [sym_identifier] = ACTIONS(840),
    [anon_sym_LPAREN] = ACTIONS(842),
    [anon_sym_LBRACE] = ACTIONS(842),
    [anon_sym_LBRACK] = ACTIONS(842),
    [anon_sym_STAR] = ACTIONS(842),
    [anon_sym_u8] = ACTIONS(840),
    [anon_sym_i8] = ACTIONS(840),
    [anon_sym_u16] = ACTIONS(840),
    [anon_sym_i16] = ACTIONS(840),
    [anon_sym_u32] = ACTIONS(840),
    [anon_sym_i32] = ACTIONS(840),
    [anon_sym_u64] = ACTIONS(840),
    [anon_sym_i64] = ACTIONS(840),
    [anon_sym_u128] = ACTIONS(840),
    [anon_sym_i128] = ACTIONS(840),
    [anon_sym_isize] = ACTIONS(840),
    [anon_sym_usize] = ACTIONS(840),
    [anon_sym_f32] = ACTIONS(840),
    [anon_sym_f64] = ACTIONS(840),
    [anon_sym_bool] = ACTIONS(840),
    [anon_sym_str] = ACTIONS(840),
    [anon_sym_char] = ACTIONS(840),
    [anon_sym_SQUOTE] = ACTIONS(840),
    [anon_sym_async] = ACTIONS(840),
    [anon_sym_break] = ACTIONS(840),
    [anon_sym_const] = ACTIONS(840),
    [anon_sym_continue] = ACTIONS(840),
    [anon_sym_default] = ACTIONS(840),
    [anon_sym_for] = ACTIONS(840),
    [anon_sym_if] = ACTIONS(840),
    [anon_sym_loop] = ACTIONS(840),
    [anon_sym_match] = ACTIONS(840),
    [anon_sym_return] = ACTIONS(840),
    [anon_sym_union] = ACTIONS(840),
    [anon_sym_unsafe] = ACTIONS(840),
    [anon_sym_while] = ACTIONS(840),
    [anon_sym_BANG] = ACTIONS(842),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_DASH_GT] = ACTIONS(842),
    [anon_sym_LT] = ACTIONS(842),
    [anon_sym_COLON_COLON] = ACTIONS(842),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(842),
    [sym_mutable_specifier] = ACTIONS(796),
    [anon_sym_DOT_DOT] = ACTIONS(842),
    [anon_sym_DASH] = ACTIONS(840),
    [anon_sym_PIPE] = ACTIONS(842),
    [anon_sym_yield] = ACTIONS(840),
    [anon_sym_move] = ACTIONS(840),
    [sym_integer_literal] = ACTIONS(842),
    [aux_sym_string_literal_token1] = ACTIONS(842),
    [sym_char_literal] = ACTIONS(842),
    [anon_sym_true] = ACTIONS(840),
    [anon_sym_false] = ACTIONS(840),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(840),
    [sym_super] = ACTIONS(840),
    [sym_crate] = ACTIONS(840),
    [sym_metavariable] = ACTIONS(842),
    [sym_raw_string_literal] = ACTIONS(842),
    [sym_float_literal] = ACTIONS(842),
    [sym_block_comment] = ACTIONS(3),
  },
  [210] = {
    [sym_else_clause] = STATE(221),
    [sym_identifier] = ACTIONS(402),
    [anon_sym_LPAREN] = ACTIONS(400),
    [anon_sym_RBRACE] = ACTIONS(400),
    [anon_sym_LBRACK] = ACTIONS(400),
    [anon_sym_PLUS] = ACTIONS(402),
    [anon_sym_STAR] = ACTIONS(402),
    [anon_sym_QMARK] = ACTIONS(400),
    [anon_sym_u8] = ACTIONS(402),
    [anon_sym_i8] = ACTIONS(402),
    [anon_sym_u16] = ACTIONS(402),
    [anon_sym_i16] = ACTIONS(402),
    [anon_sym_u32] = ACTIONS(402),
    [anon_sym_i32] = ACTIONS(402),
    [anon_sym_u64] = ACTIONS(402),
    [anon_sym_i64] = ACTIONS(402),
    [anon_sym_u128] = ACTIONS(402),
    [anon_sym_i128] = ACTIONS(402),
    [anon_sym_isize] = ACTIONS(402),
    [anon_sym_usize] = ACTIONS(402),
    [anon_sym_f32] = ACTIONS(402),
    [anon_sym_f64] = ACTIONS(402),
    [anon_sym_bool] = ACTIONS(402),
    [anon_sym_str] = ACTIONS(402),
    [anon_sym_char] = ACTIONS(402),
    [anon_sym_as] = ACTIONS(402),
    [anon_sym_const] = ACTIONS(402),
    [anon_sym_default] = ACTIONS(402),
    [anon_sym_union] = ACTIONS(402),
    [anon_sym_POUND] = ACTIONS(400),
    [anon_sym_EQ] = ACTIONS(402),
    [anon_sym_COMMA] = ACTIONS(400),
    [anon_sym_ref] = ACTIONS(402),
    [anon_sym_LT] = ACTIONS(402),
    [anon_sym_GT] = ACTIONS(402),
    [anon_sym_else] = ACTIONS(844),
    [anon_sym_COLON_COLON] = ACTIONS(400),
    [anon_sym__] = ACTIONS(402),
    [anon_sym_AMP] = ACTIONS(402),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(400),
    [sym_mutable_specifier] = ACTIONS(402),
    [anon_sym_DOT_DOT] = ACTIONS(402),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(400),
    [anon_sym_DASH] = ACTIONS(402),
    [anon_sym_AMP_AMP] = ACTIONS(400),
    [anon_sym_PIPE_PIPE] = ACTIONS(400),
    [anon_sym_PIPE] = ACTIONS(402),
    [anon_sym_CARET] = ACTIONS(402),
    [anon_sym_EQ_EQ] = ACTIONS(400),
    [anon_sym_BANG_EQ] = ACTIONS(400),
    [anon_sym_LT_EQ] = ACTIONS(400),
    [anon_sym_GT_EQ] = ACTIONS(400),
    [anon_sym_LT_LT] = ACTIONS(402),
    [anon_sym_GT_GT] = ACTIONS(402),
    [anon_sym_SLASH] = ACTIONS(402),
    [anon_sym_PERCENT] = ACTIONS(402),
    [anon_sym_PLUS_EQ] = ACTIONS(400),
    [anon_sym_DASH_EQ] = ACTIONS(400),
    [anon_sym_STAR_EQ] = ACTIONS(400),
    [anon_sym_SLASH_EQ] = ACTIONS(400),
    [anon_sym_PERCENT_EQ] = ACTIONS(400),
    [anon_sym_AMP_EQ] = ACTIONS(400),
    [anon_sym_PIPE_EQ] = ACTIONS(400),
    [anon_sym_CARET_EQ] = ACTIONS(400),
    [anon_sym_LT_LT_EQ] = ACTIONS(400),
    [anon_sym_GT_GT_EQ] = ACTIONS(400),
    [anon_sym_DOT] = ACTIONS(402),
    [sym_integer_literal] = ACTIONS(400),
    [aux_sym_string_literal_token1] = ACTIONS(400),
    [sym_char_literal] = ACTIONS(400),
    [anon_sym_true] = ACTIONS(402),
    [anon_sym_false] = ACTIONS(402),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(402),
    [sym_super] = ACTIONS(402),
    [sym_crate] = ACTIONS(402),
    [sym_metavariable] = ACTIONS(400),
    [sym_raw_string_literal] = ACTIONS(400),
    [sym_float_literal] = ACTIONS(400),
    [sym_block_comment] = ACTIONS(3),
  },
  [211] = {
    [sym_identifier] = ACTIONS(408),
    [anon_sym_LPAREN] = ACTIONS(406),
    [anon_sym_RBRACE] = ACTIONS(406),
    [anon_sym_LBRACK] = ACTIONS(406),
    [anon_sym_PLUS] = ACTIONS(408),
    [anon_sym_STAR] = ACTIONS(408),
    [anon_sym_QMARK] = ACTIONS(406),
    [anon_sym_u8] = ACTIONS(408),
    [anon_sym_i8] = ACTIONS(408),
    [anon_sym_u16] = ACTIONS(408),
    [anon_sym_i16] = ACTIONS(408),
    [anon_sym_u32] = ACTIONS(408),
    [anon_sym_i32] = ACTIONS(408),
    [anon_sym_u64] = ACTIONS(408),
    [anon_sym_i64] = ACTIONS(408),
    [anon_sym_u128] = ACTIONS(408),
    [anon_sym_i128] = ACTIONS(408),
    [anon_sym_isize] = ACTIONS(408),
    [anon_sym_usize] = ACTIONS(408),
    [anon_sym_f32] = ACTIONS(408),
    [anon_sym_f64] = ACTIONS(408),
    [anon_sym_bool] = ACTIONS(408),
    [anon_sym_str] = ACTIONS(408),
    [anon_sym_char] = ACTIONS(408),
    [anon_sym_as] = ACTIONS(408),
    [anon_sym_const] = ACTIONS(408),
    [anon_sym_default] = ACTIONS(408),
    [anon_sym_union] = ACTIONS(408),
    [anon_sym_POUND] = ACTIONS(406),
    [anon_sym_EQ] = ACTIONS(408),
    [anon_sym_COMMA] = ACTIONS(406),
    [anon_sym_ref] = ACTIONS(408),
    [anon_sym_LT] = ACTIONS(408),
    [anon_sym_GT] = ACTIONS(408),
    [anon_sym_else] = ACTIONS(408),
    [anon_sym_COLON_COLON] = ACTIONS(406),
    [anon_sym__] = ACTIONS(408),
    [anon_sym_AMP] = ACTIONS(408),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(406),
    [sym_mutable_specifier] = ACTIONS(408),
    [anon_sym_DOT_DOT] = ACTIONS(408),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(406),
    [anon_sym_DASH] = ACTIONS(408),
    [anon_sym_AMP_AMP] = ACTIONS(406),
    [anon_sym_PIPE_PIPE] = ACTIONS(406),
    [anon_sym_PIPE] = ACTIONS(408),
    [anon_sym_CARET] = ACTIONS(408),
    [anon_sym_EQ_EQ] = ACTIONS(406),
    [anon_sym_BANG_EQ] = ACTIONS(406),
    [anon_sym_LT_EQ] = ACTIONS(406),
    [anon_sym_GT_EQ] = ACTIONS(406),
    [anon_sym_LT_LT] = ACTIONS(408),
    [anon_sym_GT_GT] = ACTIONS(408),
    [anon_sym_SLASH] = ACTIONS(408),
    [anon_sym_PERCENT] = ACTIONS(408),
    [anon_sym_PLUS_EQ] = ACTIONS(406),
    [anon_sym_DASH_EQ] = ACTIONS(406),
    [anon_sym_STAR_EQ] = ACTIONS(406),
    [anon_sym_SLASH_EQ] = ACTIONS(406),
    [anon_sym_PERCENT_EQ] = ACTIONS(406),
    [anon_sym_AMP_EQ] = ACTIONS(406),
    [anon_sym_PIPE_EQ] = ACTIONS(406),
    [anon_sym_CARET_EQ] = ACTIONS(406),
    [anon_sym_LT_LT_EQ] = ACTIONS(406),
    [anon_sym_GT_GT_EQ] = ACTIONS(406),
    [anon_sym_DOT] = ACTIONS(408),
    [sym_integer_literal] = ACTIONS(406),
    [aux_sym_string_literal_token1] = ACTIONS(406),
    [sym_char_literal] = ACTIONS(406),
    [anon_sym_true] = ACTIONS(408),
    [anon_sym_false] = ACTIONS(408),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(408),
    [sym_super] = ACTIONS(408),
    [sym_crate] = ACTIONS(408),
    [sym_metavariable] = ACTIONS(406),
    [sym_raw_string_literal] = ACTIONS(406),
    [sym_float_literal] = ACTIONS(406),
    [sym_block_comment] = ACTIONS(3),
  },
  [212] = {
    [sym_identifier] = ACTIONS(416),
    [anon_sym_LPAREN] = ACTIONS(414),
    [anon_sym_RBRACE] = ACTIONS(414),
    [anon_sym_LBRACK] = ACTIONS(414),
    [anon_sym_PLUS] = ACTIONS(416),
    [anon_sym_STAR] = ACTIONS(416),
    [anon_sym_QMARK] = ACTIONS(414),
    [anon_sym_u8] = ACTIONS(416),
    [anon_sym_i8] = ACTIONS(416),
    [anon_sym_u16] = ACTIONS(416),
    [anon_sym_i16] = ACTIONS(416),
    [anon_sym_u32] = ACTIONS(416),
    [anon_sym_i32] = ACTIONS(416),
    [anon_sym_u64] = ACTIONS(416),
    [anon_sym_i64] = ACTIONS(416),
    [anon_sym_u128] = ACTIONS(416),
    [anon_sym_i128] = ACTIONS(416),
    [anon_sym_isize] = ACTIONS(416),
    [anon_sym_usize] = ACTIONS(416),
    [anon_sym_f32] = ACTIONS(416),
    [anon_sym_f64] = ACTIONS(416),
    [anon_sym_bool] = ACTIONS(416),
    [anon_sym_str] = ACTIONS(416),
    [anon_sym_char] = ACTIONS(416),
    [anon_sym_as] = ACTIONS(416),
    [anon_sym_const] = ACTIONS(416),
    [anon_sym_default] = ACTIONS(416),
    [anon_sym_union] = ACTIONS(416),
    [anon_sym_POUND] = ACTIONS(414),
    [anon_sym_EQ] = ACTIONS(416),
    [anon_sym_COMMA] = ACTIONS(414),
    [anon_sym_ref] = ACTIONS(416),
    [anon_sym_LT] = ACTIONS(416),
    [anon_sym_GT] = ACTIONS(416),
    [anon_sym_else] = ACTIONS(416),
    [anon_sym_COLON_COLON] = ACTIONS(414),
    [anon_sym__] = ACTIONS(416),
    [anon_sym_AMP] = ACTIONS(416),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(414),
    [sym_mutable_specifier] = ACTIONS(416),
    [anon_sym_DOT_DOT] = ACTIONS(416),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(414),
    [anon_sym_DASH] = ACTIONS(416),
    [anon_sym_AMP_AMP] = ACTIONS(414),
    [anon_sym_PIPE_PIPE] = ACTIONS(414),
    [anon_sym_PIPE] = ACTIONS(416),
    [anon_sym_CARET] = ACTIONS(416),
    [anon_sym_EQ_EQ] = ACTIONS(414),
    [anon_sym_BANG_EQ] = ACTIONS(414),
    [anon_sym_LT_EQ] = ACTIONS(414),
    [anon_sym_GT_EQ] = ACTIONS(414),
    [anon_sym_LT_LT] = ACTIONS(416),
    [anon_sym_GT_GT] = ACTIONS(416),
    [anon_sym_SLASH] = ACTIONS(416),
    [anon_sym_PERCENT] = ACTIONS(416),
    [anon_sym_PLUS_EQ] = ACTIONS(414),
    [anon_sym_DASH_EQ] = ACTIONS(414),
    [anon_sym_STAR_EQ] = ACTIONS(414),
    [anon_sym_SLASH_EQ] = ACTIONS(414),
    [anon_sym_PERCENT_EQ] = ACTIONS(414),
    [anon_sym_AMP_EQ] = ACTIONS(414),
    [anon_sym_PIPE_EQ] = ACTIONS(414),
    [anon_sym_CARET_EQ] = ACTIONS(414),
    [anon_sym_LT_LT_EQ] = ACTIONS(414),
    [anon_sym_GT_GT_EQ] = ACTIONS(414),
    [anon_sym_DOT] = ACTIONS(416),
    [sym_integer_literal] = ACTIONS(414),
    [aux_sym_string_literal_token1] = ACTIONS(414),
    [sym_char_literal] = ACTIONS(414),
    [anon_sym_true] = ACTIONS(416),
    [anon_sym_false] = ACTIONS(416),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(416),
    [sym_super] = ACTIONS(416),
    [sym_crate] = ACTIONS(416),
    [sym_metavariable] = ACTIONS(414),
    [sym_raw_string_literal] = ACTIONS(414),
    [sym_float_literal] = ACTIONS(414),
    [sym_block_comment] = ACTIONS(3),
  },
  [213] = {
    [sym_identifier] = ACTIONS(412),
    [anon_sym_LPAREN] = ACTIONS(410),
    [anon_sym_RBRACE] = ACTIONS(410),
    [anon_sym_LBRACK] = ACTIONS(410),
    [anon_sym_PLUS] = ACTIONS(412),
    [anon_sym_STAR] = ACTIONS(412),
    [anon_sym_QMARK] = ACTIONS(410),
    [anon_sym_u8] = ACTIONS(412),
    [anon_sym_i8] = ACTIONS(412),
    [anon_sym_u16] = ACTIONS(412),
    [anon_sym_i16] = ACTIONS(412),
    [anon_sym_u32] = ACTIONS(412),
    [anon_sym_i32] = ACTIONS(412),
    [anon_sym_u64] = ACTIONS(412),
    [anon_sym_i64] = ACTIONS(412),
    [anon_sym_u128] = ACTIONS(412),
    [anon_sym_i128] = ACTIONS(412),
    [anon_sym_isize] = ACTIONS(412),
    [anon_sym_usize] = ACTIONS(412),
    [anon_sym_f32] = ACTIONS(412),
    [anon_sym_f64] = ACTIONS(412),
    [anon_sym_bool] = ACTIONS(412),
    [anon_sym_str] = ACTIONS(412),
    [anon_sym_char] = ACTIONS(412),
    [anon_sym_as] = ACTIONS(412),
    [anon_sym_const] = ACTIONS(412),
    [anon_sym_default] = ACTIONS(412),
    [anon_sym_union] = ACTIONS(412),
    [anon_sym_POUND] = ACTIONS(410),
    [anon_sym_EQ] = ACTIONS(412),
    [anon_sym_COMMA] = ACTIONS(410),
    [anon_sym_ref] = ACTIONS(412),
    [anon_sym_LT] = ACTIONS(412),
    [anon_sym_GT] = ACTIONS(412),
    [anon_sym_else] = ACTIONS(412),
    [anon_sym_COLON_COLON] = ACTIONS(410),
    [anon_sym__] = ACTIONS(412),
    [anon_sym_AMP] = ACTIONS(412),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(410),
    [sym_mutable_specifier] = ACTIONS(412),
    [anon_sym_DOT_DOT] = ACTIONS(412),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(410),
    [anon_sym_DASH] = ACTIONS(412),
    [anon_sym_AMP_AMP] = ACTIONS(410),
    [anon_sym_PIPE_PIPE] = ACTIONS(410),
    [anon_sym_PIPE] = ACTIONS(412),
    [anon_sym_CARET] = ACTIONS(412),
    [anon_sym_EQ_EQ] = ACTIONS(410),
    [anon_sym_BANG_EQ] = ACTIONS(410),
    [anon_sym_LT_EQ] = ACTIONS(410),
    [anon_sym_GT_EQ] = ACTIONS(410),
    [anon_sym_LT_LT] = ACTIONS(412),
    [anon_sym_GT_GT] = ACTIONS(412),
    [anon_sym_SLASH] = ACTIONS(412),
    [anon_sym_PERCENT] = ACTIONS(412),
    [anon_sym_PLUS_EQ] = ACTIONS(410),
    [anon_sym_DASH_EQ] = ACTIONS(410),
    [anon_sym_STAR_EQ] = ACTIONS(410),
    [anon_sym_SLASH_EQ] = ACTIONS(410),
    [anon_sym_PERCENT_EQ] = ACTIONS(410),
    [anon_sym_AMP_EQ] = ACTIONS(410),
    [anon_sym_PIPE_EQ] = ACTIONS(410),
    [anon_sym_CARET_EQ] = ACTIONS(410),
    [anon_sym_LT_LT_EQ] = ACTIONS(410),
    [anon_sym_GT_GT_EQ] = ACTIONS(410),
    [anon_sym_DOT] = ACTIONS(412),
    [sym_integer_literal] = ACTIONS(410),
    [aux_sym_string_literal_token1] = ACTIONS(410),
    [sym_char_literal] = ACTIONS(410),
    [anon_sym_true] = ACTIONS(412),
    [anon_sym_false] = ACTIONS(412),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(412),
    [sym_super] = ACTIONS(412),
    [sym_crate] = ACTIONS(412),
    [sym_metavariable] = ACTIONS(410),
    [sym_raw_string_literal] = ACTIONS(410),
    [sym_float_literal] = ACTIONS(410),
    [sym_block_comment] = ACTIONS(3),
  },
  [214] = {
    [sym_identifier] = ACTIONS(846),
    [anon_sym_LPAREN] = ACTIONS(848),
    [anon_sym_RBRACE] = ACTIONS(458),
    [anon_sym_LBRACK] = ACTIONS(848),
    [anon_sym_PLUS] = ACTIONS(460),
    [anon_sym_STAR] = ACTIONS(460),
    [anon_sym_QMARK] = ACTIONS(458),
    [anon_sym_u8] = ACTIONS(846),
    [anon_sym_i8] = ACTIONS(846),
    [anon_sym_u16] = ACTIONS(846),
    [anon_sym_i16] = ACTIONS(846),
    [anon_sym_u32] = ACTIONS(846),
    [anon_sym_i32] = ACTIONS(846),
    [anon_sym_u64] = ACTIONS(846),
    [anon_sym_i64] = ACTIONS(846),
    [anon_sym_u128] = ACTIONS(846),
    [anon_sym_i128] = ACTIONS(846),
    [anon_sym_isize] = ACTIONS(846),
    [anon_sym_usize] = ACTIONS(846),
    [anon_sym_f32] = ACTIONS(846),
    [anon_sym_f64] = ACTIONS(846),
    [anon_sym_bool] = ACTIONS(846),
    [anon_sym_str] = ACTIONS(846),
    [anon_sym_char] = ACTIONS(846),
    [anon_sym_as] = ACTIONS(460),
    [anon_sym_const] = ACTIONS(846),
    [anon_sym_default] = ACTIONS(846),
    [anon_sym_union] = ACTIONS(846),
    [anon_sym_POUND] = ACTIONS(848),
    [anon_sym_EQ] = ACTIONS(460),
    [anon_sym_COMMA] = ACTIONS(458),
    [anon_sym_ref] = ACTIONS(846),
    [anon_sym_LT] = ACTIONS(846),
    [anon_sym_GT] = ACTIONS(460),
    [anon_sym_COLON_COLON] = ACTIONS(848),
    [anon_sym__] = ACTIONS(846),
    [anon_sym_AMP] = ACTIONS(846),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(458),
    [sym_mutable_specifier] = ACTIONS(846),
    [anon_sym_DOT_DOT] = ACTIONS(846),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(458),
    [anon_sym_DASH] = ACTIONS(846),
    [anon_sym_AMP_AMP] = ACTIONS(458),
    [anon_sym_PIPE_PIPE] = ACTIONS(458),
    [anon_sym_PIPE] = ACTIONS(460),
    [anon_sym_CARET] = ACTIONS(460),
    [anon_sym_EQ_EQ] = ACTIONS(458),
    [anon_sym_BANG_EQ] = ACTIONS(458),
    [anon_sym_LT_EQ] = ACTIONS(458),
    [anon_sym_GT_EQ] = ACTIONS(458),
    [anon_sym_LT_LT] = ACTIONS(460),
    [anon_sym_GT_GT] = ACTIONS(460),
    [anon_sym_SLASH] = ACTIONS(460),
    [anon_sym_PERCENT] = ACTIONS(460),
    [anon_sym_PLUS_EQ] = ACTIONS(458),
    [anon_sym_DASH_EQ] = ACTIONS(458),
    [anon_sym_STAR_EQ] = ACTIONS(458),
    [anon_sym_SLASH_EQ] = ACTIONS(458),
    [anon_sym_PERCENT_EQ] = ACTIONS(458),
    [anon_sym_AMP_EQ] = ACTIONS(458),
    [anon_sym_PIPE_EQ] = ACTIONS(458),
    [anon_sym_CARET_EQ] = ACTIONS(458),
    [anon_sym_LT_LT_EQ] = ACTIONS(458),
    [anon_sym_GT_GT_EQ] = ACTIONS(458),
    [anon_sym_DOT] = ACTIONS(460),
    [sym_integer_literal] = ACTIONS(848),
    [aux_sym_string_literal_token1] = ACTIONS(848),
    [sym_char_literal] = ACTIONS(848),
    [anon_sym_true] = ACTIONS(846),
    [anon_sym_false] = ACTIONS(846),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(846),
    [sym_super] = ACTIONS(846),
    [sym_crate] = ACTIONS(846),
    [sym_metavariable] = ACTIONS(848),
    [sym_raw_string_literal] = ACTIONS(848),
    [sym_float_literal] = ACTIONS(848),
    [sym_block_comment] = ACTIONS(3),
  },
  [215] = {
    [sym_identifier] = ACTIONS(424),
    [anon_sym_LPAREN] = ACTIONS(422),
    [anon_sym_RBRACE] = ACTIONS(422),
    [anon_sym_LBRACK] = ACTIONS(422),
    [anon_sym_PLUS] = ACTIONS(424),
    [anon_sym_STAR] = ACTIONS(424),
    [anon_sym_QMARK] = ACTIONS(422),
    [anon_sym_u8] = ACTIONS(424),
    [anon_sym_i8] = ACTIONS(424),
    [anon_sym_u16] = ACTIONS(424),
    [anon_sym_i16] = ACTIONS(424),
    [anon_sym_u32] = ACTIONS(424),
    [anon_sym_i32] = ACTIONS(424),
    [anon_sym_u64] = ACTIONS(424),
    [anon_sym_i64] = ACTIONS(424),
    [anon_sym_u128] = ACTIONS(424),
    [anon_sym_i128] = ACTIONS(424),
    [anon_sym_isize] = ACTIONS(424),
    [anon_sym_usize] = ACTIONS(424),
    [anon_sym_f32] = ACTIONS(424),
    [anon_sym_f64] = ACTIONS(424),
    [anon_sym_bool] = ACTIONS(424),
    [anon_sym_str] = ACTIONS(424),
    [anon_sym_char] = ACTIONS(424),
    [anon_sym_as] = ACTIONS(424),
    [anon_sym_const] = ACTIONS(424),
    [anon_sym_default] = ACTIONS(424),
    [anon_sym_union] = ACTIONS(424),
    [anon_sym_POUND] = ACTIONS(422),
    [anon_sym_EQ] = ACTIONS(424),
    [anon_sym_COMMA] = ACTIONS(422),
    [anon_sym_ref] = ACTIONS(424),
    [anon_sym_LT] = ACTIONS(424),
    [anon_sym_GT] = ACTIONS(424),
    [anon_sym_COLON_COLON] = ACTIONS(422),
    [anon_sym__] = ACTIONS(424),
    [anon_sym_AMP] = ACTIONS(424),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(422),
    [sym_mutable_specifier] = ACTIONS(424),
    [anon_sym_DOT_DOT] = ACTIONS(424),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(422),
    [anon_sym_DASH] = ACTIONS(424),
    [anon_sym_AMP_AMP] = ACTIONS(422),
    [anon_sym_PIPE_PIPE] = ACTIONS(422),
    [anon_sym_PIPE] = ACTIONS(424),
    [anon_sym_CARET] = ACTIONS(424),
    [anon_sym_EQ_EQ] = ACTIONS(422),
    [anon_sym_BANG_EQ] = ACTIONS(422),
    [anon_sym_LT_EQ] = ACTIONS(422),
    [anon_sym_GT_EQ] = ACTIONS(422),
    [anon_sym_LT_LT] = ACTIONS(424),
    [anon_sym_GT_GT] = ACTIONS(424),
    [anon_sym_SLASH] = ACTIONS(424),
    [anon_sym_PERCENT] = ACTIONS(424),
    [anon_sym_PLUS_EQ] = ACTIONS(422),
    [anon_sym_DASH_EQ] = ACTIONS(422),
    [anon_sym_STAR_EQ] = ACTIONS(422),
    [anon_sym_SLASH_EQ] = ACTIONS(422),
    [anon_sym_PERCENT_EQ] = ACTIONS(422),
    [anon_sym_AMP_EQ] = ACTIONS(422),
    [anon_sym_PIPE_EQ] = ACTIONS(422),
    [anon_sym_CARET_EQ] = ACTIONS(422),
    [anon_sym_LT_LT_EQ] = ACTIONS(422),
    [anon_sym_GT_GT_EQ] = ACTIONS(422),
    [anon_sym_DOT] = ACTIONS(424),
    [sym_integer_literal] = ACTIONS(422),
    [aux_sym_string_literal_token1] = ACTIONS(422),
    [sym_char_literal] = ACTIONS(422),
    [anon_sym_true] = ACTIONS(424),
    [anon_sym_false] = ACTIONS(424),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(424),
    [sym_super] = ACTIONS(424),
    [sym_crate] = ACTIONS(424),
    [sym_metavariable] = ACTIONS(422),
    [sym_raw_string_literal] = ACTIONS(422),
    [sym_float_literal] = ACTIONS(422),
    [sym_block_comment] = ACTIONS(3),
  },
  [216] = {
    [sym_identifier] = ACTIONS(610),
    [anon_sym_LPAREN] = ACTIONS(608),
    [anon_sym_RBRACE] = ACTIONS(608),
    [anon_sym_LBRACK] = ACTIONS(608),
    [anon_sym_PLUS] = ACTIONS(610),
    [anon_sym_STAR] = ACTIONS(610),
    [anon_sym_QMARK] = ACTIONS(608),
    [anon_sym_u8] = ACTIONS(610),
    [anon_sym_i8] = ACTIONS(610),
    [anon_sym_u16] = ACTIONS(610),
    [anon_sym_i16] = ACTIONS(610),
    [anon_sym_u32] = ACTIONS(610),
    [anon_sym_i32] = ACTIONS(610),
    [anon_sym_u64] = ACTIONS(610),
    [anon_sym_i64] = ACTIONS(610),
    [anon_sym_u128] = ACTIONS(610),
    [anon_sym_i128] = ACTIONS(610),
    [anon_sym_isize] = ACTIONS(610),
    [anon_sym_usize] = ACTIONS(610),
    [anon_sym_f32] = ACTIONS(610),
    [anon_sym_f64] = ACTIONS(610),
    [anon_sym_bool] = ACTIONS(610),
    [anon_sym_str] = ACTIONS(610),
    [anon_sym_char] = ACTIONS(610),
    [anon_sym_as] = ACTIONS(610),
    [anon_sym_const] = ACTIONS(610),
    [anon_sym_default] = ACTIONS(610),
    [anon_sym_union] = ACTIONS(610),
    [anon_sym_POUND] = ACTIONS(608),
    [anon_sym_EQ] = ACTIONS(610),
    [anon_sym_COMMA] = ACTIONS(608),
    [anon_sym_ref] = ACTIONS(610),
    [anon_sym_LT] = ACTIONS(610),
    [anon_sym_GT] = ACTIONS(610),
    [anon_sym_COLON_COLON] = ACTIONS(608),
    [anon_sym__] = ACTIONS(610),
    [anon_sym_AMP] = ACTIONS(610),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(608),
    [sym_mutable_specifier] = ACTIONS(610),
    [anon_sym_DOT_DOT] = ACTIONS(610),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(608),
    [anon_sym_DASH] = ACTIONS(610),
    [anon_sym_AMP_AMP] = ACTIONS(608),
    [anon_sym_PIPE_PIPE] = ACTIONS(608),
    [anon_sym_PIPE] = ACTIONS(610),
    [anon_sym_CARET] = ACTIONS(610),
    [anon_sym_EQ_EQ] = ACTIONS(608),
    [anon_sym_BANG_EQ] = ACTIONS(608),
    [anon_sym_LT_EQ] = ACTIONS(608),
    [anon_sym_GT_EQ] = ACTIONS(608),
    [anon_sym_LT_LT] = ACTIONS(610),
    [anon_sym_GT_GT] = ACTIONS(610),
    [anon_sym_SLASH] = ACTIONS(610),
    [anon_sym_PERCENT] = ACTIONS(610),
    [anon_sym_PLUS_EQ] = ACTIONS(608),
    [anon_sym_DASH_EQ] = ACTIONS(608),
    [anon_sym_STAR_EQ] = ACTIONS(608),
    [anon_sym_SLASH_EQ] = ACTIONS(608),
    [anon_sym_PERCENT_EQ] = ACTIONS(608),
    [anon_sym_AMP_EQ] = ACTIONS(608),
    [anon_sym_PIPE_EQ] = ACTIONS(608),
    [anon_sym_CARET_EQ] = ACTIONS(608),
    [anon_sym_LT_LT_EQ] = ACTIONS(608),
    [anon_sym_GT_GT_EQ] = ACTIONS(608),
    [anon_sym_DOT] = ACTIONS(610),
    [sym_integer_literal] = ACTIONS(608),
    [aux_sym_string_literal_token1] = ACTIONS(608),
    [sym_char_literal] = ACTIONS(608),
    [anon_sym_true] = ACTIONS(610),
    [anon_sym_false] = ACTIONS(610),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(610),
    [sym_super] = ACTIONS(610),
    [sym_crate] = ACTIONS(610),
    [sym_metavariable] = ACTIONS(608),
    [sym_raw_string_literal] = ACTIONS(608),
    [sym_float_literal] = ACTIONS(608),
    [sym_block_comment] = ACTIONS(3),
  },
  [217] = {
    [sym_identifier] = ACTIONS(474),
    [anon_sym_LPAREN] = ACTIONS(472),
    [anon_sym_RBRACE] = ACTIONS(472),
    [anon_sym_LBRACK] = ACTIONS(472),
    [anon_sym_PLUS] = ACTIONS(474),
    [anon_sym_STAR] = ACTIONS(474),
    [anon_sym_QMARK] = ACTIONS(472),
    [anon_sym_u8] = ACTIONS(474),
    [anon_sym_i8] = ACTIONS(474),
    [anon_sym_u16] = ACTIONS(474),
    [anon_sym_i16] = ACTIONS(474),
    [anon_sym_u32] = ACTIONS(474),
    [anon_sym_i32] = ACTIONS(474),
    [anon_sym_u64] = ACTIONS(474),
    [anon_sym_i64] = ACTIONS(474),
    [anon_sym_u128] = ACTIONS(474),
    [anon_sym_i128] = ACTIONS(474),
    [anon_sym_isize] = ACTIONS(474),
    [anon_sym_usize] = ACTIONS(474),
    [anon_sym_f32] = ACTIONS(474),
    [anon_sym_f64] = ACTIONS(474),
    [anon_sym_bool] = ACTIONS(474),
    [anon_sym_str] = ACTIONS(474),
    [anon_sym_char] = ACTIONS(474),
    [anon_sym_as] = ACTIONS(474),
    [anon_sym_const] = ACTIONS(474),
    [anon_sym_default] = ACTIONS(474),
    [anon_sym_union] = ACTIONS(474),
    [anon_sym_POUND] = ACTIONS(472),
    [anon_sym_EQ] = ACTIONS(474),
    [anon_sym_COMMA] = ACTIONS(472),
    [anon_sym_ref] = ACTIONS(474),
    [anon_sym_LT] = ACTIONS(474),
    [anon_sym_GT] = ACTIONS(474),
    [anon_sym_COLON_COLON] = ACTIONS(472),
    [anon_sym__] = ACTIONS(474),
    [anon_sym_AMP] = ACTIONS(474),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(472),
    [sym_mutable_specifier] = ACTIONS(474),
    [anon_sym_DOT_DOT] = ACTIONS(474),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(472),
    [anon_sym_DASH] = ACTIONS(474),
    [anon_sym_AMP_AMP] = ACTIONS(472),
    [anon_sym_PIPE_PIPE] = ACTIONS(472),
    [anon_sym_PIPE] = ACTIONS(474),
    [anon_sym_CARET] = ACTIONS(474),
    [anon_sym_EQ_EQ] = ACTIONS(472),
    [anon_sym_BANG_EQ] = ACTIONS(472),
    [anon_sym_LT_EQ] = ACTIONS(472),
    [anon_sym_GT_EQ] = ACTIONS(472),
    [anon_sym_LT_LT] = ACTIONS(474),
    [anon_sym_GT_GT] = ACTIONS(474),
    [anon_sym_SLASH] = ACTIONS(474),
    [anon_sym_PERCENT] = ACTIONS(474),
    [anon_sym_PLUS_EQ] = ACTIONS(472),
    [anon_sym_DASH_EQ] = ACTIONS(472),
    [anon_sym_STAR_EQ] = ACTIONS(472),
    [anon_sym_SLASH_EQ] = ACTIONS(472),
    [anon_sym_PERCENT_EQ] = ACTIONS(472),
    [anon_sym_AMP_EQ] = ACTIONS(472),
    [anon_sym_PIPE_EQ] = ACTIONS(472),
    [anon_sym_CARET_EQ] = ACTIONS(472),
    [anon_sym_LT_LT_EQ] = ACTIONS(472),
    [anon_sym_GT_GT_EQ] = ACTIONS(472),
    [anon_sym_DOT] = ACTIONS(474),
    [sym_integer_literal] = ACTIONS(472),
    [aux_sym_string_literal_token1] = ACTIONS(472),
    [sym_char_literal] = ACTIONS(472),
    [anon_sym_true] = ACTIONS(474),
    [anon_sym_false] = ACTIONS(474),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(474),
    [sym_super] = ACTIONS(474),
    [sym_crate] = ACTIONS(474),
    [sym_metavariable] = ACTIONS(472),
    [sym_raw_string_literal] = ACTIONS(472),
    [sym_float_literal] = ACTIONS(472),
    [sym_block_comment] = ACTIONS(3),
  },
  [218] = {
    [sym_identifier] = ACTIONS(850),
    [anon_sym_LPAREN] = ACTIONS(852),
    [anon_sym_RBRACE] = ACTIONS(458),
    [anon_sym_LBRACK] = ACTIONS(852),
    [anon_sym_PLUS] = ACTIONS(460),
    [anon_sym_STAR] = ACTIONS(460),
    [anon_sym_QMARK] = ACTIONS(458),
    [anon_sym_u8] = ACTIONS(850),
    [anon_sym_i8] = ACTIONS(850),
    [anon_sym_u16] = ACTIONS(850),
    [anon_sym_i16] = ACTIONS(850),
    [anon_sym_u32] = ACTIONS(850),
    [anon_sym_i32] = ACTIONS(850),
    [anon_sym_u64] = ACTIONS(850),
    [anon_sym_i64] = ACTIONS(850),
    [anon_sym_u128] = ACTIONS(850),
    [anon_sym_i128] = ACTIONS(850),
    [anon_sym_isize] = ACTIONS(850),
    [anon_sym_usize] = ACTIONS(850),
    [anon_sym_f32] = ACTIONS(850),
    [anon_sym_f64] = ACTIONS(850),
    [anon_sym_bool] = ACTIONS(850),
    [anon_sym_str] = ACTIONS(850),
    [anon_sym_char] = ACTIONS(850),
    [anon_sym_as] = ACTIONS(460),
    [anon_sym_const] = ACTIONS(850),
    [anon_sym_default] = ACTIONS(850),
    [anon_sym_union] = ACTIONS(850),
    [anon_sym_POUND] = ACTIONS(852),
    [anon_sym_EQ] = ACTIONS(460),
    [anon_sym_COMMA] = ACTIONS(458),
    [anon_sym_ref] = ACTIONS(850),
    [anon_sym_LT] = ACTIONS(850),
    [anon_sym_GT] = ACTIONS(460),
    [anon_sym_COLON_COLON] = ACTIONS(852),
    [anon_sym__] = ACTIONS(850),
    [anon_sym_AMP] = ACTIONS(850),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(458),
    [sym_mutable_specifier] = ACTIONS(850),
    [anon_sym_DOT_DOT] = ACTIONS(850),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(458),
    [anon_sym_DASH] = ACTIONS(850),
    [anon_sym_AMP_AMP] = ACTIONS(458),
    [anon_sym_PIPE_PIPE] = ACTIONS(458),
    [anon_sym_PIPE] = ACTIONS(460),
    [anon_sym_CARET] = ACTIONS(460),
    [anon_sym_EQ_EQ] = ACTIONS(458),
    [anon_sym_BANG_EQ] = ACTIONS(458),
    [anon_sym_LT_EQ] = ACTIONS(458),
    [anon_sym_GT_EQ] = ACTIONS(458),
    [anon_sym_LT_LT] = ACTIONS(460),
    [anon_sym_GT_GT] = ACTIONS(460),
    [anon_sym_SLASH] = ACTIONS(460),
    [anon_sym_PERCENT] = ACTIONS(460),
    [anon_sym_PLUS_EQ] = ACTIONS(458),
    [anon_sym_DASH_EQ] = ACTIONS(458),
    [anon_sym_STAR_EQ] = ACTIONS(458),
    [anon_sym_SLASH_EQ] = ACTIONS(458),
    [anon_sym_PERCENT_EQ] = ACTIONS(458),
    [anon_sym_AMP_EQ] = ACTIONS(458),
    [anon_sym_PIPE_EQ] = ACTIONS(458),
    [anon_sym_CARET_EQ] = ACTIONS(458),
    [anon_sym_LT_LT_EQ] = ACTIONS(458),
    [anon_sym_GT_GT_EQ] = ACTIONS(458),
    [anon_sym_DOT] = ACTIONS(460),
    [sym_integer_literal] = ACTIONS(852),
    [aux_sym_string_literal_token1] = ACTIONS(852),
    [sym_char_literal] = ACTIONS(852),
    [anon_sym_true] = ACTIONS(850),
    [anon_sym_false] = ACTIONS(850),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(850),
    [sym_super] = ACTIONS(850),
    [sym_crate] = ACTIONS(850),
    [sym_metavariable] = ACTIONS(852),
    [sym_raw_string_literal] = ACTIONS(852),
    [sym_float_literal] = ACTIONS(852),
    [sym_block_comment] = ACTIONS(3),
  },
  [219] = {
    [sym_identifier] = ACTIONS(486),
    [anon_sym_LPAREN] = ACTIONS(484),
    [anon_sym_RBRACE] = ACTIONS(484),
    [anon_sym_LBRACK] = ACTIONS(484),
    [anon_sym_PLUS] = ACTIONS(486),
    [anon_sym_STAR] = ACTIONS(486),
    [anon_sym_QMARK] = ACTIONS(484),
    [anon_sym_u8] = ACTIONS(486),
    [anon_sym_i8] = ACTIONS(486),
    [anon_sym_u16] = ACTIONS(486),
    [anon_sym_i16] = ACTIONS(486),
    [anon_sym_u32] = ACTIONS(486),
    [anon_sym_i32] = ACTIONS(486),
    [anon_sym_u64] = ACTIONS(486),
    [anon_sym_i64] = ACTIONS(486),
    [anon_sym_u128] = ACTIONS(486),
    [anon_sym_i128] = ACTIONS(486),
    [anon_sym_isize] = ACTIONS(486),
    [anon_sym_usize] = ACTIONS(486),
    [anon_sym_f32] = ACTIONS(486),
    [anon_sym_f64] = ACTIONS(486),
    [anon_sym_bool] = ACTIONS(486),
    [anon_sym_str] = ACTIONS(486),
    [anon_sym_char] = ACTIONS(486),
    [anon_sym_as] = ACTIONS(486),
    [anon_sym_const] = ACTIONS(486),
    [anon_sym_default] = ACTIONS(486),
    [anon_sym_union] = ACTIONS(486),
    [anon_sym_POUND] = ACTIONS(484),
    [anon_sym_EQ] = ACTIONS(486),
    [anon_sym_COMMA] = ACTIONS(484),
    [anon_sym_ref] = ACTIONS(486),
    [anon_sym_LT] = ACTIONS(486),
    [anon_sym_GT] = ACTIONS(486),
    [anon_sym_COLON_COLON] = ACTIONS(484),
    [anon_sym__] = ACTIONS(486),
    [anon_sym_AMP] = ACTIONS(486),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(484),
    [sym_mutable_specifier] = ACTIONS(486),
    [anon_sym_DOT_DOT] = ACTIONS(486),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(484),
    [anon_sym_DASH] = ACTIONS(486),
    [anon_sym_AMP_AMP] = ACTIONS(484),
    [anon_sym_PIPE_PIPE] = ACTIONS(484),
    [anon_sym_PIPE] = ACTIONS(486),
    [anon_sym_CARET] = ACTIONS(486),
    [anon_sym_EQ_EQ] = ACTIONS(484),
    [anon_sym_BANG_EQ] = ACTIONS(484),
    [anon_sym_LT_EQ] = ACTIONS(484),
    [anon_sym_GT_EQ] = ACTIONS(484),
    [anon_sym_LT_LT] = ACTIONS(486),
    [anon_sym_GT_GT] = ACTIONS(486),
    [anon_sym_SLASH] = ACTIONS(486),
    [anon_sym_PERCENT] = ACTIONS(486),
    [anon_sym_PLUS_EQ] = ACTIONS(484),
    [anon_sym_DASH_EQ] = ACTIONS(484),
    [anon_sym_STAR_EQ] = ACTIONS(484),
    [anon_sym_SLASH_EQ] = ACTIONS(484),
    [anon_sym_PERCENT_EQ] = ACTIONS(484),
    [anon_sym_AMP_EQ] = ACTIONS(484),
    [anon_sym_PIPE_EQ] = ACTIONS(484),
    [anon_sym_CARET_EQ] = ACTIONS(484),
    [anon_sym_LT_LT_EQ] = ACTIONS(484),
    [anon_sym_GT_GT_EQ] = ACTIONS(484),
    [anon_sym_DOT] = ACTIONS(486),
    [sym_integer_literal] = ACTIONS(484),
    [aux_sym_string_literal_token1] = ACTIONS(484),
    [sym_char_literal] = ACTIONS(484),
    [anon_sym_true] = ACTIONS(486),
    [anon_sym_false] = ACTIONS(486),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(486),
    [sym_super] = ACTIONS(486),
    [sym_crate] = ACTIONS(486),
    [sym_metavariable] = ACTIONS(484),
    [sym_raw_string_literal] = ACTIONS(484),
    [sym_float_literal] = ACTIONS(484),
    [sym_block_comment] = ACTIONS(3),
  },
  [220] = {
    [sym_identifier] = ACTIONS(432),
    [anon_sym_LPAREN] = ACTIONS(430),
    [anon_sym_RBRACE] = ACTIONS(430),
    [anon_sym_LBRACK] = ACTIONS(430),
    [anon_sym_PLUS] = ACTIONS(432),
    [anon_sym_STAR] = ACTIONS(432),
    [anon_sym_QMARK] = ACTIONS(430),
    [anon_sym_u8] = ACTIONS(432),
    [anon_sym_i8] = ACTIONS(432),
    [anon_sym_u16] = ACTIONS(432),
    [anon_sym_i16] = ACTIONS(432),
    [anon_sym_u32] = ACTIONS(432),
    [anon_sym_i32] = ACTIONS(432),
    [anon_sym_u64] = ACTIONS(432),
    [anon_sym_i64] = ACTIONS(432),
    [anon_sym_u128] = ACTIONS(432),
    [anon_sym_i128] = ACTIONS(432),
    [anon_sym_isize] = ACTIONS(432),
    [anon_sym_usize] = ACTIONS(432),
    [anon_sym_f32] = ACTIONS(432),
    [anon_sym_f64] = ACTIONS(432),
    [anon_sym_bool] = ACTIONS(432),
    [anon_sym_str] = ACTIONS(432),
    [anon_sym_char] = ACTIONS(432),
    [anon_sym_as] = ACTIONS(432),
    [anon_sym_const] = ACTIONS(432),
    [anon_sym_default] = ACTIONS(432),
    [anon_sym_union] = ACTIONS(432),
    [anon_sym_POUND] = ACTIONS(430),
    [anon_sym_EQ] = ACTIONS(432),
    [anon_sym_COMMA] = ACTIONS(430),
    [anon_sym_ref] = ACTIONS(432),
    [anon_sym_LT] = ACTIONS(432),
    [anon_sym_GT] = ACTIONS(432),
    [anon_sym_COLON_COLON] = ACTIONS(430),
    [anon_sym__] = ACTIONS(432),
    [anon_sym_AMP] = ACTIONS(432),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(430),
    [sym_mutable_specifier] = ACTIONS(432),
    [anon_sym_DOT_DOT] = ACTIONS(432),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(430),
    [anon_sym_DASH] = ACTIONS(432),
    [anon_sym_AMP_AMP] = ACTIONS(430),
    [anon_sym_PIPE_PIPE] = ACTIONS(430),
    [anon_sym_PIPE] = ACTIONS(432),
    [anon_sym_CARET] = ACTIONS(432),
    [anon_sym_EQ_EQ] = ACTIONS(430),
    [anon_sym_BANG_EQ] = ACTIONS(430),
    [anon_sym_LT_EQ] = ACTIONS(430),
    [anon_sym_GT_EQ] = ACTIONS(430),
    [anon_sym_LT_LT] = ACTIONS(432),
    [anon_sym_GT_GT] = ACTIONS(432),
    [anon_sym_SLASH] = ACTIONS(432),
    [anon_sym_PERCENT] = ACTIONS(432),
    [anon_sym_PLUS_EQ] = ACTIONS(430),
    [anon_sym_DASH_EQ] = ACTIONS(430),
    [anon_sym_STAR_EQ] = ACTIONS(430),
    [anon_sym_SLASH_EQ] = ACTIONS(430),
    [anon_sym_PERCENT_EQ] = ACTIONS(430),
    [anon_sym_AMP_EQ] = ACTIONS(430),
    [anon_sym_PIPE_EQ] = ACTIONS(430),
    [anon_sym_CARET_EQ] = ACTIONS(430),
    [anon_sym_LT_LT_EQ] = ACTIONS(430),
    [anon_sym_GT_GT_EQ] = ACTIONS(430),
    [anon_sym_DOT] = ACTIONS(432),
    [sym_integer_literal] = ACTIONS(430),
    [aux_sym_string_literal_token1] = ACTIONS(430),
    [sym_char_literal] = ACTIONS(430),
    [anon_sym_true] = ACTIONS(432),
    [anon_sym_false] = ACTIONS(432),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(432),
    [sym_super] = ACTIONS(432),
    [sym_crate] = ACTIONS(432),
    [sym_metavariable] = ACTIONS(430),
    [sym_raw_string_literal] = ACTIONS(430),
    [sym_float_literal] = ACTIONS(430),
    [sym_block_comment] = ACTIONS(3),
  },
  [221] = {
    [sym_identifier] = ACTIONS(482),
    [anon_sym_LPAREN] = ACTIONS(480),
    [anon_sym_RBRACE] = ACTIONS(480),
    [anon_sym_LBRACK] = ACTIONS(480),
    [anon_sym_PLUS] = ACTIONS(482),
    [anon_sym_STAR] = ACTIONS(482),
    [anon_sym_QMARK] = ACTIONS(480),
    [anon_sym_u8] = ACTIONS(482),
    [anon_sym_i8] = ACTIONS(482),
    [anon_sym_u16] = ACTIONS(482),
    [anon_sym_i16] = ACTIONS(482),
    [anon_sym_u32] = ACTIONS(482),
    [anon_sym_i32] = ACTIONS(482),
    [anon_sym_u64] = ACTIONS(482),
    [anon_sym_i64] = ACTIONS(482),
    [anon_sym_u128] = ACTIONS(482),
    [anon_sym_i128] = ACTIONS(482),
    [anon_sym_isize] = ACTIONS(482),
    [anon_sym_usize] = ACTIONS(482),
    [anon_sym_f32] = ACTIONS(482),
    [anon_sym_f64] = ACTIONS(482),
    [anon_sym_bool] = ACTIONS(482),
    [anon_sym_str] = ACTIONS(482),
    [anon_sym_char] = ACTIONS(482),
    [anon_sym_as] = ACTIONS(482),
    [anon_sym_const] = ACTIONS(482),
    [anon_sym_default] = ACTIONS(482),
    [anon_sym_union] = ACTIONS(482),
    [anon_sym_POUND] = ACTIONS(480),
    [anon_sym_EQ] = ACTIONS(482),
    [anon_sym_COMMA] = ACTIONS(480),
    [anon_sym_ref] = ACTIONS(482),
    [anon_sym_LT] = ACTIONS(482),
    [anon_sym_GT] = ACTIONS(482),
    [anon_sym_COLON_COLON] = ACTIONS(480),
    [anon_sym__] = ACTIONS(482),
    [anon_sym_AMP] = ACTIONS(482),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(480),
    [sym_mutable_specifier] = ACTIONS(482),
    [anon_sym_DOT_DOT] = ACTIONS(482),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(480),
    [anon_sym_DASH] = ACTIONS(482),
    [anon_sym_AMP_AMP] = ACTIONS(480),
    [anon_sym_PIPE_PIPE] = ACTIONS(480),
    [anon_sym_PIPE] = ACTIONS(482),
    [anon_sym_CARET] = ACTIONS(482),
    [anon_sym_EQ_EQ] = ACTIONS(480),
    [anon_sym_BANG_EQ] = ACTIONS(480),
    [anon_sym_LT_EQ] = ACTIONS(480),
    [anon_sym_GT_EQ] = ACTIONS(480),
    [anon_sym_LT_LT] = ACTIONS(482),
    [anon_sym_GT_GT] = ACTIONS(482),
    [anon_sym_SLASH] = ACTIONS(482),
    [anon_sym_PERCENT] = ACTIONS(482),
    [anon_sym_PLUS_EQ] = ACTIONS(480),
    [anon_sym_DASH_EQ] = ACTIONS(480),
    [anon_sym_STAR_EQ] = ACTIONS(480),
    [anon_sym_SLASH_EQ] = ACTIONS(480),
    [anon_sym_PERCENT_EQ] = ACTIONS(480),
    [anon_sym_AMP_EQ] = ACTIONS(480),
    [anon_sym_PIPE_EQ] = ACTIONS(480),
    [anon_sym_CARET_EQ] = ACTIONS(480),
    [anon_sym_LT_LT_EQ] = ACTIONS(480),
    [anon_sym_GT_GT_EQ] = ACTIONS(480),
    [anon_sym_DOT] = ACTIONS(482),
    [sym_integer_literal] = ACTIONS(480),
    [aux_sym_string_literal_token1] = ACTIONS(480),
    [sym_char_literal] = ACTIONS(480),
    [anon_sym_true] = ACTIONS(482),
    [anon_sym_false] = ACTIONS(482),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(482),
    [sym_super] = ACTIONS(482),
    [sym_crate] = ACTIONS(482),
    [sym_metavariable] = ACTIONS(480),
    [sym_raw_string_literal] = ACTIONS(480),
    [sym_float_literal] = ACTIONS(480),
    [sym_block_comment] = ACTIONS(3),
  },
  [222] = {
    [sym_identifier] = ACTIONS(494),
    [anon_sym_LPAREN] = ACTIONS(492),
    [anon_sym_RBRACE] = ACTIONS(492),
    [anon_sym_LBRACK] = ACTIONS(492),
    [anon_sym_PLUS] = ACTIONS(494),
    [anon_sym_STAR] = ACTIONS(494),
    [anon_sym_QMARK] = ACTIONS(492),
    [anon_sym_u8] = ACTIONS(494),
    [anon_sym_i8] = ACTIONS(494),
    [anon_sym_u16] = ACTIONS(494),
    [anon_sym_i16] = ACTIONS(494),
    [anon_sym_u32] = ACTIONS(494),
    [anon_sym_i32] = ACTIONS(494),
    [anon_sym_u64] = ACTIONS(494),
    [anon_sym_i64] = ACTIONS(494),
    [anon_sym_u128] = ACTIONS(494),
    [anon_sym_i128] = ACTIONS(494),
    [anon_sym_isize] = ACTIONS(494),
    [anon_sym_usize] = ACTIONS(494),
    [anon_sym_f32] = ACTIONS(494),
    [anon_sym_f64] = ACTIONS(494),
    [anon_sym_bool] = ACTIONS(494),
    [anon_sym_str] = ACTIONS(494),
    [anon_sym_char] = ACTIONS(494),
    [anon_sym_as] = ACTIONS(494),
    [anon_sym_const] = ACTIONS(494),
    [anon_sym_default] = ACTIONS(494),
    [anon_sym_union] = ACTIONS(494),
    [anon_sym_POUND] = ACTIONS(492),
    [anon_sym_EQ] = ACTIONS(494),
    [anon_sym_COMMA] = ACTIONS(492),
    [anon_sym_ref] = ACTIONS(494),
    [anon_sym_LT] = ACTIONS(494),
    [anon_sym_GT] = ACTIONS(494),
    [anon_sym_COLON_COLON] = ACTIONS(492),
    [anon_sym__] = ACTIONS(494),
    [anon_sym_AMP] = ACTIONS(494),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(492),
    [sym_mutable_specifier] = ACTIONS(494),
    [anon_sym_DOT_DOT] = ACTIONS(494),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(492),
    [anon_sym_DASH] = ACTIONS(494),
    [anon_sym_AMP_AMP] = ACTIONS(492),
    [anon_sym_PIPE_PIPE] = ACTIONS(492),
    [anon_sym_PIPE] = ACTIONS(494),
    [anon_sym_CARET] = ACTIONS(494),
    [anon_sym_EQ_EQ] = ACTIONS(492),
    [anon_sym_BANG_EQ] = ACTIONS(492),
    [anon_sym_LT_EQ] = ACTIONS(492),
    [anon_sym_GT_EQ] = ACTIONS(492),
    [anon_sym_LT_LT] = ACTIONS(494),
    [anon_sym_GT_GT] = ACTIONS(494),
    [anon_sym_SLASH] = ACTIONS(494),
    [anon_sym_PERCENT] = ACTIONS(494),
    [anon_sym_PLUS_EQ] = ACTIONS(492),
    [anon_sym_DASH_EQ] = ACTIONS(492),
    [anon_sym_STAR_EQ] = ACTIONS(492),
    [anon_sym_SLASH_EQ] = ACTIONS(492),
    [anon_sym_PERCENT_EQ] = ACTIONS(492),
    [anon_sym_AMP_EQ] = ACTIONS(492),
    [anon_sym_PIPE_EQ] = ACTIONS(492),
    [anon_sym_CARET_EQ] = ACTIONS(492),
    [anon_sym_LT_LT_EQ] = ACTIONS(492),
    [anon_sym_GT_GT_EQ] = ACTIONS(492),
    [anon_sym_DOT] = ACTIONS(494),
    [sym_integer_literal] = ACTIONS(492),
    [aux_sym_string_literal_token1] = ACTIONS(492),
    [sym_char_literal] = ACTIONS(492),
    [anon_sym_true] = ACTIONS(494),
    [anon_sym_false] = ACTIONS(494),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(494),
    [sym_super] = ACTIONS(494),
    [sym_crate] = ACTIONS(494),
    [sym_metavariable] = ACTIONS(492),
    [sym_raw_string_literal] = ACTIONS(492),
    [sym_float_literal] = ACTIONS(492),
    [sym_block_comment] = ACTIONS(3),
  },
  [223] = {
    [sym_identifier] = ACTIONS(490),
    [anon_sym_LPAREN] = ACTIONS(488),
    [anon_sym_RBRACE] = ACTIONS(488),
    [anon_sym_LBRACK] = ACTIONS(488),
    [anon_sym_PLUS] = ACTIONS(490),
    [anon_sym_STAR] = ACTIONS(490),
    [anon_sym_QMARK] = ACTIONS(488),
    [anon_sym_u8] = ACTIONS(490),
    [anon_sym_i8] = ACTIONS(490),
    [anon_sym_u16] = ACTIONS(490),
    [anon_sym_i16] = ACTIONS(490),
    [anon_sym_u32] = ACTIONS(490),
    [anon_sym_i32] = ACTIONS(490),
    [anon_sym_u64] = ACTIONS(490),
    [anon_sym_i64] = ACTIONS(490),
    [anon_sym_u128] = ACTIONS(490),
    [anon_sym_i128] = ACTIONS(490),
    [anon_sym_isize] = ACTIONS(490),
    [anon_sym_usize] = ACTIONS(490),
    [anon_sym_f32] = ACTIONS(490),
    [anon_sym_f64] = ACTIONS(490),
    [anon_sym_bool] = ACTIONS(490),
    [anon_sym_str] = ACTIONS(490),
    [anon_sym_char] = ACTIONS(490),
    [anon_sym_as] = ACTIONS(490),
    [anon_sym_const] = ACTIONS(490),
    [anon_sym_default] = ACTIONS(490),
    [anon_sym_union] = ACTIONS(490),
    [anon_sym_POUND] = ACTIONS(488),
    [anon_sym_EQ] = ACTIONS(490),
    [anon_sym_COMMA] = ACTIONS(488),
    [anon_sym_ref] = ACTIONS(490),
    [anon_sym_LT] = ACTIONS(490),
    [anon_sym_GT] = ACTIONS(490),
    [anon_sym_COLON_COLON] = ACTIONS(488),
    [anon_sym__] = ACTIONS(490),
    [anon_sym_AMP] = ACTIONS(490),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(488),
    [sym_mutable_specifier] = ACTIONS(490),
    [anon_sym_DOT_DOT] = ACTIONS(490),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(488),
    [anon_sym_DASH] = ACTIONS(490),
    [anon_sym_AMP_AMP] = ACTIONS(488),
    [anon_sym_PIPE_PIPE] = ACTIONS(488),
    [anon_sym_PIPE] = ACTIONS(490),
    [anon_sym_CARET] = ACTIONS(490),
    [anon_sym_EQ_EQ] = ACTIONS(488),
    [anon_sym_BANG_EQ] = ACTIONS(488),
    [anon_sym_LT_EQ] = ACTIONS(488),
    [anon_sym_GT_EQ] = ACTIONS(488),
    [anon_sym_LT_LT] = ACTIONS(490),
    [anon_sym_GT_GT] = ACTIONS(490),
    [anon_sym_SLASH] = ACTIONS(490),
    [anon_sym_PERCENT] = ACTIONS(490),
    [anon_sym_PLUS_EQ] = ACTIONS(488),
    [anon_sym_DASH_EQ] = ACTIONS(488),
    [anon_sym_STAR_EQ] = ACTIONS(488),
    [anon_sym_SLASH_EQ] = ACTIONS(488),
    [anon_sym_PERCENT_EQ] = ACTIONS(488),
    [anon_sym_AMP_EQ] = ACTIONS(488),
    [anon_sym_PIPE_EQ] = ACTIONS(488),
    [anon_sym_CARET_EQ] = ACTIONS(488),
    [anon_sym_LT_LT_EQ] = ACTIONS(488),
    [anon_sym_GT_GT_EQ] = ACTIONS(488),
    [anon_sym_DOT] = ACTIONS(490),
    [sym_integer_literal] = ACTIONS(488),
    [aux_sym_string_literal_token1] = ACTIONS(488),
    [sym_char_literal] = ACTIONS(488),
    [anon_sym_true] = ACTIONS(490),
    [anon_sym_false] = ACTIONS(490),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(490),
    [sym_super] = ACTIONS(490),
    [sym_crate] = ACTIONS(490),
    [sym_metavariable] = ACTIONS(488),
    [sym_raw_string_literal] = ACTIONS(488),
    [sym_float_literal] = ACTIONS(488),
    [sym_block_comment] = ACTIONS(3),
  },
  [224] = {
    [sym_identifier] = ACTIONS(464),
    [anon_sym_LPAREN] = ACTIONS(462),
    [anon_sym_RBRACE] = ACTIONS(462),
    [anon_sym_LBRACK] = ACTIONS(462),
    [anon_sym_PLUS] = ACTIONS(464),
    [anon_sym_STAR] = ACTIONS(464),
    [anon_sym_QMARK] = ACTIONS(462),
    [anon_sym_u8] = ACTIONS(464),
    [anon_sym_i8] = ACTIONS(464),
    [anon_sym_u16] = ACTIONS(464),
    [anon_sym_i16] = ACTIONS(464),
    [anon_sym_u32] = ACTIONS(464),
    [anon_sym_i32] = ACTIONS(464),
    [anon_sym_u64] = ACTIONS(464),
    [anon_sym_i64] = ACTIONS(464),
    [anon_sym_u128] = ACTIONS(464),
    [anon_sym_i128] = ACTIONS(464),
    [anon_sym_isize] = ACTIONS(464),
    [anon_sym_usize] = ACTIONS(464),
    [anon_sym_f32] = ACTIONS(464),
    [anon_sym_f64] = ACTIONS(464),
    [anon_sym_bool] = ACTIONS(464),
    [anon_sym_str] = ACTIONS(464),
    [anon_sym_char] = ACTIONS(464),
    [anon_sym_as] = ACTIONS(464),
    [anon_sym_const] = ACTIONS(464),
    [anon_sym_default] = ACTIONS(464),
    [anon_sym_union] = ACTIONS(464),
    [anon_sym_POUND] = ACTIONS(462),
    [anon_sym_EQ] = ACTIONS(464),
    [anon_sym_COMMA] = ACTIONS(462),
    [anon_sym_ref] = ACTIONS(464),
    [anon_sym_LT] = ACTIONS(464),
    [anon_sym_GT] = ACTIONS(464),
    [anon_sym_COLON_COLON] = ACTIONS(462),
    [anon_sym__] = ACTIONS(464),
    [anon_sym_AMP] = ACTIONS(464),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(462),
    [sym_mutable_specifier] = ACTIONS(464),
    [anon_sym_DOT_DOT] = ACTIONS(464),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(462),
    [anon_sym_DASH] = ACTIONS(464),
    [anon_sym_AMP_AMP] = ACTIONS(462),
    [anon_sym_PIPE_PIPE] = ACTIONS(462),
    [anon_sym_PIPE] = ACTIONS(464),
    [anon_sym_CARET] = ACTIONS(464),
    [anon_sym_EQ_EQ] = ACTIONS(462),
    [anon_sym_BANG_EQ] = ACTIONS(462),
    [anon_sym_LT_EQ] = ACTIONS(462),
    [anon_sym_GT_EQ] = ACTIONS(462),
    [anon_sym_LT_LT] = ACTIONS(464),
    [anon_sym_GT_GT] = ACTIONS(464),
    [anon_sym_SLASH] = ACTIONS(464),
    [anon_sym_PERCENT] = ACTIONS(464),
    [anon_sym_PLUS_EQ] = ACTIONS(462),
    [anon_sym_DASH_EQ] = ACTIONS(462),
    [anon_sym_STAR_EQ] = ACTIONS(462),
    [anon_sym_SLASH_EQ] = ACTIONS(462),
    [anon_sym_PERCENT_EQ] = ACTIONS(462),
    [anon_sym_AMP_EQ] = ACTIONS(462),
    [anon_sym_PIPE_EQ] = ACTIONS(462),
    [anon_sym_CARET_EQ] = ACTIONS(462),
    [anon_sym_LT_LT_EQ] = ACTIONS(462),
    [anon_sym_GT_GT_EQ] = ACTIONS(462),
    [anon_sym_DOT] = ACTIONS(464),
    [sym_integer_literal] = ACTIONS(462),
    [aux_sym_string_literal_token1] = ACTIONS(462),
    [sym_char_literal] = ACTIONS(462),
    [anon_sym_true] = ACTIONS(464),
    [anon_sym_false] = ACTIONS(464),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(464),
    [sym_super] = ACTIONS(464),
    [sym_crate] = ACTIONS(464),
    [sym_metavariable] = ACTIONS(462),
    [sym_raw_string_literal] = ACTIONS(462),
    [sym_float_literal] = ACTIONS(462),
    [sym_block_comment] = ACTIONS(3),
  },
  [225] = {
    [sym_identifier] = ACTIONS(478),
    [anon_sym_LPAREN] = ACTIONS(476),
    [anon_sym_RBRACE] = ACTIONS(476),
    [anon_sym_LBRACK] = ACTIONS(476),
    [anon_sym_PLUS] = ACTIONS(478),
    [anon_sym_STAR] = ACTIONS(478),
    [anon_sym_QMARK] = ACTIONS(476),
    [anon_sym_u8] = ACTIONS(478),
    [anon_sym_i8] = ACTIONS(478),
    [anon_sym_u16] = ACTIONS(478),
    [anon_sym_i16] = ACTIONS(478),
    [anon_sym_u32] = ACTIONS(478),
    [anon_sym_i32] = ACTIONS(478),
    [anon_sym_u64] = ACTIONS(478),
    [anon_sym_i64] = ACTIONS(478),
    [anon_sym_u128] = ACTIONS(478),
    [anon_sym_i128] = ACTIONS(478),
    [anon_sym_isize] = ACTIONS(478),
    [anon_sym_usize] = ACTIONS(478),
    [anon_sym_f32] = ACTIONS(478),
    [anon_sym_f64] = ACTIONS(478),
    [anon_sym_bool] = ACTIONS(478),
    [anon_sym_str] = ACTIONS(478),
    [anon_sym_char] = ACTIONS(478),
    [anon_sym_as] = ACTIONS(478),
    [anon_sym_const] = ACTIONS(478),
    [anon_sym_default] = ACTIONS(478),
    [anon_sym_union] = ACTIONS(478),
    [anon_sym_POUND] = ACTIONS(476),
    [anon_sym_EQ] = ACTIONS(478),
    [anon_sym_COMMA] = ACTIONS(476),
    [anon_sym_ref] = ACTIONS(478),
    [anon_sym_LT] = ACTIONS(478),
    [anon_sym_GT] = ACTIONS(478),
    [anon_sym_COLON_COLON] = ACTIONS(476),
    [anon_sym__] = ACTIONS(478),
    [anon_sym_AMP] = ACTIONS(478),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(476),
    [sym_mutable_specifier] = ACTIONS(478),
    [anon_sym_DOT_DOT] = ACTIONS(478),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(476),
    [anon_sym_DASH] = ACTIONS(478),
    [anon_sym_AMP_AMP] = ACTIONS(476),
    [anon_sym_PIPE_PIPE] = ACTIONS(476),
    [anon_sym_PIPE] = ACTIONS(478),
    [anon_sym_CARET] = ACTIONS(478),
    [anon_sym_EQ_EQ] = ACTIONS(476),
    [anon_sym_BANG_EQ] = ACTIONS(476),
    [anon_sym_LT_EQ] = ACTIONS(476),
    [anon_sym_GT_EQ] = ACTIONS(476),
    [anon_sym_LT_LT] = ACTIONS(478),
    [anon_sym_GT_GT] = ACTIONS(478),
    [anon_sym_SLASH] = ACTIONS(478),
    [anon_sym_PERCENT] = ACTIONS(478),
    [anon_sym_PLUS_EQ] = ACTIONS(476),
    [anon_sym_DASH_EQ] = ACTIONS(476),
    [anon_sym_STAR_EQ] = ACTIONS(476),
    [anon_sym_SLASH_EQ] = ACTIONS(476),
    [anon_sym_PERCENT_EQ] = ACTIONS(476),
    [anon_sym_AMP_EQ] = ACTIONS(476),
    [anon_sym_PIPE_EQ] = ACTIONS(476),
    [anon_sym_CARET_EQ] = ACTIONS(476),
    [anon_sym_LT_LT_EQ] = ACTIONS(476),
    [anon_sym_GT_GT_EQ] = ACTIONS(476),
    [anon_sym_DOT] = ACTIONS(478),
    [sym_integer_literal] = ACTIONS(476),
    [aux_sym_string_literal_token1] = ACTIONS(476),
    [sym_char_literal] = ACTIONS(476),
    [anon_sym_true] = ACTIONS(478),
    [anon_sym_false] = ACTIONS(478),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(478),
    [sym_super] = ACTIONS(478),
    [sym_crate] = ACTIONS(478),
    [sym_metavariable] = ACTIONS(476),
    [sym_raw_string_literal] = ACTIONS(476),
    [sym_float_literal] = ACTIONS(476),
    [sym_block_comment] = ACTIONS(3),
  },
  [226] = {
    [sym_identifier] = ACTIONS(436),
    [anon_sym_LPAREN] = ACTIONS(434),
    [anon_sym_RBRACE] = ACTIONS(434),
    [anon_sym_LBRACK] = ACTIONS(434),
    [anon_sym_PLUS] = ACTIONS(436),
    [anon_sym_STAR] = ACTIONS(436),
    [anon_sym_QMARK] = ACTIONS(434),
    [anon_sym_u8] = ACTIONS(436),
    [anon_sym_i8] = ACTIONS(436),
    [anon_sym_u16] = ACTIONS(436),
    [anon_sym_i16] = ACTIONS(436),
    [anon_sym_u32] = ACTIONS(436),
    [anon_sym_i32] = ACTIONS(436),
    [anon_sym_u64] = ACTIONS(436),
    [anon_sym_i64] = ACTIONS(436),
    [anon_sym_u128] = ACTIONS(436),
    [anon_sym_i128] = ACTIONS(436),
    [anon_sym_isize] = ACTIONS(436),
    [anon_sym_usize] = ACTIONS(436),
    [anon_sym_f32] = ACTIONS(436),
    [anon_sym_f64] = ACTIONS(436),
    [anon_sym_bool] = ACTIONS(436),
    [anon_sym_str] = ACTIONS(436),
    [anon_sym_char] = ACTIONS(436),
    [anon_sym_as] = ACTIONS(436),
    [anon_sym_const] = ACTIONS(436),
    [anon_sym_default] = ACTIONS(436),
    [anon_sym_union] = ACTIONS(436),
    [anon_sym_POUND] = ACTIONS(434),
    [anon_sym_EQ] = ACTIONS(436),
    [anon_sym_COMMA] = ACTIONS(434),
    [anon_sym_ref] = ACTIONS(436),
    [anon_sym_LT] = ACTIONS(436),
    [anon_sym_GT] = ACTIONS(436),
    [anon_sym_COLON_COLON] = ACTIONS(434),
    [anon_sym__] = ACTIONS(436),
    [anon_sym_AMP] = ACTIONS(436),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(434),
    [sym_mutable_specifier] = ACTIONS(436),
    [anon_sym_DOT_DOT] = ACTIONS(436),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(434),
    [anon_sym_DASH] = ACTIONS(436),
    [anon_sym_AMP_AMP] = ACTIONS(434),
    [anon_sym_PIPE_PIPE] = ACTIONS(434),
    [anon_sym_PIPE] = ACTIONS(436),
    [anon_sym_CARET] = ACTIONS(436),
    [anon_sym_EQ_EQ] = ACTIONS(434),
    [anon_sym_BANG_EQ] = ACTIONS(434),
    [anon_sym_LT_EQ] = ACTIONS(434),
    [anon_sym_GT_EQ] = ACTIONS(434),
    [anon_sym_LT_LT] = ACTIONS(436),
    [anon_sym_GT_GT] = ACTIONS(436),
    [anon_sym_SLASH] = ACTIONS(436),
    [anon_sym_PERCENT] = ACTIONS(436),
    [anon_sym_PLUS_EQ] = ACTIONS(434),
    [anon_sym_DASH_EQ] = ACTIONS(434),
    [anon_sym_STAR_EQ] = ACTIONS(434),
    [anon_sym_SLASH_EQ] = ACTIONS(434),
    [anon_sym_PERCENT_EQ] = ACTIONS(434),
    [anon_sym_AMP_EQ] = ACTIONS(434),
    [anon_sym_PIPE_EQ] = ACTIONS(434),
    [anon_sym_CARET_EQ] = ACTIONS(434),
    [anon_sym_LT_LT_EQ] = ACTIONS(434),
    [anon_sym_GT_GT_EQ] = ACTIONS(434),
    [anon_sym_DOT] = ACTIONS(436),
    [sym_integer_literal] = ACTIONS(434),
    [aux_sym_string_literal_token1] = ACTIONS(434),
    [sym_char_literal] = ACTIONS(434),
    [anon_sym_true] = ACTIONS(436),
    [anon_sym_false] = ACTIONS(436),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(436),
    [sym_super] = ACTIONS(436),
    [sym_crate] = ACTIONS(436),
    [sym_metavariable] = ACTIONS(434),
    [sym_raw_string_literal] = ACTIONS(434),
    [sym_float_literal] = ACTIONS(434),
    [sym_block_comment] = ACTIONS(3),
  },
  [227] = {
    [sym_identifier] = ACTIONS(452),
    [anon_sym_LPAREN] = ACTIONS(450),
    [anon_sym_RBRACE] = ACTIONS(450),
    [anon_sym_LBRACK] = ACTIONS(450),
    [anon_sym_PLUS] = ACTIONS(452),
    [anon_sym_STAR] = ACTIONS(452),
    [anon_sym_QMARK] = ACTIONS(450),
    [anon_sym_u8] = ACTIONS(452),
    [anon_sym_i8] = ACTIONS(452),
    [anon_sym_u16] = ACTIONS(452),
    [anon_sym_i16] = ACTIONS(452),
    [anon_sym_u32] = ACTIONS(452),
    [anon_sym_i32] = ACTIONS(452),
    [anon_sym_u64] = ACTIONS(452),
    [anon_sym_i64] = ACTIONS(452),
    [anon_sym_u128] = ACTIONS(452),
    [anon_sym_i128] = ACTIONS(452),
    [anon_sym_isize] = ACTIONS(452),
    [anon_sym_usize] = ACTIONS(452),
    [anon_sym_f32] = ACTIONS(452),
    [anon_sym_f64] = ACTIONS(452),
    [anon_sym_bool] = ACTIONS(452),
    [anon_sym_str] = ACTIONS(452),
    [anon_sym_char] = ACTIONS(452),
    [anon_sym_as] = ACTIONS(452),
    [anon_sym_const] = ACTIONS(452),
    [anon_sym_default] = ACTIONS(452),
    [anon_sym_union] = ACTIONS(452),
    [anon_sym_POUND] = ACTIONS(450),
    [anon_sym_EQ] = ACTIONS(452),
    [anon_sym_COMMA] = ACTIONS(450),
    [anon_sym_ref] = ACTIONS(452),
    [anon_sym_LT] = ACTIONS(452),
    [anon_sym_GT] = ACTIONS(452),
    [anon_sym_COLON_COLON] = ACTIONS(450),
    [anon_sym__] = ACTIONS(452),
    [anon_sym_AMP] = ACTIONS(452),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(450),
    [sym_mutable_specifier] = ACTIONS(452),
    [anon_sym_DOT_DOT] = ACTIONS(452),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(450),
    [anon_sym_DASH] = ACTIONS(452),
    [anon_sym_AMP_AMP] = ACTIONS(450),
    [anon_sym_PIPE_PIPE] = ACTIONS(450),
    [anon_sym_PIPE] = ACTIONS(452),
    [anon_sym_CARET] = ACTIONS(452),
    [anon_sym_EQ_EQ] = ACTIONS(450),
    [anon_sym_BANG_EQ] = ACTIONS(450),
    [anon_sym_LT_EQ] = ACTIONS(450),
    [anon_sym_GT_EQ] = ACTIONS(450),
    [anon_sym_LT_LT] = ACTIONS(452),
    [anon_sym_GT_GT] = ACTIONS(452),
    [anon_sym_SLASH] = ACTIONS(452),
    [anon_sym_PERCENT] = ACTIONS(452),
    [anon_sym_PLUS_EQ] = ACTIONS(450),
    [anon_sym_DASH_EQ] = ACTIONS(450),
    [anon_sym_STAR_EQ] = ACTIONS(450),
    [anon_sym_SLASH_EQ] = ACTIONS(450),
    [anon_sym_PERCENT_EQ] = ACTIONS(450),
    [anon_sym_AMP_EQ] = ACTIONS(450),
    [anon_sym_PIPE_EQ] = ACTIONS(450),
    [anon_sym_CARET_EQ] = ACTIONS(450),
    [anon_sym_LT_LT_EQ] = ACTIONS(450),
    [anon_sym_GT_GT_EQ] = ACTIONS(450),
    [anon_sym_DOT] = ACTIONS(452),
    [sym_integer_literal] = ACTIONS(450),
    [aux_sym_string_literal_token1] = ACTIONS(450),
    [sym_char_literal] = ACTIONS(450),
    [anon_sym_true] = ACTIONS(452),
    [anon_sym_false] = ACTIONS(452),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(452),
    [sym_super] = ACTIONS(452),
    [sym_crate] = ACTIONS(452),
    [sym_metavariable] = ACTIONS(450),
    [sym_raw_string_literal] = ACTIONS(450),
    [sym_float_literal] = ACTIONS(450),
    [sym_block_comment] = ACTIONS(3),
  },
  [228] = {
    [sym_identifier] = ACTIONS(428),
    [anon_sym_LPAREN] = ACTIONS(426),
    [anon_sym_RBRACE] = ACTIONS(426),
    [anon_sym_LBRACK] = ACTIONS(426),
    [anon_sym_PLUS] = ACTIONS(428),
    [anon_sym_STAR] = ACTIONS(428),
    [anon_sym_QMARK] = ACTIONS(426),
    [anon_sym_u8] = ACTIONS(428),
    [anon_sym_i8] = ACTIONS(428),
    [anon_sym_u16] = ACTIONS(428),
    [anon_sym_i16] = ACTIONS(428),
    [anon_sym_u32] = ACTIONS(428),
    [anon_sym_i32] = ACTIONS(428),
    [anon_sym_u64] = ACTIONS(428),
    [anon_sym_i64] = ACTIONS(428),
    [anon_sym_u128] = ACTIONS(428),
    [anon_sym_i128] = ACTIONS(428),
    [anon_sym_isize] = ACTIONS(428),
    [anon_sym_usize] = ACTIONS(428),
    [anon_sym_f32] = ACTIONS(428),
    [anon_sym_f64] = ACTIONS(428),
    [anon_sym_bool] = ACTIONS(428),
    [anon_sym_str] = ACTIONS(428),
    [anon_sym_char] = ACTIONS(428),
    [anon_sym_as] = ACTIONS(428),
    [anon_sym_const] = ACTIONS(428),
    [anon_sym_default] = ACTIONS(428),
    [anon_sym_union] = ACTIONS(428),
    [anon_sym_POUND] = ACTIONS(426),
    [anon_sym_EQ] = ACTIONS(428),
    [anon_sym_COMMA] = ACTIONS(426),
    [anon_sym_ref] = ACTIONS(428),
    [anon_sym_LT] = ACTIONS(428),
    [anon_sym_GT] = ACTIONS(428),
    [anon_sym_COLON_COLON] = ACTIONS(426),
    [anon_sym__] = ACTIONS(428),
    [anon_sym_AMP] = ACTIONS(428),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(426),
    [sym_mutable_specifier] = ACTIONS(428),
    [anon_sym_DOT_DOT] = ACTIONS(428),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(426),
    [anon_sym_DASH] = ACTIONS(428),
    [anon_sym_AMP_AMP] = ACTIONS(426),
    [anon_sym_PIPE_PIPE] = ACTIONS(426),
    [anon_sym_PIPE] = ACTIONS(428),
    [anon_sym_CARET] = ACTIONS(428),
    [anon_sym_EQ_EQ] = ACTIONS(426),
    [anon_sym_BANG_EQ] = ACTIONS(426),
    [anon_sym_LT_EQ] = ACTIONS(426),
    [anon_sym_GT_EQ] = ACTIONS(426),
    [anon_sym_LT_LT] = ACTIONS(428),
    [anon_sym_GT_GT] = ACTIONS(428),
    [anon_sym_SLASH] = ACTIONS(428),
    [anon_sym_PERCENT] = ACTIONS(428),
    [anon_sym_PLUS_EQ] = ACTIONS(426),
    [anon_sym_DASH_EQ] = ACTIONS(426),
    [anon_sym_STAR_EQ] = ACTIONS(426),
    [anon_sym_SLASH_EQ] = ACTIONS(426),
    [anon_sym_PERCENT_EQ] = ACTIONS(426),
    [anon_sym_AMP_EQ] = ACTIONS(426),
    [anon_sym_PIPE_EQ] = ACTIONS(426),
    [anon_sym_CARET_EQ] = ACTIONS(426),
    [anon_sym_LT_LT_EQ] = ACTIONS(426),
    [anon_sym_GT_GT_EQ] = ACTIONS(426),
    [anon_sym_DOT] = ACTIONS(428),
    [sym_integer_literal] = ACTIONS(426),
    [aux_sym_string_literal_token1] = ACTIONS(426),
    [sym_char_literal] = ACTIONS(426),
    [anon_sym_true] = ACTIONS(428),
    [anon_sym_false] = ACTIONS(428),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(428),
    [sym_super] = ACTIONS(428),
    [sym_crate] = ACTIONS(428),
    [sym_metavariable] = ACTIONS(426),
    [sym_raw_string_literal] = ACTIONS(426),
    [sym_float_literal] = ACTIONS(426),
    [sym_block_comment] = ACTIONS(3),
  },
  [229] = {
    [sym_identifier] = ACTIONS(444),
    [anon_sym_LPAREN] = ACTIONS(442),
    [anon_sym_RBRACE] = ACTIONS(442),
    [anon_sym_LBRACK] = ACTIONS(442),
    [anon_sym_PLUS] = ACTIONS(444),
    [anon_sym_STAR] = ACTIONS(444),
    [anon_sym_QMARK] = ACTIONS(442),
    [anon_sym_u8] = ACTIONS(444),
    [anon_sym_i8] = ACTIONS(444),
    [anon_sym_u16] = ACTIONS(444),
    [anon_sym_i16] = ACTIONS(444),
    [anon_sym_u32] = ACTIONS(444),
    [anon_sym_i32] = ACTIONS(444),
    [anon_sym_u64] = ACTIONS(444),
    [anon_sym_i64] = ACTIONS(444),
    [anon_sym_u128] = ACTIONS(444),
    [anon_sym_i128] = ACTIONS(444),
    [anon_sym_isize] = ACTIONS(444),
    [anon_sym_usize] = ACTIONS(444),
    [anon_sym_f32] = ACTIONS(444),
    [anon_sym_f64] = ACTIONS(444),
    [anon_sym_bool] = ACTIONS(444),
    [anon_sym_str] = ACTIONS(444),
    [anon_sym_char] = ACTIONS(444),
    [anon_sym_as] = ACTIONS(444),
    [anon_sym_const] = ACTIONS(444),
    [anon_sym_default] = ACTIONS(444),
    [anon_sym_union] = ACTIONS(444),
    [anon_sym_POUND] = ACTIONS(442),
    [anon_sym_EQ] = ACTIONS(444),
    [anon_sym_COMMA] = ACTIONS(442),
    [anon_sym_ref] = ACTIONS(444),
    [anon_sym_LT] = ACTIONS(444),
    [anon_sym_GT] = ACTIONS(444),
    [anon_sym_COLON_COLON] = ACTIONS(442),
    [anon_sym__] = ACTIONS(444),
    [anon_sym_AMP] = ACTIONS(444),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(442),
    [sym_mutable_specifier] = ACTIONS(444),
    [anon_sym_DOT_DOT] = ACTIONS(444),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(442),
    [anon_sym_DASH] = ACTIONS(444),
    [anon_sym_AMP_AMP] = ACTIONS(442),
    [anon_sym_PIPE_PIPE] = ACTIONS(442),
    [anon_sym_PIPE] = ACTIONS(444),
    [anon_sym_CARET] = ACTIONS(444),
    [anon_sym_EQ_EQ] = ACTIONS(442),
    [anon_sym_BANG_EQ] = ACTIONS(442),
    [anon_sym_LT_EQ] = ACTIONS(442),
    [anon_sym_GT_EQ] = ACTIONS(442),
    [anon_sym_LT_LT] = ACTIONS(444),
    [anon_sym_GT_GT] = ACTIONS(444),
    [anon_sym_SLASH] = ACTIONS(444),
    [anon_sym_PERCENT] = ACTIONS(444),
    [anon_sym_PLUS_EQ] = ACTIONS(442),
    [anon_sym_DASH_EQ] = ACTIONS(442),
    [anon_sym_STAR_EQ] = ACTIONS(442),
    [anon_sym_SLASH_EQ] = ACTIONS(442),
    [anon_sym_PERCENT_EQ] = ACTIONS(442),
    [anon_sym_AMP_EQ] = ACTIONS(442),
    [anon_sym_PIPE_EQ] = ACTIONS(442),
    [anon_sym_CARET_EQ] = ACTIONS(442),
    [anon_sym_LT_LT_EQ] = ACTIONS(442),
    [anon_sym_GT_GT_EQ] = ACTIONS(442),
    [anon_sym_DOT] = ACTIONS(444),
    [sym_integer_literal] = ACTIONS(442),
    [aux_sym_string_literal_token1] = ACTIONS(442),
    [sym_char_literal] = ACTIONS(442),
    [anon_sym_true] = ACTIONS(444),
    [anon_sym_false] = ACTIONS(444),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(444),
    [sym_super] = ACTIONS(444),
    [sym_crate] = ACTIONS(444),
    [sym_metavariable] = ACTIONS(442),
    [sym_raw_string_literal] = ACTIONS(442),
    [sym_float_literal] = ACTIONS(442),
    [sym_block_comment] = ACTIONS(3),
  },
  [230] = {
    [sym_function_modifiers] = STATE(2472),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(1935),
    [sym_bracketed_type] = STATE(2378),
    [sym_lifetime] = STATE(1934),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2379),
    [sym_bounded_type] = STATE(1390),
    [sym_type_binding] = STATE(2264),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1390),
    [sym_scoped_identifier] = STATE(2157),
    [sym_scoped_type_identifier] = STATE(1331),
    [sym_block] = STATE(2264),
    [sym__literal] = STATE(2264),
    [sym_string_literal] = STATE(2313),
    [sym_boolean_literal] = STATE(2313),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(854),
    [anon_sym_LPAREN] = ACTIONS(856),
    [anon_sym_LBRACE] = ACTIONS(858),
    [anon_sym_LBRACK] = ACTIONS(860),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(862),
    [anon_sym_i8] = ACTIONS(862),
    [anon_sym_u16] = ACTIONS(862),
    [anon_sym_i16] = ACTIONS(862),
    [anon_sym_u32] = ACTIONS(862),
    [anon_sym_i32] = ACTIONS(862),
    [anon_sym_u64] = ACTIONS(862),
    [anon_sym_i64] = ACTIONS(862),
    [anon_sym_u128] = ACTIONS(862),
    [anon_sym_i128] = ACTIONS(862),
    [anon_sym_isize] = ACTIONS(862),
    [anon_sym_usize] = ACTIONS(862),
    [anon_sym_f32] = ACTIONS(862),
    [anon_sym_f64] = ACTIONS(862),
    [anon_sym_bool] = ACTIONS(862),
    [anon_sym_str] = ACTIONS(862),
    [anon_sym_char] = ACTIONS(862),
    [anon_sym_SQUOTE] = ACTIONS(662),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(664),
    [anon_sym_default] = ACTIONS(864),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(866),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_GT] = ACTIONS(868),
    [anon_sym_COLON_COLON] = ACTIONS(870),
    [anon_sym_AMP] = ACTIONS(872),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_integer_literal] = ACTIONS(874),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(874),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(876),
    [sym_super] = ACTIONS(876),
    [sym_crate] = ACTIONS(876),
    [sym_metavariable] = ACTIONS(878),
    [sym_raw_string_literal] = ACTIONS(874),
    [sym_float_literal] = ACTIONS(874),
    [sym_block_comment] = ACTIONS(3),
  },
  [231] = {
    [sym_function_modifiers] = STATE(2472),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(1935),
    [sym_bracketed_type] = STATE(2378),
    [sym_lifetime] = STATE(1934),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2379),
    [sym_bounded_type] = STATE(1390),
    [sym_type_binding] = STATE(2264),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1390),
    [sym_scoped_identifier] = STATE(2157),
    [sym_scoped_type_identifier] = STATE(1331),
    [sym_block] = STATE(2264),
    [sym__literal] = STATE(2264),
    [sym_string_literal] = STATE(2313),
    [sym_boolean_literal] = STATE(2313),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(854),
    [anon_sym_LPAREN] = ACTIONS(856),
    [anon_sym_LBRACE] = ACTIONS(858),
    [anon_sym_LBRACK] = ACTIONS(860),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(862),
    [anon_sym_i8] = ACTIONS(862),
    [anon_sym_u16] = ACTIONS(862),
    [anon_sym_i16] = ACTIONS(862),
    [anon_sym_u32] = ACTIONS(862),
    [anon_sym_i32] = ACTIONS(862),
    [anon_sym_u64] = ACTIONS(862),
    [anon_sym_i64] = ACTIONS(862),
    [anon_sym_u128] = ACTIONS(862),
    [anon_sym_i128] = ACTIONS(862),
    [anon_sym_isize] = ACTIONS(862),
    [anon_sym_usize] = ACTIONS(862),
    [anon_sym_f32] = ACTIONS(862),
    [anon_sym_f64] = ACTIONS(862),
    [anon_sym_bool] = ACTIONS(862),
    [anon_sym_str] = ACTIONS(862),
    [anon_sym_char] = ACTIONS(862),
    [anon_sym_SQUOTE] = ACTIONS(662),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(664),
    [anon_sym_default] = ACTIONS(864),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(866),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_GT] = ACTIONS(880),
    [anon_sym_COLON_COLON] = ACTIONS(870),
    [anon_sym_AMP] = ACTIONS(872),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_integer_literal] = ACTIONS(874),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(874),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(876),
    [sym_super] = ACTIONS(876),
    [sym_crate] = ACTIONS(876),
    [sym_metavariable] = ACTIONS(878),
    [sym_raw_string_literal] = ACTIONS(874),
    [sym_float_literal] = ACTIONS(874),
    [sym_block_comment] = ACTIONS(3),
  },
  [232] = {
    [sym_function_modifiers] = STATE(2472),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(1935),
    [sym_bracketed_type] = STATE(2378),
    [sym_lifetime] = STATE(1934),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2379),
    [sym_bounded_type] = STATE(1390),
    [sym_type_binding] = STATE(2264),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1390),
    [sym_scoped_identifier] = STATE(2157),
    [sym_scoped_type_identifier] = STATE(1331),
    [sym_block] = STATE(2264),
    [sym__literal] = STATE(2264),
    [sym_string_literal] = STATE(2313),
    [sym_boolean_literal] = STATE(2313),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(854),
    [anon_sym_LPAREN] = ACTIONS(856),
    [anon_sym_LBRACE] = ACTIONS(858),
    [anon_sym_LBRACK] = ACTIONS(860),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(862),
    [anon_sym_i8] = ACTIONS(862),
    [anon_sym_u16] = ACTIONS(862),
    [anon_sym_i16] = ACTIONS(862),
    [anon_sym_u32] = ACTIONS(862),
    [anon_sym_i32] = ACTIONS(862),
    [anon_sym_u64] = ACTIONS(862),
    [anon_sym_i64] = ACTIONS(862),
    [anon_sym_u128] = ACTIONS(862),
    [anon_sym_i128] = ACTIONS(862),
    [anon_sym_isize] = ACTIONS(862),
    [anon_sym_usize] = ACTIONS(862),
    [anon_sym_f32] = ACTIONS(862),
    [anon_sym_f64] = ACTIONS(862),
    [anon_sym_bool] = ACTIONS(862),
    [anon_sym_str] = ACTIONS(862),
    [anon_sym_char] = ACTIONS(862),
    [anon_sym_SQUOTE] = ACTIONS(662),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(664),
    [anon_sym_default] = ACTIONS(864),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(866),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_GT] = ACTIONS(882),
    [anon_sym_COLON_COLON] = ACTIONS(870),
    [anon_sym_AMP] = ACTIONS(872),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_integer_literal] = ACTIONS(874),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(874),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(876),
    [sym_super] = ACTIONS(876),
    [sym_crate] = ACTIONS(876),
    [sym_metavariable] = ACTIONS(878),
    [sym_raw_string_literal] = ACTIONS(874),
    [sym_float_literal] = ACTIONS(874),
    [sym_block_comment] = ACTIONS(3),
  },
  [233] = {
    [sym_identifier] = ACTIONS(448),
    [anon_sym_LPAREN] = ACTIONS(446),
    [anon_sym_RBRACE] = ACTIONS(446),
    [anon_sym_LBRACK] = ACTIONS(446),
    [anon_sym_PLUS] = ACTIONS(448),
    [anon_sym_STAR] = ACTIONS(448),
    [anon_sym_QMARK] = ACTIONS(446),
    [anon_sym_u8] = ACTIONS(448),
    [anon_sym_i8] = ACTIONS(448),
    [anon_sym_u16] = ACTIONS(448),
    [anon_sym_i16] = ACTIONS(448),
    [anon_sym_u32] = ACTIONS(448),
    [anon_sym_i32] = ACTIONS(448),
    [anon_sym_u64] = ACTIONS(448),
    [anon_sym_i64] = ACTIONS(448),
    [anon_sym_u128] = ACTIONS(448),
    [anon_sym_i128] = ACTIONS(448),
    [anon_sym_isize] = ACTIONS(448),
    [anon_sym_usize] = ACTIONS(448),
    [anon_sym_f32] = ACTIONS(448),
    [anon_sym_f64] = ACTIONS(448),
    [anon_sym_bool] = ACTIONS(448),
    [anon_sym_str] = ACTIONS(448),
    [anon_sym_char] = ACTIONS(448),
    [anon_sym_as] = ACTIONS(448),
    [anon_sym_const] = ACTIONS(448),
    [anon_sym_default] = ACTIONS(448),
    [anon_sym_union] = ACTIONS(448),
    [anon_sym_POUND] = ACTIONS(446),
    [anon_sym_EQ] = ACTIONS(448),
    [anon_sym_COMMA] = ACTIONS(446),
    [anon_sym_ref] = ACTIONS(448),
    [anon_sym_LT] = ACTIONS(448),
    [anon_sym_GT] = ACTIONS(448),
    [anon_sym_COLON_COLON] = ACTIONS(446),
    [anon_sym__] = ACTIONS(448),
    [anon_sym_AMP] = ACTIONS(448),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(446),
    [sym_mutable_specifier] = ACTIONS(448),
    [anon_sym_DOT_DOT] = ACTIONS(448),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(446),
    [anon_sym_DASH] = ACTIONS(448),
    [anon_sym_AMP_AMP] = ACTIONS(446),
    [anon_sym_PIPE_PIPE] = ACTIONS(446),
    [anon_sym_PIPE] = ACTIONS(448),
    [anon_sym_CARET] = ACTIONS(448),
    [anon_sym_EQ_EQ] = ACTIONS(446),
    [anon_sym_BANG_EQ] = ACTIONS(446),
    [anon_sym_LT_EQ] = ACTIONS(446),
    [anon_sym_GT_EQ] = ACTIONS(446),
    [anon_sym_LT_LT] = ACTIONS(448),
    [anon_sym_GT_GT] = ACTIONS(448),
    [anon_sym_SLASH] = ACTIONS(448),
    [anon_sym_PERCENT] = ACTIONS(448),
    [anon_sym_PLUS_EQ] = ACTIONS(446),
    [anon_sym_DASH_EQ] = ACTIONS(446),
    [anon_sym_STAR_EQ] = ACTIONS(446),
    [anon_sym_SLASH_EQ] = ACTIONS(446),
    [anon_sym_PERCENT_EQ] = ACTIONS(446),
    [anon_sym_AMP_EQ] = ACTIONS(446),
    [anon_sym_PIPE_EQ] = ACTIONS(446),
    [anon_sym_CARET_EQ] = ACTIONS(446),
    [anon_sym_LT_LT_EQ] = ACTIONS(446),
    [anon_sym_GT_GT_EQ] = ACTIONS(446),
    [anon_sym_DOT] = ACTIONS(448),
    [sym_integer_literal] = ACTIONS(446),
    [aux_sym_string_literal_token1] = ACTIONS(446),
    [sym_char_literal] = ACTIONS(446),
    [anon_sym_true] = ACTIONS(448),
    [anon_sym_false] = ACTIONS(448),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(448),
    [sym_super] = ACTIONS(448),
    [sym_crate] = ACTIONS(448),
    [sym_metavariable] = ACTIONS(446),
    [sym_raw_string_literal] = ACTIONS(446),
    [sym_float_literal] = ACTIONS(446),
    [sym_block_comment] = ACTIONS(3),
  },
  [234] = {
    [sym_identifier] = ACTIONS(606),
    [anon_sym_LPAREN] = ACTIONS(604),
    [anon_sym_RBRACE] = ACTIONS(604),
    [anon_sym_LBRACK] = ACTIONS(604),
    [anon_sym_PLUS] = ACTIONS(606),
    [anon_sym_STAR] = ACTIONS(606),
    [anon_sym_QMARK] = ACTIONS(604),
    [anon_sym_u8] = ACTIONS(606),
    [anon_sym_i8] = ACTIONS(606),
    [anon_sym_u16] = ACTIONS(606),
    [anon_sym_i16] = ACTIONS(606),
    [anon_sym_u32] = ACTIONS(606),
    [anon_sym_i32] = ACTIONS(606),
    [anon_sym_u64] = ACTIONS(606),
    [anon_sym_i64] = ACTIONS(606),
    [anon_sym_u128] = ACTIONS(606),
    [anon_sym_i128] = ACTIONS(606),
    [anon_sym_isize] = ACTIONS(606),
    [anon_sym_usize] = ACTIONS(606),
    [anon_sym_f32] = ACTIONS(606),
    [anon_sym_f64] = ACTIONS(606),
    [anon_sym_bool] = ACTIONS(606),
    [anon_sym_str] = ACTIONS(606),
    [anon_sym_char] = ACTIONS(606),
    [anon_sym_as] = ACTIONS(606),
    [anon_sym_const] = ACTIONS(606),
    [anon_sym_default] = ACTIONS(606),
    [anon_sym_union] = ACTIONS(606),
    [anon_sym_POUND] = ACTIONS(604),
    [anon_sym_EQ] = ACTIONS(606),
    [anon_sym_COMMA] = ACTIONS(604),
    [anon_sym_ref] = ACTIONS(606),
    [anon_sym_LT] = ACTIONS(606),
    [anon_sym_GT] = ACTIONS(606),
    [anon_sym_COLON_COLON] = ACTIONS(604),
    [anon_sym__] = ACTIONS(606),
    [anon_sym_AMP] = ACTIONS(606),
    [anon_sym_DOT_DOT_DOT] = ACTIONS(604),
    [sym_mutable_specifier] = ACTIONS(606),
    [anon_sym_DOT_DOT] = ACTIONS(606),
    [anon_sym_DOT_DOT_EQ] = ACTIONS(604),
    [anon_sym_DASH] = ACTIONS(606),
    [anon_sym_AMP_AMP] = ACTIONS(604),
    [anon_sym_PIPE_PIPE] = ACTIONS(604),
    [anon_sym_PIPE] = ACTIONS(606),
    [anon_sym_CARET] = ACTIONS(606),
    [anon_sym_EQ_EQ] = ACTIONS(604),
    [anon_sym_BANG_EQ] = ACTIONS(604),
    [anon_sym_LT_EQ] = ACTIONS(604),
    [anon_sym_GT_EQ] = ACTIONS(604),
    [anon_sym_LT_LT] = ACTIONS(606),
    [anon_sym_GT_GT] = ACTIONS(606),
    [anon_sym_SLASH] = ACTIONS(606),
    [anon_sym_PERCENT] = ACTIONS(606),
    [anon_sym_PLUS_EQ] = ACTIONS(604),
    [anon_sym_DASH_EQ] = ACTIONS(604),
    [anon_sym_STAR_EQ] = ACTIONS(604),
    [anon_sym_SLASH_EQ] = ACTIONS(604),
    [anon_sym_PERCENT_EQ] = ACTIONS(604),
    [anon_sym_AMP_EQ] = ACTIONS(604),
    [anon_sym_PIPE_EQ] = ACTIONS(604),
    [anon_sym_CARET_EQ] = ACTIONS(604),
    [anon_sym_LT_LT_EQ] = ACTIONS(604),
    [anon_sym_GT_GT_EQ] = ACTIONS(604),
    [anon_sym_DOT] = ACTIONS(606),
    [sym_integer_literal] = ACTIONS(604),
    [aux_sym_string_literal_token1] = ACTIONS(604),
    [sym_char_literal] = ACTIONS(604),
    [anon_sym_true] = ACTIONS(606),
    [anon_sym_false] = ACTIONS(606),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(606),
    [sym_super] = ACTIONS(606),
    [sym_crate] = ACTIONS(606),
    [sym_metavariable] = ACTIONS(604),
    [sym_raw_string_literal] = ACTIONS(604),
    [sym_float_literal] = ACTIONS(604),
    [sym_block_comment] = ACTIONS(3),
  },
  [235] = {
    [sym_function_modifiers] = STATE(2472),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(1935),
    [sym_bracketed_type] = STATE(2378),
    [sym_lifetime] = STATE(1934),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2379),
    [sym_bounded_type] = STATE(1390),
    [sym_type_binding] = STATE(2264),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1390),
    [sym_scoped_identifier] = STATE(2157),
    [sym_scoped_type_identifier] = STATE(1331),
    [sym_block] = STATE(2264),
    [sym__literal] = STATE(2264),
    [sym_string_literal] = STATE(2313),
    [sym_boolean_literal] = STATE(2313),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(854),
    [anon_sym_LPAREN] = ACTIONS(856),
    [anon_sym_LBRACE] = ACTIONS(858),
    [anon_sym_LBRACK] = ACTIONS(860),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(862),
    [anon_sym_i8] = ACTIONS(862),
    [anon_sym_u16] = ACTIONS(862),
    [anon_sym_i16] = ACTIONS(862),
    [anon_sym_u32] = ACTIONS(862),
    [anon_sym_i32] = ACTIONS(862),
    [anon_sym_u64] = ACTIONS(862),
    [anon_sym_i64] = ACTIONS(862),
    [anon_sym_u128] = ACTIONS(862),
    [anon_sym_i128] = ACTIONS(862),
    [anon_sym_isize] = ACTIONS(862),
    [anon_sym_usize] = ACTIONS(862),
    [anon_sym_f32] = ACTIONS(862),
    [anon_sym_f64] = ACTIONS(862),
    [anon_sym_bool] = ACTIONS(862),
    [anon_sym_str] = ACTIONS(862),
    [anon_sym_char] = ACTIONS(862),
    [anon_sym_SQUOTE] = ACTIONS(662),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(664),
    [anon_sym_default] = ACTIONS(864),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(866),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_GT] = ACTIONS(884),
    [anon_sym_COLON_COLON] = ACTIONS(870),
    [anon_sym_AMP] = ACTIONS(872),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_integer_literal] = ACTIONS(874),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(874),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(876),
    [sym_super] = ACTIONS(876),
    [sym_crate] = ACTIONS(876),
    [sym_metavariable] = ACTIONS(878),
    [sym_raw_string_literal] = ACTIONS(874),
    [sym_float_literal] = ACTIONS(874),
    [sym_block_comment] = ACTIONS(3),
  },
  [236] = {
    [sym_function_modifiers] = STATE(2472),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(1935),
    [sym_bracketed_type] = STATE(2378),
    [sym_lifetime] = STATE(1934),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2379),
    [sym_bounded_type] = STATE(1390),
    [sym_type_binding] = STATE(2264),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1390),
    [sym_scoped_identifier] = STATE(2157),
    [sym_scoped_type_identifier] = STATE(1331),
    [sym_block] = STATE(2264),
    [sym__literal] = STATE(2264),
    [sym_string_literal] = STATE(2313),
    [sym_boolean_literal] = STATE(2313),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(854),
    [anon_sym_LPAREN] = ACTIONS(856),
    [anon_sym_LBRACE] = ACTIONS(858),
    [anon_sym_LBRACK] = ACTIONS(860),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(862),
    [anon_sym_i8] = ACTIONS(862),
    [anon_sym_u16] = ACTIONS(862),
    [anon_sym_i16] = ACTIONS(862),
    [anon_sym_u32] = ACTIONS(862),
    [anon_sym_i32] = ACTIONS(862),
    [anon_sym_u64] = ACTIONS(862),
    [anon_sym_i64] = ACTIONS(862),
    [anon_sym_u128] = ACTIONS(862),
    [anon_sym_i128] = ACTIONS(862),
    [anon_sym_isize] = ACTIONS(862),
    [anon_sym_usize] = ACTIONS(862),
    [anon_sym_f32] = ACTIONS(862),
    [anon_sym_f64] = ACTIONS(862),
    [anon_sym_bool] = ACTIONS(862),
    [anon_sym_str] = ACTIONS(862),
    [anon_sym_char] = ACTIONS(862),
    [anon_sym_SQUOTE] = ACTIONS(662),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(664),
    [anon_sym_default] = ACTIONS(864),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(866),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(870),
    [anon_sym_AMP] = ACTIONS(872),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_integer_literal] = ACTIONS(874),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(874),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(876),
    [sym_super] = ACTIONS(876),
    [sym_crate] = ACTIONS(876),
    [sym_metavariable] = ACTIONS(878),
    [sym_raw_string_literal] = ACTIONS(874),
    [sym_float_literal] = ACTIONS(874),
    [sym_block_comment] = ACTIONS(3),
  },
  [237] = {
    [sym_function_modifiers] = STATE(2472),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(1786),
    [sym_bracketed_type] = STATE(2378),
    [sym_lifetime] = STATE(1785),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2379),
    [sym_bounded_type] = STATE(1390),
    [sym_type_binding] = STATE(2086),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1390),
    [sym_scoped_identifier] = STATE(2157),
    [sym_scoped_type_identifier] = STATE(1331),
    [sym_block] = STATE(2086),
    [sym__literal] = STATE(2086),
    [sym_string_literal] = STATE(2313),
    [sym_boolean_literal] = STATE(2313),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(854),
    [anon_sym_LPAREN] = ACTIONS(856),
    [anon_sym_LBRACE] = ACTIONS(858),
    [anon_sym_LBRACK] = ACTIONS(860),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(862),
    [anon_sym_i8] = ACTIONS(862),
    [anon_sym_u16] = ACTIONS(862),
    [anon_sym_i16] = ACTIONS(862),
    [anon_sym_u32] = ACTIONS(862),
    [anon_sym_i32] = ACTIONS(862),
    [anon_sym_u64] = ACTIONS(862),
    [anon_sym_i64] = ACTIONS(862),
    [anon_sym_u128] = ACTIONS(862),
    [anon_sym_i128] = ACTIONS(862),
    [anon_sym_isize] = ACTIONS(862),
    [anon_sym_usize] = ACTIONS(862),
    [anon_sym_f32] = ACTIONS(862),
    [anon_sym_f64] = ACTIONS(862),
    [anon_sym_bool] = ACTIONS(862),
    [anon_sym_str] = ACTIONS(862),
    [anon_sym_char] = ACTIONS(862),
    [anon_sym_SQUOTE] = ACTIONS(662),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(664),
    [anon_sym_default] = ACTIONS(864),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(866),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(870),
    [anon_sym_AMP] = ACTIONS(872),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_integer_literal] = ACTIONS(874),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(874),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(876),
    [sym_super] = ACTIONS(876),
    [sym_crate] = ACTIONS(876),
    [sym_metavariable] = ACTIONS(878),
    [sym_raw_string_literal] = ACTIONS(874),
    [sym_float_literal] = ACTIONS(874),
    [sym_block_comment] = ACTIONS(3),
  },
  [238] = {
    [sym_function_modifiers] = STATE(2472),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(1795),
    [sym_bracketed_type] = STATE(2378),
    [sym_lifetime] = STATE(1799),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2379),
    [sym_bounded_type] = STATE(1390),
    [sym_type_binding] = STATE(1948),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1390),
    [sym_scoped_identifier] = STATE(2157),
    [sym_scoped_type_identifier] = STATE(1331),
    [sym_block] = STATE(1948),
    [sym__literal] = STATE(1948),
    [sym_string_literal] = STATE(2313),
    [sym_boolean_literal] = STATE(2313),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(854),
    [anon_sym_LPAREN] = ACTIONS(856),
    [anon_sym_LBRACE] = ACTIONS(858),
    [anon_sym_LBRACK] = ACTIONS(860),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(862),
    [anon_sym_i8] = ACTIONS(862),
    [anon_sym_u16] = ACTIONS(862),
    [anon_sym_i16] = ACTIONS(862),
    [anon_sym_u32] = ACTIONS(862),
    [anon_sym_i32] = ACTIONS(862),
    [anon_sym_u64] = ACTIONS(862),
    [anon_sym_i64] = ACTIONS(862),
    [anon_sym_u128] = ACTIONS(862),
    [anon_sym_i128] = ACTIONS(862),
    [anon_sym_isize] = ACTIONS(862),
    [anon_sym_usize] = ACTIONS(862),
    [anon_sym_f32] = ACTIONS(862),
    [anon_sym_f64] = ACTIONS(862),
    [anon_sym_bool] = ACTIONS(862),
    [anon_sym_str] = ACTIONS(862),
    [anon_sym_char] = ACTIONS(862),
    [anon_sym_SQUOTE] = ACTIONS(662),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(664),
    [anon_sym_default] = ACTIONS(864),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(866),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(870),
    [anon_sym_AMP] = ACTIONS(872),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_integer_literal] = ACTIONS(874),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(874),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(876),
    [sym_super] = ACTIONS(876),
    [sym_crate] = ACTIONS(876),
    [sym_metavariable] = ACTIONS(878),
    [sym_raw_string_literal] = ACTIONS(874),
    [sym_float_literal] = ACTIONS(874),
    [sym_block_comment] = ACTIONS(3),
  },
  [239] = {
    [sym__token_pattern] = STATE(239),
    [sym_token_tree_pattern] = STATE(239),
    [sym_token_binding_pattern] = STATE(239),
    [sym_token_repetition_pattern] = STATE(239),
    [sym__literal] = STATE(239),
    [sym_string_literal] = STATE(585),
    [sym_boolean_literal] = STATE(585),
    [aux_sym_token_tree_pattern_repeat1] = STATE(239),
    [sym_identifier] = ACTIONS(886),
    [anon_sym_LPAREN] = ACTIONS(889),
    [anon_sym_RPAREN] = ACTIONS(892),
    [anon_sym_LBRACE] = ACTIONS(894),
    [anon_sym_RBRACE] = ACTIONS(892),
    [anon_sym_LBRACK] = ACTIONS(897),
    [anon_sym_RBRACK] = ACTIONS(892),
    [anon_sym_DOLLAR] = ACTIONS(900),
    [anon_sym_u8] = ACTIONS(886),
    [anon_sym_i8] = ACTIONS(886),
    [anon_sym_u16] = ACTIONS(886),
    [anon_sym_i16] = ACTIONS(886),
    [anon_sym_u32] = ACTIONS(886),
    [anon_sym_i32] = ACTIONS(886),
    [anon_sym_u64] = ACTIONS(886),
    [anon_sym_i64] = ACTIONS(886),
    [anon_sym_u128] = ACTIONS(886),
    [anon_sym_i128] = ACTIONS(886),
    [anon_sym_isize] = ACTIONS(886),
    [anon_sym_usize] = ACTIONS(886),
    [anon_sym_f32] = ACTIONS(886),
    [anon_sym_f64] = ACTIONS(886),
    [anon_sym_bool] = ACTIONS(886),
    [anon_sym_str] = ACTIONS(886),
    [anon_sym_char] = ACTIONS(886),
    [aux_sym__non_special_token_token1] = ACTIONS(886),
    [anon_sym_SQUOTE] = ACTIONS(886),
    [anon_sym_as] = ACTIONS(886),
    [anon_sym_async] = ACTIONS(886),
    [anon_sym_await] = ACTIONS(886),
    [anon_sym_break] = ACTIONS(886),
    [anon_sym_const] = ACTIONS(886),
    [anon_sym_continue] = ACTIONS(886),
    [anon_sym_default] = ACTIONS(886),
    [anon_sym_enum] = ACTIONS(886),
    [anon_sym_fn] = ACTIONS(886),
    [anon_sym_for] = ACTIONS(886),
    [anon_sym_if] = ACTIONS(886),
    [anon_sym_impl] = ACTIONS(886),
    [anon_sym_let] = ACTIONS(886),
    [anon_sym_loop] = ACTIONS(886),
    [anon_sym_match] = ACTIONS(886),
    [anon_sym_mod] = ACTIONS(886),
    [anon_sym_pub] = ACTIONS(886),
    [anon_sym_return] = ACTIONS(886),
    [anon_sym_static] = ACTIONS(886),
    [anon_sym_struct] = ACTIONS(886),
    [anon_sym_trait] = ACTIONS(886),
    [anon_sym_type] = ACTIONS(886),
    [anon_sym_union] = ACTIONS(886),
    [anon_sym_unsafe] = ACTIONS(886),
    [anon_sym_use] = ACTIONS(886),
    [anon_sym_where] = ACTIONS(886),
    [anon_sym_while] = ACTIONS(886),
    [sym_mutable_specifier] = ACTIONS(886),
    [sym_integer_literal] = ACTIONS(903),
    [aux_sym_string_literal_token1] = ACTIONS(906),
    [sym_char_literal] = ACTIONS(903),
    [anon_sym_true] = ACTIONS(909),
    [anon_sym_false] = ACTIONS(909),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(886),
    [sym_super] = ACTIONS(886),
    [sym_crate] = ACTIONS(886),
    [sym_metavariable] = ACTIONS(914),
    [sym_raw_string_literal] = ACTIONS(903),
    [sym_float_literal] = ACTIONS(903),
    [sym_block_comment] = ACTIONS(3),
  },
  [240] = {
    [sym_empty_statement] = STATE(240),
    [sym_macro_definition] = STATE(240),
    [sym_attribute_item] = STATE(240),
    [sym_inner_attribute_item] = STATE(240),
    [sym_mod_item] = STATE(240),
    [sym_foreign_mod_item] = STATE(240),
    [sym_struct_item] = STATE(240),
    [sym_union_item] = STATE(240),
    [sym_enum_item] = STATE(240),
    [sym_extern_crate_declaration] = STATE(240),
    [sym_const_item] = STATE(240),
    [sym_static_item] = STATE(240),
    [sym_type_item] = STATE(240),
    [sym_function_item] = STATE(240),
    [sym_function_signature_item] = STATE(240),
    [sym_function_modifiers] = STATE(2552),
    [sym_impl_item] = STATE(240),
    [sym_trait_item] = STATE(240),
    [sym_associated_type] = STATE(240),
    [sym_let_declaration] = STATE(240),
    [sym_use_declaration] = STATE(240),
    [sym_extern_modifier] = STATE(1501),
    [sym_visibility_modifier] = STATE(1334),
    [sym_bracketed_type] = STATE(2369),
    [sym_generic_type_with_turbofish] = STATE(2457),
    [sym_macro_invocation] = STATE(240),
    [sym_scoped_identifier] = STATE(2290),
    [aux_sym_declaration_list_repeat1] = STATE(240),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(917),
    [anon_sym_SEMI] = ACTIONS(920),
    [anon_sym_macro_rules_BANG] = ACTIONS(923),
    [anon_sym_RBRACE] = ACTIONS(926),
    [anon_sym_u8] = ACTIONS(928),
    [anon_sym_i8] = ACTIONS(928),
    [anon_sym_u16] = ACTIONS(928),
    [anon_sym_i16] = ACTIONS(928),
    [anon_sym_u32] = ACTIONS(928),
    [anon_sym_i32] = ACTIONS(928),
    [anon_sym_u64] = ACTIONS(928),
    [anon_sym_i64] = ACTIONS(928),
    [anon_sym_u128] = ACTIONS(928),
    [anon_sym_i128] = ACTIONS(928),
    [anon_sym_isize] = ACTIONS(928),
    [anon_sym_usize] = ACTIONS(928),
    [anon_sym_f32] = ACTIONS(928),
    [anon_sym_f64] = ACTIONS(928),
    [anon_sym_bool] = ACTIONS(928),
    [anon_sym_str] = ACTIONS(928),
    [anon_sym_char] = ACTIONS(928),
    [anon_sym_async] = ACTIONS(931),
    [anon_sym_const] = ACTIONS(934),
    [anon_sym_default] = ACTIONS(937),
    [anon_sym_enum] = ACTIONS(940),
    [anon_sym_fn] = ACTIONS(943),
    [anon_sym_impl] = ACTIONS(946),
    [anon_sym_let] = ACTIONS(949),
    [anon_sym_mod] = ACTIONS(952),
    [anon_sym_pub] = ACTIONS(955),
    [anon_sym_static] = ACTIONS(958),
    [anon_sym_struct] = ACTIONS(961),
    [anon_sym_trait] = ACTIONS(964),
    [anon_sym_type] = ACTIONS(967),
    [anon_sym_union] = ACTIONS(970),
    [anon_sym_unsafe] = ACTIONS(973),
    [anon_sym_use] = ACTIONS(976),
    [anon_sym_POUND] = ACTIONS(979),
    [anon_sym_extern] = ACTIONS(982),
    [anon_sym_LT] = ACTIONS(985),
    [anon_sym_COLON_COLON] = ACTIONS(988),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(991),
    [sym_super] = ACTIONS(991),
    [sym_crate] = ACTIONS(994),
    [sym_metavariable] = ACTIONS(997),
    [sym_block_comment] = ACTIONS(3),
  },
  [241] = {
    [sym_empty_statement] = STATE(240),
    [sym_macro_definition] = STATE(240),
    [sym_attribute_item] = STATE(240),
    [sym_inner_attribute_item] = STATE(240),
    [sym_mod_item] = STATE(240),
    [sym_foreign_mod_item] = STATE(240),
    [sym_struct_item] = STATE(240),
    [sym_union_item] = STATE(240),
    [sym_enum_item] = STATE(240),
    [sym_extern_crate_declaration] = STATE(240),
    [sym_const_item] = STATE(240),
    [sym_static_item] = STATE(240),
    [sym_type_item] = STATE(240),
    [sym_function_item] = STATE(240),
    [sym_function_signature_item] = STATE(240),
    [sym_function_modifiers] = STATE(2552),
    [sym_impl_item] = STATE(240),
    [sym_trait_item] = STATE(240),
    [sym_associated_type] = STATE(240),
    [sym_let_declaration] = STATE(240),
    [sym_use_declaration] = STATE(240),
    [sym_extern_modifier] = STATE(1501),
    [sym_visibility_modifier] = STATE(1334),
    [sym_bracketed_type] = STATE(2369),
    [sym_generic_type_with_turbofish] = STATE(2457),
    [sym_macro_invocation] = STATE(240),
    [sym_scoped_identifier] = STATE(2290),
    [aux_sym_declaration_list_repeat1] = STATE(240),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(1000),
    [anon_sym_SEMI] = ACTIONS(1002),
    [anon_sym_macro_rules_BANG] = ACTIONS(1004),
    [anon_sym_RBRACE] = ACTIONS(1006),
    [anon_sym_u8] = ACTIONS(1008),
    [anon_sym_i8] = ACTIONS(1008),
    [anon_sym_u16] = ACTIONS(1008),
    [anon_sym_i16] = ACTIONS(1008),
    [anon_sym_u32] = ACTIONS(1008),
    [anon_sym_i32] = ACTIONS(1008),
    [anon_sym_u64] = ACTIONS(1008),
    [anon_sym_i64] = ACTIONS(1008),
    [anon_sym_u128] = ACTIONS(1008),
    [anon_sym_i128] = ACTIONS(1008),
    [anon_sym_isize] = ACTIONS(1008),
    [anon_sym_usize] = ACTIONS(1008),
    [anon_sym_f32] = ACTIONS(1008),
    [anon_sym_f64] = ACTIONS(1008),
    [anon_sym_bool] = ACTIONS(1008),
    [anon_sym_str] = ACTIONS(1008),
    [anon_sym_char] = ACTIONS(1008),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(1010),
    [anon_sym_default] = ACTIONS(1012),
    [anon_sym_enum] = ACTIONS(1014),
    [anon_sym_fn] = ACTIONS(1016),
    [anon_sym_impl] = ACTIONS(1018),
    [anon_sym_let] = ACTIONS(1020),
    [anon_sym_mod] = ACTIONS(1022),
    [anon_sym_pub] = ACTIONS(53),
    [anon_sym_static] = ACTIONS(1024),
    [anon_sym_struct] = ACTIONS(1026),
    [anon_sym_trait] = ACTIONS(1028),
    [anon_sym_type] = ACTIONS(1030),
    [anon_sym_union] = ACTIONS(1032),
    [anon_sym_unsafe] = ACTIONS(1034),
    [anon_sym_use] = ACTIONS(1036),
    [anon_sym_POUND] = ACTIONS(1038),
    [anon_sym_extern] = ACTIONS(1040),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(1042),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1044),
    [sym_super] = ACTIONS(1044),
    [sym_crate] = ACTIONS(1046),
    [sym_metavariable] = ACTIONS(1048),
    [sym_block_comment] = ACTIONS(3),
  },
  [242] = {
    [sym_empty_statement] = STATE(241),
    [sym_macro_definition] = STATE(241),
    [sym_attribute_item] = STATE(241),
    [sym_inner_attribute_item] = STATE(241),
    [sym_mod_item] = STATE(241),
    [sym_foreign_mod_item] = STATE(241),
    [sym_struct_item] = STATE(241),
    [sym_union_item] = STATE(241),
    [sym_enum_item] = STATE(241),
    [sym_extern_crate_declaration] = STATE(241),
    [sym_const_item] = STATE(241),
    [sym_static_item] = STATE(241),
    [sym_type_item] = STATE(241),
    [sym_function_item] = STATE(241),
    [sym_function_signature_item] = STATE(241),
    [sym_function_modifiers] = STATE(2552),
    [sym_impl_item] = STATE(241),
    [sym_trait_item] = STATE(241),
    [sym_associated_type] = STATE(241),
    [sym_let_declaration] = STATE(241),
    [sym_use_declaration] = STATE(241),
    [sym_extern_modifier] = STATE(1501),
    [sym_visibility_modifier] = STATE(1334),
    [sym_bracketed_type] = STATE(2369),
    [sym_generic_type_with_turbofish] = STATE(2457),
    [sym_macro_invocation] = STATE(241),
    [sym_scoped_identifier] = STATE(2290),
    [aux_sym_declaration_list_repeat1] = STATE(241),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(1000),
    [anon_sym_SEMI] = ACTIONS(1002),
    [anon_sym_macro_rules_BANG] = ACTIONS(1004),
    [anon_sym_RBRACE] = ACTIONS(1050),
    [anon_sym_u8] = ACTIONS(1008),
    [anon_sym_i8] = ACTIONS(1008),
    [anon_sym_u16] = ACTIONS(1008),
    [anon_sym_i16] = ACTIONS(1008),
    [anon_sym_u32] = ACTIONS(1008),
    [anon_sym_i32] = ACTIONS(1008),
    [anon_sym_u64] = ACTIONS(1008),
    [anon_sym_i64] = ACTIONS(1008),
    [anon_sym_u128] = ACTIONS(1008),
    [anon_sym_i128] = ACTIONS(1008),
    [anon_sym_isize] = ACTIONS(1008),
    [anon_sym_usize] = ACTIONS(1008),
    [anon_sym_f32] = ACTIONS(1008),
    [anon_sym_f64] = ACTIONS(1008),
    [anon_sym_bool] = ACTIONS(1008),
    [anon_sym_str] = ACTIONS(1008),
    [anon_sym_char] = ACTIONS(1008),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(1010),
    [anon_sym_default] = ACTIONS(1012),
    [anon_sym_enum] = ACTIONS(1014),
    [anon_sym_fn] = ACTIONS(1016),
    [anon_sym_impl] = ACTIONS(1018),
    [anon_sym_let] = ACTIONS(1020),
    [anon_sym_mod] = ACTIONS(1022),
    [anon_sym_pub] = ACTIONS(53),
    [anon_sym_static] = ACTIONS(1024),
    [anon_sym_struct] = ACTIONS(1026),
    [anon_sym_trait] = ACTIONS(1028),
    [anon_sym_type] = ACTIONS(1030),
    [anon_sym_union] = ACTIONS(1032),
    [anon_sym_unsafe] = ACTIONS(1034),
    [anon_sym_use] = ACTIONS(1036),
    [anon_sym_POUND] = ACTIONS(1038),
    [anon_sym_extern] = ACTIONS(1040),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(1042),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1044),
    [sym_super] = ACTIONS(1044),
    [sym_crate] = ACTIONS(1046),
    [sym_metavariable] = ACTIONS(1048),
    [sym_block_comment] = ACTIONS(3),
  },
  [243] = {
    [sym_empty_statement] = STATE(244),
    [sym_macro_definition] = STATE(244),
    [sym_attribute_item] = STATE(244),
    [sym_inner_attribute_item] = STATE(244),
    [sym_mod_item] = STATE(244),
    [sym_foreign_mod_item] = STATE(244),
    [sym_struct_item] = STATE(244),
    [sym_union_item] = STATE(244),
    [sym_enum_item] = STATE(244),
    [sym_extern_crate_declaration] = STATE(244),
    [sym_const_item] = STATE(244),
    [sym_static_item] = STATE(244),
    [sym_type_item] = STATE(244),
    [sym_function_item] = STATE(244),
    [sym_function_signature_item] = STATE(244),
    [sym_function_modifiers] = STATE(2552),
    [sym_impl_item] = STATE(244),
    [sym_trait_item] = STATE(244),
    [sym_associated_type] = STATE(244),
    [sym_let_declaration] = STATE(244),
    [sym_use_declaration] = STATE(244),
    [sym_extern_modifier] = STATE(1501),
    [sym_visibility_modifier] = STATE(1334),
    [sym_bracketed_type] = STATE(2369),
    [sym_generic_type_with_turbofish] = STATE(2457),
    [sym_macro_invocation] = STATE(244),
    [sym_scoped_identifier] = STATE(2290),
    [aux_sym_declaration_list_repeat1] = STATE(244),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(1000),
    [anon_sym_SEMI] = ACTIONS(1002),
    [anon_sym_macro_rules_BANG] = ACTIONS(1004),
    [anon_sym_RBRACE] = ACTIONS(1052),
    [anon_sym_u8] = ACTIONS(1008),
    [anon_sym_i8] = ACTIONS(1008),
    [anon_sym_u16] = ACTIONS(1008),
    [anon_sym_i16] = ACTIONS(1008),
    [anon_sym_u32] = ACTIONS(1008),
    [anon_sym_i32] = ACTIONS(1008),
    [anon_sym_u64] = ACTIONS(1008),
    [anon_sym_i64] = ACTIONS(1008),
    [anon_sym_u128] = ACTIONS(1008),
    [anon_sym_i128] = ACTIONS(1008),
    [anon_sym_isize] = ACTIONS(1008),
    [anon_sym_usize] = ACTIONS(1008),
    [anon_sym_f32] = ACTIONS(1008),
    [anon_sym_f64] = ACTIONS(1008),
    [anon_sym_bool] = ACTIONS(1008),
    [anon_sym_str] = ACTIONS(1008),
    [anon_sym_char] = ACTIONS(1008),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(1010),
    [anon_sym_default] = ACTIONS(1012),
    [anon_sym_enum] = ACTIONS(1014),
    [anon_sym_fn] = ACTIONS(1016),
    [anon_sym_impl] = ACTIONS(1018),
    [anon_sym_let] = ACTIONS(1020),
    [anon_sym_mod] = ACTIONS(1022),
    [anon_sym_pub] = ACTIONS(53),
    [anon_sym_static] = ACTIONS(1024),
    [anon_sym_struct] = ACTIONS(1026),
    [anon_sym_trait] = ACTIONS(1028),
    [anon_sym_type] = ACTIONS(1030),
    [anon_sym_union] = ACTIONS(1032),
    [anon_sym_unsafe] = ACTIONS(1034),
    [anon_sym_use] = ACTIONS(1036),
    [anon_sym_POUND] = ACTIONS(1038),
    [anon_sym_extern] = ACTIONS(1040),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(1042),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1044),
    [sym_super] = ACTIONS(1044),
    [sym_crate] = ACTIONS(1046),
    [sym_metavariable] = ACTIONS(1048),
    [sym_block_comment] = ACTIONS(3),
  },
  [244] = {
    [sym_empty_statement] = STATE(240),
    [sym_macro_definition] = STATE(240),
    [sym_attribute_item] = STATE(240),
    [sym_inner_attribute_item] = STATE(240),
    [sym_mod_item] = STATE(240),
    [sym_foreign_mod_item] = STATE(240),
    [sym_struct_item] = STATE(240),
    [sym_union_item] = STATE(240),
    [sym_enum_item] = STATE(240),
    [sym_extern_crate_declaration] = STATE(240),
    [sym_const_item] = STATE(240),
    [sym_static_item] = STATE(240),
    [sym_type_item] = STATE(240),
    [sym_function_item] = STATE(240),
    [sym_function_signature_item] = STATE(240),
    [sym_function_modifiers] = STATE(2552),
    [sym_impl_item] = STATE(240),
    [sym_trait_item] = STATE(240),
    [sym_associated_type] = STATE(240),
    [sym_let_declaration] = STATE(240),
    [sym_use_declaration] = STATE(240),
    [sym_extern_modifier] = STATE(1501),
    [sym_visibility_modifier] = STATE(1334),
    [sym_bracketed_type] = STATE(2369),
    [sym_generic_type_with_turbofish] = STATE(2457),
    [sym_macro_invocation] = STATE(240),
    [sym_scoped_identifier] = STATE(2290),
    [aux_sym_declaration_list_repeat1] = STATE(240),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(1000),
    [anon_sym_SEMI] = ACTIONS(1002),
    [anon_sym_macro_rules_BANG] = ACTIONS(1004),
    [anon_sym_RBRACE] = ACTIONS(1054),
    [anon_sym_u8] = ACTIONS(1008),
    [anon_sym_i8] = ACTIONS(1008),
    [anon_sym_u16] = ACTIONS(1008),
    [anon_sym_i16] = ACTIONS(1008),
    [anon_sym_u32] = ACTIONS(1008),
    [anon_sym_i32] = ACTIONS(1008),
    [anon_sym_u64] = ACTIONS(1008),
    [anon_sym_i64] = ACTIONS(1008),
    [anon_sym_u128] = ACTIONS(1008),
    [anon_sym_i128] = ACTIONS(1008),
    [anon_sym_isize] = ACTIONS(1008),
    [anon_sym_usize] = ACTIONS(1008),
    [anon_sym_f32] = ACTIONS(1008),
    [anon_sym_f64] = ACTIONS(1008),
    [anon_sym_bool] = ACTIONS(1008),
    [anon_sym_str] = ACTIONS(1008),
    [anon_sym_char] = ACTIONS(1008),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(1010),
    [anon_sym_default] = ACTIONS(1012),
    [anon_sym_enum] = ACTIONS(1014),
    [anon_sym_fn] = ACTIONS(1016),
    [anon_sym_impl] = ACTIONS(1018),
    [anon_sym_let] = ACTIONS(1020),
    [anon_sym_mod] = ACTIONS(1022),
    [anon_sym_pub] = ACTIONS(53),
    [anon_sym_static] = ACTIONS(1024),
    [anon_sym_struct] = ACTIONS(1026),
    [anon_sym_trait] = ACTIONS(1028),
    [anon_sym_type] = ACTIONS(1030),
    [anon_sym_union] = ACTIONS(1032),
    [anon_sym_unsafe] = ACTIONS(1034),
    [anon_sym_use] = ACTIONS(1036),
    [anon_sym_POUND] = ACTIONS(1038),
    [anon_sym_extern] = ACTIONS(1040),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(1042),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1044),
    [sym_super] = ACTIONS(1044),
    [sym_crate] = ACTIONS(1046),
    [sym_metavariable] = ACTIONS(1048),
    [sym_block_comment] = ACTIONS(3),
  },
  [245] = {
    [ts_builtin_sym_end] = ACTIONS(414),
    [sym_identifier] = ACTIONS(416),
    [anon_sym_SEMI] = ACTIONS(414),
    [anon_sym_macro_rules_BANG] = ACTIONS(414),
    [anon_sym_LPAREN] = ACTIONS(414),
    [anon_sym_LBRACE] = ACTIONS(414),
    [anon_sym_RBRACE] = ACTIONS(414),
    [anon_sym_LBRACK] = ACTIONS(414),
    [anon_sym_STAR] = ACTIONS(414),
    [anon_sym_u8] = ACTIONS(416),
    [anon_sym_i8] = ACTIONS(416),
    [anon_sym_u16] = ACTIONS(416),
    [anon_sym_i16] = ACTIONS(416),
    [anon_sym_u32] = ACTIONS(416),
    [anon_sym_i32] = ACTIONS(416),
    [anon_sym_u64] = ACTIONS(416),
    [anon_sym_i64] = ACTIONS(416),
    [anon_sym_u128] = ACTIONS(416),
    [anon_sym_i128] = ACTIONS(416),
    [anon_sym_isize] = ACTIONS(416),
    [anon_sym_usize] = ACTIONS(416),
    [anon_sym_f32] = ACTIONS(416),
    [anon_sym_f64] = ACTIONS(416),
    [anon_sym_bool] = ACTIONS(416),
    [anon_sym_str] = ACTIONS(416),
    [anon_sym_char] = ACTIONS(416),
    [anon_sym_SQUOTE] = ACTIONS(416),
    [anon_sym_async] = ACTIONS(416),
    [anon_sym_break] = ACTIONS(416),
    [anon_sym_const] = ACTIONS(416),
    [anon_sym_continue] = ACTIONS(416),
    [anon_sym_default] = ACTIONS(416),
    [anon_sym_enum] = ACTIONS(416),
    [anon_sym_fn] = ACTIONS(416),
    [anon_sym_for] = ACTIONS(416),
    [anon_sym_if] = ACTIONS(416),
    [anon_sym_impl] = ACTIONS(416),
    [anon_sym_let] = ACTIONS(416),
    [anon_sym_loop] = ACTIONS(416),
    [anon_sym_match] = ACTIONS(416),
    [anon_sym_mod] = ACTIONS(416),
    [anon_sym_pub] = ACTIONS(416),
    [anon_sym_return] = ACTIONS(416),
    [anon_sym_static] = ACTIONS(416),
    [anon_sym_struct] = ACTIONS(416),
    [anon_sym_trait] = ACTIONS(416),
    [anon_sym_type] = ACTIONS(416),
    [anon_sym_union] = ACTIONS(416),
    [anon_sym_unsafe] = ACTIONS(416),
    [anon_sym_use] = ACTIONS(416),
    [anon_sym_while] = ACTIONS(416),
    [anon_sym_POUND] = ACTIONS(414),
    [anon_sym_BANG] = ACTIONS(414),
    [anon_sym_extern] = ACTIONS(416),
    [anon_sym_LT] = ACTIONS(414),
    [anon_sym_COLON_COLON] = ACTIONS(414),
    [anon_sym_AMP] = ACTIONS(414),
    [anon_sym_DOT_DOT] = ACTIONS(414),
    [anon_sym_DASH] = ACTIONS(414),
    [anon_sym_PIPE] = ACTIONS(414),
    [anon_sym_yield] = ACTIONS(416),
    [anon_sym_move] = ACTIONS(416),
    [sym_integer_literal] = ACTIONS(414),
    [aux_sym_string_literal_token1] = ACTIONS(414),
    [sym_char_literal] = ACTIONS(414),
    [anon_sym_true] = ACTIONS(416),
    [anon_sym_false] = ACTIONS(416),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(416),
    [sym_super] = ACTIONS(416),
    [sym_crate] = ACTIONS(416),
    [sym_metavariable] = ACTIONS(414),
    [sym_raw_string_literal] = ACTIONS(414),
    [sym_float_literal] = ACTIONS(414),
    [sym_block_comment] = ACTIONS(3),
  },
  [246] = {
    [ts_builtin_sym_end] = ACTIONS(1056),
    [sym_identifier] = ACTIONS(1058),
    [anon_sym_SEMI] = ACTIONS(1056),
    [anon_sym_macro_rules_BANG] = ACTIONS(1056),
    [anon_sym_LPAREN] = ACTIONS(1056),
    [anon_sym_LBRACE] = ACTIONS(1056),
    [anon_sym_RBRACE] = ACTIONS(1056),
    [anon_sym_LBRACK] = ACTIONS(1056),
    [anon_sym_STAR] = ACTIONS(1056),
    [anon_sym_u8] = ACTIONS(1058),
    [anon_sym_i8] = ACTIONS(1058),
    [anon_sym_u16] = ACTIONS(1058),
    [anon_sym_i16] = ACTIONS(1058),
    [anon_sym_u32] = ACTIONS(1058),
    [anon_sym_i32] = ACTIONS(1058),
    [anon_sym_u64] = ACTIONS(1058),
    [anon_sym_i64] = ACTIONS(1058),
    [anon_sym_u128] = ACTIONS(1058),
    [anon_sym_i128] = ACTIONS(1058),
    [anon_sym_isize] = ACTIONS(1058),
    [anon_sym_usize] = ACTIONS(1058),
    [anon_sym_f32] = ACTIONS(1058),
    [anon_sym_f64] = ACTIONS(1058),
    [anon_sym_bool] = ACTIONS(1058),
    [anon_sym_str] = ACTIONS(1058),
    [anon_sym_char] = ACTIONS(1058),
    [anon_sym_SQUOTE] = ACTIONS(1058),
    [anon_sym_async] = ACTIONS(1058),
    [anon_sym_break] = ACTIONS(1058),
    [anon_sym_const] = ACTIONS(1058),
    [anon_sym_continue] = ACTIONS(1058),
    [anon_sym_default] = ACTIONS(1058),
    [anon_sym_enum] = ACTIONS(1058),
    [anon_sym_fn] = ACTIONS(1058),
    [anon_sym_for] = ACTIONS(1058),
    [anon_sym_if] = ACTIONS(1058),
    [anon_sym_impl] = ACTIONS(1058),
    [anon_sym_let] = ACTIONS(1058),
    [anon_sym_loop] = ACTIONS(1058),
    [anon_sym_match] = ACTIONS(1058),
    [anon_sym_mod] = ACTIONS(1058),
    [anon_sym_pub] = ACTIONS(1058),
    [anon_sym_return] = ACTIONS(1058),
    [anon_sym_static] = ACTIONS(1058),
    [anon_sym_struct] = ACTIONS(1058),
    [anon_sym_trait] = ACTIONS(1058),
    [anon_sym_type] = ACTIONS(1058),
    [anon_sym_union] = ACTIONS(1058),
    [anon_sym_unsafe] = ACTIONS(1058),
    [anon_sym_use] = ACTIONS(1058),
    [anon_sym_while] = ACTIONS(1058),
    [anon_sym_POUND] = ACTIONS(1056),
    [anon_sym_BANG] = ACTIONS(1056),
    [anon_sym_extern] = ACTIONS(1058),
    [anon_sym_LT] = ACTIONS(1056),
    [anon_sym_COLON_COLON] = ACTIONS(1056),
    [anon_sym_AMP] = ACTIONS(1056),
    [anon_sym_DOT_DOT] = ACTIONS(1056),
    [anon_sym_DASH] = ACTIONS(1056),
    [anon_sym_PIPE] = ACTIONS(1056),
    [anon_sym_yield] = ACTIONS(1058),
    [anon_sym_move] = ACTIONS(1058),
    [sym_integer_literal] = ACTIONS(1056),
    [aux_sym_string_literal_token1] = ACTIONS(1056),
    [sym_char_literal] = ACTIONS(1056),
    [anon_sym_true] = ACTIONS(1058),
    [anon_sym_false] = ACTIONS(1058),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1058),
    [sym_super] = ACTIONS(1058),
    [sym_crate] = ACTIONS(1058),
    [sym_metavariable] = ACTIONS(1056),
    [sym_raw_string_literal] = ACTIONS(1056),
    [sym_float_literal] = ACTIONS(1056),
    [sym_block_comment] = ACTIONS(3),
  },
  [247] = {
    [ts_builtin_sym_end] = ACTIONS(1060),
    [sym_identifier] = ACTIONS(1062),
    [anon_sym_SEMI] = ACTIONS(1060),
    [anon_sym_macro_rules_BANG] = ACTIONS(1060),
    [anon_sym_LPAREN] = ACTIONS(1060),
    [anon_sym_LBRACE] = ACTIONS(1060),
    [anon_sym_RBRACE] = ACTIONS(1060),
    [anon_sym_LBRACK] = ACTIONS(1060),
    [anon_sym_STAR] = ACTIONS(1060),
    [anon_sym_u8] = ACTIONS(1062),
    [anon_sym_i8] = ACTIONS(1062),
    [anon_sym_u16] = ACTIONS(1062),
    [anon_sym_i16] = ACTIONS(1062),
    [anon_sym_u32] = ACTIONS(1062),
    [anon_sym_i32] = ACTIONS(1062),
    [anon_sym_u64] = ACTIONS(1062),
    [anon_sym_i64] = ACTIONS(1062),
    [anon_sym_u128] = ACTIONS(1062),
    [anon_sym_i128] = ACTIONS(1062),
    [anon_sym_isize] = ACTIONS(1062),
    [anon_sym_usize] = ACTIONS(1062),
    [anon_sym_f32] = ACTIONS(1062),
    [anon_sym_f64] = ACTIONS(1062),
    [anon_sym_bool] = ACTIONS(1062),
    [anon_sym_str] = ACTIONS(1062),
    [anon_sym_char] = ACTIONS(1062),
    [anon_sym_SQUOTE] = ACTIONS(1062),
    [anon_sym_async] = ACTIONS(1062),
    [anon_sym_break] = ACTIONS(1062),
    [anon_sym_const] = ACTIONS(1062),
    [anon_sym_continue] = ACTIONS(1062),
    [anon_sym_default] = ACTIONS(1062),
    [anon_sym_enum] = ACTIONS(1062),
    [anon_sym_fn] = ACTIONS(1062),
    [anon_sym_for] = ACTIONS(1062),
    [anon_sym_if] = ACTIONS(1062),
    [anon_sym_impl] = ACTIONS(1062),
    [anon_sym_let] = ACTIONS(1062),
    [anon_sym_loop] = ACTIONS(1062),
    [anon_sym_match] = ACTIONS(1062),
    [anon_sym_mod] = ACTIONS(1062),
    [anon_sym_pub] = ACTIONS(1062),
    [anon_sym_return] = ACTIONS(1062),
    [anon_sym_static] = ACTIONS(1062),
    [anon_sym_struct] = ACTIONS(1062),
    [anon_sym_trait] = ACTIONS(1062),
    [anon_sym_type] = ACTIONS(1062),
    [anon_sym_union] = ACTIONS(1062),
    [anon_sym_unsafe] = ACTIONS(1062),
    [anon_sym_use] = ACTIONS(1062),
    [anon_sym_while] = ACTIONS(1062),
    [anon_sym_POUND] = ACTIONS(1060),
    [anon_sym_BANG] = ACTIONS(1060),
    [anon_sym_extern] = ACTIONS(1062),
    [anon_sym_LT] = ACTIONS(1060),
    [anon_sym_COLON_COLON] = ACTIONS(1060),
    [anon_sym_AMP] = ACTIONS(1060),
    [anon_sym_DOT_DOT] = ACTIONS(1060),
    [anon_sym_DASH] = ACTIONS(1060),
    [anon_sym_PIPE] = ACTIONS(1060),
    [anon_sym_yield] = ACTIONS(1062),
    [anon_sym_move] = ACTIONS(1062),
    [sym_integer_literal] = ACTIONS(1060),
    [aux_sym_string_literal_token1] = ACTIONS(1060),
    [sym_char_literal] = ACTIONS(1060),
    [anon_sym_true] = ACTIONS(1062),
    [anon_sym_false] = ACTIONS(1062),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1062),
    [sym_super] = ACTIONS(1062),
    [sym_crate] = ACTIONS(1062),
    [sym_metavariable] = ACTIONS(1060),
    [sym_raw_string_literal] = ACTIONS(1060),
    [sym_float_literal] = ACTIONS(1060),
    [sym_block_comment] = ACTIONS(3),
  },
  [248] = {
    [ts_builtin_sym_end] = ACTIONS(1064),
    [sym_identifier] = ACTIONS(1066),
    [anon_sym_SEMI] = ACTIONS(1064),
    [anon_sym_macro_rules_BANG] = ACTIONS(1064),
    [anon_sym_LPAREN] = ACTIONS(1064),
    [anon_sym_LBRACE] = ACTIONS(1064),
    [anon_sym_RBRACE] = ACTIONS(1064),
    [anon_sym_LBRACK] = ACTIONS(1064),
    [anon_sym_STAR] = ACTIONS(1064),
    [anon_sym_u8] = ACTIONS(1066),
    [anon_sym_i8] = ACTIONS(1066),
    [anon_sym_u16] = ACTIONS(1066),
    [anon_sym_i16] = ACTIONS(1066),
    [anon_sym_u32] = ACTIONS(1066),
    [anon_sym_i32] = ACTIONS(1066),
    [anon_sym_u64] = ACTIONS(1066),
    [anon_sym_i64] = ACTIONS(1066),
    [anon_sym_u128] = ACTIONS(1066),
    [anon_sym_i128] = ACTIONS(1066),
    [anon_sym_isize] = ACTIONS(1066),
    [anon_sym_usize] = ACTIONS(1066),
    [anon_sym_f32] = ACTIONS(1066),
    [anon_sym_f64] = ACTIONS(1066),
    [anon_sym_bool] = ACTIONS(1066),
    [anon_sym_str] = ACTIONS(1066),
    [anon_sym_char] = ACTIONS(1066),
    [anon_sym_SQUOTE] = ACTIONS(1066),
    [anon_sym_async] = ACTIONS(1066),
    [anon_sym_break] = ACTIONS(1066),
    [anon_sym_const] = ACTIONS(1066),
    [anon_sym_continue] = ACTIONS(1066),
    [anon_sym_default] = ACTIONS(1066),
    [anon_sym_enum] = ACTIONS(1066),
    [anon_sym_fn] = ACTIONS(1066),
    [anon_sym_for] = ACTIONS(1066),
    [anon_sym_if] = ACTIONS(1066),
    [anon_sym_impl] = ACTIONS(1066),
    [anon_sym_let] = ACTIONS(1066),
    [anon_sym_loop] = ACTIONS(1066),
    [anon_sym_match] = ACTIONS(1066),
    [anon_sym_mod] = ACTIONS(1066),
    [anon_sym_pub] = ACTIONS(1066),
    [anon_sym_return] = ACTIONS(1066),
    [anon_sym_static] = ACTIONS(1066),
    [anon_sym_struct] = ACTIONS(1066),
    [anon_sym_trait] = ACTIONS(1066),
    [anon_sym_type] = ACTIONS(1066),
    [anon_sym_union] = ACTIONS(1066),
    [anon_sym_unsafe] = ACTIONS(1066),
    [anon_sym_use] = ACTIONS(1066),
    [anon_sym_while] = ACTIONS(1066),
    [anon_sym_POUND] = ACTIONS(1064),
    [anon_sym_BANG] = ACTIONS(1064),
    [anon_sym_extern] = ACTIONS(1066),
    [anon_sym_LT] = ACTIONS(1064),
    [anon_sym_COLON_COLON] = ACTIONS(1064),
    [anon_sym_AMP] = ACTIONS(1064),
    [anon_sym_DOT_DOT] = ACTIONS(1064),
    [anon_sym_DASH] = ACTIONS(1064),
    [anon_sym_PIPE] = ACTIONS(1064),
    [anon_sym_yield] = ACTIONS(1066),
    [anon_sym_move] = ACTIONS(1066),
    [sym_integer_literal] = ACTIONS(1064),
    [aux_sym_string_literal_token1] = ACTIONS(1064),
    [sym_char_literal] = ACTIONS(1064),
    [anon_sym_true] = ACTIONS(1066),
    [anon_sym_false] = ACTIONS(1066),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1066),
    [sym_super] = ACTIONS(1066),
    [sym_crate] = ACTIONS(1066),
    [sym_metavariable] = ACTIONS(1064),
    [sym_raw_string_literal] = ACTIONS(1064),
    [sym_float_literal] = ACTIONS(1064),
    [sym_block_comment] = ACTIONS(3),
  },
  [249] = {
    [ts_builtin_sym_end] = ACTIONS(1068),
    [sym_identifier] = ACTIONS(1070),
    [anon_sym_SEMI] = ACTIONS(1068),
    [anon_sym_macro_rules_BANG] = ACTIONS(1068),
    [anon_sym_LPAREN] = ACTIONS(1068),
    [anon_sym_LBRACE] = ACTIONS(1068),
    [anon_sym_RBRACE] = ACTIONS(1068),
    [anon_sym_LBRACK] = ACTIONS(1068),
    [anon_sym_STAR] = ACTIONS(1068),
    [anon_sym_u8] = ACTIONS(1070),
    [anon_sym_i8] = ACTIONS(1070),
    [anon_sym_u16] = ACTIONS(1070),
    [anon_sym_i16] = ACTIONS(1070),
    [anon_sym_u32] = ACTIONS(1070),
    [anon_sym_i32] = ACTIONS(1070),
    [anon_sym_u64] = ACTIONS(1070),
    [anon_sym_i64] = ACTIONS(1070),
    [anon_sym_u128] = ACTIONS(1070),
    [anon_sym_i128] = ACTIONS(1070),
    [anon_sym_isize] = ACTIONS(1070),
    [anon_sym_usize] = ACTIONS(1070),
    [anon_sym_f32] = ACTIONS(1070),
    [anon_sym_f64] = ACTIONS(1070),
    [anon_sym_bool] = ACTIONS(1070),
    [anon_sym_str] = ACTIONS(1070),
    [anon_sym_char] = ACTIONS(1070),
    [anon_sym_SQUOTE] = ACTIONS(1070),
    [anon_sym_async] = ACTIONS(1070),
    [anon_sym_break] = ACTIONS(1070),
    [anon_sym_const] = ACTIONS(1070),
    [anon_sym_continue] = ACTIONS(1070),
    [anon_sym_default] = ACTIONS(1070),
    [anon_sym_enum] = ACTIONS(1070),
    [anon_sym_fn] = ACTIONS(1070),
    [anon_sym_for] = ACTIONS(1070),
    [anon_sym_if] = ACTIONS(1070),
    [anon_sym_impl] = ACTIONS(1070),
    [anon_sym_let] = ACTIONS(1070),
    [anon_sym_loop] = ACTIONS(1070),
    [anon_sym_match] = ACTIONS(1070),
    [anon_sym_mod] = ACTIONS(1070),
    [anon_sym_pub] = ACTIONS(1070),
    [anon_sym_return] = ACTIONS(1070),
    [anon_sym_static] = ACTIONS(1070),
    [anon_sym_struct] = ACTIONS(1070),
    [anon_sym_trait] = ACTIONS(1070),
    [anon_sym_type] = ACTIONS(1070),
    [anon_sym_union] = ACTIONS(1070),
    [anon_sym_unsafe] = ACTIONS(1070),
    [anon_sym_use] = ACTIONS(1070),
    [anon_sym_while] = ACTIONS(1070),
    [anon_sym_POUND] = ACTIONS(1068),
    [anon_sym_BANG] = ACTIONS(1068),
    [anon_sym_extern] = ACTIONS(1070),
    [anon_sym_LT] = ACTIONS(1068),
    [anon_sym_COLON_COLON] = ACTIONS(1068),
    [anon_sym_AMP] = ACTIONS(1068),
    [anon_sym_DOT_DOT] = ACTIONS(1068),
    [anon_sym_DASH] = ACTIONS(1068),
    [anon_sym_PIPE] = ACTIONS(1068),
    [anon_sym_yield] = ACTIONS(1070),
    [anon_sym_move] = ACTIONS(1070),
    [sym_integer_literal] = ACTIONS(1068),
    [aux_sym_string_literal_token1] = ACTIONS(1068),
    [sym_char_literal] = ACTIONS(1068),
    [anon_sym_true] = ACTIONS(1070),
    [anon_sym_false] = ACTIONS(1070),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1070),
    [sym_super] = ACTIONS(1070),
    [sym_crate] = ACTIONS(1070),
    [sym_metavariable] = ACTIONS(1068),
    [sym_raw_string_literal] = ACTIONS(1068),
    [sym_float_literal] = ACTIONS(1068),
    [sym_block_comment] = ACTIONS(3),
  },
  [250] = {
    [ts_builtin_sym_end] = ACTIONS(1072),
    [sym_identifier] = ACTIONS(1074),
    [anon_sym_SEMI] = ACTIONS(1072),
    [anon_sym_macro_rules_BANG] = ACTIONS(1072),
    [anon_sym_LPAREN] = ACTIONS(1072),
    [anon_sym_LBRACE] = ACTIONS(1072),
    [anon_sym_RBRACE] = ACTIONS(1072),
    [anon_sym_LBRACK] = ACTIONS(1072),
    [anon_sym_STAR] = ACTIONS(1072),
    [anon_sym_u8] = ACTIONS(1074),
    [anon_sym_i8] = ACTIONS(1074),
    [anon_sym_u16] = ACTIONS(1074),
    [anon_sym_i16] = ACTIONS(1074),
    [anon_sym_u32] = ACTIONS(1074),
    [anon_sym_i32] = ACTIONS(1074),
    [anon_sym_u64] = ACTIONS(1074),
    [anon_sym_i64] = ACTIONS(1074),
    [anon_sym_u128] = ACTIONS(1074),
    [anon_sym_i128] = ACTIONS(1074),
    [anon_sym_isize] = ACTIONS(1074),
    [anon_sym_usize] = ACTIONS(1074),
    [anon_sym_f32] = ACTIONS(1074),
    [anon_sym_f64] = ACTIONS(1074),
    [anon_sym_bool] = ACTIONS(1074),
    [anon_sym_str] = ACTIONS(1074),
    [anon_sym_char] = ACTIONS(1074),
    [anon_sym_SQUOTE] = ACTIONS(1074),
    [anon_sym_async] = ACTIONS(1074),
    [anon_sym_break] = ACTIONS(1074),
    [anon_sym_const] = ACTIONS(1074),
    [anon_sym_continue] = ACTIONS(1074),
    [anon_sym_default] = ACTIONS(1074),
    [anon_sym_enum] = ACTIONS(1074),
    [anon_sym_fn] = ACTIONS(1074),
    [anon_sym_for] = ACTIONS(1074),
    [anon_sym_if] = ACTIONS(1074),
    [anon_sym_impl] = ACTIONS(1074),
    [anon_sym_let] = ACTIONS(1074),
    [anon_sym_loop] = ACTIONS(1074),
    [anon_sym_match] = ACTIONS(1074),
    [anon_sym_mod] = ACTIONS(1074),
    [anon_sym_pub] = ACTIONS(1074),
    [anon_sym_return] = ACTIONS(1074),
    [anon_sym_static] = ACTIONS(1074),
    [anon_sym_struct] = ACTIONS(1074),
    [anon_sym_trait] = ACTIONS(1074),
    [anon_sym_type] = ACTIONS(1074),
    [anon_sym_union] = ACTIONS(1074),
    [anon_sym_unsafe] = ACTIONS(1074),
    [anon_sym_use] = ACTIONS(1074),
    [anon_sym_while] = ACTIONS(1074),
    [anon_sym_POUND] = ACTIONS(1072),
    [anon_sym_BANG] = ACTIONS(1072),
    [anon_sym_extern] = ACTIONS(1074),
    [anon_sym_LT] = ACTIONS(1072),
    [anon_sym_COLON_COLON] = ACTIONS(1072),
    [anon_sym_AMP] = ACTIONS(1072),
    [anon_sym_DOT_DOT] = ACTIONS(1072),
    [anon_sym_DASH] = ACTIONS(1072),
    [anon_sym_PIPE] = ACTIONS(1072),
    [anon_sym_yield] = ACTIONS(1074),
    [anon_sym_move] = ACTIONS(1074),
    [sym_integer_literal] = ACTIONS(1072),
    [aux_sym_string_literal_token1] = ACTIONS(1072),
    [sym_char_literal] = ACTIONS(1072),
    [anon_sym_true] = ACTIONS(1074),
    [anon_sym_false] = ACTIONS(1074),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1074),
    [sym_super] = ACTIONS(1074),
    [sym_crate] = ACTIONS(1074),
    [sym_metavariable] = ACTIONS(1072),
    [sym_raw_string_literal] = ACTIONS(1072),
    [sym_float_literal] = ACTIONS(1072),
    [sym_block_comment] = ACTIONS(3),
  },
  [251] = {
    [ts_builtin_sym_end] = ACTIONS(1076),
    [sym_identifier] = ACTIONS(1078),
    [anon_sym_SEMI] = ACTIONS(1076),
    [anon_sym_macro_rules_BANG] = ACTIONS(1076),
    [anon_sym_LPAREN] = ACTIONS(1076),
    [anon_sym_LBRACE] = ACTIONS(1076),
    [anon_sym_RBRACE] = ACTIONS(1076),
    [anon_sym_LBRACK] = ACTIONS(1076),
    [anon_sym_STAR] = ACTIONS(1076),
    [anon_sym_u8] = ACTIONS(1078),
    [anon_sym_i8] = ACTIONS(1078),
    [anon_sym_u16] = ACTIONS(1078),
    [anon_sym_i16] = ACTIONS(1078),
    [anon_sym_u32] = ACTIONS(1078),
    [anon_sym_i32] = ACTIONS(1078),
    [anon_sym_u64] = ACTIONS(1078),
    [anon_sym_i64] = ACTIONS(1078),
    [anon_sym_u128] = ACTIONS(1078),
    [anon_sym_i128] = ACTIONS(1078),
    [anon_sym_isize] = ACTIONS(1078),
    [anon_sym_usize] = ACTIONS(1078),
    [anon_sym_f32] = ACTIONS(1078),
    [anon_sym_f64] = ACTIONS(1078),
    [anon_sym_bool] = ACTIONS(1078),
    [anon_sym_str] = ACTIONS(1078),
    [anon_sym_char] = ACTIONS(1078),
    [anon_sym_SQUOTE] = ACTIONS(1078),
    [anon_sym_async] = ACTIONS(1078),
    [anon_sym_break] = ACTIONS(1078),
    [anon_sym_const] = ACTIONS(1078),
    [anon_sym_continue] = ACTIONS(1078),
    [anon_sym_default] = ACTIONS(1078),
    [anon_sym_enum] = ACTIONS(1078),
    [anon_sym_fn] = ACTIONS(1078),
    [anon_sym_for] = ACTIONS(1078),
    [anon_sym_if] = ACTIONS(1078),
    [anon_sym_impl] = ACTIONS(1078),
    [anon_sym_let] = ACTIONS(1078),
    [anon_sym_loop] = ACTIONS(1078),
    [anon_sym_match] = ACTIONS(1078),
    [anon_sym_mod] = ACTIONS(1078),
    [anon_sym_pub] = ACTIONS(1078),
    [anon_sym_return] = ACTIONS(1078),
    [anon_sym_static] = ACTIONS(1078),
    [anon_sym_struct] = ACTIONS(1078),
    [anon_sym_trait] = ACTIONS(1078),
    [anon_sym_type] = ACTIONS(1078),
    [anon_sym_union] = ACTIONS(1078),
    [anon_sym_unsafe] = ACTIONS(1078),
    [anon_sym_use] = ACTIONS(1078),
    [anon_sym_while] = ACTIONS(1078),
    [anon_sym_POUND] = ACTIONS(1076),
    [anon_sym_BANG] = ACTIONS(1076),
    [anon_sym_extern] = ACTIONS(1078),
    [anon_sym_LT] = ACTIONS(1076),
    [anon_sym_COLON_COLON] = ACTIONS(1076),
    [anon_sym_AMP] = ACTIONS(1076),
    [anon_sym_DOT_DOT] = ACTIONS(1076),
    [anon_sym_DASH] = ACTIONS(1076),
    [anon_sym_PIPE] = ACTIONS(1076),
    [anon_sym_yield] = ACTIONS(1078),
    [anon_sym_move] = ACTIONS(1078),
    [sym_integer_literal] = ACTIONS(1076),
    [aux_sym_string_literal_token1] = ACTIONS(1076),
    [sym_char_literal] = ACTIONS(1076),
    [anon_sym_true] = ACTIONS(1078),
    [anon_sym_false] = ACTIONS(1078),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1078),
    [sym_super] = ACTIONS(1078),
    [sym_crate] = ACTIONS(1078),
    [sym_metavariable] = ACTIONS(1076),
    [sym_raw_string_literal] = ACTIONS(1076),
    [sym_float_literal] = ACTIONS(1076),
    [sym_block_comment] = ACTIONS(3),
  },
  [252] = {
    [ts_builtin_sym_end] = ACTIONS(1080),
    [sym_identifier] = ACTIONS(1082),
    [anon_sym_SEMI] = ACTIONS(1080),
    [anon_sym_macro_rules_BANG] = ACTIONS(1080),
    [anon_sym_LPAREN] = ACTIONS(1080),
    [anon_sym_LBRACE] = ACTIONS(1080),
    [anon_sym_RBRACE] = ACTIONS(1080),
    [anon_sym_LBRACK] = ACTIONS(1080),
    [anon_sym_STAR] = ACTIONS(1080),
    [anon_sym_u8] = ACTIONS(1082),
    [anon_sym_i8] = ACTIONS(1082),
    [anon_sym_u16] = ACTIONS(1082),
    [anon_sym_i16] = ACTIONS(1082),
    [anon_sym_u32] = ACTIONS(1082),
    [anon_sym_i32] = ACTIONS(1082),
    [anon_sym_u64] = ACTIONS(1082),
    [anon_sym_i64] = ACTIONS(1082),
    [anon_sym_u128] = ACTIONS(1082),
    [anon_sym_i128] = ACTIONS(1082),
    [anon_sym_isize] = ACTIONS(1082),
    [anon_sym_usize] = ACTIONS(1082),
    [anon_sym_f32] = ACTIONS(1082),
    [anon_sym_f64] = ACTIONS(1082),
    [anon_sym_bool] = ACTIONS(1082),
    [anon_sym_str] = ACTIONS(1082),
    [anon_sym_char] = ACTIONS(1082),
    [anon_sym_SQUOTE] = ACTIONS(1082),
    [anon_sym_async] = ACTIONS(1082),
    [anon_sym_break] = ACTIONS(1082),
    [anon_sym_const] = ACTIONS(1082),
    [anon_sym_continue] = ACTIONS(1082),
    [anon_sym_default] = ACTIONS(1082),
    [anon_sym_enum] = ACTIONS(1082),
    [anon_sym_fn] = ACTIONS(1082),
    [anon_sym_for] = ACTIONS(1082),
    [anon_sym_if] = ACTIONS(1082),
    [anon_sym_impl] = ACTIONS(1082),
    [anon_sym_let] = ACTIONS(1082),
    [anon_sym_loop] = ACTIONS(1082),
    [anon_sym_match] = ACTIONS(1082),
    [anon_sym_mod] = ACTIONS(1082),
    [anon_sym_pub] = ACTIONS(1082),
    [anon_sym_return] = ACTIONS(1082),
    [anon_sym_static] = ACTIONS(1082),
    [anon_sym_struct] = ACTIONS(1082),
    [anon_sym_trait] = ACTIONS(1082),
    [anon_sym_type] = ACTIONS(1082),
    [anon_sym_union] = ACTIONS(1082),
    [anon_sym_unsafe] = ACTIONS(1082),
    [anon_sym_use] = ACTIONS(1082),
    [anon_sym_while] = ACTIONS(1082),
    [anon_sym_POUND] = ACTIONS(1080),
    [anon_sym_BANG] = ACTIONS(1080),
    [anon_sym_extern] = ACTIONS(1082),
    [anon_sym_LT] = ACTIONS(1080),
    [anon_sym_COLON_COLON] = ACTIONS(1080),
    [anon_sym_AMP] = ACTIONS(1080),
    [anon_sym_DOT_DOT] = ACTIONS(1080),
    [anon_sym_DASH] = ACTIONS(1080),
    [anon_sym_PIPE] = ACTIONS(1080),
    [anon_sym_yield] = ACTIONS(1082),
    [anon_sym_move] = ACTIONS(1082),
    [sym_integer_literal] = ACTIONS(1080),
    [aux_sym_string_literal_token1] = ACTIONS(1080),
    [sym_char_literal] = ACTIONS(1080),
    [anon_sym_true] = ACTIONS(1082),
    [anon_sym_false] = ACTIONS(1082),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1082),
    [sym_super] = ACTIONS(1082),
    [sym_crate] = ACTIONS(1082),
    [sym_metavariable] = ACTIONS(1080),
    [sym_raw_string_literal] = ACTIONS(1080),
    [sym_float_literal] = ACTIONS(1080),
    [sym_block_comment] = ACTIONS(3),
  },
  [253] = {
    [ts_builtin_sym_end] = ACTIONS(1084),
    [sym_identifier] = ACTIONS(1086),
    [anon_sym_SEMI] = ACTIONS(1084),
    [anon_sym_macro_rules_BANG] = ACTIONS(1084),
    [anon_sym_LPAREN] = ACTIONS(1084),
    [anon_sym_LBRACE] = ACTIONS(1084),
    [anon_sym_RBRACE] = ACTIONS(1084),
    [anon_sym_LBRACK] = ACTIONS(1084),
    [anon_sym_STAR] = ACTIONS(1084),
    [anon_sym_u8] = ACTIONS(1086),
    [anon_sym_i8] = ACTIONS(1086),
    [anon_sym_u16] = ACTIONS(1086),
    [anon_sym_i16] = ACTIONS(1086),
    [anon_sym_u32] = ACTIONS(1086),
    [anon_sym_i32] = ACTIONS(1086),
    [anon_sym_u64] = ACTIONS(1086),
    [anon_sym_i64] = ACTIONS(1086),
    [anon_sym_u128] = ACTIONS(1086),
    [anon_sym_i128] = ACTIONS(1086),
    [anon_sym_isize] = ACTIONS(1086),
    [anon_sym_usize] = ACTIONS(1086),
    [anon_sym_f32] = ACTIONS(1086),
    [anon_sym_f64] = ACTIONS(1086),
    [anon_sym_bool] = ACTIONS(1086),
    [anon_sym_str] = ACTIONS(1086),
    [anon_sym_char] = ACTIONS(1086),
    [anon_sym_SQUOTE] = ACTIONS(1086),
    [anon_sym_async] = ACTIONS(1086),
    [anon_sym_break] = ACTIONS(1086),
    [anon_sym_const] = ACTIONS(1086),
    [anon_sym_continue] = ACTIONS(1086),
    [anon_sym_default] = ACTIONS(1086),
    [anon_sym_enum] = ACTIONS(1086),
    [anon_sym_fn] = ACTIONS(1086),
    [anon_sym_for] = ACTIONS(1086),
    [anon_sym_if] = ACTIONS(1086),
    [anon_sym_impl] = ACTIONS(1086),
    [anon_sym_let] = ACTIONS(1086),
    [anon_sym_loop] = ACTIONS(1086),
    [anon_sym_match] = ACTIONS(1086),
    [anon_sym_mod] = ACTIONS(1086),
    [anon_sym_pub] = ACTIONS(1086),
    [anon_sym_return] = ACTIONS(1086),
    [anon_sym_static] = ACTIONS(1086),
    [anon_sym_struct] = ACTIONS(1086),
    [anon_sym_trait] = ACTIONS(1086),
    [anon_sym_type] = ACTIONS(1086),
    [anon_sym_union] = ACTIONS(1086),
    [anon_sym_unsafe] = ACTIONS(1086),
    [anon_sym_use] = ACTIONS(1086),
    [anon_sym_while] = ACTIONS(1086),
    [anon_sym_POUND] = ACTIONS(1084),
    [anon_sym_BANG] = ACTIONS(1084),
    [anon_sym_extern] = ACTIONS(1086),
    [anon_sym_LT] = ACTIONS(1084),
    [anon_sym_COLON_COLON] = ACTIONS(1084),
    [anon_sym_AMP] = ACTIONS(1084),
    [anon_sym_DOT_DOT] = ACTIONS(1084),
    [anon_sym_DASH] = ACTIONS(1084),
    [anon_sym_PIPE] = ACTIONS(1084),
    [anon_sym_yield] = ACTIONS(1086),
    [anon_sym_move] = ACTIONS(1086),
    [sym_integer_literal] = ACTIONS(1084),
    [aux_sym_string_literal_token1] = ACTIONS(1084),
    [sym_char_literal] = ACTIONS(1084),
    [anon_sym_true] = ACTIONS(1086),
    [anon_sym_false] = ACTIONS(1086),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1086),
    [sym_super] = ACTIONS(1086),
    [sym_crate] = ACTIONS(1086),
    [sym_metavariable] = ACTIONS(1084),
    [sym_raw_string_literal] = ACTIONS(1084),
    [sym_float_literal] = ACTIONS(1084),
    [sym_block_comment] = ACTIONS(3),
  },
  [254] = {
    [ts_builtin_sym_end] = ACTIONS(1088),
    [sym_identifier] = ACTIONS(1090),
    [anon_sym_SEMI] = ACTIONS(1088),
    [anon_sym_macro_rules_BANG] = ACTIONS(1088),
    [anon_sym_LPAREN] = ACTIONS(1088),
    [anon_sym_LBRACE] = ACTIONS(1088),
    [anon_sym_RBRACE] = ACTIONS(1088),
    [anon_sym_LBRACK] = ACTIONS(1088),
    [anon_sym_STAR] = ACTIONS(1088),
    [anon_sym_u8] = ACTIONS(1090),
    [anon_sym_i8] = ACTIONS(1090),
    [anon_sym_u16] = ACTIONS(1090),
    [anon_sym_i16] = ACTIONS(1090),
    [anon_sym_u32] = ACTIONS(1090),
    [anon_sym_i32] = ACTIONS(1090),
    [anon_sym_u64] = ACTIONS(1090),
    [anon_sym_i64] = ACTIONS(1090),
    [anon_sym_u128] = ACTIONS(1090),
    [anon_sym_i128] = ACTIONS(1090),
    [anon_sym_isize] = ACTIONS(1090),
    [anon_sym_usize] = ACTIONS(1090),
    [anon_sym_f32] = ACTIONS(1090),
    [anon_sym_f64] = ACTIONS(1090),
    [anon_sym_bool] = ACTIONS(1090),
    [anon_sym_str] = ACTIONS(1090),
    [anon_sym_char] = ACTIONS(1090),
    [anon_sym_SQUOTE] = ACTIONS(1090),
    [anon_sym_async] = ACTIONS(1090),
    [anon_sym_break] = ACTIONS(1090),
    [anon_sym_const] = ACTIONS(1090),
    [anon_sym_continue] = ACTIONS(1090),
    [anon_sym_default] = ACTIONS(1090),
    [anon_sym_enum] = ACTIONS(1090),
    [anon_sym_fn] = ACTIONS(1090),
    [anon_sym_for] = ACTIONS(1090),
    [anon_sym_if] = ACTIONS(1090),
    [anon_sym_impl] = ACTIONS(1090),
    [anon_sym_let] = ACTIONS(1090),
    [anon_sym_loop] = ACTIONS(1090),
    [anon_sym_match] = ACTIONS(1090),
    [anon_sym_mod] = ACTIONS(1090),
    [anon_sym_pub] = ACTIONS(1090),
    [anon_sym_return] = ACTIONS(1090),
    [anon_sym_static] = ACTIONS(1090),
    [anon_sym_struct] = ACTIONS(1090),
    [anon_sym_trait] = ACTIONS(1090),
    [anon_sym_type] = ACTIONS(1090),
    [anon_sym_union] = ACTIONS(1090),
    [anon_sym_unsafe] = ACTIONS(1090),
    [anon_sym_use] = ACTIONS(1090),
    [anon_sym_while] = ACTIONS(1090),
    [anon_sym_POUND] = ACTIONS(1088),
    [anon_sym_BANG] = ACTIONS(1088),
    [anon_sym_extern] = ACTIONS(1090),
    [anon_sym_LT] = ACTIONS(1088),
    [anon_sym_COLON_COLON] = ACTIONS(1088),
    [anon_sym_AMP] = ACTIONS(1088),
    [anon_sym_DOT_DOT] = ACTIONS(1088),
    [anon_sym_DASH] = ACTIONS(1088),
    [anon_sym_PIPE] = ACTIONS(1088),
    [anon_sym_yield] = ACTIONS(1090),
    [anon_sym_move] = ACTIONS(1090),
    [sym_integer_literal] = ACTIONS(1088),
    [aux_sym_string_literal_token1] = ACTIONS(1088),
    [sym_char_literal] = ACTIONS(1088),
    [anon_sym_true] = ACTIONS(1090),
    [anon_sym_false] = ACTIONS(1090),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1090),
    [sym_super] = ACTIONS(1090),
    [sym_crate] = ACTIONS(1090),
    [sym_metavariable] = ACTIONS(1088),
    [sym_raw_string_literal] = ACTIONS(1088),
    [sym_float_literal] = ACTIONS(1088),
    [sym_block_comment] = ACTIONS(3),
  },
  [255] = {
    [ts_builtin_sym_end] = ACTIONS(1092),
    [sym_identifier] = ACTIONS(1094),
    [anon_sym_SEMI] = ACTIONS(1092),
    [anon_sym_macro_rules_BANG] = ACTIONS(1092),
    [anon_sym_LPAREN] = ACTIONS(1092),
    [anon_sym_LBRACE] = ACTIONS(1092),
    [anon_sym_RBRACE] = ACTIONS(1092),
    [anon_sym_LBRACK] = ACTIONS(1092),
    [anon_sym_STAR] = ACTIONS(1092),
    [anon_sym_u8] = ACTIONS(1094),
    [anon_sym_i8] = ACTIONS(1094),
    [anon_sym_u16] = ACTIONS(1094),
    [anon_sym_i16] = ACTIONS(1094),
    [anon_sym_u32] = ACTIONS(1094),
    [anon_sym_i32] = ACTIONS(1094),
    [anon_sym_u64] = ACTIONS(1094),
    [anon_sym_i64] = ACTIONS(1094),
    [anon_sym_u128] = ACTIONS(1094),
    [anon_sym_i128] = ACTIONS(1094),
    [anon_sym_isize] = ACTIONS(1094),
    [anon_sym_usize] = ACTIONS(1094),
    [anon_sym_f32] = ACTIONS(1094),
    [anon_sym_f64] = ACTIONS(1094),
    [anon_sym_bool] = ACTIONS(1094),
    [anon_sym_str] = ACTIONS(1094),
    [anon_sym_char] = ACTIONS(1094),
    [anon_sym_SQUOTE] = ACTIONS(1094),
    [anon_sym_async] = ACTIONS(1094),
    [anon_sym_break] = ACTIONS(1094),
    [anon_sym_const] = ACTIONS(1094),
    [anon_sym_continue] = ACTIONS(1094),
    [anon_sym_default] = ACTIONS(1094),
    [anon_sym_enum] = ACTIONS(1094),
    [anon_sym_fn] = ACTIONS(1094),
    [anon_sym_for] = ACTIONS(1094),
    [anon_sym_if] = ACTIONS(1094),
    [anon_sym_impl] = ACTIONS(1094),
    [anon_sym_let] = ACTIONS(1094),
    [anon_sym_loop] = ACTIONS(1094),
    [anon_sym_match] = ACTIONS(1094),
    [anon_sym_mod] = ACTIONS(1094),
    [anon_sym_pub] = ACTIONS(1094),
    [anon_sym_return] = ACTIONS(1094),
    [anon_sym_static] = ACTIONS(1094),
    [anon_sym_struct] = ACTIONS(1094),
    [anon_sym_trait] = ACTIONS(1094),
    [anon_sym_type] = ACTIONS(1094),
    [anon_sym_union] = ACTIONS(1094),
    [anon_sym_unsafe] = ACTIONS(1094),
    [anon_sym_use] = ACTIONS(1094),
    [anon_sym_while] = ACTIONS(1094),
    [anon_sym_POUND] = ACTIONS(1092),
    [anon_sym_BANG] = ACTIONS(1092),
    [anon_sym_extern] = ACTIONS(1094),
    [anon_sym_LT] = ACTIONS(1092),
    [anon_sym_COLON_COLON] = ACTIONS(1092),
    [anon_sym_AMP] = ACTIONS(1092),
    [anon_sym_DOT_DOT] = ACTIONS(1092),
    [anon_sym_DASH] = ACTIONS(1092),
    [anon_sym_PIPE] = ACTIONS(1092),
    [anon_sym_yield] = ACTIONS(1094),
    [anon_sym_move] = ACTIONS(1094),
    [sym_integer_literal] = ACTIONS(1092),
    [aux_sym_string_literal_token1] = ACTIONS(1092),
    [sym_char_literal] = ACTIONS(1092),
    [anon_sym_true] = ACTIONS(1094),
    [anon_sym_false] = ACTIONS(1094),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1094),
    [sym_super] = ACTIONS(1094),
    [sym_crate] = ACTIONS(1094),
    [sym_metavariable] = ACTIONS(1092),
    [sym_raw_string_literal] = ACTIONS(1092),
    [sym_float_literal] = ACTIONS(1092),
    [sym_block_comment] = ACTIONS(3),
  },
  [256] = {
    [ts_builtin_sym_end] = ACTIONS(1096),
    [sym_identifier] = ACTIONS(1098),
    [anon_sym_SEMI] = ACTIONS(1096),
    [anon_sym_macro_rules_BANG] = ACTIONS(1096),
    [anon_sym_LPAREN] = ACTIONS(1096),
    [anon_sym_LBRACE] = ACTIONS(1096),
    [anon_sym_RBRACE] = ACTIONS(1096),
    [anon_sym_LBRACK] = ACTIONS(1096),
    [anon_sym_STAR] = ACTIONS(1096),
    [anon_sym_u8] = ACTIONS(1098),
    [anon_sym_i8] = ACTIONS(1098),
    [anon_sym_u16] = ACTIONS(1098),
    [anon_sym_i16] = ACTIONS(1098),
    [anon_sym_u32] = ACTIONS(1098),
    [anon_sym_i32] = ACTIONS(1098),
    [anon_sym_u64] = ACTIONS(1098),
    [anon_sym_i64] = ACTIONS(1098),
    [anon_sym_u128] = ACTIONS(1098),
    [anon_sym_i128] = ACTIONS(1098),
    [anon_sym_isize] = ACTIONS(1098),
    [anon_sym_usize] = ACTIONS(1098),
    [anon_sym_f32] = ACTIONS(1098),
    [anon_sym_f64] = ACTIONS(1098),
    [anon_sym_bool] = ACTIONS(1098),
    [anon_sym_str] = ACTIONS(1098),
    [anon_sym_char] = ACTIONS(1098),
    [anon_sym_SQUOTE] = ACTIONS(1098),
    [anon_sym_async] = ACTIONS(1098),
    [anon_sym_break] = ACTIONS(1098),
    [anon_sym_const] = ACTIONS(1098),
    [anon_sym_continue] = ACTIONS(1098),
    [anon_sym_default] = ACTIONS(1098),
    [anon_sym_enum] = ACTIONS(1098),
    [anon_sym_fn] = ACTIONS(1098),
    [anon_sym_for] = ACTIONS(1098),
    [anon_sym_if] = ACTIONS(1098),
    [anon_sym_impl] = ACTIONS(1098),
    [anon_sym_let] = ACTIONS(1098),
    [anon_sym_loop] = ACTIONS(1098),
    [anon_sym_match] = ACTIONS(1098),
    [anon_sym_mod] = ACTIONS(1098),
    [anon_sym_pub] = ACTIONS(1098),
    [anon_sym_return] = ACTIONS(1098),
    [anon_sym_static] = ACTIONS(1098),
    [anon_sym_struct] = ACTIONS(1098),
    [anon_sym_trait] = ACTIONS(1098),
    [anon_sym_type] = ACTIONS(1098),
    [anon_sym_union] = ACTIONS(1098),
    [anon_sym_unsafe] = ACTIONS(1098),
    [anon_sym_use] = ACTIONS(1098),
    [anon_sym_while] = ACTIONS(1098),
    [anon_sym_POUND] = ACTIONS(1096),
    [anon_sym_BANG] = ACTIONS(1096),
    [anon_sym_extern] = ACTIONS(1098),
    [anon_sym_LT] = ACTIONS(1096),
    [anon_sym_COLON_COLON] = ACTIONS(1096),
    [anon_sym_AMP] = ACTIONS(1096),
    [anon_sym_DOT_DOT] = ACTIONS(1096),
    [anon_sym_DASH] = ACTIONS(1096),
    [anon_sym_PIPE] = ACTIONS(1096),
    [anon_sym_yield] = ACTIONS(1098),
    [anon_sym_move] = ACTIONS(1098),
    [sym_integer_literal] = ACTIONS(1096),
    [aux_sym_string_literal_token1] = ACTIONS(1096),
    [sym_char_literal] = ACTIONS(1096),
    [anon_sym_true] = ACTIONS(1098),
    [anon_sym_false] = ACTIONS(1098),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1098),
    [sym_super] = ACTIONS(1098),
    [sym_crate] = ACTIONS(1098),
    [sym_metavariable] = ACTIONS(1096),
    [sym_raw_string_literal] = ACTIONS(1096),
    [sym_float_literal] = ACTIONS(1096),
    [sym_block_comment] = ACTIONS(3),
  },
  [257] = {
    [ts_builtin_sym_end] = ACTIONS(1100),
    [sym_identifier] = ACTIONS(1102),
    [anon_sym_SEMI] = ACTIONS(1100),
    [anon_sym_macro_rules_BANG] = ACTIONS(1100),
    [anon_sym_LPAREN] = ACTIONS(1100),
    [anon_sym_LBRACE] = ACTIONS(1100),
    [anon_sym_RBRACE] = ACTIONS(1100),
    [anon_sym_LBRACK] = ACTIONS(1100),
    [anon_sym_STAR] = ACTIONS(1100),
    [anon_sym_u8] = ACTIONS(1102),
    [anon_sym_i8] = ACTIONS(1102),
    [anon_sym_u16] = ACTIONS(1102),
    [anon_sym_i16] = ACTIONS(1102),
    [anon_sym_u32] = ACTIONS(1102),
    [anon_sym_i32] = ACTIONS(1102),
    [anon_sym_u64] = ACTIONS(1102),
    [anon_sym_i64] = ACTIONS(1102),
    [anon_sym_u128] = ACTIONS(1102),
    [anon_sym_i128] = ACTIONS(1102),
    [anon_sym_isize] = ACTIONS(1102),
    [anon_sym_usize] = ACTIONS(1102),
    [anon_sym_f32] = ACTIONS(1102),
    [anon_sym_f64] = ACTIONS(1102),
    [anon_sym_bool] = ACTIONS(1102),
    [anon_sym_str] = ACTIONS(1102),
    [anon_sym_char] = ACTIONS(1102),
    [anon_sym_SQUOTE] = ACTIONS(1102),
    [anon_sym_async] = ACTIONS(1102),
    [anon_sym_break] = ACTIONS(1102),
    [anon_sym_const] = ACTIONS(1102),
    [anon_sym_continue] = ACTIONS(1102),
    [anon_sym_default] = ACTIONS(1102),
    [anon_sym_enum] = ACTIONS(1102),
    [anon_sym_fn] = ACTIONS(1102),
    [anon_sym_for] = ACTIONS(1102),
    [anon_sym_if] = ACTIONS(1102),
    [anon_sym_impl] = ACTIONS(1102),
    [anon_sym_let] = ACTIONS(1102),
    [anon_sym_loop] = ACTIONS(1102),
    [anon_sym_match] = ACTIONS(1102),
    [anon_sym_mod] = ACTIONS(1102),
    [anon_sym_pub] = ACTIONS(1102),
    [anon_sym_return] = ACTIONS(1102),
    [anon_sym_static] = ACTIONS(1102),
    [anon_sym_struct] = ACTIONS(1102),
    [anon_sym_trait] = ACTIONS(1102),
    [anon_sym_type] = ACTIONS(1102),
    [anon_sym_union] = ACTIONS(1102),
    [anon_sym_unsafe] = ACTIONS(1102),
    [anon_sym_use] = ACTIONS(1102),
    [anon_sym_while] = ACTIONS(1102),
    [anon_sym_POUND] = ACTIONS(1100),
    [anon_sym_BANG] = ACTIONS(1100),
    [anon_sym_extern] = ACTIONS(1102),
    [anon_sym_LT] = ACTIONS(1100),
    [anon_sym_COLON_COLON] = ACTIONS(1100),
    [anon_sym_AMP] = ACTIONS(1100),
    [anon_sym_DOT_DOT] = ACTIONS(1100),
    [anon_sym_DASH] = ACTIONS(1100),
    [anon_sym_PIPE] = ACTIONS(1100),
    [anon_sym_yield] = ACTIONS(1102),
    [anon_sym_move] = ACTIONS(1102),
    [sym_integer_literal] = ACTIONS(1100),
    [aux_sym_string_literal_token1] = ACTIONS(1100),
    [sym_char_literal] = ACTIONS(1100),
    [anon_sym_true] = ACTIONS(1102),
    [anon_sym_false] = ACTIONS(1102),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1102),
    [sym_super] = ACTIONS(1102),
    [sym_crate] = ACTIONS(1102),
    [sym_metavariable] = ACTIONS(1100),
    [sym_raw_string_literal] = ACTIONS(1100),
    [sym_float_literal] = ACTIONS(1100),
    [sym_block_comment] = ACTIONS(3),
  },
  [258] = {
    [ts_builtin_sym_end] = ACTIONS(1104),
    [sym_identifier] = ACTIONS(1106),
    [anon_sym_SEMI] = ACTIONS(1104),
    [anon_sym_macro_rules_BANG] = ACTIONS(1104),
    [anon_sym_LPAREN] = ACTIONS(1104),
    [anon_sym_LBRACE] = ACTIONS(1104),
    [anon_sym_RBRACE] = ACTIONS(1104),
    [anon_sym_LBRACK] = ACTIONS(1104),
    [anon_sym_STAR] = ACTIONS(1104),
    [anon_sym_u8] = ACTIONS(1106),
    [anon_sym_i8] = ACTIONS(1106),
    [anon_sym_u16] = ACTIONS(1106),
    [anon_sym_i16] = ACTIONS(1106),
    [anon_sym_u32] = ACTIONS(1106),
    [anon_sym_i32] = ACTIONS(1106),
    [anon_sym_u64] = ACTIONS(1106),
    [anon_sym_i64] = ACTIONS(1106),
    [anon_sym_u128] = ACTIONS(1106),
    [anon_sym_i128] = ACTIONS(1106),
    [anon_sym_isize] = ACTIONS(1106),
    [anon_sym_usize] = ACTIONS(1106),
    [anon_sym_f32] = ACTIONS(1106),
    [anon_sym_f64] = ACTIONS(1106),
    [anon_sym_bool] = ACTIONS(1106),
    [anon_sym_str] = ACTIONS(1106),
    [anon_sym_char] = ACTIONS(1106),
    [anon_sym_SQUOTE] = ACTIONS(1106),
    [anon_sym_async] = ACTIONS(1106),
    [anon_sym_break] = ACTIONS(1106),
    [anon_sym_const] = ACTIONS(1106),
    [anon_sym_continue] = ACTIONS(1106),
    [anon_sym_default] = ACTIONS(1106),
    [anon_sym_enum] = ACTIONS(1106),
    [anon_sym_fn] = ACTIONS(1106),
    [anon_sym_for] = ACTIONS(1106),
    [anon_sym_if] = ACTIONS(1106),
    [anon_sym_impl] = ACTIONS(1106),
    [anon_sym_let] = ACTIONS(1106),
    [anon_sym_loop] = ACTIONS(1106),
    [anon_sym_match] = ACTIONS(1106),
    [anon_sym_mod] = ACTIONS(1106),
    [anon_sym_pub] = ACTIONS(1106),
    [anon_sym_return] = ACTIONS(1106),
    [anon_sym_static] = ACTIONS(1106),
    [anon_sym_struct] = ACTIONS(1106),
    [anon_sym_trait] = ACTIONS(1106),
    [anon_sym_type] = ACTIONS(1106),
    [anon_sym_union] = ACTIONS(1106),
    [anon_sym_unsafe] = ACTIONS(1106),
    [anon_sym_use] = ACTIONS(1106),
    [anon_sym_while] = ACTIONS(1106),
    [anon_sym_POUND] = ACTIONS(1104),
    [anon_sym_BANG] = ACTIONS(1104),
    [anon_sym_extern] = ACTIONS(1106),
    [anon_sym_LT] = ACTIONS(1104),
    [anon_sym_COLON_COLON] = ACTIONS(1104),
    [anon_sym_AMP] = ACTIONS(1104),
    [anon_sym_DOT_DOT] = ACTIONS(1104),
    [anon_sym_DASH] = ACTIONS(1104),
    [anon_sym_PIPE] = ACTIONS(1104),
    [anon_sym_yield] = ACTIONS(1106),
    [anon_sym_move] = ACTIONS(1106),
    [sym_integer_literal] = ACTIONS(1104),
    [aux_sym_string_literal_token1] = ACTIONS(1104),
    [sym_char_literal] = ACTIONS(1104),
    [anon_sym_true] = ACTIONS(1106),
    [anon_sym_false] = ACTIONS(1106),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1106),
    [sym_super] = ACTIONS(1106),
    [sym_crate] = ACTIONS(1106),
    [sym_metavariable] = ACTIONS(1104),
    [sym_raw_string_literal] = ACTIONS(1104),
    [sym_float_literal] = ACTIONS(1104),
    [sym_block_comment] = ACTIONS(3),
  },
  [259] = {
    [ts_builtin_sym_end] = ACTIONS(1108),
    [sym_identifier] = ACTIONS(1110),
    [anon_sym_SEMI] = ACTIONS(1108),
    [anon_sym_macro_rules_BANG] = ACTIONS(1108),
    [anon_sym_LPAREN] = ACTIONS(1108),
    [anon_sym_LBRACE] = ACTIONS(1108),
    [anon_sym_RBRACE] = ACTIONS(1108),
    [anon_sym_LBRACK] = ACTIONS(1108),
    [anon_sym_STAR] = ACTIONS(1108),
    [anon_sym_u8] = ACTIONS(1110),
    [anon_sym_i8] = ACTIONS(1110),
    [anon_sym_u16] = ACTIONS(1110),
    [anon_sym_i16] = ACTIONS(1110),
    [anon_sym_u32] = ACTIONS(1110),
    [anon_sym_i32] = ACTIONS(1110),
    [anon_sym_u64] = ACTIONS(1110),
    [anon_sym_i64] = ACTIONS(1110),
    [anon_sym_u128] = ACTIONS(1110),
    [anon_sym_i128] = ACTIONS(1110),
    [anon_sym_isize] = ACTIONS(1110),
    [anon_sym_usize] = ACTIONS(1110),
    [anon_sym_f32] = ACTIONS(1110),
    [anon_sym_f64] = ACTIONS(1110),
    [anon_sym_bool] = ACTIONS(1110),
    [anon_sym_str] = ACTIONS(1110),
    [anon_sym_char] = ACTIONS(1110),
    [anon_sym_SQUOTE] = ACTIONS(1110),
    [anon_sym_async] = ACTIONS(1110),
    [anon_sym_break] = ACTIONS(1110),
    [anon_sym_const] = ACTIONS(1110),
    [anon_sym_continue] = ACTIONS(1110),
    [anon_sym_default] = ACTIONS(1110),
    [anon_sym_enum] = ACTIONS(1110),
    [anon_sym_fn] = ACTIONS(1110),
    [anon_sym_for] = ACTIONS(1110),
    [anon_sym_if] = ACTIONS(1110),
    [anon_sym_impl] = ACTIONS(1110),
    [anon_sym_let] = ACTIONS(1110),
    [anon_sym_loop] = ACTIONS(1110),
    [anon_sym_match] = ACTIONS(1110),
    [anon_sym_mod] = ACTIONS(1110),
    [anon_sym_pub] = ACTIONS(1110),
    [anon_sym_return] = ACTIONS(1110),
    [anon_sym_static] = ACTIONS(1110),
    [anon_sym_struct] = ACTIONS(1110),
    [anon_sym_trait] = ACTIONS(1110),
    [anon_sym_type] = ACTIONS(1110),
    [anon_sym_union] = ACTIONS(1110),
    [anon_sym_unsafe] = ACTIONS(1110),
    [anon_sym_use] = ACTIONS(1110),
    [anon_sym_while] = ACTIONS(1110),
    [anon_sym_POUND] = ACTIONS(1108),
    [anon_sym_BANG] = ACTIONS(1108),
    [anon_sym_extern] = ACTIONS(1110),
    [anon_sym_LT] = ACTIONS(1108),
    [anon_sym_COLON_COLON] = ACTIONS(1108),
    [anon_sym_AMP] = ACTIONS(1108),
    [anon_sym_DOT_DOT] = ACTIONS(1108),
    [anon_sym_DASH] = ACTIONS(1108),
    [anon_sym_PIPE] = ACTIONS(1108),
    [anon_sym_yield] = ACTIONS(1110),
    [anon_sym_move] = ACTIONS(1110),
    [sym_integer_literal] = ACTIONS(1108),
    [aux_sym_string_literal_token1] = ACTIONS(1108),
    [sym_char_literal] = ACTIONS(1108),
    [anon_sym_true] = ACTIONS(1110),
    [anon_sym_false] = ACTIONS(1110),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1110),
    [sym_super] = ACTIONS(1110),
    [sym_crate] = ACTIONS(1110),
    [sym_metavariable] = ACTIONS(1108),
    [sym_raw_string_literal] = ACTIONS(1108),
    [sym_float_literal] = ACTIONS(1108),
    [sym_block_comment] = ACTIONS(3),
  },
  [260] = {
    [ts_builtin_sym_end] = ACTIONS(1112),
    [sym_identifier] = ACTIONS(1114),
    [anon_sym_SEMI] = ACTIONS(1112),
    [anon_sym_macro_rules_BANG] = ACTIONS(1112),
    [anon_sym_LPAREN] = ACTIONS(1112),
    [anon_sym_LBRACE] = ACTIONS(1112),
    [anon_sym_RBRACE] = ACTIONS(1112),
    [anon_sym_LBRACK] = ACTIONS(1112),
    [anon_sym_STAR] = ACTIONS(1112),
    [anon_sym_u8] = ACTIONS(1114),
    [anon_sym_i8] = ACTIONS(1114),
    [anon_sym_u16] = ACTIONS(1114),
    [anon_sym_i16] = ACTIONS(1114),
    [anon_sym_u32] = ACTIONS(1114),
    [anon_sym_i32] = ACTIONS(1114),
    [anon_sym_u64] = ACTIONS(1114),
    [anon_sym_i64] = ACTIONS(1114),
    [anon_sym_u128] = ACTIONS(1114),
    [anon_sym_i128] = ACTIONS(1114),
    [anon_sym_isize] = ACTIONS(1114),
    [anon_sym_usize] = ACTIONS(1114),
    [anon_sym_f32] = ACTIONS(1114),
    [anon_sym_f64] = ACTIONS(1114),
    [anon_sym_bool] = ACTIONS(1114),
    [anon_sym_str] = ACTIONS(1114),
    [anon_sym_char] = ACTIONS(1114),
    [anon_sym_SQUOTE] = ACTIONS(1114),
    [anon_sym_async] = ACTIONS(1114),
    [anon_sym_break] = ACTIONS(1114),
    [anon_sym_const] = ACTIONS(1114),
    [anon_sym_continue] = ACTIONS(1114),
    [anon_sym_default] = ACTIONS(1114),
    [anon_sym_enum] = ACTIONS(1114),
    [anon_sym_fn] = ACTIONS(1114),
    [anon_sym_for] = ACTIONS(1114),
    [anon_sym_if] = ACTIONS(1114),
    [anon_sym_impl] = ACTIONS(1114),
    [anon_sym_let] = ACTIONS(1114),
    [anon_sym_loop] = ACTIONS(1114),
    [anon_sym_match] = ACTIONS(1114),
    [anon_sym_mod] = ACTIONS(1114),
    [anon_sym_pub] = ACTIONS(1114),
    [anon_sym_return] = ACTIONS(1114),
    [anon_sym_static] = ACTIONS(1114),
    [anon_sym_struct] = ACTIONS(1114),
    [anon_sym_trait] = ACTIONS(1114),
    [anon_sym_type] = ACTIONS(1114),
    [anon_sym_union] = ACTIONS(1114),
    [anon_sym_unsafe] = ACTIONS(1114),
    [anon_sym_use] = ACTIONS(1114),
    [anon_sym_while] = ACTIONS(1114),
    [anon_sym_POUND] = ACTIONS(1112),
    [anon_sym_BANG] = ACTIONS(1112),
    [anon_sym_extern] = ACTIONS(1114),
    [anon_sym_LT] = ACTIONS(1112),
    [anon_sym_COLON_COLON] = ACTIONS(1112),
    [anon_sym_AMP] = ACTIONS(1112),
    [anon_sym_DOT_DOT] = ACTIONS(1112),
    [anon_sym_DASH] = ACTIONS(1112),
    [anon_sym_PIPE] = ACTIONS(1112),
    [anon_sym_yield] = ACTIONS(1114),
    [anon_sym_move] = ACTIONS(1114),
    [sym_integer_literal] = ACTIONS(1112),
    [aux_sym_string_literal_token1] = ACTIONS(1112),
    [sym_char_literal] = ACTIONS(1112),
    [anon_sym_true] = ACTIONS(1114),
    [anon_sym_false] = ACTIONS(1114),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1114),
    [sym_super] = ACTIONS(1114),
    [sym_crate] = ACTIONS(1114),
    [sym_metavariable] = ACTIONS(1112),
    [sym_raw_string_literal] = ACTIONS(1112),
    [sym_float_literal] = ACTIONS(1112),
    [sym_block_comment] = ACTIONS(3),
  },
  [261] = {
    [ts_builtin_sym_end] = ACTIONS(1116),
    [sym_identifier] = ACTIONS(1118),
    [anon_sym_SEMI] = ACTIONS(1116),
    [anon_sym_macro_rules_BANG] = ACTIONS(1116),
    [anon_sym_LPAREN] = ACTIONS(1116),
    [anon_sym_LBRACE] = ACTIONS(1116),
    [anon_sym_RBRACE] = ACTIONS(1116),
    [anon_sym_LBRACK] = ACTIONS(1116),
    [anon_sym_STAR] = ACTIONS(1116),
    [anon_sym_u8] = ACTIONS(1118),
    [anon_sym_i8] = ACTIONS(1118),
    [anon_sym_u16] = ACTIONS(1118),
    [anon_sym_i16] = ACTIONS(1118),
    [anon_sym_u32] = ACTIONS(1118),
    [anon_sym_i32] = ACTIONS(1118),
    [anon_sym_u64] = ACTIONS(1118),
    [anon_sym_i64] = ACTIONS(1118),
    [anon_sym_u128] = ACTIONS(1118),
    [anon_sym_i128] = ACTIONS(1118),
    [anon_sym_isize] = ACTIONS(1118),
    [anon_sym_usize] = ACTIONS(1118),
    [anon_sym_f32] = ACTIONS(1118),
    [anon_sym_f64] = ACTIONS(1118),
    [anon_sym_bool] = ACTIONS(1118),
    [anon_sym_str] = ACTIONS(1118),
    [anon_sym_char] = ACTIONS(1118),
    [anon_sym_SQUOTE] = ACTIONS(1118),
    [anon_sym_async] = ACTIONS(1118),
    [anon_sym_break] = ACTIONS(1118),
    [anon_sym_const] = ACTIONS(1118),
    [anon_sym_continue] = ACTIONS(1118),
    [anon_sym_default] = ACTIONS(1118),
    [anon_sym_enum] = ACTIONS(1118),
    [anon_sym_fn] = ACTIONS(1118),
    [anon_sym_for] = ACTIONS(1118),
    [anon_sym_if] = ACTIONS(1118),
    [anon_sym_impl] = ACTIONS(1118),
    [anon_sym_let] = ACTIONS(1118),
    [anon_sym_loop] = ACTIONS(1118),
    [anon_sym_match] = ACTIONS(1118),
    [anon_sym_mod] = ACTIONS(1118),
    [anon_sym_pub] = ACTIONS(1118),
    [anon_sym_return] = ACTIONS(1118),
    [anon_sym_static] = ACTIONS(1118),
    [anon_sym_struct] = ACTIONS(1118),
    [anon_sym_trait] = ACTIONS(1118),
    [anon_sym_type] = ACTIONS(1118),
    [anon_sym_union] = ACTIONS(1118),
    [anon_sym_unsafe] = ACTIONS(1118),
    [anon_sym_use] = ACTIONS(1118),
    [anon_sym_while] = ACTIONS(1118),
    [anon_sym_POUND] = ACTIONS(1116),
    [anon_sym_BANG] = ACTIONS(1116),
    [anon_sym_extern] = ACTIONS(1118),
    [anon_sym_LT] = ACTIONS(1116),
    [anon_sym_COLON_COLON] = ACTIONS(1116),
    [anon_sym_AMP] = ACTIONS(1116),
    [anon_sym_DOT_DOT] = ACTIONS(1116),
    [anon_sym_DASH] = ACTIONS(1116),
    [anon_sym_PIPE] = ACTIONS(1116),
    [anon_sym_yield] = ACTIONS(1118),
    [anon_sym_move] = ACTIONS(1118),
    [sym_integer_literal] = ACTIONS(1116),
    [aux_sym_string_literal_token1] = ACTIONS(1116),
    [sym_char_literal] = ACTIONS(1116),
    [anon_sym_true] = ACTIONS(1118),
    [anon_sym_false] = ACTIONS(1118),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1118),
    [sym_super] = ACTIONS(1118),
    [sym_crate] = ACTIONS(1118),
    [sym_metavariable] = ACTIONS(1116),
    [sym_raw_string_literal] = ACTIONS(1116),
    [sym_float_literal] = ACTIONS(1116),
    [sym_block_comment] = ACTIONS(3),
  },
  [262] = {
    [ts_builtin_sym_end] = ACTIONS(1120),
    [sym_identifier] = ACTIONS(1122),
    [anon_sym_SEMI] = ACTIONS(1120),
    [anon_sym_macro_rules_BANG] = ACTIONS(1120),
    [anon_sym_LPAREN] = ACTIONS(1120),
    [anon_sym_LBRACE] = ACTIONS(1120),
    [anon_sym_RBRACE] = ACTIONS(1120),
    [anon_sym_LBRACK] = ACTIONS(1120),
    [anon_sym_STAR] = ACTIONS(1120),
    [anon_sym_u8] = ACTIONS(1122),
    [anon_sym_i8] = ACTIONS(1122),
    [anon_sym_u16] = ACTIONS(1122),
    [anon_sym_i16] = ACTIONS(1122),
    [anon_sym_u32] = ACTIONS(1122),
    [anon_sym_i32] = ACTIONS(1122),
    [anon_sym_u64] = ACTIONS(1122),
    [anon_sym_i64] = ACTIONS(1122),
    [anon_sym_u128] = ACTIONS(1122),
    [anon_sym_i128] = ACTIONS(1122),
    [anon_sym_isize] = ACTIONS(1122),
    [anon_sym_usize] = ACTIONS(1122),
    [anon_sym_f32] = ACTIONS(1122),
    [anon_sym_f64] = ACTIONS(1122),
    [anon_sym_bool] = ACTIONS(1122),
    [anon_sym_str] = ACTIONS(1122),
    [anon_sym_char] = ACTIONS(1122),
    [anon_sym_SQUOTE] = ACTIONS(1122),
    [anon_sym_async] = ACTIONS(1122),
    [anon_sym_break] = ACTIONS(1122),
    [anon_sym_const] = ACTIONS(1122),
    [anon_sym_continue] = ACTIONS(1122),
    [anon_sym_default] = ACTIONS(1122),
    [anon_sym_enum] = ACTIONS(1122),
    [anon_sym_fn] = ACTIONS(1122),
    [anon_sym_for] = ACTIONS(1122),
    [anon_sym_if] = ACTIONS(1122),
    [anon_sym_impl] = ACTIONS(1122),
    [anon_sym_let] = ACTIONS(1122),
    [anon_sym_loop] = ACTIONS(1122),
    [anon_sym_match] = ACTIONS(1122),
    [anon_sym_mod] = ACTIONS(1122),
    [anon_sym_pub] = ACTIONS(1122),
    [anon_sym_return] = ACTIONS(1122),
    [anon_sym_static] = ACTIONS(1122),
    [anon_sym_struct] = ACTIONS(1122),
    [anon_sym_trait] = ACTIONS(1122),
    [anon_sym_type] = ACTIONS(1122),
    [anon_sym_union] = ACTIONS(1122),
    [anon_sym_unsafe] = ACTIONS(1122),
    [anon_sym_use] = ACTIONS(1122),
    [anon_sym_while] = ACTIONS(1122),
    [anon_sym_POUND] = ACTIONS(1120),
    [anon_sym_BANG] = ACTIONS(1120),
    [anon_sym_extern] = ACTIONS(1122),
    [anon_sym_LT] = ACTIONS(1120),
    [anon_sym_COLON_COLON] = ACTIONS(1120),
    [anon_sym_AMP] = ACTIONS(1120),
    [anon_sym_DOT_DOT] = ACTIONS(1120),
    [anon_sym_DASH] = ACTIONS(1120),
    [anon_sym_PIPE] = ACTIONS(1120),
    [anon_sym_yield] = ACTIONS(1122),
    [anon_sym_move] = ACTIONS(1122),
    [sym_integer_literal] = ACTIONS(1120),
    [aux_sym_string_literal_token1] = ACTIONS(1120),
    [sym_char_literal] = ACTIONS(1120),
    [anon_sym_true] = ACTIONS(1122),
    [anon_sym_false] = ACTIONS(1122),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1122),
    [sym_super] = ACTIONS(1122),
    [sym_crate] = ACTIONS(1122),
    [sym_metavariable] = ACTIONS(1120),
    [sym_raw_string_literal] = ACTIONS(1120),
    [sym_float_literal] = ACTIONS(1120),
    [sym_block_comment] = ACTIONS(3),
  },
  [263] = {
    [ts_builtin_sym_end] = ACTIONS(1124),
    [sym_identifier] = ACTIONS(1126),
    [anon_sym_SEMI] = ACTIONS(1124),
    [anon_sym_macro_rules_BANG] = ACTIONS(1124),
    [anon_sym_LPAREN] = ACTIONS(1124),
    [anon_sym_LBRACE] = ACTIONS(1124),
    [anon_sym_RBRACE] = ACTIONS(1124),
    [anon_sym_LBRACK] = ACTIONS(1124),
    [anon_sym_STAR] = ACTIONS(1124),
    [anon_sym_u8] = ACTIONS(1126),
    [anon_sym_i8] = ACTIONS(1126),
    [anon_sym_u16] = ACTIONS(1126),
    [anon_sym_i16] = ACTIONS(1126),
    [anon_sym_u32] = ACTIONS(1126),
    [anon_sym_i32] = ACTIONS(1126),
    [anon_sym_u64] = ACTIONS(1126),
    [anon_sym_i64] = ACTIONS(1126),
    [anon_sym_u128] = ACTIONS(1126),
    [anon_sym_i128] = ACTIONS(1126),
    [anon_sym_isize] = ACTIONS(1126),
    [anon_sym_usize] = ACTIONS(1126),
    [anon_sym_f32] = ACTIONS(1126),
    [anon_sym_f64] = ACTIONS(1126),
    [anon_sym_bool] = ACTIONS(1126),
    [anon_sym_str] = ACTIONS(1126),
    [anon_sym_char] = ACTIONS(1126),
    [anon_sym_SQUOTE] = ACTIONS(1126),
    [anon_sym_async] = ACTIONS(1126),
    [anon_sym_break] = ACTIONS(1126),
    [anon_sym_const] = ACTIONS(1126),
    [anon_sym_continue] = ACTIONS(1126),
    [anon_sym_default] = ACTIONS(1126),
    [anon_sym_enum] = ACTIONS(1126),
    [anon_sym_fn] = ACTIONS(1126),
    [anon_sym_for] = ACTIONS(1126),
    [anon_sym_if] = ACTIONS(1126),
    [anon_sym_impl] = ACTIONS(1126),
    [anon_sym_let] = ACTIONS(1126),
    [anon_sym_loop] = ACTIONS(1126),
    [anon_sym_match] = ACTIONS(1126),
    [anon_sym_mod] = ACTIONS(1126),
    [anon_sym_pub] = ACTIONS(1126),
    [anon_sym_return] = ACTIONS(1126),
    [anon_sym_static] = ACTIONS(1126),
    [anon_sym_struct] = ACTIONS(1126),
    [anon_sym_trait] = ACTIONS(1126),
    [anon_sym_type] = ACTIONS(1126),
    [anon_sym_union] = ACTIONS(1126),
    [anon_sym_unsafe] = ACTIONS(1126),
    [anon_sym_use] = ACTIONS(1126),
    [anon_sym_while] = ACTIONS(1126),
    [anon_sym_POUND] = ACTIONS(1124),
    [anon_sym_BANG] = ACTIONS(1124),
    [anon_sym_extern] = ACTIONS(1126),
    [anon_sym_LT] = ACTIONS(1124),
    [anon_sym_COLON_COLON] = ACTIONS(1124),
    [anon_sym_AMP] = ACTIONS(1124),
    [anon_sym_DOT_DOT] = ACTIONS(1124),
    [anon_sym_DASH] = ACTIONS(1124),
    [anon_sym_PIPE] = ACTIONS(1124),
    [anon_sym_yield] = ACTIONS(1126),
    [anon_sym_move] = ACTIONS(1126),
    [sym_integer_literal] = ACTIONS(1124),
    [aux_sym_string_literal_token1] = ACTIONS(1124),
    [sym_char_literal] = ACTIONS(1124),
    [anon_sym_true] = ACTIONS(1126),
    [anon_sym_false] = ACTIONS(1126),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1126),
    [sym_super] = ACTIONS(1126),
    [sym_crate] = ACTIONS(1126),
    [sym_metavariable] = ACTIONS(1124),
    [sym_raw_string_literal] = ACTIONS(1124),
    [sym_float_literal] = ACTIONS(1124),
    [sym_block_comment] = ACTIONS(3),
  },
  [264] = {
    [ts_builtin_sym_end] = ACTIONS(1128),
    [sym_identifier] = ACTIONS(1130),
    [anon_sym_SEMI] = ACTIONS(1128),
    [anon_sym_macro_rules_BANG] = ACTIONS(1128),
    [anon_sym_LPAREN] = ACTIONS(1128),
    [anon_sym_LBRACE] = ACTIONS(1128),
    [anon_sym_RBRACE] = ACTIONS(1128),
    [anon_sym_LBRACK] = ACTIONS(1128),
    [anon_sym_STAR] = ACTIONS(1128),
    [anon_sym_u8] = ACTIONS(1130),
    [anon_sym_i8] = ACTIONS(1130),
    [anon_sym_u16] = ACTIONS(1130),
    [anon_sym_i16] = ACTIONS(1130),
    [anon_sym_u32] = ACTIONS(1130),
    [anon_sym_i32] = ACTIONS(1130),
    [anon_sym_u64] = ACTIONS(1130),
    [anon_sym_i64] = ACTIONS(1130),
    [anon_sym_u128] = ACTIONS(1130),
    [anon_sym_i128] = ACTIONS(1130),
    [anon_sym_isize] = ACTIONS(1130),
    [anon_sym_usize] = ACTIONS(1130),
    [anon_sym_f32] = ACTIONS(1130),
    [anon_sym_f64] = ACTIONS(1130),
    [anon_sym_bool] = ACTIONS(1130),
    [anon_sym_str] = ACTIONS(1130),
    [anon_sym_char] = ACTIONS(1130),
    [anon_sym_SQUOTE] = ACTIONS(1130),
    [anon_sym_async] = ACTIONS(1130),
    [anon_sym_break] = ACTIONS(1130),
    [anon_sym_const] = ACTIONS(1130),
    [anon_sym_continue] = ACTIONS(1130),
    [anon_sym_default] = ACTIONS(1130),
    [anon_sym_enum] = ACTIONS(1130),
    [anon_sym_fn] = ACTIONS(1130),
    [anon_sym_for] = ACTIONS(1130),
    [anon_sym_if] = ACTIONS(1130),
    [anon_sym_impl] = ACTIONS(1130),
    [anon_sym_let] = ACTIONS(1130),
    [anon_sym_loop] = ACTIONS(1130),
    [anon_sym_match] = ACTIONS(1130),
    [anon_sym_mod] = ACTIONS(1130),
    [anon_sym_pub] = ACTIONS(1130),
    [anon_sym_return] = ACTIONS(1130),
    [anon_sym_static] = ACTIONS(1130),
    [anon_sym_struct] = ACTIONS(1130),
    [anon_sym_trait] = ACTIONS(1130),
    [anon_sym_type] = ACTIONS(1130),
    [anon_sym_union] = ACTIONS(1130),
    [anon_sym_unsafe] = ACTIONS(1130),
    [anon_sym_use] = ACTIONS(1130),
    [anon_sym_while] = ACTIONS(1130),
    [anon_sym_POUND] = ACTIONS(1128),
    [anon_sym_BANG] = ACTIONS(1128),
    [anon_sym_extern] = ACTIONS(1130),
    [anon_sym_LT] = ACTIONS(1128),
    [anon_sym_COLON_COLON] = ACTIONS(1128),
    [anon_sym_AMP] = ACTIONS(1128),
    [anon_sym_DOT_DOT] = ACTIONS(1128),
    [anon_sym_DASH] = ACTIONS(1128),
    [anon_sym_PIPE] = ACTIONS(1128),
    [anon_sym_yield] = ACTIONS(1130),
    [anon_sym_move] = ACTIONS(1130),
    [sym_integer_literal] = ACTIONS(1128),
    [aux_sym_string_literal_token1] = ACTIONS(1128),
    [sym_char_literal] = ACTIONS(1128),
    [anon_sym_true] = ACTIONS(1130),
    [anon_sym_false] = ACTIONS(1130),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1130),
    [sym_super] = ACTIONS(1130),
    [sym_crate] = ACTIONS(1130),
    [sym_metavariable] = ACTIONS(1128),
    [sym_raw_string_literal] = ACTIONS(1128),
    [sym_float_literal] = ACTIONS(1128),
    [sym_block_comment] = ACTIONS(3),
  },
  [265] = {
    [ts_builtin_sym_end] = ACTIONS(1132),
    [sym_identifier] = ACTIONS(1134),
    [anon_sym_SEMI] = ACTIONS(1132),
    [anon_sym_macro_rules_BANG] = ACTIONS(1132),
    [anon_sym_LPAREN] = ACTIONS(1132),
    [anon_sym_LBRACE] = ACTIONS(1132),
    [anon_sym_RBRACE] = ACTIONS(1132),
    [anon_sym_LBRACK] = ACTIONS(1132),
    [anon_sym_STAR] = ACTIONS(1132),
    [anon_sym_u8] = ACTIONS(1134),
    [anon_sym_i8] = ACTIONS(1134),
    [anon_sym_u16] = ACTIONS(1134),
    [anon_sym_i16] = ACTIONS(1134),
    [anon_sym_u32] = ACTIONS(1134),
    [anon_sym_i32] = ACTIONS(1134),
    [anon_sym_u64] = ACTIONS(1134),
    [anon_sym_i64] = ACTIONS(1134),
    [anon_sym_u128] = ACTIONS(1134),
    [anon_sym_i128] = ACTIONS(1134),
    [anon_sym_isize] = ACTIONS(1134),
    [anon_sym_usize] = ACTIONS(1134),
    [anon_sym_f32] = ACTIONS(1134),
    [anon_sym_f64] = ACTIONS(1134),
    [anon_sym_bool] = ACTIONS(1134),
    [anon_sym_str] = ACTIONS(1134),
    [anon_sym_char] = ACTIONS(1134),
    [anon_sym_SQUOTE] = ACTIONS(1134),
    [anon_sym_async] = ACTIONS(1134),
    [anon_sym_break] = ACTIONS(1134),
    [anon_sym_const] = ACTIONS(1134),
    [anon_sym_continue] = ACTIONS(1134),
    [anon_sym_default] = ACTIONS(1134),
    [anon_sym_enum] = ACTIONS(1134),
    [anon_sym_fn] = ACTIONS(1134),
    [anon_sym_for] = ACTIONS(1134),
    [anon_sym_if] = ACTIONS(1134),
    [anon_sym_impl] = ACTIONS(1134),
    [anon_sym_let] = ACTIONS(1134),
    [anon_sym_loop] = ACTIONS(1134),
    [anon_sym_match] = ACTIONS(1134),
    [anon_sym_mod] = ACTIONS(1134),
    [anon_sym_pub] = ACTIONS(1134),
    [anon_sym_return] = ACTIONS(1134),
    [anon_sym_static] = ACTIONS(1134),
    [anon_sym_struct] = ACTIONS(1134),
    [anon_sym_trait] = ACTIONS(1134),
    [anon_sym_type] = ACTIONS(1134),
    [anon_sym_union] = ACTIONS(1134),
    [anon_sym_unsafe] = ACTIONS(1134),
    [anon_sym_use] = ACTIONS(1134),
    [anon_sym_while] = ACTIONS(1134),
    [anon_sym_POUND] = ACTIONS(1132),
    [anon_sym_BANG] = ACTIONS(1132),
    [anon_sym_extern] = ACTIONS(1134),
    [anon_sym_LT] = ACTIONS(1132),
    [anon_sym_COLON_COLON] = ACTIONS(1132),
    [anon_sym_AMP] = ACTIONS(1132),
    [anon_sym_DOT_DOT] = ACTIONS(1132),
    [anon_sym_DASH] = ACTIONS(1132),
    [anon_sym_PIPE] = ACTIONS(1132),
    [anon_sym_yield] = ACTIONS(1134),
    [anon_sym_move] = ACTIONS(1134),
    [sym_integer_literal] = ACTIONS(1132),
    [aux_sym_string_literal_token1] = ACTIONS(1132),
    [sym_char_literal] = ACTIONS(1132),
    [anon_sym_true] = ACTIONS(1134),
    [anon_sym_false] = ACTIONS(1134),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1134),
    [sym_super] = ACTIONS(1134),
    [sym_crate] = ACTIONS(1134),
    [sym_metavariable] = ACTIONS(1132),
    [sym_raw_string_literal] = ACTIONS(1132),
    [sym_float_literal] = ACTIONS(1132),
    [sym_block_comment] = ACTIONS(3),
  },
  [266] = {
    [ts_builtin_sym_end] = ACTIONS(1136),
    [sym_identifier] = ACTIONS(1138),
    [anon_sym_SEMI] = ACTIONS(1136),
    [anon_sym_macro_rules_BANG] = ACTIONS(1136),
    [anon_sym_LPAREN] = ACTIONS(1136),
    [anon_sym_LBRACE] = ACTIONS(1136),
    [anon_sym_RBRACE] = ACTIONS(1136),
    [anon_sym_LBRACK] = ACTIONS(1136),
    [anon_sym_STAR] = ACTIONS(1136),
    [anon_sym_u8] = ACTIONS(1138),
    [anon_sym_i8] = ACTIONS(1138),
    [anon_sym_u16] = ACTIONS(1138),
    [anon_sym_i16] = ACTIONS(1138),
    [anon_sym_u32] = ACTIONS(1138),
    [anon_sym_i32] = ACTIONS(1138),
    [anon_sym_u64] = ACTIONS(1138),
    [anon_sym_i64] = ACTIONS(1138),
    [anon_sym_u128] = ACTIONS(1138),
    [anon_sym_i128] = ACTIONS(1138),
    [anon_sym_isize] = ACTIONS(1138),
    [anon_sym_usize] = ACTIONS(1138),
    [anon_sym_f32] = ACTIONS(1138),
    [anon_sym_f64] = ACTIONS(1138),
    [anon_sym_bool] = ACTIONS(1138),
    [anon_sym_str] = ACTIONS(1138),
    [anon_sym_char] = ACTIONS(1138),
    [anon_sym_SQUOTE] = ACTIONS(1138),
    [anon_sym_async] = ACTIONS(1138),
    [anon_sym_break] = ACTIONS(1138),
    [anon_sym_const] = ACTIONS(1138),
    [anon_sym_continue] = ACTIONS(1138),
    [anon_sym_default] = ACTIONS(1138),
    [anon_sym_enum] = ACTIONS(1138),
    [anon_sym_fn] = ACTIONS(1138),
    [anon_sym_for] = ACTIONS(1138),
    [anon_sym_if] = ACTIONS(1138),
    [anon_sym_impl] = ACTIONS(1138),
    [anon_sym_let] = ACTIONS(1138),
    [anon_sym_loop] = ACTIONS(1138),
    [anon_sym_match] = ACTIONS(1138),
    [anon_sym_mod] = ACTIONS(1138),
    [anon_sym_pub] = ACTIONS(1138),
    [anon_sym_return] = ACTIONS(1138),
    [anon_sym_static] = ACTIONS(1138),
    [anon_sym_struct] = ACTIONS(1138),
    [anon_sym_trait] = ACTIONS(1138),
    [anon_sym_type] = ACTIONS(1138),
    [anon_sym_union] = ACTIONS(1138),
    [anon_sym_unsafe] = ACTIONS(1138),
    [anon_sym_use] = ACTIONS(1138),
    [anon_sym_while] = ACTIONS(1138),
    [anon_sym_POUND] = ACTIONS(1136),
    [anon_sym_BANG] = ACTIONS(1136),
    [anon_sym_extern] = ACTIONS(1138),
    [anon_sym_LT] = ACTIONS(1136),
    [anon_sym_COLON_COLON] = ACTIONS(1136),
    [anon_sym_AMP] = ACTIONS(1136),
    [anon_sym_DOT_DOT] = ACTIONS(1136),
    [anon_sym_DASH] = ACTIONS(1136),
    [anon_sym_PIPE] = ACTIONS(1136),
    [anon_sym_yield] = ACTIONS(1138),
    [anon_sym_move] = ACTIONS(1138),
    [sym_integer_literal] = ACTIONS(1136),
    [aux_sym_string_literal_token1] = ACTIONS(1136),
    [sym_char_literal] = ACTIONS(1136),
    [anon_sym_true] = ACTIONS(1138),
    [anon_sym_false] = ACTIONS(1138),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1138),
    [sym_super] = ACTIONS(1138),
    [sym_crate] = ACTIONS(1138),
    [sym_metavariable] = ACTIONS(1136),
    [sym_raw_string_literal] = ACTIONS(1136),
    [sym_float_literal] = ACTIONS(1136),
    [sym_block_comment] = ACTIONS(3),
  },
  [267] = {
    [ts_builtin_sym_end] = ACTIONS(1140),
    [sym_identifier] = ACTIONS(1142),
    [anon_sym_SEMI] = ACTIONS(1140),
    [anon_sym_macro_rules_BANG] = ACTIONS(1140),
    [anon_sym_LPAREN] = ACTIONS(1140),
    [anon_sym_LBRACE] = ACTIONS(1140),
    [anon_sym_RBRACE] = ACTIONS(1140),
    [anon_sym_LBRACK] = ACTIONS(1140),
    [anon_sym_STAR] = ACTIONS(1140),
    [anon_sym_u8] = ACTIONS(1142),
    [anon_sym_i8] = ACTIONS(1142),
    [anon_sym_u16] = ACTIONS(1142),
    [anon_sym_i16] = ACTIONS(1142),
    [anon_sym_u32] = ACTIONS(1142),
    [anon_sym_i32] = ACTIONS(1142),
    [anon_sym_u64] = ACTIONS(1142),
    [anon_sym_i64] = ACTIONS(1142),
    [anon_sym_u128] = ACTIONS(1142),
    [anon_sym_i128] = ACTIONS(1142),
    [anon_sym_isize] = ACTIONS(1142),
    [anon_sym_usize] = ACTIONS(1142),
    [anon_sym_f32] = ACTIONS(1142),
    [anon_sym_f64] = ACTIONS(1142),
    [anon_sym_bool] = ACTIONS(1142),
    [anon_sym_str] = ACTIONS(1142),
    [anon_sym_char] = ACTIONS(1142),
    [anon_sym_SQUOTE] = ACTIONS(1142),
    [anon_sym_async] = ACTIONS(1142),
    [anon_sym_break] = ACTIONS(1142),
    [anon_sym_const] = ACTIONS(1142),
    [anon_sym_continue] = ACTIONS(1142),
    [anon_sym_default] = ACTIONS(1142),
    [anon_sym_enum] = ACTIONS(1142),
    [anon_sym_fn] = ACTIONS(1142),
    [anon_sym_for] = ACTIONS(1142),
    [anon_sym_if] = ACTIONS(1142),
    [anon_sym_impl] = ACTIONS(1142),
    [anon_sym_let] = ACTIONS(1142),
    [anon_sym_loop] = ACTIONS(1142),
    [anon_sym_match] = ACTIONS(1142),
    [anon_sym_mod] = ACTIONS(1142),
    [anon_sym_pub] = ACTIONS(1142),
    [anon_sym_return] = ACTIONS(1142),
    [anon_sym_static] = ACTIONS(1142),
    [anon_sym_struct] = ACTIONS(1142),
    [anon_sym_trait] = ACTIONS(1142),
    [anon_sym_type] = ACTIONS(1142),
    [anon_sym_union] = ACTIONS(1142),
    [anon_sym_unsafe] = ACTIONS(1142),
    [anon_sym_use] = ACTIONS(1142),
    [anon_sym_while] = ACTIONS(1142),
    [anon_sym_POUND] = ACTIONS(1140),
    [anon_sym_BANG] = ACTIONS(1140),
    [anon_sym_extern] = ACTIONS(1142),
    [anon_sym_LT] = ACTIONS(1140),
    [anon_sym_COLON_COLON] = ACTIONS(1140),
    [anon_sym_AMP] = ACTIONS(1140),
    [anon_sym_DOT_DOT] = ACTIONS(1140),
    [anon_sym_DASH] = ACTIONS(1140),
    [anon_sym_PIPE] = ACTIONS(1140),
    [anon_sym_yield] = ACTIONS(1142),
    [anon_sym_move] = ACTIONS(1142),
    [sym_integer_literal] = ACTIONS(1140),
    [aux_sym_string_literal_token1] = ACTIONS(1140),
    [sym_char_literal] = ACTIONS(1140),
    [anon_sym_true] = ACTIONS(1142),
    [anon_sym_false] = ACTIONS(1142),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1142),
    [sym_super] = ACTIONS(1142),
    [sym_crate] = ACTIONS(1142),
    [sym_metavariable] = ACTIONS(1140),
    [sym_raw_string_literal] = ACTIONS(1140),
    [sym_float_literal] = ACTIONS(1140),
    [sym_block_comment] = ACTIONS(3),
  },
  [268] = {
    [ts_builtin_sym_end] = ACTIONS(1144),
    [sym_identifier] = ACTIONS(1146),
    [anon_sym_SEMI] = ACTIONS(1144),
    [anon_sym_macro_rules_BANG] = ACTIONS(1144),
    [anon_sym_LPAREN] = ACTIONS(1144),
    [anon_sym_LBRACE] = ACTIONS(1144),
    [anon_sym_RBRACE] = ACTIONS(1144),
    [anon_sym_LBRACK] = ACTIONS(1144),
    [anon_sym_STAR] = ACTIONS(1144),
    [anon_sym_u8] = ACTIONS(1146),
    [anon_sym_i8] = ACTIONS(1146),
    [anon_sym_u16] = ACTIONS(1146),
    [anon_sym_i16] = ACTIONS(1146),
    [anon_sym_u32] = ACTIONS(1146),
    [anon_sym_i32] = ACTIONS(1146),
    [anon_sym_u64] = ACTIONS(1146),
    [anon_sym_i64] = ACTIONS(1146),
    [anon_sym_u128] = ACTIONS(1146),
    [anon_sym_i128] = ACTIONS(1146),
    [anon_sym_isize] = ACTIONS(1146),
    [anon_sym_usize] = ACTIONS(1146),
    [anon_sym_f32] = ACTIONS(1146),
    [anon_sym_f64] = ACTIONS(1146),
    [anon_sym_bool] = ACTIONS(1146),
    [anon_sym_str] = ACTIONS(1146),
    [anon_sym_char] = ACTIONS(1146),
    [anon_sym_SQUOTE] = ACTIONS(1146),
    [anon_sym_async] = ACTIONS(1146),
    [anon_sym_break] = ACTIONS(1146),
    [anon_sym_const] = ACTIONS(1146),
    [anon_sym_continue] = ACTIONS(1146),
    [anon_sym_default] = ACTIONS(1146),
    [anon_sym_enum] = ACTIONS(1146),
    [anon_sym_fn] = ACTIONS(1146),
    [anon_sym_for] = ACTIONS(1146),
    [anon_sym_if] = ACTIONS(1146),
    [anon_sym_impl] = ACTIONS(1146),
    [anon_sym_let] = ACTIONS(1146),
    [anon_sym_loop] = ACTIONS(1146),
    [anon_sym_match] = ACTIONS(1146),
    [anon_sym_mod] = ACTIONS(1146),
    [anon_sym_pub] = ACTIONS(1146),
    [anon_sym_return] = ACTIONS(1146),
    [anon_sym_static] = ACTIONS(1146),
    [anon_sym_struct] = ACTIONS(1146),
    [anon_sym_trait] = ACTIONS(1146),
    [anon_sym_type] = ACTIONS(1146),
    [anon_sym_union] = ACTIONS(1146),
    [anon_sym_unsafe] = ACTIONS(1146),
    [anon_sym_use] = ACTIONS(1146),
    [anon_sym_while] = ACTIONS(1146),
    [anon_sym_POUND] = ACTIONS(1144),
    [anon_sym_BANG] = ACTIONS(1144),
    [anon_sym_extern] = ACTIONS(1146),
    [anon_sym_LT] = ACTIONS(1144),
    [anon_sym_COLON_COLON] = ACTIONS(1144),
    [anon_sym_AMP] = ACTIONS(1144),
    [anon_sym_DOT_DOT] = ACTIONS(1144),
    [anon_sym_DASH] = ACTIONS(1144),
    [anon_sym_PIPE] = ACTIONS(1144),
    [anon_sym_yield] = ACTIONS(1146),
    [anon_sym_move] = ACTIONS(1146),
    [sym_integer_literal] = ACTIONS(1144),
    [aux_sym_string_literal_token1] = ACTIONS(1144),
    [sym_char_literal] = ACTIONS(1144),
    [anon_sym_true] = ACTIONS(1146),
    [anon_sym_false] = ACTIONS(1146),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1146),
    [sym_super] = ACTIONS(1146),
    [sym_crate] = ACTIONS(1146),
    [sym_metavariable] = ACTIONS(1144),
    [sym_raw_string_literal] = ACTIONS(1144),
    [sym_float_literal] = ACTIONS(1144),
    [sym_block_comment] = ACTIONS(3),
  },
  [269] = {
    [ts_builtin_sym_end] = ACTIONS(1148),
    [sym_identifier] = ACTIONS(1150),
    [anon_sym_SEMI] = ACTIONS(1148),
    [anon_sym_macro_rules_BANG] = ACTIONS(1148),
    [anon_sym_LPAREN] = ACTIONS(1148),
    [anon_sym_LBRACE] = ACTIONS(1148),
    [anon_sym_RBRACE] = ACTIONS(1148),
    [anon_sym_LBRACK] = ACTIONS(1148),
    [anon_sym_STAR] = ACTIONS(1148),
    [anon_sym_u8] = ACTIONS(1150),
    [anon_sym_i8] = ACTIONS(1150),
    [anon_sym_u16] = ACTIONS(1150),
    [anon_sym_i16] = ACTIONS(1150),
    [anon_sym_u32] = ACTIONS(1150),
    [anon_sym_i32] = ACTIONS(1150),
    [anon_sym_u64] = ACTIONS(1150),
    [anon_sym_i64] = ACTIONS(1150),
    [anon_sym_u128] = ACTIONS(1150),
    [anon_sym_i128] = ACTIONS(1150),
    [anon_sym_isize] = ACTIONS(1150),
    [anon_sym_usize] = ACTIONS(1150),
    [anon_sym_f32] = ACTIONS(1150),
    [anon_sym_f64] = ACTIONS(1150),
    [anon_sym_bool] = ACTIONS(1150),
    [anon_sym_str] = ACTIONS(1150),
    [anon_sym_char] = ACTIONS(1150),
    [anon_sym_SQUOTE] = ACTIONS(1150),
    [anon_sym_async] = ACTIONS(1150),
    [anon_sym_break] = ACTIONS(1150),
    [anon_sym_const] = ACTIONS(1150),
    [anon_sym_continue] = ACTIONS(1150),
    [anon_sym_default] = ACTIONS(1150),
    [anon_sym_enum] = ACTIONS(1150),
    [anon_sym_fn] = ACTIONS(1150),
    [anon_sym_for] = ACTIONS(1150),
    [anon_sym_if] = ACTIONS(1150),
    [anon_sym_impl] = ACTIONS(1150),
    [anon_sym_let] = ACTIONS(1150),
    [anon_sym_loop] = ACTIONS(1150),
    [anon_sym_match] = ACTIONS(1150),
    [anon_sym_mod] = ACTIONS(1150),
    [anon_sym_pub] = ACTIONS(1150),
    [anon_sym_return] = ACTIONS(1150),
    [anon_sym_static] = ACTIONS(1150),
    [anon_sym_struct] = ACTIONS(1150),
    [anon_sym_trait] = ACTIONS(1150),
    [anon_sym_type] = ACTIONS(1150),
    [anon_sym_union] = ACTIONS(1150),
    [anon_sym_unsafe] = ACTIONS(1150),
    [anon_sym_use] = ACTIONS(1150),
    [anon_sym_while] = ACTIONS(1150),
    [anon_sym_POUND] = ACTIONS(1148),
    [anon_sym_BANG] = ACTIONS(1148),
    [anon_sym_extern] = ACTIONS(1150),
    [anon_sym_LT] = ACTIONS(1148),
    [anon_sym_COLON_COLON] = ACTIONS(1148),
    [anon_sym_AMP] = ACTIONS(1148),
    [anon_sym_DOT_DOT] = ACTIONS(1148),
    [anon_sym_DASH] = ACTIONS(1148),
    [anon_sym_PIPE] = ACTIONS(1148),
    [anon_sym_yield] = ACTIONS(1150),
    [anon_sym_move] = ACTIONS(1150),
    [sym_integer_literal] = ACTIONS(1148),
    [aux_sym_string_literal_token1] = ACTIONS(1148),
    [sym_char_literal] = ACTIONS(1148),
    [anon_sym_true] = ACTIONS(1150),
    [anon_sym_false] = ACTIONS(1150),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1150),
    [sym_super] = ACTIONS(1150),
    [sym_crate] = ACTIONS(1150),
    [sym_metavariable] = ACTIONS(1148),
    [sym_raw_string_literal] = ACTIONS(1148),
    [sym_float_literal] = ACTIONS(1148),
    [sym_block_comment] = ACTIONS(3),
  },
  [270] = {
    [ts_builtin_sym_end] = ACTIONS(1152),
    [sym_identifier] = ACTIONS(1154),
    [anon_sym_SEMI] = ACTIONS(1152),
    [anon_sym_macro_rules_BANG] = ACTIONS(1152),
    [anon_sym_LPAREN] = ACTIONS(1152),
    [anon_sym_LBRACE] = ACTIONS(1152),
    [anon_sym_RBRACE] = ACTIONS(1152),
    [anon_sym_LBRACK] = ACTIONS(1152),
    [anon_sym_STAR] = ACTIONS(1152),
    [anon_sym_u8] = ACTIONS(1154),
    [anon_sym_i8] = ACTIONS(1154),
    [anon_sym_u16] = ACTIONS(1154),
    [anon_sym_i16] = ACTIONS(1154),
    [anon_sym_u32] = ACTIONS(1154),
    [anon_sym_i32] = ACTIONS(1154),
    [anon_sym_u64] = ACTIONS(1154),
    [anon_sym_i64] = ACTIONS(1154),
    [anon_sym_u128] = ACTIONS(1154),
    [anon_sym_i128] = ACTIONS(1154),
    [anon_sym_isize] = ACTIONS(1154),
    [anon_sym_usize] = ACTIONS(1154),
    [anon_sym_f32] = ACTIONS(1154),
    [anon_sym_f64] = ACTIONS(1154),
    [anon_sym_bool] = ACTIONS(1154),
    [anon_sym_str] = ACTIONS(1154),
    [anon_sym_char] = ACTIONS(1154),
    [anon_sym_SQUOTE] = ACTIONS(1154),
    [anon_sym_async] = ACTIONS(1154),
    [anon_sym_break] = ACTIONS(1154),
    [anon_sym_const] = ACTIONS(1154),
    [anon_sym_continue] = ACTIONS(1154),
    [anon_sym_default] = ACTIONS(1154),
    [anon_sym_enum] = ACTIONS(1154),
    [anon_sym_fn] = ACTIONS(1154),
    [anon_sym_for] = ACTIONS(1154),
    [anon_sym_if] = ACTIONS(1154),
    [anon_sym_impl] = ACTIONS(1154),
    [anon_sym_let] = ACTIONS(1154),
    [anon_sym_loop] = ACTIONS(1154),
    [anon_sym_match] = ACTIONS(1154),
    [anon_sym_mod] = ACTIONS(1154),
    [anon_sym_pub] = ACTIONS(1154),
    [anon_sym_return] = ACTIONS(1154),
    [anon_sym_static] = ACTIONS(1154),
    [anon_sym_struct] = ACTIONS(1154),
    [anon_sym_trait] = ACTIONS(1154),
    [anon_sym_type] = ACTIONS(1154),
    [anon_sym_union] = ACTIONS(1154),
    [anon_sym_unsafe] = ACTIONS(1154),
    [anon_sym_use] = ACTIONS(1154),
    [anon_sym_while] = ACTIONS(1154),
    [anon_sym_POUND] = ACTIONS(1152),
    [anon_sym_BANG] = ACTIONS(1152),
    [anon_sym_extern] = ACTIONS(1154),
    [anon_sym_LT] = ACTIONS(1152),
    [anon_sym_COLON_COLON] = ACTIONS(1152),
    [anon_sym_AMP] = ACTIONS(1152),
    [anon_sym_DOT_DOT] = ACTIONS(1152),
    [anon_sym_DASH] = ACTIONS(1152),
    [anon_sym_PIPE] = ACTIONS(1152),
    [anon_sym_yield] = ACTIONS(1154),
    [anon_sym_move] = ACTIONS(1154),
    [sym_integer_literal] = ACTIONS(1152),
    [aux_sym_string_literal_token1] = ACTIONS(1152),
    [sym_char_literal] = ACTIONS(1152),
    [anon_sym_true] = ACTIONS(1154),
    [anon_sym_false] = ACTIONS(1154),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1154),
    [sym_super] = ACTIONS(1154),
    [sym_crate] = ACTIONS(1154),
    [sym_metavariable] = ACTIONS(1152),
    [sym_raw_string_literal] = ACTIONS(1152),
    [sym_float_literal] = ACTIONS(1152),
    [sym_block_comment] = ACTIONS(3),
  },
  [271] = {
    [ts_builtin_sym_end] = ACTIONS(1156),
    [sym_identifier] = ACTIONS(1158),
    [anon_sym_SEMI] = ACTIONS(1156),
    [anon_sym_macro_rules_BANG] = ACTIONS(1156),
    [anon_sym_LPAREN] = ACTIONS(1156),
    [anon_sym_LBRACE] = ACTIONS(1156),
    [anon_sym_RBRACE] = ACTIONS(1156),
    [anon_sym_LBRACK] = ACTIONS(1156),
    [anon_sym_STAR] = ACTIONS(1156),
    [anon_sym_u8] = ACTIONS(1158),
    [anon_sym_i8] = ACTIONS(1158),
    [anon_sym_u16] = ACTIONS(1158),
    [anon_sym_i16] = ACTIONS(1158),
    [anon_sym_u32] = ACTIONS(1158),
    [anon_sym_i32] = ACTIONS(1158),
    [anon_sym_u64] = ACTIONS(1158),
    [anon_sym_i64] = ACTIONS(1158),
    [anon_sym_u128] = ACTIONS(1158),
    [anon_sym_i128] = ACTIONS(1158),
    [anon_sym_isize] = ACTIONS(1158),
    [anon_sym_usize] = ACTIONS(1158),
    [anon_sym_f32] = ACTIONS(1158),
    [anon_sym_f64] = ACTIONS(1158),
    [anon_sym_bool] = ACTIONS(1158),
    [anon_sym_str] = ACTIONS(1158),
    [anon_sym_char] = ACTIONS(1158),
    [anon_sym_SQUOTE] = ACTIONS(1158),
    [anon_sym_async] = ACTIONS(1158),
    [anon_sym_break] = ACTIONS(1158),
    [anon_sym_const] = ACTIONS(1158),
    [anon_sym_continue] = ACTIONS(1158),
    [anon_sym_default] = ACTIONS(1158),
    [anon_sym_enum] = ACTIONS(1158),
    [anon_sym_fn] = ACTIONS(1158),
    [anon_sym_for] = ACTIONS(1158),
    [anon_sym_if] = ACTIONS(1158),
    [anon_sym_impl] = ACTIONS(1158),
    [anon_sym_let] = ACTIONS(1158),
    [anon_sym_loop] = ACTIONS(1158),
    [anon_sym_match] = ACTIONS(1158),
    [anon_sym_mod] = ACTIONS(1158),
    [anon_sym_pub] = ACTIONS(1158),
    [anon_sym_return] = ACTIONS(1158),
    [anon_sym_static] = ACTIONS(1158),
    [anon_sym_struct] = ACTIONS(1158),
    [anon_sym_trait] = ACTIONS(1158),
    [anon_sym_type] = ACTIONS(1158),
    [anon_sym_union] = ACTIONS(1158),
    [anon_sym_unsafe] = ACTIONS(1158),
    [anon_sym_use] = ACTIONS(1158),
    [anon_sym_while] = ACTIONS(1158),
    [anon_sym_POUND] = ACTIONS(1156),
    [anon_sym_BANG] = ACTIONS(1156),
    [anon_sym_extern] = ACTIONS(1158),
    [anon_sym_LT] = ACTIONS(1156),
    [anon_sym_COLON_COLON] = ACTIONS(1156),
    [anon_sym_AMP] = ACTIONS(1156),
    [anon_sym_DOT_DOT] = ACTIONS(1156),
    [anon_sym_DASH] = ACTIONS(1156),
    [anon_sym_PIPE] = ACTIONS(1156),
    [anon_sym_yield] = ACTIONS(1158),
    [anon_sym_move] = ACTIONS(1158),
    [sym_integer_literal] = ACTIONS(1156),
    [aux_sym_string_literal_token1] = ACTIONS(1156),
    [sym_char_literal] = ACTIONS(1156),
    [anon_sym_true] = ACTIONS(1158),
    [anon_sym_false] = ACTIONS(1158),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1158),
    [sym_super] = ACTIONS(1158),
    [sym_crate] = ACTIONS(1158),
    [sym_metavariable] = ACTIONS(1156),
    [sym_raw_string_literal] = ACTIONS(1156),
    [sym_float_literal] = ACTIONS(1156),
    [sym_block_comment] = ACTIONS(3),
  },
  [272] = {
    [ts_builtin_sym_end] = ACTIONS(1160),
    [sym_identifier] = ACTIONS(1162),
    [anon_sym_SEMI] = ACTIONS(1160),
    [anon_sym_macro_rules_BANG] = ACTIONS(1160),
    [anon_sym_LPAREN] = ACTIONS(1160),
    [anon_sym_LBRACE] = ACTIONS(1160),
    [anon_sym_RBRACE] = ACTIONS(1160),
    [anon_sym_LBRACK] = ACTIONS(1160),
    [anon_sym_STAR] = ACTIONS(1160),
    [anon_sym_u8] = ACTIONS(1162),
    [anon_sym_i8] = ACTIONS(1162),
    [anon_sym_u16] = ACTIONS(1162),
    [anon_sym_i16] = ACTIONS(1162),
    [anon_sym_u32] = ACTIONS(1162),
    [anon_sym_i32] = ACTIONS(1162),
    [anon_sym_u64] = ACTIONS(1162),
    [anon_sym_i64] = ACTIONS(1162),
    [anon_sym_u128] = ACTIONS(1162),
    [anon_sym_i128] = ACTIONS(1162),
    [anon_sym_isize] = ACTIONS(1162),
    [anon_sym_usize] = ACTIONS(1162),
    [anon_sym_f32] = ACTIONS(1162),
    [anon_sym_f64] = ACTIONS(1162),
    [anon_sym_bool] = ACTIONS(1162),
    [anon_sym_str] = ACTIONS(1162),
    [anon_sym_char] = ACTIONS(1162),
    [anon_sym_SQUOTE] = ACTIONS(1162),
    [anon_sym_async] = ACTIONS(1162),
    [anon_sym_break] = ACTIONS(1162),
    [anon_sym_const] = ACTIONS(1162),
    [anon_sym_continue] = ACTIONS(1162),
    [anon_sym_default] = ACTIONS(1162),
    [anon_sym_enum] = ACTIONS(1162),
    [anon_sym_fn] = ACTIONS(1162),
    [anon_sym_for] = ACTIONS(1162),
    [anon_sym_if] = ACTIONS(1162),
    [anon_sym_impl] = ACTIONS(1162),
    [anon_sym_let] = ACTIONS(1162),
    [anon_sym_loop] = ACTIONS(1162),
    [anon_sym_match] = ACTIONS(1162),
    [anon_sym_mod] = ACTIONS(1162),
    [anon_sym_pub] = ACTIONS(1162),
    [anon_sym_return] = ACTIONS(1162),
    [anon_sym_static] = ACTIONS(1162),
    [anon_sym_struct] = ACTIONS(1162),
    [anon_sym_trait] = ACTIONS(1162),
    [anon_sym_type] = ACTIONS(1162),
    [anon_sym_union] = ACTIONS(1162),
    [anon_sym_unsafe] = ACTIONS(1162),
    [anon_sym_use] = ACTIONS(1162),
    [anon_sym_while] = ACTIONS(1162),
    [anon_sym_POUND] = ACTIONS(1160),
    [anon_sym_BANG] = ACTIONS(1160),
    [anon_sym_extern] = ACTIONS(1162),
    [anon_sym_LT] = ACTIONS(1160),
    [anon_sym_COLON_COLON] = ACTIONS(1160),
    [anon_sym_AMP] = ACTIONS(1160),
    [anon_sym_DOT_DOT] = ACTIONS(1160),
    [anon_sym_DASH] = ACTIONS(1160),
    [anon_sym_PIPE] = ACTIONS(1160),
    [anon_sym_yield] = ACTIONS(1162),
    [anon_sym_move] = ACTIONS(1162),
    [sym_integer_literal] = ACTIONS(1160),
    [aux_sym_string_literal_token1] = ACTIONS(1160),
    [sym_char_literal] = ACTIONS(1160),
    [anon_sym_true] = ACTIONS(1162),
    [anon_sym_false] = ACTIONS(1162),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1162),
    [sym_super] = ACTIONS(1162),
    [sym_crate] = ACTIONS(1162),
    [sym_metavariable] = ACTIONS(1160),
    [sym_raw_string_literal] = ACTIONS(1160),
    [sym_float_literal] = ACTIONS(1160),
    [sym_block_comment] = ACTIONS(3),
  },
  [273] = {
    [ts_builtin_sym_end] = ACTIONS(1164),
    [sym_identifier] = ACTIONS(1166),
    [anon_sym_SEMI] = ACTIONS(1164),
    [anon_sym_macro_rules_BANG] = ACTIONS(1164),
    [anon_sym_LPAREN] = ACTIONS(1164),
    [anon_sym_LBRACE] = ACTIONS(1164),
    [anon_sym_RBRACE] = ACTIONS(1164),
    [anon_sym_LBRACK] = ACTIONS(1164),
    [anon_sym_STAR] = ACTIONS(1164),
    [anon_sym_u8] = ACTIONS(1166),
    [anon_sym_i8] = ACTIONS(1166),
    [anon_sym_u16] = ACTIONS(1166),
    [anon_sym_i16] = ACTIONS(1166),
    [anon_sym_u32] = ACTIONS(1166),
    [anon_sym_i32] = ACTIONS(1166),
    [anon_sym_u64] = ACTIONS(1166),
    [anon_sym_i64] = ACTIONS(1166),
    [anon_sym_u128] = ACTIONS(1166),
    [anon_sym_i128] = ACTIONS(1166),
    [anon_sym_isize] = ACTIONS(1166),
    [anon_sym_usize] = ACTIONS(1166),
    [anon_sym_f32] = ACTIONS(1166),
    [anon_sym_f64] = ACTIONS(1166),
    [anon_sym_bool] = ACTIONS(1166),
    [anon_sym_str] = ACTIONS(1166),
    [anon_sym_char] = ACTIONS(1166),
    [anon_sym_SQUOTE] = ACTIONS(1166),
    [anon_sym_async] = ACTIONS(1166),
    [anon_sym_break] = ACTIONS(1166),
    [anon_sym_const] = ACTIONS(1166),
    [anon_sym_continue] = ACTIONS(1166),
    [anon_sym_default] = ACTIONS(1166),
    [anon_sym_enum] = ACTIONS(1166),
    [anon_sym_fn] = ACTIONS(1166),
    [anon_sym_for] = ACTIONS(1166),
    [anon_sym_if] = ACTIONS(1166),
    [anon_sym_impl] = ACTIONS(1166),
    [anon_sym_let] = ACTIONS(1166),
    [anon_sym_loop] = ACTIONS(1166),
    [anon_sym_match] = ACTIONS(1166),
    [anon_sym_mod] = ACTIONS(1166),
    [anon_sym_pub] = ACTIONS(1166),
    [anon_sym_return] = ACTIONS(1166),
    [anon_sym_static] = ACTIONS(1166),
    [anon_sym_struct] = ACTIONS(1166),
    [anon_sym_trait] = ACTIONS(1166),
    [anon_sym_type] = ACTIONS(1166),
    [anon_sym_union] = ACTIONS(1166),
    [anon_sym_unsafe] = ACTIONS(1166),
    [anon_sym_use] = ACTIONS(1166),
    [anon_sym_while] = ACTIONS(1166),
    [anon_sym_POUND] = ACTIONS(1164),
    [anon_sym_BANG] = ACTIONS(1164),
    [anon_sym_extern] = ACTIONS(1166),
    [anon_sym_LT] = ACTIONS(1164),
    [anon_sym_COLON_COLON] = ACTIONS(1164),
    [anon_sym_AMP] = ACTIONS(1164),
    [anon_sym_DOT_DOT] = ACTIONS(1164),
    [anon_sym_DASH] = ACTIONS(1164),
    [anon_sym_PIPE] = ACTIONS(1164),
    [anon_sym_yield] = ACTIONS(1166),
    [anon_sym_move] = ACTIONS(1166),
    [sym_integer_literal] = ACTIONS(1164),
    [aux_sym_string_literal_token1] = ACTIONS(1164),
    [sym_char_literal] = ACTIONS(1164),
    [anon_sym_true] = ACTIONS(1166),
    [anon_sym_false] = ACTIONS(1166),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1166),
    [sym_super] = ACTIONS(1166),
    [sym_crate] = ACTIONS(1166),
    [sym_metavariable] = ACTIONS(1164),
    [sym_raw_string_literal] = ACTIONS(1164),
    [sym_float_literal] = ACTIONS(1164),
    [sym_block_comment] = ACTIONS(3),
  },
  [274] = {
    [ts_builtin_sym_end] = ACTIONS(1168),
    [sym_identifier] = ACTIONS(1170),
    [anon_sym_SEMI] = ACTIONS(1168),
    [anon_sym_macro_rules_BANG] = ACTIONS(1168),
    [anon_sym_LPAREN] = ACTIONS(1168),
    [anon_sym_LBRACE] = ACTIONS(1168),
    [anon_sym_RBRACE] = ACTIONS(1168),
    [anon_sym_LBRACK] = ACTIONS(1168),
    [anon_sym_STAR] = ACTIONS(1168),
    [anon_sym_u8] = ACTIONS(1170),
    [anon_sym_i8] = ACTIONS(1170),
    [anon_sym_u16] = ACTIONS(1170),
    [anon_sym_i16] = ACTIONS(1170),
    [anon_sym_u32] = ACTIONS(1170),
    [anon_sym_i32] = ACTIONS(1170),
    [anon_sym_u64] = ACTIONS(1170),
    [anon_sym_i64] = ACTIONS(1170),
    [anon_sym_u128] = ACTIONS(1170),
    [anon_sym_i128] = ACTIONS(1170),
    [anon_sym_isize] = ACTIONS(1170),
    [anon_sym_usize] = ACTIONS(1170),
    [anon_sym_f32] = ACTIONS(1170),
    [anon_sym_f64] = ACTIONS(1170),
    [anon_sym_bool] = ACTIONS(1170),
    [anon_sym_str] = ACTIONS(1170),
    [anon_sym_char] = ACTIONS(1170),
    [anon_sym_SQUOTE] = ACTIONS(1170),
    [anon_sym_async] = ACTIONS(1170),
    [anon_sym_break] = ACTIONS(1170),
    [anon_sym_const] = ACTIONS(1170),
    [anon_sym_continue] = ACTIONS(1170),
    [anon_sym_default] = ACTIONS(1170),
    [anon_sym_enum] = ACTIONS(1170),
    [anon_sym_fn] = ACTIONS(1170),
    [anon_sym_for] = ACTIONS(1170),
    [anon_sym_if] = ACTIONS(1170),
    [anon_sym_impl] = ACTIONS(1170),
    [anon_sym_let] = ACTIONS(1170),
    [anon_sym_loop] = ACTIONS(1170),
    [anon_sym_match] = ACTIONS(1170),
    [anon_sym_mod] = ACTIONS(1170),
    [anon_sym_pub] = ACTIONS(1170),
    [anon_sym_return] = ACTIONS(1170),
    [anon_sym_static] = ACTIONS(1170),
    [anon_sym_struct] = ACTIONS(1170),
    [anon_sym_trait] = ACTIONS(1170),
    [anon_sym_type] = ACTIONS(1170),
    [anon_sym_union] = ACTIONS(1170),
    [anon_sym_unsafe] = ACTIONS(1170),
    [anon_sym_use] = ACTIONS(1170),
    [anon_sym_while] = ACTIONS(1170),
    [anon_sym_POUND] = ACTIONS(1168),
    [anon_sym_BANG] = ACTIONS(1168),
    [anon_sym_extern] = ACTIONS(1170),
    [anon_sym_LT] = ACTIONS(1168),
    [anon_sym_COLON_COLON] = ACTIONS(1168),
    [anon_sym_AMP] = ACTIONS(1168),
    [anon_sym_DOT_DOT] = ACTIONS(1168),
    [anon_sym_DASH] = ACTIONS(1168),
    [anon_sym_PIPE] = ACTIONS(1168),
    [anon_sym_yield] = ACTIONS(1170),
    [anon_sym_move] = ACTIONS(1170),
    [sym_integer_literal] = ACTIONS(1168),
    [aux_sym_string_literal_token1] = ACTIONS(1168),
    [sym_char_literal] = ACTIONS(1168),
    [anon_sym_true] = ACTIONS(1170),
    [anon_sym_false] = ACTIONS(1170),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1170),
    [sym_super] = ACTIONS(1170),
    [sym_crate] = ACTIONS(1170),
    [sym_metavariable] = ACTIONS(1168),
    [sym_raw_string_literal] = ACTIONS(1168),
    [sym_float_literal] = ACTIONS(1168),
    [sym_block_comment] = ACTIONS(3),
  },
  [275] = {
    [ts_builtin_sym_end] = ACTIONS(1172),
    [sym_identifier] = ACTIONS(1174),
    [anon_sym_SEMI] = ACTIONS(1172),
    [anon_sym_macro_rules_BANG] = ACTIONS(1172),
    [anon_sym_LPAREN] = ACTIONS(1172),
    [anon_sym_LBRACE] = ACTIONS(1172),
    [anon_sym_RBRACE] = ACTIONS(1172),
    [anon_sym_LBRACK] = ACTIONS(1172),
    [anon_sym_STAR] = ACTIONS(1172),
    [anon_sym_u8] = ACTIONS(1174),
    [anon_sym_i8] = ACTIONS(1174),
    [anon_sym_u16] = ACTIONS(1174),
    [anon_sym_i16] = ACTIONS(1174),
    [anon_sym_u32] = ACTIONS(1174),
    [anon_sym_i32] = ACTIONS(1174),
    [anon_sym_u64] = ACTIONS(1174),
    [anon_sym_i64] = ACTIONS(1174),
    [anon_sym_u128] = ACTIONS(1174),
    [anon_sym_i128] = ACTIONS(1174),
    [anon_sym_isize] = ACTIONS(1174),
    [anon_sym_usize] = ACTIONS(1174),
    [anon_sym_f32] = ACTIONS(1174),
    [anon_sym_f64] = ACTIONS(1174),
    [anon_sym_bool] = ACTIONS(1174),
    [anon_sym_str] = ACTIONS(1174),
    [anon_sym_char] = ACTIONS(1174),
    [anon_sym_SQUOTE] = ACTIONS(1174),
    [anon_sym_async] = ACTIONS(1174),
    [anon_sym_break] = ACTIONS(1174),
    [anon_sym_const] = ACTIONS(1174),
    [anon_sym_continue] = ACTIONS(1174),
    [anon_sym_default] = ACTIONS(1174),
    [anon_sym_enum] = ACTIONS(1174),
    [anon_sym_fn] = ACTIONS(1174),
    [anon_sym_for] = ACTIONS(1174),
    [anon_sym_if] = ACTIONS(1174),
    [anon_sym_impl] = ACTIONS(1174),
    [anon_sym_let] = ACTIONS(1174),
    [anon_sym_loop] = ACTIONS(1174),
    [anon_sym_match] = ACTIONS(1174),
    [anon_sym_mod] = ACTIONS(1174),
    [anon_sym_pub] = ACTIONS(1174),
    [anon_sym_return] = ACTIONS(1174),
    [anon_sym_static] = ACTIONS(1174),
    [anon_sym_struct] = ACTIONS(1174),
    [anon_sym_trait] = ACTIONS(1174),
    [anon_sym_type] = ACTIONS(1174),
    [anon_sym_union] = ACTIONS(1174),
    [anon_sym_unsafe] = ACTIONS(1174),
    [anon_sym_use] = ACTIONS(1174),
    [anon_sym_while] = ACTIONS(1174),
    [anon_sym_POUND] = ACTIONS(1172),
    [anon_sym_BANG] = ACTIONS(1172),
    [anon_sym_extern] = ACTIONS(1174),
    [anon_sym_LT] = ACTIONS(1172),
    [anon_sym_COLON_COLON] = ACTIONS(1172),
    [anon_sym_AMP] = ACTIONS(1172),
    [anon_sym_DOT_DOT] = ACTIONS(1172),
    [anon_sym_DASH] = ACTIONS(1172),
    [anon_sym_PIPE] = ACTIONS(1172),
    [anon_sym_yield] = ACTIONS(1174),
    [anon_sym_move] = ACTIONS(1174),
    [sym_integer_literal] = ACTIONS(1172),
    [aux_sym_string_literal_token1] = ACTIONS(1172),
    [sym_char_literal] = ACTIONS(1172),
    [anon_sym_true] = ACTIONS(1174),
    [anon_sym_false] = ACTIONS(1174),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1174),
    [sym_super] = ACTIONS(1174),
    [sym_crate] = ACTIONS(1174),
    [sym_metavariable] = ACTIONS(1172),
    [sym_raw_string_literal] = ACTIONS(1172),
    [sym_float_literal] = ACTIONS(1172),
    [sym_block_comment] = ACTIONS(3),
  },
  [276] = {
    [ts_builtin_sym_end] = ACTIONS(1176),
    [sym_identifier] = ACTIONS(1178),
    [anon_sym_SEMI] = ACTIONS(1176),
    [anon_sym_macro_rules_BANG] = ACTIONS(1176),
    [anon_sym_LPAREN] = ACTIONS(1176),
    [anon_sym_LBRACE] = ACTIONS(1176),
    [anon_sym_RBRACE] = ACTIONS(1176),
    [anon_sym_LBRACK] = ACTIONS(1176),
    [anon_sym_STAR] = ACTIONS(1176),
    [anon_sym_u8] = ACTIONS(1178),
    [anon_sym_i8] = ACTIONS(1178),
    [anon_sym_u16] = ACTIONS(1178),
    [anon_sym_i16] = ACTIONS(1178),
    [anon_sym_u32] = ACTIONS(1178),
    [anon_sym_i32] = ACTIONS(1178),
    [anon_sym_u64] = ACTIONS(1178),
    [anon_sym_i64] = ACTIONS(1178),
    [anon_sym_u128] = ACTIONS(1178),
    [anon_sym_i128] = ACTIONS(1178),
    [anon_sym_isize] = ACTIONS(1178),
    [anon_sym_usize] = ACTIONS(1178),
    [anon_sym_f32] = ACTIONS(1178),
    [anon_sym_f64] = ACTIONS(1178),
    [anon_sym_bool] = ACTIONS(1178),
    [anon_sym_str] = ACTIONS(1178),
    [anon_sym_char] = ACTIONS(1178),
    [anon_sym_SQUOTE] = ACTIONS(1178),
    [anon_sym_async] = ACTIONS(1178),
    [anon_sym_break] = ACTIONS(1178),
    [anon_sym_const] = ACTIONS(1178),
    [anon_sym_continue] = ACTIONS(1178),
    [anon_sym_default] = ACTIONS(1178),
    [anon_sym_enum] = ACTIONS(1178),
    [anon_sym_fn] = ACTIONS(1178),
    [anon_sym_for] = ACTIONS(1178),
    [anon_sym_if] = ACTIONS(1178),
    [anon_sym_impl] = ACTIONS(1178),
    [anon_sym_let] = ACTIONS(1178),
    [anon_sym_loop] = ACTIONS(1178),
    [anon_sym_match] = ACTIONS(1178),
    [anon_sym_mod] = ACTIONS(1178),
    [anon_sym_pub] = ACTIONS(1178),
    [anon_sym_return] = ACTIONS(1178),
    [anon_sym_static] = ACTIONS(1178),
    [anon_sym_struct] = ACTIONS(1178),
    [anon_sym_trait] = ACTIONS(1178),
    [anon_sym_type] = ACTIONS(1178),
    [anon_sym_union] = ACTIONS(1178),
    [anon_sym_unsafe] = ACTIONS(1178),
    [anon_sym_use] = ACTIONS(1178),
    [anon_sym_while] = ACTIONS(1178),
    [anon_sym_POUND] = ACTIONS(1176),
    [anon_sym_BANG] = ACTIONS(1176),
    [anon_sym_extern] = ACTIONS(1178),
    [anon_sym_LT] = ACTIONS(1176),
    [anon_sym_COLON_COLON] = ACTIONS(1176),
    [anon_sym_AMP] = ACTIONS(1176),
    [anon_sym_DOT_DOT] = ACTIONS(1176),
    [anon_sym_DASH] = ACTIONS(1176),
    [anon_sym_PIPE] = ACTIONS(1176),
    [anon_sym_yield] = ACTIONS(1178),
    [anon_sym_move] = ACTIONS(1178),
    [sym_integer_literal] = ACTIONS(1176),
    [aux_sym_string_literal_token1] = ACTIONS(1176),
    [sym_char_literal] = ACTIONS(1176),
    [anon_sym_true] = ACTIONS(1178),
    [anon_sym_false] = ACTIONS(1178),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1178),
    [sym_super] = ACTIONS(1178),
    [sym_crate] = ACTIONS(1178),
    [sym_metavariable] = ACTIONS(1176),
    [sym_raw_string_literal] = ACTIONS(1176),
    [sym_float_literal] = ACTIONS(1176),
    [sym_block_comment] = ACTIONS(3),
  },
  [277] = {
    [ts_builtin_sym_end] = ACTIONS(1180),
    [sym_identifier] = ACTIONS(1182),
    [anon_sym_SEMI] = ACTIONS(1180),
    [anon_sym_macro_rules_BANG] = ACTIONS(1180),
    [anon_sym_LPAREN] = ACTIONS(1180),
    [anon_sym_LBRACE] = ACTIONS(1180),
    [anon_sym_RBRACE] = ACTIONS(1180),
    [anon_sym_LBRACK] = ACTIONS(1180),
    [anon_sym_STAR] = ACTIONS(1180),
    [anon_sym_u8] = ACTIONS(1182),
    [anon_sym_i8] = ACTIONS(1182),
    [anon_sym_u16] = ACTIONS(1182),
    [anon_sym_i16] = ACTIONS(1182),
    [anon_sym_u32] = ACTIONS(1182),
    [anon_sym_i32] = ACTIONS(1182),
    [anon_sym_u64] = ACTIONS(1182),
    [anon_sym_i64] = ACTIONS(1182),
    [anon_sym_u128] = ACTIONS(1182),
    [anon_sym_i128] = ACTIONS(1182),
    [anon_sym_isize] = ACTIONS(1182),
    [anon_sym_usize] = ACTIONS(1182),
    [anon_sym_f32] = ACTIONS(1182),
    [anon_sym_f64] = ACTIONS(1182),
    [anon_sym_bool] = ACTIONS(1182),
    [anon_sym_str] = ACTIONS(1182),
    [anon_sym_char] = ACTIONS(1182),
    [anon_sym_SQUOTE] = ACTIONS(1182),
    [anon_sym_async] = ACTIONS(1182),
    [anon_sym_break] = ACTIONS(1182),
    [anon_sym_const] = ACTIONS(1182),
    [anon_sym_continue] = ACTIONS(1182),
    [anon_sym_default] = ACTIONS(1182),
    [anon_sym_enum] = ACTIONS(1182),
    [anon_sym_fn] = ACTIONS(1182),
    [anon_sym_for] = ACTIONS(1182),
    [anon_sym_if] = ACTIONS(1182),
    [anon_sym_impl] = ACTIONS(1182),
    [anon_sym_let] = ACTIONS(1182),
    [anon_sym_loop] = ACTIONS(1182),
    [anon_sym_match] = ACTIONS(1182),
    [anon_sym_mod] = ACTIONS(1182),
    [anon_sym_pub] = ACTIONS(1182),
    [anon_sym_return] = ACTIONS(1182),
    [anon_sym_static] = ACTIONS(1182),
    [anon_sym_struct] = ACTIONS(1182),
    [anon_sym_trait] = ACTIONS(1182),
    [anon_sym_type] = ACTIONS(1182),
    [anon_sym_union] = ACTIONS(1182),
    [anon_sym_unsafe] = ACTIONS(1182),
    [anon_sym_use] = ACTIONS(1182),
    [anon_sym_while] = ACTIONS(1182),
    [anon_sym_POUND] = ACTIONS(1180),
    [anon_sym_BANG] = ACTIONS(1180),
    [anon_sym_extern] = ACTIONS(1182),
    [anon_sym_LT] = ACTIONS(1180),
    [anon_sym_COLON_COLON] = ACTIONS(1180),
    [anon_sym_AMP] = ACTIONS(1180),
    [anon_sym_DOT_DOT] = ACTIONS(1180),
    [anon_sym_DASH] = ACTIONS(1180),
    [anon_sym_PIPE] = ACTIONS(1180),
    [anon_sym_yield] = ACTIONS(1182),
    [anon_sym_move] = ACTIONS(1182),
    [sym_integer_literal] = ACTIONS(1180),
    [aux_sym_string_literal_token1] = ACTIONS(1180),
    [sym_char_literal] = ACTIONS(1180),
    [anon_sym_true] = ACTIONS(1182),
    [anon_sym_false] = ACTIONS(1182),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1182),
    [sym_super] = ACTIONS(1182),
    [sym_crate] = ACTIONS(1182),
    [sym_metavariable] = ACTIONS(1180),
    [sym_raw_string_literal] = ACTIONS(1180),
    [sym_float_literal] = ACTIONS(1180),
    [sym_block_comment] = ACTIONS(3),
  },
  [278] = {
    [ts_builtin_sym_end] = ACTIONS(1184),
    [sym_identifier] = ACTIONS(1186),
    [anon_sym_SEMI] = ACTIONS(1184),
    [anon_sym_macro_rules_BANG] = ACTIONS(1184),
    [anon_sym_LPAREN] = ACTIONS(1184),
    [anon_sym_LBRACE] = ACTIONS(1184),
    [anon_sym_RBRACE] = ACTIONS(1184),
    [anon_sym_LBRACK] = ACTIONS(1184),
    [anon_sym_STAR] = ACTIONS(1184),
    [anon_sym_u8] = ACTIONS(1186),
    [anon_sym_i8] = ACTIONS(1186),
    [anon_sym_u16] = ACTIONS(1186),
    [anon_sym_i16] = ACTIONS(1186),
    [anon_sym_u32] = ACTIONS(1186),
    [anon_sym_i32] = ACTIONS(1186),
    [anon_sym_u64] = ACTIONS(1186),
    [anon_sym_i64] = ACTIONS(1186),
    [anon_sym_u128] = ACTIONS(1186),
    [anon_sym_i128] = ACTIONS(1186),
    [anon_sym_isize] = ACTIONS(1186),
    [anon_sym_usize] = ACTIONS(1186),
    [anon_sym_f32] = ACTIONS(1186),
    [anon_sym_f64] = ACTIONS(1186),
    [anon_sym_bool] = ACTIONS(1186),
    [anon_sym_str] = ACTIONS(1186),
    [anon_sym_char] = ACTIONS(1186),
    [anon_sym_SQUOTE] = ACTIONS(1186),
    [anon_sym_async] = ACTIONS(1186),
    [anon_sym_break] = ACTIONS(1186),
    [anon_sym_const] = ACTIONS(1186),
    [anon_sym_continue] = ACTIONS(1186),
    [anon_sym_default] = ACTIONS(1186),
    [anon_sym_enum] = ACTIONS(1186),
    [anon_sym_fn] = ACTIONS(1186),
    [anon_sym_for] = ACTIONS(1186),
    [anon_sym_if] = ACTIONS(1186),
    [anon_sym_impl] = ACTIONS(1186),
    [anon_sym_let] = ACTIONS(1186),
    [anon_sym_loop] = ACTIONS(1186),
    [anon_sym_match] = ACTIONS(1186),
    [anon_sym_mod] = ACTIONS(1186),
    [anon_sym_pub] = ACTIONS(1186),
    [anon_sym_return] = ACTIONS(1186),
    [anon_sym_static] = ACTIONS(1186),
    [anon_sym_struct] = ACTIONS(1186),
    [anon_sym_trait] = ACTIONS(1186),
    [anon_sym_type] = ACTIONS(1186),
    [anon_sym_union] = ACTIONS(1186),
    [anon_sym_unsafe] = ACTIONS(1186),
    [anon_sym_use] = ACTIONS(1186),
    [anon_sym_while] = ACTIONS(1186),
    [anon_sym_POUND] = ACTIONS(1184),
    [anon_sym_BANG] = ACTIONS(1184),
    [anon_sym_extern] = ACTIONS(1186),
    [anon_sym_LT] = ACTIONS(1184),
    [anon_sym_COLON_COLON] = ACTIONS(1184),
    [anon_sym_AMP] = ACTIONS(1184),
    [anon_sym_DOT_DOT] = ACTIONS(1184),
    [anon_sym_DASH] = ACTIONS(1184),
    [anon_sym_PIPE] = ACTIONS(1184),
    [anon_sym_yield] = ACTIONS(1186),
    [anon_sym_move] = ACTIONS(1186),
    [sym_integer_literal] = ACTIONS(1184),
    [aux_sym_string_literal_token1] = ACTIONS(1184),
    [sym_char_literal] = ACTIONS(1184),
    [anon_sym_true] = ACTIONS(1186),
    [anon_sym_false] = ACTIONS(1186),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1186),
    [sym_super] = ACTIONS(1186),
    [sym_crate] = ACTIONS(1186),
    [sym_metavariable] = ACTIONS(1184),
    [sym_raw_string_literal] = ACTIONS(1184),
    [sym_float_literal] = ACTIONS(1184),
    [sym_block_comment] = ACTIONS(3),
  },
  [279] = {
    [ts_builtin_sym_end] = ACTIONS(1188),
    [sym_identifier] = ACTIONS(1190),
    [anon_sym_SEMI] = ACTIONS(1188),
    [anon_sym_macro_rules_BANG] = ACTIONS(1188),
    [anon_sym_LPAREN] = ACTIONS(1188),
    [anon_sym_LBRACE] = ACTIONS(1188),
    [anon_sym_RBRACE] = ACTIONS(1188),
    [anon_sym_LBRACK] = ACTIONS(1188),
    [anon_sym_STAR] = ACTIONS(1188),
    [anon_sym_u8] = ACTIONS(1190),
    [anon_sym_i8] = ACTIONS(1190),
    [anon_sym_u16] = ACTIONS(1190),
    [anon_sym_i16] = ACTIONS(1190),
    [anon_sym_u32] = ACTIONS(1190),
    [anon_sym_i32] = ACTIONS(1190),
    [anon_sym_u64] = ACTIONS(1190),
    [anon_sym_i64] = ACTIONS(1190),
    [anon_sym_u128] = ACTIONS(1190),
    [anon_sym_i128] = ACTIONS(1190),
    [anon_sym_isize] = ACTIONS(1190),
    [anon_sym_usize] = ACTIONS(1190),
    [anon_sym_f32] = ACTIONS(1190),
    [anon_sym_f64] = ACTIONS(1190),
    [anon_sym_bool] = ACTIONS(1190),
    [anon_sym_str] = ACTIONS(1190),
    [anon_sym_char] = ACTIONS(1190),
    [anon_sym_SQUOTE] = ACTIONS(1190),
    [anon_sym_async] = ACTIONS(1190),
    [anon_sym_break] = ACTIONS(1190),
    [anon_sym_const] = ACTIONS(1190),
    [anon_sym_continue] = ACTIONS(1190),
    [anon_sym_default] = ACTIONS(1190),
    [anon_sym_enum] = ACTIONS(1190),
    [anon_sym_fn] = ACTIONS(1190),
    [anon_sym_for] = ACTIONS(1190),
    [anon_sym_if] = ACTIONS(1190),
    [anon_sym_impl] = ACTIONS(1190),
    [anon_sym_let] = ACTIONS(1190),
    [anon_sym_loop] = ACTIONS(1190),
    [anon_sym_match] = ACTIONS(1190),
    [anon_sym_mod] = ACTIONS(1190),
    [anon_sym_pub] = ACTIONS(1190),
    [anon_sym_return] = ACTIONS(1190),
    [anon_sym_static] = ACTIONS(1190),
    [anon_sym_struct] = ACTIONS(1190),
    [anon_sym_trait] = ACTIONS(1190),
    [anon_sym_type] = ACTIONS(1190),
    [anon_sym_union] = ACTIONS(1190),
    [anon_sym_unsafe] = ACTIONS(1190),
    [anon_sym_use] = ACTIONS(1190),
    [anon_sym_while] = ACTIONS(1190),
    [anon_sym_POUND] = ACTIONS(1188),
    [anon_sym_BANG] = ACTIONS(1188),
    [anon_sym_extern] = ACTIONS(1190),
    [anon_sym_LT] = ACTIONS(1188),
    [anon_sym_COLON_COLON] = ACTIONS(1188),
    [anon_sym_AMP] = ACTIONS(1188),
    [anon_sym_DOT_DOT] = ACTIONS(1188),
    [anon_sym_DASH] = ACTIONS(1188),
    [anon_sym_PIPE] = ACTIONS(1188),
    [anon_sym_yield] = ACTIONS(1190),
    [anon_sym_move] = ACTIONS(1190),
    [sym_integer_literal] = ACTIONS(1188),
    [aux_sym_string_literal_token1] = ACTIONS(1188),
    [sym_char_literal] = ACTIONS(1188),
    [anon_sym_true] = ACTIONS(1190),
    [anon_sym_false] = ACTIONS(1190),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1190),
    [sym_super] = ACTIONS(1190),
    [sym_crate] = ACTIONS(1190),
    [sym_metavariable] = ACTIONS(1188),
    [sym_raw_string_literal] = ACTIONS(1188),
    [sym_float_literal] = ACTIONS(1188),
    [sym_block_comment] = ACTIONS(3),
  },
  [280] = {
    [ts_builtin_sym_end] = ACTIONS(1192),
    [sym_identifier] = ACTIONS(1194),
    [anon_sym_SEMI] = ACTIONS(1192),
    [anon_sym_macro_rules_BANG] = ACTIONS(1192),
    [anon_sym_LPAREN] = ACTIONS(1192),
    [anon_sym_LBRACE] = ACTIONS(1192),
    [anon_sym_RBRACE] = ACTIONS(1192),
    [anon_sym_LBRACK] = ACTIONS(1192),
    [anon_sym_STAR] = ACTIONS(1192),
    [anon_sym_u8] = ACTIONS(1194),
    [anon_sym_i8] = ACTIONS(1194),
    [anon_sym_u16] = ACTIONS(1194),
    [anon_sym_i16] = ACTIONS(1194),
    [anon_sym_u32] = ACTIONS(1194),
    [anon_sym_i32] = ACTIONS(1194),
    [anon_sym_u64] = ACTIONS(1194),
    [anon_sym_i64] = ACTIONS(1194),
    [anon_sym_u128] = ACTIONS(1194),
    [anon_sym_i128] = ACTIONS(1194),
    [anon_sym_isize] = ACTIONS(1194),
    [anon_sym_usize] = ACTIONS(1194),
    [anon_sym_f32] = ACTIONS(1194),
    [anon_sym_f64] = ACTIONS(1194),
    [anon_sym_bool] = ACTIONS(1194),
    [anon_sym_str] = ACTIONS(1194),
    [anon_sym_char] = ACTIONS(1194),
    [anon_sym_SQUOTE] = ACTIONS(1194),
    [anon_sym_async] = ACTIONS(1194),
    [anon_sym_break] = ACTIONS(1194),
    [anon_sym_const] = ACTIONS(1194),
    [anon_sym_continue] = ACTIONS(1194),
    [anon_sym_default] = ACTIONS(1194),
    [anon_sym_enum] = ACTIONS(1194),
    [anon_sym_fn] = ACTIONS(1194),
    [anon_sym_for] = ACTIONS(1194),
    [anon_sym_if] = ACTIONS(1194),
    [anon_sym_impl] = ACTIONS(1194),
    [anon_sym_let] = ACTIONS(1194),
    [anon_sym_loop] = ACTIONS(1194),
    [anon_sym_match] = ACTIONS(1194),
    [anon_sym_mod] = ACTIONS(1194),
    [anon_sym_pub] = ACTIONS(1194),
    [anon_sym_return] = ACTIONS(1194),
    [anon_sym_static] = ACTIONS(1194),
    [anon_sym_struct] = ACTIONS(1194),
    [anon_sym_trait] = ACTIONS(1194),
    [anon_sym_type] = ACTIONS(1194),
    [anon_sym_union] = ACTIONS(1194),
    [anon_sym_unsafe] = ACTIONS(1194),
    [anon_sym_use] = ACTIONS(1194),
    [anon_sym_while] = ACTIONS(1194),
    [anon_sym_POUND] = ACTIONS(1192),
    [anon_sym_BANG] = ACTIONS(1192),
    [anon_sym_extern] = ACTIONS(1194),
    [anon_sym_LT] = ACTIONS(1192),
    [anon_sym_COLON_COLON] = ACTIONS(1192),
    [anon_sym_AMP] = ACTIONS(1192),
    [anon_sym_DOT_DOT] = ACTIONS(1192),
    [anon_sym_DASH] = ACTIONS(1192),
    [anon_sym_PIPE] = ACTIONS(1192),
    [anon_sym_yield] = ACTIONS(1194),
    [anon_sym_move] = ACTIONS(1194),
    [sym_integer_literal] = ACTIONS(1192),
    [aux_sym_string_literal_token1] = ACTIONS(1192),
    [sym_char_literal] = ACTIONS(1192),
    [anon_sym_true] = ACTIONS(1194),
    [anon_sym_false] = ACTIONS(1194),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1194),
    [sym_super] = ACTIONS(1194),
    [sym_crate] = ACTIONS(1194),
    [sym_metavariable] = ACTIONS(1192),
    [sym_raw_string_literal] = ACTIONS(1192),
    [sym_float_literal] = ACTIONS(1192),
    [sym_block_comment] = ACTIONS(3),
  },
  [281] = {
    [ts_builtin_sym_end] = ACTIONS(1196),
    [sym_identifier] = ACTIONS(1198),
    [anon_sym_SEMI] = ACTIONS(1196),
    [anon_sym_macro_rules_BANG] = ACTIONS(1196),
    [anon_sym_LPAREN] = ACTIONS(1196),
    [anon_sym_LBRACE] = ACTIONS(1196),
    [anon_sym_RBRACE] = ACTIONS(1196),
    [anon_sym_LBRACK] = ACTIONS(1196),
    [anon_sym_STAR] = ACTIONS(1196),
    [anon_sym_u8] = ACTIONS(1198),
    [anon_sym_i8] = ACTIONS(1198),
    [anon_sym_u16] = ACTIONS(1198),
    [anon_sym_i16] = ACTIONS(1198),
    [anon_sym_u32] = ACTIONS(1198),
    [anon_sym_i32] = ACTIONS(1198),
    [anon_sym_u64] = ACTIONS(1198),
    [anon_sym_i64] = ACTIONS(1198),
    [anon_sym_u128] = ACTIONS(1198),
    [anon_sym_i128] = ACTIONS(1198),
    [anon_sym_isize] = ACTIONS(1198),
    [anon_sym_usize] = ACTIONS(1198),
    [anon_sym_f32] = ACTIONS(1198),
    [anon_sym_f64] = ACTIONS(1198),
    [anon_sym_bool] = ACTIONS(1198),
    [anon_sym_str] = ACTIONS(1198),
    [anon_sym_char] = ACTIONS(1198),
    [anon_sym_SQUOTE] = ACTIONS(1198),
    [anon_sym_async] = ACTIONS(1198),
    [anon_sym_break] = ACTIONS(1198),
    [anon_sym_const] = ACTIONS(1198),
    [anon_sym_continue] = ACTIONS(1198),
    [anon_sym_default] = ACTIONS(1198),
    [anon_sym_enum] = ACTIONS(1198),
    [anon_sym_fn] = ACTIONS(1198),
    [anon_sym_for] = ACTIONS(1198),
    [anon_sym_if] = ACTIONS(1198),
    [anon_sym_impl] = ACTIONS(1198),
    [anon_sym_let] = ACTIONS(1198),
    [anon_sym_loop] = ACTIONS(1198),
    [anon_sym_match] = ACTIONS(1198),
    [anon_sym_mod] = ACTIONS(1198),
    [anon_sym_pub] = ACTIONS(1198),
    [anon_sym_return] = ACTIONS(1198),
    [anon_sym_static] = ACTIONS(1198),
    [anon_sym_struct] = ACTIONS(1198),
    [anon_sym_trait] = ACTIONS(1198),
    [anon_sym_type] = ACTIONS(1198),
    [anon_sym_union] = ACTIONS(1198),
    [anon_sym_unsafe] = ACTIONS(1198),
    [anon_sym_use] = ACTIONS(1198),
    [anon_sym_while] = ACTIONS(1198),
    [anon_sym_POUND] = ACTIONS(1196),
    [anon_sym_BANG] = ACTIONS(1196),
    [anon_sym_extern] = ACTIONS(1198),
    [anon_sym_LT] = ACTIONS(1196),
    [anon_sym_COLON_COLON] = ACTIONS(1196),
    [anon_sym_AMP] = ACTIONS(1196),
    [anon_sym_DOT_DOT] = ACTIONS(1196),
    [anon_sym_DASH] = ACTIONS(1196),
    [anon_sym_PIPE] = ACTIONS(1196),
    [anon_sym_yield] = ACTIONS(1198),
    [anon_sym_move] = ACTIONS(1198),
    [sym_integer_literal] = ACTIONS(1196),
    [aux_sym_string_literal_token1] = ACTIONS(1196),
    [sym_char_literal] = ACTIONS(1196),
    [anon_sym_true] = ACTIONS(1198),
    [anon_sym_false] = ACTIONS(1198),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1198),
    [sym_super] = ACTIONS(1198),
    [sym_crate] = ACTIONS(1198),
    [sym_metavariable] = ACTIONS(1196),
    [sym_raw_string_literal] = ACTIONS(1196),
    [sym_float_literal] = ACTIONS(1196),
    [sym_block_comment] = ACTIONS(3),
  },
  [282] = {
    [ts_builtin_sym_end] = ACTIONS(1200),
    [sym_identifier] = ACTIONS(1202),
    [anon_sym_SEMI] = ACTIONS(1200),
    [anon_sym_macro_rules_BANG] = ACTIONS(1200),
    [anon_sym_LPAREN] = ACTIONS(1200),
    [anon_sym_LBRACE] = ACTIONS(1200),
    [anon_sym_RBRACE] = ACTIONS(1200),
    [anon_sym_LBRACK] = ACTIONS(1200),
    [anon_sym_STAR] = ACTIONS(1200),
    [anon_sym_u8] = ACTIONS(1202),
    [anon_sym_i8] = ACTIONS(1202),
    [anon_sym_u16] = ACTIONS(1202),
    [anon_sym_i16] = ACTIONS(1202),
    [anon_sym_u32] = ACTIONS(1202),
    [anon_sym_i32] = ACTIONS(1202),
    [anon_sym_u64] = ACTIONS(1202),
    [anon_sym_i64] = ACTIONS(1202),
    [anon_sym_u128] = ACTIONS(1202),
    [anon_sym_i128] = ACTIONS(1202),
    [anon_sym_isize] = ACTIONS(1202),
    [anon_sym_usize] = ACTIONS(1202),
    [anon_sym_f32] = ACTIONS(1202),
    [anon_sym_f64] = ACTIONS(1202),
    [anon_sym_bool] = ACTIONS(1202),
    [anon_sym_str] = ACTIONS(1202),
    [anon_sym_char] = ACTIONS(1202),
    [anon_sym_SQUOTE] = ACTIONS(1202),
    [anon_sym_async] = ACTIONS(1202),
    [anon_sym_break] = ACTIONS(1202),
    [anon_sym_const] = ACTIONS(1202),
    [anon_sym_continue] = ACTIONS(1202),
    [anon_sym_default] = ACTIONS(1202),
    [anon_sym_enum] = ACTIONS(1202),
    [anon_sym_fn] = ACTIONS(1202),
    [anon_sym_for] = ACTIONS(1202),
    [anon_sym_if] = ACTIONS(1202),
    [anon_sym_impl] = ACTIONS(1202),
    [anon_sym_let] = ACTIONS(1202),
    [anon_sym_loop] = ACTIONS(1202),
    [anon_sym_match] = ACTIONS(1202),
    [anon_sym_mod] = ACTIONS(1202),
    [anon_sym_pub] = ACTIONS(1202),
    [anon_sym_return] = ACTIONS(1202),
    [anon_sym_static] = ACTIONS(1202),
    [anon_sym_struct] = ACTIONS(1202),
    [anon_sym_trait] = ACTIONS(1202),
    [anon_sym_type] = ACTIONS(1202),
    [anon_sym_union] = ACTIONS(1202),
    [anon_sym_unsafe] = ACTIONS(1202),
    [anon_sym_use] = ACTIONS(1202),
    [anon_sym_while] = ACTIONS(1202),
    [anon_sym_POUND] = ACTIONS(1200),
    [anon_sym_BANG] = ACTIONS(1200),
    [anon_sym_extern] = ACTIONS(1202),
    [anon_sym_LT] = ACTIONS(1200),
    [anon_sym_COLON_COLON] = ACTIONS(1200),
    [anon_sym_AMP] = ACTIONS(1200),
    [anon_sym_DOT_DOT] = ACTIONS(1200),
    [anon_sym_DASH] = ACTIONS(1200),
    [anon_sym_PIPE] = ACTIONS(1200),
    [anon_sym_yield] = ACTIONS(1202),
    [anon_sym_move] = ACTIONS(1202),
    [sym_integer_literal] = ACTIONS(1200),
    [aux_sym_string_literal_token1] = ACTIONS(1200),
    [sym_char_literal] = ACTIONS(1200),
    [anon_sym_true] = ACTIONS(1202),
    [anon_sym_false] = ACTIONS(1202),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1202),
    [sym_super] = ACTIONS(1202),
    [sym_crate] = ACTIONS(1202),
    [sym_metavariable] = ACTIONS(1200),
    [sym_raw_string_literal] = ACTIONS(1200),
    [sym_float_literal] = ACTIONS(1200),
    [sym_block_comment] = ACTIONS(3),
  },
  [283] = {
    [ts_builtin_sym_end] = ACTIONS(1204),
    [sym_identifier] = ACTIONS(1206),
    [anon_sym_SEMI] = ACTIONS(1204),
    [anon_sym_macro_rules_BANG] = ACTIONS(1204),
    [anon_sym_LPAREN] = ACTIONS(1204),
    [anon_sym_LBRACE] = ACTIONS(1204),
    [anon_sym_RBRACE] = ACTIONS(1204),
    [anon_sym_LBRACK] = ACTIONS(1204),
    [anon_sym_STAR] = ACTIONS(1204),
    [anon_sym_u8] = ACTIONS(1206),
    [anon_sym_i8] = ACTIONS(1206),
    [anon_sym_u16] = ACTIONS(1206),
    [anon_sym_i16] = ACTIONS(1206),
    [anon_sym_u32] = ACTIONS(1206),
    [anon_sym_i32] = ACTIONS(1206),
    [anon_sym_u64] = ACTIONS(1206),
    [anon_sym_i64] = ACTIONS(1206),
    [anon_sym_u128] = ACTIONS(1206),
    [anon_sym_i128] = ACTIONS(1206),
    [anon_sym_isize] = ACTIONS(1206),
    [anon_sym_usize] = ACTIONS(1206),
    [anon_sym_f32] = ACTIONS(1206),
    [anon_sym_f64] = ACTIONS(1206),
    [anon_sym_bool] = ACTIONS(1206),
    [anon_sym_str] = ACTIONS(1206),
    [anon_sym_char] = ACTIONS(1206),
    [anon_sym_SQUOTE] = ACTIONS(1206),
    [anon_sym_async] = ACTIONS(1206),
    [anon_sym_break] = ACTIONS(1206),
    [anon_sym_const] = ACTIONS(1206),
    [anon_sym_continue] = ACTIONS(1206),
    [anon_sym_default] = ACTIONS(1206),
    [anon_sym_enum] = ACTIONS(1206),
    [anon_sym_fn] = ACTIONS(1206),
    [anon_sym_for] = ACTIONS(1206),
    [anon_sym_if] = ACTIONS(1206),
    [anon_sym_impl] = ACTIONS(1206),
    [anon_sym_let] = ACTIONS(1206),
    [anon_sym_loop] = ACTIONS(1206),
    [anon_sym_match] = ACTIONS(1206),
    [anon_sym_mod] = ACTIONS(1206),
    [anon_sym_pub] = ACTIONS(1206),
    [anon_sym_return] = ACTIONS(1206),
    [anon_sym_static] = ACTIONS(1206),
    [anon_sym_struct] = ACTIONS(1206),
    [anon_sym_trait] = ACTIONS(1206),
    [anon_sym_type] = ACTIONS(1206),
    [anon_sym_union] = ACTIONS(1206),
    [anon_sym_unsafe] = ACTIONS(1206),
    [anon_sym_use] = ACTIONS(1206),
    [anon_sym_while] = ACTIONS(1206),
    [anon_sym_POUND] = ACTIONS(1204),
    [anon_sym_BANG] = ACTIONS(1204),
    [anon_sym_extern] = ACTIONS(1206),
    [anon_sym_LT] = ACTIONS(1204),
    [anon_sym_COLON_COLON] = ACTIONS(1204),
    [anon_sym_AMP] = ACTIONS(1204),
    [anon_sym_DOT_DOT] = ACTIONS(1204),
    [anon_sym_DASH] = ACTIONS(1204),
    [anon_sym_PIPE] = ACTIONS(1204),
    [anon_sym_yield] = ACTIONS(1206),
    [anon_sym_move] = ACTIONS(1206),
    [sym_integer_literal] = ACTIONS(1204),
    [aux_sym_string_literal_token1] = ACTIONS(1204),
    [sym_char_literal] = ACTIONS(1204),
    [anon_sym_true] = ACTIONS(1206),
    [anon_sym_false] = ACTIONS(1206),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1206),
    [sym_super] = ACTIONS(1206),
    [sym_crate] = ACTIONS(1206),
    [sym_metavariable] = ACTIONS(1204),
    [sym_raw_string_literal] = ACTIONS(1204),
    [sym_float_literal] = ACTIONS(1204),
    [sym_block_comment] = ACTIONS(3),
  },
  [284] = {
    [ts_builtin_sym_end] = ACTIONS(1208),
    [sym_identifier] = ACTIONS(1210),
    [anon_sym_SEMI] = ACTIONS(1208),
    [anon_sym_macro_rules_BANG] = ACTIONS(1208),
    [anon_sym_LPAREN] = ACTIONS(1208),
    [anon_sym_LBRACE] = ACTIONS(1208),
    [anon_sym_RBRACE] = ACTIONS(1208),
    [anon_sym_LBRACK] = ACTIONS(1208),
    [anon_sym_STAR] = ACTIONS(1208),
    [anon_sym_u8] = ACTIONS(1210),
    [anon_sym_i8] = ACTIONS(1210),
    [anon_sym_u16] = ACTIONS(1210),
    [anon_sym_i16] = ACTIONS(1210),
    [anon_sym_u32] = ACTIONS(1210),
    [anon_sym_i32] = ACTIONS(1210),
    [anon_sym_u64] = ACTIONS(1210),
    [anon_sym_i64] = ACTIONS(1210),
    [anon_sym_u128] = ACTIONS(1210),
    [anon_sym_i128] = ACTIONS(1210),
    [anon_sym_isize] = ACTIONS(1210),
    [anon_sym_usize] = ACTIONS(1210),
    [anon_sym_f32] = ACTIONS(1210),
    [anon_sym_f64] = ACTIONS(1210),
    [anon_sym_bool] = ACTIONS(1210),
    [anon_sym_str] = ACTIONS(1210),
    [anon_sym_char] = ACTIONS(1210),
    [anon_sym_SQUOTE] = ACTIONS(1210),
    [anon_sym_async] = ACTIONS(1210),
    [anon_sym_break] = ACTIONS(1210),
    [anon_sym_const] = ACTIONS(1210),
    [anon_sym_continue] = ACTIONS(1210),
    [anon_sym_default] = ACTIONS(1210),
    [anon_sym_enum] = ACTIONS(1210),
    [anon_sym_fn] = ACTIONS(1210),
    [anon_sym_for] = ACTIONS(1210),
    [anon_sym_if] = ACTIONS(1210),
    [anon_sym_impl] = ACTIONS(1210),
    [anon_sym_let] = ACTIONS(1210),
    [anon_sym_loop] = ACTIONS(1210),
    [anon_sym_match] = ACTIONS(1210),
    [anon_sym_mod] = ACTIONS(1210),
    [anon_sym_pub] = ACTIONS(1210),
    [anon_sym_return] = ACTIONS(1210),
    [anon_sym_static] = ACTIONS(1210),
    [anon_sym_struct] = ACTIONS(1210),
    [anon_sym_trait] = ACTIONS(1210),
    [anon_sym_type] = ACTIONS(1210),
    [anon_sym_union] = ACTIONS(1210),
    [anon_sym_unsafe] = ACTIONS(1210),
    [anon_sym_use] = ACTIONS(1210),
    [anon_sym_while] = ACTIONS(1210),
    [anon_sym_POUND] = ACTIONS(1208),
    [anon_sym_BANG] = ACTIONS(1208),
    [anon_sym_extern] = ACTIONS(1210),
    [anon_sym_LT] = ACTIONS(1208),
    [anon_sym_COLON_COLON] = ACTIONS(1208),
    [anon_sym_AMP] = ACTIONS(1208),
    [anon_sym_DOT_DOT] = ACTIONS(1208),
    [anon_sym_DASH] = ACTIONS(1208),
    [anon_sym_PIPE] = ACTIONS(1208),
    [anon_sym_yield] = ACTIONS(1210),
    [anon_sym_move] = ACTIONS(1210),
    [sym_integer_literal] = ACTIONS(1208),
    [aux_sym_string_literal_token1] = ACTIONS(1208),
    [sym_char_literal] = ACTIONS(1208),
    [anon_sym_true] = ACTIONS(1210),
    [anon_sym_false] = ACTIONS(1210),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1210),
    [sym_super] = ACTIONS(1210),
    [sym_crate] = ACTIONS(1210),
    [sym_metavariable] = ACTIONS(1208),
    [sym_raw_string_literal] = ACTIONS(1208),
    [sym_float_literal] = ACTIONS(1208),
    [sym_block_comment] = ACTIONS(3),
  },
  [285] = {
    [ts_builtin_sym_end] = ACTIONS(1212),
    [sym_identifier] = ACTIONS(1214),
    [anon_sym_SEMI] = ACTIONS(1212),
    [anon_sym_macro_rules_BANG] = ACTIONS(1212),
    [anon_sym_LPAREN] = ACTIONS(1212),
    [anon_sym_LBRACE] = ACTIONS(1212),
    [anon_sym_RBRACE] = ACTIONS(1212),
    [anon_sym_LBRACK] = ACTIONS(1212),
    [anon_sym_STAR] = ACTIONS(1212),
    [anon_sym_u8] = ACTIONS(1214),
    [anon_sym_i8] = ACTIONS(1214),
    [anon_sym_u16] = ACTIONS(1214),
    [anon_sym_i16] = ACTIONS(1214),
    [anon_sym_u32] = ACTIONS(1214),
    [anon_sym_i32] = ACTIONS(1214),
    [anon_sym_u64] = ACTIONS(1214),
    [anon_sym_i64] = ACTIONS(1214),
    [anon_sym_u128] = ACTIONS(1214),
    [anon_sym_i128] = ACTIONS(1214),
    [anon_sym_isize] = ACTIONS(1214),
    [anon_sym_usize] = ACTIONS(1214),
    [anon_sym_f32] = ACTIONS(1214),
    [anon_sym_f64] = ACTIONS(1214),
    [anon_sym_bool] = ACTIONS(1214),
    [anon_sym_str] = ACTIONS(1214),
    [anon_sym_char] = ACTIONS(1214),
    [anon_sym_SQUOTE] = ACTIONS(1214),
    [anon_sym_async] = ACTIONS(1214),
    [anon_sym_break] = ACTIONS(1214),
    [anon_sym_const] = ACTIONS(1214),
    [anon_sym_continue] = ACTIONS(1214),
    [anon_sym_default] = ACTIONS(1214),
    [anon_sym_enum] = ACTIONS(1214),
    [anon_sym_fn] = ACTIONS(1214),
    [anon_sym_for] = ACTIONS(1214),
    [anon_sym_if] = ACTIONS(1214),
    [anon_sym_impl] = ACTIONS(1214),
    [anon_sym_let] = ACTIONS(1214),
    [anon_sym_loop] = ACTIONS(1214),
    [anon_sym_match] = ACTIONS(1214),
    [anon_sym_mod] = ACTIONS(1214),
    [anon_sym_pub] = ACTIONS(1214),
    [anon_sym_return] = ACTIONS(1214),
    [anon_sym_static] = ACTIONS(1214),
    [anon_sym_struct] = ACTIONS(1214),
    [anon_sym_trait] = ACTIONS(1214),
    [anon_sym_type] = ACTIONS(1214),
    [anon_sym_union] = ACTIONS(1214),
    [anon_sym_unsafe] = ACTIONS(1214),
    [anon_sym_use] = ACTIONS(1214),
    [anon_sym_while] = ACTIONS(1214),
    [anon_sym_POUND] = ACTIONS(1212),
    [anon_sym_BANG] = ACTIONS(1212),
    [anon_sym_extern] = ACTIONS(1214),
    [anon_sym_LT] = ACTIONS(1212),
    [anon_sym_COLON_COLON] = ACTIONS(1212),
    [anon_sym_AMP] = ACTIONS(1212),
    [anon_sym_DOT_DOT] = ACTIONS(1212),
    [anon_sym_DASH] = ACTIONS(1212),
    [anon_sym_PIPE] = ACTIONS(1212),
    [anon_sym_yield] = ACTIONS(1214),
    [anon_sym_move] = ACTIONS(1214),
    [sym_integer_literal] = ACTIONS(1212),
    [aux_sym_string_literal_token1] = ACTIONS(1212),
    [sym_char_literal] = ACTIONS(1212),
    [anon_sym_true] = ACTIONS(1214),
    [anon_sym_false] = ACTIONS(1214),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1214),
    [sym_super] = ACTIONS(1214),
    [sym_crate] = ACTIONS(1214),
    [sym_metavariable] = ACTIONS(1212),
    [sym_raw_string_literal] = ACTIONS(1212),
    [sym_float_literal] = ACTIONS(1212),
    [sym_block_comment] = ACTIONS(3),
  },
  [286] = {
    [ts_builtin_sym_end] = ACTIONS(1216),
    [sym_identifier] = ACTIONS(1218),
    [anon_sym_SEMI] = ACTIONS(1216),
    [anon_sym_macro_rules_BANG] = ACTIONS(1216),
    [anon_sym_LPAREN] = ACTIONS(1216),
    [anon_sym_LBRACE] = ACTIONS(1216),
    [anon_sym_RBRACE] = ACTIONS(1216),
    [anon_sym_LBRACK] = ACTIONS(1216),
    [anon_sym_STAR] = ACTIONS(1216),
    [anon_sym_u8] = ACTIONS(1218),
    [anon_sym_i8] = ACTIONS(1218),
    [anon_sym_u16] = ACTIONS(1218),
    [anon_sym_i16] = ACTIONS(1218),
    [anon_sym_u32] = ACTIONS(1218),
    [anon_sym_i32] = ACTIONS(1218),
    [anon_sym_u64] = ACTIONS(1218),
    [anon_sym_i64] = ACTIONS(1218),
    [anon_sym_u128] = ACTIONS(1218),
    [anon_sym_i128] = ACTIONS(1218),
    [anon_sym_isize] = ACTIONS(1218),
    [anon_sym_usize] = ACTIONS(1218),
    [anon_sym_f32] = ACTIONS(1218),
    [anon_sym_f64] = ACTIONS(1218),
    [anon_sym_bool] = ACTIONS(1218),
    [anon_sym_str] = ACTIONS(1218),
    [anon_sym_char] = ACTIONS(1218),
    [anon_sym_SQUOTE] = ACTIONS(1218),
    [anon_sym_async] = ACTIONS(1218),
    [anon_sym_break] = ACTIONS(1218),
    [anon_sym_const] = ACTIONS(1218),
    [anon_sym_continue] = ACTIONS(1218),
    [anon_sym_default] = ACTIONS(1218),
    [anon_sym_enum] = ACTIONS(1218),
    [anon_sym_fn] = ACTIONS(1218),
    [anon_sym_for] = ACTIONS(1218),
    [anon_sym_if] = ACTIONS(1218),
    [anon_sym_impl] = ACTIONS(1218),
    [anon_sym_let] = ACTIONS(1218),
    [anon_sym_loop] = ACTIONS(1218),
    [anon_sym_match] = ACTIONS(1218),
    [anon_sym_mod] = ACTIONS(1218),
    [anon_sym_pub] = ACTIONS(1218),
    [anon_sym_return] = ACTIONS(1218),
    [anon_sym_static] = ACTIONS(1218),
    [anon_sym_struct] = ACTIONS(1218),
    [anon_sym_trait] = ACTIONS(1218),
    [anon_sym_type] = ACTIONS(1218),
    [anon_sym_union] = ACTIONS(1218),
    [anon_sym_unsafe] = ACTIONS(1218),
    [anon_sym_use] = ACTIONS(1218),
    [anon_sym_while] = ACTIONS(1218),
    [anon_sym_POUND] = ACTIONS(1216),
    [anon_sym_BANG] = ACTIONS(1216),
    [anon_sym_extern] = ACTIONS(1218),
    [anon_sym_LT] = ACTIONS(1216),
    [anon_sym_COLON_COLON] = ACTIONS(1216),
    [anon_sym_AMP] = ACTIONS(1216),
    [anon_sym_DOT_DOT] = ACTIONS(1216),
    [anon_sym_DASH] = ACTIONS(1216),
    [anon_sym_PIPE] = ACTIONS(1216),
    [anon_sym_yield] = ACTIONS(1218),
    [anon_sym_move] = ACTIONS(1218),
    [sym_integer_literal] = ACTIONS(1216),
    [aux_sym_string_literal_token1] = ACTIONS(1216),
    [sym_char_literal] = ACTIONS(1216),
    [anon_sym_true] = ACTIONS(1218),
    [anon_sym_false] = ACTIONS(1218),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1218),
    [sym_super] = ACTIONS(1218),
    [sym_crate] = ACTIONS(1218),
    [sym_metavariable] = ACTIONS(1216),
    [sym_raw_string_literal] = ACTIONS(1216),
    [sym_float_literal] = ACTIONS(1216),
    [sym_block_comment] = ACTIONS(3),
  },
  [287] = {
    [ts_builtin_sym_end] = ACTIONS(1220),
    [sym_identifier] = ACTIONS(1222),
    [anon_sym_SEMI] = ACTIONS(1220),
    [anon_sym_macro_rules_BANG] = ACTIONS(1220),
    [anon_sym_LPAREN] = ACTIONS(1220),
    [anon_sym_LBRACE] = ACTIONS(1220),
    [anon_sym_RBRACE] = ACTIONS(1220),
    [anon_sym_LBRACK] = ACTIONS(1220),
    [anon_sym_STAR] = ACTIONS(1220),
    [anon_sym_u8] = ACTIONS(1222),
    [anon_sym_i8] = ACTIONS(1222),
    [anon_sym_u16] = ACTIONS(1222),
    [anon_sym_i16] = ACTIONS(1222),
    [anon_sym_u32] = ACTIONS(1222),
    [anon_sym_i32] = ACTIONS(1222),
    [anon_sym_u64] = ACTIONS(1222),
    [anon_sym_i64] = ACTIONS(1222),
    [anon_sym_u128] = ACTIONS(1222),
    [anon_sym_i128] = ACTIONS(1222),
    [anon_sym_isize] = ACTIONS(1222),
    [anon_sym_usize] = ACTIONS(1222),
    [anon_sym_f32] = ACTIONS(1222),
    [anon_sym_f64] = ACTIONS(1222),
    [anon_sym_bool] = ACTIONS(1222),
    [anon_sym_str] = ACTIONS(1222),
    [anon_sym_char] = ACTIONS(1222),
    [anon_sym_SQUOTE] = ACTIONS(1222),
    [anon_sym_async] = ACTIONS(1222),
    [anon_sym_break] = ACTIONS(1222),
    [anon_sym_const] = ACTIONS(1222),
    [anon_sym_continue] = ACTIONS(1222),
    [anon_sym_default] = ACTIONS(1222),
    [anon_sym_enum] = ACTIONS(1222),
    [anon_sym_fn] = ACTIONS(1222),
    [anon_sym_for] = ACTIONS(1222),
    [anon_sym_if] = ACTIONS(1222),
    [anon_sym_impl] = ACTIONS(1222),
    [anon_sym_let] = ACTIONS(1222),
    [anon_sym_loop] = ACTIONS(1222),
    [anon_sym_match] = ACTIONS(1222),
    [anon_sym_mod] = ACTIONS(1222),
    [anon_sym_pub] = ACTIONS(1222),
    [anon_sym_return] = ACTIONS(1222),
    [anon_sym_static] = ACTIONS(1222),
    [anon_sym_struct] = ACTIONS(1222),
    [anon_sym_trait] = ACTIONS(1222),
    [anon_sym_type] = ACTIONS(1222),
    [anon_sym_union] = ACTIONS(1222),
    [anon_sym_unsafe] = ACTIONS(1222),
    [anon_sym_use] = ACTIONS(1222),
    [anon_sym_while] = ACTIONS(1222),
    [anon_sym_POUND] = ACTIONS(1220),
    [anon_sym_BANG] = ACTIONS(1220),
    [anon_sym_extern] = ACTIONS(1222),
    [anon_sym_LT] = ACTIONS(1220),
    [anon_sym_COLON_COLON] = ACTIONS(1220),
    [anon_sym_AMP] = ACTIONS(1220),
    [anon_sym_DOT_DOT] = ACTIONS(1220),
    [anon_sym_DASH] = ACTIONS(1220),
    [anon_sym_PIPE] = ACTIONS(1220),
    [anon_sym_yield] = ACTIONS(1222),
    [anon_sym_move] = ACTIONS(1222),
    [sym_integer_literal] = ACTIONS(1220),
    [aux_sym_string_literal_token1] = ACTIONS(1220),
    [sym_char_literal] = ACTIONS(1220),
    [anon_sym_true] = ACTIONS(1222),
    [anon_sym_false] = ACTIONS(1222),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1222),
    [sym_super] = ACTIONS(1222),
    [sym_crate] = ACTIONS(1222),
    [sym_metavariable] = ACTIONS(1220),
    [sym_raw_string_literal] = ACTIONS(1220),
    [sym_float_literal] = ACTIONS(1220),
    [sym_block_comment] = ACTIONS(3),
  },
  [288] = {
    [ts_builtin_sym_end] = ACTIONS(1224),
    [sym_identifier] = ACTIONS(1226),
    [anon_sym_SEMI] = ACTIONS(1224),
    [anon_sym_macro_rules_BANG] = ACTIONS(1224),
    [anon_sym_LPAREN] = ACTIONS(1224),
    [anon_sym_LBRACE] = ACTIONS(1224),
    [anon_sym_RBRACE] = ACTIONS(1224),
    [anon_sym_LBRACK] = ACTIONS(1224),
    [anon_sym_STAR] = ACTIONS(1224),
    [anon_sym_u8] = ACTIONS(1226),
    [anon_sym_i8] = ACTIONS(1226),
    [anon_sym_u16] = ACTIONS(1226),
    [anon_sym_i16] = ACTIONS(1226),
    [anon_sym_u32] = ACTIONS(1226),
    [anon_sym_i32] = ACTIONS(1226),
    [anon_sym_u64] = ACTIONS(1226),
    [anon_sym_i64] = ACTIONS(1226),
    [anon_sym_u128] = ACTIONS(1226),
    [anon_sym_i128] = ACTIONS(1226),
    [anon_sym_isize] = ACTIONS(1226),
    [anon_sym_usize] = ACTIONS(1226),
    [anon_sym_f32] = ACTIONS(1226),
    [anon_sym_f64] = ACTIONS(1226),
    [anon_sym_bool] = ACTIONS(1226),
    [anon_sym_str] = ACTIONS(1226),
    [anon_sym_char] = ACTIONS(1226),
    [anon_sym_SQUOTE] = ACTIONS(1226),
    [anon_sym_async] = ACTIONS(1226),
    [anon_sym_break] = ACTIONS(1226),
    [anon_sym_const] = ACTIONS(1226),
    [anon_sym_continue] = ACTIONS(1226),
    [anon_sym_default] = ACTIONS(1226),
    [anon_sym_enum] = ACTIONS(1226),
    [anon_sym_fn] = ACTIONS(1226),
    [anon_sym_for] = ACTIONS(1226),
    [anon_sym_if] = ACTIONS(1226),
    [anon_sym_impl] = ACTIONS(1226),
    [anon_sym_let] = ACTIONS(1226),
    [anon_sym_loop] = ACTIONS(1226),
    [anon_sym_match] = ACTIONS(1226),
    [anon_sym_mod] = ACTIONS(1226),
    [anon_sym_pub] = ACTIONS(1226),
    [anon_sym_return] = ACTIONS(1226),
    [anon_sym_static] = ACTIONS(1226),
    [anon_sym_struct] = ACTIONS(1226),
    [anon_sym_trait] = ACTIONS(1226),
    [anon_sym_type] = ACTIONS(1226),
    [anon_sym_union] = ACTIONS(1226),
    [anon_sym_unsafe] = ACTIONS(1226),
    [anon_sym_use] = ACTIONS(1226),
    [anon_sym_while] = ACTIONS(1226),
    [anon_sym_POUND] = ACTIONS(1224),
    [anon_sym_BANG] = ACTIONS(1224),
    [anon_sym_extern] = ACTIONS(1226),
    [anon_sym_LT] = ACTIONS(1224),
    [anon_sym_COLON_COLON] = ACTIONS(1224),
    [anon_sym_AMP] = ACTIONS(1224),
    [anon_sym_DOT_DOT] = ACTIONS(1224),
    [anon_sym_DASH] = ACTIONS(1224),
    [anon_sym_PIPE] = ACTIONS(1224),
    [anon_sym_yield] = ACTIONS(1226),
    [anon_sym_move] = ACTIONS(1226),
    [sym_integer_literal] = ACTIONS(1224),
    [aux_sym_string_literal_token1] = ACTIONS(1224),
    [sym_char_literal] = ACTIONS(1224),
    [anon_sym_true] = ACTIONS(1226),
    [anon_sym_false] = ACTIONS(1226),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1226),
    [sym_super] = ACTIONS(1226),
    [sym_crate] = ACTIONS(1226),
    [sym_metavariable] = ACTIONS(1224),
    [sym_raw_string_literal] = ACTIONS(1224),
    [sym_float_literal] = ACTIONS(1224),
    [sym_block_comment] = ACTIONS(3),
  },
  [289] = {
    [ts_builtin_sym_end] = ACTIONS(1228),
    [sym_identifier] = ACTIONS(1230),
    [anon_sym_SEMI] = ACTIONS(1228),
    [anon_sym_macro_rules_BANG] = ACTIONS(1228),
    [anon_sym_LPAREN] = ACTIONS(1228),
    [anon_sym_LBRACE] = ACTIONS(1228),
    [anon_sym_RBRACE] = ACTIONS(1228),
    [anon_sym_LBRACK] = ACTIONS(1228),
    [anon_sym_STAR] = ACTIONS(1228),
    [anon_sym_u8] = ACTIONS(1230),
    [anon_sym_i8] = ACTIONS(1230),
    [anon_sym_u16] = ACTIONS(1230),
    [anon_sym_i16] = ACTIONS(1230),
    [anon_sym_u32] = ACTIONS(1230),
    [anon_sym_i32] = ACTIONS(1230),
    [anon_sym_u64] = ACTIONS(1230),
    [anon_sym_i64] = ACTIONS(1230),
    [anon_sym_u128] = ACTIONS(1230),
    [anon_sym_i128] = ACTIONS(1230),
    [anon_sym_isize] = ACTIONS(1230),
    [anon_sym_usize] = ACTIONS(1230),
    [anon_sym_f32] = ACTIONS(1230),
    [anon_sym_f64] = ACTIONS(1230),
    [anon_sym_bool] = ACTIONS(1230),
    [anon_sym_str] = ACTIONS(1230),
    [anon_sym_char] = ACTIONS(1230),
    [anon_sym_SQUOTE] = ACTIONS(1230),
    [anon_sym_async] = ACTIONS(1230),
    [anon_sym_break] = ACTIONS(1230),
    [anon_sym_const] = ACTIONS(1230),
    [anon_sym_continue] = ACTIONS(1230),
    [anon_sym_default] = ACTIONS(1230),
    [anon_sym_enum] = ACTIONS(1230),
    [anon_sym_fn] = ACTIONS(1230),
    [anon_sym_for] = ACTIONS(1230),
    [anon_sym_if] = ACTIONS(1230),
    [anon_sym_impl] = ACTIONS(1230),
    [anon_sym_let] = ACTIONS(1230),
    [anon_sym_loop] = ACTIONS(1230),
    [anon_sym_match] = ACTIONS(1230),
    [anon_sym_mod] = ACTIONS(1230),
    [anon_sym_pub] = ACTIONS(1230),
    [anon_sym_return] = ACTIONS(1230),
    [anon_sym_static] = ACTIONS(1230),
    [anon_sym_struct] = ACTIONS(1230),
    [anon_sym_trait] = ACTIONS(1230),
    [anon_sym_type] = ACTIONS(1230),
    [anon_sym_union] = ACTIONS(1230),
    [anon_sym_unsafe] = ACTIONS(1230),
    [anon_sym_use] = ACTIONS(1230),
    [anon_sym_while] = ACTIONS(1230),
    [anon_sym_POUND] = ACTIONS(1228),
    [anon_sym_BANG] = ACTIONS(1228),
    [anon_sym_extern] = ACTIONS(1230),
    [anon_sym_LT] = ACTIONS(1228),
    [anon_sym_COLON_COLON] = ACTIONS(1228),
    [anon_sym_AMP] = ACTIONS(1228),
    [anon_sym_DOT_DOT] = ACTIONS(1228),
    [anon_sym_DASH] = ACTIONS(1228),
    [anon_sym_PIPE] = ACTIONS(1228),
    [anon_sym_yield] = ACTIONS(1230),
    [anon_sym_move] = ACTIONS(1230),
    [sym_integer_literal] = ACTIONS(1228),
    [aux_sym_string_literal_token1] = ACTIONS(1228),
    [sym_char_literal] = ACTIONS(1228),
    [anon_sym_true] = ACTIONS(1230),
    [anon_sym_false] = ACTIONS(1230),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1230),
    [sym_super] = ACTIONS(1230),
    [sym_crate] = ACTIONS(1230),
    [sym_metavariable] = ACTIONS(1228),
    [sym_raw_string_literal] = ACTIONS(1228),
    [sym_float_literal] = ACTIONS(1228),
    [sym_block_comment] = ACTIONS(3),
  },
  [290] = {
    [ts_builtin_sym_end] = ACTIONS(1232),
    [sym_identifier] = ACTIONS(1234),
    [anon_sym_SEMI] = ACTIONS(1232),
    [anon_sym_macro_rules_BANG] = ACTIONS(1232),
    [anon_sym_LPAREN] = ACTIONS(1232),
    [anon_sym_LBRACE] = ACTIONS(1232),
    [anon_sym_RBRACE] = ACTIONS(1232),
    [anon_sym_LBRACK] = ACTIONS(1232),
    [anon_sym_STAR] = ACTIONS(1232),
    [anon_sym_u8] = ACTIONS(1234),
    [anon_sym_i8] = ACTIONS(1234),
    [anon_sym_u16] = ACTIONS(1234),
    [anon_sym_i16] = ACTIONS(1234),
    [anon_sym_u32] = ACTIONS(1234),
    [anon_sym_i32] = ACTIONS(1234),
    [anon_sym_u64] = ACTIONS(1234),
    [anon_sym_i64] = ACTIONS(1234),
    [anon_sym_u128] = ACTIONS(1234),
    [anon_sym_i128] = ACTIONS(1234),
    [anon_sym_isize] = ACTIONS(1234),
    [anon_sym_usize] = ACTIONS(1234),
    [anon_sym_f32] = ACTIONS(1234),
    [anon_sym_f64] = ACTIONS(1234),
    [anon_sym_bool] = ACTIONS(1234),
    [anon_sym_str] = ACTIONS(1234),
    [anon_sym_char] = ACTIONS(1234),
    [anon_sym_SQUOTE] = ACTIONS(1234),
    [anon_sym_async] = ACTIONS(1234),
    [anon_sym_break] = ACTIONS(1234),
    [anon_sym_const] = ACTIONS(1234),
    [anon_sym_continue] = ACTIONS(1234),
    [anon_sym_default] = ACTIONS(1234),
    [anon_sym_enum] = ACTIONS(1234),
    [anon_sym_fn] = ACTIONS(1234),
    [anon_sym_for] = ACTIONS(1234),
    [anon_sym_if] = ACTIONS(1234),
    [anon_sym_impl] = ACTIONS(1234),
    [anon_sym_let] = ACTIONS(1234),
    [anon_sym_loop] = ACTIONS(1234),
    [anon_sym_match] = ACTIONS(1234),
    [anon_sym_mod] = ACTIONS(1234),
    [anon_sym_pub] = ACTIONS(1234),
    [anon_sym_return] = ACTIONS(1234),
    [anon_sym_static] = ACTIONS(1234),
    [anon_sym_struct] = ACTIONS(1234),
    [anon_sym_trait] = ACTIONS(1234),
    [anon_sym_type] = ACTIONS(1234),
    [anon_sym_union] = ACTIONS(1234),
    [anon_sym_unsafe] = ACTIONS(1234),
    [anon_sym_use] = ACTIONS(1234),
    [anon_sym_while] = ACTIONS(1234),
    [anon_sym_POUND] = ACTIONS(1232),
    [anon_sym_BANG] = ACTIONS(1232),
    [anon_sym_extern] = ACTIONS(1234),
    [anon_sym_LT] = ACTIONS(1232),
    [anon_sym_COLON_COLON] = ACTIONS(1232),
    [anon_sym_AMP] = ACTIONS(1232),
    [anon_sym_DOT_DOT] = ACTIONS(1232),
    [anon_sym_DASH] = ACTIONS(1232),
    [anon_sym_PIPE] = ACTIONS(1232),
    [anon_sym_yield] = ACTIONS(1234),
    [anon_sym_move] = ACTIONS(1234),
    [sym_integer_literal] = ACTIONS(1232),
    [aux_sym_string_literal_token1] = ACTIONS(1232),
    [sym_char_literal] = ACTIONS(1232),
    [anon_sym_true] = ACTIONS(1234),
    [anon_sym_false] = ACTIONS(1234),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1234),
    [sym_super] = ACTIONS(1234),
    [sym_crate] = ACTIONS(1234),
    [sym_metavariable] = ACTIONS(1232),
    [sym_raw_string_literal] = ACTIONS(1232),
    [sym_float_literal] = ACTIONS(1232),
    [sym_block_comment] = ACTIONS(3),
  },
  [291] = {
    [ts_builtin_sym_end] = ACTIONS(1236),
    [sym_identifier] = ACTIONS(1238),
    [anon_sym_SEMI] = ACTIONS(1236),
    [anon_sym_macro_rules_BANG] = ACTIONS(1236),
    [anon_sym_LPAREN] = ACTIONS(1236),
    [anon_sym_LBRACE] = ACTIONS(1236),
    [anon_sym_RBRACE] = ACTIONS(1236),
    [anon_sym_LBRACK] = ACTIONS(1236),
    [anon_sym_STAR] = ACTIONS(1236),
    [anon_sym_u8] = ACTIONS(1238),
    [anon_sym_i8] = ACTIONS(1238),
    [anon_sym_u16] = ACTIONS(1238),
    [anon_sym_i16] = ACTIONS(1238),
    [anon_sym_u32] = ACTIONS(1238),
    [anon_sym_i32] = ACTIONS(1238),
    [anon_sym_u64] = ACTIONS(1238),
    [anon_sym_i64] = ACTIONS(1238),
    [anon_sym_u128] = ACTIONS(1238),
    [anon_sym_i128] = ACTIONS(1238),
    [anon_sym_isize] = ACTIONS(1238),
    [anon_sym_usize] = ACTIONS(1238),
    [anon_sym_f32] = ACTIONS(1238),
    [anon_sym_f64] = ACTIONS(1238),
    [anon_sym_bool] = ACTIONS(1238),
    [anon_sym_str] = ACTIONS(1238),
    [anon_sym_char] = ACTIONS(1238),
    [anon_sym_SQUOTE] = ACTIONS(1238),
    [anon_sym_async] = ACTIONS(1238),
    [anon_sym_break] = ACTIONS(1238),
    [anon_sym_const] = ACTIONS(1238),
    [anon_sym_continue] = ACTIONS(1238),
    [anon_sym_default] = ACTIONS(1238),
    [anon_sym_enum] = ACTIONS(1238),
    [anon_sym_fn] = ACTIONS(1238),
    [anon_sym_for] = ACTIONS(1238),
    [anon_sym_if] = ACTIONS(1238),
    [anon_sym_impl] = ACTIONS(1238),
    [anon_sym_let] = ACTIONS(1238),
    [anon_sym_loop] = ACTIONS(1238),
    [anon_sym_match] = ACTIONS(1238),
    [anon_sym_mod] = ACTIONS(1238),
    [anon_sym_pub] = ACTIONS(1238),
    [anon_sym_return] = ACTIONS(1238),
    [anon_sym_static] = ACTIONS(1238),
    [anon_sym_struct] = ACTIONS(1238),
    [anon_sym_trait] = ACTIONS(1238),
    [anon_sym_type] = ACTIONS(1238),
    [anon_sym_union] = ACTIONS(1238),
    [anon_sym_unsafe] = ACTIONS(1238),
    [anon_sym_use] = ACTIONS(1238),
    [anon_sym_while] = ACTIONS(1238),
    [anon_sym_POUND] = ACTIONS(1236),
    [anon_sym_BANG] = ACTIONS(1236),
    [anon_sym_extern] = ACTIONS(1238),
    [anon_sym_LT] = ACTIONS(1236),
    [anon_sym_COLON_COLON] = ACTIONS(1236),
    [anon_sym_AMP] = ACTIONS(1236),
    [anon_sym_DOT_DOT] = ACTIONS(1236),
    [anon_sym_DASH] = ACTIONS(1236),
    [anon_sym_PIPE] = ACTIONS(1236),
    [anon_sym_yield] = ACTIONS(1238),
    [anon_sym_move] = ACTIONS(1238),
    [sym_integer_literal] = ACTIONS(1236),
    [aux_sym_string_literal_token1] = ACTIONS(1236),
    [sym_char_literal] = ACTIONS(1236),
    [anon_sym_true] = ACTIONS(1238),
    [anon_sym_false] = ACTIONS(1238),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1238),
    [sym_super] = ACTIONS(1238),
    [sym_crate] = ACTIONS(1238),
    [sym_metavariable] = ACTIONS(1236),
    [sym_raw_string_literal] = ACTIONS(1236),
    [sym_float_literal] = ACTIONS(1236),
    [sym_block_comment] = ACTIONS(3),
  },
  [292] = {
    [ts_builtin_sym_end] = ACTIONS(1240),
    [sym_identifier] = ACTIONS(1242),
    [anon_sym_SEMI] = ACTIONS(1240),
    [anon_sym_macro_rules_BANG] = ACTIONS(1240),
    [anon_sym_LPAREN] = ACTIONS(1240),
    [anon_sym_LBRACE] = ACTIONS(1240),
    [anon_sym_RBRACE] = ACTIONS(1240),
    [anon_sym_LBRACK] = ACTIONS(1240),
    [anon_sym_STAR] = ACTIONS(1240),
    [anon_sym_u8] = ACTIONS(1242),
    [anon_sym_i8] = ACTIONS(1242),
    [anon_sym_u16] = ACTIONS(1242),
    [anon_sym_i16] = ACTIONS(1242),
    [anon_sym_u32] = ACTIONS(1242),
    [anon_sym_i32] = ACTIONS(1242),
    [anon_sym_u64] = ACTIONS(1242),
    [anon_sym_i64] = ACTIONS(1242),
    [anon_sym_u128] = ACTIONS(1242),
    [anon_sym_i128] = ACTIONS(1242),
    [anon_sym_isize] = ACTIONS(1242),
    [anon_sym_usize] = ACTIONS(1242),
    [anon_sym_f32] = ACTIONS(1242),
    [anon_sym_f64] = ACTIONS(1242),
    [anon_sym_bool] = ACTIONS(1242),
    [anon_sym_str] = ACTIONS(1242),
    [anon_sym_char] = ACTIONS(1242),
    [anon_sym_SQUOTE] = ACTIONS(1242),
    [anon_sym_async] = ACTIONS(1242),
    [anon_sym_break] = ACTIONS(1242),
    [anon_sym_const] = ACTIONS(1242),
    [anon_sym_continue] = ACTIONS(1242),
    [anon_sym_default] = ACTIONS(1242),
    [anon_sym_enum] = ACTIONS(1242),
    [anon_sym_fn] = ACTIONS(1242),
    [anon_sym_for] = ACTIONS(1242),
    [anon_sym_if] = ACTIONS(1242),
    [anon_sym_impl] = ACTIONS(1242),
    [anon_sym_let] = ACTIONS(1242),
    [anon_sym_loop] = ACTIONS(1242),
    [anon_sym_match] = ACTIONS(1242),
    [anon_sym_mod] = ACTIONS(1242),
    [anon_sym_pub] = ACTIONS(1242),
    [anon_sym_return] = ACTIONS(1242),
    [anon_sym_static] = ACTIONS(1242),
    [anon_sym_struct] = ACTIONS(1242),
    [anon_sym_trait] = ACTIONS(1242),
    [anon_sym_type] = ACTIONS(1242),
    [anon_sym_union] = ACTIONS(1242),
    [anon_sym_unsafe] = ACTIONS(1242),
    [anon_sym_use] = ACTIONS(1242),
    [anon_sym_while] = ACTIONS(1242),
    [anon_sym_POUND] = ACTIONS(1240),
    [anon_sym_BANG] = ACTIONS(1240),
    [anon_sym_extern] = ACTIONS(1242),
    [anon_sym_LT] = ACTIONS(1240),
    [anon_sym_COLON_COLON] = ACTIONS(1240),
    [anon_sym_AMP] = ACTIONS(1240),
    [anon_sym_DOT_DOT] = ACTIONS(1240),
    [anon_sym_DASH] = ACTIONS(1240),
    [anon_sym_PIPE] = ACTIONS(1240),
    [anon_sym_yield] = ACTIONS(1242),
    [anon_sym_move] = ACTIONS(1242),
    [sym_integer_literal] = ACTIONS(1240),
    [aux_sym_string_literal_token1] = ACTIONS(1240),
    [sym_char_literal] = ACTIONS(1240),
    [anon_sym_true] = ACTIONS(1242),
    [anon_sym_false] = ACTIONS(1242),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1242),
    [sym_super] = ACTIONS(1242),
    [sym_crate] = ACTIONS(1242),
    [sym_metavariable] = ACTIONS(1240),
    [sym_raw_string_literal] = ACTIONS(1240),
    [sym_float_literal] = ACTIONS(1240),
    [sym_block_comment] = ACTIONS(3),
  },
  [293] = {
    [ts_builtin_sym_end] = ACTIONS(1244),
    [sym_identifier] = ACTIONS(1246),
    [anon_sym_SEMI] = ACTIONS(1244),
    [anon_sym_macro_rules_BANG] = ACTIONS(1244),
    [anon_sym_LPAREN] = ACTIONS(1244),
    [anon_sym_LBRACE] = ACTIONS(1244),
    [anon_sym_RBRACE] = ACTIONS(1244),
    [anon_sym_LBRACK] = ACTIONS(1244),
    [anon_sym_STAR] = ACTIONS(1244),
    [anon_sym_u8] = ACTIONS(1246),
    [anon_sym_i8] = ACTIONS(1246),
    [anon_sym_u16] = ACTIONS(1246),
    [anon_sym_i16] = ACTIONS(1246),
    [anon_sym_u32] = ACTIONS(1246),
    [anon_sym_i32] = ACTIONS(1246),
    [anon_sym_u64] = ACTIONS(1246),
    [anon_sym_i64] = ACTIONS(1246),
    [anon_sym_u128] = ACTIONS(1246),
    [anon_sym_i128] = ACTIONS(1246),
    [anon_sym_isize] = ACTIONS(1246),
    [anon_sym_usize] = ACTIONS(1246),
    [anon_sym_f32] = ACTIONS(1246),
    [anon_sym_f64] = ACTIONS(1246),
    [anon_sym_bool] = ACTIONS(1246),
    [anon_sym_str] = ACTIONS(1246),
    [anon_sym_char] = ACTIONS(1246),
    [anon_sym_SQUOTE] = ACTIONS(1246),
    [anon_sym_async] = ACTIONS(1246),
    [anon_sym_break] = ACTIONS(1246),
    [anon_sym_const] = ACTIONS(1246),
    [anon_sym_continue] = ACTIONS(1246),
    [anon_sym_default] = ACTIONS(1246),
    [anon_sym_enum] = ACTIONS(1246),
    [anon_sym_fn] = ACTIONS(1246),
    [anon_sym_for] = ACTIONS(1246),
    [anon_sym_if] = ACTIONS(1246),
    [anon_sym_impl] = ACTIONS(1246),
    [anon_sym_let] = ACTIONS(1246),
    [anon_sym_loop] = ACTIONS(1246),
    [anon_sym_match] = ACTIONS(1246),
    [anon_sym_mod] = ACTIONS(1246),
    [anon_sym_pub] = ACTIONS(1246),
    [anon_sym_return] = ACTIONS(1246),
    [anon_sym_static] = ACTIONS(1246),
    [anon_sym_struct] = ACTIONS(1246),
    [anon_sym_trait] = ACTIONS(1246),
    [anon_sym_type] = ACTIONS(1246),
    [anon_sym_union] = ACTIONS(1246),
    [anon_sym_unsafe] = ACTIONS(1246),
    [anon_sym_use] = ACTIONS(1246),
    [anon_sym_while] = ACTIONS(1246),
    [anon_sym_POUND] = ACTIONS(1244),
    [anon_sym_BANG] = ACTIONS(1244),
    [anon_sym_extern] = ACTIONS(1246),
    [anon_sym_LT] = ACTIONS(1244),
    [anon_sym_COLON_COLON] = ACTIONS(1244),
    [anon_sym_AMP] = ACTIONS(1244),
    [anon_sym_DOT_DOT] = ACTIONS(1244),
    [anon_sym_DASH] = ACTIONS(1244),
    [anon_sym_PIPE] = ACTIONS(1244),
    [anon_sym_yield] = ACTIONS(1246),
    [anon_sym_move] = ACTIONS(1246),
    [sym_integer_literal] = ACTIONS(1244),
    [aux_sym_string_literal_token1] = ACTIONS(1244),
    [sym_char_literal] = ACTIONS(1244),
    [anon_sym_true] = ACTIONS(1246),
    [anon_sym_false] = ACTIONS(1246),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1246),
    [sym_super] = ACTIONS(1246),
    [sym_crate] = ACTIONS(1246),
    [sym_metavariable] = ACTIONS(1244),
    [sym_raw_string_literal] = ACTIONS(1244),
    [sym_float_literal] = ACTIONS(1244),
    [sym_block_comment] = ACTIONS(3),
  },
  [294] = {
    [ts_builtin_sym_end] = ACTIONS(1248),
    [sym_identifier] = ACTIONS(1250),
    [anon_sym_SEMI] = ACTIONS(1248),
    [anon_sym_macro_rules_BANG] = ACTIONS(1248),
    [anon_sym_LPAREN] = ACTIONS(1248),
    [anon_sym_LBRACE] = ACTIONS(1248),
    [anon_sym_RBRACE] = ACTIONS(1248),
    [anon_sym_LBRACK] = ACTIONS(1248),
    [anon_sym_STAR] = ACTIONS(1248),
    [anon_sym_u8] = ACTIONS(1250),
    [anon_sym_i8] = ACTIONS(1250),
    [anon_sym_u16] = ACTIONS(1250),
    [anon_sym_i16] = ACTIONS(1250),
    [anon_sym_u32] = ACTIONS(1250),
    [anon_sym_i32] = ACTIONS(1250),
    [anon_sym_u64] = ACTIONS(1250),
    [anon_sym_i64] = ACTIONS(1250),
    [anon_sym_u128] = ACTIONS(1250),
    [anon_sym_i128] = ACTIONS(1250),
    [anon_sym_isize] = ACTIONS(1250),
    [anon_sym_usize] = ACTIONS(1250),
    [anon_sym_f32] = ACTIONS(1250),
    [anon_sym_f64] = ACTIONS(1250),
    [anon_sym_bool] = ACTIONS(1250),
    [anon_sym_str] = ACTIONS(1250),
    [anon_sym_char] = ACTIONS(1250),
    [anon_sym_SQUOTE] = ACTIONS(1250),
    [anon_sym_async] = ACTIONS(1250),
    [anon_sym_break] = ACTIONS(1250),
    [anon_sym_const] = ACTIONS(1250),
    [anon_sym_continue] = ACTIONS(1250),
    [anon_sym_default] = ACTIONS(1250),
    [anon_sym_enum] = ACTIONS(1250),
    [anon_sym_fn] = ACTIONS(1250),
    [anon_sym_for] = ACTIONS(1250),
    [anon_sym_if] = ACTIONS(1250),
    [anon_sym_impl] = ACTIONS(1250),
    [anon_sym_let] = ACTIONS(1250),
    [anon_sym_loop] = ACTIONS(1250),
    [anon_sym_match] = ACTIONS(1250),
    [anon_sym_mod] = ACTIONS(1250),
    [anon_sym_pub] = ACTIONS(1250),
    [anon_sym_return] = ACTIONS(1250),
    [anon_sym_static] = ACTIONS(1250),
    [anon_sym_struct] = ACTIONS(1250),
    [anon_sym_trait] = ACTIONS(1250),
    [anon_sym_type] = ACTIONS(1250),
    [anon_sym_union] = ACTIONS(1250),
    [anon_sym_unsafe] = ACTIONS(1250),
    [anon_sym_use] = ACTIONS(1250),
    [anon_sym_while] = ACTIONS(1250),
    [anon_sym_POUND] = ACTIONS(1248),
    [anon_sym_BANG] = ACTIONS(1248),
    [anon_sym_extern] = ACTIONS(1250),
    [anon_sym_LT] = ACTIONS(1248),
    [anon_sym_COLON_COLON] = ACTIONS(1248),
    [anon_sym_AMP] = ACTIONS(1248),
    [anon_sym_DOT_DOT] = ACTIONS(1248),
    [anon_sym_DASH] = ACTIONS(1248),
    [anon_sym_PIPE] = ACTIONS(1248),
    [anon_sym_yield] = ACTIONS(1250),
    [anon_sym_move] = ACTIONS(1250),
    [sym_integer_literal] = ACTIONS(1248),
    [aux_sym_string_literal_token1] = ACTIONS(1248),
    [sym_char_literal] = ACTIONS(1248),
    [anon_sym_true] = ACTIONS(1250),
    [anon_sym_false] = ACTIONS(1250),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1250),
    [sym_super] = ACTIONS(1250),
    [sym_crate] = ACTIONS(1250),
    [sym_metavariable] = ACTIONS(1248),
    [sym_raw_string_literal] = ACTIONS(1248),
    [sym_float_literal] = ACTIONS(1248),
    [sym_block_comment] = ACTIONS(3),
  },
  [295] = {
    [ts_builtin_sym_end] = ACTIONS(1252),
    [sym_identifier] = ACTIONS(1254),
    [anon_sym_SEMI] = ACTIONS(1252),
    [anon_sym_macro_rules_BANG] = ACTIONS(1252),
    [anon_sym_LPAREN] = ACTIONS(1252),
    [anon_sym_LBRACE] = ACTIONS(1252),
    [anon_sym_RBRACE] = ACTIONS(1252),
    [anon_sym_LBRACK] = ACTIONS(1252),
    [anon_sym_STAR] = ACTIONS(1252),
    [anon_sym_u8] = ACTIONS(1254),
    [anon_sym_i8] = ACTIONS(1254),
    [anon_sym_u16] = ACTIONS(1254),
    [anon_sym_i16] = ACTIONS(1254),
    [anon_sym_u32] = ACTIONS(1254),
    [anon_sym_i32] = ACTIONS(1254),
    [anon_sym_u64] = ACTIONS(1254),
    [anon_sym_i64] = ACTIONS(1254),
    [anon_sym_u128] = ACTIONS(1254),
    [anon_sym_i128] = ACTIONS(1254),
    [anon_sym_isize] = ACTIONS(1254),
    [anon_sym_usize] = ACTIONS(1254),
    [anon_sym_f32] = ACTIONS(1254),
    [anon_sym_f64] = ACTIONS(1254),
    [anon_sym_bool] = ACTIONS(1254),
    [anon_sym_str] = ACTIONS(1254),
    [anon_sym_char] = ACTIONS(1254),
    [anon_sym_SQUOTE] = ACTIONS(1254),
    [anon_sym_async] = ACTIONS(1254),
    [anon_sym_break] = ACTIONS(1254),
    [anon_sym_const] = ACTIONS(1254),
    [anon_sym_continue] = ACTIONS(1254),
    [anon_sym_default] = ACTIONS(1254),
    [anon_sym_enum] = ACTIONS(1254),
    [anon_sym_fn] = ACTIONS(1254),
    [anon_sym_for] = ACTIONS(1254),
    [anon_sym_if] = ACTIONS(1254),
    [anon_sym_impl] = ACTIONS(1254),
    [anon_sym_let] = ACTIONS(1254),
    [anon_sym_loop] = ACTIONS(1254),
    [anon_sym_match] = ACTIONS(1254),
    [anon_sym_mod] = ACTIONS(1254),
    [anon_sym_pub] = ACTIONS(1254),
    [anon_sym_return] = ACTIONS(1254),
    [anon_sym_static] = ACTIONS(1254),
    [anon_sym_struct] = ACTIONS(1254),
    [anon_sym_trait] = ACTIONS(1254),
    [anon_sym_type] = ACTIONS(1254),
    [anon_sym_union] = ACTIONS(1254),
    [anon_sym_unsafe] = ACTIONS(1254),
    [anon_sym_use] = ACTIONS(1254),
    [anon_sym_while] = ACTIONS(1254),
    [anon_sym_POUND] = ACTIONS(1252),
    [anon_sym_BANG] = ACTIONS(1252),
    [anon_sym_extern] = ACTIONS(1254),
    [anon_sym_LT] = ACTIONS(1252),
    [anon_sym_COLON_COLON] = ACTIONS(1252),
    [anon_sym_AMP] = ACTIONS(1252),
    [anon_sym_DOT_DOT] = ACTIONS(1252),
    [anon_sym_DASH] = ACTIONS(1252),
    [anon_sym_PIPE] = ACTIONS(1252),
    [anon_sym_yield] = ACTIONS(1254),
    [anon_sym_move] = ACTIONS(1254),
    [sym_integer_literal] = ACTIONS(1252),
    [aux_sym_string_literal_token1] = ACTIONS(1252),
    [sym_char_literal] = ACTIONS(1252),
    [anon_sym_true] = ACTIONS(1254),
    [anon_sym_false] = ACTIONS(1254),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1254),
    [sym_super] = ACTIONS(1254),
    [sym_crate] = ACTIONS(1254),
    [sym_metavariable] = ACTIONS(1252),
    [sym_raw_string_literal] = ACTIONS(1252),
    [sym_float_literal] = ACTIONS(1252),
    [sym_block_comment] = ACTIONS(3),
  },
  [296] = {
    [ts_builtin_sym_end] = ACTIONS(1256),
    [sym_identifier] = ACTIONS(1258),
    [anon_sym_SEMI] = ACTIONS(1256),
    [anon_sym_macro_rules_BANG] = ACTIONS(1256),
    [anon_sym_LPAREN] = ACTIONS(1256),
    [anon_sym_LBRACE] = ACTIONS(1256),
    [anon_sym_RBRACE] = ACTIONS(1256),
    [anon_sym_LBRACK] = ACTIONS(1256),
    [anon_sym_STAR] = ACTIONS(1256),
    [anon_sym_u8] = ACTIONS(1258),
    [anon_sym_i8] = ACTIONS(1258),
    [anon_sym_u16] = ACTIONS(1258),
    [anon_sym_i16] = ACTIONS(1258),
    [anon_sym_u32] = ACTIONS(1258),
    [anon_sym_i32] = ACTIONS(1258),
    [anon_sym_u64] = ACTIONS(1258),
    [anon_sym_i64] = ACTIONS(1258),
    [anon_sym_u128] = ACTIONS(1258),
    [anon_sym_i128] = ACTIONS(1258),
    [anon_sym_isize] = ACTIONS(1258),
    [anon_sym_usize] = ACTIONS(1258),
    [anon_sym_f32] = ACTIONS(1258),
    [anon_sym_f64] = ACTIONS(1258),
    [anon_sym_bool] = ACTIONS(1258),
    [anon_sym_str] = ACTIONS(1258),
    [anon_sym_char] = ACTIONS(1258),
    [anon_sym_SQUOTE] = ACTIONS(1258),
    [anon_sym_async] = ACTIONS(1258),
    [anon_sym_break] = ACTIONS(1258),
    [anon_sym_const] = ACTIONS(1258),
    [anon_sym_continue] = ACTIONS(1258),
    [anon_sym_default] = ACTIONS(1258),
    [anon_sym_enum] = ACTIONS(1258),
    [anon_sym_fn] = ACTIONS(1258),
    [anon_sym_for] = ACTIONS(1258),
    [anon_sym_if] = ACTIONS(1258),
    [anon_sym_impl] = ACTIONS(1258),
    [anon_sym_let] = ACTIONS(1258),
    [anon_sym_loop] = ACTIONS(1258),
    [anon_sym_match] = ACTIONS(1258),
    [anon_sym_mod] = ACTIONS(1258),
    [anon_sym_pub] = ACTIONS(1258),
    [anon_sym_return] = ACTIONS(1258),
    [anon_sym_static] = ACTIONS(1258),
    [anon_sym_struct] = ACTIONS(1258),
    [anon_sym_trait] = ACTIONS(1258),
    [anon_sym_type] = ACTIONS(1258),
    [anon_sym_union] = ACTIONS(1258),
    [anon_sym_unsafe] = ACTIONS(1258),
    [anon_sym_use] = ACTIONS(1258),
    [anon_sym_while] = ACTIONS(1258),
    [anon_sym_POUND] = ACTIONS(1256),
    [anon_sym_BANG] = ACTIONS(1256),
    [anon_sym_extern] = ACTIONS(1258),
    [anon_sym_LT] = ACTIONS(1256),
    [anon_sym_COLON_COLON] = ACTIONS(1256),
    [anon_sym_AMP] = ACTIONS(1256),
    [anon_sym_DOT_DOT] = ACTIONS(1256),
    [anon_sym_DASH] = ACTIONS(1256),
    [anon_sym_PIPE] = ACTIONS(1256),
    [anon_sym_yield] = ACTIONS(1258),
    [anon_sym_move] = ACTIONS(1258),
    [sym_integer_literal] = ACTIONS(1256),
    [aux_sym_string_literal_token1] = ACTIONS(1256),
    [sym_char_literal] = ACTIONS(1256),
    [anon_sym_true] = ACTIONS(1258),
    [anon_sym_false] = ACTIONS(1258),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1258),
    [sym_super] = ACTIONS(1258),
    [sym_crate] = ACTIONS(1258),
    [sym_metavariable] = ACTIONS(1256),
    [sym_raw_string_literal] = ACTIONS(1256),
    [sym_float_literal] = ACTIONS(1256),
    [sym_block_comment] = ACTIONS(3),
  },
  [297] = {
    [ts_builtin_sym_end] = ACTIONS(1260),
    [sym_identifier] = ACTIONS(1262),
    [anon_sym_SEMI] = ACTIONS(1260),
    [anon_sym_macro_rules_BANG] = ACTIONS(1260),
    [anon_sym_LPAREN] = ACTIONS(1260),
    [anon_sym_LBRACE] = ACTIONS(1260),
    [anon_sym_RBRACE] = ACTIONS(1260),
    [anon_sym_LBRACK] = ACTIONS(1260),
    [anon_sym_STAR] = ACTIONS(1260),
    [anon_sym_u8] = ACTIONS(1262),
    [anon_sym_i8] = ACTIONS(1262),
    [anon_sym_u16] = ACTIONS(1262),
    [anon_sym_i16] = ACTIONS(1262),
    [anon_sym_u32] = ACTIONS(1262),
    [anon_sym_i32] = ACTIONS(1262),
    [anon_sym_u64] = ACTIONS(1262),
    [anon_sym_i64] = ACTIONS(1262),
    [anon_sym_u128] = ACTIONS(1262),
    [anon_sym_i128] = ACTIONS(1262),
    [anon_sym_isize] = ACTIONS(1262),
    [anon_sym_usize] = ACTIONS(1262),
    [anon_sym_f32] = ACTIONS(1262),
    [anon_sym_f64] = ACTIONS(1262),
    [anon_sym_bool] = ACTIONS(1262),
    [anon_sym_str] = ACTIONS(1262),
    [anon_sym_char] = ACTIONS(1262),
    [anon_sym_SQUOTE] = ACTIONS(1262),
    [anon_sym_async] = ACTIONS(1262),
    [anon_sym_break] = ACTIONS(1262),
    [anon_sym_const] = ACTIONS(1262),
    [anon_sym_continue] = ACTIONS(1262),
    [anon_sym_default] = ACTIONS(1262),
    [anon_sym_enum] = ACTIONS(1262),
    [anon_sym_fn] = ACTIONS(1262),
    [anon_sym_for] = ACTIONS(1262),
    [anon_sym_if] = ACTIONS(1262),
    [anon_sym_impl] = ACTIONS(1262),
    [anon_sym_let] = ACTIONS(1262),
    [anon_sym_loop] = ACTIONS(1262),
    [anon_sym_match] = ACTIONS(1262),
    [anon_sym_mod] = ACTIONS(1262),
    [anon_sym_pub] = ACTIONS(1262),
    [anon_sym_return] = ACTIONS(1262),
    [anon_sym_static] = ACTIONS(1262),
    [anon_sym_struct] = ACTIONS(1262),
    [anon_sym_trait] = ACTIONS(1262),
    [anon_sym_type] = ACTIONS(1262),
    [anon_sym_union] = ACTIONS(1262),
    [anon_sym_unsafe] = ACTIONS(1262),
    [anon_sym_use] = ACTIONS(1262),
    [anon_sym_while] = ACTIONS(1262),
    [anon_sym_POUND] = ACTIONS(1260),
    [anon_sym_BANG] = ACTIONS(1260),
    [anon_sym_extern] = ACTIONS(1262),
    [anon_sym_LT] = ACTIONS(1260),
    [anon_sym_COLON_COLON] = ACTIONS(1260),
    [anon_sym_AMP] = ACTIONS(1260),
    [anon_sym_DOT_DOT] = ACTIONS(1260),
    [anon_sym_DASH] = ACTIONS(1260),
    [anon_sym_PIPE] = ACTIONS(1260),
    [anon_sym_yield] = ACTIONS(1262),
    [anon_sym_move] = ACTIONS(1262),
    [sym_integer_literal] = ACTIONS(1260),
    [aux_sym_string_literal_token1] = ACTIONS(1260),
    [sym_char_literal] = ACTIONS(1260),
    [anon_sym_true] = ACTIONS(1262),
    [anon_sym_false] = ACTIONS(1262),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1262),
    [sym_super] = ACTIONS(1262),
    [sym_crate] = ACTIONS(1262),
    [sym_metavariable] = ACTIONS(1260),
    [sym_raw_string_literal] = ACTIONS(1260),
    [sym_float_literal] = ACTIONS(1260),
    [sym_block_comment] = ACTIONS(3),
  },
  [298] = {
    [ts_builtin_sym_end] = ACTIONS(1264),
    [sym_identifier] = ACTIONS(1266),
    [anon_sym_SEMI] = ACTIONS(1264),
    [anon_sym_macro_rules_BANG] = ACTIONS(1264),
    [anon_sym_LPAREN] = ACTIONS(1264),
    [anon_sym_LBRACE] = ACTIONS(1264),
    [anon_sym_RBRACE] = ACTIONS(1264),
    [anon_sym_LBRACK] = ACTIONS(1264),
    [anon_sym_STAR] = ACTIONS(1264),
    [anon_sym_u8] = ACTIONS(1266),
    [anon_sym_i8] = ACTIONS(1266),
    [anon_sym_u16] = ACTIONS(1266),
    [anon_sym_i16] = ACTIONS(1266),
    [anon_sym_u32] = ACTIONS(1266),
    [anon_sym_i32] = ACTIONS(1266),
    [anon_sym_u64] = ACTIONS(1266),
    [anon_sym_i64] = ACTIONS(1266),
    [anon_sym_u128] = ACTIONS(1266),
    [anon_sym_i128] = ACTIONS(1266),
    [anon_sym_isize] = ACTIONS(1266),
    [anon_sym_usize] = ACTIONS(1266),
    [anon_sym_f32] = ACTIONS(1266),
    [anon_sym_f64] = ACTIONS(1266),
    [anon_sym_bool] = ACTIONS(1266),
    [anon_sym_str] = ACTIONS(1266),
    [anon_sym_char] = ACTIONS(1266),
    [anon_sym_SQUOTE] = ACTIONS(1266),
    [anon_sym_async] = ACTIONS(1266),
    [anon_sym_break] = ACTIONS(1266),
    [anon_sym_const] = ACTIONS(1266),
    [anon_sym_continue] = ACTIONS(1266),
    [anon_sym_default] = ACTIONS(1266),
    [anon_sym_enum] = ACTIONS(1266),
    [anon_sym_fn] = ACTIONS(1266),
    [anon_sym_for] = ACTIONS(1266),
    [anon_sym_if] = ACTIONS(1266),
    [anon_sym_impl] = ACTIONS(1266),
    [anon_sym_let] = ACTIONS(1266),
    [anon_sym_loop] = ACTIONS(1266),
    [anon_sym_match] = ACTIONS(1266),
    [anon_sym_mod] = ACTIONS(1266),
    [anon_sym_pub] = ACTIONS(1266),
    [anon_sym_return] = ACTIONS(1266),
    [anon_sym_static] = ACTIONS(1266),
    [anon_sym_struct] = ACTIONS(1266),
    [anon_sym_trait] = ACTIONS(1266),
    [anon_sym_type] = ACTIONS(1266),
    [anon_sym_union] = ACTIONS(1266),
    [anon_sym_unsafe] = ACTIONS(1266),
    [anon_sym_use] = ACTIONS(1266),
    [anon_sym_while] = ACTIONS(1266),
    [anon_sym_POUND] = ACTIONS(1264),
    [anon_sym_BANG] = ACTIONS(1264),
    [anon_sym_extern] = ACTIONS(1266),
    [anon_sym_LT] = ACTIONS(1264),
    [anon_sym_COLON_COLON] = ACTIONS(1264),
    [anon_sym_AMP] = ACTIONS(1264),
    [anon_sym_DOT_DOT] = ACTIONS(1264),
    [anon_sym_DASH] = ACTIONS(1264),
    [anon_sym_PIPE] = ACTIONS(1264),
    [anon_sym_yield] = ACTIONS(1266),
    [anon_sym_move] = ACTIONS(1266),
    [sym_integer_literal] = ACTIONS(1264),
    [aux_sym_string_literal_token1] = ACTIONS(1264),
    [sym_char_literal] = ACTIONS(1264),
    [anon_sym_true] = ACTIONS(1266),
    [anon_sym_false] = ACTIONS(1266),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1266),
    [sym_super] = ACTIONS(1266),
    [sym_crate] = ACTIONS(1266),
    [sym_metavariable] = ACTIONS(1264),
    [sym_raw_string_literal] = ACTIONS(1264),
    [sym_float_literal] = ACTIONS(1264),
    [sym_block_comment] = ACTIONS(3),
  },
  [299] = {
    [ts_builtin_sym_end] = ACTIONS(1268),
    [sym_identifier] = ACTIONS(1270),
    [anon_sym_SEMI] = ACTIONS(1268),
    [anon_sym_macro_rules_BANG] = ACTIONS(1268),
    [anon_sym_LPAREN] = ACTIONS(1268),
    [anon_sym_LBRACE] = ACTIONS(1268),
    [anon_sym_RBRACE] = ACTIONS(1268),
    [anon_sym_LBRACK] = ACTIONS(1268),
    [anon_sym_STAR] = ACTIONS(1268),
    [anon_sym_u8] = ACTIONS(1270),
    [anon_sym_i8] = ACTIONS(1270),
    [anon_sym_u16] = ACTIONS(1270),
    [anon_sym_i16] = ACTIONS(1270),
    [anon_sym_u32] = ACTIONS(1270),
    [anon_sym_i32] = ACTIONS(1270),
    [anon_sym_u64] = ACTIONS(1270),
    [anon_sym_i64] = ACTIONS(1270),
    [anon_sym_u128] = ACTIONS(1270),
    [anon_sym_i128] = ACTIONS(1270),
    [anon_sym_isize] = ACTIONS(1270),
    [anon_sym_usize] = ACTIONS(1270),
    [anon_sym_f32] = ACTIONS(1270),
    [anon_sym_f64] = ACTIONS(1270),
    [anon_sym_bool] = ACTIONS(1270),
    [anon_sym_str] = ACTIONS(1270),
    [anon_sym_char] = ACTIONS(1270),
    [anon_sym_SQUOTE] = ACTIONS(1270),
    [anon_sym_async] = ACTIONS(1270),
    [anon_sym_break] = ACTIONS(1270),
    [anon_sym_const] = ACTIONS(1270),
    [anon_sym_continue] = ACTIONS(1270),
    [anon_sym_default] = ACTIONS(1270),
    [anon_sym_enum] = ACTIONS(1270),
    [anon_sym_fn] = ACTIONS(1270),
    [anon_sym_for] = ACTIONS(1270),
    [anon_sym_if] = ACTIONS(1270),
    [anon_sym_impl] = ACTIONS(1270),
    [anon_sym_let] = ACTIONS(1270),
    [anon_sym_loop] = ACTIONS(1270),
    [anon_sym_match] = ACTIONS(1270),
    [anon_sym_mod] = ACTIONS(1270),
    [anon_sym_pub] = ACTIONS(1270),
    [anon_sym_return] = ACTIONS(1270),
    [anon_sym_static] = ACTIONS(1270),
    [anon_sym_struct] = ACTIONS(1270),
    [anon_sym_trait] = ACTIONS(1270),
    [anon_sym_type] = ACTIONS(1270),
    [anon_sym_union] = ACTIONS(1270),
    [anon_sym_unsafe] = ACTIONS(1270),
    [anon_sym_use] = ACTIONS(1270),
    [anon_sym_while] = ACTIONS(1270),
    [anon_sym_POUND] = ACTIONS(1268),
    [anon_sym_BANG] = ACTIONS(1268),
    [anon_sym_extern] = ACTIONS(1270),
    [anon_sym_LT] = ACTIONS(1268),
    [anon_sym_COLON_COLON] = ACTIONS(1268),
    [anon_sym_AMP] = ACTIONS(1268),
    [anon_sym_DOT_DOT] = ACTIONS(1268),
    [anon_sym_DASH] = ACTIONS(1268),
    [anon_sym_PIPE] = ACTIONS(1268),
    [anon_sym_yield] = ACTIONS(1270),
    [anon_sym_move] = ACTIONS(1270),
    [sym_integer_literal] = ACTIONS(1268),
    [aux_sym_string_literal_token1] = ACTIONS(1268),
    [sym_char_literal] = ACTIONS(1268),
    [anon_sym_true] = ACTIONS(1270),
    [anon_sym_false] = ACTIONS(1270),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1270),
    [sym_super] = ACTIONS(1270),
    [sym_crate] = ACTIONS(1270),
    [sym_metavariable] = ACTIONS(1268),
    [sym_raw_string_literal] = ACTIONS(1268),
    [sym_float_literal] = ACTIONS(1268),
    [sym_block_comment] = ACTIONS(3),
  },
  [300] = {
    [ts_builtin_sym_end] = ACTIONS(1272),
    [sym_identifier] = ACTIONS(1274),
    [anon_sym_SEMI] = ACTIONS(1272),
    [anon_sym_macro_rules_BANG] = ACTIONS(1272),
    [anon_sym_LPAREN] = ACTIONS(1272),
    [anon_sym_LBRACE] = ACTIONS(1272),
    [anon_sym_RBRACE] = ACTIONS(1272),
    [anon_sym_LBRACK] = ACTIONS(1272),
    [anon_sym_STAR] = ACTIONS(1272),
    [anon_sym_u8] = ACTIONS(1274),
    [anon_sym_i8] = ACTIONS(1274),
    [anon_sym_u16] = ACTIONS(1274),
    [anon_sym_i16] = ACTIONS(1274),
    [anon_sym_u32] = ACTIONS(1274),
    [anon_sym_i32] = ACTIONS(1274),
    [anon_sym_u64] = ACTIONS(1274),
    [anon_sym_i64] = ACTIONS(1274),
    [anon_sym_u128] = ACTIONS(1274),
    [anon_sym_i128] = ACTIONS(1274),
    [anon_sym_isize] = ACTIONS(1274),
    [anon_sym_usize] = ACTIONS(1274),
    [anon_sym_f32] = ACTIONS(1274),
    [anon_sym_f64] = ACTIONS(1274),
    [anon_sym_bool] = ACTIONS(1274),
    [anon_sym_str] = ACTIONS(1274),
    [anon_sym_char] = ACTIONS(1274),
    [anon_sym_SQUOTE] = ACTIONS(1274),
    [anon_sym_async] = ACTIONS(1274),
    [anon_sym_break] = ACTIONS(1274),
    [anon_sym_const] = ACTIONS(1274),
    [anon_sym_continue] = ACTIONS(1274),
    [anon_sym_default] = ACTIONS(1274),
    [anon_sym_enum] = ACTIONS(1274),
    [anon_sym_fn] = ACTIONS(1274),
    [anon_sym_for] = ACTIONS(1274),
    [anon_sym_if] = ACTIONS(1274),
    [anon_sym_impl] = ACTIONS(1274),
    [anon_sym_let] = ACTIONS(1274),
    [anon_sym_loop] = ACTIONS(1274),
    [anon_sym_match] = ACTIONS(1274),
    [anon_sym_mod] = ACTIONS(1274),
    [anon_sym_pub] = ACTIONS(1274),
    [anon_sym_return] = ACTIONS(1274),
    [anon_sym_static] = ACTIONS(1274),
    [anon_sym_struct] = ACTIONS(1274),
    [anon_sym_trait] = ACTIONS(1274),
    [anon_sym_type] = ACTIONS(1274),
    [anon_sym_union] = ACTIONS(1274),
    [anon_sym_unsafe] = ACTIONS(1274),
    [anon_sym_use] = ACTIONS(1274),
    [anon_sym_while] = ACTIONS(1274),
    [anon_sym_POUND] = ACTIONS(1272),
    [anon_sym_BANG] = ACTIONS(1272),
    [anon_sym_extern] = ACTIONS(1274),
    [anon_sym_LT] = ACTIONS(1272),
    [anon_sym_COLON_COLON] = ACTIONS(1272),
    [anon_sym_AMP] = ACTIONS(1272),
    [anon_sym_DOT_DOT] = ACTIONS(1272),
    [anon_sym_DASH] = ACTIONS(1272),
    [anon_sym_PIPE] = ACTIONS(1272),
    [anon_sym_yield] = ACTIONS(1274),
    [anon_sym_move] = ACTIONS(1274),
    [sym_integer_literal] = ACTIONS(1272),
    [aux_sym_string_literal_token1] = ACTIONS(1272),
    [sym_char_literal] = ACTIONS(1272),
    [anon_sym_true] = ACTIONS(1274),
    [anon_sym_false] = ACTIONS(1274),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1274),
    [sym_super] = ACTIONS(1274),
    [sym_crate] = ACTIONS(1274),
    [sym_metavariable] = ACTIONS(1272),
    [sym_raw_string_literal] = ACTIONS(1272),
    [sym_float_literal] = ACTIONS(1272),
    [sym_block_comment] = ACTIONS(3),
  },
  [301] = {
    [ts_builtin_sym_end] = ACTIONS(1276),
    [sym_identifier] = ACTIONS(1278),
    [anon_sym_SEMI] = ACTIONS(1276),
    [anon_sym_macro_rules_BANG] = ACTIONS(1276),
    [anon_sym_LPAREN] = ACTIONS(1276),
    [anon_sym_LBRACE] = ACTIONS(1276),
    [anon_sym_RBRACE] = ACTIONS(1276),
    [anon_sym_LBRACK] = ACTIONS(1276),
    [anon_sym_STAR] = ACTIONS(1276),
    [anon_sym_u8] = ACTIONS(1278),
    [anon_sym_i8] = ACTIONS(1278),
    [anon_sym_u16] = ACTIONS(1278),
    [anon_sym_i16] = ACTIONS(1278),
    [anon_sym_u32] = ACTIONS(1278),
    [anon_sym_i32] = ACTIONS(1278),
    [anon_sym_u64] = ACTIONS(1278),
    [anon_sym_i64] = ACTIONS(1278),
    [anon_sym_u128] = ACTIONS(1278),
    [anon_sym_i128] = ACTIONS(1278),
    [anon_sym_isize] = ACTIONS(1278),
    [anon_sym_usize] = ACTIONS(1278),
    [anon_sym_f32] = ACTIONS(1278),
    [anon_sym_f64] = ACTIONS(1278),
    [anon_sym_bool] = ACTIONS(1278),
    [anon_sym_str] = ACTIONS(1278),
    [anon_sym_char] = ACTIONS(1278),
    [anon_sym_SQUOTE] = ACTIONS(1278),
    [anon_sym_async] = ACTIONS(1278),
    [anon_sym_break] = ACTIONS(1278),
    [anon_sym_const] = ACTIONS(1278),
    [anon_sym_continue] = ACTIONS(1278),
    [anon_sym_default] = ACTIONS(1278),
    [anon_sym_enum] = ACTIONS(1278),
    [anon_sym_fn] = ACTIONS(1278),
    [anon_sym_for] = ACTIONS(1278),
    [anon_sym_if] = ACTIONS(1278),
    [anon_sym_impl] = ACTIONS(1278),
    [anon_sym_let] = ACTIONS(1278),
    [anon_sym_loop] = ACTIONS(1278),
    [anon_sym_match] = ACTIONS(1278),
    [anon_sym_mod] = ACTIONS(1278),
    [anon_sym_pub] = ACTIONS(1278),
    [anon_sym_return] = ACTIONS(1278),
    [anon_sym_static] = ACTIONS(1278),
    [anon_sym_struct] = ACTIONS(1278),
    [anon_sym_trait] = ACTIONS(1278),
    [anon_sym_type] = ACTIONS(1278),
    [anon_sym_union] = ACTIONS(1278),
    [anon_sym_unsafe] = ACTIONS(1278),
    [anon_sym_use] = ACTIONS(1278),
    [anon_sym_while] = ACTIONS(1278),
    [anon_sym_POUND] = ACTIONS(1276),
    [anon_sym_BANG] = ACTIONS(1276),
    [anon_sym_extern] = ACTIONS(1278),
    [anon_sym_LT] = ACTIONS(1276),
    [anon_sym_COLON_COLON] = ACTIONS(1276),
    [anon_sym_AMP] = ACTIONS(1276),
    [anon_sym_DOT_DOT] = ACTIONS(1276),
    [anon_sym_DASH] = ACTIONS(1276),
    [anon_sym_PIPE] = ACTIONS(1276),
    [anon_sym_yield] = ACTIONS(1278),
    [anon_sym_move] = ACTIONS(1278),
    [sym_integer_literal] = ACTIONS(1276),
    [aux_sym_string_literal_token1] = ACTIONS(1276),
    [sym_char_literal] = ACTIONS(1276),
    [anon_sym_true] = ACTIONS(1278),
    [anon_sym_false] = ACTIONS(1278),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1278),
    [sym_super] = ACTIONS(1278),
    [sym_crate] = ACTIONS(1278),
    [sym_metavariable] = ACTIONS(1276),
    [sym_raw_string_literal] = ACTIONS(1276),
    [sym_float_literal] = ACTIONS(1276),
    [sym_block_comment] = ACTIONS(3),
  },
  [302] = {
    [ts_builtin_sym_end] = ACTIONS(1280),
    [sym_identifier] = ACTIONS(1282),
    [anon_sym_SEMI] = ACTIONS(1280),
    [anon_sym_macro_rules_BANG] = ACTIONS(1280),
    [anon_sym_LPAREN] = ACTIONS(1280),
    [anon_sym_LBRACE] = ACTIONS(1280),
    [anon_sym_RBRACE] = ACTIONS(1280),
    [anon_sym_LBRACK] = ACTIONS(1280),
    [anon_sym_STAR] = ACTIONS(1280),
    [anon_sym_u8] = ACTIONS(1282),
    [anon_sym_i8] = ACTIONS(1282),
    [anon_sym_u16] = ACTIONS(1282),
    [anon_sym_i16] = ACTIONS(1282),
    [anon_sym_u32] = ACTIONS(1282),
    [anon_sym_i32] = ACTIONS(1282),
    [anon_sym_u64] = ACTIONS(1282),
    [anon_sym_i64] = ACTIONS(1282),
    [anon_sym_u128] = ACTIONS(1282),
    [anon_sym_i128] = ACTIONS(1282),
    [anon_sym_isize] = ACTIONS(1282),
    [anon_sym_usize] = ACTIONS(1282),
    [anon_sym_f32] = ACTIONS(1282),
    [anon_sym_f64] = ACTIONS(1282),
    [anon_sym_bool] = ACTIONS(1282),
    [anon_sym_str] = ACTIONS(1282),
    [anon_sym_char] = ACTIONS(1282),
    [anon_sym_SQUOTE] = ACTIONS(1282),
    [anon_sym_async] = ACTIONS(1282),
    [anon_sym_break] = ACTIONS(1282),
    [anon_sym_const] = ACTIONS(1282),
    [anon_sym_continue] = ACTIONS(1282),
    [anon_sym_default] = ACTIONS(1282),
    [anon_sym_enum] = ACTIONS(1282),
    [anon_sym_fn] = ACTIONS(1282),
    [anon_sym_for] = ACTIONS(1282),
    [anon_sym_if] = ACTIONS(1282),
    [anon_sym_impl] = ACTIONS(1282),
    [anon_sym_let] = ACTIONS(1282),
    [anon_sym_loop] = ACTIONS(1282),
    [anon_sym_match] = ACTIONS(1282),
    [anon_sym_mod] = ACTIONS(1282),
    [anon_sym_pub] = ACTIONS(1282),
    [anon_sym_return] = ACTIONS(1282),
    [anon_sym_static] = ACTIONS(1282),
    [anon_sym_struct] = ACTIONS(1282),
    [anon_sym_trait] = ACTIONS(1282),
    [anon_sym_type] = ACTIONS(1282),
    [anon_sym_union] = ACTIONS(1282),
    [anon_sym_unsafe] = ACTIONS(1282),
    [anon_sym_use] = ACTIONS(1282),
    [anon_sym_while] = ACTIONS(1282),
    [anon_sym_POUND] = ACTIONS(1280),
    [anon_sym_BANG] = ACTIONS(1280),
    [anon_sym_extern] = ACTIONS(1282),
    [anon_sym_LT] = ACTIONS(1280),
    [anon_sym_COLON_COLON] = ACTIONS(1280),
    [anon_sym_AMP] = ACTIONS(1280),
    [anon_sym_DOT_DOT] = ACTIONS(1280),
    [anon_sym_DASH] = ACTIONS(1280),
    [anon_sym_PIPE] = ACTIONS(1280),
    [anon_sym_yield] = ACTIONS(1282),
    [anon_sym_move] = ACTIONS(1282),
    [sym_integer_literal] = ACTIONS(1280),
    [aux_sym_string_literal_token1] = ACTIONS(1280),
    [sym_char_literal] = ACTIONS(1280),
    [anon_sym_true] = ACTIONS(1282),
    [anon_sym_false] = ACTIONS(1282),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1282),
    [sym_super] = ACTIONS(1282),
    [sym_crate] = ACTIONS(1282),
    [sym_metavariable] = ACTIONS(1280),
    [sym_raw_string_literal] = ACTIONS(1280),
    [sym_float_literal] = ACTIONS(1280),
    [sym_block_comment] = ACTIONS(3),
  },
  [303] = {
    [ts_builtin_sym_end] = ACTIONS(1284),
    [sym_identifier] = ACTIONS(1286),
    [anon_sym_SEMI] = ACTIONS(1284),
    [anon_sym_macro_rules_BANG] = ACTIONS(1284),
    [anon_sym_LPAREN] = ACTIONS(1284),
    [anon_sym_LBRACE] = ACTIONS(1284),
    [anon_sym_RBRACE] = ACTIONS(1284),
    [anon_sym_LBRACK] = ACTIONS(1284),
    [anon_sym_STAR] = ACTIONS(1284),
    [anon_sym_u8] = ACTIONS(1286),
    [anon_sym_i8] = ACTIONS(1286),
    [anon_sym_u16] = ACTIONS(1286),
    [anon_sym_i16] = ACTIONS(1286),
    [anon_sym_u32] = ACTIONS(1286),
    [anon_sym_i32] = ACTIONS(1286),
    [anon_sym_u64] = ACTIONS(1286),
    [anon_sym_i64] = ACTIONS(1286),
    [anon_sym_u128] = ACTIONS(1286),
    [anon_sym_i128] = ACTIONS(1286),
    [anon_sym_isize] = ACTIONS(1286),
    [anon_sym_usize] = ACTIONS(1286),
    [anon_sym_f32] = ACTIONS(1286),
    [anon_sym_f64] = ACTIONS(1286),
    [anon_sym_bool] = ACTIONS(1286),
    [anon_sym_str] = ACTIONS(1286),
    [anon_sym_char] = ACTIONS(1286),
    [anon_sym_SQUOTE] = ACTIONS(1286),
    [anon_sym_async] = ACTIONS(1286),
    [anon_sym_break] = ACTIONS(1286),
    [anon_sym_const] = ACTIONS(1286),
    [anon_sym_continue] = ACTIONS(1286),
    [anon_sym_default] = ACTIONS(1286),
    [anon_sym_enum] = ACTIONS(1286),
    [anon_sym_fn] = ACTIONS(1286),
    [anon_sym_for] = ACTIONS(1286),
    [anon_sym_if] = ACTIONS(1286),
    [anon_sym_impl] = ACTIONS(1286),
    [anon_sym_let] = ACTIONS(1286),
    [anon_sym_loop] = ACTIONS(1286),
    [anon_sym_match] = ACTIONS(1286),
    [anon_sym_mod] = ACTIONS(1286),
    [anon_sym_pub] = ACTIONS(1286),
    [anon_sym_return] = ACTIONS(1286),
    [anon_sym_static] = ACTIONS(1286),
    [anon_sym_struct] = ACTIONS(1286),
    [anon_sym_trait] = ACTIONS(1286),
    [anon_sym_type] = ACTIONS(1286),
    [anon_sym_union] = ACTIONS(1286),
    [anon_sym_unsafe] = ACTIONS(1286),
    [anon_sym_use] = ACTIONS(1286),
    [anon_sym_while] = ACTIONS(1286),
    [anon_sym_POUND] = ACTIONS(1284),
    [anon_sym_BANG] = ACTIONS(1284),
    [anon_sym_extern] = ACTIONS(1286),
    [anon_sym_LT] = ACTIONS(1284),
    [anon_sym_COLON_COLON] = ACTIONS(1284),
    [anon_sym_AMP] = ACTIONS(1284),
    [anon_sym_DOT_DOT] = ACTIONS(1284),
    [anon_sym_DASH] = ACTIONS(1284),
    [anon_sym_PIPE] = ACTIONS(1284),
    [anon_sym_yield] = ACTIONS(1286),
    [anon_sym_move] = ACTIONS(1286),
    [sym_integer_literal] = ACTIONS(1284),
    [aux_sym_string_literal_token1] = ACTIONS(1284),
    [sym_char_literal] = ACTIONS(1284),
    [anon_sym_true] = ACTIONS(1286),
    [anon_sym_false] = ACTIONS(1286),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1286),
    [sym_super] = ACTIONS(1286),
    [sym_crate] = ACTIONS(1286),
    [sym_metavariable] = ACTIONS(1284),
    [sym_raw_string_literal] = ACTIONS(1284),
    [sym_float_literal] = ACTIONS(1284),
    [sym_block_comment] = ACTIONS(3),
  },
  [304] = {
    [ts_builtin_sym_end] = ACTIONS(1288),
    [sym_identifier] = ACTIONS(1290),
    [anon_sym_SEMI] = ACTIONS(1288),
    [anon_sym_macro_rules_BANG] = ACTIONS(1288),
    [anon_sym_LPAREN] = ACTIONS(1288),
    [anon_sym_LBRACE] = ACTIONS(1288),
    [anon_sym_RBRACE] = ACTIONS(1288),
    [anon_sym_LBRACK] = ACTIONS(1288),
    [anon_sym_STAR] = ACTIONS(1288),
    [anon_sym_u8] = ACTIONS(1290),
    [anon_sym_i8] = ACTIONS(1290),
    [anon_sym_u16] = ACTIONS(1290),
    [anon_sym_i16] = ACTIONS(1290),
    [anon_sym_u32] = ACTIONS(1290),
    [anon_sym_i32] = ACTIONS(1290),
    [anon_sym_u64] = ACTIONS(1290),
    [anon_sym_i64] = ACTIONS(1290),
    [anon_sym_u128] = ACTIONS(1290),
    [anon_sym_i128] = ACTIONS(1290),
    [anon_sym_isize] = ACTIONS(1290),
    [anon_sym_usize] = ACTIONS(1290),
    [anon_sym_f32] = ACTIONS(1290),
    [anon_sym_f64] = ACTIONS(1290),
    [anon_sym_bool] = ACTIONS(1290),
    [anon_sym_str] = ACTIONS(1290),
    [anon_sym_char] = ACTIONS(1290),
    [anon_sym_SQUOTE] = ACTIONS(1290),
    [anon_sym_async] = ACTIONS(1290),
    [anon_sym_break] = ACTIONS(1290),
    [anon_sym_const] = ACTIONS(1290),
    [anon_sym_continue] = ACTIONS(1290),
    [anon_sym_default] = ACTIONS(1290),
    [anon_sym_enum] = ACTIONS(1290),
    [anon_sym_fn] = ACTIONS(1290),
    [anon_sym_for] = ACTIONS(1290),
    [anon_sym_if] = ACTIONS(1290),
    [anon_sym_impl] = ACTIONS(1290),
    [anon_sym_let] = ACTIONS(1290),
    [anon_sym_loop] = ACTIONS(1290),
    [anon_sym_match] = ACTIONS(1290),
    [anon_sym_mod] = ACTIONS(1290),
    [anon_sym_pub] = ACTIONS(1290),
    [anon_sym_return] = ACTIONS(1290),
    [anon_sym_static] = ACTIONS(1290),
    [anon_sym_struct] = ACTIONS(1290),
    [anon_sym_trait] = ACTIONS(1290),
    [anon_sym_type] = ACTIONS(1290),
    [anon_sym_union] = ACTIONS(1290),
    [anon_sym_unsafe] = ACTIONS(1290),
    [anon_sym_use] = ACTIONS(1290),
    [anon_sym_while] = ACTIONS(1290),
    [anon_sym_POUND] = ACTIONS(1288),
    [anon_sym_BANG] = ACTIONS(1288),
    [anon_sym_extern] = ACTIONS(1290),
    [anon_sym_LT] = ACTIONS(1288),
    [anon_sym_COLON_COLON] = ACTIONS(1288),
    [anon_sym_AMP] = ACTIONS(1288),
    [anon_sym_DOT_DOT] = ACTIONS(1288),
    [anon_sym_DASH] = ACTIONS(1288),
    [anon_sym_PIPE] = ACTIONS(1288),
    [anon_sym_yield] = ACTIONS(1290),
    [anon_sym_move] = ACTIONS(1290),
    [sym_integer_literal] = ACTIONS(1288),
    [aux_sym_string_literal_token1] = ACTIONS(1288),
    [sym_char_literal] = ACTIONS(1288),
    [anon_sym_true] = ACTIONS(1290),
    [anon_sym_false] = ACTIONS(1290),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1290),
    [sym_super] = ACTIONS(1290),
    [sym_crate] = ACTIONS(1290),
    [sym_metavariable] = ACTIONS(1288),
    [sym_raw_string_literal] = ACTIONS(1288),
    [sym_float_literal] = ACTIONS(1288),
    [sym_block_comment] = ACTIONS(3),
  },
  [305] = {
    [ts_builtin_sym_end] = ACTIONS(1292),
    [sym_identifier] = ACTIONS(1294),
    [anon_sym_SEMI] = ACTIONS(1292),
    [anon_sym_macro_rules_BANG] = ACTIONS(1292),
    [anon_sym_LPAREN] = ACTIONS(1292),
    [anon_sym_LBRACE] = ACTIONS(1292),
    [anon_sym_RBRACE] = ACTIONS(1292),
    [anon_sym_LBRACK] = ACTIONS(1292),
    [anon_sym_STAR] = ACTIONS(1292),
    [anon_sym_u8] = ACTIONS(1294),
    [anon_sym_i8] = ACTIONS(1294),
    [anon_sym_u16] = ACTIONS(1294),
    [anon_sym_i16] = ACTIONS(1294),
    [anon_sym_u32] = ACTIONS(1294),
    [anon_sym_i32] = ACTIONS(1294),
    [anon_sym_u64] = ACTIONS(1294),
    [anon_sym_i64] = ACTIONS(1294),
    [anon_sym_u128] = ACTIONS(1294),
    [anon_sym_i128] = ACTIONS(1294),
    [anon_sym_isize] = ACTIONS(1294),
    [anon_sym_usize] = ACTIONS(1294),
    [anon_sym_f32] = ACTIONS(1294),
    [anon_sym_f64] = ACTIONS(1294),
    [anon_sym_bool] = ACTIONS(1294),
    [anon_sym_str] = ACTIONS(1294),
    [anon_sym_char] = ACTIONS(1294),
    [anon_sym_SQUOTE] = ACTIONS(1294),
    [anon_sym_async] = ACTIONS(1294),
    [anon_sym_break] = ACTIONS(1294),
    [anon_sym_const] = ACTIONS(1294),
    [anon_sym_continue] = ACTIONS(1294),
    [anon_sym_default] = ACTIONS(1294),
    [anon_sym_enum] = ACTIONS(1294),
    [anon_sym_fn] = ACTIONS(1294),
    [anon_sym_for] = ACTIONS(1294),
    [anon_sym_if] = ACTIONS(1294),
    [anon_sym_impl] = ACTIONS(1294),
    [anon_sym_let] = ACTIONS(1294),
    [anon_sym_loop] = ACTIONS(1294),
    [anon_sym_match] = ACTIONS(1294),
    [anon_sym_mod] = ACTIONS(1294),
    [anon_sym_pub] = ACTIONS(1294),
    [anon_sym_return] = ACTIONS(1294),
    [anon_sym_static] = ACTIONS(1294),
    [anon_sym_struct] = ACTIONS(1294),
    [anon_sym_trait] = ACTIONS(1294),
    [anon_sym_type] = ACTIONS(1294),
    [anon_sym_union] = ACTIONS(1294),
    [anon_sym_unsafe] = ACTIONS(1294),
    [anon_sym_use] = ACTIONS(1294),
    [anon_sym_while] = ACTIONS(1294),
    [anon_sym_POUND] = ACTIONS(1292),
    [anon_sym_BANG] = ACTIONS(1292),
    [anon_sym_extern] = ACTIONS(1294),
    [anon_sym_LT] = ACTIONS(1292),
    [anon_sym_COLON_COLON] = ACTIONS(1292),
    [anon_sym_AMP] = ACTIONS(1292),
    [anon_sym_DOT_DOT] = ACTIONS(1292),
    [anon_sym_DASH] = ACTIONS(1292),
    [anon_sym_PIPE] = ACTIONS(1292),
    [anon_sym_yield] = ACTIONS(1294),
    [anon_sym_move] = ACTIONS(1294),
    [sym_integer_literal] = ACTIONS(1292),
    [aux_sym_string_literal_token1] = ACTIONS(1292),
    [sym_char_literal] = ACTIONS(1292),
    [anon_sym_true] = ACTIONS(1294),
    [anon_sym_false] = ACTIONS(1294),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1294),
    [sym_super] = ACTIONS(1294),
    [sym_crate] = ACTIONS(1294),
    [sym_metavariable] = ACTIONS(1292),
    [sym_raw_string_literal] = ACTIONS(1292),
    [sym_float_literal] = ACTIONS(1292),
    [sym_block_comment] = ACTIONS(3),
  },
  [306] = {
    [ts_builtin_sym_end] = ACTIONS(1296),
    [sym_identifier] = ACTIONS(1298),
    [anon_sym_SEMI] = ACTIONS(1296),
    [anon_sym_macro_rules_BANG] = ACTIONS(1296),
    [anon_sym_LPAREN] = ACTIONS(1296),
    [anon_sym_LBRACE] = ACTIONS(1296),
    [anon_sym_RBRACE] = ACTIONS(1296),
    [anon_sym_LBRACK] = ACTIONS(1296),
    [anon_sym_STAR] = ACTIONS(1296),
    [anon_sym_u8] = ACTIONS(1298),
    [anon_sym_i8] = ACTIONS(1298),
    [anon_sym_u16] = ACTIONS(1298),
    [anon_sym_i16] = ACTIONS(1298),
    [anon_sym_u32] = ACTIONS(1298),
    [anon_sym_i32] = ACTIONS(1298),
    [anon_sym_u64] = ACTIONS(1298),
    [anon_sym_i64] = ACTIONS(1298),
    [anon_sym_u128] = ACTIONS(1298),
    [anon_sym_i128] = ACTIONS(1298),
    [anon_sym_isize] = ACTIONS(1298),
    [anon_sym_usize] = ACTIONS(1298),
    [anon_sym_f32] = ACTIONS(1298),
    [anon_sym_f64] = ACTIONS(1298),
    [anon_sym_bool] = ACTIONS(1298),
    [anon_sym_str] = ACTIONS(1298),
    [anon_sym_char] = ACTIONS(1298),
    [anon_sym_SQUOTE] = ACTIONS(1298),
    [anon_sym_async] = ACTIONS(1298),
    [anon_sym_break] = ACTIONS(1298),
    [anon_sym_const] = ACTIONS(1298),
    [anon_sym_continue] = ACTIONS(1298),
    [anon_sym_default] = ACTIONS(1298),
    [anon_sym_enum] = ACTIONS(1298),
    [anon_sym_fn] = ACTIONS(1298),
    [anon_sym_for] = ACTIONS(1298),
    [anon_sym_if] = ACTIONS(1298),
    [anon_sym_impl] = ACTIONS(1298),
    [anon_sym_let] = ACTIONS(1298),
    [anon_sym_loop] = ACTIONS(1298),
    [anon_sym_match] = ACTIONS(1298),
    [anon_sym_mod] = ACTIONS(1298),
    [anon_sym_pub] = ACTIONS(1298),
    [anon_sym_return] = ACTIONS(1298),
    [anon_sym_static] = ACTIONS(1298),
    [anon_sym_struct] = ACTIONS(1298),
    [anon_sym_trait] = ACTIONS(1298),
    [anon_sym_type] = ACTIONS(1298),
    [anon_sym_union] = ACTIONS(1298),
    [anon_sym_unsafe] = ACTIONS(1298),
    [anon_sym_use] = ACTIONS(1298),
    [anon_sym_while] = ACTIONS(1298),
    [anon_sym_POUND] = ACTIONS(1296),
    [anon_sym_BANG] = ACTIONS(1296),
    [anon_sym_extern] = ACTIONS(1298),
    [anon_sym_LT] = ACTIONS(1296),
    [anon_sym_COLON_COLON] = ACTIONS(1296),
    [anon_sym_AMP] = ACTIONS(1296),
    [anon_sym_DOT_DOT] = ACTIONS(1296),
    [anon_sym_DASH] = ACTIONS(1296),
    [anon_sym_PIPE] = ACTIONS(1296),
    [anon_sym_yield] = ACTIONS(1298),
    [anon_sym_move] = ACTIONS(1298),
    [sym_integer_literal] = ACTIONS(1296),
    [aux_sym_string_literal_token1] = ACTIONS(1296),
    [sym_char_literal] = ACTIONS(1296),
    [anon_sym_true] = ACTIONS(1298),
    [anon_sym_false] = ACTIONS(1298),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1298),
    [sym_super] = ACTIONS(1298),
    [sym_crate] = ACTIONS(1298),
    [sym_metavariable] = ACTIONS(1296),
    [sym_raw_string_literal] = ACTIONS(1296),
    [sym_float_literal] = ACTIONS(1296),
    [sym_block_comment] = ACTIONS(3),
  },
  [307] = {
    [ts_builtin_sym_end] = ACTIONS(1300),
    [sym_identifier] = ACTIONS(1302),
    [anon_sym_SEMI] = ACTIONS(1300),
    [anon_sym_macro_rules_BANG] = ACTIONS(1300),
    [anon_sym_LPAREN] = ACTIONS(1300),
    [anon_sym_LBRACE] = ACTIONS(1300),
    [anon_sym_RBRACE] = ACTIONS(1300),
    [anon_sym_LBRACK] = ACTIONS(1300),
    [anon_sym_STAR] = ACTIONS(1300),
    [anon_sym_u8] = ACTIONS(1302),
    [anon_sym_i8] = ACTIONS(1302),
    [anon_sym_u16] = ACTIONS(1302),
    [anon_sym_i16] = ACTIONS(1302),
    [anon_sym_u32] = ACTIONS(1302),
    [anon_sym_i32] = ACTIONS(1302),
    [anon_sym_u64] = ACTIONS(1302),
    [anon_sym_i64] = ACTIONS(1302),
    [anon_sym_u128] = ACTIONS(1302),
    [anon_sym_i128] = ACTIONS(1302),
    [anon_sym_isize] = ACTIONS(1302),
    [anon_sym_usize] = ACTIONS(1302),
    [anon_sym_f32] = ACTIONS(1302),
    [anon_sym_f64] = ACTIONS(1302),
    [anon_sym_bool] = ACTIONS(1302),
    [anon_sym_str] = ACTIONS(1302),
    [anon_sym_char] = ACTIONS(1302),
    [anon_sym_SQUOTE] = ACTIONS(1302),
    [anon_sym_async] = ACTIONS(1302),
    [anon_sym_break] = ACTIONS(1302),
    [anon_sym_const] = ACTIONS(1302),
    [anon_sym_continue] = ACTIONS(1302),
    [anon_sym_default] = ACTIONS(1302),
    [anon_sym_enum] = ACTIONS(1302),
    [anon_sym_fn] = ACTIONS(1302),
    [anon_sym_for] = ACTIONS(1302),
    [anon_sym_if] = ACTIONS(1302),
    [anon_sym_impl] = ACTIONS(1302),
    [anon_sym_let] = ACTIONS(1302),
    [anon_sym_loop] = ACTIONS(1302),
    [anon_sym_match] = ACTIONS(1302),
    [anon_sym_mod] = ACTIONS(1302),
    [anon_sym_pub] = ACTIONS(1302),
    [anon_sym_return] = ACTIONS(1302),
    [anon_sym_static] = ACTIONS(1302),
    [anon_sym_struct] = ACTIONS(1302),
    [anon_sym_trait] = ACTIONS(1302),
    [anon_sym_type] = ACTIONS(1302),
    [anon_sym_union] = ACTIONS(1302),
    [anon_sym_unsafe] = ACTIONS(1302),
    [anon_sym_use] = ACTIONS(1302),
    [anon_sym_while] = ACTIONS(1302),
    [anon_sym_POUND] = ACTIONS(1300),
    [anon_sym_BANG] = ACTIONS(1300),
    [anon_sym_extern] = ACTIONS(1302),
    [anon_sym_LT] = ACTIONS(1300),
    [anon_sym_COLON_COLON] = ACTIONS(1300),
    [anon_sym_AMP] = ACTIONS(1300),
    [anon_sym_DOT_DOT] = ACTIONS(1300),
    [anon_sym_DASH] = ACTIONS(1300),
    [anon_sym_PIPE] = ACTIONS(1300),
    [anon_sym_yield] = ACTIONS(1302),
    [anon_sym_move] = ACTIONS(1302),
    [sym_integer_literal] = ACTIONS(1300),
    [aux_sym_string_literal_token1] = ACTIONS(1300),
    [sym_char_literal] = ACTIONS(1300),
    [anon_sym_true] = ACTIONS(1302),
    [anon_sym_false] = ACTIONS(1302),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1302),
    [sym_super] = ACTIONS(1302),
    [sym_crate] = ACTIONS(1302),
    [sym_metavariable] = ACTIONS(1300),
    [sym_raw_string_literal] = ACTIONS(1300),
    [sym_float_literal] = ACTIONS(1300),
    [sym_block_comment] = ACTIONS(3),
  },
  [308] = {
    [ts_builtin_sym_end] = ACTIONS(1304),
    [sym_identifier] = ACTIONS(1306),
    [anon_sym_SEMI] = ACTIONS(1304),
    [anon_sym_macro_rules_BANG] = ACTIONS(1304),
    [anon_sym_LPAREN] = ACTIONS(1304),
    [anon_sym_LBRACE] = ACTIONS(1304),
    [anon_sym_RBRACE] = ACTIONS(1304),
    [anon_sym_LBRACK] = ACTIONS(1304),
    [anon_sym_STAR] = ACTIONS(1304),
    [anon_sym_u8] = ACTIONS(1306),
    [anon_sym_i8] = ACTIONS(1306),
    [anon_sym_u16] = ACTIONS(1306),
    [anon_sym_i16] = ACTIONS(1306),
    [anon_sym_u32] = ACTIONS(1306),
    [anon_sym_i32] = ACTIONS(1306),
    [anon_sym_u64] = ACTIONS(1306),
    [anon_sym_i64] = ACTIONS(1306),
    [anon_sym_u128] = ACTIONS(1306),
    [anon_sym_i128] = ACTIONS(1306),
    [anon_sym_isize] = ACTIONS(1306),
    [anon_sym_usize] = ACTIONS(1306),
    [anon_sym_f32] = ACTIONS(1306),
    [anon_sym_f64] = ACTIONS(1306),
    [anon_sym_bool] = ACTIONS(1306),
    [anon_sym_str] = ACTIONS(1306),
    [anon_sym_char] = ACTIONS(1306),
    [anon_sym_SQUOTE] = ACTIONS(1306),
    [anon_sym_async] = ACTIONS(1306),
    [anon_sym_break] = ACTIONS(1306),
    [anon_sym_const] = ACTIONS(1306),
    [anon_sym_continue] = ACTIONS(1306),
    [anon_sym_default] = ACTIONS(1306),
    [anon_sym_enum] = ACTIONS(1306),
    [anon_sym_fn] = ACTIONS(1306),
    [anon_sym_for] = ACTIONS(1306),
    [anon_sym_if] = ACTIONS(1306),
    [anon_sym_impl] = ACTIONS(1306),
    [anon_sym_let] = ACTIONS(1306),
    [anon_sym_loop] = ACTIONS(1306),
    [anon_sym_match] = ACTIONS(1306),
    [anon_sym_mod] = ACTIONS(1306),
    [anon_sym_pub] = ACTIONS(1306),
    [anon_sym_return] = ACTIONS(1306),
    [anon_sym_static] = ACTIONS(1306),
    [anon_sym_struct] = ACTIONS(1306),
    [anon_sym_trait] = ACTIONS(1306),
    [anon_sym_type] = ACTIONS(1306),
    [anon_sym_union] = ACTIONS(1306),
    [anon_sym_unsafe] = ACTIONS(1306),
    [anon_sym_use] = ACTIONS(1306),
    [anon_sym_while] = ACTIONS(1306),
    [anon_sym_POUND] = ACTIONS(1304),
    [anon_sym_BANG] = ACTIONS(1304),
    [anon_sym_extern] = ACTIONS(1306),
    [anon_sym_LT] = ACTIONS(1304),
    [anon_sym_COLON_COLON] = ACTIONS(1304),
    [anon_sym_AMP] = ACTIONS(1304),
    [anon_sym_DOT_DOT] = ACTIONS(1304),
    [anon_sym_DASH] = ACTIONS(1304),
    [anon_sym_PIPE] = ACTIONS(1304),
    [anon_sym_yield] = ACTIONS(1306),
    [anon_sym_move] = ACTIONS(1306),
    [sym_integer_literal] = ACTIONS(1304),
    [aux_sym_string_literal_token1] = ACTIONS(1304),
    [sym_char_literal] = ACTIONS(1304),
    [anon_sym_true] = ACTIONS(1306),
    [anon_sym_false] = ACTIONS(1306),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1306),
    [sym_super] = ACTIONS(1306),
    [sym_crate] = ACTIONS(1306),
    [sym_metavariable] = ACTIONS(1304),
    [sym_raw_string_literal] = ACTIONS(1304),
    [sym_float_literal] = ACTIONS(1304),
    [sym_block_comment] = ACTIONS(3),
  },
  [309] = {
    [ts_builtin_sym_end] = ACTIONS(1308),
    [sym_identifier] = ACTIONS(1310),
    [anon_sym_SEMI] = ACTIONS(1308),
    [anon_sym_macro_rules_BANG] = ACTIONS(1308),
    [anon_sym_LPAREN] = ACTIONS(1308),
    [anon_sym_LBRACE] = ACTIONS(1308),
    [anon_sym_RBRACE] = ACTIONS(1308),
    [anon_sym_LBRACK] = ACTIONS(1308),
    [anon_sym_STAR] = ACTIONS(1308),
    [anon_sym_u8] = ACTIONS(1310),
    [anon_sym_i8] = ACTIONS(1310),
    [anon_sym_u16] = ACTIONS(1310),
    [anon_sym_i16] = ACTIONS(1310),
    [anon_sym_u32] = ACTIONS(1310),
    [anon_sym_i32] = ACTIONS(1310),
    [anon_sym_u64] = ACTIONS(1310),
    [anon_sym_i64] = ACTIONS(1310),
    [anon_sym_u128] = ACTIONS(1310),
    [anon_sym_i128] = ACTIONS(1310),
    [anon_sym_isize] = ACTIONS(1310),
    [anon_sym_usize] = ACTIONS(1310),
    [anon_sym_f32] = ACTIONS(1310),
    [anon_sym_f64] = ACTIONS(1310),
    [anon_sym_bool] = ACTIONS(1310),
    [anon_sym_str] = ACTIONS(1310),
    [anon_sym_char] = ACTIONS(1310),
    [anon_sym_SQUOTE] = ACTIONS(1310),
    [anon_sym_async] = ACTIONS(1310),
    [anon_sym_break] = ACTIONS(1310),
    [anon_sym_const] = ACTIONS(1310),
    [anon_sym_continue] = ACTIONS(1310),
    [anon_sym_default] = ACTIONS(1310),
    [anon_sym_enum] = ACTIONS(1310),
    [anon_sym_fn] = ACTIONS(1310),
    [anon_sym_for] = ACTIONS(1310),
    [anon_sym_if] = ACTIONS(1310),
    [anon_sym_impl] = ACTIONS(1310),
    [anon_sym_let] = ACTIONS(1310),
    [anon_sym_loop] = ACTIONS(1310),
    [anon_sym_match] = ACTIONS(1310),
    [anon_sym_mod] = ACTIONS(1310),
    [anon_sym_pub] = ACTIONS(1310),
    [anon_sym_return] = ACTIONS(1310),
    [anon_sym_static] = ACTIONS(1310),
    [anon_sym_struct] = ACTIONS(1310),
    [anon_sym_trait] = ACTIONS(1310),
    [anon_sym_type] = ACTIONS(1310),
    [anon_sym_union] = ACTIONS(1310),
    [anon_sym_unsafe] = ACTIONS(1310),
    [anon_sym_use] = ACTIONS(1310),
    [anon_sym_while] = ACTIONS(1310),
    [anon_sym_POUND] = ACTIONS(1308),
    [anon_sym_BANG] = ACTIONS(1308),
    [anon_sym_extern] = ACTIONS(1310),
    [anon_sym_LT] = ACTIONS(1308),
    [anon_sym_COLON_COLON] = ACTIONS(1308),
    [anon_sym_AMP] = ACTIONS(1308),
    [anon_sym_DOT_DOT] = ACTIONS(1308),
    [anon_sym_DASH] = ACTIONS(1308),
    [anon_sym_PIPE] = ACTIONS(1308),
    [anon_sym_yield] = ACTIONS(1310),
    [anon_sym_move] = ACTIONS(1310),
    [sym_integer_literal] = ACTIONS(1308),
    [aux_sym_string_literal_token1] = ACTIONS(1308),
    [sym_char_literal] = ACTIONS(1308),
    [anon_sym_true] = ACTIONS(1310),
    [anon_sym_false] = ACTIONS(1310),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1310),
    [sym_super] = ACTIONS(1310),
    [sym_crate] = ACTIONS(1310),
    [sym_metavariable] = ACTIONS(1308),
    [sym_raw_string_literal] = ACTIONS(1308),
    [sym_float_literal] = ACTIONS(1308),
    [sym_block_comment] = ACTIONS(3),
  },
  [310] = {
    [ts_builtin_sym_end] = ACTIONS(1312),
    [sym_identifier] = ACTIONS(1314),
    [anon_sym_SEMI] = ACTIONS(1312),
    [anon_sym_macro_rules_BANG] = ACTIONS(1312),
    [anon_sym_LPAREN] = ACTIONS(1312),
    [anon_sym_LBRACE] = ACTIONS(1312),
    [anon_sym_RBRACE] = ACTIONS(1312),
    [anon_sym_LBRACK] = ACTIONS(1312),
    [anon_sym_STAR] = ACTIONS(1312),
    [anon_sym_u8] = ACTIONS(1314),
    [anon_sym_i8] = ACTIONS(1314),
    [anon_sym_u16] = ACTIONS(1314),
    [anon_sym_i16] = ACTIONS(1314),
    [anon_sym_u32] = ACTIONS(1314),
    [anon_sym_i32] = ACTIONS(1314),
    [anon_sym_u64] = ACTIONS(1314),
    [anon_sym_i64] = ACTIONS(1314),
    [anon_sym_u128] = ACTIONS(1314),
    [anon_sym_i128] = ACTIONS(1314),
    [anon_sym_isize] = ACTIONS(1314),
    [anon_sym_usize] = ACTIONS(1314),
    [anon_sym_f32] = ACTIONS(1314),
    [anon_sym_f64] = ACTIONS(1314),
    [anon_sym_bool] = ACTIONS(1314),
    [anon_sym_str] = ACTIONS(1314),
    [anon_sym_char] = ACTIONS(1314),
    [anon_sym_SQUOTE] = ACTIONS(1314),
    [anon_sym_async] = ACTIONS(1314),
    [anon_sym_break] = ACTIONS(1314),
    [anon_sym_const] = ACTIONS(1314),
    [anon_sym_continue] = ACTIONS(1314),
    [anon_sym_default] = ACTIONS(1314),
    [anon_sym_enum] = ACTIONS(1314),
    [anon_sym_fn] = ACTIONS(1314),
    [anon_sym_for] = ACTIONS(1314),
    [anon_sym_if] = ACTIONS(1314),
    [anon_sym_impl] = ACTIONS(1314),
    [anon_sym_let] = ACTIONS(1314),
    [anon_sym_loop] = ACTIONS(1314),
    [anon_sym_match] = ACTIONS(1314),
    [anon_sym_mod] = ACTIONS(1314),
    [anon_sym_pub] = ACTIONS(1314),
    [anon_sym_return] = ACTIONS(1314),
    [anon_sym_static] = ACTIONS(1314),
    [anon_sym_struct] = ACTIONS(1314),
    [anon_sym_trait] = ACTIONS(1314),
    [anon_sym_type] = ACTIONS(1314),
    [anon_sym_union] = ACTIONS(1314),
    [anon_sym_unsafe] = ACTIONS(1314),
    [anon_sym_use] = ACTIONS(1314),
    [anon_sym_while] = ACTIONS(1314),
    [anon_sym_POUND] = ACTIONS(1312),
    [anon_sym_BANG] = ACTIONS(1312),
    [anon_sym_extern] = ACTIONS(1314),
    [anon_sym_LT] = ACTIONS(1312),
    [anon_sym_COLON_COLON] = ACTIONS(1312),
    [anon_sym_AMP] = ACTIONS(1312),
    [anon_sym_DOT_DOT] = ACTIONS(1312),
    [anon_sym_DASH] = ACTIONS(1312),
    [anon_sym_PIPE] = ACTIONS(1312),
    [anon_sym_yield] = ACTIONS(1314),
    [anon_sym_move] = ACTIONS(1314),
    [sym_integer_literal] = ACTIONS(1312),
    [aux_sym_string_literal_token1] = ACTIONS(1312),
    [sym_char_literal] = ACTIONS(1312),
    [anon_sym_true] = ACTIONS(1314),
    [anon_sym_false] = ACTIONS(1314),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1314),
    [sym_super] = ACTIONS(1314),
    [sym_crate] = ACTIONS(1314),
    [sym_metavariable] = ACTIONS(1312),
    [sym_raw_string_literal] = ACTIONS(1312),
    [sym_float_literal] = ACTIONS(1312),
    [sym_block_comment] = ACTIONS(3),
  },
  [311] = {
    [ts_builtin_sym_end] = ACTIONS(1316),
    [sym_identifier] = ACTIONS(1318),
    [anon_sym_SEMI] = ACTIONS(1316),
    [anon_sym_macro_rules_BANG] = ACTIONS(1316),
    [anon_sym_LPAREN] = ACTIONS(1316),
    [anon_sym_LBRACE] = ACTIONS(1316),
    [anon_sym_RBRACE] = ACTIONS(1316),
    [anon_sym_LBRACK] = ACTIONS(1316),
    [anon_sym_STAR] = ACTIONS(1316),
    [anon_sym_u8] = ACTIONS(1318),
    [anon_sym_i8] = ACTIONS(1318),
    [anon_sym_u16] = ACTIONS(1318),
    [anon_sym_i16] = ACTIONS(1318),
    [anon_sym_u32] = ACTIONS(1318),
    [anon_sym_i32] = ACTIONS(1318),
    [anon_sym_u64] = ACTIONS(1318),
    [anon_sym_i64] = ACTIONS(1318),
    [anon_sym_u128] = ACTIONS(1318),
    [anon_sym_i128] = ACTIONS(1318),
    [anon_sym_isize] = ACTIONS(1318),
    [anon_sym_usize] = ACTIONS(1318),
    [anon_sym_f32] = ACTIONS(1318),
    [anon_sym_f64] = ACTIONS(1318),
    [anon_sym_bool] = ACTIONS(1318),
    [anon_sym_str] = ACTIONS(1318),
    [anon_sym_char] = ACTIONS(1318),
    [anon_sym_SQUOTE] = ACTIONS(1318),
    [anon_sym_async] = ACTIONS(1318),
    [anon_sym_break] = ACTIONS(1318),
    [anon_sym_const] = ACTIONS(1318),
    [anon_sym_continue] = ACTIONS(1318),
    [anon_sym_default] = ACTIONS(1318),
    [anon_sym_enum] = ACTIONS(1318),
    [anon_sym_fn] = ACTIONS(1318),
    [anon_sym_for] = ACTIONS(1318),
    [anon_sym_if] = ACTIONS(1318),
    [anon_sym_impl] = ACTIONS(1318),
    [anon_sym_let] = ACTIONS(1318),
    [anon_sym_loop] = ACTIONS(1318),
    [anon_sym_match] = ACTIONS(1318),
    [anon_sym_mod] = ACTIONS(1318),
    [anon_sym_pub] = ACTIONS(1318),
    [anon_sym_return] = ACTIONS(1318),
    [anon_sym_static] = ACTIONS(1318),
    [anon_sym_struct] = ACTIONS(1318),
    [anon_sym_trait] = ACTIONS(1318),
    [anon_sym_type] = ACTIONS(1318),
    [anon_sym_union] = ACTIONS(1318),
    [anon_sym_unsafe] = ACTIONS(1318),
    [anon_sym_use] = ACTIONS(1318),
    [anon_sym_while] = ACTIONS(1318),
    [anon_sym_POUND] = ACTIONS(1316),
    [anon_sym_BANG] = ACTIONS(1316),
    [anon_sym_extern] = ACTIONS(1318),
    [anon_sym_LT] = ACTIONS(1316),
    [anon_sym_COLON_COLON] = ACTIONS(1316),
    [anon_sym_AMP] = ACTIONS(1316),
    [anon_sym_DOT_DOT] = ACTIONS(1316),
    [anon_sym_DASH] = ACTIONS(1316),
    [anon_sym_PIPE] = ACTIONS(1316),
    [anon_sym_yield] = ACTIONS(1318),
    [anon_sym_move] = ACTIONS(1318),
    [sym_integer_literal] = ACTIONS(1316),
    [aux_sym_string_literal_token1] = ACTIONS(1316),
    [sym_char_literal] = ACTIONS(1316),
    [anon_sym_true] = ACTIONS(1318),
    [anon_sym_false] = ACTIONS(1318),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1318),
    [sym_super] = ACTIONS(1318),
    [sym_crate] = ACTIONS(1318),
    [sym_metavariable] = ACTIONS(1316),
    [sym_raw_string_literal] = ACTIONS(1316),
    [sym_float_literal] = ACTIONS(1316),
    [sym_block_comment] = ACTIONS(3),
  },
  [312] = {
    [ts_builtin_sym_end] = ACTIONS(1320),
    [sym_identifier] = ACTIONS(1322),
    [anon_sym_SEMI] = ACTIONS(1320),
    [anon_sym_macro_rules_BANG] = ACTIONS(1320),
    [anon_sym_LPAREN] = ACTIONS(1320),
    [anon_sym_LBRACE] = ACTIONS(1320),
    [anon_sym_RBRACE] = ACTIONS(1320),
    [anon_sym_LBRACK] = ACTIONS(1320),
    [anon_sym_STAR] = ACTIONS(1320),
    [anon_sym_u8] = ACTIONS(1322),
    [anon_sym_i8] = ACTIONS(1322),
    [anon_sym_u16] = ACTIONS(1322),
    [anon_sym_i16] = ACTIONS(1322),
    [anon_sym_u32] = ACTIONS(1322),
    [anon_sym_i32] = ACTIONS(1322),
    [anon_sym_u64] = ACTIONS(1322),
    [anon_sym_i64] = ACTIONS(1322),
    [anon_sym_u128] = ACTIONS(1322),
    [anon_sym_i128] = ACTIONS(1322),
    [anon_sym_isize] = ACTIONS(1322),
    [anon_sym_usize] = ACTIONS(1322),
    [anon_sym_f32] = ACTIONS(1322),
    [anon_sym_f64] = ACTIONS(1322),
    [anon_sym_bool] = ACTIONS(1322),
    [anon_sym_str] = ACTIONS(1322),
    [anon_sym_char] = ACTIONS(1322),
    [anon_sym_SQUOTE] = ACTIONS(1322),
    [anon_sym_async] = ACTIONS(1322),
    [anon_sym_break] = ACTIONS(1322),
    [anon_sym_const] = ACTIONS(1322),
    [anon_sym_continue] = ACTIONS(1322),
    [anon_sym_default] = ACTIONS(1322),
    [anon_sym_enum] = ACTIONS(1322),
    [anon_sym_fn] = ACTIONS(1322),
    [anon_sym_for] = ACTIONS(1322),
    [anon_sym_if] = ACTIONS(1322),
    [anon_sym_impl] = ACTIONS(1322),
    [anon_sym_let] = ACTIONS(1322),
    [anon_sym_loop] = ACTIONS(1322),
    [anon_sym_match] = ACTIONS(1322),
    [anon_sym_mod] = ACTIONS(1322),
    [anon_sym_pub] = ACTIONS(1322),
    [anon_sym_return] = ACTIONS(1322),
    [anon_sym_static] = ACTIONS(1322),
    [anon_sym_struct] = ACTIONS(1322),
    [anon_sym_trait] = ACTIONS(1322),
    [anon_sym_type] = ACTIONS(1322),
    [anon_sym_union] = ACTIONS(1322),
    [anon_sym_unsafe] = ACTIONS(1322),
    [anon_sym_use] = ACTIONS(1322),
    [anon_sym_while] = ACTIONS(1322),
    [anon_sym_POUND] = ACTIONS(1320),
    [anon_sym_BANG] = ACTIONS(1320),
    [anon_sym_extern] = ACTIONS(1322),
    [anon_sym_LT] = ACTIONS(1320),
    [anon_sym_COLON_COLON] = ACTIONS(1320),
    [anon_sym_AMP] = ACTIONS(1320),
    [anon_sym_DOT_DOT] = ACTIONS(1320),
    [anon_sym_DASH] = ACTIONS(1320),
    [anon_sym_PIPE] = ACTIONS(1320),
    [anon_sym_yield] = ACTIONS(1322),
    [anon_sym_move] = ACTIONS(1322),
    [sym_integer_literal] = ACTIONS(1320),
    [aux_sym_string_literal_token1] = ACTIONS(1320),
    [sym_char_literal] = ACTIONS(1320),
    [anon_sym_true] = ACTIONS(1322),
    [anon_sym_false] = ACTIONS(1322),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1322),
    [sym_super] = ACTIONS(1322),
    [sym_crate] = ACTIONS(1322),
    [sym_metavariable] = ACTIONS(1320),
    [sym_raw_string_literal] = ACTIONS(1320),
    [sym_float_literal] = ACTIONS(1320),
    [sym_block_comment] = ACTIONS(3),
  },
  [313] = {
    [ts_builtin_sym_end] = ACTIONS(1324),
    [sym_identifier] = ACTIONS(1326),
    [anon_sym_SEMI] = ACTIONS(1324),
    [anon_sym_macro_rules_BANG] = ACTIONS(1324),
    [anon_sym_LPAREN] = ACTIONS(1324),
    [anon_sym_LBRACE] = ACTIONS(1324),
    [anon_sym_RBRACE] = ACTIONS(1324),
    [anon_sym_LBRACK] = ACTIONS(1324),
    [anon_sym_STAR] = ACTIONS(1324),
    [anon_sym_u8] = ACTIONS(1326),
    [anon_sym_i8] = ACTIONS(1326),
    [anon_sym_u16] = ACTIONS(1326),
    [anon_sym_i16] = ACTIONS(1326),
    [anon_sym_u32] = ACTIONS(1326),
    [anon_sym_i32] = ACTIONS(1326),
    [anon_sym_u64] = ACTIONS(1326),
    [anon_sym_i64] = ACTIONS(1326),
    [anon_sym_u128] = ACTIONS(1326),
    [anon_sym_i128] = ACTIONS(1326),
    [anon_sym_isize] = ACTIONS(1326),
    [anon_sym_usize] = ACTIONS(1326),
    [anon_sym_f32] = ACTIONS(1326),
    [anon_sym_f64] = ACTIONS(1326),
    [anon_sym_bool] = ACTIONS(1326),
    [anon_sym_str] = ACTIONS(1326),
    [anon_sym_char] = ACTIONS(1326),
    [anon_sym_SQUOTE] = ACTIONS(1326),
    [anon_sym_async] = ACTIONS(1326),
    [anon_sym_break] = ACTIONS(1326),
    [anon_sym_const] = ACTIONS(1326),
    [anon_sym_continue] = ACTIONS(1326),
    [anon_sym_default] = ACTIONS(1326),
    [anon_sym_enum] = ACTIONS(1326),
    [anon_sym_fn] = ACTIONS(1326),
    [anon_sym_for] = ACTIONS(1326),
    [anon_sym_if] = ACTIONS(1326),
    [anon_sym_impl] = ACTIONS(1326),
    [anon_sym_let] = ACTIONS(1326),
    [anon_sym_loop] = ACTIONS(1326),
    [anon_sym_match] = ACTIONS(1326),
    [anon_sym_mod] = ACTIONS(1326),
    [anon_sym_pub] = ACTIONS(1326),
    [anon_sym_return] = ACTIONS(1326),
    [anon_sym_static] = ACTIONS(1326),
    [anon_sym_struct] = ACTIONS(1326),
    [anon_sym_trait] = ACTIONS(1326),
    [anon_sym_type] = ACTIONS(1326),
    [anon_sym_union] = ACTIONS(1326),
    [anon_sym_unsafe] = ACTIONS(1326),
    [anon_sym_use] = ACTIONS(1326),
    [anon_sym_while] = ACTIONS(1326),
    [anon_sym_POUND] = ACTIONS(1324),
    [anon_sym_BANG] = ACTIONS(1324),
    [anon_sym_extern] = ACTIONS(1326),
    [anon_sym_LT] = ACTIONS(1324),
    [anon_sym_COLON_COLON] = ACTIONS(1324),
    [anon_sym_AMP] = ACTIONS(1324),
    [anon_sym_DOT_DOT] = ACTIONS(1324),
    [anon_sym_DASH] = ACTIONS(1324),
    [anon_sym_PIPE] = ACTIONS(1324),
    [anon_sym_yield] = ACTIONS(1326),
    [anon_sym_move] = ACTIONS(1326),
    [sym_integer_literal] = ACTIONS(1324),
    [aux_sym_string_literal_token1] = ACTIONS(1324),
    [sym_char_literal] = ACTIONS(1324),
    [anon_sym_true] = ACTIONS(1326),
    [anon_sym_false] = ACTIONS(1326),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1326),
    [sym_super] = ACTIONS(1326),
    [sym_crate] = ACTIONS(1326),
    [sym_metavariable] = ACTIONS(1324),
    [sym_raw_string_literal] = ACTIONS(1324),
    [sym_float_literal] = ACTIONS(1324),
    [sym_block_comment] = ACTIONS(3),
  },
  [314] = {
    [ts_builtin_sym_end] = ACTIONS(1328),
    [sym_identifier] = ACTIONS(1330),
    [anon_sym_SEMI] = ACTIONS(1328),
    [anon_sym_macro_rules_BANG] = ACTIONS(1328),
    [anon_sym_LPAREN] = ACTIONS(1328),
    [anon_sym_LBRACE] = ACTIONS(1328),
    [anon_sym_RBRACE] = ACTIONS(1328),
    [anon_sym_LBRACK] = ACTIONS(1328),
    [anon_sym_STAR] = ACTIONS(1328),
    [anon_sym_u8] = ACTIONS(1330),
    [anon_sym_i8] = ACTIONS(1330),
    [anon_sym_u16] = ACTIONS(1330),
    [anon_sym_i16] = ACTIONS(1330),
    [anon_sym_u32] = ACTIONS(1330),
    [anon_sym_i32] = ACTIONS(1330),
    [anon_sym_u64] = ACTIONS(1330),
    [anon_sym_i64] = ACTIONS(1330),
    [anon_sym_u128] = ACTIONS(1330),
    [anon_sym_i128] = ACTIONS(1330),
    [anon_sym_isize] = ACTIONS(1330),
    [anon_sym_usize] = ACTIONS(1330),
    [anon_sym_f32] = ACTIONS(1330),
    [anon_sym_f64] = ACTIONS(1330),
    [anon_sym_bool] = ACTIONS(1330),
    [anon_sym_str] = ACTIONS(1330),
    [anon_sym_char] = ACTIONS(1330),
    [anon_sym_SQUOTE] = ACTIONS(1330),
    [anon_sym_async] = ACTIONS(1330),
    [anon_sym_break] = ACTIONS(1330),
    [anon_sym_const] = ACTIONS(1330),
    [anon_sym_continue] = ACTIONS(1330),
    [anon_sym_default] = ACTIONS(1330),
    [anon_sym_enum] = ACTIONS(1330),
    [anon_sym_fn] = ACTIONS(1330),
    [anon_sym_for] = ACTIONS(1330),
    [anon_sym_if] = ACTIONS(1330),
    [anon_sym_impl] = ACTIONS(1330),
    [anon_sym_let] = ACTIONS(1330),
    [anon_sym_loop] = ACTIONS(1330),
    [anon_sym_match] = ACTIONS(1330),
    [anon_sym_mod] = ACTIONS(1330),
    [anon_sym_pub] = ACTIONS(1330),
    [anon_sym_return] = ACTIONS(1330),
    [anon_sym_static] = ACTIONS(1330),
    [anon_sym_struct] = ACTIONS(1330),
    [anon_sym_trait] = ACTIONS(1330),
    [anon_sym_type] = ACTIONS(1330),
    [anon_sym_union] = ACTIONS(1330),
    [anon_sym_unsafe] = ACTIONS(1330),
    [anon_sym_use] = ACTIONS(1330),
    [anon_sym_while] = ACTIONS(1330),
    [anon_sym_POUND] = ACTIONS(1328),
    [anon_sym_BANG] = ACTIONS(1328),
    [anon_sym_extern] = ACTIONS(1330),
    [anon_sym_LT] = ACTIONS(1328),
    [anon_sym_COLON_COLON] = ACTIONS(1328),
    [anon_sym_AMP] = ACTIONS(1328),
    [anon_sym_DOT_DOT] = ACTIONS(1328),
    [anon_sym_DASH] = ACTIONS(1328),
    [anon_sym_PIPE] = ACTIONS(1328),
    [anon_sym_yield] = ACTIONS(1330),
    [anon_sym_move] = ACTIONS(1330),
    [sym_integer_literal] = ACTIONS(1328),
    [aux_sym_string_literal_token1] = ACTIONS(1328),
    [sym_char_literal] = ACTIONS(1328),
    [anon_sym_true] = ACTIONS(1330),
    [anon_sym_false] = ACTIONS(1330),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1330),
    [sym_super] = ACTIONS(1330),
    [sym_crate] = ACTIONS(1330),
    [sym_metavariable] = ACTIONS(1328),
    [sym_raw_string_literal] = ACTIONS(1328),
    [sym_float_literal] = ACTIONS(1328),
    [sym_block_comment] = ACTIONS(3),
  },
  [315] = {
    [ts_builtin_sym_end] = ACTIONS(1332),
    [sym_identifier] = ACTIONS(1334),
    [anon_sym_SEMI] = ACTIONS(1332),
    [anon_sym_macro_rules_BANG] = ACTIONS(1332),
    [anon_sym_LPAREN] = ACTIONS(1332),
    [anon_sym_LBRACE] = ACTIONS(1332),
    [anon_sym_RBRACE] = ACTIONS(1332),
    [anon_sym_LBRACK] = ACTIONS(1332),
    [anon_sym_STAR] = ACTIONS(1332),
    [anon_sym_u8] = ACTIONS(1334),
    [anon_sym_i8] = ACTIONS(1334),
    [anon_sym_u16] = ACTIONS(1334),
    [anon_sym_i16] = ACTIONS(1334),
    [anon_sym_u32] = ACTIONS(1334),
    [anon_sym_i32] = ACTIONS(1334),
    [anon_sym_u64] = ACTIONS(1334),
    [anon_sym_i64] = ACTIONS(1334),
    [anon_sym_u128] = ACTIONS(1334),
    [anon_sym_i128] = ACTIONS(1334),
    [anon_sym_isize] = ACTIONS(1334),
    [anon_sym_usize] = ACTIONS(1334),
    [anon_sym_f32] = ACTIONS(1334),
    [anon_sym_f64] = ACTIONS(1334),
    [anon_sym_bool] = ACTIONS(1334),
    [anon_sym_str] = ACTIONS(1334),
    [anon_sym_char] = ACTIONS(1334),
    [anon_sym_SQUOTE] = ACTIONS(1334),
    [anon_sym_async] = ACTIONS(1334),
    [anon_sym_break] = ACTIONS(1334),
    [anon_sym_const] = ACTIONS(1334),
    [anon_sym_continue] = ACTIONS(1334),
    [anon_sym_default] = ACTIONS(1334),
    [anon_sym_enum] = ACTIONS(1334),
    [anon_sym_fn] = ACTIONS(1334),
    [anon_sym_for] = ACTIONS(1334),
    [anon_sym_if] = ACTIONS(1334),
    [anon_sym_impl] = ACTIONS(1334),
    [anon_sym_let] = ACTIONS(1334),
    [anon_sym_loop] = ACTIONS(1334),
    [anon_sym_match] = ACTIONS(1334),
    [anon_sym_mod] = ACTIONS(1334),
    [anon_sym_pub] = ACTIONS(1334),
    [anon_sym_return] = ACTIONS(1334),
    [anon_sym_static] = ACTIONS(1334),
    [anon_sym_struct] = ACTIONS(1334),
    [anon_sym_trait] = ACTIONS(1334),
    [anon_sym_type] = ACTIONS(1334),
    [anon_sym_union] = ACTIONS(1334),
    [anon_sym_unsafe] = ACTIONS(1334),
    [anon_sym_use] = ACTIONS(1334),
    [anon_sym_while] = ACTIONS(1334),
    [anon_sym_POUND] = ACTIONS(1332),
    [anon_sym_BANG] = ACTIONS(1332),
    [anon_sym_extern] = ACTIONS(1334),
    [anon_sym_LT] = ACTIONS(1332),
    [anon_sym_COLON_COLON] = ACTIONS(1332),
    [anon_sym_AMP] = ACTIONS(1332),
    [anon_sym_DOT_DOT] = ACTIONS(1332),
    [anon_sym_DASH] = ACTIONS(1332),
    [anon_sym_PIPE] = ACTIONS(1332),
    [anon_sym_yield] = ACTIONS(1334),
    [anon_sym_move] = ACTIONS(1334),
    [sym_integer_literal] = ACTIONS(1332),
    [aux_sym_string_literal_token1] = ACTIONS(1332),
    [sym_char_literal] = ACTIONS(1332),
    [anon_sym_true] = ACTIONS(1334),
    [anon_sym_false] = ACTIONS(1334),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1334),
    [sym_super] = ACTIONS(1334),
    [sym_crate] = ACTIONS(1334),
    [sym_metavariable] = ACTIONS(1332),
    [sym_raw_string_literal] = ACTIONS(1332),
    [sym_float_literal] = ACTIONS(1332),
    [sym_block_comment] = ACTIONS(3),
  },
  [316] = {
    [ts_builtin_sym_end] = ACTIONS(1336),
    [sym_identifier] = ACTIONS(1338),
    [anon_sym_SEMI] = ACTIONS(1336),
    [anon_sym_macro_rules_BANG] = ACTIONS(1336),
    [anon_sym_LPAREN] = ACTIONS(1336),
    [anon_sym_LBRACE] = ACTIONS(1336),
    [anon_sym_RBRACE] = ACTIONS(1336),
    [anon_sym_LBRACK] = ACTIONS(1336),
    [anon_sym_STAR] = ACTIONS(1336),
    [anon_sym_u8] = ACTIONS(1338),
    [anon_sym_i8] = ACTIONS(1338),
    [anon_sym_u16] = ACTIONS(1338),
    [anon_sym_i16] = ACTIONS(1338),
    [anon_sym_u32] = ACTIONS(1338),
    [anon_sym_i32] = ACTIONS(1338),
    [anon_sym_u64] = ACTIONS(1338),
    [anon_sym_i64] = ACTIONS(1338),
    [anon_sym_u128] = ACTIONS(1338),
    [anon_sym_i128] = ACTIONS(1338),
    [anon_sym_isize] = ACTIONS(1338),
    [anon_sym_usize] = ACTIONS(1338),
    [anon_sym_f32] = ACTIONS(1338),
    [anon_sym_f64] = ACTIONS(1338),
    [anon_sym_bool] = ACTIONS(1338),
    [anon_sym_str] = ACTIONS(1338),
    [anon_sym_char] = ACTIONS(1338),
    [anon_sym_SQUOTE] = ACTIONS(1338),
    [anon_sym_async] = ACTIONS(1338),
    [anon_sym_break] = ACTIONS(1338),
    [anon_sym_const] = ACTIONS(1338),
    [anon_sym_continue] = ACTIONS(1338),
    [anon_sym_default] = ACTIONS(1338),
    [anon_sym_enum] = ACTIONS(1338),
    [anon_sym_fn] = ACTIONS(1338),
    [anon_sym_for] = ACTIONS(1338),
    [anon_sym_if] = ACTIONS(1338),
    [anon_sym_impl] = ACTIONS(1338),
    [anon_sym_let] = ACTIONS(1338),
    [anon_sym_loop] = ACTIONS(1338),
    [anon_sym_match] = ACTIONS(1338),
    [anon_sym_mod] = ACTIONS(1338),
    [anon_sym_pub] = ACTIONS(1338),
    [anon_sym_return] = ACTIONS(1338),
    [anon_sym_static] = ACTIONS(1338),
    [anon_sym_struct] = ACTIONS(1338),
    [anon_sym_trait] = ACTIONS(1338),
    [anon_sym_type] = ACTIONS(1338),
    [anon_sym_union] = ACTIONS(1338),
    [anon_sym_unsafe] = ACTIONS(1338),
    [anon_sym_use] = ACTIONS(1338),
    [anon_sym_while] = ACTIONS(1338),
    [anon_sym_POUND] = ACTIONS(1336),
    [anon_sym_BANG] = ACTIONS(1336),
    [anon_sym_extern] = ACTIONS(1338),
    [anon_sym_LT] = ACTIONS(1336),
    [anon_sym_COLON_COLON] = ACTIONS(1336),
    [anon_sym_AMP] = ACTIONS(1336),
    [anon_sym_DOT_DOT] = ACTIONS(1336),
    [anon_sym_DASH] = ACTIONS(1336),
    [anon_sym_PIPE] = ACTIONS(1336),
    [anon_sym_yield] = ACTIONS(1338),
    [anon_sym_move] = ACTIONS(1338),
    [sym_integer_literal] = ACTIONS(1336),
    [aux_sym_string_literal_token1] = ACTIONS(1336),
    [sym_char_literal] = ACTIONS(1336),
    [anon_sym_true] = ACTIONS(1338),
    [anon_sym_false] = ACTIONS(1338),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1338),
    [sym_super] = ACTIONS(1338),
    [sym_crate] = ACTIONS(1338),
    [sym_metavariable] = ACTIONS(1336),
    [sym_raw_string_literal] = ACTIONS(1336),
    [sym_float_literal] = ACTIONS(1336),
    [sym_block_comment] = ACTIONS(3),
  },
  [317] = {
    [ts_builtin_sym_end] = ACTIONS(1340),
    [sym_identifier] = ACTIONS(1342),
    [anon_sym_SEMI] = ACTIONS(1340),
    [anon_sym_macro_rules_BANG] = ACTIONS(1340),
    [anon_sym_LPAREN] = ACTIONS(1340),
    [anon_sym_LBRACE] = ACTIONS(1340),
    [anon_sym_RBRACE] = ACTIONS(1340),
    [anon_sym_LBRACK] = ACTIONS(1340),
    [anon_sym_STAR] = ACTIONS(1340),
    [anon_sym_u8] = ACTIONS(1342),
    [anon_sym_i8] = ACTIONS(1342),
    [anon_sym_u16] = ACTIONS(1342),
    [anon_sym_i16] = ACTIONS(1342),
    [anon_sym_u32] = ACTIONS(1342),
    [anon_sym_i32] = ACTIONS(1342),
    [anon_sym_u64] = ACTIONS(1342),
    [anon_sym_i64] = ACTIONS(1342),
    [anon_sym_u128] = ACTIONS(1342),
    [anon_sym_i128] = ACTIONS(1342),
    [anon_sym_isize] = ACTIONS(1342),
    [anon_sym_usize] = ACTIONS(1342),
    [anon_sym_f32] = ACTIONS(1342),
    [anon_sym_f64] = ACTIONS(1342),
    [anon_sym_bool] = ACTIONS(1342),
    [anon_sym_str] = ACTIONS(1342),
    [anon_sym_char] = ACTIONS(1342),
    [anon_sym_SQUOTE] = ACTIONS(1342),
    [anon_sym_async] = ACTIONS(1342),
    [anon_sym_break] = ACTIONS(1342),
    [anon_sym_const] = ACTIONS(1342),
    [anon_sym_continue] = ACTIONS(1342),
    [anon_sym_default] = ACTIONS(1342),
    [anon_sym_enum] = ACTIONS(1342),
    [anon_sym_fn] = ACTIONS(1342),
    [anon_sym_for] = ACTIONS(1342),
    [anon_sym_if] = ACTIONS(1342),
    [anon_sym_impl] = ACTIONS(1342),
    [anon_sym_let] = ACTIONS(1342),
    [anon_sym_loop] = ACTIONS(1342),
    [anon_sym_match] = ACTIONS(1342),
    [anon_sym_mod] = ACTIONS(1342),
    [anon_sym_pub] = ACTIONS(1342),
    [anon_sym_return] = ACTIONS(1342),
    [anon_sym_static] = ACTIONS(1342),
    [anon_sym_struct] = ACTIONS(1342),
    [anon_sym_trait] = ACTIONS(1342),
    [anon_sym_type] = ACTIONS(1342),
    [anon_sym_union] = ACTIONS(1342),
    [anon_sym_unsafe] = ACTIONS(1342),
    [anon_sym_use] = ACTIONS(1342),
    [anon_sym_while] = ACTIONS(1342),
    [anon_sym_POUND] = ACTIONS(1340),
    [anon_sym_BANG] = ACTIONS(1340),
    [anon_sym_extern] = ACTIONS(1342),
    [anon_sym_LT] = ACTIONS(1340),
    [anon_sym_COLON_COLON] = ACTIONS(1340),
    [anon_sym_AMP] = ACTIONS(1340),
    [anon_sym_DOT_DOT] = ACTIONS(1340),
    [anon_sym_DASH] = ACTIONS(1340),
    [anon_sym_PIPE] = ACTIONS(1340),
    [anon_sym_yield] = ACTIONS(1342),
    [anon_sym_move] = ACTIONS(1342),
    [sym_integer_literal] = ACTIONS(1340),
    [aux_sym_string_literal_token1] = ACTIONS(1340),
    [sym_char_literal] = ACTIONS(1340),
    [anon_sym_true] = ACTIONS(1342),
    [anon_sym_false] = ACTIONS(1342),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1342),
    [sym_super] = ACTIONS(1342),
    [sym_crate] = ACTIONS(1342),
    [sym_metavariable] = ACTIONS(1340),
    [sym_raw_string_literal] = ACTIONS(1340),
    [sym_float_literal] = ACTIONS(1340),
    [sym_block_comment] = ACTIONS(3),
  },
  [318] = {
    [ts_builtin_sym_end] = ACTIONS(1344),
    [sym_identifier] = ACTIONS(1346),
    [anon_sym_SEMI] = ACTIONS(1344),
    [anon_sym_macro_rules_BANG] = ACTIONS(1344),
    [anon_sym_LPAREN] = ACTIONS(1344),
    [anon_sym_LBRACE] = ACTIONS(1344),
    [anon_sym_RBRACE] = ACTIONS(1344),
    [anon_sym_LBRACK] = ACTIONS(1344),
    [anon_sym_STAR] = ACTIONS(1344),
    [anon_sym_u8] = ACTIONS(1346),
    [anon_sym_i8] = ACTIONS(1346),
    [anon_sym_u16] = ACTIONS(1346),
    [anon_sym_i16] = ACTIONS(1346),
    [anon_sym_u32] = ACTIONS(1346),
    [anon_sym_i32] = ACTIONS(1346),
    [anon_sym_u64] = ACTIONS(1346),
    [anon_sym_i64] = ACTIONS(1346),
    [anon_sym_u128] = ACTIONS(1346),
    [anon_sym_i128] = ACTIONS(1346),
    [anon_sym_isize] = ACTIONS(1346),
    [anon_sym_usize] = ACTIONS(1346),
    [anon_sym_f32] = ACTIONS(1346),
    [anon_sym_f64] = ACTIONS(1346),
    [anon_sym_bool] = ACTIONS(1346),
    [anon_sym_str] = ACTIONS(1346),
    [anon_sym_char] = ACTIONS(1346),
    [anon_sym_SQUOTE] = ACTIONS(1346),
    [anon_sym_async] = ACTIONS(1346),
    [anon_sym_break] = ACTIONS(1346),
    [anon_sym_const] = ACTIONS(1346),
    [anon_sym_continue] = ACTIONS(1346),
    [anon_sym_default] = ACTIONS(1346),
    [anon_sym_enum] = ACTIONS(1346),
    [anon_sym_fn] = ACTIONS(1346),
    [anon_sym_for] = ACTIONS(1346),
    [anon_sym_if] = ACTIONS(1346),
    [anon_sym_impl] = ACTIONS(1346),
    [anon_sym_let] = ACTIONS(1346),
    [anon_sym_loop] = ACTIONS(1346),
    [anon_sym_match] = ACTIONS(1346),
    [anon_sym_mod] = ACTIONS(1346),
    [anon_sym_pub] = ACTIONS(1346),
    [anon_sym_return] = ACTIONS(1346),
    [anon_sym_static] = ACTIONS(1346),
    [anon_sym_struct] = ACTIONS(1346),
    [anon_sym_trait] = ACTIONS(1346),
    [anon_sym_type] = ACTIONS(1346),
    [anon_sym_union] = ACTIONS(1346),
    [anon_sym_unsafe] = ACTIONS(1346),
    [anon_sym_use] = ACTIONS(1346),
    [anon_sym_while] = ACTIONS(1346),
    [anon_sym_POUND] = ACTIONS(1344),
    [anon_sym_BANG] = ACTIONS(1344),
    [anon_sym_extern] = ACTIONS(1346),
    [anon_sym_LT] = ACTIONS(1344),
    [anon_sym_COLON_COLON] = ACTIONS(1344),
    [anon_sym_AMP] = ACTIONS(1344),
    [anon_sym_DOT_DOT] = ACTIONS(1344),
    [anon_sym_DASH] = ACTIONS(1344),
    [anon_sym_PIPE] = ACTIONS(1344),
    [anon_sym_yield] = ACTIONS(1346),
    [anon_sym_move] = ACTIONS(1346),
    [sym_integer_literal] = ACTIONS(1344),
    [aux_sym_string_literal_token1] = ACTIONS(1344),
    [sym_char_literal] = ACTIONS(1344),
    [anon_sym_true] = ACTIONS(1346),
    [anon_sym_false] = ACTIONS(1346),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1346),
    [sym_super] = ACTIONS(1346),
    [sym_crate] = ACTIONS(1346),
    [sym_metavariable] = ACTIONS(1344),
    [sym_raw_string_literal] = ACTIONS(1344),
    [sym_float_literal] = ACTIONS(1344),
    [sym_block_comment] = ACTIONS(3),
  },
  [319] = {
    [ts_builtin_sym_end] = ACTIONS(1348),
    [sym_identifier] = ACTIONS(1350),
    [anon_sym_SEMI] = ACTIONS(1348),
    [anon_sym_macro_rules_BANG] = ACTIONS(1348),
    [anon_sym_LPAREN] = ACTIONS(1348),
    [anon_sym_LBRACE] = ACTIONS(1348),
    [anon_sym_RBRACE] = ACTIONS(1348),
    [anon_sym_LBRACK] = ACTIONS(1348),
    [anon_sym_STAR] = ACTIONS(1348),
    [anon_sym_u8] = ACTIONS(1350),
    [anon_sym_i8] = ACTIONS(1350),
    [anon_sym_u16] = ACTIONS(1350),
    [anon_sym_i16] = ACTIONS(1350),
    [anon_sym_u32] = ACTIONS(1350),
    [anon_sym_i32] = ACTIONS(1350),
    [anon_sym_u64] = ACTIONS(1350),
    [anon_sym_i64] = ACTIONS(1350),
    [anon_sym_u128] = ACTIONS(1350),
    [anon_sym_i128] = ACTIONS(1350),
    [anon_sym_isize] = ACTIONS(1350),
    [anon_sym_usize] = ACTIONS(1350),
    [anon_sym_f32] = ACTIONS(1350),
    [anon_sym_f64] = ACTIONS(1350),
    [anon_sym_bool] = ACTIONS(1350),
    [anon_sym_str] = ACTIONS(1350),
    [anon_sym_char] = ACTIONS(1350),
    [anon_sym_SQUOTE] = ACTIONS(1350),
    [anon_sym_async] = ACTIONS(1350),
    [anon_sym_break] = ACTIONS(1350),
    [anon_sym_const] = ACTIONS(1350),
    [anon_sym_continue] = ACTIONS(1350),
    [anon_sym_default] = ACTIONS(1350),
    [anon_sym_enum] = ACTIONS(1350),
    [anon_sym_fn] = ACTIONS(1350),
    [anon_sym_for] = ACTIONS(1350),
    [anon_sym_if] = ACTIONS(1350),
    [anon_sym_impl] = ACTIONS(1350),
    [anon_sym_let] = ACTIONS(1350),
    [anon_sym_loop] = ACTIONS(1350),
    [anon_sym_match] = ACTIONS(1350),
    [anon_sym_mod] = ACTIONS(1350),
    [anon_sym_pub] = ACTIONS(1350),
    [anon_sym_return] = ACTIONS(1350),
    [anon_sym_static] = ACTIONS(1350),
    [anon_sym_struct] = ACTIONS(1350),
    [anon_sym_trait] = ACTIONS(1350),
    [anon_sym_type] = ACTIONS(1350),
    [anon_sym_union] = ACTIONS(1350),
    [anon_sym_unsafe] = ACTIONS(1350),
    [anon_sym_use] = ACTIONS(1350),
    [anon_sym_while] = ACTIONS(1350),
    [anon_sym_POUND] = ACTIONS(1348),
    [anon_sym_BANG] = ACTIONS(1348),
    [anon_sym_extern] = ACTIONS(1350),
    [anon_sym_LT] = ACTIONS(1348),
    [anon_sym_COLON_COLON] = ACTIONS(1348),
    [anon_sym_AMP] = ACTIONS(1348),
    [anon_sym_DOT_DOT] = ACTIONS(1348),
    [anon_sym_DASH] = ACTIONS(1348),
    [anon_sym_PIPE] = ACTIONS(1348),
    [anon_sym_yield] = ACTIONS(1350),
    [anon_sym_move] = ACTIONS(1350),
    [sym_integer_literal] = ACTIONS(1348),
    [aux_sym_string_literal_token1] = ACTIONS(1348),
    [sym_char_literal] = ACTIONS(1348),
    [anon_sym_true] = ACTIONS(1350),
    [anon_sym_false] = ACTIONS(1350),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1350),
    [sym_super] = ACTIONS(1350),
    [sym_crate] = ACTIONS(1350),
    [sym_metavariable] = ACTIONS(1348),
    [sym_raw_string_literal] = ACTIONS(1348),
    [sym_float_literal] = ACTIONS(1348),
    [sym_block_comment] = ACTIONS(3),
  },
  [320] = {
    [ts_builtin_sym_end] = ACTIONS(1352),
    [sym_identifier] = ACTIONS(1354),
    [anon_sym_SEMI] = ACTIONS(1352),
    [anon_sym_macro_rules_BANG] = ACTIONS(1352),
    [anon_sym_LPAREN] = ACTIONS(1352),
    [anon_sym_LBRACE] = ACTIONS(1352),
    [anon_sym_RBRACE] = ACTIONS(1352),
    [anon_sym_LBRACK] = ACTIONS(1352),
    [anon_sym_STAR] = ACTIONS(1352),
    [anon_sym_u8] = ACTIONS(1354),
    [anon_sym_i8] = ACTIONS(1354),
    [anon_sym_u16] = ACTIONS(1354),
    [anon_sym_i16] = ACTIONS(1354),
    [anon_sym_u32] = ACTIONS(1354),
    [anon_sym_i32] = ACTIONS(1354),
    [anon_sym_u64] = ACTIONS(1354),
    [anon_sym_i64] = ACTIONS(1354),
    [anon_sym_u128] = ACTIONS(1354),
    [anon_sym_i128] = ACTIONS(1354),
    [anon_sym_isize] = ACTIONS(1354),
    [anon_sym_usize] = ACTIONS(1354),
    [anon_sym_f32] = ACTIONS(1354),
    [anon_sym_f64] = ACTIONS(1354),
    [anon_sym_bool] = ACTIONS(1354),
    [anon_sym_str] = ACTIONS(1354),
    [anon_sym_char] = ACTIONS(1354),
    [anon_sym_SQUOTE] = ACTIONS(1354),
    [anon_sym_async] = ACTIONS(1354),
    [anon_sym_break] = ACTIONS(1354),
    [anon_sym_const] = ACTIONS(1354),
    [anon_sym_continue] = ACTIONS(1354),
    [anon_sym_default] = ACTIONS(1354),
    [anon_sym_enum] = ACTIONS(1354),
    [anon_sym_fn] = ACTIONS(1354),
    [anon_sym_for] = ACTIONS(1354),
    [anon_sym_if] = ACTIONS(1354),
    [anon_sym_impl] = ACTIONS(1354),
    [anon_sym_let] = ACTIONS(1354),
    [anon_sym_loop] = ACTIONS(1354),
    [anon_sym_match] = ACTIONS(1354),
    [anon_sym_mod] = ACTIONS(1354),
    [anon_sym_pub] = ACTIONS(1354),
    [anon_sym_return] = ACTIONS(1354),
    [anon_sym_static] = ACTIONS(1354),
    [anon_sym_struct] = ACTIONS(1354),
    [anon_sym_trait] = ACTIONS(1354),
    [anon_sym_type] = ACTIONS(1354),
    [anon_sym_union] = ACTIONS(1354),
    [anon_sym_unsafe] = ACTIONS(1354),
    [anon_sym_use] = ACTIONS(1354),
    [anon_sym_while] = ACTIONS(1354),
    [anon_sym_POUND] = ACTIONS(1352),
    [anon_sym_BANG] = ACTIONS(1352),
    [anon_sym_extern] = ACTIONS(1354),
    [anon_sym_LT] = ACTIONS(1352),
    [anon_sym_COLON_COLON] = ACTIONS(1352),
    [anon_sym_AMP] = ACTIONS(1352),
    [anon_sym_DOT_DOT] = ACTIONS(1352),
    [anon_sym_DASH] = ACTIONS(1352),
    [anon_sym_PIPE] = ACTIONS(1352),
    [anon_sym_yield] = ACTIONS(1354),
    [anon_sym_move] = ACTIONS(1354),
    [sym_integer_literal] = ACTIONS(1352),
    [aux_sym_string_literal_token1] = ACTIONS(1352),
    [sym_char_literal] = ACTIONS(1352),
    [anon_sym_true] = ACTIONS(1354),
    [anon_sym_false] = ACTIONS(1354),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1354),
    [sym_super] = ACTIONS(1354),
    [sym_crate] = ACTIONS(1354),
    [sym_metavariable] = ACTIONS(1352),
    [sym_raw_string_literal] = ACTIONS(1352),
    [sym_float_literal] = ACTIONS(1352),
    [sym_block_comment] = ACTIONS(3),
  },
  [321] = {
    [ts_builtin_sym_end] = ACTIONS(1356),
    [sym_identifier] = ACTIONS(1358),
    [anon_sym_SEMI] = ACTIONS(1356),
    [anon_sym_macro_rules_BANG] = ACTIONS(1356),
    [anon_sym_LPAREN] = ACTIONS(1356),
    [anon_sym_LBRACE] = ACTIONS(1356),
    [anon_sym_RBRACE] = ACTIONS(1356),
    [anon_sym_LBRACK] = ACTIONS(1356),
    [anon_sym_STAR] = ACTIONS(1356),
    [anon_sym_u8] = ACTIONS(1358),
    [anon_sym_i8] = ACTIONS(1358),
    [anon_sym_u16] = ACTIONS(1358),
    [anon_sym_i16] = ACTIONS(1358),
    [anon_sym_u32] = ACTIONS(1358),
    [anon_sym_i32] = ACTIONS(1358),
    [anon_sym_u64] = ACTIONS(1358),
    [anon_sym_i64] = ACTIONS(1358),
    [anon_sym_u128] = ACTIONS(1358),
    [anon_sym_i128] = ACTIONS(1358),
    [anon_sym_isize] = ACTIONS(1358),
    [anon_sym_usize] = ACTIONS(1358),
    [anon_sym_f32] = ACTIONS(1358),
    [anon_sym_f64] = ACTIONS(1358),
    [anon_sym_bool] = ACTIONS(1358),
    [anon_sym_str] = ACTIONS(1358),
    [anon_sym_char] = ACTIONS(1358),
    [anon_sym_SQUOTE] = ACTIONS(1358),
    [anon_sym_async] = ACTIONS(1358),
    [anon_sym_break] = ACTIONS(1358),
    [anon_sym_const] = ACTIONS(1358),
    [anon_sym_continue] = ACTIONS(1358),
    [anon_sym_default] = ACTIONS(1358),
    [anon_sym_enum] = ACTIONS(1358),
    [anon_sym_fn] = ACTIONS(1358),
    [anon_sym_for] = ACTIONS(1358),
    [anon_sym_if] = ACTIONS(1358),
    [anon_sym_impl] = ACTIONS(1358),
    [anon_sym_let] = ACTIONS(1358),
    [anon_sym_loop] = ACTIONS(1358),
    [anon_sym_match] = ACTIONS(1358),
    [anon_sym_mod] = ACTIONS(1358),
    [anon_sym_pub] = ACTIONS(1358),
    [anon_sym_return] = ACTIONS(1358),
    [anon_sym_static] = ACTIONS(1358),
    [anon_sym_struct] = ACTIONS(1358),
    [anon_sym_trait] = ACTIONS(1358),
    [anon_sym_type] = ACTIONS(1358),
    [anon_sym_union] = ACTIONS(1358),
    [anon_sym_unsafe] = ACTIONS(1358),
    [anon_sym_use] = ACTIONS(1358),
    [anon_sym_while] = ACTIONS(1358),
    [anon_sym_POUND] = ACTIONS(1356),
    [anon_sym_BANG] = ACTIONS(1356),
    [anon_sym_extern] = ACTIONS(1358),
    [anon_sym_LT] = ACTIONS(1356),
    [anon_sym_COLON_COLON] = ACTIONS(1356),
    [anon_sym_AMP] = ACTIONS(1356),
    [anon_sym_DOT_DOT] = ACTIONS(1356),
    [anon_sym_DASH] = ACTIONS(1356),
    [anon_sym_PIPE] = ACTIONS(1356),
    [anon_sym_yield] = ACTIONS(1358),
    [anon_sym_move] = ACTIONS(1358),
    [sym_integer_literal] = ACTIONS(1356),
    [aux_sym_string_literal_token1] = ACTIONS(1356),
    [sym_char_literal] = ACTIONS(1356),
    [anon_sym_true] = ACTIONS(1358),
    [anon_sym_false] = ACTIONS(1358),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1358),
    [sym_super] = ACTIONS(1358),
    [sym_crate] = ACTIONS(1358),
    [sym_metavariable] = ACTIONS(1356),
    [sym_raw_string_literal] = ACTIONS(1356),
    [sym_float_literal] = ACTIONS(1356),
    [sym_block_comment] = ACTIONS(3),
  },
  [322] = {
    [ts_builtin_sym_end] = ACTIONS(1360),
    [sym_identifier] = ACTIONS(1362),
    [anon_sym_SEMI] = ACTIONS(1360),
    [anon_sym_macro_rules_BANG] = ACTIONS(1360),
    [anon_sym_LPAREN] = ACTIONS(1360),
    [anon_sym_LBRACE] = ACTIONS(1360),
    [anon_sym_RBRACE] = ACTIONS(1360),
    [anon_sym_LBRACK] = ACTIONS(1360),
    [anon_sym_STAR] = ACTIONS(1360),
    [anon_sym_u8] = ACTIONS(1362),
    [anon_sym_i8] = ACTIONS(1362),
    [anon_sym_u16] = ACTIONS(1362),
    [anon_sym_i16] = ACTIONS(1362),
    [anon_sym_u32] = ACTIONS(1362),
    [anon_sym_i32] = ACTIONS(1362),
    [anon_sym_u64] = ACTIONS(1362),
    [anon_sym_i64] = ACTIONS(1362),
    [anon_sym_u128] = ACTIONS(1362),
    [anon_sym_i128] = ACTIONS(1362),
    [anon_sym_isize] = ACTIONS(1362),
    [anon_sym_usize] = ACTIONS(1362),
    [anon_sym_f32] = ACTIONS(1362),
    [anon_sym_f64] = ACTIONS(1362),
    [anon_sym_bool] = ACTIONS(1362),
    [anon_sym_str] = ACTIONS(1362),
    [anon_sym_char] = ACTIONS(1362),
    [anon_sym_SQUOTE] = ACTIONS(1362),
    [anon_sym_async] = ACTIONS(1362),
    [anon_sym_break] = ACTIONS(1362),
    [anon_sym_const] = ACTIONS(1362),
    [anon_sym_continue] = ACTIONS(1362),
    [anon_sym_default] = ACTIONS(1362),
    [anon_sym_enum] = ACTIONS(1362),
    [anon_sym_fn] = ACTIONS(1362),
    [anon_sym_for] = ACTIONS(1362),
    [anon_sym_if] = ACTIONS(1362),
    [anon_sym_impl] = ACTIONS(1362),
    [anon_sym_let] = ACTIONS(1362),
    [anon_sym_loop] = ACTIONS(1362),
    [anon_sym_match] = ACTIONS(1362),
    [anon_sym_mod] = ACTIONS(1362),
    [anon_sym_pub] = ACTIONS(1362),
    [anon_sym_return] = ACTIONS(1362),
    [anon_sym_static] = ACTIONS(1362),
    [anon_sym_struct] = ACTIONS(1362),
    [anon_sym_trait] = ACTIONS(1362),
    [anon_sym_type] = ACTIONS(1362),
    [anon_sym_union] = ACTIONS(1362),
    [anon_sym_unsafe] = ACTIONS(1362),
    [anon_sym_use] = ACTIONS(1362),
    [anon_sym_while] = ACTIONS(1362),
    [anon_sym_POUND] = ACTIONS(1360),
    [anon_sym_BANG] = ACTIONS(1360),
    [anon_sym_extern] = ACTIONS(1362),
    [anon_sym_LT] = ACTIONS(1360),
    [anon_sym_COLON_COLON] = ACTIONS(1360),
    [anon_sym_AMP] = ACTIONS(1360),
    [anon_sym_DOT_DOT] = ACTIONS(1360),
    [anon_sym_DASH] = ACTIONS(1360),
    [anon_sym_PIPE] = ACTIONS(1360),
    [anon_sym_yield] = ACTIONS(1362),
    [anon_sym_move] = ACTIONS(1362),
    [sym_integer_literal] = ACTIONS(1360),
    [aux_sym_string_literal_token1] = ACTIONS(1360),
    [sym_char_literal] = ACTIONS(1360),
    [anon_sym_true] = ACTIONS(1362),
    [anon_sym_false] = ACTIONS(1362),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1362),
    [sym_super] = ACTIONS(1362),
    [sym_crate] = ACTIONS(1362),
    [sym_metavariable] = ACTIONS(1360),
    [sym_raw_string_literal] = ACTIONS(1360),
    [sym_float_literal] = ACTIONS(1360),
    [sym_block_comment] = ACTIONS(3),
  },
  [323] = {
    [ts_builtin_sym_end] = ACTIONS(1364),
    [sym_identifier] = ACTIONS(1366),
    [anon_sym_SEMI] = ACTIONS(1364),
    [anon_sym_macro_rules_BANG] = ACTIONS(1364),
    [anon_sym_LPAREN] = ACTIONS(1364),
    [anon_sym_LBRACE] = ACTIONS(1364),
    [anon_sym_RBRACE] = ACTIONS(1364),
    [anon_sym_LBRACK] = ACTIONS(1364),
    [anon_sym_STAR] = ACTIONS(1364),
    [anon_sym_u8] = ACTIONS(1366),
    [anon_sym_i8] = ACTIONS(1366),
    [anon_sym_u16] = ACTIONS(1366),
    [anon_sym_i16] = ACTIONS(1366),
    [anon_sym_u32] = ACTIONS(1366),
    [anon_sym_i32] = ACTIONS(1366),
    [anon_sym_u64] = ACTIONS(1366),
    [anon_sym_i64] = ACTIONS(1366),
    [anon_sym_u128] = ACTIONS(1366),
    [anon_sym_i128] = ACTIONS(1366),
    [anon_sym_isize] = ACTIONS(1366),
    [anon_sym_usize] = ACTIONS(1366),
    [anon_sym_f32] = ACTIONS(1366),
    [anon_sym_f64] = ACTIONS(1366),
    [anon_sym_bool] = ACTIONS(1366),
    [anon_sym_str] = ACTIONS(1366),
    [anon_sym_char] = ACTIONS(1366),
    [anon_sym_SQUOTE] = ACTIONS(1366),
    [anon_sym_async] = ACTIONS(1366),
    [anon_sym_break] = ACTIONS(1366),
    [anon_sym_const] = ACTIONS(1366),
    [anon_sym_continue] = ACTIONS(1366),
    [anon_sym_default] = ACTIONS(1366),
    [anon_sym_enum] = ACTIONS(1366),
    [anon_sym_fn] = ACTIONS(1366),
    [anon_sym_for] = ACTIONS(1366),
    [anon_sym_if] = ACTIONS(1366),
    [anon_sym_impl] = ACTIONS(1366),
    [anon_sym_let] = ACTIONS(1366),
    [anon_sym_loop] = ACTIONS(1366),
    [anon_sym_match] = ACTIONS(1366),
    [anon_sym_mod] = ACTIONS(1366),
    [anon_sym_pub] = ACTIONS(1366),
    [anon_sym_return] = ACTIONS(1366),
    [anon_sym_static] = ACTIONS(1366),
    [anon_sym_struct] = ACTIONS(1366),
    [anon_sym_trait] = ACTIONS(1366),
    [anon_sym_type] = ACTIONS(1366),
    [anon_sym_union] = ACTIONS(1366),
    [anon_sym_unsafe] = ACTIONS(1366),
    [anon_sym_use] = ACTIONS(1366),
    [anon_sym_while] = ACTIONS(1366),
    [anon_sym_POUND] = ACTIONS(1364),
    [anon_sym_BANG] = ACTIONS(1364),
    [anon_sym_extern] = ACTIONS(1366),
    [anon_sym_LT] = ACTIONS(1364),
    [anon_sym_COLON_COLON] = ACTIONS(1364),
    [anon_sym_AMP] = ACTIONS(1364),
    [anon_sym_DOT_DOT] = ACTIONS(1364),
    [anon_sym_DASH] = ACTIONS(1364),
    [anon_sym_PIPE] = ACTIONS(1364),
    [anon_sym_yield] = ACTIONS(1366),
    [anon_sym_move] = ACTIONS(1366),
    [sym_integer_literal] = ACTIONS(1364),
    [aux_sym_string_literal_token1] = ACTIONS(1364),
    [sym_char_literal] = ACTIONS(1364),
    [anon_sym_true] = ACTIONS(1366),
    [anon_sym_false] = ACTIONS(1366),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1366),
    [sym_super] = ACTIONS(1366),
    [sym_crate] = ACTIONS(1366),
    [sym_metavariable] = ACTIONS(1364),
    [sym_raw_string_literal] = ACTIONS(1364),
    [sym_float_literal] = ACTIONS(1364),
    [sym_block_comment] = ACTIONS(3),
  },
  [324] = {
    [ts_builtin_sym_end] = ACTIONS(1368),
    [sym_identifier] = ACTIONS(1370),
    [anon_sym_SEMI] = ACTIONS(1368),
    [anon_sym_macro_rules_BANG] = ACTIONS(1368),
    [anon_sym_LPAREN] = ACTIONS(1368),
    [anon_sym_LBRACE] = ACTIONS(1368),
    [anon_sym_RBRACE] = ACTIONS(1368),
    [anon_sym_LBRACK] = ACTIONS(1368),
    [anon_sym_STAR] = ACTIONS(1368),
    [anon_sym_u8] = ACTIONS(1370),
    [anon_sym_i8] = ACTIONS(1370),
    [anon_sym_u16] = ACTIONS(1370),
    [anon_sym_i16] = ACTIONS(1370),
    [anon_sym_u32] = ACTIONS(1370),
    [anon_sym_i32] = ACTIONS(1370),
    [anon_sym_u64] = ACTIONS(1370),
    [anon_sym_i64] = ACTIONS(1370),
    [anon_sym_u128] = ACTIONS(1370),
    [anon_sym_i128] = ACTIONS(1370),
    [anon_sym_isize] = ACTIONS(1370),
    [anon_sym_usize] = ACTIONS(1370),
    [anon_sym_f32] = ACTIONS(1370),
    [anon_sym_f64] = ACTIONS(1370),
    [anon_sym_bool] = ACTIONS(1370),
    [anon_sym_str] = ACTIONS(1370),
    [anon_sym_char] = ACTIONS(1370),
    [anon_sym_SQUOTE] = ACTIONS(1370),
    [anon_sym_async] = ACTIONS(1370),
    [anon_sym_break] = ACTIONS(1370),
    [anon_sym_const] = ACTIONS(1370),
    [anon_sym_continue] = ACTIONS(1370),
    [anon_sym_default] = ACTIONS(1370),
    [anon_sym_enum] = ACTIONS(1370),
    [anon_sym_fn] = ACTIONS(1370),
    [anon_sym_for] = ACTIONS(1370),
    [anon_sym_if] = ACTIONS(1370),
    [anon_sym_impl] = ACTIONS(1370),
    [anon_sym_let] = ACTIONS(1370),
    [anon_sym_loop] = ACTIONS(1370),
    [anon_sym_match] = ACTIONS(1370),
    [anon_sym_mod] = ACTIONS(1370),
    [anon_sym_pub] = ACTIONS(1370),
    [anon_sym_return] = ACTIONS(1370),
    [anon_sym_static] = ACTIONS(1370),
    [anon_sym_struct] = ACTIONS(1370),
    [anon_sym_trait] = ACTIONS(1370),
    [anon_sym_type] = ACTIONS(1370),
    [anon_sym_union] = ACTIONS(1370),
    [anon_sym_unsafe] = ACTIONS(1370),
    [anon_sym_use] = ACTIONS(1370),
    [anon_sym_while] = ACTIONS(1370),
    [anon_sym_POUND] = ACTIONS(1368),
    [anon_sym_BANG] = ACTIONS(1368),
    [anon_sym_extern] = ACTIONS(1370),
    [anon_sym_LT] = ACTIONS(1368),
    [anon_sym_COLON_COLON] = ACTIONS(1368),
    [anon_sym_AMP] = ACTIONS(1368),
    [anon_sym_DOT_DOT] = ACTIONS(1368),
    [anon_sym_DASH] = ACTIONS(1368),
    [anon_sym_PIPE] = ACTIONS(1368),
    [anon_sym_yield] = ACTIONS(1370),
    [anon_sym_move] = ACTIONS(1370),
    [sym_integer_literal] = ACTIONS(1368),
    [aux_sym_string_literal_token1] = ACTIONS(1368),
    [sym_char_literal] = ACTIONS(1368),
    [anon_sym_true] = ACTIONS(1370),
    [anon_sym_false] = ACTIONS(1370),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1370),
    [sym_super] = ACTIONS(1370),
    [sym_crate] = ACTIONS(1370),
    [sym_metavariable] = ACTIONS(1368),
    [sym_raw_string_literal] = ACTIONS(1368),
    [sym_float_literal] = ACTIONS(1368),
    [sym_block_comment] = ACTIONS(3),
  },
  [325] = {
    [ts_builtin_sym_end] = ACTIONS(1372),
    [sym_identifier] = ACTIONS(1374),
    [anon_sym_SEMI] = ACTIONS(1372),
    [anon_sym_macro_rules_BANG] = ACTIONS(1372),
    [anon_sym_LPAREN] = ACTIONS(1372),
    [anon_sym_LBRACE] = ACTIONS(1372),
    [anon_sym_RBRACE] = ACTIONS(1372),
    [anon_sym_LBRACK] = ACTIONS(1372),
    [anon_sym_STAR] = ACTIONS(1372),
    [anon_sym_u8] = ACTIONS(1374),
    [anon_sym_i8] = ACTIONS(1374),
    [anon_sym_u16] = ACTIONS(1374),
    [anon_sym_i16] = ACTIONS(1374),
    [anon_sym_u32] = ACTIONS(1374),
    [anon_sym_i32] = ACTIONS(1374),
    [anon_sym_u64] = ACTIONS(1374),
    [anon_sym_i64] = ACTIONS(1374),
    [anon_sym_u128] = ACTIONS(1374),
    [anon_sym_i128] = ACTIONS(1374),
    [anon_sym_isize] = ACTIONS(1374),
    [anon_sym_usize] = ACTIONS(1374),
    [anon_sym_f32] = ACTIONS(1374),
    [anon_sym_f64] = ACTIONS(1374),
    [anon_sym_bool] = ACTIONS(1374),
    [anon_sym_str] = ACTIONS(1374),
    [anon_sym_char] = ACTIONS(1374),
    [anon_sym_SQUOTE] = ACTIONS(1374),
    [anon_sym_async] = ACTIONS(1374),
    [anon_sym_break] = ACTIONS(1374),
    [anon_sym_const] = ACTIONS(1374),
    [anon_sym_continue] = ACTIONS(1374),
    [anon_sym_default] = ACTIONS(1374),
    [anon_sym_enum] = ACTIONS(1374),
    [anon_sym_fn] = ACTIONS(1374),
    [anon_sym_for] = ACTIONS(1374),
    [anon_sym_if] = ACTIONS(1374),
    [anon_sym_impl] = ACTIONS(1374),
    [anon_sym_let] = ACTIONS(1374),
    [anon_sym_loop] = ACTIONS(1374),
    [anon_sym_match] = ACTIONS(1374),
    [anon_sym_mod] = ACTIONS(1374),
    [anon_sym_pub] = ACTIONS(1374),
    [anon_sym_return] = ACTIONS(1374),
    [anon_sym_static] = ACTIONS(1374),
    [anon_sym_struct] = ACTIONS(1374),
    [anon_sym_trait] = ACTIONS(1374),
    [anon_sym_type] = ACTIONS(1374),
    [anon_sym_union] = ACTIONS(1374),
    [anon_sym_unsafe] = ACTIONS(1374),
    [anon_sym_use] = ACTIONS(1374),
    [anon_sym_while] = ACTIONS(1374),
    [anon_sym_POUND] = ACTIONS(1372),
    [anon_sym_BANG] = ACTIONS(1372),
    [anon_sym_extern] = ACTIONS(1374),
    [anon_sym_LT] = ACTIONS(1372),
    [anon_sym_COLON_COLON] = ACTIONS(1372),
    [anon_sym_AMP] = ACTIONS(1372),
    [anon_sym_DOT_DOT] = ACTIONS(1372),
    [anon_sym_DASH] = ACTIONS(1372),
    [anon_sym_PIPE] = ACTIONS(1372),
    [anon_sym_yield] = ACTIONS(1374),
    [anon_sym_move] = ACTIONS(1374),
    [sym_integer_literal] = ACTIONS(1372),
    [aux_sym_string_literal_token1] = ACTIONS(1372),
    [sym_char_literal] = ACTIONS(1372),
    [anon_sym_true] = ACTIONS(1374),
    [anon_sym_false] = ACTIONS(1374),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1374),
    [sym_super] = ACTIONS(1374),
    [sym_crate] = ACTIONS(1374),
    [sym_metavariable] = ACTIONS(1372),
    [sym_raw_string_literal] = ACTIONS(1372),
    [sym_float_literal] = ACTIONS(1372),
    [sym_block_comment] = ACTIONS(3),
  },
  [326] = {
    [ts_builtin_sym_end] = ACTIONS(1376),
    [sym_identifier] = ACTIONS(1378),
    [anon_sym_SEMI] = ACTIONS(1376),
    [anon_sym_macro_rules_BANG] = ACTIONS(1376),
    [anon_sym_LPAREN] = ACTIONS(1376),
    [anon_sym_LBRACE] = ACTIONS(1376),
    [anon_sym_RBRACE] = ACTIONS(1376),
    [anon_sym_LBRACK] = ACTIONS(1376),
    [anon_sym_STAR] = ACTIONS(1376),
    [anon_sym_u8] = ACTIONS(1378),
    [anon_sym_i8] = ACTIONS(1378),
    [anon_sym_u16] = ACTIONS(1378),
    [anon_sym_i16] = ACTIONS(1378),
    [anon_sym_u32] = ACTIONS(1378),
    [anon_sym_i32] = ACTIONS(1378),
    [anon_sym_u64] = ACTIONS(1378),
    [anon_sym_i64] = ACTIONS(1378),
    [anon_sym_u128] = ACTIONS(1378),
    [anon_sym_i128] = ACTIONS(1378),
    [anon_sym_isize] = ACTIONS(1378),
    [anon_sym_usize] = ACTIONS(1378),
    [anon_sym_f32] = ACTIONS(1378),
    [anon_sym_f64] = ACTIONS(1378),
    [anon_sym_bool] = ACTIONS(1378),
    [anon_sym_str] = ACTIONS(1378),
    [anon_sym_char] = ACTIONS(1378),
    [anon_sym_SQUOTE] = ACTIONS(1378),
    [anon_sym_async] = ACTIONS(1378),
    [anon_sym_break] = ACTIONS(1378),
    [anon_sym_const] = ACTIONS(1378),
    [anon_sym_continue] = ACTIONS(1378),
    [anon_sym_default] = ACTIONS(1378),
    [anon_sym_enum] = ACTIONS(1378),
    [anon_sym_fn] = ACTIONS(1378),
    [anon_sym_for] = ACTIONS(1378),
    [anon_sym_if] = ACTIONS(1378),
    [anon_sym_impl] = ACTIONS(1378),
    [anon_sym_let] = ACTIONS(1378),
    [anon_sym_loop] = ACTIONS(1378),
    [anon_sym_match] = ACTIONS(1378),
    [anon_sym_mod] = ACTIONS(1378),
    [anon_sym_pub] = ACTIONS(1378),
    [anon_sym_return] = ACTIONS(1378),
    [anon_sym_static] = ACTIONS(1378),
    [anon_sym_struct] = ACTIONS(1378),
    [anon_sym_trait] = ACTIONS(1378),
    [anon_sym_type] = ACTIONS(1378),
    [anon_sym_union] = ACTIONS(1378),
    [anon_sym_unsafe] = ACTIONS(1378),
    [anon_sym_use] = ACTIONS(1378),
    [anon_sym_while] = ACTIONS(1378),
    [anon_sym_POUND] = ACTIONS(1376),
    [anon_sym_BANG] = ACTIONS(1376),
    [anon_sym_extern] = ACTIONS(1378),
    [anon_sym_LT] = ACTIONS(1376),
    [anon_sym_COLON_COLON] = ACTIONS(1376),
    [anon_sym_AMP] = ACTIONS(1376),
    [anon_sym_DOT_DOT] = ACTIONS(1376),
    [anon_sym_DASH] = ACTIONS(1376),
    [anon_sym_PIPE] = ACTIONS(1376),
    [anon_sym_yield] = ACTIONS(1378),
    [anon_sym_move] = ACTIONS(1378),
    [sym_integer_literal] = ACTIONS(1376),
    [aux_sym_string_literal_token1] = ACTIONS(1376),
    [sym_char_literal] = ACTIONS(1376),
    [anon_sym_true] = ACTIONS(1378),
    [anon_sym_false] = ACTIONS(1378),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1378),
    [sym_super] = ACTIONS(1378),
    [sym_crate] = ACTIONS(1378),
    [sym_metavariable] = ACTIONS(1376),
    [sym_raw_string_literal] = ACTIONS(1376),
    [sym_float_literal] = ACTIONS(1376),
    [sym_block_comment] = ACTIONS(3),
  },
  [327] = {
    [ts_builtin_sym_end] = ACTIONS(1380),
    [sym_identifier] = ACTIONS(1382),
    [anon_sym_SEMI] = ACTIONS(1380),
    [anon_sym_macro_rules_BANG] = ACTIONS(1380),
    [anon_sym_LPAREN] = ACTIONS(1380),
    [anon_sym_LBRACE] = ACTIONS(1380),
    [anon_sym_RBRACE] = ACTIONS(1380),
    [anon_sym_LBRACK] = ACTIONS(1380),
    [anon_sym_STAR] = ACTIONS(1380),
    [anon_sym_u8] = ACTIONS(1382),
    [anon_sym_i8] = ACTIONS(1382),
    [anon_sym_u16] = ACTIONS(1382),
    [anon_sym_i16] = ACTIONS(1382),
    [anon_sym_u32] = ACTIONS(1382),
    [anon_sym_i32] = ACTIONS(1382),
    [anon_sym_u64] = ACTIONS(1382),
    [anon_sym_i64] = ACTIONS(1382),
    [anon_sym_u128] = ACTIONS(1382),
    [anon_sym_i128] = ACTIONS(1382),
    [anon_sym_isize] = ACTIONS(1382),
    [anon_sym_usize] = ACTIONS(1382),
    [anon_sym_f32] = ACTIONS(1382),
    [anon_sym_f64] = ACTIONS(1382),
    [anon_sym_bool] = ACTIONS(1382),
    [anon_sym_str] = ACTIONS(1382),
    [anon_sym_char] = ACTIONS(1382),
    [anon_sym_SQUOTE] = ACTIONS(1382),
    [anon_sym_async] = ACTIONS(1382),
    [anon_sym_break] = ACTIONS(1382),
    [anon_sym_const] = ACTIONS(1382),
    [anon_sym_continue] = ACTIONS(1382),
    [anon_sym_default] = ACTIONS(1382),
    [anon_sym_enum] = ACTIONS(1382),
    [anon_sym_fn] = ACTIONS(1382),
    [anon_sym_for] = ACTIONS(1382),
    [anon_sym_if] = ACTIONS(1382),
    [anon_sym_impl] = ACTIONS(1382),
    [anon_sym_let] = ACTIONS(1382),
    [anon_sym_loop] = ACTIONS(1382),
    [anon_sym_match] = ACTIONS(1382),
    [anon_sym_mod] = ACTIONS(1382),
    [anon_sym_pub] = ACTIONS(1382),
    [anon_sym_return] = ACTIONS(1382),
    [anon_sym_static] = ACTIONS(1382),
    [anon_sym_struct] = ACTIONS(1382),
    [anon_sym_trait] = ACTIONS(1382),
    [anon_sym_type] = ACTIONS(1382),
    [anon_sym_union] = ACTIONS(1382),
    [anon_sym_unsafe] = ACTIONS(1382),
    [anon_sym_use] = ACTIONS(1382),
    [anon_sym_while] = ACTIONS(1382),
    [anon_sym_POUND] = ACTIONS(1380),
    [anon_sym_BANG] = ACTIONS(1380),
    [anon_sym_extern] = ACTIONS(1382),
    [anon_sym_LT] = ACTIONS(1380),
    [anon_sym_COLON_COLON] = ACTIONS(1380),
    [anon_sym_AMP] = ACTIONS(1380),
    [anon_sym_DOT_DOT] = ACTIONS(1380),
    [anon_sym_DASH] = ACTIONS(1380),
    [anon_sym_PIPE] = ACTIONS(1380),
    [anon_sym_yield] = ACTIONS(1382),
    [anon_sym_move] = ACTIONS(1382),
    [sym_integer_literal] = ACTIONS(1380),
    [aux_sym_string_literal_token1] = ACTIONS(1380),
    [sym_char_literal] = ACTIONS(1380),
    [anon_sym_true] = ACTIONS(1382),
    [anon_sym_false] = ACTIONS(1382),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1382),
    [sym_super] = ACTIONS(1382),
    [sym_crate] = ACTIONS(1382),
    [sym_metavariable] = ACTIONS(1380),
    [sym_raw_string_literal] = ACTIONS(1380),
    [sym_float_literal] = ACTIONS(1380),
    [sym_block_comment] = ACTIONS(3),
  },
  [328] = {
    [ts_builtin_sym_end] = ACTIONS(1384),
    [sym_identifier] = ACTIONS(1386),
    [anon_sym_SEMI] = ACTIONS(1384),
    [anon_sym_macro_rules_BANG] = ACTIONS(1384),
    [anon_sym_LPAREN] = ACTIONS(1384),
    [anon_sym_LBRACE] = ACTIONS(1384),
    [anon_sym_RBRACE] = ACTIONS(1384),
    [anon_sym_LBRACK] = ACTIONS(1384),
    [anon_sym_STAR] = ACTIONS(1384),
    [anon_sym_u8] = ACTIONS(1386),
    [anon_sym_i8] = ACTIONS(1386),
    [anon_sym_u16] = ACTIONS(1386),
    [anon_sym_i16] = ACTIONS(1386),
    [anon_sym_u32] = ACTIONS(1386),
    [anon_sym_i32] = ACTIONS(1386),
    [anon_sym_u64] = ACTIONS(1386),
    [anon_sym_i64] = ACTIONS(1386),
    [anon_sym_u128] = ACTIONS(1386),
    [anon_sym_i128] = ACTIONS(1386),
    [anon_sym_isize] = ACTIONS(1386),
    [anon_sym_usize] = ACTIONS(1386),
    [anon_sym_f32] = ACTIONS(1386),
    [anon_sym_f64] = ACTIONS(1386),
    [anon_sym_bool] = ACTIONS(1386),
    [anon_sym_str] = ACTIONS(1386),
    [anon_sym_char] = ACTIONS(1386),
    [anon_sym_SQUOTE] = ACTIONS(1386),
    [anon_sym_async] = ACTIONS(1386),
    [anon_sym_break] = ACTIONS(1386),
    [anon_sym_const] = ACTIONS(1386),
    [anon_sym_continue] = ACTIONS(1386),
    [anon_sym_default] = ACTIONS(1386),
    [anon_sym_enum] = ACTIONS(1386),
    [anon_sym_fn] = ACTIONS(1386),
    [anon_sym_for] = ACTIONS(1386),
    [anon_sym_if] = ACTIONS(1386),
    [anon_sym_impl] = ACTIONS(1386),
    [anon_sym_let] = ACTIONS(1386),
    [anon_sym_loop] = ACTIONS(1386),
    [anon_sym_match] = ACTIONS(1386),
    [anon_sym_mod] = ACTIONS(1386),
    [anon_sym_pub] = ACTIONS(1386),
    [anon_sym_return] = ACTIONS(1386),
    [anon_sym_static] = ACTIONS(1386),
    [anon_sym_struct] = ACTIONS(1386),
    [anon_sym_trait] = ACTIONS(1386),
    [anon_sym_type] = ACTIONS(1386),
    [anon_sym_union] = ACTIONS(1386),
    [anon_sym_unsafe] = ACTIONS(1386),
    [anon_sym_use] = ACTIONS(1386),
    [anon_sym_while] = ACTIONS(1386),
    [anon_sym_POUND] = ACTIONS(1384),
    [anon_sym_BANG] = ACTIONS(1384),
    [anon_sym_extern] = ACTIONS(1386),
    [anon_sym_LT] = ACTIONS(1384),
    [anon_sym_COLON_COLON] = ACTIONS(1384),
    [anon_sym_AMP] = ACTIONS(1384),
    [anon_sym_DOT_DOT] = ACTIONS(1384),
    [anon_sym_DASH] = ACTIONS(1384),
    [anon_sym_PIPE] = ACTIONS(1384),
    [anon_sym_yield] = ACTIONS(1386),
    [anon_sym_move] = ACTIONS(1386),
    [sym_integer_literal] = ACTIONS(1384),
    [aux_sym_string_literal_token1] = ACTIONS(1384),
    [sym_char_literal] = ACTIONS(1384),
    [anon_sym_true] = ACTIONS(1386),
    [anon_sym_false] = ACTIONS(1386),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1386),
    [sym_super] = ACTIONS(1386),
    [sym_crate] = ACTIONS(1386),
    [sym_metavariable] = ACTIONS(1384),
    [sym_raw_string_literal] = ACTIONS(1384),
    [sym_float_literal] = ACTIONS(1384),
    [sym_block_comment] = ACTIONS(3),
  },
  [329] = {
    [ts_builtin_sym_end] = ACTIONS(1388),
    [sym_identifier] = ACTIONS(1390),
    [anon_sym_SEMI] = ACTIONS(1388),
    [anon_sym_macro_rules_BANG] = ACTIONS(1388),
    [anon_sym_LPAREN] = ACTIONS(1388),
    [anon_sym_LBRACE] = ACTIONS(1388),
    [anon_sym_RBRACE] = ACTIONS(1388),
    [anon_sym_LBRACK] = ACTIONS(1388),
    [anon_sym_STAR] = ACTIONS(1388),
    [anon_sym_u8] = ACTIONS(1390),
    [anon_sym_i8] = ACTIONS(1390),
    [anon_sym_u16] = ACTIONS(1390),
    [anon_sym_i16] = ACTIONS(1390),
    [anon_sym_u32] = ACTIONS(1390),
    [anon_sym_i32] = ACTIONS(1390),
    [anon_sym_u64] = ACTIONS(1390),
    [anon_sym_i64] = ACTIONS(1390),
    [anon_sym_u128] = ACTIONS(1390),
    [anon_sym_i128] = ACTIONS(1390),
    [anon_sym_isize] = ACTIONS(1390),
    [anon_sym_usize] = ACTIONS(1390),
    [anon_sym_f32] = ACTIONS(1390),
    [anon_sym_f64] = ACTIONS(1390),
    [anon_sym_bool] = ACTIONS(1390),
    [anon_sym_str] = ACTIONS(1390),
    [anon_sym_char] = ACTIONS(1390),
    [anon_sym_SQUOTE] = ACTIONS(1390),
    [anon_sym_async] = ACTIONS(1390),
    [anon_sym_break] = ACTIONS(1390),
    [anon_sym_const] = ACTIONS(1390),
    [anon_sym_continue] = ACTIONS(1390),
    [anon_sym_default] = ACTIONS(1390),
    [anon_sym_enum] = ACTIONS(1390),
    [anon_sym_fn] = ACTIONS(1390),
    [anon_sym_for] = ACTIONS(1390),
    [anon_sym_if] = ACTIONS(1390),
    [anon_sym_impl] = ACTIONS(1390),
    [anon_sym_let] = ACTIONS(1390),
    [anon_sym_loop] = ACTIONS(1390),
    [anon_sym_match] = ACTIONS(1390),
    [anon_sym_mod] = ACTIONS(1390),
    [anon_sym_pub] = ACTIONS(1390),
    [anon_sym_return] = ACTIONS(1390),
    [anon_sym_static] = ACTIONS(1390),
    [anon_sym_struct] = ACTIONS(1390),
    [anon_sym_trait] = ACTIONS(1390),
    [anon_sym_type] = ACTIONS(1390),
    [anon_sym_union] = ACTIONS(1390),
    [anon_sym_unsafe] = ACTIONS(1390),
    [anon_sym_use] = ACTIONS(1390),
    [anon_sym_while] = ACTIONS(1390),
    [anon_sym_POUND] = ACTIONS(1388),
    [anon_sym_BANG] = ACTIONS(1388),
    [anon_sym_extern] = ACTIONS(1390),
    [anon_sym_LT] = ACTIONS(1388),
    [anon_sym_COLON_COLON] = ACTIONS(1388),
    [anon_sym_AMP] = ACTIONS(1388),
    [anon_sym_DOT_DOT] = ACTIONS(1388),
    [anon_sym_DASH] = ACTIONS(1388),
    [anon_sym_PIPE] = ACTIONS(1388),
    [anon_sym_yield] = ACTIONS(1390),
    [anon_sym_move] = ACTIONS(1390),
    [sym_integer_literal] = ACTIONS(1388),
    [aux_sym_string_literal_token1] = ACTIONS(1388),
    [sym_char_literal] = ACTIONS(1388),
    [anon_sym_true] = ACTIONS(1390),
    [anon_sym_false] = ACTIONS(1390),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1390),
    [sym_super] = ACTIONS(1390),
    [sym_crate] = ACTIONS(1390),
    [sym_metavariable] = ACTIONS(1388),
    [sym_raw_string_literal] = ACTIONS(1388),
    [sym_float_literal] = ACTIONS(1388),
    [sym_block_comment] = ACTIONS(3),
  },
  [330] = {
    [ts_builtin_sym_end] = ACTIONS(406),
    [sym_identifier] = ACTIONS(408),
    [anon_sym_SEMI] = ACTIONS(406),
    [anon_sym_macro_rules_BANG] = ACTIONS(406),
    [anon_sym_LPAREN] = ACTIONS(406),
    [anon_sym_LBRACE] = ACTIONS(406),
    [anon_sym_RBRACE] = ACTIONS(406),
    [anon_sym_LBRACK] = ACTIONS(406),
    [anon_sym_STAR] = ACTIONS(406),
    [anon_sym_u8] = ACTIONS(408),
    [anon_sym_i8] = ACTIONS(408),
    [anon_sym_u16] = ACTIONS(408),
    [anon_sym_i16] = ACTIONS(408),
    [anon_sym_u32] = ACTIONS(408),
    [anon_sym_i32] = ACTIONS(408),
    [anon_sym_u64] = ACTIONS(408),
    [anon_sym_i64] = ACTIONS(408),
    [anon_sym_u128] = ACTIONS(408),
    [anon_sym_i128] = ACTIONS(408),
    [anon_sym_isize] = ACTIONS(408),
    [anon_sym_usize] = ACTIONS(408),
    [anon_sym_f32] = ACTIONS(408),
    [anon_sym_f64] = ACTIONS(408),
    [anon_sym_bool] = ACTIONS(408),
    [anon_sym_str] = ACTIONS(408),
    [anon_sym_char] = ACTIONS(408),
    [anon_sym_SQUOTE] = ACTIONS(408),
    [anon_sym_async] = ACTIONS(408),
    [anon_sym_break] = ACTIONS(408),
    [anon_sym_const] = ACTIONS(408),
    [anon_sym_continue] = ACTIONS(408),
    [anon_sym_default] = ACTIONS(408),
    [anon_sym_enum] = ACTIONS(408),
    [anon_sym_fn] = ACTIONS(408),
    [anon_sym_for] = ACTIONS(408),
    [anon_sym_if] = ACTIONS(408),
    [anon_sym_impl] = ACTIONS(408),
    [anon_sym_let] = ACTIONS(408),
    [anon_sym_loop] = ACTIONS(408),
    [anon_sym_match] = ACTIONS(408),
    [anon_sym_mod] = ACTIONS(408),
    [anon_sym_pub] = ACTIONS(408),
    [anon_sym_return] = ACTIONS(408),
    [anon_sym_static] = ACTIONS(408),
    [anon_sym_struct] = ACTIONS(408),
    [anon_sym_trait] = ACTIONS(408),
    [anon_sym_type] = ACTIONS(408),
    [anon_sym_union] = ACTIONS(408),
    [anon_sym_unsafe] = ACTIONS(408),
    [anon_sym_use] = ACTIONS(408),
    [anon_sym_while] = ACTIONS(408),
    [anon_sym_POUND] = ACTIONS(406),
    [anon_sym_BANG] = ACTIONS(406),
    [anon_sym_extern] = ACTIONS(408),
    [anon_sym_LT] = ACTIONS(406),
    [anon_sym_COLON_COLON] = ACTIONS(406),
    [anon_sym_AMP] = ACTIONS(406),
    [anon_sym_DOT_DOT] = ACTIONS(406),
    [anon_sym_DASH] = ACTIONS(406),
    [anon_sym_PIPE] = ACTIONS(406),
    [anon_sym_yield] = ACTIONS(408),
    [anon_sym_move] = ACTIONS(408),
    [sym_integer_literal] = ACTIONS(406),
    [aux_sym_string_literal_token1] = ACTIONS(406),
    [sym_char_literal] = ACTIONS(406),
    [anon_sym_true] = ACTIONS(408),
    [anon_sym_false] = ACTIONS(408),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(408),
    [sym_super] = ACTIONS(408),
    [sym_crate] = ACTIONS(408),
    [sym_metavariable] = ACTIONS(406),
    [sym_raw_string_literal] = ACTIONS(406),
    [sym_float_literal] = ACTIONS(406),
    [sym_block_comment] = ACTIONS(3),
  },
  [331] = {
    [ts_builtin_sym_end] = ACTIONS(1392),
    [sym_identifier] = ACTIONS(1394),
    [anon_sym_SEMI] = ACTIONS(1392),
    [anon_sym_macro_rules_BANG] = ACTIONS(1392),
    [anon_sym_LPAREN] = ACTIONS(1392),
    [anon_sym_LBRACE] = ACTIONS(1392),
    [anon_sym_RBRACE] = ACTIONS(1392),
    [anon_sym_LBRACK] = ACTIONS(1392),
    [anon_sym_STAR] = ACTIONS(1392),
    [anon_sym_u8] = ACTIONS(1394),
    [anon_sym_i8] = ACTIONS(1394),
    [anon_sym_u16] = ACTIONS(1394),
    [anon_sym_i16] = ACTIONS(1394),
    [anon_sym_u32] = ACTIONS(1394),
    [anon_sym_i32] = ACTIONS(1394),
    [anon_sym_u64] = ACTIONS(1394),
    [anon_sym_i64] = ACTIONS(1394),
    [anon_sym_u128] = ACTIONS(1394),
    [anon_sym_i128] = ACTIONS(1394),
    [anon_sym_isize] = ACTIONS(1394),
    [anon_sym_usize] = ACTIONS(1394),
    [anon_sym_f32] = ACTIONS(1394),
    [anon_sym_f64] = ACTIONS(1394),
    [anon_sym_bool] = ACTIONS(1394),
    [anon_sym_str] = ACTIONS(1394),
    [anon_sym_char] = ACTIONS(1394),
    [anon_sym_SQUOTE] = ACTIONS(1394),
    [anon_sym_async] = ACTIONS(1394),
    [anon_sym_break] = ACTIONS(1394),
    [anon_sym_const] = ACTIONS(1394),
    [anon_sym_continue] = ACTIONS(1394),
    [anon_sym_default] = ACTIONS(1394),
    [anon_sym_enum] = ACTIONS(1394),
    [anon_sym_fn] = ACTIONS(1394),
    [anon_sym_for] = ACTIONS(1394),
    [anon_sym_if] = ACTIONS(1394),
    [anon_sym_impl] = ACTIONS(1394),
    [anon_sym_let] = ACTIONS(1394),
    [anon_sym_loop] = ACTIONS(1394),
    [anon_sym_match] = ACTIONS(1394),
    [anon_sym_mod] = ACTIONS(1394),
    [anon_sym_pub] = ACTIONS(1394),
    [anon_sym_return] = ACTIONS(1394),
    [anon_sym_static] = ACTIONS(1394),
    [anon_sym_struct] = ACTIONS(1394),
    [anon_sym_trait] = ACTIONS(1394),
    [anon_sym_type] = ACTIONS(1394),
    [anon_sym_union] = ACTIONS(1394),
    [anon_sym_unsafe] = ACTIONS(1394),
    [anon_sym_use] = ACTIONS(1394),
    [anon_sym_while] = ACTIONS(1394),
    [anon_sym_POUND] = ACTIONS(1392),
    [anon_sym_BANG] = ACTIONS(1392),
    [anon_sym_extern] = ACTIONS(1394),
    [anon_sym_LT] = ACTIONS(1392),
    [anon_sym_COLON_COLON] = ACTIONS(1392),
    [anon_sym_AMP] = ACTIONS(1392),
    [anon_sym_DOT_DOT] = ACTIONS(1392),
    [anon_sym_DASH] = ACTIONS(1392),
    [anon_sym_PIPE] = ACTIONS(1392),
    [anon_sym_yield] = ACTIONS(1394),
    [anon_sym_move] = ACTIONS(1394),
    [sym_integer_literal] = ACTIONS(1392),
    [aux_sym_string_literal_token1] = ACTIONS(1392),
    [sym_char_literal] = ACTIONS(1392),
    [anon_sym_true] = ACTIONS(1394),
    [anon_sym_false] = ACTIONS(1394),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1394),
    [sym_super] = ACTIONS(1394),
    [sym_crate] = ACTIONS(1394),
    [sym_metavariable] = ACTIONS(1392),
    [sym_raw_string_literal] = ACTIONS(1392),
    [sym_float_literal] = ACTIONS(1392),
    [sym_block_comment] = ACTIONS(3),
  },
  [332] = {
    [ts_builtin_sym_end] = ACTIONS(1396),
    [sym_identifier] = ACTIONS(1398),
    [anon_sym_SEMI] = ACTIONS(1396),
    [anon_sym_macro_rules_BANG] = ACTIONS(1396),
    [anon_sym_LPAREN] = ACTIONS(1396),
    [anon_sym_LBRACE] = ACTIONS(1396),
    [anon_sym_RBRACE] = ACTIONS(1396),
    [anon_sym_LBRACK] = ACTIONS(1396),
    [anon_sym_STAR] = ACTIONS(1396),
    [anon_sym_u8] = ACTIONS(1398),
    [anon_sym_i8] = ACTIONS(1398),
    [anon_sym_u16] = ACTIONS(1398),
    [anon_sym_i16] = ACTIONS(1398),
    [anon_sym_u32] = ACTIONS(1398),
    [anon_sym_i32] = ACTIONS(1398),
    [anon_sym_u64] = ACTIONS(1398),
    [anon_sym_i64] = ACTIONS(1398),
    [anon_sym_u128] = ACTIONS(1398),
    [anon_sym_i128] = ACTIONS(1398),
    [anon_sym_isize] = ACTIONS(1398),
    [anon_sym_usize] = ACTIONS(1398),
    [anon_sym_f32] = ACTIONS(1398),
    [anon_sym_f64] = ACTIONS(1398),
    [anon_sym_bool] = ACTIONS(1398),
    [anon_sym_str] = ACTIONS(1398),
    [anon_sym_char] = ACTIONS(1398),
    [anon_sym_SQUOTE] = ACTIONS(1398),
    [anon_sym_async] = ACTIONS(1398),
    [anon_sym_break] = ACTIONS(1398),
    [anon_sym_const] = ACTIONS(1398),
    [anon_sym_continue] = ACTIONS(1398),
    [anon_sym_default] = ACTIONS(1398),
    [anon_sym_enum] = ACTIONS(1398),
    [anon_sym_fn] = ACTIONS(1398),
    [anon_sym_for] = ACTIONS(1398),
    [anon_sym_if] = ACTIONS(1398),
    [anon_sym_impl] = ACTIONS(1398),
    [anon_sym_let] = ACTIONS(1398),
    [anon_sym_loop] = ACTIONS(1398),
    [anon_sym_match] = ACTIONS(1398),
    [anon_sym_mod] = ACTIONS(1398),
    [anon_sym_pub] = ACTIONS(1398),
    [anon_sym_return] = ACTIONS(1398),
    [anon_sym_static] = ACTIONS(1398),
    [anon_sym_struct] = ACTIONS(1398),
    [anon_sym_trait] = ACTIONS(1398),
    [anon_sym_type] = ACTIONS(1398),
    [anon_sym_union] = ACTIONS(1398),
    [anon_sym_unsafe] = ACTIONS(1398),
    [anon_sym_use] = ACTIONS(1398),
    [anon_sym_while] = ACTIONS(1398),
    [anon_sym_POUND] = ACTIONS(1396),
    [anon_sym_BANG] = ACTIONS(1396),
    [anon_sym_extern] = ACTIONS(1398),
    [anon_sym_LT] = ACTIONS(1396),
    [anon_sym_COLON_COLON] = ACTIONS(1396),
    [anon_sym_AMP] = ACTIONS(1396),
    [anon_sym_DOT_DOT] = ACTIONS(1396),
    [anon_sym_DASH] = ACTIONS(1396),
    [anon_sym_PIPE] = ACTIONS(1396),
    [anon_sym_yield] = ACTIONS(1398),
    [anon_sym_move] = ACTIONS(1398),
    [sym_integer_literal] = ACTIONS(1396),
    [aux_sym_string_literal_token1] = ACTIONS(1396),
    [sym_char_literal] = ACTIONS(1396),
    [anon_sym_true] = ACTIONS(1398),
    [anon_sym_false] = ACTIONS(1398),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1398),
    [sym_super] = ACTIONS(1398),
    [sym_crate] = ACTIONS(1398),
    [sym_metavariable] = ACTIONS(1396),
    [sym_raw_string_literal] = ACTIONS(1396),
    [sym_float_literal] = ACTIONS(1396),
    [sym_block_comment] = ACTIONS(3),
  },
  [333] = {
    [ts_builtin_sym_end] = ACTIONS(1400),
    [sym_identifier] = ACTIONS(1402),
    [anon_sym_SEMI] = ACTIONS(1400),
    [anon_sym_macro_rules_BANG] = ACTIONS(1400),
    [anon_sym_LPAREN] = ACTIONS(1400),
    [anon_sym_LBRACE] = ACTIONS(1400),
    [anon_sym_RBRACE] = ACTIONS(1400),
    [anon_sym_LBRACK] = ACTIONS(1400),
    [anon_sym_STAR] = ACTIONS(1400),
    [anon_sym_u8] = ACTIONS(1402),
    [anon_sym_i8] = ACTIONS(1402),
    [anon_sym_u16] = ACTIONS(1402),
    [anon_sym_i16] = ACTIONS(1402),
    [anon_sym_u32] = ACTIONS(1402),
    [anon_sym_i32] = ACTIONS(1402),
    [anon_sym_u64] = ACTIONS(1402),
    [anon_sym_i64] = ACTIONS(1402),
    [anon_sym_u128] = ACTIONS(1402),
    [anon_sym_i128] = ACTIONS(1402),
    [anon_sym_isize] = ACTIONS(1402),
    [anon_sym_usize] = ACTIONS(1402),
    [anon_sym_f32] = ACTIONS(1402),
    [anon_sym_f64] = ACTIONS(1402),
    [anon_sym_bool] = ACTIONS(1402),
    [anon_sym_str] = ACTIONS(1402),
    [anon_sym_char] = ACTIONS(1402),
    [anon_sym_SQUOTE] = ACTIONS(1402),
    [anon_sym_async] = ACTIONS(1402),
    [anon_sym_break] = ACTIONS(1402),
    [anon_sym_const] = ACTIONS(1402),
    [anon_sym_continue] = ACTIONS(1402),
    [anon_sym_default] = ACTIONS(1402),
    [anon_sym_enum] = ACTIONS(1402),
    [anon_sym_fn] = ACTIONS(1402),
    [anon_sym_for] = ACTIONS(1402),
    [anon_sym_if] = ACTIONS(1402),
    [anon_sym_impl] = ACTIONS(1402),
    [anon_sym_let] = ACTIONS(1402),
    [anon_sym_loop] = ACTIONS(1402),
    [anon_sym_match] = ACTIONS(1402),
    [anon_sym_mod] = ACTIONS(1402),
    [anon_sym_pub] = ACTIONS(1402),
    [anon_sym_return] = ACTIONS(1402),
    [anon_sym_static] = ACTIONS(1402),
    [anon_sym_struct] = ACTIONS(1402),
    [anon_sym_trait] = ACTIONS(1402),
    [anon_sym_type] = ACTIONS(1402),
    [anon_sym_union] = ACTIONS(1402),
    [anon_sym_unsafe] = ACTIONS(1402),
    [anon_sym_use] = ACTIONS(1402),
    [anon_sym_while] = ACTIONS(1402),
    [anon_sym_POUND] = ACTIONS(1400),
    [anon_sym_BANG] = ACTIONS(1400),
    [anon_sym_extern] = ACTIONS(1402),
    [anon_sym_LT] = ACTIONS(1400),
    [anon_sym_COLON_COLON] = ACTIONS(1400),
    [anon_sym_AMP] = ACTIONS(1400),
    [anon_sym_DOT_DOT] = ACTIONS(1400),
    [anon_sym_DASH] = ACTIONS(1400),
    [anon_sym_PIPE] = ACTIONS(1400),
    [anon_sym_yield] = ACTIONS(1402),
    [anon_sym_move] = ACTIONS(1402),
    [sym_integer_literal] = ACTIONS(1400),
    [aux_sym_string_literal_token1] = ACTIONS(1400),
    [sym_char_literal] = ACTIONS(1400),
    [anon_sym_true] = ACTIONS(1402),
    [anon_sym_false] = ACTIONS(1402),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1402),
    [sym_super] = ACTIONS(1402),
    [sym_crate] = ACTIONS(1402),
    [sym_metavariable] = ACTIONS(1400),
    [sym_raw_string_literal] = ACTIONS(1400),
    [sym_float_literal] = ACTIONS(1400),
    [sym_block_comment] = ACTIONS(3),
  },
  [334] = {
    [ts_builtin_sym_end] = ACTIONS(1404),
    [sym_identifier] = ACTIONS(1406),
    [anon_sym_SEMI] = ACTIONS(1404),
    [anon_sym_macro_rules_BANG] = ACTIONS(1404),
    [anon_sym_LPAREN] = ACTIONS(1404),
    [anon_sym_LBRACE] = ACTIONS(1404),
    [anon_sym_RBRACE] = ACTIONS(1404),
    [anon_sym_LBRACK] = ACTIONS(1404),
    [anon_sym_STAR] = ACTIONS(1404),
    [anon_sym_u8] = ACTIONS(1406),
    [anon_sym_i8] = ACTIONS(1406),
    [anon_sym_u16] = ACTIONS(1406),
    [anon_sym_i16] = ACTIONS(1406),
    [anon_sym_u32] = ACTIONS(1406),
    [anon_sym_i32] = ACTIONS(1406),
    [anon_sym_u64] = ACTIONS(1406),
    [anon_sym_i64] = ACTIONS(1406),
    [anon_sym_u128] = ACTIONS(1406),
    [anon_sym_i128] = ACTIONS(1406),
    [anon_sym_isize] = ACTIONS(1406),
    [anon_sym_usize] = ACTIONS(1406),
    [anon_sym_f32] = ACTIONS(1406),
    [anon_sym_f64] = ACTIONS(1406),
    [anon_sym_bool] = ACTIONS(1406),
    [anon_sym_str] = ACTIONS(1406),
    [anon_sym_char] = ACTIONS(1406),
    [anon_sym_SQUOTE] = ACTIONS(1406),
    [anon_sym_async] = ACTIONS(1406),
    [anon_sym_break] = ACTIONS(1406),
    [anon_sym_const] = ACTIONS(1406),
    [anon_sym_continue] = ACTIONS(1406),
    [anon_sym_default] = ACTIONS(1406),
    [anon_sym_enum] = ACTIONS(1406),
    [anon_sym_fn] = ACTIONS(1406),
    [anon_sym_for] = ACTIONS(1406),
    [anon_sym_if] = ACTIONS(1406),
    [anon_sym_impl] = ACTIONS(1406),
    [anon_sym_let] = ACTIONS(1406),
    [anon_sym_loop] = ACTIONS(1406),
    [anon_sym_match] = ACTIONS(1406),
    [anon_sym_mod] = ACTIONS(1406),
    [anon_sym_pub] = ACTIONS(1406),
    [anon_sym_return] = ACTIONS(1406),
    [anon_sym_static] = ACTIONS(1406),
    [anon_sym_struct] = ACTIONS(1406),
    [anon_sym_trait] = ACTIONS(1406),
    [anon_sym_type] = ACTIONS(1406),
    [anon_sym_union] = ACTIONS(1406),
    [anon_sym_unsafe] = ACTIONS(1406),
    [anon_sym_use] = ACTIONS(1406),
    [anon_sym_while] = ACTIONS(1406),
    [anon_sym_POUND] = ACTIONS(1404),
    [anon_sym_BANG] = ACTIONS(1404),
    [anon_sym_extern] = ACTIONS(1406),
    [anon_sym_LT] = ACTIONS(1404),
    [anon_sym_COLON_COLON] = ACTIONS(1404),
    [anon_sym_AMP] = ACTIONS(1404),
    [anon_sym_DOT_DOT] = ACTIONS(1404),
    [anon_sym_DASH] = ACTIONS(1404),
    [anon_sym_PIPE] = ACTIONS(1404),
    [anon_sym_yield] = ACTIONS(1406),
    [anon_sym_move] = ACTIONS(1406),
    [sym_integer_literal] = ACTIONS(1404),
    [aux_sym_string_literal_token1] = ACTIONS(1404),
    [sym_char_literal] = ACTIONS(1404),
    [anon_sym_true] = ACTIONS(1406),
    [anon_sym_false] = ACTIONS(1406),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1406),
    [sym_super] = ACTIONS(1406),
    [sym_crate] = ACTIONS(1406),
    [sym_metavariable] = ACTIONS(1404),
    [sym_raw_string_literal] = ACTIONS(1404),
    [sym_float_literal] = ACTIONS(1404),
    [sym_block_comment] = ACTIONS(3),
  },
  [335] = {
    [ts_builtin_sym_end] = ACTIONS(1408),
    [sym_identifier] = ACTIONS(1410),
    [anon_sym_SEMI] = ACTIONS(1408),
    [anon_sym_macro_rules_BANG] = ACTIONS(1408),
    [anon_sym_LPAREN] = ACTIONS(1408),
    [anon_sym_LBRACE] = ACTIONS(1408),
    [anon_sym_RBRACE] = ACTIONS(1408),
    [anon_sym_LBRACK] = ACTIONS(1408),
    [anon_sym_STAR] = ACTIONS(1408),
    [anon_sym_u8] = ACTIONS(1410),
    [anon_sym_i8] = ACTIONS(1410),
    [anon_sym_u16] = ACTIONS(1410),
    [anon_sym_i16] = ACTIONS(1410),
    [anon_sym_u32] = ACTIONS(1410),
    [anon_sym_i32] = ACTIONS(1410),
    [anon_sym_u64] = ACTIONS(1410),
    [anon_sym_i64] = ACTIONS(1410),
    [anon_sym_u128] = ACTIONS(1410),
    [anon_sym_i128] = ACTIONS(1410),
    [anon_sym_isize] = ACTIONS(1410),
    [anon_sym_usize] = ACTIONS(1410),
    [anon_sym_f32] = ACTIONS(1410),
    [anon_sym_f64] = ACTIONS(1410),
    [anon_sym_bool] = ACTIONS(1410),
    [anon_sym_str] = ACTIONS(1410),
    [anon_sym_char] = ACTIONS(1410),
    [anon_sym_SQUOTE] = ACTIONS(1410),
    [anon_sym_async] = ACTIONS(1410),
    [anon_sym_break] = ACTIONS(1410),
    [anon_sym_const] = ACTIONS(1410),
    [anon_sym_continue] = ACTIONS(1410),
    [anon_sym_default] = ACTIONS(1410),
    [anon_sym_enum] = ACTIONS(1410),
    [anon_sym_fn] = ACTIONS(1410),
    [anon_sym_for] = ACTIONS(1410),
    [anon_sym_if] = ACTIONS(1410),
    [anon_sym_impl] = ACTIONS(1410),
    [anon_sym_let] = ACTIONS(1410),
    [anon_sym_loop] = ACTIONS(1410),
    [anon_sym_match] = ACTIONS(1410),
    [anon_sym_mod] = ACTIONS(1410),
    [anon_sym_pub] = ACTIONS(1410),
    [anon_sym_return] = ACTIONS(1410),
    [anon_sym_static] = ACTIONS(1410),
    [anon_sym_struct] = ACTIONS(1410),
    [anon_sym_trait] = ACTIONS(1410),
    [anon_sym_type] = ACTIONS(1410),
    [anon_sym_union] = ACTIONS(1410),
    [anon_sym_unsafe] = ACTIONS(1410),
    [anon_sym_use] = ACTIONS(1410),
    [anon_sym_while] = ACTIONS(1410),
    [anon_sym_POUND] = ACTIONS(1408),
    [anon_sym_BANG] = ACTIONS(1408),
    [anon_sym_extern] = ACTIONS(1410),
    [anon_sym_LT] = ACTIONS(1408),
    [anon_sym_COLON_COLON] = ACTIONS(1408),
    [anon_sym_AMP] = ACTIONS(1408),
    [anon_sym_DOT_DOT] = ACTIONS(1408),
    [anon_sym_DASH] = ACTIONS(1408),
    [anon_sym_PIPE] = ACTIONS(1408),
    [anon_sym_yield] = ACTIONS(1410),
    [anon_sym_move] = ACTIONS(1410),
    [sym_integer_literal] = ACTIONS(1408),
    [aux_sym_string_literal_token1] = ACTIONS(1408),
    [sym_char_literal] = ACTIONS(1408),
    [anon_sym_true] = ACTIONS(1410),
    [anon_sym_false] = ACTIONS(1410),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1410),
    [sym_super] = ACTIONS(1410),
    [sym_crate] = ACTIONS(1410),
    [sym_metavariable] = ACTIONS(1408),
    [sym_raw_string_literal] = ACTIONS(1408),
    [sym_float_literal] = ACTIONS(1408),
    [sym_block_comment] = ACTIONS(3),
  },
  [336] = {
    [ts_builtin_sym_end] = ACTIONS(1412),
    [sym_identifier] = ACTIONS(1414),
    [anon_sym_SEMI] = ACTIONS(1412),
    [anon_sym_macro_rules_BANG] = ACTIONS(1412),
    [anon_sym_LPAREN] = ACTIONS(1412),
    [anon_sym_LBRACE] = ACTIONS(1412),
    [anon_sym_RBRACE] = ACTIONS(1412),
    [anon_sym_LBRACK] = ACTIONS(1412),
    [anon_sym_STAR] = ACTIONS(1412),
    [anon_sym_u8] = ACTIONS(1414),
    [anon_sym_i8] = ACTIONS(1414),
    [anon_sym_u16] = ACTIONS(1414),
    [anon_sym_i16] = ACTIONS(1414),
    [anon_sym_u32] = ACTIONS(1414),
    [anon_sym_i32] = ACTIONS(1414),
    [anon_sym_u64] = ACTIONS(1414),
    [anon_sym_i64] = ACTIONS(1414),
    [anon_sym_u128] = ACTIONS(1414),
    [anon_sym_i128] = ACTIONS(1414),
    [anon_sym_isize] = ACTIONS(1414),
    [anon_sym_usize] = ACTIONS(1414),
    [anon_sym_f32] = ACTIONS(1414),
    [anon_sym_f64] = ACTIONS(1414),
    [anon_sym_bool] = ACTIONS(1414),
    [anon_sym_str] = ACTIONS(1414),
    [anon_sym_char] = ACTIONS(1414),
    [anon_sym_SQUOTE] = ACTIONS(1414),
    [anon_sym_async] = ACTIONS(1414),
    [anon_sym_break] = ACTIONS(1414),
    [anon_sym_const] = ACTIONS(1414),
    [anon_sym_continue] = ACTIONS(1414),
    [anon_sym_default] = ACTIONS(1414),
    [anon_sym_enum] = ACTIONS(1414),
    [anon_sym_fn] = ACTIONS(1414),
    [anon_sym_for] = ACTIONS(1414),
    [anon_sym_if] = ACTIONS(1414),
    [anon_sym_impl] = ACTIONS(1414),
    [anon_sym_let] = ACTIONS(1414),
    [anon_sym_loop] = ACTIONS(1414),
    [anon_sym_match] = ACTIONS(1414),
    [anon_sym_mod] = ACTIONS(1414),
    [anon_sym_pub] = ACTIONS(1414),
    [anon_sym_return] = ACTIONS(1414),
    [anon_sym_static] = ACTIONS(1414),
    [anon_sym_struct] = ACTIONS(1414),
    [anon_sym_trait] = ACTIONS(1414),
    [anon_sym_type] = ACTIONS(1414),
    [anon_sym_union] = ACTIONS(1414),
    [anon_sym_unsafe] = ACTIONS(1414),
    [anon_sym_use] = ACTIONS(1414),
    [anon_sym_while] = ACTIONS(1414),
    [anon_sym_POUND] = ACTIONS(1412),
    [anon_sym_BANG] = ACTIONS(1412),
    [anon_sym_extern] = ACTIONS(1414),
    [anon_sym_LT] = ACTIONS(1412),
    [anon_sym_COLON_COLON] = ACTIONS(1412),
    [anon_sym_AMP] = ACTIONS(1412),
    [anon_sym_DOT_DOT] = ACTIONS(1412),
    [anon_sym_DASH] = ACTIONS(1412),
    [anon_sym_PIPE] = ACTIONS(1412),
    [anon_sym_yield] = ACTIONS(1414),
    [anon_sym_move] = ACTIONS(1414),
    [sym_integer_literal] = ACTIONS(1412),
    [aux_sym_string_literal_token1] = ACTIONS(1412),
    [sym_char_literal] = ACTIONS(1412),
    [anon_sym_true] = ACTIONS(1414),
    [anon_sym_false] = ACTIONS(1414),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1414),
    [sym_super] = ACTIONS(1414),
    [sym_crate] = ACTIONS(1414),
    [sym_metavariable] = ACTIONS(1412),
    [sym_raw_string_literal] = ACTIONS(1412),
    [sym_float_literal] = ACTIONS(1412),
    [sym_block_comment] = ACTIONS(3),
  },
  [337] = {
    [ts_builtin_sym_end] = ACTIONS(410),
    [sym_identifier] = ACTIONS(412),
    [anon_sym_SEMI] = ACTIONS(410),
    [anon_sym_macro_rules_BANG] = ACTIONS(410),
    [anon_sym_LPAREN] = ACTIONS(410),
    [anon_sym_LBRACE] = ACTIONS(410),
    [anon_sym_RBRACE] = ACTIONS(410),
    [anon_sym_LBRACK] = ACTIONS(410),
    [anon_sym_STAR] = ACTIONS(410),
    [anon_sym_u8] = ACTIONS(412),
    [anon_sym_i8] = ACTIONS(412),
    [anon_sym_u16] = ACTIONS(412),
    [anon_sym_i16] = ACTIONS(412),
    [anon_sym_u32] = ACTIONS(412),
    [anon_sym_i32] = ACTIONS(412),
    [anon_sym_u64] = ACTIONS(412),
    [anon_sym_i64] = ACTIONS(412),
    [anon_sym_u128] = ACTIONS(412),
    [anon_sym_i128] = ACTIONS(412),
    [anon_sym_isize] = ACTIONS(412),
    [anon_sym_usize] = ACTIONS(412),
    [anon_sym_f32] = ACTIONS(412),
    [anon_sym_f64] = ACTIONS(412),
    [anon_sym_bool] = ACTIONS(412),
    [anon_sym_str] = ACTIONS(412),
    [anon_sym_char] = ACTIONS(412),
    [anon_sym_SQUOTE] = ACTIONS(412),
    [anon_sym_async] = ACTIONS(412),
    [anon_sym_break] = ACTIONS(412),
    [anon_sym_const] = ACTIONS(412),
    [anon_sym_continue] = ACTIONS(412),
    [anon_sym_default] = ACTIONS(412),
    [anon_sym_enum] = ACTIONS(412),
    [anon_sym_fn] = ACTIONS(412),
    [anon_sym_for] = ACTIONS(412),
    [anon_sym_if] = ACTIONS(412),
    [anon_sym_impl] = ACTIONS(412),
    [anon_sym_let] = ACTIONS(412),
    [anon_sym_loop] = ACTIONS(412),
    [anon_sym_match] = ACTIONS(412),
    [anon_sym_mod] = ACTIONS(412),
    [anon_sym_pub] = ACTIONS(412),
    [anon_sym_return] = ACTIONS(412),
    [anon_sym_static] = ACTIONS(412),
    [anon_sym_struct] = ACTIONS(412),
    [anon_sym_trait] = ACTIONS(412),
    [anon_sym_type] = ACTIONS(412),
    [anon_sym_union] = ACTIONS(412),
    [anon_sym_unsafe] = ACTIONS(412),
    [anon_sym_use] = ACTIONS(412),
    [anon_sym_while] = ACTIONS(412),
    [anon_sym_POUND] = ACTIONS(410),
    [anon_sym_BANG] = ACTIONS(410),
    [anon_sym_extern] = ACTIONS(412),
    [anon_sym_LT] = ACTIONS(410),
    [anon_sym_COLON_COLON] = ACTIONS(410),
    [anon_sym_AMP] = ACTIONS(410),
    [anon_sym_DOT_DOT] = ACTIONS(410),
    [anon_sym_DASH] = ACTIONS(410),
    [anon_sym_PIPE] = ACTIONS(410),
    [anon_sym_yield] = ACTIONS(412),
    [anon_sym_move] = ACTIONS(412),
    [sym_integer_literal] = ACTIONS(410),
    [aux_sym_string_literal_token1] = ACTIONS(410),
    [sym_char_literal] = ACTIONS(410),
    [anon_sym_true] = ACTIONS(412),
    [anon_sym_false] = ACTIONS(412),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(412),
    [sym_super] = ACTIONS(412),
    [sym_crate] = ACTIONS(412),
    [sym_metavariable] = ACTIONS(410),
    [sym_raw_string_literal] = ACTIONS(410),
    [sym_float_literal] = ACTIONS(410),
    [sym_block_comment] = ACTIONS(3),
  },
  [338] = {
    [ts_builtin_sym_end] = ACTIONS(1416),
    [sym_identifier] = ACTIONS(1418),
    [anon_sym_SEMI] = ACTIONS(1416),
    [anon_sym_macro_rules_BANG] = ACTIONS(1416),
    [anon_sym_LPAREN] = ACTIONS(1416),
    [anon_sym_LBRACE] = ACTIONS(1416),
    [anon_sym_RBRACE] = ACTIONS(1416),
    [anon_sym_LBRACK] = ACTIONS(1416),
    [anon_sym_STAR] = ACTIONS(1416),
    [anon_sym_u8] = ACTIONS(1418),
    [anon_sym_i8] = ACTIONS(1418),
    [anon_sym_u16] = ACTIONS(1418),
    [anon_sym_i16] = ACTIONS(1418),
    [anon_sym_u32] = ACTIONS(1418),
    [anon_sym_i32] = ACTIONS(1418),
    [anon_sym_u64] = ACTIONS(1418),
    [anon_sym_i64] = ACTIONS(1418),
    [anon_sym_u128] = ACTIONS(1418),
    [anon_sym_i128] = ACTIONS(1418),
    [anon_sym_isize] = ACTIONS(1418),
    [anon_sym_usize] = ACTIONS(1418),
    [anon_sym_f32] = ACTIONS(1418),
    [anon_sym_f64] = ACTIONS(1418),
    [anon_sym_bool] = ACTIONS(1418),
    [anon_sym_str] = ACTIONS(1418),
    [anon_sym_char] = ACTIONS(1418),
    [anon_sym_SQUOTE] = ACTIONS(1418),
    [anon_sym_async] = ACTIONS(1418),
    [anon_sym_break] = ACTIONS(1418),
    [anon_sym_const] = ACTIONS(1418),
    [anon_sym_continue] = ACTIONS(1418),
    [anon_sym_default] = ACTIONS(1418),
    [anon_sym_enum] = ACTIONS(1418),
    [anon_sym_fn] = ACTIONS(1418),
    [anon_sym_for] = ACTIONS(1418),
    [anon_sym_if] = ACTIONS(1418),
    [anon_sym_impl] = ACTIONS(1418),
    [anon_sym_let] = ACTIONS(1418),
    [anon_sym_loop] = ACTIONS(1418),
    [anon_sym_match] = ACTIONS(1418),
    [anon_sym_mod] = ACTIONS(1418),
    [anon_sym_pub] = ACTIONS(1418),
    [anon_sym_return] = ACTIONS(1418),
    [anon_sym_static] = ACTIONS(1418),
    [anon_sym_struct] = ACTIONS(1418),
    [anon_sym_trait] = ACTIONS(1418),
    [anon_sym_type] = ACTIONS(1418),
    [anon_sym_union] = ACTIONS(1418),
    [anon_sym_unsafe] = ACTIONS(1418),
    [anon_sym_use] = ACTIONS(1418),
    [anon_sym_while] = ACTIONS(1418),
    [anon_sym_POUND] = ACTIONS(1416),
    [anon_sym_BANG] = ACTIONS(1416),
    [anon_sym_extern] = ACTIONS(1418),
    [anon_sym_LT] = ACTIONS(1416),
    [anon_sym_COLON_COLON] = ACTIONS(1416),
    [anon_sym_AMP] = ACTIONS(1416),
    [anon_sym_DOT_DOT] = ACTIONS(1416),
    [anon_sym_DASH] = ACTIONS(1416),
    [anon_sym_PIPE] = ACTIONS(1416),
    [anon_sym_yield] = ACTIONS(1418),
    [anon_sym_move] = ACTIONS(1418),
    [sym_integer_literal] = ACTIONS(1416),
    [aux_sym_string_literal_token1] = ACTIONS(1416),
    [sym_char_literal] = ACTIONS(1416),
    [anon_sym_true] = ACTIONS(1418),
    [anon_sym_false] = ACTIONS(1418),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1418),
    [sym_super] = ACTIONS(1418),
    [sym_crate] = ACTIONS(1418),
    [sym_metavariable] = ACTIONS(1416),
    [sym_raw_string_literal] = ACTIONS(1416),
    [sym_float_literal] = ACTIONS(1416),
    [sym_block_comment] = ACTIONS(3),
  },
  [339] = {
    [ts_builtin_sym_end] = ACTIONS(1420),
    [sym_identifier] = ACTIONS(1422),
    [anon_sym_SEMI] = ACTIONS(1420),
    [anon_sym_macro_rules_BANG] = ACTIONS(1420),
    [anon_sym_LPAREN] = ACTIONS(1420),
    [anon_sym_LBRACE] = ACTIONS(1420),
    [anon_sym_RBRACE] = ACTIONS(1420),
    [anon_sym_LBRACK] = ACTIONS(1420),
    [anon_sym_STAR] = ACTIONS(1420),
    [anon_sym_u8] = ACTIONS(1422),
    [anon_sym_i8] = ACTIONS(1422),
    [anon_sym_u16] = ACTIONS(1422),
    [anon_sym_i16] = ACTIONS(1422),
    [anon_sym_u32] = ACTIONS(1422),
    [anon_sym_i32] = ACTIONS(1422),
    [anon_sym_u64] = ACTIONS(1422),
    [anon_sym_i64] = ACTIONS(1422),
    [anon_sym_u128] = ACTIONS(1422),
    [anon_sym_i128] = ACTIONS(1422),
    [anon_sym_isize] = ACTIONS(1422),
    [anon_sym_usize] = ACTIONS(1422),
    [anon_sym_f32] = ACTIONS(1422),
    [anon_sym_f64] = ACTIONS(1422),
    [anon_sym_bool] = ACTIONS(1422),
    [anon_sym_str] = ACTIONS(1422),
    [anon_sym_char] = ACTIONS(1422),
    [anon_sym_SQUOTE] = ACTIONS(1422),
    [anon_sym_async] = ACTIONS(1422),
    [anon_sym_break] = ACTIONS(1422),
    [anon_sym_const] = ACTIONS(1422),
    [anon_sym_continue] = ACTIONS(1422),
    [anon_sym_default] = ACTIONS(1422),
    [anon_sym_enum] = ACTIONS(1422),
    [anon_sym_fn] = ACTIONS(1422),
    [anon_sym_for] = ACTIONS(1422),
    [anon_sym_if] = ACTIONS(1422),
    [anon_sym_impl] = ACTIONS(1422),
    [anon_sym_let] = ACTIONS(1422),
    [anon_sym_loop] = ACTIONS(1422),
    [anon_sym_match] = ACTIONS(1422),
    [anon_sym_mod] = ACTIONS(1422),
    [anon_sym_pub] = ACTIONS(1422),
    [anon_sym_return] = ACTIONS(1422),
    [anon_sym_static] = ACTIONS(1422),
    [anon_sym_struct] = ACTIONS(1422),
    [anon_sym_trait] = ACTIONS(1422),
    [anon_sym_type] = ACTIONS(1422),
    [anon_sym_union] = ACTIONS(1422),
    [anon_sym_unsafe] = ACTIONS(1422),
    [anon_sym_use] = ACTIONS(1422),
    [anon_sym_while] = ACTIONS(1422),
    [anon_sym_POUND] = ACTIONS(1420),
    [anon_sym_BANG] = ACTIONS(1420),
    [anon_sym_extern] = ACTIONS(1422),
    [anon_sym_LT] = ACTIONS(1420),
    [anon_sym_COLON_COLON] = ACTIONS(1420),
    [anon_sym_AMP] = ACTIONS(1420),
    [anon_sym_DOT_DOT] = ACTIONS(1420),
    [anon_sym_DASH] = ACTIONS(1420),
    [anon_sym_PIPE] = ACTIONS(1420),
    [anon_sym_yield] = ACTIONS(1422),
    [anon_sym_move] = ACTIONS(1422),
    [sym_integer_literal] = ACTIONS(1420),
    [aux_sym_string_literal_token1] = ACTIONS(1420),
    [sym_char_literal] = ACTIONS(1420),
    [anon_sym_true] = ACTIONS(1422),
    [anon_sym_false] = ACTIONS(1422),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1422),
    [sym_super] = ACTIONS(1422),
    [sym_crate] = ACTIONS(1422),
    [sym_metavariable] = ACTIONS(1420),
    [sym_raw_string_literal] = ACTIONS(1420),
    [sym_float_literal] = ACTIONS(1420),
    [sym_block_comment] = ACTIONS(3),
  },
  [340] = {
    [ts_builtin_sym_end] = ACTIONS(1424),
    [sym_identifier] = ACTIONS(1426),
    [anon_sym_SEMI] = ACTIONS(1424),
    [anon_sym_macro_rules_BANG] = ACTIONS(1424),
    [anon_sym_LPAREN] = ACTIONS(1424),
    [anon_sym_LBRACE] = ACTIONS(1424),
    [anon_sym_RBRACE] = ACTIONS(1424),
    [anon_sym_LBRACK] = ACTIONS(1424),
    [anon_sym_STAR] = ACTIONS(1424),
    [anon_sym_u8] = ACTIONS(1426),
    [anon_sym_i8] = ACTIONS(1426),
    [anon_sym_u16] = ACTIONS(1426),
    [anon_sym_i16] = ACTIONS(1426),
    [anon_sym_u32] = ACTIONS(1426),
    [anon_sym_i32] = ACTIONS(1426),
    [anon_sym_u64] = ACTIONS(1426),
    [anon_sym_i64] = ACTIONS(1426),
    [anon_sym_u128] = ACTIONS(1426),
    [anon_sym_i128] = ACTIONS(1426),
    [anon_sym_isize] = ACTIONS(1426),
    [anon_sym_usize] = ACTIONS(1426),
    [anon_sym_f32] = ACTIONS(1426),
    [anon_sym_f64] = ACTIONS(1426),
    [anon_sym_bool] = ACTIONS(1426),
    [anon_sym_str] = ACTIONS(1426),
    [anon_sym_char] = ACTIONS(1426),
    [anon_sym_SQUOTE] = ACTIONS(1426),
    [anon_sym_async] = ACTIONS(1426),
    [anon_sym_break] = ACTIONS(1426),
    [anon_sym_const] = ACTIONS(1426),
    [anon_sym_continue] = ACTIONS(1426),
    [anon_sym_default] = ACTIONS(1426),
    [anon_sym_enum] = ACTIONS(1426),
    [anon_sym_fn] = ACTIONS(1426),
    [anon_sym_for] = ACTIONS(1426),
    [anon_sym_if] = ACTIONS(1426),
    [anon_sym_impl] = ACTIONS(1426),
    [anon_sym_let] = ACTIONS(1426),
    [anon_sym_loop] = ACTIONS(1426),
    [anon_sym_match] = ACTIONS(1426),
    [anon_sym_mod] = ACTIONS(1426),
    [anon_sym_pub] = ACTIONS(1426),
    [anon_sym_return] = ACTIONS(1426),
    [anon_sym_static] = ACTIONS(1426),
    [anon_sym_struct] = ACTIONS(1426),
    [anon_sym_trait] = ACTIONS(1426),
    [anon_sym_type] = ACTIONS(1426),
    [anon_sym_union] = ACTIONS(1426),
    [anon_sym_unsafe] = ACTIONS(1426),
    [anon_sym_use] = ACTIONS(1426),
    [anon_sym_while] = ACTIONS(1426),
    [anon_sym_POUND] = ACTIONS(1424),
    [anon_sym_BANG] = ACTIONS(1424),
    [anon_sym_extern] = ACTIONS(1426),
    [anon_sym_LT] = ACTIONS(1424),
    [anon_sym_COLON_COLON] = ACTIONS(1424),
    [anon_sym_AMP] = ACTIONS(1424),
    [anon_sym_DOT_DOT] = ACTIONS(1424),
    [anon_sym_DASH] = ACTIONS(1424),
    [anon_sym_PIPE] = ACTIONS(1424),
    [anon_sym_yield] = ACTIONS(1426),
    [anon_sym_move] = ACTIONS(1426),
    [sym_integer_literal] = ACTIONS(1424),
    [aux_sym_string_literal_token1] = ACTIONS(1424),
    [sym_char_literal] = ACTIONS(1424),
    [anon_sym_true] = ACTIONS(1426),
    [anon_sym_false] = ACTIONS(1426),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1426),
    [sym_super] = ACTIONS(1426),
    [sym_crate] = ACTIONS(1426),
    [sym_metavariable] = ACTIONS(1424),
    [sym_raw_string_literal] = ACTIONS(1424),
    [sym_float_literal] = ACTIONS(1424),
    [sym_block_comment] = ACTIONS(3),
  },
  [341] = {
    [ts_builtin_sym_end] = ACTIONS(1428),
    [sym_identifier] = ACTIONS(1430),
    [anon_sym_SEMI] = ACTIONS(1428),
    [anon_sym_macro_rules_BANG] = ACTIONS(1428),
    [anon_sym_LPAREN] = ACTIONS(1428),
    [anon_sym_LBRACE] = ACTIONS(1428),
    [anon_sym_RBRACE] = ACTIONS(1428),
    [anon_sym_LBRACK] = ACTIONS(1428),
    [anon_sym_STAR] = ACTIONS(1428),
    [anon_sym_u8] = ACTIONS(1430),
    [anon_sym_i8] = ACTIONS(1430),
    [anon_sym_u16] = ACTIONS(1430),
    [anon_sym_i16] = ACTIONS(1430),
    [anon_sym_u32] = ACTIONS(1430),
    [anon_sym_i32] = ACTIONS(1430),
    [anon_sym_u64] = ACTIONS(1430),
    [anon_sym_i64] = ACTIONS(1430),
    [anon_sym_u128] = ACTIONS(1430),
    [anon_sym_i128] = ACTIONS(1430),
    [anon_sym_isize] = ACTIONS(1430),
    [anon_sym_usize] = ACTIONS(1430),
    [anon_sym_f32] = ACTIONS(1430),
    [anon_sym_f64] = ACTIONS(1430),
    [anon_sym_bool] = ACTIONS(1430),
    [anon_sym_str] = ACTIONS(1430),
    [anon_sym_char] = ACTIONS(1430),
    [anon_sym_SQUOTE] = ACTIONS(1430),
    [anon_sym_async] = ACTIONS(1430),
    [anon_sym_break] = ACTIONS(1430),
    [anon_sym_const] = ACTIONS(1430),
    [anon_sym_continue] = ACTIONS(1430),
    [anon_sym_default] = ACTIONS(1430),
    [anon_sym_enum] = ACTIONS(1430),
    [anon_sym_fn] = ACTIONS(1430),
    [anon_sym_for] = ACTIONS(1430),
    [anon_sym_if] = ACTIONS(1430),
    [anon_sym_impl] = ACTIONS(1430),
    [anon_sym_let] = ACTIONS(1430),
    [anon_sym_loop] = ACTIONS(1430),
    [anon_sym_match] = ACTIONS(1430),
    [anon_sym_mod] = ACTIONS(1430),
    [anon_sym_pub] = ACTIONS(1430),
    [anon_sym_return] = ACTIONS(1430),
    [anon_sym_static] = ACTIONS(1430),
    [anon_sym_struct] = ACTIONS(1430),
    [anon_sym_trait] = ACTIONS(1430),
    [anon_sym_type] = ACTIONS(1430),
    [anon_sym_union] = ACTIONS(1430),
    [anon_sym_unsafe] = ACTIONS(1430),
    [anon_sym_use] = ACTIONS(1430),
    [anon_sym_while] = ACTIONS(1430),
    [anon_sym_POUND] = ACTIONS(1428),
    [anon_sym_BANG] = ACTIONS(1428),
    [anon_sym_extern] = ACTIONS(1430),
    [anon_sym_LT] = ACTIONS(1428),
    [anon_sym_COLON_COLON] = ACTIONS(1428),
    [anon_sym_AMP] = ACTIONS(1428),
    [anon_sym_DOT_DOT] = ACTIONS(1428),
    [anon_sym_DASH] = ACTIONS(1428),
    [anon_sym_PIPE] = ACTIONS(1428),
    [anon_sym_yield] = ACTIONS(1430),
    [anon_sym_move] = ACTIONS(1430),
    [sym_integer_literal] = ACTIONS(1428),
    [aux_sym_string_literal_token1] = ACTIONS(1428),
    [sym_char_literal] = ACTIONS(1428),
    [anon_sym_true] = ACTIONS(1430),
    [anon_sym_false] = ACTIONS(1430),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1430),
    [sym_super] = ACTIONS(1430),
    [sym_crate] = ACTIONS(1430),
    [sym_metavariable] = ACTIONS(1428),
    [sym_raw_string_literal] = ACTIONS(1428),
    [sym_float_literal] = ACTIONS(1428),
    [sym_block_comment] = ACTIONS(3),
  },
  [342] = {
    [ts_builtin_sym_end] = ACTIONS(1432),
    [sym_identifier] = ACTIONS(1434),
    [anon_sym_SEMI] = ACTIONS(1432),
    [anon_sym_macro_rules_BANG] = ACTIONS(1432),
    [anon_sym_LPAREN] = ACTIONS(1432),
    [anon_sym_LBRACE] = ACTIONS(1432),
    [anon_sym_RBRACE] = ACTIONS(1432),
    [anon_sym_LBRACK] = ACTIONS(1432),
    [anon_sym_STAR] = ACTIONS(1432),
    [anon_sym_u8] = ACTIONS(1434),
    [anon_sym_i8] = ACTIONS(1434),
    [anon_sym_u16] = ACTIONS(1434),
    [anon_sym_i16] = ACTIONS(1434),
    [anon_sym_u32] = ACTIONS(1434),
    [anon_sym_i32] = ACTIONS(1434),
    [anon_sym_u64] = ACTIONS(1434),
    [anon_sym_i64] = ACTIONS(1434),
    [anon_sym_u128] = ACTIONS(1434),
    [anon_sym_i128] = ACTIONS(1434),
    [anon_sym_isize] = ACTIONS(1434),
    [anon_sym_usize] = ACTIONS(1434),
    [anon_sym_f32] = ACTIONS(1434),
    [anon_sym_f64] = ACTIONS(1434),
    [anon_sym_bool] = ACTIONS(1434),
    [anon_sym_str] = ACTIONS(1434),
    [anon_sym_char] = ACTIONS(1434),
    [anon_sym_SQUOTE] = ACTIONS(1434),
    [anon_sym_async] = ACTIONS(1434),
    [anon_sym_break] = ACTIONS(1434),
    [anon_sym_const] = ACTIONS(1434),
    [anon_sym_continue] = ACTIONS(1434),
    [anon_sym_default] = ACTIONS(1434),
    [anon_sym_enum] = ACTIONS(1434),
    [anon_sym_fn] = ACTIONS(1434),
    [anon_sym_for] = ACTIONS(1434),
    [anon_sym_if] = ACTIONS(1434),
    [anon_sym_impl] = ACTIONS(1434),
    [anon_sym_let] = ACTIONS(1434),
    [anon_sym_loop] = ACTIONS(1434),
    [anon_sym_match] = ACTIONS(1434),
    [anon_sym_mod] = ACTIONS(1434),
    [anon_sym_pub] = ACTIONS(1434),
    [anon_sym_return] = ACTIONS(1434),
    [anon_sym_static] = ACTIONS(1434),
    [anon_sym_struct] = ACTIONS(1434),
    [anon_sym_trait] = ACTIONS(1434),
    [anon_sym_type] = ACTIONS(1434),
    [anon_sym_union] = ACTIONS(1434),
    [anon_sym_unsafe] = ACTIONS(1434),
    [anon_sym_use] = ACTIONS(1434),
    [anon_sym_while] = ACTIONS(1434),
    [anon_sym_POUND] = ACTIONS(1432),
    [anon_sym_BANG] = ACTIONS(1432),
    [anon_sym_extern] = ACTIONS(1434),
    [anon_sym_LT] = ACTIONS(1432),
    [anon_sym_COLON_COLON] = ACTIONS(1432),
    [anon_sym_AMP] = ACTIONS(1432),
    [anon_sym_DOT_DOT] = ACTIONS(1432),
    [anon_sym_DASH] = ACTIONS(1432),
    [anon_sym_PIPE] = ACTIONS(1432),
    [anon_sym_yield] = ACTIONS(1434),
    [anon_sym_move] = ACTIONS(1434),
    [sym_integer_literal] = ACTIONS(1432),
    [aux_sym_string_literal_token1] = ACTIONS(1432),
    [sym_char_literal] = ACTIONS(1432),
    [anon_sym_true] = ACTIONS(1434),
    [anon_sym_false] = ACTIONS(1434),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1434),
    [sym_super] = ACTIONS(1434),
    [sym_crate] = ACTIONS(1434),
    [sym_metavariable] = ACTIONS(1432),
    [sym_raw_string_literal] = ACTIONS(1432),
    [sym_float_literal] = ACTIONS(1432),
    [sym_block_comment] = ACTIONS(3),
  },
  [343] = {
    [ts_builtin_sym_end] = ACTIONS(1436),
    [sym_identifier] = ACTIONS(1438),
    [anon_sym_SEMI] = ACTIONS(1436),
    [anon_sym_macro_rules_BANG] = ACTIONS(1436),
    [anon_sym_LPAREN] = ACTIONS(1436),
    [anon_sym_LBRACE] = ACTIONS(1436),
    [anon_sym_RBRACE] = ACTIONS(1436),
    [anon_sym_LBRACK] = ACTIONS(1436),
    [anon_sym_STAR] = ACTIONS(1436),
    [anon_sym_u8] = ACTIONS(1438),
    [anon_sym_i8] = ACTIONS(1438),
    [anon_sym_u16] = ACTIONS(1438),
    [anon_sym_i16] = ACTIONS(1438),
    [anon_sym_u32] = ACTIONS(1438),
    [anon_sym_i32] = ACTIONS(1438),
    [anon_sym_u64] = ACTIONS(1438),
    [anon_sym_i64] = ACTIONS(1438),
    [anon_sym_u128] = ACTIONS(1438),
    [anon_sym_i128] = ACTIONS(1438),
    [anon_sym_isize] = ACTIONS(1438),
    [anon_sym_usize] = ACTIONS(1438),
    [anon_sym_f32] = ACTIONS(1438),
    [anon_sym_f64] = ACTIONS(1438),
    [anon_sym_bool] = ACTIONS(1438),
    [anon_sym_str] = ACTIONS(1438),
    [anon_sym_char] = ACTIONS(1438),
    [anon_sym_SQUOTE] = ACTIONS(1438),
    [anon_sym_async] = ACTIONS(1438),
    [anon_sym_break] = ACTIONS(1438),
    [anon_sym_const] = ACTIONS(1438),
    [anon_sym_continue] = ACTIONS(1438),
    [anon_sym_default] = ACTIONS(1438),
    [anon_sym_enum] = ACTIONS(1438),
    [anon_sym_fn] = ACTIONS(1438),
    [anon_sym_for] = ACTIONS(1438),
    [anon_sym_if] = ACTIONS(1438),
    [anon_sym_impl] = ACTIONS(1438),
    [anon_sym_let] = ACTIONS(1438),
    [anon_sym_loop] = ACTIONS(1438),
    [anon_sym_match] = ACTIONS(1438),
    [anon_sym_mod] = ACTIONS(1438),
    [anon_sym_pub] = ACTIONS(1438),
    [anon_sym_return] = ACTIONS(1438),
    [anon_sym_static] = ACTIONS(1438),
    [anon_sym_struct] = ACTIONS(1438),
    [anon_sym_trait] = ACTIONS(1438),
    [anon_sym_type] = ACTIONS(1438),
    [anon_sym_union] = ACTIONS(1438),
    [anon_sym_unsafe] = ACTIONS(1438),
    [anon_sym_use] = ACTIONS(1438),
    [anon_sym_while] = ACTIONS(1438),
    [anon_sym_POUND] = ACTIONS(1436),
    [anon_sym_BANG] = ACTIONS(1436),
    [anon_sym_extern] = ACTIONS(1438),
    [anon_sym_LT] = ACTIONS(1436),
    [anon_sym_COLON_COLON] = ACTIONS(1436),
    [anon_sym_AMP] = ACTIONS(1436),
    [anon_sym_DOT_DOT] = ACTIONS(1436),
    [anon_sym_DASH] = ACTIONS(1436),
    [anon_sym_PIPE] = ACTIONS(1436),
    [anon_sym_yield] = ACTIONS(1438),
    [anon_sym_move] = ACTIONS(1438),
    [sym_integer_literal] = ACTIONS(1436),
    [aux_sym_string_literal_token1] = ACTIONS(1436),
    [sym_char_literal] = ACTIONS(1436),
    [anon_sym_true] = ACTIONS(1438),
    [anon_sym_false] = ACTIONS(1438),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1438),
    [sym_super] = ACTIONS(1438),
    [sym_crate] = ACTIONS(1438),
    [sym_metavariable] = ACTIONS(1436),
    [sym_raw_string_literal] = ACTIONS(1436),
    [sym_float_literal] = ACTIONS(1436),
    [sym_block_comment] = ACTIONS(3),
  },
  [344] = {
    [ts_builtin_sym_end] = ACTIONS(1440),
    [sym_identifier] = ACTIONS(1442),
    [anon_sym_SEMI] = ACTIONS(1440),
    [anon_sym_macro_rules_BANG] = ACTIONS(1440),
    [anon_sym_LPAREN] = ACTIONS(1440),
    [anon_sym_LBRACE] = ACTIONS(1440),
    [anon_sym_RBRACE] = ACTIONS(1440),
    [anon_sym_LBRACK] = ACTIONS(1440),
    [anon_sym_STAR] = ACTIONS(1440),
    [anon_sym_u8] = ACTIONS(1442),
    [anon_sym_i8] = ACTIONS(1442),
    [anon_sym_u16] = ACTIONS(1442),
    [anon_sym_i16] = ACTIONS(1442),
    [anon_sym_u32] = ACTIONS(1442),
    [anon_sym_i32] = ACTIONS(1442),
    [anon_sym_u64] = ACTIONS(1442),
    [anon_sym_i64] = ACTIONS(1442),
    [anon_sym_u128] = ACTIONS(1442),
    [anon_sym_i128] = ACTIONS(1442),
    [anon_sym_isize] = ACTIONS(1442),
    [anon_sym_usize] = ACTIONS(1442),
    [anon_sym_f32] = ACTIONS(1442),
    [anon_sym_f64] = ACTIONS(1442),
    [anon_sym_bool] = ACTIONS(1442),
    [anon_sym_str] = ACTIONS(1442),
    [anon_sym_char] = ACTIONS(1442),
    [anon_sym_SQUOTE] = ACTIONS(1442),
    [anon_sym_async] = ACTIONS(1442),
    [anon_sym_break] = ACTIONS(1442),
    [anon_sym_const] = ACTIONS(1442),
    [anon_sym_continue] = ACTIONS(1442),
    [anon_sym_default] = ACTIONS(1442),
    [anon_sym_enum] = ACTIONS(1442),
    [anon_sym_fn] = ACTIONS(1442),
    [anon_sym_for] = ACTIONS(1442),
    [anon_sym_if] = ACTIONS(1442),
    [anon_sym_impl] = ACTIONS(1442),
    [anon_sym_let] = ACTIONS(1442),
    [anon_sym_loop] = ACTIONS(1442),
    [anon_sym_match] = ACTIONS(1442),
    [anon_sym_mod] = ACTIONS(1442),
    [anon_sym_pub] = ACTIONS(1442),
    [anon_sym_return] = ACTIONS(1442),
    [anon_sym_static] = ACTIONS(1442),
    [anon_sym_struct] = ACTIONS(1442),
    [anon_sym_trait] = ACTIONS(1442),
    [anon_sym_type] = ACTIONS(1442),
    [anon_sym_union] = ACTIONS(1442),
    [anon_sym_unsafe] = ACTIONS(1442),
    [anon_sym_use] = ACTIONS(1442),
    [anon_sym_while] = ACTIONS(1442),
    [anon_sym_POUND] = ACTIONS(1440),
    [anon_sym_BANG] = ACTIONS(1440),
    [anon_sym_extern] = ACTIONS(1442),
    [anon_sym_LT] = ACTIONS(1440),
    [anon_sym_COLON_COLON] = ACTIONS(1440),
    [anon_sym_AMP] = ACTIONS(1440),
    [anon_sym_DOT_DOT] = ACTIONS(1440),
    [anon_sym_DASH] = ACTIONS(1440),
    [anon_sym_PIPE] = ACTIONS(1440),
    [anon_sym_yield] = ACTIONS(1442),
    [anon_sym_move] = ACTIONS(1442),
    [sym_integer_literal] = ACTIONS(1440),
    [aux_sym_string_literal_token1] = ACTIONS(1440),
    [sym_char_literal] = ACTIONS(1440),
    [anon_sym_true] = ACTIONS(1442),
    [anon_sym_false] = ACTIONS(1442),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1442),
    [sym_super] = ACTIONS(1442),
    [sym_crate] = ACTIONS(1442),
    [sym_metavariable] = ACTIONS(1440),
    [sym_raw_string_literal] = ACTIONS(1440),
    [sym_float_literal] = ACTIONS(1440),
    [sym_block_comment] = ACTIONS(3),
  },
  [345] = {
    [ts_builtin_sym_end] = ACTIONS(1444),
    [sym_identifier] = ACTIONS(1446),
    [anon_sym_SEMI] = ACTIONS(1444),
    [anon_sym_macro_rules_BANG] = ACTIONS(1444),
    [anon_sym_LPAREN] = ACTIONS(1444),
    [anon_sym_LBRACE] = ACTIONS(1444),
    [anon_sym_RBRACE] = ACTIONS(1444),
    [anon_sym_LBRACK] = ACTIONS(1444),
    [anon_sym_STAR] = ACTIONS(1444),
    [anon_sym_u8] = ACTIONS(1446),
    [anon_sym_i8] = ACTIONS(1446),
    [anon_sym_u16] = ACTIONS(1446),
    [anon_sym_i16] = ACTIONS(1446),
    [anon_sym_u32] = ACTIONS(1446),
    [anon_sym_i32] = ACTIONS(1446),
    [anon_sym_u64] = ACTIONS(1446),
    [anon_sym_i64] = ACTIONS(1446),
    [anon_sym_u128] = ACTIONS(1446),
    [anon_sym_i128] = ACTIONS(1446),
    [anon_sym_isize] = ACTIONS(1446),
    [anon_sym_usize] = ACTIONS(1446),
    [anon_sym_f32] = ACTIONS(1446),
    [anon_sym_f64] = ACTIONS(1446),
    [anon_sym_bool] = ACTIONS(1446),
    [anon_sym_str] = ACTIONS(1446),
    [anon_sym_char] = ACTIONS(1446),
    [anon_sym_SQUOTE] = ACTIONS(1446),
    [anon_sym_async] = ACTIONS(1446),
    [anon_sym_break] = ACTIONS(1446),
    [anon_sym_const] = ACTIONS(1446),
    [anon_sym_continue] = ACTIONS(1446),
    [anon_sym_default] = ACTIONS(1446),
    [anon_sym_enum] = ACTIONS(1446),
    [anon_sym_fn] = ACTIONS(1446),
    [anon_sym_for] = ACTIONS(1446),
    [anon_sym_if] = ACTIONS(1446),
    [anon_sym_impl] = ACTIONS(1446),
    [anon_sym_let] = ACTIONS(1446),
    [anon_sym_loop] = ACTIONS(1446),
    [anon_sym_match] = ACTIONS(1446),
    [anon_sym_mod] = ACTIONS(1446),
    [anon_sym_pub] = ACTIONS(1446),
    [anon_sym_return] = ACTIONS(1446),
    [anon_sym_static] = ACTIONS(1446),
    [anon_sym_struct] = ACTIONS(1446),
    [anon_sym_trait] = ACTIONS(1446),
    [anon_sym_type] = ACTIONS(1446),
    [anon_sym_union] = ACTIONS(1446),
    [anon_sym_unsafe] = ACTIONS(1446),
    [anon_sym_use] = ACTIONS(1446),
    [anon_sym_while] = ACTIONS(1446),
    [anon_sym_POUND] = ACTIONS(1444),
    [anon_sym_BANG] = ACTIONS(1444),
    [anon_sym_extern] = ACTIONS(1446),
    [anon_sym_LT] = ACTIONS(1444),
    [anon_sym_COLON_COLON] = ACTIONS(1444),
    [anon_sym_AMP] = ACTIONS(1444),
    [anon_sym_DOT_DOT] = ACTIONS(1444),
    [anon_sym_DASH] = ACTIONS(1444),
    [anon_sym_PIPE] = ACTIONS(1444),
    [anon_sym_yield] = ACTIONS(1446),
    [anon_sym_move] = ACTIONS(1446),
    [sym_integer_literal] = ACTIONS(1444),
    [aux_sym_string_literal_token1] = ACTIONS(1444),
    [sym_char_literal] = ACTIONS(1444),
    [anon_sym_true] = ACTIONS(1446),
    [anon_sym_false] = ACTIONS(1446),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1446),
    [sym_super] = ACTIONS(1446),
    [sym_crate] = ACTIONS(1446),
    [sym_metavariable] = ACTIONS(1444),
    [sym_raw_string_literal] = ACTIONS(1444),
    [sym_float_literal] = ACTIONS(1444),
    [sym_block_comment] = ACTIONS(3),
  },
  [346] = {
    [ts_builtin_sym_end] = ACTIONS(1448),
    [sym_identifier] = ACTIONS(1450),
    [anon_sym_SEMI] = ACTIONS(1448),
    [anon_sym_macro_rules_BANG] = ACTIONS(1448),
    [anon_sym_LPAREN] = ACTIONS(1448),
    [anon_sym_LBRACE] = ACTIONS(1448),
    [anon_sym_RBRACE] = ACTIONS(1448),
    [anon_sym_LBRACK] = ACTIONS(1448),
    [anon_sym_STAR] = ACTIONS(1448),
    [anon_sym_u8] = ACTIONS(1450),
    [anon_sym_i8] = ACTIONS(1450),
    [anon_sym_u16] = ACTIONS(1450),
    [anon_sym_i16] = ACTIONS(1450),
    [anon_sym_u32] = ACTIONS(1450),
    [anon_sym_i32] = ACTIONS(1450),
    [anon_sym_u64] = ACTIONS(1450),
    [anon_sym_i64] = ACTIONS(1450),
    [anon_sym_u128] = ACTIONS(1450),
    [anon_sym_i128] = ACTIONS(1450),
    [anon_sym_isize] = ACTIONS(1450),
    [anon_sym_usize] = ACTIONS(1450),
    [anon_sym_f32] = ACTIONS(1450),
    [anon_sym_f64] = ACTIONS(1450),
    [anon_sym_bool] = ACTIONS(1450),
    [anon_sym_str] = ACTIONS(1450),
    [anon_sym_char] = ACTIONS(1450),
    [anon_sym_SQUOTE] = ACTIONS(1450),
    [anon_sym_async] = ACTIONS(1450),
    [anon_sym_break] = ACTIONS(1450),
    [anon_sym_const] = ACTIONS(1450),
    [anon_sym_continue] = ACTIONS(1450),
    [anon_sym_default] = ACTIONS(1450),
    [anon_sym_enum] = ACTIONS(1450),
    [anon_sym_fn] = ACTIONS(1450),
    [anon_sym_for] = ACTIONS(1450),
    [anon_sym_if] = ACTIONS(1450),
    [anon_sym_impl] = ACTIONS(1450),
    [anon_sym_let] = ACTIONS(1450),
    [anon_sym_loop] = ACTIONS(1450),
    [anon_sym_match] = ACTIONS(1450),
    [anon_sym_mod] = ACTIONS(1450),
    [anon_sym_pub] = ACTIONS(1450),
    [anon_sym_return] = ACTIONS(1450),
    [anon_sym_static] = ACTIONS(1450),
    [anon_sym_struct] = ACTIONS(1450),
    [anon_sym_trait] = ACTIONS(1450),
    [anon_sym_type] = ACTIONS(1450),
    [anon_sym_union] = ACTIONS(1450),
    [anon_sym_unsafe] = ACTIONS(1450),
    [anon_sym_use] = ACTIONS(1450),
    [anon_sym_while] = ACTIONS(1450),
    [anon_sym_POUND] = ACTIONS(1448),
    [anon_sym_BANG] = ACTIONS(1448),
    [anon_sym_extern] = ACTIONS(1450),
    [anon_sym_LT] = ACTIONS(1448),
    [anon_sym_COLON_COLON] = ACTIONS(1448),
    [anon_sym_AMP] = ACTIONS(1448),
    [anon_sym_DOT_DOT] = ACTIONS(1448),
    [anon_sym_DASH] = ACTIONS(1448),
    [anon_sym_PIPE] = ACTIONS(1448),
    [anon_sym_yield] = ACTIONS(1450),
    [anon_sym_move] = ACTIONS(1450),
    [sym_integer_literal] = ACTIONS(1448),
    [aux_sym_string_literal_token1] = ACTIONS(1448),
    [sym_char_literal] = ACTIONS(1448),
    [anon_sym_true] = ACTIONS(1450),
    [anon_sym_false] = ACTIONS(1450),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1450),
    [sym_super] = ACTIONS(1450),
    [sym_crate] = ACTIONS(1450),
    [sym_metavariable] = ACTIONS(1448),
    [sym_raw_string_literal] = ACTIONS(1448),
    [sym_float_literal] = ACTIONS(1448),
    [sym_block_comment] = ACTIONS(3),
  },
  [347] = {
    [ts_builtin_sym_end] = ACTIONS(1452),
    [sym_identifier] = ACTIONS(1454),
    [anon_sym_SEMI] = ACTIONS(1452),
    [anon_sym_macro_rules_BANG] = ACTIONS(1452),
    [anon_sym_LPAREN] = ACTIONS(1452),
    [anon_sym_LBRACE] = ACTIONS(1452),
    [anon_sym_RBRACE] = ACTIONS(1452),
    [anon_sym_LBRACK] = ACTIONS(1452),
    [anon_sym_STAR] = ACTIONS(1452),
    [anon_sym_u8] = ACTIONS(1454),
    [anon_sym_i8] = ACTIONS(1454),
    [anon_sym_u16] = ACTIONS(1454),
    [anon_sym_i16] = ACTIONS(1454),
    [anon_sym_u32] = ACTIONS(1454),
    [anon_sym_i32] = ACTIONS(1454),
    [anon_sym_u64] = ACTIONS(1454),
    [anon_sym_i64] = ACTIONS(1454),
    [anon_sym_u128] = ACTIONS(1454),
    [anon_sym_i128] = ACTIONS(1454),
    [anon_sym_isize] = ACTIONS(1454),
    [anon_sym_usize] = ACTIONS(1454),
    [anon_sym_f32] = ACTIONS(1454),
    [anon_sym_f64] = ACTIONS(1454),
    [anon_sym_bool] = ACTIONS(1454),
    [anon_sym_str] = ACTIONS(1454),
    [anon_sym_char] = ACTIONS(1454),
    [anon_sym_SQUOTE] = ACTIONS(1454),
    [anon_sym_async] = ACTIONS(1454),
    [anon_sym_break] = ACTIONS(1454),
    [anon_sym_const] = ACTIONS(1454),
    [anon_sym_continue] = ACTIONS(1454),
    [anon_sym_default] = ACTIONS(1454),
    [anon_sym_enum] = ACTIONS(1454),
    [anon_sym_fn] = ACTIONS(1454),
    [anon_sym_for] = ACTIONS(1454),
    [anon_sym_if] = ACTIONS(1454),
    [anon_sym_impl] = ACTIONS(1454),
    [anon_sym_let] = ACTIONS(1454),
    [anon_sym_loop] = ACTIONS(1454),
    [anon_sym_match] = ACTIONS(1454),
    [anon_sym_mod] = ACTIONS(1454),
    [anon_sym_pub] = ACTIONS(1454),
    [anon_sym_return] = ACTIONS(1454),
    [anon_sym_static] = ACTIONS(1454),
    [anon_sym_struct] = ACTIONS(1454),
    [anon_sym_trait] = ACTIONS(1454),
    [anon_sym_type] = ACTIONS(1454),
    [anon_sym_union] = ACTIONS(1454),
    [anon_sym_unsafe] = ACTIONS(1454),
    [anon_sym_use] = ACTIONS(1454),
    [anon_sym_while] = ACTIONS(1454),
    [anon_sym_POUND] = ACTIONS(1452),
    [anon_sym_BANG] = ACTIONS(1452),
    [anon_sym_extern] = ACTIONS(1454),
    [anon_sym_LT] = ACTIONS(1452),
    [anon_sym_COLON_COLON] = ACTIONS(1452),
    [anon_sym_AMP] = ACTIONS(1452),
    [anon_sym_DOT_DOT] = ACTIONS(1452),
    [anon_sym_DASH] = ACTIONS(1452),
    [anon_sym_PIPE] = ACTIONS(1452),
    [anon_sym_yield] = ACTIONS(1454),
    [anon_sym_move] = ACTIONS(1454),
    [sym_integer_literal] = ACTIONS(1452),
    [aux_sym_string_literal_token1] = ACTIONS(1452),
    [sym_char_literal] = ACTIONS(1452),
    [anon_sym_true] = ACTIONS(1454),
    [anon_sym_false] = ACTIONS(1454),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1454),
    [sym_super] = ACTIONS(1454),
    [sym_crate] = ACTIONS(1454),
    [sym_metavariable] = ACTIONS(1452),
    [sym_raw_string_literal] = ACTIONS(1452),
    [sym_float_literal] = ACTIONS(1452),
    [sym_block_comment] = ACTIONS(3),
  },
  [348] = {
    [ts_builtin_sym_end] = ACTIONS(1456),
    [sym_identifier] = ACTIONS(1458),
    [anon_sym_SEMI] = ACTIONS(1456),
    [anon_sym_macro_rules_BANG] = ACTIONS(1456),
    [anon_sym_LPAREN] = ACTIONS(1456),
    [anon_sym_LBRACE] = ACTIONS(1456),
    [anon_sym_RBRACE] = ACTIONS(1456),
    [anon_sym_LBRACK] = ACTIONS(1456),
    [anon_sym_STAR] = ACTIONS(1456),
    [anon_sym_u8] = ACTIONS(1458),
    [anon_sym_i8] = ACTIONS(1458),
    [anon_sym_u16] = ACTIONS(1458),
    [anon_sym_i16] = ACTIONS(1458),
    [anon_sym_u32] = ACTIONS(1458),
    [anon_sym_i32] = ACTIONS(1458),
    [anon_sym_u64] = ACTIONS(1458),
    [anon_sym_i64] = ACTIONS(1458),
    [anon_sym_u128] = ACTIONS(1458),
    [anon_sym_i128] = ACTIONS(1458),
    [anon_sym_isize] = ACTIONS(1458),
    [anon_sym_usize] = ACTIONS(1458),
    [anon_sym_f32] = ACTIONS(1458),
    [anon_sym_f64] = ACTIONS(1458),
    [anon_sym_bool] = ACTIONS(1458),
    [anon_sym_str] = ACTIONS(1458),
    [anon_sym_char] = ACTIONS(1458),
    [anon_sym_SQUOTE] = ACTIONS(1458),
    [anon_sym_async] = ACTIONS(1458),
    [anon_sym_break] = ACTIONS(1458),
    [anon_sym_const] = ACTIONS(1458),
    [anon_sym_continue] = ACTIONS(1458),
    [anon_sym_default] = ACTIONS(1458),
    [anon_sym_enum] = ACTIONS(1458),
    [anon_sym_fn] = ACTIONS(1458),
    [anon_sym_for] = ACTIONS(1458),
    [anon_sym_if] = ACTIONS(1458),
    [anon_sym_impl] = ACTIONS(1458),
    [anon_sym_let] = ACTIONS(1458),
    [anon_sym_loop] = ACTIONS(1458),
    [anon_sym_match] = ACTIONS(1458),
    [anon_sym_mod] = ACTIONS(1458),
    [anon_sym_pub] = ACTIONS(1458),
    [anon_sym_return] = ACTIONS(1458),
    [anon_sym_static] = ACTIONS(1458),
    [anon_sym_struct] = ACTIONS(1458),
    [anon_sym_trait] = ACTIONS(1458),
    [anon_sym_type] = ACTIONS(1458),
    [anon_sym_union] = ACTIONS(1458),
    [anon_sym_unsafe] = ACTIONS(1458),
    [anon_sym_use] = ACTIONS(1458),
    [anon_sym_while] = ACTIONS(1458),
    [anon_sym_POUND] = ACTIONS(1456),
    [anon_sym_BANG] = ACTIONS(1456),
    [anon_sym_extern] = ACTIONS(1458),
    [anon_sym_LT] = ACTIONS(1456),
    [anon_sym_COLON_COLON] = ACTIONS(1456),
    [anon_sym_AMP] = ACTIONS(1456),
    [anon_sym_DOT_DOT] = ACTIONS(1456),
    [anon_sym_DASH] = ACTIONS(1456),
    [anon_sym_PIPE] = ACTIONS(1456),
    [anon_sym_yield] = ACTIONS(1458),
    [anon_sym_move] = ACTIONS(1458),
    [sym_integer_literal] = ACTIONS(1456),
    [aux_sym_string_literal_token1] = ACTIONS(1456),
    [sym_char_literal] = ACTIONS(1456),
    [anon_sym_true] = ACTIONS(1458),
    [anon_sym_false] = ACTIONS(1458),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1458),
    [sym_super] = ACTIONS(1458),
    [sym_crate] = ACTIONS(1458),
    [sym_metavariable] = ACTIONS(1456),
    [sym_raw_string_literal] = ACTIONS(1456),
    [sym_float_literal] = ACTIONS(1456),
    [sym_block_comment] = ACTIONS(3),
  },
  [349] = {
    [ts_builtin_sym_end] = ACTIONS(1460),
    [sym_identifier] = ACTIONS(1462),
    [anon_sym_SEMI] = ACTIONS(1460),
    [anon_sym_macro_rules_BANG] = ACTIONS(1460),
    [anon_sym_LPAREN] = ACTIONS(1460),
    [anon_sym_LBRACE] = ACTIONS(1460),
    [anon_sym_RBRACE] = ACTIONS(1460),
    [anon_sym_LBRACK] = ACTIONS(1460),
    [anon_sym_STAR] = ACTIONS(1460),
    [anon_sym_u8] = ACTIONS(1462),
    [anon_sym_i8] = ACTIONS(1462),
    [anon_sym_u16] = ACTIONS(1462),
    [anon_sym_i16] = ACTIONS(1462),
    [anon_sym_u32] = ACTIONS(1462),
    [anon_sym_i32] = ACTIONS(1462),
    [anon_sym_u64] = ACTIONS(1462),
    [anon_sym_i64] = ACTIONS(1462),
    [anon_sym_u128] = ACTIONS(1462),
    [anon_sym_i128] = ACTIONS(1462),
    [anon_sym_isize] = ACTIONS(1462),
    [anon_sym_usize] = ACTIONS(1462),
    [anon_sym_f32] = ACTIONS(1462),
    [anon_sym_f64] = ACTIONS(1462),
    [anon_sym_bool] = ACTIONS(1462),
    [anon_sym_str] = ACTIONS(1462),
    [anon_sym_char] = ACTIONS(1462),
    [anon_sym_SQUOTE] = ACTIONS(1462),
    [anon_sym_async] = ACTIONS(1462),
    [anon_sym_break] = ACTIONS(1462),
    [anon_sym_const] = ACTIONS(1462),
    [anon_sym_continue] = ACTIONS(1462),
    [anon_sym_default] = ACTIONS(1462),
    [anon_sym_enum] = ACTIONS(1462),
    [anon_sym_fn] = ACTIONS(1462),
    [anon_sym_for] = ACTIONS(1462),
    [anon_sym_if] = ACTIONS(1462),
    [anon_sym_impl] = ACTIONS(1462),
    [anon_sym_let] = ACTIONS(1462),
    [anon_sym_loop] = ACTIONS(1462),
    [anon_sym_match] = ACTIONS(1462),
    [anon_sym_mod] = ACTIONS(1462),
    [anon_sym_pub] = ACTIONS(1462),
    [anon_sym_return] = ACTIONS(1462),
    [anon_sym_static] = ACTIONS(1462),
    [anon_sym_struct] = ACTIONS(1462),
    [anon_sym_trait] = ACTIONS(1462),
    [anon_sym_type] = ACTIONS(1462),
    [anon_sym_union] = ACTIONS(1462),
    [anon_sym_unsafe] = ACTIONS(1462),
    [anon_sym_use] = ACTIONS(1462),
    [anon_sym_while] = ACTIONS(1462),
    [anon_sym_POUND] = ACTIONS(1460),
    [anon_sym_BANG] = ACTIONS(1460),
    [anon_sym_extern] = ACTIONS(1462),
    [anon_sym_LT] = ACTIONS(1460),
    [anon_sym_COLON_COLON] = ACTIONS(1460),
    [anon_sym_AMP] = ACTIONS(1460),
    [anon_sym_DOT_DOT] = ACTIONS(1460),
    [anon_sym_DASH] = ACTIONS(1460),
    [anon_sym_PIPE] = ACTIONS(1460),
    [anon_sym_yield] = ACTIONS(1462),
    [anon_sym_move] = ACTIONS(1462),
    [sym_integer_literal] = ACTIONS(1460),
    [aux_sym_string_literal_token1] = ACTIONS(1460),
    [sym_char_literal] = ACTIONS(1460),
    [anon_sym_true] = ACTIONS(1462),
    [anon_sym_false] = ACTIONS(1462),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1462),
    [sym_super] = ACTIONS(1462),
    [sym_crate] = ACTIONS(1462),
    [sym_metavariable] = ACTIONS(1460),
    [sym_raw_string_literal] = ACTIONS(1460),
    [sym_float_literal] = ACTIONS(1460),
    [sym_block_comment] = ACTIONS(3),
  },
  [350] = {
    [ts_builtin_sym_end] = ACTIONS(1464),
    [sym_identifier] = ACTIONS(1466),
    [anon_sym_SEMI] = ACTIONS(1464),
    [anon_sym_macro_rules_BANG] = ACTIONS(1464),
    [anon_sym_LPAREN] = ACTIONS(1464),
    [anon_sym_LBRACE] = ACTIONS(1464),
    [anon_sym_RBRACE] = ACTIONS(1464),
    [anon_sym_LBRACK] = ACTIONS(1464),
    [anon_sym_STAR] = ACTIONS(1464),
    [anon_sym_u8] = ACTIONS(1466),
    [anon_sym_i8] = ACTIONS(1466),
    [anon_sym_u16] = ACTIONS(1466),
    [anon_sym_i16] = ACTIONS(1466),
    [anon_sym_u32] = ACTIONS(1466),
    [anon_sym_i32] = ACTIONS(1466),
    [anon_sym_u64] = ACTIONS(1466),
    [anon_sym_i64] = ACTIONS(1466),
    [anon_sym_u128] = ACTIONS(1466),
    [anon_sym_i128] = ACTIONS(1466),
    [anon_sym_isize] = ACTIONS(1466),
    [anon_sym_usize] = ACTIONS(1466),
    [anon_sym_f32] = ACTIONS(1466),
    [anon_sym_f64] = ACTIONS(1466),
    [anon_sym_bool] = ACTIONS(1466),
    [anon_sym_str] = ACTIONS(1466),
    [anon_sym_char] = ACTIONS(1466),
    [anon_sym_SQUOTE] = ACTIONS(1466),
    [anon_sym_async] = ACTIONS(1466),
    [anon_sym_break] = ACTIONS(1466),
    [anon_sym_const] = ACTIONS(1466),
    [anon_sym_continue] = ACTIONS(1466),
    [anon_sym_default] = ACTIONS(1466),
    [anon_sym_enum] = ACTIONS(1466),
    [anon_sym_fn] = ACTIONS(1466),
    [anon_sym_for] = ACTIONS(1466),
    [anon_sym_if] = ACTIONS(1466),
    [anon_sym_impl] = ACTIONS(1466),
    [anon_sym_let] = ACTIONS(1466),
    [anon_sym_loop] = ACTIONS(1466),
    [anon_sym_match] = ACTIONS(1466),
    [anon_sym_mod] = ACTIONS(1466),
    [anon_sym_pub] = ACTIONS(1466),
    [anon_sym_return] = ACTIONS(1466),
    [anon_sym_static] = ACTIONS(1466),
    [anon_sym_struct] = ACTIONS(1466),
    [anon_sym_trait] = ACTIONS(1466),
    [anon_sym_type] = ACTIONS(1466),
    [anon_sym_union] = ACTIONS(1466),
    [anon_sym_unsafe] = ACTIONS(1466),
    [anon_sym_use] = ACTIONS(1466),
    [anon_sym_while] = ACTIONS(1466),
    [anon_sym_POUND] = ACTIONS(1464),
    [anon_sym_BANG] = ACTIONS(1464),
    [anon_sym_extern] = ACTIONS(1466),
    [anon_sym_LT] = ACTIONS(1464),
    [anon_sym_COLON_COLON] = ACTIONS(1464),
    [anon_sym_AMP] = ACTIONS(1464),
    [anon_sym_DOT_DOT] = ACTIONS(1464),
    [anon_sym_DASH] = ACTIONS(1464),
    [anon_sym_PIPE] = ACTIONS(1464),
    [anon_sym_yield] = ACTIONS(1466),
    [anon_sym_move] = ACTIONS(1466),
    [sym_integer_literal] = ACTIONS(1464),
    [aux_sym_string_literal_token1] = ACTIONS(1464),
    [sym_char_literal] = ACTIONS(1464),
    [anon_sym_true] = ACTIONS(1466),
    [anon_sym_false] = ACTIONS(1466),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1466),
    [sym_super] = ACTIONS(1466),
    [sym_crate] = ACTIONS(1466),
    [sym_metavariable] = ACTIONS(1464),
    [sym_raw_string_literal] = ACTIONS(1464),
    [sym_float_literal] = ACTIONS(1464),
    [sym_block_comment] = ACTIONS(3),
  },
  [351] = {
    [ts_builtin_sym_end] = ACTIONS(1468),
    [sym_identifier] = ACTIONS(1470),
    [anon_sym_SEMI] = ACTIONS(1468),
    [anon_sym_macro_rules_BANG] = ACTIONS(1468),
    [anon_sym_LPAREN] = ACTIONS(1468),
    [anon_sym_LBRACE] = ACTIONS(1468),
    [anon_sym_RBRACE] = ACTIONS(1468),
    [anon_sym_LBRACK] = ACTIONS(1468),
    [anon_sym_STAR] = ACTIONS(1468),
    [anon_sym_u8] = ACTIONS(1470),
    [anon_sym_i8] = ACTIONS(1470),
    [anon_sym_u16] = ACTIONS(1470),
    [anon_sym_i16] = ACTIONS(1470),
    [anon_sym_u32] = ACTIONS(1470),
    [anon_sym_i32] = ACTIONS(1470),
    [anon_sym_u64] = ACTIONS(1470),
    [anon_sym_i64] = ACTIONS(1470),
    [anon_sym_u128] = ACTIONS(1470),
    [anon_sym_i128] = ACTIONS(1470),
    [anon_sym_isize] = ACTIONS(1470),
    [anon_sym_usize] = ACTIONS(1470),
    [anon_sym_f32] = ACTIONS(1470),
    [anon_sym_f64] = ACTIONS(1470),
    [anon_sym_bool] = ACTIONS(1470),
    [anon_sym_str] = ACTIONS(1470),
    [anon_sym_char] = ACTIONS(1470),
    [anon_sym_SQUOTE] = ACTIONS(1470),
    [anon_sym_async] = ACTIONS(1470),
    [anon_sym_break] = ACTIONS(1470),
    [anon_sym_const] = ACTIONS(1470),
    [anon_sym_continue] = ACTIONS(1470),
    [anon_sym_default] = ACTIONS(1470),
    [anon_sym_enum] = ACTIONS(1470),
    [anon_sym_fn] = ACTIONS(1470),
    [anon_sym_for] = ACTIONS(1470),
    [anon_sym_if] = ACTIONS(1470),
    [anon_sym_impl] = ACTIONS(1470),
    [anon_sym_let] = ACTIONS(1470),
    [anon_sym_loop] = ACTIONS(1470),
    [anon_sym_match] = ACTIONS(1470),
    [anon_sym_mod] = ACTIONS(1470),
    [anon_sym_pub] = ACTIONS(1470),
    [anon_sym_return] = ACTIONS(1470),
    [anon_sym_static] = ACTIONS(1470),
    [anon_sym_struct] = ACTIONS(1470),
    [anon_sym_trait] = ACTIONS(1470),
    [anon_sym_type] = ACTIONS(1470),
    [anon_sym_union] = ACTIONS(1470),
    [anon_sym_unsafe] = ACTIONS(1470),
    [anon_sym_use] = ACTIONS(1470),
    [anon_sym_while] = ACTIONS(1470),
    [anon_sym_POUND] = ACTIONS(1468),
    [anon_sym_BANG] = ACTIONS(1468),
    [anon_sym_extern] = ACTIONS(1470),
    [anon_sym_LT] = ACTIONS(1468),
    [anon_sym_COLON_COLON] = ACTIONS(1468),
    [anon_sym_AMP] = ACTIONS(1468),
    [anon_sym_DOT_DOT] = ACTIONS(1468),
    [anon_sym_DASH] = ACTIONS(1468),
    [anon_sym_PIPE] = ACTIONS(1468),
    [anon_sym_yield] = ACTIONS(1470),
    [anon_sym_move] = ACTIONS(1470),
    [sym_integer_literal] = ACTIONS(1468),
    [aux_sym_string_literal_token1] = ACTIONS(1468),
    [sym_char_literal] = ACTIONS(1468),
    [anon_sym_true] = ACTIONS(1470),
    [anon_sym_false] = ACTIONS(1470),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1470),
    [sym_super] = ACTIONS(1470),
    [sym_crate] = ACTIONS(1470),
    [sym_metavariable] = ACTIONS(1468),
    [sym_raw_string_literal] = ACTIONS(1468),
    [sym_float_literal] = ACTIONS(1468),
    [sym_block_comment] = ACTIONS(3),
  },
  [352] = {
    [ts_builtin_sym_end] = ACTIONS(1472),
    [sym_identifier] = ACTIONS(1474),
    [anon_sym_SEMI] = ACTIONS(1472),
    [anon_sym_macro_rules_BANG] = ACTIONS(1472),
    [anon_sym_LPAREN] = ACTIONS(1472),
    [anon_sym_LBRACE] = ACTIONS(1472),
    [anon_sym_RBRACE] = ACTIONS(1472),
    [anon_sym_LBRACK] = ACTIONS(1472),
    [anon_sym_STAR] = ACTIONS(1472),
    [anon_sym_u8] = ACTIONS(1474),
    [anon_sym_i8] = ACTIONS(1474),
    [anon_sym_u16] = ACTIONS(1474),
    [anon_sym_i16] = ACTIONS(1474),
    [anon_sym_u32] = ACTIONS(1474),
    [anon_sym_i32] = ACTIONS(1474),
    [anon_sym_u64] = ACTIONS(1474),
    [anon_sym_i64] = ACTIONS(1474),
    [anon_sym_u128] = ACTIONS(1474),
    [anon_sym_i128] = ACTIONS(1474),
    [anon_sym_isize] = ACTIONS(1474),
    [anon_sym_usize] = ACTIONS(1474),
    [anon_sym_f32] = ACTIONS(1474),
    [anon_sym_f64] = ACTIONS(1474),
    [anon_sym_bool] = ACTIONS(1474),
    [anon_sym_str] = ACTIONS(1474),
    [anon_sym_char] = ACTIONS(1474),
    [anon_sym_SQUOTE] = ACTIONS(1474),
    [anon_sym_async] = ACTIONS(1474),
    [anon_sym_break] = ACTIONS(1474),
    [anon_sym_const] = ACTIONS(1474),
    [anon_sym_continue] = ACTIONS(1474),
    [anon_sym_default] = ACTIONS(1474),
    [anon_sym_enum] = ACTIONS(1474),
    [anon_sym_fn] = ACTIONS(1474),
    [anon_sym_for] = ACTIONS(1474),
    [anon_sym_if] = ACTIONS(1474),
    [anon_sym_impl] = ACTIONS(1474),
    [anon_sym_let] = ACTIONS(1474),
    [anon_sym_loop] = ACTIONS(1474),
    [anon_sym_match] = ACTIONS(1474),
    [anon_sym_mod] = ACTIONS(1474),
    [anon_sym_pub] = ACTIONS(1474),
    [anon_sym_return] = ACTIONS(1474),
    [anon_sym_static] = ACTIONS(1474),
    [anon_sym_struct] = ACTIONS(1474),
    [anon_sym_trait] = ACTIONS(1474),
    [anon_sym_type] = ACTIONS(1474),
    [anon_sym_union] = ACTIONS(1474),
    [anon_sym_unsafe] = ACTIONS(1474),
    [anon_sym_use] = ACTIONS(1474),
    [anon_sym_while] = ACTIONS(1474),
    [anon_sym_POUND] = ACTIONS(1472),
    [anon_sym_BANG] = ACTIONS(1472),
    [anon_sym_extern] = ACTIONS(1474),
    [anon_sym_LT] = ACTIONS(1472),
    [anon_sym_COLON_COLON] = ACTIONS(1472),
    [anon_sym_AMP] = ACTIONS(1472),
    [anon_sym_DOT_DOT] = ACTIONS(1472),
    [anon_sym_DASH] = ACTIONS(1472),
    [anon_sym_PIPE] = ACTIONS(1472),
    [anon_sym_yield] = ACTIONS(1474),
    [anon_sym_move] = ACTIONS(1474),
    [sym_integer_literal] = ACTIONS(1472),
    [aux_sym_string_literal_token1] = ACTIONS(1472),
    [sym_char_literal] = ACTIONS(1472),
    [anon_sym_true] = ACTIONS(1474),
    [anon_sym_false] = ACTIONS(1474),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1474),
    [sym_super] = ACTIONS(1474),
    [sym_crate] = ACTIONS(1474),
    [sym_metavariable] = ACTIONS(1472),
    [sym_raw_string_literal] = ACTIONS(1472),
    [sym_float_literal] = ACTIONS(1472),
    [sym_block_comment] = ACTIONS(3),
  },
  [353] = {
    [ts_builtin_sym_end] = ACTIONS(1476),
    [sym_identifier] = ACTIONS(1478),
    [anon_sym_SEMI] = ACTIONS(1476),
    [anon_sym_macro_rules_BANG] = ACTIONS(1476),
    [anon_sym_LPAREN] = ACTIONS(1476),
    [anon_sym_LBRACE] = ACTIONS(1476),
    [anon_sym_RBRACE] = ACTIONS(1476),
    [anon_sym_LBRACK] = ACTIONS(1476),
    [anon_sym_STAR] = ACTIONS(1476),
    [anon_sym_u8] = ACTIONS(1478),
    [anon_sym_i8] = ACTIONS(1478),
    [anon_sym_u16] = ACTIONS(1478),
    [anon_sym_i16] = ACTIONS(1478),
    [anon_sym_u32] = ACTIONS(1478),
    [anon_sym_i32] = ACTIONS(1478),
    [anon_sym_u64] = ACTIONS(1478),
    [anon_sym_i64] = ACTIONS(1478),
    [anon_sym_u128] = ACTIONS(1478),
    [anon_sym_i128] = ACTIONS(1478),
    [anon_sym_isize] = ACTIONS(1478),
    [anon_sym_usize] = ACTIONS(1478),
    [anon_sym_f32] = ACTIONS(1478),
    [anon_sym_f64] = ACTIONS(1478),
    [anon_sym_bool] = ACTIONS(1478),
    [anon_sym_str] = ACTIONS(1478),
    [anon_sym_char] = ACTIONS(1478),
    [anon_sym_SQUOTE] = ACTIONS(1478),
    [anon_sym_async] = ACTIONS(1478),
    [anon_sym_break] = ACTIONS(1478),
    [anon_sym_const] = ACTIONS(1478),
    [anon_sym_continue] = ACTIONS(1478),
    [anon_sym_default] = ACTIONS(1478),
    [anon_sym_enum] = ACTIONS(1478),
    [anon_sym_fn] = ACTIONS(1478),
    [anon_sym_for] = ACTIONS(1478),
    [anon_sym_if] = ACTIONS(1478),
    [anon_sym_impl] = ACTIONS(1478),
    [anon_sym_let] = ACTIONS(1478),
    [anon_sym_loop] = ACTIONS(1478),
    [anon_sym_match] = ACTIONS(1478),
    [anon_sym_mod] = ACTIONS(1478),
    [anon_sym_pub] = ACTIONS(1478),
    [anon_sym_return] = ACTIONS(1478),
    [anon_sym_static] = ACTIONS(1478),
    [anon_sym_struct] = ACTIONS(1478),
    [anon_sym_trait] = ACTIONS(1478),
    [anon_sym_type] = ACTIONS(1478),
    [anon_sym_union] = ACTIONS(1478),
    [anon_sym_unsafe] = ACTIONS(1478),
    [anon_sym_use] = ACTIONS(1478),
    [anon_sym_while] = ACTIONS(1478),
    [anon_sym_POUND] = ACTIONS(1476),
    [anon_sym_BANG] = ACTIONS(1476),
    [anon_sym_extern] = ACTIONS(1478),
    [anon_sym_LT] = ACTIONS(1476),
    [anon_sym_COLON_COLON] = ACTIONS(1476),
    [anon_sym_AMP] = ACTIONS(1476),
    [anon_sym_DOT_DOT] = ACTIONS(1476),
    [anon_sym_DASH] = ACTIONS(1476),
    [anon_sym_PIPE] = ACTIONS(1476),
    [anon_sym_yield] = ACTIONS(1478),
    [anon_sym_move] = ACTIONS(1478),
    [sym_integer_literal] = ACTIONS(1476),
    [aux_sym_string_literal_token1] = ACTIONS(1476),
    [sym_char_literal] = ACTIONS(1476),
    [anon_sym_true] = ACTIONS(1478),
    [anon_sym_false] = ACTIONS(1478),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1478),
    [sym_super] = ACTIONS(1478),
    [sym_crate] = ACTIONS(1478),
    [sym_metavariable] = ACTIONS(1476),
    [sym_raw_string_literal] = ACTIONS(1476),
    [sym_float_literal] = ACTIONS(1476),
    [sym_block_comment] = ACTIONS(3),
  },
  [354] = {
    [ts_builtin_sym_end] = ACTIONS(1480),
    [sym_identifier] = ACTIONS(1482),
    [anon_sym_SEMI] = ACTIONS(1480),
    [anon_sym_macro_rules_BANG] = ACTIONS(1480),
    [anon_sym_LPAREN] = ACTIONS(1480),
    [anon_sym_LBRACE] = ACTIONS(1480),
    [anon_sym_RBRACE] = ACTIONS(1480),
    [anon_sym_LBRACK] = ACTIONS(1480),
    [anon_sym_STAR] = ACTIONS(1480),
    [anon_sym_u8] = ACTIONS(1482),
    [anon_sym_i8] = ACTIONS(1482),
    [anon_sym_u16] = ACTIONS(1482),
    [anon_sym_i16] = ACTIONS(1482),
    [anon_sym_u32] = ACTIONS(1482),
    [anon_sym_i32] = ACTIONS(1482),
    [anon_sym_u64] = ACTIONS(1482),
    [anon_sym_i64] = ACTIONS(1482),
    [anon_sym_u128] = ACTIONS(1482),
    [anon_sym_i128] = ACTIONS(1482),
    [anon_sym_isize] = ACTIONS(1482),
    [anon_sym_usize] = ACTIONS(1482),
    [anon_sym_f32] = ACTIONS(1482),
    [anon_sym_f64] = ACTIONS(1482),
    [anon_sym_bool] = ACTIONS(1482),
    [anon_sym_str] = ACTIONS(1482),
    [anon_sym_char] = ACTIONS(1482),
    [anon_sym_SQUOTE] = ACTIONS(1482),
    [anon_sym_async] = ACTIONS(1482),
    [anon_sym_break] = ACTIONS(1482),
    [anon_sym_const] = ACTIONS(1482),
    [anon_sym_continue] = ACTIONS(1482),
    [anon_sym_default] = ACTIONS(1482),
    [anon_sym_enum] = ACTIONS(1482),
    [anon_sym_fn] = ACTIONS(1482),
    [anon_sym_for] = ACTIONS(1482),
    [anon_sym_if] = ACTIONS(1482),
    [anon_sym_impl] = ACTIONS(1482),
    [anon_sym_let] = ACTIONS(1482),
    [anon_sym_loop] = ACTIONS(1482),
    [anon_sym_match] = ACTIONS(1482),
    [anon_sym_mod] = ACTIONS(1482),
    [anon_sym_pub] = ACTIONS(1482),
    [anon_sym_return] = ACTIONS(1482),
    [anon_sym_static] = ACTIONS(1482),
    [anon_sym_struct] = ACTIONS(1482),
    [anon_sym_trait] = ACTIONS(1482),
    [anon_sym_type] = ACTIONS(1482),
    [anon_sym_union] = ACTIONS(1482),
    [anon_sym_unsafe] = ACTIONS(1482),
    [anon_sym_use] = ACTIONS(1482),
    [anon_sym_while] = ACTIONS(1482),
    [anon_sym_POUND] = ACTIONS(1480),
    [anon_sym_BANG] = ACTIONS(1480),
    [anon_sym_extern] = ACTIONS(1482),
    [anon_sym_LT] = ACTIONS(1480),
    [anon_sym_COLON_COLON] = ACTIONS(1480),
    [anon_sym_AMP] = ACTIONS(1480),
    [anon_sym_DOT_DOT] = ACTIONS(1480),
    [anon_sym_DASH] = ACTIONS(1480),
    [anon_sym_PIPE] = ACTIONS(1480),
    [anon_sym_yield] = ACTIONS(1482),
    [anon_sym_move] = ACTIONS(1482),
    [sym_integer_literal] = ACTIONS(1480),
    [aux_sym_string_literal_token1] = ACTIONS(1480),
    [sym_char_literal] = ACTIONS(1480),
    [anon_sym_true] = ACTIONS(1482),
    [anon_sym_false] = ACTIONS(1482),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1482),
    [sym_super] = ACTIONS(1482),
    [sym_crate] = ACTIONS(1482),
    [sym_metavariable] = ACTIONS(1480),
    [sym_raw_string_literal] = ACTIONS(1480),
    [sym_float_literal] = ACTIONS(1480),
    [sym_block_comment] = ACTIONS(3),
  },
  [355] = {
    [ts_builtin_sym_end] = ACTIONS(1484),
    [sym_identifier] = ACTIONS(1486),
    [anon_sym_SEMI] = ACTIONS(1484),
    [anon_sym_macro_rules_BANG] = ACTIONS(1484),
    [anon_sym_LPAREN] = ACTIONS(1484),
    [anon_sym_LBRACE] = ACTIONS(1484),
    [anon_sym_RBRACE] = ACTIONS(1484),
    [anon_sym_LBRACK] = ACTIONS(1484),
    [anon_sym_STAR] = ACTIONS(1484),
    [anon_sym_u8] = ACTIONS(1486),
    [anon_sym_i8] = ACTIONS(1486),
    [anon_sym_u16] = ACTIONS(1486),
    [anon_sym_i16] = ACTIONS(1486),
    [anon_sym_u32] = ACTIONS(1486),
    [anon_sym_i32] = ACTIONS(1486),
    [anon_sym_u64] = ACTIONS(1486),
    [anon_sym_i64] = ACTIONS(1486),
    [anon_sym_u128] = ACTIONS(1486),
    [anon_sym_i128] = ACTIONS(1486),
    [anon_sym_isize] = ACTIONS(1486),
    [anon_sym_usize] = ACTIONS(1486),
    [anon_sym_f32] = ACTIONS(1486),
    [anon_sym_f64] = ACTIONS(1486),
    [anon_sym_bool] = ACTIONS(1486),
    [anon_sym_str] = ACTIONS(1486),
    [anon_sym_char] = ACTIONS(1486),
    [anon_sym_SQUOTE] = ACTIONS(1486),
    [anon_sym_async] = ACTIONS(1486),
    [anon_sym_break] = ACTIONS(1486),
    [anon_sym_const] = ACTIONS(1486),
    [anon_sym_continue] = ACTIONS(1486),
    [anon_sym_default] = ACTIONS(1486),
    [anon_sym_enum] = ACTIONS(1486),
    [anon_sym_fn] = ACTIONS(1486),
    [anon_sym_for] = ACTIONS(1486),
    [anon_sym_if] = ACTIONS(1486),
    [anon_sym_impl] = ACTIONS(1486),
    [anon_sym_let] = ACTIONS(1486),
    [anon_sym_loop] = ACTIONS(1486),
    [anon_sym_match] = ACTIONS(1486),
    [anon_sym_mod] = ACTIONS(1486),
    [anon_sym_pub] = ACTIONS(1486),
    [anon_sym_return] = ACTIONS(1486),
    [anon_sym_static] = ACTIONS(1486),
    [anon_sym_struct] = ACTIONS(1486),
    [anon_sym_trait] = ACTIONS(1486),
    [anon_sym_type] = ACTIONS(1486),
    [anon_sym_union] = ACTIONS(1486),
    [anon_sym_unsafe] = ACTIONS(1486),
    [anon_sym_use] = ACTIONS(1486),
    [anon_sym_while] = ACTIONS(1486),
    [anon_sym_POUND] = ACTIONS(1484),
    [anon_sym_BANG] = ACTIONS(1484),
    [anon_sym_extern] = ACTIONS(1486),
    [anon_sym_LT] = ACTIONS(1484),
    [anon_sym_COLON_COLON] = ACTIONS(1484),
    [anon_sym_AMP] = ACTIONS(1484),
    [anon_sym_DOT_DOT] = ACTIONS(1484),
    [anon_sym_DASH] = ACTIONS(1484),
    [anon_sym_PIPE] = ACTIONS(1484),
    [anon_sym_yield] = ACTIONS(1486),
    [anon_sym_move] = ACTIONS(1486),
    [sym_integer_literal] = ACTIONS(1484),
    [aux_sym_string_literal_token1] = ACTIONS(1484),
    [sym_char_literal] = ACTIONS(1484),
    [anon_sym_true] = ACTIONS(1486),
    [anon_sym_false] = ACTIONS(1486),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1486),
    [sym_super] = ACTIONS(1486),
    [sym_crate] = ACTIONS(1486),
    [sym_metavariable] = ACTIONS(1484),
    [sym_raw_string_literal] = ACTIONS(1484),
    [sym_float_literal] = ACTIONS(1484),
    [sym_block_comment] = ACTIONS(3),
  },
  [356] = {
    [ts_builtin_sym_end] = ACTIONS(1488),
    [sym_identifier] = ACTIONS(1490),
    [anon_sym_SEMI] = ACTIONS(1488),
    [anon_sym_macro_rules_BANG] = ACTIONS(1488),
    [anon_sym_LPAREN] = ACTIONS(1488),
    [anon_sym_LBRACE] = ACTIONS(1488),
    [anon_sym_RBRACE] = ACTIONS(1488),
    [anon_sym_LBRACK] = ACTIONS(1488),
    [anon_sym_STAR] = ACTIONS(1488),
    [anon_sym_u8] = ACTIONS(1490),
    [anon_sym_i8] = ACTIONS(1490),
    [anon_sym_u16] = ACTIONS(1490),
    [anon_sym_i16] = ACTIONS(1490),
    [anon_sym_u32] = ACTIONS(1490),
    [anon_sym_i32] = ACTIONS(1490),
    [anon_sym_u64] = ACTIONS(1490),
    [anon_sym_i64] = ACTIONS(1490),
    [anon_sym_u128] = ACTIONS(1490),
    [anon_sym_i128] = ACTIONS(1490),
    [anon_sym_isize] = ACTIONS(1490),
    [anon_sym_usize] = ACTIONS(1490),
    [anon_sym_f32] = ACTIONS(1490),
    [anon_sym_f64] = ACTIONS(1490),
    [anon_sym_bool] = ACTIONS(1490),
    [anon_sym_str] = ACTIONS(1490),
    [anon_sym_char] = ACTIONS(1490),
    [anon_sym_SQUOTE] = ACTIONS(1490),
    [anon_sym_async] = ACTIONS(1490),
    [anon_sym_break] = ACTIONS(1490),
    [anon_sym_const] = ACTIONS(1490),
    [anon_sym_continue] = ACTIONS(1490),
    [anon_sym_default] = ACTIONS(1490),
    [anon_sym_enum] = ACTIONS(1490),
    [anon_sym_fn] = ACTIONS(1490),
    [anon_sym_for] = ACTIONS(1490),
    [anon_sym_if] = ACTIONS(1490),
    [anon_sym_impl] = ACTIONS(1490),
    [anon_sym_let] = ACTIONS(1490),
    [anon_sym_loop] = ACTIONS(1490),
    [anon_sym_match] = ACTIONS(1490),
    [anon_sym_mod] = ACTIONS(1490),
    [anon_sym_pub] = ACTIONS(1490),
    [anon_sym_return] = ACTIONS(1490),
    [anon_sym_static] = ACTIONS(1490),
    [anon_sym_struct] = ACTIONS(1490),
    [anon_sym_trait] = ACTIONS(1490),
    [anon_sym_type] = ACTIONS(1490),
    [anon_sym_union] = ACTIONS(1490),
    [anon_sym_unsafe] = ACTIONS(1490),
    [anon_sym_use] = ACTIONS(1490),
    [anon_sym_while] = ACTIONS(1490),
    [anon_sym_POUND] = ACTIONS(1488),
    [anon_sym_BANG] = ACTIONS(1488),
    [anon_sym_extern] = ACTIONS(1490),
    [anon_sym_LT] = ACTIONS(1488),
    [anon_sym_COLON_COLON] = ACTIONS(1488),
    [anon_sym_AMP] = ACTIONS(1488),
    [anon_sym_DOT_DOT] = ACTIONS(1488),
    [anon_sym_DASH] = ACTIONS(1488),
    [anon_sym_PIPE] = ACTIONS(1488),
    [anon_sym_yield] = ACTIONS(1490),
    [anon_sym_move] = ACTIONS(1490),
    [sym_integer_literal] = ACTIONS(1488),
    [aux_sym_string_literal_token1] = ACTIONS(1488),
    [sym_char_literal] = ACTIONS(1488),
    [anon_sym_true] = ACTIONS(1490),
    [anon_sym_false] = ACTIONS(1490),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1490),
    [sym_super] = ACTIONS(1490),
    [sym_crate] = ACTIONS(1490),
    [sym_metavariable] = ACTIONS(1488),
    [sym_raw_string_literal] = ACTIONS(1488),
    [sym_float_literal] = ACTIONS(1488),
    [sym_block_comment] = ACTIONS(3),
  },
  [357] = {
    [ts_builtin_sym_end] = ACTIONS(1492),
    [sym_identifier] = ACTIONS(1494),
    [anon_sym_SEMI] = ACTIONS(1492),
    [anon_sym_macro_rules_BANG] = ACTIONS(1492),
    [anon_sym_LPAREN] = ACTIONS(1492),
    [anon_sym_LBRACE] = ACTIONS(1492),
    [anon_sym_RBRACE] = ACTIONS(1492),
    [anon_sym_LBRACK] = ACTIONS(1492),
    [anon_sym_STAR] = ACTIONS(1492),
    [anon_sym_u8] = ACTIONS(1494),
    [anon_sym_i8] = ACTIONS(1494),
    [anon_sym_u16] = ACTIONS(1494),
    [anon_sym_i16] = ACTIONS(1494),
    [anon_sym_u32] = ACTIONS(1494),
    [anon_sym_i32] = ACTIONS(1494),
    [anon_sym_u64] = ACTIONS(1494),
    [anon_sym_i64] = ACTIONS(1494),
    [anon_sym_u128] = ACTIONS(1494),
    [anon_sym_i128] = ACTIONS(1494),
    [anon_sym_isize] = ACTIONS(1494),
    [anon_sym_usize] = ACTIONS(1494),
    [anon_sym_f32] = ACTIONS(1494),
    [anon_sym_f64] = ACTIONS(1494),
    [anon_sym_bool] = ACTIONS(1494),
    [anon_sym_str] = ACTIONS(1494),
    [anon_sym_char] = ACTIONS(1494),
    [anon_sym_SQUOTE] = ACTIONS(1494),
    [anon_sym_async] = ACTIONS(1494),
    [anon_sym_break] = ACTIONS(1494),
    [anon_sym_const] = ACTIONS(1494),
    [anon_sym_continue] = ACTIONS(1494),
    [anon_sym_default] = ACTIONS(1494),
    [anon_sym_enum] = ACTIONS(1494),
    [anon_sym_fn] = ACTIONS(1494),
    [anon_sym_for] = ACTIONS(1494),
    [anon_sym_if] = ACTIONS(1494),
    [anon_sym_impl] = ACTIONS(1494),
    [anon_sym_let] = ACTIONS(1494),
    [anon_sym_loop] = ACTIONS(1494),
    [anon_sym_match] = ACTIONS(1494),
    [anon_sym_mod] = ACTIONS(1494),
    [anon_sym_pub] = ACTIONS(1494),
    [anon_sym_return] = ACTIONS(1494),
    [anon_sym_static] = ACTIONS(1494),
    [anon_sym_struct] = ACTIONS(1494),
    [anon_sym_trait] = ACTIONS(1494),
    [anon_sym_type] = ACTIONS(1494),
    [anon_sym_union] = ACTIONS(1494),
    [anon_sym_unsafe] = ACTIONS(1494),
    [anon_sym_use] = ACTIONS(1494),
    [anon_sym_while] = ACTIONS(1494),
    [anon_sym_POUND] = ACTIONS(1492),
    [anon_sym_BANG] = ACTIONS(1492),
    [anon_sym_extern] = ACTIONS(1494),
    [anon_sym_LT] = ACTIONS(1492),
    [anon_sym_COLON_COLON] = ACTIONS(1492),
    [anon_sym_AMP] = ACTIONS(1492),
    [anon_sym_DOT_DOT] = ACTIONS(1492),
    [anon_sym_DASH] = ACTIONS(1492),
    [anon_sym_PIPE] = ACTIONS(1492),
    [anon_sym_yield] = ACTIONS(1494),
    [anon_sym_move] = ACTIONS(1494),
    [sym_integer_literal] = ACTIONS(1492),
    [aux_sym_string_literal_token1] = ACTIONS(1492),
    [sym_char_literal] = ACTIONS(1492),
    [anon_sym_true] = ACTIONS(1494),
    [anon_sym_false] = ACTIONS(1494),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1494),
    [sym_super] = ACTIONS(1494),
    [sym_crate] = ACTIONS(1494),
    [sym_metavariable] = ACTIONS(1492),
    [sym_raw_string_literal] = ACTIONS(1492),
    [sym_float_literal] = ACTIONS(1492),
    [sym_block_comment] = ACTIONS(3),
  },
  [358] = {
    [ts_builtin_sym_end] = ACTIONS(1496),
    [sym_identifier] = ACTIONS(1498),
    [anon_sym_SEMI] = ACTIONS(1496),
    [anon_sym_macro_rules_BANG] = ACTIONS(1496),
    [anon_sym_LPAREN] = ACTIONS(1496),
    [anon_sym_LBRACE] = ACTIONS(1496),
    [anon_sym_RBRACE] = ACTIONS(1496),
    [anon_sym_LBRACK] = ACTIONS(1496),
    [anon_sym_STAR] = ACTIONS(1496),
    [anon_sym_u8] = ACTIONS(1498),
    [anon_sym_i8] = ACTIONS(1498),
    [anon_sym_u16] = ACTIONS(1498),
    [anon_sym_i16] = ACTIONS(1498),
    [anon_sym_u32] = ACTIONS(1498),
    [anon_sym_i32] = ACTIONS(1498),
    [anon_sym_u64] = ACTIONS(1498),
    [anon_sym_i64] = ACTIONS(1498),
    [anon_sym_u128] = ACTIONS(1498),
    [anon_sym_i128] = ACTIONS(1498),
    [anon_sym_isize] = ACTIONS(1498),
    [anon_sym_usize] = ACTIONS(1498),
    [anon_sym_f32] = ACTIONS(1498),
    [anon_sym_f64] = ACTIONS(1498),
    [anon_sym_bool] = ACTIONS(1498),
    [anon_sym_str] = ACTIONS(1498),
    [anon_sym_char] = ACTIONS(1498),
    [anon_sym_SQUOTE] = ACTIONS(1498),
    [anon_sym_async] = ACTIONS(1498),
    [anon_sym_break] = ACTIONS(1498),
    [anon_sym_const] = ACTIONS(1498),
    [anon_sym_continue] = ACTIONS(1498),
    [anon_sym_default] = ACTIONS(1498),
    [anon_sym_enum] = ACTIONS(1498),
    [anon_sym_fn] = ACTIONS(1498),
    [anon_sym_for] = ACTIONS(1498),
    [anon_sym_if] = ACTIONS(1498),
    [anon_sym_impl] = ACTIONS(1498),
    [anon_sym_let] = ACTIONS(1498),
    [anon_sym_loop] = ACTIONS(1498),
    [anon_sym_match] = ACTIONS(1498),
    [anon_sym_mod] = ACTIONS(1498),
    [anon_sym_pub] = ACTIONS(1498),
    [anon_sym_return] = ACTIONS(1498),
    [anon_sym_static] = ACTIONS(1498),
    [anon_sym_struct] = ACTIONS(1498),
    [anon_sym_trait] = ACTIONS(1498),
    [anon_sym_type] = ACTIONS(1498),
    [anon_sym_union] = ACTIONS(1498),
    [anon_sym_unsafe] = ACTIONS(1498),
    [anon_sym_use] = ACTIONS(1498),
    [anon_sym_while] = ACTIONS(1498),
    [anon_sym_POUND] = ACTIONS(1496),
    [anon_sym_BANG] = ACTIONS(1496),
    [anon_sym_extern] = ACTIONS(1498),
    [anon_sym_LT] = ACTIONS(1496),
    [anon_sym_COLON_COLON] = ACTIONS(1496),
    [anon_sym_AMP] = ACTIONS(1496),
    [anon_sym_DOT_DOT] = ACTIONS(1496),
    [anon_sym_DASH] = ACTIONS(1496),
    [anon_sym_PIPE] = ACTIONS(1496),
    [anon_sym_yield] = ACTIONS(1498),
    [anon_sym_move] = ACTIONS(1498),
    [sym_integer_literal] = ACTIONS(1496),
    [aux_sym_string_literal_token1] = ACTIONS(1496),
    [sym_char_literal] = ACTIONS(1496),
    [anon_sym_true] = ACTIONS(1498),
    [anon_sym_false] = ACTIONS(1498),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1498),
    [sym_super] = ACTIONS(1498),
    [sym_crate] = ACTIONS(1498),
    [sym_metavariable] = ACTIONS(1496),
    [sym_raw_string_literal] = ACTIONS(1496),
    [sym_float_literal] = ACTIONS(1496),
    [sym_block_comment] = ACTIONS(3),
  },
  [359] = {
    [ts_builtin_sym_end] = ACTIONS(1500),
    [sym_identifier] = ACTIONS(1502),
    [anon_sym_SEMI] = ACTIONS(1500),
    [anon_sym_macro_rules_BANG] = ACTIONS(1500),
    [anon_sym_LPAREN] = ACTIONS(1500),
    [anon_sym_LBRACE] = ACTIONS(1500),
    [anon_sym_RBRACE] = ACTIONS(1500),
    [anon_sym_LBRACK] = ACTIONS(1500),
    [anon_sym_STAR] = ACTIONS(1500),
    [anon_sym_u8] = ACTIONS(1502),
    [anon_sym_i8] = ACTIONS(1502),
    [anon_sym_u16] = ACTIONS(1502),
    [anon_sym_i16] = ACTIONS(1502),
    [anon_sym_u32] = ACTIONS(1502),
    [anon_sym_i32] = ACTIONS(1502),
    [anon_sym_u64] = ACTIONS(1502),
    [anon_sym_i64] = ACTIONS(1502),
    [anon_sym_u128] = ACTIONS(1502),
    [anon_sym_i128] = ACTIONS(1502),
    [anon_sym_isize] = ACTIONS(1502),
    [anon_sym_usize] = ACTIONS(1502),
    [anon_sym_f32] = ACTIONS(1502),
    [anon_sym_f64] = ACTIONS(1502),
    [anon_sym_bool] = ACTIONS(1502),
    [anon_sym_str] = ACTIONS(1502),
    [anon_sym_char] = ACTIONS(1502),
    [anon_sym_SQUOTE] = ACTIONS(1502),
    [anon_sym_async] = ACTIONS(1502),
    [anon_sym_break] = ACTIONS(1502),
    [anon_sym_const] = ACTIONS(1502),
    [anon_sym_continue] = ACTIONS(1502),
    [anon_sym_default] = ACTIONS(1502),
    [anon_sym_enum] = ACTIONS(1502),
    [anon_sym_fn] = ACTIONS(1502),
    [anon_sym_for] = ACTIONS(1502),
    [anon_sym_if] = ACTIONS(1502),
    [anon_sym_impl] = ACTIONS(1502),
    [anon_sym_let] = ACTIONS(1502),
    [anon_sym_loop] = ACTIONS(1502),
    [anon_sym_match] = ACTIONS(1502),
    [anon_sym_mod] = ACTIONS(1502),
    [anon_sym_pub] = ACTIONS(1502),
    [anon_sym_return] = ACTIONS(1502),
    [anon_sym_static] = ACTIONS(1502),
    [anon_sym_struct] = ACTIONS(1502),
    [anon_sym_trait] = ACTIONS(1502),
    [anon_sym_type] = ACTIONS(1502),
    [anon_sym_union] = ACTIONS(1502),
    [anon_sym_unsafe] = ACTIONS(1502),
    [anon_sym_use] = ACTIONS(1502),
    [anon_sym_while] = ACTIONS(1502),
    [anon_sym_POUND] = ACTIONS(1500),
    [anon_sym_BANG] = ACTIONS(1500),
    [anon_sym_extern] = ACTIONS(1502),
    [anon_sym_LT] = ACTIONS(1500),
    [anon_sym_COLON_COLON] = ACTIONS(1500),
    [anon_sym_AMP] = ACTIONS(1500),
    [anon_sym_DOT_DOT] = ACTIONS(1500),
    [anon_sym_DASH] = ACTIONS(1500),
    [anon_sym_PIPE] = ACTIONS(1500),
    [anon_sym_yield] = ACTIONS(1502),
    [anon_sym_move] = ACTIONS(1502),
    [sym_integer_literal] = ACTIONS(1500),
    [aux_sym_string_literal_token1] = ACTIONS(1500),
    [sym_char_literal] = ACTIONS(1500),
    [anon_sym_true] = ACTIONS(1502),
    [anon_sym_false] = ACTIONS(1502),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1502),
    [sym_super] = ACTIONS(1502),
    [sym_crate] = ACTIONS(1502),
    [sym_metavariable] = ACTIONS(1500),
    [sym_raw_string_literal] = ACTIONS(1500),
    [sym_float_literal] = ACTIONS(1500),
    [sym_block_comment] = ACTIONS(3),
  },
  [360] = {
    [ts_builtin_sym_end] = ACTIONS(1504),
    [sym_identifier] = ACTIONS(1506),
    [anon_sym_SEMI] = ACTIONS(1504),
    [anon_sym_macro_rules_BANG] = ACTIONS(1504),
    [anon_sym_LPAREN] = ACTIONS(1504),
    [anon_sym_LBRACE] = ACTIONS(1504),
    [anon_sym_RBRACE] = ACTIONS(1504),
    [anon_sym_LBRACK] = ACTIONS(1504),
    [anon_sym_STAR] = ACTIONS(1504),
    [anon_sym_u8] = ACTIONS(1506),
    [anon_sym_i8] = ACTIONS(1506),
    [anon_sym_u16] = ACTIONS(1506),
    [anon_sym_i16] = ACTIONS(1506),
    [anon_sym_u32] = ACTIONS(1506),
    [anon_sym_i32] = ACTIONS(1506),
    [anon_sym_u64] = ACTIONS(1506),
    [anon_sym_i64] = ACTIONS(1506),
    [anon_sym_u128] = ACTIONS(1506),
    [anon_sym_i128] = ACTIONS(1506),
    [anon_sym_isize] = ACTIONS(1506),
    [anon_sym_usize] = ACTIONS(1506),
    [anon_sym_f32] = ACTIONS(1506),
    [anon_sym_f64] = ACTIONS(1506),
    [anon_sym_bool] = ACTIONS(1506),
    [anon_sym_str] = ACTIONS(1506),
    [anon_sym_char] = ACTIONS(1506),
    [anon_sym_SQUOTE] = ACTIONS(1506),
    [anon_sym_async] = ACTIONS(1506),
    [anon_sym_break] = ACTIONS(1506),
    [anon_sym_const] = ACTIONS(1506),
    [anon_sym_continue] = ACTIONS(1506),
    [anon_sym_default] = ACTIONS(1506),
    [anon_sym_enum] = ACTIONS(1506),
    [anon_sym_fn] = ACTIONS(1506),
    [anon_sym_for] = ACTIONS(1506),
    [anon_sym_if] = ACTIONS(1506),
    [anon_sym_impl] = ACTIONS(1506),
    [anon_sym_let] = ACTIONS(1506),
    [anon_sym_loop] = ACTIONS(1506),
    [anon_sym_match] = ACTIONS(1506),
    [anon_sym_mod] = ACTIONS(1506),
    [anon_sym_pub] = ACTIONS(1506),
    [anon_sym_return] = ACTIONS(1506),
    [anon_sym_static] = ACTIONS(1506),
    [anon_sym_struct] = ACTIONS(1506),
    [anon_sym_trait] = ACTIONS(1506),
    [anon_sym_type] = ACTIONS(1506),
    [anon_sym_union] = ACTIONS(1506),
    [anon_sym_unsafe] = ACTIONS(1506),
    [anon_sym_use] = ACTIONS(1506),
    [anon_sym_while] = ACTIONS(1506),
    [anon_sym_POUND] = ACTIONS(1504),
    [anon_sym_BANG] = ACTIONS(1504),
    [anon_sym_extern] = ACTIONS(1506),
    [anon_sym_LT] = ACTIONS(1504),
    [anon_sym_COLON_COLON] = ACTIONS(1504),
    [anon_sym_AMP] = ACTIONS(1504),
    [anon_sym_DOT_DOT] = ACTIONS(1504),
    [anon_sym_DASH] = ACTIONS(1504),
    [anon_sym_PIPE] = ACTIONS(1504),
    [anon_sym_yield] = ACTIONS(1506),
    [anon_sym_move] = ACTIONS(1506),
    [sym_integer_literal] = ACTIONS(1504),
    [aux_sym_string_literal_token1] = ACTIONS(1504),
    [sym_char_literal] = ACTIONS(1504),
    [anon_sym_true] = ACTIONS(1506),
    [anon_sym_false] = ACTIONS(1506),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1506),
    [sym_super] = ACTIONS(1506),
    [sym_crate] = ACTIONS(1506),
    [sym_metavariable] = ACTIONS(1504),
    [sym_raw_string_literal] = ACTIONS(1504),
    [sym_float_literal] = ACTIONS(1504),
    [sym_block_comment] = ACTIONS(3),
  },
  [361] = {
    [ts_builtin_sym_end] = ACTIONS(1508),
    [sym_identifier] = ACTIONS(1510),
    [anon_sym_SEMI] = ACTIONS(1508),
    [anon_sym_macro_rules_BANG] = ACTIONS(1508),
    [anon_sym_LPAREN] = ACTIONS(1508),
    [anon_sym_LBRACE] = ACTIONS(1508),
    [anon_sym_RBRACE] = ACTIONS(1508),
    [anon_sym_LBRACK] = ACTIONS(1508),
    [anon_sym_STAR] = ACTIONS(1508),
    [anon_sym_u8] = ACTIONS(1510),
    [anon_sym_i8] = ACTIONS(1510),
    [anon_sym_u16] = ACTIONS(1510),
    [anon_sym_i16] = ACTIONS(1510),
    [anon_sym_u32] = ACTIONS(1510),
    [anon_sym_i32] = ACTIONS(1510),
    [anon_sym_u64] = ACTIONS(1510),
    [anon_sym_i64] = ACTIONS(1510),
    [anon_sym_u128] = ACTIONS(1510),
    [anon_sym_i128] = ACTIONS(1510),
    [anon_sym_isize] = ACTIONS(1510),
    [anon_sym_usize] = ACTIONS(1510),
    [anon_sym_f32] = ACTIONS(1510),
    [anon_sym_f64] = ACTIONS(1510),
    [anon_sym_bool] = ACTIONS(1510),
    [anon_sym_str] = ACTIONS(1510),
    [anon_sym_char] = ACTIONS(1510),
    [anon_sym_SQUOTE] = ACTIONS(1510),
    [anon_sym_async] = ACTIONS(1510),
    [anon_sym_break] = ACTIONS(1510),
    [anon_sym_const] = ACTIONS(1510),
    [anon_sym_continue] = ACTIONS(1510),
    [anon_sym_default] = ACTIONS(1510),
    [anon_sym_enum] = ACTIONS(1510),
    [anon_sym_fn] = ACTIONS(1510),
    [anon_sym_for] = ACTIONS(1510),
    [anon_sym_if] = ACTIONS(1510),
    [anon_sym_impl] = ACTIONS(1510),
    [anon_sym_let] = ACTIONS(1510),
    [anon_sym_loop] = ACTIONS(1510),
    [anon_sym_match] = ACTIONS(1510),
    [anon_sym_mod] = ACTIONS(1510),
    [anon_sym_pub] = ACTIONS(1510),
    [anon_sym_return] = ACTIONS(1510),
    [anon_sym_static] = ACTIONS(1510),
    [anon_sym_struct] = ACTIONS(1510),
    [anon_sym_trait] = ACTIONS(1510),
    [anon_sym_type] = ACTIONS(1510),
    [anon_sym_union] = ACTIONS(1510),
    [anon_sym_unsafe] = ACTIONS(1510),
    [anon_sym_use] = ACTIONS(1510),
    [anon_sym_while] = ACTIONS(1510),
    [anon_sym_POUND] = ACTIONS(1508),
    [anon_sym_BANG] = ACTIONS(1508),
    [anon_sym_extern] = ACTIONS(1510),
    [anon_sym_LT] = ACTIONS(1508),
    [anon_sym_COLON_COLON] = ACTIONS(1508),
    [anon_sym_AMP] = ACTIONS(1508),
    [anon_sym_DOT_DOT] = ACTIONS(1508),
    [anon_sym_DASH] = ACTIONS(1508),
    [anon_sym_PIPE] = ACTIONS(1508),
    [anon_sym_yield] = ACTIONS(1510),
    [anon_sym_move] = ACTIONS(1510),
    [sym_integer_literal] = ACTIONS(1508),
    [aux_sym_string_literal_token1] = ACTIONS(1508),
    [sym_char_literal] = ACTIONS(1508),
    [anon_sym_true] = ACTIONS(1510),
    [anon_sym_false] = ACTIONS(1510),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1510),
    [sym_super] = ACTIONS(1510),
    [sym_crate] = ACTIONS(1510),
    [sym_metavariable] = ACTIONS(1508),
    [sym_raw_string_literal] = ACTIONS(1508),
    [sym_float_literal] = ACTIONS(1508),
    [sym_block_comment] = ACTIONS(3),
  },
  [362] = {
    [ts_builtin_sym_end] = ACTIONS(1512),
    [sym_identifier] = ACTIONS(1514),
    [anon_sym_SEMI] = ACTIONS(1512),
    [anon_sym_macro_rules_BANG] = ACTIONS(1512),
    [anon_sym_LPAREN] = ACTIONS(1512),
    [anon_sym_LBRACE] = ACTIONS(1512),
    [anon_sym_RBRACE] = ACTIONS(1512),
    [anon_sym_LBRACK] = ACTIONS(1512),
    [anon_sym_STAR] = ACTIONS(1512),
    [anon_sym_u8] = ACTIONS(1514),
    [anon_sym_i8] = ACTIONS(1514),
    [anon_sym_u16] = ACTIONS(1514),
    [anon_sym_i16] = ACTIONS(1514),
    [anon_sym_u32] = ACTIONS(1514),
    [anon_sym_i32] = ACTIONS(1514),
    [anon_sym_u64] = ACTIONS(1514),
    [anon_sym_i64] = ACTIONS(1514),
    [anon_sym_u128] = ACTIONS(1514),
    [anon_sym_i128] = ACTIONS(1514),
    [anon_sym_isize] = ACTIONS(1514),
    [anon_sym_usize] = ACTIONS(1514),
    [anon_sym_f32] = ACTIONS(1514),
    [anon_sym_f64] = ACTIONS(1514),
    [anon_sym_bool] = ACTIONS(1514),
    [anon_sym_str] = ACTIONS(1514),
    [anon_sym_char] = ACTIONS(1514),
    [anon_sym_SQUOTE] = ACTIONS(1514),
    [anon_sym_async] = ACTIONS(1514),
    [anon_sym_break] = ACTIONS(1514),
    [anon_sym_const] = ACTIONS(1514),
    [anon_sym_continue] = ACTIONS(1514),
    [anon_sym_default] = ACTIONS(1514),
    [anon_sym_enum] = ACTIONS(1514),
    [anon_sym_fn] = ACTIONS(1514),
    [anon_sym_for] = ACTIONS(1514),
    [anon_sym_if] = ACTIONS(1514),
    [anon_sym_impl] = ACTIONS(1514),
    [anon_sym_let] = ACTIONS(1514),
    [anon_sym_loop] = ACTIONS(1514),
    [anon_sym_match] = ACTIONS(1514),
    [anon_sym_mod] = ACTIONS(1514),
    [anon_sym_pub] = ACTIONS(1514),
    [anon_sym_return] = ACTIONS(1514),
    [anon_sym_static] = ACTIONS(1514),
    [anon_sym_struct] = ACTIONS(1514),
    [anon_sym_trait] = ACTIONS(1514),
    [anon_sym_type] = ACTIONS(1514),
    [anon_sym_union] = ACTIONS(1514),
    [anon_sym_unsafe] = ACTIONS(1514),
    [anon_sym_use] = ACTIONS(1514),
    [anon_sym_while] = ACTIONS(1514),
    [anon_sym_POUND] = ACTIONS(1512),
    [anon_sym_BANG] = ACTIONS(1512),
    [anon_sym_extern] = ACTIONS(1514),
    [anon_sym_LT] = ACTIONS(1512),
    [anon_sym_COLON_COLON] = ACTIONS(1512),
    [anon_sym_AMP] = ACTIONS(1512),
    [anon_sym_DOT_DOT] = ACTIONS(1512),
    [anon_sym_DASH] = ACTIONS(1512),
    [anon_sym_PIPE] = ACTIONS(1512),
    [anon_sym_yield] = ACTIONS(1514),
    [anon_sym_move] = ACTIONS(1514),
    [sym_integer_literal] = ACTIONS(1512),
    [aux_sym_string_literal_token1] = ACTIONS(1512),
    [sym_char_literal] = ACTIONS(1512),
    [anon_sym_true] = ACTIONS(1514),
    [anon_sym_false] = ACTIONS(1514),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1514),
    [sym_super] = ACTIONS(1514),
    [sym_crate] = ACTIONS(1514),
    [sym_metavariable] = ACTIONS(1512),
    [sym_raw_string_literal] = ACTIONS(1512),
    [sym_float_literal] = ACTIONS(1512),
    [sym_block_comment] = ACTIONS(3),
  },
  [363] = {
    [ts_builtin_sym_end] = ACTIONS(1516),
    [sym_identifier] = ACTIONS(1518),
    [anon_sym_SEMI] = ACTIONS(1516),
    [anon_sym_macro_rules_BANG] = ACTIONS(1516),
    [anon_sym_LPAREN] = ACTIONS(1516),
    [anon_sym_LBRACE] = ACTIONS(1516),
    [anon_sym_RBRACE] = ACTIONS(1516),
    [anon_sym_LBRACK] = ACTIONS(1516),
    [anon_sym_STAR] = ACTIONS(1516),
    [anon_sym_u8] = ACTIONS(1518),
    [anon_sym_i8] = ACTIONS(1518),
    [anon_sym_u16] = ACTIONS(1518),
    [anon_sym_i16] = ACTIONS(1518),
    [anon_sym_u32] = ACTIONS(1518),
    [anon_sym_i32] = ACTIONS(1518),
    [anon_sym_u64] = ACTIONS(1518),
    [anon_sym_i64] = ACTIONS(1518),
    [anon_sym_u128] = ACTIONS(1518),
    [anon_sym_i128] = ACTIONS(1518),
    [anon_sym_isize] = ACTIONS(1518),
    [anon_sym_usize] = ACTIONS(1518),
    [anon_sym_f32] = ACTIONS(1518),
    [anon_sym_f64] = ACTIONS(1518),
    [anon_sym_bool] = ACTIONS(1518),
    [anon_sym_str] = ACTIONS(1518),
    [anon_sym_char] = ACTIONS(1518),
    [anon_sym_SQUOTE] = ACTIONS(1518),
    [anon_sym_async] = ACTIONS(1518),
    [anon_sym_break] = ACTIONS(1518),
    [anon_sym_const] = ACTIONS(1518),
    [anon_sym_continue] = ACTIONS(1518),
    [anon_sym_default] = ACTIONS(1518),
    [anon_sym_enum] = ACTIONS(1518),
    [anon_sym_fn] = ACTIONS(1518),
    [anon_sym_for] = ACTIONS(1518),
    [anon_sym_if] = ACTIONS(1518),
    [anon_sym_impl] = ACTIONS(1518),
    [anon_sym_let] = ACTIONS(1518),
    [anon_sym_loop] = ACTIONS(1518),
    [anon_sym_match] = ACTIONS(1518),
    [anon_sym_mod] = ACTIONS(1518),
    [anon_sym_pub] = ACTIONS(1518),
    [anon_sym_return] = ACTIONS(1518),
    [anon_sym_static] = ACTIONS(1518),
    [anon_sym_struct] = ACTIONS(1518),
    [anon_sym_trait] = ACTIONS(1518),
    [anon_sym_type] = ACTIONS(1518),
    [anon_sym_union] = ACTIONS(1518),
    [anon_sym_unsafe] = ACTIONS(1518),
    [anon_sym_use] = ACTIONS(1518),
    [anon_sym_while] = ACTIONS(1518),
    [anon_sym_POUND] = ACTIONS(1516),
    [anon_sym_BANG] = ACTIONS(1516),
    [anon_sym_extern] = ACTIONS(1518),
    [anon_sym_LT] = ACTIONS(1516),
    [anon_sym_COLON_COLON] = ACTIONS(1516),
    [anon_sym_AMP] = ACTIONS(1516),
    [anon_sym_DOT_DOT] = ACTIONS(1516),
    [anon_sym_DASH] = ACTIONS(1516),
    [anon_sym_PIPE] = ACTIONS(1516),
    [anon_sym_yield] = ACTIONS(1518),
    [anon_sym_move] = ACTIONS(1518),
    [sym_integer_literal] = ACTIONS(1516),
    [aux_sym_string_literal_token1] = ACTIONS(1516),
    [sym_char_literal] = ACTIONS(1516),
    [anon_sym_true] = ACTIONS(1518),
    [anon_sym_false] = ACTIONS(1518),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1518),
    [sym_super] = ACTIONS(1518),
    [sym_crate] = ACTIONS(1518),
    [sym_metavariable] = ACTIONS(1516),
    [sym_raw_string_literal] = ACTIONS(1516),
    [sym_float_literal] = ACTIONS(1516),
    [sym_block_comment] = ACTIONS(3),
  },
  [364] = {
    [ts_builtin_sym_end] = ACTIONS(1520),
    [sym_identifier] = ACTIONS(1522),
    [anon_sym_SEMI] = ACTIONS(1520),
    [anon_sym_macro_rules_BANG] = ACTIONS(1520),
    [anon_sym_LPAREN] = ACTIONS(1520),
    [anon_sym_LBRACE] = ACTIONS(1520),
    [anon_sym_RBRACE] = ACTIONS(1520),
    [anon_sym_LBRACK] = ACTIONS(1520),
    [anon_sym_STAR] = ACTIONS(1520),
    [anon_sym_u8] = ACTIONS(1522),
    [anon_sym_i8] = ACTIONS(1522),
    [anon_sym_u16] = ACTIONS(1522),
    [anon_sym_i16] = ACTIONS(1522),
    [anon_sym_u32] = ACTIONS(1522),
    [anon_sym_i32] = ACTIONS(1522),
    [anon_sym_u64] = ACTIONS(1522),
    [anon_sym_i64] = ACTIONS(1522),
    [anon_sym_u128] = ACTIONS(1522),
    [anon_sym_i128] = ACTIONS(1522),
    [anon_sym_isize] = ACTIONS(1522),
    [anon_sym_usize] = ACTIONS(1522),
    [anon_sym_f32] = ACTIONS(1522),
    [anon_sym_f64] = ACTIONS(1522),
    [anon_sym_bool] = ACTIONS(1522),
    [anon_sym_str] = ACTIONS(1522),
    [anon_sym_char] = ACTIONS(1522),
    [anon_sym_SQUOTE] = ACTIONS(1522),
    [anon_sym_async] = ACTIONS(1522),
    [anon_sym_break] = ACTIONS(1522),
    [anon_sym_const] = ACTIONS(1522),
    [anon_sym_continue] = ACTIONS(1522),
    [anon_sym_default] = ACTIONS(1522),
    [anon_sym_enum] = ACTIONS(1522),
    [anon_sym_fn] = ACTIONS(1522),
    [anon_sym_for] = ACTIONS(1522),
    [anon_sym_if] = ACTIONS(1522),
    [anon_sym_impl] = ACTIONS(1522),
    [anon_sym_let] = ACTIONS(1522),
    [anon_sym_loop] = ACTIONS(1522),
    [anon_sym_match] = ACTIONS(1522),
    [anon_sym_mod] = ACTIONS(1522),
    [anon_sym_pub] = ACTIONS(1522),
    [anon_sym_return] = ACTIONS(1522),
    [anon_sym_static] = ACTIONS(1522),
    [anon_sym_struct] = ACTIONS(1522),
    [anon_sym_trait] = ACTIONS(1522),
    [anon_sym_type] = ACTIONS(1522),
    [anon_sym_union] = ACTIONS(1522),
    [anon_sym_unsafe] = ACTIONS(1522),
    [anon_sym_use] = ACTIONS(1522),
    [anon_sym_while] = ACTIONS(1522),
    [anon_sym_POUND] = ACTIONS(1520),
    [anon_sym_BANG] = ACTIONS(1520),
    [anon_sym_extern] = ACTIONS(1522),
    [anon_sym_LT] = ACTIONS(1520),
    [anon_sym_COLON_COLON] = ACTIONS(1520),
    [anon_sym_AMP] = ACTIONS(1520),
    [anon_sym_DOT_DOT] = ACTIONS(1520),
    [anon_sym_DASH] = ACTIONS(1520),
    [anon_sym_PIPE] = ACTIONS(1520),
    [anon_sym_yield] = ACTIONS(1522),
    [anon_sym_move] = ACTIONS(1522),
    [sym_integer_literal] = ACTIONS(1520),
    [aux_sym_string_literal_token1] = ACTIONS(1520),
    [sym_char_literal] = ACTIONS(1520),
    [anon_sym_true] = ACTIONS(1522),
    [anon_sym_false] = ACTIONS(1522),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1522),
    [sym_super] = ACTIONS(1522),
    [sym_crate] = ACTIONS(1522),
    [sym_metavariable] = ACTIONS(1520),
    [sym_raw_string_literal] = ACTIONS(1520),
    [sym_float_literal] = ACTIONS(1520),
    [sym_block_comment] = ACTIONS(3),
  },
  [365] = {
    [sym_attribute_item] = STATE(547),
    [sym_bracketed_type] = STATE(2489),
    [sym_generic_type] = STATE(2488),
    [sym_generic_type_with_turbofish] = STATE(2487),
    [sym_macro_invocation] = STATE(1422),
    [sym_scoped_identifier] = STATE(1337),
    [sym_scoped_type_identifier] = STATE(2013),
    [sym_match_arm] = STATE(490),
    [sym_last_match_arm] = STATE(2385),
    [sym_match_pattern] = STATE(2387),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1981),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [aux_sym_enum_variant_list_repeat1] = STATE(547),
    [aux_sym_match_block_repeat1] = STATE(490),
    [sym_identifier] = ACTIONS(1524),
    [anon_sym_LPAREN] = ACTIONS(1526),
    [anon_sym_RBRACE] = ACTIONS(1528),
    [anon_sym_LBRACK] = ACTIONS(1530),
    [anon_sym_u8] = ACTIONS(1532),
    [anon_sym_i8] = ACTIONS(1532),
    [anon_sym_u16] = ACTIONS(1532),
    [anon_sym_i16] = ACTIONS(1532),
    [anon_sym_u32] = ACTIONS(1532),
    [anon_sym_i32] = ACTIONS(1532),
    [anon_sym_u64] = ACTIONS(1532),
    [anon_sym_i64] = ACTIONS(1532),
    [anon_sym_u128] = ACTIONS(1532),
    [anon_sym_i128] = ACTIONS(1532),
    [anon_sym_isize] = ACTIONS(1532),
    [anon_sym_usize] = ACTIONS(1532),
    [anon_sym_f32] = ACTIONS(1532),
    [anon_sym_f64] = ACTIONS(1532),
    [anon_sym_bool] = ACTIONS(1532),
    [anon_sym_str] = ACTIONS(1532),
    [anon_sym_char] = ACTIONS(1532),
    [anon_sym_const] = ACTIONS(1534),
    [anon_sym_default] = ACTIONS(1536),
    [anon_sym_union] = ACTIONS(1536),
    [anon_sym_POUND] = ACTIONS(370),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(1538),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(1540),
    [sym_mutable_specifier] = ACTIONS(796),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1542),
    [sym_super] = ACTIONS(1542),
    [sym_crate] = ACTIONS(1542),
    [sym_metavariable] = ACTIONS(1544),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [366] = {
    [ts_builtin_sym_end] = ACTIONS(1546),
    [sym_identifier] = ACTIONS(1548),
    [anon_sym_SEMI] = ACTIONS(1546),
    [anon_sym_macro_rules_BANG] = ACTIONS(1546),
    [anon_sym_LPAREN] = ACTIONS(1546),
    [anon_sym_LBRACE] = ACTIONS(1546),
    [anon_sym_RBRACE] = ACTIONS(1546),
    [anon_sym_LBRACK] = ACTIONS(1546),
    [anon_sym_STAR] = ACTIONS(1546),
    [anon_sym_u8] = ACTIONS(1548),
    [anon_sym_i8] = ACTIONS(1548),
    [anon_sym_u16] = ACTIONS(1548),
    [anon_sym_i16] = ACTIONS(1548),
    [anon_sym_u32] = ACTIONS(1548),
    [anon_sym_i32] = ACTIONS(1548),
    [anon_sym_u64] = ACTIONS(1548),
    [anon_sym_i64] = ACTIONS(1548),
    [anon_sym_u128] = ACTIONS(1548),
    [anon_sym_i128] = ACTIONS(1548),
    [anon_sym_isize] = ACTIONS(1548),
    [anon_sym_usize] = ACTIONS(1548),
    [anon_sym_f32] = ACTIONS(1548),
    [anon_sym_f64] = ACTIONS(1548),
    [anon_sym_bool] = ACTIONS(1548),
    [anon_sym_str] = ACTIONS(1548),
    [anon_sym_char] = ACTIONS(1548),
    [anon_sym_SQUOTE] = ACTIONS(1548),
    [anon_sym_async] = ACTIONS(1548),
    [anon_sym_break] = ACTIONS(1548),
    [anon_sym_const] = ACTIONS(1548),
    [anon_sym_continue] = ACTIONS(1548),
    [anon_sym_default] = ACTIONS(1548),
    [anon_sym_enum] = ACTIONS(1548),
    [anon_sym_fn] = ACTIONS(1548),
    [anon_sym_for] = ACTIONS(1548),
    [anon_sym_if] = ACTIONS(1548),
    [anon_sym_impl] = ACTIONS(1548),
    [anon_sym_let] = ACTIONS(1548),
    [anon_sym_loop] = ACTIONS(1548),
    [anon_sym_match] = ACTIONS(1548),
    [anon_sym_mod] = ACTIONS(1548),
    [anon_sym_pub] = ACTIONS(1548),
    [anon_sym_return] = ACTIONS(1548),
    [anon_sym_static] = ACTIONS(1548),
    [anon_sym_struct] = ACTIONS(1548),
    [anon_sym_trait] = ACTIONS(1548),
    [anon_sym_type] = ACTIONS(1548),
    [anon_sym_union] = ACTIONS(1548),
    [anon_sym_unsafe] = ACTIONS(1548),
    [anon_sym_use] = ACTIONS(1548),
    [anon_sym_while] = ACTIONS(1548),
    [anon_sym_POUND] = ACTIONS(1546),
    [anon_sym_BANG] = ACTIONS(1546),
    [anon_sym_extern] = ACTIONS(1548),
    [anon_sym_LT] = ACTIONS(1546),
    [anon_sym_COLON_COLON] = ACTIONS(1546),
    [anon_sym_AMP] = ACTIONS(1546),
    [anon_sym_DOT_DOT] = ACTIONS(1546),
    [anon_sym_DASH] = ACTIONS(1546),
    [anon_sym_PIPE] = ACTIONS(1546),
    [anon_sym_yield] = ACTIONS(1548),
    [anon_sym_move] = ACTIONS(1548),
    [sym_integer_literal] = ACTIONS(1546),
    [aux_sym_string_literal_token1] = ACTIONS(1546),
    [sym_char_literal] = ACTIONS(1546),
    [anon_sym_true] = ACTIONS(1548),
    [anon_sym_false] = ACTIONS(1548),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1548),
    [sym_super] = ACTIONS(1548),
    [sym_crate] = ACTIONS(1548),
    [sym_metavariable] = ACTIONS(1546),
    [sym_raw_string_literal] = ACTIONS(1546),
    [sym_float_literal] = ACTIONS(1546),
    [sym_block_comment] = ACTIONS(3),
  },
  [367] = {
    [ts_builtin_sym_end] = ACTIONS(1550),
    [sym_identifier] = ACTIONS(1552),
    [anon_sym_SEMI] = ACTIONS(1550),
    [anon_sym_macro_rules_BANG] = ACTIONS(1550),
    [anon_sym_LPAREN] = ACTIONS(1550),
    [anon_sym_LBRACE] = ACTIONS(1550),
    [anon_sym_RBRACE] = ACTIONS(1550),
    [anon_sym_LBRACK] = ACTIONS(1550),
    [anon_sym_STAR] = ACTIONS(1550),
    [anon_sym_u8] = ACTIONS(1552),
    [anon_sym_i8] = ACTIONS(1552),
    [anon_sym_u16] = ACTIONS(1552),
    [anon_sym_i16] = ACTIONS(1552),
    [anon_sym_u32] = ACTIONS(1552),
    [anon_sym_i32] = ACTIONS(1552),
    [anon_sym_u64] = ACTIONS(1552),
    [anon_sym_i64] = ACTIONS(1552),
    [anon_sym_u128] = ACTIONS(1552),
    [anon_sym_i128] = ACTIONS(1552),
    [anon_sym_isize] = ACTIONS(1552),
    [anon_sym_usize] = ACTIONS(1552),
    [anon_sym_f32] = ACTIONS(1552),
    [anon_sym_f64] = ACTIONS(1552),
    [anon_sym_bool] = ACTIONS(1552),
    [anon_sym_str] = ACTIONS(1552),
    [anon_sym_char] = ACTIONS(1552),
    [anon_sym_SQUOTE] = ACTIONS(1552),
    [anon_sym_async] = ACTIONS(1552),
    [anon_sym_break] = ACTIONS(1552),
    [anon_sym_const] = ACTIONS(1552),
    [anon_sym_continue] = ACTIONS(1552),
    [anon_sym_default] = ACTIONS(1552),
    [anon_sym_enum] = ACTIONS(1552),
    [anon_sym_fn] = ACTIONS(1552),
    [anon_sym_for] = ACTIONS(1552),
    [anon_sym_if] = ACTIONS(1552),
    [anon_sym_impl] = ACTIONS(1552),
    [anon_sym_let] = ACTIONS(1552),
    [anon_sym_loop] = ACTIONS(1552),
    [anon_sym_match] = ACTIONS(1552),
    [anon_sym_mod] = ACTIONS(1552),
    [anon_sym_pub] = ACTIONS(1552),
    [anon_sym_return] = ACTIONS(1552),
    [anon_sym_static] = ACTIONS(1552),
    [anon_sym_struct] = ACTIONS(1552),
    [anon_sym_trait] = ACTIONS(1552),
    [anon_sym_type] = ACTIONS(1552),
    [anon_sym_union] = ACTIONS(1552),
    [anon_sym_unsafe] = ACTIONS(1552),
    [anon_sym_use] = ACTIONS(1552),
    [anon_sym_while] = ACTIONS(1552),
    [anon_sym_POUND] = ACTIONS(1550),
    [anon_sym_BANG] = ACTIONS(1550),
    [anon_sym_extern] = ACTIONS(1552),
    [anon_sym_LT] = ACTIONS(1550),
    [anon_sym_COLON_COLON] = ACTIONS(1550),
    [anon_sym_AMP] = ACTIONS(1550),
    [anon_sym_DOT_DOT] = ACTIONS(1550),
    [anon_sym_DASH] = ACTIONS(1550),
    [anon_sym_PIPE] = ACTIONS(1550),
    [anon_sym_yield] = ACTIONS(1552),
    [anon_sym_move] = ACTIONS(1552),
    [sym_integer_literal] = ACTIONS(1550),
    [aux_sym_string_literal_token1] = ACTIONS(1550),
    [sym_char_literal] = ACTIONS(1550),
    [anon_sym_true] = ACTIONS(1552),
    [anon_sym_false] = ACTIONS(1552),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1552),
    [sym_super] = ACTIONS(1552),
    [sym_crate] = ACTIONS(1552),
    [sym_metavariable] = ACTIONS(1550),
    [sym_raw_string_literal] = ACTIONS(1550),
    [sym_float_literal] = ACTIONS(1550),
    [sym_block_comment] = ACTIONS(3),
  },
  [368] = {
    [ts_builtin_sym_end] = ACTIONS(1554),
    [sym_identifier] = ACTIONS(1556),
    [anon_sym_SEMI] = ACTIONS(1554),
    [anon_sym_macro_rules_BANG] = ACTIONS(1554),
    [anon_sym_LPAREN] = ACTIONS(1554),
    [anon_sym_LBRACE] = ACTIONS(1554),
    [anon_sym_RBRACE] = ACTIONS(1554),
    [anon_sym_LBRACK] = ACTIONS(1554),
    [anon_sym_STAR] = ACTIONS(1554),
    [anon_sym_u8] = ACTIONS(1556),
    [anon_sym_i8] = ACTIONS(1556),
    [anon_sym_u16] = ACTIONS(1556),
    [anon_sym_i16] = ACTIONS(1556),
    [anon_sym_u32] = ACTIONS(1556),
    [anon_sym_i32] = ACTIONS(1556),
    [anon_sym_u64] = ACTIONS(1556),
    [anon_sym_i64] = ACTIONS(1556),
    [anon_sym_u128] = ACTIONS(1556),
    [anon_sym_i128] = ACTIONS(1556),
    [anon_sym_isize] = ACTIONS(1556),
    [anon_sym_usize] = ACTIONS(1556),
    [anon_sym_f32] = ACTIONS(1556),
    [anon_sym_f64] = ACTIONS(1556),
    [anon_sym_bool] = ACTIONS(1556),
    [anon_sym_str] = ACTIONS(1556),
    [anon_sym_char] = ACTIONS(1556),
    [anon_sym_SQUOTE] = ACTIONS(1556),
    [anon_sym_async] = ACTIONS(1556),
    [anon_sym_break] = ACTIONS(1556),
    [anon_sym_const] = ACTIONS(1556),
    [anon_sym_continue] = ACTIONS(1556),
    [anon_sym_default] = ACTIONS(1556),
    [anon_sym_enum] = ACTIONS(1556),
    [anon_sym_fn] = ACTIONS(1556),
    [anon_sym_for] = ACTIONS(1556),
    [anon_sym_if] = ACTIONS(1556),
    [anon_sym_impl] = ACTIONS(1556),
    [anon_sym_let] = ACTIONS(1556),
    [anon_sym_loop] = ACTIONS(1556),
    [anon_sym_match] = ACTIONS(1556),
    [anon_sym_mod] = ACTIONS(1556),
    [anon_sym_pub] = ACTIONS(1556),
    [anon_sym_return] = ACTIONS(1556),
    [anon_sym_static] = ACTIONS(1556),
    [anon_sym_struct] = ACTIONS(1556),
    [anon_sym_trait] = ACTIONS(1556),
    [anon_sym_type] = ACTIONS(1556),
    [anon_sym_union] = ACTIONS(1556),
    [anon_sym_unsafe] = ACTIONS(1556),
    [anon_sym_use] = ACTIONS(1556),
    [anon_sym_while] = ACTIONS(1556),
    [anon_sym_POUND] = ACTIONS(1554),
    [anon_sym_BANG] = ACTIONS(1554),
    [anon_sym_extern] = ACTIONS(1556),
    [anon_sym_LT] = ACTIONS(1554),
    [anon_sym_COLON_COLON] = ACTIONS(1554),
    [anon_sym_AMP] = ACTIONS(1554),
    [anon_sym_DOT_DOT] = ACTIONS(1554),
    [anon_sym_DASH] = ACTIONS(1554),
    [anon_sym_PIPE] = ACTIONS(1554),
    [anon_sym_yield] = ACTIONS(1556),
    [anon_sym_move] = ACTIONS(1556),
    [sym_integer_literal] = ACTIONS(1554),
    [aux_sym_string_literal_token1] = ACTIONS(1554),
    [sym_char_literal] = ACTIONS(1554),
    [anon_sym_true] = ACTIONS(1556),
    [anon_sym_false] = ACTIONS(1556),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1556),
    [sym_super] = ACTIONS(1556),
    [sym_crate] = ACTIONS(1556),
    [sym_metavariable] = ACTIONS(1554),
    [sym_raw_string_literal] = ACTIONS(1554),
    [sym_float_literal] = ACTIONS(1554),
    [sym_block_comment] = ACTIONS(3),
  },
  [369] = {
    [ts_builtin_sym_end] = ACTIONS(1558),
    [sym_identifier] = ACTIONS(1560),
    [anon_sym_SEMI] = ACTIONS(1558),
    [anon_sym_macro_rules_BANG] = ACTIONS(1558),
    [anon_sym_LPAREN] = ACTIONS(1558),
    [anon_sym_LBRACE] = ACTIONS(1558),
    [anon_sym_RBRACE] = ACTIONS(1558),
    [anon_sym_LBRACK] = ACTIONS(1558),
    [anon_sym_STAR] = ACTIONS(1558),
    [anon_sym_u8] = ACTIONS(1560),
    [anon_sym_i8] = ACTIONS(1560),
    [anon_sym_u16] = ACTIONS(1560),
    [anon_sym_i16] = ACTIONS(1560),
    [anon_sym_u32] = ACTIONS(1560),
    [anon_sym_i32] = ACTIONS(1560),
    [anon_sym_u64] = ACTIONS(1560),
    [anon_sym_i64] = ACTIONS(1560),
    [anon_sym_u128] = ACTIONS(1560),
    [anon_sym_i128] = ACTIONS(1560),
    [anon_sym_isize] = ACTIONS(1560),
    [anon_sym_usize] = ACTIONS(1560),
    [anon_sym_f32] = ACTIONS(1560),
    [anon_sym_f64] = ACTIONS(1560),
    [anon_sym_bool] = ACTIONS(1560),
    [anon_sym_str] = ACTIONS(1560),
    [anon_sym_char] = ACTIONS(1560),
    [anon_sym_SQUOTE] = ACTIONS(1560),
    [anon_sym_async] = ACTIONS(1560),
    [anon_sym_break] = ACTIONS(1560),
    [anon_sym_const] = ACTIONS(1560),
    [anon_sym_continue] = ACTIONS(1560),
    [anon_sym_default] = ACTIONS(1560),
    [anon_sym_enum] = ACTIONS(1560),
    [anon_sym_fn] = ACTIONS(1560),
    [anon_sym_for] = ACTIONS(1560),
    [anon_sym_if] = ACTIONS(1560),
    [anon_sym_impl] = ACTIONS(1560),
    [anon_sym_let] = ACTIONS(1560),
    [anon_sym_loop] = ACTIONS(1560),
    [anon_sym_match] = ACTIONS(1560),
    [anon_sym_mod] = ACTIONS(1560),
    [anon_sym_pub] = ACTIONS(1560),
    [anon_sym_return] = ACTIONS(1560),
    [anon_sym_static] = ACTIONS(1560),
    [anon_sym_struct] = ACTIONS(1560),
    [anon_sym_trait] = ACTIONS(1560),
    [anon_sym_type] = ACTIONS(1560),
    [anon_sym_union] = ACTIONS(1560),
    [anon_sym_unsafe] = ACTIONS(1560),
    [anon_sym_use] = ACTIONS(1560),
    [anon_sym_while] = ACTIONS(1560),
    [anon_sym_POUND] = ACTIONS(1558),
    [anon_sym_BANG] = ACTIONS(1558),
    [anon_sym_extern] = ACTIONS(1560),
    [anon_sym_LT] = ACTIONS(1558),
    [anon_sym_COLON_COLON] = ACTIONS(1558),
    [anon_sym_AMP] = ACTIONS(1558),
    [anon_sym_DOT_DOT] = ACTIONS(1558),
    [anon_sym_DASH] = ACTIONS(1558),
    [anon_sym_PIPE] = ACTIONS(1558),
    [anon_sym_yield] = ACTIONS(1560),
    [anon_sym_move] = ACTIONS(1560),
    [sym_integer_literal] = ACTIONS(1558),
    [aux_sym_string_literal_token1] = ACTIONS(1558),
    [sym_char_literal] = ACTIONS(1558),
    [anon_sym_true] = ACTIONS(1560),
    [anon_sym_false] = ACTIONS(1560),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1560),
    [sym_super] = ACTIONS(1560),
    [sym_crate] = ACTIONS(1560),
    [sym_metavariable] = ACTIONS(1558),
    [sym_raw_string_literal] = ACTIONS(1558),
    [sym_float_literal] = ACTIONS(1558),
    [sym_block_comment] = ACTIONS(3),
  },
  [370] = {
    [ts_builtin_sym_end] = ACTIONS(1562),
    [sym_identifier] = ACTIONS(1564),
    [anon_sym_SEMI] = ACTIONS(1562),
    [anon_sym_macro_rules_BANG] = ACTIONS(1562),
    [anon_sym_LPAREN] = ACTIONS(1562),
    [anon_sym_LBRACE] = ACTIONS(1562),
    [anon_sym_RBRACE] = ACTIONS(1562),
    [anon_sym_LBRACK] = ACTIONS(1562),
    [anon_sym_STAR] = ACTIONS(1562),
    [anon_sym_u8] = ACTIONS(1564),
    [anon_sym_i8] = ACTIONS(1564),
    [anon_sym_u16] = ACTIONS(1564),
    [anon_sym_i16] = ACTIONS(1564),
    [anon_sym_u32] = ACTIONS(1564),
    [anon_sym_i32] = ACTIONS(1564),
    [anon_sym_u64] = ACTIONS(1564),
    [anon_sym_i64] = ACTIONS(1564),
    [anon_sym_u128] = ACTIONS(1564),
    [anon_sym_i128] = ACTIONS(1564),
    [anon_sym_isize] = ACTIONS(1564),
    [anon_sym_usize] = ACTIONS(1564),
    [anon_sym_f32] = ACTIONS(1564),
    [anon_sym_f64] = ACTIONS(1564),
    [anon_sym_bool] = ACTIONS(1564),
    [anon_sym_str] = ACTIONS(1564),
    [anon_sym_char] = ACTIONS(1564),
    [anon_sym_SQUOTE] = ACTIONS(1564),
    [anon_sym_async] = ACTIONS(1564),
    [anon_sym_break] = ACTIONS(1564),
    [anon_sym_const] = ACTIONS(1564),
    [anon_sym_continue] = ACTIONS(1564),
    [anon_sym_default] = ACTIONS(1564),
    [anon_sym_enum] = ACTIONS(1564),
    [anon_sym_fn] = ACTIONS(1564),
    [anon_sym_for] = ACTIONS(1564),
    [anon_sym_if] = ACTIONS(1564),
    [anon_sym_impl] = ACTIONS(1564),
    [anon_sym_let] = ACTIONS(1564),
    [anon_sym_loop] = ACTIONS(1564),
    [anon_sym_match] = ACTIONS(1564),
    [anon_sym_mod] = ACTIONS(1564),
    [anon_sym_pub] = ACTIONS(1564),
    [anon_sym_return] = ACTIONS(1564),
    [anon_sym_static] = ACTIONS(1564),
    [anon_sym_struct] = ACTIONS(1564),
    [anon_sym_trait] = ACTIONS(1564),
    [anon_sym_type] = ACTIONS(1564),
    [anon_sym_union] = ACTIONS(1564),
    [anon_sym_unsafe] = ACTIONS(1564),
    [anon_sym_use] = ACTIONS(1564),
    [anon_sym_while] = ACTIONS(1564),
    [anon_sym_POUND] = ACTIONS(1562),
    [anon_sym_BANG] = ACTIONS(1562),
    [anon_sym_extern] = ACTIONS(1564),
    [anon_sym_LT] = ACTIONS(1562),
    [anon_sym_COLON_COLON] = ACTIONS(1562),
    [anon_sym_AMP] = ACTIONS(1562),
    [anon_sym_DOT_DOT] = ACTIONS(1562),
    [anon_sym_DASH] = ACTIONS(1562),
    [anon_sym_PIPE] = ACTIONS(1562),
    [anon_sym_yield] = ACTIONS(1564),
    [anon_sym_move] = ACTIONS(1564),
    [sym_integer_literal] = ACTIONS(1562),
    [aux_sym_string_literal_token1] = ACTIONS(1562),
    [sym_char_literal] = ACTIONS(1562),
    [anon_sym_true] = ACTIONS(1564),
    [anon_sym_false] = ACTIONS(1564),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1564),
    [sym_super] = ACTIONS(1564),
    [sym_crate] = ACTIONS(1564),
    [sym_metavariable] = ACTIONS(1562),
    [sym_raw_string_literal] = ACTIONS(1562),
    [sym_float_literal] = ACTIONS(1562),
    [sym_block_comment] = ACTIONS(3),
  },
  [371] = {
    [ts_builtin_sym_end] = ACTIONS(1566),
    [sym_identifier] = ACTIONS(1568),
    [anon_sym_SEMI] = ACTIONS(1566),
    [anon_sym_macro_rules_BANG] = ACTIONS(1566),
    [anon_sym_LPAREN] = ACTIONS(1566),
    [anon_sym_LBRACE] = ACTIONS(1566),
    [anon_sym_RBRACE] = ACTIONS(1566),
    [anon_sym_LBRACK] = ACTIONS(1566),
    [anon_sym_STAR] = ACTIONS(1566),
    [anon_sym_u8] = ACTIONS(1568),
    [anon_sym_i8] = ACTIONS(1568),
    [anon_sym_u16] = ACTIONS(1568),
    [anon_sym_i16] = ACTIONS(1568),
    [anon_sym_u32] = ACTIONS(1568),
    [anon_sym_i32] = ACTIONS(1568),
    [anon_sym_u64] = ACTIONS(1568),
    [anon_sym_i64] = ACTIONS(1568),
    [anon_sym_u128] = ACTIONS(1568),
    [anon_sym_i128] = ACTIONS(1568),
    [anon_sym_isize] = ACTIONS(1568),
    [anon_sym_usize] = ACTIONS(1568),
    [anon_sym_f32] = ACTIONS(1568),
    [anon_sym_f64] = ACTIONS(1568),
    [anon_sym_bool] = ACTIONS(1568),
    [anon_sym_str] = ACTIONS(1568),
    [anon_sym_char] = ACTIONS(1568),
    [anon_sym_SQUOTE] = ACTIONS(1568),
    [anon_sym_async] = ACTIONS(1568),
    [anon_sym_break] = ACTIONS(1568),
    [anon_sym_const] = ACTIONS(1568),
    [anon_sym_continue] = ACTIONS(1568),
    [anon_sym_default] = ACTIONS(1568),
    [anon_sym_enum] = ACTIONS(1568),
    [anon_sym_fn] = ACTIONS(1568),
    [anon_sym_for] = ACTIONS(1568),
    [anon_sym_if] = ACTIONS(1568),
    [anon_sym_impl] = ACTIONS(1568),
    [anon_sym_let] = ACTIONS(1568),
    [anon_sym_loop] = ACTIONS(1568),
    [anon_sym_match] = ACTIONS(1568),
    [anon_sym_mod] = ACTIONS(1568),
    [anon_sym_pub] = ACTIONS(1568),
    [anon_sym_return] = ACTIONS(1568),
    [anon_sym_static] = ACTIONS(1568),
    [anon_sym_struct] = ACTIONS(1568),
    [anon_sym_trait] = ACTIONS(1568),
    [anon_sym_type] = ACTIONS(1568),
    [anon_sym_union] = ACTIONS(1568),
    [anon_sym_unsafe] = ACTIONS(1568),
    [anon_sym_use] = ACTIONS(1568),
    [anon_sym_while] = ACTIONS(1568),
    [anon_sym_POUND] = ACTIONS(1566),
    [anon_sym_BANG] = ACTIONS(1566),
    [anon_sym_extern] = ACTIONS(1568),
    [anon_sym_LT] = ACTIONS(1566),
    [anon_sym_COLON_COLON] = ACTIONS(1566),
    [anon_sym_AMP] = ACTIONS(1566),
    [anon_sym_DOT_DOT] = ACTIONS(1566),
    [anon_sym_DASH] = ACTIONS(1566),
    [anon_sym_PIPE] = ACTIONS(1566),
    [anon_sym_yield] = ACTIONS(1568),
    [anon_sym_move] = ACTIONS(1568),
    [sym_integer_literal] = ACTIONS(1566),
    [aux_sym_string_literal_token1] = ACTIONS(1566),
    [sym_char_literal] = ACTIONS(1566),
    [anon_sym_true] = ACTIONS(1568),
    [anon_sym_false] = ACTIONS(1568),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1568),
    [sym_super] = ACTIONS(1568),
    [sym_crate] = ACTIONS(1568),
    [sym_metavariable] = ACTIONS(1566),
    [sym_raw_string_literal] = ACTIONS(1566),
    [sym_float_literal] = ACTIONS(1566),
    [sym_block_comment] = ACTIONS(3),
  },
  [372] = {
    [ts_builtin_sym_end] = ACTIONS(1570),
    [sym_identifier] = ACTIONS(1572),
    [anon_sym_SEMI] = ACTIONS(1570),
    [anon_sym_macro_rules_BANG] = ACTIONS(1570),
    [anon_sym_LPAREN] = ACTIONS(1570),
    [anon_sym_LBRACE] = ACTIONS(1570),
    [anon_sym_RBRACE] = ACTIONS(1570),
    [anon_sym_LBRACK] = ACTIONS(1570),
    [anon_sym_STAR] = ACTIONS(1570),
    [anon_sym_u8] = ACTIONS(1572),
    [anon_sym_i8] = ACTIONS(1572),
    [anon_sym_u16] = ACTIONS(1572),
    [anon_sym_i16] = ACTIONS(1572),
    [anon_sym_u32] = ACTIONS(1572),
    [anon_sym_i32] = ACTIONS(1572),
    [anon_sym_u64] = ACTIONS(1572),
    [anon_sym_i64] = ACTIONS(1572),
    [anon_sym_u128] = ACTIONS(1572),
    [anon_sym_i128] = ACTIONS(1572),
    [anon_sym_isize] = ACTIONS(1572),
    [anon_sym_usize] = ACTIONS(1572),
    [anon_sym_f32] = ACTIONS(1572),
    [anon_sym_f64] = ACTIONS(1572),
    [anon_sym_bool] = ACTIONS(1572),
    [anon_sym_str] = ACTIONS(1572),
    [anon_sym_char] = ACTIONS(1572),
    [anon_sym_SQUOTE] = ACTIONS(1572),
    [anon_sym_async] = ACTIONS(1572),
    [anon_sym_break] = ACTIONS(1572),
    [anon_sym_const] = ACTIONS(1572),
    [anon_sym_continue] = ACTIONS(1572),
    [anon_sym_default] = ACTIONS(1572),
    [anon_sym_enum] = ACTIONS(1572),
    [anon_sym_fn] = ACTIONS(1572),
    [anon_sym_for] = ACTIONS(1572),
    [anon_sym_if] = ACTIONS(1572),
    [anon_sym_impl] = ACTIONS(1572),
    [anon_sym_let] = ACTIONS(1572),
    [anon_sym_loop] = ACTIONS(1572),
    [anon_sym_match] = ACTIONS(1572),
    [anon_sym_mod] = ACTIONS(1572),
    [anon_sym_pub] = ACTIONS(1572),
    [anon_sym_return] = ACTIONS(1572),
    [anon_sym_static] = ACTIONS(1572),
    [anon_sym_struct] = ACTIONS(1572),
    [anon_sym_trait] = ACTIONS(1572),
    [anon_sym_type] = ACTIONS(1572),
    [anon_sym_union] = ACTIONS(1572),
    [anon_sym_unsafe] = ACTIONS(1572),
    [anon_sym_use] = ACTIONS(1572),
    [anon_sym_while] = ACTIONS(1572),
    [anon_sym_POUND] = ACTIONS(1570),
    [anon_sym_BANG] = ACTIONS(1570),
    [anon_sym_extern] = ACTIONS(1572),
    [anon_sym_LT] = ACTIONS(1570),
    [anon_sym_COLON_COLON] = ACTIONS(1570),
    [anon_sym_AMP] = ACTIONS(1570),
    [anon_sym_DOT_DOT] = ACTIONS(1570),
    [anon_sym_DASH] = ACTIONS(1570),
    [anon_sym_PIPE] = ACTIONS(1570),
    [anon_sym_yield] = ACTIONS(1572),
    [anon_sym_move] = ACTIONS(1572),
    [sym_integer_literal] = ACTIONS(1570),
    [aux_sym_string_literal_token1] = ACTIONS(1570),
    [sym_char_literal] = ACTIONS(1570),
    [anon_sym_true] = ACTIONS(1572),
    [anon_sym_false] = ACTIONS(1572),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1572),
    [sym_super] = ACTIONS(1572),
    [sym_crate] = ACTIONS(1572),
    [sym_metavariable] = ACTIONS(1570),
    [sym_raw_string_literal] = ACTIONS(1570),
    [sym_float_literal] = ACTIONS(1570),
    [sym_block_comment] = ACTIONS(3),
  },
  [373] = {
    [ts_builtin_sym_end] = ACTIONS(1574),
    [sym_identifier] = ACTIONS(1576),
    [anon_sym_SEMI] = ACTIONS(1574),
    [anon_sym_macro_rules_BANG] = ACTIONS(1574),
    [anon_sym_LPAREN] = ACTIONS(1574),
    [anon_sym_LBRACE] = ACTIONS(1574),
    [anon_sym_RBRACE] = ACTIONS(1574),
    [anon_sym_LBRACK] = ACTIONS(1574),
    [anon_sym_STAR] = ACTIONS(1574),
    [anon_sym_u8] = ACTIONS(1576),
    [anon_sym_i8] = ACTIONS(1576),
    [anon_sym_u16] = ACTIONS(1576),
    [anon_sym_i16] = ACTIONS(1576),
    [anon_sym_u32] = ACTIONS(1576),
    [anon_sym_i32] = ACTIONS(1576),
    [anon_sym_u64] = ACTIONS(1576),
    [anon_sym_i64] = ACTIONS(1576),
    [anon_sym_u128] = ACTIONS(1576),
    [anon_sym_i128] = ACTIONS(1576),
    [anon_sym_isize] = ACTIONS(1576),
    [anon_sym_usize] = ACTIONS(1576),
    [anon_sym_f32] = ACTIONS(1576),
    [anon_sym_f64] = ACTIONS(1576),
    [anon_sym_bool] = ACTIONS(1576),
    [anon_sym_str] = ACTIONS(1576),
    [anon_sym_char] = ACTIONS(1576),
    [anon_sym_SQUOTE] = ACTIONS(1576),
    [anon_sym_async] = ACTIONS(1576),
    [anon_sym_break] = ACTIONS(1576),
    [anon_sym_const] = ACTIONS(1576),
    [anon_sym_continue] = ACTIONS(1576),
    [anon_sym_default] = ACTIONS(1576),
    [anon_sym_enum] = ACTIONS(1576),
    [anon_sym_fn] = ACTIONS(1576),
    [anon_sym_for] = ACTIONS(1576),
    [anon_sym_if] = ACTIONS(1576),
    [anon_sym_impl] = ACTIONS(1576),
    [anon_sym_let] = ACTIONS(1576),
    [anon_sym_loop] = ACTIONS(1576),
    [anon_sym_match] = ACTIONS(1576),
    [anon_sym_mod] = ACTIONS(1576),
    [anon_sym_pub] = ACTIONS(1576),
    [anon_sym_return] = ACTIONS(1576),
    [anon_sym_static] = ACTIONS(1576),
    [anon_sym_struct] = ACTIONS(1576),
    [anon_sym_trait] = ACTIONS(1576),
    [anon_sym_type] = ACTIONS(1576),
    [anon_sym_union] = ACTIONS(1576),
    [anon_sym_unsafe] = ACTIONS(1576),
    [anon_sym_use] = ACTIONS(1576),
    [anon_sym_while] = ACTIONS(1576),
    [anon_sym_POUND] = ACTIONS(1574),
    [anon_sym_BANG] = ACTIONS(1574),
    [anon_sym_extern] = ACTIONS(1576),
    [anon_sym_LT] = ACTIONS(1574),
    [anon_sym_COLON_COLON] = ACTIONS(1574),
    [anon_sym_AMP] = ACTIONS(1574),
    [anon_sym_DOT_DOT] = ACTIONS(1574),
    [anon_sym_DASH] = ACTIONS(1574),
    [anon_sym_PIPE] = ACTIONS(1574),
    [anon_sym_yield] = ACTIONS(1576),
    [anon_sym_move] = ACTIONS(1576),
    [sym_integer_literal] = ACTIONS(1574),
    [aux_sym_string_literal_token1] = ACTIONS(1574),
    [sym_char_literal] = ACTIONS(1574),
    [anon_sym_true] = ACTIONS(1576),
    [anon_sym_false] = ACTIONS(1576),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1576),
    [sym_super] = ACTIONS(1576),
    [sym_crate] = ACTIONS(1576),
    [sym_metavariable] = ACTIONS(1574),
    [sym_raw_string_literal] = ACTIONS(1574),
    [sym_float_literal] = ACTIONS(1574),
    [sym_block_comment] = ACTIONS(3),
  },
  [374] = {
    [ts_builtin_sym_end] = ACTIONS(1578),
    [sym_identifier] = ACTIONS(1580),
    [anon_sym_SEMI] = ACTIONS(1578),
    [anon_sym_macro_rules_BANG] = ACTIONS(1578),
    [anon_sym_LPAREN] = ACTIONS(1578),
    [anon_sym_LBRACE] = ACTIONS(1578),
    [anon_sym_RBRACE] = ACTIONS(1578),
    [anon_sym_LBRACK] = ACTIONS(1578),
    [anon_sym_STAR] = ACTIONS(1578),
    [anon_sym_u8] = ACTIONS(1580),
    [anon_sym_i8] = ACTIONS(1580),
    [anon_sym_u16] = ACTIONS(1580),
    [anon_sym_i16] = ACTIONS(1580),
    [anon_sym_u32] = ACTIONS(1580),
    [anon_sym_i32] = ACTIONS(1580),
    [anon_sym_u64] = ACTIONS(1580),
    [anon_sym_i64] = ACTIONS(1580),
    [anon_sym_u128] = ACTIONS(1580),
    [anon_sym_i128] = ACTIONS(1580),
    [anon_sym_isize] = ACTIONS(1580),
    [anon_sym_usize] = ACTIONS(1580),
    [anon_sym_f32] = ACTIONS(1580),
    [anon_sym_f64] = ACTIONS(1580),
    [anon_sym_bool] = ACTIONS(1580),
    [anon_sym_str] = ACTIONS(1580),
    [anon_sym_char] = ACTIONS(1580),
    [anon_sym_SQUOTE] = ACTIONS(1580),
    [anon_sym_async] = ACTIONS(1580),
    [anon_sym_break] = ACTIONS(1580),
    [anon_sym_const] = ACTIONS(1580),
    [anon_sym_continue] = ACTIONS(1580),
    [anon_sym_default] = ACTIONS(1580),
    [anon_sym_enum] = ACTIONS(1580),
    [anon_sym_fn] = ACTIONS(1580),
    [anon_sym_for] = ACTIONS(1580),
    [anon_sym_if] = ACTIONS(1580),
    [anon_sym_impl] = ACTIONS(1580),
    [anon_sym_let] = ACTIONS(1580),
    [anon_sym_loop] = ACTIONS(1580),
    [anon_sym_match] = ACTIONS(1580),
    [anon_sym_mod] = ACTIONS(1580),
    [anon_sym_pub] = ACTIONS(1580),
    [anon_sym_return] = ACTIONS(1580),
    [anon_sym_static] = ACTIONS(1580),
    [anon_sym_struct] = ACTIONS(1580),
    [anon_sym_trait] = ACTIONS(1580),
    [anon_sym_type] = ACTIONS(1580),
    [anon_sym_union] = ACTIONS(1580),
    [anon_sym_unsafe] = ACTIONS(1580),
    [anon_sym_use] = ACTIONS(1580),
    [anon_sym_while] = ACTIONS(1580),
    [anon_sym_POUND] = ACTIONS(1578),
    [anon_sym_BANG] = ACTIONS(1578),
    [anon_sym_extern] = ACTIONS(1580),
    [anon_sym_LT] = ACTIONS(1578),
    [anon_sym_COLON_COLON] = ACTIONS(1578),
    [anon_sym_AMP] = ACTIONS(1578),
    [anon_sym_DOT_DOT] = ACTIONS(1578),
    [anon_sym_DASH] = ACTIONS(1578),
    [anon_sym_PIPE] = ACTIONS(1578),
    [anon_sym_yield] = ACTIONS(1580),
    [anon_sym_move] = ACTIONS(1580),
    [sym_integer_literal] = ACTIONS(1578),
    [aux_sym_string_literal_token1] = ACTIONS(1578),
    [sym_char_literal] = ACTIONS(1578),
    [anon_sym_true] = ACTIONS(1580),
    [anon_sym_false] = ACTIONS(1580),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1580),
    [sym_super] = ACTIONS(1580),
    [sym_crate] = ACTIONS(1580),
    [sym_metavariable] = ACTIONS(1578),
    [sym_raw_string_literal] = ACTIONS(1578),
    [sym_float_literal] = ACTIONS(1578),
    [sym_block_comment] = ACTIONS(3),
  },
  [375] = {
    [ts_builtin_sym_end] = ACTIONS(1582),
    [sym_identifier] = ACTIONS(1584),
    [anon_sym_SEMI] = ACTIONS(1582),
    [anon_sym_macro_rules_BANG] = ACTIONS(1582),
    [anon_sym_LPAREN] = ACTIONS(1582),
    [anon_sym_LBRACE] = ACTIONS(1582),
    [anon_sym_RBRACE] = ACTIONS(1582),
    [anon_sym_LBRACK] = ACTIONS(1582),
    [anon_sym_STAR] = ACTIONS(1582),
    [anon_sym_u8] = ACTIONS(1584),
    [anon_sym_i8] = ACTIONS(1584),
    [anon_sym_u16] = ACTIONS(1584),
    [anon_sym_i16] = ACTIONS(1584),
    [anon_sym_u32] = ACTIONS(1584),
    [anon_sym_i32] = ACTIONS(1584),
    [anon_sym_u64] = ACTIONS(1584),
    [anon_sym_i64] = ACTIONS(1584),
    [anon_sym_u128] = ACTIONS(1584),
    [anon_sym_i128] = ACTIONS(1584),
    [anon_sym_isize] = ACTIONS(1584),
    [anon_sym_usize] = ACTIONS(1584),
    [anon_sym_f32] = ACTIONS(1584),
    [anon_sym_f64] = ACTIONS(1584),
    [anon_sym_bool] = ACTIONS(1584),
    [anon_sym_str] = ACTIONS(1584),
    [anon_sym_char] = ACTIONS(1584),
    [anon_sym_SQUOTE] = ACTIONS(1584),
    [anon_sym_async] = ACTIONS(1584),
    [anon_sym_break] = ACTIONS(1584),
    [anon_sym_const] = ACTIONS(1584),
    [anon_sym_continue] = ACTIONS(1584),
    [anon_sym_default] = ACTIONS(1584),
    [anon_sym_enum] = ACTIONS(1584),
    [anon_sym_fn] = ACTIONS(1584),
    [anon_sym_for] = ACTIONS(1584),
    [anon_sym_if] = ACTIONS(1584),
    [anon_sym_impl] = ACTIONS(1584),
    [anon_sym_let] = ACTIONS(1584),
    [anon_sym_loop] = ACTIONS(1584),
    [anon_sym_match] = ACTIONS(1584),
    [anon_sym_mod] = ACTIONS(1584),
    [anon_sym_pub] = ACTIONS(1584),
    [anon_sym_return] = ACTIONS(1584),
    [anon_sym_static] = ACTIONS(1584),
    [anon_sym_struct] = ACTIONS(1584),
    [anon_sym_trait] = ACTIONS(1584),
    [anon_sym_type] = ACTIONS(1584),
    [anon_sym_union] = ACTIONS(1584),
    [anon_sym_unsafe] = ACTIONS(1584),
    [anon_sym_use] = ACTIONS(1584),
    [anon_sym_while] = ACTIONS(1584),
    [anon_sym_POUND] = ACTIONS(1582),
    [anon_sym_BANG] = ACTIONS(1582),
    [anon_sym_extern] = ACTIONS(1584),
    [anon_sym_LT] = ACTIONS(1582),
    [anon_sym_COLON_COLON] = ACTIONS(1582),
    [anon_sym_AMP] = ACTIONS(1582),
    [anon_sym_DOT_DOT] = ACTIONS(1582),
    [anon_sym_DASH] = ACTIONS(1582),
    [anon_sym_PIPE] = ACTIONS(1582),
    [anon_sym_yield] = ACTIONS(1584),
    [anon_sym_move] = ACTIONS(1584),
    [sym_integer_literal] = ACTIONS(1582),
    [aux_sym_string_literal_token1] = ACTIONS(1582),
    [sym_char_literal] = ACTIONS(1582),
    [anon_sym_true] = ACTIONS(1584),
    [anon_sym_false] = ACTIONS(1584),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1584),
    [sym_super] = ACTIONS(1584),
    [sym_crate] = ACTIONS(1584),
    [sym_metavariable] = ACTIONS(1582),
    [sym_raw_string_literal] = ACTIONS(1582),
    [sym_float_literal] = ACTIONS(1582),
    [sym_block_comment] = ACTIONS(3),
  },
  [376] = {
    [ts_builtin_sym_end] = ACTIONS(1586),
    [sym_identifier] = ACTIONS(1588),
    [anon_sym_SEMI] = ACTIONS(1586),
    [anon_sym_macro_rules_BANG] = ACTIONS(1586),
    [anon_sym_LPAREN] = ACTIONS(1586),
    [anon_sym_LBRACE] = ACTIONS(1586),
    [anon_sym_RBRACE] = ACTIONS(1586),
    [anon_sym_LBRACK] = ACTIONS(1586),
    [anon_sym_STAR] = ACTIONS(1586),
    [anon_sym_u8] = ACTIONS(1588),
    [anon_sym_i8] = ACTIONS(1588),
    [anon_sym_u16] = ACTIONS(1588),
    [anon_sym_i16] = ACTIONS(1588),
    [anon_sym_u32] = ACTIONS(1588),
    [anon_sym_i32] = ACTIONS(1588),
    [anon_sym_u64] = ACTIONS(1588),
    [anon_sym_i64] = ACTIONS(1588),
    [anon_sym_u128] = ACTIONS(1588),
    [anon_sym_i128] = ACTIONS(1588),
    [anon_sym_isize] = ACTIONS(1588),
    [anon_sym_usize] = ACTIONS(1588),
    [anon_sym_f32] = ACTIONS(1588),
    [anon_sym_f64] = ACTIONS(1588),
    [anon_sym_bool] = ACTIONS(1588),
    [anon_sym_str] = ACTIONS(1588),
    [anon_sym_char] = ACTIONS(1588),
    [anon_sym_SQUOTE] = ACTIONS(1588),
    [anon_sym_async] = ACTIONS(1588),
    [anon_sym_break] = ACTIONS(1588),
    [anon_sym_const] = ACTIONS(1588),
    [anon_sym_continue] = ACTIONS(1588),
    [anon_sym_default] = ACTIONS(1588),
    [anon_sym_enum] = ACTIONS(1588),
    [anon_sym_fn] = ACTIONS(1588),
    [anon_sym_for] = ACTIONS(1588),
    [anon_sym_if] = ACTIONS(1588),
    [anon_sym_impl] = ACTIONS(1588),
    [anon_sym_let] = ACTIONS(1588),
    [anon_sym_loop] = ACTIONS(1588),
    [anon_sym_match] = ACTIONS(1588),
    [anon_sym_mod] = ACTIONS(1588),
    [anon_sym_pub] = ACTIONS(1588),
    [anon_sym_return] = ACTIONS(1588),
    [anon_sym_static] = ACTIONS(1588),
    [anon_sym_struct] = ACTIONS(1588),
    [anon_sym_trait] = ACTIONS(1588),
    [anon_sym_type] = ACTIONS(1588),
    [anon_sym_union] = ACTIONS(1588),
    [anon_sym_unsafe] = ACTIONS(1588),
    [anon_sym_use] = ACTIONS(1588),
    [anon_sym_while] = ACTIONS(1588),
    [anon_sym_POUND] = ACTIONS(1586),
    [anon_sym_BANG] = ACTIONS(1586),
    [anon_sym_extern] = ACTIONS(1588),
    [anon_sym_LT] = ACTIONS(1586),
    [anon_sym_COLON_COLON] = ACTIONS(1586),
    [anon_sym_AMP] = ACTIONS(1586),
    [anon_sym_DOT_DOT] = ACTIONS(1586),
    [anon_sym_DASH] = ACTIONS(1586),
    [anon_sym_PIPE] = ACTIONS(1586),
    [anon_sym_yield] = ACTIONS(1588),
    [anon_sym_move] = ACTIONS(1588),
    [sym_integer_literal] = ACTIONS(1586),
    [aux_sym_string_literal_token1] = ACTIONS(1586),
    [sym_char_literal] = ACTIONS(1586),
    [anon_sym_true] = ACTIONS(1588),
    [anon_sym_false] = ACTIONS(1588),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1588),
    [sym_super] = ACTIONS(1588),
    [sym_crate] = ACTIONS(1588),
    [sym_metavariable] = ACTIONS(1586),
    [sym_raw_string_literal] = ACTIONS(1586),
    [sym_float_literal] = ACTIONS(1586),
    [sym_block_comment] = ACTIONS(3),
  },
  [377] = {
    [ts_builtin_sym_end] = ACTIONS(1590),
    [sym_identifier] = ACTIONS(1592),
    [anon_sym_SEMI] = ACTIONS(1590),
    [anon_sym_macro_rules_BANG] = ACTIONS(1590),
    [anon_sym_LPAREN] = ACTIONS(1590),
    [anon_sym_LBRACE] = ACTIONS(1590),
    [anon_sym_RBRACE] = ACTIONS(1590),
    [anon_sym_LBRACK] = ACTIONS(1590),
    [anon_sym_STAR] = ACTIONS(1590),
    [anon_sym_u8] = ACTIONS(1592),
    [anon_sym_i8] = ACTIONS(1592),
    [anon_sym_u16] = ACTIONS(1592),
    [anon_sym_i16] = ACTIONS(1592),
    [anon_sym_u32] = ACTIONS(1592),
    [anon_sym_i32] = ACTIONS(1592),
    [anon_sym_u64] = ACTIONS(1592),
    [anon_sym_i64] = ACTIONS(1592),
    [anon_sym_u128] = ACTIONS(1592),
    [anon_sym_i128] = ACTIONS(1592),
    [anon_sym_isize] = ACTIONS(1592),
    [anon_sym_usize] = ACTIONS(1592),
    [anon_sym_f32] = ACTIONS(1592),
    [anon_sym_f64] = ACTIONS(1592),
    [anon_sym_bool] = ACTIONS(1592),
    [anon_sym_str] = ACTIONS(1592),
    [anon_sym_char] = ACTIONS(1592),
    [anon_sym_SQUOTE] = ACTIONS(1592),
    [anon_sym_async] = ACTIONS(1592),
    [anon_sym_break] = ACTIONS(1592),
    [anon_sym_const] = ACTIONS(1592),
    [anon_sym_continue] = ACTIONS(1592),
    [anon_sym_default] = ACTIONS(1592),
    [anon_sym_enum] = ACTIONS(1592),
    [anon_sym_fn] = ACTIONS(1592),
    [anon_sym_for] = ACTIONS(1592),
    [anon_sym_if] = ACTIONS(1592),
    [anon_sym_impl] = ACTIONS(1592),
    [anon_sym_let] = ACTIONS(1592),
    [anon_sym_loop] = ACTIONS(1592),
    [anon_sym_match] = ACTIONS(1592),
    [anon_sym_mod] = ACTIONS(1592),
    [anon_sym_pub] = ACTIONS(1592),
    [anon_sym_return] = ACTIONS(1592),
    [anon_sym_static] = ACTIONS(1592),
    [anon_sym_struct] = ACTIONS(1592),
    [anon_sym_trait] = ACTIONS(1592),
    [anon_sym_type] = ACTIONS(1592),
    [anon_sym_union] = ACTIONS(1592),
    [anon_sym_unsafe] = ACTIONS(1592),
    [anon_sym_use] = ACTIONS(1592),
    [anon_sym_while] = ACTIONS(1592),
    [anon_sym_POUND] = ACTIONS(1590),
    [anon_sym_BANG] = ACTIONS(1590),
    [anon_sym_extern] = ACTIONS(1592),
    [anon_sym_LT] = ACTIONS(1590),
    [anon_sym_COLON_COLON] = ACTIONS(1590),
    [anon_sym_AMP] = ACTIONS(1590),
    [anon_sym_DOT_DOT] = ACTIONS(1590),
    [anon_sym_DASH] = ACTIONS(1590),
    [anon_sym_PIPE] = ACTIONS(1590),
    [anon_sym_yield] = ACTIONS(1592),
    [anon_sym_move] = ACTIONS(1592),
    [sym_integer_literal] = ACTIONS(1590),
    [aux_sym_string_literal_token1] = ACTIONS(1590),
    [sym_char_literal] = ACTIONS(1590),
    [anon_sym_true] = ACTIONS(1592),
    [anon_sym_false] = ACTIONS(1592),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1592),
    [sym_super] = ACTIONS(1592),
    [sym_crate] = ACTIONS(1592),
    [sym_metavariable] = ACTIONS(1590),
    [sym_raw_string_literal] = ACTIONS(1590),
    [sym_float_literal] = ACTIONS(1590),
    [sym_block_comment] = ACTIONS(3),
  },
  [378] = {
    [ts_builtin_sym_end] = ACTIONS(1594),
    [sym_identifier] = ACTIONS(1596),
    [anon_sym_SEMI] = ACTIONS(1594),
    [anon_sym_macro_rules_BANG] = ACTIONS(1594),
    [anon_sym_LPAREN] = ACTIONS(1594),
    [anon_sym_LBRACE] = ACTIONS(1594),
    [anon_sym_RBRACE] = ACTIONS(1594),
    [anon_sym_LBRACK] = ACTIONS(1594),
    [anon_sym_STAR] = ACTIONS(1594),
    [anon_sym_u8] = ACTIONS(1596),
    [anon_sym_i8] = ACTIONS(1596),
    [anon_sym_u16] = ACTIONS(1596),
    [anon_sym_i16] = ACTIONS(1596),
    [anon_sym_u32] = ACTIONS(1596),
    [anon_sym_i32] = ACTIONS(1596),
    [anon_sym_u64] = ACTIONS(1596),
    [anon_sym_i64] = ACTIONS(1596),
    [anon_sym_u128] = ACTIONS(1596),
    [anon_sym_i128] = ACTIONS(1596),
    [anon_sym_isize] = ACTIONS(1596),
    [anon_sym_usize] = ACTIONS(1596),
    [anon_sym_f32] = ACTIONS(1596),
    [anon_sym_f64] = ACTIONS(1596),
    [anon_sym_bool] = ACTIONS(1596),
    [anon_sym_str] = ACTIONS(1596),
    [anon_sym_char] = ACTIONS(1596),
    [anon_sym_SQUOTE] = ACTIONS(1596),
    [anon_sym_async] = ACTIONS(1596),
    [anon_sym_break] = ACTIONS(1596),
    [anon_sym_const] = ACTIONS(1596),
    [anon_sym_continue] = ACTIONS(1596),
    [anon_sym_default] = ACTIONS(1596),
    [anon_sym_enum] = ACTIONS(1596),
    [anon_sym_fn] = ACTIONS(1596),
    [anon_sym_for] = ACTIONS(1596),
    [anon_sym_if] = ACTIONS(1596),
    [anon_sym_impl] = ACTIONS(1596),
    [anon_sym_let] = ACTIONS(1596),
    [anon_sym_loop] = ACTIONS(1596),
    [anon_sym_match] = ACTIONS(1596),
    [anon_sym_mod] = ACTIONS(1596),
    [anon_sym_pub] = ACTIONS(1596),
    [anon_sym_return] = ACTIONS(1596),
    [anon_sym_static] = ACTIONS(1596),
    [anon_sym_struct] = ACTIONS(1596),
    [anon_sym_trait] = ACTIONS(1596),
    [anon_sym_type] = ACTIONS(1596),
    [anon_sym_union] = ACTIONS(1596),
    [anon_sym_unsafe] = ACTIONS(1596),
    [anon_sym_use] = ACTIONS(1596),
    [anon_sym_while] = ACTIONS(1596),
    [anon_sym_POUND] = ACTIONS(1594),
    [anon_sym_BANG] = ACTIONS(1594),
    [anon_sym_extern] = ACTIONS(1596),
    [anon_sym_LT] = ACTIONS(1594),
    [anon_sym_COLON_COLON] = ACTIONS(1594),
    [anon_sym_AMP] = ACTIONS(1594),
    [anon_sym_DOT_DOT] = ACTIONS(1594),
    [anon_sym_DASH] = ACTIONS(1594),
    [anon_sym_PIPE] = ACTIONS(1594),
    [anon_sym_yield] = ACTIONS(1596),
    [anon_sym_move] = ACTIONS(1596),
    [sym_integer_literal] = ACTIONS(1594),
    [aux_sym_string_literal_token1] = ACTIONS(1594),
    [sym_char_literal] = ACTIONS(1594),
    [anon_sym_true] = ACTIONS(1596),
    [anon_sym_false] = ACTIONS(1596),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1596),
    [sym_super] = ACTIONS(1596),
    [sym_crate] = ACTIONS(1596),
    [sym_metavariable] = ACTIONS(1594),
    [sym_raw_string_literal] = ACTIONS(1594),
    [sym_float_literal] = ACTIONS(1594),
    [sym_block_comment] = ACTIONS(3),
  },
  [379] = {
    [ts_builtin_sym_end] = ACTIONS(1598),
    [sym_identifier] = ACTIONS(1600),
    [anon_sym_SEMI] = ACTIONS(1598),
    [anon_sym_macro_rules_BANG] = ACTIONS(1598),
    [anon_sym_LPAREN] = ACTIONS(1598),
    [anon_sym_LBRACE] = ACTIONS(1598),
    [anon_sym_RBRACE] = ACTIONS(1598),
    [anon_sym_LBRACK] = ACTIONS(1598),
    [anon_sym_STAR] = ACTIONS(1598),
    [anon_sym_u8] = ACTIONS(1600),
    [anon_sym_i8] = ACTIONS(1600),
    [anon_sym_u16] = ACTIONS(1600),
    [anon_sym_i16] = ACTIONS(1600),
    [anon_sym_u32] = ACTIONS(1600),
    [anon_sym_i32] = ACTIONS(1600),
    [anon_sym_u64] = ACTIONS(1600),
    [anon_sym_i64] = ACTIONS(1600),
    [anon_sym_u128] = ACTIONS(1600),
    [anon_sym_i128] = ACTIONS(1600),
    [anon_sym_isize] = ACTIONS(1600),
    [anon_sym_usize] = ACTIONS(1600),
    [anon_sym_f32] = ACTIONS(1600),
    [anon_sym_f64] = ACTIONS(1600),
    [anon_sym_bool] = ACTIONS(1600),
    [anon_sym_str] = ACTIONS(1600),
    [anon_sym_char] = ACTIONS(1600),
    [anon_sym_SQUOTE] = ACTIONS(1600),
    [anon_sym_async] = ACTIONS(1600),
    [anon_sym_break] = ACTIONS(1600),
    [anon_sym_const] = ACTIONS(1600),
    [anon_sym_continue] = ACTIONS(1600),
    [anon_sym_default] = ACTIONS(1600),
    [anon_sym_enum] = ACTIONS(1600),
    [anon_sym_fn] = ACTIONS(1600),
    [anon_sym_for] = ACTIONS(1600),
    [anon_sym_if] = ACTIONS(1600),
    [anon_sym_impl] = ACTIONS(1600),
    [anon_sym_let] = ACTIONS(1600),
    [anon_sym_loop] = ACTIONS(1600),
    [anon_sym_match] = ACTIONS(1600),
    [anon_sym_mod] = ACTIONS(1600),
    [anon_sym_pub] = ACTIONS(1600),
    [anon_sym_return] = ACTIONS(1600),
    [anon_sym_static] = ACTIONS(1600),
    [anon_sym_struct] = ACTIONS(1600),
    [anon_sym_trait] = ACTIONS(1600),
    [anon_sym_type] = ACTIONS(1600),
    [anon_sym_union] = ACTIONS(1600),
    [anon_sym_unsafe] = ACTIONS(1600),
    [anon_sym_use] = ACTIONS(1600),
    [anon_sym_while] = ACTIONS(1600),
    [anon_sym_POUND] = ACTIONS(1598),
    [anon_sym_BANG] = ACTIONS(1598),
    [anon_sym_extern] = ACTIONS(1600),
    [anon_sym_LT] = ACTIONS(1598),
    [anon_sym_COLON_COLON] = ACTIONS(1598),
    [anon_sym_AMP] = ACTIONS(1598),
    [anon_sym_DOT_DOT] = ACTIONS(1598),
    [anon_sym_DASH] = ACTIONS(1598),
    [anon_sym_PIPE] = ACTIONS(1598),
    [anon_sym_yield] = ACTIONS(1600),
    [anon_sym_move] = ACTIONS(1600),
    [sym_integer_literal] = ACTIONS(1598),
    [aux_sym_string_literal_token1] = ACTIONS(1598),
    [sym_char_literal] = ACTIONS(1598),
    [anon_sym_true] = ACTIONS(1600),
    [anon_sym_false] = ACTIONS(1600),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1600),
    [sym_super] = ACTIONS(1600),
    [sym_crate] = ACTIONS(1600),
    [sym_metavariable] = ACTIONS(1598),
    [sym_raw_string_literal] = ACTIONS(1598),
    [sym_float_literal] = ACTIONS(1598),
    [sym_block_comment] = ACTIONS(3),
  },
  [380] = {
    [ts_builtin_sym_end] = ACTIONS(1602),
    [sym_identifier] = ACTIONS(1604),
    [anon_sym_SEMI] = ACTIONS(1602),
    [anon_sym_macro_rules_BANG] = ACTIONS(1602),
    [anon_sym_LPAREN] = ACTIONS(1602),
    [anon_sym_LBRACE] = ACTIONS(1602),
    [anon_sym_RBRACE] = ACTIONS(1602),
    [anon_sym_LBRACK] = ACTIONS(1602),
    [anon_sym_STAR] = ACTIONS(1602),
    [anon_sym_u8] = ACTIONS(1604),
    [anon_sym_i8] = ACTIONS(1604),
    [anon_sym_u16] = ACTIONS(1604),
    [anon_sym_i16] = ACTIONS(1604),
    [anon_sym_u32] = ACTIONS(1604),
    [anon_sym_i32] = ACTIONS(1604),
    [anon_sym_u64] = ACTIONS(1604),
    [anon_sym_i64] = ACTIONS(1604),
    [anon_sym_u128] = ACTIONS(1604),
    [anon_sym_i128] = ACTIONS(1604),
    [anon_sym_isize] = ACTIONS(1604),
    [anon_sym_usize] = ACTIONS(1604),
    [anon_sym_f32] = ACTIONS(1604),
    [anon_sym_f64] = ACTIONS(1604),
    [anon_sym_bool] = ACTIONS(1604),
    [anon_sym_str] = ACTIONS(1604),
    [anon_sym_char] = ACTIONS(1604),
    [anon_sym_SQUOTE] = ACTIONS(1604),
    [anon_sym_async] = ACTIONS(1604),
    [anon_sym_break] = ACTIONS(1604),
    [anon_sym_const] = ACTIONS(1604),
    [anon_sym_continue] = ACTIONS(1604),
    [anon_sym_default] = ACTIONS(1604),
    [anon_sym_enum] = ACTIONS(1604),
    [anon_sym_fn] = ACTIONS(1604),
    [anon_sym_for] = ACTIONS(1604),
    [anon_sym_if] = ACTIONS(1604),
    [anon_sym_impl] = ACTIONS(1604),
    [anon_sym_let] = ACTIONS(1604),
    [anon_sym_loop] = ACTIONS(1604),
    [anon_sym_match] = ACTIONS(1604),
    [anon_sym_mod] = ACTIONS(1604),
    [anon_sym_pub] = ACTIONS(1604),
    [anon_sym_return] = ACTIONS(1604),
    [anon_sym_static] = ACTIONS(1604),
    [anon_sym_struct] = ACTIONS(1604),
    [anon_sym_trait] = ACTIONS(1604),
    [anon_sym_type] = ACTIONS(1604),
    [anon_sym_union] = ACTIONS(1604),
    [anon_sym_unsafe] = ACTIONS(1604),
    [anon_sym_use] = ACTIONS(1604),
    [anon_sym_while] = ACTIONS(1604),
    [anon_sym_POUND] = ACTIONS(1602),
    [anon_sym_BANG] = ACTIONS(1602),
    [anon_sym_extern] = ACTIONS(1604),
    [anon_sym_LT] = ACTIONS(1602),
    [anon_sym_COLON_COLON] = ACTIONS(1602),
    [anon_sym_AMP] = ACTIONS(1602),
    [anon_sym_DOT_DOT] = ACTIONS(1602),
    [anon_sym_DASH] = ACTIONS(1602),
    [anon_sym_PIPE] = ACTIONS(1602),
    [anon_sym_yield] = ACTIONS(1604),
    [anon_sym_move] = ACTIONS(1604),
    [sym_integer_literal] = ACTIONS(1602),
    [aux_sym_string_literal_token1] = ACTIONS(1602),
    [sym_char_literal] = ACTIONS(1602),
    [anon_sym_true] = ACTIONS(1604),
    [anon_sym_false] = ACTIONS(1604),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1604),
    [sym_super] = ACTIONS(1604),
    [sym_crate] = ACTIONS(1604),
    [sym_metavariable] = ACTIONS(1602),
    [sym_raw_string_literal] = ACTIONS(1602),
    [sym_float_literal] = ACTIONS(1602),
    [sym_block_comment] = ACTIONS(3),
  },
  [381] = {
    [ts_builtin_sym_end] = ACTIONS(1606),
    [sym_identifier] = ACTIONS(1608),
    [anon_sym_SEMI] = ACTIONS(1606),
    [anon_sym_macro_rules_BANG] = ACTIONS(1606),
    [anon_sym_LPAREN] = ACTIONS(1606),
    [anon_sym_LBRACE] = ACTIONS(1606),
    [anon_sym_RBRACE] = ACTIONS(1606),
    [anon_sym_LBRACK] = ACTIONS(1606),
    [anon_sym_STAR] = ACTIONS(1606),
    [anon_sym_u8] = ACTIONS(1608),
    [anon_sym_i8] = ACTIONS(1608),
    [anon_sym_u16] = ACTIONS(1608),
    [anon_sym_i16] = ACTIONS(1608),
    [anon_sym_u32] = ACTIONS(1608),
    [anon_sym_i32] = ACTIONS(1608),
    [anon_sym_u64] = ACTIONS(1608),
    [anon_sym_i64] = ACTIONS(1608),
    [anon_sym_u128] = ACTIONS(1608),
    [anon_sym_i128] = ACTIONS(1608),
    [anon_sym_isize] = ACTIONS(1608),
    [anon_sym_usize] = ACTIONS(1608),
    [anon_sym_f32] = ACTIONS(1608),
    [anon_sym_f64] = ACTIONS(1608),
    [anon_sym_bool] = ACTIONS(1608),
    [anon_sym_str] = ACTIONS(1608),
    [anon_sym_char] = ACTIONS(1608),
    [anon_sym_SQUOTE] = ACTIONS(1608),
    [anon_sym_async] = ACTIONS(1608),
    [anon_sym_break] = ACTIONS(1608),
    [anon_sym_const] = ACTIONS(1608),
    [anon_sym_continue] = ACTIONS(1608),
    [anon_sym_default] = ACTIONS(1608),
    [anon_sym_enum] = ACTIONS(1608),
    [anon_sym_fn] = ACTIONS(1608),
    [anon_sym_for] = ACTIONS(1608),
    [anon_sym_if] = ACTIONS(1608),
    [anon_sym_impl] = ACTIONS(1608),
    [anon_sym_let] = ACTIONS(1608),
    [anon_sym_loop] = ACTIONS(1608),
    [anon_sym_match] = ACTIONS(1608),
    [anon_sym_mod] = ACTIONS(1608),
    [anon_sym_pub] = ACTIONS(1608),
    [anon_sym_return] = ACTIONS(1608),
    [anon_sym_static] = ACTIONS(1608),
    [anon_sym_struct] = ACTIONS(1608),
    [anon_sym_trait] = ACTIONS(1608),
    [anon_sym_type] = ACTIONS(1608),
    [anon_sym_union] = ACTIONS(1608),
    [anon_sym_unsafe] = ACTIONS(1608),
    [anon_sym_use] = ACTIONS(1608),
    [anon_sym_while] = ACTIONS(1608),
    [anon_sym_POUND] = ACTIONS(1606),
    [anon_sym_BANG] = ACTIONS(1606),
    [anon_sym_extern] = ACTIONS(1608),
    [anon_sym_LT] = ACTIONS(1606),
    [anon_sym_COLON_COLON] = ACTIONS(1606),
    [anon_sym_AMP] = ACTIONS(1606),
    [anon_sym_DOT_DOT] = ACTIONS(1606),
    [anon_sym_DASH] = ACTIONS(1606),
    [anon_sym_PIPE] = ACTIONS(1606),
    [anon_sym_yield] = ACTIONS(1608),
    [anon_sym_move] = ACTIONS(1608),
    [sym_integer_literal] = ACTIONS(1606),
    [aux_sym_string_literal_token1] = ACTIONS(1606),
    [sym_char_literal] = ACTIONS(1606),
    [anon_sym_true] = ACTIONS(1608),
    [anon_sym_false] = ACTIONS(1608),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1608),
    [sym_super] = ACTIONS(1608),
    [sym_crate] = ACTIONS(1608),
    [sym_metavariable] = ACTIONS(1606),
    [sym_raw_string_literal] = ACTIONS(1606),
    [sym_float_literal] = ACTIONS(1606),
    [sym_block_comment] = ACTIONS(3),
  },
  [382] = {
    [ts_builtin_sym_end] = ACTIONS(1610),
    [sym_identifier] = ACTIONS(1612),
    [anon_sym_SEMI] = ACTIONS(1610),
    [anon_sym_macro_rules_BANG] = ACTIONS(1610),
    [anon_sym_LPAREN] = ACTIONS(1610),
    [anon_sym_LBRACE] = ACTIONS(1610),
    [anon_sym_RBRACE] = ACTIONS(1610),
    [anon_sym_LBRACK] = ACTIONS(1610),
    [anon_sym_STAR] = ACTIONS(1610),
    [anon_sym_u8] = ACTIONS(1612),
    [anon_sym_i8] = ACTIONS(1612),
    [anon_sym_u16] = ACTIONS(1612),
    [anon_sym_i16] = ACTIONS(1612),
    [anon_sym_u32] = ACTIONS(1612),
    [anon_sym_i32] = ACTIONS(1612),
    [anon_sym_u64] = ACTIONS(1612),
    [anon_sym_i64] = ACTIONS(1612),
    [anon_sym_u128] = ACTIONS(1612),
    [anon_sym_i128] = ACTIONS(1612),
    [anon_sym_isize] = ACTIONS(1612),
    [anon_sym_usize] = ACTIONS(1612),
    [anon_sym_f32] = ACTIONS(1612),
    [anon_sym_f64] = ACTIONS(1612),
    [anon_sym_bool] = ACTIONS(1612),
    [anon_sym_str] = ACTIONS(1612),
    [anon_sym_char] = ACTIONS(1612),
    [anon_sym_SQUOTE] = ACTIONS(1612),
    [anon_sym_async] = ACTIONS(1612),
    [anon_sym_break] = ACTIONS(1612),
    [anon_sym_const] = ACTIONS(1612),
    [anon_sym_continue] = ACTIONS(1612),
    [anon_sym_default] = ACTIONS(1612),
    [anon_sym_enum] = ACTIONS(1612),
    [anon_sym_fn] = ACTIONS(1612),
    [anon_sym_for] = ACTIONS(1612),
    [anon_sym_if] = ACTIONS(1612),
    [anon_sym_impl] = ACTIONS(1612),
    [anon_sym_let] = ACTIONS(1612),
    [anon_sym_loop] = ACTIONS(1612),
    [anon_sym_match] = ACTIONS(1612),
    [anon_sym_mod] = ACTIONS(1612),
    [anon_sym_pub] = ACTIONS(1612),
    [anon_sym_return] = ACTIONS(1612),
    [anon_sym_static] = ACTIONS(1612),
    [anon_sym_struct] = ACTIONS(1612),
    [anon_sym_trait] = ACTIONS(1612),
    [anon_sym_type] = ACTIONS(1612),
    [anon_sym_union] = ACTIONS(1612),
    [anon_sym_unsafe] = ACTIONS(1612),
    [anon_sym_use] = ACTIONS(1612),
    [anon_sym_while] = ACTIONS(1612),
    [anon_sym_POUND] = ACTIONS(1610),
    [anon_sym_BANG] = ACTIONS(1610),
    [anon_sym_extern] = ACTIONS(1612),
    [anon_sym_LT] = ACTIONS(1610),
    [anon_sym_COLON_COLON] = ACTIONS(1610),
    [anon_sym_AMP] = ACTIONS(1610),
    [anon_sym_DOT_DOT] = ACTIONS(1610),
    [anon_sym_DASH] = ACTIONS(1610),
    [anon_sym_PIPE] = ACTIONS(1610),
    [anon_sym_yield] = ACTIONS(1612),
    [anon_sym_move] = ACTIONS(1612),
    [sym_integer_literal] = ACTIONS(1610),
    [aux_sym_string_literal_token1] = ACTIONS(1610),
    [sym_char_literal] = ACTIONS(1610),
    [anon_sym_true] = ACTIONS(1612),
    [anon_sym_false] = ACTIONS(1612),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1612),
    [sym_super] = ACTIONS(1612),
    [sym_crate] = ACTIONS(1612),
    [sym_metavariable] = ACTIONS(1610),
    [sym_raw_string_literal] = ACTIONS(1610),
    [sym_float_literal] = ACTIONS(1610),
    [sym_block_comment] = ACTIONS(3),
  },
  [383] = {
    [ts_builtin_sym_end] = ACTIONS(1614),
    [sym_identifier] = ACTIONS(1616),
    [anon_sym_SEMI] = ACTIONS(1614),
    [anon_sym_macro_rules_BANG] = ACTIONS(1614),
    [anon_sym_LPAREN] = ACTIONS(1614),
    [anon_sym_LBRACE] = ACTIONS(1614),
    [anon_sym_RBRACE] = ACTIONS(1614),
    [anon_sym_LBRACK] = ACTIONS(1614),
    [anon_sym_STAR] = ACTIONS(1614),
    [anon_sym_u8] = ACTIONS(1616),
    [anon_sym_i8] = ACTIONS(1616),
    [anon_sym_u16] = ACTIONS(1616),
    [anon_sym_i16] = ACTIONS(1616),
    [anon_sym_u32] = ACTIONS(1616),
    [anon_sym_i32] = ACTIONS(1616),
    [anon_sym_u64] = ACTIONS(1616),
    [anon_sym_i64] = ACTIONS(1616),
    [anon_sym_u128] = ACTIONS(1616),
    [anon_sym_i128] = ACTIONS(1616),
    [anon_sym_isize] = ACTIONS(1616),
    [anon_sym_usize] = ACTIONS(1616),
    [anon_sym_f32] = ACTIONS(1616),
    [anon_sym_f64] = ACTIONS(1616),
    [anon_sym_bool] = ACTIONS(1616),
    [anon_sym_str] = ACTIONS(1616),
    [anon_sym_char] = ACTIONS(1616),
    [anon_sym_SQUOTE] = ACTIONS(1616),
    [anon_sym_async] = ACTIONS(1616),
    [anon_sym_break] = ACTIONS(1616),
    [anon_sym_const] = ACTIONS(1616),
    [anon_sym_continue] = ACTIONS(1616),
    [anon_sym_default] = ACTIONS(1616),
    [anon_sym_enum] = ACTIONS(1616),
    [anon_sym_fn] = ACTIONS(1616),
    [anon_sym_for] = ACTIONS(1616),
    [anon_sym_if] = ACTIONS(1616),
    [anon_sym_impl] = ACTIONS(1616),
    [anon_sym_let] = ACTIONS(1616),
    [anon_sym_loop] = ACTIONS(1616),
    [anon_sym_match] = ACTIONS(1616),
    [anon_sym_mod] = ACTIONS(1616),
    [anon_sym_pub] = ACTIONS(1616),
    [anon_sym_return] = ACTIONS(1616),
    [anon_sym_static] = ACTIONS(1616),
    [anon_sym_struct] = ACTIONS(1616),
    [anon_sym_trait] = ACTIONS(1616),
    [anon_sym_type] = ACTIONS(1616),
    [anon_sym_union] = ACTIONS(1616),
    [anon_sym_unsafe] = ACTIONS(1616),
    [anon_sym_use] = ACTIONS(1616),
    [anon_sym_while] = ACTIONS(1616),
    [anon_sym_POUND] = ACTIONS(1614),
    [anon_sym_BANG] = ACTIONS(1614),
    [anon_sym_extern] = ACTIONS(1616),
    [anon_sym_LT] = ACTIONS(1614),
    [anon_sym_COLON_COLON] = ACTIONS(1614),
    [anon_sym_AMP] = ACTIONS(1614),
    [anon_sym_DOT_DOT] = ACTIONS(1614),
    [anon_sym_DASH] = ACTIONS(1614),
    [anon_sym_PIPE] = ACTIONS(1614),
    [anon_sym_yield] = ACTIONS(1616),
    [anon_sym_move] = ACTIONS(1616),
    [sym_integer_literal] = ACTIONS(1614),
    [aux_sym_string_literal_token1] = ACTIONS(1614),
    [sym_char_literal] = ACTIONS(1614),
    [anon_sym_true] = ACTIONS(1616),
    [anon_sym_false] = ACTIONS(1616),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1616),
    [sym_super] = ACTIONS(1616),
    [sym_crate] = ACTIONS(1616),
    [sym_metavariable] = ACTIONS(1614),
    [sym_raw_string_literal] = ACTIONS(1614),
    [sym_float_literal] = ACTIONS(1614),
    [sym_block_comment] = ACTIONS(3),
  },
  [384] = {
    [ts_builtin_sym_end] = ACTIONS(1618),
    [sym_identifier] = ACTIONS(1620),
    [anon_sym_SEMI] = ACTIONS(1618),
    [anon_sym_macro_rules_BANG] = ACTIONS(1618),
    [anon_sym_LPAREN] = ACTIONS(1618),
    [anon_sym_LBRACE] = ACTIONS(1618),
    [anon_sym_RBRACE] = ACTIONS(1618),
    [anon_sym_LBRACK] = ACTIONS(1618),
    [anon_sym_STAR] = ACTIONS(1618),
    [anon_sym_u8] = ACTIONS(1620),
    [anon_sym_i8] = ACTIONS(1620),
    [anon_sym_u16] = ACTIONS(1620),
    [anon_sym_i16] = ACTIONS(1620),
    [anon_sym_u32] = ACTIONS(1620),
    [anon_sym_i32] = ACTIONS(1620),
    [anon_sym_u64] = ACTIONS(1620),
    [anon_sym_i64] = ACTIONS(1620),
    [anon_sym_u128] = ACTIONS(1620),
    [anon_sym_i128] = ACTIONS(1620),
    [anon_sym_isize] = ACTIONS(1620),
    [anon_sym_usize] = ACTIONS(1620),
    [anon_sym_f32] = ACTIONS(1620),
    [anon_sym_f64] = ACTIONS(1620),
    [anon_sym_bool] = ACTIONS(1620),
    [anon_sym_str] = ACTIONS(1620),
    [anon_sym_char] = ACTIONS(1620),
    [anon_sym_SQUOTE] = ACTIONS(1620),
    [anon_sym_async] = ACTIONS(1620),
    [anon_sym_break] = ACTIONS(1620),
    [anon_sym_const] = ACTIONS(1620),
    [anon_sym_continue] = ACTIONS(1620),
    [anon_sym_default] = ACTIONS(1620),
    [anon_sym_enum] = ACTIONS(1620),
    [anon_sym_fn] = ACTIONS(1620),
    [anon_sym_for] = ACTIONS(1620),
    [anon_sym_if] = ACTIONS(1620),
    [anon_sym_impl] = ACTIONS(1620),
    [anon_sym_let] = ACTIONS(1620),
    [anon_sym_loop] = ACTIONS(1620),
    [anon_sym_match] = ACTIONS(1620),
    [anon_sym_mod] = ACTIONS(1620),
    [anon_sym_pub] = ACTIONS(1620),
    [anon_sym_return] = ACTIONS(1620),
    [anon_sym_static] = ACTIONS(1620),
    [anon_sym_struct] = ACTIONS(1620),
    [anon_sym_trait] = ACTIONS(1620),
    [anon_sym_type] = ACTIONS(1620),
    [anon_sym_union] = ACTIONS(1620),
    [anon_sym_unsafe] = ACTIONS(1620),
    [anon_sym_use] = ACTIONS(1620),
    [anon_sym_while] = ACTIONS(1620),
    [anon_sym_POUND] = ACTIONS(1618),
    [anon_sym_BANG] = ACTIONS(1618),
    [anon_sym_extern] = ACTIONS(1620),
    [anon_sym_LT] = ACTIONS(1618),
    [anon_sym_COLON_COLON] = ACTIONS(1618),
    [anon_sym_AMP] = ACTIONS(1618),
    [anon_sym_DOT_DOT] = ACTIONS(1618),
    [anon_sym_DASH] = ACTIONS(1618),
    [anon_sym_PIPE] = ACTIONS(1618),
    [anon_sym_yield] = ACTIONS(1620),
    [anon_sym_move] = ACTIONS(1620),
    [sym_integer_literal] = ACTIONS(1618),
    [aux_sym_string_literal_token1] = ACTIONS(1618),
    [sym_char_literal] = ACTIONS(1618),
    [anon_sym_true] = ACTIONS(1620),
    [anon_sym_false] = ACTIONS(1620),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1620),
    [sym_super] = ACTIONS(1620),
    [sym_crate] = ACTIONS(1620),
    [sym_metavariable] = ACTIONS(1618),
    [sym_raw_string_literal] = ACTIONS(1618),
    [sym_float_literal] = ACTIONS(1618),
    [sym_block_comment] = ACTIONS(3),
  },
  [385] = {
    [ts_builtin_sym_end] = ACTIONS(1622),
    [sym_identifier] = ACTIONS(1624),
    [anon_sym_SEMI] = ACTIONS(1622),
    [anon_sym_macro_rules_BANG] = ACTIONS(1622),
    [anon_sym_LPAREN] = ACTIONS(1622),
    [anon_sym_LBRACE] = ACTIONS(1622),
    [anon_sym_RBRACE] = ACTIONS(1622),
    [anon_sym_LBRACK] = ACTIONS(1622),
    [anon_sym_STAR] = ACTIONS(1622),
    [anon_sym_u8] = ACTIONS(1624),
    [anon_sym_i8] = ACTIONS(1624),
    [anon_sym_u16] = ACTIONS(1624),
    [anon_sym_i16] = ACTIONS(1624),
    [anon_sym_u32] = ACTIONS(1624),
    [anon_sym_i32] = ACTIONS(1624),
    [anon_sym_u64] = ACTIONS(1624),
    [anon_sym_i64] = ACTIONS(1624),
    [anon_sym_u128] = ACTIONS(1624),
    [anon_sym_i128] = ACTIONS(1624),
    [anon_sym_isize] = ACTIONS(1624),
    [anon_sym_usize] = ACTIONS(1624),
    [anon_sym_f32] = ACTIONS(1624),
    [anon_sym_f64] = ACTIONS(1624),
    [anon_sym_bool] = ACTIONS(1624),
    [anon_sym_str] = ACTIONS(1624),
    [anon_sym_char] = ACTIONS(1624),
    [anon_sym_SQUOTE] = ACTIONS(1624),
    [anon_sym_async] = ACTIONS(1624),
    [anon_sym_break] = ACTIONS(1624),
    [anon_sym_const] = ACTIONS(1624),
    [anon_sym_continue] = ACTIONS(1624),
    [anon_sym_default] = ACTIONS(1624),
    [anon_sym_enum] = ACTIONS(1624),
    [anon_sym_fn] = ACTIONS(1624),
    [anon_sym_for] = ACTIONS(1624),
    [anon_sym_if] = ACTIONS(1624),
    [anon_sym_impl] = ACTIONS(1624),
    [anon_sym_let] = ACTIONS(1624),
    [anon_sym_loop] = ACTIONS(1624),
    [anon_sym_match] = ACTIONS(1624),
    [anon_sym_mod] = ACTIONS(1624),
    [anon_sym_pub] = ACTIONS(1624),
    [anon_sym_return] = ACTIONS(1624),
    [anon_sym_static] = ACTIONS(1624),
    [anon_sym_struct] = ACTIONS(1624),
    [anon_sym_trait] = ACTIONS(1624),
    [anon_sym_type] = ACTIONS(1624),
    [anon_sym_union] = ACTIONS(1624),
    [anon_sym_unsafe] = ACTIONS(1624),
    [anon_sym_use] = ACTIONS(1624),
    [anon_sym_while] = ACTIONS(1624),
    [anon_sym_POUND] = ACTIONS(1622),
    [anon_sym_BANG] = ACTIONS(1622),
    [anon_sym_extern] = ACTIONS(1624),
    [anon_sym_LT] = ACTIONS(1622),
    [anon_sym_COLON_COLON] = ACTIONS(1622),
    [anon_sym_AMP] = ACTIONS(1622),
    [anon_sym_DOT_DOT] = ACTIONS(1622),
    [anon_sym_DASH] = ACTIONS(1622),
    [anon_sym_PIPE] = ACTIONS(1622),
    [anon_sym_yield] = ACTIONS(1624),
    [anon_sym_move] = ACTIONS(1624),
    [sym_integer_literal] = ACTIONS(1622),
    [aux_sym_string_literal_token1] = ACTIONS(1622),
    [sym_char_literal] = ACTIONS(1622),
    [anon_sym_true] = ACTIONS(1624),
    [anon_sym_false] = ACTIONS(1624),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1624),
    [sym_super] = ACTIONS(1624),
    [sym_crate] = ACTIONS(1624),
    [sym_metavariable] = ACTIONS(1622),
    [sym_raw_string_literal] = ACTIONS(1622),
    [sym_float_literal] = ACTIONS(1622),
    [sym_block_comment] = ACTIONS(3),
  },
  [386] = {
    [ts_builtin_sym_end] = ACTIONS(1626),
    [sym_identifier] = ACTIONS(1628),
    [anon_sym_SEMI] = ACTIONS(1626),
    [anon_sym_macro_rules_BANG] = ACTIONS(1626),
    [anon_sym_LPAREN] = ACTIONS(1626),
    [anon_sym_LBRACE] = ACTIONS(1626),
    [anon_sym_RBRACE] = ACTIONS(1626),
    [anon_sym_LBRACK] = ACTIONS(1626),
    [anon_sym_STAR] = ACTIONS(1626),
    [anon_sym_u8] = ACTIONS(1628),
    [anon_sym_i8] = ACTIONS(1628),
    [anon_sym_u16] = ACTIONS(1628),
    [anon_sym_i16] = ACTIONS(1628),
    [anon_sym_u32] = ACTIONS(1628),
    [anon_sym_i32] = ACTIONS(1628),
    [anon_sym_u64] = ACTIONS(1628),
    [anon_sym_i64] = ACTIONS(1628),
    [anon_sym_u128] = ACTIONS(1628),
    [anon_sym_i128] = ACTIONS(1628),
    [anon_sym_isize] = ACTIONS(1628),
    [anon_sym_usize] = ACTIONS(1628),
    [anon_sym_f32] = ACTIONS(1628),
    [anon_sym_f64] = ACTIONS(1628),
    [anon_sym_bool] = ACTIONS(1628),
    [anon_sym_str] = ACTIONS(1628),
    [anon_sym_char] = ACTIONS(1628),
    [anon_sym_SQUOTE] = ACTIONS(1628),
    [anon_sym_async] = ACTIONS(1628),
    [anon_sym_break] = ACTIONS(1628),
    [anon_sym_const] = ACTIONS(1628),
    [anon_sym_continue] = ACTIONS(1628),
    [anon_sym_default] = ACTIONS(1628),
    [anon_sym_enum] = ACTIONS(1628),
    [anon_sym_fn] = ACTIONS(1628),
    [anon_sym_for] = ACTIONS(1628),
    [anon_sym_if] = ACTIONS(1628),
    [anon_sym_impl] = ACTIONS(1628),
    [anon_sym_let] = ACTIONS(1628),
    [anon_sym_loop] = ACTIONS(1628),
    [anon_sym_match] = ACTIONS(1628),
    [anon_sym_mod] = ACTIONS(1628),
    [anon_sym_pub] = ACTIONS(1628),
    [anon_sym_return] = ACTIONS(1628),
    [anon_sym_static] = ACTIONS(1628),
    [anon_sym_struct] = ACTIONS(1628),
    [anon_sym_trait] = ACTIONS(1628),
    [anon_sym_type] = ACTIONS(1628),
    [anon_sym_union] = ACTIONS(1628),
    [anon_sym_unsafe] = ACTIONS(1628),
    [anon_sym_use] = ACTIONS(1628),
    [anon_sym_while] = ACTIONS(1628),
    [anon_sym_POUND] = ACTIONS(1626),
    [anon_sym_BANG] = ACTIONS(1626),
    [anon_sym_extern] = ACTIONS(1628),
    [anon_sym_LT] = ACTIONS(1626),
    [anon_sym_COLON_COLON] = ACTIONS(1626),
    [anon_sym_AMP] = ACTIONS(1626),
    [anon_sym_DOT_DOT] = ACTIONS(1626),
    [anon_sym_DASH] = ACTIONS(1626),
    [anon_sym_PIPE] = ACTIONS(1626),
    [anon_sym_yield] = ACTIONS(1628),
    [anon_sym_move] = ACTIONS(1628),
    [sym_integer_literal] = ACTIONS(1626),
    [aux_sym_string_literal_token1] = ACTIONS(1626),
    [sym_char_literal] = ACTIONS(1626),
    [anon_sym_true] = ACTIONS(1628),
    [anon_sym_false] = ACTIONS(1628),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1628),
    [sym_super] = ACTIONS(1628),
    [sym_crate] = ACTIONS(1628),
    [sym_metavariable] = ACTIONS(1626),
    [sym_raw_string_literal] = ACTIONS(1626),
    [sym_float_literal] = ACTIONS(1626),
    [sym_block_comment] = ACTIONS(3),
  },
  [387] = {
    [ts_builtin_sym_end] = ACTIONS(1630),
    [sym_identifier] = ACTIONS(1632),
    [anon_sym_SEMI] = ACTIONS(1630),
    [anon_sym_macro_rules_BANG] = ACTIONS(1630),
    [anon_sym_LPAREN] = ACTIONS(1630),
    [anon_sym_LBRACE] = ACTIONS(1630),
    [anon_sym_RBRACE] = ACTIONS(1630),
    [anon_sym_LBRACK] = ACTIONS(1630),
    [anon_sym_STAR] = ACTIONS(1630),
    [anon_sym_u8] = ACTIONS(1632),
    [anon_sym_i8] = ACTIONS(1632),
    [anon_sym_u16] = ACTIONS(1632),
    [anon_sym_i16] = ACTIONS(1632),
    [anon_sym_u32] = ACTIONS(1632),
    [anon_sym_i32] = ACTIONS(1632),
    [anon_sym_u64] = ACTIONS(1632),
    [anon_sym_i64] = ACTIONS(1632),
    [anon_sym_u128] = ACTIONS(1632),
    [anon_sym_i128] = ACTIONS(1632),
    [anon_sym_isize] = ACTIONS(1632),
    [anon_sym_usize] = ACTIONS(1632),
    [anon_sym_f32] = ACTIONS(1632),
    [anon_sym_f64] = ACTIONS(1632),
    [anon_sym_bool] = ACTIONS(1632),
    [anon_sym_str] = ACTIONS(1632),
    [anon_sym_char] = ACTIONS(1632),
    [anon_sym_SQUOTE] = ACTIONS(1632),
    [anon_sym_async] = ACTIONS(1632),
    [anon_sym_break] = ACTIONS(1632),
    [anon_sym_const] = ACTIONS(1632),
    [anon_sym_continue] = ACTIONS(1632),
    [anon_sym_default] = ACTIONS(1632),
    [anon_sym_enum] = ACTIONS(1632),
    [anon_sym_fn] = ACTIONS(1632),
    [anon_sym_for] = ACTIONS(1632),
    [anon_sym_if] = ACTIONS(1632),
    [anon_sym_impl] = ACTIONS(1632),
    [anon_sym_let] = ACTIONS(1632),
    [anon_sym_loop] = ACTIONS(1632),
    [anon_sym_match] = ACTIONS(1632),
    [anon_sym_mod] = ACTIONS(1632),
    [anon_sym_pub] = ACTIONS(1632),
    [anon_sym_return] = ACTIONS(1632),
    [anon_sym_static] = ACTIONS(1632),
    [anon_sym_struct] = ACTIONS(1632),
    [anon_sym_trait] = ACTIONS(1632),
    [anon_sym_type] = ACTIONS(1632),
    [anon_sym_union] = ACTIONS(1632),
    [anon_sym_unsafe] = ACTIONS(1632),
    [anon_sym_use] = ACTIONS(1632),
    [anon_sym_while] = ACTIONS(1632),
    [anon_sym_POUND] = ACTIONS(1630),
    [anon_sym_BANG] = ACTIONS(1630),
    [anon_sym_extern] = ACTIONS(1632),
    [anon_sym_LT] = ACTIONS(1630),
    [anon_sym_COLON_COLON] = ACTIONS(1630),
    [anon_sym_AMP] = ACTIONS(1630),
    [anon_sym_DOT_DOT] = ACTIONS(1630),
    [anon_sym_DASH] = ACTIONS(1630),
    [anon_sym_PIPE] = ACTIONS(1630),
    [anon_sym_yield] = ACTIONS(1632),
    [anon_sym_move] = ACTIONS(1632),
    [sym_integer_literal] = ACTIONS(1630),
    [aux_sym_string_literal_token1] = ACTIONS(1630),
    [sym_char_literal] = ACTIONS(1630),
    [anon_sym_true] = ACTIONS(1632),
    [anon_sym_false] = ACTIONS(1632),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1632),
    [sym_super] = ACTIONS(1632),
    [sym_crate] = ACTIONS(1632),
    [sym_metavariable] = ACTIONS(1630),
    [sym_raw_string_literal] = ACTIONS(1630),
    [sym_float_literal] = ACTIONS(1630),
    [sym_block_comment] = ACTIONS(3),
  },
  [388] = {
    [ts_builtin_sym_end] = ACTIONS(1634),
    [sym_identifier] = ACTIONS(1636),
    [anon_sym_SEMI] = ACTIONS(1634),
    [anon_sym_macro_rules_BANG] = ACTIONS(1634),
    [anon_sym_LPAREN] = ACTIONS(1634),
    [anon_sym_LBRACE] = ACTIONS(1634),
    [anon_sym_RBRACE] = ACTIONS(1634),
    [anon_sym_LBRACK] = ACTIONS(1634),
    [anon_sym_STAR] = ACTIONS(1634),
    [anon_sym_u8] = ACTIONS(1636),
    [anon_sym_i8] = ACTIONS(1636),
    [anon_sym_u16] = ACTIONS(1636),
    [anon_sym_i16] = ACTIONS(1636),
    [anon_sym_u32] = ACTIONS(1636),
    [anon_sym_i32] = ACTIONS(1636),
    [anon_sym_u64] = ACTIONS(1636),
    [anon_sym_i64] = ACTIONS(1636),
    [anon_sym_u128] = ACTIONS(1636),
    [anon_sym_i128] = ACTIONS(1636),
    [anon_sym_isize] = ACTIONS(1636),
    [anon_sym_usize] = ACTIONS(1636),
    [anon_sym_f32] = ACTIONS(1636),
    [anon_sym_f64] = ACTIONS(1636),
    [anon_sym_bool] = ACTIONS(1636),
    [anon_sym_str] = ACTIONS(1636),
    [anon_sym_char] = ACTIONS(1636),
    [anon_sym_SQUOTE] = ACTIONS(1636),
    [anon_sym_async] = ACTIONS(1636),
    [anon_sym_break] = ACTIONS(1636),
    [anon_sym_const] = ACTIONS(1636),
    [anon_sym_continue] = ACTIONS(1636),
    [anon_sym_default] = ACTIONS(1636),
    [anon_sym_enum] = ACTIONS(1636),
    [anon_sym_fn] = ACTIONS(1636),
    [anon_sym_for] = ACTIONS(1636),
    [anon_sym_if] = ACTIONS(1636),
    [anon_sym_impl] = ACTIONS(1636),
    [anon_sym_let] = ACTIONS(1636),
    [anon_sym_loop] = ACTIONS(1636),
    [anon_sym_match] = ACTIONS(1636),
    [anon_sym_mod] = ACTIONS(1636),
    [anon_sym_pub] = ACTIONS(1636),
    [anon_sym_return] = ACTIONS(1636),
    [anon_sym_static] = ACTIONS(1636),
    [anon_sym_struct] = ACTIONS(1636),
    [anon_sym_trait] = ACTIONS(1636),
    [anon_sym_type] = ACTIONS(1636),
    [anon_sym_union] = ACTIONS(1636),
    [anon_sym_unsafe] = ACTIONS(1636),
    [anon_sym_use] = ACTIONS(1636),
    [anon_sym_while] = ACTIONS(1636),
    [anon_sym_POUND] = ACTIONS(1634),
    [anon_sym_BANG] = ACTIONS(1634),
    [anon_sym_extern] = ACTIONS(1636),
    [anon_sym_LT] = ACTIONS(1634),
    [anon_sym_COLON_COLON] = ACTIONS(1634),
    [anon_sym_AMP] = ACTIONS(1634),
    [anon_sym_DOT_DOT] = ACTIONS(1634),
    [anon_sym_DASH] = ACTIONS(1634),
    [anon_sym_PIPE] = ACTIONS(1634),
    [anon_sym_yield] = ACTIONS(1636),
    [anon_sym_move] = ACTIONS(1636),
    [sym_integer_literal] = ACTIONS(1634),
    [aux_sym_string_literal_token1] = ACTIONS(1634),
    [sym_char_literal] = ACTIONS(1634),
    [anon_sym_true] = ACTIONS(1636),
    [anon_sym_false] = ACTIONS(1636),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1636),
    [sym_super] = ACTIONS(1636),
    [sym_crate] = ACTIONS(1636),
    [sym_metavariable] = ACTIONS(1634),
    [sym_raw_string_literal] = ACTIONS(1634),
    [sym_float_literal] = ACTIONS(1634),
    [sym_block_comment] = ACTIONS(3),
  },
  [389] = {
    [ts_builtin_sym_end] = ACTIONS(1638),
    [sym_identifier] = ACTIONS(1640),
    [anon_sym_SEMI] = ACTIONS(1638),
    [anon_sym_macro_rules_BANG] = ACTIONS(1638),
    [anon_sym_LPAREN] = ACTIONS(1638),
    [anon_sym_LBRACE] = ACTIONS(1638),
    [anon_sym_RBRACE] = ACTIONS(1638),
    [anon_sym_LBRACK] = ACTIONS(1638),
    [anon_sym_STAR] = ACTIONS(1638),
    [anon_sym_u8] = ACTIONS(1640),
    [anon_sym_i8] = ACTIONS(1640),
    [anon_sym_u16] = ACTIONS(1640),
    [anon_sym_i16] = ACTIONS(1640),
    [anon_sym_u32] = ACTIONS(1640),
    [anon_sym_i32] = ACTIONS(1640),
    [anon_sym_u64] = ACTIONS(1640),
    [anon_sym_i64] = ACTIONS(1640),
    [anon_sym_u128] = ACTIONS(1640),
    [anon_sym_i128] = ACTIONS(1640),
    [anon_sym_isize] = ACTIONS(1640),
    [anon_sym_usize] = ACTIONS(1640),
    [anon_sym_f32] = ACTIONS(1640),
    [anon_sym_f64] = ACTIONS(1640),
    [anon_sym_bool] = ACTIONS(1640),
    [anon_sym_str] = ACTIONS(1640),
    [anon_sym_char] = ACTIONS(1640),
    [anon_sym_SQUOTE] = ACTIONS(1640),
    [anon_sym_async] = ACTIONS(1640),
    [anon_sym_break] = ACTIONS(1640),
    [anon_sym_const] = ACTIONS(1640),
    [anon_sym_continue] = ACTIONS(1640),
    [anon_sym_default] = ACTIONS(1640),
    [anon_sym_enum] = ACTIONS(1640),
    [anon_sym_fn] = ACTIONS(1640),
    [anon_sym_for] = ACTIONS(1640),
    [anon_sym_if] = ACTIONS(1640),
    [anon_sym_impl] = ACTIONS(1640),
    [anon_sym_let] = ACTIONS(1640),
    [anon_sym_loop] = ACTIONS(1640),
    [anon_sym_match] = ACTIONS(1640),
    [anon_sym_mod] = ACTIONS(1640),
    [anon_sym_pub] = ACTIONS(1640),
    [anon_sym_return] = ACTIONS(1640),
    [anon_sym_static] = ACTIONS(1640),
    [anon_sym_struct] = ACTIONS(1640),
    [anon_sym_trait] = ACTIONS(1640),
    [anon_sym_type] = ACTIONS(1640),
    [anon_sym_union] = ACTIONS(1640),
    [anon_sym_unsafe] = ACTIONS(1640),
    [anon_sym_use] = ACTIONS(1640),
    [anon_sym_while] = ACTIONS(1640),
    [anon_sym_POUND] = ACTIONS(1638),
    [anon_sym_BANG] = ACTIONS(1638),
    [anon_sym_extern] = ACTIONS(1640),
    [anon_sym_LT] = ACTIONS(1638),
    [anon_sym_COLON_COLON] = ACTIONS(1638),
    [anon_sym_AMP] = ACTIONS(1638),
    [anon_sym_DOT_DOT] = ACTIONS(1638),
    [anon_sym_DASH] = ACTIONS(1638),
    [anon_sym_PIPE] = ACTIONS(1638),
    [anon_sym_yield] = ACTIONS(1640),
    [anon_sym_move] = ACTIONS(1640),
    [sym_integer_literal] = ACTIONS(1638),
    [aux_sym_string_literal_token1] = ACTIONS(1638),
    [sym_char_literal] = ACTIONS(1638),
    [anon_sym_true] = ACTIONS(1640),
    [anon_sym_false] = ACTIONS(1640),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1640),
    [sym_super] = ACTIONS(1640),
    [sym_crate] = ACTIONS(1640),
    [sym_metavariable] = ACTIONS(1638),
    [sym_raw_string_literal] = ACTIONS(1638),
    [sym_float_literal] = ACTIONS(1638),
    [sym_block_comment] = ACTIONS(3),
  },
  [390] = {
    [ts_builtin_sym_end] = ACTIONS(1642),
    [sym_identifier] = ACTIONS(1644),
    [anon_sym_SEMI] = ACTIONS(1642),
    [anon_sym_macro_rules_BANG] = ACTIONS(1642),
    [anon_sym_LPAREN] = ACTIONS(1642),
    [anon_sym_LBRACE] = ACTIONS(1642),
    [anon_sym_RBRACE] = ACTIONS(1642),
    [anon_sym_LBRACK] = ACTIONS(1642),
    [anon_sym_STAR] = ACTIONS(1642),
    [anon_sym_u8] = ACTIONS(1644),
    [anon_sym_i8] = ACTIONS(1644),
    [anon_sym_u16] = ACTIONS(1644),
    [anon_sym_i16] = ACTIONS(1644),
    [anon_sym_u32] = ACTIONS(1644),
    [anon_sym_i32] = ACTIONS(1644),
    [anon_sym_u64] = ACTIONS(1644),
    [anon_sym_i64] = ACTIONS(1644),
    [anon_sym_u128] = ACTIONS(1644),
    [anon_sym_i128] = ACTIONS(1644),
    [anon_sym_isize] = ACTIONS(1644),
    [anon_sym_usize] = ACTIONS(1644),
    [anon_sym_f32] = ACTIONS(1644),
    [anon_sym_f64] = ACTIONS(1644),
    [anon_sym_bool] = ACTIONS(1644),
    [anon_sym_str] = ACTIONS(1644),
    [anon_sym_char] = ACTIONS(1644),
    [anon_sym_SQUOTE] = ACTIONS(1644),
    [anon_sym_async] = ACTIONS(1644),
    [anon_sym_break] = ACTIONS(1644),
    [anon_sym_const] = ACTIONS(1644),
    [anon_sym_continue] = ACTIONS(1644),
    [anon_sym_default] = ACTIONS(1644),
    [anon_sym_enum] = ACTIONS(1644),
    [anon_sym_fn] = ACTIONS(1644),
    [anon_sym_for] = ACTIONS(1644),
    [anon_sym_if] = ACTIONS(1644),
    [anon_sym_impl] = ACTIONS(1644),
    [anon_sym_let] = ACTIONS(1644),
    [anon_sym_loop] = ACTIONS(1644),
    [anon_sym_match] = ACTIONS(1644),
    [anon_sym_mod] = ACTIONS(1644),
    [anon_sym_pub] = ACTIONS(1644),
    [anon_sym_return] = ACTIONS(1644),
    [anon_sym_static] = ACTIONS(1644),
    [anon_sym_struct] = ACTIONS(1644),
    [anon_sym_trait] = ACTIONS(1644),
    [anon_sym_type] = ACTIONS(1644),
    [anon_sym_union] = ACTIONS(1644),
    [anon_sym_unsafe] = ACTIONS(1644),
    [anon_sym_use] = ACTIONS(1644),
    [anon_sym_while] = ACTIONS(1644),
    [anon_sym_POUND] = ACTIONS(1642),
    [anon_sym_BANG] = ACTIONS(1642),
    [anon_sym_extern] = ACTIONS(1644),
    [anon_sym_LT] = ACTIONS(1642),
    [anon_sym_COLON_COLON] = ACTIONS(1642),
    [anon_sym_AMP] = ACTIONS(1642),
    [anon_sym_DOT_DOT] = ACTIONS(1642),
    [anon_sym_DASH] = ACTIONS(1642),
    [anon_sym_PIPE] = ACTIONS(1642),
    [anon_sym_yield] = ACTIONS(1644),
    [anon_sym_move] = ACTIONS(1644),
    [sym_integer_literal] = ACTIONS(1642),
    [aux_sym_string_literal_token1] = ACTIONS(1642),
    [sym_char_literal] = ACTIONS(1642),
    [anon_sym_true] = ACTIONS(1644),
    [anon_sym_false] = ACTIONS(1644),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1644),
    [sym_super] = ACTIONS(1644),
    [sym_crate] = ACTIONS(1644),
    [sym_metavariable] = ACTIONS(1642),
    [sym_raw_string_literal] = ACTIONS(1642),
    [sym_float_literal] = ACTIONS(1642),
    [sym_block_comment] = ACTIONS(3),
  },
  [391] = {
    [sym_attribute_item] = STATE(547),
    [sym_bracketed_type] = STATE(2489),
    [sym_generic_type] = STATE(2488),
    [sym_generic_type_with_turbofish] = STATE(2487),
    [sym_macro_invocation] = STATE(1422),
    [sym_scoped_identifier] = STATE(1337),
    [sym_scoped_type_identifier] = STATE(2013),
    [sym_match_arm] = STATE(484),
    [sym_last_match_arm] = STATE(2338),
    [sym_match_pattern] = STATE(2387),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1981),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [aux_sym_enum_variant_list_repeat1] = STATE(547),
    [aux_sym_match_block_repeat1] = STATE(484),
    [sym_identifier] = ACTIONS(1524),
    [anon_sym_LPAREN] = ACTIONS(1526),
    [anon_sym_RBRACE] = ACTIONS(1646),
    [anon_sym_LBRACK] = ACTIONS(1530),
    [anon_sym_u8] = ACTIONS(1532),
    [anon_sym_i8] = ACTIONS(1532),
    [anon_sym_u16] = ACTIONS(1532),
    [anon_sym_i16] = ACTIONS(1532),
    [anon_sym_u32] = ACTIONS(1532),
    [anon_sym_i32] = ACTIONS(1532),
    [anon_sym_u64] = ACTIONS(1532),
    [anon_sym_i64] = ACTIONS(1532),
    [anon_sym_u128] = ACTIONS(1532),
    [anon_sym_i128] = ACTIONS(1532),
    [anon_sym_isize] = ACTIONS(1532),
    [anon_sym_usize] = ACTIONS(1532),
    [anon_sym_f32] = ACTIONS(1532),
    [anon_sym_f64] = ACTIONS(1532),
    [anon_sym_bool] = ACTIONS(1532),
    [anon_sym_str] = ACTIONS(1532),
    [anon_sym_char] = ACTIONS(1532),
    [anon_sym_const] = ACTIONS(1534),
    [anon_sym_default] = ACTIONS(1536),
    [anon_sym_union] = ACTIONS(1536),
    [anon_sym_POUND] = ACTIONS(370),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(1538),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(1540),
    [sym_mutable_specifier] = ACTIONS(796),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1542),
    [sym_super] = ACTIONS(1542),
    [sym_crate] = ACTIONS(1542),
    [sym_metavariable] = ACTIONS(1544),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [392] = {
    [ts_builtin_sym_end] = ACTIONS(1648),
    [sym_identifier] = ACTIONS(1650),
    [anon_sym_SEMI] = ACTIONS(1648),
    [anon_sym_macro_rules_BANG] = ACTIONS(1648),
    [anon_sym_LPAREN] = ACTIONS(1648),
    [anon_sym_LBRACE] = ACTIONS(1648),
    [anon_sym_RBRACE] = ACTIONS(1648),
    [anon_sym_LBRACK] = ACTIONS(1648),
    [anon_sym_STAR] = ACTIONS(1648),
    [anon_sym_u8] = ACTIONS(1650),
    [anon_sym_i8] = ACTIONS(1650),
    [anon_sym_u16] = ACTIONS(1650),
    [anon_sym_i16] = ACTIONS(1650),
    [anon_sym_u32] = ACTIONS(1650),
    [anon_sym_i32] = ACTIONS(1650),
    [anon_sym_u64] = ACTIONS(1650),
    [anon_sym_i64] = ACTIONS(1650),
    [anon_sym_u128] = ACTIONS(1650),
    [anon_sym_i128] = ACTIONS(1650),
    [anon_sym_isize] = ACTIONS(1650),
    [anon_sym_usize] = ACTIONS(1650),
    [anon_sym_f32] = ACTIONS(1650),
    [anon_sym_f64] = ACTIONS(1650),
    [anon_sym_bool] = ACTIONS(1650),
    [anon_sym_str] = ACTIONS(1650),
    [anon_sym_char] = ACTIONS(1650),
    [anon_sym_SQUOTE] = ACTIONS(1650),
    [anon_sym_async] = ACTIONS(1650),
    [anon_sym_break] = ACTIONS(1650),
    [anon_sym_const] = ACTIONS(1650),
    [anon_sym_continue] = ACTIONS(1650),
    [anon_sym_default] = ACTIONS(1650),
    [anon_sym_enum] = ACTIONS(1650),
    [anon_sym_fn] = ACTIONS(1650),
    [anon_sym_for] = ACTIONS(1650),
    [anon_sym_if] = ACTIONS(1650),
    [anon_sym_impl] = ACTIONS(1650),
    [anon_sym_let] = ACTIONS(1650),
    [anon_sym_loop] = ACTIONS(1650),
    [anon_sym_match] = ACTIONS(1650),
    [anon_sym_mod] = ACTIONS(1650),
    [anon_sym_pub] = ACTIONS(1650),
    [anon_sym_return] = ACTIONS(1650),
    [anon_sym_static] = ACTIONS(1650),
    [anon_sym_struct] = ACTIONS(1650),
    [anon_sym_trait] = ACTIONS(1650),
    [anon_sym_type] = ACTIONS(1650),
    [anon_sym_union] = ACTIONS(1650),
    [anon_sym_unsafe] = ACTIONS(1650),
    [anon_sym_use] = ACTIONS(1650),
    [anon_sym_while] = ACTIONS(1650),
    [anon_sym_POUND] = ACTIONS(1648),
    [anon_sym_BANG] = ACTIONS(1648),
    [anon_sym_extern] = ACTIONS(1650),
    [anon_sym_LT] = ACTIONS(1648),
    [anon_sym_COLON_COLON] = ACTIONS(1648),
    [anon_sym_AMP] = ACTIONS(1648),
    [anon_sym_DOT_DOT] = ACTIONS(1648),
    [anon_sym_DASH] = ACTIONS(1648),
    [anon_sym_PIPE] = ACTIONS(1648),
    [anon_sym_yield] = ACTIONS(1650),
    [anon_sym_move] = ACTIONS(1650),
    [sym_integer_literal] = ACTIONS(1648),
    [aux_sym_string_literal_token1] = ACTIONS(1648),
    [sym_char_literal] = ACTIONS(1648),
    [anon_sym_true] = ACTIONS(1650),
    [anon_sym_false] = ACTIONS(1650),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1650),
    [sym_super] = ACTIONS(1650),
    [sym_crate] = ACTIONS(1650),
    [sym_metavariable] = ACTIONS(1648),
    [sym_raw_string_literal] = ACTIONS(1648),
    [sym_float_literal] = ACTIONS(1648),
    [sym_block_comment] = ACTIONS(3),
  },
  [393] = {
    [ts_builtin_sym_end] = ACTIONS(1652),
    [sym_identifier] = ACTIONS(1654),
    [anon_sym_SEMI] = ACTIONS(1652),
    [anon_sym_macro_rules_BANG] = ACTIONS(1652),
    [anon_sym_LPAREN] = ACTIONS(1652),
    [anon_sym_LBRACE] = ACTIONS(1652),
    [anon_sym_RBRACE] = ACTIONS(1652),
    [anon_sym_LBRACK] = ACTIONS(1652),
    [anon_sym_STAR] = ACTIONS(1652),
    [anon_sym_u8] = ACTIONS(1654),
    [anon_sym_i8] = ACTIONS(1654),
    [anon_sym_u16] = ACTIONS(1654),
    [anon_sym_i16] = ACTIONS(1654),
    [anon_sym_u32] = ACTIONS(1654),
    [anon_sym_i32] = ACTIONS(1654),
    [anon_sym_u64] = ACTIONS(1654),
    [anon_sym_i64] = ACTIONS(1654),
    [anon_sym_u128] = ACTIONS(1654),
    [anon_sym_i128] = ACTIONS(1654),
    [anon_sym_isize] = ACTIONS(1654),
    [anon_sym_usize] = ACTIONS(1654),
    [anon_sym_f32] = ACTIONS(1654),
    [anon_sym_f64] = ACTIONS(1654),
    [anon_sym_bool] = ACTIONS(1654),
    [anon_sym_str] = ACTIONS(1654),
    [anon_sym_char] = ACTIONS(1654),
    [anon_sym_SQUOTE] = ACTIONS(1654),
    [anon_sym_async] = ACTIONS(1654),
    [anon_sym_break] = ACTIONS(1654),
    [anon_sym_const] = ACTIONS(1654),
    [anon_sym_continue] = ACTIONS(1654),
    [anon_sym_default] = ACTIONS(1654),
    [anon_sym_enum] = ACTIONS(1654),
    [anon_sym_fn] = ACTIONS(1654),
    [anon_sym_for] = ACTIONS(1654),
    [anon_sym_if] = ACTIONS(1654),
    [anon_sym_impl] = ACTIONS(1654),
    [anon_sym_let] = ACTIONS(1654),
    [anon_sym_loop] = ACTIONS(1654),
    [anon_sym_match] = ACTIONS(1654),
    [anon_sym_mod] = ACTIONS(1654),
    [anon_sym_pub] = ACTIONS(1654),
    [anon_sym_return] = ACTIONS(1654),
    [anon_sym_static] = ACTIONS(1654),
    [anon_sym_struct] = ACTIONS(1654),
    [anon_sym_trait] = ACTIONS(1654),
    [anon_sym_type] = ACTIONS(1654),
    [anon_sym_union] = ACTIONS(1654),
    [anon_sym_unsafe] = ACTIONS(1654),
    [anon_sym_use] = ACTIONS(1654),
    [anon_sym_while] = ACTIONS(1654),
    [anon_sym_POUND] = ACTIONS(1652),
    [anon_sym_BANG] = ACTIONS(1652),
    [anon_sym_extern] = ACTIONS(1654),
    [anon_sym_LT] = ACTIONS(1652),
    [anon_sym_COLON_COLON] = ACTIONS(1652),
    [anon_sym_AMP] = ACTIONS(1652),
    [anon_sym_DOT_DOT] = ACTIONS(1652),
    [anon_sym_DASH] = ACTIONS(1652),
    [anon_sym_PIPE] = ACTIONS(1652),
    [anon_sym_yield] = ACTIONS(1654),
    [anon_sym_move] = ACTIONS(1654),
    [sym_integer_literal] = ACTIONS(1652),
    [aux_sym_string_literal_token1] = ACTIONS(1652),
    [sym_char_literal] = ACTIONS(1652),
    [anon_sym_true] = ACTIONS(1654),
    [anon_sym_false] = ACTIONS(1654),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1654),
    [sym_super] = ACTIONS(1654),
    [sym_crate] = ACTIONS(1654),
    [sym_metavariable] = ACTIONS(1652),
    [sym_raw_string_literal] = ACTIONS(1652),
    [sym_float_literal] = ACTIONS(1652),
    [sym_block_comment] = ACTIONS(3),
  },
  [394] = {
    [ts_builtin_sym_end] = ACTIONS(1656),
    [sym_identifier] = ACTIONS(1658),
    [anon_sym_SEMI] = ACTIONS(1656),
    [anon_sym_macro_rules_BANG] = ACTIONS(1656),
    [anon_sym_LPAREN] = ACTIONS(1656),
    [anon_sym_LBRACE] = ACTIONS(1656),
    [anon_sym_RBRACE] = ACTIONS(1656),
    [anon_sym_LBRACK] = ACTIONS(1656),
    [anon_sym_STAR] = ACTIONS(1656),
    [anon_sym_u8] = ACTIONS(1658),
    [anon_sym_i8] = ACTIONS(1658),
    [anon_sym_u16] = ACTIONS(1658),
    [anon_sym_i16] = ACTIONS(1658),
    [anon_sym_u32] = ACTIONS(1658),
    [anon_sym_i32] = ACTIONS(1658),
    [anon_sym_u64] = ACTIONS(1658),
    [anon_sym_i64] = ACTIONS(1658),
    [anon_sym_u128] = ACTIONS(1658),
    [anon_sym_i128] = ACTIONS(1658),
    [anon_sym_isize] = ACTIONS(1658),
    [anon_sym_usize] = ACTIONS(1658),
    [anon_sym_f32] = ACTIONS(1658),
    [anon_sym_f64] = ACTIONS(1658),
    [anon_sym_bool] = ACTIONS(1658),
    [anon_sym_str] = ACTIONS(1658),
    [anon_sym_char] = ACTIONS(1658),
    [anon_sym_SQUOTE] = ACTIONS(1658),
    [anon_sym_async] = ACTIONS(1658),
    [anon_sym_break] = ACTIONS(1658),
    [anon_sym_const] = ACTIONS(1658),
    [anon_sym_continue] = ACTIONS(1658),
    [anon_sym_default] = ACTIONS(1658),
    [anon_sym_enum] = ACTIONS(1658),
    [anon_sym_fn] = ACTIONS(1658),
    [anon_sym_for] = ACTIONS(1658),
    [anon_sym_if] = ACTIONS(1658),
    [anon_sym_impl] = ACTIONS(1658),
    [anon_sym_let] = ACTIONS(1658),
    [anon_sym_loop] = ACTIONS(1658),
    [anon_sym_match] = ACTIONS(1658),
    [anon_sym_mod] = ACTIONS(1658),
    [anon_sym_pub] = ACTIONS(1658),
    [anon_sym_return] = ACTIONS(1658),
    [anon_sym_static] = ACTIONS(1658),
    [anon_sym_struct] = ACTIONS(1658),
    [anon_sym_trait] = ACTIONS(1658),
    [anon_sym_type] = ACTIONS(1658),
    [anon_sym_union] = ACTIONS(1658),
    [anon_sym_unsafe] = ACTIONS(1658),
    [anon_sym_use] = ACTIONS(1658),
    [anon_sym_while] = ACTIONS(1658),
    [anon_sym_POUND] = ACTIONS(1656),
    [anon_sym_BANG] = ACTIONS(1656),
    [anon_sym_extern] = ACTIONS(1658),
    [anon_sym_LT] = ACTIONS(1656),
    [anon_sym_COLON_COLON] = ACTIONS(1656),
    [anon_sym_AMP] = ACTIONS(1656),
    [anon_sym_DOT_DOT] = ACTIONS(1656),
    [anon_sym_DASH] = ACTIONS(1656),
    [anon_sym_PIPE] = ACTIONS(1656),
    [anon_sym_yield] = ACTIONS(1658),
    [anon_sym_move] = ACTIONS(1658),
    [sym_integer_literal] = ACTIONS(1656),
    [aux_sym_string_literal_token1] = ACTIONS(1656),
    [sym_char_literal] = ACTIONS(1656),
    [anon_sym_true] = ACTIONS(1658),
    [anon_sym_false] = ACTIONS(1658),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1658),
    [sym_super] = ACTIONS(1658),
    [sym_crate] = ACTIONS(1658),
    [sym_metavariable] = ACTIONS(1656),
    [sym_raw_string_literal] = ACTIONS(1656),
    [sym_float_literal] = ACTIONS(1656),
    [sym_block_comment] = ACTIONS(3),
  },
  [395] = {
    [ts_builtin_sym_end] = ACTIONS(1660),
    [sym_identifier] = ACTIONS(1662),
    [anon_sym_SEMI] = ACTIONS(1660),
    [anon_sym_macro_rules_BANG] = ACTIONS(1660),
    [anon_sym_LPAREN] = ACTIONS(1660),
    [anon_sym_LBRACE] = ACTIONS(1660),
    [anon_sym_RBRACE] = ACTIONS(1660),
    [anon_sym_LBRACK] = ACTIONS(1660),
    [anon_sym_STAR] = ACTIONS(1660),
    [anon_sym_u8] = ACTIONS(1662),
    [anon_sym_i8] = ACTIONS(1662),
    [anon_sym_u16] = ACTIONS(1662),
    [anon_sym_i16] = ACTIONS(1662),
    [anon_sym_u32] = ACTIONS(1662),
    [anon_sym_i32] = ACTIONS(1662),
    [anon_sym_u64] = ACTIONS(1662),
    [anon_sym_i64] = ACTIONS(1662),
    [anon_sym_u128] = ACTIONS(1662),
    [anon_sym_i128] = ACTIONS(1662),
    [anon_sym_isize] = ACTIONS(1662),
    [anon_sym_usize] = ACTIONS(1662),
    [anon_sym_f32] = ACTIONS(1662),
    [anon_sym_f64] = ACTIONS(1662),
    [anon_sym_bool] = ACTIONS(1662),
    [anon_sym_str] = ACTIONS(1662),
    [anon_sym_char] = ACTIONS(1662),
    [anon_sym_SQUOTE] = ACTIONS(1662),
    [anon_sym_async] = ACTIONS(1662),
    [anon_sym_break] = ACTIONS(1662),
    [anon_sym_const] = ACTIONS(1662),
    [anon_sym_continue] = ACTIONS(1662),
    [anon_sym_default] = ACTIONS(1662),
    [anon_sym_enum] = ACTIONS(1662),
    [anon_sym_fn] = ACTIONS(1662),
    [anon_sym_for] = ACTIONS(1662),
    [anon_sym_if] = ACTIONS(1662),
    [anon_sym_impl] = ACTIONS(1662),
    [anon_sym_let] = ACTIONS(1662),
    [anon_sym_loop] = ACTIONS(1662),
    [anon_sym_match] = ACTIONS(1662),
    [anon_sym_mod] = ACTIONS(1662),
    [anon_sym_pub] = ACTIONS(1662),
    [anon_sym_return] = ACTIONS(1662),
    [anon_sym_static] = ACTIONS(1662),
    [anon_sym_struct] = ACTIONS(1662),
    [anon_sym_trait] = ACTIONS(1662),
    [anon_sym_type] = ACTIONS(1662),
    [anon_sym_union] = ACTIONS(1662),
    [anon_sym_unsafe] = ACTIONS(1662),
    [anon_sym_use] = ACTIONS(1662),
    [anon_sym_while] = ACTIONS(1662),
    [anon_sym_POUND] = ACTIONS(1660),
    [anon_sym_BANG] = ACTIONS(1660),
    [anon_sym_extern] = ACTIONS(1662),
    [anon_sym_LT] = ACTIONS(1660),
    [anon_sym_COLON_COLON] = ACTIONS(1660),
    [anon_sym_AMP] = ACTIONS(1660),
    [anon_sym_DOT_DOT] = ACTIONS(1660),
    [anon_sym_DASH] = ACTIONS(1660),
    [anon_sym_PIPE] = ACTIONS(1660),
    [anon_sym_yield] = ACTIONS(1662),
    [anon_sym_move] = ACTIONS(1662),
    [sym_integer_literal] = ACTIONS(1660),
    [aux_sym_string_literal_token1] = ACTIONS(1660),
    [sym_char_literal] = ACTIONS(1660),
    [anon_sym_true] = ACTIONS(1662),
    [anon_sym_false] = ACTIONS(1662),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1662),
    [sym_super] = ACTIONS(1662),
    [sym_crate] = ACTIONS(1662),
    [sym_metavariable] = ACTIONS(1660),
    [sym_raw_string_literal] = ACTIONS(1660),
    [sym_float_literal] = ACTIONS(1660),
    [sym_block_comment] = ACTIONS(3),
  },
  [396] = {
    [ts_builtin_sym_end] = ACTIONS(1664),
    [sym_identifier] = ACTIONS(1666),
    [anon_sym_SEMI] = ACTIONS(1664),
    [anon_sym_macro_rules_BANG] = ACTIONS(1664),
    [anon_sym_LPAREN] = ACTIONS(1664),
    [anon_sym_LBRACE] = ACTIONS(1664),
    [anon_sym_RBRACE] = ACTIONS(1664),
    [anon_sym_LBRACK] = ACTIONS(1664),
    [anon_sym_STAR] = ACTIONS(1664),
    [anon_sym_u8] = ACTIONS(1666),
    [anon_sym_i8] = ACTIONS(1666),
    [anon_sym_u16] = ACTIONS(1666),
    [anon_sym_i16] = ACTIONS(1666),
    [anon_sym_u32] = ACTIONS(1666),
    [anon_sym_i32] = ACTIONS(1666),
    [anon_sym_u64] = ACTIONS(1666),
    [anon_sym_i64] = ACTIONS(1666),
    [anon_sym_u128] = ACTIONS(1666),
    [anon_sym_i128] = ACTIONS(1666),
    [anon_sym_isize] = ACTIONS(1666),
    [anon_sym_usize] = ACTIONS(1666),
    [anon_sym_f32] = ACTIONS(1666),
    [anon_sym_f64] = ACTIONS(1666),
    [anon_sym_bool] = ACTIONS(1666),
    [anon_sym_str] = ACTIONS(1666),
    [anon_sym_char] = ACTIONS(1666),
    [anon_sym_SQUOTE] = ACTIONS(1666),
    [anon_sym_async] = ACTIONS(1666),
    [anon_sym_break] = ACTIONS(1666),
    [anon_sym_const] = ACTIONS(1666),
    [anon_sym_continue] = ACTIONS(1666),
    [anon_sym_default] = ACTIONS(1666),
    [anon_sym_enum] = ACTIONS(1666),
    [anon_sym_fn] = ACTIONS(1666),
    [anon_sym_for] = ACTIONS(1666),
    [anon_sym_if] = ACTIONS(1666),
    [anon_sym_impl] = ACTIONS(1666),
    [anon_sym_let] = ACTIONS(1666),
    [anon_sym_loop] = ACTIONS(1666),
    [anon_sym_match] = ACTIONS(1666),
    [anon_sym_mod] = ACTIONS(1666),
    [anon_sym_pub] = ACTIONS(1666),
    [anon_sym_return] = ACTIONS(1666),
    [anon_sym_static] = ACTIONS(1666),
    [anon_sym_struct] = ACTIONS(1666),
    [anon_sym_trait] = ACTIONS(1666),
    [anon_sym_type] = ACTIONS(1666),
    [anon_sym_union] = ACTIONS(1666),
    [anon_sym_unsafe] = ACTIONS(1666),
    [anon_sym_use] = ACTIONS(1666),
    [anon_sym_while] = ACTIONS(1666),
    [anon_sym_POUND] = ACTIONS(1664),
    [anon_sym_BANG] = ACTIONS(1664),
    [anon_sym_extern] = ACTIONS(1666),
    [anon_sym_LT] = ACTIONS(1664),
    [anon_sym_COLON_COLON] = ACTIONS(1664),
    [anon_sym_AMP] = ACTIONS(1664),
    [anon_sym_DOT_DOT] = ACTIONS(1664),
    [anon_sym_DASH] = ACTIONS(1664),
    [anon_sym_PIPE] = ACTIONS(1664),
    [anon_sym_yield] = ACTIONS(1666),
    [anon_sym_move] = ACTIONS(1666),
    [sym_integer_literal] = ACTIONS(1664),
    [aux_sym_string_literal_token1] = ACTIONS(1664),
    [sym_char_literal] = ACTIONS(1664),
    [anon_sym_true] = ACTIONS(1666),
    [anon_sym_false] = ACTIONS(1666),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1666),
    [sym_super] = ACTIONS(1666),
    [sym_crate] = ACTIONS(1666),
    [sym_metavariable] = ACTIONS(1664),
    [sym_raw_string_literal] = ACTIONS(1664),
    [sym_float_literal] = ACTIONS(1664),
    [sym_block_comment] = ACTIONS(3),
  },
  [397] = {
    [ts_builtin_sym_end] = ACTIONS(1668),
    [sym_identifier] = ACTIONS(1670),
    [anon_sym_SEMI] = ACTIONS(1668),
    [anon_sym_macro_rules_BANG] = ACTIONS(1668),
    [anon_sym_LPAREN] = ACTIONS(1668),
    [anon_sym_LBRACE] = ACTIONS(1668),
    [anon_sym_RBRACE] = ACTIONS(1668),
    [anon_sym_LBRACK] = ACTIONS(1668),
    [anon_sym_STAR] = ACTIONS(1668),
    [anon_sym_u8] = ACTIONS(1670),
    [anon_sym_i8] = ACTIONS(1670),
    [anon_sym_u16] = ACTIONS(1670),
    [anon_sym_i16] = ACTIONS(1670),
    [anon_sym_u32] = ACTIONS(1670),
    [anon_sym_i32] = ACTIONS(1670),
    [anon_sym_u64] = ACTIONS(1670),
    [anon_sym_i64] = ACTIONS(1670),
    [anon_sym_u128] = ACTIONS(1670),
    [anon_sym_i128] = ACTIONS(1670),
    [anon_sym_isize] = ACTIONS(1670),
    [anon_sym_usize] = ACTIONS(1670),
    [anon_sym_f32] = ACTIONS(1670),
    [anon_sym_f64] = ACTIONS(1670),
    [anon_sym_bool] = ACTIONS(1670),
    [anon_sym_str] = ACTIONS(1670),
    [anon_sym_char] = ACTIONS(1670),
    [anon_sym_SQUOTE] = ACTIONS(1670),
    [anon_sym_async] = ACTIONS(1670),
    [anon_sym_break] = ACTIONS(1670),
    [anon_sym_const] = ACTIONS(1670),
    [anon_sym_continue] = ACTIONS(1670),
    [anon_sym_default] = ACTIONS(1670),
    [anon_sym_enum] = ACTIONS(1670),
    [anon_sym_fn] = ACTIONS(1670),
    [anon_sym_for] = ACTIONS(1670),
    [anon_sym_if] = ACTIONS(1670),
    [anon_sym_impl] = ACTIONS(1670),
    [anon_sym_let] = ACTIONS(1670),
    [anon_sym_loop] = ACTIONS(1670),
    [anon_sym_match] = ACTIONS(1670),
    [anon_sym_mod] = ACTIONS(1670),
    [anon_sym_pub] = ACTIONS(1670),
    [anon_sym_return] = ACTIONS(1670),
    [anon_sym_static] = ACTIONS(1670),
    [anon_sym_struct] = ACTIONS(1670),
    [anon_sym_trait] = ACTIONS(1670),
    [anon_sym_type] = ACTIONS(1670),
    [anon_sym_union] = ACTIONS(1670),
    [anon_sym_unsafe] = ACTIONS(1670),
    [anon_sym_use] = ACTIONS(1670),
    [anon_sym_while] = ACTIONS(1670),
    [anon_sym_POUND] = ACTIONS(1668),
    [anon_sym_BANG] = ACTIONS(1668),
    [anon_sym_extern] = ACTIONS(1670),
    [anon_sym_LT] = ACTIONS(1668),
    [anon_sym_COLON_COLON] = ACTIONS(1668),
    [anon_sym_AMP] = ACTIONS(1668),
    [anon_sym_DOT_DOT] = ACTIONS(1668),
    [anon_sym_DASH] = ACTIONS(1668),
    [anon_sym_PIPE] = ACTIONS(1668),
    [anon_sym_yield] = ACTIONS(1670),
    [anon_sym_move] = ACTIONS(1670),
    [sym_integer_literal] = ACTIONS(1668),
    [aux_sym_string_literal_token1] = ACTIONS(1668),
    [sym_char_literal] = ACTIONS(1668),
    [anon_sym_true] = ACTIONS(1670),
    [anon_sym_false] = ACTIONS(1670),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1670),
    [sym_super] = ACTIONS(1670),
    [sym_crate] = ACTIONS(1670),
    [sym_metavariable] = ACTIONS(1668),
    [sym_raw_string_literal] = ACTIONS(1668),
    [sym_float_literal] = ACTIONS(1668),
    [sym_block_comment] = ACTIONS(3),
  },
  [398] = {
    [ts_builtin_sym_end] = ACTIONS(1672),
    [sym_identifier] = ACTIONS(1674),
    [anon_sym_SEMI] = ACTIONS(1672),
    [anon_sym_macro_rules_BANG] = ACTIONS(1672),
    [anon_sym_LPAREN] = ACTIONS(1672),
    [anon_sym_LBRACE] = ACTIONS(1672),
    [anon_sym_RBRACE] = ACTIONS(1672),
    [anon_sym_LBRACK] = ACTIONS(1672),
    [anon_sym_STAR] = ACTIONS(1672),
    [anon_sym_u8] = ACTIONS(1674),
    [anon_sym_i8] = ACTIONS(1674),
    [anon_sym_u16] = ACTIONS(1674),
    [anon_sym_i16] = ACTIONS(1674),
    [anon_sym_u32] = ACTIONS(1674),
    [anon_sym_i32] = ACTIONS(1674),
    [anon_sym_u64] = ACTIONS(1674),
    [anon_sym_i64] = ACTIONS(1674),
    [anon_sym_u128] = ACTIONS(1674),
    [anon_sym_i128] = ACTIONS(1674),
    [anon_sym_isize] = ACTIONS(1674),
    [anon_sym_usize] = ACTIONS(1674),
    [anon_sym_f32] = ACTIONS(1674),
    [anon_sym_f64] = ACTIONS(1674),
    [anon_sym_bool] = ACTIONS(1674),
    [anon_sym_str] = ACTIONS(1674),
    [anon_sym_char] = ACTIONS(1674),
    [anon_sym_SQUOTE] = ACTIONS(1674),
    [anon_sym_async] = ACTIONS(1674),
    [anon_sym_break] = ACTIONS(1674),
    [anon_sym_const] = ACTIONS(1674),
    [anon_sym_continue] = ACTIONS(1674),
    [anon_sym_default] = ACTIONS(1674),
    [anon_sym_enum] = ACTIONS(1674),
    [anon_sym_fn] = ACTIONS(1674),
    [anon_sym_for] = ACTIONS(1674),
    [anon_sym_if] = ACTIONS(1674),
    [anon_sym_impl] = ACTIONS(1674),
    [anon_sym_let] = ACTIONS(1674),
    [anon_sym_loop] = ACTIONS(1674),
    [anon_sym_match] = ACTIONS(1674),
    [anon_sym_mod] = ACTIONS(1674),
    [anon_sym_pub] = ACTIONS(1674),
    [anon_sym_return] = ACTIONS(1674),
    [anon_sym_static] = ACTIONS(1674),
    [anon_sym_struct] = ACTIONS(1674),
    [anon_sym_trait] = ACTIONS(1674),
    [anon_sym_type] = ACTIONS(1674),
    [anon_sym_union] = ACTIONS(1674),
    [anon_sym_unsafe] = ACTIONS(1674),
    [anon_sym_use] = ACTIONS(1674),
    [anon_sym_while] = ACTIONS(1674),
    [anon_sym_POUND] = ACTIONS(1672),
    [anon_sym_BANG] = ACTIONS(1672),
    [anon_sym_extern] = ACTIONS(1674),
    [anon_sym_LT] = ACTIONS(1672),
    [anon_sym_COLON_COLON] = ACTIONS(1672),
    [anon_sym_AMP] = ACTIONS(1672),
    [anon_sym_DOT_DOT] = ACTIONS(1672),
    [anon_sym_DASH] = ACTIONS(1672),
    [anon_sym_PIPE] = ACTIONS(1672),
    [anon_sym_yield] = ACTIONS(1674),
    [anon_sym_move] = ACTIONS(1674),
    [sym_integer_literal] = ACTIONS(1672),
    [aux_sym_string_literal_token1] = ACTIONS(1672),
    [sym_char_literal] = ACTIONS(1672),
    [anon_sym_true] = ACTIONS(1674),
    [anon_sym_false] = ACTIONS(1674),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1674),
    [sym_super] = ACTIONS(1674),
    [sym_crate] = ACTIONS(1674),
    [sym_metavariable] = ACTIONS(1672),
    [sym_raw_string_literal] = ACTIONS(1672),
    [sym_float_literal] = ACTIONS(1672),
    [sym_block_comment] = ACTIONS(3),
  },
  [399] = {
    [ts_builtin_sym_end] = ACTIONS(1676),
    [sym_identifier] = ACTIONS(1678),
    [anon_sym_SEMI] = ACTIONS(1676),
    [anon_sym_macro_rules_BANG] = ACTIONS(1676),
    [anon_sym_LPAREN] = ACTIONS(1676),
    [anon_sym_LBRACE] = ACTIONS(1676),
    [anon_sym_RBRACE] = ACTIONS(1676),
    [anon_sym_LBRACK] = ACTIONS(1676),
    [anon_sym_STAR] = ACTIONS(1676),
    [anon_sym_u8] = ACTIONS(1678),
    [anon_sym_i8] = ACTIONS(1678),
    [anon_sym_u16] = ACTIONS(1678),
    [anon_sym_i16] = ACTIONS(1678),
    [anon_sym_u32] = ACTIONS(1678),
    [anon_sym_i32] = ACTIONS(1678),
    [anon_sym_u64] = ACTIONS(1678),
    [anon_sym_i64] = ACTIONS(1678),
    [anon_sym_u128] = ACTIONS(1678),
    [anon_sym_i128] = ACTIONS(1678),
    [anon_sym_isize] = ACTIONS(1678),
    [anon_sym_usize] = ACTIONS(1678),
    [anon_sym_f32] = ACTIONS(1678),
    [anon_sym_f64] = ACTIONS(1678),
    [anon_sym_bool] = ACTIONS(1678),
    [anon_sym_str] = ACTIONS(1678),
    [anon_sym_char] = ACTIONS(1678),
    [anon_sym_SQUOTE] = ACTIONS(1678),
    [anon_sym_async] = ACTIONS(1678),
    [anon_sym_break] = ACTIONS(1678),
    [anon_sym_const] = ACTIONS(1678),
    [anon_sym_continue] = ACTIONS(1678),
    [anon_sym_default] = ACTIONS(1678),
    [anon_sym_enum] = ACTIONS(1678),
    [anon_sym_fn] = ACTIONS(1678),
    [anon_sym_for] = ACTIONS(1678),
    [anon_sym_if] = ACTIONS(1678),
    [anon_sym_impl] = ACTIONS(1678),
    [anon_sym_let] = ACTIONS(1678),
    [anon_sym_loop] = ACTIONS(1678),
    [anon_sym_match] = ACTIONS(1678),
    [anon_sym_mod] = ACTIONS(1678),
    [anon_sym_pub] = ACTIONS(1678),
    [anon_sym_return] = ACTIONS(1678),
    [anon_sym_static] = ACTIONS(1678),
    [anon_sym_struct] = ACTIONS(1678),
    [anon_sym_trait] = ACTIONS(1678),
    [anon_sym_type] = ACTIONS(1678),
    [anon_sym_union] = ACTIONS(1678),
    [anon_sym_unsafe] = ACTIONS(1678),
    [anon_sym_use] = ACTIONS(1678),
    [anon_sym_while] = ACTIONS(1678),
    [anon_sym_POUND] = ACTIONS(1676),
    [anon_sym_BANG] = ACTIONS(1676),
    [anon_sym_extern] = ACTIONS(1678),
    [anon_sym_LT] = ACTIONS(1676),
    [anon_sym_COLON_COLON] = ACTIONS(1676),
    [anon_sym_AMP] = ACTIONS(1676),
    [anon_sym_DOT_DOT] = ACTIONS(1676),
    [anon_sym_DASH] = ACTIONS(1676),
    [anon_sym_PIPE] = ACTIONS(1676),
    [anon_sym_yield] = ACTIONS(1678),
    [anon_sym_move] = ACTIONS(1678),
    [sym_integer_literal] = ACTIONS(1676),
    [aux_sym_string_literal_token1] = ACTIONS(1676),
    [sym_char_literal] = ACTIONS(1676),
    [anon_sym_true] = ACTIONS(1678),
    [anon_sym_false] = ACTIONS(1678),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1678),
    [sym_super] = ACTIONS(1678),
    [sym_crate] = ACTIONS(1678),
    [sym_metavariable] = ACTIONS(1676),
    [sym_raw_string_literal] = ACTIONS(1676),
    [sym_float_literal] = ACTIONS(1676),
    [sym_block_comment] = ACTIONS(3),
  },
  [400] = {
    [ts_builtin_sym_end] = ACTIONS(1680),
    [sym_identifier] = ACTIONS(1682),
    [anon_sym_SEMI] = ACTIONS(1680),
    [anon_sym_macro_rules_BANG] = ACTIONS(1680),
    [anon_sym_LPAREN] = ACTIONS(1680),
    [anon_sym_LBRACE] = ACTIONS(1680),
    [anon_sym_RBRACE] = ACTIONS(1680),
    [anon_sym_LBRACK] = ACTIONS(1680),
    [anon_sym_STAR] = ACTIONS(1680),
    [anon_sym_u8] = ACTIONS(1682),
    [anon_sym_i8] = ACTIONS(1682),
    [anon_sym_u16] = ACTIONS(1682),
    [anon_sym_i16] = ACTIONS(1682),
    [anon_sym_u32] = ACTIONS(1682),
    [anon_sym_i32] = ACTIONS(1682),
    [anon_sym_u64] = ACTIONS(1682),
    [anon_sym_i64] = ACTIONS(1682),
    [anon_sym_u128] = ACTIONS(1682),
    [anon_sym_i128] = ACTIONS(1682),
    [anon_sym_isize] = ACTIONS(1682),
    [anon_sym_usize] = ACTIONS(1682),
    [anon_sym_f32] = ACTIONS(1682),
    [anon_sym_f64] = ACTIONS(1682),
    [anon_sym_bool] = ACTIONS(1682),
    [anon_sym_str] = ACTIONS(1682),
    [anon_sym_char] = ACTIONS(1682),
    [anon_sym_SQUOTE] = ACTIONS(1682),
    [anon_sym_async] = ACTIONS(1682),
    [anon_sym_break] = ACTIONS(1682),
    [anon_sym_const] = ACTIONS(1682),
    [anon_sym_continue] = ACTIONS(1682),
    [anon_sym_default] = ACTIONS(1682),
    [anon_sym_enum] = ACTIONS(1682),
    [anon_sym_fn] = ACTIONS(1682),
    [anon_sym_for] = ACTIONS(1682),
    [anon_sym_if] = ACTIONS(1682),
    [anon_sym_impl] = ACTIONS(1682),
    [anon_sym_let] = ACTIONS(1682),
    [anon_sym_loop] = ACTIONS(1682),
    [anon_sym_match] = ACTIONS(1682),
    [anon_sym_mod] = ACTIONS(1682),
    [anon_sym_pub] = ACTIONS(1682),
    [anon_sym_return] = ACTIONS(1682),
    [anon_sym_static] = ACTIONS(1682),
    [anon_sym_struct] = ACTIONS(1682),
    [anon_sym_trait] = ACTIONS(1682),
    [anon_sym_type] = ACTIONS(1682),
    [anon_sym_union] = ACTIONS(1682),
    [anon_sym_unsafe] = ACTIONS(1682),
    [anon_sym_use] = ACTIONS(1682),
    [anon_sym_while] = ACTIONS(1682),
    [anon_sym_POUND] = ACTIONS(1680),
    [anon_sym_BANG] = ACTIONS(1680),
    [anon_sym_extern] = ACTIONS(1682),
    [anon_sym_LT] = ACTIONS(1680),
    [anon_sym_COLON_COLON] = ACTIONS(1680),
    [anon_sym_AMP] = ACTIONS(1680),
    [anon_sym_DOT_DOT] = ACTIONS(1680),
    [anon_sym_DASH] = ACTIONS(1680),
    [anon_sym_PIPE] = ACTIONS(1680),
    [anon_sym_yield] = ACTIONS(1682),
    [anon_sym_move] = ACTIONS(1682),
    [sym_integer_literal] = ACTIONS(1680),
    [aux_sym_string_literal_token1] = ACTIONS(1680),
    [sym_char_literal] = ACTIONS(1680),
    [anon_sym_true] = ACTIONS(1682),
    [anon_sym_false] = ACTIONS(1682),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1682),
    [sym_super] = ACTIONS(1682),
    [sym_crate] = ACTIONS(1682),
    [sym_metavariable] = ACTIONS(1680),
    [sym_raw_string_literal] = ACTIONS(1680),
    [sym_float_literal] = ACTIONS(1680),
    [sym_block_comment] = ACTIONS(3),
  },
  [401] = {
    [ts_builtin_sym_end] = ACTIONS(1684),
    [sym_identifier] = ACTIONS(1686),
    [anon_sym_SEMI] = ACTIONS(1684),
    [anon_sym_macro_rules_BANG] = ACTIONS(1684),
    [anon_sym_LPAREN] = ACTIONS(1684),
    [anon_sym_LBRACE] = ACTIONS(1684),
    [anon_sym_RBRACE] = ACTIONS(1684),
    [anon_sym_LBRACK] = ACTIONS(1684),
    [anon_sym_STAR] = ACTIONS(1684),
    [anon_sym_u8] = ACTIONS(1686),
    [anon_sym_i8] = ACTIONS(1686),
    [anon_sym_u16] = ACTIONS(1686),
    [anon_sym_i16] = ACTIONS(1686),
    [anon_sym_u32] = ACTIONS(1686),
    [anon_sym_i32] = ACTIONS(1686),
    [anon_sym_u64] = ACTIONS(1686),
    [anon_sym_i64] = ACTIONS(1686),
    [anon_sym_u128] = ACTIONS(1686),
    [anon_sym_i128] = ACTIONS(1686),
    [anon_sym_isize] = ACTIONS(1686),
    [anon_sym_usize] = ACTIONS(1686),
    [anon_sym_f32] = ACTIONS(1686),
    [anon_sym_f64] = ACTIONS(1686),
    [anon_sym_bool] = ACTIONS(1686),
    [anon_sym_str] = ACTIONS(1686),
    [anon_sym_char] = ACTIONS(1686),
    [anon_sym_SQUOTE] = ACTIONS(1686),
    [anon_sym_async] = ACTIONS(1686),
    [anon_sym_break] = ACTIONS(1686),
    [anon_sym_const] = ACTIONS(1686),
    [anon_sym_continue] = ACTIONS(1686),
    [anon_sym_default] = ACTIONS(1686),
    [anon_sym_enum] = ACTIONS(1686),
    [anon_sym_fn] = ACTIONS(1686),
    [anon_sym_for] = ACTIONS(1686),
    [anon_sym_if] = ACTIONS(1686),
    [anon_sym_impl] = ACTIONS(1686),
    [anon_sym_let] = ACTIONS(1686),
    [anon_sym_loop] = ACTIONS(1686),
    [anon_sym_match] = ACTIONS(1686),
    [anon_sym_mod] = ACTIONS(1686),
    [anon_sym_pub] = ACTIONS(1686),
    [anon_sym_return] = ACTIONS(1686),
    [anon_sym_static] = ACTIONS(1686),
    [anon_sym_struct] = ACTIONS(1686),
    [anon_sym_trait] = ACTIONS(1686),
    [anon_sym_type] = ACTIONS(1686),
    [anon_sym_union] = ACTIONS(1686),
    [anon_sym_unsafe] = ACTIONS(1686),
    [anon_sym_use] = ACTIONS(1686),
    [anon_sym_while] = ACTIONS(1686),
    [anon_sym_POUND] = ACTIONS(1684),
    [anon_sym_BANG] = ACTIONS(1684),
    [anon_sym_extern] = ACTIONS(1686),
    [anon_sym_LT] = ACTIONS(1684),
    [anon_sym_COLON_COLON] = ACTIONS(1684),
    [anon_sym_AMP] = ACTIONS(1684),
    [anon_sym_DOT_DOT] = ACTIONS(1684),
    [anon_sym_DASH] = ACTIONS(1684),
    [anon_sym_PIPE] = ACTIONS(1684),
    [anon_sym_yield] = ACTIONS(1686),
    [anon_sym_move] = ACTIONS(1686),
    [sym_integer_literal] = ACTIONS(1684),
    [aux_sym_string_literal_token1] = ACTIONS(1684),
    [sym_char_literal] = ACTIONS(1684),
    [anon_sym_true] = ACTIONS(1686),
    [anon_sym_false] = ACTIONS(1686),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1686),
    [sym_super] = ACTIONS(1686),
    [sym_crate] = ACTIONS(1686),
    [sym_metavariable] = ACTIONS(1684),
    [sym_raw_string_literal] = ACTIONS(1684),
    [sym_float_literal] = ACTIONS(1684),
    [sym_block_comment] = ACTIONS(3),
  },
  [402] = {
    [ts_builtin_sym_end] = ACTIONS(1688),
    [sym_identifier] = ACTIONS(1690),
    [anon_sym_SEMI] = ACTIONS(1688),
    [anon_sym_macro_rules_BANG] = ACTIONS(1688),
    [anon_sym_LPAREN] = ACTIONS(1688),
    [anon_sym_LBRACE] = ACTIONS(1688),
    [anon_sym_RBRACE] = ACTIONS(1688),
    [anon_sym_LBRACK] = ACTIONS(1688),
    [anon_sym_STAR] = ACTIONS(1688),
    [anon_sym_u8] = ACTIONS(1690),
    [anon_sym_i8] = ACTIONS(1690),
    [anon_sym_u16] = ACTIONS(1690),
    [anon_sym_i16] = ACTIONS(1690),
    [anon_sym_u32] = ACTIONS(1690),
    [anon_sym_i32] = ACTIONS(1690),
    [anon_sym_u64] = ACTIONS(1690),
    [anon_sym_i64] = ACTIONS(1690),
    [anon_sym_u128] = ACTIONS(1690),
    [anon_sym_i128] = ACTIONS(1690),
    [anon_sym_isize] = ACTIONS(1690),
    [anon_sym_usize] = ACTIONS(1690),
    [anon_sym_f32] = ACTIONS(1690),
    [anon_sym_f64] = ACTIONS(1690),
    [anon_sym_bool] = ACTIONS(1690),
    [anon_sym_str] = ACTIONS(1690),
    [anon_sym_char] = ACTIONS(1690),
    [anon_sym_SQUOTE] = ACTIONS(1690),
    [anon_sym_async] = ACTIONS(1690),
    [anon_sym_break] = ACTIONS(1690),
    [anon_sym_const] = ACTIONS(1690),
    [anon_sym_continue] = ACTIONS(1690),
    [anon_sym_default] = ACTIONS(1690),
    [anon_sym_enum] = ACTIONS(1690),
    [anon_sym_fn] = ACTIONS(1690),
    [anon_sym_for] = ACTIONS(1690),
    [anon_sym_if] = ACTIONS(1690),
    [anon_sym_impl] = ACTIONS(1690),
    [anon_sym_let] = ACTIONS(1690),
    [anon_sym_loop] = ACTIONS(1690),
    [anon_sym_match] = ACTIONS(1690),
    [anon_sym_mod] = ACTIONS(1690),
    [anon_sym_pub] = ACTIONS(1690),
    [anon_sym_return] = ACTIONS(1690),
    [anon_sym_static] = ACTIONS(1690),
    [anon_sym_struct] = ACTIONS(1690),
    [anon_sym_trait] = ACTIONS(1690),
    [anon_sym_type] = ACTIONS(1690),
    [anon_sym_union] = ACTIONS(1690),
    [anon_sym_unsafe] = ACTIONS(1690),
    [anon_sym_use] = ACTIONS(1690),
    [anon_sym_while] = ACTIONS(1690),
    [anon_sym_POUND] = ACTIONS(1688),
    [anon_sym_BANG] = ACTIONS(1688),
    [anon_sym_extern] = ACTIONS(1690),
    [anon_sym_LT] = ACTIONS(1688),
    [anon_sym_COLON_COLON] = ACTIONS(1688),
    [anon_sym_AMP] = ACTIONS(1688),
    [anon_sym_DOT_DOT] = ACTIONS(1688),
    [anon_sym_DASH] = ACTIONS(1688),
    [anon_sym_PIPE] = ACTIONS(1688),
    [anon_sym_yield] = ACTIONS(1690),
    [anon_sym_move] = ACTIONS(1690),
    [sym_integer_literal] = ACTIONS(1688),
    [aux_sym_string_literal_token1] = ACTIONS(1688),
    [sym_char_literal] = ACTIONS(1688),
    [anon_sym_true] = ACTIONS(1690),
    [anon_sym_false] = ACTIONS(1690),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1690),
    [sym_super] = ACTIONS(1690),
    [sym_crate] = ACTIONS(1690),
    [sym_metavariable] = ACTIONS(1688),
    [sym_raw_string_literal] = ACTIONS(1688),
    [sym_float_literal] = ACTIONS(1688),
    [sym_block_comment] = ACTIONS(3),
  },
  [403] = {
    [ts_builtin_sym_end] = ACTIONS(1692),
    [sym_identifier] = ACTIONS(1694),
    [anon_sym_SEMI] = ACTIONS(1692),
    [anon_sym_macro_rules_BANG] = ACTIONS(1692),
    [anon_sym_LPAREN] = ACTIONS(1692),
    [anon_sym_LBRACE] = ACTIONS(1692),
    [anon_sym_RBRACE] = ACTIONS(1692),
    [anon_sym_LBRACK] = ACTIONS(1692),
    [anon_sym_STAR] = ACTIONS(1692),
    [anon_sym_u8] = ACTIONS(1694),
    [anon_sym_i8] = ACTIONS(1694),
    [anon_sym_u16] = ACTIONS(1694),
    [anon_sym_i16] = ACTIONS(1694),
    [anon_sym_u32] = ACTIONS(1694),
    [anon_sym_i32] = ACTIONS(1694),
    [anon_sym_u64] = ACTIONS(1694),
    [anon_sym_i64] = ACTIONS(1694),
    [anon_sym_u128] = ACTIONS(1694),
    [anon_sym_i128] = ACTIONS(1694),
    [anon_sym_isize] = ACTIONS(1694),
    [anon_sym_usize] = ACTIONS(1694),
    [anon_sym_f32] = ACTIONS(1694),
    [anon_sym_f64] = ACTIONS(1694),
    [anon_sym_bool] = ACTIONS(1694),
    [anon_sym_str] = ACTIONS(1694),
    [anon_sym_char] = ACTIONS(1694),
    [anon_sym_SQUOTE] = ACTIONS(1694),
    [anon_sym_async] = ACTIONS(1694),
    [anon_sym_break] = ACTIONS(1694),
    [anon_sym_const] = ACTIONS(1694),
    [anon_sym_continue] = ACTIONS(1694),
    [anon_sym_default] = ACTIONS(1694),
    [anon_sym_enum] = ACTIONS(1694),
    [anon_sym_fn] = ACTIONS(1694),
    [anon_sym_for] = ACTIONS(1694),
    [anon_sym_if] = ACTIONS(1694),
    [anon_sym_impl] = ACTIONS(1694),
    [anon_sym_let] = ACTIONS(1694),
    [anon_sym_loop] = ACTIONS(1694),
    [anon_sym_match] = ACTIONS(1694),
    [anon_sym_mod] = ACTIONS(1694),
    [anon_sym_pub] = ACTIONS(1694),
    [anon_sym_return] = ACTIONS(1694),
    [anon_sym_static] = ACTIONS(1694),
    [anon_sym_struct] = ACTIONS(1694),
    [anon_sym_trait] = ACTIONS(1694),
    [anon_sym_type] = ACTIONS(1694),
    [anon_sym_union] = ACTIONS(1694),
    [anon_sym_unsafe] = ACTIONS(1694),
    [anon_sym_use] = ACTIONS(1694),
    [anon_sym_while] = ACTIONS(1694),
    [anon_sym_POUND] = ACTIONS(1692),
    [anon_sym_BANG] = ACTIONS(1692),
    [anon_sym_extern] = ACTIONS(1694),
    [anon_sym_LT] = ACTIONS(1692),
    [anon_sym_COLON_COLON] = ACTIONS(1692),
    [anon_sym_AMP] = ACTIONS(1692),
    [anon_sym_DOT_DOT] = ACTIONS(1692),
    [anon_sym_DASH] = ACTIONS(1692),
    [anon_sym_PIPE] = ACTIONS(1692),
    [anon_sym_yield] = ACTIONS(1694),
    [anon_sym_move] = ACTIONS(1694),
    [sym_integer_literal] = ACTIONS(1692),
    [aux_sym_string_literal_token1] = ACTIONS(1692),
    [sym_char_literal] = ACTIONS(1692),
    [anon_sym_true] = ACTIONS(1694),
    [anon_sym_false] = ACTIONS(1694),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1694),
    [sym_super] = ACTIONS(1694),
    [sym_crate] = ACTIONS(1694),
    [sym_metavariable] = ACTIONS(1692),
    [sym_raw_string_literal] = ACTIONS(1692),
    [sym_float_literal] = ACTIONS(1692),
    [sym_block_comment] = ACTIONS(3),
  },
  [404] = {
    [ts_builtin_sym_end] = ACTIONS(1696),
    [sym_identifier] = ACTIONS(1698),
    [anon_sym_SEMI] = ACTIONS(1696),
    [anon_sym_macro_rules_BANG] = ACTIONS(1696),
    [anon_sym_LPAREN] = ACTIONS(1696),
    [anon_sym_LBRACE] = ACTIONS(1696),
    [anon_sym_RBRACE] = ACTIONS(1696),
    [anon_sym_LBRACK] = ACTIONS(1696),
    [anon_sym_STAR] = ACTIONS(1696),
    [anon_sym_u8] = ACTIONS(1698),
    [anon_sym_i8] = ACTIONS(1698),
    [anon_sym_u16] = ACTIONS(1698),
    [anon_sym_i16] = ACTIONS(1698),
    [anon_sym_u32] = ACTIONS(1698),
    [anon_sym_i32] = ACTIONS(1698),
    [anon_sym_u64] = ACTIONS(1698),
    [anon_sym_i64] = ACTIONS(1698),
    [anon_sym_u128] = ACTIONS(1698),
    [anon_sym_i128] = ACTIONS(1698),
    [anon_sym_isize] = ACTIONS(1698),
    [anon_sym_usize] = ACTIONS(1698),
    [anon_sym_f32] = ACTIONS(1698),
    [anon_sym_f64] = ACTIONS(1698),
    [anon_sym_bool] = ACTIONS(1698),
    [anon_sym_str] = ACTIONS(1698),
    [anon_sym_char] = ACTIONS(1698),
    [anon_sym_SQUOTE] = ACTIONS(1698),
    [anon_sym_async] = ACTIONS(1698),
    [anon_sym_break] = ACTIONS(1698),
    [anon_sym_const] = ACTIONS(1698),
    [anon_sym_continue] = ACTIONS(1698),
    [anon_sym_default] = ACTIONS(1698),
    [anon_sym_enum] = ACTIONS(1698),
    [anon_sym_fn] = ACTIONS(1698),
    [anon_sym_for] = ACTIONS(1698),
    [anon_sym_if] = ACTIONS(1698),
    [anon_sym_impl] = ACTIONS(1698),
    [anon_sym_let] = ACTIONS(1698),
    [anon_sym_loop] = ACTIONS(1698),
    [anon_sym_match] = ACTIONS(1698),
    [anon_sym_mod] = ACTIONS(1698),
    [anon_sym_pub] = ACTIONS(1698),
    [anon_sym_return] = ACTIONS(1698),
    [anon_sym_static] = ACTIONS(1698),
    [anon_sym_struct] = ACTIONS(1698),
    [anon_sym_trait] = ACTIONS(1698),
    [anon_sym_type] = ACTIONS(1698),
    [anon_sym_union] = ACTIONS(1698),
    [anon_sym_unsafe] = ACTIONS(1698),
    [anon_sym_use] = ACTIONS(1698),
    [anon_sym_while] = ACTIONS(1698),
    [anon_sym_POUND] = ACTIONS(1696),
    [anon_sym_BANG] = ACTIONS(1696),
    [anon_sym_extern] = ACTIONS(1698),
    [anon_sym_LT] = ACTIONS(1696),
    [anon_sym_COLON_COLON] = ACTIONS(1696),
    [anon_sym_AMP] = ACTIONS(1696),
    [anon_sym_DOT_DOT] = ACTIONS(1696),
    [anon_sym_DASH] = ACTIONS(1696),
    [anon_sym_PIPE] = ACTIONS(1696),
    [anon_sym_yield] = ACTIONS(1698),
    [anon_sym_move] = ACTIONS(1698),
    [sym_integer_literal] = ACTIONS(1696),
    [aux_sym_string_literal_token1] = ACTIONS(1696),
    [sym_char_literal] = ACTIONS(1696),
    [anon_sym_true] = ACTIONS(1698),
    [anon_sym_false] = ACTIONS(1698),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1698),
    [sym_super] = ACTIONS(1698),
    [sym_crate] = ACTIONS(1698),
    [sym_metavariable] = ACTIONS(1696),
    [sym_raw_string_literal] = ACTIONS(1696),
    [sym_float_literal] = ACTIONS(1696),
    [sym_block_comment] = ACTIONS(3),
  },
  [405] = {
    [ts_builtin_sym_end] = ACTIONS(1700),
    [sym_identifier] = ACTIONS(1702),
    [anon_sym_SEMI] = ACTIONS(1700),
    [anon_sym_macro_rules_BANG] = ACTIONS(1700),
    [anon_sym_LPAREN] = ACTIONS(1700),
    [anon_sym_LBRACE] = ACTIONS(1700),
    [anon_sym_RBRACE] = ACTIONS(1700),
    [anon_sym_LBRACK] = ACTIONS(1700),
    [anon_sym_STAR] = ACTIONS(1700),
    [anon_sym_u8] = ACTIONS(1702),
    [anon_sym_i8] = ACTIONS(1702),
    [anon_sym_u16] = ACTIONS(1702),
    [anon_sym_i16] = ACTIONS(1702),
    [anon_sym_u32] = ACTIONS(1702),
    [anon_sym_i32] = ACTIONS(1702),
    [anon_sym_u64] = ACTIONS(1702),
    [anon_sym_i64] = ACTIONS(1702),
    [anon_sym_u128] = ACTIONS(1702),
    [anon_sym_i128] = ACTIONS(1702),
    [anon_sym_isize] = ACTIONS(1702),
    [anon_sym_usize] = ACTIONS(1702),
    [anon_sym_f32] = ACTIONS(1702),
    [anon_sym_f64] = ACTIONS(1702),
    [anon_sym_bool] = ACTIONS(1702),
    [anon_sym_str] = ACTIONS(1702),
    [anon_sym_char] = ACTIONS(1702),
    [anon_sym_SQUOTE] = ACTIONS(1702),
    [anon_sym_async] = ACTIONS(1702),
    [anon_sym_break] = ACTIONS(1702),
    [anon_sym_const] = ACTIONS(1702),
    [anon_sym_continue] = ACTIONS(1702),
    [anon_sym_default] = ACTIONS(1702),
    [anon_sym_enum] = ACTIONS(1702),
    [anon_sym_fn] = ACTIONS(1702),
    [anon_sym_for] = ACTIONS(1702),
    [anon_sym_if] = ACTIONS(1702),
    [anon_sym_impl] = ACTIONS(1702),
    [anon_sym_let] = ACTIONS(1702),
    [anon_sym_loop] = ACTIONS(1702),
    [anon_sym_match] = ACTIONS(1702),
    [anon_sym_mod] = ACTIONS(1702),
    [anon_sym_pub] = ACTIONS(1702),
    [anon_sym_return] = ACTIONS(1702),
    [anon_sym_static] = ACTIONS(1702),
    [anon_sym_struct] = ACTIONS(1702),
    [anon_sym_trait] = ACTIONS(1702),
    [anon_sym_type] = ACTIONS(1702),
    [anon_sym_union] = ACTIONS(1702),
    [anon_sym_unsafe] = ACTIONS(1702),
    [anon_sym_use] = ACTIONS(1702),
    [anon_sym_while] = ACTIONS(1702),
    [anon_sym_POUND] = ACTIONS(1700),
    [anon_sym_BANG] = ACTIONS(1700),
    [anon_sym_extern] = ACTIONS(1702),
    [anon_sym_LT] = ACTIONS(1700),
    [anon_sym_COLON_COLON] = ACTIONS(1700),
    [anon_sym_AMP] = ACTIONS(1700),
    [anon_sym_DOT_DOT] = ACTIONS(1700),
    [anon_sym_DASH] = ACTIONS(1700),
    [anon_sym_PIPE] = ACTIONS(1700),
    [anon_sym_yield] = ACTIONS(1702),
    [anon_sym_move] = ACTIONS(1702),
    [sym_integer_literal] = ACTIONS(1700),
    [aux_sym_string_literal_token1] = ACTIONS(1700),
    [sym_char_literal] = ACTIONS(1700),
    [anon_sym_true] = ACTIONS(1702),
    [anon_sym_false] = ACTIONS(1702),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1702),
    [sym_super] = ACTIONS(1702),
    [sym_crate] = ACTIONS(1702),
    [sym_metavariable] = ACTIONS(1700),
    [sym_raw_string_literal] = ACTIONS(1700),
    [sym_float_literal] = ACTIONS(1700),
    [sym_block_comment] = ACTIONS(3),
  },
  [406] = {
    [ts_builtin_sym_end] = ACTIONS(1704),
    [sym_identifier] = ACTIONS(1706),
    [anon_sym_SEMI] = ACTIONS(1704),
    [anon_sym_macro_rules_BANG] = ACTIONS(1704),
    [anon_sym_LPAREN] = ACTIONS(1704),
    [anon_sym_LBRACE] = ACTIONS(1704),
    [anon_sym_RBRACE] = ACTIONS(1704),
    [anon_sym_LBRACK] = ACTIONS(1704),
    [anon_sym_STAR] = ACTIONS(1704),
    [anon_sym_u8] = ACTIONS(1706),
    [anon_sym_i8] = ACTIONS(1706),
    [anon_sym_u16] = ACTIONS(1706),
    [anon_sym_i16] = ACTIONS(1706),
    [anon_sym_u32] = ACTIONS(1706),
    [anon_sym_i32] = ACTIONS(1706),
    [anon_sym_u64] = ACTIONS(1706),
    [anon_sym_i64] = ACTIONS(1706),
    [anon_sym_u128] = ACTIONS(1706),
    [anon_sym_i128] = ACTIONS(1706),
    [anon_sym_isize] = ACTIONS(1706),
    [anon_sym_usize] = ACTIONS(1706),
    [anon_sym_f32] = ACTIONS(1706),
    [anon_sym_f64] = ACTIONS(1706),
    [anon_sym_bool] = ACTIONS(1706),
    [anon_sym_str] = ACTIONS(1706),
    [anon_sym_char] = ACTIONS(1706),
    [anon_sym_SQUOTE] = ACTIONS(1706),
    [anon_sym_async] = ACTIONS(1706),
    [anon_sym_break] = ACTIONS(1706),
    [anon_sym_const] = ACTIONS(1706),
    [anon_sym_continue] = ACTIONS(1706),
    [anon_sym_default] = ACTIONS(1706),
    [anon_sym_enum] = ACTIONS(1706),
    [anon_sym_fn] = ACTIONS(1706),
    [anon_sym_for] = ACTIONS(1706),
    [anon_sym_if] = ACTIONS(1706),
    [anon_sym_impl] = ACTIONS(1706),
    [anon_sym_let] = ACTIONS(1706),
    [anon_sym_loop] = ACTIONS(1706),
    [anon_sym_match] = ACTIONS(1706),
    [anon_sym_mod] = ACTIONS(1706),
    [anon_sym_pub] = ACTIONS(1706),
    [anon_sym_return] = ACTIONS(1706),
    [anon_sym_static] = ACTIONS(1706),
    [anon_sym_struct] = ACTIONS(1706),
    [anon_sym_trait] = ACTIONS(1706),
    [anon_sym_type] = ACTIONS(1706),
    [anon_sym_union] = ACTIONS(1706),
    [anon_sym_unsafe] = ACTIONS(1706),
    [anon_sym_use] = ACTIONS(1706),
    [anon_sym_while] = ACTIONS(1706),
    [anon_sym_POUND] = ACTIONS(1704),
    [anon_sym_BANG] = ACTIONS(1704),
    [anon_sym_extern] = ACTIONS(1706),
    [anon_sym_LT] = ACTIONS(1704),
    [anon_sym_COLON_COLON] = ACTIONS(1704),
    [anon_sym_AMP] = ACTIONS(1704),
    [anon_sym_DOT_DOT] = ACTIONS(1704),
    [anon_sym_DASH] = ACTIONS(1704),
    [anon_sym_PIPE] = ACTIONS(1704),
    [anon_sym_yield] = ACTIONS(1706),
    [anon_sym_move] = ACTIONS(1706),
    [sym_integer_literal] = ACTIONS(1704),
    [aux_sym_string_literal_token1] = ACTIONS(1704),
    [sym_char_literal] = ACTIONS(1704),
    [anon_sym_true] = ACTIONS(1706),
    [anon_sym_false] = ACTIONS(1706),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1706),
    [sym_super] = ACTIONS(1706),
    [sym_crate] = ACTIONS(1706),
    [sym_metavariable] = ACTIONS(1704),
    [sym_raw_string_literal] = ACTIONS(1704),
    [sym_float_literal] = ACTIONS(1704),
    [sym_block_comment] = ACTIONS(3),
  },
  [407] = {
    [ts_builtin_sym_end] = ACTIONS(1708),
    [sym_identifier] = ACTIONS(1710),
    [anon_sym_SEMI] = ACTIONS(1708),
    [anon_sym_macro_rules_BANG] = ACTIONS(1708),
    [anon_sym_LPAREN] = ACTIONS(1708),
    [anon_sym_LBRACE] = ACTIONS(1708),
    [anon_sym_RBRACE] = ACTIONS(1708),
    [anon_sym_LBRACK] = ACTIONS(1708),
    [anon_sym_STAR] = ACTIONS(1708),
    [anon_sym_u8] = ACTIONS(1710),
    [anon_sym_i8] = ACTIONS(1710),
    [anon_sym_u16] = ACTIONS(1710),
    [anon_sym_i16] = ACTIONS(1710),
    [anon_sym_u32] = ACTIONS(1710),
    [anon_sym_i32] = ACTIONS(1710),
    [anon_sym_u64] = ACTIONS(1710),
    [anon_sym_i64] = ACTIONS(1710),
    [anon_sym_u128] = ACTIONS(1710),
    [anon_sym_i128] = ACTIONS(1710),
    [anon_sym_isize] = ACTIONS(1710),
    [anon_sym_usize] = ACTIONS(1710),
    [anon_sym_f32] = ACTIONS(1710),
    [anon_sym_f64] = ACTIONS(1710),
    [anon_sym_bool] = ACTIONS(1710),
    [anon_sym_str] = ACTIONS(1710),
    [anon_sym_char] = ACTIONS(1710),
    [anon_sym_SQUOTE] = ACTIONS(1710),
    [anon_sym_async] = ACTIONS(1710),
    [anon_sym_break] = ACTIONS(1710),
    [anon_sym_const] = ACTIONS(1710),
    [anon_sym_continue] = ACTIONS(1710),
    [anon_sym_default] = ACTIONS(1710),
    [anon_sym_enum] = ACTIONS(1710),
    [anon_sym_fn] = ACTIONS(1710),
    [anon_sym_for] = ACTIONS(1710),
    [anon_sym_if] = ACTIONS(1710),
    [anon_sym_impl] = ACTIONS(1710),
    [anon_sym_let] = ACTIONS(1710),
    [anon_sym_loop] = ACTIONS(1710),
    [anon_sym_match] = ACTIONS(1710),
    [anon_sym_mod] = ACTIONS(1710),
    [anon_sym_pub] = ACTIONS(1710),
    [anon_sym_return] = ACTIONS(1710),
    [anon_sym_static] = ACTIONS(1710),
    [anon_sym_struct] = ACTIONS(1710),
    [anon_sym_trait] = ACTIONS(1710),
    [anon_sym_type] = ACTIONS(1710),
    [anon_sym_union] = ACTIONS(1710),
    [anon_sym_unsafe] = ACTIONS(1710),
    [anon_sym_use] = ACTIONS(1710),
    [anon_sym_while] = ACTIONS(1710),
    [anon_sym_POUND] = ACTIONS(1708),
    [anon_sym_BANG] = ACTIONS(1708),
    [anon_sym_extern] = ACTIONS(1710),
    [anon_sym_LT] = ACTIONS(1708),
    [anon_sym_COLON_COLON] = ACTIONS(1708),
    [anon_sym_AMP] = ACTIONS(1708),
    [anon_sym_DOT_DOT] = ACTIONS(1708),
    [anon_sym_DASH] = ACTIONS(1708),
    [anon_sym_PIPE] = ACTIONS(1708),
    [anon_sym_yield] = ACTIONS(1710),
    [anon_sym_move] = ACTIONS(1710),
    [sym_integer_literal] = ACTIONS(1708),
    [aux_sym_string_literal_token1] = ACTIONS(1708),
    [sym_char_literal] = ACTIONS(1708),
    [anon_sym_true] = ACTIONS(1710),
    [anon_sym_false] = ACTIONS(1710),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1710),
    [sym_super] = ACTIONS(1710),
    [sym_crate] = ACTIONS(1710),
    [sym_metavariable] = ACTIONS(1708),
    [sym_raw_string_literal] = ACTIONS(1708),
    [sym_float_literal] = ACTIONS(1708),
    [sym_block_comment] = ACTIONS(3),
  },
  [408] = {
    [ts_builtin_sym_end] = ACTIONS(1712),
    [sym_identifier] = ACTIONS(1714),
    [anon_sym_SEMI] = ACTIONS(1712),
    [anon_sym_macro_rules_BANG] = ACTIONS(1712),
    [anon_sym_LPAREN] = ACTIONS(1712),
    [anon_sym_LBRACE] = ACTIONS(1712),
    [anon_sym_RBRACE] = ACTIONS(1712),
    [anon_sym_LBRACK] = ACTIONS(1712),
    [anon_sym_STAR] = ACTIONS(1712),
    [anon_sym_u8] = ACTIONS(1714),
    [anon_sym_i8] = ACTIONS(1714),
    [anon_sym_u16] = ACTIONS(1714),
    [anon_sym_i16] = ACTIONS(1714),
    [anon_sym_u32] = ACTIONS(1714),
    [anon_sym_i32] = ACTIONS(1714),
    [anon_sym_u64] = ACTIONS(1714),
    [anon_sym_i64] = ACTIONS(1714),
    [anon_sym_u128] = ACTIONS(1714),
    [anon_sym_i128] = ACTIONS(1714),
    [anon_sym_isize] = ACTIONS(1714),
    [anon_sym_usize] = ACTIONS(1714),
    [anon_sym_f32] = ACTIONS(1714),
    [anon_sym_f64] = ACTIONS(1714),
    [anon_sym_bool] = ACTIONS(1714),
    [anon_sym_str] = ACTIONS(1714),
    [anon_sym_char] = ACTIONS(1714),
    [anon_sym_SQUOTE] = ACTIONS(1714),
    [anon_sym_async] = ACTIONS(1714),
    [anon_sym_break] = ACTIONS(1714),
    [anon_sym_const] = ACTIONS(1714),
    [anon_sym_continue] = ACTIONS(1714),
    [anon_sym_default] = ACTIONS(1714),
    [anon_sym_enum] = ACTIONS(1714),
    [anon_sym_fn] = ACTIONS(1714),
    [anon_sym_for] = ACTIONS(1714),
    [anon_sym_if] = ACTIONS(1714),
    [anon_sym_impl] = ACTIONS(1714),
    [anon_sym_let] = ACTIONS(1714),
    [anon_sym_loop] = ACTIONS(1714),
    [anon_sym_match] = ACTIONS(1714),
    [anon_sym_mod] = ACTIONS(1714),
    [anon_sym_pub] = ACTIONS(1714),
    [anon_sym_return] = ACTIONS(1714),
    [anon_sym_static] = ACTIONS(1714),
    [anon_sym_struct] = ACTIONS(1714),
    [anon_sym_trait] = ACTIONS(1714),
    [anon_sym_type] = ACTIONS(1714),
    [anon_sym_union] = ACTIONS(1714),
    [anon_sym_unsafe] = ACTIONS(1714),
    [anon_sym_use] = ACTIONS(1714),
    [anon_sym_while] = ACTIONS(1714),
    [anon_sym_POUND] = ACTIONS(1712),
    [anon_sym_BANG] = ACTIONS(1712),
    [anon_sym_extern] = ACTIONS(1714),
    [anon_sym_LT] = ACTIONS(1712),
    [anon_sym_COLON_COLON] = ACTIONS(1712),
    [anon_sym_AMP] = ACTIONS(1712),
    [anon_sym_DOT_DOT] = ACTIONS(1712),
    [anon_sym_DASH] = ACTIONS(1712),
    [anon_sym_PIPE] = ACTIONS(1712),
    [anon_sym_yield] = ACTIONS(1714),
    [anon_sym_move] = ACTIONS(1714),
    [sym_integer_literal] = ACTIONS(1712),
    [aux_sym_string_literal_token1] = ACTIONS(1712),
    [sym_char_literal] = ACTIONS(1712),
    [anon_sym_true] = ACTIONS(1714),
    [anon_sym_false] = ACTIONS(1714),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1714),
    [sym_super] = ACTIONS(1714),
    [sym_crate] = ACTIONS(1714),
    [sym_metavariable] = ACTIONS(1712),
    [sym_raw_string_literal] = ACTIONS(1712),
    [sym_float_literal] = ACTIONS(1712),
    [sym_block_comment] = ACTIONS(3),
  },
  [409] = {
    [ts_builtin_sym_end] = ACTIONS(1716),
    [sym_identifier] = ACTIONS(1718),
    [anon_sym_SEMI] = ACTIONS(1716),
    [anon_sym_macro_rules_BANG] = ACTIONS(1716),
    [anon_sym_LPAREN] = ACTIONS(1716),
    [anon_sym_LBRACE] = ACTIONS(1716),
    [anon_sym_RBRACE] = ACTIONS(1716),
    [anon_sym_LBRACK] = ACTIONS(1716),
    [anon_sym_STAR] = ACTIONS(1716),
    [anon_sym_u8] = ACTIONS(1718),
    [anon_sym_i8] = ACTIONS(1718),
    [anon_sym_u16] = ACTIONS(1718),
    [anon_sym_i16] = ACTIONS(1718),
    [anon_sym_u32] = ACTIONS(1718),
    [anon_sym_i32] = ACTIONS(1718),
    [anon_sym_u64] = ACTIONS(1718),
    [anon_sym_i64] = ACTIONS(1718),
    [anon_sym_u128] = ACTIONS(1718),
    [anon_sym_i128] = ACTIONS(1718),
    [anon_sym_isize] = ACTIONS(1718),
    [anon_sym_usize] = ACTIONS(1718),
    [anon_sym_f32] = ACTIONS(1718),
    [anon_sym_f64] = ACTIONS(1718),
    [anon_sym_bool] = ACTIONS(1718),
    [anon_sym_str] = ACTIONS(1718),
    [anon_sym_char] = ACTIONS(1718),
    [anon_sym_SQUOTE] = ACTIONS(1718),
    [anon_sym_async] = ACTIONS(1718),
    [anon_sym_break] = ACTIONS(1718),
    [anon_sym_const] = ACTIONS(1718),
    [anon_sym_continue] = ACTIONS(1718),
    [anon_sym_default] = ACTIONS(1718),
    [anon_sym_enum] = ACTIONS(1718),
    [anon_sym_fn] = ACTIONS(1718),
    [anon_sym_for] = ACTIONS(1718),
    [anon_sym_if] = ACTIONS(1718),
    [anon_sym_impl] = ACTIONS(1718),
    [anon_sym_let] = ACTIONS(1718),
    [anon_sym_loop] = ACTIONS(1718),
    [anon_sym_match] = ACTIONS(1718),
    [anon_sym_mod] = ACTIONS(1718),
    [anon_sym_pub] = ACTIONS(1718),
    [anon_sym_return] = ACTIONS(1718),
    [anon_sym_static] = ACTIONS(1718),
    [anon_sym_struct] = ACTIONS(1718),
    [anon_sym_trait] = ACTIONS(1718),
    [anon_sym_type] = ACTIONS(1718),
    [anon_sym_union] = ACTIONS(1718),
    [anon_sym_unsafe] = ACTIONS(1718),
    [anon_sym_use] = ACTIONS(1718),
    [anon_sym_while] = ACTIONS(1718),
    [anon_sym_POUND] = ACTIONS(1716),
    [anon_sym_BANG] = ACTIONS(1716),
    [anon_sym_extern] = ACTIONS(1718),
    [anon_sym_LT] = ACTIONS(1716),
    [anon_sym_COLON_COLON] = ACTIONS(1716),
    [anon_sym_AMP] = ACTIONS(1716),
    [anon_sym_DOT_DOT] = ACTIONS(1716),
    [anon_sym_DASH] = ACTIONS(1716),
    [anon_sym_PIPE] = ACTIONS(1716),
    [anon_sym_yield] = ACTIONS(1718),
    [anon_sym_move] = ACTIONS(1718),
    [sym_integer_literal] = ACTIONS(1716),
    [aux_sym_string_literal_token1] = ACTIONS(1716),
    [sym_char_literal] = ACTIONS(1716),
    [anon_sym_true] = ACTIONS(1718),
    [anon_sym_false] = ACTIONS(1718),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1718),
    [sym_super] = ACTIONS(1718),
    [sym_crate] = ACTIONS(1718),
    [sym_metavariable] = ACTIONS(1716),
    [sym_raw_string_literal] = ACTIONS(1716),
    [sym_float_literal] = ACTIONS(1716),
    [sym_block_comment] = ACTIONS(3),
  },
  [410] = {
    [ts_builtin_sym_end] = ACTIONS(1720),
    [sym_identifier] = ACTIONS(1722),
    [anon_sym_SEMI] = ACTIONS(1720),
    [anon_sym_macro_rules_BANG] = ACTIONS(1720),
    [anon_sym_LPAREN] = ACTIONS(1720),
    [anon_sym_LBRACE] = ACTIONS(1720),
    [anon_sym_RBRACE] = ACTIONS(1720),
    [anon_sym_LBRACK] = ACTIONS(1720),
    [anon_sym_STAR] = ACTIONS(1720),
    [anon_sym_u8] = ACTIONS(1722),
    [anon_sym_i8] = ACTIONS(1722),
    [anon_sym_u16] = ACTIONS(1722),
    [anon_sym_i16] = ACTIONS(1722),
    [anon_sym_u32] = ACTIONS(1722),
    [anon_sym_i32] = ACTIONS(1722),
    [anon_sym_u64] = ACTIONS(1722),
    [anon_sym_i64] = ACTIONS(1722),
    [anon_sym_u128] = ACTIONS(1722),
    [anon_sym_i128] = ACTIONS(1722),
    [anon_sym_isize] = ACTIONS(1722),
    [anon_sym_usize] = ACTIONS(1722),
    [anon_sym_f32] = ACTIONS(1722),
    [anon_sym_f64] = ACTIONS(1722),
    [anon_sym_bool] = ACTIONS(1722),
    [anon_sym_str] = ACTIONS(1722),
    [anon_sym_char] = ACTIONS(1722),
    [anon_sym_SQUOTE] = ACTIONS(1722),
    [anon_sym_async] = ACTIONS(1722),
    [anon_sym_break] = ACTIONS(1722),
    [anon_sym_const] = ACTIONS(1722),
    [anon_sym_continue] = ACTIONS(1722),
    [anon_sym_default] = ACTIONS(1722),
    [anon_sym_enum] = ACTIONS(1722),
    [anon_sym_fn] = ACTIONS(1722),
    [anon_sym_for] = ACTIONS(1722),
    [anon_sym_if] = ACTIONS(1722),
    [anon_sym_impl] = ACTIONS(1722),
    [anon_sym_let] = ACTIONS(1722),
    [anon_sym_loop] = ACTIONS(1722),
    [anon_sym_match] = ACTIONS(1722),
    [anon_sym_mod] = ACTIONS(1722),
    [anon_sym_pub] = ACTIONS(1722),
    [anon_sym_return] = ACTIONS(1722),
    [anon_sym_static] = ACTIONS(1722),
    [anon_sym_struct] = ACTIONS(1722),
    [anon_sym_trait] = ACTIONS(1722),
    [anon_sym_type] = ACTIONS(1722),
    [anon_sym_union] = ACTIONS(1722),
    [anon_sym_unsafe] = ACTIONS(1722),
    [anon_sym_use] = ACTIONS(1722),
    [anon_sym_while] = ACTIONS(1722),
    [anon_sym_POUND] = ACTIONS(1720),
    [anon_sym_BANG] = ACTIONS(1720),
    [anon_sym_extern] = ACTIONS(1722),
    [anon_sym_LT] = ACTIONS(1720),
    [anon_sym_COLON_COLON] = ACTIONS(1720),
    [anon_sym_AMP] = ACTIONS(1720),
    [anon_sym_DOT_DOT] = ACTIONS(1720),
    [anon_sym_DASH] = ACTIONS(1720),
    [anon_sym_PIPE] = ACTIONS(1720),
    [anon_sym_yield] = ACTIONS(1722),
    [anon_sym_move] = ACTIONS(1722),
    [sym_integer_literal] = ACTIONS(1720),
    [aux_sym_string_literal_token1] = ACTIONS(1720),
    [sym_char_literal] = ACTIONS(1720),
    [anon_sym_true] = ACTIONS(1722),
    [anon_sym_false] = ACTIONS(1722),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1722),
    [sym_super] = ACTIONS(1722),
    [sym_crate] = ACTIONS(1722),
    [sym_metavariable] = ACTIONS(1720),
    [sym_raw_string_literal] = ACTIONS(1720),
    [sym_float_literal] = ACTIONS(1720),
    [sym_block_comment] = ACTIONS(3),
  },
  [411] = {
    [ts_builtin_sym_end] = ACTIONS(1724),
    [sym_identifier] = ACTIONS(1726),
    [anon_sym_SEMI] = ACTIONS(1724),
    [anon_sym_macro_rules_BANG] = ACTIONS(1724),
    [anon_sym_LPAREN] = ACTIONS(1724),
    [anon_sym_LBRACE] = ACTIONS(1724),
    [anon_sym_RBRACE] = ACTIONS(1724),
    [anon_sym_LBRACK] = ACTIONS(1724),
    [anon_sym_STAR] = ACTIONS(1724),
    [anon_sym_u8] = ACTIONS(1726),
    [anon_sym_i8] = ACTIONS(1726),
    [anon_sym_u16] = ACTIONS(1726),
    [anon_sym_i16] = ACTIONS(1726),
    [anon_sym_u32] = ACTIONS(1726),
    [anon_sym_i32] = ACTIONS(1726),
    [anon_sym_u64] = ACTIONS(1726),
    [anon_sym_i64] = ACTIONS(1726),
    [anon_sym_u128] = ACTIONS(1726),
    [anon_sym_i128] = ACTIONS(1726),
    [anon_sym_isize] = ACTIONS(1726),
    [anon_sym_usize] = ACTIONS(1726),
    [anon_sym_f32] = ACTIONS(1726),
    [anon_sym_f64] = ACTIONS(1726),
    [anon_sym_bool] = ACTIONS(1726),
    [anon_sym_str] = ACTIONS(1726),
    [anon_sym_char] = ACTIONS(1726),
    [anon_sym_SQUOTE] = ACTIONS(1726),
    [anon_sym_async] = ACTIONS(1726),
    [anon_sym_break] = ACTIONS(1726),
    [anon_sym_const] = ACTIONS(1726),
    [anon_sym_continue] = ACTIONS(1726),
    [anon_sym_default] = ACTIONS(1726),
    [anon_sym_enum] = ACTIONS(1726),
    [anon_sym_fn] = ACTIONS(1726),
    [anon_sym_for] = ACTIONS(1726),
    [anon_sym_if] = ACTIONS(1726),
    [anon_sym_impl] = ACTIONS(1726),
    [anon_sym_let] = ACTIONS(1726),
    [anon_sym_loop] = ACTIONS(1726),
    [anon_sym_match] = ACTIONS(1726),
    [anon_sym_mod] = ACTIONS(1726),
    [anon_sym_pub] = ACTIONS(1726),
    [anon_sym_return] = ACTIONS(1726),
    [anon_sym_static] = ACTIONS(1726),
    [anon_sym_struct] = ACTIONS(1726),
    [anon_sym_trait] = ACTIONS(1726),
    [anon_sym_type] = ACTIONS(1726),
    [anon_sym_union] = ACTIONS(1726),
    [anon_sym_unsafe] = ACTIONS(1726),
    [anon_sym_use] = ACTIONS(1726),
    [anon_sym_while] = ACTIONS(1726),
    [anon_sym_POUND] = ACTIONS(1724),
    [anon_sym_BANG] = ACTIONS(1724),
    [anon_sym_extern] = ACTIONS(1726),
    [anon_sym_LT] = ACTIONS(1724),
    [anon_sym_COLON_COLON] = ACTIONS(1724),
    [anon_sym_AMP] = ACTIONS(1724),
    [anon_sym_DOT_DOT] = ACTIONS(1724),
    [anon_sym_DASH] = ACTIONS(1724),
    [anon_sym_PIPE] = ACTIONS(1724),
    [anon_sym_yield] = ACTIONS(1726),
    [anon_sym_move] = ACTIONS(1726),
    [sym_integer_literal] = ACTIONS(1724),
    [aux_sym_string_literal_token1] = ACTIONS(1724),
    [sym_char_literal] = ACTIONS(1724),
    [anon_sym_true] = ACTIONS(1726),
    [anon_sym_false] = ACTIONS(1726),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1726),
    [sym_super] = ACTIONS(1726),
    [sym_crate] = ACTIONS(1726),
    [sym_metavariable] = ACTIONS(1724),
    [sym_raw_string_literal] = ACTIONS(1724),
    [sym_float_literal] = ACTIONS(1724),
    [sym_block_comment] = ACTIONS(3),
  },
  [412] = {
    [ts_builtin_sym_end] = ACTIONS(1728),
    [sym_identifier] = ACTIONS(1730),
    [anon_sym_SEMI] = ACTIONS(1728),
    [anon_sym_macro_rules_BANG] = ACTIONS(1728),
    [anon_sym_LPAREN] = ACTIONS(1728),
    [anon_sym_LBRACE] = ACTIONS(1728),
    [anon_sym_RBRACE] = ACTIONS(1728),
    [anon_sym_LBRACK] = ACTIONS(1728),
    [anon_sym_STAR] = ACTIONS(1728),
    [anon_sym_u8] = ACTIONS(1730),
    [anon_sym_i8] = ACTIONS(1730),
    [anon_sym_u16] = ACTIONS(1730),
    [anon_sym_i16] = ACTIONS(1730),
    [anon_sym_u32] = ACTIONS(1730),
    [anon_sym_i32] = ACTIONS(1730),
    [anon_sym_u64] = ACTIONS(1730),
    [anon_sym_i64] = ACTIONS(1730),
    [anon_sym_u128] = ACTIONS(1730),
    [anon_sym_i128] = ACTIONS(1730),
    [anon_sym_isize] = ACTIONS(1730),
    [anon_sym_usize] = ACTIONS(1730),
    [anon_sym_f32] = ACTIONS(1730),
    [anon_sym_f64] = ACTIONS(1730),
    [anon_sym_bool] = ACTIONS(1730),
    [anon_sym_str] = ACTIONS(1730),
    [anon_sym_char] = ACTIONS(1730),
    [anon_sym_SQUOTE] = ACTIONS(1730),
    [anon_sym_async] = ACTIONS(1730),
    [anon_sym_break] = ACTIONS(1730),
    [anon_sym_const] = ACTIONS(1730),
    [anon_sym_continue] = ACTIONS(1730),
    [anon_sym_default] = ACTIONS(1730),
    [anon_sym_enum] = ACTIONS(1730),
    [anon_sym_fn] = ACTIONS(1730),
    [anon_sym_for] = ACTIONS(1730),
    [anon_sym_if] = ACTIONS(1730),
    [anon_sym_impl] = ACTIONS(1730),
    [anon_sym_let] = ACTIONS(1730),
    [anon_sym_loop] = ACTIONS(1730),
    [anon_sym_match] = ACTIONS(1730),
    [anon_sym_mod] = ACTIONS(1730),
    [anon_sym_pub] = ACTIONS(1730),
    [anon_sym_return] = ACTIONS(1730),
    [anon_sym_static] = ACTIONS(1730),
    [anon_sym_struct] = ACTIONS(1730),
    [anon_sym_trait] = ACTIONS(1730),
    [anon_sym_type] = ACTIONS(1730),
    [anon_sym_union] = ACTIONS(1730),
    [anon_sym_unsafe] = ACTIONS(1730),
    [anon_sym_use] = ACTIONS(1730),
    [anon_sym_while] = ACTIONS(1730),
    [anon_sym_POUND] = ACTIONS(1728),
    [anon_sym_BANG] = ACTIONS(1728),
    [anon_sym_extern] = ACTIONS(1730),
    [anon_sym_LT] = ACTIONS(1728),
    [anon_sym_COLON_COLON] = ACTIONS(1728),
    [anon_sym_AMP] = ACTIONS(1728),
    [anon_sym_DOT_DOT] = ACTIONS(1728),
    [anon_sym_DASH] = ACTIONS(1728),
    [anon_sym_PIPE] = ACTIONS(1728),
    [anon_sym_yield] = ACTIONS(1730),
    [anon_sym_move] = ACTIONS(1730),
    [sym_integer_literal] = ACTIONS(1728),
    [aux_sym_string_literal_token1] = ACTIONS(1728),
    [sym_char_literal] = ACTIONS(1728),
    [anon_sym_true] = ACTIONS(1730),
    [anon_sym_false] = ACTIONS(1730),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1730),
    [sym_super] = ACTIONS(1730),
    [sym_crate] = ACTIONS(1730),
    [sym_metavariable] = ACTIONS(1728),
    [sym_raw_string_literal] = ACTIONS(1728),
    [sym_float_literal] = ACTIONS(1728),
    [sym_block_comment] = ACTIONS(3),
  },
  [413] = {
    [ts_builtin_sym_end] = ACTIONS(1732),
    [sym_identifier] = ACTIONS(1734),
    [anon_sym_SEMI] = ACTIONS(1732),
    [anon_sym_macro_rules_BANG] = ACTIONS(1732),
    [anon_sym_LPAREN] = ACTIONS(1732),
    [anon_sym_LBRACE] = ACTIONS(1732),
    [anon_sym_RBRACE] = ACTIONS(1732),
    [anon_sym_LBRACK] = ACTIONS(1732),
    [anon_sym_STAR] = ACTIONS(1732),
    [anon_sym_u8] = ACTIONS(1734),
    [anon_sym_i8] = ACTIONS(1734),
    [anon_sym_u16] = ACTIONS(1734),
    [anon_sym_i16] = ACTIONS(1734),
    [anon_sym_u32] = ACTIONS(1734),
    [anon_sym_i32] = ACTIONS(1734),
    [anon_sym_u64] = ACTIONS(1734),
    [anon_sym_i64] = ACTIONS(1734),
    [anon_sym_u128] = ACTIONS(1734),
    [anon_sym_i128] = ACTIONS(1734),
    [anon_sym_isize] = ACTIONS(1734),
    [anon_sym_usize] = ACTIONS(1734),
    [anon_sym_f32] = ACTIONS(1734),
    [anon_sym_f64] = ACTIONS(1734),
    [anon_sym_bool] = ACTIONS(1734),
    [anon_sym_str] = ACTIONS(1734),
    [anon_sym_char] = ACTIONS(1734),
    [anon_sym_SQUOTE] = ACTIONS(1734),
    [anon_sym_async] = ACTIONS(1734),
    [anon_sym_break] = ACTIONS(1734),
    [anon_sym_const] = ACTIONS(1734),
    [anon_sym_continue] = ACTIONS(1734),
    [anon_sym_default] = ACTIONS(1734),
    [anon_sym_enum] = ACTIONS(1734),
    [anon_sym_fn] = ACTIONS(1734),
    [anon_sym_for] = ACTIONS(1734),
    [anon_sym_if] = ACTIONS(1734),
    [anon_sym_impl] = ACTIONS(1734),
    [anon_sym_let] = ACTIONS(1734),
    [anon_sym_loop] = ACTIONS(1734),
    [anon_sym_match] = ACTIONS(1734),
    [anon_sym_mod] = ACTIONS(1734),
    [anon_sym_pub] = ACTIONS(1734),
    [anon_sym_return] = ACTIONS(1734),
    [anon_sym_static] = ACTIONS(1734),
    [anon_sym_struct] = ACTIONS(1734),
    [anon_sym_trait] = ACTIONS(1734),
    [anon_sym_type] = ACTIONS(1734),
    [anon_sym_union] = ACTIONS(1734),
    [anon_sym_unsafe] = ACTIONS(1734),
    [anon_sym_use] = ACTIONS(1734),
    [anon_sym_while] = ACTIONS(1734),
    [anon_sym_POUND] = ACTIONS(1732),
    [anon_sym_BANG] = ACTIONS(1732),
    [anon_sym_extern] = ACTIONS(1734),
    [anon_sym_LT] = ACTIONS(1732),
    [anon_sym_COLON_COLON] = ACTIONS(1732),
    [anon_sym_AMP] = ACTIONS(1732),
    [anon_sym_DOT_DOT] = ACTIONS(1732),
    [anon_sym_DASH] = ACTIONS(1732),
    [anon_sym_PIPE] = ACTIONS(1732),
    [anon_sym_yield] = ACTIONS(1734),
    [anon_sym_move] = ACTIONS(1734),
    [sym_integer_literal] = ACTIONS(1732),
    [aux_sym_string_literal_token1] = ACTIONS(1732),
    [sym_char_literal] = ACTIONS(1732),
    [anon_sym_true] = ACTIONS(1734),
    [anon_sym_false] = ACTIONS(1734),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1734),
    [sym_super] = ACTIONS(1734),
    [sym_crate] = ACTIONS(1734),
    [sym_metavariable] = ACTIONS(1732),
    [sym_raw_string_literal] = ACTIONS(1732),
    [sym_float_literal] = ACTIONS(1732),
    [sym_block_comment] = ACTIONS(3),
  },
  [414] = {
    [ts_builtin_sym_end] = ACTIONS(1736),
    [sym_identifier] = ACTIONS(1738),
    [anon_sym_SEMI] = ACTIONS(1736),
    [anon_sym_macro_rules_BANG] = ACTIONS(1736),
    [anon_sym_LPAREN] = ACTIONS(1736),
    [anon_sym_LBRACE] = ACTIONS(1736),
    [anon_sym_RBRACE] = ACTIONS(1736),
    [anon_sym_LBRACK] = ACTIONS(1736),
    [anon_sym_STAR] = ACTIONS(1736),
    [anon_sym_u8] = ACTIONS(1738),
    [anon_sym_i8] = ACTIONS(1738),
    [anon_sym_u16] = ACTIONS(1738),
    [anon_sym_i16] = ACTIONS(1738),
    [anon_sym_u32] = ACTIONS(1738),
    [anon_sym_i32] = ACTIONS(1738),
    [anon_sym_u64] = ACTIONS(1738),
    [anon_sym_i64] = ACTIONS(1738),
    [anon_sym_u128] = ACTIONS(1738),
    [anon_sym_i128] = ACTIONS(1738),
    [anon_sym_isize] = ACTIONS(1738),
    [anon_sym_usize] = ACTIONS(1738),
    [anon_sym_f32] = ACTIONS(1738),
    [anon_sym_f64] = ACTIONS(1738),
    [anon_sym_bool] = ACTIONS(1738),
    [anon_sym_str] = ACTIONS(1738),
    [anon_sym_char] = ACTIONS(1738),
    [anon_sym_SQUOTE] = ACTIONS(1738),
    [anon_sym_async] = ACTIONS(1738),
    [anon_sym_break] = ACTIONS(1738),
    [anon_sym_const] = ACTIONS(1738),
    [anon_sym_continue] = ACTIONS(1738),
    [anon_sym_default] = ACTIONS(1738),
    [anon_sym_enum] = ACTIONS(1738),
    [anon_sym_fn] = ACTIONS(1738),
    [anon_sym_for] = ACTIONS(1738),
    [anon_sym_if] = ACTIONS(1738),
    [anon_sym_impl] = ACTIONS(1738),
    [anon_sym_let] = ACTIONS(1738),
    [anon_sym_loop] = ACTIONS(1738),
    [anon_sym_match] = ACTIONS(1738),
    [anon_sym_mod] = ACTIONS(1738),
    [anon_sym_pub] = ACTIONS(1738),
    [anon_sym_return] = ACTIONS(1738),
    [anon_sym_static] = ACTIONS(1738),
    [anon_sym_struct] = ACTIONS(1738),
    [anon_sym_trait] = ACTIONS(1738),
    [anon_sym_type] = ACTIONS(1738),
    [anon_sym_union] = ACTIONS(1738),
    [anon_sym_unsafe] = ACTIONS(1738),
    [anon_sym_use] = ACTIONS(1738),
    [anon_sym_while] = ACTIONS(1738),
    [anon_sym_POUND] = ACTIONS(1736),
    [anon_sym_BANG] = ACTIONS(1736),
    [anon_sym_extern] = ACTIONS(1738),
    [anon_sym_LT] = ACTIONS(1736),
    [anon_sym_COLON_COLON] = ACTIONS(1736),
    [anon_sym_AMP] = ACTIONS(1736),
    [anon_sym_DOT_DOT] = ACTIONS(1736),
    [anon_sym_DASH] = ACTIONS(1736),
    [anon_sym_PIPE] = ACTIONS(1736),
    [anon_sym_yield] = ACTIONS(1738),
    [anon_sym_move] = ACTIONS(1738),
    [sym_integer_literal] = ACTIONS(1736),
    [aux_sym_string_literal_token1] = ACTIONS(1736),
    [sym_char_literal] = ACTIONS(1736),
    [anon_sym_true] = ACTIONS(1738),
    [anon_sym_false] = ACTIONS(1738),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1738),
    [sym_super] = ACTIONS(1738),
    [sym_crate] = ACTIONS(1738),
    [sym_metavariable] = ACTIONS(1736),
    [sym_raw_string_literal] = ACTIONS(1736),
    [sym_float_literal] = ACTIONS(1736),
    [sym_block_comment] = ACTIONS(3),
  },
  [415] = {
    [ts_builtin_sym_end] = ACTIONS(1740),
    [sym_identifier] = ACTIONS(1742),
    [anon_sym_SEMI] = ACTIONS(1740),
    [anon_sym_macro_rules_BANG] = ACTIONS(1740),
    [anon_sym_LPAREN] = ACTIONS(1740),
    [anon_sym_LBRACE] = ACTIONS(1740),
    [anon_sym_RBRACE] = ACTIONS(1740),
    [anon_sym_LBRACK] = ACTIONS(1740),
    [anon_sym_STAR] = ACTIONS(1740),
    [anon_sym_u8] = ACTIONS(1742),
    [anon_sym_i8] = ACTIONS(1742),
    [anon_sym_u16] = ACTIONS(1742),
    [anon_sym_i16] = ACTIONS(1742),
    [anon_sym_u32] = ACTIONS(1742),
    [anon_sym_i32] = ACTIONS(1742),
    [anon_sym_u64] = ACTIONS(1742),
    [anon_sym_i64] = ACTIONS(1742),
    [anon_sym_u128] = ACTIONS(1742),
    [anon_sym_i128] = ACTIONS(1742),
    [anon_sym_isize] = ACTIONS(1742),
    [anon_sym_usize] = ACTIONS(1742),
    [anon_sym_f32] = ACTIONS(1742),
    [anon_sym_f64] = ACTIONS(1742),
    [anon_sym_bool] = ACTIONS(1742),
    [anon_sym_str] = ACTIONS(1742),
    [anon_sym_char] = ACTIONS(1742),
    [anon_sym_SQUOTE] = ACTIONS(1742),
    [anon_sym_async] = ACTIONS(1742),
    [anon_sym_break] = ACTIONS(1742),
    [anon_sym_const] = ACTIONS(1742),
    [anon_sym_continue] = ACTIONS(1742),
    [anon_sym_default] = ACTIONS(1742),
    [anon_sym_enum] = ACTIONS(1742),
    [anon_sym_fn] = ACTIONS(1742),
    [anon_sym_for] = ACTIONS(1742),
    [anon_sym_if] = ACTIONS(1742),
    [anon_sym_impl] = ACTIONS(1742),
    [anon_sym_let] = ACTIONS(1742),
    [anon_sym_loop] = ACTIONS(1742),
    [anon_sym_match] = ACTIONS(1742),
    [anon_sym_mod] = ACTIONS(1742),
    [anon_sym_pub] = ACTIONS(1742),
    [anon_sym_return] = ACTIONS(1742),
    [anon_sym_static] = ACTIONS(1742),
    [anon_sym_struct] = ACTIONS(1742),
    [anon_sym_trait] = ACTIONS(1742),
    [anon_sym_type] = ACTIONS(1742),
    [anon_sym_union] = ACTIONS(1742),
    [anon_sym_unsafe] = ACTIONS(1742),
    [anon_sym_use] = ACTIONS(1742),
    [anon_sym_while] = ACTIONS(1742),
    [anon_sym_POUND] = ACTIONS(1740),
    [anon_sym_BANG] = ACTIONS(1740),
    [anon_sym_extern] = ACTIONS(1742),
    [anon_sym_LT] = ACTIONS(1740),
    [anon_sym_COLON_COLON] = ACTIONS(1740),
    [anon_sym_AMP] = ACTIONS(1740),
    [anon_sym_DOT_DOT] = ACTIONS(1740),
    [anon_sym_DASH] = ACTIONS(1740),
    [anon_sym_PIPE] = ACTIONS(1740),
    [anon_sym_yield] = ACTIONS(1742),
    [anon_sym_move] = ACTIONS(1742),
    [sym_integer_literal] = ACTIONS(1740),
    [aux_sym_string_literal_token1] = ACTIONS(1740),
    [sym_char_literal] = ACTIONS(1740),
    [anon_sym_true] = ACTIONS(1742),
    [anon_sym_false] = ACTIONS(1742),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1742),
    [sym_super] = ACTIONS(1742),
    [sym_crate] = ACTIONS(1742),
    [sym_metavariable] = ACTIONS(1740),
    [sym_raw_string_literal] = ACTIONS(1740),
    [sym_float_literal] = ACTIONS(1740),
    [sym_block_comment] = ACTIONS(3),
  },
  [416] = {
    [ts_builtin_sym_end] = ACTIONS(1744),
    [sym_identifier] = ACTIONS(1746),
    [anon_sym_SEMI] = ACTIONS(1744),
    [anon_sym_macro_rules_BANG] = ACTIONS(1744),
    [anon_sym_LPAREN] = ACTIONS(1744),
    [anon_sym_LBRACE] = ACTIONS(1744),
    [anon_sym_RBRACE] = ACTIONS(1744),
    [anon_sym_LBRACK] = ACTIONS(1744),
    [anon_sym_STAR] = ACTIONS(1744),
    [anon_sym_u8] = ACTIONS(1746),
    [anon_sym_i8] = ACTIONS(1746),
    [anon_sym_u16] = ACTIONS(1746),
    [anon_sym_i16] = ACTIONS(1746),
    [anon_sym_u32] = ACTIONS(1746),
    [anon_sym_i32] = ACTIONS(1746),
    [anon_sym_u64] = ACTIONS(1746),
    [anon_sym_i64] = ACTIONS(1746),
    [anon_sym_u128] = ACTIONS(1746),
    [anon_sym_i128] = ACTIONS(1746),
    [anon_sym_isize] = ACTIONS(1746),
    [anon_sym_usize] = ACTIONS(1746),
    [anon_sym_f32] = ACTIONS(1746),
    [anon_sym_f64] = ACTIONS(1746),
    [anon_sym_bool] = ACTIONS(1746),
    [anon_sym_str] = ACTIONS(1746),
    [anon_sym_char] = ACTIONS(1746),
    [anon_sym_SQUOTE] = ACTIONS(1746),
    [anon_sym_async] = ACTIONS(1746),
    [anon_sym_break] = ACTIONS(1746),
    [anon_sym_const] = ACTIONS(1746),
    [anon_sym_continue] = ACTIONS(1746),
    [anon_sym_default] = ACTIONS(1746),
    [anon_sym_enum] = ACTIONS(1746),
    [anon_sym_fn] = ACTIONS(1746),
    [anon_sym_for] = ACTIONS(1746),
    [anon_sym_if] = ACTIONS(1746),
    [anon_sym_impl] = ACTIONS(1746),
    [anon_sym_let] = ACTIONS(1746),
    [anon_sym_loop] = ACTIONS(1746),
    [anon_sym_match] = ACTIONS(1746),
    [anon_sym_mod] = ACTIONS(1746),
    [anon_sym_pub] = ACTIONS(1746),
    [anon_sym_return] = ACTIONS(1746),
    [anon_sym_static] = ACTIONS(1746),
    [anon_sym_struct] = ACTIONS(1746),
    [anon_sym_trait] = ACTIONS(1746),
    [anon_sym_type] = ACTIONS(1746),
    [anon_sym_union] = ACTIONS(1746),
    [anon_sym_unsafe] = ACTIONS(1746),
    [anon_sym_use] = ACTIONS(1746),
    [anon_sym_while] = ACTIONS(1746),
    [anon_sym_POUND] = ACTIONS(1744),
    [anon_sym_BANG] = ACTIONS(1744),
    [anon_sym_extern] = ACTIONS(1746),
    [anon_sym_LT] = ACTIONS(1744),
    [anon_sym_COLON_COLON] = ACTIONS(1744),
    [anon_sym_AMP] = ACTIONS(1744),
    [anon_sym_DOT_DOT] = ACTIONS(1744),
    [anon_sym_DASH] = ACTIONS(1744),
    [anon_sym_PIPE] = ACTIONS(1744),
    [anon_sym_yield] = ACTIONS(1746),
    [anon_sym_move] = ACTIONS(1746),
    [sym_integer_literal] = ACTIONS(1744),
    [aux_sym_string_literal_token1] = ACTIONS(1744),
    [sym_char_literal] = ACTIONS(1744),
    [anon_sym_true] = ACTIONS(1746),
    [anon_sym_false] = ACTIONS(1746),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1746),
    [sym_super] = ACTIONS(1746),
    [sym_crate] = ACTIONS(1746),
    [sym_metavariable] = ACTIONS(1744),
    [sym_raw_string_literal] = ACTIONS(1744),
    [sym_float_literal] = ACTIONS(1744),
    [sym_block_comment] = ACTIONS(3),
  },
  [417] = {
    [ts_builtin_sym_end] = ACTIONS(1748),
    [sym_identifier] = ACTIONS(1750),
    [anon_sym_SEMI] = ACTIONS(1748),
    [anon_sym_macro_rules_BANG] = ACTIONS(1748),
    [anon_sym_LPAREN] = ACTIONS(1748),
    [anon_sym_LBRACE] = ACTIONS(1748),
    [anon_sym_RBRACE] = ACTIONS(1748),
    [anon_sym_LBRACK] = ACTIONS(1748),
    [anon_sym_STAR] = ACTIONS(1748),
    [anon_sym_u8] = ACTIONS(1750),
    [anon_sym_i8] = ACTIONS(1750),
    [anon_sym_u16] = ACTIONS(1750),
    [anon_sym_i16] = ACTIONS(1750),
    [anon_sym_u32] = ACTIONS(1750),
    [anon_sym_i32] = ACTIONS(1750),
    [anon_sym_u64] = ACTIONS(1750),
    [anon_sym_i64] = ACTIONS(1750),
    [anon_sym_u128] = ACTIONS(1750),
    [anon_sym_i128] = ACTIONS(1750),
    [anon_sym_isize] = ACTIONS(1750),
    [anon_sym_usize] = ACTIONS(1750),
    [anon_sym_f32] = ACTIONS(1750),
    [anon_sym_f64] = ACTIONS(1750),
    [anon_sym_bool] = ACTIONS(1750),
    [anon_sym_str] = ACTIONS(1750),
    [anon_sym_char] = ACTIONS(1750),
    [anon_sym_SQUOTE] = ACTIONS(1750),
    [anon_sym_async] = ACTIONS(1750),
    [anon_sym_break] = ACTIONS(1750),
    [anon_sym_const] = ACTIONS(1750),
    [anon_sym_continue] = ACTIONS(1750),
    [anon_sym_default] = ACTIONS(1750),
    [anon_sym_enum] = ACTIONS(1750),
    [anon_sym_fn] = ACTIONS(1750),
    [anon_sym_for] = ACTIONS(1750),
    [anon_sym_if] = ACTIONS(1750),
    [anon_sym_impl] = ACTIONS(1750),
    [anon_sym_let] = ACTIONS(1750),
    [anon_sym_loop] = ACTIONS(1750),
    [anon_sym_match] = ACTIONS(1750),
    [anon_sym_mod] = ACTIONS(1750),
    [anon_sym_pub] = ACTIONS(1750),
    [anon_sym_return] = ACTIONS(1750),
    [anon_sym_static] = ACTIONS(1750),
    [anon_sym_struct] = ACTIONS(1750),
    [anon_sym_trait] = ACTIONS(1750),
    [anon_sym_type] = ACTIONS(1750),
    [anon_sym_union] = ACTIONS(1750),
    [anon_sym_unsafe] = ACTIONS(1750),
    [anon_sym_use] = ACTIONS(1750),
    [anon_sym_while] = ACTIONS(1750),
    [anon_sym_POUND] = ACTIONS(1748),
    [anon_sym_BANG] = ACTIONS(1748),
    [anon_sym_extern] = ACTIONS(1750),
    [anon_sym_LT] = ACTIONS(1748),
    [anon_sym_COLON_COLON] = ACTIONS(1748),
    [anon_sym_AMP] = ACTIONS(1748),
    [anon_sym_DOT_DOT] = ACTIONS(1748),
    [anon_sym_DASH] = ACTIONS(1748),
    [anon_sym_PIPE] = ACTIONS(1748),
    [anon_sym_yield] = ACTIONS(1750),
    [anon_sym_move] = ACTIONS(1750),
    [sym_integer_literal] = ACTIONS(1748),
    [aux_sym_string_literal_token1] = ACTIONS(1748),
    [sym_char_literal] = ACTIONS(1748),
    [anon_sym_true] = ACTIONS(1750),
    [anon_sym_false] = ACTIONS(1750),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1750),
    [sym_super] = ACTIONS(1750),
    [sym_crate] = ACTIONS(1750),
    [sym_metavariable] = ACTIONS(1748),
    [sym_raw_string_literal] = ACTIONS(1748),
    [sym_float_literal] = ACTIONS(1748),
    [sym_block_comment] = ACTIONS(3),
  },
  [418] = {
    [ts_builtin_sym_end] = ACTIONS(1752),
    [sym_identifier] = ACTIONS(1754),
    [anon_sym_SEMI] = ACTIONS(1752),
    [anon_sym_macro_rules_BANG] = ACTIONS(1752),
    [anon_sym_LPAREN] = ACTIONS(1752),
    [anon_sym_LBRACE] = ACTIONS(1752),
    [anon_sym_RBRACE] = ACTIONS(1752),
    [anon_sym_LBRACK] = ACTIONS(1752),
    [anon_sym_STAR] = ACTIONS(1752),
    [anon_sym_u8] = ACTIONS(1754),
    [anon_sym_i8] = ACTIONS(1754),
    [anon_sym_u16] = ACTIONS(1754),
    [anon_sym_i16] = ACTIONS(1754),
    [anon_sym_u32] = ACTIONS(1754),
    [anon_sym_i32] = ACTIONS(1754),
    [anon_sym_u64] = ACTIONS(1754),
    [anon_sym_i64] = ACTIONS(1754),
    [anon_sym_u128] = ACTIONS(1754),
    [anon_sym_i128] = ACTIONS(1754),
    [anon_sym_isize] = ACTIONS(1754),
    [anon_sym_usize] = ACTIONS(1754),
    [anon_sym_f32] = ACTIONS(1754),
    [anon_sym_f64] = ACTIONS(1754),
    [anon_sym_bool] = ACTIONS(1754),
    [anon_sym_str] = ACTIONS(1754),
    [anon_sym_char] = ACTIONS(1754),
    [anon_sym_SQUOTE] = ACTIONS(1754),
    [anon_sym_async] = ACTIONS(1754),
    [anon_sym_break] = ACTIONS(1754),
    [anon_sym_const] = ACTIONS(1754),
    [anon_sym_continue] = ACTIONS(1754),
    [anon_sym_default] = ACTIONS(1754),
    [anon_sym_enum] = ACTIONS(1754),
    [anon_sym_fn] = ACTIONS(1754),
    [anon_sym_for] = ACTIONS(1754),
    [anon_sym_if] = ACTIONS(1754),
    [anon_sym_impl] = ACTIONS(1754),
    [anon_sym_let] = ACTIONS(1754),
    [anon_sym_loop] = ACTIONS(1754),
    [anon_sym_match] = ACTIONS(1754),
    [anon_sym_mod] = ACTIONS(1754),
    [anon_sym_pub] = ACTIONS(1754),
    [anon_sym_return] = ACTIONS(1754),
    [anon_sym_static] = ACTIONS(1754),
    [anon_sym_struct] = ACTIONS(1754),
    [anon_sym_trait] = ACTIONS(1754),
    [anon_sym_type] = ACTIONS(1754),
    [anon_sym_union] = ACTIONS(1754),
    [anon_sym_unsafe] = ACTIONS(1754),
    [anon_sym_use] = ACTIONS(1754),
    [anon_sym_while] = ACTIONS(1754),
    [anon_sym_POUND] = ACTIONS(1752),
    [anon_sym_BANG] = ACTIONS(1752),
    [anon_sym_extern] = ACTIONS(1754),
    [anon_sym_LT] = ACTIONS(1752),
    [anon_sym_COLON_COLON] = ACTIONS(1752),
    [anon_sym_AMP] = ACTIONS(1752),
    [anon_sym_DOT_DOT] = ACTIONS(1752),
    [anon_sym_DASH] = ACTIONS(1752),
    [anon_sym_PIPE] = ACTIONS(1752),
    [anon_sym_yield] = ACTIONS(1754),
    [anon_sym_move] = ACTIONS(1754),
    [sym_integer_literal] = ACTIONS(1752),
    [aux_sym_string_literal_token1] = ACTIONS(1752),
    [sym_char_literal] = ACTIONS(1752),
    [anon_sym_true] = ACTIONS(1754),
    [anon_sym_false] = ACTIONS(1754),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1754),
    [sym_super] = ACTIONS(1754),
    [sym_crate] = ACTIONS(1754),
    [sym_metavariable] = ACTIONS(1752),
    [sym_raw_string_literal] = ACTIONS(1752),
    [sym_float_literal] = ACTIONS(1752),
    [sym_block_comment] = ACTIONS(3),
  },
  [419] = {
    [ts_builtin_sym_end] = ACTIONS(1756),
    [sym_identifier] = ACTIONS(1758),
    [anon_sym_SEMI] = ACTIONS(1756),
    [anon_sym_macro_rules_BANG] = ACTIONS(1756),
    [anon_sym_LPAREN] = ACTIONS(1756),
    [anon_sym_LBRACE] = ACTIONS(1756),
    [anon_sym_RBRACE] = ACTIONS(1756),
    [anon_sym_LBRACK] = ACTIONS(1756),
    [anon_sym_STAR] = ACTIONS(1756),
    [anon_sym_u8] = ACTIONS(1758),
    [anon_sym_i8] = ACTIONS(1758),
    [anon_sym_u16] = ACTIONS(1758),
    [anon_sym_i16] = ACTIONS(1758),
    [anon_sym_u32] = ACTIONS(1758),
    [anon_sym_i32] = ACTIONS(1758),
    [anon_sym_u64] = ACTIONS(1758),
    [anon_sym_i64] = ACTIONS(1758),
    [anon_sym_u128] = ACTIONS(1758),
    [anon_sym_i128] = ACTIONS(1758),
    [anon_sym_isize] = ACTIONS(1758),
    [anon_sym_usize] = ACTIONS(1758),
    [anon_sym_f32] = ACTIONS(1758),
    [anon_sym_f64] = ACTIONS(1758),
    [anon_sym_bool] = ACTIONS(1758),
    [anon_sym_str] = ACTIONS(1758),
    [anon_sym_char] = ACTIONS(1758),
    [anon_sym_SQUOTE] = ACTIONS(1758),
    [anon_sym_async] = ACTIONS(1758),
    [anon_sym_break] = ACTIONS(1758),
    [anon_sym_const] = ACTIONS(1758),
    [anon_sym_continue] = ACTIONS(1758),
    [anon_sym_default] = ACTIONS(1758),
    [anon_sym_enum] = ACTIONS(1758),
    [anon_sym_fn] = ACTIONS(1758),
    [anon_sym_for] = ACTIONS(1758),
    [anon_sym_if] = ACTIONS(1758),
    [anon_sym_impl] = ACTIONS(1758),
    [anon_sym_let] = ACTIONS(1758),
    [anon_sym_loop] = ACTIONS(1758),
    [anon_sym_match] = ACTIONS(1758),
    [anon_sym_mod] = ACTIONS(1758),
    [anon_sym_pub] = ACTIONS(1758),
    [anon_sym_return] = ACTIONS(1758),
    [anon_sym_static] = ACTIONS(1758),
    [anon_sym_struct] = ACTIONS(1758),
    [anon_sym_trait] = ACTIONS(1758),
    [anon_sym_type] = ACTIONS(1758),
    [anon_sym_union] = ACTIONS(1758),
    [anon_sym_unsafe] = ACTIONS(1758),
    [anon_sym_use] = ACTIONS(1758),
    [anon_sym_while] = ACTIONS(1758),
    [anon_sym_POUND] = ACTIONS(1756),
    [anon_sym_BANG] = ACTIONS(1756),
    [anon_sym_extern] = ACTIONS(1758),
    [anon_sym_LT] = ACTIONS(1756),
    [anon_sym_COLON_COLON] = ACTIONS(1756),
    [anon_sym_AMP] = ACTIONS(1756),
    [anon_sym_DOT_DOT] = ACTIONS(1756),
    [anon_sym_DASH] = ACTIONS(1756),
    [anon_sym_PIPE] = ACTIONS(1756),
    [anon_sym_yield] = ACTIONS(1758),
    [anon_sym_move] = ACTIONS(1758),
    [sym_integer_literal] = ACTIONS(1756),
    [aux_sym_string_literal_token1] = ACTIONS(1756),
    [sym_char_literal] = ACTIONS(1756),
    [anon_sym_true] = ACTIONS(1758),
    [anon_sym_false] = ACTIONS(1758),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1758),
    [sym_super] = ACTIONS(1758),
    [sym_crate] = ACTIONS(1758),
    [sym_metavariable] = ACTIONS(1756),
    [sym_raw_string_literal] = ACTIONS(1756),
    [sym_float_literal] = ACTIONS(1756),
    [sym_block_comment] = ACTIONS(3),
  },
  [420] = {
    [ts_builtin_sym_end] = ACTIONS(1760),
    [sym_identifier] = ACTIONS(1762),
    [anon_sym_SEMI] = ACTIONS(1760),
    [anon_sym_macro_rules_BANG] = ACTIONS(1760),
    [anon_sym_LPAREN] = ACTIONS(1760),
    [anon_sym_LBRACE] = ACTIONS(1760),
    [anon_sym_RBRACE] = ACTIONS(1760),
    [anon_sym_LBRACK] = ACTIONS(1760),
    [anon_sym_STAR] = ACTIONS(1760),
    [anon_sym_u8] = ACTIONS(1762),
    [anon_sym_i8] = ACTIONS(1762),
    [anon_sym_u16] = ACTIONS(1762),
    [anon_sym_i16] = ACTIONS(1762),
    [anon_sym_u32] = ACTIONS(1762),
    [anon_sym_i32] = ACTIONS(1762),
    [anon_sym_u64] = ACTIONS(1762),
    [anon_sym_i64] = ACTIONS(1762),
    [anon_sym_u128] = ACTIONS(1762),
    [anon_sym_i128] = ACTIONS(1762),
    [anon_sym_isize] = ACTIONS(1762),
    [anon_sym_usize] = ACTIONS(1762),
    [anon_sym_f32] = ACTIONS(1762),
    [anon_sym_f64] = ACTIONS(1762),
    [anon_sym_bool] = ACTIONS(1762),
    [anon_sym_str] = ACTIONS(1762),
    [anon_sym_char] = ACTIONS(1762),
    [anon_sym_SQUOTE] = ACTIONS(1762),
    [anon_sym_async] = ACTIONS(1762),
    [anon_sym_break] = ACTIONS(1762),
    [anon_sym_const] = ACTIONS(1762),
    [anon_sym_continue] = ACTIONS(1762),
    [anon_sym_default] = ACTIONS(1762),
    [anon_sym_enum] = ACTIONS(1762),
    [anon_sym_fn] = ACTIONS(1762),
    [anon_sym_for] = ACTIONS(1762),
    [anon_sym_if] = ACTIONS(1762),
    [anon_sym_impl] = ACTIONS(1762),
    [anon_sym_let] = ACTIONS(1762),
    [anon_sym_loop] = ACTIONS(1762),
    [anon_sym_match] = ACTIONS(1762),
    [anon_sym_mod] = ACTIONS(1762),
    [anon_sym_pub] = ACTIONS(1762),
    [anon_sym_return] = ACTIONS(1762),
    [anon_sym_static] = ACTIONS(1762),
    [anon_sym_struct] = ACTIONS(1762),
    [anon_sym_trait] = ACTIONS(1762),
    [anon_sym_type] = ACTIONS(1762),
    [anon_sym_union] = ACTIONS(1762),
    [anon_sym_unsafe] = ACTIONS(1762),
    [anon_sym_use] = ACTIONS(1762),
    [anon_sym_while] = ACTIONS(1762),
    [anon_sym_POUND] = ACTIONS(1760),
    [anon_sym_BANG] = ACTIONS(1760),
    [anon_sym_extern] = ACTIONS(1762),
    [anon_sym_LT] = ACTIONS(1760),
    [anon_sym_COLON_COLON] = ACTIONS(1760),
    [anon_sym_AMP] = ACTIONS(1760),
    [anon_sym_DOT_DOT] = ACTIONS(1760),
    [anon_sym_DASH] = ACTIONS(1760),
    [anon_sym_PIPE] = ACTIONS(1760),
    [anon_sym_yield] = ACTIONS(1762),
    [anon_sym_move] = ACTIONS(1762),
    [sym_integer_literal] = ACTIONS(1760),
    [aux_sym_string_literal_token1] = ACTIONS(1760),
    [sym_char_literal] = ACTIONS(1760),
    [anon_sym_true] = ACTIONS(1762),
    [anon_sym_false] = ACTIONS(1762),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1762),
    [sym_super] = ACTIONS(1762),
    [sym_crate] = ACTIONS(1762),
    [sym_metavariable] = ACTIONS(1760),
    [sym_raw_string_literal] = ACTIONS(1760),
    [sym_float_literal] = ACTIONS(1760),
    [sym_block_comment] = ACTIONS(3),
  },
  [421] = {
    [ts_builtin_sym_end] = ACTIONS(1764),
    [sym_identifier] = ACTIONS(1766),
    [anon_sym_SEMI] = ACTIONS(1764),
    [anon_sym_macro_rules_BANG] = ACTIONS(1764),
    [anon_sym_LPAREN] = ACTIONS(1764),
    [anon_sym_LBRACE] = ACTIONS(1764),
    [anon_sym_RBRACE] = ACTIONS(1764),
    [anon_sym_LBRACK] = ACTIONS(1764),
    [anon_sym_STAR] = ACTIONS(1764),
    [anon_sym_u8] = ACTIONS(1766),
    [anon_sym_i8] = ACTIONS(1766),
    [anon_sym_u16] = ACTIONS(1766),
    [anon_sym_i16] = ACTIONS(1766),
    [anon_sym_u32] = ACTIONS(1766),
    [anon_sym_i32] = ACTIONS(1766),
    [anon_sym_u64] = ACTIONS(1766),
    [anon_sym_i64] = ACTIONS(1766),
    [anon_sym_u128] = ACTIONS(1766),
    [anon_sym_i128] = ACTIONS(1766),
    [anon_sym_isize] = ACTIONS(1766),
    [anon_sym_usize] = ACTIONS(1766),
    [anon_sym_f32] = ACTIONS(1766),
    [anon_sym_f64] = ACTIONS(1766),
    [anon_sym_bool] = ACTIONS(1766),
    [anon_sym_str] = ACTIONS(1766),
    [anon_sym_char] = ACTIONS(1766),
    [anon_sym_SQUOTE] = ACTIONS(1766),
    [anon_sym_async] = ACTIONS(1766),
    [anon_sym_break] = ACTIONS(1766),
    [anon_sym_const] = ACTIONS(1766),
    [anon_sym_continue] = ACTIONS(1766),
    [anon_sym_default] = ACTIONS(1766),
    [anon_sym_enum] = ACTIONS(1766),
    [anon_sym_fn] = ACTIONS(1766),
    [anon_sym_for] = ACTIONS(1766),
    [anon_sym_if] = ACTIONS(1766),
    [anon_sym_impl] = ACTIONS(1766),
    [anon_sym_let] = ACTIONS(1766),
    [anon_sym_loop] = ACTIONS(1766),
    [anon_sym_match] = ACTIONS(1766),
    [anon_sym_mod] = ACTIONS(1766),
    [anon_sym_pub] = ACTIONS(1766),
    [anon_sym_return] = ACTIONS(1766),
    [anon_sym_static] = ACTIONS(1766),
    [anon_sym_struct] = ACTIONS(1766),
    [anon_sym_trait] = ACTIONS(1766),
    [anon_sym_type] = ACTIONS(1766),
    [anon_sym_union] = ACTIONS(1766),
    [anon_sym_unsafe] = ACTIONS(1766),
    [anon_sym_use] = ACTIONS(1766),
    [anon_sym_while] = ACTIONS(1766),
    [anon_sym_POUND] = ACTIONS(1764),
    [anon_sym_BANG] = ACTIONS(1764),
    [anon_sym_extern] = ACTIONS(1766),
    [anon_sym_LT] = ACTIONS(1764),
    [anon_sym_COLON_COLON] = ACTIONS(1764),
    [anon_sym_AMP] = ACTIONS(1764),
    [anon_sym_DOT_DOT] = ACTIONS(1764),
    [anon_sym_DASH] = ACTIONS(1764),
    [anon_sym_PIPE] = ACTIONS(1764),
    [anon_sym_yield] = ACTIONS(1766),
    [anon_sym_move] = ACTIONS(1766),
    [sym_integer_literal] = ACTIONS(1764),
    [aux_sym_string_literal_token1] = ACTIONS(1764),
    [sym_char_literal] = ACTIONS(1764),
    [anon_sym_true] = ACTIONS(1766),
    [anon_sym_false] = ACTIONS(1766),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1766),
    [sym_super] = ACTIONS(1766),
    [sym_crate] = ACTIONS(1766),
    [sym_metavariable] = ACTIONS(1764),
    [sym_raw_string_literal] = ACTIONS(1764),
    [sym_float_literal] = ACTIONS(1764),
    [sym_block_comment] = ACTIONS(3),
  },
  [422] = {
    [ts_builtin_sym_end] = ACTIONS(1768),
    [sym_identifier] = ACTIONS(1770),
    [anon_sym_SEMI] = ACTIONS(1768),
    [anon_sym_macro_rules_BANG] = ACTIONS(1768),
    [anon_sym_LPAREN] = ACTIONS(1768),
    [anon_sym_LBRACE] = ACTIONS(1768),
    [anon_sym_RBRACE] = ACTIONS(1768),
    [anon_sym_LBRACK] = ACTIONS(1768),
    [anon_sym_STAR] = ACTIONS(1768),
    [anon_sym_u8] = ACTIONS(1770),
    [anon_sym_i8] = ACTIONS(1770),
    [anon_sym_u16] = ACTIONS(1770),
    [anon_sym_i16] = ACTIONS(1770),
    [anon_sym_u32] = ACTIONS(1770),
    [anon_sym_i32] = ACTIONS(1770),
    [anon_sym_u64] = ACTIONS(1770),
    [anon_sym_i64] = ACTIONS(1770),
    [anon_sym_u128] = ACTIONS(1770),
    [anon_sym_i128] = ACTIONS(1770),
    [anon_sym_isize] = ACTIONS(1770),
    [anon_sym_usize] = ACTIONS(1770),
    [anon_sym_f32] = ACTIONS(1770),
    [anon_sym_f64] = ACTIONS(1770),
    [anon_sym_bool] = ACTIONS(1770),
    [anon_sym_str] = ACTIONS(1770),
    [anon_sym_char] = ACTIONS(1770),
    [anon_sym_SQUOTE] = ACTIONS(1770),
    [anon_sym_async] = ACTIONS(1770),
    [anon_sym_break] = ACTIONS(1770),
    [anon_sym_const] = ACTIONS(1770),
    [anon_sym_continue] = ACTIONS(1770),
    [anon_sym_default] = ACTIONS(1770),
    [anon_sym_enum] = ACTIONS(1770),
    [anon_sym_fn] = ACTIONS(1770),
    [anon_sym_for] = ACTIONS(1770),
    [anon_sym_if] = ACTIONS(1770),
    [anon_sym_impl] = ACTIONS(1770),
    [anon_sym_let] = ACTIONS(1770),
    [anon_sym_loop] = ACTIONS(1770),
    [anon_sym_match] = ACTIONS(1770),
    [anon_sym_mod] = ACTIONS(1770),
    [anon_sym_pub] = ACTIONS(1770),
    [anon_sym_return] = ACTIONS(1770),
    [anon_sym_static] = ACTIONS(1770),
    [anon_sym_struct] = ACTIONS(1770),
    [anon_sym_trait] = ACTIONS(1770),
    [anon_sym_type] = ACTIONS(1770),
    [anon_sym_union] = ACTIONS(1770),
    [anon_sym_unsafe] = ACTIONS(1770),
    [anon_sym_use] = ACTIONS(1770),
    [anon_sym_while] = ACTIONS(1770),
    [anon_sym_POUND] = ACTIONS(1768),
    [anon_sym_BANG] = ACTIONS(1768),
    [anon_sym_extern] = ACTIONS(1770),
    [anon_sym_LT] = ACTIONS(1768),
    [anon_sym_COLON_COLON] = ACTIONS(1768),
    [anon_sym_AMP] = ACTIONS(1768),
    [anon_sym_DOT_DOT] = ACTIONS(1768),
    [anon_sym_DASH] = ACTIONS(1768),
    [anon_sym_PIPE] = ACTIONS(1768),
    [anon_sym_yield] = ACTIONS(1770),
    [anon_sym_move] = ACTIONS(1770),
    [sym_integer_literal] = ACTIONS(1768),
    [aux_sym_string_literal_token1] = ACTIONS(1768),
    [sym_char_literal] = ACTIONS(1768),
    [anon_sym_true] = ACTIONS(1770),
    [anon_sym_false] = ACTIONS(1770),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1770),
    [sym_super] = ACTIONS(1770),
    [sym_crate] = ACTIONS(1770),
    [sym_metavariable] = ACTIONS(1768),
    [sym_raw_string_literal] = ACTIONS(1768),
    [sym_float_literal] = ACTIONS(1768),
    [sym_block_comment] = ACTIONS(3),
  },
  [423] = {
    [ts_builtin_sym_end] = ACTIONS(1772),
    [sym_identifier] = ACTIONS(1774),
    [anon_sym_SEMI] = ACTIONS(1772),
    [anon_sym_macro_rules_BANG] = ACTIONS(1772),
    [anon_sym_LPAREN] = ACTIONS(1772),
    [anon_sym_LBRACE] = ACTIONS(1772),
    [anon_sym_RBRACE] = ACTIONS(1772),
    [anon_sym_LBRACK] = ACTIONS(1772),
    [anon_sym_STAR] = ACTIONS(1772),
    [anon_sym_u8] = ACTIONS(1774),
    [anon_sym_i8] = ACTIONS(1774),
    [anon_sym_u16] = ACTIONS(1774),
    [anon_sym_i16] = ACTIONS(1774),
    [anon_sym_u32] = ACTIONS(1774),
    [anon_sym_i32] = ACTIONS(1774),
    [anon_sym_u64] = ACTIONS(1774),
    [anon_sym_i64] = ACTIONS(1774),
    [anon_sym_u128] = ACTIONS(1774),
    [anon_sym_i128] = ACTIONS(1774),
    [anon_sym_isize] = ACTIONS(1774),
    [anon_sym_usize] = ACTIONS(1774),
    [anon_sym_f32] = ACTIONS(1774),
    [anon_sym_f64] = ACTIONS(1774),
    [anon_sym_bool] = ACTIONS(1774),
    [anon_sym_str] = ACTIONS(1774),
    [anon_sym_char] = ACTIONS(1774),
    [anon_sym_SQUOTE] = ACTIONS(1774),
    [anon_sym_async] = ACTIONS(1774),
    [anon_sym_break] = ACTIONS(1774),
    [anon_sym_const] = ACTIONS(1774),
    [anon_sym_continue] = ACTIONS(1774),
    [anon_sym_default] = ACTIONS(1774),
    [anon_sym_enum] = ACTIONS(1774),
    [anon_sym_fn] = ACTIONS(1774),
    [anon_sym_for] = ACTIONS(1774),
    [anon_sym_if] = ACTIONS(1774),
    [anon_sym_impl] = ACTIONS(1774),
    [anon_sym_let] = ACTIONS(1774),
    [anon_sym_loop] = ACTIONS(1774),
    [anon_sym_match] = ACTIONS(1774),
    [anon_sym_mod] = ACTIONS(1774),
    [anon_sym_pub] = ACTIONS(1774),
    [anon_sym_return] = ACTIONS(1774),
    [anon_sym_static] = ACTIONS(1774),
    [anon_sym_struct] = ACTIONS(1774),
    [anon_sym_trait] = ACTIONS(1774),
    [anon_sym_type] = ACTIONS(1774),
    [anon_sym_union] = ACTIONS(1774),
    [anon_sym_unsafe] = ACTIONS(1774),
    [anon_sym_use] = ACTIONS(1774),
    [anon_sym_while] = ACTIONS(1774),
    [anon_sym_POUND] = ACTIONS(1772),
    [anon_sym_BANG] = ACTIONS(1772),
    [anon_sym_extern] = ACTIONS(1774),
    [anon_sym_LT] = ACTIONS(1772),
    [anon_sym_COLON_COLON] = ACTIONS(1772),
    [anon_sym_AMP] = ACTIONS(1772),
    [anon_sym_DOT_DOT] = ACTIONS(1772),
    [anon_sym_DASH] = ACTIONS(1772),
    [anon_sym_PIPE] = ACTIONS(1772),
    [anon_sym_yield] = ACTIONS(1774),
    [anon_sym_move] = ACTIONS(1774),
    [sym_integer_literal] = ACTIONS(1772),
    [aux_sym_string_literal_token1] = ACTIONS(1772),
    [sym_char_literal] = ACTIONS(1772),
    [anon_sym_true] = ACTIONS(1774),
    [anon_sym_false] = ACTIONS(1774),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1774),
    [sym_super] = ACTIONS(1774),
    [sym_crate] = ACTIONS(1774),
    [sym_metavariable] = ACTIONS(1772),
    [sym_raw_string_literal] = ACTIONS(1772),
    [sym_float_literal] = ACTIONS(1772),
    [sym_block_comment] = ACTIONS(3),
  },
  [424] = {
    [ts_builtin_sym_end] = ACTIONS(1776),
    [sym_identifier] = ACTIONS(1778),
    [anon_sym_SEMI] = ACTIONS(1776),
    [anon_sym_macro_rules_BANG] = ACTIONS(1776),
    [anon_sym_LPAREN] = ACTIONS(1776),
    [anon_sym_LBRACE] = ACTIONS(1776),
    [anon_sym_RBRACE] = ACTIONS(1776),
    [anon_sym_LBRACK] = ACTIONS(1776),
    [anon_sym_STAR] = ACTIONS(1776),
    [anon_sym_u8] = ACTIONS(1778),
    [anon_sym_i8] = ACTIONS(1778),
    [anon_sym_u16] = ACTIONS(1778),
    [anon_sym_i16] = ACTIONS(1778),
    [anon_sym_u32] = ACTIONS(1778),
    [anon_sym_i32] = ACTIONS(1778),
    [anon_sym_u64] = ACTIONS(1778),
    [anon_sym_i64] = ACTIONS(1778),
    [anon_sym_u128] = ACTIONS(1778),
    [anon_sym_i128] = ACTIONS(1778),
    [anon_sym_isize] = ACTIONS(1778),
    [anon_sym_usize] = ACTIONS(1778),
    [anon_sym_f32] = ACTIONS(1778),
    [anon_sym_f64] = ACTIONS(1778),
    [anon_sym_bool] = ACTIONS(1778),
    [anon_sym_str] = ACTIONS(1778),
    [anon_sym_char] = ACTIONS(1778),
    [anon_sym_SQUOTE] = ACTIONS(1778),
    [anon_sym_async] = ACTIONS(1778),
    [anon_sym_break] = ACTIONS(1778),
    [anon_sym_const] = ACTIONS(1778),
    [anon_sym_continue] = ACTIONS(1778),
    [anon_sym_default] = ACTIONS(1778),
    [anon_sym_enum] = ACTIONS(1778),
    [anon_sym_fn] = ACTIONS(1778),
    [anon_sym_for] = ACTIONS(1778),
    [anon_sym_if] = ACTIONS(1778),
    [anon_sym_impl] = ACTIONS(1778),
    [anon_sym_let] = ACTIONS(1778),
    [anon_sym_loop] = ACTIONS(1778),
    [anon_sym_match] = ACTIONS(1778),
    [anon_sym_mod] = ACTIONS(1778),
    [anon_sym_pub] = ACTIONS(1778),
    [anon_sym_return] = ACTIONS(1778),
    [anon_sym_static] = ACTIONS(1778),
    [anon_sym_struct] = ACTIONS(1778),
    [anon_sym_trait] = ACTIONS(1778),
    [anon_sym_type] = ACTIONS(1778),
    [anon_sym_union] = ACTIONS(1778),
    [anon_sym_unsafe] = ACTIONS(1778),
    [anon_sym_use] = ACTIONS(1778),
    [anon_sym_while] = ACTIONS(1778),
    [anon_sym_POUND] = ACTIONS(1776),
    [anon_sym_BANG] = ACTIONS(1776),
    [anon_sym_extern] = ACTIONS(1778),
    [anon_sym_LT] = ACTIONS(1776),
    [anon_sym_COLON_COLON] = ACTIONS(1776),
    [anon_sym_AMP] = ACTIONS(1776),
    [anon_sym_DOT_DOT] = ACTIONS(1776),
    [anon_sym_DASH] = ACTIONS(1776),
    [anon_sym_PIPE] = ACTIONS(1776),
    [anon_sym_yield] = ACTIONS(1778),
    [anon_sym_move] = ACTIONS(1778),
    [sym_integer_literal] = ACTIONS(1776),
    [aux_sym_string_literal_token1] = ACTIONS(1776),
    [sym_char_literal] = ACTIONS(1776),
    [anon_sym_true] = ACTIONS(1778),
    [anon_sym_false] = ACTIONS(1778),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1778),
    [sym_super] = ACTIONS(1778),
    [sym_crate] = ACTIONS(1778),
    [sym_metavariable] = ACTIONS(1776),
    [sym_raw_string_literal] = ACTIONS(1776),
    [sym_float_literal] = ACTIONS(1776),
    [sym_block_comment] = ACTIONS(3),
  },
  [425] = {
    [ts_builtin_sym_end] = ACTIONS(1780),
    [sym_identifier] = ACTIONS(1782),
    [anon_sym_SEMI] = ACTIONS(1780),
    [anon_sym_macro_rules_BANG] = ACTIONS(1780),
    [anon_sym_LPAREN] = ACTIONS(1780),
    [anon_sym_LBRACE] = ACTIONS(1780),
    [anon_sym_RBRACE] = ACTIONS(1780),
    [anon_sym_LBRACK] = ACTIONS(1780),
    [anon_sym_STAR] = ACTIONS(1780),
    [anon_sym_u8] = ACTIONS(1782),
    [anon_sym_i8] = ACTIONS(1782),
    [anon_sym_u16] = ACTIONS(1782),
    [anon_sym_i16] = ACTIONS(1782),
    [anon_sym_u32] = ACTIONS(1782),
    [anon_sym_i32] = ACTIONS(1782),
    [anon_sym_u64] = ACTIONS(1782),
    [anon_sym_i64] = ACTIONS(1782),
    [anon_sym_u128] = ACTIONS(1782),
    [anon_sym_i128] = ACTIONS(1782),
    [anon_sym_isize] = ACTIONS(1782),
    [anon_sym_usize] = ACTIONS(1782),
    [anon_sym_f32] = ACTIONS(1782),
    [anon_sym_f64] = ACTIONS(1782),
    [anon_sym_bool] = ACTIONS(1782),
    [anon_sym_str] = ACTIONS(1782),
    [anon_sym_char] = ACTIONS(1782),
    [anon_sym_SQUOTE] = ACTIONS(1782),
    [anon_sym_async] = ACTIONS(1782),
    [anon_sym_break] = ACTIONS(1782),
    [anon_sym_const] = ACTIONS(1782),
    [anon_sym_continue] = ACTIONS(1782),
    [anon_sym_default] = ACTIONS(1782),
    [anon_sym_enum] = ACTIONS(1782),
    [anon_sym_fn] = ACTIONS(1782),
    [anon_sym_for] = ACTIONS(1782),
    [anon_sym_if] = ACTIONS(1782),
    [anon_sym_impl] = ACTIONS(1782),
    [anon_sym_let] = ACTIONS(1782),
    [anon_sym_loop] = ACTIONS(1782),
    [anon_sym_match] = ACTIONS(1782),
    [anon_sym_mod] = ACTIONS(1782),
    [anon_sym_pub] = ACTIONS(1782),
    [anon_sym_return] = ACTIONS(1782),
    [anon_sym_static] = ACTIONS(1782),
    [anon_sym_struct] = ACTIONS(1782),
    [anon_sym_trait] = ACTIONS(1782),
    [anon_sym_type] = ACTIONS(1782),
    [anon_sym_union] = ACTIONS(1782),
    [anon_sym_unsafe] = ACTIONS(1782),
    [anon_sym_use] = ACTIONS(1782),
    [anon_sym_while] = ACTIONS(1782),
    [anon_sym_POUND] = ACTIONS(1780),
    [anon_sym_BANG] = ACTIONS(1780),
    [anon_sym_extern] = ACTIONS(1782),
    [anon_sym_LT] = ACTIONS(1780),
    [anon_sym_COLON_COLON] = ACTIONS(1780),
    [anon_sym_AMP] = ACTIONS(1780),
    [anon_sym_DOT_DOT] = ACTIONS(1780),
    [anon_sym_DASH] = ACTIONS(1780),
    [anon_sym_PIPE] = ACTIONS(1780),
    [anon_sym_yield] = ACTIONS(1782),
    [anon_sym_move] = ACTIONS(1782),
    [sym_integer_literal] = ACTIONS(1780),
    [aux_sym_string_literal_token1] = ACTIONS(1780),
    [sym_char_literal] = ACTIONS(1780),
    [anon_sym_true] = ACTIONS(1782),
    [anon_sym_false] = ACTIONS(1782),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1782),
    [sym_super] = ACTIONS(1782),
    [sym_crate] = ACTIONS(1782),
    [sym_metavariable] = ACTIONS(1780),
    [sym_raw_string_literal] = ACTIONS(1780),
    [sym_float_literal] = ACTIONS(1780),
    [sym_block_comment] = ACTIONS(3),
  },
  [426] = {
    [ts_builtin_sym_end] = ACTIONS(1784),
    [sym_identifier] = ACTIONS(1786),
    [anon_sym_SEMI] = ACTIONS(1784),
    [anon_sym_macro_rules_BANG] = ACTIONS(1784),
    [anon_sym_LPAREN] = ACTIONS(1784),
    [anon_sym_LBRACE] = ACTIONS(1784),
    [anon_sym_RBRACE] = ACTIONS(1784),
    [anon_sym_LBRACK] = ACTIONS(1784),
    [anon_sym_STAR] = ACTIONS(1784),
    [anon_sym_u8] = ACTIONS(1786),
    [anon_sym_i8] = ACTIONS(1786),
    [anon_sym_u16] = ACTIONS(1786),
    [anon_sym_i16] = ACTIONS(1786),
    [anon_sym_u32] = ACTIONS(1786),
    [anon_sym_i32] = ACTIONS(1786),
    [anon_sym_u64] = ACTIONS(1786),
    [anon_sym_i64] = ACTIONS(1786),
    [anon_sym_u128] = ACTIONS(1786),
    [anon_sym_i128] = ACTIONS(1786),
    [anon_sym_isize] = ACTIONS(1786),
    [anon_sym_usize] = ACTIONS(1786),
    [anon_sym_f32] = ACTIONS(1786),
    [anon_sym_f64] = ACTIONS(1786),
    [anon_sym_bool] = ACTIONS(1786),
    [anon_sym_str] = ACTIONS(1786),
    [anon_sym_char] = ACTIONS(1786),
    [anon_sym_SQUOTE] = ACTIONS(1786),
    [anon_sym_async] = ACTIONS(1786),
    [anon_sym_break] = ACTIONS(1786),
    [anon_sym_const] = ACTIONS(1786),
    [anon_sym_continue] = ACTIONS(1786),
    [anon_sym_default] = ACTIONS(1786),
    [anon_sym_enum] = ACTIONS(1786),
    [anon_sym_fn] = ACTIONS(1786),
    [anon_sym_for] = ACTIONS(1786),
    [anon_sym_if] = ACTIONS(1786),
    [anon_sym_impl] = ACTIONS(1786),
    [anon_sym_let] = ACTIONS(1786),
    [anon_sym_loop] = ACTIONS(1786),
    [anon_sym_match] = ACTIONS(1786),
    [anon_sym_mod] = ACTIONS(1786),
    [anon_sym_pub] = ACTIONS(1786),
    [anon_sym_return] = ACTIONS(1786),
    [anon_sym_static] = ACTIONS(1786),
    [anon_sym_struct] = ACTIONS(1786),
    [anon_sym_trait] = ACTIONS(1786),
    [anon_sym_type] = ACTIONS(1786),
    [anon_sym_union] = ACTIONS(1786),
    [anon_sym_unsafe] = ACTIONS(1786),
    [anon_sym_use] = ACTIONS(1786),
    [anon_sym_while] = ACTIONS(1786),
    [anon_sym_POUND] = ACTIONS(1784),
    [anon_sym_BANG] = ACTIONS(1784),
    [anon_sym_extern] = ACTIONS(1786),
    [anon_sym_LT] = ACTIONS(1784),
    [anon_sym_COLON_COLON] = ACTIONS(1784),
    [anon_sym_AMP] = ACTIONS(1784),
    [anon_sym_DOT_DOT] = ACTIONS(1784),
    [anon_sym_DASH] = ACTIONS(1784),
    [anon_sym_PIPE] = ACTIONS(1784),
    [anon_sym_yield] = ACTIONS(1786),
    [anon_sym_move] = ACTIONS(1786),
    [sym_integer_literal] = ACTIONS(1784),
    [aux_sym_string_literal_token1] = ACTIONS(1784),
    [sym_char_literal] = ACTIONS(1784),
    [anon_sym_true] = ACTIONS(1786),
    [anon_sym_false] = ACTIONS(1786),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1786),
    [sym_super] = ACTIONS(1786),
    [sym_crate] = ACTIONS(1786),
    [sym_metavariable] = ACTIONS(1784),
    [sym_raw_string_literal] = ACTIONS(1784),
    [sym_float_literal] = ACTIONS(1784),
    [sym_block_comment] = ACTIONS(3),
  },
  [427] = {
    [ts_builtin_sym_end] = ACTIONS(1788),
    [sym_identifier] = ACTIONS(1790),
    [anon_sym_SEMI] = ACTIONS(1788),
    [anon_sym_macro_rules_BANG] = ACTIONS(1788),
    [anon_sym_LPAREN] = ACTIONS(1788),
    [anon_sym_LBRACE] = ACTIONS(1788),
    [anon_sym_RBRACE] = ACTIONS(1788),
    [anon_sym_LBRACK] = ACTIONS(1788),
    [anon_sym_STAR] = ACTIONS(1788),
    [anon_sym_u8] = ACTIONS(1790),
    [anon_sym_i8] = ACTIONS(1790),
    [anon_sym_u16] = ACTIONS(1790),
    [anon_sym_i16] = ACTIONS(1790),
    [anon_sym_u32] = ACTIONS(1790),
    [anon_sym_i32] = ACTIONS(1790),
    [anon_sym_u64] = ACTIONS(1790),
    [anon_sym_i64] = ACTIONS(1790),
    [anon_sym_u128] = ACTIONS(1790),
    [anon_sym_i128] = ACTIONS(1790),
    [anon_sym_isize] = ACTIONS(1790),
    [anon_sym_usize] = ACTIONS(1790),
    [anon_sym_f32] = ACTIONS(1790),
    [anon_sym_f64] = ACTIONS(1790),
    [anon_sym_bool] = ACTIONS(1790),
    [anon_sym_str] = ACTIONS(1790),
    [anon_sym_char] = ACTIONS(1790),
    [anon_sym_SQUOTE] = ACTIONS(1790),
    [anon_sym_async] = ACTIONS(1790),
    [anon_sym_break] = ACTIONS(1790),
    [anon_sym_const] = ACTIONS(1790),
    [anon_sym_continue] = ACTIONS(1790),
    [anon_sym_default] = ACTIONS(1790),
    [anon_sym_enum] = ACTIONS(1790),
    [anon_sym_fn] = ACTIONS(1790),
    [anon_sym_for] = ACTIONS(1790),
    [anon_sym_if] = ACTIONS(1790),
    [anon_sym_impl] = ACTIONS(1790),
    [anon_sym_let] = ACTIONS(1790),
    [anon_sym_loop] = ACTIONS(1790),
    [anon_sym_match] = ACTIONS(1790),
    [anon_sym_mod] = ACTIONS(1790),
    [anon_sym_pub] = ACTIONS(1790),
    [anon_sym_return] = ACTIONS(1790),
    [anon_sym_static] = ACTIONS(1790),
    [anon_sym_struct] = ACTIONS(1790),
    [anon_sym_trait] = ACTIONS(1790),
    [anon_sym_type] = ACTIONS(1790),
    [anon_sym_union] = ACTIONS(1790),
    [anon_sym_unsafe] = ACTIONS(1790),
    [anon_sym_use] = ACTIONS(1790),
    [anon_sym_while] = ACTIONS(1790),
    [anon_sym_POUND] = ACTIONS(1788),
    [anon_sym_BANG] = ACTIONS(1788),
    [anon_sym_extern] = ACTIONS(1790),
    [anon_sym_LT] = ACTIONS(1788),
    [anon_sym_COLON_COLON] = ACTIONS(1788),
    [anon_sym_AMP] = ACTIONS(1788),
    [anon_sym_DOT_DOT] = ACTIONS(1788),
    [anon_sym_DASH] = ACTIONS(1788),
    [anon_sym_PIPE] = ACTIONS(1788),
    [anon_sym_yield] = ACTIONS(1790),
    [anon_sym_move] = ACTIONS(1790),
    [sym_integer_literal] = ACTIONS(1788),
    [aux_sym_string_literal_token1] = ACTIONS(1788),
    [sym_char_literal] = ACTIONS(1788),
    [anon_sym_true] = ACTIONS(1790),
    [anon_sym_false] = ACTIONS(1790),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1790),
    [sym_super] = ACTIONS(1790),
    [sym_crate] = ACTIONS(1790),
    [sym_metavariable] = ACTIONS(1788),
    [sym_raw_string_literal] = ACTIONS(1788),
    [sym_float_literal] = ACTIONS(1788),
    [sym_block_comment] = ACTIONS(3),
  },
  [428] = {
    [ts_builtin_sym_end] = ACTIONS(1792),
    [sym_identifier] = ACTIONS(1794),
    [anon_sym_SEMI] = ACTIONS(1792),
    [anon_sym_macro_rules_BANG] = ACTIONS(1792),
    [anon_sym_LPAREN] = ACTIONS(1792),
    [anon_sym_LBRACE] = ACTIONS(1792),
    [anon_sym_RBRACE] = ACTIONS(1792),
    [anon_sym_LBRACK] = ACTIONS(1792),
    [anon_sym_STAR] = ACTIONS(1792),
    [anon_sym_u8] = ACTIONS(1794),
    [anon_sym_i8] = ACTIONS(1794),
    [anon_sym_u16] = ACTIONS(1794),
    [anon_sym_i16] = ACTIONS(1794),
    [anon_sym_u32] = ACTIONS(1794),
    [anon_sym_i32] = ACTIONS(1794),
    [anon_sym_u64] = ACTIONS(1794),
    [anon_sym_i64] = ACTIONS(1794),
    [anon_sym_u128] = ACTIONS(1794),
    [anon_sym_i128] = ACTIONS(1794),
    [anon_sym_isize] = ACTIONS(1794),
    [anon_sym_usize] = ACTIONS(1794),
    [anon_sym_f32] = ACTIONS(1794),
    [anon_sym_f64] = ACTIONS(1794),
    [anon_sym_bool] = ACTIONS(1794),
    [anon_sym_str] = ACTIONS(1794),
    [anon_sym_char] = ACTIONS(1794),
    [anon_sym_SQUOTE] = ACTIONS(1794),
    [anon_sym_async] = ACTIONS(1794),
    [anon_sym_break] = ACTIONS(1794),
    [anon_sym_const] = ACTIONS(1794),
    [anon_sym_continue] = ACTIONS(1794),
    [anon_sym_default] = ACTIONS(1794),
    [anon_sym_enum] = ACTIONS(1794),
    [anon_sym_fn] = ACTIONS(1794),
    [anon_sym_for] = ACTIONS(1794),
    [anon_sym_if] = ACTIONS(1794),
    [anon_sym_impl] = ACTIONS(1794),
    [anon_sym_let] = ACTIONS(1794),
    [anon_sym_loop] = ACTIONS(1794),
    [anon_sym_match] = ACTIONS(1794),
    [anon_sym_mod] = ACTIONS(1794),
    [anon_sym_pub] = ACTIONS(1794),
    [anon_sym_return] = ACTIONS(1794),
    [anon_sym_static] = ACTIONS(1794),
    [anon_sym_struct] = ACTIONS(1794),
    [anon_sym_trait] = ACTIONS(1794),
    [anon_sym_type] = ACTIONS(1794),
    [anon_sym_union] = ACTIONS(1794),
    [anon_sym_unsafe] = ACTIONS(1794),
    [anon_sym_use] = ACTIONS(1794),
    [anon_sym_while] = ACTIONS(1794),
    [anon_sym_POUND] = ACTIONS(1792),
    [anon_sym_BANG] = ACTIONS(1792),
    [anon_sym_extern] = ACTIONS(1794),
    [anon_sym_LT] = ACTIONS(1792),
    [anon_sym_COLON_COLON] = ACTIONS(1792),
    [anon_sym_AMP] = ACTIONS(1792),
    [anon_sym_DOT_DOT] = ACTIONS(1792),
    [anon_sym_DASH] = ACTIONS(1792),
    [anon_sym_PIPE] = ACTIONS(1792),
    [anon_sym_yield] = ACTIONS(1794),
    [anon_sym_move] = ACTIONS(1794),
    [sym_integer_literal] = ACTIONS(1792),
    [aux_sym_string_literal_token1] = ACTIONS(1792),
    [sym_char_literal] = ACTIONS(1792),
    [anon_sym_true] = ACTIONS(1794),
    [anon_sym_false] = ACTIONS(1794),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1794),
    [sym_super] = ACTIONS(1794),
    [sym_crate] = ACTIONS(1794),
    [sym_metavariable] = ACTIONS(1792),
    [sym_raw_string_literal] = ACTIONS(1792),
    [sym_float_literal] = ACTIONS(1792),
    [sym_block_comment] = ACTIONS(3),
  },
  [429] = {
    [ts_builtin_sym_end] = ACTIONS(1796),
    [sym_identifier] = ACTIONS(1798),
    [anon_sym_SEMI] = ACTIONS(1796),
    [anon_sym_macro_rules_BANG] = ACTIONS(1796),
    [anon_sym_LPAREN] = ACTIONS(1796),
    [anon_sym_LBRACE] = ACTIONS(1796),
    [anon_sym_RBRACE] = ACTIONS(1796),
    [anon_sym_LBRACK] = ACTIONS(1796),
    [anon_sym_STAR] = ACTIONS(1796),
    [anon_sym_u8] = ACTIONS(1798),
    [anon_sym_i8] = ACTIONS(1798),
    [anon_sym_u16] = ACTIONS(1798),
    [anon_sym_i16] = ACTIONS(1798),
    [anon_sym_u32] = ACTIONS(1798),
    [anon_sym_i32] = ACTIONS(1798),
    [anon_sym_u64] = ACTIONS(1798),
    [anon_sym_i64] = ACTIONS(1798),
    [anon_sym_u128] = ACTIONS(1798),
    [anon_sym_i128] = ACTIONS(1798),
    [anon_sym_isize] = ACTIONS(1798),
    [anon_sym_usize] = ACTIONS(1798),
    [anon_sym_f32] = ACTIONS(1798),
    [anon_sym_f64] = ACTIONS(1798),
    [anon_sym_bool] = ACTIONS(1798),
    [anon_sym_str] = ACTIONS(1798),
    [anon_sym_char] = ACTIONS(1798),
    [anon_sym_SQUOTE] = ACTIONS(1798),
    [anon_sym_async] = ACTIONS(1798),
    [anon_sym_break] = ACTIONS(1798),
    [anon_sym_const] = ACTIONS(1798),
    [anon_sym_continue] = ACTIONS(1798),
    [anon_sym_default] = ACTIONS(1798),
    [anon_sym_enum] = ACTIONS(1798),
    [anon_sym_fn] = ACTIONS(1798),
    [anon_sym_for] = ACTIONS(1798),
    [anon_sym_if] = ACTIONS(1798),
    [anon_sym_impl] = ACTIONS(1798),
    [anon_sym_let] = ACTIONS(1798),
    [anon_sym_loop] = ACTIONS(1798),
    [anon_sym_match] = ACTIONS(1798),
    [anon_sym_mod] = ACTIONS(1798),
    [anon_sym_pub] = ACTIONS(1798),
    [anon_sym_return] = ACTIONS(1798),
    [anon_sym_static] = ACTIONS(1798),
    [anon_sym_struct] = ACTIONS(1798),
    [anon_sym_trait] = ACTIONS(1798),
    [anon_sym_type] = ACTIONS(1798),
    [anon_sym_union] = ACTIONS(1798),
    [anon_sym_unsafe] = ACTIONS(1798),
    [anon_sym_use] = ACTIONS(1798),
    [anon_sym_while] = ACTIONS(1798),
    [anon_sym_POUND] = ACTIONS(1796),
    [anon_sym_BANG] = ACTIONS(1796),
    [anon_sym_extern] = ACTIONS(1798),
    [anon_sym_LT] = ACTIONS(1796),
    [anon_sym_COLON_COLON] = ACTIONS(1796),
    [anon_sym_AMP] = ACTIONS(1796),
    [anon_sym_DOT_DOT] = ACTIONS(1796),
    [anon_sym_DASH] = ACTIONS(1796),
    [anon_sym_PIPE] = ACTIONS(1796),
    [anon_sym_yield] = ACTIONS(1798),
    [anon_sym_move] = ACTIONS(1798),
    [sym_integer_literal] = ACTIONS(1796),
    [aux_sym_string_literal_token1] = ACTIONS(1796),
    [sym_char_literal] = ACTIONS(1796),
    [anon_sym_true] = ACTIONS(1798),
    [anon_sym_false] = ACTIONS(1798),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1798),
    [sym_super] = ACTIONS(1798),
    [sym_crate] = ACTIONS(1798),
    [sym_metavariable] = ACTIONS(1796),
    [sym_raw_string_literal] = ACTIONS(1796),
    [sym_float_literal] = ACTIONS(1796),
    [sym_block_comment] = ACTIONS(3),
  },
  [430] = {
    [ts_builtin_sym_end] = ACTIONS(1800),
    [sym_identifier] = ACTIONS(1802),
    [anon_sym_SEMI] = ACTIONS(1800),
    [anon_sym_macro_rules_BANG] = ACTIONS(1800),
    [anon_sym_LPAREN] = ACTIONS(1800),
    [anon_sym_LBRACE] = ACTIONS(1800),
    [anon_sym_RBRACE] = ACTIONS(1800),
    [anon_sym_LBRACK] = ACTIONS(1800),
    [anon_sym_STAR] = ACTIONS(1800),
    [anon_sym_u8] = ACTIONS(1802),
    [anon_sym_i8] = ACTIONS(1802),
    [anon_sym_u16] = ACTIONS(1802),
    [anon_sym_i16] = ACTIONS(1802),
    [anon_sym_u32] = ACTIONS(1802),
    [anon_sym_i32] = ACTIONS(1802),
    [anon_sym_u64] = ACTIONS(1802),
    [anon_sym_i64] = ACTIONS(1802),
    [anon_sym_u128] = ACTIONS(1802),
    [anon_sym_i128] = ACTIONS(1802),
    [anon_sym_isize] = ACTIONS(1802),
    [anon_sym_usize] = ACTIONS(1802),
    [anon_sym_f32] = ACTIONS(1802),
    [anon_sym_f64] = ACTIONS(1802),
    [anon_sym_bool] = ACTIONS(1802),
    [anon_sym_str] = ACTIONS(1802),
    [anon_sym_char] = ACTIONS(1802),
    [anon_sym_SQUOTE] = ACTIONS(1802),
    [anon_sym_async] = ACTIONS(1802),
    [anon_sym_break] = ACTIONS(1802),
    [anon_sym_const] = ACTIONS(1802),
    [anon_sym_continue] = ACTIONS(1802),
    [anon_sym_default] = ACTIONS(1802),
    [anon_sym_enum] = ACTIONS(1802),
    [anon_sym_fn] = ACTIONS(1802),
    [anon_sym_for] = ACTIONS(1802),
    [anon_sym_if] = ACTIONS(1802),
    [anon_sym_impl] = ACTIONS(1802),
    [anon_sym_let] = ACTIONS(1802),
    [anon_sym_loop] = ACTIONS(1802),
    [anon_sym_match] = ACTIONS(1802),
    [anon_sym_mod] = ACTIONS(1802),
    [anon_sym_pub] = ACTIONS(1802),
    [anon_sym_return] = ACTIONS(1802),
    [anon_sym_static] = ACTIONS(1802),
    [anon_sym_struct] = ACTIONS(1802),
    [anon_sym_trait] = ACTIONS(1802),
    [anon_sym_type] = ACTIONS(1802),
    [anon_sym_union] = ACTIONS(1802),
    [anon_sym_unsafe] = ACTIONS(1802),
    [anon_sym_use] = ACTIONS(1802),
    [anon_sym_while] = ACTIONS(1802),
    [anon_sym_POUND] = ACTIONS(1800),
    [anon_sym_BANG] = ACTIONS(1800),
    [anon_sym_extern] = ACTIONS(1802),
    [anon_sym_LT] = ACTIONS(1800),
    [anon_sym_COLON_COLON] = ACTIONS(1800),
    [anon_sym_AMP] = ACTIONS(1800),
    [anon_sym_DOT_DOT] = ACTIONS(1800),
    [anon_sym_DASH] = ACTIONS(1800),
    [anon_sym_PIPE] = ACTIONS(1800),
    [anon_sym_yield] = ACTIONS(1802),
    [anon_sym_move] = ACTIONS(1802),
    [sym_integer_literal] = ACTIONS(1800),
    [aux_sym_string_literal_token1] = ACTIONS(1800),
    [sym_char_literal] = ACTIONS(1800),
    [anon_sym_true] = ACTIONS(1802),
    [anon_sym_false] = ACTIONS(1802),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1802),
    [sym_super] = ACTIONS(1802),
    [sym_crate] = ACTIONS(1802),
    [sym_metavariable] = ACTIONS(1800),
    [sym_raw_string_literal] = ACTIONS(1800),
    [sym_float_literal] = ACTIONS(1800),
    [sym_block_comment] = ACTIONS(3),
  },
  [431] = {
    [ts_builtin_sym_end] = ACTIONS(1804),
    [sym_identifier] = ACTIONS(1806),
    [anon_sym_SEMI] = ACTIONS(1804),
    [anon_sym_macro_rules_BANG] = ACTIONS(1804),
    [anon_sym_LPAREN] = ACTIONS(1804),
    [anon_sym_LBRACE] = ACTIONS(1804),
    [anon_sym_RBRACE] = ACTIONS(1804),
    [anon_sym_LBRACK] = ACTIONS(1804),
    [anon_sym_STAR] = ACTIONS(1804),
    [anon_sym_u8] = ACTIONS(1806),
    [anon_sym_i8] = ACTIONS(1806),
    [anon_sym_u16] = ACTIONS(1806),
    [anon_sym_i16] = ACTIONS(1806),
    [anon_sym_u32] = ACTIONS(1806),
    [anon_sym_i32] = ACTIONS(1806),
    [anon_sym_u64] = ACTIONS(1806),
    [anon_sym_i64] = ACTIONS(1806),
    [anon_sym_u128] = ACTIONS(1806),
    [anon_sym_i128] = ACTIONS(1806),
    [anon_sym_isize] = ACTIONS(1806),
    [anon_sym_usize] = ACTIONS(1806),
    [anon_sym_f32] = ACTIONS(1806),
    [anon_sym_f64] = ACTIONS(1806),
    [anon_sym_bool] = ACTIONS(1806),
    [anon_sym_str] = ACTIONS(1806),
    [anon_sym_char] = ACTIONS(1806),
    [anon_sym_SQUOTE] = ACTIONS(1806),
    [anon_sym_async] = ACTIONS(1806),
    [anon_sym_break] = ACTIONS(1806),
    [anon_sym_const] = ACTIONS(1806),
    [anon_sym_continue] = ACTIONS(1806),
    [anon_sym_default] = ACTIONS(1806),
    [anon_sym_enum] = ACTIONS(1806),
    [anon_sym_fn] = ACTIONS(1806),
    [anon_sym_for] = ACTIONS(1806),
    [anon_sym_if] = ACTIONS(1806),
    [anon_sym_impl] = ACTIONS(1806),
    [anon_sym_let] = ACTIONS(1806),
    [anon_sym_loop] = ACTIONS(1806),
    [anon_sym_match] = ACTIONS(1806),
    [anon_sym_mod] = ACTIONS(1806),
    [anon_sym_pub] = ACTIONS(1806),
    [anon_sym_return] = ACTIONS(1806),
    [anon_sym_static] = ACTIONS(1806),
    [anon_sym_struct] = ACTIONS(1806),
    [anon_sym_trait] = ACTIONS(1806),
    [anon_sym_type] = ACTIONS(1806),
    [anon_sym_union] = ACTIONS(1806),
    [anon_sym_unsafe] = ACTIONS(1806),
    [anon_sym_use] = ACTIONS(1806),
    [anon_sym_while] = ACTIONS(1806),
    [anon_sym_POUND] = ACTIONS(1804),
    [anon_sym_BANG] = ACTIONS(1804),
    [anon_sym_extern] = ACTIONS(1806),
    [anon_sym_LT] = ACTIONS(1804),
    [anon_sym_COLON_COLON] = ACTIONS(1804),
    [anon_sym_AMP] = ACTIONS(1804),
    [anon_sym_DOT_DOT] = ACTIONS(1804),
    [anon_sym_DASH] = ACTIONS(1804),
    [anon_sym_PIPE] = ACTIONS(1804),
    [anon_sym_yield] = ACTIONS(1806),
    [anon_sym_move] = ACTIONS(1806),
    [sym_integer_literal] = ACTIONS(1804),
    [aux_sym_string_literal_token1] = ACTIONS(1804),
    [sym_char_literal] = ACTIONS(1804),
    [anon_sym_true] = ACTIONS(1806),
    [anon_sym_false] = ACTIONS(1806),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1806),
    [sym_super] = ACTIONS(1806),
    [sym_crate] = ACTIONS(1806),
    [sym_metavariable] = ACTIONS(1804),
    [sym_raw_string_literal] = ACTIONS(1804),
    [sym_float_literal] = ACTIONS(1804),
    [sym_block_comment] = ACTIONS(3),
  },
  [432] = {
    [ts_builtin_sym_end] = ACTIONS(1808),
    [sym_identifier] = ACTIONS(1810),
    [anon_sym_SEMI] = ACTIONS(1808),
    [anon_sym_macro_rules_BANG] = ACTIONS(1808),
    [anon_sym_LPAREN] = ACTIONS(1808),
    [anon_sym_LBRACE] = ACTIONS(1808),
    [anon_sym_RBRACE] = ACTIONS(1808),
    [anon_sym_LBRACK] = ACTIONS(1808),
    [anon_sym_STAR] = ACTIONS(1808),
    [anon_sym_u8] = ACTIONS(1810),
    [anon_sym_i8] = ACTIONS(1810),
    [anon_sym_u16] = ACTIONS(1810),
    [anon_sym_i16] = ACTIONS(1810),
    [anon_sym_u32] = ACTIONS(1810),
    [anon_sym_i32] = ACTIONS(1810),
    [anon_sym_u64] = ACTIONS(1810),
    [anon_sym_i64] = ACTIONS(1810),
    [anon_sym_u128] = ACTIONS(1810),
    [anon_sym_i128] = ACTIONS(1810),
    [anon_sym_isize] = ACTIONS(1810),
    [anon_sym_usize] = ACTIONS(1810),
    [anon_sym_f32] = ACTIONS(1810),
    [anon_sym_f64] = ACTIONS(1810),
    [anon_sym_bool] = ACTIONS(1810),
    [anon_sym_str] = ACTIONS(1810),
    [anon_sym_char] = ACTIONS(1810),
    [anon_sym_SQUOTE] = ACTIONS(1810),
    [anon_sym_async] = ACTIONS(1810),
    [anon_sym_break] = ACTIONS(1810),
    [anon_sym_const] = ACTIONS(1810),
    [anon_sym_continue] = ACTIONS(1810),
    [anon_sym_default] = ACTIONS(1810),
    [anon_sym_enum] = ACTIONS(1810),
    [anon_sym_fn] = ACTIONS(1810),
    [anon_sym_for] = ACTIONS(1810),
    [anon_sym_if] = ACTIONS(1810),
    [anon_sym_impl] = ACTIONS(1810),
    [anon_sym_let] = ACTIONS(1810),
    [anon_sym_loop] = ACTIONS(1810),
    [anon_sym_match] = ACTIONS(1810),
    [anon_sym_mod] = ACTIONS(1810),
    [anon_sym_pub] = ACTIONS(1810),
    [anon_sym_return] = ACTIONS(1810),
    [anon_sym_static] = ACTIONS(1810),
    [anon_sym_struct] = ACTIONS(1810),
    [anon_sym_trait] = ACTIONS(1810),
    [anon_sym_type] = ACTIONS(1810),
    [anon_sym_union] = ACTIONS(1810),
    [anon_sym_unsafe] = ACTIONS(1810),
    [anon_sym_use] = ACTIONS(1810),
    [anon_sym_while] = ACTIONS(1810),
    [anon_sym_POUND] = ACTIONS(1808),
    [anon_sym_BANG] = ACTIONS(1808),
    [anon_sym_extern] = ACTIONS(1810),
    [anon_sym_LT] = ACTIONS(1808),
    [anon_sym_COLON_COLON] = ACTIONS(1808),
    [anon_sym_AMP] = ACTIONS(1808),
    [anon_sym_DOT_DOT] = ACTIONS(1808),
    [anon_sym_DASH] = ACTIONS(1808),
    [anon_sym_PIPE] = ACTIONS(1808),
    [anon_sym_yield] = ACTIONS(1810),
    [anon_sym_move] = ACTIONS(1810),
    [sym_integer_literal] = ACTIONS(1808),
    [aux_sym_string_literal_token1] = ACTIONS(1808),
    [sym_char_literal] = ACTIONS(1808),
    [anon_sym_true] = ACTIONS(1810),
    [anon_sym_false] = ACTIONS(1810),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1810),
    [sym_super] = ACTIONS(1810),
    [sym_crate] = ACTIONS(1810),
    [sym_metavariable] = ACTIONS(1808),
    [sym_raw_string_literal] = ACTIONS(1808),
    [sym_float_literal] = ACTIONS(1808),
    [sym_block_comment] = ACTIONS(3),
  },
  [433] = {
    [ts_builtin_sym_end] = ACTIONS(1812),
    [sym_identifier] = ACTIONS(1814),
    [anon_sym_SEMI] = ACTIONS(1812),
    [anon_sym_macro_rules_BANG] = ACTIONS(1812),
    [anon_sym_LPAREN] = ACTIONS(1812),
    [anon_sym_LBRACE] = ACTIONS(1812),
    [anon_sym_RBRACE] = ACTIONS(1812),
    [anon_sym_LBRACK] = ACTIONS(1812),
    [anon_sym_STAR] = ACTIONS(1812),
    [anon_sym_u8] = ACTIONS(1814),
    [anon_sym_i8] = ACTIONS(1814),
    [anon_sym_u16] = ACTIONS(1814),
    [anon_sym_i16] = ACTIONS(1814),
    [anon_sym_u32] = ACTIONS(1814),
    [anon_sym_i32] = ACTIONS(1814),
    [anon_sym_u64] = ACTIONS(1814),
    [anon_sym_i64] = ACTIONS(1814),
    [anon_sym_u128] = ACTIONS(1814),
    [anon_sym_i128] = ACTIONS(1814),
    [anon_sym_isize] = ACTIONS(1814),
    [anon_sym_usize] = ACTIONS(1814),
    [anon_sym_f32] = ACTIONS(1814),
    [anon_sym_f64] = ACTIONS(1814),
    [anon_sym_bool] = ACTIONS(1814),
    [anon_sym_str] = ACTIONS(1814),
    [anon_sym_char] = ACTIONS(1814),
    [anon_sym_SQUOTE] = ACTIONS(1814),
    [anon_sym_async] = ACTIONS(1814),
    [anon_sym_break] = ACTIONS(1814),
    [anon_sym_const] = ACTIONS(1814),
    [anon_sym_continue] = ACTIONS(1814),
    [anon_sym_default] = ACTIONS(1814),
    [anon_sym_enum] = ACTIONS(1814),
    [anon_sym_fn] = ACTIONS(1814),
    [anon_sym_for] = ACTIONS(1814),
    [anon_sym_if] = ACTIONS(1814),
    [anon_sym_impl] = ACTIONS(1814),
    [anon_sym_let] = ACTIONS(1814),
    [anon_sym_loop] = ACTIONS(1814),
    [anon_sym_match] = ACTIONS(1814),
    [anon_sym_mod] = ACTIONS(1814),
    [anon_sym_pub] = ACTIONS(1814),
    [anon_sym_return] = ACTIONS(1814),
    [anon_sym_static] = ACTIONS(1814),
    [anon_sym_struct] = ACTIONS(1814),
    [anon_sym_trait] = ACTIONS(1814),
    [anon_sym_type] = ACTIONS(1814),
    [anon_sym_union] = ACTIONS(1814),
    [anon_sym_unsafe] = ACTIONS(1814),
    [anon_sym_use] = ACTIONS(1814),
    [anon_sym_while] = ACTIONS(1814),
    [anon_sym_POUND] = ACTIONS(1812),
    [anon_sym_BANG] = ACTIONS(1812),
    [anon_sym_extern] = ACTIONS(1814),
    [anon_sym_LT] = ACTIONS(1812),
    [anon_sym_COLON_COLON] = ACTIONS(1812),
    [anon_sym_AMP] = ACTIONS(1812),
    [anon_sym_DOT_DOT] = ACTIONS(1812),
    [anon_sym_DASH] = ACTIONS(1812),
    [anon_sym_PIPE] = ACTIONS(1812),
    [anon_sym_yield] = ACTIONS(1814),
    [anon_sym_move] = ACTIONS(1814),
    [sym_integer_literal] = ACTIONS(1812),
    [aux_sym_string_literal_token1] = ACTIONS(1812),
    [sym_char_literal] = ACTIONS(1812),
    [anon_sym_true] = ACTIONS(1814),
    [anon_sym_false] = ACTIONS(1814),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1814),
    [sym_super] = ACTIONS(1814),
    [sym_crate] = ACTIONS(1814),
    [sym_metavariable] = ACTIONS(1812),
    [sym_raw_string_literal] = ACTIONS(1812),
    [sym_float_literal] = ACTIONS(1812),
    [sym_block_comment] = ACTIONS(3),
  },
  [434] = {
    [ts_builtin_sym_end] = ACTIONS(1816),
    [sym_identifier] = ACTIONS(1818),
    [anon_sym_SEMI] = ACTIONS(1816),
    [anon_sym_macro_rules_BANG] = ACTIONS(1816),
    [anon_sym_LPAREN] = ACTIONS(1816),
    [anon_sym_LBRACE] = ACTIONS(1816),
    [anon_sym_RBRACE] = ACTIONS(1816),
    [anon_sym_LBRACK] = ACTIONS(1816),
    [anon_sym_STAR] = ACTIONS(1816),
    [anon_sym_u8] = ACTIONS(1818),
    [anon_sym_i8] = ACTIONS(1818),
    [anon_sym_u16] = ACTIONS(1818),
    [anon_sym_i16] = ACTIONS(1818),
    [anon_sym_u32] = ACTIONS(1818),
    [anon_sym_i32] = ACTIONS(1818),
    [anon_sym_u64] = ACTIONS(1818),
    [anon_sym_i64] = ACTIONS(1818),
    [anon_sym_u128] = ACTIONS(1818),
    [anon_sym_i128] = ACTIONS(1818),
    [anon_sym_isize] = ACTIONS(1818),
    [anon_sym_usize] = ACTIONS(1818),
    [anon_sym_f32] = ACTIONS(1818),
    [anon_sym_f64] = ACTIONS(1818),
    [anon_sym_bool] = ACTIONS(1818),
    [anon_sym_str] = ACTIONS(1818),
    [anon_sym_char] = ACTIONS(1818),
    [anon_sym_SQUOTE] = ACTIONS(1818),
    [anon_sym_async] = ACTIONS(1818),
    [anon_sym_break] = ACTIONS(1818),
    [anon_sym_const] = ACTIONS(1818),
    [anon_sym_continue] = ACTIONS(1818),
    [anon_sym_default] = ACTIONS(1818),
    [anon_sym_enum] = ACTIONS(1818),
    [anon_sym_fn] = ACTIONS(1818),
    [anon_sym_for] = ACTIONS(1818),
    [anon_sym_if] = ACTIONS(1818),
    [anon_sym_impl] = ACTIONS(1818),
    [anon_sym_let] = ACTIONS(1818),
    [anon_sym_loop] = ACTIONS(1818),
    [anon_sym_match] = ACTIONS(1818),
    [anon_sym_mod] = ACTIONS(1818),
    [anon_sym_pub] = ACTIONS(1818),
    [anon_sym_return] = ACTIONS(1818),
    [anon_sym_static] = ACTIONS(1818),
    [anon_sym_struct] = ACTIONS(1818),
    [anon_sym_trait] = ACTIONS(1818),
    [anon_sym_type] = ACTIONS(1818),
    [anon_sym_union] = ACTIONS(1818),
    [anon_sym_unsafe] = ACTIONS(1818),
    [anon_sym_use] = ACTIONS(1818),
    [anon_sym_while] = ACTIONS(1818),
    [anon_sym_POUND] = ACTIONS(1816),
    [anon_sym_BANG] = ACTIONS(1816),
    [anon_sym_extern] = ACTIONS(1818),
    [anon_sym_LT] = ACTIONS(1816),
    [anon_sym_COLON_COLON] = ACTIONS(1816),
    [anon_sym_AMP] = ACTIONS(1816),
    [anon_sym_DOT_DOT] = ACTIONS(1816),
    [anon_sym_DASH] = ACTIONS(1816),
    [anon_sym_PIPE] = ACTIONS(1816),
    [anon_sym_yield] = ACTIONS(1818),
    [anon_sym_move] = ACTIONS(1818),
    [sym_integer_literal] = ACTIONS(1816),
    [aux_sym_string_literal_token1] = ACTIONS(1816),
    [sym_char_literal] = ACTIONS(1816),
    [anon_sym_true] = ACTIONS(1818),
    [anon_sym_false] = ACTIONS(1818),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1818),
    [sym_super] = ACTIONS(1818),
    [sym_crate] = ACTIONS(1818),
    [sym_metavariable] = ACTIONS(1816),
    [sym_raw_string_literal] = ACTIONS(1816),
    [sym_float_literal] = ACTIONS(1816),
    [sym_block_comment] = ACTIONS(3),
  },
  [435] = {
    [ts_builtin_sym_end] = ACTIONS(1820),
    [sym_identifier] = ACTIONS(1822),
    [anon_sym_SEMI] = ACTIONS(1820),
    [anon_sym_macro_rules_BANG] = ACTIONS(1820),
    [anon_sym_LPAREN] = ACTIONS(1820),
    [anon_sym_LBRACE] = ACTIONS(1820),
    [anon_sym_RBRACE] = ACTIONS(1820),
    [anon_sym_LBRACK] = ACTIONS(1820),
    [anon_sym_STAR] = ACTIONS(1820),
    [anon_sym_u8] = ACTIONS(1822),
    [anon_sym_i8] = ACTIONS(1822),
    [anon_sym_u16] = ACTIONS(1822),
    [anon_sym_i16] = ACTIONS(1822),
    [anon_sym_u32] = ACTIONS(1822),
    [anon_sym_i32] = ACTIONS(1822),
    [anon_sym_u64] = ACTIONS(1822),
    [anon_sym_i64] = ACTIONS(1822),
    [anon_sym_u128] = ACTIONS(1822),
    [anon_sym_i128] = ACTIONS(1822),
    [anon_sym_isize] = ACTIONS(1822),
    [anon_sym_usize] = ACTIONS(1822),
    [anon_sym_f32] = ACTIONS(1822),
    [anon_sym_f64] = ACTIONS(1822),
    [anon_sym_bool] = ACTIONS(1822),
    [anon_sym_str] = ACTIONS(1822),
    [anon_sym_char] = ACTIONS(1822),
    [anon_sym_SQUOTE] = ACTIONS(1822),
    [anon_sym_async] = ACTIONS(1822),
    [anon_sym_break] = ACTIONS(1822),
    [anon_sym_const] = ACTIONS(1822),
    [anon_sym_continue] = ACTIONS(1822),
    [anon_sym_default] = ACTIONS(1822),
    [anon_sym_enum] = ACTIONS(1822),
    [anon_sym_fn] = ACTIONS(1822),
    [anon_sym_for] = ACTIONS(1822),
    [anon_sym_if] = ACTIONS(1822),
    [anon_sym_impl] = ACTIONS(1822),
    [anon_sym_let] = ACTIONS(1822),
    [anon_sym_loop] = ACTIONS(1822),
    [anon_sym_match] = ACTIONS(1822),
    [anon_sym_mod] = ACTIONS(1822),
    [anon_sym_pub] = ACTIONS(1822),
    [anon_sym_return] = ACTIONS(1822),
    [anon_sym_static] = ACTIONS(1822),
    [anon_sym_struct] = ACTIONS(1822),
    [anon_sym_trait] = ACTIONS(1822),
    [anon_sym_type] = ACTIONS(1822),
    [anon_sym_union] = ACTIONS(1822),
    [anon_sym_unsafe] = ACTIONS(1822),
    [anon_sym_use] = ACTIONS(1822),
    [anon_sym_while] = ACTIONS(1822),
    [anon_sym_POUND] = ACTIONS(1820),
    [anon_sym_BANG] = ACTIONS(1820),
    [anon_sym_extern] = ACTIONS(1822),
    [anon_sym_LT] = ACTIONS(1820),
    [anon_sym_COLON_COLON] = ACTIONS(1820),
    [anon_sym_AMP] = ACTIONS(1820),
    [anon_sym_DOT_DOT] = ACTIONS(1820),
    [anon_sym_DASH] = ACTIONS(1820),
    [anon_sym_PIPE] = ACTIONS(1820),
    [anon_sym_yield] = ACTIONS(1822),
    [anon_sym_move] = ACTIONS(1822),
    [sym_integer_literal] = ACTIONS(1820),
    [aux_sym_string_literal_token1] = ACTIONS(1820),
    [sym_char_literal] = ACTIONS(1820),
    [anon_sym_true] = ACTIONS(1822),
    [anon_sym_false] = ACTIONS(1822),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1822),
    [sym_super] = ACTIONS(1822),
    [sym_crate] = ACTIONS(1822),
    [sym_metavariable] = ACTIONS(1820),
    [sym_raw_string_literal] = ACTIONS(1820),
    [sym_float_literal] = ACTIONS(1820),
    [sym_block_comment] = ACTIONS(3),
  },
  [436] = {
    [ts_builtin_sym_end] = ACTIONS(1824),
    [sym_identifier] = ACTIONS(1826),
    [anon_sym_SEMI] = ACTIONS(1824),
    [anon_sym_macro_rules_BANG] = ACTIONS(1824),
    [anon_sym_LPAREN] = ACTIONS(1824),
    [anon_sym_LBRACE] = ACTIONS(1824),
    [anon_sym_RBRACE] = ACTIONS(1824),
    [anon_sym_LBRACK] = ACTIONS(1824),
    [anon_sym_STAR] = ACTIONS(1824),
    [anon_sym_u8] = ACTIONS(1826),
    [anon_sym_i8] = ACTIONS(1826),
    [anon_sym_u16] = ACTIONS(1826),
    [anon_sym_i16] = ACTIONS(1826),
    [anon_sym_u32] = ACTIONS(1826),
    [anon_sym_i32] = ACTIONS(1826),
    [anon_sym_u64] = ACTIONS(1826),
    [anon_sym_i64] = ACTIONS(1826),
    [anon_sym_u128] = ACTIONS(1826),
    [anon_sym_i128] = ACTIONS(1826),
    [anon_sym_isize] = ACTIONS(1826),
    [anon_sym_usize] = ACTIONS(1826),
    [anon_sym_f32] = ACTIONS(1826),
    [anon_sym_f64] = ACTIONS(1826),
    [anon_sym_bool] = ACTIONS(1826),
    [anon_sym_str] = ACTIONS(1826),
    [anon_sym_char] = ACTIONS(1826),
    [anon_sym_SQUOTE] = ACTIONS(1826),
    [anon_sym_async] = ACTIONS(1826),
    [anon_sym_break] = ACTIONS(1826),
    [anon_sym_const] = ACTIONS(1826),
    [anon_sym_continue] = ACTIONS(1826),
    [anon_sym_default] = ACTIONS(1826),
    [anon_sym_enum] = ACTIONS(1826),
    [anon_sym_fn] = ACTIONS(1826),
    [anon_sym_for] = ACTIONS(1826),
    [anon_sym_if] = ACTIONS(1826),
    [anon_sym_impl] = ACTIONS(1826),
    [anon_sym_let] = ACTIONS(1826),
    [anon_sym_loop] = ACTIONS(1826),
    [anon_sym_match] = ACTIONS(1826),
    [anon_sym_mod] = ACTIONS(1826),
    [anon_sym_pub] = ACTIONS(1826),
    [anon_sym_return] = ACTIONS(1826),
    [anon_sym_static] = ACTIONS(1826),
    [anon_sym_struct] = ACTIONS(1826),
    [anon_sym_trait] = ACTIONS(1826),
    [anon_sym_type] = ACTIONS(1826),
    [anon_sym_union] = ACTIONS(1826),
    [anon_sym_unsafe] = ACTIONS(1826),
    [anon_sym_use] = ACTIONS(1826),
    [anon_sym_while] = ACTIONS(1826),
    [anon_sym_POUND] = ACTIONS(1824),
    [anon_sym_BANG] = ACTIONS(1824),
    [anon_sym_extern] = ACTIONS(1826),
    [anon_sym_LT] = ACTIONS(1824),
    [anon_sym_COLON_COLON] = ACTIONS(1824),
    [anon_sym_AMP] = ACTIONS(1824),
    [anon_sym_DOT_DOT] = ACTIONS(1824),
    [anon_sym_DASH] = ACTIONS(1824),
    [anon_sym_PIPE] = ACTIONS(1824),
    [anon_sym_yield] = ACTIONS(1826),
    [anon_sym_move] = ACTIONS(1826),
    [sym_integer_literal] = ACTIONS(1824),
    [aux_sym_string_literal_token1] = ACTIONS(1824),
    [sym_char_literal] = ACTIONS(1824),
    [anon_sym_true] = ACTIONS(1826),
    [anon_sym_false] = ACTIONS(1826),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1826),
    [sym_super] = ACTIONS(1826),
    [sym_crate] = ACTIONS(1826),
    [sym_metavariable] = ACTIONS(1824),
    [sym_raw_string_literal] = ACTIONS(1824),
    [sym_float_literal] = ACTIONS(1824),
    [sym_block_comment] = ACTIONS(3),
  },
  [437] = {
    [ts_builtin_sym_end] = ACTIONS(1828),
    [sym_identifier] = ACTIONS(1830),
    [anon_sym_SEMI] = ACTIONS(1828),
    [anon_sym_macro_rules_BANG] = ACTIONS(1828),
    [anon_sym_LPAREN] = ACTIONS(1828),
    [anon_sym_LBRACE] = ACTIONS(1828),
    [anon_sym_RBRACE] = ACTIONS(1828),
    [anon_sym_LBRACK] = ACTIONS(1828),
    [anon_sym_STAR] = ACTIONS(1828),
    [anon_sym_u8] = ACTIONS(1830),
    [anon_sym_i8] = ACTIONS(1830),
    [anon_sym_u16] = ACTIONS(1830),
    [anon_sym_i16] = ACTIONS(1830),
    [anon_sym_u32] = ACTIONS(1830),
    [anon_sym_i32] = ACTIONS(1830),
    [anon_sym_u64] = ACTIONS(1830),
    [anon_sym_i64] = ACTIONS(1830),
    [anon_sym_u128] = ACTIONS(1830),
    [anon_sym_i128] = ACTIONS(1830),
    [anon_sym_isize] = ACTIONS(1830),
    [anon_sym_usize] = ACTIONS(1830),
    [anon_sym_f32] = ACTIONS(1830),
    [anon_sym_f64] = ACTIONS(1830),
    [anon_sym_bool] = ACTIONS(1830),
    [anon_sym_str] = ACTIONS(1830),
    [anon_sym_char] = ACTIONS(1830),
    [anon_sym_SQUOTE] = ACTIONS(1830),
    [anon_sym_async] = ACTIONS(1830),
    [anon_sym_break] = ACTIONS(1830),
    [anon_sym_const] = ACTIONS(1830),
    [anon_sym_continue] = ACTIONS(1830),
    [anon_sym_default] = ACTIONS(1830),
    [anon_sym_enum] = ACTIONS(1830),
    [anon_sym_fn] = ACTIONS(1830),
    [anon_sym_for] = ACTIONS(1830),
    [anon_sym_if] = ACTIONS(1830),
    [anon_sym_impl] = ACTIONS(1830),
    [anon_sym_let] = ACTIONS(1830),
    [anon_sym_loop] = ACTIONS(1830),
    [anon_sym_match] = ACTIONS(1830),
    [anon_sym_mod] = ACTIONS(1830),
    [anon_sym_pub] = ACTIONS(1830),
    [anon_sym_return] = ACTIONS(1830),
    [anon_sym_static] = ACTIONS(1830),
    [anon_sym_struct] = ACTIONS(1830),
    [anon_sym_trait] = ACTIONS(1830),
    [anon_sym_type] = ACTIONS(1830),
    [anon_sym_union] = ACTIONS(1830),
    [anon_sym_unsafe] = ACTIONS(1830),
    [anon_sym_use] = ACTIONS(1830),
    [anon_sym_while] = ACTIONS(1830),
    [anon_sym_POUND] = ACTIONS(1828),
    [anon_sym_BANG] = ACTIONS(1828),
    [anon_sym_extern] = ACTIONS(1830),
    [anon_sym_LT] = ACTIONS(1828),
    [anon_sym_COLON_COLON] = ACTIONS(1828),
    [anon_sym_AMP] = ACTIONS(1828),
    [anon_sym_DOT_DOT] = ACTIONS(1828),
    [anon_sym_DASH] = ACTIONS(1828),
    [anon_sym_PIPE] = ACTIONS(1828),
    [anon_sym_yield] = ACTIONS(1830),
    [anon_sym_move] = ACTIONS(1830),
    [sym_integer_literal] = ACTIONS(1828),
    [aux_sym_string_literal_token1] = ACTIONS(1828),
    [sym_char_literal] = ACTIONS(1828),
    [anon_sym_true] = ACTIONS(1830),
    [anon_sym_false] = ACTIONS(1830),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1830),
    [sym_super] = ACTIONS(1830),
    [sym_crate] = ACTIONS(1830),
    [sym_metavariable] = ACTIONS(1828),
    [sym_raw_string_literal] = ACTIONS(1828),
    [sym_float_literal] = ACTIONS(1828),
    [sym_block_comment] = ACTIONS(3),
  },
  [438] = {
    [ts_builtin_sym_end] = ACTIONS(1832),
    [sym_identifier] = ACTIONS(1834),
    [anon_sym_SEMI] = ACTIONS(1832),
    [anon_sym_macro_rules_BANG] = ACTIONS(1832),
    [anon_sym_LPAREN] = ACTIONS(1832),
    [anon_sym_LBRACE] = ACTIONS(1832),
    [anon_sym_RBRACE] = ACTIONS(1832),
    [anon_sym_LBRACK] = ACTIONS(1832),
    [anon_sym_STAR] = ACTIONS(1832),
    [anon_sym_u8] = ACTIONS(1834),
    [anon_sym_i8] = ACTIONS(1834),
    [anon_sym_u16] = ACTIONS(1834),
    [anon_sym_i16] = ACTIONS(1834),
    [anon_sym_u32] = ACTIONS(1834),
    [anon_sym_i32] = ACTIONS(1834),
    [anon_sym_u64] = ACTIONS(1834),
    [anon_sym_i64] = ACTIONS(1834),
    [anon_sym_u128] = ACTIONS(1834),
    [anon_sym_i128] = ACTIONS(1834),
    [anon_sym_isize] = ACTIONS(1834),
    [anon_sym_usize] = ACTIONS(1834),
    [anon_sym_f32] = ACTIONS(1834),
    [anon_sym_f64] = ACTIONS(1834),
    [anon_sym_bool] = ACTIONS(1834),
    [anon_sym_str] = ACTIONS(1834),
    [anon_sym_char] = ACTIONS(1834),
    [anon_sym_SQUOTE] = ACTIONS(1834),
    [anon_sym_async] = ACTIONS(1834),
    [anon_sym_break] = ACTIONS(1834),
    [anon_sym_const] = ACTIONS(1834),
    [anon_sym_continue] = ACTIONS(1834),
    [anon_sym_default] = ACTIONS(1834),
    [anon_sym_enum] = ACTIONS(1834),
    [anon_sym_fn] = ACTIONS(1834),
    [anon_sym_for] = ACTIONS(1834),
    [anon_sym_if] = ACTIONS(1834),
    [anon_sym_impl] = ACTIONS(1834),
    [anon_sym_let] = ACTIONS(1834),
    [anon_sym_loop] = ACTIONS(1834),
    [anon_sym_match] = ACTIONS(1834),
    [anon_sym_mod] = ACTIONS(1834),
    [anon_sym_pub] = ACTIONS(1834),
    [anon_sym_return] = ACTIONS(1834),
    [anon_sym_static] = ACTIONS(1834),
    [anon_sym_struct] = ACTIONS(1834),
    [anon_sym_trait] = ACTIONS(1834),
    [anon_sym_type] = ACTIONS(1834),
    [anon_sym_union] = ACTIONS(1834),
    [anon_sym_unsafe] = ACTIONS(1834),
    [anon_sym_use] = ACTIONS(1834),
    [anon_sym_while] = ACTIONS(1834),
    [anon_sym_POUND] = ACTIONS(1832),
    [anon_sym_BANG] = ACTIONS(1832),
    [anon_sym_extern] = ACTIONS(1834),
    [anon_sym_LT] = ACTIONS(1832),
    [anon_sym_COLON_COLON] = ACTIONS(1832),
    [anon_sym_AMP] = ACTIONS(1832),
    [anon_sym_DOT_DOT] = ACTIONS(1832),
    [anon_sym_DASH] = ACTIONS(1832),
    [anon_sym_PIPE] = ACTIONS(1832),
    [anon_sym_yield] = ACTIONS(1834),
    [anon_sym_move] = ACTIONS(1834),
    [sym_integer_literal] = ACTIONS(1832),
    [aux_sym_string_literal_token1] = ACTIONS(1832),
    [sym_char_literal] = ACTIONS(1832),
    [anon_sym_true] = ACTIONS(1834),
    [anon_sym_false] = ACTIONS(1834),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1834),
    [sym_super] = ACTIONS(1834),
    [sym_crate] = ACTIONS(1834),
    [sym_metavariable] = ACTIONS(1832),
    [sym_raw_string_literal] = ACTIONS(1832),
    [sym_float_literal] = ACTIONS(1832),
    [sym_block_comment] = ACTIONS(3),
  },
  [439] = {
    [ts_builtin_sym_end] = ACTIONS(1836),
    [sym_identifier] = ACTIONS(1838),
    [anon_sym_SEMI] = ACTIONS(1836),
    [anon_sym_macro_rules_BANG] = ACTIONS(1836),
    [anon_sym_LPAREN] = ACTIONS(1836),
    [anon_sym_LBRACE] = ACTIONS(1836),
    [anon_sym_RBRACE] = ACTIONS(1836),
    [anon_sym_LBRACK] = ACTIONS(1836),
    [anon_sym_STAR] = ACTIONS(1836),
    [anon_sym_u8] = ACTIONS(1838),
    [anon_sym_i8] = ACTIONS(1838),
    [anon_sym_u16] = ACTIONS(1838),
    [anon_sym_i16] = ACTIONS(1838),
    [anon_sym_u32] = ACTIONS(1838),
    [anon_sym_i32] = ACTIONS(1838),
    [anon_sym_u64] = ACTIONS(1838),
    [anon_sym_i64] = ACTIONS(1838),
    [anon_sym_u128] = ACTIONS(1838),
    [anon_sym_i128] = ACTIONS(1838),
    [anon_sym_isize] = ACTIONS(1838),
    [anon_sym_usize] = ACTIONS(1838),
    [anon_sym_f32] = ACTIONS(1838),
    [anon_sym_f64] = ACTIONS(1838),
    [anon_sym_bool] = ACTIONS(1838),
    [anon_sym_str] = ACTIONS(1838),
    [anon_sym_char] = ACTIONS(1838),
    [anon_sym_SQUOTE] = ACTIONS(1838),
    [anon_sym_async] = ACTIONS(1838),
    [anon_sym_break] = ACTIONS(1838),
    [anon_sym_const] = ACTIONS(1838),
    [anon_sym_continue] = ACTIONS(1838),
    [anon_sym_default] = ACTIONS(1838),
    [anon_sym_enum] = ACTIONS(1838),
    [anon_sym_fn] = ACTIONS(1838),
    [anon_sym_for] = ACTIONS(1838),
    [anon_sym_if] = ACTIONS(1838),
    [anon_sym_impl] = ACTIONS(1838),
    [anon_sym_let] = ACTIONS(1838),
    [anon_sym_loop] = ACTIONS(1838),
    [anon_sym_match] = ACTIONS(1838),
    [anon_sym_mod] = ACTIONS(1838),
    [anon_sym_pub] = ACTIONS(1838),
    [anon_sym_return] = ACTIONS(1838),
    [anon_sym_static] = ACTIONS(1838),
    [anon_sym_struct] = ACTIONS(1838),
    [anon_sym_trait] = ACTIONS(1838),
    [anon_sym_type] = ACTIONS(1838),
    [anon_sym_union] = ACTIONS(1838),
    [anon_sym_unsafe] = ACTIONS(1838),
    [anon_sym_use] = ACTIONS(1838),
    [anon_sym_while] = ACTIONS(1838),
    [anon_sym_POUND] = ACTIONS(1836),
    [anon_sym_BANG] = ACTIONS(1836),
    [anon_sym_extern] = ACTIONS(1838),
    [anon_sym_LT] = ACTIONS(1836),
    [anon_sym_COLON_COLON] = ACTIONS(1836),
    [anon_sym_AMP] = ACTIONS(1836),
    [anon_sym_DOT_DOT] = ACTIONS(1836),
    [anon_sym_DASH] = ACTIONS(1836),
    [anon_sym_PIPE] = ACTIONS(1836),
    [anon_sym_yield] = ACTIONS(1838),
    [anon_sym_move] = ACTIONS(1838),
    [sym_integer_literal] = ACTIONS(1836),
    [aux_sym_string_literal_token1] = ACTIONS(1836),
    [sym_char_literal] = ACTIONS(1836),
    [anon_sym_true] = ACTIONS(1838),
    [anon_sym_false] = ACTIONS(1838),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1838),
    [sym_super] = ACTIONS(1838),
    [sym_crate] = ACTIONS(1838),
    [sym_metavariable] = ACTIONS(1836),
    [sym_raw_string_literal] = ACTIONS(1836),
    [sym_float_literal] = ACTIONS(1836),
    [sym_block_comment] = ACTIONS(3),
  },
  [440] = {
    [ts_builtin_sym_end] = ACTIONS(1840),
    [sym_identifier] = ACTIONS(1842),
    [anon_sym_SEMI] = ACTIONS(1840),
    [anon_sym_macro_rules_BANG] = ACTIONS(1840),
    [anon_sym_LPAREN] = ACTIONS(1840),
    [anon_sym_LBRACE] = ACTIONS(1840),
    [anon_sym_RBRACE] = ACTIONS(1840),
    [anon_sym_LBRACK] = ACTIONS(1840),
    [anon_sym_STAR] = ACTIONS(1840),
    [anon_sym_u8] = ACTIONS(1842),
    [anon_sym_i8] = ACTIONS(1842),
    [anon_sym_u16] = ACTIONS(1842),
    [anon_sym_i16] = ACTIONS(1842),
    [anon_sym_u32] = ACTIONS(1842),
    [anon_sym_i32] = ACTIONS(1842),
    [anon_sym_u64] = ACTIONS(1842),
    [anon_sym_i64] = ACTIONS(1842),
    [anon_sym_u128] = ACTIONS(1842),
    [anon_sym_i128] = ACTIONS(1842),
    [anon_sym_isize] = ACTIONS(1842),
    [anon_sym_usize] = ACTIONS(1842),
    [anon_sym_f32] = ACTIONS(1842),
    [anon_sym_f64] = ACTIONS(1842),
    [anon_sym_bool] = ACTIONS(1842),
    [anon_sym_str] = ACTIONS(1842),
    [anon_sym_char] = ACTIONS(1842),
    [anon_sym_SQUOTE] = ACTIONS(1842),
    [anon_sym_async] = ACTIONS(1842),
    [anon_sym_break] = ACTIONS(1842),
    [anon_sym_const] = ACTIONS(1842),
    [anon_sym_continue] = ACTIONS(1842),
    [anon_sym_default] = ACTIONS(1842),
    [anon_sym_enum] = ACTIONS(1842),
    [anon_sym_fn] = ACTIONS(1842),
    [anon_sym_for] = ACTIONS(1842),
    [anon_sym_if] = ACTIONS(1842),
    [anon_sym_impl] = ACTIONS(1842),
    [anon_sym_let] = ACTIONS(1842),
    [anon_sym_loop] = ACTIONS(1842),
    [anon_sym_match] = ACTIONS(1842),
    [anon_sym_mod] = ACTIONS(1842),
    [anon_sym_pub] = ACTIONS(1842),
    [anon_sym_return] = ACTIONS(1842),
    [anon_sym_static] = ACTIONS(1842),
    [anon_sym_struct] = ACTIONS(1842),
    [anon_sym_trait] = ACTIONS(1842),
    [anon_sym_type] = ACTIONS(1842),
    [anon_sym_union] = ACTIONS(1842),
    [anon_sym_unsafe] = ACTIONS(1842),
    [anon_sym_use] = ACTIONS(1842),
    [anon_sym_while] = ACTIONS(1842),
    [anon_sym_POUND] = ACTIONS(1840),
    [anon_sym_BANG] = ACTIONS(1840),
    [anon_sym_extern] = ACTIONS(1842),
    [anon_sym_LT] = ACTIONS(1840),
    [anon_sym_COLON_COLON] = ACTIONS(1840),
    [anon_sym_AMP] = ACTIONS(1840),
    [anon_sym_DOT_DOT] = ACTIONS(1840),
    [anon_sym_DASH] = ACTIONS(1840),
    [anon_sym_PIPE] = ACTIONS(1840),
    [anon_sym_yield] = ACTIONS(1842),
    [anon_sym_move] = ACTIONS(1842),
    [sym_integer_literal] = ACTIONS(1840),
    [aux_sym_string_literal_token1] = ACTIONS(1840),
    [sym_char_literal] = ACTIONS(1840),
    [anon_sym_true] = ACTIONS(1842),
    [anon_sym_false] = ACTIONS(1842),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1842),
    [sym_super] = ACTIONS(1842),
    [sym_crate] = ACTIONS(1842),
    [sym_metavariable] = ACTIONS(1840),
    [sym_raw_string_literal] = ACTIONS(1840),
    [sym_float_literal] = ACTIONS(1840),
    [sym_block_comment] = ACTIONS(3),
  },
  [441] = {
    [ts_builtin_sym_end] = ACTIONS(1844),
    [sym_identifier] = ACTIONS(1846),
    [anon_sym_SEMI] = ACTIONS(1844),
    [anon_sym_macro_rules_BANG] = ACTIONS(1844),
    [anon_sym_LPAREN] = ACTIONS(1844),
    [anon_sym_LBRACE] = ACTIONS(1844),
    [anon_sym_RBRACE] = ACTIONS(1844),
    [anon_sym_LBRACK] = ACTIONS(1844),
    [anon_sym_STAR] = ACTIONS(1844),
    [anon_sym_u8] = ACTIONS(1846),
    [anon_sym_i8] = ACTIONS(1846),
    [anon_sym_u16] = ACTIONS(1846),
    [anon_sym_i16] = ACTIONS(1846),
    [anon_sym_u32] = ACTIONS(1846),
    [anon_sym_i32] = ACTIONS(1846),
    [anon_sym_u64] = ACTIONS(1846),
    [anon_sym_i64] = ACTIONS(1846),
    [anon_sym_u128] = ACTIONS(1846),
    [anon_sym_i128] = ACTIONS(1846),
    [anon_sym_isize] = ACTIONS(1846),
    [anon_sym_usize] = ACTIONS(1846),
    [anon_sym_f32] = ACTIONS(1846),
    [anon_sym_f64] = ACTIONS(1846),
    [anon_sym_bool] = ACTIONS(1846),
    [anon_sym_str] = ACTIONS(1846),
    [anon_sym_char] = ACTIONS(1846),
    [anon_sym_SQUOTE] = ACTIONS(1846),
    [anon_sym_async] = ACTIONS(1846),
    [anon_sym_break] = ACTIONS(1846),
    [anon_sym_const] = ACTIONS(1846),
    [anon_sym_continue] = ACTIONS(1846),
    [anon_sym_default] = ACTIONS(1846),
    [anon_sym_enum] = ACTIONS(1846),
    [anon_sym_fn] = ACTIONS(1846),
    [anon_sym_for] = ACTIONS(1846),
    [anon_sym_if] = ACTIONS(1846),
    [anon_sym_impl] = ACTIONS(1846),
    [anon_sym_let] = ACTIONS(1846),
    [anon_sym_loop] = ACTIONS(1846),
    [anon_sym_match] = ACTIONS(1846),
    [anon_sym_mod] = ACTIONS(1846),
    [anon_sym_pub] = ACTIONS(1846),
    [anon_sym_return] = ACTIONS(1846),
    [anon_sym_static] = ACTIONS(1846),
    [anon_sym_struct] = ACTIONS(1846),
    [anon_sym_trait] = ACTIONS(1846),
    [anon_sym_type] = ACTIONS(1846),
    [anon_sym_union] = ACTIONS(1846),
    [anon_sym_unsafe] = ACTIONS(1846),
    [anon_sym_use] = ACTIONS(1846),
    [anon_sym_while] = ACTIONS(1846),
    [anon_sym_POUND] = ACTIONS(1844),
    [anon_sym_BANG] = ACTIONS(1844),
    [anon_sym_extern] = ACTIONS(1846),
    [anon_sym_LT] = ACTIONS(1844),
    [anon_sym_COLON_COLON] = ACTIONS(1844),
    [anon_sym_AMP] = ACTIONS(1844),
    [anon_sym_DOT_DOT] = ACTIONS(1844),
    [anon_sym_DASH] = ACTIONS(1844),
    [anon_sym_PIPE] = ACTIONS(1844),
    [anon_sym_yield] = ACTIONS(1846),
    [anon_sym_move] = ACTIONS(1846),
    [sym_integer_literal] = ACTIONS(1844),
    [aux_sym_string_literal_token1] = ACTIONS(1844),
    [sym_char_literal] = ACTIONS(1844),
    [anon_sym_true] = ACTIONS(1846),
    [anon_sym_false] = ACTIONS(1846),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1846),
    [sym_super] = ACTIONS(1846),
    [sym_crate] = ACTIONS(1846),
    [sym_metavariable] = ACTIONS(1844),
    [sym_raw_string_literal] = ACTIONS(1844),
    [sym_float_literal] = ACTIONS(1844),
    [sym_block_comment] = ACTIONS(3),
  },
  [442] = {
    [ts_builtin_sym_end] = ACTIONS(1848),
    [sym_identifier] = ACTIONS(1850),
    [anon_sym_SEMI] = ACTIONS(1848),
    [anon_sym_macro_rules_BANG] = ACTIONS(1848),
    [anon_sym_LPAREN] = ACTIONS(1848),
    [anon_sym_LBRACE] = ACTIONS(1848),
    [anon_sym_RBRACE] = ACTIONS(1848),
    [anon_sym_LBRACK] = ACTIONS(1848),
    [anon_sym_STAR] = ACTIONS(1848),
    [anon_sym_u8] = ACTIONS(1850),
    [anon_sym_i8] = ACTIONS(1850),
    [anon_sym_u16] = ACTIONS(1850),
    [anon_sym_i16] = ACTIONS(1850),
    [anon_sym_u32] = ACTIONS(1850),
    [anon_sym_i32] = ACTIONS(1850),
    [anon_sym_u64] = ACTIONS(1850),
    [anon_sym_i64] = ACTIONS(1850),
    [anon_sym_u128] = ACTIONS(1850),
    [anon_sym_i128] = ACTIONS(1850),
    [anon_sym_isize] = ACTIONS(1850),
    [anon_sym_usize] = ACTIONS(1850),
    [anon_sym_f32] = ACTIONS(1850),
    [anon_sym_f64] = ACTIONS(1850),
    [anon_sym_bool] = ACTIONS(1850),
    [anon_sym_str] = ACTIONS(1850),
    [anon_sym_char] = ACTIONS(1850),
    [anon_sym_SQUOTE] = ACTIONS(1850),
    [anon_sym_async] = ACTIONS(1850),
    [anon_sym_break] = ACTIONS(1850),
    [anon_sym_const] = ACTIONS(1850),
    [anon_sym_continue] = ACTIONS(1850),
    [anon_sym_default] = ACTIONS(1850),
    [anon_sym_enum] = ACTIONS(1850),
    [anon_sym_fn] = ACTIONS(1850),
    [anon_sym_for] = ACTIONS(1850),
    [anon_sym_if] = ACTIONS(1850),
    [anon_sym_impl] = ACTIONS(1850),
    [anon_sym_let] = ACTIONS(1850),
    [anon_sym_loop] = ACTIONS(1850),
    [anon_sym_match] = ACTIONS(1850),
    [anon_sym_mod] = ACTIONS(1850),
    [anon_sym_pub] = ACTIONS(1850),
    [anon_sym_return] = ACTIONS(1850),
    [anon_sym_static] = ACTIONS(1850),
    [anon_sym_struct] = ACTIONS(1850),
    [anon_sym_trait] = ACTIONS(1850),
    [anon_sym_type] = ACTIONS(1850),
    [anon_sym_union] = ACTIONS(1850),
    [anon_sym_unsafe] = ACTIONS(1850),
    [anon_sym_use] = ACTIONS(1850),
    [anon_sym_while] = ACTIONS(1850),
    [anon_sym_POUND] = ACTIONS(1848),
    [anon_sym_BANG] = ACTIONS(1848),
    [anon_sym_extern] = ACTIONS(1850),
    [anon_sym_LT] = ACTIONS(1848),
    [anon_sym_COLON_COLON] = ACTIONS(1848),
    [anon_sym_AMP] = ACTIONS(1848),
    [anon_sym_DOT_DOT] = ACTIONS(1848),
    [anon_sym_DASH] = ACTIONS(1848),
    [anon_sym_PIPE] = ACTIONS(1848),
    [anon_sym_yield] = ACTIONS(1850),
    [anon_sym_move] = ACTIONS(1850),
    [sym_integer_literal] = ACTIONS(1848),
    [aux_sym_string_literal_token1] = ACTIONS(1848),
    [sym_char_literal] = ACTIONS(1848),
    [anon_sym_true] = ACTIONS(1850),
    [anon_sym_false] = ACTIONS(1850),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1850),
    [sym_super] = ACTIONS(1850),
    [sym_crate] = ACTIONS(1850),
    [sym_metavariable] = ACTIONS(1848),
    [sym_raw_string_literal] = ACTIONS(1848),
    [sym_float_literal] = ACTIONS(1848),
    [sym_block_comment] = ACTIONS(3),
  },
  [443] = {
    [ts_builtin_sym_end] = ACTIONS(1852),
    [sym_identifier] = ACTIONS(1854),
    [anon_sym_SEMI] = ACTIONS(1852),
    [anon_sym_macro_rules_BANG] = ACTIONS(1852),
    [anon_sym_LPAREN] = ACTIONS(1852),
    [anon_sym_LBRACE] = ACTIONS(1852),
    [anon_sym_RBRACE] = ACTIONS(1852),
    [anon_sym_LBRACK] = ACTIONS(1852),
    [anon_sym_STAR] = ACTIONS(1852),
    [anon_sym_u8] = ACTIONS(1854),
    [anon_sym_i8] = ACTIONS(1854),
    [anon_sym_u16] = ACTIONS(1854),
    [anon_sym_i16] = ACTIONS(1854),
    [anon_sym_u32] = ACTIONS(1854),
    [anon_sym_i32] = ACTIONS(1854),
    [anon_sym_u64] = ACTIONS(1854),
    [anon_sym_i64] = ACTIONS(1854),
    [anon_sym_u128] = ACTIONS(1854),
    [anon_sym_i128] = ACTIONS(1854),
    [anon_sym_isize] = ACTIONS(1854),
    [anon_sym_usize] = ACTIONS(1854),
    [anon_sym_f32] = ACTIONS(1854),
    [anon_sym_f64] = ACTIONS(1854),
    [anon_sym_bool] = ACTIONS(1854),
    [anon_sym_str] = ACTIONS(1854),
    [anon_sym_char] = ACTIONS(1854),
    [anon_sym_SQUOTE] = ACTIONS(1854),
    [anon_sym_async] = ACTIONS(1854),
    [anon_sym_break] = ACTIONS(1854),
    [anon_sym_const] = ACTIONS(1854),
    [anon_sym_continue] = ACTIONS(1854),
    [anon_sym_default] = ACTIONS(1854),
    [anon_sym_enum] = ACTIONS(1854),
    [anon_sym_fn] = ACTIONS(1854),
    [anon_sym_for] = ACTIONS(1854),
    [anon_sym_if] = ACTIONS(1854),
    [anon_sym_impl] = ACTIONS(1854),
    [anon_sym_let] = ACTIONS(1854),
    [anon_sym_loop] = ACTIONS(1854),
    [anon_sym_match] = ACTIONS(1854),
    [anon_sym_mod] = ACTIONS(1854),
    [anon_sym_pub] = ACTIONS(1854),
    [anon_sym_return] = ACTIONS(1854),
    [anon_sym_static] = ACTIONS(1854),
    [anon_sym_struct] = ACTIONS(1854),
    [anon_sym_trait] = ACTIONS(1854),
    [anon_sym_type] = ACTIONS(1854),
    [anon_sym_union] = ACTIONS(1854),
    [anon_sym_unsafe] = ACTIONS(1854),
    [anon_sym_use] = ACTIONS(1854),
    [anon_sym_while] = ACTIONS(1854),
    [anon_sym_POUND] = ACTIONS(1852),
    [anon_sym_BANG] = ACTIONS(1852),
    [anon_sym_extern] = ACTIONS(1854),
    [anon_sym_LT] = ACTIONS(1852),
    [anon_sym_COLON_COLON] = ACTIONS(1852),
    [anon_sym_AMP] = ACTIONS(1852),
    [anon_sym_DOT_DOT] = ACTIONS(1852),
    [anon_sym_DASH] = ACTIONS(1852),
    [anon_sym_PIPE] = ACTIONS(1852),
    [anon_sym_yield] = ACTIONS(1854),
    [anon_sym_move] = ACTIONS(1854),
    [sym_integer_literal] = ACTIONS(1852),
    [aux_sym_string_literal_token1] = ACTIONS(1852),
    [sym_char_literal] = ACTIONS(1852),
    [anon_sym_true] = ACTIONS(1854),
    [anon_sym_false] = ACTIONS(1854),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1854),
    [sym_super] = ACTIONS(1854),
    [sym_crate] = ACTIONS(1854),
    [sym_metavariable] = ACTIONS(1852),
    [sym_raw_string_literal] = ACTIONS(1852),
    [sym_float_literal] = ACTIONS(1852),
    [sym_block_comment] = ACTIONS(3),
  },
  [444] = {
    [ts_builtin_sym_end] = ACTIONS(1856),
    [sym_identifier] = ACTIONS(1858),
    [anon_sym_SEMI] = ACTIONS(1856),
    [anon_sym_macro_rules_BANG] = ACTIONS(1856),
    [anon_sym_LPAREN] = ACTIONS(1856),
    [anon_sym_LBRACE] = ACTIONS(1856),
    [anon_sym_RBRACE] = ACTIONS(1856),
    [anon_sym_LBRACK] = ACTIONS(1856),
    [anon_sym_STAR] = ACTIONS(1856),
    [anon_sym_u8] = ACTIONS(1858),
    [anon_sym_i8] = ACTIONS(1858),
    [anon_sym_u16] = ACTIONS(1858),
    [anon_sym_i16] = ACTIONS(1858),
    [anon_sym_u32] = ACTIONS(1858),
    [anon_sym_i32] = ACTIONS(1858),
    [anon_sym_u64] = ACTIONS(1858),
    [anon_sym_i64] = ACTIONS(1858),
    [anon_sym_u128] = ACTIONS(1858),
    [anon_sym_i128] = ACTIONS(1858),
    [anon_sym_isize] = ACTIONS(1858),
    [anon_sym_usize] = ACTIONS(1858),
    [anon_sym_f32] = ACTIONS(1858),
    [anon_sym_f64] = ACTIONS(1858),
    [anon_sym_bool] = ACTIONS(1858),
    [anon_sym_str] = ACTIONS(1858),
    [anon_sym_char] = ACTIONS(1858),
    [anon_sym_SQUOTE] = ACTIONS(1858),
    [anon_sym_async] = ACTIONS(1858),
    [anon_sym_break] = ACTIONS(1858),
    [anon_sym_const] = ACTIONS(1858),
    [anon_sym_continue] = ACTIONS(1858),
    [anon_sym_default] = ACTIONS(1858),
    [anon_sym_enum] = ACTIONS(1858),
    [anon_sym_fn] = ACTIONS(1858),
    [anon_sym_for] = ACTIONS(1858),
    [anon_sym_if] = ACTIONS(1858),
    [anon_sym_impl] = ACTIONS(1858),
    [anon_sym_let] = ACTIONS(1858),
    [anon_sym_loop] = ACTIONS(1858),
    [anon_sym_match] = ACTIONS(1858),
    [anon_sym_mod] = ACTIONS(1858),
    [anon_sym_pub] = ACTIONS(1858),
    [anon_sym_return] = ACTIONS(1858),
    [anon_sym_static] = ACTIONS(1858),
    [anon_sym_struct] = ACTIONS(1858),
    [anon_sym_trait] = ACTIONS(1858),
    [anon_sym_type] = ACTIONS(1858),
    [anon_sym_union] = ACTIONS(1858),
    [anon_sym_unsafe] = ACTIONS(1858),
    [anon_sym_use] = ACTIONS(1858),
    [anon_sym_while] = ACTIONS(1858),
    [anon_sym_POUND] = ACTIONS(1856),
    [anon_sym_BANG] = ACTIONS(1856),
    [anon_sym_extern] = ACTIONS(1858),
    [anon_sym_LT] = ACTIONS(1856),
    [anon_sym_COLON_COLON] = ACTIONS(1856),
    [anon_sym_AMP] = ACTIONS(1856),
    [anon_sym_DOT_DOT] = ACTIONS(1856),
    [anon_sym_DASH] = ACTIONS(1856),
    [anon_sym_PIPE] = ACTIONS(1856),
    [anon_sym_yield] = ACTIONS(1858),
    [anon_sym_move] = ACTIONS(1858),
    [sym_integer_literal] = ACTIONS(1856),
    [aux_sym_string_literal_token1] = ACTIONS(1856),
    [sym_char_literal] = ACTIONS(1856),
    [anon_sym_true] = ACTIONS(1858),
    [anon_sym_false] = ACTIONS(1858),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1858),
    [sym_super] = ACTIONS(1858),
    [sym_crate] = ACTIONS(1858),
    [sym_metavariable] = ACTIONS(1856),
    [sym_raw_string_literal] = ACTIONS(1856),
    [sym_float_literal] = ACTIONS(1856),
    [sym_block_comment] = ACTIONS(3),
  },
  [445] = {
    [ts_builtin_sym_end] = ACTIONS(1860),
    [sym_identifier] = ACTIONS(1862),
    [anon_sym_SEMI] = ACTIONS(1860),
    [anon_sym_macro_rules_BANG] = ACTIONS(1860),
    [anon_sym_LPAREN] = ACTIONS(1860),
    [anon_sym_LBRACE] = ACTIONS(1860),
    [anon_sym_RBRACE] = ACTIONS(1860),
    [anon_sym_LBRACK] = ACTIONS(1860),
    [anon_sym_STAR] = ACTIONS(1860),
    [anon_sym_u8] = ACTIONS(1862),
    [anon_sym_i8] = ACTIONS(1862),
    [anon_sym_u16] = ACTIONS(1862),
    [anon_sym_i16] = ACTIONS(1862),
    [anon_sym_u32] = ACTIONS(1862),
    [anon_sym_i32] = ACTIONS(1862),
    [anon_sym_u64] = ACTIONS(1862),
    [anon_sym_i64] = ACTIONS(1862),
    [anon_sym_u128] = ACTIONS(1862),
    [anon_sym_i128] = ACTIONS(1862),
    [anon_sym_isize] = ACTIONS(1862),
    [anon_sym_usize] = ACTIONS(1862),
    [anon_sym_f32] = ACTIONS(1862),
    [anon_sym_f64] = ACTIONS(1862),
    [anon_sym_bool] = ACTIONS(1862),
    [anon_sym_str] = ACTIONS(1862),
    [anon_sym_char] = ACTIONS(1862),
    [anon_sym_SQUOTE] = ACTIONS(1862),
    [anon_sym_async] = ACTIONS(1862),
    [anon_sym_break] = ACTIONS(1862),
    [anon_sym_const] = ACTIONS(1862),
    [anon_sym_continue] = ACTIONS(1862),
    [anon_sym_default] = ACTIONS(1862),
    [anon_sym_enum] = ACTIONS(1862),
    [anon_sym_fn] = ACTIONS(1862),
    [anon_sym_for] = ACTIONS(1862),
    [anon_sym_if] = ACTIONS(1862),
    [anon_sym_impl] = ACTIONS(1862),
    [anon_sym_let] = ACTIONS(1862),
    [anon_sym_loop] = ACTIONS(1862),
    [anon_sym_match] = ACTIONS(1862),
    [anon_sym_mod] = ACTIONS(1862),
    [anon_sym_pub] = ACTIONS(1862),
    [anon_sym_return] = ACTIONS(1862),
    [anon_sym_static] = ACTIONS(1862),
    [anon_sym_struct] = ACTIONS(1862),
    [anon_sym_trait] = ACTIONS(1862),
    [anon_sym_type] = ACTIONS(1862),
    [anon_sym_union] = ACTIONS(1862),
    [anon_sym_unsafe] = ACTIONS(1862),
    [anon_sym_use] = ACTIONS(1862),
    [anon_sym_while] = ACTIONS(1862),
    [anon_sym_POUND] = ACTIONS(1860),
    [anon_sym_BANG] = ACTIONS(1860),
    [anon_sym_extern] = ACTIONS(1862),
    [anon_sym_LT] = ACTIONS(1860),
    [anon_sym_COLON_COLON] = ACTIONS(1860),
    [anon_sym_AMP] = ACTIONS(1860),
    [anon_sym_DOT_DOT] = ACTIONS(1860),
    [anon_sym_DASH] = ACTIONS(1860),
    [anon_sym_PIPE] = ACTIONS(1860),
    [anon_sym_yield] = ACTIONS(1862),
    [anon_sym_move] = ACTIONS(1862),
    [sym_integer_literal] = ACTIONS(1860),
    [aux_sym_string_literal_token1] = ACTIONS(1860),
    [sym_char_literal] = ACTIONS(1860),
    [anon_sym_true] = ACTIONS(1862),
    [anon_sym_false] = ACTIONS(1862),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1862),
    [sym_super] = ACTIONS(1862),
    [sym_crate] = ACTIONS(1862),
    [sym_metavariable] = ACTIONS(1860),
    [sym_raw_string_literal] = ACTIONS(1860),
    [sym_float_literal] = ACTIONS(1860),
    [sym_block_comment] = ACTIONS(3),
  },
  [446] = {
    [ts_builtin_sym_end] = ACTIONS(1864),
    [sym_identifier] = ACTIONS(1866),
    [anon_sym_SEMI] = ACTIONS(1864),
    [anon_sym_macro_rules_BANG] = ACTIONS(1864),
    [anon_sym_LPAREN] = ACTIONS(1864),
    [anon_sym_LBRACE] = ACTIONS(1864),
    [anon_sym_RBRACE] = ACTIONS(1864),
    [anon_sym_LBRACK] = ACTIONS(1864),
    [anon_sym_STAR] = ACTIONS(1864),
    [anon_sym_u8] = ACTIONS(1866),
    [anon_sym_i8] = ACTIONS(1866),
    [anon_sym_u16] = ACTIONS(1866),
    [anon_sym_i16] = ACTIONS(1866),
    [anon_sym_u32] = ACTIONS(1866),
    [anon_sym_i32] = ACTIONS(1866),
    [anon_sym_u64] = ACTIONS(1866),
    [anon_sym_i64] = ACTIONS(1866),
    [anon_sym_u128] = ACTIONS(1866),
    [anon_sym_i128] = ACTIONS(1866),
    [anon_sym_isize] = ACTIONS(1866),
    [anon_sym_usize] = ACTIONS(1866),
    [anon_sym_f32] = ACTIONS(1866),
    [anon_sym_f64] = ACTIONS(1866),
    [anon_sym_bool] = ACTIONS(1866),
    [anon_sym_str] = ACTIONS(1866),
    [anon_sym_char] = ACTIONS(1866),
    [anon_sym_SQUOTE] = ACTIONS(1866),
    [anon_sym_async] = ACTIONS(1866),
    [anon_sym_break] = ACTIONS(1866),
    [anon_sym_const] = ACTIONS(1866),
    [anon_sym_continue] = ACTIONS(1866),
    [anon_sym_default] = ACTIONS(1866),
    [anon_sym_enum] = ACTIONS(1866),
    [anon_sym_fn] = ACTIONS(1866),
    [anon_sym_for] = ACTIONS(1866),
    [anon_sym_if] = ACTIONS(1866),
    [anon_sym_impl] = ACTIONS(1866),
    [anon_sym_let] = ACTIONS(1866),
    [anon_sym_loop] = ACTIONS(1866),
    [anon_sym_match] = ACTIONS(1866),
    [anon_sym_mod] = ACTIONS(1866),
    [anon_sym_pub] = ACTIONS(1866),
    [anon_sym_return] = ACTIONS(1866),
    [anon_sym_static] = ACTIONS(1866),
    [anon_sym_struct] = ACTIONS(1866),
    [anon_sym_trait] = ACTIONS(1866),
    [anon_sym_type] = ACTIONS(1866),
    [anon_sym_union] = ACTIONS(1866),
    [anon_sym_unsafe] = ACTIONS(1866),
    [anon_sym_use] = ACTIONS(1866),
    [anon_sym_while] = ACTIONS(1866),
    [anon_sym_POUND] = ACTIONS(1864),
    [anon_sym_BANG] = ACTIONS(1864),
    [anon_sym_extern] = ACTIONS(1866),
    [anon_sym_LT] = ACTIONS(1864),
    [anon_sym_COLON_COLON] = ACTIONS(1864),
    [anon_sym_AMP] = ACTIONS(1864),
    [anon_sym_DOT_DOT] = ACTIONS(1864),
    [anon_sym_DASH] = ACTIONS(1864),
    [anon_sym_PIPE] = ACTIONS(1864),
    [anon_sym_yield] = ACTIONS(1866),
    [anon_sym_move] = ACTIONS(1866),
    [sym_integer_literal] = ACTIONS(1864),
    [aux_sym_string_literal_token1] = ACTIONS(1864),
    [sym_char_literal] = ACTIONS(1864),
    [anon_sym_true] = ACTIONS(1866),
    [anon_sym_false] = ACTIONS(1866),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1866),
    [sym_super] = ACTIONS(1866),
    [sym_crate] = ACTIONS(1866),
    [sym_metavariable] = ACTIONS(1864),
    [sym_raw_string_literal] = ACTIONS(1864),
    [sym_float_literal] = ACTIONS(1864),
    [sym_block_comment] = ACTIONS(3),
  },
  [447] = {
    [ts_builtin_sym_end] = ACTIONS(1868),
    [sym_identifier] = ACTIONS(1870),
    [anon_sym_SEMI] = ACTIONS(1868),
    [anon_sym_macro_rules_BANG] = ACTIONS(1868),
    [anon_sym_LPAREN] = ACTIONS(1868),
    [anon_sym_LBRACE] = ACTIONS(1868),
    [anon_sym_RBRACE] = ACTIONS(1868),
    [anon_sym_LBRACK] = ACTIONS(1868),
    [anon_sym_STAR] = ACTIONS(1868),
    [anon_sym_u8] = ACTIONS(1870),
    [anon_sym_i8] = ACTIONS(1870),
    [anon_sym_u16] = ACTIONS(1870),
    [anon_sym_i16] = ACTIONS(1870),
    [anon_sym_u32] = ACTIONS(1870),
    [anon_sym_i32] = ACTIONS(1870),
    [anon_sym_u64] = ACTIONS(1870),
    [anon_sym_i64] = ACTIONS(1870),
    [anon_sym_u128] = ACTIONS(1870),
    [anon_sym_i128] = ACTIONS(1870),
    [anon_sym_isize] = ACTIONS(1870),
    [anon_sym_usize] = ACTIONS(1870),
    [anon_sym_f32] = ACTIONS(1870),
    [anon_sym_f64] = ACTIONS(1870),
    [anon_sym_bool] = ACTIONS(1870),
    [anon_sym_str] = ACTIONS(1870),
    [anon_sym_char] = ACTIONS(1870),
    [anon_sym_SQUOTE] = ACTIONS(1870),
    [anon_sym_async] = ACTIONS(1870),
    [anon_sym_break] = ACTIONS(1870),
    [anon_sym_const] = ACTIONS(1870),
    [anon_sym_continue] = ACTIONS(1870),
    [anon_sym_default] = ACTIONS(1870),
    [anon_sym_enum] = ACTIONS(1870),
    [anon_sym_fn] = ACTIONS(1870),
    [anon_sym_for] = ACTIONS(1870),
    [anon_sym_if] = ACTIONS(1870),
    [anon_sym_impl] = ACTIONS(1870),
    [anon_sym_let] = ACTIONS(1870),
    [anon_sym_loop] = ACTIONS(1870),
    [anon_sym_match] = ACTIONS(1870),
    [anon_sym_mod] = ACTIONS(1870),
    [anon_sym_pub] = ACTIONS(1870),
    [anon_sym_return] = ACTIONS(1870),
    [anon_sym_static] = ACTIONS(1870),
    [anon_sym_struct] = ACTIONS(1870),
    [anon_sym_trait] = ACTIONS(1870),
    [anon_sym_type] = ACTIONS(1870),
    [anon_sym_union] = ACTIONS(1870),
    [anon_sym_unsafe] = ACTIONS(1870),
    [anon_sym_use] = ACTIONS(1870),
    [anon_sym_while] = ACTIONS(1870),
    [anon_sym_POUND] = ACTIONS(1868),
    [anon_sym_BANG] = ACTIONS(1868),
    [anon_sym_extern] = ACTIONS(1870),
    [anon_sym_LT] = ACTIONS(1868),
    [anon_sym_COLON_COLON] = ACTIONS(1868),
    [anon_sym_AMP] = ACTIONS(1868),
    [anon_sym_DOT_DOT] = ACTIONS(1868),
    [anon_sym_DASH] = ACTIONS(1868),
    [anon_sym_PIPE] = ACTIONS(1868),
    [anon_sym_yield] = ACTIONS(1870),
    [anon_sym_move] = ACTIONS(1870),
    [sym_integer_literal] = ACTIONS(1868),
    [aux_sym_string_literal_token1] = ACTIONS(1868),
    [sym_char_literal] = ACTIONS(1868),
    [anon_sym_true] = ACTIONS(1870),
    [anon_sym_false] = ACTIONS(1870),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1870),
    [sym_super] = ACTIONS(1870),
    [sym_crate] = ACTIONS(1870),
    [sym_metavariable] = ACTIONS(1868),
    [sym_raw_string_literal] = ACTIONS(1868),
    [sym_float_literal] = ACTIONS(1868),
    [sym_block_comment] = ACTIONS(3),
  },
  [448] = {
    [ts_builtin_sym_end] = ACTIONS(1872),
    [sym_identifier] = ACTIONS(1874),
    [anon_sym_SEMI] = ACTIONS(1872),
    [anon_sym_macro_rules_BANG] = ACTIONS(1872),
    [anon_sym_LPAREN] = ACTIONS(1872),
    [anon_sym_LBRACE] = ACTIONS(1872),
    [anon_sym_RBRACE] = ACTIONS(1872),
    [anon_sym_LBRACK] = ACTIONS(1872),
    [anon_sym_STAR] = ACTIONS(1872),
    [anon_sym_u8] = ACTIONS(1874),
    [anon_sym_i8] = ACTIONS(1874),
    [anon_sym_u16] = ACTIONS(1874),
    [anon_sym_i16] = ACTIONS(1874),
    [anon_sym_u32] = ACTIONS(1874),
    [anon_sym_i32] = ACTIONS(1874),
    [anon_sym_u64] = ACTIONS(1874),
    [anon_sym_i64] = ACTIONS(1874),
    [anon_sym_u128] = ACTIONS(1874),
    [anon_sym_i128] = ACTIONS(1874),
    [anon_sym_isize] = ACTIONS(1874),
    [anon_sym_usize] = ACTIONS(1874),
    [anon_sym_f32] = ACTIONS(1874),
    [anon_sym_f64] = ACTIONS(1874),
    [anon_sym_bool] = ACTIONS(1874),
    [anon_sym_str] = ACTIONS(1874),
    [anon_sym_char] = ACTIONS(1874),
    [anon_sym_SQUOTE] = ACTIONS(1874),
    [anon_sym_async] = ACTIONS(1874),
    [anon_sym_break] = ACTIONS(1874),
    [anon_sym_const] = ACTIONS(1874),
    [anon_sym_continue] = ACTIONS(1874),
    [anon_sym_default] = ACTIONS(1874),
    [anon_sym_enum] = ACTIONS(1874),
    [anon_sym_fn] = ACTIONS(1874),
    [anon_sym_for] = ACTIONS(1874),
    [anon_sym_if] = ACTIONS(1874),
    [anon_sym_impl] = ACTIONS(1874),
    [anon_sym_let] = ACTIONS(1874),
    [anon_sym_loop] = ACTIONS(1874),
    [anon_sym_match] = ACTIONS(1874),
    [anon_sym_mod] = ACTIONS(1874),
    [anon_sym_pub] = ACTIONS(1874),
    [anon_sym_return] = ACTIONS(1874),
    [anon_sym_static] = ACTIONS(1874),
    [anon_sym_struct] = ACTIONS(1874),
    [anon_sym_trait] = ACTIONS(1874),
    [anon_sym_type] = ACTIONS(1874),
    [anon_sym_union] = ACTIONS(1874),
    [anon_sym_unsafe] = ACTIONS(1874),
    [anon_sym_use] = ACTIONS(1874),
    [anon_sym_while] = ACTIONS(1874),
    [anon_sym_POUND] = ACTIONS(1872),
    [anon_sym_BANG] = ACTIONS(1872),
    [anon_sym_extern] = ACTIONS(1874),
    [anon_sym_LT] = ACTIONS(1872),
    [anon_sym_COLON_COLON] = ACTIONS(1872),
    [anon_sym_AMP] = ACTIONS(1872),
    [anon_sym_DOT_DOT] = ACTIONS(1872),
    [anon_sym_DASH] = ACTIONS(1872),
    [anon_sym_PIPE] = ACTIONS(1872),
    [anon_sym_yield] = ACTIONS(1874),
    [anon_sym_move] = ACTIONS(1874),
    [sym_integer_literal] = ACTIONS(1872),
    [aux_sym_string_literal_token1] = ACTIONS(1872),
    [sym_char_literal] = ACTIONS(1872),
    [anon_sym_true] = ACTIONS(1874),
    [anon_sym_false] = ACTIONS(1874),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1874),
    [sym_super] = ACTIONS(1874),
    [sym_crate] = ACTIONS(1874),
    [sym_metavariable] = ACTIONS(1872),
    [sym_raw_string_literal] = ACTIONS(1872),
    [sym_float_literal] = ACTIONS(1872),
    [sym_block_comment] = ACTIONS(3),
  },
  [449] = {
    [ts_builtin_sym_end] = ACTIONS(1876),
    [sym_identifier] = ACTIONS(1878),
    [anon_sym_SEMI] = ACTIONS(1876),
    [anon_sym_macro_rules_BANG] = ACTIONS(1876),
    [anon_sym_LPAREN] = ACTIONS(1876),
    [anon_sym_LBRACE] = ACTIONS(1876),
    [anon_sym_RBRACE] = ACTIONS(1876),
    [anon_sym_LBRACK] = ACTIONS(1876),
    [anon_sym_STAR] = ACTIONS(1876),
    [anon_sym_u8] = ACTIONS(1878),
    [anon_sym_i8] = ACTIONS(1878),
    [anon_sym_u16] = ACTIONS(1878),
    [anon_sym_i16] = ACTIONS(1878),
    [anon_sym_u32] = ACTIONS(1878),
    [anon_sym_i32] = ACTIONS(1878),
    [anon_sym_u64] = ACTIONS(1878),
    [anon_sym_i64] = ACTIONS(1878),
    [anon_sym_u128] = ACTIONS(1878),
    [anon_sym_i128] = ACTIONS(1878),
    [anon_sym_isize] = ACTIONS(1878),
    [anon_sym_usize] = ACTIONS(1878),
    [anon_sym_f32] = ACTIONS(1878),
    [anon_sym_f64] = ACTIONS(1878),
    [anon_sym_bool] = ACTIONS(1878),
    [anon_sym_str] = ACTIONS(1878),
    [anon_sym_char] = ACTIONS(1878),
    [anon_sym_SQUOTE] = ACTIONS(1878),
    [anon_sym_async] = ACTIONS(1878),
    [anon_sym_break] = ACTIONS(1878),
    [anon_sym_const] = ACTIONS(1878),
    [anon_sym_continue] = ACTIONS(1878),
    [anon_sym_default] = ACTIONS(1878),
    [anon_sym_enum] = ACTIONS(1878),
    [anon_sym_fn] = ACTIONS(1878),
    [anon_sym_for] = ACTIONS(1878),
    [anon_sym_if] = ACTIONS(1878),
    [anon_sym_impl] = ACTIONS(1878),
    [anon_sym_let] = ACTIONS(1878),
    [anon_sym_loop] = ACTIONS(1878),
    [anon_sym_match] = ACTIONS(1878),
    [anon_sym_mod] = ACTIONS(1878),
    [anon_sym_pub] = ACTIONS(1878),
    [anon_sym_return] = ACTIONS(1878),
    [anon_sym_static] = ACTIONS(1878),
    [anon_sym_struct] = ACTIONS(1878),
    [anon_sym_trait] = ACTIONS(1878),
    [anon_sym_type] = ACTIONS(1878),
    [anon_sym_union] = ACTIONS(1878),
    [anon_sym_unsafe] = ACTIONS(1878),
    [anon_sym_use] = ACTIONS(1878),
    [anon_sym_while] = ACTIONS(1878),
    [anon_sym_POUND] = ACTIONS(1876),
    [anon_sym_BANG] = ACTIONS(1876),
    [anon_sym_extern] = ACTIONS(1878),
    [anon_sym_LT] = ACTIONS(1876),
    [anon_sym_COLON_COLON] = ACTIONS(1876),
    [anon_sym_AMP] = ACTIONS(1876),
    [anon_sym_DOT_DOT] = ACTIONS(1876),
    [anon_sym_DASH] = ACTIONS(1876),
    [anon_sym_PIPE] = ACTIONS(1876),
    [anon_sym_yield] = ACTIONS(1878),
    [anon_sym_move] = ACTIONS(1878),
    [sym_integer_literal] = ACTIONS(1876),
    [aux_sym_string_literal_token1] = ACTIONS(1876),
    [sym_char_literal] = ACTIONS(1876),
    [anon_sym_true] = ACTIONS(1878),
    [anon_sym_false] = ACTIONS(1878),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1878),
    [sym_super] = ACTIONS(1878),
    [sym_crate] = ACTIONS(1878),
    [sym_metavariable] = ACTIONS(1876),
    [sym_raw_string_literal] = ACTIONS(1876),
    [sym_float_literal] = ACTIONS(1876),
    [sym_block_comment] = ACTIONS(3),
  },
  [450] = {
    [ts_builtin_sym_end] = ACTIONS(1880),
    [sym_identifier] = ACTIONS(1882),
    [anon_sym_SEMI] = ACTIONS(1880),
    [anon_sym_macro_rules_BANG] = ACTIONS(1880),
    [anon_sym_LPAREN] = ACTIONS(1880),
    [anon_sym_LBRACE] = ACTIONS(1880),
    [anon_sym_RBRACE] = ACTIONS(1880),
    [anon_sym_LBRACK] = ACTIONS(1880),
    [anon_sym_STAR] = ACTIONS(1880),
    [anon_sym_u8] = ACTIONS(1882),
    [anon_sym_i8] = ACTIONS(1882),
    [anon_sym_u16] = ACTIONS(1882),
    [anon_sym_i16] = ACTIONS(1882),
    [anon_sym_u32] = ACTIONS(1882),
    [anon_sym_i32] = ACTIONS(1882),
    [anon_sym_u64] = ACTIONS(1882),
    [anon_sym_i64] = ACTIONS(1882),
    [anon_sym_u128] = ACTIONS(1882),
    [anon_sym_i128] = ACTIONS(1882),
    [anon_sym_isize] = ACTIONS(1882),
    [anon_sym_usize] = ACTIONS(1882),
    [anon_sym_f32] = ACTIONS(1882),
    [anon_sym_f64] = ACTIONS(1882),
    [anon_sym_bool] = ACTIONS(1882),
    [anon_sym_str] = ACTIONS(1882),
    [anon_sym_char] = ACTIONS(1882),
    [anon_sym_SQUOTE] = ACTIONS(1882),
    [anon_sym_async] = ACTIONS(1882),
    [anon_sym_break] = ACTIONS(1882),
    [anon_sym_const] = ACTIONS(1882),
    [anon_sym_continue] = ACTIONS(1882),
    [anon_sym_default] = ACTIONS(1882),
    [anon_sym_enum] = ACTIONS(1882),
    [anon_sym_fn] = ACTIONS(1882),
    [anon_sym_for] = ACTIONS(1882),
    [anon_sym_if] = ACTIONS(1882),
    [anon_sym_impl] = ACTIONS(1882),
    [anon_sym_let] = ACTIONS(1882),
    [anon_sym_loop] = ACTIONS(1882),
    [anon_sym_match] = ACTIONS(1882),
    [anon_sym_mod] = ACTIONS(1882),
    [anon_sym_pub] = ACTIONS(1882),
    [anon_sym_return] = ACTIONS(1882),
    [anon_sym_static] = ACTIONS(1882),
    [anon_sym_struct] = ACTIONS(1882),
    [anon_sym_trait] = ACTIONS(1882),
    [anon_sym_type] = ACTIONS(1882),
    [anon_sym_union] = ACTIONS(1882),
    [anon_sym_unsafe] = ACTIONS(1882),
    [anon_sym_use] = ACTIONS(1882),
    [anon_sym_while] = ACTIONS(1882),
    [anon_sym_POUND] = ACTIONS(1880),
    [anon_sym_BANG] = ACTIONS(1880),
    [anon_sym_extern] = ACTIONS(1882),
    [anon_sym_LT] = ACTIONS(1880),
    [anon_sym_COLON_COLON] = ACTIONS(1880),
    [anon_sym_AMP] = ACTIONS(1880),
    [anon_sym_DOT_DOT] = ACTIONS(1880),
    [anon_sym_DASH] = ACTIONS(1880),
    [anon_sym_PIPE] = ACTIONS(1880),
    [anon_sym_yield] = ACTIONS(1882),
    [anon_sym_move] = ACTIONS(1882),
    [sym_integer_literal] = ACTIONS(1880),
    [aux_sym_string_literal_token1] = ACTIONS(1880),
    [sym_char_literal] = ACTIONS(1880),
    [anon_sym_true] = ACTIONS(1882),
    [anon_sym_false] = ACTIONS(1882),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1882),
    [sym_super] = ACTIONS(1882),
    [sym_crate] = ACTIONS(1882),
    [sym_metavariable] = ACTIONS(1880),
    [sym_raw_string_literal] = ACTIONS(1880),
    [sym_float_literal] = ACTIONS(1880),
    [sym_block_comment] = ACTIONS(3),
  },
  [451] = {
    [ts_builtin_sym_end] = ACTIONS(1884),
    [sym_identifier] = ACTIONS(1886),
    [anon_sym_SEMI] = ACTIONS(1884),
    [anon_sym_macro_rules_BANG] = ACTIONS(1884),
    [anon_sym_LPAREN] = ACTIONS(1884),
    [anon_sym_LBRACE] = ACTIONS(1884),
    [anon_sym_RBRACE] = ACTIONS(1884),
    [anon_sym_LBRACK] = ACTIONS(1884),
    [anon_sym_STAR] = ACTIONS(1884),
    [anon_sym_u8] = ACTIONS(1886),
    [anon_sym_i8] = ACTIONS(1886),
    [anon_sym_u16] = ACTIONS(1886),
    [anon_sym_i16] = ACTIONS(1886),
    [anon_sym_u32] = ACTIONS(1886),
    [anon_sym_i32] = ACTIONS(1886),
    [anon_sym_u64] = ACTIONS(1886),
    [anon_sym_i64] = ACTIONS(1886),
    [anon_sym_u128] = ACTIONS(1886),
    [anon_sym_i128] = ACTIONS(1886),
    [anon_sym_isize] = ACTIONS(1886),
    [anon_sym_usize] = ACTIONS(1886),
    [anon_sym_f32] = ACTIONS(1886),
    [anon_sym_f64] = ACTIONS(1886),
    [anon_sym_bool] = ACTIONS(1886),
    [anon_sym_str] = ACTIONS(1886),
    [anon_sym_char] = ACTIONS(1886),
    [anon_sym_SQUOTE] = ACTIONS(1886),
    [anon_sym_async] = ACTIONS(1886),
    [anon_sym_break] = ACTIONS(1886),
    [anon_sym_const] = ACTIONS(1886),
    [anon_sym_continue] = ACTIONS(1886),
    [anon_sym_default] = ACTIONS(1886),
    [anon_sym_enum] = ACTIONS(1886),
    [anon_sym_fn] = ACTIONS(1886),
    [anon_sym_for] = ACTIONS(1886),
    [anon_sym_if] = ACTIONS(1886),
    [anon_sym_impl] = ACTIONS(1886),
    [anon_sym_let] = ACTIONS(1886),
    [anon_sym_loop] = ACTIONS(1886),
    [anon_sym_match] = ACTIONS(1886),
    [anon_sym_mod] = ACTIONS(1886),
    [anon_sym_pub] = ACTIONS(1886),
    [anon_sym_return] = ACTIONS(1886),
    [anon_sym_static] = ACTIONS(1886),
    [anon_sym_struct] = ACTIONS(1886),
    [anon_sym_trait] = ACTIONS(1886),
    [anon_sym_type] = ACTIONS(1886),
    [anon_sym_union] = ACTIONS(1886),
    [anon_sym_unsafe] = ACTIONS(1886),
    [anon_sym_use] = ACTIONS(1886),
    [anon_sym_while] = ACTIONS(1886),
    [anon_sym_POUND] = ACTIONS(1884),
    [anon_sym_BANG] = ACTIONS(1884),
    [anon_sym_extern] = ACTIONS(1886),
    [anon_sym_LT] = ACTIONS(1884),
    [anon_sym_COLON_COLON] = ACTIONS(1884),
    [anon_sym_AMP] = ACTIONS(1884),
    [anon_sym_DOT_DOT] = ACTIONS(1884),
    [anon_sym_DASH] = ACTIONS(1884),
    [anon_sym_PIPE] = ACTIONS(1884),
    [anon_sym_yield] = ACTIONS(1886),
    [anon_sym_move] = ACTIONS(1886),
    [sym_integer_literal] = ACTIONS(1884),
    [aux_sym_string_literal_token1] = ACTIONS(1884),
    [sym_char_literal] = ACTIONS(1884),
    [anon_sym_true] = ACTIONS(1886),
    [anon_sym_false] = ACTIONS(1886),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1886),
    [sym_super] = ACTIONS(1886),
    [sym_crate] = ACTIONS(1886),
    [sym_metavariable] = ACTIONS(1884),
    [sym_raw_string_literal] = ACTIONS(1884),
    [sym_float_literal] = ACTIONS(1884),
    [sym_block_comment] = ACTIONS(3),
  },
  [452] = {
    [ts_builtin_sym_end] = ACTIONS(1888),
    [sym_identifier] = ACTIONS(1890),
    [anon_sym_SEMI] = ACTIONS(1888),
    [anon_sym_macro_rules_BANG] = ACTIONS(1888),
    [anon_sym_LPAREN] = ACTIONS(1888),
    [anon_sym_LBRACE] = ACTIONS(1888),
    [anon_sym_RBRACE] = ACTIONS(1888),
    [anon_sym_LBRACK] = ACTIONS(1888),
    [anon_sym_STAR] = ACTIONS(1888),
    [anon_sym_u8] = ACTIONS(1890),
    [anon_sym_i8] = ACTIONS(1890),
    [anon_sym_u16] = ACTIONS(1890),
    [anon_sym_i16] = ACTIONS(1890),
    [anon_sym_u32] = ACTIONS(1890),
    [anon_sym_i32] = ACTIONS(1890),
    [anon_sym_u64] = ACTIONS(1890),
    [anon_sym_i64] = ACTIONS(1890),
    [anon_sym_u128] = ACTIONS(1890),
    [anon_sym_i128] = ACTIONS(1890),
    [anon_sym_isize] = ACTIONS(1890),
    [anon_sym_usize] = ACTIONS(1890),
    [anon_sym_f32] = ACTIONS(1890),
    [anon_sym_f64] = ACTIONS(1890),
    [anon_sym_bool] = ACTIONS(1890),
    [anon_sym_str] = ACTIONS(1890),
    [anon_sym_char] = ACTIONS(1890),
    [anon_sym_SQUOTE] = ACTIONS(1890),
    [anon_sym_async] = ACTIONS(1890),
    [anon_sym_break] = ACTIONS(1890),
    [anon_sym_const] = ACTIONS(1890),
    [anon_sym_continue] = ACTIONS(1890),
    [anon_sym_default] = ACTIONS(1890),
    [anon_sym_enum] = ACTIONS(1890),
    [anon_sym_fn] = ACTIONS(1890),
    [anon_sym_for] = ACTIONS(1890),
    [anon_sym_if] = ACTIONS(1890),
    [anon_sym_impl] = ACTIONS(1890),
    [anon_sym_let] = ACTIONS(1890),
    [anon_sym_loop] = ACTIONS(1890),
    [anon_sym_match] = ACTIONS(1890),
    [anon_sym_mod] = ACTIONS(1890),
    [anon_sym_pub] = ACTIONS(1890),
    [anon_sym_return] = ACTIONS(1890),
    [anon_sym_static] = ACTIONS(1890),
    [anon_sym_struct] = ACTIONS(1890),
    [anon_sym_trait] = ACTIONS(1890),
    [anon_sym_type] = ACTIONS(1890),
    [anon_sym_union] = ACTIONS(1890),
    [anon_sym_unsafe] = ACTIONS(1890),
    [anon_sym_use] = ACTIONS(1890),
    [anon_sym_while] = ACTIONS(1890),
    [anon_sym_POUND] = ACTIONS(1888),
    [anon_sym_BANG] = ACTIONS(1888),
    [anon_sym_extern] = ACTIONS(1890),
    [anon_sym_LT] = ACTIONS(1888),
    [anon_sym_COLON_COLON] = ACTIONS(1888),
    [anon_sym_AMP] = ACTIONS(1888),
    [anon_sym_DOT_DOT] = ACTIONS(1888),
    [anon_sym_DASH] = ACTIONS(1888),
    [anon_sym_PIPE] = ACTIONS(1888),
    [anon_sym_yield] = ACTIONS(1890),
    [anon_sym_move] = ACTIONS(1890),
    [sym_integer_literal] = ACTIONS(1888),
    [aux_sym_string_literal_token1] = ACTIONS(1888),
    [sym_char_literal] = ACTIONS(1888),
    [anon_sym_true] = ACTIONS(1890),
    [anon_sym_false] = ACTIONS(1890),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1890),
    [sym_super] = ACTIONS(1890),
    [sym_crate] = ACTIONS(1890),
    [sym_metavariable] = ACTIONS(1888),
    [sym_raw_string_literal] = ACTIONS(1888),
    [sym_float_literal] = ACTIONS(1888),
    [sym_block_comment] = ACTIONS(3),
  },
  [453] = {
    [ts_builtin_sym_end] = ACTIONS(1892),
    [sym_identifier] = ACTIONS(1894),
    [anon_sym_SEMI] = ACTIONS(1892),
    [anon_sym_macro_rules_BANG] = ACTIONS(1892),
    [anon_sym_LPAREN] = ACTIONS(1892),
    [anon_sym_LBRACE] = ACTIONS(1892),
    [anon_sym_RBRACE] = ACTIONS(1892),
    [anon_sym_LBRACK] = ACTIONS(1892),
    [anon_sym_STAR] = ACTIONS(1892),
    [anon_sym_u8] = ACTIONS(1894),
    [anon_sym_i8] = ACTIONS(1894),
    [anon_sym_u16] = ACTIONS(1894),
    [anon_sym_i16] = ACTIONS(1894),
    [anon_sym_u32] = ACTIONS(1894),
    [anon_sym_i32] = ACTIONS(1894),
    [anon_sym_u64] = ACTIONS(1894),
    [anon_sym_i64] = ACTIONS(1894),
    [anon_sym_u128] = ACTIONS(1894),
    [anon_sym_i128] = ACTIONS(1894),
    [anon_sym_isize] = ACTIONS(1894),
    [anon_sym_usize] = ACTIONS(1894),
    [anon_sym_f32] = ACTIONS(1894),
    [anon_sym_f64] = ACTIONS(1894),
    [anon_sym_bool] = ACTIONS(1894),
    [anon_sym_str] = ACTIONS(1894),
    [anon_sym_char] = ACTIONS(1894),
    [anon_sym_SQUOTE] = ACTIONS(1894),
    [anon_sym_async] = ACTIONS(1894),
    [anon_sym_break] = ACTIONS(1894),
    [anon_sym_const] = ACTIONS(1894),
    [anon_sym_continue] = ACTIONS(1894),
    [anon_sym_default] = ACTIONS(1894),
    [anon_sym_enum] = ACTIONS(1894),
    [anon_sym_fn] = ACTIONS(1894),
    [anon_sym_for] = ACTIONS(1894),
    [anon_sym_if] = ACTIONS(1894),
    [anon_sym_impl] = ACTIONS(1894),
    [anon_sym_let] = ACTIONS(1894),
    [anon_sym_loop] = ACTIONS(1894),
    [anon_sym_match] = ACTIONS(1894),
    [anon_sym_mod] = ACTIONS(1894),
    [anon_sym_pub] = ACTIONS(1894),
    [anon_sym_return] = ACTIONS(1894),
    [anon_sym_static] = ACTIONS(1894),
    [anon_sym_struct] = ACTIONS(1894),
    [anon_sym_trait] = ACTIONS(1894),
    [anon_sym_type] = ACTIONS(1894),
    [anon_sym_union] = ACTIONS(1894),
    [anon_sym_unsafe] = ACTIONS(1894),
    [anon_sym_use] = ACTIONS(1894),
    [anon_sym_while] = ACTIONS(1894),
    [anon_sym_POUND] = ACTIONS(1892),
    [anon_sym_BANG] = ACTIONS(1892),
    [anon_sym_extern] = ACTIONS(1894),
    [anon_sym_LT] = ACTIONS(1892),
    [anon_sym_COLON_COLON] = ACTIONS(1892),
    [anon_sym_AMP] = ACTIONS(1892),
    [anon_sym_DOT_DOT] = ACTIONS(1892),
    [anon_sym_DASH] = ACTIONS(1892),
    [anon_sym_PIPE] = ACTIONS(1892),
    [anon_sym_yield] = ACTIONS(1894),
    [anon_sym_move] = ACTIONS(1894),
    [sym_integer_literal] = ACTIONS(1892),
    [aux_sym_string_literal_token1] = ACTIONS(1892),
    [sym_char_literal] = ACTIONS(1892),
    [anon_sym_true] = ACTIONS(1894),
    [anon_sym_false] = ACTIONS(1894),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1894),
    [sym_super] = ACTIONS(1894),
    [sym_crate] = ACTIONS(1894),
    [sym_metavariable] = ACTIONS(1892),
    [sym_raw_string_literal] = ACTIONS(1892),
    [sym_float_literal] = ACTIONS(1892),
    [sym_block_comment] = ACTIONS(3),
  },
  [454] = {
    [ts_builtin_sym_end] = ACTIONS(1896),
    [sym_identifier] = ACTIONS(1898),
    [anon_sym_SEMI] = ACTIONS(1896),
    [anon_sym_macro_rules_BANG] = ACTIONS(1896),
    [anon_sym_LPAREN] = ACTIONS(1896),
    [anon_sym_LBRACE] = ACTIONS(1896),
    [anon_sym_RBRACE] = ACTIONS(1896),
    [anon_sym_LBRACK] = ACTIONS(1896),
    [anon_sym_STAR] = ACTIONS(1896),
    [anon_sym_u8] = ACTIONS(1898),
    [anon_sym_i8] = ACTIONS(1898),
    [anon_sym_u16] = ACTIONS(1898),
    [anon_sym_i16] = ACTIONS(1898),
    [anon_sym_u32] = ACTIONS(1898),
    [anon_sym_i32] = ACTIONS(1898),
    [anon_sym_u64] = ACTIONS(1898),
    [anon_sym_i64] = ACTIONS(1898),
    [anon_sym_u128] = ACTIONS(1898),
    [anon_sym_i128] = ACTIONS(1898),
    [anon_sym_isize] = ACTIONS(1898),
    [anon_sym_usize] = ACTIONS(1898),
    [anon_sym_f32] = ACTIONS(1898),
    [anon_sym_f64] = ACTIONS(1898),
    [anon_sym_bool] = ACTIONS(1898),
    [anon_sym_str] = ACTIONS(1898),
    [anon_sym_char] = ACTIONS(1898),
    [anon_sym_SQUOTE] = ACTIONS(1898),
    [anon_sym_async] = ACTIONS(1898),
    [anon_sym_break] = ACTIONS(1898),
    [anon_sym_const] = ACTIONS(1898),
    [anon_sym_continue] = ACTIONS(1898),
    [anon_sym_default] = ACTIONS(1898),
    [anon_sym_enum] = ACTIONS(1898),
    [anon_sym_fn] = ACTIONS(1898),
    [anon_sym_for] = ACTIONS(1898),
    [anon_sym_if] = ACTIONS(1898),
    [anon_sym_impl] = ACTIONS(1898),
    [anon_sym_let] = ACTIONS(1898),
    [anon_sym_loop] = ACTIONS(1898),
    [anon_sym_match] = ACTIONS(1898),
    [anon_sym_mod] = ACTIONS(1898),
    [anon_sym_pub] = ACTIONS(1898),
    [anon_sym_return] = ACTIONS(1898),
    [anon_sym_static] = ACTIONS(1898),
    [anon_sym_struct] = ACTIONS(1898),
    [anon_sym_trait] = ACTIONS(1898),
    [anon_sym_type] = ACTIONS(1898),
    [anon_sym_union] = ACTIONS(1898),
    [anon_sym_unsafe] = ACTIONS(1898),
    [anon_sym_use] = ACTIONS(1898),
    [anon_sym_while] = ACTIONS(1898),
    [anon_sym_POUND] = ACTIONS(1896),
    [anon_sym_BANG] = ACTIONS(1896),
    [anon_sym_extern] = ACTIONS(1898),
    [anon_sym_LT] = ACTIONS(1896),
    [anon_sym_COLON_COLON] = ACTIONS(1896),
    [anon_sym_AMP] = ACTIONS(1896),
    [anon_sym_DOT_DOT] = ACTIONS(1896),
    [anon_sym_DASH] = ACTIONS(1896),
    [anon_sym_PIPE] = ACTIONS(1896),
    [anon_sym_yield] = ACTIONS(1898),
    [anon_sym_move] = ACTIONS(1898),
    [sym_integer_literal] = ACTIONS(1896),
    [aux_sym_string_literal_token1] = ACTIONS(1896),
    [sym_char_literal] = ACTIONS(1896),
    [anon_sym_true] = ACTIONS(1898),
    [anon_sym_false] = ACTIONS(1898),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1898),
    [sym_super] = ACTIONS(1898),
    [sym_crate] = ACTIONS(1898),
    [sym_metavariable] = ACTIONS(1896),
    [sym_raw_string_literal] = ACTIONS(1896),
    [sym_float_literal] = ACTIONS(1896),
    [sym_block_comment] = ACTIONS(3),
  },
  [455] = {
    [ts_builtin_sym_end] = ACTIONS(1900),
    [sym_identifier] = ACTIONS(1902),
    [anon_sym_SEMI] = ACTIONS(1900),
    [anon_sym_macro_rules_BANG] = ACTIONS(1900),
    [anon_sym_LPAREN] = ACTIONS(1900),
    [anon_sym_LBRACE] = ACTIONS(1900),
    [anon_sym_RBRACE] = ACTIONS(1900),
    [anon_sym_LBRACK] = ACTIONS(1900),
    [anon_sym_STAR] = ACTIONS(1900),
    [anon_sym_u8] = ACTIONS(1902),
    [anon_sym_i8] = ACTIONS(1902),
    [anon_sym_u16] = ACTIONS(1902),
    [anon_sym_i16] = ACTIONS(1902),
    [anon_sym_u32] = ACTIONS(1902),
    [anon_sym_i32] = ACTIONS(1902),
    [anon_sym_u64] = ACTIONS(1902),
    [anon_sym_i64] = ACTIONS(1902),
    [anon_sym_u128] = ACTIONS(1902),
    [anon_sym_i128] = ACTIONS(1902),
    [anon_sym_isize] = ACTIONS(1902),
    [anon_sym_usize] = ACTIONS(1902),
    [anon_sym_f32] = ACTIONS(1902),
    [anon_sym_f64] = ACTIONS(1902),
    [anon_sym_bool] = ACTIONS(1902),
    [anon_sym_str] = ACTIONS(1902),
    [anon_sym_char] = ACTIONS(1902),
    [anon_sym_SQUOTE] = ACTIONS(1902),
    [anon_sym_async] = ACTIONS(1902),
    [anon_sym_break] = ACTIONS(1902),
    [anon_sym_const] = ACTIONS(1902),
    [anon_sym_continue] = ACTIONS(1902),
    [anon_sym_default] = ACTIONS(1902),
    [anon_sym_enum] = ACTIONS(1902),
    [anon_sym_fn] = ACTIONS(1902),
    [anon_sym_for] = ACTIONS(1902),
    [anon_sym_if] = ACTIONS(1902),
    [anon_sym_impl] = ACTIONS(1902),
    [anon_sym_let] = ACTIONS(1902),
    [anon_sym_loop] = ACTIONS(1902),
    [anon_sym_match] = ACTIONS(1902),
    [anon_sym_mod] = ACTIONS(1902),
    [anon_sym_pub] = ACTIONS(1902),
    [anon_sym_return] = ACTIONS(1902),
    [anon_sym_static] = ACTIONS(1902),
    [anon_sym_struct] = ACTIONS(1902),
    [anon_sym_trait] = ACTIONS(1902),
    [anon_sym_type] = ACTIONS(1902),
    [anon_sym_union] = ACTIONS(1902),
    [anon_sym_unsafe] = ACTIONS(1902),
    [anon_sym_use] = ACTIONS(1902),
    [anon_sym_while] = ACTIONS(1902),
    [anon_sym_POUND] = ACTIONS(1900),
    [anon_sym_BANG] = ACTIONS(1900),
    [anon_sym_extern] = ACTIONS(1902),
    [anon_sym_LT] = ACTIONS(1900),
    [anon_sym_COLON_COLON] = ACTIONS(1900),
    [anon_sym_AMP] = ACTIONS(1900),
    [anon_sym_DOT_DOT] = ACTIONS(1900),
    [anon_sym_DASH] = ACTIONS(1900),
    [anon_sym_PIPE] = ACTIONS(1900),
    [anon_sym_yield] = ACTIONS(1902),
    [anon_sym_move] = ACTIONS(1902),
    [sym_integer_literal] = ACTIONS(1900),
    [aux_sym_string_literal_token1] = ACTIONS(1900),
    [sym_char_literal] = ACTIONS(1900),
    [anon_sym_true] = ACTIONS(1902),
    [anon_sym_false] = ACTIONS(1902),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1902),
    [sym_super] = ACTIONS(1902),
    [sym_crate] = ACTIONS(1902),
    [sym_metavariable] = ACTIONS(1900),
    [sym_raw_string_literal] = ACTIONS(1900),
    [sym_float_literal] = ACTIONS(1900),
    [sym_block_comment] = ACTIONS(3),
  },
  [456] = {
    [ts_builtin_sym_end] = ACTIONS(1904),
    [sym_identifier] = ACTIONS(1906),
    [anon_sym_SEMI] = ACTIONS(1904),
    [anon_sym_macro_rules_BANG] = ACTIONS(1904),
    [anon_sym_LPAREN] = ACTIONS(1904),
    [anon_sym_LBRACE] = ACTIONS(1904),
    [anon_sym_RBRACE] = ACTIONS(1904),
    [anon_sym_LBRACK] = ACTIONS(1904),
    [anon_sym_STAR] = ACTIONS(1904),
    [anon_sym_u8] = ACTIONS(1906),
    [anon_sym_i8] = ACTIONS(1906),
    [anon_sym_u16] = ACTIONS(1906),
    [anon_sym_i16] = ACTIONS(1906),
    [anon_sym_u32] = ACTIONS(1906),
    [anon_sym_i32] = ACTIONS(1906),
    [anon_sym_u64] = ACTIONS(1906),
    [anon_sym_i64] = ACTIONS(1906),
    [anon_sym_u128] = ACTIONS(1906),
    [anon_sym_i128] = ACTIONS(1906),
    [anon_sym_isize] = ACTIONS(1906),
    [anon_sym_usize] = ACTIONS(1906),
    [anon_sym_f32] = ACTIONS(1906),
    [anon_sym_f64] = ACTIONS(1906),
    [anon_sym_bool] = ACTIONS(1906),
    [anon_sym_str] = ACTIONS(1906),
    [anon_sym_char] = ACTIONS(1906),
    [anon_sym_SQUOTE] = ACTIONS(1906),
    [anon_sym_async] = ACTIONS(1906),
    [anon_sym_break] = ACTIONS(1906),
    [anon_sym_const] = ACTIONS(1906),
    [anon_sym_continue] = ACTIONS(1906),
    [anon_sym_default] = ACTIONS(1906),
    [anon_sym_enum] = ACTIONS(1906),
    [anon_sym_fn] = ACTIONS(1906),
    [anon_sym_for] = ACTIONS(1906),
    [anon_sym_if] = ACTIONS(1906),
    [anon_sym_impl] = ACTIONS(1906),
    [anon_sym_let] = ACTIONS(1906),
    [anon_sym_loop] = ACTIONS(1906),
    [anon_sym_match] = ACTIONS(1906),
    [anon_sym_mod] = ACTIONS(1906),
    [anon_sym_pub] = ACTIONS(1906),
    [anon_sym_return] = ACTIONS(1906),
    [anon_sym_static] = ACTIONS(1906),
    [anon_sym_struct] = ACTIONS(1906),
    [anon_sym_trait] = ACTIONS(1906),
    [anon_sym_type] = ACTIONS(1906),
    [anon_sym_union] = ACTIONS(1906),
    [anon_sym_unsafe] = ACTIONS(1906),
    [anon_sym_use] = ACTIONS(1906),
    [anon_sym_while] = ACTIONS(1906),
    [anon_sym_POUND] = ACTIONS(1904),
    [anon_sym_BANG] = ACTIONS(1904),
    [anon_sym_extern] = ACTIONS(1906),
    [anon_sym_LT] = ACTIONS(1904),
    [anon_sym_COLON_COLON] = ACTIONS(1904),
    [anon_sym_AMP] = ACTIONS(1904),
    [anon_sym_DOT_DOT] = ACTIONS(1904),
    [anon_sym_DASH] = ACTIONS(1904),
    [anon_sym_PIPE] = ACTIONS(1904),
    [anon_sym_yield] = ACTIONS(1906),
    [anon_sym_move] = ACTIONS(1906),
    [sym_integer_literal] = ACTIONS(1904),
    [aux_sym_string_literal_token1] = ACTIONS(1904),
    [sym_char_literal] = ACTIONS(1904),
    [anon_sym_true] = ACTIONS(1906),
    [anon_sym_false] = ACTIONS(1906),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1906),
    [sym_super] = ACTIONS(1906),
    [sym_crate] = ACTIONS(1906),
    [sym_metavariable] = ACTIONS(1904),
    [sym_raw_string_literal] = ACTIONS(1904),
    [sym_float_literal] = ACTIONS(1904),
    [sym_block_comment] = ACTIONS(3),
  },
  [457] = {
    [ts_builtin_sym_end] = ACTIONS(1908),
    [sym_identifier] = ACTIONS(1910),
    [anon_sym_SEMI] = ACTIONS(1908),
    [anon_sym_macro_rules_BANG] = ACTIONS(1908),
    [anon_sym_LPAREN] = ACTIONS(1908),
    [anon_sym_LBRACE] = ACTIONS(1908),
    [anon_sym_RBRACE] = ACTIONS(1908),
    [anon_sym_LBRACK] = ACTIONS(1908),
    [anon_sym_STAR] = ACTIONS(1908),
    [anon_sym_u8] = ACTIONS(1910),
    [anon_sym_i8] = ACTIONS(1910),
    [anon_sym_u16] = ACTIONS(1910),
    [anon_sym_i16] = ACTIONS(1910),
    [anon_sym_u32] = ACTIONS(1910),
    [anon_sym_i32] = ACTIONS(1910),
    [anon_sym_u64] = ACTIONS(1910),
    [anon_sym_i64] = ACTIONS(1910),
    [anon_sym_u128] = ACTIONS(1910),
    [anon_sym_i128] = ACTIONS(1910),
    [anon_sym_isize] = ACTIONS(1910),
    [anon_sym_usize] = ACTIONS(1910),
    [anon_sym_f32] = ACTIONS(1910),
    [anon_sym_f64] = ACTIONS(1910),
    [anon_sym_bool] = ACTIONS(1910),
    [anon_sym_str] = ACTIONS(1910),
    [anon_sym_char] = ACTIONS(1910),
    [anon_sym_SQUOTE] = ACTIONS(1910),
    [anon_sym_async] = ACTIONS(1910),
    [anon_sym_break] = ACTIONS(1910),
    [anon_sym_const] = ACTIONS(1910),
    [anon_sym_continue] = ACTIONS(1910),
    [anon_sym_default] = ACTIONS(1910),
    [anon_sym_enum] = ACTIONS(1910),
    [anon_sym_fn] = ACTIONS(1910),
    [anon_sym_for] = ACTIONS(1910),
    [anon_sym_if] = ACTIONS(1910),
    [anon_sym_impl] = ACTIONS(1910),
    [anon_sym_let] = ACTIONS(1910),
    [anon_sym_loop] = ACTIONS(1910),
    [anon_sym_match] = ACTIONS(1910),
    [anon_sym_mod] = ACTIONS(1910),
    [anon_sym_pub] = ACTIONS(1910),
    [anon_sym_return] = ACTIONS(1910),
    [anon_sym_static] = ACTIONS(1910),
    [anon_sym_struct] = ACTIONS(1910),
    [anon_sym_trait] = ACTIONS(1910),
    [anon_sym_type] = ACTIONS(1910),
    [anon_sym_union] = ACTIONS(1910),
    [anon_sym_unsafe] = ACTIONS(1910),
    [anon_sym_use] = ACTIONS(1910),
    [anon_sym_while] = ACTIONS(1910),
    [anon_sym_POUND] = ACTIONS(1908),
    [anon_sym_BANG] = ACTIONS(1908),
    [anon_sym_extern] = ACTIONS(1910),
    [anon_sym_LT] = ACTIONS(1908),
    [anon_sym_COLON_COLON] = ACTIONS(1908),
    [anon_sym_AMP] = ACTIONS(1908),
    [anon_sym_DOT_DOT] = ACTIONS(1908),
    [anon_sym_DASH] = ACTIONS(1908),
    [anon_sym_PIPE] = ACTIONS(1908),
    [anon_sym_yield] = ACTIONS(1910),
    [anon_sym_move] = ACTIONS(1910),
    [sym_integer_literal] = ACTIONS(1908),
    [aux_sym_string_literal_token1] = ACTIONS(1908),
    [sym_char_literal] = ACTIONS(1908),
    [anon_sym_true] = ACTIONS(1910),
    [anon_sym_false] = ACTIONS(1910),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1910),
    [sym_super] = ACTIONS(1910),
    [sym_crate] = ACTIONS(1910),
    [sym_metavariable] = ACTIONS(1908),
    [sym_raw_string_literal] = ACTIONS(1908),
    [sym_float_literal] = ACTIONS(1908),
    [sym_block_comment] = ACTIONS(3),
  },
  [458] = {
    [ts_builtin_sym_end] = ACTIONS(1912),
    [sym_identifier] = ACTIONS(1914),
    [anon_sym_SEMI] = ACTIONS(1912),
    [anon_sym_macro_rules_BANG] = ACTIONS(1912),
    [anon_sym_LPAREN] = ACTIONS(1912),
    [anon_sym_LBRACE] = ACTIONS(1912),
    [anon_sym_RBRACE] = ACTIONS(1912),
    [anon_sym_LBRACK] = ACTIONS(1912),
    [anon_sym_STAR] = ACTIONS(1912),
    [anon_sym_u8] = ACTIONS(1914),
    [anon_sym_i8] = ACTIONS(1914),
    [anon_sym_u16] = ACTIONS(1914),
    [anon_sym_i16] = ACTIONS(1914),
    [anon_sym_u32] = ACTIONS(1914),
    [anon_sym_i32] = ACTIONS(1914),
    [anon_sym_u64] = ACTIONS(1914),
    [anon_sym_i64] = ACTIONS(1914),
    [anon_sym_u128] = ACTIONS(1914),
    [anon_sym_i128] = ACTIONS(1914),
    [anon_sym_isize] = ACTIONS(1914),
    [anon_sym_usize] = ACTIONS(1914),
    [anon_sym_f32] = ACTIONS(1914),
    [anon_sym_f64] = ACTIONS(1914),
    [anon_sym_bool] = ACTIONS(1914),
    [anon_sym_str] = ACTIONS(1914),
    [anon_sym_char] = ACTIONS(1914),
    [anon_sym_SQUOTE] = ACTIONS(1914),
    [anon_sym_async] = ACTIONS(1914),
    [anon_sym_break] = ACTIONS(1914),
    [anon_sym_const] = ACTIONS(1914),
    [anon_sym_continue] = ACTIONS(1914),
    [anon_sym_default] = ACTIONS(1914),
    [anon_sym_enum] = ACTIONS(1914),
    [anon_sym_fn] = ACTIONS(1914),
    [anon_sym_for] = ACTIONS(1914),
    [anon_sym_if] = ACTIONS(1914),
    [anon_sym_impl] = ACTIONS(1914),
    [anon_sym_let] = ACTIONS(1914),
    [anon_sym_loop] = ACTIONS(1914),
    [anon_sym_match] = ACTIONS(1914),
    [anon_sym_mod] = ACTIONS(1914),
    [anon_sym_pub] = ACTIONS(1914),
    [anon_sym_return] = ACTIONS(1914),
    [anon_sym_static] = ACTIONS(1914),
    [anon_sym_struct] = ACTIONS(1914),
    [anon_sym_trait] = ACTIONS(1914),
    [anon_sym_type] = ACTIONS(1914),
    [anon_sym_union] = ACTIONS(1914),
    [anon_sym_unsafe] = ACTIONS(1914),
    [anon_sym_use] = ACTIONS(1914),
    [anon_sym_while] = ACTIONS(1914),
    [anon_sym_POUND] = ACTIONS(1912),
    [anon_sym_BANG] = ACTIONS(1912),
    [anon_sym_extern] = ACTIONS(1914),
    [anon_sym_LT] = ACTIONS(1912),
    [anon_sym_COLON_COLON] = ACTIONS(1912),
    [anon_sym_AMP] = ACTIONS(1912),
    [anon_sym_DOT_DOT] = ACTIONS(1912),
    [anon_sym_DASH] = ACTIONS(1912),
    [anon_sym_PIPE] = ACTIONS(1912),
    [anon_sym_yield] = ACTIONS(1914),
    [anon_sym_move] = ACTIONS(1914),
    [sym_integer_literal] = ACTIONS(1912),
    [aux_sym_string_literal_token1] = ACTIONS(1912),
    [sym_char_literal] = ACTIONS(1912),
    [anon_sym_true] = ACTIONS(1914),
    [anon_sym_false] = ACTIONS(1914),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1914),
    [sym_super] = ACTIONS(1914),
    [sym_crate] = ACTIONS(1914),
    [sym_metavariable] = ACTIONS(1912),
    [sym_raw_string_literal] = ACTIONS(1912),
    [sym_float_literal] = ACTIONS(1912),
    [sym_block_comment] = ACTIONS(3),
  },
  [459] = {
    [ts_builtin_sym_end] = ACTIONS(1916),
    [sym_identifier] = ACTIONS(1918),
    [anon_sym_SEMI] = ACTIONS(1916),
    [anon_sym_macro_rules_BANG] = ACTIONS(1916),
    [anon_sym_LPAREN] = ACTIONS(1916),
    [anon_sym_LBRACE] = ACTIONS(1916),
    [anon_sym_RBRACE] = ACTIONS(1916),
    [anon_sym_LBRACK] = ACTIONS(1916),
    [anon_sym_STAR] = ACTIONS(1916),
    [anon_sym_u8] = ACTIONS(1918),
    [anon_sym_i8] = ACTIONS(1918),
    [anon_sym_u16] = ACTIONS(1918),
    [anon_sym_i16] = ACTIONS(1918),
    [anon_sym_u32] = ACTIONS(1918),
    [anon_sym_i32] = ACTIONS(1918),
    [anon_sym_u64] = ACTIONS(1918),
    [anon_sym_i64] = ACTIONS(1918),
    [anon_sym_u128] = ACTIONS(1918),
    [anon_sym_i128] = ACTIONS(1918),
    [anon_sym_isize] = ACTIONS(1918),
    [anon_sym_usize] = ACTIONS(1918),
    [anon_sym_f32] = ACTIONS(1918),
    [anon_sym_f64] = ACTIONS(1918),
    [anon_sym_bool] = ACTIONS(1918),
    [anon_sym_str] = ACTIONS(1918),
    [anon_sym_char] = ACTIONS(1918),
    [anon_sym_SQUOTE] = ACTIONS(1918),
    [anon_sym_async] = ACTIONS(1918),
    [anon_sym_break] = ACTIONS(1918),
    [anon_sym_const] = ACTIONS(1918),
    [anon_sym_continue] = ACTIONS(1918),
    [anon_sym_default] = ACTIONS(1918),
    [anon_sym_enum] = ACTIONS(1918),
    [anon_sym_fn] = ACTIONS(1918),
    [anon_sym_for] = ACTIONS(1918),
    [anon_sym_if] = ACTIONS(1918),
    [anon_sym_impl] = ACTIONS(1918),
    [anon_sym_let] = ACTIONS(1918),
    [anon_sym_loop] = ACTIONS(1918),
    [anon_sym_match] = ACTIONS(1918),
    [anon_sym_mod] = ACTIONS(1918),
    [anon_sym_pub] = ACTIONS(1918),
    [anon_sym_return] = ACTIONS(1918),
    [anon_sym_static] = ACTIONS(1918),
    [anon_sym_struct] = ACTIONS(1918),
    [anon_sym_trait] = ACTIONS(1918),
    [anon_sym_type] = ACTIONS(1918),
    [anon_sym_union] = ACTIONS(1918),
    [anon_sym_unsafe] = ACTIONS(1918),
    [anon_sym_use] = ACTIONS(1918),
    [anon_sym_while] = ACTIONS(1918),
    [anon_sym_POUND] = ACTIONS(1916),
    [anon_sym_BANG] = ACTIONS(1916),
    [anon_sym_extern] = ACTIONS(1918),
    [anon_sym_LT] = ACTIONS(1916),
    [anon_sym_COLON_COLON] = ACTIONS(1916),
    [anon_sym_AMP] = ACTIONS(1916),
    [anon_sym_DOT_DOT] = ACTIONS(1916),
    [anon_sym_DASH] = ACTIONS(1916),
    [anon_sym_PIPE] = ACTIONS(1916),
    [anon_sym_yield] = ACTIONS(1918),
    [anon_sym_move] = ACTIONS(1918),
    [sym_integer_literal] = ACTIONS(1916),
    [aux_sym_string_literal_token1] = ACTIONS(1916),
    [sym_char_literal] = ACTIONS(1916),
    [anon_sym_true] = ACTIONS(1918),
    [anon_sym_false] = ACTIONS(1918),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1918),
    [sym_super] = ACTIONS(1918),
    [sym_crate] = ACTIONS(1918),
    [sym_metavariable] = ACTIONS(1916),
    [sym_raw_string_literal] = ACTIONS(1916),
    [sym_float_literal] = ACTIONS(1916),
    [sym_block_comment] = ACTIONS(3),
  },
  [460] = {
    [ts_builtin_sym_end] = ACTIONS(1920),
    [sym_identifier] = ACTIONS(1922),
    [anon_sym_SEMI] = ACTIONS(1920),
    [anon_sym_macro_rules_BANG] = ACTIONS(1920),
    [anon_sym_LPAREN] = ACTIONS(1920),
    [anon_sym_LBRACE] = ACTIONS(1920),
    [anon_sym_RBRACE] = ACTIONS(1920),
    [anon_sym_LBRACK] = ACTIONS(1920),
    [anon_sym_STAR] = ACTIONS(1920),
    [anon_sym_u8] = ACTIONS(1922),
    [anon_sym_i8] = ACTIONS(1922),
    [anon_sym_u16] = ACTIONS(1922),
    [anon_sym_i16] = ACTIONS(1922),
    [anon_sym_u32] = ACTIONS(1922),
    [anon_sym_i32] = ACTIONS(1922),
    [anon_sym_u64] = ACTIONS(1922),
    [anon_sym_i64] = ACTIONS(1922),
    [anon_sym_u128] = ACTIONS(1922),
    [anon_sym_i128] = ACTIONS(1922),
    [anon_sym_isize] = ACTIONS(1922),
    [anon_sym_usize] = ACTIONS(1922),
    [anon_sym_f32] = ACTIONS(1922),
    [anon_sym_f64] = ACTIONS(1922),
    [anon_sym_bool] = ACTIONS(1922),
    [anon_sym_str] = ACTIONS(1922),
    [anon_sym_char] = ACTIONS(1922),
    [anon_sym_SQUOTE] = ACTIONS(1922),
    [anon_sym_async] = ACTIONS(1922),
    [anon_sym_break] = ACTIONS(1922),
    [anon_sym_const] = ACTIONS(1922),
    [anon_sym_continue] = ACTIONS(1922),
    [anon_sym_default] = ACTIONS(1922),
    [anon_sym_enum] = ACTIONS(1922),
    [anon_sym_fn] = ACTIONS(1922),
    [anon_sym_for] = ACTIONS(1922),
    [anon_sym_if] = ACTIONS(1922),
    [anon_sym_impl] = ACTIONS(1922),
    [anon_sym_let] = ACTIONS(1922),
    [anon_sym_loop] = ACTIONS(1922),
    [anon_sym_match] = ACTIONS(1922),
    [anon_sym_mod] = ACTIONS(1922),
    [anon_sym_pub] = ACTIONS(1922),
    [anon_sym_return] = ACTIONS(1922),
    [anon_sym_static] = ACTIONS(1922),
    [anon_sym_struct] = ACTIONS(1922),
    [anon_sym_trait] = ACTIONS(1922),
    [anon_sym_type] = ACTIONS(1922),
    [anon_sym_union] = ACTIONS(1922),
    [anon_sym_unsafe] = ACTIONS(1922),
    [anon_sym_use] = ACTIONS(1922),
    [anon_sym_while] = ACTIONS(1922),
    [anon_sym_POUND] = ACTIONS(1920),
    [anon_sym_BANG] = ACTIONS(1920),
    [anon_sym_extern] = ACTIONS(1922),
    [anon_sym_LT] = ACTIONS(1920),
    [anon_sym_COLON_COLON] = ACTIONS(1920),
    [anon_sym_AMP] = ACTIONS(1920),
    [anon_sym_DOT_DOT] = ACTIONS(1920),
    [anon_sym_DASH] = ACTIONS(1920),
    [anon_sym_PIPE] = ACTIONS(1920),
    [anon_sym_yield] = ACTIONS(1922),
    [anon_sym_move] = ACTIONS(1922),
    [sym_integer_literal] = ACTIONS(1920),
    [aux_sym_string_literal_token1] = ACTIONS(1920),
    [sym_char_literal] = ACTIONS(1920),
    [anon_sym_true] = ACTIONS(1922),
    [anon_sym_false] = ACTIONS(1922),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1922),
    [sym_super] = ACTIONS(1922),
    [sym_crate] = ACTIONS(1922),
    [sym_metavariable] = ACTIONS(1920),
    [sym_raw_string_literal] = ACTIONS(1920),
    [sym_float_literal] = ACTIONS(1920),
    [sym_block_comment] = ACTIONS(3),
  },
  [461] = {
    [ts_builtin_sym_end] = ACTIONS(1924),
    [sym_identifier] = ACTIONS(1926),
    [anon_sym_SEMI] = ACTIONS(1924),
    [anon_sym_macro_rules_BANG] = ACTIONS(1924),
    [anon_sym_LPAREN] = ACTIONS(1924),
    [anon_sym_LBRACE] = ACTIONS(1924),
    [anon_sym_RBRACE] = ACTIONS(1924),
    [anon_sym_LBRACK] = ACTIONS(1924),
    [anon_sym_STAR] = ACTIONS(1924),
    [anon_sym_u8] = ACTIONS(1926),
    [anon_sym_i8] = ACTIONS(1926),
    [anon_sym_u16] = ACTIONS(1926),
    [anon_sym_i16] = ACTIONS(1926),
    [anon_sym_u32] = ACTIONS(1926),
    [anon_sym_i32] = ACTIONS(1926),
    [anon_sym_u64] = ACTIONS(1926),
    [anon_sym_i64] = ACTIONS(1926),
    [anon_sym_u128] = ACTIONS(1926),
    [anon_sym_i128] = ACTIONS(1926),
    [anon_sym_isize] = ACTIONS(1926),
    [anon_sym_usize] = ACTIONS(1926),
    [anon_sym_f32] = ACTIONS(1926),
    [anon_sym_f64] = ACTIONS(1926),
    [anon_sym_bool] = ACTIONS(1926),
    [anon_sym_str] = ACTIONS(1926),
    [anon_sym_char] = ACTIONS(1926),
    [anon_sym_SQUOTE] = ACTIONS(1926),
    [anon_sym_async] = ACTIONS(1926),
    [anon_sym_break] = ACTIONS(1926),
    [anon_sym_const] = ACTIONS(1926),
    [anon_sym_continue] = ACTIONS(1926),
    [anon_sym_default] = ACTIONS(1926),
    [anon_sym_enum] = ACTIONS(1926),
    [anon_sym_fn] = ACTIONS(1926),
    [anon_sym_for] = ACTIONS(1926),
    [anon_sym_if] = ACTIONS(1926),
    [anon_sym_impl] = ACTIONS(1926),
    [anon_sym_let] = ACTIONS(1926),
    [anon_sym_loop] = ACTIONS(1926),
    [anon_sym_match] = ACTIONS(1926),
    [anon_sym_mod] = ACTIONS(1926),
    [anon_sym_pub] = ACTIONS(1926),
    [anon_sym_return] = ACTIONS(1926),
    [anon_sym_static] = ACTIONS(1926),
    [anon_sym_struct] = ACTIONS(1926),
    [anon_sym_trait] = ACTIONS(1926),
    [anon_sym_type] = ACTIONS(1926),
    [anon_sym_union] = ACTIONS(1926),
    [anon_sym_unsafe] = ACTIONS(1926),
    [anon_sym_use] = ACTIONS(1926),
    [anon_sym_while] = ACTIONS(1926),
    [anon_sym_POUND] = ACTIONS(1924),
    [anon_sym_BANG] = ACTIONS(1924),
    [anon_sym_extern] = ACTIONS(1926),
    [anon_sym_LT] = ACTIONS(1924),
    [anon_sym_COLON_COLON] = ACTIONS(1924),
    [anon_sym_AMP] = ACTIONS(1924),
    [anon_sym_DOT_DOT] = ACTIONS(1924),
    [anon_sym_DASH] = ACTIONS(1924),
    [anon_sym_PIPE] = ACTIONS(1924),
    [anon_sym_yield] = ACTIONS(1926),
    [anon_sym_move] = ACTIONS(1926),
    [sym_integer_literal] = ACTIONS(1924),
    [aux_sym_string_literal_token1] = ACTIONS(1924),
    [sym_char_literal] = ACTIONS(1924),
    [anon_sym_true] = ACTIONS(1926),
    [anon_sym_false] = ACTIONS(1926),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1926),
    [sym_super] = ACTIONS(1926),
    [sym_crate] = ACTIONS(1926),
    [sym_metavariable] = ACTIONS(1924),
    [sym_raw_string_literal] = ACTIONS(1924),
    [sym_float_literal] = ACTIONS(1924),
    [sym_block_comment] = ACTIONS(3),
  },
  [462] = {
    [ts_builtin_sym_end] = ACTIONS(1928),
    [sym_identifier] = ACTIONS(1930),
    [anon_sym_SEMI] = ACTIONS(1928),
    [anon_sym_macro_rules_BANG] = ACTIONS(1928),
    [anon_sym_LPAREN] = ACTIONS(1928),
    [anon_sym_LBRACE] = ACTIONS(1928),
    [anon_sym_RBRACE] = ACTIONS(1928),
    [anon_sym_LBRACK] = ACTIONS(1928),
    [anon_sym_STAR] = ACTIONS(1928),
    [anon_sym_u8] = ACTIONS(1930),
    [anon_sym_i8] = ACTIONS(1930),
    [anon_sym_u16] = ACTIONS(1930),
    [anon_sym_i16] = ACTIONS(1930),
    [anon_sym_u32] = ACTIONS(1930),
    [anon_sym_i32] = ACTIONS(1930),
    [anon_sym_u64] = ACTIONS(1930),
    [anon_sym_i64] = ACTIONS(1930),
    [anon_sym_u128] = ACTIONS(1930),
    [anon_sym_i128] = ACTIONS(1930),
    [anon_sym_isize] = ACTIONS(1930),
    [anon_sym_usize] = ACTIONS(1930),
    [anon_sym_f32] = ACTIONS(1930),
    [anon_sym_f64] = ACTIONS(1930),
    [anon_sym_bool] = ACTIONS(1930),
    [anon_sym_str] = ACTIONS(1930),
    [anon_sym_char] = ACTIONS(1930),
    [anon_sym_SQUOTE] = ACTIONS(1930),
    [anon_sym_async] = ACTIONS(1930),
    [anon_sym_break] = ACTIONS(1930),
    [anon_sym_const] = ACTIONS(1930),
    [anon_sym_continue] = ACTIONS(1930),
    [anon_sym_default] = ACTIONS(1930),
    [anon_sym_enum] = ACTIONS(1930),
    [anon_sym_fn] = ACTIONS(1930),
    [anon_sym_for] = ACTIONS(1930),
    [anon_sym_if] = ACTIONS(1930),
    [anon_sym_impl] = ACTIONS(1930),
    [anon_sym_let] = ACTIONS(1930),
    [anon_sym_loop] = ACTIONS(1930),
    [anon_sym_match] = ACTIONS(1930),
    [anon_sym_mod] = ACTIONS(1930),
    [anon_sym_pub] = ACTIONS(1930),
    [anon_sym_return] = ACTIONS(1930),
    [anon_sym_static] = ACTIONS(1930),
    [anon_sym_struct] = ACTIONS(1930),
    [anon_sym_trait] = ACTIONS(1930),
    [anon_sym_type] = ACTIONS(1930),
    [anon_sym_union] = ACTIONS(1930),
    [anon_sym_unsafe] = ACTIONS(1930),
    [anon_sym_use] = ACTIONS(1930),
    [anon_sym_while] = ACTIONS(1930),
    [anon_sym_POUND] = ACTIONS(1928),
    [anon_sym_BANG] = ACTIONS(1928),
    [anon_sym_extern] = ACTIONS(1930),
    [anon_sym_LT] = ACTIONS(1928),
    [anon_sym_COLON_COLON] = ACTIONS(1928),
    [anon_sym_AMP] = ACTIONS(1928),
    [anon_sym_DOT_DOT] = ACTIONS(1928),
    [anon_sym_DASH] = ACTIONS(1928),
    [anon_sym_PIPE] = ACTIONS(1928),
    [anon_sym_yield] = ACTIONS(1930),
    [anon_sym_move] = ACTIONS(1930),
    [sym_integer_literal] = ACTIONS(1928),
    [aux_sym_string_literal_token1] = ACTIONS(1928),
    [sym_char_literal] = ACTIONS(1928),
    [anon_sym_true] = ACTIONS(1930),
    [anon_sym_false] = ACTIONS(1930),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1930),
    [sym_super] = ACTIONS(1930),
    [sym_crate] = ACTIONS(1930),
    [sym_metavariable] = ACTIONS(1928),
    [sym_raw_string_literal] = ACTIONS(1928),
    [sym_float_literal] = ACTIONS(1928),
    [sym_block_comment] = ACTIONS(3),
  },
  [463] = {
    [ts_builtin_sym_end] = ACTIONS(1932),
    [sym_identifier] = ACTIONS(1934),
    [anon_sym_SEMI] = ACTIONS(1932),
    [anon_sym_macro_rules_BANG] = ACTIONS(1932),
    [anon_sym_LPAREN] = ACTIONS(1932),
    [anon_sym_LBRACE] = ACTIONS(1932),
    [anon_sym_RBRACE] = ACTIONS(1932),
    [anon_sym_LBRACK] = ACTIONS(1932),
    [anon_sym_STAR] = ACTIONS(1932),
    [anon_sym_u8] = ACTIONS(1934),
    [anon_sym_i8] = ACTIONS(1934),
    [anon_sym_u16] = ACTIONS(1934),
    [anon_sym_i16] = ACTIONS(1934),
    [anon_sym_u32] = ACTIONS(1934),
    [anon_sym_i32] = ACTIONS(1934),
    [anon_sym_u64] = ACTIONS(1934),
    [anon_sym_i64] = ACTIONS(1934),
    [anon_sym_u128] = ACTIONS(1934),
    [anon_sym_i128] = ACTIONS(1934),
    [anon_sym_isize] = ACTIONS(1934),
    [anon_sym_usize] = ACTIONS(1934),
    [anon_sym_f32] = ACTIONS(1934),
    [anon_sym_f64] = ACTIONS(1934),
    [anon_sym_bool] = ACTIONS(1934),
    [anon_sym_str] = ACTIONS(1934),
    [anon_sym_char] = ACTIONS(1934),
    [anon_sym_SQUOTE] = ACTIONS(1934),
    [anon_sym_async] = ACTIONS(1934),
    [anon_sym_break] = ACTIONS(1934),
    [anon_sym_const] = ACTIONS(1934),
    [anon_sym_continue] = ACTIONS(1934),
    [anon_sym_default] = ACTIONS(1934),
    [anon_sym_enum] = ACTIONS(1934),
    [anon_sym_fn] = ACTIONS(1934),
    [anon_sym_for] = ACTIONS(1934),
    [anon_sym_if] = ACTIONS(1934),
    [anon_sym_impl] = ACTIONS(1934),
    [anon_sym_let] = ACTIONS(1934),
    [anon_sym_loop] = ACTIONS(1934),
    [anon_sym_match] = ACTIONS(1934),
    [anon_sym_mod] = ACTIONS(1934),
    [anon_sym_pub] = ACTIONS(1934),
    [anon_sym_return] = ACTIONS(1934),
    [anon_sym_static] = ACTIONS(1934),
    [anon_sym_struct] = ACTIONS(1934),
    [anon_sym_trait] = ACTIONS(1934),
    [anon_sym_type] = ACTIONS(1934),
    [anon_sym_union] = ACTIONS(1934),
    [anon_sym_unsafe] = ACTIONS(1934),
    [anon_sym_use] = ACTIONS(1934),
    [anon_sym_while] = ACTIONS(1934),
    [anon_sym_POUND] = ACTIONS(1932),
    [anon_sym_BANG] = ACTIONS(1932),
    [anon_sym_extern] = ACTIONS(1934),
    [anon_sym_LT] = ACTIONS(1932),
    [anon_sym_COLON_COLON] = ACTIONS(1932),
    [anon_sym_AMP] = ACTIONS(1932),
    [anon_sym_DOT_DOT] = ACTIONS(1932),
    [anon_sym_DASH] = ACTIONS(1932),
    [anon_sym_PIPE] = ACTIONS(1932),
    [anon_sym_yield] = ACTIONS(1934),
    [anon_sym_move] = ACTIONS(1934),
    [sym_integer_literal] = ACTIONS(1932),
    [aux_sym_string_literal_token1] = ACTIONS(1932),
    [sym_char_literal] = ACTIONS(1932),
    [anon_sym_true] = ACTIONS(1934),
    [anon_sym_false] = ACTIONS(1934),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1934),
    [sym_super] = ACTIONS(1934),
    [sym_crate] = ACTIONS(1934),
    [sym_metavariable] = ACTIONS(1932),
    [sym_raw_string_literal] = ACTIONS(1932),
    [sym_float_literal] = ACTIONS(1932),
    [sym_block_comment] = ACTIONS(3),
  },
  [464] = {
    [ts_builtin_sym_end] = ACTIONS(1936),
    [sym_identifier] = ACTIONS(1938),
    [anon_sym_SEMI] = ACTIONS(1936),
    [anon_sym_macro_rules_BANG] = ACTIONS(1936),
    [anon_sym_LPAREN] = ACTIONS(1936),
    [anon_sym_LBRACE] = ACTIONS(1936),
    [anon_sym_RBRACE] = ACTIONS(1936),
    [anon_sym_LBRACK] = ACTIONS(1936),
    [anon_sym_STAR] = ACTIONS(1936),
    [anon_sym_u8] = ACTIONS(1938),
    [anon_sym_i8] = ACTIONS(1938),
    [anon_sym_u16] = ACTIONS(1938),
    [anon_sym_i16] = ACTIONS(1938),
    [anon_sym_u32] = ACTIONS(1938),
    [anon_sym_i32] = ACTIONS(1938),
    [anon_sym_u64] = ACTIONS(1938),
    [anon_sym_i64] = ACTIONS(1938),
    [anon_sym_u128] = ACTIONS(1938),
    [anon_sym_i128] = ACTIONS(1938),
    [anon_sym_isize] = ACTIONS(1938),
    [anon_sym_usize] = ACTIONS(1938),
    [anon_sym_f32] = ACTIONS(1938),
    [anon_sym_f64] = ACTIONS(1938),
    [anon_sym_bool] = ACTIONS(1938),
    [anon_sym_str] = ACTIONS(1938),
    [anon_sym_char] = ACTIONS(1938),
    [anon_sym_SQUOTE] = ACTIONS(1938),
    [anon_sym_async] = ACTIONS(1938),
    [anon_sym_break] = ACTIONS(1938),
    [anon_sym_const] = ACTIONS(1938),
    [anon_sym_continue] = ACTIONS(1938),
    [anon_sym_default] = ACTIONS(1938),
    [anon_sym_enum] = ACTIONS(1938),
    [anon_sym_fn] = ACTIONS(1938),
    [anon_sym_for] = ACTIONS(1938),
    [anon_sym_if] = ACTIONS(1938),
    [anon_sym_impl] = ACTIONS(1938),
    [anon_sym_let] = ACTIONS(1938),
    [anon_sym_loop] = ACTIONS(1938),
    [anon_sym_match] = ACTIONS(1938),
    [anon_sym_mod] = ACTIONS(1938),
    [anon_sym_pub] = ACTIONS(1938),
    [anon_sym_return] = ACTIONS(1938),
    [anon_sym_static] = ACTIONS(1938),
    [anon_sym_struct] = ACTIONS(1938),
    [anon_sym_trait] = ACTIONS(1938),
    [anon_sym_type] = ACTIONS(1938),
    [anon_sym_union] = ACTIONS(1938),
    [anon_sym_unsafe] = ACTIONS(1938),
    [anon_sym_use] = ACTIONS(1938),
    [anon_sym_while] = ACTIONS(1938),
    [anon_sym_POUND] = ACTIONS(1936),
    [anon_sym_BANG] = ACTIONS(1936),
    [anon_sym_extern] = ACTIONS(1938),
    [anon_sym_LT] = ACTIONS(1936),
    [anon_sym_COLON_COLON] = ACTIONS(1936),
    [anon_sym_AMP] = ACTIONS(1936),
    [anon_sym_DOT_DOT] = ACTIONS(1936),
    [anon_sym_DASH] = ACTIONS(1936),
    [anon_sym_PIPE] = ACTIONS(1936),
    [anon_sym_yield] = ACTIONS(1938),
    [anon_sym_move] = ACTIONS(1938),
    [sym_integer_literal] = ACTIONS(1936),
    [aux_sym_string_literal_token1] = ACTIONS(1936),
    [sym_char_literal] = ACTIONS(1936),
    [anon_sym_true] = ACTIONS(1938),
    [anon_sym_false] = ACTIONS(1938),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1938),
    [sym_super] = ACTIONS(1938),
    [sym_crate] = ACTIONS(1938),
    [sym_metavariable] = ACTIONS(1936),
    [sym_raw_string_literal] = ACTIONS(1936),
    [sym_float_literal] = ACTIONS(1936),
    [sym_block_comment] = ACTIONS(3),
  },
  [465] = {
    [ts_builtin_sym_end] = ACTIONS(1940),
    [sym_identifier] = ACTIONS(1942),
    [anon_sym_SEMI] = ACTIONS(1940),
    [anon_sym_macro_rules_BANG] = ACTIONS(1940),
    [anon_sym_LPAREN] = ACTIONS(1940),
    [anon_sym_LBRACE] = ACTIONS(1940),
    [anon_sym_RBRACE] = ACTIONS(1940),
    [anon_sym_LBRACK] = ACTIONS(1940),
    [anon_sym_STAR] = ACTIONS(1940),
    [anon_sym_u8] = ACTIONS(1942),
    [anon_sym_i8] = ACTIONS(1942),
    [anon_sym_u16] = ACTIONS(1942),
    [anon_sym_i16] = ACTIONS(1942),
    [anon_sym_u32] = ACTIONS(1942),
    [anon_sym_i32] = ACTIONS(1942),
    [anon_sym_u64] = ACTIONS(1942),
    [anon_sym_i64] = ACTIONS(1942),
    [anon_sym_u128] = ACTIONS(1942),
    [anon_sym_i128] = ACTIONS(1942),
    [anon_sym_isize] = ACTIONS(1942),
    [anon_sym_usize] = ACTIONS(1942),
    [anon_sym_f32] = ACTIONS(1942),
    [anon_sym_f64] = ACTIONS(1942),
    [anon_sym_bool] = ACTIONS(1942),
    [anon_sym_str] = ACTIONS(1942),
    [anon_sym_char] = ACTIONS(1942),
    [anon_sym_SQUOTE] = ACTIONS(1942),
    [anon_sym_async] = ACTIONS(1942),
    [anon_sym_break] = ACTIONS(1942),
    [anon_sym_const] = ACTIONS(1942),
    [anon_sym_continue] = ACTIONS(1942),
    [anon_sym_default] = ACTIONS(1942),
    [anon_sym_enum] = ACTIONS(1942),
    [anon_sym_fn] = ACTIONS(1942),
    [anon_sym_for] = ACTIONS(1942),
    [anon_sym_if] = ACTIONS(1942),
    [anon_sym_impl] = ACTIONS(1942),
    [anon_sym_let] = ACTIONS(1942),
    [anon_sym_loop] = ACTIONS(1942),
    [anon_sym_match] = ACTIONS(1942),
    [anon_sym_mod] = ACTIONS(1942),
    [anon_sym_pub] = ACTIONS(1942),
    [anon_sym_return] = ACTIONS(1942),
    [anon_sym_static] = ACTIONS(1942),
    [anon_sym_struct] = ACTIONS(1942),
    [anon_sym_trait] = ACTIONS(1942),
    [anon_sym_type] = ACTIONS(1942),
    [anon_sym_union] = ACTIONS(1942),
    [anon_sym_unsafe] = ACTIONS(1942),
    [anon_sym_use] = ACTIONS(1942),
    [anon_sym_while] = ACTIONS(1942),
    [anon_sym_POUND] = ACTIONS(1940),
    [anon_sym_BANG] = ACTIONS(1940),
    [anon_sym_extern] = ACTIONS(1942),
    [anon_sym_LT] = ACTIONS(1940),
    [anon_sym_COLON_COLON] = ACTIONS(1940),
    [anon_sym_AMP] = ACTIONS(1940),
    [anon_sym_DOT_DOT] = ACTIONS(1940),
    [anon_sym_DASH] = ACTIONS(1940),
    [anon_sym_PIPE] = ACTIONS(1940),
    [anon_sym_yield] = ACTIONS(1942),
    [anon_sym_move] = ACTIONS(1942),
    [sym_integer_literal] = ACTIONS(1940),
    [aux_sym_string_literal_token1] = ACTIONS(1940),
    [sym_char_literal] = ACTIONS(1940),
    [anon_sym_true] = ACTIONS(1942),
    [anon_sym_false] = ACTIONS(1942),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1942),
    [sym_super] = ACTIONS(1942),
    [sym_crate] = ACTIONS(1942),
    [sym_metavariable] = ACTIONS(1940),
    [sym_raw_string_literal] = ACTIONS(1940),
    [sym_float_literal] = ACTIONS(1940),
    [sym_block_comment] = ACTIONS(3),
  },
  [466] = {
    [ts_builtin_sym_end] = ACTIONS(1944),
    [sym_identifier] = ACTIONS(1946),
    [anon_sym_SEMI] = ACTIONS(1944),
    [anon_sym_macro_rules_BANG] = ACTIONS(1944),
    [anon_sym_LPAREN] = ACTIONS(1944),
    [anon_sym_LBRACE] = ACTIONS(1944),
    [anon_sym_RBRACE] = ACTIONS(1944),
    [anon_sym_LBRACK] = ACTIONS(1944),
    [anon_sym_STAR] = ACTIONS(1944),
    [anon_sym_u8] = ACTIONS(1946),
    [anon_sym_i8] = ACTIONS(1946),
    [anon_sym_u16] = ACTIONS(1946),
    [anon_sym_i16] = ACTIONS(1946),
    [anon_sym_u32] = ACTIONS(1946),
    [anon_sym_i32] = ACTIONS(1946),
    [anon_sym_u64] = ACTIONS(1946),
    [anon_sym_i64] = ACTIONS(1946),
    [anon_sym_u128] = ACTIONS(1946),
    [anon_sym_i128] = ACTIONS(1946),
    [anon_sym_isize] = ACTIONS(1946),
    [anon_sym_usize] = ACTIONS(1946),
    [anon_sym_f32] = ACTIONS(1946),
    [anon_sym_f64] = ACTIONS(1946),
    [anon_sym_bool] = ACTIONS(1946),
    [anon_sym_str] = ACTIONS(1946),
    [anon_sym_char] = ACTIONS(1946),
    [anon_sym_SQUOTE] = ACTIONS(1946),
    [anon_sym_async] = ACTIONS(1946),
    [anon_sym_break] = ACTIONS(1946),
    [anon_sym_const] = ACTIONS(1946),
    [anon_sym_continue] = ACTIONS(1946),
    [anon_sym_default] = ACTIONS(1946),
    [anon_sym_enum] = ACTIONS(1946),
    [anon_sym_fn] = ACTIONS(1946),
    [anon_sym_for] = ACTIONS(1946),
    [anon_sym_if] = ACTIONS(1946),
    [anon_sym_impl] = ACTIONS(1946),
    [anon_sym_let] = ACTIONS(1946),
    [anon_sym_loop] = ACTIONS(1946),
    [anon_sym_match] = ACTIONS(1946),
    [anon_sym_mod] = ACTIONS(1946),
    [anon_sym_pub] = ACTIONS(1946),
    [anon_sym_return] = ACTIONS(1946),
    [anon_sym_static] = ACTIONS(1946),
    [anon_sym_struct] = ACTIONS(1946),
    [anon_sym_trait] = ACTIONS(1946),
    [anon_sym_type] = ACTIONS(1946),
    [anon_sym_union] = ACTIONS(1946),
    [anon_sym_unsafe] = ACTIONS(1946),
    [anon_sym_use] = ACTIONS(1946),
    [anon_sym_while] = ACTIONS(1946),
    [anon_sym_POUND] = ACTIONS(1944),
    [anon_sym_BANG] = ACTIONS(1944),
    [anon_sym_extern] = ACTIONS(1946),
    [anon_sym_LT] = ACTIONS(1944),
    [anon_sym_COLON_COLON] = ACTIONS(1944),
    [anon_sym_AMP] = ACTIONS(1944),
    [anon_sym_DOT_DOT] = ACTIONS(1944),
    [anon_sym_DASH] = ACTIONS(1944),
    [anon_sym_PIPE] = ACTIONS(1944),
    [anon_sym_yield] = ACTIONS(1946),
    [anon_sym_move] = ACTIONS(1946),
    [sym_integer_literal] = ACTIONS(1944),
    [aux_sym_string_literal_token1] = ACTIONS(1944),
    [sym_char_literal] = ACTIONS(1944),
    [anon_sym_true] = ACTIONS(1946),
    [anon_sym_false] = ACTIONS(1946),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1946),
    [sym_super] = ACTIONS(1946),
    [sym_crate] = ACTIONS(1946),
    [sym_metavariable] = ACTIONS(1944),
    [sym_raw_string_literal] = ACTIONS(1944),
    [sym_float_literal] = ACTIONS(1944),
    [sym_block_comment] = ACTIONS(3),
  },
  [467] = {
    [ts_builtin_sym_end] = ACTIONS(1948),
    [sym_identifier] = ACTIONS(1950),
    [anon_sym_SEMI] = ACTIONS(1948),
    [anon_sym_macro_rules_BANG] = ACTIONS(1948),
    [anon_sym_LPAREN] = ACTIONS(1948),
    [anon_sym_LBRACE] = ACTIONS(1948),
    [anon_sym_RBRACE] = ACTIONS(1948),
    [anon_sym_LBRACK] = ACTIONS(1948),
    [anon_sym_STAR] = ACTIONS(1948),
    [anon_sym_u8] = ACTIONS(1950),
    [anon_sym_i8] = ACTIONS(1950),
    [anon_sym_u16] = ACTIONS(1950),
    [anon_sym_i16] = ACTIONS(1950),
    [anon_sym_u32] = ACTIONS(1950),
    [anon_sym_i32] = ACTIONS(1950),
    [anon_sym_u64] = ACTIONS(1950),
    [anon_sym_i64] = ACTIONS(1950),
    [anon_sym_u128] = ACTIONS(1950),
    [anon_sym_i128] = ACTIONS(1950),
    [anon_sym_isize] = ACTIONS(1950),
    [anon_sym_usize] = ACTIONS(1950),
    [anon_sym_f32] = ACTIONS(1950),
    [anon_sym_f64] = ACTIONS(1950),
    [anon_sym_bool] = ACTIONS(1950),
    [anon_sym_str] = ACTIONS(1950),
    [anon_sym_char] = ACTIONS(1950),
    [anon_sym_SQUOTE] = ACTIONS(1950),
    [anon_sym_async] = ACTIONS(1950),
    [anon_sym_break] = ACTIONS(1950),
    [anon_sym_const] = ACTIONS(1950),
    [anon_sym_continue] = ACTIONS(1950),
    [anon_sym_default] = ACTIONS(1950),
    [anon_sym_enum] = ACTIONS(1950),
    [anon_sym_fn] = ACTIONS(1950),
    [anon_sym_for] = ACTIONS(1950),
    [anon_sym_if] = ACTIONS(1950),
    [anon_sym_impl] = ACTIONS(1950),
    [anon_sym_let] = ACTIONS(1950),
    [anon_sym_loop] = ACTIONS(1950),
    [anon_sym_match] = ACTIONS(1950),
    [anon_sym_mod] = ACTIONS(1950),
    [anon_sym_pub] = ACTIONS(1950),
    [anon_sym_return] = ACTIONS(1950),
    [anon_sym_static] = ACTIONS(1950),
    [anon_sym_struct] = ACTIONS(1950),
    [anon_sym_trait] = ACTIONS(1950),
    [anon_sym_type] = ACTIONS(1950),
    [anon_sym_union] = ACTIONS(1950),
    [anon_sym_unsafe] = ACTIONS(1950),
    [anon_sym_use] = ACTIONS(1950),
    [anon_sym_while] = ACTIONS(1950),
    [anon_sym_POUND] = ACTIONS(1948),
    [anon_sym_BANG] = ACTIONS(1948),
    [anon_sym_extern] = ACTIONS(1950),
    [anon_sym_LT] = ACTIONS(1948),
    [anon_sym_COLON_COLON] = ACTIONS(1948),
    [anon_sym_AMP] = ACTIONS(1948),
    [anon_sym_DOT_DOT] = ACTIONS(1948),
    [anon_sym_DASH] = ACTIONS(1948),
    [anon_sym_PIPE] = ACTIONS(1948),
    [anon_sym_yield] = ACTIONS(1950),
    [anon_sym_move] = ACTIONS(1950),
    [sym_integer_literal] = ACTIONS(1948),
    [aux_sym_string_literal_token1] = ACTIONS(1948),
    [sym_char_literal] = ACTIONS(1948),
    [anon_sym_true] = ACTIONS(1950),
    [anon_sym_false] = ACTIONS(1950),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1950),
    [sym_super] = ACTIONS(1950),
    [sym_crate] = ACTIONS(1950),
    [sym_metavariable] = ACTIONS(1948),
    [sym_raw_string_literal] = ACTIONS(1948),
    [sym_float_literal] = ACTIONS(1948),
    [sym_block_comment] = ACTIONS(3),
  },
  [468] = {
    [ts_builtin_sym_end] = ACTIONS(1952),
    [sym_identifier] = ACTIONS(1954),
    [anon_sym_SEMI] = ACTIONS(1952),
    [anon_sym_macro_rules_BANG] = ACTIONS(1952),
    [anon_sym_LPAREN] = ACTIONS(1952),
    [anon_sym_LBRACE] = ACTIONS(1952),
    [anon_sym_RBRACE] = ACTIONS(1952),
    [anon_sym_LBRACK] = ACTIONS(1952),
    [anon_sym_STAR] = ACTIONS(1952),
    [anon_sym_u8] = ACTIONS(1954),
    [anon_sym_i8] = ACTIONS(1954),
    [anon_sym_u16] = ACTIONS(1954),
    [anon_sym_i16] = ACTIONS(1954),
    [anon_sym_u32] = ACTIONS(1954),
    [anon_sym_i32] = ACTIONS(1954),
    [anon_sym_u64] = ACTIONS(1954),
    [anon_sym_i64] = ACTIONS(1954),
    [anon_sym_u128] = ACTIONS(1954),
    [anon_sym_i128] = ACTIONS(1954),
    [anon_sym_isize] = ACTIONS(1954),
    [anon_sym_usize] = ACTIONS(1954),
    [anon_sym_f32] = ACTIONS(1954),
    [anon_sym_f64] = ACTIONS(1954),
    [anon_sym_bool] = ACTIONS(1954),
    [anon_sym_str] = ACTIONS(1954),
    [anon_sym_char] = ACTIONS(1954),
    [anon_sym_SQUOTE] = ACTIONS(1954),
    [anon_sym_async] = ACTIONS(1954),
    [anon_sym_break] = ACTIONS(1954),
    [anon_sym_const] = ACTIONS(1954),
    [anon_sym_continue] = ACTIONS(1954),
    [anon_sym_default] = ACTIONS(1954),
    [anon_sym_enum] = ACTIONS(1954),
    [anon_sym_fn] = ACTIONS(1954),
    [anon_sym_for] = ACTIONS(1954),
    [anon_sym_if] = ACTIONS(1954),
    [anon_sym_impl] = ACTIONS(1954),
    [anon_sym_let] = ACTIONS(1954),
    [anon_sym_loop] = ACTIONS(1954),
    [anon_sym_match] = ACTIONS(1954),
    [anon_sym_mod] = ACTIONS(1954),
    [anon_sym_pub] = ACTIONS(1954),
    [anon_sym_return] = ACTIONS(1954),
    [anon_sym_static] = ACTIONS(1954),
    [anon_sym_struct] = ACTIONS(1954),
    [anon_sym_trait] = ACTIONS(1954),
    [anon_sym_type] = ACTIONS(1954),
    [anon_sym_union] = ACTIONS(1954),
    [anon_sym_unsafe] = ACTIONS(1954),
    [anon_sym_use] = ACTIONS(1954),
    [anon_sym_while] = ACTIONS(1954),
    [anon_sym_POUND] = ACTIONS(1952),
    [anon_sym_BANG] = ACTIONS(1952),
    [anon_sym_extern] = ACTIONS(1954),
    [anon_sym_LT] = ACTIONS(1952),
    [anon_sym_COLON_COLON] = ACTIONS(1952),
    [anon_sym_AMP] = ACTIONS(1952),
    [anon_sym_DOT_DOT] = ACTIONS(1952),
    [anon_sym_DASH] = ACTIONS(1952),
    [anon_sym_PIPE] = ACTIONS(1952),
    [anon_sym_yield] = ACTIONS(1954),
    [anon_sym_move] = ACTIONS(1954),
    [sym_integer_literal] = ACTIONS(1952),
    [aux_sym_string_literal_token1] = ACTIONS(1952),
    [sym_char_literal] = ACTIONS(1952),
    [anon_sym_true] = ACTIONS(1954),
    [anon_sym_false] = ACTIONS(1954),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1954),
    [sym_super] = ACTIONS(1954),
    [sym_crate] = ACTIONS(1954),
    [sym_metavariable] = ACTIONS(1952),
    [sym_raw_string_literal] = ACTIONS(1952),
    [sym_float_literal] = ACTIONS(1952),
    [sym_block_comment] = ACTIONS(3),
  },
  [469] = {
    [sym_attribute_item] = STATE(547),
    [sym_bracketed_type] = STATE(2489),
    [sym_generic_type] = STATE(2488),
    [sym_generic_type_with_turbofish] = STATE(2487),
    [sym_macro_invocation] = STATE(1422),
    [sym_scoped_identifier] = STATE(1337),
    [sym_scoped_type_identifier] = STATE(2013),
    [sym_match_arm] = STATE(486),
    [sym_last_match_arm] = STATE(2454),
    [sym_match_pattern] = STATE(2387),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1981),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [aux_sym_enum_variant_list_repeat1] = STATE(547),
    [aux_sym_match_block_repeat1] = STATE(486),
    [sym_identifier] = ACTIONS(1524),
    [anon_sym_LPAREN] = ACTIONS(1526),
    [anon_sym_RBRACE] = ACTIONS(1956),
    [anon_sym_LBRACK] = ACTIONS(1530),
    [anon_sym_u8] = ACTIONS(1532),
    [anon_sym_i8] = ACTIONS(1532),
    [anon_sym_u16] = ACTIONS(1532),
    [anon_sym_i16] = ACTIONS(1532),
    [anon_sym_u32] = ACTIONS(1532),
    [anon_sym_i32] = ACTIONS(1532),
    [anon_sym_u64] = ACTIONS(1532),
    [anon_sym_i64] = ACTIONS(1532),
    [anon_sym_u128] = ACTIONS(1532),
    [anon_sym_i128] = ACTIONS(1532),
    [anon_sym_isize] = ACTIONS(1532),
    [anon_sym_usize] = ACTIONS(1532),
    [anon_sym_f32] = ACTIONS(1532),
    [anon_sym_f64] = ACTIONS(1532),
    [anon_sym_bool] = ACTIONS(1532),
    [anon_sym_str] = ACTIONS(1532),
    [anon_sym_char] = ACTIONS(1532),
    [anon_sym_const] = ACTIONS(1534),
    [anon_sym_default] = ACTIONS(1536),
    [anon_sym_union] = ACTIONS(1536),
    [anon_sym_POUND] = ACTIONS(370),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(1538),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(1540),
    [sym_mutable_specifier] = ACTIONS(796),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1542),
    [sym_super] = ACTIONS(1542),
    [sym_crate] = ACTIONS(1542),
    [sym_metavariable] = ACTIONS(1544),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [470] = {
    [ts_builtin_sym_end] = ACTIONS(1958),
    [sym_identifier] = ACTIONS(1960),
    [anon_sym_SEMI] = ACTIONS(1958),
    [anon_sym_macro_rules_BANG] = ACTIONS(1958),
    [anon_sym_LPAREN] = ACTIONS(1958),
    [anon_sym_LBRACE] = ACTIONS(1958),
    [anon_sym_RBRACE] = ACTIONS(1958),
    [anon_sym_LBRACK] = ACTIONS(1958),
    [anon_sym_STAR] = ACTIONS(1958),
    [anon_sym_u8] = ACTIONS(1960),
    [anon_sym_i8] = ACTIONS(1960),
    [anon_sym_u16] = ACTIONS(1960),
    [anon_sym_i16] = ACTIONS(1960),
    [anon_sym_u32] = ACTIONS(1960),
    [anon_sym_i32] = ACTIONS(1960),
    [anon_sym_u64] = ACTIONS(1960),
    [anon_sym_i64] = ACTIONS(1960),
    [anon_sym_u128] = ACTIONS(1960),
    [anon_sym_i128] = ACTIONS(1960),
    [anon_sym_isize] = ACTIONS(1960),
    [anon_sym_usize] = ACTIONS(1960),
    [anon_sym_f32] = ACTIONS(1960),
    [anon_sym_f64] = ACTIONS(1960),
    [anon_sym_bool] = ACTIONS(1960),
    [anon_sym_str] = ACTIONS(1960),
    [anon_sym_char] = ACTIONS(1960),
    [anon_sym_SQUOTE] = ACTIONS(1960),
    [anon_sym_async] = ACTIONS(1960),
    [anon_sym_break] = ACTIONS(1960),
    [anon_sym_const] = ACTIONS(1960),
    [anon_sym_continue] = ACTIONS(1960),
    [anon_sym_default] = ACTIONS(1960),
    [anon_sym_enum] = ACTIONS(1960),
    [anon_sym_fn] = ACTIONS(1960),
    [anon_sym_for] = ACTIONS(1960),
    [anon_sym_if] = ACTIONS(1960),
    [anon_sym_impl] = ACTIONS(1960),
    [anon_sym_let] = ACTIONS(1960),
    [anon_sym_loop] = ACTIONS(1960),
    [anon_sym_match] = ACTIONS(1960),
    [anon_sym_mod] = ACTIONS(1960),
    [anon_sym_pub] = ACTIONS(1960),
    [anon_sym_return] = ACTIONS(1960),
    [anon_sym_static] = ACTIONS(1960),
    [anon_sym_struct] = ACTIONS(1960),
    [anon_sym_trait] = ACTIONS(1960),
    [anon_sym_type] = ACTIONS(1960),
    [anon_sym_union] = ACTIONS(1960),
    [anon_sym_unsafe] = ACTIONS(1960),
    [anon_sym_use] = ACTIONS(1960),
    [anon_sym_while] = ACTIONS(1960),
    [anon_sym_POUND] = ACTIONS(1958),
    [anon_sym_BANG] = ACTIONS(1958),
    [anon_sym_extern] = ACTIONS(1960),
    [anon_sym_LT] = ACTIONS(1958),
    [anon_sym_COLON_COLON] = ACTIONS(1958),
    [anon_sym_AMP] = ACTIONS(1958),
    [anon_sym_DOT_DOT] = ACTIONS(1958),
    [anon_sym_DASH] = ACTIONS(1958),
    [anon_sym_PIPE] = ACTIONS(1958),
    [anon_sym_yield] = ACTIONS(1960),
    [anon_sym_move] = ACTIONS(1960),
    [sym_integer_literal] = ACTIONS(1958),
    [aux_sym_string_literal_token1] = ACTIONS(1958),
    [sym_char_literal] = ACTIONS(1958),
    [anon_sym_true] = ACTIONS(1960),
    [anon_sym_false] = ACTIONS(1960),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1960),
    [sym_super] = ACTIONS(1960),
    [sym_crate] = ACTIONS(1960),
    [sym_metavariable] = ACTIONS(1958),
    [sym_raw_string_literal] = ACTIONS(1958),
    [sym_float_literal] = ACTIONS(1958),
    [sym_block_comment] = ACTIONS(3),
  },
  [471] = {
    [ts_builtin_sym_end] = ACTIONS(1962),
    [sym_identifier] = ACTIONS(1964),
    [anon_sym_SEMI] = ACTIONS(1962),
    [anon_sym_macro_rules_BANG] = ACTIONS(1962),
    [anon_sym_LPAREN] = ACTIONS(1962),
    [anon_sym_LBRACE] = ACTIONS(1962),
    [anon_sym_RBRACE] = ACTIONS(1962),
    [anon_sym_LBRACK] = ACTIONS(1962),
    [anon_sym_STAR] = ACTIONS(1962),
    [anon_sym_u8] = ACTIONS(1964),
    [anon_sym_i8] = ACTIONS(1964),
    [anon_sym_u16] = ACTIONS(1964),
    [anon_sym_i16] = ACTIONS(1964),
    [anon_sym_u32] = ACTIONS(1964),
    [anon_sym_i32] = ACTIONS(1964),
    [anon_sym_u64] = ACTIONS(1964),
    [anon_sym_i64] = ACTIONS(1964),
    [anon_sym_u128] = ACTIONS(1964),
    [anon_sym_i128] = ACTIONS(1964),
    [anon_sym_isize] = ACTIONS(1964),
    [anon_sym_usize] = ACTIONS(1964),
    [anon_sym_f32] = ACTIONS(1964),
    [anon_sym_f64] = ACTIONS(1964),
    [anon_sym_bool] = ACTIONS(1964),
    [anon_sym_str] = ACTIONS(1964),
    [anon_sym_char] = ACTIONS(1964),
    [anon_sym_SQUOTE] = ACTIONS(1964),
    [anon_sym_async] = ACTIONS(1964),
    [anon_sym_break] = ACTIONS(1964),
    [anon_sym_const] = ACTIONS(1964),
    [anon_sym_continue] = ACTIONS(1964),
    [anon_sym_default] = ACTIONS(1964),
    [anon_sym_enum] = ACTIONS(1964),
    [anon_sym_fn] = ACTIONS(1964),
    [anon_sym_for] = ACTIONS(1964),
    [anon_sym_if] = ACTIONS(1964),
    [anon_sym_impl] = ACTIONS(1964),
    [anon_sym_let] = ACTIONS(1964),
    [anon_sym_loop] = ACTIONS(1964),
    [anon_sym_match] = ACTIONS(1964),
    [anon_sym_mod] = ACTIONS(1964),
    [anon_sym_pub] = ACTIONS(1964),
    [anon_sym_return] = ACTIONS(1964),
    [anon_sym_static] = ACTIONS(1964),
    [anon_sym_struct] = ACTIONS(1964),
    [anon_sym_trait] = ACTIONS(1964),
    [anon_sym_type] = ACTIONS(1964),
    [anon_sym_union] = ACTIONS(1964),
    [anon_sym_unsafe] = ACTIONS(1964),
    [anon_sym_use] = ACTIONS(1964),
    [anon_sym_while] = ACTIONS(1964),
    [anon_sym_POUND] = ACTIONS(1962),
    [anon_sym_BANG] = ACTIONS(1962),
    [anon_sym_extern] = ACTIONS(1964),
    [anon_sym_LT] = ACTIONS(1962),
    [anon_sym_COLON_COLON] = ACTIONS(1962),
    [anon_sym_AMP] = ACTIONS(1962),
    [anon_sym_DOT_DOT] = ACTIONS(1962),
    [anon_sym_DASH] = ACTIONS(1962),
    [anon_sym_PIPE] = ACTIONS(1962),
    [anon_sym_yield] = ACTIONS(1964),
    [anon_sym_move] = ACTIONS(1964),
    [sym_integer_literal] = ACTIONS(1962),
    [aux_sym_string_literal_token1] = ACTIONS(1962),
    [sym_char_literal] = ACTIONS(1962),
    [anon_sym_true] = ACTIONS(1964),
    [anon_sym_false] = ACTIONS(1964),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1964),
    [sym_super] = ACTIONS(1964),
    [sym_crate] = ACTIONS(1964),
    [sym_metavariable] = ACTIONS(1962),
    [sym_raw_string_literal] = ACTIONS(1962),
    [sym_float_literal] = ACTIONS(1962),
    [sym_block_comment] = ACTIONS(3),
  },
  [472] = {
    [ts_builtin_sym_end] = ACTIONS(1966),
    [sym_identifier] = ACTIONS(1968),
    [anon_sym_SEMI] = ACTIONS(1966),
    [anon_sym_macro_rules_BANG] = ACTIONS(1966),
    [anon_sym_LPAREN] = ACTIONS(1966),
    [anon_sym_LBRACE] = ACTIONS(1966),
    [anon_sym_RBRACE] = ACTIONS(1966),
    [anon_sym_LBRACK] = ACTIONS(1966),
    [anon_sym_STAR] = ACTIONS(1966),
    [anon_sym_u8] = ACTIONS(1968),
    [anon_sym_i8] = ACTIONS(1968),
    [anon_sym_u16] = ACTIONS(1968),
    [anon_sym_i16] = ACTIONS(1968),
    [anon_sym_u32] = ACTIONS(1968),
    [anon_sym_i32] = ACTIONS(1968),
    [anon_sym_u64] = ACTIONS(1968),
    [anon_sym_i64] = ACTIONS(1968),
    [anon_sym_u128] = ACTIONS(1968),
    [anon_sym_i128] = ACTIONS(1968),
    [anon_sym_isize] = ACTIONS(1968),
    [anon_sym_usize] = ACTIONS(1968),
    [anon_sym_f32] = ACTIONS(1968),
    [anon_sym_f64] = ACTIONS(1968),
    [anon_sym_bool] = ACTIONS(1968),
    [anon_sym_str] = ACTIONS(1968),
    [anon_sym_char] = ACTIONS(1968),
    [anon_sym_SQUOTE] = ACTIONS(1968),
    [anon_sym_async] = ACTIONS(1968),
    [anon_sym_break] = ACTIONS(1968),
    [anon_sym_const] = ACTIONS(1968),
    [anon_sym_continue] = ACTIONS(1968),
    [anon_sym_default] = ACTIONS(1968),
    [anon_sym_enum] = ACTIONS(1968),
    [anon_sym_fn] = ACTIONS(1968),
    [anon_sym_for] = ACTIONS(1968),
    [anon_sym_if] = ACTIONS(1968),
    [anon_sym_impl] = ACTIONS(1968),
    [anon_sym_let] = ACTIONS(1968),
    [anon_sym_loop] = ACTIONS(1968),
    [anon_sym_match] = ACTIONS(1968),
    [anon_sym_mod] = ACTIONS(1968),
    [anon_sym_pub] = ACTIONS(1968),
    [anon_sym_return] = ACTIONS(1968),
    [anon_sym_static] = ACTIONS(1968),
    [anon_sym_struct] = ACTIONS(1968),
    [anon_sym_trait] = ACTIONS(1968),
    [anon_sym_type] = ACTIONS(1968),
    [anon_sym_union] = ACTIONS(1968),
    [anon_sym_unsafe] = ACTIONS(1968),
    [anon_sym_use] = ACTIONS(1968),
    [anon_sym_while] = ACTIONS(1968),
    [anon_sym_POUND] = ACTIONS(1966),
    [anon_sym_BANG] = ACTIONS(1966),
    [anon_sym_extern] = ACTIONS(1968),
    [anon_sym_LT] = ACTIONS(1966),
    [anon_sym_COLON_COLON] = ACTIONS(1966),
    [anon_sym_AMP] = ACTIONS(1966),
    [anon_sym_DOT_DOT] = ACTIONS(1966),
    [anon_sym_DASH] = ACTIONS(1966),
    [anon_sym_PIPE] = ACTIONS(1966),
    [anon_sym_yield] = ACTIONS(1968),
    [anon_sym_move] = ACTIONS(1968),
    [sym_integer_literal] = ACTIONS(1966),
    [aux_sym_string_literal_token1] = ACTIONS(1966),
    [sym_char_literal] = ACTIONS(1966),
    [anon_sym_true] = ACTIONS(1968),
    [anon_sym_false] = ACTIONS(1968),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1968),
    [sym_super] = ACTIONS(1968),
    [sym_crate] = ACTIONS(1968),
    [sym_metavariable] = ACTIONS(1966),
    [sym_raw_string_literal] = ACTIONS(1966),
    [sym_float_literal] = ACTIONS(1966),
    [sym_block_comment] = ACTIONS(3),
  },
  [473] = {
    [ts_builtin_sym_end] = ACTIONS(1970),
    [sym_identifier] = ACTIONS(1972),
    [anon_sym_SEMI] = ACTIONS(1970),
    [anon_sym_macro_rules_BANG] = ACTIONS(1970),
    [anon_sym_LPAREN] = ACTIONS(1970),
    [anon_sym_LBRACE] = ACTIONS(1970),
    [anon_sym_RBRACE] = ACTIONS(1970),
    [anon_sym_LBRACK] = ACTIONS(1970),
    [anon_sym_STAR] = ACTIONS(1970),
    [anon_sym_u8] = ACTIONS(1972),
    [anon_sym_i8] = ACTIONS(1972),
    [anon_sym_u16] = ACTIONS(1972),
    [anon_sym_i16] = ACTIONS(1972),
    [anon_sym_u32] = ACTIONS(1972),
    [anon_sym_i32] = ACTIONS(1972),
    [anon_sym_u64] = ACTIONS(1972),
    [anon_sym_i64] = ACTIONS(1972),
    [anon_sym_u128] = ACTIONS(1972),
    [anon_sym_i128] = ACTIONS(1972),
    [anon_sym_isize] = ACTIONS(1972),
    [anon_sym_usize] = ACTIONS(1972),
    [anon_sym_f32] = ACTIONS(1972),
    [anon_sym_f64] = ACTIONS(1972),
    [anon_sym_bool] = ACTIONS(1972),
    [anon_sym_str] = ACTIONS(1972),
    [anon_sym_char] = ACTIONS(1972),
    [anon_sym_SQUOTE] = ACTIONS(1972),
    [anon_sym_async] = ACTIONS(1972),
    [anon_sym_break] = ACTIONS(1972),
    [anon_sym_const] = ACTIONS(1972),
    [anon_sym_continue] = ACTIONS(1972),
    [anon_sym_default] = ACTIONS(1972),
    [anon_sym_enum] = ACTIONS(1972),
    [anon_sym_fn] = ACTIONS(1972),
    [anon_sym_for] = ACTIONS(1972),
    [anon_sym_if] = ACTIONS(1972),
    [anon_sym_impl] = ACTIONS(1972),
    [anon_sym_let] = ACTIONS(1972),
    [anon_sym_loop] = ACTIONS(1972),
    [anon_sym_match] = ACTIONS(1972),
    [anon_sym_mod] = ACTIONS(1972),
    [anon_sym_pub] = ACTIONS(1972),
    [anon_sym_return] = ACTIONS(1972),
    [anon_sym_static] = ACTIONS(1972),
    [anon_sym_struct] = ACTIONS(1972),
    [anon_sym_trait] = ACTIONS(1972),
    [anon_sym_type] = ACTIONS(1972),
    [anon_sym_union] = ACTIONS(1972),
    [anon_sym_unsafe] = ACTIONS(1972),
    [anon_sym_use] = ACTIONS(1972),
    [anon_sym_while] = ACTIONS(1972),
    [anon_sym_POUND] = ACTIONS(1970),
    [anon_sym_BANG] = ACTIONS(1970),
    [anon_sym_extern] = ACTIONS(1972),
    [anon_sym_LT] = ACTIONS(1970),
    [anon_sym_COLON_COLON] = ACTIONS(1970),
    [anon_sym_AMP] = ACTIONS(1970),
    [anon_sym_DOT_DOT] = ACTIONS(1970),
    [anon_sym_DASH] = ACTIONS(1970),
    [anon_sym_PIPE] = ACTIONS(1970),
    [anon_sym_yield] = ACTIONS(1972),
    [anon_sym_move] = ACTIONS(1972),
    [sym_integer_literal] = ACTIONS(1970),
    [aux_sym_string_literal_token1] = ACTIONS(1970),
    [sym_char_literal] = ACTIONS(1970),
    [anon_sym_true] = ACTIONS(1972),
    [anon_sym_false] = ACTIONS(1972),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1972),
    [sym_super] = ACTIONS(1972),
    [sym_crate] = ACTIONS(1972),
    [sym_metavariable] = ACTIONS(1970),
    [sym_raw_string_literal] = ACTIONS(1970),
    [sym_float_literal] = ACTIONS(1970),
    [sym_block_comment] = ACTIONS(3),
  },
  [474] = {
    [ts_builtin_sym_end] = ACTIONS(1974),
    [sym_identifier] = ACTIONS(1976),
    [anon_sym_SEMI] = ACTIONS(1974),
    [anon_sym_macro_rules_BANG] = ACTIONS(1974),
    [anon_sym_LPAREN] = ACTIONS(1974),
    [anon_sym_LBRACE] = ACTIONS(1974),
    [anon_sym_RBRACE] = ACTIONS(1974),
    [anon_sym_LBRACK] = ACTIONS(1974),
    [anon_sym_STAR] = ACTIONS(1974),
    [anon_sym_u8] = ACTIONS(1976),
    [anon_sym_i8] = ACTIONS(1976),
    [anon_sym_u16] = ACTIONS(1976),
    [anon_sym_i16] = ACTIONS(1976),
    [anon_sym_u32] = ACTIONS(1976),
    [anon_sym_i32] = ACTIONS(1976),
    [anon_sym_u64] = ACTIONS(1976),
    [anon_sym_i64] = ACTIONS(1976),
    [anon_sym_u128] = ACTIONS(1976),
    [anon_sym_i128] = ACTIONS(1976),
    [anon_sym_isize] = ACTIONS(1976),
    [anon_sym_usize] = ACTIONS(1976),
    [anon_sym_f32] = ACTIONS(1976),
    [anon_sym_f64] = ACTIONS(1976),
    [anon_sym_bool] = ACTIONS(1976),
    [anon_sym_str] = ACTIONS(1976),
    [anon_sym_char] = ACTIONS(1976),
    [anon_sym_SQUOTE] = ACTIONS(1976),
    [anon_sym_async] = ACTIONS(1976),
    [anon_sym_break] = ACTIONS(1976),
    [anon_sym_const] = ACTIONS(1976),
    [anon_sym_continue] = ACTIONS(1976),
    [anon_sym_default] = ACTIONS(1976),
    [anon_sym_enum] = ACTIONS(1976),
    [anon_sym_fn] = ACTIONS(1976),
    [anon_sym_for] = ACTIONS(1976),
    [anon_sym_if] = ACTIONS(1976),
    [anon_sym_impl] = ACTIONS(1976),
    [anon_sym_let] = ACTIONS(1976),
    [anon_sym_loop] = ACTIONS(1976),
    [anon_sym_match] = ACTIONS(1976),
    [anon_sym_mod] = ACTIONS(1976),
    [anon_sym_pub] = ACTIONS(1976),
    [anon_sym_return] = ACTIONS(1976),
    [anon_sym_static] = ACTIONS(1976),
    [anon_sym_struct] = ACTIONS(1976),
    [anon_sym_trait] = ACTIONS(1976),
    [anon_sym_type] = ACTIONS(1976),
    [anon_sym_union] = ACTIONS(1976),
    [anon_sym_unsafe] = ACTIONS(1976),
    [anon_sym_use] = ACTIONS(1976),
    [anon_sym_while] = ACTIONS(1976),
    [anon_sym_POUND] = ACTIONS(1974),
    [anon_sym_BANG] = ACTIONS(1974),
    [anon_sym_extern] = ACTIONS(1976),
    [anon_sym_LT] = ACTIONS(1974),
    [anon_sym_COLON_COLON] = ACTIONS(1974),
    [anon_sym_AMP] = ACTIONS(1974),
    [anon_sym_DOT_DOT] = ACTIONS(1974),
    [anon_sym_DASH] = ACTIONS(1974),
    [anon_sym_PIPE] = ACTIONS(1974),
    [anon_sym_yield] = ACTIONS(1976),
    [anon_sym_move] = ACTIONS(1976),
    [sym_integer_literal] = ACTIONS(1974),
    [aux_sym_string_literal_token1] = ACTIONS(1974),
    [sym_char_literal] = ACTIONS(1974),
    [anon_sym_true] = ACTIONS(1976),
    [anon_sym_false] = ACTIONS(1976),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1976),
    [sym_super] = ACTIONS(1976),
    [sym_crate] = ACTIONS(1976),
    [sym_metavariable] = ACTIONS(1974),
    [sym_raw_string_literal] = ACTIONS(1974),
    [sym_float_literal] = ACTIONS(1974),
    [sym_block_comment] = ACTIONS(3),
  },
  [475] = {
    [ts_builtin_sym_end] = ACTIONS(1978),
    [sym_identifier] = ACTIONS(1980),
    [anon_sym_SEMI] = ACTIONS(1978),
    [anon_sym_macro_rules_BANG] = ACTIONS(1978),
    [anon_sym_LPAREN] = ACTIONS(1978),
    [anon_sym_LBRACE] = ACTIONS(1978),
    [anon_sym_RBRACE] = ACTIONS(1978),
    [anon_sym_LBRACK] = ACTIONS(1978),
    [anon_sym_STAR] = ACTIONS(1978),
    [anon_sym_u8] = ACTIONS(1980),
    [anon_sym_i8] = ACTIONS(1980),
    [anon_sym_u16] = ACTIONS(1980),
    [anon_sym_i16] = ACTIONS(1980),
    [anon_sym_u32] = ACTIONS(1980),
    [anon_sym_i32] = ACTIONS(1980),
    [anon_sym_u64] = ACTIONS(1980),
    [anon_sym_i64] = ACTIONS(1980),
    [anon_sym_u128] = ACTIONS(1980),
    [anon_sym_i128] = ACTIONS(1980),
    [anon_sym_isize] = ACTIONS(1980),
    [anon_sym_usize] = ACTIONS(1980),
    [anon_sym_f32] = ACTIONS(1980),
    [anon_sym_f64] = ACTIONS(1980),
    [anon_sym_bool] = ACTIONS(1980),
    [anon_sym_str] = ACTIONS(1980),
    [anon_sym_char] = ACTIONS(1980),
    [anon_sym_SQUOTE] = ACTIONS(1980),
    [anon_sym_async] = ACTIONS(1980),
    [anon_sym_break] = ACTIONS(1980),
    [anon_sym_const] = ACTIONS(1980),
    [anon_sym_continue] = ACTIONS(1980),
    [anon_sym_default] = ACTIONS(1980),
    [anon_sym_enum] = ACTIONS(1980),
    [anon_sym_fn] = ACTIONS(1980),
    [anon_sym_for] = ACTIONS(1980),
    [anon_sym_if] = ACTIONS(1980),
    [anon_sym_impl] = ACTIONS(1980),
    [anon_sym_let] = ACTIONS(1980),
    [anon_sym_loop] = ACTIONS(1980),
    [anon_sym_match] = ACTIONS(1980),
    [anon_sym_mod] = ACTIONS(1980),
    [anon_sym_pub] = ACTIONS(1980),
    [anon_sym_return] = ACTIONS(1980),
    [anon_sym_static] = ACTIONS(1980),
    [anon_sym_struct] = ACTIONS(1980),
    [anon_sym_trait] = ACTIONS(1980),
    [anon_sym_type] = ACTIONS(1980),
    [anon_sym_union] = ACTIONS(1980),
    [anon_sym_unsafe] = ACTIONS(1980),
    [anon_sym_use] = ACTIONS(1980),
    [anon_sym_while] = ACTIONS(1980),
    [anon_sym_POUND] = ACTIONS(1978),
    [anon_sym_BANG] = ACTIONS(1978),
    [anon_sym_extern] = ACTIONS(1980),
    [anon_sym_LT] = ACTIONS(1978),
    [anon_sym_COLON_COLON] = ACTIONS(1978),
    [anon_sym_AMP] = ACTIONS(1978),
    [anon_sym_DOT_DOT] = ACTIONS(1978),
    [anon_sym_DASH] = ACTIONS(1978),
    [anon_sym_PIPE] = ACTIONS(1978),
    [anon_sym_yield] = ACTIONS(1980),
    [anon_sym_move] = ACTIONS(1980),
    [sym_integer_literal] = ACTIONS(1978),
    [aux_sym_string_literal_token1] = ACTIONS(1978),
    [sym_char_literal] = ACTIONS(1978),
    [anon_sym_true] = ACTIONS(1980),
    [anon_sym_false] = ACTIONS(1980),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1980),
    [sym_super] = ACTIONS(1980),
    [sym_crate] = ACTIONS(1980),
    [sym_metavariable] = ACTIONS(1978),
    [sym_raw_string_literal] = ACTIONS(1978),
    [sym_float_literal] = ACTIONS(1978),
    [sym_block_comment] = ACTIONS(3),
  },
  [476] = {
    [ts_builtin_sym_end] = ACTIONS(1982),
    [sym_identifier] = ACTIONS(1984),
    [anon_sym_SEMI] = ACTIONS(1982),
    [anon_sym_macro_rules_BANG] = ACTIONS(1982),
    [anon_sym_LPAREN] = ACTIONS(1982),
    [anon_sym_LBRACE] = ACTIONS(1982),
    [anon_sym_RBRACE] = ACTIONS(1982),
    [anon_sym_LBRACK] = ACTIONS(1982),
    [anon_sym_STAR] = ACTIONS(1982),
    [anon_sym_u8] = ACTIONS(1984),
    [anon_sym_i8] = ACTIONS(1984),
    [anon_sym_u16] = ACTIONS(1984),
    [anon_sym_i16] = ACTIONS(1984),
    [anon_sym_u32] = ACTIONS(1984),
    [anon_sym_i32] = ACTIONS(1984),
    [anon_sym_u64] = ACTIONS(1984),
    [anon_sym_i64] = ACTIONS(1984),
    [anon_sym_u128] = ACTIONS(1984),
    [anon_sym_i128] = ACTIONS(1984),
    [anon_sym_isize] = ACTIONS(1984),
    [anon_sym_usize] = ACTIONS(1984),
    [anon_sym_f32] = ACTIONS(1984),
    [anon_sym_f64] = ACTIONS(1984),
    [anon_sym_bool] = ACTIONS(1984),
    [anon_sym_str] = ACTIONS(1984),
    [anon_sym_char] = ACTIONS(1984),
    [anon_sym_SQUOTE] = ACTIONS(1984),
    [anon_sym_async] = ACTIONS(1984),
    [anon_sym_break] = ACTIONS(1984),
    [anon_sym_const] = ACTIONS(1984),
    [anon_sym_continue] = ACTIONS(1984),
    [anon_sym_default] = ACTIONS(1984),
    [anon_sym_enum] = ACTIONS(1984),
    [anon_sym_fn] = ACTIONS(1984),
    [anon_sym_for] = ACTIONS(1984),
    [anon_sym_if] = ACTIONS(1984),
    [anon_sym_impl] = ACTIONS(1984),
    [anon_sym_let] = ACTIONS(1984),
    [anon_sym_loop] = ACTIONS(1984),
    [anon_sym_match] = ACTIONS(1984),
    [anon_sym_mod] = ACTIONS(1984),
    [anon_sym_pub] = ACTIONS(1984),
    [anon_sym_return] = ACTIONS(1984),
    [anon_sym_static] = ACTIONS(1984),
    [anon_sym_struct] = ACTIONS(1984),
    [anon_sym_trait] = ACTIONS(1984),
    [anon_sym_type] = ACTIONS(1984),
    [anon_sym_union] = ACTIONS(1984),
    [anon_sym_unsafe] = ACTIONS(1984),
    [anon_sym_use] = ACTIONS(1984),
    [anon_sym_while] = ACTIONS(1984),
    [anon_sym_POUND] = ACTIONS(1982),
    [anon_sym_BANG] = ACTIONS(1982),
    [anon_sym_extern] = ACTIONS(1984),
    [anon_sym_LT] = ACTIONS(1982),
    [anon_sym_COLON_COLON] = ACTIONS(1982),
    [anon_sym_AMP] = ACTIONS(1982),
    [anon_sym_DOT_DOT] = ACTIONS(1982),
    [anon_sym_DASH] = ACTIONS(1982),
    [anon_sym_PIPE] = ACTIONS(1982),
    [anon_sym_yield] = ACTIONS(1984),
    [anon_sym_move] = ACTIONS(1984),
    [sym_integer_literal] = ACTIONS(1982),
    [aux_sym_string_literal_token1] = ACTIONS(1982),
    [sym_char_literal] = ACTIONS(1982),
    [anon_sym_true] = ACTIONS(1984),
    [anon_sym_false] = ACTIONS(1984),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1984),
    [sym_super] = ACTIONS(1984),
    [sym_crate] = ACTIONS(1984),
    [sym_metavariable] = ACTIONS(1982),
    [sym_raw_string_literal] = ACTIONS(1982),
    [sym_float_literal] = ACTIONS(1982),
    [sym_block_comment] = ACTIONS(3),
  },
  [477] = {
    [ts_builtin_sym_end] = ACTIONS(1986),
    [sym_identifier] = ACTIONS(1988),
    [anon_sym_SEMI] = ACTIONS(1986),
    [anon_sym_macro_rules_BANG] = ACTIONS(1986),
    [anon_sym_LPAREN] = ACTIONS(1986),
    [anon_sym_LBRACE] = ACTIONS(1986),
    [anon_sym_RBRACE] = ACTIONS(1986),
    [anon_sym_LBRACK] = ACTIONS(1986),
    [anon_sym_STAR] = ACTIONS(1986),
    [anon_sym_u8] = ACTIONS(1988),
    [anon_sym_i8] = ACTIONS(1988),
    [anon_sym_u16] = ACTIONS(1988),
    [anon_sym_i16] = ACTIONS(1988),
    [anon_sym_u32] = ACTIONS(1988),
    [anon_sym_i32] = ACTIONS(1988),
    [anon_sym_u64] = ACTIONS(1988),
    [anon_sym_i64] = ACTIONS(1988),
    [anon_sym_u128] = ACTIONS(1988),
    [anon_sym_i128] = ACTIONS(1988),
    [anon_sym_isize] = ACTIONS(1988),
    [anon_sym_usize] = ACTIONS(1988),
    [anon_sym_f32] = ACTIONS(1988),
    [anon_sym_f64] = ACTIONS(1988),
    [anon_sym_bool] = ACTIONS(1988),
    [anon_sym_str] = ACTIONS(1988),
    [anon_sym_char] = ACTIONS(1988),
    [anon_sym_SQUOTE] = ACTIONS(1988),
    [anon_sym_async] = ACTIONS(1988),
    [anon_sym_break] = ACTIONS(1988),
    [anon_sym_const] = ACTIONS(1988),
    [anon_sym_continue] = ACTIONS(1988),
    [anon_sym_default] = ACTIONS(1988),
    [anon_sym_enum] = ACTIONS(1988),
    [anon_sym_fn] = ACTIONS(1988),
    [anon_sym_for] = ACTIONS(1988),
    [anon_sym_if] = ACTIONS(1988),
    [anon_sym_impl] = ACTIONS(1988),
    [anon_sym_let] = ACTIONS(1988),
    [anon_sym_loop] = ACTIONS(1988),
    [anon_sym_match] = ACTIONS(1988),
    [anon_sym_mod] = ACTIONS(1988),
    [anon_sym_pub] = ACTIONS(1988),
    [anon_sym_return] = ACTIONS(1988),
    [anon_sym_static] = ACTIONS(1988),
    [anon_sym_struct] = ACTIONS(1988),
    [anon_sym_trait] = ACTIONS(1988),
    [anon_sym_type] = ACTIONS(1988),
    [anon_sym_union] = ACTIONS(1988),
    [anon_sym_unsafe] = ACTIONS(1988),
    [anon_sym_use] = ACTIONS(1988),
    [anon_sym_while] = ACTIONS(1988),
    [anon_sym_POUND] = ACTIONS(1986),
    [anon_sym_BANG] = ACTIONS(1986),
    [anon_sym_extern] = ACTIONS(1988),
    [anon_sym_LT] = ACTIONS(1986),
    [anon_sym_COLON_COLON] = ACTIONS(1986),
    [anon_sym_AMP] = ACTIONS(1986),
    [anon_sym_DOT_DOT] = ACTIONS(1986),
    [anon_sym_DASH] = ACTIONS(1986),
    [anon_sym_PIPE] = ACTIONS(1986),
    [anon_sym_yield] = ACTIONS(1988),
    [anon_sym_move] = ACTIONS(1988),
    [sym_integer_literal] = ACTIONS(1986),
    [aux_sym_string_literal_token1] = ACTIONS(1986),
    [sym_char_literal] = ACTIONS(1986),
    [anon_sym_true] = ACTIONS(1988),
    [anon_sym_false] = ACTIONS(1988),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1988),
    [sym_super] = ACTIONS(1988),
    [sym_crate] = ACTIONS(1988),
    [sym_metavariable] = ACTIONS(1986),
    [sym_raw_string_literal] = ACTIONS(1986),
    [sym_float_literal] = ACTIONS(1986),
    [sym_block_comment] = ACTIONS(3),
  },
  [478] = {
    [ts_builtin_sym_end] = ACTIONS(1990),
    [sym_identifier] = ACTIONS(1992),
    [anon_sym_SEMI] = ACTIONS(1990),
    [anon_sym_macro_rules_BANG] = ACTIONS(1990),
    [anon_sym_LPAREN] = ACTIONS(1990),
    [anon_sym_LBRACE] = ACTIONS(1990),
    [anon_sym_RBRACE] = ACTIONS(1990),
    [anon_sym_LBRACK] = ACTIONS(1990),
    [anon_sym_STAR] = ACTIONS(1990),
    [anon_sym_u8] = ACTIONS(1992),
    [anon_sym_i8] = ACTIONS(1992),
    [anon_sym_u16] = ACTIONS(1992),
    [anon_sym_i16] = ACTIONS(1992),
    [anon_sym_u32] = ACTIONS(1992),
    [anon_sym_i32] = ACTIONS(1992),
    [anon_sym_u64] = ACTIONS(1992),
    [anon_sym_i64] = ACTIONS(1992),
    [anon_sym_u128] = ACTIONS(1992),
    [anon_sym_i128] = ACTIONS(1992),
    [anon_sym_isize] = ACTIONS(1992),
    [anon_sym_usize] = ACTIONS(1992),
    [anon_sym_f32] = ACTIONS(1992),
    [anon_sym_f64] = ACTIONS(1992),
    [anon_sym_bool] = ACTIONS(1992),
    [anon_sym_str] = ACTIONS(1992),
    [anon_sym_char] = ACTIONS(1992),
    [anon_sym_SQUOTE] = ACTIONS(1992),
    [anon_sym_async] = ACTIONS(1992),
    [anon_sym_break] = ACTIONS(1992),
    [anon_sym_const] = ACTIONS(1992),
    [anon_sym_continue] = ACTIONS(1992),
    [anon_sym_default] = ACTIONS(1992),
    [anon_sym_enum] = ACTIONS(1992),
    [anon_sym_fn] = ACTIONS(1992),
    [anon_sym_for] = ACTIONS(1992),
    [anon_sym_if] = ACTIONS(1992),
    [anon_sym_impl] = ACTIONS(1992),
    [anon_sym_let] = ACTIONS(1992),
    [anon_sym_loop] = ACTIONS(1992),
    [anon_sym_match] = ACTIONS(1992),
    [anon_sym_mod] = ACTIONS(1992),
    [anon_sym_pub] = ACTIONS(1992),
    [anon_sym_return] = ACTIONS(1992),
    [anon_sym_static] = ACTIONS(1992),
    [anon_sym_struct] = ACTIONS(1992),
    [anon_sym_trait] = ACTIONS(1992),
    [anon_sym_type] = ACTIONS(1992),
    [anon_sym_union] = ACTIONS(1992),
    [anon_sym_unsafe] = ACTIONS(1992),
    [anon_sym_use] = ACTIONS(1992),
    [anon_sym_while] = ACTIONS(1992),
    [anon_sym_POUND] = ACTIONS(1990),
    [anon_sym_BANG] = ACTIONS(1990),
    [anon_sym_extern] = ACTIONS(1992),
    [anon_sym_LT] = ACTIONS(1990),
    [anon_sym_COLON_COLON] = ACTIONS(1990),
    [anon_sym_AMP] = ACTIONS(1990),
    [anon_sym_DOT_DOT] = ACTIONS(1990),
    [anon_sym_DASH] = ACTIONS(1990),
    [anon_sym_PIPE] = ACTIONS(1990),
    [anon_sym_yield] = ACTIONS(1992),
    [anon_sym_move] = ACTIONS(1992),
    [sym_integer_literal] = ACTIONS(1990),
    [aux_sym_string_literal_token1] = ACTIONS(1990),
    [sym_char_literal] = ACTIONS(1990),
    [anon_sym_true] = ACTIONS(1992),
    [anon_sym_false] = ACTIONS(1992),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1992),
    [sym_super] = ACTIONS(1992),
    [sym_crate] = ACTIONS(1992),
    [sym_metavariable] = ACTIONS(1990),
    [sym_raw_string_literal] = ACTIONS(1990),
    [sym_float_literal] = ACTIONS(1990),
    [sym_block_comment] = ACTIONS(3),
  },
  [479] = {
    [ts_builtin_sym_end] = ACTIONS(1994),
    [sym_identifier] = ACTIONS(1996),
    [anon_sym_SEMI] = ACTIONS(1994),
    [anon_sym_macro_rules_BANG] = ACTIONS(1994),
    [anon_sym_LPAREN] = ACTIONS(1994),
    [anon_sym_LBRACE] = ACTIONS(1994),
    [anon_sym_RBRACE] = ACTIONS(1994),
    [anon_sym_LBRACK] = ACTIONS(1994),
    [anon_sym_STAR] = ACTIONS(1994),
    [anon_sym_u8] = ACTIONS(1996),
    [anon_sym_i8] = ACTIONS(1996),
    [anon_sym_u16] = ACTIONS(1996),
    [anon_sym_i16] = ACTIONS(1996),
    [anon_sym_u32] = ACTIONS(1996),
    [anon_sym_i32] = ACTIONS(1996),
    [anon_sym_u64] = ACTIONS(1996),
    [anon_sym_i64] = ACTIONS(1996),
    [anon_sym_u128] = ACTIONS(1996),
    [anon_sym_i128] = ACTIONS(1996),
    [anon_sym_isize] = ACTIONS(1996),
    [anon_sym_usize] = ACTIONS(1996),
    [anon_sym_f32] = ACTIONS(1996),
    [anon_sym_f64] = ACTIONS(1996),
    [anon_sym_bool] = ACTIONS(1996),
    [anon_sym_str] = ACTIONS(1996),
    [anon_sym_char] = ACTIONS(1996),
    [anon_sym_SQUOTE] = ACTIONS(1996),
    [anon_sym_async] = ACTIONS(1996),
    [anon_sym_break] = ACTIONS(1996),
    [anon_sym_const] = ACTIONS(1996),
    [anon_sym_continue] = ACTIONS(1996),
    [anon_sym_default] = ACTIONS(1996),
    [anon_sym_enum] = ACTIONS(1996),
    [anon_sym_fn] = ACTIONS(1996),
    [anon_sym_for] = ACTIONS(1996),
    [anon_sym_if] = ACTIONS(1996),
    [anon_sym_impl] = ACTIONS(1996),
    [anon_sym_let] = ACTIONS(1996),
    [anon_sym_loop] = ACTIONS(1996),
    [anon_sym_match] = ACTIONS(1996),
    [anon_sym_mod] = ACTIONS(1996),
    [anon_sym_pub] = ACTIONS(1996),
    [anon_sym_return] = ACTIONS(1996),
    [anon_sym_static] = ACTIONS(1996),
    [anon_sym_struct] = ACTIONS(1996),
    [anon_sym_trait] = ACTIONS(1996),
    [anon_sym_type] = ACTIONS(1996),
    [anon_sym_union] = ACTIONS(1996),
    [anon_sym_unsafe] = ACTIONS(1996),
    [anon_sym_use] = ACTIONS(1996),
    [anon_sym_while] = ACTIONS(1996),
    [anon_sym_POUND] = ACTIONS(1994),
    [anon_sym_BANG] = ACTIONS(1994),
    [anon_sym_extern] = ACTIONS(1996),
    [anon_sym_LT] = ACTIONS(1994),
    [anon_sym_COLON_COLON] = ACTIONS(1994),
    [anon_sym_AMP] = ACTIONS(1994),
    [anon_sym_DOT_DOT] = ACTIONS(1994),
    [anon_sym_DASH] = ACTIONS(1994),
    [anon_sym_PIPE] = ACTIONS(1994),
    [anon_sym_yield] = ACTIONS(1996),
    [anon_sym_move] = ACTIONS(1996),
    [sym_integer_literal] = ACTIONS(1994),
    [aux_sym_string_literal_token1] = ACTIONS(1994),
    [sym_char_literal] = ACTIONS(1994),
    [anon_sym_true] = ACTIONS(1996),
    [anon_sym_false] = ACTIONS(1996),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1996),
    [sym_super] = ACTIONS(1996),
    [sym_crate] = ACTIONS(1996),
    [sym_metavariable] = ACTIONS(1994),
    [sym_raw_string_literal] = ACTIONS(1994),
    [sym_float_literal] = ACTIONS(1994),
    [sym_block_comment] = ACTIONS(3),
  },
  [480] = {
    [ts_builtin_sym_end] = ACTIONS(1998),
    [sym_identifier] = ACTIONS(2000),
    [anon_sym_SEMI] = ACTIONS(1998),
    [anon_sym_macro_rules_BANG] = ACTIONS(1998),
    [anon_sym_LPAREN] = ACTIONS(1998),
    [anon_sym_LBRACE] = ACTIONS(1998),
    [anon_sym_RBRACE] = ACTIONS(1998),
    [anon_sym_LBRACK] = ACTIONS(1998),
    [anon_sym_STAR] = ACTIONS(1998),
    [anon_sym_u8] = ACTIONS(2000),
    [anon_sym_i8] = ACTIONS(2000),
    [anon_sym_u16] = ACTIONS(2000),
    [anon_sym_i16] = ACTIONS(2000),
    [anon_sym_u32] = ACTIONS(2000),
    [anon_sym_i32] = ACTIONS(2000),
    [anon_sym_u64] = ACTIONS(2000),
    [anon_sym_i64] = ACTIONS(2000),
    [anon_sym_u128] = ACTIONS(2000),
    [anon_sym_i128] = ACTIONS(2000),
    [anon_sym_isize] = ACTIONS(2000),
    [anon_sym_usize] = ACTIONS(2000),
    [anon_sym_f32] = ACTIONS(2000),
    [anon_sym_f64] = ACTIONS(2000),
    [anon_sym_bool] = ACTIONS(2000),
    [anon_sym_str] = ACTIONS(2000),
    [anon_sym_char] = ACTIONS(2000),
    [anon_sym_SQUOTE] = ACTIONS(2000),
    [anon_sym_async] = ACTIONS(2000),
    [anon_sym_break] = ACTIONS(2000),
    [anon_sym_const] = ACTIONS(2000),
    [anon_sym_continue] = ACTIONS(2000),
    [anon_sym_default] = ACTIONS(2000),
    [anon_sym_enum] = ACTIONS(2000),
    [anon_sym_fn] = ACTIONS(2000),
    [anon_sym_for] = ACTIONS(2000),
    [anon_sym_if] = ACTIONS(2000),
    [anon_sym_impl] = ACTIONS(2000),
    [anon_sym_let] = ACTIONS(2000),
    [anon_sym_loop] = ACTIONS(2000),
    [anon_sym_match] = ACTIONS(2000),
    [anon_sym_mod] = ACTIONS(2000),
    [anon_sym_pub] = ACTIONS(2000),
    [anon_sym_return] = ACTIONS(2000),
    [anon_sym_static] = ACTIONS(2000),
    [anon_sym_struct] = ACTIONS(2000),
    [anon_sym_trait] = ACTIONS(2000),
    [anon_sym_type] = ACTIONS(2000),
    [anon_sym_union] = ACTIONS(2000),
    [anon_sym_unsafe] = ACTIONS(2000),
    [anon_sym_use] = ACTIONS(2000),
    [anon_sym_while] = ACTIONS(2000),
    [anon_sym_POUND] = ACTIONS(1998),
    [anon_sym_BANG] = ACTIONS(1998),
    [anon_sym_extern] = ACTIONS(2000),
    [anon_sym_LT] = ACTIONS(1998),
    [anon_sym_COLON_COLON] = ACTIONS(1998),
    [anon_sym_AMP] = ACTIONS(1998),
    [anon_sym_DOT_DOT] = ACTIONS(1998),
    [anon_sym_DASH] = ACTIONS(1998),
    [anon_sym_PIPE] = ACTIONS(1998),
    [anon_sym_yield] = ACTIONS(2000),
    [anon_sym_move] = ACTIONS(2000),
    [sym_integer_literal] = ACTIONS(1998),
    [aux_sym_string_literal_token1] = ACTIONS(1998),
    [sym_char_literal] = ACTIONS(1998),
    [anon_sym_true] = ACTIONS(2000),
    [anon_sym_false] = ACTIONS(2000),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(2000),
    [sym_super] = ACTIONS(2000),
    [sym_crate] = ACTIONS(2000),
    [sym_metavariable] = ACTIONS(1998),
    [sym_raw_string_literal] = ACTIONS(1998),
    [sym_float_literal] = ACTIONS(1998),
    [sym_block_comment] = ACTIONS(3),
  },
  [481] = {
    [sym__token_pattern] = STATE(497),
    [sym_token_tree_pattern] = STATE(497),
    [sym_token_binding_pattern] = STATE(497),
    [sym_token_repetition_pattern] = STATE(497),
    [sym__literal] = STATE(497),
    [sym_string_literal] = STATE(585),
    [sym_boolean_literal] = STATE(585),
    [aux_sym_token_tree_pattern_repeat1] = STATE(497),
    [sym_identifier] = ACTIONS(2002),
    [anon_sym_LPAREN] = ACTIONS(2004),
    [anon_sym_LBRACE] = ACTIONS(2006),
    [anon_sym_RBRACE] = ACTIONS(2008),
    [anon_sym_LBRACK] = ACTIONS(2010),
    [anon_sym_DOLLAR] = ACTIONS(2012),
    [anon_sym_u8] = ACTIONS(2002),
    [anon_sym_i8] = ACTIONS(2002),
    [anon_sym_u16] = ACTIONS(2002),
    [anon_sym_i16] = ACTIONS(2002),
    [anon_sym_u32] = ACTIONS(2002),
    [anon_sym_i32] = ACTIONS(2002),
    [anon_sym_u64] = ACTIONS(2002),
    [anon_sym_i64] = ACTIONS(2002),
    [anon_sym_u128] = ACTIONS(2002),
    [anon_sym_i128] = ACTIONS(2002),
    [anon_sym_isize] = ACTIONS(2002),
    [anon_sym_usize] = ACTIONS(2002),
    [anon_sym_f32] = ACTIONS(2002),
    [anon_sym_f64] = ACTIONS(2002),
    [anon_sym_bool] = ACTIONS(2002),
    [anon_sym_str] = ACTIONS(2002),
    [anon_sym_char] = ACTIONS(2002),
    [aux_sym__non_special_token_token1] = ACTIONS(2002),
    [anon_sym_SQUOTE] = ACTIONS(2002),
    [anon_sym_as] = ACTIONS(2002),
    [anon_sym_async] = ACTIONS(2002),
    [anon_sym_await] = ACTIONS(2002),
    [anon_sym_break] = ACTIONS(2002),
    [anon_sym_const] = ACTIONS(2002),
    [anon_sym_continue] = ACTIONS(2002),
    [anon_sym_default] = ACTIONS(2002),
    [anon_sym_enum] = ACTIONS(2002),
    [anon_sym_fn] = ACTIONS(2002),
    [anon_sym_for] = ACTIONS(2002),
    [anon_sym_if] = ACTIONS(2002),
    [anon_sym_impl] = ACTIONS(2002),
    [anon_sym_let] = ACTIONS(2002),
    [anon_sym_loop] = ACTIONS(2002),
    [anon_sym_match] = ACTIONS(2002),
    [anon_sym_mod] = ACTIONS(2002),
    [anon_sym_pub] = ACTIONS(2002),
    [anon_sym_return] = ACTIONS(2002),
    [anon_sym_static] = ACTIONS(2002),
    [anon_sym_struct] = ACTIONS(2002),
    [anon_sym_trait] = ACTIONS(2002),
    [anon_sym_type] = ACTIONS(2002),
    [anon_sym_union] = ACTIONS(2002),
    [anon_sym_unsafe] = ACTIONS(2002),
    [anon_sym_use] = ACTIONS(2002),
    [anon_sym_where] = ACTIONS(2002),
    [anon_sym_while] = ACTIONS(2002),
    [sym_mutable_specifier] = ACTIONS(2002),
    [sym_integer_literal] = ACTIONS(2014),
    [aux_sym_string_literal_token1] = ACTIONS(2016),
    [sym_char_literal] = ACTIONS(2014),
    [anon_sym_true] = ACTIONS(2018),
    [anon_sym_false] = ACTIONS(2018),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2002),
    [sym_super] = ACTIONS(2002),
    [sym_crate] = ACTIONS(2002),
    [sym_metavariable] = ACTIONS(2020),
    [sym_raw_string_literal] = ACTIONS(2014),
    [sym_float_literal] = ACTIONS(2014),
    [sym_block_comment] = ACTIONS(3),
  },
  [482] = {
    [sym__token_pattern] = STATE(491),
    [sym_token_tree_pattern] = STATE(491),
    [sym_token_binding_pattern] = STATE(491),
    [sym_token_repetition_pattern] = STATE(491),
    [sym__literal] = STATE(491),
    [sym_string_literal] = STATE(585),
    [sym_boolean_literal] = STATE(585),
    [aux_sym_token_tree_pattern_repeat1] = STATE(491),
    [sym_identifier] = ACTIONS(2022),
    [anon_sym_LPAREN] = ACTIONS(2004),
    [anon_sym_RPAREN] = ACTIONS(2024),
    [anon_sym_LBRACE] = ACTIONS(2006),
    [anon_sym_LBRACK] = ACTIONS(2010),
    [anon_sym_DOLLAR] = ACTIONS(2012),
    [anon_sym_u8] = ACTIONS(2022),
    [anon_sym_i8] = ACTIONS(2022),
    [anon_sym_u16] = ACTIONS(2022),
    [anon_sym_i16] = ACTIONS(2022),
    [anon_sym_u32] = ACTIONS(2022),
    [anon_sym_i32] = ACTIONS(2022),
    [anon_sym_u64] = ACTIONS(2022),
    [anon_sym_i64] = ACTIONS(2022),
    [anon_sym_u128] = ACTIONS(2022),
    [anon_sym_i128] = ACTIONS(2022),
    [anon_sym_isize] = ACTIONS(2022),
    [anon_sym_usize] = ACTIONS(2022),
    [anon_sym_f32] = ACTIONS(2022),
    [anon_sym_f64] = ACTIONS(2022),
    [anon_sym_bool] = ACTIONS(2022),
    [anon_sym_str] = ACTIONS(2022),
    [anon_sym_char] = ACTIONS(2022),
    [aux_sym__non_special_token_token1] = ACTIONS(2022),
    [anon_sym_SQUOTE] = ACTIONS(2022),
    [anon_sym_as] = ACTIONS(2022),
    [anon_sym_async] = ACTIONS(2022),
    [anon_sym_await] = ACTIONS(2022),
    [anon_sym_break] = ACTIONS(2022),
    [anon_sym_const] = ACTIONS(2022),
    [anon_sym_continue] = ACTIONS(2022),
    [anon_sym_default] = ACTIONS(2022),
    [anon_sym_enum] = ACTIONS(2022),
    [anon_sym_fn] = ACTIONS(2022),
    [anon_sym_for] = ACTIONS(2022),
    [anon_sym_if] = ACTIONS(2022),
    [anon_sym_impl] = ACTIONS(2022),
    [anon_sym_let] = ACTIONS(2022),
    [anon_sym_loop] = ACTIONS(2022),
    [anon_sym_match] = ACTIONS(2022),
    [anon_sym_mod] = ACTIONS(2022),
    [anon_sym_pub] = ACTIONS(2022),
    [anon_sym_return] = ACTIONS(2022),
    [anon_sym_static] = ACTIONS(2022),
    [anon_sym_struct] = ACTIONS(2022),
    [anon_sym_trait] = ACTIONS(2022),
    [anon_sym_type] = ACTIONS(2022),
    [anon_sym_union] = ACTIONS(2022),
    [anon_sym_unsafe] = ACTIONS(2022),
    [anon_sym_use] = ACTIONS(2022),
    [anon_sym_where] = ACTIONS(2022),
    [anon_sym_while] = ACTIONS(2022),
    [sym_mutable_specifier] = ACTIONS(2022),
    [sym_integer_literal] = ACTIONS(2014),
    [aux_sym_string_literal_token1] = ACTIONS(2016),
    [sym_char_literal] = ACTIONS(2014),
    [anon_sym_true] = ACTIONS(2018),
    [anon_sym_false] = ACTIONS(2018),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2022),
    [sym_super] = ACTIONS(2022),
    [sym_crate] = ACTIONS(2022),
    [sym_metavariable] = ACTIONS(2020),
    [sym_raw_string_literal] = ACTIONS(2014),
    [sym_float_literal] = ACTIONS(2014),
    [sym_block_comment] = ACTIONS(3),
  },
  [483] = {
    [sym_delim_token_tree] = STATE(483),
    [sym__delim_tokens] = STATE(483),
    [sym__non_delim_token] = STATE(483),
    [sym__literal] = STATE(483),
    [sym_string_literal] = STATE(592),
    [sym_boolean_literal] = STATE(592),
    [aux_sym_delim_token_tree_repeat1] = STATE(483),
    [sym_identifier] = ACTIONS(2026),
    [anon_sym_LPAREN] = ACTIONS(2029),
    [anon_sym_RPAREN] = ACTIONS(2032),
    [anon_sym_LBRACE] = ACTIONS(2034),
    [anon_sym_RBRACE] = ACTIONS(2032),
    [anon_sym_LBRACK] = ACTIONS(2037),
    [anon_sym_RBRACK] = ACTIONS(2032),
    [anon_sym_DOLLAR] = ACTIONS(2040),
    [anon_sym_u8] = ACTIONS(2026),
    [anon_sym_i8] = ACTIONS(2026),
    [anon_sym_u16] = ACTIONS(2026),
    [anon_sym_i16] = ACTIONS(2026),
    [anon_sym_u32] = ACTIONS(2026),
    [anon_sym_i32] = ACTIONS(2026),
    [anon_sym_u64] = ACTIONS(2026),
    [anon_sym_i64] = ACTIONS(2026),
    [anon_sym_u128] = ACTIONS(2026),
    [anon_sym_i128] = ACTIONS(2026),
    [anon_sym_isize] = ACTIONS(2026),
    [anon_sym_usize] = ACTIONS(2026),
    [anon_sym_f32] = ACTIONS(2026),
    [anon_sym_f64] = ACTIONS(2026),
    [anon_sym_bool] = ACTIONS(2026),
    [anon_sym_str] = ACTIONS(2026),
    [anon_sym_char] = ACTIONS(2026),
    [aux_sym__non_special_token_token1] = ACTIONS(2026),
    [anon_sym_SQUOTE] = ACTIONS(2026),
    [anon_sym_as] = ACTIONS(2026),
    [anon_sym_async] = ACTIONS(2026),
    [anon_sym_await] = ACTIONS(2026),
    [anon_sym_break] = ACTIONS(2026),
    [anon_sym_const] = ACTIONS(2026),
    [anon_sym_continue] = ACTIONS(2026),
    [anon_sym_default] = ACTIONS(2026),
    [anon_sym_enum] = ACTIONS(2026),
    [anon_sym_fn] = ACTIONS(2026),
    [anon_sym_for] = ACTIONS(2026),
    [anon_sym_if] = ACTIONS(2026),
    [anon_sym_impl] = ACTIONS(2026),
    [anon_sym_let] = ACTIONS(2026),
    [anon_sym_loop] = ACTIONS(2026),
    [anon_sym_match] = ACTIONS(2026),
    [anon_sym_mod] = ACTIONS(2026),
    [anon_sym_pub] = ACTIONS(2026),
    [anon_sym_return] = ACTIONS(2026),
    [anon_sym_static] = ACTIONS(2026),
    [anon_sym_struct] = ACTIONS(2026),
    [anon_sym_trait] = ACTIONS(2026),
    [anon_sym_type] = ACTIONS(2026),
    [anon_sym_union] = ACTIONS(2026),
    [anon_sym_unsafe] = ACTIONS(2026),
    [anon_sym_use] = ACTIONS(2026),
    [anon_sym_where] = ACTIONS(2026),
    [anon_sym_while] = ACTIONS(2026),
    [sym_mutable_specifier] = ACTIONS(2026),
    [sym_integer_literal] = ACTIONS(2043),
    [aux_sym_string_literal_token1] = ACTIONS(2046),
    [sym_char_literal] = ACTIONS(2043),
    [anon_sym_true] = ACTIONS(2049),
    [anon_sym_false] = ACTIONS(2049),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2026),
    [sym_super] = ACTIONS(2026),
    [sym_crate] = ACTIONS(2026),
    [sym_raw_string_literal] = ACTIONS(2043),
    [sym_float_literal] = ACTIONS(2043),
    [sym_block_comment] = ACTIONS(3),
  },
  [484] = {
    [sym_attribute_item] = STATE(547),
    [sym_bracketed_type] = STATE(2489),
    [sym_generic_type] = STATE(2488),
    [sym_generic_type_with_turbofish] = STATE(2487),
    [sym_macro_invocation] = STATE(1422),
    [sym_scoped_identifier] = STATE(1337),
    [sym_scoped_type_identifier] = STATE(2013),
    [sym_match_arm] = STATE(500),
    [sym_last_match_arm] = STATE(2333),
    [sym_match_pattern] = STATE(2387),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1981),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [aux_sym_enum_variant_list_repeat1] = STATE(547),
    [aux_sym_match_block_repeat1] = STATE(500),
    [sym_identifier] = ACTIONS(1524),
    [anon_sym_LPAREN] = ACTIONS(1526),
    [anon_sym_LBRACK] = ACTIONS(1530),
    [anon_sym_u8] = ACTIONS(1532),
    [anon_sym_i8] = ACTIONS(1532),
    [anon_sym_u16] = ACTIONS(1532),
    [anon_sym_i16] = ACTIONS(1532),
    [anon_sym_u32] = ACTIONS(1532),
    [anon_sym_i32] = ACTIONS(1532),
    [anon_sym_u64] = ACTIONS(1532),
    [anon_sym_i64] = ACTIONS(1532),
    [anon_sym_u128] = ACTIONS(1532),
    [anon_sym_i128] = ACTIONS(1532),
    [anon_sym_isize] = ACTIONS(1532),
    [anon_sym_usize] = ACTIONS(1532),
    [anon_sym_f32] = ACTIONS(1532),
    [anon_sym_f64] = ACTIONS(1532),
    [anon_sym_bool] = ACTIONS(1532),
    [anon_sym_str] = ACTIONS(1532),
    [anon_sym_char] = ACTIONS(1532),
    [anon_sym_const] = ACTIONS(1534),
    [anon_sym_default] = ACTIONS(1536),
    [anon_sym_union] = ACTIONS(1536),
    [anon_sym_POUND] = ACTIONS(370),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(1538),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(1540),
    [sym_mutable_specifier] = ACTIONS(796),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1542),
    [sym_super] = ACTIONS(1542),
    [sym_crate] = ACTIONS(1542),
    [sym_metavariable] = ACTIONS(1544),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [485] = {
    [sym_token_tree] = STATE(485),
    [sym_token_repetition] = STATE(485),
    [sym__literal] = STATE(485),
    [sym_string_literal] = STATE(585),
    [sym_boolean_literal] = STATE(585),
    [aux_sym_token_tree_repeat1] = STATE(485),
    [sym_identifier] = ACTIONS(2052),
    [anon_sym_LPAREN] = ACTIONS(2055),
    [anon_sym_RPAREN] = ACTIONS(2058),
    [anon_sym_LBRACE] = ACTIONS(2060),
    [anon_sym_RBRACE] = ACTIONS(2058),
    [anon_sym_LBRACK] = ACTIONS(2063),
    [anon_sym_RBRACK] = ACTIONS(2058),
    [anon_sym_DOLLAR] = ACTIONS(2066),
    [anon_sym_u8] = ACTIONS(2052),
    [anon_sym_i8] = ACTIONS(2052),
    [anon_sym_u16] = ACTIONS(2052),
    [anon_sym_i16] = ACTIONS(2052),
    [anon_sym_u32] = ACTIONS(2052),
    [anon_sym_i32] = ACTIONS(2052),
    [anon_sym_u64] = ACTIONS(2052),
    [anon_sym_i64] = ACTIONS(2052),
    [anon_sym_u128] = ACTIONS(2052),
    [anon_sym_i128] = ACTIONS(2052),
    [anon_sym_isize] = ACTIONS(2052),
    [anon_sym_usize] = ACTIONS(2052),
    [anon_sym_f32] = ACTIONS(2052),
    [anon_sym_f64] = ACTIONS(2052),
    [anon_sym_bool] = ACTIONS(2052),
    [anon_sym_str] = ACTIONS(2052),
    [anon_sym_char] = ACTIONS(2052),
    [aux_sym__non_special_token_token1] = ACTIONS(2052),
    [anon_sym_SQUOTE] = ACTIONS(2052),
    [anon_sym_as] = ACTIONS(2052),
    [anon_sym_async] = ACTIONS(2052),
    [anon_sym_await] = ACTIONS(2052),
    [anon_sym_break] = ACTIONS(2052),
    [anon_sym_const] = ACTIONS(2052),
    [anon_sym_continue] = ACTIONS(2052),
    [anon_sym_default] = ACTIONS(2052),
    [anon_sym_enum] = ACTIONS(2052),
    [anon_sym_fn] = ACTIONS(2052),
    [anon_sym_for] = ACTIONS(2052),
    [anon_sym_if] = ACTIONS(2052),
    [anon_sym_impl] = ACTIONS(2052),
    [anon_sym_let] = ACTIONS(2052),
    [anon_sym_loop] = ACTIONS(2052),
    [anon_sym_match] = ACTIONS(2052),
    [anon_sym_mod] = ACTIONS(2052),
    [anon_sym_pub] = ACTIONS(2052),
    [anon_sym_return] = ACTIONS(2052),
    [anon_sym_static] = ACTIONS(2052),
    [anon_sym_struct] = ACTIONS(2052),
    [anon_sym_trait] = ACTIONS(2052),
    [anon_sym_type] = ACTIONS(2052),
    [anon_sym_union] = ACTIONS(2052),
    [anon_sym_unsafe] = ACTIONS(2052),
    [anon_sym_use] = ACTIONS(2052),
    [anon_sym_where] = ACTIONS(2052),
    [anon_sym_while] = ACTIONS(2052),
    [sym_mutable_specifier] = ACTIONS(2052),
    [sym_integer_literal] = ACTIONS(2069),
    [aux_sym_string_literal_token1] = ACTIONS(2072),
    [sym_char_literal] = ACTIONS(2069),
    [anon_sym_true] = ACTIONS(2075),
    [anon_sym_false] = ACTIONS(2075),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2052),
    [sym_super] = ACTIONS(2052),
    [sym_crate] = ACTIONS(2052),
    [sym_metavariable] = ACTIONS(2078),
    [sym_raw_string_literal] = ACTIONS(2069),
    [sym_float_literal] = ACTIONS(2069),
    [sym_block_comment] = ACTIONS(3),
  },
  [486] = {
    [sym_attribute_item] = STATE(547),
    [sym_bracketed_type] = STATE(2489),
    [sym_generic_type] = STATE(2488),
    [sym_generic_type_with_turbofish] = STATE(2487),
    [sym_macro_invocation] = STATE(1422),
    [sym_scoped_identifier] = STATE(1337),
    [sym_scoped_type_identifier] = STATE(2013),
    [sym_match_arm] = STATE(500),
    [sym_last_match_arm] = STATE(2427),
    [sym_match_pattern] = STATE(2387),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1981),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [aux_sym_enum_variant_list_repeat1] = STATE(547),
    [aux_sym_match_block_repeat1] = STATE(500),
    [sym_identifier] = ACTIONS(1524),
    [anon_sym_LPAREN] = ACTIONS(1526),
    [anon_sym_LBRACK] = ACTIONS(1530),
    [anon_sym_u8] = ACTIONS(1532),
    [anon_sym_i8] = ACTIONS(1532),
    [anon_sym_u16] = ACTIONS(1532),
    [anon_sym_i16] = ACTIONS(1532),
    [anon_sym_u32] = ACTIONS(1532),
    [anon_sym_i32] = ACTIONS(1532),
    [anon_sym_u64] = ACTIONS(1532),
    [anon_sym_i64] = ACTIONS(1532),
    [anon_sym_u128] = ACTIONS(1532),
    [anon_sym_i128] = ACTIONS(1532),
    [anon_sym_isize] = ACTIONS(1532),
    [anon_sym_usize] = ACTIONS(1532),
    [anon_sym_f32] = ACTIONS(1532),
    [anon_sym_f64] = ACTIONS(1532),
    [anon_sym_bool] = ACTIONS(1532),
    [anon_sym_str] = ACTIONS(1532),
    [anon_sym_char] = ACTIONS(1532),
    [anon_sym_const] = ACTIONS(1534),
    [anon_sym_default] = ACTIONS(1536),
    [anon_sym_union] = ACTIONS(1536),
    [anon_sym_POUND] = ACTIONS(370),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(1538),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(1540),
    [sym_mutable_specifier] = ACTIONS(796),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1542),
    [sym_super] = ACTIONS(1542),
    [sym_crate] = ACTIONS(1542),
    [sym_metavariable] = ACTIONS(1544),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [487] = {
    [sym__token_pattern] = STATE(239),
    [sym_token_tree_pattern] = STATE(239),
    [sym_token_binding_pattern] = STATE(239),
    [sym_token_repetition_pattern] = STATE(239),
    [sym__literal] = STATE(239),
    [sym_string_literal] = STATE(585),
    [sym_boolean_literal] = STATE(585),
    [aux_sym_token_tree_pattern_repeat1] = STATE(239),
    [sym_identifier] = ACTIONS(2081),
    [anon_sym_LPAREN] = ACTIONS(2004),
    [anon_sym_LBRACE] = ACTIONS(2006),
    [anon_sym_LBRACK] = ACTIONS(2010),
    [anon_sym_RBRACK] = ACTIONS(2083),
    [anon_sym_DOLLAR] = ACTIONS(2012),
    [anon_sym_u8] = ACTIONS(2081),
    [anon_sym_i8] = ACTIONS(2081),
    [anon_sym_u16] = ACTIONS(2081),
    [anon_sym_i16] = ACTIONS(2081),
    [anon_sym_u32] = ACTIONS(2081),
    [anon_sym_i32] = ACTIONS(2081),
    [anon_sym_u64] = ACTIONS(2081),
    [anon_sym_i64] = ACTIONS(2081),
    [anon_sym_u128] = ACTIONS(2081),
    [anon_sym_i128] = ACTIONS(2081),
    [anon_sym_isize] = ACTIONS(2081),
    [anon_sym_usize] = ACTIONS(2081),
    [anon_sym_f32] = ACTIONS(2081),
    [anon_sym_f64] = ACTIONS(2081),
    [anon_sym_bool] = ACTIONS(2081),
    [anon_sym_str] = ACTIONS(2081),
    [anon_sym_char] = ACTIONS(2081),
    [aux_sym__non_special_token_token1] = ACTIONS(2081),
    [anon_sym_SQUOTE] = ACTIONS(2081),
    [anon_sym_as] = ACTIONS(2081),
    [anon_sym_async] = ACTIONS(2081),
    [anon_sym_await] = ACTIONS(2081),
    [anon_sym_break] = ACTIONS(2081),
    [anon_sym_const] = ACTIONS(2081),
    [anon_sym_continue] = ACTIONS(2081),
    [anon_sym_default] = ACTIONS(2081),
    [anon_sym_enum] = ACTIONS(2081),
    [anon_sym_fn] = ACTIONS(2081),
    [anon_sym_for] = ACTIONS(2081),
    [anon_sym_if] = ACTIONS(2081),
    [anon_sym_impl] = ACTIONS(2081),
    [anon_sym_let] = ACTIONS(2081),
    [anon_sym_loop] = ACTIONS(2081),
    [anon_sym_match] = ACTIONS(2081),
    [anon_sym_mod] = ACTIONS(2081),
    [anon_sym_pub] = ACTIONS(2081),
    [anon_sym_return] = ACTIONS(2081),
    [anon_sym_static] = ACTIONS(2081),
    [anon_sym_struct] = ACTIONS(2081),
    [anon_sym_trait] = ACTIONS(2081),
    [anon_sym_type] = ACTIONS(2081),
    [anon_sym_union] = ACTIONS(2081),
    [anon_sym_unsafe] = ACTIONS(2081),
    [anon_sym_use] = ACTIONS(2081),
    [anon_sym_where] = ACTIONS(2081),
    [anon_sym_while] = ACTIONS(2081),
    [sym_mutable_specifier] = ACTIONS(2081),
    [sym_integer_literal] = ACTIONS(2014),
    [aux_sym_string_literal_token1] = ACTIONS(2016),
    [sym_char_literal] = ACTIONS(2014),
    [anon_sym_true] = ACTIONS(2018),
    [anon_sym_false] = ACTIONS(2018),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2081),
    [sym_super] = ACTIONS(2081),
    [sym_crate] = ACTIONS(2081),
    [sym_metavariable] = ACTIONS(2020),
    [sym_raw_string_literal] = ACTIONS(2014),
    [sym_float_literal] = ACTIONS(2014),
    [sym_block_comment] = ACTIONS(3),
  },
  [488] = {
    [sym__token_pattern] = STATE(239),
    [sym_token_tree_pattern] = STATE(239),
    [sym_token_binding_pattern] = STATE(239),
    [sym_token_repetition_pattern] = STATE(239),
    [sym__literal] = STATE(239),
    [sym_string_literal] = STATE(585),
    [sym_boolean_literal] = STATE(585),
    [aux_sym_token_tree_pattern_repeat1] = STATE(239),
    [sym_identifier] = ACTIONS(2081),
    [anon_sym_LPAREN] = ACTIONS(2004),
    [anon_sym_LBRACE] = ACTIONS(2006),
    [anon_sym_RBRACE] = ACTIONS(2083),
    [anon_sym_LBRACK] = ACTIONS(2010),
    [anon_sym_DOLLAR] = ACTIONS(2012),
    [anon_sym_u8] = ACTIONS(2081),
    [anon_sym_i8] = ACTIONS(2081),
    [anon_sym_u16] = ACTIONS(2081),
    [anon_sym_i16] = ACTIONS(2081),
    [anon_sym_u32] = ACTIONS(2081),
    [anon_sym_i32] = ACTIONS(2081),
    [anon_sym_u64] = ACTIONS(2081),
    [anon_sym_i64] = ACTIONS(2081),
    [anon_sym_u128] = ACTIONS(2081),
    [anon_sym_i128] = ACTIONS(2081),
    [anon_sym_isize] = ACTIONS(2081),
    [anon_sym_usize] = ACTIONS(2081),
    [anon_sym_f32] = ACTIONS(2081),
    [anon_sym_f64] = ACTIONS(2081),
    [anon_sym_bool] = ACTIONS(2081),
    [anon_sym_str] = ACTIONS(2081),
    [anon_sym_char] = ACTIONS(2081),
    [aux_sym__non_special_token_token1] = ACTIONS(2081),
    [anon_sym_SQUOTE] = ACTIONS(2081),
    [anon_sym_as] = ACTIONS(2081),
    [anon_sym_async] = ACTIONS(2081),
    [anon_sym_await] = ACTIONS(2081),
    [anon_sym_break] = ACTIONS(2081),
    [anon_sym_const] = ACTIONS(2081),
    [anon_sym_continue] = ACTIONS(2081),
    [anon_sym_default] = ACTIONS(2081),
    [anon_sym_enum] = ACTIONS(2081),
    [anon_sym_fn] = ACTIONS(2081),
    [anon_sym_for] = ACTIONS(2081),
    [anon_sym_if] = ACTIONS(2081),
    [anon_sym_impl] = ACTIONS(2081),
    [anon_sym_let] = ACTIONS(2081),
    [anon_sym_loop] = ACTIONS(2081),
    [anon_sym_match] = ACTIONS(2081),
    [anon_sym_mod] = ACTIONS(2081),
    [anon_sym_pub] = ACTIONS(2081),
    [anon_sym_return] = ACTIONS(2081),
    [anon_sym_static] = ACTIONS(2081),
    [anon_sym_struct] = ACTIONS(2081),
    [anon_sym_trait] = ACTIONS(2081),
    [anon_sym_type] = ACTIONS(2081),
    [anon_sym_union] = ACTIONS(2081),
    [anon_sym_unsafe] = ACTIONS(2081),
    [anon_sym_use] = ACTIONS(2081),
    [anon_sym_where] = ACTIONS(2081),
    [anon_sym_while] = ACTIONS(2081),
    [sym_mutable_specifier] = ACTIONS(2081),
    [sym_integer_literal] = ACTIONS(2014),
    [aux_sym_string_literal_token1] = ACTIONS(2016),
    [sym_char_literal] = ACTIONS(2014),
    [anon_sym_true] = ACTIONS(2018),
    [anon_sym_false] = ACTIONS(2018),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2081),
    [sym_super] = ACTIONS(2081),
    [sym_crate] = ACTIONS(2081),
    [sym_metavariable] = ACTIONS(2020),
    [sym_raw_string_literal] = ACTIONS(2014),
    [sym_float_literal] = ACTIONS(2014),
    [sym_block_comment] = ACTIONS(3),
  },
  [489] = {
    [sym__token_pattern] = STATE(239),
    [sym_token_tree_pattern] = STATE(239),
    [sym_token_binding_pattern] = STATE(239),
    [sym_token_repetition_pattern] = STATE(239),
    [sym__literal] = STATE(239),
    [sym_string_literal] = STATE(585),
    [sym_boolean_literal] = STATE(585),
    [aux_sym_token_tree_pattern_repeat1] = STATE(239),
    [sym_identifier] = ACTIONS(2081),
    [anon_sym_LPAREN] = ACTIONS(2004),
    [anon_sym_RPAREN] = ACTIONS(2083),
    [anon_sym_LBRACE] = ACTIONS(2006),
    [anon_sym_LBRACK] = ACTIONS(2010),
    [anon_sym_DOLLAR] = ACTIONS(2012),
    [anon_sym_u8] = ACTIONS(2081),
    [anon_sym_i8] = ACTIONS(2081),
    [anon_sym_u16] = ACTIONS(2081),
    [anon_sym_i16] = ACTIONS(2081),
    [anon_sym_u32] = ACTIONS(2081),
    [anon_sym_i32] = ACTIONS(2081),
    [anon_sym_u64] = ACTIONS(2081),
    [anon_sym_i64] = ACTIONS(2081),
    [anon_sym_u128] = ACTIONS(2081),
    [anon_sym_i128] = ACTIONS(2081),
    [anon_sym_isize] = ACTIONS(2081),
    [anon_sym_usize] = ACTIONS(2081),
    [anon_sym_f32] = ACTIONS(2081),
    [anon_sym_f64] = ACTIONS(2081),
    [anon_sym_bool] = ACTIONS(2081),
    [anon_sym_str] = ACTIONS(2081),
    [anon_sym_char] = ACTIONS(2081),
    [aux_sym__non_special_token_token1] = ACTIONS(2081),
    [anon_sym_SQUOTE] = ACTIONS(2081),
    [anon_sym_as] = ACTIONS(2081),
    [anon_sym_async] = ACTIONS(2081),
    [anon_sym_await] = ACTIONS(2081),
    [anon_sym_break] = ACTIONS(2081),
    [anon_sym_const] = ACTIONS(2081),
    [anon_sym_continue] = ACTIONS(2081),
    [anon_sym_default] = ACTIONS(2081),
    [anon_sym_enum] = ACTIONS(2081),
    [anon_sym_fn] = ACTIONS(2081),
    [anon_sym_for] = ACTIONS(2081),
    [anon_sym_if] = ACTIONS(2081),
    [anon_sym_impl] = ACTIONS(2081),
    [anon_sym_let] = ACTIONS(2081),
    [anon_sym_loop] = ACTIONS(2081),
    [anon_sym_match] = ACTIONS(2081),
    [anon_sym_mod] = ACTIONS(2081),
    [anon_sym_pub] = ACTIONS(2081),
    [anon_sym_return] = ACTIONS(2081),
    [anon_sym_static] = ACTIONS(2081),
    [anon_sym_struct] = ACTIONS(2081),
    [anon_sym_trait] = ACTIONS(2081),
    [anon_sym_type] = ACTIONS(2081),
    [anon_sym_union] = ACTIONS(2081),
    [anon_sym_unsafe] = ACTIONS(2081),
    [anon_sym_use] = ACTIONS(2081),
    [anon_sym_where] = ACTIONS(2081),
    [anon_sym_while] = ACTIONS(2081),
    [sym_mutable_specifier] = ACTIONS(2081),
    [sym_integer_literal] = ACTIONS(2014),
    [aux_sym_string_literal_token1] = ACTIONS(2016),
    [sym_char_literal] = ACTIONS(2014),
    [anon_sym_true] = ACTIONS(2018),
    [anon_sym_false] = ACTIONS(2018),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2081),
    [sym_super] = ACTIONS(2081),
    [sym_crate] = ACTIONS(2081),
    [sym_metavariable] = ACTIONS(2020),
    [sym_raw_string_literal] = ACTIONS(2014),
    [sym_float_literal] = ACTIONS(2014),
    [sym_block_comment] = ACTIONS(3),
  },
  [490] = {
    [sym_attribute_item] = STATE(547),
    [sym_bracketed_type] = STATE(2489),
    [sym_generic_type] = STATE(2488),
    [sym_generic_type_with_turbofish] = STATE(2487),
    [sym_macro_invocation] = STATE(1422),
    [sym_scoped_identifier] = STATE(1337),
    [sym_scoped_type_identifier] = STATE(2013),
    [sym_match_arm] = STATE(500),
    [sym_last_match_arm] = STATE(2482),
    [sym_match_pattern] = STATE(2387),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1981),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [aux_sym_enum_variant_list_repeat1] = STATE(547),
    [aux_sym_match_block_repeat1] = STATE(500),
    [sym_identifier] = ACTIONS(1524),
    [anon_sym_LPAREN] = ACTIONS(1526),
    [anon_sym_LBRACK] = ACTIONS(1530),
    [anon_sym_u8] = ACTIONS(1532),
    [anon_sym_i8] = ACTIONS(1532),
    [anon_sym_u16] = ACTIONS(1532),
    [anon_sym_i16] = ACTIONS(1532),
    [anon_sym_u32] = ACTIONS(1532),
    [anon_sym_i32] = ACTIONS(1532),
    [anon_sym_u64] = ACTIONS(1532),
    [anon_sym_i64] = ACTIONS(1532),
    [anon_sym_u128] = ACTIONS(1532),
    [anon_sym_i128] = ACTIONS(1532),
    [anon_sym_isize] = ACTIONS(1532),
    [anon_sym_usize] = ACTIONS(1532),
    [anon_sym_f32] = ACTIONS(1532),
    [anon_sym_f64] = ACTIONS(1532),
    [anon_sym_bool] = ACTIONS(1532),
    [anon_sym_str] = ACTIONS(1532),
    [anon_sym_char] = ACTIONS(1532),
    [anon_sym_const] = ACTIONS(1534),
    [anon_sym_default] = ACTIONS(1536),
    [anon_sym_union] = ACTIONS(1536),
    [anon_sym_POUND] = ACTIONS(370),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(1538),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(1540),
    [sym_mutable_specifier] = ACTIONS(796),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1542),
    [sym_super] = ACTIONS(1542),
    [sym_crate] = ACTIONS(1542),
    [sym_metavariable] = ACTIONS(1544),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [491] = {
    [sym__token_pattern] = STATE(239),
    [sym_token_tree_pattern] = STATE(239),
    [sym_token_binding_pattern] = STATE(239),
    [sym_token_repetition_pattern] = STATE(239),
    [sym__literal] = STATE(239),
    [sym_string_literal] = STATE(585),
    [sym_boolean_literal] = STATE(585),
    [aux_sym_token_tree_pattern_repeat1] = STATE(239),
    [sym_identifier] = ACTIONS(2081),
    [anon_sym_LPAREN] = ACTIONS(2004),
    [anon_sym_RPAREN] = ACTIONS(2085),
    [anon_sym_LBRACE] = ACTIONS(2006),
    [anon_sym_LBRACK] = ACTIONS(2010),
    [anon_sym_DOLLAR] = ACTIONS(2012),
    [anon_sym_u8] = ACTIONS(2081),
    [anon_sym_i8] = ACTIONS(2081),
    [anon_sym_u16] = ACTIONS(2081),
    [anon_sym_i16] = ACTIONS(2081),
    [anon_sym_u32] = ACTIONS(2081),
    [anon_sym_i32] = ACTIONS(2081),
    [anon_sym_u64] = ACTIONS(2081),
    [anon_sym_i64] = ACTIONS(2081),
    [anon_sym_u128] = ACTIONS(2081),
    [anon_sym_i128] = ACTIONS(2081),
    [anon_sym_isize] = ACTIONS(2081),
    [anon_sym_usize] = ACTIONS(2081),
    [anon_sym_f32] = ACTIONS(2081),
    [anon_sym_f64] = ACTIONS(2081),
    [anon_sym_bool] = ACTIONS(2081),
    [anon_sym_str] = ACTIONS(2081),
    [anon_sym_char] = ACTIONS(2081),
    [aux_sym__non_special_token_token1] = ACTIONS(2081),
    [anon_sym_SQUOTE] = ACTIONS(2081),
    [anon_sym_as] = ACTIONS(2081),
    [anon_sym_async] = ACTIONS(2081),
    [anon_sym_await] = ACTIONS(2081),
    [anon_sym_break] = ACTIONS(2081),
    [anon_sym_const] = ACTIONS(2081),
    [anon_sym_continue] = ACTIONS(2081),
    [anon_sym_default] = ACTIONS(2081),
    [anon_sym_enum] = ACTIONS(2081),
    [anon_sym_fn] = ACTIONS(2081),
    [anon_sym_for] = ACTIONS(2081),
    [anon_sym_if] = ACTIONS(2081),
    [anon_sym_impl] = ACTIONS(2081),
    [anon_sym_let] = ACTIONS(2081),
    [anon_sym_loop] = ACTIONS(2081),
    [anon_sym_match] = ACTIONS(2081),
    [anon_sym_mod] = ACTIONS(2081),
    [anon_sym_pub] = ACTIONS(2081),
    [anon_sym_return] = ACTIONS(2081),
    [anon_sym_static] = ACTIONS(2081),
    [anon_sym_struct] = ACTIONS(2081),
    [anon_sym_trait] = ACTIONS(2081),
    [anon_sym_type] = ACTIONS(2081),
    [anon_sym_union] = ACTIONS(2081),
    [anon_sym_unsafe] = ACTIONS(2081),
    [anon_sym_use] = ACTIONS(2081),
    [anon_sym_where] = ACTIONS(2081),
    [anon_sym_while] = ACTIONS(2081),
    [sym_mutable_specifier] = ACTIONS(2081),
    [sym_integer_literal] = ACTIONS(2014),
    [aux_sym_string_literal_token1] = ACTIONS(2016),
    [sym_char_literal] = ACTIONS(2014),
    [anon_sym_true] = ACTIONS(2018),
    [anon_sym_false] = ACTIONS(2018),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2081),
    [sym_super] = ACTIONS(2081),
    [sym_crate] = ACTIONS(2081),
    [sym_metavariable] = ACTIONS(2020),
    [sym_raw_string_literal] = ACTIONS(2014),
    [sym_float_literal] = ACTIONS(2014),
    [sym_block_comment] = ACTIONS(3),
  },
  [492] = {
    [sym__token_pattern] = STATE(493),
    [sym_token_tree_pattern] = STATE(493),
    [sym_token_binding_pattern] = STATE(493),
    [sym_token_repetition_pattern] = STATE(493),
    [sym__literal] = STATE(493),
    [sym_string_literal] = STATE(585),
    [sym_boolean_literal] = STATE(585),
    [aux_sym_token_tree_pattern_repeat1] = STATE(493),
    [sym_identifier] = ACTIONS(2087),
    [anon_sym_LPAREN] = ACTIONS(2004),
    [anon_sym_RPAREN] = ACTIONS(2008),
    [anon_sym_LBRACE] = ACTIONS(2006),
    [anon_sym_LBRACK] = ACTIONS(2010),
    [anon_sym_DOLLAR] = ACTIONS(2012),
    [anon_sym_u8] = ACTIONS(2087),
    [anon_sym_i8] = ACTIONS(2087),
    [anon_sym_u16] = ACTIONS(2087),
    [anon_sym_i16] = ACTIONS(2087),
    [anon_sym_u32] = ACTIONS(2087),
    [anon_sym_i32] = ACTIONS(2087),
    [anon_sym_u64] = ACTIONS(2087),
    [anon_sym_i64] = ACTIONS(2087),
    [anon_sym_u128] = ACTIONS(2087),
    [anon_sym_i128] = ACTIONS(2087),
    [anon_sym_isize] = ACTIONS(2087),
    [anon_sym_usize] = ACTIONS(2087),
    [anon_sym_f32] = ACTIONS(2087),
    [anon_sym_f64] = ACTIONS(2087),
    [anon_sym_bool] = ACTIONS(2087),
    [anon_sym_str] = ACTIONS(2087),
    [anon_sym_char] = ACTIONS(2087),
    [aux_sym__non_special_token_token1] = ACTIONS(2087),
    [anon_sym_SQUOTE] = ACTIONS(2087),
    [anon_sym_as] = ACTIONS(2087),
    [anon_sym_async] = ACTIONS(2087),
    [anon_sym_await] = ACTIONS(2087),
    [anon_sym_break] = ACTIONS(2087),
    [anon_sym_const] = ACTIONS(2087),
    [anon_sym_continue] = ACTIONS(2087),
    [anon_sym_default] = ACTIONS(2087),
    [anon_sym_enum] = ACTIONS(2087),
    [anon_sym_fn] = ACTIONS(2087),
    [anon_sym_for] = ACTIONS(2087),
    [anon_sym_if] = ACTIONS(2087),
    [anon_sym_impl] = ACTIONS(2087),
    [anon_sym_let] = ACTIONS(2087),
    [anon_sym_loop] = ACTIONS(2087),
    [anon_sym_match] = ACTIONS(2087),
    [anon_sym_mod] = ACTIONS(2087),
    [anon_sym_pub] = ACTIONS(2087),
    [anon_sym_return] = ACTIONS(2087),
    [anon_sym_static] = ACTIONS(2087),
    [anon_sym_struct] = ACTIONS(2087),
    [anon_sym_trait] = ACTIONS(2087),
    [anon_sym_type] = ACTIONS(2087),
    [anon_sym_union] = ACTIONS(2087),
    [anon_sym_unsafe] = ACTIONS(2087),
    [anon_sym_use] = ACTIONS(2087),
    [anon_sym_where] = ACTIONS(2087),
    [anon_sym_while] = ACTIONS(2087),
    [sym_mutable_specifier] = ACTIONS(2087),
    [sym_integer_literal] = ACTIONS(2014),
    [aux_sym_string_literal_token1] = ACTIONS(2016),
    [sym_char_literal] = ACTIONS(2014),
    [anon_sym_true] = ACTIONS(2018),
    [anon_sym_false] = ACTIONS(2018),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2087),
    [sym_super] = ACTIONS(2087),
    [sym_crate] = ACTIONS(2087),
    [sym_metavariable] = ACTIONS(2020),
    [sym_raw_string_literal] = ACTIONS(2014),
    [sym_float_literal] = ACTIONS(2014),
    [sym_block_comment] = ACTIONS(3),
  },
  [493] = {
    [sym__token_pattern] = STATE(239),
    [sym_token_tree_pattern] = STATE(239),
    [sym_token_binding_pattern] = STATE(239),
    [sym_token_repetition_pattern] = STATE(239),
    [sym__literal] = STATE(239),
    [sym_string_literal] = STATE(585),
    [sym_boolean_literal] = STATE(585),
    [aux_sym_token_tree_pattern_repeat1] = STATE(239),
    [sym_identifier] = ACTIONS(2081),
    [anon_sym_LPAREN] = ACTIONS(2004),
    [anon_sym_RPAREN] = ACTIONS(2089),
    [anon_sym_LBRACE] = ACTIONS(2006),
    [anon_sym_LBRACK] = ACTIONS(2010),
    [anon_sym_DOLLAR] = ACTIONS(2012),
    [anon_sym_u8] = ACTIONS(2081),
    [anon_sym_i8] = ACTIONS(2081),
    [anon_sym_u16] = ACTIONS(2081),
    [anon_sym_i16] = ACTIONS(2081),
    [anon_sym_u32] = ACTIONS(2081),
    [anon_sym_i32] = ACTIONS(2081),
    [anon_sym_u64] = ACTIONS(2081),
    [anon_sym_i64] = ACTIONS(2081),
    [anon_sym_u128] = ACTIONS(2081),
    [anon_sym_i128] = ACTIONS(2081),
    [anon_sym_isize] = ACTIONS(2081),
    [anon_sym_usize] = ACTIONS(2081),
    [anon_sym_f32] = ACTIONS(2081),
    [anon_sym_f64] = ACTIONS(2081),
    [anon_sym_bool] = ACTIONS(2081),
    [anon_sym_str] = ACTIONS(2081),
    [anon_sym_char] = ACTIONS(2081),
    [aux_sym__non_special_token_token1] = ACTIONS(2081),
    [anon_sym_SQUOTE] = ACTIONS(2081),
    [anon_sym_as] = ACTIONS(2081),
    [anon_sym_async] = ACTIONS(2081),
    [anon_sym_await] = ACTIONS(2081),
    [anon_sym_break] = ACTIONS(2081),
    [anon_sym_const] = ACTIONS(2081),
    [anon_sym_continue] = ACTIONS(2081),
    [anon_sym_default] = ACTIONS(2081),
    [anon_sym_enum] = ACTIONS(2081),
    [anon_sym_fn] = ACTIONS(2081),
    [anon_sym_for] = ACTIONS(2081),
    [anon_sym_if] = ACTIONS(2081),
    [anon_sym_impl] = ACTIONS(2081),
    [anon_sym_let] = ACTIONS(2081),
    [anon_sym_loop] = ACTIONS(2081),
    [anon_sym_match] = ACTIONS(2081),
    [anon_sym_mod] = ACTIONS(2081),
    [anon_sym_pub] = ACTIONS(2081),
    [anon_sym_return] = ACTIONS(2081),
    [anon_sym_static] = ACTIONS(2081),
    [anon_sym_struct] = ACTIONS(2081),
    [anon_sym_trait] = ACTIONS(2081),
    [anon_sym_type] = ACTIONS(2081),
    [anon_sym_union] = ACTIONS(2081),
    [anon_sym_unsafe] = ACTIONS(2081),
    [anon_sym_use] = ACTIONS(2081),
    [anon_sym_where] = ACTIONS(2081),
    [anon_sym_while] = ACTIONS(2081),
    [sym_mutable_specifier] = ACTIONS(2081),
    [sym_integer_literal] = ACTIONS(2014),
    [aux_sym_string_literal_token1] = ACTIONS(2016),
    [sym_char_literal] = ACTIONS(2014),
    [anon_sym_true] = ACTIONS(2018),
    [anon_sym_false] = ACTIONS(2018),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2081),
    [sym_super] = ACTIONS(2081),
    [sym_crate] = ACTIONS(2081),
    [sym_metavariable] = ACTIONS(2020),
    [sym_raw_string_literal] = ACTIONS(2014),
    [sym_float_literal] = ACTIONS(2014),
    [sym_block_comment] = ACTIONS(3),
  },
  [494] = {
    [sym__token_pattern] = STATE(495),
    [sym_token_tree_pattern] = STATE(495),
    [sym_token_binding_pattern] = STATE(495),
    [sym_token_repetition_pattern] = STATE(495),
    [sym__literal] = STATE(495),
    [sym_string_literal] = STATE(585),
    [sym_boolean_literal] = STATE(585),
    [aux_sym_token_tree_pattern_repeat1] = STATE(495),
    [sym_identifier] = ACTIONS(2091),
    [anon_sym_LPAREN] = ACTIONS(2004),
    [anon_sym_LBRACE] = ACTIONS(2006),
    [anon_sym_LBRACK] = ACTIONS(2010),
    [anon_sym_RBRACK] = ACTIONS(2008),
    [anon_sym_DOLLAR] = ACTIONS(2012),
    [anon_sym_u8] = ACTIONS(2091),
    [anon_sym_i8] = ACTIONS(2091),
    [anon_sym_u16] = ACTIONS(2091),
    [anon_sym_i16] = ACTIONS(2091),
    [anon_sym_u32] = ACTIONS(2091),
    [anon_sym_i32] = ACTIONS(2091),
    [anon_sym_u64] = ACTIONS(2091),
    [anon_sym_i64] = ACTIONS(2091),
    [anon_sym_u128] = ACTIONS(2091),
    [anon_sym_i128] = ACTIONS(2091),
    [anon_sym_isize] = ACTIONS(2091),
    [anon_sym_usize] = ACTIONS(2091),
    [anon_sym_f32] = ACTIONS(2091),
    [anon_sym_f64] = ACTIONS(2091),
    [anon_sym_bool] = ACTIONS(2091),
    [anon_sym_str] = ACTIONS(2091),
    [anon_sym_char] = ACTIONS(2091),
    [aux_sym__non_special_token_token1] = ACTIONS(2091),
    [anon_sym_SQUOTE] = ACTIONS(2091),
    [anon_sym_as] = ACTIONS(2091),
    [anon_sym_async] = ACTIONS(2091),
    [anon_sym_await] = ACTIONS(2091),
    [anon_sym_break] = ACTIONS(2091),
    [anon_sym_const] = ACTIONS(2091),
    [anon_sym_continue] = ACTIONS(2091),
    [anon_sym_default] = ACTIONS(2091),
    [anon_sym_enum] = ACTIONS(2091),
    [anon_sym_fn] = ACTIONS(2091),
    [anon_sym_for] = ACTIONS(2091),
    [anon_sym_if] = ACTIONS(2091),
    [anon_sym_impl] = ACTIONS(2091),
    [anon_sym_let] = ACTIONS(2091),
    [anon_sym_loop] = ACTIONS(2091),
    [anon_sym_match] = ACTIONS(2091),
    [anon_sym_mod] = ACTIONS(2091),
    [anon_sym_pub] = ACTIONS(2091),
    [anon_sym_return] = ACTIONS(2091),
    [anon_sym_static] = ACTIONS(2091),
    [anon_sym_struct] = ACTIONS(2091),
    [anon_sym_trait] = ACTIONS(2091),
    [anon_sym_type] = ACTIONS(2091),
    [anon_sym_union] = ACTIONS(2091),
    [anon_sym_unsafe] = ACTIONS(2091),
    [anon_sym_use] = ACTIONS(2091),
    [anon_sym_where] = ACTIONS(2091),
    [anon_sym_while] = ACTIONS(2091),
    [sym_mutable_specifier] = ACTIONS(2091),
    [sym_integer_literal] = ACTIONS(2014),
    [aux_sym_string_literal_token1] = ACTIONS(2016),
    [sym_char_literal] = ACTIONS(2014),
    [anon_sym_true] = ACTIONS(2018),
    [anon_sym_false] = ACTIONS(2018),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2091),
    [sym_super] = ACTIONS(2091),
    [sym_crate] = ACTIONS(2091),
    [sym_metavariable] = ACTIONS(2020),
    [sym_raw_string_literal] = ACTIONS(2014),
    [sym_float_literal] = ACTIONS(2014),
    [sym_block_comment] = ACTIONS(3),
  },
  [495] = {
    [sym__token_pattern] = STATE(239),
    [sym_token_tree_pattern] = STATE(239),
    [sym_token_binding_pattern] = STATE(239),
    [sym_token_repetition_pattern] = STATE(239),
    [sym__literal] = STATE(239),
    [sym_string_literal] = STATE(585),
    [sym_boolean_literal] = STATE(585),
    [aux_sym_token_tree_pattern_repeat1] = STATE(239),
    [sym_identifier] = ACTIONS(2081),
    [anon_sym_LPAREN] = ACTIONS(2004),
    [anon_sym_LBRACE] = ACTIONS(2006),
    [anon_sym_LBRACK] = ACTIONS(2010),
    [anon_sym_RBRACK] = ACTIONS(2089),
    [anon_sym_DOLLAR] = ACTIONS(2012),
    [anon_sym_u8] = ACTIONS(2081),
    [anon_sym_i8] = ACTIONS(2081),
    [anon_sym_u16] = ACTIONS(2081),
    [anon_sym_i16] = ACTIONS(2081),
    [anon_sym_u32] = ACTIONS(2081),
    [anon_sym_i32] = ACTIONS(2081),
    [anon_sym_u64] = ACTIONS(2081),
    [anon_sym_i64] = ACTIONS(2081),
    [anon_sym_u128] = ACTIONS(2081),
    [anon_sym_i128] = ACTIONS(2081),
    [anon_sym_isize] = ACTIONS(2081),
    [anon_sym_usize] = ACTIONS(2081),
    [anon_sym_f32] = ACTIONS(2081),
    [anon_sym_f64] = ACTIONS(2081),
    [anon_sym_bool] = ACTIONS(2081),
    [anon_sym_str] = ACTIONS(2081),
    [anon_sym_char] = ACTIONS(2081),
    [aux_sym__non_special_token_token1] = ACTIONS(2081),
    [anon_sym_SQUOTE] = ACTIONS(2081),
    [anon_sym_as] = ACTIONS(2081),
    [anon_sym_async] = ACTIONS(2081),
    [anon_sym_await] = ACTIONS(2081),
    [anon_sym_break] = ACTIONS(2081),
    [anon_sym_const] = ACTIONS(2081),
    [anon_sym_continue] = ACTIONS(2081),
    [anon_sym_default] = ACTIONS(2081),
    [anon_sym_enum] = ACTIONS(2081),
    [anon_sym_fn] = ACTIONS(2081),
    [anon_sym_for] = ACTIONS(2081),
    [anon_sym_if] = ACTIONS(2081),
    [anon_sym_impl] = ACTIONS(2081),
    [anon_sym_let] = ACTIONS(2081),
    [anon_sym_loop] = ACTIONS(2081),
    [anon_sym_match] = ACTIONS(2081),
    [anon_sym_mod] = ACTIONS(2081),
    [anon_sym_pub] = ACTIONS(2081),
    [anon_sym_return] = ACTIONS(2081),
    [anon_sym_static] = ACTIONS(2081),
    [anon_sym_struct] = ACTIONS(2081),
    [anon_sym_trait] = ACTIONS(2081),
    [anon_sym_type] = ACTIONS(2081),
    [anon_sym_union] = ACTIONS(2081),
    [anon_sym_unsafe] = ACTIONS(2081),
    [anon_sym_use] = ACTIONS(2081),
    [anon_sym_where] = ACTIONS(2081),
    [anon_sym_while] = ACTIONS(2081),
    [sym_mutable_specifier] = ACTIONS(2081),
    [sym_integer_literal] = ACTIONS(2014),
    [aux_sym_string_literal_token1] = ACTIONS(2016),
    [sym_char_literal] = ACTIONS(2014),
    [anon_sym_true] = ACTIONS(2018),
    [anon_sym_false] = ACTIONS(2018),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2081),
    [sym_super] = ACTIONS(2081),
    [sym_crate] = ACTIONS(2081),
    [sym_metavariable] = ACTIONS(2020),
    [sym_raw_string_literal] = ACTIONS(2014),
    [sym_float_literal] = ACTIONS(2014),
    [sym_block_comment] = ACTIONS(3),
  },
  [496] = {
    [sym__token_pattern] = STATE(489),
    [sym_token_tree_pattern] = STATE(489),
    [sym_token_binding_pattern] = STATE(489),
    [sym_token_repetition_pattern] = STATE(489),
    [sym__literal] = STATE(489),
    [sym_string_literal] = STATE(585),
    [sym_boolean_literal] = STATE(585),
    [aux_sym_token_tree_pattern_repeat1] = STATE(489),
    [sym_identifier] = ACTIONS(2093),
    [anon_sym_LPAREN] = ACTIONS(2004),
    [anon_sym_RPAREN] = ACTIONS(2095),
    [anon_sym_LBRACE] = ACTIONS(2006),
    [anon_sym_LBRACK] = ACTIONS(2010),
    [anon_sym_DOLLAR] = ACTIONS(2012),
    [anon_sym_u8] = ACTIONS(2093),
    [anon_sym_i8] = ACTIONS(2093),
    [anon_sym_u16] = ACTIONS(2093),
    [anon_sym_i16] = ACTIONS(2093),
    [anon_sym_u32] = ACTIONS(2093),
    [anon_sym_i32] = ACTIONS(2093),
    [anon_sym_u64] = ACTIONS(2093),
    [anon_sym_i64] = ACTIONS(2093),
    [anon_sym_u128] = ACTIONS(2093),
    [anon_sym_i128] = ACTIONS(2093),
    [anon_sym_isize] = ACTIONS(2093),
    [anon_sym_usize] = ACTIONS(2093),
    [anon_sym_f32] = ACTIONS(2093),
    [anon_sym_f64] = ACTIONS(2093),
    [anon_sym_bool] = ACTIONS(2093),
    [anon_sym_str] = ACTIONS(2093),
    [anon_sym_char] = ACTIONS(2093),
    [aux_sym__non_special_token_token1] = ACTIONS(2093),
    [anon_sym_SQUOTE] = ACTIONS(2093),
    [anon_sym_as] = ACTIONS(2093),
    [anon_sym_async] = ACTIONS(2093),
    [anon_sym_await] = ACTIONS(2093),
    [anon_sym_break] = ACTIONS(2093),
    [anon_sym_const] = ACTIONS(2093),
    [anon_sym_continue] = ACTIONS(2093),
    [anon_sym_default] = ACTIONS(2093),
    [anon_sym_enum] = ACTIONS(2093),
    [anon_sym_fn] = ACTIONS(2093),
    [anon_sym_for] = ACTIONS(2093),
    [anon_sym_if] = ACTIONS(2093),
    [anon_sym_impl] = ACTIONS(2093),
    [anon_sym_let] = ACTIONS(2093),
    [anon_sym_loop] = ACTIONS(2093),
    [anon_sym_match] = ACTIONS(2093),
    [anon_sym_mod] = ACTIONS(2093),
    [anon_sym_pub] = ACTIONS(2093),
    [anon_sym_return] = ACTIONS(2093),
    [anon_sym_static] = ACTIONS(2093),
    [anon_sym_struct] = ACTIONS(2093),
    [anon_sym_trait] = ACTIONS(2093),
    [anon_sym_type] = ACTIONS(2093),
    [anon_sym_union] = ACTIONS(2093),
    [anon_sym_unsafe] = ACTIONS(2093),
    [anon_sym_use] = ACTIONS(2093),
    [anon_sym_where] = ACTIONS(2093),
    [anon_sym_while] = ACTIONS(2093),
    [sym_mutable_specifier] = ACTIONS(2093),
    [sym_integer_literal] = ACTIONS(2014),
    [aux_sym_string_literal_token1] = ACTIONS(2016),
    [sym_char_literal] = ACTIONS(2014),
    [anon_sym_true] = ACTIONS(2018),
    [anon_sym_false] = ACTIONS(2018),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2093),
    [sym_super] = ACTIONS(2093),
    [sym_crate] = ACTIONS(2093),
    [sym_metavariable] = ACTIONS(2020),
    [sym_raw_string_literal] = ACTIONS(2014),
    [sym_float_literal] = ACTIONS(2014),
    [sym_block_comment] = ACTIONS(3),
  },
  [497] = {
    [sym__token_pattern] = STATE(239),
    [sym_token_tree_pattern] = STATE(239),
    [sym_token_binding_pattern] = STATE(239),
    [sym_token_repetition_pattern] = STATE(239),
    [sym__literal] = STATE(239),
    [sym_string_literal] = STATE(585),
    [sym_boolean_literal] = STATE(585),
    [aux_sym_token_tree_pattern_repeat1] = STATE(239),
    [sym_identifier] = ACTIONS(2081),
    [anon_sym_LPAREN] = ACTIONS(2004),
    [anon_sym_LBRACE] = ACTIONS(2006),
    [anon_sym_RBRACE] = ACTIONS(2089),
    [anon_sym_LBRACK] = ACTIONS(2010),
    [anon_sym_DOLLAR] = ACTIONS(2012),
    [anon_sym_u8] = ACTIONS(2081),
    [anon_sym_i8] = ACTIONS(2081),
    [anon_sym_u16] = ACTIONS(2081),
    [anon_sym_i16] = ACTIONS(2081),
    [anon_sym_u32] = ACTIONS(2081),
    [anon_sym_i32] = ACTIONS(2081),
    [anon_sym_u64] = ACTIONS(2081),
    [anon_sym_i64] = ACTIONS(2081),
    [anon_sym_u128] = ACTIONS(2081),
    [anon_sym_i128] = ACTIONS(2081),
    [anon_sym_isize] = ACTIONS(2081),
    [anon_sym_usize] = ACTIONS(2081),
    [anon_sym_f32] = ACTIONS(2081),
    [anon_sym_f64] = ACTIONS(2081),
    [anon_sym_bool] = ACTIONS(2081),
    [anon_sym_str] = ACTIONS(2081),
    [anon_sym_char] = ACTIONS(2081),
    [aux_sym__non_special_token_token1] = ACTIONS(2081),
    [anon_sym_SQUOTE] = ACTIONS(2081),
    [anon_sym_as] = ACTIONS(2081),
    [anon_sym_async] = ACTIONS(2081),
    [anon_sym_await] = ACTIONS(2081),
    [anon_sym_break] = ACTIONS(2081),
    [anon_sym_const] = ACTIONS(2081),
    [anon_sym_continue] = ACTIONS(2081),
    [anon_sym_default] = ACTIONS(2081),
    [anon_sym_enum] = ACTIONS(2081),
    [anon_sym_fn] = ACTIONS(2081),
    [anon_sym_for] = ACTIONS(2081),
    [anon_sym_if] = ACTIONS(2081),
    [anon_sym_impl] = ACTIONS(2081),
    [anon_sym_let] = ACTIONS(2081),
    [anon_sym_loop] = ACTIONS(2081),
    [anon_sym_match] = ACTIONS(2081),
    [anon_sym_mod] = ACTIONS(2081),
    [anon_sym_pub] = ACTIONS(2081),
    [anon_sym_return] = ACTIONS(2081),
    [anon_sym_static] = ACTIONS(2081),
    [anon_sym_struct] = ACTIONS(2081),
    [anon_sym_trait] = ACTIONS(2081),
    [anon_sym_type] = ACTIONS(2081),
    [anon_sym_union] = ACTIONS(2081),
    [anon_sym_unsafe] = ACTIONS(2081),
    [anon_sym_use] = ACTIONS(2081),
    [anon_sym_where] = ACTIONS(2081),
    [anon_sym_while] = ACTIONS(2081),
    [sym_mutable_specifier] = ACTIONS(2081),
    [sym_integer_literal] = ACTIONS(2014),
    [aux_sym_string_literal_token1] = ACTIONS(2016),
    [sym_char_literal] = ACTIONS(2014),
    [anon_sym_true] = ACTIONS(2018),
    [anon_sym_false] = ACTIONS(2018),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2081),
    [sym_super] = ACTIONS(2081),
    [sym_crate] = ACTIONS(2081),
    [sym_metavariable] = ACTIONS(2020),
    [sym_raw_string_literal] = ACTIONS(2014),
    [sym_float_literal] = ACTIONS(2014),
    [sym_block_comment] = ACTIONS(3),
  },
  [498] = {
    [sym__token_pattern] = STATE(488),
    [sym_token_tree_pattern] = STATE(488),
    [sym_token_binding_pattern] = STATE(488),
    [sym_token_repetition_pattern] = STATE(488),
    [sym__literal] = STATE(488),
    [sym_string_literal] = STATE(585),
    [sym_boolean_literal] = STATE(585),
    [aux_sym_token_tree_pattern_repeat1] = STATE(488),
    [sym_identifier] = ACTIONS(2097),
    [anon_sym_LPAREN] = ACTIONS(2004),
    [anon_sym_LBRACE] = ACTIONS(2006),
    [anon_sym_RBRACE] = ACTIONS(2095),
    [anon_sym_LBRACK] = ACTIONS(2010),
    [anon_sym_DOLLAR] = ACTIONS(2012),
    [anon_sym_u8] = ACTIONS(2097),
    [anon_sym_i8] = ACTIONS(2097),
    [anon_sym_u16] = ACTIONS(2097),
    [anon_sym_i16] = ACTIONS(2097),
    [anon_sym_u32] = ACTIONS(2097),
    [anon_sym_i32] = ACTIONS(2097),
    [anon_sym_u64] = ACTIONS(2097),
    [anon_sym_i64] = ACTIONS(2097),
    [anon_sym_u128] = ACTIONS(2097),
    [anon_sym_i128] = ACTIONS(2097),
    [anon_sym_isize] = ACTIONS(2097),
    [anon_sym_usize] = ACTIONS(2097),
    [anon_sym_f32] = ACTIONS(2097),
    [anon_sym_f64] = ACTIONS(2097),
    [anon_sym_bool] = ACTIONS(2097),
    [anon_sym_str] = ACTIONS(2097),
    [anon_sym_char] = ACTIONS(2097),
    [aux_sym__non_special_token_token1] = ACTIONS(2097),
    [anon_sym_SQUOTE] = ACTIONS(2097),
    [anon_sym_as] = ACTIONS(2097),
    [anon_sym_async] = ACTIONS(2097),
    [anon_sym_await] = ACTIONS(2097),
    [anon_sym_break] = ACTIONS(2097),
    [anon_sym_const] = ACTIONS(2097),
    [anon_sym_continue] = ACTIONS(2097),
    [anon_sym_default] = ACTIONS(2097),
    [anon_sym_enum] = ACTIONS(2097),
    [anon_sym_fn] = ACTIONS(2097),
    [anon_sym_for] = ACTIONS(2097),
    [anon_sym_if] = ACTIONS(2097),
    [anon_sym_impl] = ACTIONS(2097),
    [anon_sym_let] = ACTIONS(2097),
    [anon_sym_loop] = ACTIONS(2097),
    [anon_sym_match] = ACTIONS(2097),
    [anon_sym_mod] = ACTIONS(2097),
    [anon_sym_pub] = ACTIONS(2097),
    [anon_sym_return] = ACTIONS(2097),
    [anon_sym_static] = ACTIONS(2097),
    [anon_sym_struct] = ACTIONS(2097),
    [anon_sym_trait] = ACTIONS(2097),
    [anon_sym_type] = ACTIONS(2097),
    [anon_sym_union] = ACTIONS(2097),
    [anon_sym_unsafe] = ACTIONS(2097),
    [anon_sym_use] = ACTIONS(2097),
    [anon_sym_where] = ACTIONS(2097),
    [anon_sym_while] = ACTIONS(2097),
    [sym_mutable_specifier] = ACTIONS(2097),
    [sym_integer_literal] = ACTIONS(2014),
    [aux_sym_string_literal_token1] = ACTIONS(2016),
    [sym_char_literal] = ACTIONS(2014),
    [anon_sym_true] = ACTIONS(2018),
    [anon_sym_false] = ACTIONS(2018),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2097),
    [sym_super] = ACTIONS(2097),
    [sym_crate] = ACTIONS(2097),
    [sym_metavariable] = ACTIONS(2020),
    [sym_raw_string_literal] = ACTIONS(2014),
    [sym_float_literal] = ACTIONS(2014),
    [sym_block_comment] = ACTIONS(3),
  },
  [499] = {
    [sym__token_pattern] = STATE(487),
    [sym_token_tree_pattern] = STATE(487),
    [sym_token_binding_pattern] = STATE(487),
    [sym_token_repetition_pattern] = STATE(487),
    [sym__literal] = STATE(487),
    [sym_string_literal] = STATE(585),
    [sym_boolean_literal] = STATE(585),
    [aux_sym_token_tree_pattern_repeat1] = STATE(487),
    [sym_identifier] = ACTIONS(2099),
    [anon_sym_LPAREN] = ACTIONS(2004),
    [anon_sym_LBRACE] = ACTIONS(2006),
    [anon_sym_LBRACK] = ACTIONS(2010),
    [anon_sym_RBRACK] = ACTIONS(2095),
    [anon_sym_DOLLAR] = ACTIONS(2012),
    [anon_sym_u8] = ACTIONS(2099),
    [anon_sym_i8] = ACTIONS(2099),
    [anon_sym_u16] = ACTIONS(2099),
    [anon_sym_i16] = ACTIONS(2099),
    [anon_sym_u32] = ACTIONS(2099),
    [anon_sym_i32] = ACTIONS(2099),
    [anon_sym_u64] = ACTIONS(2099),
    [anon_sym_i64] = ACTIONS(2099),
    [anon_sym_u128] = ACTIONS(2099),
    [anon_sym_i128] = ACTIONS(2099),
    [anon_sym_isize] = ACTIONS(2099),
    [anon_sym_usize] = ACTIONS(2099),
    [anon_sym_f32] = ACTIONS(2099),
    [anon_sym_f64] = ACTIONS(2099),
    [anon_sym_bool] = ACTIONS(2099),
    [anon_sym_str] = ACTIONS(2099),
    [anon_sym_char] = ACTIONS(2099),
    [aux_sym__non_special_token_token1] = ACTIONS(2099),
    [anon_sym_SQUOTE] = ACTIONS(2099),
    [anon_sym_as] = ACTIONS(2099),
    [anon_sym_async] = ACTIONS(2099),
    [anon_sym_await] = ACTIONS(2099),
    [anon_sym_break] = ACTIONS(2099),
    [anon_sym_const] = ACTIONS(2099),
    [anon_sym_continue] = ACTIONS(2099),
    [anon_sym_default] = ACTIONS(2099),
    [anon_sym_enum] = ACTIONS(2099),
    [anon_sym_fn] = ACTIONS(2099),
    [anon_sym_for] = ACTIONS(2099),
    [anon_sym_if] = ACTIONS(2099),
    [anon_sym_impl] = ACTIONS(2099),
    [anon_sym_let] = ACTIONS(2099),
    [anon_sym_loop] = ACTIONS(2099),
    [anon_sym_match] = ACTIONS(2099),
    [anon_sym_mod] = ACTIONS(2099),
    [anon_sym_pub] = ACTIONS(2099),
    [anon_sym_return] = ACTIONS(2099),
    [anon_sym_static] = ACTIONS(2099),
    [anon_sym_struct] = ACTIONS(2099),
    [anon_sym_trait] = ACTIONS(2099),
    [anon_sym_type] = ACTIONS(2099),
    [anon_sym_union] = ACTIONS(2099),
    [anon_sym_unsafe] = ACTIONS(2099),
    [anon_sym_use] = ACTIONS(2099),
    [anon_sym_where] = ACTIONS(2099),
    [anon_sym_while] = ACTIONS(2099),
    [sym_mutable_specifier] = ACTIONS(2099),
    [sym_integer_literal] = ACTIONS(2014),
    [aux_sym_string_literal_token1] = ACTIONS(2016),
    [sym_char_literal] = ACTIONS(2014),
    [anon_sym_true] = ACTIONS(2018),
    [anon_sym_false] = ACTIONS(2018),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2099),
    [sym_super] = ACTIONS(2099),
    [sym_crate] = ACTIONS(2099),
    [sym_metavariable] = ACTIONS(2020),
    [sym_raw_string_literal] = ACTIONS(2014),
    [sym_float_literal] = ACTIONS(2014),
    [sym_block_comment] = ACTIONS(3),
  },
  [500] = {
    [sym_attribute_item] = STATE(546),
    [sym_bracketed_type] = STATE(2489),
    [sym_generic_type] = STATE(2488),
    [sym_generic_type_with_turbofish] = STATE(2487),
    [sym_macro_invocation] = STATE(1422),
    [sym_scoped_identifier] = STATE(1337),
    [sym_scoped_type_identifier] = STATE(2013),
    [sym_match_arm] = STATE(500),
    [sym_match_pattern] = STATE(2548),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1981),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [aux_sym_enum_variant_list_repeat1] = STATE(546),
    [aux_sym_match_block_repeat1] = STATE(500),
    [sym_identifier] = ACTIONS(2101),
    [anon_sym_LPAREN] = ACTIONS(2104),
    [anon_sym_LBRACK] = ACTIONS(2107),
    [anon_sym_u8] = ACTIONS(2110),
    [anon_sym_i8] = ACTIONS(2110),
    [anon_sym_u16] = ACTIONS(2110),
    [anon_sym_i16] = ACTIONS(2110),
    [anon_sym_u32] = ACTIONS(2110),
    [anon_sym_i32] = ACTIONS(2110),
    [anon_sym_u64] = ACTIONS(2110),
    [anon_sym_i64] = ACTIONS(2110),
    [anon_sym_u128] = ACTIONS(2110),
    [anon_sym_i128] = ACTIONS(2110),
    [anon_sym_isize] = ACTIONS(2110),
    [anon_sym_usize] = ACTIONS(2110),
    [anon_sym_f32] = ACTIONS(2110),
    [anon_sym_f64] = ACTIONS(2110),
    [anon_sym_bool] = ACTIONS(2110),
    [anon_sym_str] = ACTIONS(2110),
    [anon_sym_char] = ACTIONS(2110),
    [anon_sym_const] = ACTIONS(2113),
    [anon_sym_default] = ACTIONS(2116),
    [anon_sym_union] = ACTIONS(2116),
    [anon_sym_POUND] = ACTIONS(2119),
    [anon_sym_ref] = ACTIONS(2122),
    [anon_sym_LT] = ACTIONS(2125),
    [anon_sym_COLON_COLON] = ACTIONS(2128),
    [anon_sym__] = ACTIONS(2131),
    [anon_sym_AMP] = ACTIONS(2134),
    [sym_mutable_specifier] = ACTIONS(2137),
    [anon_sym_DOT_DOT] = ACTIONS(2140),
    [anon_sym_DASH] = ACTIONS(2143),
    [sym_integer_literal] = ACTIONS(2146),
    [aux_sym_string_literal_token1] = ACTIONS(2149),
    [sym_char_literal] = ACTIONS(2146),
    [anon_sym_true] = ACTIONS(2152),
    [anon_sym_false] = ACTIONS(2152),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(2155),
    [sym_super] = ACTIONS(2155),
    [sym_crate] = ACTIONS(2155),
    [sym_metavariable] = ACTIONS(2158),
    [sym_raw_string_literal] = ACTIONS(2146),
    [sym_float_literal] = ACTIONS(2146),
    [sym_block_comment] = ACTIONS(3),
  },
  [501] = {
    [sym_delim_token_tree] = STATE(483),
    [sym__delim_tokens] = STATE(483),
    [sym__non_delim_token] = STATE(483),
    [sym__literal] = STATE(483),
    [sym_string_literal] = STATE(592),
    [sym_boolean_literal] = STATE(592),
    [aux_sym_delim_token_tree_repeat1] = STATE(483),
    [sym_identifier] = ACTIONS(2161),
    [anon_sym_LPAREN] = ACTIONS(2163),
    [anon_sym_LBRACE] = ACTIONS(2165),
    [anon_sym_RBRACE] = ACTIONS(2167),
    [anon_sym_LBRACK] = ACTIONS(2169),
    [anon_sym_DOLLAR] = ACTIONS(2171),
    [anon_sym_u8] = ACTIONS(2161),
    [anon_sym_i8] = ACTIONS(2161),
    [anon_sym_u16] = ACTIONS(2161),
    [anon_sym_i16] = ACTIONS(2161),
    [anon_sym_u32] = ACTIONS(2161),
    [anon_sym_i32] = ACTIONS(2161),
    [anon_sym_u64] = ACTIONS(2161),
    [anon_sym_i64] = ACTIONS(2161),
    [anon_sym_u128] = ACTIONS(2161),
    [anon_sym_i128] = ACTIONS(2161),
    [anon_sym_isize] = ACTIONS(2161),
    [anon_sym_usize] = ACTIONS(2161),
    [anon_sym_f32] = ACTIONS(2161),
    [anon_sym_f64] = ACTIONS(2161),
    [anon_sym_bool] = ACTIONS(2161),
    [anon_sym_str] = ACTIONS(2161),
    [anon_sym_char] = ACTIONS(2161),
    [aux_sym__non_special_token_token1] = ACTIONS(2161),
    [anon_sym_SQUOTE] = ACTIONS(2161),
    [anon_sym_as] = ACTIONS(2161),
    [anon_sym_async] = ACTIONS(2161),
    [anon_sym_await] = ACTIONS(2161),
    [anon_sym_break] = ACTIONS(2161),
    [anon_sym_const] = ACTIONS(2161),
    [anon_sym_continue] = ACTIONS(2161),
    [anon_sym_default] = ACTIONS(2161),
    [anon_sym_enum] = ACTIONS(2161),
    [anon_sym_fn] = ACTIONS(2161),
    [anon_sym_for] = ACTIONS(2161),
    [anon_sym_if] = ACTIONS(2161),
    [anon_sym_impl] = ACTIONS(2161),
    [anon_sym_let] = ACTIONS(2161),
    [anon_sym_loop] = ACTIONS(2161),
    [anon_sym_match] = ACTIONS(2161),
    [anon_sym_mod] = ACTIONS(2161),
    [anon_sym_pub] = ACTIONS(2161),
    [anon_sym_return] = ACTIONS(2161),
    [anon_sym_static] = ACTIONS(2161),
    [anon_sym_struct] = ACTIONS(2161),
    [anon_sym_trait] = ACTIONS(2161),
    [anon_sym_type] = ACTIONS(2161),
    [anon_sym_union] = ACTIONS(2161),
    [anon_sym_unsafe] = ACTIONS(2161),
    [anon_sym_use] = ACTIONS(2161),
    [anon_sym_where] = ACTIONS(2161),
    [anon_sym_while] = ACTIONS(2161),
    [sym_mutable_specifier] = ACTIONS(2161),
    [sym_integer_literal] = ACTIONS(2173),
    [aux_sym_string_literal_token1] = ACTIONS(2175),
    [sym_char_literal] = ACTIONS(2173),
    [anon_sym_true] = ACTIONS(2177),
    [anon_sym_false] = ACTIONS(2177),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2161),
    [sym_super] = ACTIONS(2161),
    [sym_crate] = ACTIONS(2161),
    [sym_raw_string_literal] = ACTIONS(2173),
    [sym_float_literal] = ACTIONS(2173),
    [sym_block_comment] = ACTIONS(3),
  },
  [502] = {
    [sym_delim_token_tree] = STATE(483),
    [sym__delim_tokens] = STATE(483),
    [sym__non_delim_token] = STATE(483),
    [sym__literal] = STATE(483),
    [sym_string_literal] = STATE(592),
    [sym_boolean_literal] = STATE(592),
    [aux_sym_delim_token_tree_repeat1] = STATE(483),
    [sym_identifier] = ACTIONS(2161),
    [anon_sym_LPAREN] = ACTIONS(2163),
    [anon_sym_LBRACE] = ACTIONS(2165),
    [anon_sym_LBRACK] = ACTIONS(2169),
    [anon_sym_RBRACK] = ACTIONS(2179),
    [anon_sym_DOLLAR] = ACTIONS(2171),
    [anon_sym_u8] = ACTIONS(2161),
    [anon_sym_i8] = ACTIONS(2161),
    [anon_sym_u16] = ACTIONS(2161),
    [anon_sym_i16] = ACTIONS(2161),
    [anon_sym_u32] = ACTIONS(2161),
    [anon_sym_i32] = ACTIONS(2161),
    [anon_sym_u64] = ACTIONS(2161),
    [anon_sym_i64] = ACTIONS(2161),
    [anon_sym_u128] = ACTIONS(2161),
    [anon_sym_i128] = ACTIONS(2161),
    [anon_sym_isize] = ACTIONS(2161),
    [anon_sym_usize] = ACTIONS(2161),
    [anon_sym_f32] = ACTIONS(2161),
    [anon_sym_f64] = ACTIONS(2161),
    [anon_sym_bool] = ACTIONS(2161),
    [anon_sym_str] = ACTIONS(2161),
    [anon_sym_char] = ACTIONS(2161),
    [aux_sym__non_special_token_token1] = ACTIONS(2161),
    [anon_sym_SQUOTE] = ACTIONS(2161),
    [anon_sym_as] = ACTIONS(2161),
    [anon_sym_async] = ACTIONS(2161),
    [anon_sym_await] = ACTIONS(2161),
    [anon_sym_break] = ACTIONS(2161),
    [anon_sym_const] = ACTIONS(2161),
    [anon_sym_continue] = ACTIONS(2161),
    [anon_sym_default] = ACTIONS(2161),
    [anon_sym_enum] = ACTIONS(2161),
    [anon_sym_fn] = ACTIONS(2161),
    [anon_sym_for] = ACTIONS(2161),
    [anon_sym_if] = ACTIONS(2161),
    [anon_sym_impl] = ACTIONS(2161),
    [anon_sym_let] = ACTIONS(2161),
    [anon_sym_loop] = ACTIONS(2161),
    [anon_sym_match] = ACTIONS(2161),
    [anon_sym_mod] = ACTIONS(2161),
    [anon_sym_pub] = ACTIONS(2161),
    [anon_sym_return] = ACTIONS(2161),
    [anon_sym_static] = ACTIONS(2161),
    [anon_sym_struct] = ACTIONS(2161),
    [anon_sym_trait] = ACTIONS(2161),
    [anon_sym_type] = ACTIONS(2161),
    [anon_sym_union] = ACTIONS(2161),
    [anon_sym_unsafe] = ACTIONS(2161),
    [anon_sym_use] = ACTIONS(2161),
    [anon_sym_where] = ACTIONS(2161),
    [anon_sym_while] = ACTIONS(2161),
    [sym_mutable_specifier] = ACTIONS(2161),
    [sym_integer_literal] = ACTIONS(2173),
    [aux_sym_string_literal_token1] = ACTIONS(2175),
    [sym_char_literal] = ACTIONS(2173),
    [anon_sym_true] = ACTIONS(2177),
    [anon_sym_false] = ACTIONS(2177),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2161),
    [sym_super] = ACTIONS(2161),
    [sym_crate] = ACTIONS(2161),
    [sym_raw_string_literal] = ACTIONS(2173),
    [sym_float_literal] = ACTIONS(2173),
    [sym_block_comment] = ACTIONS(3),
  },
  [503] = {
    [sym_token_tree] = STATE(485),
    [sym_token_repetition] = STATE(485),
    [sym__literal] = STATE(485),
    [sym_string_literal] = STATE(585),
    [sym_boolean_literal] = STATE(585),
    [aux_sym_token_tree_repeat1] = STATE(485),
    [sym_identifier] = ACTIONS(2181),
    [anon_sym_LPAREN] = ACTIONS(2183),
    [anon_sym_LBRACE] = ACTIONS(2185),
    [anon_sym_RBRACE] = ACTIONS(2187),
    [anon_sym_LBRACK] = ACTIONS(2189),
    [anon_sym_DOLLAR] = ACTIONS(2191),
    [anon_sym_u8] = ACTIONS(2181),
    [anon_sym_i8] = ACTIONS(2181),
    [anon_sym_u16] = ACTIONS(2181),
    [anon_sym_i16] = ACTIONS(2181),
    [anon_sym_u32] = ACTIONS(2181),
    [anon_sym_i32] = ACTIONS(2181),
    [anon_sym_u64] = ACTIONS(2181),
    [anon_sym_i64] = ACTIONS(2181),
    [anon_sym_u128] = ACTIONS(2181),
    [anon_sym_i128] = ACTIONS(2181),
    [anon_sym_isize] = ACTIONS(2181),
    [anon_sym_usize] = ACTIONS(2181),
    [anon_sym_f32] = ACTIONS(2181),
    [anon_sym_f64] = ACTIONS(2181),
    [anon_sym_bool] = ACTIONS(2181),
    [anon_sym_str] = ACTIONS(2181),
    [anon_sym_char] = ACTIONS(2181),
    [aux_sym__non_special_token_token1] = ACTIONS(2181),
    [anon_sym_SQUOTE] = ACTIONS(2181),
    [anon_sym_as] = ACTIONS(2181),
    [anon_sym_async] = ACTIONS(2181),
    [anon_sym_await] = ACTIONS(2181),
    [anon_sym_break] = ACTIONS(2181),
    [anon_sym_const] = ACTIONS(2181),
    [anon_sym_continue] = ACTIONS(2181),
    [anon_sym_default] = ACTIONS(2181),
    [anon_sym_enum] = ACTIONS(2181),
    [anon_sym_fn] = ACTIONS(2181),
    [anon_sym_for] = ACTIONS(2181),
    [anon_sym_if] = ACTIONS(2181),
    [anon_sym_impl] = ACTIONS(2181),
    [anon_sym_let] = ACTIONS(2181),
    [anon_sym_loop] = ACTIONS(2181),
    [anon_sym_match] = ACTIONS(2181),
    [anon_sym_mod] = ACTIONS(2181),
    [anon_sym_pub] = ACTIONS(2181),
    [anon_sym_return] = ACTIONS(2181),
    [anon_sym_static] = ACTIONS(2181),
    [anon_sym_struct] = ACTIONS(2181),
    [anon_sym_trait] = ACTIONS(2181),
    [anon_sym_type] = ACTIONS(2181),
    [anon_sym_union] = ACTIONS(2181),
    [anon_sym_unsafe] = ACTIONS(2181),
    [anon_sym_use] = ACTIONS(2181),
    [anon_sym_where] = ACTIONS(2181),
    [anon_sym_while] = ACTIONS(2181),
    [sym_mutable_specifier] = ACTIONS(2181),
    [sym_integer_literal] = ACTIONS(2014),
    [aux_sym_string_literal_token1] = ACTIONS(2016),
    [sym_char_literal] = ACTIONS(2014),
    [anon_sym_true] = ACTIONS(2018),
    [anon_sym_false] = ACTIONS(2018),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2181),
    [sym_super] = ACTIONS(2181),
    [sym_crate] = ACTIONS(2181),
    [sym_metavariable] = ACTIONS(2193),
    [sym_raw_string_literal] = ACTIONS(2014),
    [sym_float_literal] = ACTIONS(2014),
    [sym_block_comment] = ACTIONS(3),
  },
  [504] = {
    [sym_delim_token_tree] = STATE(483),
    [sym__delim_tokens] = STATE(483),
    [sym__non_delim_token] = STATE(483),
    [sym__literal] = STATE(483),
    [sym_string_literal] = STATE(592),
    [sym_boolean_literal] = STATE(592),
    [aux_sym_delim_token_tree_repeat1] = STATE(483),
    [sym_identifier] = ACTIONS(2161),
    [anon_sym_LPAREN] = ACTIONS(2163),
    [anon_sym_RPAREN] = ACTIONS(2195),
    [anon_sym_LBRACE] = ACTIONS(2165),
    [anon_sym_LBRACK] = ACTIONS(2169),
    [anon_sym_DOLLAR] = ACTIONS(2171),
    [anon_sym_u8] = ACTIONS(2161),
    [anon_sym_i8] = ACTIONS(2161),
    [anon_sym_u16] = ACTIONS(2161),
    [anon_sym_i16] = ACTIONS(2161),
    [anon_sym_u32] = ACTIONS(2161),
    [anon_sym_i32] = ACTIONS(2161),
    [anon_sym_u64] = ACTIONS(2161),
    [anon_sym_i64] = ACTIONS(2161),
    [anon_sym_u128] = ACTIONS(2161),
    [anon_sym_i128] = ACTIONS(2161),
    [anon_sym_isize] = ACTIONS(2161),
    [anon_sym_usize] = ACTIONS(2161),
    [anon_sym_f32] = ACTIONS(2161),
    [anon_sym_f64] = ACTIONS(2161),
    [anon_sym_bool] = ACTIONS(2161),
    [anon_sym_str] = ACTIONS(2161),
    [anon_sym_char] = ACTIONS(2161),
    [aux_sym__non_special_token_token1] = ACTIONS(2161),
    [anon_sym_SQUOTE] = ACTIONS(2161),
    [anon_sym_as] = ACTIONS(2161),
    [anon_sym_async] = ACTIONS(2161),
    [anon_sym_await] = ACTIONS(2161),
    [anon_sym_break] = ACTIONS(2161),
    [anon_sym_const] = ACTIONS(2161),
    [anon_sym_continue] = ACTIONS(2161),
    [anon_sym_default] = ACTIONS(2161),
    [anon_sym_enum] = ACTIONS(2161),
    [anon_sym_fn] = ACTIONS(2161),
    [anon_sym_for] = ACTIONS(2161),
    [anon_sym_if] = ACTIONS(2161),
    [anon_sym_impl] = ACTIONS(2161),
    [anon_sym_let] = ACTIONS(2161),
    [anon_sym_loop] = ACTIONS(2161),
    [anon_sym_match] = ACTIONS(2161),
    [anon_sym_mod] = ACTIONS(2161),
    [anon_sym_pub] = ACTIONS(2161),
    [anon_sym_return] = ACTIONS(2161),
    [anon_sym_static] = ACTIONS(2161),
    [anon_sym_struct] = ACTIONS(2161),
    [anon_sym_trait] = ACTIONS(2161),
    [anon_sym_type] = ACTIONS(2161),
    [anon_sym_union] = ACTIONS(2161),
    [anon_sym_unsafe] = ACTIONS(2161),
    [anon_sym_use] = ACTIONS(2161),
    [anon_sym_where] = ACTIONS(2161),
    [anon_sym_while] = ACTIONS(2161),
    [sym_mutable_specifier] = ACTIONS(2161),
    [sym_integer_literal] = ACTIONS(2173),
    [aux_sym_string_literal_token1] = ACTIONS(2175),
    [sym_char_literal] = ACTIONS(2173),
    [anon_sym_true] = ACTIONS(2177),
    [anon_sym_false] = ACTIONS(2177),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2161),
    [sym_super] = ACTIONS(2161),
    [sym_crate] = ACTIONS(2161),
    [sym_raw_string_literal] = ACTIONS(2173),
    [sym_float_literal] = ACTIONS(2173),
    [sym_block_comment] = ACTIONS(3),
  },
  [505] = {
    [sym_delim_token_tree] = STATE(513),
    [sym__delim_tokens] = STATE(513),
    [sym__non_delim_token] = STATE(513),
    [sym__literal] = STATE(513),
    [sym_string_literal] = STATE(592),
    [sym_boolean_literal] = STATE(592),
    [aux_sym_delim_token_tree_repeat1] = STATE(513),
    [sym_identifier] = ACTIONS(2197),
    [anon_sym_LPAREN] = ACTIONS(2163),
    [anon_sym_LBRACE] = ACTIONS(2165),
    [anon_sym_LBRACK] = ACTIONS(2169),
    [anon_sym_RBRACK] = ACTIONS(2199),
    [anon_sym_DOLLAR] = ACTIONS(2201),
    [anon_sym_u8] = ACTIONS(2197),
    [anon_sym_i8] = ACTIONS(2197),
    [anon_sym_u16] = ACTIONS(2197),
    [anon_sym_i16] = ACTIONS(2197),
    [anon_sym_u32] = ACTIONS(2197),
    [anon_sym_i32] = ACTIONS(2197),
    [anon_sym_u64] = ACTIONS(2197),
    [anon_sym_i64] = ACTIONS(2197),
    [anon_sym_u128] = ACTIONS(2197),
    [anon_sym_i128] = ACTIONS(2197),
    [anon_sym_isize] = ACTIONS(2197),
    [anon_sym_usize] = ACTIONS(2197),
    [anon_sym_f32] = ACTIONS(2197),
    [anon_sym_f64] = ACTIONS(2197),
    [anon_sym_bool] = ACTIONS(2197),
    [anon_sym_str] = ACTIONS(2197),
    [anon_sym_char] = ACTIONS(2197),
    [aux_sym__non_special_token_token1] = ACTIONS(2197),
    [anon_sym_SQUOTE] = ACTIONS(2197),
    [anon_sym_as] = ACTIONS(2197),
    [anon_sym_async] = ACTIONS(2197),
    [anon_sym_await] = ACTIONS(2197),
    [anon_sym_break] = ACTIONS(2197),
    [anon_sym_const] = ACTIONS(2197),
    [anon_sym_continue] = ACTIONS(2197),
    [anon_sym_default] = ACTIONS(2197),
    [anon_sym_enum] = ACTIONS(2197),
    [anon_sym_fn] = ACTIONS(2197),
    [anon_sym_for] = ACTIONS(2197),
    [anon_sym_if] = ACTIONS(2197),
    [anon_sym_impl] = ACTIONS(2197),
    [anon_sym_let] = ACTIONS(2197),
    [anon_sym_loop] = ACTIONS(2197),
    [anon_sym_match] = ACTIONS(2197),
    [anon_sym_mod] = ACTIONS(2197),
    [anon_sym_pub] = ACTIONS(2197),
    [anon_sym_return] = ACTIONS(2197),
    [anon_sym_static] = ACTIONS(2197),
    [anon_sym_struct] = ACTIONS(2197),
    [anon_sym_trait] = ACTIONS(2197),
    [anon_sym_type] = ACTIONS(2197),
    [anon_sym_union] = ACTIONS(2197),
    [anon_sym_unsafe] = ACTIONS(2197),
    [anon_sym_use] = ACTIONS(2197),
    [anon_sym_where] = ACTIONS(2197),
    [anon_sym_while] = ACTIONS(2197),
    [sym_mutable_specifier] = ACTIONS(2197),
    [sym_integer_literal] = ACTIONS(2173),
    [aux_sym_string_literal_token1] = ACTIONS(2175),
    [sym_char_literal] = ACTIONS(2173),
    [anon_sym_true] = ACTIONS(2177),
    [anon_sym_false] = ACTIONS(2177),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2197),
    [sym_super] = ACTIONS(2197),
    [sym_crate] = ACTIONS(2197),
    [sym_raw_string_literal] = ACTIONS(2173),
    [sym_float_literal] = ACTIONS(2173),
    [sym_block_comment] = ACTIONS(3),
  },
  [506] = {
    [sym_token_tree] = STATE(485),
    [sym_token_repetition] = STATE(485),
    [sym__literal] = STATE(485),
    [sym_string_literal] = STATE(585),
    [sym_boolean_literal] = STATE(585),
    [aux_sym_token_tree_repeat1] = STATE(485),
    [sym_identifier] = ACTIONS(2181),
    [anon_sym_LPAREN] = ACTIONS(2183),
    [anon_sym_RPAREN] = ACTIONS(2187),
    [anon_sym_LBRACE] = ACTIONS(2185),
    [anon_sym_LBRACK] = ACTIONS(2189),
    [anon_sym_DOLLAR] = ACTIONS(2191),
    [anon_sym_u8] = ACTIONS(2181),
    [anon_sym_i8] = ACTIONS(2181),
    [anon_sym_u16] = ACTIONS(2181),
    [anon_sym_i16] = ACTIONS(2181),
    [anon_sym_u32] = ACTIONS(2181),
    [anon_sym_i32] = ACTIONS(2181),
    [anon_sym_u64] = ACTIONS(2181),
    [anon_sym_i64] = ACTIONS(2181),
    [anon_sym_u128] = ACTIONS(2181),
    [anon_sym_i128] = ACTIONS(2181),
    [anon_sym_isize] = ACTIONS(2181),
    [anon_sym_usize] = ACTIONS(2181),
    [anon_sym_f32] = ACTIONS(2181),
    [anon_sym_f64] = ACTIONS(2181),
    [anon_sym_bool] = ACTIONS(2181),
    [anon_sym_str] = ACTIONS(2181),
    [anon_sym_char] = ACTIONS(2181),
    [aux_sym__non_special_token_token1] = ACTIONS(2181),
    [anon_sym_SQUOTE] = ACTIONS(2181),
    [anon_sym_as] = ACTIONS(2181),
    [anon_sym_async] = ACTIONS(2181),
    [anon_sym_await] = ACTIONS(2181),
    [anon_sym_break] = ACTIONS(2181),
    [anon_sym_const] = ACTIONS(2181),
    [anon_sym_continue] = ACTIONS(2181),
    [anon_sym_default] = ACTIONS(2181),
    [anon_sym_enum] = ACTIONS(2181),
    [anon_sym_fn] = ACTIONS(2181),
    [anon_sym_for] = ACTIONS(2181),
    [anon_sym_if] = ACTIONS(2181),
    [anon_sym_impl] = ACTIONS(2181),
    [anon_sym_let] = ACTIONS(2181),
    [anon_sym_loop] = ACTIONS(2181),
    [anon_sym_match] = ACTIONS(2181),
    [anon_sym_mod] = ACTIONS(2181),
    [anon_sym_pub] = ACTIONS(2181),
    [anon_sym_return] = ACTIONS(2181),
    [anon_sym_static] = ACTIONS(2181),
    [anon_sym_struct] = ACTIONS(2181),
    [anon_sym_trait] = ACTIONS(2181),
    [anon_sym_type] = ACTIONS(2181),
    [anon_sym_union] = ACTIONS(2181),
    [anon_sym_unsafe] = ACTIONS(2181),
    [anon_sym_use] = ACTIONS(2181),
    [anon_sym_where] = ACTIONS(2181),
    [anon_sym_while] = ACTIONS(2181),
    [sym_mutable_specifier] = ACTIONS(2181),
    [sym_integer_literal] = ACTIONS(2014),
    [aux_sym_string_literal_token1] = ACTIONS(2016),
    [sym_char_literal] = ACTIONS(2014),
    [anon_sym_true] = ACTIONS(2018),
    [anon_sym_false] = ACTIONS(2018),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2181),
    [sym_super] = ACTIONS(2181),
    [sym_crate] = ACTIONS(2181),
    [sym_metavariable] = ACTIONS(2193),
    [sym_raw_string_literal] = ACTIONS(2014),
    [sym_float_literal] = ACTIONS(2014),
    [sym_block_comment] = ACTIONS(3),
  },
  [507] = {
    [sym_delim_token_tree] = STATE(519),
    [sym__delim_tokens] = STATE(519),
    [sym__non_delim_token] = STATE(519),
    [sym__literal] = STATE(519),
    [sym_string_literal] = STATE(592),
    [sym_boolean_literal] = STATE(592),
    [aux_sym_delim_token_tree_repeat1] = STATE(519),
    [sym_identifier] = ACTIONS(2203),
    [anon_sym_LPAREN] = ACTIONS(2163),
    [anon_sym_LBRACE] = ACTIONS(2165),
    [anon_sym_LBRACK] = ACTIONS(2169),
    [anon_sym_RBRACK] = ACTIONS(2205),
    [anon_sym_DOLLAR] = ACTIONS(2207),
    [anon_sym_u8] = ACTIONS(2203),
    [anon_sym_i8] = ACTIONS(2203),
    [anon_sym_u16] = ACTIONS(2203),
    [anon_sym_i16] = ACTIONS(2203),
    [anon_sym_u32] = ACTIONS(2203),
    [anon_sym_i32] = ACTIONS(2203),
    [anon_sym_u64] = ACTIONS(2203),
    [anon_sym_i64] = ACTIONS(2203),
    [anon_sym_u128] = ACTIONS(2203),
    [anon_sym_i128] = ACTIONS(2203),
    [anon_sym_isize] = ACTIONS(2203),
    [anon_sym_usize] = ACTIONS(2203),
    [anon_sym_f32] = ACTIONS(2203),
    [anon_sym_f64] = ACTIONS(2203),
    [anon_sym_bool] = ACTIONS(2203),
    [anon_sym_str] = ACTIONS(2203),
    [anon_sym_char] = ACTIONS(2203),
    [aux_sym__non_special_token_token1] = ACTIONS(2203),
    [anon_sym_SQUOTE] = ACTIONS(2203),
    [anon_sym_as] = ACTIONS(2203),
    [anon_sym_async] = ACTIONS(2203),
    [anon_sym_await] = ACTIONS(2203),
    [anon_sym_break] = ACTIONS(2203),
    [anon_sym_const] = ACTIONS(2203),
    [anon_sym_continue] = ACTIONS(2203),
    [anon_sym_default] = ACTIONS(2203),
    [anon_sym_enum] = ACTIONS(2203),
    [anon_sym_fn] = ACTIONS(2203),
    [anon_sym_for] = ACTIONS(2203),
    [anon_sym_if] = ACTIONS(2203),
    [anon_sym_impl] = ACTIONS(2203),
    [anon_sym_let] = ACTIONS(2203),
    [anon_sym_loop] = ACTIONS(2203),
    [anon_sym_match] = ACTIONS(2203),
    [anon_sym_mod] = ACTIONS(2203),
    [anon_sym_pub] = ACTIONS(2203),
    [anon_sym_return] = ACTIONS(2203),
    [anon_sym_static] = ACTIONS(2203),
    [anon_sym_struct] = ACTIONS(2203),
    [anon_sym_trait] = ACTIONS(2203),
    [anon_sym_type] = ACTIONS(2203),
    [anon_sym_union] = ACTIONS(2203),
    [anon_sym_unsafe] = ACTIONS(2203),
    [anon_sym_use] = ACTIONS(2203),
    [anon_sym_where] = ACTIONS(2203),
    [anon_sym_while] = ACTIONS(2203),
    [sym_mutable_specifier] = ACTIONS(2203),
    [sym_integer_literal] = ACTIONS(2173),
    [aux_sym_string_literal_token1] = ACTIONS(2175),
    [sym_char_literal] = ACTIONS(2173),
    [anon_sym_true] = ACTIONS(2177),
    [anon_sym_false] = ACTIONS(2177),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2203),
    [sym_super] = ACTIONS(2203),
    [sym_crate] = ACTIONS(2203),
    [sym_raw_string_literal] = ACTIONS(2173),
    [sym_float_literal] = ACTIONS(2173),
    [sym_block_comment] = ACTIONS(3),
  },
  [508] = {
    [sym_delim_token_tree] = STATE(542),
    [sym__delim_tokens] = STATE(542),
    [sym__non_delim_token] = STATE(542),
    [sym__literal] = STATE(542),
    [sym_string_literal] = STATE(592),
    [sym_boolean_literal] = STATE(592),
    [aux_sym_delim_token_tree_repeat1] = STATE(542),
    [sym_identifier] = ACTIONS(2209),
    [anon_sym_LPAREN] = ACTIONS(2163),
    [anon_sym_LBRACE] = ACTIONS(2165),
    [anon_sym_RBRACE] = ACTIONS(2205),
    [anon_sym_LBRACK] = ACTIONS(2169),
    [anon_sym_DOLLAR] = ACTIONS(2211),
    [anon_sym_u8] = ACTIONS(2209),
    [anon_sym_i8] = ACTIONS(2209),
    [anon_sym_u16] = ACTIONS(2209),
    [anon_sym_i16] = ACTIONS(2209),
    [anon_sym_u32] = ACTIONS(2209),
    [anon_sym_i32] = ACTIONS(2209),
    [anon_sym_u64] = ACTIONS(2209),
    [anon_sym_i64] = ACTIONS(2209),
    [anon_sym_u128] = ACTIONS(2209),
    [anon_sym_i128] = ACTIONS(2209),
    [anon_sym_isize] = ACTIONS(2209),
    [anon_sym_usize] = ACTIONS(2209),
    [anon_sym_f32] = ACTIONS(2209),
    [anon_sym_f64] = ACTIONS(2209),
    [anon_sym_bool] = ACTIONS(2209),
    [anon_sym_str] = ACTIONS(2209),
    [anon_sym_char] = ACTIONS(2209),
    [aux_sym__non_special_token_token1] = ACTIONS(2209),
    [anon_sym_SQUOTE] = ACTIONS(2209),
    [anon_sym_as] = ACTIONS(2209),
    [anon_sym_async] = ACTIONS(2209),
    [anon_sym_await] = ACTIONS(2209),
    [anon_sym_break] = ACTIONS(2209),
    [anon_sym_const] = ACTIONS(2209),
    [anon_sym_continue] = ACTIONS(2209),
    [anon_sym_default] = ACTIONS(2209),
    [anon_sym_enum] = ACTIONS(2209),
    [anon_sym_fn] = ACTIONS(2209),
    [anon_sym_for] = ACTIONS(2209),
    [anon_sym_if] = ACTIONS(2209),
    [anon_sym_impl] = ACTIONS(2209),
    [anon_sym_let] = ACTIONS(2209),
    [anon_sym_loop] = ACTIONS(2209),
    [anon_sym_match] = ACTIONS(2209),
    [anon_sym_mod] = ACTIONS(2209),
    [anon_sym_pub] = ACTIONS(2209),
    [anon_sym_return] = ACTIONS(2209),
    [anon_sym_static] = ACTIONS(2209),
    [anon_sym_struct] = ACTIONS(2209),
    [anon_sym_trait] = ACTIONS(2209),
    [anon_sym_type] = ACTIONS(2209),
    [anon_sym_union] = ACTIONS(2209),
    [anon_sym_unsafe] = ACTIONS(2209),
    [anon_sym_use] = ACTIONS(2209),
    [anon_sym_where] = ACTIONS(2209),
    [anon_sym_while] = ACTIONS(2209),
    [sym_mutable_specifier] = ACTIONS(2209),
    [sym_integer_literal] = ACTIONS(2173),
    [aux_sym_string_literal_token1] = ACTIONS(2175),
    [sym_char_literal] = ACTIONS(2173),
    [anon_sym_true] = ACTIONS(2177),
    [anon_sym_false] = ACTIONS(2177),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2209),
    [sym_super] = ACTIONS(2209),
    [sym_crate] = ACTIONS(2209),
    [sym_raw_string_literal] = ACTIONS(2173),
    [sym_float_literal] = ACTIONS(2173),
    [sym_block_comment] = ACTIONS(3),
  },
  [509] = {
    [sym_delim_token_tree] = STATE(504),
    [sym__delim_tokens] = STATE(504),
    [sym__non_delim_token] = STATE(504),
    [sym__literal] = STATE(504),
    [sym_string_literal] = STATE(592),
    [sym_boolean_literal] = STATE(592),
    [aux_sym_delim_token_tree_repeat1] = STATE(504),
    [sym_identifier] = ACTIONS(2213),
    [anon_sym_LPAREN] = ACTIONS(2163),
    [anon_sym_RPAREN] = ACTIONS(2205),
    [anon_sym_LBRACE] = ACTIONS(2165),
    [anon_sym_LBRACK] = ACTIONS(2169),
    [anon_sym_DOLLAR] = ACTIONS(2215),
    [anon_sym_u8] = ACTIONS(2213),
    [anon_sym_i8] = ACTIONS(2213),
    [anon_sym_u16] = ACTIONS(2213),
    [anon_sym_i16] = ACTIONS(2213),
    [anon_sym_u32] = ACTIONS(2213),
    [anon_sym_i32] = ACTIONS(2213),
    [anon_sym_u64] = ACTIONS(2213),
    [anon_sym_i64] = ACTIONS(2213),
    [anon_sym_u128] = ACTIONS(2213),
    [anon_sym_i128] = ACTIONS(2213),
    [anon_sym_isize] = ACTIONS(2213),
    [anon_sym_usize] = ACTIONS(2213),
    [anon_sym_f32] = ACTIONS(2213),
    [anon_sym_f64] = ACTIONS(2213),
    [anon_sym_bool] = ACTIONS(2213),
    [anon_sym_str] = ACTIONS(2213),
    [anon_sym_char] = ACTIONS(2213),
    [aux_sym__non_special_token_token1] = ACTIONS(2213),
    [anon_sym_SQUOTE] = ACTIONS(2213),
    [anon_sym_as] = ACTIONS(2213),
    [anon_sym_async] = ACTIONS(2213),
    [anon_sym_await] = ACTIONS(2213),
    [anon_sym_break] = ACTIONS(2213),
    [anon_sym_const] = ACTIONS(2213),
    [anon_sym_continue] = ACTIONS(2213),
    [anon_sym_default] = ACTIONS(2213),
    [anon_sym_enum] = ACTIONS(2213),
    [anon_sym_fn] = ACTIONS(2213),
    [anon_sym_for] = ACTIONS(2213),
    [anon_sym_if] = ACTIONS(2213),
    [anon_sym_impl] = ACTIONS(2213),
    [anon_sym_let] = ACTIONS(2213),
    [anon_sym_loop] = ACTIONS(2213),
    [anon_sym_match] = ACTIONS(2213),
    [anon_sym_mod] = ACTIONS(2213),
    [anon_sym_pub] = ACTIONS(2213),
    [anon_sym_return] = ACTIONS(2213),
    [anon_sym_static] = ACTIONS(2213),
    [anon_sym_struct] = ACTIONS(2213),
    [anon_sym_trait] = ACTIONS(2213),
    [anon_sym_type] = ACTIONS(2213),
    [anon_sym_union] = ACTIONS(2213),
    [anon_sym_unsafe] = ACTIONS(2213),
    [anon_sym_use] = ACTIONS(2213),
    [anon_sym_where] = ACTIONS(2213),
    [anon_sym_while] = ACTIONS(2213),
    [sym_mutable_specifier] = ACTIONS(2213),
    [sym_integer_literal] = ACTIONS(2173),
    [aux_sym_string_literal_token1] = ACTIONS(2175),
    [sym_char_literal] = ACTIONS(2173),
    [anon_sym_true] = ACTIONS(2177),
    [anon_sym_false] = ACTIONS(2177),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2213),
    [sym_super] = ACTIONS(2213),
    [sym_crate] = ACTIONS(2213),
    [sym_raw_string_literal] = ACTIONS(2173),
    [sym_float_literal] = ACTIONS(2173),
    [sym_block_comment] = ACTIONS(3),
  },
  [510] = {
    [sym_delim_token_tree] = STATE(483),
    [sym__delim_tokens] = STATE(483),
    [sym__non_delim_token] = STATE(483),
    [sym__literal] = STATE(483),
    [sym_string_literal] = STATE(592),
    [sym_boolean_literal] = STATE(592),
    [aux_sym_delim_token_tree_repeat1] = STATE(483),
    [sym_identifier] = ACTIONS(2161),
    [anon_sym_LPAREN] = ACTIONS(2163),
    [anon_sym_RPAREN] = ACTIONS(2217),
    [anon_sym_LBRACE] = ACTIONS(2165),
    [anon_sym_LBRACK] = ACTIONS(2169),
    [anon_sym_DOLLAR] = ACTIONS(2171),
    [anon_sym_u8] = ACTIONS(2161),
    [anon_sym_i8] = ACTIONS(2161),
    [anon_sym_u16] = ACTIONS(2161),
    [anon_sym_i16] = ACTIONS(2161),
    [anon_sym_u32] = ACTIONS(2161),
    [anon_sym_i32] = ACTIONS(2161),
    [anon_sym_u64] = ACTIONS(2161),
    [anon_sym_i64] = ACTIONS(2161),
    [anon_sym_u128] = ACTIONS(2161),
    [anon_sym_i128] = ACTIONS(2161),
    [anon_sym_isize] = ACTIONS(2161),
    [anon_sym_usize] = ACTIONS(2161),
    [anon_sym_f32] = ACTIONS(2161),
    [anon_sym_f64] = ACTIONS(2161),
    [anon_sym_bool] = ACTIONS(2161),
    [anon_sym_str] = ACTIONS(2161),
    [anon_sym_char] = ACTIONS(2161),
    [aux_sym__non_special_token_token1] = ACTIONS(2161),
    [anon_sym_SQUOTE] = ACTIONS(2161),
    [anon_sym_as] = ACTIONS(2161),
    [anon_sym_async] = ACTIONS(2161),
    [anon_sym_await] = ACTIONS(2161),
    [anon_sym_break] = ACTIONS(2161),
    [anon_sym_const] = ACTIONS(2161),
    [anon_sym_continue] = ACTIONS(2161),
    [anon_sym_default] = ACTIONS(2161),
    [anon_sym_enum] = ACTIONS(2161),
    [anon_sym_fn] = ACTIONS(2161),
    [anon_sym_for] = ACTIONS(2161),
    [anon_sym_if] = ACTIONS(2161),
    [anon_sym_impl] = ACTIONS(2161),
    [anon_sym_let] = ACTIONS(2161),
    [anon_sym_loop] = ACTIONS(2161),
    [anon_sym_match] = ACTIONS(2161),
    [anon_sym_mod] = ACTIONS(2161),
    [anon_sym_pub] = ACTIONS(2161),
    [anon_sym_return] = ACTIONS(2161),
    [anon_sym_static] = ACTIONS(2161),
    [anon_sym_struct] = ACTIONS(2161),
    [anon_sym_trait] = ACTIONS(2161),
    [anon_sym_type] = ACTIONS(2161),
    [anon_sym_union] = ACTIONS(2161),
    [anon_sym_unsafe] = ACTIONS(2161),
    [anon_sym_use] = ACTIONS(2161),
    [anon_sym_where] = ACTIONS(2161),
    [anon_sym_while] = ACTIONS(2161),
    [sym_mutable_specifier] = ACTIONS(2161),
    [sym_integer_literal] = ACTIONS(2173),
    [aux_sym_string_literal_token1] = ACTIONS(2175),
    [sym_char_literal] = ACTIONS(2173),
    [anon_sym_true] = ACTIONS(2177),
    [anon_sym_false] = ACTIONS(2177),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2161),
    [sym_super] = ACTIONS(2161),
    [sym_crate] = ACTIONS(2161),
    [sym_raw_string_literal] = ACTIONS(2173),
    [sym_float_literal] = ACTIONS(2173),
    [sym_block_comment] = ACTIONS(3),
  },
  [511] = {
    [sym_delim_token_tree] = STATE(529),
    [sym__delim_tokens] = STATE(529),
    [sym__non_delim_token] = STATE(529),
    [sym__literal] = STATE(529),
    [sym_string_literal] = STATE(592),
    [sym_boolean_literal] = STATE(592),
    [aux_sym_delim_token_tree_repeat1] = STATE(529),
    [sym_identifier] = ACTIONS(2219),
    [anon_sym_LPAREN] = ACTIONS(2163),
    [anon_sym_LBRACE] = ACTIONS(2165),
    [anon_sym_LBRACK] = ACTIONS(2169),
    [anon_sym_RBRACK] = ACTIONS(2221),
    [anon_sym_DOLLAR] = ACTIONS(2223),
    [anon_sym_u8] = ACTIONS(2219),
    [anon_sym_i8] = ACTIONS(2219),
    [anon_sym_u16] = ACTIONS(2219),
    [anon_sym_i16] = ACTIONS(2219),
    [anon_sym_u32] = ACTIONS(2219),
    [anon_sym_i32] = ACTIONS(2219),
    [anon_sym_u64] = ACTIONS(2219),
    [anon_sym_i64] = ACTIONS(2219),
    [anon_sym_u128] = ACTIONS(2219),
    [anon_sym_i128] = ACTIONS(2219),
    [anon_sym_isize] = ACTIONS(2219),
    [anon_sym_usize] = ACTIONS(2219),
    [anon_sym_f32] = ACTIONS(2219),
    [anon_sym_f64] = ACTIONS(2219),
    [anon_sym_bool] = ACTIONS(2219),
    [anon_sym_str] = ACTIONS(2219),
    [anon_sym_char] = ACTIONS(2219),
    [aux_sym__non_special_token_token1] = ACTIONS(2219),
    [anon_sym_SQUOTE] = ACTIONS(2219),
    [anon_sym_as] = ACTIONS(2219),
    [anon_sym_async] = ACTIONS(2219),
    [anon_sym_await] = ACTIONS(2219),
    [anon_sym_break] = ACTIONS(2219),
    [anon_sym_const] = ACTIONS(2219),
    [anon_sym_continue] = ACTIONS(2219),
    [anon_sym_default] = ACTIONS(2219),
    [anon_sym_enum] = ACTIONS(2219),
    [anon_sym_fn] = ACTIONS(2219),
    [anon_sym_for] = ACTIONS(2219),
    [anon_sym_if] = ACTIONS(2219),
    [anon_sym_impl] = ACTIONS(2219),
    [anon_sym_let] = ACTIONS(2219),
    [anon_sym_loop] = ACTIONS(2219),
    [anon_sym_match] = ACTIONS(2219),
    [anon_sym_mod] = ACTIONS(2219),
    [anon_sym_pub] = ACTIONS(2219),
    [anon_sym_return] = ACTIONS(2219),
    [anon_sym_static] = ACTIONS(2219),
    [anon_sym_struct] = ACTIONS(2219),
    [anon_sym_trait] = ACTIONS(2219),
    [anon_sym_type] = ACTIONS(2219),
    [anon_sym_union] = ACTIONS(2219),
    [anon_sym_unsafe] = ACTIONS(2219),
    [anon_sym_use] = ACTIONS(2219),
    [anon_sym_where] = ACTIONS(2219),
    [anon_sym_while] = ACTIONS(2219),
    [sym_mutable_specifier] = ACTIONS(2219),
    [sym_integer_literal] = ACTIONS(2173),
    [aux_sym_string_literal_token1] = ACTIONS(2175),
    [sym_char_literal] = ACTIONS(2173),
    [anon_sym_true] = ACTIONS(2177),
    [anon_sym_false] = ACTIONS(2177),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2219),
    [sym_super] = ACTIONS(2219),
    [sym_crate] = ACTIONS(2219),
    [sym_raw_string_literal] = ACTIONS(2173),
    [sym_float_literal] = ACTIONS(2173),
    [sym_block_comment] = ACTIONS(3),
  },
  [512] = {
    [sym_delim_token_tree] = STATE(483),
    [sym__delim_tokens] = STATE(483),
    [sym__non_delim_token] = STATE(483),
    [sym__literal] = STATE(483),
    [sym_string_literal] = STATE(592),
    [sym_boolean_literal] = STATE(592),
    [aux_sym_delim_token_tree_repeat1] = STATE(483),
    [sym_identifier] = ACTIONS(2161),
    [anon_sym_LPAREN] = ACTIONS(2163),
    [anon_sym_LBRACE] = ACTIONS(2165),
    [anon_sym_RBRACE] = ACTIONS(2217),
    [anon_sym_LBRACK] = ACTIONS(2169),
    [anon_sym_DOLLAR] = ACTIONS(2171),
    [anon_sym_u8] = ACTIONS(2161),
    [anon_sym_i8] = ACTIONS(2161),
    [anon_sym_u16] = ACTIONS(2161),
    [anon_sym_i16] = ACTIONS(2161),
    [anon_sym_u32] = ACTIONS(2161),
    [anon_sym_i32] = ACTIONS(2161),
    [anon_sym_u64] = ACTIONS(2161),
    [anon_sym_i64] = ACTIONS(2161),
    [anon_sym_u128] = ACTIONS(2161),
    [anon_sym_i128] = ACTIONS(2161),
    [anon_sym_isize] = ACTIONS(2161),
    [anon_sym_usize] = ACTIONS(2161),
    [anon_sym_f32] = ACTIONS(2161),
    [anon_sym_f64] = ACTIONS(2161),
    [anon_sym_bool] = ACTIONS(2161),
    [anon_sym_str] = ACTIONS(2161),
    [anon_sym_char] = ACTIONS(2161),
    [aux_sym__non_special_token_token1] = ACTIONS(2161),
    [anon_sym_SQUOTE] = ACTIONS(2161),
    [anon_sym_as] = ACTIONS(2161),
    [anon_sym_async] = ACTIONS(2161),
    [anon_sym_await] = ACTIONS(2161),
    [anon_sym_break] = ACTIONS(2161),
    [anon_sym_const] = ACTIONS(2161),
    [anon_sym_continue] = ACTIONS(2161),
    [anon_sym_default] = ACTIONS(2161),
    [anon_sym_enum] = ACTIONS(2161),
    [anon_sym_fn] = ACTIONS(2161),
    [anon_sym_for] = ACTIONS(2161),
    [anon_sym_if] = ACTIONS(2161),
    [anon_sym_impl] = ACTIONS(2161),
    [anon_sym_let] = ACTIONS(2161),
    [anon_sym_loop] = ACTIONS(2161),
    [anon_sym_match] = ACTIONS(2161),
    [anon_sym_mod] = ACTIONS(2161),
    [anon_sym_pub] = ACTIONS(2161),
    [anon_sym_return] = ACTIONS(2161),
    [anon_sym_static] = ACTIONS(2161),
    [anon_sym_struct] = ACTIONS(2161),
    [anon_sym_trait] = ACTIONS(2161),
    [anon_sym_type] = ACTIONS(2161),
    [anon_sym_union] = ACTIONS(2161),
    [anon_sym_unsafe] = ACTIONS(2161),
    [anon_sym_use] = ACTIONS(2161),
    [anon_sym_where] = ACTIONS(2161),
    [anon_sym_while] = ACTIONS(2161),
    [sym_mutable_specifier] = ACTIONS(2161),
    [sym_integer_literal] = ACTIONS(2173),
    [aux_sym_string_literal_token1] = ACTIONS(2175),
    [sym_char_literal] = ACTIONS(2173),
    [anon_sym_true] = ACTIONS(2177),
    [anon_sym_false] = ACTIONS(2177),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2161),
    [sym_super] = ACTIONS(2161),
    [sym_crate] = ACTIONS(2161),
    [sym_raw_string_literal] = ACTIONS(2173),
    [sym_float_literal] = ACTIONS(2173),
    [sym_block_comment] = ACTIONS(3),
  },
  [513] = {
    [sym_delim_token_tree] = STATE(483),
    [sym__delim_tokens] = STATE(483),
    [sym__non_delim_token] = STATE(483),
    [sym__literal] = STATE(483),
    [sym_string_literal] = STATE(592),
    [sym_boolean_literal] = STATE(592),
    [aux_sym_delim_token_tree_repeat1] = STATE(483),
    [sym_identifier] = ACTIONS(2161),
    [anon_sym_LPAREN] = ACTIONS(2163),
    [anon_sym_LBRACE] = ACTIONS(2165),
    [anon_sym_LBRACK] = ACTIONS(2169),
    [anon_sym_RBRACK] = ACTIONS(2217),
    [anon_sym_DOLLAR] = ACTIONS(2171),
    [anon_sym_u8] = ACTIONS(2161),
    [anon_sym_i8] = ACTIONS(2161),
    [anon_sym_u16] = ACTIONS(2161),
    [anon_sym_i16] = ACTIONS(2161),
    [anon_sym_u32] = ACTIONS(2161),
    [anon_sym_i32] = ACTIONS(2161),
    [anon_sym_u64] = ACTIONS(2161),
    [anon_sym_i64] = ACTIONS(2161),
    [anon_sym_u128] = ACTIONS(2161),
    [anon_sym_i128] = ACTIONS(2161),
    [anon_sym_isize] = ACTIONS(2161),
    [anon_sym_usize] = ACTIONS(2161),
    [anon_sym_f32] = ACTIONS(2161),
    [anon_sym_f64] = ACTIONS(2161),
    [anon_sym_bool] = ACTIONS(2161),
    [anon_sym_str] = ACTIONS(2161),
    [anon_sym_char] = ACTIONS(2161),
    [aux_sym__non_special_token_token1] = ACTIONS(2161),
    [anon_sym_SQUOTE] = ACTIONS(2161),
    [anon_sym_as] = ACTIONS(2161),
    [anon_sym_async] = ACTIONS(2161),
    [anon_sym_await] = ACTIONS(2161),
    [anon_sym_break] = ACTIONS(2161),
    [anon_sym_const] = ACTIONS(2161),
    [anon_sym_continue] = ACTIONS(2161),
    [anon_sym_default] = ACTIONS(2161),
    [anon_sym_enum] = ACTIONS(2161),
    [anon_sym_fn] = ACTIONS(2161),
    [anon_sym_for] = ACTIONS(2161),
    [anon_sym_if] = ACTIONS(2161),
    [anon_sym_impl] = ACTIONS(2161),
    [anon_sym_let] = ACTIONS(2161),
    [anon_sym_loop] = ACTIONS(2161),
    [anon_sym_match] = ACTIONS(2161),
    [anon_sym_mod] = ACTIONS(2161),
    [anon_sym_pub] = ACTIONS(2161),
    [anon_sym_return] = ACTIONS(2161),
    [anon_sym_static] = ACTIONS(2161),
    [anon_sym_struct] = ACTIONS(2161),
    [anon_sym_trait] = ACTIONS(2161),
    [anon_sym_type] = ACTIONS(2161),
    [anon_sym_union] = ACTIONS(2161),
    [anon_sym_unsafe] = ACTIONS(2161),
    [anon_sym_use] = ACTIONS(2161),
    [anon_sym_where] = ACTIONS(2161),
    [anon_sym_while] = ACTIONS(2161),
    [sym_mutable_specifier] = ACTIONS(2161),
    [sym_integer_literal] = ACTIONS(2173),
    [aux_sym_string_literal_token1] = ACTIONS(2175),
    [sym_char_literal] = ACTIONS(2173),
    [anon_sym_true] = ACTIONS(2177),
    [anon_sym_false] = ACTIONS(2177),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2161),
    [sym_super] = ACTIONS(2161),
    [sym_crate] = ACTIONS(2161),
    [sym_raw_string_literal] = ACTIONS(2173),
    [sym_float_literal] = ACTIONS(2173),
    [sym_block_comment] = ACTIONS(3),
  },
  [514] = {
    [sym_token_tree] = STATE(525),
    [sym_token_repetition] = STATE(525),
    [sym__literal] = STATE(525),
    [sym_string_literal] = STATE(585),
    [sym_boolean_literal] = STATE(585),
    [aux_sym_token_tree_repeat1] = STATE(525),
    [sym_identifier] = ACTIONS(2225),
    [anon_sym_LPAREN] = ACTIONS(2183),
    [anon_sym_LBRACE] = ACTIONS(2185),
    [anon_sym_LBRACK] = ACTIONS(2189),
    [anon_sym_RBRACK] = ACTIONS(2227),
    [anon_sym_DOLLAR] = ACTIONS(2191),
    [anon_sym_u8] = ACTIONS(2225),
    [anon_sym_i8] = ACTIONS(2225),
    [anon_sym_u16] = ACTIONS(2225),
    [anon_sym_i16] = ACTIONS(2225),
    [anon_sym_u32] = ACTIONS(2225),
    [anon_sym_i32] = ACTIONS(2225),
    [anon_sym_u64] = ACTIONS(2225),
    [anon_sym_i64] = ACTIONS(2225),
    [anon_sym_u128] = ACTIONS(2225),
    [anon_sym_i128] = ACTIONS(2225),
    [anon_sym_isize] = ACTIONS(2225),
    [anon_sym_usize] = ACTIONS(2225),
    [anon_sym_f32] = ACTIONS(2225),
    [anon_sym_f64] = ACTIONS(2225),
    [anon_sym_bool] = ACTIONS(2225),
    [anon_sym_str] = ACTIONS(2225),
    [anon_sym_char] = ACTIONS(2225),
    [aux_sym__non_special_token_token1] = ACTIONS(2225),
    [anon_sym_SQUOTE] = ACTIONS(2225),
    [anon_sym_as] = ACTIONS(2225),
    [anon_sym_async] = ACTIONS(2225),
    [anon_sym_await] = ACTIONS(2225),
    [anon_sym_break] = ACTIONS(2225),
    [anon_sym_const] = ACTIONS(2225),
    [anon_sym_continue] = ACTIONS(2225),
    [anon_sym_default] = ACTIONS(2225),
    [anon_sym_enum] = ACTIONS(2225),
    [anon_sym_fn] = ACTIONS(2225),
    [anon_sym_for] = ACTIONS(2225),
    [anon_sym_if] = ACTIONS(2225),
    [anon_sym_impl] = ACTIONS(2225),
    [anon_sym_let] = ACTIONS(2225),
    [anon_sym_loop] = ACTIONS(2225),
    [anon_sym_match] = ACTIONS(2225),
    [anon_sym_mod] = ACTIONS(2225),
    [anon_sym_pub] = ACTIONS(2225),
    [anon_sym_return] = ACTIONS(2225),
    [anon_sym_static] = ACTIONS(2225),
    [anon_sym_struct] = ACTIONS(2225),
    [anon_sym_trait] = ACTIONS(2225),
    [anon_sym_type] = ACTIONS(2225),
    [anon_sym_union] = ACTIONS(2225),
    [anon_sym_unsafe] = ACTIONS(2225),
    [anon_sym_use] = ACTIONS(2225),
    [anon_sym_where] = ACTIONS(2225),
    [anon_sym_while] = ACTIONS(2225),
    [sym_mutable_specifier] = ACTIONS(2225),
    [sym_integer_literal] = ACTIONS(2014),
    [aux_sym_string_literal_token1] = ACTIONS(2016),
    [sym_char_literal] = ACTIONS(2014),
    [anon_sym_true] = ACTIONS(2018),
    [anon_sym_false] = ACTIONS(2018),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2225),
    [sym_super] = ACTIONS(2225),
    [sym_crate] = ACTIONS(2225),
    [sym_metavariable] = ACTIONS(2229),
    [sym_raw_string_literal] = ACTIONS(2014),
    [sym_float_literal] = ACTIONS(2014),
    [sym_block_comment] = ACTIONS(3),
  },
  [515] = {
    [sym_token_tree] = STATE(503),
    [sym_token_repetition] = STATE(503),
    [sym__literal] = STATE(503),
    [sym_string_literal] = STATE(585),
    [sym_boolean_literal] = STATE(585),
    [aux_sym_token_tree_repeat1] = STATE(503),
    [sym_identifier] = ACTIONS(2231),
    [anon_sym_LPAREN] = ACTIONS(2183),
    [anon_sym_LBRACE] = ACTIONS(2185),
    [anon_sym_RBRACE] = ACTIONS(2227),
    [anon_sym_LBRACK] = ACTIONS(2189),
    [anon_sym_DOLLAR] = ACTIONS(2191),
    [anon_sym_u8] = ACTIONS(2231),
    [anon_sym_i8] = ACTIONS(2231),
    [anon_sym_u16] = ACTIONS(2231),
    [anon_sym_i16] = ACTIONS(2231),
    [anon_sym_u32] = ACTIONS(2231),
    [anon_sym_i32] = ACTIONS(2231),
    [anon_sym_u64] = ACTIONS(2231),
    [anon_sym_i64] = ACTIONS(2231),
    [anon_sym_u128] = ACTIONS(2231),
    [anon_sym_i128] = ACTIONS(2231),
    [anon_sym_isize] = ACTIONS(2231),
    [anon_sym_usize] = ACTIONS(2231),
    [anon_sym_f32] = ACTIONS(2231),
    [anon_sym_f64] = ACTIONS(2231),
    [anon_sym_bool] = ACTIONS(2231),
    [anon_sym_str] = ACTIONS(2231),
    [anon_sym_char] = ACTIONS(2231),
    [aux_sym__non_special_token_token1] = ACTIONS(2231),
    [anon_sym_SQUOTE] = ACTIONS(2231),
    [anon_sym_as] = ACTIONS(2231),
    [anon_sym_async] = ACTIONS(2231),
    [anon_sym_await] = ACTIONS(2231),
    [anon_sym_break] = ACTIONS(2231),
    [anon_sym_const] = ACTIONS(2231),
    [anon_sym_continue] = ACTIONS(2231),
    [anon_sym_default] = ACTIONS(2231),
    [anon_sym_enum] = ACTIONS(2231),
    [anon_sym_fn] = ACTIONS(2231),
    [anon_sym_for] = ACTIONS(2231),
    [anon_sym_if] = ACTIONS(2231),
    [anon_sym_impl] = ACTIONS(2231),
    [anon_sym_let] = ACTIONS(2231),
    [anon_sym_loop] = ACTIONS(2231),
    [anon_sym_match] = ACTIONS(2231),
    [anon_sym_mod] = ACTIONS(2231),
    [anon_sym_pub] = ACTIONS(2231),
    [anon_sym_return] = ACTIONS(2231),
    [anon_sym_static] = ACTIONS(2231),
    [anon_sym_struct] = ACTIONS(2231),
    [anon_sym_trait] = ACTIONS(2231),
    [anon_sym_type] = ACTIONS(2231),
    [anon_sym_union] = ACTIONS(2231),
    [anon_sym_unsafe] = ACTIONS(2231),
    [anon_sym_use] = ACTIONS(2231),
    [anon_sym_where] = ACTIONS(2231),
    [anon_sym_while] = ACTIONS(2231),
    [sym_mutable_specifier] = ACTIONS(2231),
    [sym_integer_literal] = ACTIONS(2014),
    [aux_sym_string_literal_token1] = ACTIONS(2016),
    [sym_char_literal] = ACTIONS(2014),
    [anon_sym_true] = ACTIONS(2018),
    [anon_sym_false] = ACTIONS(2018),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2231),
    [sym_super] = ACTIONS(2231),
    [sym_crate] = ACTIONS(2231),
    [sym_metavariable] = ACTIONS(2233),
    [sym_raw_string_literal] = ACTIONS(2014),
    [sym_float_literal] = ACTIONS(2014),
    [sym_block_comment] = ACTIONS(3),
  },
  [516] = {
    [sym_token_tree] = STATE(506),
    [sym_token_repetition] = STATE(506),
    [sym__literal] = STATE(506),
    [sym_string_literal] = STATE(585),
    [sym_boolean_literal] = STATE(585),
    [aux_sym_token_tree_repeat1] = STATE(506),
    [sym_identifier] = ACTIONS(2235),
    [anon_sym_LPAREN] = ACTIONS(2183),
    [anon_sym_RPAREN] = ACTIONS(2227),
    [anon_sym_LBRACE] = ACTIONS(2185),
    [anon_sym_LBRACK] = ACTIONS(2189),
    [anon_sym_DOLLAR] = ACTIONS(2191),
    [anon_sym_u8] = ACTIONS(2235),
    [anon_sym_i8] = ACTIONS(2235),
    [anon_sym_u16] = ACTIONS(2235),
    [anon_sym_i16] = ACTIONS(2235),
    [anon_sym_u32] = ACTIONS(2235),
    [anon_sym_i32] = ACTIONS(2235),
    [anon_sym_u64] = ACTIONS(2235),
    [anon_sym_i64] = ACTIONS(2235),
    [anon_sym_u128] = ACTIONS(2235),
    [anon_sym_i128] = ACTIONS(2235),
    [anon_sym_isize] = ACTIONS(2235),
    [anon_sym_usize] = ACTIONS(2235),
    [anon_sym_f32] = ACTIONS(2235),
    [anon_sym_f64] = ACTIONS(2235),
    [anon_sym_bool] = ACTIONS(2235),
    [anon_sym_str] = ACTIONS(2235),
    [anon_sym_char] = ACTIONS(2235),
    [aux_sym__non_special_token_token1] = ACTIONS(2235),
    [anon_sym_SQUOTE] = ACTIONS(2235),
    [anon_sym_as] = ACTIONS(2235),
    [anon_sym_async] = ACTIONS(2235),
    [anon_sym_await] = ACTIONS(2235),
    [anon_sym_break] = ACTIONS(2235),
    [anon_sym_const] = ACTIONS(2235),
    [anon_sym_continue] = ACTIONS(2235),
    [anon_sym_default] = ACTIONS(2235),
    [anon_sym_enum] = ACTIONS(2235),
    [anon_sym_fn] = ACTIONS(2235),
    [anon_sym_for] = ACTIONS(2235),
    [anon_sym_if] = ACTIONS(2235),
    [anon_sym_impl] = ACTIONS(2235),
    [anon_sym_let] = ACTIONS(2235),
    [anon_sym_loop] = ACTIONS(2235),
    [anon_sym_match] = ACTIONS(2235),
    [anon_sym_mod] = ACTIONS(2235),
    [anon_sym_pub] = ACTIONS(2235),
    [anon_sym_return] = ACTIONS(2235),
    [anon_sym_static] = ACTIONS(2235),
    [anon_sym_struct] = ACTIONS(2235),
    [anon_sym_trait] = ACTIONS(2235),
    [anon_sym_type] = ACTIONS(2235),
    [anon_sym_union] = ACTIONS(2235),
    [anon_sym_unsafe] = ACTIONS(2235),
    [anon_sym_use] = ACTIONS(2235),
    [anon_sym_where] = ACTIONS(2235),
    [anon_sym_while] = ACTIONS(2235),
    [sym_mutable_specifier] = ACTIONS(2235),
    [sym_integer_literal] = ACTIONS(2014),
    [aux_sym_string_literal_token1] = ACTIONS(2016),
    [sym_char_literal] = ACTIONS(2014),
    [anon_sym_true] = ACTIONS(2018),
    [anon_sym_false] = ACTIONS(2018),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2235),
    [sym_super] = ACTIONS(2235),
    [sym_crate] = ACTIONS(2235),
    [sym_metavariable] = ACTIONS(2237),
    [sym_raw_string_literal] = ACTIONS(2014),
    [sym_float_literal] = ACTIONS(2014),
    [sym_block_comment] = ACTIONS(3),
  },
  [517] = {
    [sym_delim_token_tree] = STATE(528),
    [sym__delim_tokens] = STATE(528),
    [sym__non_delim_token] = STATE(528),
    [sym__literal] = STATE(528),
    [sym_string_literal] = STATE(592),
    [sym_boolean_literal] = STATE(592),
    [aux_sym_delim_token_tree_repeat1] = STATE(528),
    [sym_identifier] = ACTIONS(2239),
    [anon_sym_LPAREN] = ACTIONS(2163),
    [anon_sym_LBRACE] = ACTIONS(2165),
    [anon_sym_RBRACE] = ACTIONS(2221),
    [anon_sym_LBRACK] = ACTIONS(2169),
    [anon_sym_DOLLAR] = ACTIONS(2241),
    [anon_sym_u8] = ACTIONS(2239),
    [anon_sym_i8] = ACTIONS(2239),
    [anon_sym_u16] = ACTIONS(2239),
    [anon_sym_i16] = ACTIONS(2239),
    [anon_sym_u32] = ACTIONS(2239),
    [anon_sym_i32] = ACTIONS(2239),
    [anon_sym_u64] = ACTIONS(2239),
    [anon_sym_i64] = ACTIONS(2239),
    [anon_sym_u128] = ACTIONS(2239),
    [anon_sym_i128] = ACTIONS(2239),
    [anon_sym_isize] = ACTIONS(2239),
    [anon_sym_usize] = ACTIONS(2239),
    [anon_sym_f32] = ACTIONS(2239),
    [anon_sym_f64] = ACTIONS(2239),
    [anon_sym_bool] = ACTIONS(2239),
    [anon_sym_str] = ACTIONS(2239),
    [anon_sym_char] = ACTIONS(2239),
    [aux_sym__non_special_token_token1] = ACTIONS(2239),
    [anon_sym_SQUOTE] = ACTIONS(2239),
    [anon_sym_as] = ACTIONS(2239),
    [anon_sym_async] = ACTIONS(2239),
    [anon_sym_await] = ACTIONS(2239),
    [anon_sym_break] = ACTIONS(2239),
    [anon_sym_const] = ACTIONS(2239),
    [anon_sym_continue] = ACTIONS(2239),
    [anon_sym_default] = ACTIONS(2239),
    [anon_sym_enum] = ACTIONS(2239),
    [anon_sym_fn] = ACTIONS(2239),
    [anon_sym_for] = ACTIONS(2239),
    [anon_sym_if] = ACTIONS(2239),
    [anon_sym_impl] = ACTIONS(2239),
    [anon_sym_let] = ACTIONS(2239),
    [anon_sym_loop] = ACTIONS(2239),
    [anon_sym_match] = ACTIONS(2239),
    [anon_sym_mod] = ACTIONS(2239),
    [anon_sym_pub] = ACTIONS(2239),
    [anon_sym_return] = ACTIONS(2239),
    [anon_sym_static] = ACTIONS(2239),
    [anon_sym_struct] = ACTIONS(2239),
    [anon_sym_trait] = ACTIONS(2239),
    [anon_sym_type] = ACTIONS(2239),
    [anon_sym_union] = ACTIONS(2239),
    [anon_sym_unsafe] = ACTIONS(2239),
    [anon_sym_use] = ACTIONS(2239),
    [anon_sym_where] = ACTIONS(2239),
    [anon_sym_while] = ACTIONS(2239),
    [sym_mutable_specifier] = ACTIONS(2239),
    [sym_integer_literal] = ACTIONS(2173),
    [aux_sym_string_literal_token1] = ACTIONS(2175),
    [sym_char_literal] = ACTIONS(2173),
    [anon_sym_true] = ACTIONS(2177),
    [anon_sym_false] = ACTIONS(2177),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2239),
    [sym_super] = ACTIONS(2239),
    [sym_crate] = ACTIONS(2239),
    [sym_raw_string_literal] = ACTIONS(2173),
    [sym_float_literal] = ACTIONS(2173),
    [sym_block_comment] = ACTIONS(3),
  },
  [518] = {
    [sym_token_tree] = STATE(485),
    [sym_token_repetition] = STATE(485),
    [sym__literal] = STATE(485),
    [sym_string_literal] = STATE(585),
    [sym_boolean_literal] = STATE(585),
    [aux_sym_token_tree_repeat1] = STATE(485),
    [sym_identifier] = ACTIONS(2181),
    [anon_sym_LPAREN] = ACTIONS(2183),
    [anon_sym_RPAREN] = ACTIONS(2243),
    [anon_sym_LBRACE] = ACTIONS(2185),
    [anon_sym_LBRACK] = ACTIONS(2189),
    [anon_sym_DOLLAR] = ACTIONS(2191),
    [anon_sym_u8] = ACTIONS(2181),
    [anon_sym_i8] = ACTIONS(2181),
    [anon_sym_u16] = ACTIONS(2181),
    [anon_sym_i16] = ACTIONS(2181),
    [anon_sym_u32] = ACTIONS(2181),
    [anon_sym_i32] = ACTIONS(2181),
    [anon_sym_u64] = ACTIONS(2181),
    [anon_sym_i64] = ACTIONS(2181),
    [anon_sym_u128] = ACTIONS(2181),
    [anon_sym_i128] = ACTIONS(2181),
    [anon_sym_isize] = ACTIONS(2181),
    [anon_sym_usize] = ACTIONS(2181),
    [anon_sym_f32] = ACTIONS(2181),
    [anon_sym_f64] = ACTIONS(2181),
    [anon_sym_bool] = ACTIONS(2181),
    [anon_sym_str] = ACTIONS(2181),
    [anon_sym_char] = ACTIONS(2181),
    [aux_sym__non_special_token_token1] = ACTIONS(2181),
    [anon_sym_SQUOTE] = ACTIONS(2181),
    [anon_sym_as] = ACTIONS(2181),
    [anon_sym_async] = ACTIONS(2181),
    [anon_sym_await] = ACTIONS(2181),
    [anon_sym_break] = ACTIONS(2181),
    [anon_sym_const] = ACTIONS(2181),
    [anon_sym_continue] = ACTIONS(2181),
    [anon_sym_default] = ACTIONS(2181),
    [anon_sym_enum] = ACTIONS(2181),
    [anon_sym_fn] = ACTIONS(2181),
    [anon_sym_for] = ACTIONS(2181),
    [anon_sym_if] = ACTIONS(2181),
    [anon_sym_impl] = ACTIONS(2181),
    [anon_sym_let] = ACTIONS(2181),
    [anon_sym_loop] = ACTIONS(2181),
    [anon_sym_match] = ACTIONS(2181),
    [anon_sym_mod] = ACTIONS(2181),
    [anon_sym_pub] = ACTIONS(2181),
    [anon_sym_return] = ACTIONS(2181),
    [anon_sym_static] = ACTIONS(2181),
    [anon_sym_struct] = ACTIONS(2181),
    [anon_sym_trait] = ACTIONS(2181),
    [anon_sym_type] = ACTIONS(2181),
    [anon_sym_union] = ACTIONS(2181),
    [anon_sym_unsafe] = ACTIONS(2181),
    [anon_sym_use] = ACTIONS(2181),
    [anon_sym_where] = ACTIONS(2181),
    [anon_sym_while] = ACTIONS(2181),
    [sym_mutable_specifier] = ACTIONS(2181),
    [sym_integer_literal] = ACTIONS(2014),
    [aux_sym_string_literal_token1] = ACTIONS(2016),
    [sym_char_literal] = ACTIONS(2014),
    [anon_sym_true] = ACTIONS(2018),
    [anon_sym_false] = ACTIONS(2018),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2181),
    [sym_super] = ACTIONS(2181),
    [sym_crate] = ACTIONS(2181),
    [sym_metavariable] = ACTIONS(2193),
    [sym_raw_string_literal] = ACTIONS(2014),
    [sym_float_literal] = ACTIONS(2014),
    [sym_block_comment] = ACTIONS(3),
  },
  [519] = {
    [sym_delim_token_tree] = STATE(483),
    [sym__delim_tokens] = STATE(483),
    [sym__non_delim_token] = STATE(483),
    [sym__literal] = STATE(483),
    [sym_string_literal] = STATE(592),
    [sym_boolean_literal] = STATE(592),
    [aux_sym_delim_token_tree_repeat1] = STATE(483),
    [sym_identifier] = ACTIONS(2161),
    [anon_sym_LPAREN] = ACTIONS(2163),
    [anon_sym_LBRACE] = ACTIONS(2165),
    [anon_sym_LBRACK] = ACTIONS(2169),
    [anon_sym_RBRACK] = ACTIONS(2195),
    [anon_sym_DOLLAR] = ACTIONS(2171),
    [anon_sym_u8] = ACTIONS(2161),
    [anon_sym_i8] = ACTIONS(2161),
    [anon_sym_u16] = ACTIONS(2161),
    [anon_sym_i16] = ACTIONS(2161),
    [anon_sym_u32] = ACTIONS(2161),
    [anon_sym_i32] = ACTIONS(2161),
    [anon_sym_u64] = ACTIONS(2161),
    [anon_sym_i64] = ACTIONS(2161),
    [anon_sym_u128] = ACTIONS(2161),
    [anon_sym_i128] = ACTIONS(2161),
    [anon_sym_isize] = ACTIONS(2161),
    [anon_sym_usize] = ACTIONS(2161),
    [anon_sym_f32] = ACTIONS(2161),
    [anon_sym_f64] = ACTIONS(2161),
    [anon_sym_bool] = ACTIONS(2161),
    [anon_sym_str] = ACTIONS(2161),
    [anon_sym_char] = ACTIONS(2161),
    [aux_sym__non_special_token_token1] = ACTIONS(2161),
    [anon_sym_SQUOTE] = ACTIONS(2161),
    [anon_sym_as] = ACTIONS(2161),
    [anon_sym_async] = ACTIONS(2161),
    [anon_sym_await] = ACTIONS(2161),
    [anon_sym_break] = ACTIONS(2161),
    [anon_sym_const] = ACTIONS(2161),
    [anon_sym_continue] = ACTIONS(2161),
    [anon_sym_default] = ACTIONS(2161),
    [anon_sym_enum] = ACTIONS(2161),
    [anon_sym_fn] = ACTIONS(2161),
    [anon_sym_for] = ACTIONS(2161),
    [anon_sym_if] = ACTIONS(2161),
    [anon_sym_impl] = ACTIONS(2161),
    [anon_sym_let] = ACTIONS(2161),
    [anon_sym_loop] = ACTIONS(2161),
    [anon_sym_match] = ACTIONS(2161),
    [anon_sym_mod] = ACTIONS(2161),
    [anon_sym_pub] = ACTIONS(2161),
    [anon_sym_return] = ACTIONS(2161),
    [anon_sym_static] = ACTIONS(2161),
    [anon_sym_struct] = ACTIONS(2161),
    [anon_sym_trait] = ACTIONS(2161),
    [anon_sym_type] = ACTIONS(2161),
    [anon_sym_union] = ACTIONS(2161),
    [anon_sym_unsafe] = ACTIONS(2161),
    [anon_sym_use] = ACTIONS(2161),
    [anon_sym_where] = ACTIONS(2161),
    [anon_sym_while] = ACTIONS(2161),
    [sym_mutable_specifier] = ACTIONS(2161),
    [sym_integer_literal] = ACTIONS(2173),
    [aux_sym_string_literal_token1] = ACTIONS(2175),
    [sym_char_literal] = ACTIONS(2173),
    [anon_sym_true] = ACTIONS(2177),
    [anon_sym_false] = ACTIONS(2177),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2161),
    [sym_super] = ACTIONS(2161),
    [sym_crate] = ACTIONS(2161),
    [sym_raw_string_literal] = ACTIONS(2173),
    [sym_float_literal] = ACTIONS(2173),
    [sym_block_comment] = ACTIONS(3),
  },
  [520] = {
    [sym_delim_token_tree] = STATE(512),
    [sym__delim_tokens] = STATE(512),
    [sym__non_delim_token] = STATE(512),
    [sym__literal] = STATE(512),
    [sym_string_literal] = STATE(592),
    [sym_boolean_literal] = STATE(592),
    [aux_sym_delim_token_tree_repeat1] = STATE(512),
    [sym_identifier] = ACTIONS(2245),
    [anon_sym_LPAREN] = ACTIONS(2163),
    [anon_sym_LBRACE] = ACTIONS(2165),
    [anon_sym_RBRACE] = ACTIONS(2199),
    [anon_sym_LBRACK] = ACTIONS(2169),
    [anon_sym_DOLLAR] = ACTIONS(2247),
    [anon_sym_u8] = ACTIONS(2245),
    [anon_sym_i8] = ACTIONS(2245),
    [anon_sym_u16] = ACTIONS(2245),
    [anon_sym_i16] = ACTIONS(2245),
    [anon_sym_u32] = ACTIONS(2245),
    [anon_sym_i32] = ACTIONS(2245),
    [anon_sym_u64] = ACTIONS(2245),
    [anon_sym_i64] = ACTIONS(2245),
    [anon_sym_u128] = ACTIONS(2245),
    [anon_sym_i128] = ACTIONS(2245),
    [anon_sym_isize] = ACTIONS(2245),
    [anon_sym_usize] = ACTIONS(2245),
    [anon_sym_f32] = ACTIONS(2245),
    [anon_sym_f64] = ACTIONS(2245),
    [anon_sym_bool] = ACTIONS(2245),
    [anon_sym_str] = ACTIONS(2245),
    [anon_sym_char] = ACTIONS(2245),
    [aux_sym__non_special_token_token1] = ACTIONS(2245),
    [anon_sym_SQUOTE] = ACTIONS(2245),
    [anon_sym_as] = ACTIONS(2245),
    [anon_sym_async] = ACTIONS(2245),
    [anon_sym_await] = ACTIONS(2245),
    [anon_sym_break] = ACTIONS(2245),
    [anon_sym_const] = ACTIONS(2245),
    [anon_sym_continue] = ACTIONS(2245),
    [anon_sym_default] = ACTIONS(2245),
    [anon_sym_enum] = ACTIONS(2245),
    [anon_sym_fn] = ACTIONS(2245),
    [anon_sym_for] = ACTIONS(2245),
    [anon_sym_if] = ACTIONS(2245),
    [anon_sym_impl] = ACTIONS(2245),
    [anon_sym_let] = ACTIONS(2245),
    [anon_sym_loop] = ACTIONS(2245),
    [anon_sym_match] = ACTIONS(2245),
    [anon_sym_mod] = ACTIONS(2245),
    [anon_sym_pub] = ACTIONS(2245),
    [anon_sym_return] = ACTIONS(2245),
    [anon_sym_static] = ACTIONS(2245),
    [anon_sym_struct] = ACTIONS(2245),
    [anon_sym_trait] = ACTIONS(2245),
    [anon_sym_type] = ACTIONS(2245),
    [anon_sym_union] = ACTIONS(2245),
    [anon_sym_unsafe] = ACTIONS(2245),
    [anon_sym_use] = ACTIONS(2245),
    [anon_sym_where] = ACTIONS(2245),
    [anon_sym_while] = ACTIONS(2245),
    [sym_mutable_specifier] = ACTIONS(2245),
    [sym_integer_literal] = ACTIONS(2173),
    [aux_sym_string_literal_token1] = ACTIONS(2175),
    [sym_char_literal] = ACTIONS(2173),
    [anon_sym_true] = ACTIONS(2177),
    [anon_sym_false] = ACTIONS(2177),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2245),
    [sym_super] = ACTIONS(2245),
    [sym_crate] = ACTIONS(2245),
    [sym_raw_string_literal] = ACTIONS(2173),
    [sym_float_literal] = ACTIONS(2173),
    [sym_block_comment] = ACTIONS(3),
  },
  [521] = {
    [sym_delim_token_tree] = STATE(536),
    [sym__delim_tokens] = STATE(536),
    [sym__non_delim_token] = STATE(536),
    [sym__literal] = STATE(536),
    [sym_string_literal] = STATE(592),
    [sym_boolean_literal] = STATE(592),
    [aux_sym_delim_token_tree_repeat1] = STATE(536),
    [sym_identifier] = ACTIONS(2249),
    [anon_sym_LPAREN] = ACTIONS(2163),
    [anon_sym_LBRACE] = ACTIONS(2165),
    [anon_sym_LBRACK] = ACTIONS(2169),
    [anon_sym_RBRACK] = ACTIONS(2251),
    [anon_sym_DOLLAR] = ACTIONS(2253),
    [anon_sym_u8] = ACTIONS(2249),
    [anon_sym_i8] = ACTIONS(2249),
    [anon_sym_u16] = ACTIONS(2249),
    [anon_sym_i16] = ACTIONS(2249),
    [anon_sym_u32] = ACTIONS(2249),
    [anon_sym_i32] = ACTIONS(2249),
    [anon_sym_u64] = ACTIONS(2249),
    [anon_sym_i64] = ACTIONS(2249),
    [anon_sym_u128] = ACTIONS(2249),
    [anon_sym_i128] = ACTIONS(2249),
    [anon_sym_isize] = ACTIONS(2249),
    [anon_sym_usize] = ACTIONS(2249),
    [anon_sym_f32] = ACTIONS(2249),
    [anon_sym_f64] = ACTIONS(2249),
    [anon_sym_bool] = ACTIONS(2249),
    [anon_sym_str] = ACTIONS(2249),
    [anon_sym_char] = ACTIONS(2249),
    [aux_sym__non_special_token_token1] = ACTIONS(2249),
    [anon_sym_SQUOTE] = ACTIONS(2249),
    [anon_sym_as] = ACTIONS(2249),
    [anon_sym_async] = ACTIONS(2249),
    [anon_sym_await] = ACTIONS(2249),
    [anon_sym_break] = ACTIONS(2249),
    [anon_sym_const] = ACTIONS(2249),
    [anon_sym_continue] = ACTIONS(2249),
    [anon_sym_default] = ACTIONS(2249),
    [anon_sym_enum] = ACTIONS(2249),
    [anon_sym_fn] = ACTIONS(2249),
    [anon_sym_for] = ACTIONS(2249),
    [anon_sym_if] = ACTIONS(2249),
    [anon_sym_impl] = ACTIONS(2249),
    [anon_sym_let] = ACTIONS(2249),
    [anon_sym_loop] = ACTIONS(2249),
    [anon_sym_match] = ACTIONS(2249),
    [anon_sym_mod] = ACTIONS(2249),
    [anon_sym_pub] = ACTIONS(2249),
    [anon_sym_return] = ACTIONS(2249),
    [anon_sym_static] = ACTIONS(2249),
    [anon_sym_struct] = ACTIONS(2249),
    [anon_sym_trait] = ACTIONS(2249),
    [anon_sym_type] = ACTIONS(2249),
    [anon_sym_union] = ACTIONS(2249),
    [anon_sym_unsafe] = ACTIONS(2249),
    [anon_sym_use] = ACTIONS(2249),
    [anon_sym_where] = ACTIONS(2249),
    [anon_sym_while] = ACTIONS(2249),
    [sym_mutable_specifier] = ACTIONS(2249),
    [sym_integer_literal] = ACTIONS(2173),
    [aux_sym_string_literal_token1] = ACTIONS(2175),
    [sym_char_literal] = ACTIONS(2173),
    [anon_sym_true] = ACTIONS(2177),
    [anon_sym_false] = ACTIONS(2177),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2249),
    [sym_super] = ACTIONS(2249),
    [sym_crate] = ACTIONS(2249),
    [sym_raw_string_literal] = ACTIONS(2173),
    [sym_float_literal] = ACTIONS(2173),
    [sym_block_comment] = ACTIONS(3),
  },
  [522] = {
    [sym_delim_token_tree] = STATE(540),
    [sym__delim_tokens] = STATE(540),
    [sym__non_delim_token] = STATE(540),
    [sym__literal] = STATE(540),
    [sym_string_literal] = STATE(592),
    [sym_boolean_literal] = STATE(592),
    [aux_sym_delim_token_tree_repeat1] = STATE(540),
    [sym_identifier] = ACTIONS(2255),
    [anon_sym_LPAREN] = ACTIONS(2163),
    [anon_sym_RPAREN] = ACTIONS(2257),
    [anon_sym_LBRACE] = ACTIONS(2165),
    [anon_sym_LBRACK] = ACTIONS(2169),
    [anon_sym_DOLLAR] = ACTIONS(2259),
    [anon_sym_u8] = ACTIONS(2255),
    [anon_sym_i8] = ACTIONS(2255),
    [anon_sym_u16] = ACTIONS(2255),
    [anon_sym_i16] = ACTIONS(2255),
    [anon_sym_u32] = ACTIONS(2255),
    [anon_sym_i32] = ACTIONS(2255),
    [anon_sym_u64] = ACTIONS(2255),
    [anon_sym_i64] = ACTIONS(2255),
    [anon_sym_u128] = ACTIONS(2255),
    [anon_sym_i128] = ACTIONS(2255),
    [anon_sym_isize] = ACTIONS(2255),
    [anon_sym_usize] = ACTIONS(2255),
    [anon_sym_f32] = ACTIONS(2255),
    [anon_sym_f64] = ACTIONS(2255),
    [anon_sym_bool] = ACTIONS(2255),
    [anon_sym_str] = ACTIONS(2255),
    [anon_sym_char] = ACTIONS(2255),
    [aux_sym__non_special_token_token1] = ACTIONS(2255),
    [anon_sym_SQUOTE] = ACTIONS(2255),
    [anon_sym_as] = ACTIONS(2255),
    [anon_sym_async] = ACTIONS(2255),
    [anon_sym_await] = ACTIONS(2255),
    [anon_sym_break] = ACTIONS(2255),
    [anon_sym_const] = ACTIONS(2255),
    [anon_sym_continue] = ACTIONS(2255),
    [anon_sym_default] = ACTIONS(2255),
    [anon_sym_enum] = ACTIONS(2255),
    [anon_sym_fn] = ACTIONS(2255),
    [anon_sym_for] = ACTIONS(2255),
    [anon_sym_if] = ACTIONS(2255),
    [anon_sym_impl] = ACTIONS(2255),
    [anon_sym_let] = ACTIONS(2255),
    [anon_sym_loop] = ACTIONS(2255),
    [anon_sym_match] = ACTIONS(2255),
    [anon_sym_mod] = ACTIONS(2255),
    [anon_sym_pub] = ACTIONS(2255),
    [anon_sym_return] = ACTIONS(2255),
    [anon_sym_static] = ACTIONS(2255),
    [anon_sym_struct] = ACTIONS(2255),
    [anon_sym_trait] = ACTIONS(2255),
    [anon_sym_type] = ACTIONS(2255),
    [anon_sym_union] = ACTIONS(2255),
    [anon_sym_unsafe] = ACTIONS(2255),
    [anon_sym_use] = ACTIONS(2255),
    [anon_sym_where] = ACTIONS(2255),
    [anon_sym_while] = ACTIONS(2255),
    [sym_mutable_specifier] = ACTIONS(2255),
    [sym_integer_literal] = ACTIONS(2173),
    [aux_sym_string_literal_token1] = ACTIONS(2175),
    [sym_char_literal] = ACTIONS(2173),
    [anon_sym_true] = ACTIONS(2177),
    [anon_sym_false] = ACTIONS(2177),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2255),
    [sym_super] = ACTIONS(2255),
    [sym_crate] = ACTIONS(2255),
    [sym_raw_string_literal] = ACTIONS(2173),
    [sym_float_literal] = ACTIONS(2173),
    [sym_block_comment] = ACTIONS(3),
  },
  [523] = {
    [sym_delim_token_tree] = STATE(510),
    [sym__delim_tokens] = STATE(510),
    [sym__non_delim_token] = STATE(510),
    [sym__literal] = STATE(510),
    [sym_string_literal] = STATE(592),
    [sym_boolean_literal] = STATE(592),
    [aux_sym_delim_token_tree_repeat1] = STATE(510),
    [sym_identifier] = ACTIONS(2261),
    [anon_sym_LPAREN] = ACTIONS(2163),
    [anon_sym_RPAREN] = ACTIONS(2199),
    [anon_sym_LBRACE] = ACTIONS(2165),
    [anon_sym_LBRACK] = ACTIONS(2169),
    [anon_sym_DOLLAR] = ACTIONS(2263),
    [anon_sym_u8] = ACTIONS(2261),
    [anon_sym_i8] = ACTIONS(2261),
    [anon_sym_u16] = ACTIONS(2261),
    [anon_sym_i16] = ACTIONS(2261),
    [anon_sym_u32] = ACTIONS(2261),
    [anon_sym_i32] = ACTIONS(2261),
    [anon_sym_u64] = ACTIONS(2261),
    [anon_sym_i64] = ACTIONS(2261),
    [anon_sym_u128] = ACTIONS(2261),
    [anon_sym_i128] = ACTIONS(2261),
    [anon_sym_isize] = ACTIONS(2261),
    [anon_sym_usize] = ACTIONS(2261),
    [anon_sym_f32] = ACTIONS(2261),
    [anon_sym_f64] = ACTIONS(2261),
    [anon_sym_bool] = ACTIONS(2261),
    [anon_sym_str] = ACTIONS(2261),
    [anon_sym_char] = ACTIONS(2261),
    [aux_sym__non_special_token_token1] = ACTIONS(2261),
    [anon_sym_SQUOTE] = ACTIONS(2261),
    [anon_sym_as] = ACTIONS(2261),
    [anon_sym_async] = ACTIONS(2261),
    [anon_sym_await] = ACTIONS(2261),
    [anon_sym_break] = ACTIONS(2261),
    [anon_sym_const] = ACTIONS(2261),
    [anon_sym_continue] = ACTIONS(2261),
    [anon_sym_default] = ACTIONS(2261),
    [anon_sym_enum] = ACTIONS(2261),
    [anon_sym_fn] = ACTIONS(2261),
    [anon_sym_for] = ACTIONS(2261),
    [anon_sym_if] = ACTIONS(2261),
    [anon_sym_impl] = ACTIONS(2261),
    [anon_sym_let] = ACTIONS(2261),
    [anon_sym_loop] = ACTIONS(2261),
    [anon_sym_match] = ACTIONS(2261),
    [anon_sym_mod] = ACTIONS(2261),
    [anon_sym_pub] = ACTIONS(2261),
    [anon_sym_return] = ACTIONS(2261),
    [anon_sym_static] = ACTIONS(2261),
    [anon_sym_struct] = ACTIONS(2261),
    [anon_sym_trait] = ACTIONS(2261),
    [anon_sym_type] = ACTIONS(2261),
    [anon_sym_union] = ACTIONS(2261),
    [anon_sym_unsafe] = ACTIONS(2261),
    [anon_sym_use] = ACTIONS(2261),
    [anon_sym_where] = ACTIONS(2261),
    [anon_sym_while] = ACTIONS(2261),
    [sym_mutable_specifier] = ACTIONS(2261),
    [sym_integer_literal] = ACTIONS(2173),
    [aux_sym_string_literal_token1] = ACTIONS(2175),
    [sym_char_literal] = ACTIONS(2173),
    [anon_sym_true] = ACTIONS(2177),
    [anon_sym_false] = ACTIONS(2177),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2261),
    [sym_super] = ACTIONS(2261),
    [sym_crate] = ACTIONS(2261),
    [sym_raw_string_literal] = ACTIONS(2173),
    [sym_float_literal] = ACTIONS(2173),
    [sym_block_comment] = ACTIONS(3),
  },
  [524] = {
    [sym_delim_token_tree] = STATE(483),
    [sym__delim_tokens] = STATE(483),
    [sym__non_delim_token] = STATE(483),
    [sym__literal] = STATE(483),
    [sym_string_literal] = STATE(592),
    [sym_boolean_literal] = STATE(592),
    [aux_sym_delim_token_tree_repeat1] = STATE(483),
    [sym_identifier] = ACTIONS(2161),
    [anon_sym_LPAREN] = ACTIONS(2163),
    [anon_sym_RPAREN] = ACTIONS(2167),
    [anon_sym_LBRACE] = ACTIONS(2165),
    [anon_sym_LBRACK] = ACTIONS(2169),
    [anon_sym_DOLLAR] = ACTIONS(2171),
    [anon_sym_u8] = ACTIONS(2161),
    [anon_sym_i8] = ACTIONS(2161),
    [anon_sym_u16] = ACTIONS(2161),
    [anon_sym_i16] = ACTIONS(2161),
    [anon_sym_u32] = ACTIONS(2161),
    [anon_sym_i32] = ACTIONS(2161),
    [anon_sym_u64] = ACTIONS(2161),
    [anon_sym_i64] = ACTIONS(2161),
    [anon_sym_u128] = ACTIONS(2161),
    [anon_sym_i128] = ACTIONS(2161),
    [anon_sym_isize] = ACTIONS(2161),
    [anon_sym_usize] = ACTIONS(2161),
    [anon_sym_f32] = ACTIONS(2161),
    [anon_sym_f64] = ACTIONS(2161),
    [anon_sym_bool] = ACTIONS(2161),
    [anon_sym_str] = ACTIONS(2161),
    [anon_sym_char] = ACTIONS(2161),
    [aux_sym__non_special_token_token1] = ACTIONS(2161),
    [anon_sym_SQUOTE] = ACTIONS(2161),
    [anon_sym_as] = ACTIONS(2161),
    [anon_sym_async] = ACTIONS(2161),
    [anon_sym_await] = ACTIONS(2161),
    [anon_sym_break] = ACTIONS(2161),
    [anon_sym_const] = ACTIONS(2161),
    [anon_sym_continue] = ACTIONS(2161),
    [anon_sym_default] = ACTIONS(2161),
    [anon_sym_enum] = ACTIONS(2161),
    [anon_sym_fn] = ACTIONS(2161),
    [anon_sym_for] = ACTIONS(2161),
    [anon_sym_if] = ACTIONS(2161),
    [anon_sym_impl] = ACTIONS(2161),
    [anon_sym_let] = ACTIONS(2161),
    [anon_sym_loop] = ACTIONS(2161),
    [anon_sym_match] = ACTIONS(2161),
    [anon_sym_mod] = ACTIONS(2161),
    [anon_sym_pub] = ACTIONS(2161),
    [anon_sym_return] = ACTIONS(2161),
    [anon_sym_static] = ACTIONS(2161),
    [anon_sym_struct] = ACTIONS(2161),
    [anon_sym_trait] = ACTIONS(2161),
    [anon_sym_type] = ACTIONS(2161),
    [anon_sym_union] = ACTIONS(2161),
    [anon_sym_unsafe] = ACTIONS(2161),
    [anon_sym_use] = ACTIONS(2161),
    [anon_sym_where] = ACTIONS(2161),
    [anon_sym_while] = ACTIONS(2161),
    [sym_mutable_specifier] = ACTIONS(2161),
    [sym_integer_literal] = ACTIONS(2173),
    [aux_sym_string_literal_token1] = ACTIONS(2175),
    [sym_char_literal] = ACTIONS(2173),
    [anon_sym_true] = ACTIONS(2177),
    [anon_sym_false] = ACTIONS(2177),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2161),
    [sym_super] = ACTIONS(2161),
    [sym_crate] = ACTIONS(2161),
    [sym_raw_string_literal] = ACTIONS(2173),
    [sym_float_literal] = ACTIONS(2173),
    [sym_block_comment] = ACTIONS(3),
  },
  [525] = {
    [sym_token_tree] = STATE(485),
    [sym_token_repetition] = STATE(485),
    [sym__literal] = STATE(485),
    [sym_string_literal] = STATE(585),
    [sym_boolean_literal] = STATE(585),
    [aux_sym_token_tree_repeat1] = STATE(485),
    [sym_identifier] = ACTIONS(2181),
    [anon_sym_LPAREN] = ACTIONS(2183),
    [anon_sym_LBRACE] = ACTIONS(2185),
    [anon_sym_LBRACK] = ACTIONS(2189),
    [anon_sym_RBRACK] = ACTIONS(2187),
    [anon_sym_DOLLAR] = ACTIONS(2191),
    [anon_sym_u8] = ACTIONS(2181),
    [anon_sym_i8] = ACTIONS(2181),
    [anon_sym_u16] = ACTIONS(2181),
    [anon_sym_i16] = ACTIONS(2181),
    [anon_sym_u32] = ACTIONS(2181),
    [anon_sym_i32] = ACTIONS(2181),
    [anon_sym_u64] = ACTIONS(2181),
    [anon_sym_i64] = ACTIONS(2181),
    [anon_sym_u128] = ACTIONS(2181),
    [anon_sym_i128] = ACTIONS(2181),
    [anon_sym_isize] = ACTIONS(2181),
    [anon_sym_usize] = ACTIONS(2181),
    [anon_sym_f32] = ACTIONS(2181),
    [anon_sym_f64] = ACTIONS(2181),
    [anon_sym_bool] = ACTIONS(2181),
    [anon_sym_str] = ACTIONS(2181),
    [anon_sym_char] = ACTIONS(2181),
    [aux_sym__non_special_token_token1] = ACTIONS(2181),
    [anon_sym_SQUOTE] = ACTIONS(2181),
    [anon_sym_as] = ACTIONS(2181),
    [anon_sym_async] = ACTIONS(2181),
    [anon_sym_await] = ACTIONS(2181),
    [anon_sym_break] = ACTIONS(2181),
    [anon_sym_const] = ACTIONS(2181),
    [anon_sym_continue] = ACTIONS(2181),
    [anon_sym_default] = ACTIONS(2181),
    [anon_sym_enum] = ACTIONS(2181),
    [anon_sym_fn] = ACTIONS(2181),
    [anon_sym_for] = ACTIONS(2181),
    [anon_sym_if] = ACTIONS(2181),
    [anon_sym_impl] = ACTIONS(2181),
    [anon_sym_let] = ACTIONS(2181),
    [anon_sym_loop] = ACTIONS(2181),
    [anon_sym_match] = ACTIONS(2181),
    [anon_sym_mod] = ACTIONS(2181),
    [anon_sym_pub] = ACTIONS(2181),
    [anon_sym_return] = ACTIONS(2181),
    [anon_sym_static] = ACTIONS(2181),
    [anon_sym_struct] = ACTIONS(2181),
    [anon_sym_trait] = ACTIONS(2181),
    [anon_sym_type] = ACTIONS(2181),
    [anon_sym_union] = ACTIONS(2181),
    [anon_sym_unsafe] = ACTIONS(2181),
    [anon_sym_use] = ACTIONS(2181),
    [anon_sym_where] = ACTIONS(2181),
    [anon_sym_while] = ACTIONS(2181),
    [sym_mutable_specifier] = ACTIONS(2181),
    [sym_integer_literal] = ACTIONS(2014),
    [aux_sym_string_literal_token1] = ACTIONS(2016),
    [sym_char_literal] = ACTIONS(2014),
    [anon_sym_true] = ACTIONS(2018),
    [anon_sym_false] = ACTIONS(2018),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2181),
    [sym_super] = ACTIONS(2181),
    [sym_crate] = ACTIONS(2181),
    [sym_metavariable] = ACTIONS(2193),
    [sym_raw_string_literal] = ACTIONS(2014),
    [sym_float_literal] = ACTIONS(2014),
    [sym_block_comment] = ACTIONS(3),
  },
  [526] = {
    [sym_delim_token_tree] = STATE(501),
    [sym__delim_tokens] = STATE(501),
    [sym__non_delim_token] = STATE(501),
    [sym__literal] = STATE(501),
    [sym_string_literal] = STATE(592),
    [sym_boolean_literal] = STATE(592),
    [aux_sym_delim_token_tree_repeat1] = STATE(501),
    [sym_identifier] = ACTIONS(2265),
    [anon_sym_LPAREN] = ACTIONS(2163),
    [anon_sym_LBRACE] = ACTIONS(2165),
    [anon_sym_RBRACE] = ACTIONS(2251),
    [anon_sym_LBRACK] = ACTIONS(2169),
    [anon_sym_DOLLAR] = ACTIONS(2267),
    [anon_sym_u8] = ACTIONS(2265),
    [anon_sym_i8] = ACTIONS(2265),
    [anon_sym_u16] = ACTIONS(2265),
    [anon_sym_i16] = ACTIONS(2265),
    [anon_sym_u32] = ACTIONS(2265),
    [anon_sym_i32] = ACTIONS(2265),
    [anon_sym_u64] = ACTIONS(2265),
    [anon_sym_i64] = ACTIONS(2265),
    [anon_sym_u128] = ACTIONS(2265),
    [anon_sym_i128] = ACTIONS(2265),
    [anon_sym_isize] = ACTIONS(2265),
    [anon_sym_usize] = ACTIONS(2265),
    [anon_sym_f32] = ACTIONS(2265),
    [anon_sym_f64] = ACTIONS(2265),
    [anon_sym_bool] = ACTIONS(2265),
    [anon_sym_str] = ACTIONS(2265),
    [anon_sym_char] = ACTIONS(2265),
    [aux_sym__non_special_token_token1] = ACTIONS(2265),
    [anon_sym_SQUOTE] = ACTIONS(2265),
    [anon_sym_as] = ACTIONS(2265),
    [anon_sym_async] = ACTIONS(2265),
    [anon_sym_await] = ACTIONS(2265),
    [anon_sym_break] = ACTIONS(2265),
    [anon_sym_const] = ACTIONS(2265),
    [anon_sym_continue] = ACTIONS(2265),
    [anon_sym_default] = ACTIONS(2265),
    [anon_sym_enum] = ACTIONS(2265),
    [anon_sym_fn] = ACTIONS(2265),
    [anon_sym_for] = ACTIONS(2265),
    [anon_sym_if] = ACTIONS(2265),
    [anon_sym_impl] = ACTIONS(2265),
    [anon_sym_let] = ACTIONS(2265),
    [anon_sym_loop] = ACTIONS(2265),
    [anon_sym_match] = ACTIONS(2265),
    [anon_sym_mod] = ACTIONS(2265),
    [anon_sym_pub] = ACTIONS(2265),
    [anon_sym_return] = ACTIONS(2265),
    [anon_sym_static] = ACTIONS(2265),
    [anon_sym_struct] = ACTIONS(2265),
    [anon_sym_trait] = ACTIONS(2265),
    [anon_sym_type] = ACTIONS(2265),
    [anon_sym_union] = ACTIONS(2265),
    [anon_sym_unsafe] = ACTIONS(2265),
    [anon_sym_use] = ACTIONS(2265),
    [anon_sym_where] = ACTIONS(2265),
    [anon_sym_while] = ACTIONS(2265),
    [sym_mutable_specifier] = ACTIONS(2265),
    [sym_integer_literal] = ACTIONS(2173),
    [aux_sym_string_literal_token1] = ACTIONS(2175),
    [sym_char_literal] = ACTIONS(2173),
    [anon_sym_true] = ACTIONS(2177),
    [anon_sym_false] = ACTIONS(2177),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2265),
    [sym_super] = ACTIONS(2265),
    [sym_crate] = ACTIONS(2265),
    [sym_raw_string_literal] = ACTIONS(2173),
    [sym_float_literal] = ACTIONS(2173),
    [sym_block_comment] = ACTIONS(3),
  },
  [527] = {
    [sym_delim_token_tree] = STATE(483),
    [sym__delim_tokens] = STATE(483),
    [sym__non_delim_token] = STATE(483),
    [sym__literal] = STATE(483),
    [sym_string_literal] = STATE(592),
    [sym_boolean_literal] = STATE(592),
    [aux_sym_delim_token_tree_repeat1] = STATE(483),
    [sym_identifier] = ACTIONS(2161),
    [anon_sym_LPAREN] = ACTIONS(2163),
    [anon_sym_RPAREN] = ACTIONS(2269),
    [anon_sym_LBRACE] = ACTIONS(2165),
    [anon_sym_LBRACK] = ACTIONS(2169),
    [anon_sym_DOLLAR] = ACTIONS(2171),
    [anon_sym_u8] = ACTIONS(2161),
    [anon_sym_i8] = ACTIONS(2161),
    [anon_sym_u16] = ACTIONS(2161),
    [anon_sym_i16] = ACTIONS(2161),
    [anon_sym_u32] = ACTIONS(2161),
    [anon_sym_i32] = ACTIONS(2161),
    [anon_sym_u64] = ACTIONS(2161),
    [anon_sym_i64] = ACTIONS(2161),
    [anon_sym_u128] = ACTIONS(2161),
    [anon_sym_i128] = ACTIONS(2161),
    [anon_sym_isize] = ACTIONS(2161),
    [anon_sym_usize] = ACTIONS(2161),
    [anon_sym_f32] = ACTIONS(2161),
    [anon_sym_f64] = ACTIONS(2161),
    [anon_sym_bool] = ACTIONS(2161),
    [anon_sym_str] = ACTIONS(2161),
    [anon_sym_char] = ACTIONS(2161),
    [aux_sym__non_special_token_token1] = ACTIONS(2161),
    [anon_sym_SQUOTE] = ACTIONS(2161),
    [anon_sym_as] = ACTIONS(2161),
    [anon_sym_async] = ACTIONS(2161),
    [anon_sym_await] = ACTIONS(2161),
    [anon_sym_break] = ACTIONS(2161),
    [anon_sym_const] = ACTIONS(2161),
    [anon_sym_continue] = ACTIONS(2161),
    [anon_sym_default] = ACTIONS(2161),
    [anon_sym_enum] = ACTIONS(2161),
    [anon_sym_fn] = ACTIONS(2161),
    [anon_sym_for] = ACTIONS(2161),
    [anon_sym_if] = ACTIONS(2161),
    [anon_sym_impl] = ACTIONS(2161),
    [anon_sym_let] = ACTIONS(2161),
    [anon_sym_loop] = ACTIONS(2161),
    [anon_sym_match] = ACTIONS(2161),
    [anon_sym_mod] = ACTIONS(2161),
    [anon_sym_pub] = ACTIONS(2161),
    [anon_sym_return] = ACTIONS(2161),
    [anon_sym_static] = ACTIONS(2161),
    [anon_sym_struct] = ACTIONS(2161),
    [anon_sym_trait] = ACTIONS(2161),
    [anon_sym_type] = ACTIONS(2161),
    [anon_sym_union] = ACTIONS(2161),
    [anon_sym_unsafe] = ACTIONS(2161),
    [anon_sym_use] = ACTIONS(2161),
    [anon_sym_where] = ACTIONS(2161),
    [anon_sym_while] = ACTIONS(2161),
    [sym_mutable_specifier] = ACTIONS(2161),
    [sym_integer_literal] = ACTIONS(2173),
    [aux_sym_string_literal_token1] = ACTIONS(2175),
    [sym_char_literal] = ACTIONS(2173),
    [anon_sym_true] = ACTIONS(2177),
    [anon_sym_false] = ACTIONS(2177),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2161),
    [sym_super] = ACTIONS(2161),
    [sym_crate] = ACTIONS(2161),
    [sym_raw_string_literal] = ACTIONS(2173),
    [sym_float_literal] = ACTIONS(2173),
    [sym_block_comment] = ACTIONS(3),
  },
  [528] = {
    [sym_delim_token_tree] = STATE(483),
    [sym__delim_tokens] = STATE(483),
    [sym__non_delim_token] = STATE(483),
    [sym__literal] = STATE(483),
    [sym_string_literal] = STATE(592),
    [sym_boolean_literal] = STATE(592),
    [aux_sym_delim_token_tree_repeat1] = STATE(483),
    [sym_identifier] = ACTIONS(2161),
    [anon_sym_LPAREN] = ACTIONS(2163),
    [anon_sym_LBRACE] = ACTIONS(2165),
    [anon_sym_RBRACE] = ACTIONS(2269),
    [anon_sym_LBRACK] = ACTIONS(2169),
    [anon_sym_DOLLAR] = ACTIONS(2171),
    [anon_sym_u8] = ACTIONS(2161),
    [anon_sym_i8] = ACTIONS(2161),
    [anon_sym_u16] = ACTIONS(2161),
    [anon_sym_i16] = ACTIONS(2161),
    [anon_sym_u32] = ACTIONS(2161),
    [anon_sym_i32] = ACTIONS(2161),
    [anon_sym_u64] = ACTIONS(2161),
    [anon_sym_i64] = ACTIONS(2161),
    [anon_sym_u128] = ACTIONS(2161),
    [anon_sym_i128] = ACTIONS(2161),
    [anon_sym_isize] = ACTIONS(2161),
    [anon_sym_usize] = ACTIONS(2161),
    [anon_sym_f32] = ACTIONS(2161),
    [anon_sym_f64] = ACTIONS(2161),
    [anon_sym_bool] = ACTIONS(2161),
    [anon_sym_str] = ACTIONS(2161),
    [anon_sym_char] = ACTIONS(2161),
    [aux_sym__non_special_token_token1] = ACTIONS(2161),
    [anon_sym_SQUOTE] = ACTIONS(2161),
    [anon_sym_as] = ACTIONS(2161),
    [anon_sym_async] = ACTIONS(2161),
    [anon_sym_await] = ACTIONS(2161),
    [anon_sym_break] = ACTIONS(2161),
    [anon_sym_const] = ACTIONS(2161),
    [anon_sym_continue] = ACTIONS(2161),
    [anon_sym_default] = ACTIONS(2161),
    [anon_sym_enum] = ACTIONS(2161),
    [anon_sym_fn] = ACTIONS(2161),
    [anon_sym_for] = ACTIONS(2161),
    [anon_sym_if] = ACTIONS(2161),
    [anon_sym_impl] = ACTIONS(2161),
    [anon_sym_let] = ACTIONS(2161),
    [anon_sym_loop] = ACTIONS(2161),
    [anon_sym_match] = ACTIONS(2161),
    [anon_sym_mod] = ACTIONS(2161),
    [anon_sym_pub] = ACTIONS(2161),
    [anon_sym_return] = ACTIONS(2161),
    [anon_sym_static] = ACTIONS(2161),
    [anon_sym_struct] = ACTIONS(2161),
    [anon_sym_trait] = ACTIONS(2161),
    [anon_sym_type] = ACTIONS(2161),
    [anon_sym_union] = ACTIONS(2161),
    [anon_sym_unsafe] = ACTIONS(2161),
    [anon_sym_use] = ACTIONS(2161),
    [anon_sym_where] = ACTIONS(2161),
    [anon_sym_while] = ACTIONS(2161),
    [sym_mutable_specifier] = ACTIONS(2161),
    [sym_integer_literal] = ACTIONS(2173),
    [aux_sym_string_literal_token1] = ACTIONS(2175),
    [sym_char_literal] = ACTIONS(2173),
    [anon_sym_true] = ACTIONS(2177),
    [anon_sym_false] = ACTIONS(2177),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2161),
    [sym_super] = ACTIONS(2161),
    [sym_crate] = ACTIONS(2161),
    [sym_raw_string_literal] = ACTIONS(2173),
    [sym_float_literal] = ACTIONS(2173),
    [sym_block_comment] = ACTIONS(3),
  },
  [529] = {
    [sym_delim_token_tree] = STATE(483),
    [sym__delim_tokens] = STATE(483),
    [sym__non_delim_token] = STATE(483),
    [sym__literal] = STATE(483),
    [sym_string_literal] = STATE(592),
    [sym_boolean_literal] = STATE(592),
    [aux_sym_delim_token_tree_repeat1] = STATE(483),
    [sym_identifier] = ACTIONS(2161),
    [anon_sym_LPAREN] = ACTIONS(2163),
    [anon_sym_LBRACE] = ACTIONS(2165),
    [anon_sym_LBRACK] = ACTIONS(2169),
    [anon_sym_RBRACK] = ACTIONS(2269),
    [anon_sym_DOLLAR] = ACTIONS(2171),
    [anon_sym_u8] = ACTIONS(2161),
    [anon_sym_i8] = ACTIONS(2161),
    [anon_sym_u16] = ACTIONS(2161),
    [anon_sym_i16] = ACTIONS(2161),
    [anon_sym_u32] = ACTIONS(2161),
    [anon_sym_i32] = ACTIONS(2161),
    [anon_sym_u64] = ACTIONS(2161),
    [anon_sym_i64] = ACTIONS(2161),
    [anon_sym_u128] = ACTIONS(2161),
    [anon_sym_i128] = ACTIONS(2161),
    [anon_sym_isize] = ACTIONS(2161),
    [anon_sym_usize] = ACTIONS(2161),
    [anon_sym_f32] = ACTIONS(2161),
    [anon_sym_f64] = ACTIONS(2161),
    [anon_sym_bool] = ACTIONS(2161),
    [anon_sym_str] = ACTIONS(2161),
    [anon_sym_char] = ACTIONS(2161),
    [aux_sym__non_special_token_token1] = ACTIONS(2161),
    [anon_sym_SQUOTE] = ACTIONS(2161),
    [anon_sym_as] = ACTIONS(2161),
    [anon_sym_async] = ACTIONS(2161),
    [anon_sym_await] = ACTIONS(2161),
    [anon_sym_break] = ACTIONS(2161),
    [anon_sym_const] = ACTIONS(2161),
    [anon_sym_continue] = ACTIONS(2161),
    [anon_sym_default] = ACTIONS(2161),
    [anon_sym_enum] = ACTIONS(2161),
    [anon_sym_fn] = ACTIONS(2161),
    [anon_sym_for] = ACTIONS(2161),
    [anon_sym_if] = ACTIONS(2161),
    [anon_sym_impl] = ACTIONS(2161),
    [anon_sym_let] = ACTIONS(2161),
    [anon_sym_loop] = ACTIONS(2161),
    [anon_sym_match] = ACTIONS(2161),
    [anon_sym_mod] = ACTIONS(2161),
    [anon_sym_pub] = ACTIONS(2161),
    [anon_sym_return] = ACTIONS(2161),
    [anon_sym_static] = ACTIONS(2161),
    [anon_sym_struct] = ACTIONS(2161),
    [anon_sym_trait] = ACTIONS(2161),
    [anon_sym_type] = ACTIONS(2161),
    [anon_sym_union] = ACTIONS(2161),
    [anon_sym_unsafe] = ACTIONS(2161),
    [anon_sym_use] = ACTIONS(2161),
    [anon_sym_where] = ACTIONS(2161),
    [anon_sym_while] = ACTIONS(2161),
    [sym_mutable_specifier] = ACTIONS(2161),
    [sym_integer_literal] = ACTIONS(2173),
    [aux_sym_string_literal_token1] = ACTIONS(2175),
    [sym_char_literal] = ACTIONS(2173),
    [anon_sym_true] = ACTIONS(2177),
    [anon_sym_false] = ACTIONS(2177),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2161),
    [sym_super] = ACTIONS(2161),
    [sym_crate] = ACTIONS(2161),
    [sym_raw_string_literal] = ACTIONS(2173),
    [sym_float_literal] = ACTIONS(2173),
    [sym_block_comment] = ACTIONS(3),
  },
  [530] = {
    [sym_delim_token_tree] = STATE(543),
    [sym__delim_tokens] = STATE(543),
    [sym__non_delim_token] = STATE(543),
    [sym__literal] = STATE(543),
    [sym_string_literal] = STATE(592),
    [sym_boolean_literal] = STATE(592),
    [aux_sym_delim_token_tree_repeat1] = STATE(543),
    [sym_identifier] = ACTIONS(2271),
    [anon_sym_LPAREN] = ACTIONS(2163),
    [anon_sym_LBRACE] = ACTIONS(2165),
    [anon_sym_RBRACE] = ACTIONS(2257),
    [anon_sym_LBRACK] = ACTIONS(2169),
    [anon_sym_DOLLAR] = ACTIONS(2273),
    [anon_sym_u8] = ACTIONS(2271),
    [anon_sym_i8] = ACTIONS(2271),
    [anon_sym_u16] = ACTIONS(2271),
    [anon_sym_i16] = ACTIONS(2271),
    [anon_sym_u32] = ACTIONS(2271),
    [anon_sym_i32] = ACTIONS(2271),
    [anon_sym_u64] = ACTIONS(2271),
    [anon_sym_i64] = ACTIONS(2271),
    [anon_sym_u128] = ACTIONS(2271),
    [anon_sym_i128] = ACTIONS(2271),
    [anon_sym_isize] = ACTIONS(2271),
    [anon_sym_usize] = ACTIONS(2271),
    [anon_sym_f32] = ACTIONS(2271),
    [anon_sym_f64] = ACTIONS(2271),
    [anon_sym_bool] = ACTIONS(2271),
    [anon_sym_str] = ACTIONS(2271),
    [anon_sym_char] = ACTIONS(2271),
    [aux_sym__non_special_token_token1] = ACTIONS(2271),
    [anon_sym_SQUOTE] = ACTIONS(2271),
    [anon_sym_as] = ACTIONS(2271),
    [anon_sym_async] = ACTIONS(2271),
    [anon_sym_await] = ACTIONS(2271),
    [anon_sym_break] = ACTIONS(2271),
    [anon_sym_const] = ACTIONS(2271),
    [anon_sym_continue] = ACTIONS(2271),
    [anon_sym_default] = ACTIONS(2271),
    [anon_sym_enum] = ACTIONS(2271),
    [anon_sym_fn] = ACTIONS(2271),
    [anon_sym_for] = ACTIONS(2271),
    [anon_sym_if] = ACTIONS(2271),
    [anon_sym_impl] = ACTIONS(2271),
    [anon_sym_let] = ACTIONS(2271),
    [anon_sym_loop] = ACTIONS(2271),
    [anon_sym_match] = ACTIONS(2271),
    [anon_sym_mod] = ACTIONS(2271),
    [anon_sym_pub] = ACTIONS(2271),
    [anon_sym_return] = ACTIONS(2271),
    [anon_sym_static] = ACTIONS(2271),
    [anon_sym_struct] = ACTIONS(2271),
    [anon_sym_trait] = ACTIONS(2271),
    [anon_sym_type] = ACTIONS(2271),
    [anon_sym_union] = ACTIONS(2271),
    [anon_sym_unsafe] = ACTIONS(2271),
    [anon_sym_use] = ACTIONS(2271),
    [anon_sym_where] = ACTIONS(2271),
    [anon_sym_while] = ACTIONS(2271),
    [sym_mutable_specifier] = ACTIONS(2271),
    [sym_integer_literal] = ACTIONS(2173),
    [aux_sym_string_literal_token1] = ACTIONS(2175),
    [sym_char_literal] = ACTIONS(2173),
    [anon_sym_true] = ACTIONS(2177),
    [anon_sym_false] = ACTIONS(2177),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2271),
    [sym_super] = ACTIONS(2271),
    [sym_crate] = ACTIONS(2271),
    [sym_raw_string_literal] = ACTIONS(2173),
    [sym_float_literal] = ACTIONS(2173),
    [sym_block_comment] = ACTIONS(3),
  },
  [531] = {
    [sym_delim_token_tree] = STATE(524),
    [sym__delim_tokens] = STATE(524),
    [sym__non_delim_token] = STATE(524),
    [sym__literal] = STATE(524),
    [sym_string_literal] = STATE(592),
    [sym_boolean_literal] = STATE(592),
    [aux_sym_delim_token_tree_repeat1] = STATE(524),
    [sym_identifier] = ACTIONS(2275),
    [anon_sym_LPAREN] = ACTIONS(2163),
    [anon_sym_RPAREN] = ACTIONS(2251),
    [anon_sym_LBRACE] = ACTIONS(2165),
    [anon_sym_LBRACK] = ACTIONS(2169),
    [anon_sym_DOLLAR] = ACTIONS(2277),
    [anon_sym_u8] = ACTIONS(2275),
    [anon_sym_i8] = ACTIONS(2275),
    [anon_sym_u16] = ACTIONS(2275),
    [anon_sym_i16] = ACTIONS(2275),
    [anon_sym_u32] = ACTIONS(2275),
    [anon_sym_i32] = ACTIONS(2275),
    [anon_sym_u64] = ACTIONS(2275),
    [anon_sym_i64] = ACTIONS(2275),
    [anon_sym_u128] = ACTIONS(2275),
    [anon_sym_i128] = ACTIONS(2275),
    [anon_sym_isize] = ACTIONS(2275),
    [anon_sym_usize] = ACTIONS(2275),
    [anon_sym_f32] = ACTIONS(2275),
    [anon_sym_f64] = ACTIONS(2275),
    [anon_sym_bool] = ACTIONS(2275),
    [anon_sym_str] = ACTIONS(2275),
    [anon_sym_char] = ACTIONS(2275),
    [aux_sym__non_special_token_token1] = ACTIONS(2275),
    [anon_sym_SQUOTE] = ACTIONS(2275),
    [anon_sym_as] = ACTIONS(2275),
    [anon_sym_async] = ACTIONS(2275),
    [anon_sym_await] = ACTIONS(2275),
    [anon_sym_break] = ACTIONS(2275),
    [anon_sym_const] = ACTIONS(2275),
    [anon_sym_continue] = ACTIONS(2275),
    [anon_sym_default] = ACTIONS(2275),
    [anon_sym_enum] = ACTIONS(2275),
    [anon_sym_fn] = ACTIONS(2275),
    [anon_sym_for] = ACTIONS(2275),
    [anon_sym_if] = ACTIONS(2275),
    [anon_sym_impl] = ACTIONS(2275),
    [anon_sym_let] = ACTIONS(2275),
    [anon_sym_loop] = ACTIONS(2275),
    [anon_sym_match] = ACTIONS(2275),
    [anon_sym_mod] = ACTIONS(2275),
    [anon_sym_pub] = ACTIONS(2275),
    [anon_sym_return] = ACTIONS(2275),
    [anon_sym_static] = ACTIONS(2275),
    [anon_sym_struct] = ACTIONS(2275),
    [anon_sym_trait] = ACTIONS(2275),
    [anon_sym_type] = ACTIONS(2275),
    [anon_sym_union] = ACTIONS(2275),
    [anon_sym_unsafe] = ACTIONS(2275),
    [anon_sym_use] = ACTIONS(2275),
    [anon_sym_where] = ACTIONS(2275),
    [anon_sym_while] = ACTIONS(2275),
    [sym_mutable_specifier] = ACTIONS(2275),
    [sym_integer_literal] = ACTIONS(2173),
    [aux_sym_string_literal_token1] = ACTIONS(2175),
    [sym_char_literal] = ACTIONS(2173),
    [anon_sym_true] = ACTIONS(2177),
    [anon_sym_false] = ACTIONS(2177),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2275),
    [sym_super] = ACTIONS(2275),
    [sym_crate] = ACTIONS(2275),
    [sym_raw_string_literal] = ACTIONS(2173),
    [sym_float_literal] = ACTIONS(2173),
    [sym_block_comment] = ACTIONS(3),
  },
  [532] = {
    [sym_delim_token_tree] = STATE(527),
    [sym__delim_tokens] = STATE(527),
    [sym__non_delim_token] = STATE(527),
    [sym__literal] = STATE(527),
    [sym_string_literal] = STATE(592),
    [sym_boolean_literal] = STATE(592),
    [aux_sym_delim_token_tree_repeat1] = STATE(527),
    [sym_identifier] = ACTIONS(2279),
    [anon_sym_LPAREN] = ACTIONS(2163),
    [anon_sym_RPAREN] = ACTIONS(2221),
    [anon_sym_LBRACE] = ACTIONS(2165),
    [anon_sym_LBRACK] = ACTIONS(2169),
    [anon_sym_DOLLAR] = ACTIONS(2281),
    [anon_sym_u8] = ACTIONS(2279),
    [anon_sym_i8] = ACTIONS(2279),
    [anon_sym_u16] = ACTIONS(2279),
    [anon_sym_i16] = ACTIONS(2279),
    [anon_sym_u32] = ACTIONS(2279),
    [anon_sym_i32] = ACTIONS(2279),
    [anon_sym_u64] = ACTIONS(2279),
    [anon_sym_i64] = ACTIONS(2279),
    [anon_sym_u128] = ACTIONS(2279),
    [anon_sym_i128] = ACTIONS(2279),
    [anon_sym_isize] = ACTIONS(2279),
    [anon_sym_usize] = ACTIONS(2279),
    [anon_sym_f32] = ACTIONS(2279),
    [anon_sym_f64] = ACTIONS(2279),
    [anon_sym_bool] = ACTIONS(2279),
    [anon_sym_str] = ACTIONS(2279),
    [anon_sym_char] = ACTIONS(2279),
    [aux_sym__non_special_token_token1] = ACTIONS(2279),
    [anon_sym_SQUOTE] = ACTIONS(2279),
    [anon_sym_as] = ACTIONS(2279),
    [anon_sym_async] = ACTIONS(2279),
    [anon_sym_await] = ACTIONS(2279),
    [anon_sym_break] = ACTIONS(2279),
    [anon_sym_const] = ACTIONS(2279),
    [anon_sym_continue] = ACTIONS(2279),
    [anon_sym_default] = ACTIONS(2279),
    [anon_sym_enum] = ACTIONS(2279),
    [anon_sym_fn] = ACTIONS(2279),
    [anon_sym_for] = ACTIONS(2279),
    [anon_sym_if] = ACTIONS(2279),
    [anon_sym_impl] = ACTIONS(2279),
    [anon_sym_let] = ACTIONS(2279),
    [anon_sym_loop] = ACTIONS(2279),
    [anon_sym_match] = ACTIONS(2279),
    [anon_sym_mod] = ACTIONS(2279),
    [anon_sym_pub] = ACTIONS(2279),
    [anon_sym_return] = ACTIONS(2279),
    [anon_sym_static] = ACTIONS(2279),
    [anon_sym_struct] = ACTIONS(2279),
    [anon_sym_trait] = ACTIONS(2279),
    [anon_sym_type] = ACTIONS(2279),
    [anon_sym_union] = ACTIONS(2279),
    [anon_sym_unsafe] = ACTIONS(2279),
    [anon_sym_use] = ACTIONS(2279),
    [anon_sym_where] = ACTIONS(2279),
    [anon_sym_while] = ACTIONS(2279),
    [sym_mutable_specifier] = ACTIONS(2279),
    [sym_integer_literal] = ACTIONS(2173),
    [aux_sym_string_literal_token1] = ACTIONS(2175),
    [sym_char_literal] = ACTIONS(2173),
    [anon_sym_true] = ACTIONS(2177),
    [anon_sym_false] = ACTIONS(2177),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2279),
    [sym_super] = ACTIONS(2279),
    [sym_crate] = ACTIONS(2279),
    [sym_raw_string_literal] = ACTIONS(2173),
    [sym_float_literal] = ACTIONS(2173),
    [sym_block_comment] = ACTIONS(3),
  },
  [533] = {
    [sym_token_tree] = STATE(538),
    [sym_token_repetition] = STATE(538),
    [sym__literal] = STATE(538),
    [sym_string_literal] = STATE(585),
    [sym_boolean_literal] = STATE(585),
    [aux_sym_token_tree_repeat1] = STATE(538),
    [sym_identifier] = ACTIONS(2283),
    [anon_sym_LPAREN] = ACTIONS(2183),
    [anon_sym_LBRACE] = ACTIONS(2185),
    [anon_sym_LBRACK] = ACTIONS(2189),
    [anon_sym_RBRACK] = ACTIONS(2285),
    [anon_sym_DOLLAR] = ACTIONS(2191),
    [anon_sym_u8] = ACTIONS(2283),
    [anon_sym_i8] = ACTIONS(2283),
    [anon_sym_u16] = ACTIONS(2283),
    [anon_sym_i16] = ACTIONS(2283),
    [anon_sym_u32] = ACTIONS(2283),
    [anon_sym_i32] = ACTIONS(2283),
    [anon_sym_u64] = ACTIONS(2283),
    [anon_sym_i64] = ACTIONS(2283),
    [anon_sym_u128] = ACTIONS(2283),
    [anon_sym_i128] = ACTIONS(2283),
    [anon_sym_isize] = ACTIONS(2283),
    [anon_sym_usize] = ACTIONS(2283),
    [anon_sym_f32] = ACTIONS(2283),
    [anon_sym_f64] = ACTIONS(2283),
    [anon_sym_bool] = ACTIONS(2283),
    [anon_sym_str] = ACTIONS(2283),
    [anon_sym_char] = ACTIONS(2283),
    [aux_sym__non_special_token_token1] = ACTIONS(2283),
    [anon_sym_SQUOTE] = ACTIONS(2283),
    [anon_sym_as] = ACTIONS(2283),
    [anon_sym_async] = ACTIONS(2283),
    [anon_sym_await] = ACTIONS(2283),
    [anon_sym_break] = ACTIONS(2283),
    [anon_sym_const] = ACTIONS(2283),
    [anon_sym_continue] = ACTIONS(2283),
    [anon_sym_default] = ACTIONS(2283),
    [anon_sym_enum] = ACTIONS(2283),
    [anon_sym_fn] = ACTIONS(2283),
    [anon_sym_for] = ACTIONS(2283),
    [anon_sym_if] = ACTIONS(2283),
    [anon_sym_impl] = ACTIONS(2283),
    [anon_sym_let] = ACTIONS(2283),
    [anon_sym_loop] = ACTIONS(2283),
    [anon_sym_match] = ACTIONS(2283),
    [anon_sym_mod] = ACTIONS(2283),
    [anon_sym_pub] = ACTIONS(2283),
    [anon_sym_return] = ACTIONS(2283),
    [anon_sym_static] = ACTIONS(2283),
    [anon_sym_struct] = ACTIONS(2283),
    [anon_sym_trait] = ACTIONS(2283),
    [anon_sym_type] = ACTIONS(2283),
    [anon_sym_union] = ACTIONS(2283),
    [anon_sym_unsafe] = ACTIONS(2283),
    [anon_sym_use] = ACTIONS(2283),
    [anon_sym_where] = ACTIONS(2283),
    [anon_sym_while] = ACTIONS(2283),
    [sym_mutable_specifier] = ACTIONS(2283),
    [sym_integer_literal] = ACTIONS(2014),
    [aux_sym_string_literal_token1] = ACTIONS(2016),
    [sym_char_literal] = ACTIONS(2014),
    [anon_sym_true] = ACTIONS(2018),
    [anon_sym_false] = ACTIONS(2018),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2283),
    [sym_super] = ACTIONS(2283),
    [sym_crate] = ACTIONS(2283),
    [sym_metavariable] = ACTIONS(2287),
    [sym_raw_string_literal] = ACTIONS(2014),
    [sym_float_literal] = ACTIONS(2014),
    [sym_block_comment] = ACTIONS(3),
  },
  [534] = {
    [sym_delim_token_tree] = STATE(502),
    [sym__delim_tokens] = STATE(502),
    [sym__non_delim_token] = STATE(502),
    [sym__literal] = STATE(502),
    [sym_string_literal] = STATE(592),
    [sym_boolean_literal] = STATE(592),
    [aux_sym_delim_token_tree_repeat1] = STATE(502),
    [sym_identifier] = ACTIONS(2289),
    [anon_sym_LPAREN] = ACTIONS(2163),
    [anon_sym_LBRACE] = ACTIONS(2165),
    [anon_sym_LBRACK] = ACTIONS(2169),
    [anon_sym_RBRACK] = ACTIONS(2257),
    [anon_sym_DOLLAR] = ACTIONS(2291),
    [anon_sym_u8] = ACTIONS(2289),
    [anon_sym_i8] = ACTIONS(2289),
    [anon_sym_u16] = ACTIONS(2289),
    [anon_sym_i16] = ACTIONS(2289),
    [anon_sym_u32] = ACTIONS(2289),
    [anon_sym_i32] = ACTIONS(2289),
    [anon_sym_u64] = ACTIONS(2289),
    [anon_sym_i64] = ACTIONS(2289),
    [anon_sym_u128] = ACTIONS(2289),
    [anon_sym_i128] = ACTIONS(2289),
    [anon_sym_isize] = ACTIONS(2289),
    [anon_sym_usize] = ACTIONS(2289),
    [anon_sym_f32] = ACTIONS(2289),
    [anon_sym_f64] = ACTIONS(2289),
    [anon_sym_bool] = ACTIONS(2289),
    [anon_sym_str] = ACTIONS(2289),
    [anon_sym_char] = ACTIONS(2289),
    [aux_sym__non_special_token_token1] = ACTIONS(2289),
    [anon_sym_SQUOTE] = ACTIONS(2289),
    [anon_sym_as] = ACTIONS(2289),
    [anon_sym_async] = ACTIONS(2289),
    [anon_sym_await] = ACTIONS(2289),
    [anon_sym_break] = ACTIONS(2289),
    [anon_sym_const] = ACTIONS(2289),
    [anon_sym_continue] = ACTIONS(2289),
    [anon_sym_default] = ACTIONS(2289),
    [anon_sym_enum] = ACTIONS(2289),
    [anon_sym_fn] = ACTIONS(2289),
    [anon_sym_for] = ACTIONS(2289),
    [anon_sym_if] = ACTIONS(2289),
    [anon_sym_impl] = ACTIONS(2289),
    [anon_sym_let] = ACTIONS(2289),
    [anon_sym_loop] = ACTIONS(2289),
    [anon_sym_match] = ACTIONS(2289),
    [anon_sym_mod] = ACTIONS(2289),
    [anon_sym_pub] = ACTIONS(2289),
    [anon_sym_return] = ACTIONS(2289),
    [anon_sym_static] = ACTIONS(2289),
    [anon_sym_struct] = ACTIONS(2289),
    [anon_sym_trait] = ACTIONS(2289),
    [anon_sym_type] = ACTIONS(2289),
    [anon_sym_union] = ACTIONS(2289),
    [anon_sym_unsafe] = ACTIONS(2289),
    [anon_sym_use] = ACTIONS(2289),
    [anon_sym_where] = ACTIONS(2289),
    [anon_sym_while] = ACTIONS(2289),
    [sym_mutable_specifier] = ACTIONS(2289),
    [sym_integer_literal] = ACTIONS(2173),
    [aux_sym_string_literal_token1] = ACTIONS(2175),
    [sym_char_literal] = ACTIONS(2173),
    [anon_sym_true] = ACTIONS(2177),
    [anon_sym_false] = ACTIONS(2177),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2289),
    [sym_super] = ACTIONS(2289),
    [sym_crate] = ACTIONS(2289),
    [sym_raw_string_literal] = ACTIONS(2173),
    [sym_float_literal] = ACTIONS(2173),
    [sym_block_comment] = ACTIONS(3),
  },
  [535] = {
    [sym_token_tree] = STATE(518),
    [sym_token_repetition] = STATE(518),
    [sym__literal] = STATE(518),
    [sym_string_literal] = STATE(585),
    [sym_boolean_literal] = STATE(585),
    [aux_sym_token_tree_repeat1] = STATE(518),
    [sym_identifier] = ACTIONS(2293),
    [anon_sym_LPAREN] = ACTIONS(2183),
    [anon_sym_RPAREN] = ACTIONS(2285),
    [anon_sym_LBRACE] = ACTIONS(2185),
    [anon_sym_LBRACK] = ACTIONS(2189),
    [anon_sym_DOLLAR] = ACTIONS(2191),
    [anon_sym_u8] = ACTIONS(2293),
    [anon_sym_i8] = ACTIONS(2293),
    [anon_sym_u16] = ACTIONS(2293),
    [anon_sym_i16] = ACTIONS(2293),
    [anon_sym_u32] = ACTIONS(2293),
    [anon_sym_i32] = ACTIONS(2293),
    [anon_sym_u64] = ACTIONS(2293),
    [anon_sym_i64] = ACTIONS(2293),
    [anon_sym_u128] = ACTIONS(2293),
    [anon_sym_i128] = ACTIONS(2293),
    [anon_sym_isize] = ACTIONS(2293),
    [anon_sym_usize] = ACTIONS(2293),
    [anon_sym_f32] = ACTIONS(2293),
    [anon_sym_f64] = ACTIONS(2293),
    [anon_sym_bool] = ACTIONS(2293),
    [anon_sym_str] = ACTIONS(2293),
    [anon_sym_char] = ACTIONS(2293),
    [aux_sym__non_special_token_token1] = ACTIONS(2293),
    [anon_sym_SQUOTE] = ACTIONS(2293),
    [anon_sym_as] = ACTIONS(2293),
    [anon_sym_async] = ACTIONS(2293),
    [anon_sym_await] = ACTIONS(2293),
    [anon_sym_break] = ACTIONS(2293),
    [anon_sym_const] = ACTIONS(2293),
    [anon_sym_continue] = ACTIONS(2293),
    [anon_sym_default] = ACTIONS(2293),
    [anon_sym_enum] = ACTIONS(2293),
    [anon_sym_fn] = ACTIONS(2293),
    [anon_sym_for] = ACTIONS(2293),
    [anon_sym_if] = ACTIONS(2293),
    [anon_sym_impl] = ACTIONS(2293),
    [anon_sym_let] = ACTIONS(2293),
    [anon_sym_loop] = ACTIONS(2293),
    [anon_sym_match] = ACTIONS(2293),
    [anon_sym_mod] = ACTIONS(2293),
    [anon_sym_pub] = ACTIONS(2293),
    [anon_sym_return] = ACTIONS(2293),
    [anon_sym_static] = ACTIONS(2293),
    [anon_sym_struct] = ACTIONS(2293),
    [anon_sym_trait] = ACTIONS(2293),
    [anon_sym_type] = ACTIONS(2293),
    [anon_sym_union] = ACTIONS(2293),
    [anon_sym_unsafe] = ACTIONS(2293),
    [anon_sym_use] = ACTIONS(2293),
    [anon_sym_where] = ACTIONS(2293),
    [anon_sym_while] = ACTIONS(2293),
    [sym_mutable_specifier] = ACTIONS(2293),
    [sym_integer_literal] = ACTIONS(2014),
    [aux_sym_string_literal_token1] = ACTIONS(2016),
    [sym_char_literal] = ACTIONS(2014),
    [anon_sym_true] = ACTIONS(2018),
    [anon_sym_false] = ACTIONS(2018),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2293),
    [sym_super] = ACTIONS(2293),
    [sym_crate] = ACTIONS(2293),
    [sym_metavariable] = ACTIONS(2295),
    [sym_raw_string_literal] = ACTIONS(2014),
    [sym_float_literal] = ACTIONS(2014),
    [sym_block_comment] = ACTIONS(3),
  },
  [536] = {
    [sym_delim_token_tree] = STATE(483),
    [sym__delim_tokens] = STATE(483),
    [sym__non_delim_token] = STATE(483),
    [sym__literal] = STATE(483),
    [sym_string_literal] = STATE(592),
    [sym_boolean_literal] = STATE(592),
    [aux_sym_delim_token_tree_repeat1] = STATE(483),
    [sym_identifier] = ACTIONS(2161),
    [anon_sym_LPAREN] = ACTIONS(2163),
    [anon_sym_LBRACE] = ACTIONS(2165),
    [anon_sym_LBRACK] = ACTIONS(2169),
    [anon_sym_RBRACK] = ACTIONS(2167),
    [anon_sym_DOLLAR] = ACTIONS(2171),
    [anon_sym_u8] = ACTIONS(2161),
    [anon_sym_i8] = ACTIONS(2161),
    [anon_sym_u16] = ACTIONS(2161),
    [anon_sym_i16] = ACTIONS(2161),
    [anon_sym_u32] = ACTIONS(2161),
    [anon_sym_i32] = ACTIONS(2161),
    [anon_sym_u64] = ACTIONS(2161),
    [anon_sym_i64] = ACTIONS(2161),
    [anon_sym_u128] = ACTIONS(2161),
    [anon_sym_i128] = ACTIONS(2161),
    [anon_sym_isize] = ACTIONS(2161),
    [anon_sym_usize] = ACTIONS(2161),
    [anon_sym_f32] = ACTIONS(2161),
    [anon_sym_f64] = ACTIONS(2161),
    [anon_sym_bool] = ACTIONS(2161),
    [anon_sym_str] = ACTIONS(2161),
    [anon_sym_char] = ACTIONS(2161),
    [aux_sym__non_special_token_token1] = ACTIONS(2161),
    [anon_sym_SQUOTE] = ACTIONS(2161),
    [anon_sym_as] = ACTIONS(2161),
    [anon_sym_async] = ACTIONS(2161),
    [anon_sym_await] = ACTIONS(2161),
    [anon_sym_break] = ACTIONS(2161),
    [anon_sym_const] = ACTIONS(2161),
    [anon_sym_continue] = ACTIONS(2161),
    [anon_sym_default] = ACTIONS(2161),
    [anon_sym_enum] = ACTIONS(2161),
    [anon_sym_fn] = ACTIONS(2161),
    [anon_sym_for] = ACTIONS(2161),
    [anon_sym_if] = ACTIONS(2161),
    [anon_sym_impl] = ACTIONS(2161),
    [anon_sym_let] = ACTIONS(2161),
    [anon_sym_loop] = ACTIONS(2161),
    [anon_sym_match] = ACTIONS(2161),
    [anon_sym_mod] = ACTIONS(2161),
    [anon_sym_pub] = ACTIONS(2161),
    [anon_sym_return] = ACTIONS(2161),
    [anon_sym_static] = ACTIONS(2161),
    [anon_sym_struct] = ACTIONS(2161),
    [anon_sym_trait] = ACTIONS(2161),
    [anon_sym_type] = ACTIONS(2161),
    [anon_sym_union] = ACTIONS(2161),
    [anon_sym_unsafe] = ACTIONS(2161),
    [anon_sym_use] = ACTIONS(2161),
    [anon_sym_where] = ACTIONS(2161),
    [anon_sym_while] = ACTIONS(2161),
    [sym_mutable_specifier] = ACTIONS(2161),
    [sym_integer_literal] = ACTIONS(2173),
    [aux_sym_string_literal_token1] = ACTIONS(2175),
    [sym_char_literal] = ACTIONS(2173),
    [anon_sym_true] = ACTIONS(2177),
    [anon_sym_false] = ACTIONS(2177),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2161),
    [sym_super] = ACTIONS(2161),
    [sym_crate] = ACTIONS(2161),
    [sym_raw_string_literal] = ACTIONS(2173),
    [sym_float_literal] = ACTIONS(2173),
    [sym_block_comment] = ACTIONS(3),
  },
  [537] = {
    [sym_token_tree] = STATE(485),
    [sym_token_repetition] = STATE(485),
    [sym__literal] = STATE(485),
    [sym_string_literal] = STATE(585),
    [sym_boolean_literal] = STATE(585),
    [aux_sym_token_tree_repeat1] = STATE(485),
    [sym_identifier] = ACTIONS(2181),
    [anon_sym_LPAREN] = ACTIONS(2183),
    [anon_sym_RPAREN] = ACTIONS(2297),
    [anon_sym_LBRACE] = ACTIONS(2185),
    [anon_sym_LBRACK] = ACTIONS(2189),
    [anon_sym_DOLLAR] = ACTIONS(2191),
    [anon_sym_u8] = ACTIONS(2181),
    [anon_sym_i8] = ACTIONS(2181),
    [anon_sym_u16] = ACTIONS(2181),
    [anon_sym_i16] = ACTIONS(2181),
    [anon_sym_u32] = ACTIONS(2181),
    [anon_sym_i32] = ACTIONS(2181),
    [anon_sym_u64] = ACTIONS(2181),
    [anon_sym_i64] = ACTIONS(2181),
    [anon_sym_u128] = ACTIONS(2181),
    [anon_sym_i128] = ACTIONS(2181),
    [anon_sym_isize] = ACTIONS(2181),
    [anon_sym_usize] = ACTIONS(2181),
    [anon_sym_f32] = ACTIONS(2181),
    [anon_sym_f64] = ACTIONS(2181),
    [anon_sym_bool] = ACTIONS(2181),
    [anon_sym_str] = ACTIONS(2181),
    [anon_sym_char] = ACTIONS(2181),
    [aux_sym__non_special_token_token1] = ACTIONS(2181),
    [anon_sym_SQUOTE] = ACTIONS(2181),
    [anon_sym_as] = ACTIONS(2181),
    [anon_sym_async] = ACTIONS(2181),
    [anon_sym_await] = ACTIONS(2181),
    [anon_sym_break] = ACTIONS(2181),
    [anon_sym_const] = ACTIONS(2181),
    [anon_sym_continue] = ACTIONS(2181),
    [anon_sym_default] = ACTIONS(2181),
    [anon_sym_enum] = ACTIONS(2181),
    [anon_sym_fn] = ACTIONS(2181),
    [anon_sym_for] = ACTIONS(2181),
    [anon_sym_if] = ACTIONS(2181),
    [anon_sym_impl] = ACTIONS(2181),
    [anon_sym_let] = ACTIONS(2181),
    [anon_sym_loop] = ACTIONS(2181),
    [anon_sym_match] = ACTIONS(2181),
    [anon_sym_mod] = ACTIONS(2181),
    [anon_sym_pub] = ACTIONS(2181),
    [anon_sym_return] = ACTIONS(2181),
    [anon_sym_static] = ACTIONS(2181),
    [anon_sym_struct] = ACTIONS(2181),
    [anon_sym_trait] = ACTIONS(2181),
    [anon_sym_type] = ACTIONS(2181),
    [anon_sym_union] = ACTIONS(2181),
    [anon_sym_unsafe] = ACTIONS(2181),
    [anon_sym_use] = ACTIONS(2181),
    [anon_sym_where] = ACTIONS(2181),
    [anon_sym_while] = ACTIONS(2181),
    [sym_mutable_specifier] = ACTIONS(2181),
    [sym_integer_literal] = ACTIONS(2014),
    [aux_sym_string_literal_token1] = ACTIONS(2016),
    [sym_char_literal] = ACTIONS(2014),
    [anon_sym_true] = ACTIONS(2018),
    [anon_sym_false] = ACTIONS(2018),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2181),
    [sym_super] = ACTIONS(2181),
    [sym_crate] = ACTIONS(2181),
    [sym_metavariable] = ACTIONS(2193),
    [sym_raw_string_literal] = ACTIONS(2014),
    [sym_float_literal] = ACTIONS(2014),
    [sym_block_comment] = ACTIONS(3),
  },
  [538] = {
    [sym_token_tree] = STATE(485),
    [sym_token_repetition] = STATE(485),
    [sym__literal] = STATE(485),
    [sym_string_literal] = STATE(585),
    [sym_boolean_literal] = STATE(585),
    [aux_sym_token_tree_repeat1] = STATE(485),
    [sym_identifier] = ACTIONS(2181),
    [anon_sym_LPAREN] = ACTIONS(2183),
    [anon_sym_LBRACE] = ACTIONS(2185),
    [anon_sym_LBRACK] = ACTIONS(2189),
    [anon_sym_RBRACK] = ACTIONS(2243),
    [anon_sym_DOLLAR] = ACTIONS(2191),
    [anon_sym_u8] = ACTIONS(2181),
    [anon_sym_i8] = ACTIONS(2181),
    [anon_sym_u16] = ACTIONS(2181),
    [anon_sym_i16] = ACTIONS(2181),
    [anon_sym_u32] = ACTIONS(2181),
    [anon_sym_i32] = ACTIONS(2181),
    [anon_sym_u64] = ACTIONS(2181),
    [anon_sym_i64] = ACTIONS(2181),
    [anon_sym_u128] = ACTIONS(2181),
    [anon_sym_i128] = ACTIONS(2181),
    [anon_sym_isize] = ACTIONS(2181),
    [anon_sym_usize] = ACTIONS(2181),
    [anon_sym_f32] = ACTIONS(2181),
    [anon_sym_f64] = ACTIONS(2181),
    [anon_sym_bool] = ACTIONS(2181),
    [anon_sym_str] = ACTIONS(2181),
    [anon_sym_char] = ACTIONS(2181),
    [aux_sym__non_special_token_token1] = ACTIONS(2181),
    [anon_sym_SQUOTE] = ACTIONS(2181),
    [anon_sym_as] = ACTIONS(2181),
    [anon_sym_async] = ACTIONS(2181),
    [anon_sym_await] = ACTIONS(2181),
    [anon_sym_break] = ACTIONS(2181),
    [anon_sym_const] = ACTIONS(2181),
    [anon_sym_continue] = ACTIONS(2181),
    [anon_sym_default] = ACTIONS(2181),
    [anon_sym_enum] = ACTIONS(2181),
    [anon_sym_fn] = ACTIONS(2181),
    [anon_sym_for] = ACTIONS(2181),
    [anon_sym_if] = ACTIONS(2181),
    [anon_sym_impl] = ACTIONS(2181),
    [anon_sym_let] = ACTIONS(2181),
    [anon_sym_loop] = ACTIONS(2181),
    [anon_sym_match] = ACTIONS(2181),
    [anon_sym_mod] = ACTIONS(2181),
    [anon_sym_pub] = ACTIONS(2181),
    [anon_sym_return] = ACTIONS(2181),
    [anon_sym_static] = ACTIONS(2181),
    [anon_sym_struct] = ACTIONS(2181),
    [anon_sym_trait] = ACTIONS(2181),
    [anon_sym_type] = ACTIONS(2181),
    [anon_sym_union] = ACTIONS(2181),
    [anon_sym_unsafe] = ACTIONS(2181),
    [anon_sym_use] = ACTIONS(2181),
    [anon_sym_where] = ACTIONS(2181),
    [anon_sym_while] = ACTIONS(2181),
    [sym_mutable_specifier] = ACTIONS(2181),
    [sym_integer_literal] = ACTIONS(2014),
    [aux_sym_string_literal_token1] = ACTIONS(2016),
    [sym_char_literal] = ACTIONS(2014),
    [anon_sym_true] = ACTIONS(2018),
    [anon_sym_false] = ACTIONS(2018),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2181),
    [sym_super] = ACTIONS(2181),
    [sym_crate] = ACTIONS(2181),
    [sym_metavariable] = ACTIONS(2193),
    [sym_raw_string_literal] = ACTIONS(2014),
    [sym_float_literal] = ACTIONS(2014),
    [sym_block_comment] = ACTIONS(3),
  },
  [539] = {
    [sym_token_tree] = STATE(537),
    [sym_token_repetition] = STATE(537),
    [sym__literal] = STATE(537),
    [sym_string_literal] = STATE(585),
    [sym_boolean_literal] = STATE(585),
    [aux_sym_token_tree_repeat1] = STATE(537),
    [sym_identifier] = ACTIONS(2299),
    [anon_sym_LPAREN] = ACTIONS(2183),
    [anon_sym_RPAREN] = ACTIONS(2301),
    [anon_sym_LBRACE] = ACTIONS(2185),
    [anon_sym_LBRACK] = ACTIONS(2189),
    [anon_sym_DOLLAR] = ACTIONS(2191),
    [anon_sym_u8] = ACTIONS(2299),
    [anon_sym_i8] = ACTIONS(2299),
    [anon_sym_u16] = ACTIONS(2299),
    [anon_sym_i16] = ACTIONS(2299),
    [anon_sym_u32] = ACTIONS(2299),
    [anon_sym_i32] = ACTIONS(2299),
    [anon_sym_u64] = ACTIONS(2299),
    [anon_sym_i64] = ACTIONS(2299),
    [anon_sym_u128] = ACTIONS(2299),
    [anon_sym_i128] = ACTIONS(2299),
    [anon_sym_isize] = ACTIONS(2299),
    [anon_sym_usize] = ACTIONS(2299),
    [anon_sym_f32] = ACTIONS(2299),
    [anon_sym_f64] = ACTIONS(2299),
    [anon_sym_bool] = ACTIONS(2299),
    [anon_sym_str] = ACTIONS(2299),
    [anon_sym_char] = ACTIONS(2299),
    [aux_sym__non_special_token_token1] = ACTIONS(2299),
    [anon_sym_SQUOTE] = ACTIONS(2299),
    [anon_sym_as] = ACTIONS(2299),
    [anon_sym_async] = ACTIONS(2299),
    [anon_sym_await] = ACTIONS(2299),
    [anon_sym_break] = ACTIONS(2299),
    [anon_sym_const] = ACTIONS(2299),
    [anon_sym_continue] = ACTIONS(2299),
    [anon_sym_default] = ACTIONS(2299),
    [anon_sym_enum] = ACTIONS(2299),
    [anon_sym_fn] = ACTIONS(2299),
    [anon_sym_for] = ACTIONS(2299),
    [anon_sym_if] = ACTIONS(2299),
    [anon_sym_impl] = ACTIONS(2299),
    [anon_sym_let] = ACTIONS(2299),
    [anon_sym_loop] = ACTIONS(2299),
    [anon_sym_match] = ACTIONS(2299),
    [anon_sym_mod] = ACTIONS(2299),
    [anon_sym_pub] = ACTIONS(2299),
    [anon_sym_return] = ACTIONS(2299),
    [anon_sym_static] = ACTIONS(2299),
    [anon_sym_struct] = ACTIONS(2299),
    [anon_sym_trait] = ACTIONS(2299),
    [anon_sym_type] = ACTIONS(2299),
    [anon_sym_union] = ACTIONS(2299),
    [anon_sym_unsafe] = ACTIONS(2299),
    [anon_sym_use] = ACTIONS(2299),
    [anon_sym_where] = ACTIONS(2299),
    [anon_sym_while] = ACTIONS(2299),
    [sym_mutable_specifier] = ACTIONS(2299),
    [sym_integer_literal] = ACTIONS(2014),
    [aux_sym_string_literal_token1] = ACTIONS(2016),
    [sym_char_literal] = ACTIONS(2014),
    [anon_sym_true] = ACTIONS(2018),
    [anon_sym_false] = ACTIONS(2018),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2299),
    [sym_super] = ACTIONS(2299),
    [sym_crate] = ACTIONS(2299),
    [sym_metavariable] = ACTIONS(2303),
    [sym_raw_string_literal] = ACTIONS(2014),
    [sym_float_literal] = ACTIONS(2014),
    [sym_block_comment] = ACTIONS(3),
  },
  [540] = {
    [sym_delim_token_tree] = STATE(483),
    [sym__delim_tokens] = STATE(483),
    [sym__non_delim_token] = STATE(483),
    [sym__literal] = STATE(483),
    [sym_string_literal] = STATE(592),
    [sym_boolean_literal] = STATE(592),
    [aux_sym_delim_token_tree_repeat1] = STATE(483),
    [sym_identifier] = ACTIONS(2161),
    [anon_sym_LPAREN] = ACTIONS(2163),
    [anon_sym_RPAREN] = ACTIONS(2179),
    [anon_sym_LBRACE] = ACTIONS(2165),
    [anon_sym_LBRACK] = ACTIONS(2169),
    [anon_sym_DOLLAR] = ACTIONS(2171),
    [anon_sym_u8] = ACTIONS(2161),
    [anon_sym_i8] = ACTIONS(2161),
    [anon_sym_u16] = ACTIONS(2161),
    [anon_sym_i16] = ACTIONS(2161),
    [anon_sym_u32] = ACTIONS(2161),
    [anon_sym_i32] = ACTIONS(2161),
    [anon_sym_u64] = ACTIONS(2161),
    [anon_sym_i64] = ACTIONS(2161),
    [anon_sym_u128] = ACTIONS(2161),
    [anon_sym_i128] = ACTIONS(2161),
    [anon_sym_isize] = ACTIONS(2161),
    [anon_sym_usize] = ACTIONS(2161),
    [anon_sym_f32] = ACTIONS(2161),
    [anon_sym_f64] = ACTIONS(2161),
    [anon_sym_bool] = ACTIONS(2161),
    [anon_sym_str] = ACTIONS(2161),
    [anon_sym_char] = ACTIONS(2161),
    [aux_sym__non_special_token_token1] = ACTIONS(2161),
    [anon_sym_SQUOTE] = ACTIONS(2161),
    [anon_sym_as] = ACTIONS(2161),
    [anon_sym_async] = ACTIONS(2161),
    [anon_sym_await] = ACTIONS(2161),
    [anon_sym_break] = ACTIONS(2161),
    [anon_sym_const] = ACTIONS(2161),
    [anon_sym_continue] = ACTIONS(2161),
    [anon_sym_default] = ACTIONS(2161),
    [anon_sym_enum] = ACTIONS(2161),
    [anon_sym_fn] = ACTIONS(2161),
    [anon_sym_for] = ACTIONS(2161),
    [anon_sym_if] = ACTIONS(2161),
    [anon_sym_impl] = ACTIONS(2161),
    [anon_sym_let] = ACTIONS(2161),
    [anon_sym_loop] = ACTIONS(2161),
    [anon_sym_match] = ACTIONS(2161),
    [anon_sym_mod] = ACTIONS(2161),
    [anon_sym_pub] = ACTIONS(2161),
    [anon_sym_return] = ACTIONS(2161),
    [anon_sym_static] = ACTIONS(2161),
    [anon_sym_struct] = ACTIONS(2161),
    [anon_sym_trait] = ACTIONS(2161),
    [anon_sym_type] = ACTIONS(2161),
    [anon_sym_union] = ACTIONS(2161),
    [anon_sym_unsafe] = ACTIONS(2161),
    [anon_sym_use] = ACTIONS(2161),
    [anon_sym_where] = ACTIONS(2161),
    [anon_sym_while] = ACTIONS(2161),
    [sym_mutable_specifier] = ACTIONS(2161),
    [sym_integer_literal] = ACTIONS(2173),
    [aux_sym_string_literal_token1] = ACTIONS(2175),
    [sym_char_literal] = ACTIONS(2173),
    [anon_sym_true] = ACTIONS(2177),
    [anon_sym_false] = ACTIONS(2177),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2161),
    [sym_super] = ACTIONS(2161),
    [sym_crate] = ACTIONS(2161),
    [sym_raw_string_literal] = ACTIONS(2173),
    [sym_float_literal] = ACTIONS(2173),
    [sym_block_comment] = ACTIONS(3),
  },
  [541] = {
    [sym_token_tree] = STATE(485),
    [sym_token_repetition] = STATE(485),
    [sym__literal] = STATE(485),
    [sym_string_literal] = STATE(585),
    [sym_boolean_literal] = STATE(585),
    [aux_sym_token_tree_repeat1] = STATE(485),
    [sym_identifier] = ACTIONS(2181),
    [anon_sym_LPAREN] = ACTIONS(2183),
    [anon_sym_LBRACE] = ACTIONS(2185),
    [anon_sym_RBRACE] = ACTIONS(2243),
    [anon_sym_LBRACK] = ACTIONS(2189),
    [anon_sym_DOLLAR] = ACTIONS(2191),
    [anon_sym_u8] = ACTIONS(2181),
    [anon_sym_i8] = ACTIONS(2181),
    [anon_sym_u16] = ACTIONS(2181),
    [anon_sym_i16] = ACTIONS(2181),
    [anon_sym_u32] = ACTIONS(2181),
    [anon_sym_i32] = ACTIONS(2181),
    [anon_sym_u64] = ACTIONS(2181),
    [anon_sym_i64] = ACTIONS(2181),
    [anon_sym_u128] = ACTIONS(2181),
    [anon_sym_i128] = ACTIONS(2181),
    [anon_sym_isize] = ACTIONS(2181),
    [anon_sym_usize] = ACTIONS(2181),
    [anon_sym_f32] = ACTIONS(2181),
    [anon_sym_f64] = ACTIONS(2181),
    [anon_sym_bool] = ACTIONS(2181),
    [anon_sym_str] = ACTIONS(2181),
    [anon_sym_char] = ACTIONS(2181),
    [aux_sym__non_special_token_token1] = ACTIONS(2181),
    [anon_sym_SQUOTE] = ACTIONS(2181),
    [anon_sym_as] = ACTIONS(2181),
    [anon_sym_async] = ACTIONS(2181),
    [anon_sym_await] = ACTIONS(2181),
    [anon_sym_break] = ACTIONS(2181),
    [anon_sym_const] = ACTIONS(2181),
    [anon_sym_continue] = ACTIONS(2181),
    [anon_sym_default] = ACTIONS(2181),
    [anon_sym_enum] = ACTIONS(2181),
    [anon_sym_fn] = ACTIONS(2181),
    [anon_sym_for] = ACTIONS(2181),
    [anon_sym_if] = ACTIONS(2181),
    [anon_sym_impl] = ACTIONS(2181),
    [anon_sym_let] = ACTIONS(2181),
    [anon_sym_loop] = ACTIONS(2181),
    [anon_sym_match] = ACTIONS(2181),
    [anon_sym_mod] = ACTIONS(2181),
    [anon_sym_pub] = ACTIONS(2181),
    [anon_sym_return] = ACTIONS(2181),
    [anon_sym_static] = ACTIONS(2181),
    [anon_sym_struct] = ACTIONS(2181),
    [anon_sym_trait] = ACTIONS(2181),
    [anon_sym_type] = ACTIONS(2181),
    [anon_sym_union] = ACTIONS(2181),
    [anon_sym_unsafe] = ACTIONS(2181),
    [anon_sym_use] = ACTIONS(2181),
    [anon_sym_where] = ACTIONS(2181),
    [anon_sym_while] = ACTIONS(2181),
    [sym_mutable_specifier] = ACTIONS(2181),
    [sym_integer_literal] = ACTIONS(2014),
    [aux_sym_string_literal_token1] = ACTIONS(2016),
    [sym_char_literal] = ACTIONS(2014),
    [anon_sym_true] = ACTIONS(2018),
    [anon_sym_false] = ACTIONS(2018),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2181),
    [sym_super] = ACTIONS(2181),
    [sym_crate] = ACTIONS(2181),
    [sym_metavariable] = ACTIONS(2193),
    [sym_raw_string_literal] = ACTIONS(2014),
    [sym_float_literal] = ACTIONS(2014),
    [sym_block_comment] = ACTIONS(3),
  },
  [542] = {
    [sym_delim_token_tree] = STATE(483),
    [sym__delim_tokens] = STATE(483),
    [sym__non_delim_token] = STATE(483),
    [sym__literal] = STATE(483),
    [sym_string_literal] = STATE(592),
    [sym_boolean_literal] = STATE(592),
    [aux_sym_delim_token_tree_repeat1] = STATE(483),
    [sym_identifier] = ACTIONS(2161),
    [anon_sym_LPAREN] = ACTIONS(2163),
    [anon_sym_LBRACE] = ACTIONS(2165),
    [anon_sym_RBRACE] = ACTIONS(2195),
    [anon_sym_LBRACK] = ACTIONS(2169),
    [anon_sym_DOLLAR] = ACTIONS(2171),
    [anon_sym_u8] = ACTIONS(2161),
    [anon_sym_i8] = ACTIONS(2161),
    [anon_sym_u16] = ACTIONS(2161),
    [anon_sym_i16] = ACTIONS(2161),
    [anon_sym_u32] = ACTIONS(2161),
    [anon_sym_i32] = ACTIONS(2161),
    [anon_sym_u64] = ACTIONS(2161),
    [anon_sym_i64] = ACTIONS(2161),
    [anon_sym_u128] = ACTIONS(2161),
    [anon_sym_i128] = ACTIONS(2161),
    [anon_sym_isize] = ACTIONS(2161),
    [anon_sym_usize] = ACTIONS(2161),
    [anon_sym_f32] = ACTIONS(2161),
    [anon_sym_f64] = ACTIONS(2161),
    [anon_sym_bool] = ACTIONS(2161),
    [anon_sym_str] = ACTIONS(2161),
    [anon_sym_char] = ACTIONS(2161),
    [aux_sym__non_special_token_token1] = ACTIONS(2161),
    [anon_sym_SQUOTE] = ACTIONS(2161),
    [anon_sym_as] = ACTIONS(2161),
    [anon_sym_async] = ACTIONS(2161),
    [anon_sym_await] = ACTIONS(2161),
    [anon_sym_break] = ACTIONS(2161),
    [anon_sym_const] = ACTIONS(2161),
    [anon_sym_continue] = ACTIONS(2161),
    [anon_sym_default] = ACTIONS(2161),
    [anon_sym_enum] = ACTIONS(2161),
    [anon_sym_fn] = ACTIONS(2161),
    [anon_sym_for] = ACTIONS(2161),
    [anon_sym_if] = ACTIONS(2161),
    [anon_sym_impl] = ACTIONS(2161),
    [anon_sym_let] = ACTIONS(2161),
    [anon_sym_loop] = ACTIONS(2161),
    [anon_sym_match] = ACTIONS(2161),
    [anon_sym_mod] = ACTIONS(2161),
    [anon_sym_pub] = ACTIONS(2161),
    [anon_sym_return] = ACTIONS(2161),
    [anon_sym_static] = ACTIONS(2161),
    [anon_sym_struct] = ACTIONS(2161),
    [anon_sym_trait] = ACTIONS(2161),
    [anon_sym_type] = ACTIONS(2161),
    [anon_sym_union] = ACTIONS(2161),
    [anon_sym_unsafe] = ACTIONS(2161),
    [anon_sym_use] = ACTIONS(2161),
    [anon_sym_where] = ACTIONS(2161),
    [anon_sym_while] = ACTIONS(2161),
    [sym_mutable_specifier] = ACTIONS(2161),
    [sym_integer_literal] = ACTIONS(2173),
    [aux_sym_string_literal_token1] = ACTIONS(2175),
    [sym_char_literal] = ACTIONS(2173),
    [anon_sym_true] = ACTIONS(2177),
    [anon_sym_false] = ACTIONS(2177),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2161),
    [sym_super] = ACTIONS(2161),
    [sym_crate] = ACTIONS(2161),
    [sym_raw_string_literal] = ACTIONS(2173),
    [sym_float_literal] = ACTIONS(2173),
    [sym_block_comment] = ACTIONS(3),
  },
  [543] = {
    [sym_delim_token_tree] = STATE(483),
    [sym__delim_tokens] = STATE(483),
    [sym__non_delim_token] = STATE(483),
    [sym__literal] = STATE(483),
    [sym_string_literal] = STATE(592),
    [sym_boolean_literal] = STATE(592),
    [aux_sym_delim_token_tree_repeat1] = STATE(483),
    [sym_identifier] = ACTIONS(2161),
    [anon_sym_LPAREN] = ACTIONS(2163),
    [anon_sym_LBRACE] = ACTIONS(2165),
    [anon_sym_RBRACE] = ACTIONS(2179),
    [anon_sym_LBRACK] = ACTIONS(2169),
    [anon_sym_DOLLAR] = ACTIONS(2171),
    [anon_sym_u8] = ACTIONS(2161),
    [anon_sym_i8] = ACTIONS(2161),
    [anon_sym_u16] = ACTIONS(2161),
    [anon_sym_i16] = ACTIONS(2161),
    [anon_sym_u32] = ACTIONS(2161),
    [anon_sym_i32] = ACTIONS(2161),
    [anon_sym_u64] = ACTIONS(2161),
    [anon_sym_i64] = ACTIONS(2161),
    [anon_sym_u128] = ACTIONS(2161),
    [anon_sym_i128] = ACTIONS(2161),
    [anon_sym_isize] = ACTIONS(2161),
    [anon_sym_usize] = ACTIONS(2161),
    [anon_sym_f32] = ACTIONS(2161),
    [anon_sym_f64] = ACTIONS(2161),
    [anon_sym_bool] = ACTIONS(2161),
    [anon_sym_str] = ACTIONS(2161),
    [anon_sym_char] = ACTIONS(2161),
    [aux_sym__non_special_token_token1] = ACTIONS(2161),
    [anon_sym_SQUOTE] = ACTIONS(2161),
    [anon_sym_as] = ACTIONS(2161),
    [anon_sym_async] = ACTIONS(2161),
    [anon_sym_await] = ACTIONS(2161),
    [anon_sym_break] = ACTIONS(2161),
    [anon_sym_const] = ACTIONS(2161),
    [anon_sym_continue] = ACTIONS(2161),
    [anon_sym_default] = ACTIONS(2161),
    [anon_sym_enum] = ACTIONS(2161),
    [anon_sym_fn] = ACTIONS(2161),
    [anon_sym_for] = ACTIONS(2161),
    [anon_sym_if] = ACTIONS(2161),
    [anon_sym_impl] = ACTIONS(2161),
    [anon_sym_let] = ACTIONS(2161),
    [anon_sym_loop] = ACTIONS(2161),
    [anon_sym_match] = ACTIONS(2161),
    [anon_sym_mod] = ACTIONS(2161),
    [anon_sym_pub] = ACTIONS(2161),
    [anon_sym_return] = ACTIONS(2161),
    [anon_sym_static] = ACTIONS(2161),
    [anon_sym_struct] = ACTIONS(2161),
    [anon_sym_trait] = ACTIONS(2161),
    [anon_sym_type] = ACTIONS(2161),
    [anon_sym_union] = ACTIONS(2161),
    [anon_sym_unsafe] = ACTIONS(2161),
    [anon_sym_use] = ACTIONS(2161),
    [anon_sym_where] = ACTIONS(2161),
    [anon_sym_while] = ACTIONS(2161),
    [sym_mutable_specifier] = ACTIONS(2161),
    [sym_integer_literal] = ACTIONS(2173),
    [aux_sym_string_literal_token1] = ACTIONS(2175),
    [sym_char_literal] = ACTIONS(2173),
    [anon_sym_true] = ACTIONS(2177),
    [anon_sym_false] = ACTIONS(2177),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2161),
    [sym_super] = ACTIONS(2161),
    [sym_crate] = ACTIONS(2161),
    [sym_raw_string_literal] = ACTIONS(2173),
    [sym_float_literal] = ACTIONS(2173),
    [sym_block_comment] = ACTIONS(3),
  },
  [544] = {
    [sym_token_tree] = STATE(541),
    [sym_token_repetition] = STATE(541),
    [sym__literal] = STATE(541),
    [sym_string_literal] = STATE(585),
    [sym_boolean_literal] = STATE(585),
    [aux_sym_token_tree_repeat1] = STATE(541),
    [sym_identifier] = ACTIONS(2305),
    [anon_sym_LPAREN] = ACTIONS(2183),
    [anon_sym_LBRACE] = ACTIONS(2185),
    [anon_sym_RBRACE] = ACTIONS(2285),
    [anon_sym_LBRACK] = ACTIONS(2189),
    [anon_sym_DOLLAR] = ACTIONS(2191),
    [anon_sym_u8] = ACTIONS(2305),
    [anon_sym_i8] = ACTIONS(2305),
    [anon_sym_u16] = ACTIONS(2305),
    [anon_sym_i16] = ACTIONS(2305),
    [anon_sym_u32] = ACTIONS(2305),
    [anon_sym_i32] = ACTIONS(2305),
    [anon_sym_u64] = ACTIONS(2305),
    [anon_sym_i64] = ACTIONS(2305),
    [anon_sym_u128] = ACTIONS(2305),
    [anon_sym_i128] = ACTIONS(2305),
    [anon_sym_isize] = ACTIONS(2305),
    [anon_sym_usize] = ACTIONS(2305),
    [anon_sym_f32] = ACTIONS(2305),
    [anon_sym_f64] = ACTIONS(2305),
    [anon_sym_bool] = ACTIONS(2305),
    [anon_sym_str] = ACTIONS(2305),
    [anon_sym_char] = ACTIONS(2305),
    [aux_sym__non_special_token_token1] = ACTIONS(2305),
    [anon_sym_SQUOTE] = ACTIONS(2305),
    [anon_sym_as] = ACTIONS(2305),
    [anon_sym_async] = ACTIONS(2305),
    [anon_sym_await] = ACTIONS(2305),
    [anon_sym_break] = ACTIONS(2305),
    [anon_sym_const] = ACTIONS(2305),
    [anon_sym_continue] = ACTIONS(2305),
    [anon_sym_default] = ACTIONS(2305),
    [anon_sym_enum] = ACTIONS(2305),
    [anon_sym_fn] = ACTIONS(2305),
    [anon_sym_for] = ACTIONS(2305),
    [anon_sym_if] = ACTIONS(2305),
    [anon_sym_impl] = ACTIONS(2305),
    [anon_sym_let] = ACTIONS(2305),
    [anon_sym_loop] = ACTIONS(2305),
    [anon_sym_match] = ACTIONS(2305),
    [anon_sym_mod] = ACTIONS(2305),
    [anon_sym_pub] = ACTIONS(2305),
    [anon_sym_return] = ACTIONS(2305),
    [anon_sym_static] = ACTIONS(2305),
    [anon_sym_struct] = ACTIONS(2305),
    [anon_sym_trait] = ACTIONS(2305),
    [anon_sym_type] = ACTIONS(2305),
    [anon_sym_union] = ACTIONS(2305),
    [anon_sym_unsafe] = ACTIONS(2305),
    [anon_sym_use] = ACTIONS(2305),
    [anon_sym_where] = ACTIONS(2305),
    [anon_sym_while] = ACTIONS(2305),
    [sym_mutable_specifier] = ACTIONS(2305),
    [sym_integer_literal] = ACTIONS(2014),
    [aux_sym_string_literal_token1] = ACTIONS(2016),
    [sym_char_literal] = ACTIONS(2014),
    [anon_sym_true] = ACTIONS(2018),
    [anon_sym_false] = ACTIONS(2018),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2305),
    [sym_super] = ACTIONS(2305),
    [sym_crate] = ACTIONS(2305),
    [sym_metavariable] = ACTIONS(2307),
    [sym_raw_string_literal] = ACTIONS(2014),
    [sym_float_literal] = ACTIONS(2014),
    [sym_block_comment] = ACTIONS(3),
  },
  [545] = {
    [sym_attribute_item] = STATE(554),
    [sym_function_modifiers] = STATE(2472),
    [sym_extern_modifier] = STATE(1526),
    [sym_visibility_modifier] = STATE(686),
    [sym__type] = STATE(1821),
    [sym_bracketed_type] = STATE(2378),
    [sym_lifetime] = STATE(2471),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2379),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1390),
    [sym_scoped_identifier] = STATE(2157),
    [sym_scoped_type_identifier] = STATE(1331),
    [aux_sym_enum_variant_list_repeat1] = STATE(554),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(2309),
    [anon_sym_LPAREN] = ACTIONS(856),
    [anon_sym_RPAREN] = ACTIONS(2311),
    [anon_sym_LBRACK] = ACTIONS(860),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(862),
    [anon_sym_i8] = ACTIONS(862),
    [anon_sym_u16] = ACTIONS(862),
    [anon_sym_i16] = ACTIONS(862),
    [anon_sym_u32] = ACTIONS(862),
    [anon_sym_i32] = ACTIONS(862),
    [anon_sym_u64] = ACTIONS(862),
    [anon_sym_i64] = ACTIONS(862),
    [anon_sym_u128] = ACTIONS(862),
    [anon_sym_i128] = ACTIONS(862),
    [anon_sym_isize] = ACTIONS(862),
    [anon_sym_usize] = ACTIONS(862),
    [anon_sym_f32] = ACTIONS(862),
    [anon_sym_f64] = ACTIONS(862),
    [anon_sym_bool] = ACTIONS(862),
    [anon_sym_str] = ACTIONS(862),
    [anon_sym_char] = ACTIONS(862),
    [anon_sym_SQUOTE] = ACTIONS(2313),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(664),
    [anon_sym_default] = ACTIONS(864),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_pub] = ACTIONS(2315),
    [anon_sym_union] = ACTIONS(866),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_POUND] = ACTIONS(2317),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_COMMA] = ACTIONS(2319),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(870),
    [anon_sym_AMP] = ACTIONS(872),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(876),
    [sym_super] = ACTIONS(876),
    [sym_crate] = ACTIONS(2321),
    [sym_metavariable] = ACTIONS(878),
    [sym_block_comment] = ACTIONS(3),
  },
  [546] = {
    [sym_attribute_item] = STATE(621),
    [sym_bracketed_type] = STATE(2489),
    [sym_generic_type] = STATE(2488),
    [sym_generic_type_with_turbofish] = STATE(2487),
    [sym_macro_invocation] = STATE(1422),
    [sym_scoped_identifier] = STATE(1337),
    [sym_scoped_type_identifier] = STATE(2013),
    [sym_match_pattern] = STATE(2560),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1981),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [aux_sym_enum_variant_list_repeat1] = STATE(621),
    [sym_identifier] = ACTIONS(1524),
    [anon_sym_LPAREN] = ACTIONS(1526),
    [anon_sym_LBRACK] = ACTIONS(1530),
    [anon_sym_u8] = ACTIONS(1532),
    [anon_sym_i8] = ACTIONS(1532),
    [anon_sym_u16] = ACTIONS(1532),
    [anon_sym_i16] = ACTIONS(1532),
    [anon_sym_u32] = ACTIONS(1532),
    [anon_sym_i32] = ACTIONS(1532),
    [anon_sym_u64] = ACTIONS(1532),
    [anon_sym_i64] = ACTIONS(1532),
    [anon_sym_u128] = ACTIONS(1532),
    [anon_sym_i128] = ACTIONS(1532),
    [anon_sym_isize] = ACTIONS(1532),
    [anon_sym_usize] = ACTIONS(1532),
    [anon_sym_f32] = ACTIONS(1532),
    [anon_sym_f64] = ACTIONS(1532),
    [anon_sym_bool] = ACTIONS(1532),
    [anon_sym_str] = ACTIONS(1532),
    [anon_sym_char] = ACTIONS(1532),
    [anon_sym_const] = ACTIONS(1534),
    [anon_sym_default] = ACTIONS(1536),
    [anon_sym_union] = ACTIONS(1536),
    [anon_sym_POUND] = ACTIONS(370),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(1538),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(1540),
    [sym_mutable_specifier] = ACTIONS(796),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1542),
    [sym_super] = ACTIONS(1542),
    [sym_crate] = ACTIONS(1542),
    [sym_metavariable] = ACTIONS(1544),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [547] = {
    [sym_attribute_item] = STATE(621),
    [sym_bracketed_type] = STATE(2489),
    [sym_generic_type] = STATE(2488),
    [sym_generic_type_with_turbofish] = STATE(2487),
    [sym_macro_invocation] = STATE(1422),
    [sym_scoped_identifier] = STATE(1337),
    [sym_scoped_type_identifier] = STATE(2013),
    [sym_match_pattern] = STATE(2481),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1981),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [aux_sym_enum_variant_list_repeat1] = STATE(621),
    [sym_identifier] = ACTIONS(1524),
    [anon_sym_LPAREN] = ACTIONS(1526),
    [anon_sym_LBRACK] = ACTIONS(1530),
    [anon_sym_u8] = ACTIONS(1532),
    [anon_sym_i8] = ACTIONS(1532),
    [anon_sym_u16] = ACTIONS(1532),
    [anon_sym_i16] = ACTIONS(1532),
    [anon_sym_u32] = ACTIONS(1532),
    [anon_sym_i32] = ACTIONS(1532),
    [anon_sym_u64] = ACTIONS(1532),
    [anon_sym_i64] = ACTIONS(1532),
    [anon_sym_u128] = ACTIONS(1532),
    [anon_sym_i128] = ACTIONS(1532),
    [anon_sym_isize] = ACTIONS(1532),
    [anon_sym_usize] = ACTIONS(1532),
    [anon_sym_f32] = ACTIONS(1532),
    [anon_sym_f64] = ACTIONS(1532),
    [anon_sym_bool] = ACTIONS(1532),
    [anon_sym_str] = ACTIONS(1532),
    [anon_sym_char] = ACTIONS(1532),
    [anon_sym_const] = ACTIONS(1534),
    [anon_sym_default] = ACTIONS(1536),
    [anon_sym_union] = ACTIONS(1536),
    [anon_sym_POUND] = ACTIONS(370),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(1538),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(1540),
    [sym_mutable_specifier] = ACTIONS(796),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1542),
    [sym_super] = ACTIONS(1542),
    [sym_crate] = ACTIONS(1542),
    [sym_metavariable] = ACTIONS(1544),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [548] = {
    [sym_attribute_item] = STATE(557),
    [sym_function_modifiers] = STATE(2472),
    [sym_extern_modifier] = STATE(1526),
    [sym_visibility_modifier] = STATE(645),
    [sym__type] = STATE(1917),
    [sym_bracketed_type] = STATE(2378),
    [sym_lifetime] = STATE(2471),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2379),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1390),
    [sym_scoped_identifier] = STATE(2157),
    [sym_scoped_type_identifier] = STATE(1331),
    [aux_sym_enum_variant_list_repeat1] = STATE(557),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(2309),
    [anon_sym_LPAREN] = ACTIONS(856),
    [anon_sym_RPAREN] = ACTIONS(2323),
    [anon_sym_LBRACK] = ACTIONS(860),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(862),
    [anon_sym_i8] = ACTIONS(862),
    [anon_sym_u16] = ACTIONS(862),
    [anon_sym_i16] = ACTIONS(862),
    [anon_sym_u32] = ACTIONS(862),
    [anon_sym_i32] = ACTIONS(862),
    [anon_sym_u64] = ACTIONS(862),
    [anon_sym_i64] = ACTIONS(862),
    [anon_sym_u128] = ACTIONS(862),
    [anon_sym_i128] = ACTIONS(862),
    [anon_sym_isize] = ACTIONS(862),
    [anon_sym_usize] = ACTIONS(862),
    [anon_sym_f32] = ACTIONS(862),
    [anon_sym_f64] = ACTIONS(862),
    [anon_sym_bool] = ACTIONS(862),
    [anon_sym_str] = ACTIONS(862),
    [anon_sym_char] = ACTIONS(862),
    [anon_sym_SQUOTE] = ACTIONS(2313),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(664),
    [anon_sym_default] = ACTIONS(864),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_pub] = ACTIONS(2315),
    [anon_sym_union] = ACTIONS(866),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_POUND] = ACTIONS(2317),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(870),
    [anon_sym_AMP] = ACTIONS(872),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(876),
    [sym_super] = ACTIONS(876),
    [sym_crate] = ACTIONS(2321),
    [sym_metavariable] = ACTIONS(878),
    [sym_block_comment] = ACTIONS(3),
  },
  [549] = {
    [sym_attribute_item] = STATE(557),
    [sym_function_modifiers] = STATE(2472),
    [sym_extern_modifier] = STATE(1526),
    [sym_visibility_modifier] = STATE(645),
    [sym__type] = STATE(1917),
    [sym_bracketed_type] = STATE(2378),
    [sym_lifetime] = STATE(2471),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2379),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1390),
    [sym_scoped_identifier] = STATE(2157),
    [sym_scoped_type_identifier] = STATE(1331),
    [aux_sym_enum_variant_list_repeat1] = STATE(557),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(2309),
    [anon_sym_LPAREN] = ACTIONS(856),
    [anon_sym_RPAREN] = ACTIONS(2325),
    [anon_sym_LBRACK] = ACTIONS(860),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(862),
    [anon_sym_i8] = ACTIONS(862),
    [anon_sym_u16] = ACTIONS(862),
    [anon_sym_i16] = ACTIONS(862),
    [anon_sym_u32] = ACTIONS(862),
    [anon_sym_i32] = ACTIONS(862),
    [anon_sym_u64] = ACTIONS(862),
    [anon_sym_i64] = ACTIONS(862),
    [anon_sym_u128] = ACTIONS(862),
    [anon_sym_i128] = ACTIONS(862),
    [anon_sym_isize] = ACTIONS(862),
    [anon_sym_usize] = ACTIONS(862),
    [anon_sym_f32] = ACTIONS(862),
    [anon_sym_f64] = ACTIONS(862),
    [anon_sym_bool] = ACTIONS(862),
    [anon_sym_str] = ACTIONS(862),
    [anon_sym_char] = ACTIONS(862),
    [anon_sym_SQUOTE] = ACTIONS(2313),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(664),
    [anon_sym_default] = ACTIONS(864),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_pub] = ACTIONS(2315),
    [anon_sym_union] = ACTIONS(866),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_POUND] = ACTIONS(2317),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(870),
    [anon_sym_AMP] = ACTIONS(872),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(876),
    [sym_super] = ACTIONS(876),
    [sym_crate] = ACTIONS(2321),
    [sym_metavariable] = ACTIONS(878),
    [sym_block_comment] = ACTIONS(3),
  },
  [550] = {
    [sym_attribute_item] = STATE(557),
    [sym_function_modifiers] = STATE(2472),
    [sym_extern_modifier] = STATE(1526),
    [sym_visibility_modifier] = STATE(645),
    [sym__type] = STATE(1917),
    [sym_bracketed_type] = STATE(2378),
    [sym_lifetime] = STATE(2471),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2379),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1390),
    [sym_scoped_identifier] = STATE(2157),
    [sym_scoped_type_identifier] = STATE(1331),
    [aux_sym_enum_variant_list_repeat1] = STATE(557),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(2309),
    [anon_sym_LPAREN] = ACTIONS(856),
    [anon_sym_RPAREN] = ACTIONS(2327),
    [anon_sym_LBRACK] = ACTIONS(860),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(862),
    [anon_sym_i8] = ACTIONS(862),
    [anon_sym_u16] = ACTIONS(862),
    [anon_sym_i16] = ACTIONS(862),
    [anon_sym_u32] = ACTIONS(862),
    [anon_sym_i32] = ACTIONS(862),
    [anon_sym_u64] = ACTIONS(862),
    [anon_sym_i64] = ACTIONS(862),
    [anon_sym_u128] = ACTIONS(862),
    [anon_sym_i128] = ACTIONS(862),
    [anon_sym_isize] = ACTIONS(862),
    [anon_sym_usize] = ACTIONS(862),
    [anon_sym_f32] = ACTIONS(862),
    [anon_sym_f64] = ACTIONS(862),
    [anon_sym_bool] = ACTIONS(862),
    [anon_sym_str] = ACTIONS(862),
    [anon_sym_char] = ACTIONS(862),
    [anon_sym_SQUOTE] = ACTIONS(2313),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(664),
    [anon_sym_default] = ACTIONS(864),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_pub] = ACTIONS(2315),
    [anon_sym_union] = ACTIONS(866),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_POUND] = ACTIONS(2317),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(870),
    [anon_sym_AMP] = ACTIONS(872),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(876),
    [sym_super] = ACTIONS(876),
    [sym_crate] = ACTIONS(2321),
    [sym_metavariable] = ACTIONS(878),
    [sym_block_comment] = ACTIONS(3),
  },
  [551] = {
    [sym_attribute_item] = STATE(557),
    [sym_function_modifiers] = STATE(2472),
    [sym_extern_modifier] = STATE(1526),
    [sym_visibility_modifier] = STATE(645),
    [sym__type] = STATE(1917),
    [sym_bracketed_type] = STATE(2378),
    [sym_lifetime] = STATE(2471),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2379),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1390),
    [sym_scoped_identifier] = STATE(2157),
    [sym_scoped_type_identifier] = STATE(1331),
    [aux_sym_enum_variant_list_repeat1] = STATE(557),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(2309),
    [anon_sym_LPAREN] = ACTIONS(856),
    [anon_sym_RPAREN] = ACTIONS(2329),
    [anon_sym_LBRACK] = ACTIONS(860),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(862),
    [anon_sym_i8] = ACTIONS(862),
    [anon_sym_u16] = ACTIONS(862),
    [anon_sym_i16] = ACTIONS(862),
    [anon_sym_u32] = ACTIONS(862),
    [anon_sym_i32] = ACTIONS(862),
    [anon_sym_u64] = ACTIONS(862),
    [anon_sym_i64] = ACTIONS(862),
    [anon_sym_u128] = ACTIONS(862),
    [anon_sym_i128] = ACTIONS(862),
    [anon_sym_isize] = ACTIONS(862),
    [anon_sym_usize] = ACTIONS(862),
    [anon_sym_f32] = ACTIONS(862),
    [anon_sym_f64] = ACTIONS(862),
    [anon_sym_bool] = ACTIONS(862),
    [anon_sym_str] = ACTIONS(862),
    [anon_sym_char] = ACTIONS(862),
    [anon_sym_SQUOTE] = ACTIONS(2313),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(664),
    [anon_sym_default] = ACTIONS(864),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_pub] = ACTIONS(2315),
    [anon_sym_union] = ACTIONS(866),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_POUND] = ACTIONS(2317),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(870),
    [anon_sym_AMP] = ACTIONS(872),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(876),
    [sym_super] = ACTIONS(876),
    [sym_crate] = ACTIONS(2321),
    [sym_metavariable] = ACTIONS(878),
    [sym_block_comment] = ACTIONS(3),
  },
  [552] = {
    [sym_attribute_item] = STATE(557),
    [sym_function_modifiers] = STATE(2472),
    [sym_extern_modifier] = STATE(1526),
    [sym_visibility_modifier] = STATE(645),
    [sym__type] = STATE(1917),
    [sym_bracketed_type] = STATE(2378),
    [sym_lifetime] = STATE(2471),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2379),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1390),
    [sym_scoped_identifier] = STATE(2157),
    [sym_scoped_type_identifier] = STATE(1331),
    [aux_sym_enum_variant_list_repeat1] = STATE(557),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(2309),
    [anon_sym_LPAREN] = ACTIONS(856),
    [anon_sym_RPAREN] = ACTIONS(2331),
    [anon_sym_LBRACK] = ACTIONS(860),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(862),
    [anon_sym_i8] = ACTIONS(862),
    [anon_sym_u16] = ACTIONS(862),
    [anon_sym_i16] = ACTIONS(862),
    [anon_sym_u32] = ACTIONS(862),
    [anon_sym_i32] = ACTIONS(862),
    [anon_sym_u64] = ACTIONS(862),
    [anon_sym_i64] = ACTIONS(862),
    [anon_sym_u128] = ACTIONS(862),
    [anon_sym_i128] = ACTIONS(862),
    [anon_sym_isize] = ACTIONS(862),
    [anon_sym_usize] = ACTIONS(862),
    [anon_sym_f32] = ACTIONS(862),
    [anon_sym_f64] = ACTIONS(862),
    [anon_sym_bool] = ACTIONS(862),
    [anon_sym_str] = ACTIONS(862),
    [anon_sym_char] = ACTIONS(862),
    [anon_sym_SQUOTE] = ACTIONS(2313),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(664),
    [anon_sym_default] = ACTIONS(864),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_pub] = ACTIONS(2315),
    [anon_sym_union] = ACTIONS(866),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_POUND] = ACTIONS(2317),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(870),
    [anon_sym_AMP] = ACTIONS(872),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(876),
    [sym_super] = ACTIONS(876),
    [sym_crate] = ACTIONS(2321),
    [sym_metavariable] = ACTIONS(878),
    [sym_block_comment] = ACTIONS(3),
  },
  [553] = {
    [sym_attribute_item] = STATE(557),
    [sym_function_modifiers] = STATE(2472),
    [sym_extern_modifier] = STATE(1526),
    [sym_visibility_modifier] = STATE(645),
    [sym__type] = STATE(1917),
    [sym_bracketed_type] = STATE(2378),
    [sym_lifetime] = STATE(2471),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2379),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1390),
    [sym_scoped_identifier] = STATE(2157),
    [sym_scoped_type_identifier] = STATE(1331),
    [aux_sym_enum_variant_list_repeat1] = STATE(557),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(2309),
    [anon_sym_LPAREN] = ACTIONS(856),
    [anon_sym_RPAREN] = ACTIONS(2333),
    [anon_sym_LBRACK] = ACTIONS(860),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(862),
    [anon_sym_i8] = ACTIONS(862),
    [anon_sym_u16] = ACTIONS(862),
    [anon_sym_i16] = ACTIONS(862),
    [anon_sym_u32] = ACTIONS(862),
    [anon_sym_i32] = ACTIONS(862),
    [anon_sym_u64] = ACTIONS(862),
    [anon_sym_i64] = ACTIONS(862),
    [anon_sym_u128] = ACTIONS(862),
    [anon_sym_i128] = ACTIONS(862),
    [anon_sym_isize] = ACTIONS(862),
    [anon_sym_usize] = ACTIONS(862),
    [anon_sym_f32] = ACTIONS(862),
    [anon_sym_f64] = ACTIONS(862),
    [anon_sym_bool] = ACTIONS(862),
    [anon_sym_str] = ACTIONS(862),
    [anon_sym_char] = ACTIONS(862),
    [anon_sym_SQUOTE] = ACTIONS(2313),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(664),
    [anon_sym_default] = ACTIONS(864),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_pub] = ACTIONS(2315),
    [anon_sym_union] = ACTIONS(866),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_POUND] = ACTIONS(2317),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(870),
    [anon_sym_AMP] = ACTIONS(872),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(876),
    [sym_super] = ACTIONS(876),
    [sym_crate] = ACTIONS(2321),
    [sym_metavariable] = ACTIONS(878),
    [sym_block_comment] = ACTIONS(3),
  },
  [554] = {
    [sym_attribute_item] = STATE(1153),
    [sym_function_modifiers] = STATE(2472),
    [sym_extern_modifier] = STATE(1526),
    [sym_visibility_modifier] = STATE(733),
    [sym__type] = STATE(1823),
    [sym_bracketed_type] = STATE(2378),
    [sym_lifetime] = STATE(2471),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2379),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1390),
    [sym_scoped_identifier] = STATE(2157),
    [sym_scoped_type_identifier] = STATE(1331),
    [aux_sym_enum_variant_list_repeat1] = STATE(1153),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(2309),
    [anon_sym_LPAREN] = ACTIONS(856),
    [anon_sym_LBRACK] = ACTIONS(860),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(862),
    [anon_sym_i8] = ACTIONS(862),
    [anon_sym_u16] = ACTIONS(862),
    [anon_sym_i16] = ACTIONS(862),
    [anon_sym_u32] = ACTIONS(862),
    [anon_sym_i32] = ACTIONS(862),
    [anon_sym_u64] = ACTIONS(862),
    [anon_sym_i64] = ACTIONS(862),
    [anon_sym_u128] = ACTIONS(862),
    [anon_sym_i128] = ACTIONS(862),
    [anon_sym_isize] = ACTIONS(862),
    [anon_sym_usize] = ACTIONS(862),
    [anon_sym_f32] = ACTIONS(862),
    [anon_sym_f64] = ACTIONS(862),
    [anon_sym_bool] = ACTIONS(862),
    [anon_sym_str] = ACTIONS(862),
    [anon_sym_char] = ACTIONS(862),
    [anon_sym_SQUOTE] = ACTIONS(2313),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(664),
    [anon_sym_default] = ACTIONS(864),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_pub] = ACTIONS(2315),
    [anon_sym_union] = ACTIONS(866),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_POUND] = ACTIONS(2317),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(870),
    [anon_sym_AMP] = ACTIONS(872),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(876),
    [sym_super] = ACTIONS(876),
    [sym_crate] = ACTIONS(2321),
    [sym_metavariable] = ACTIONS(878),
    [sym_block_comment] = ACTIONS(3),
  },
  [555] = {
    [sym_attribute_item] = STATE(557),
    [sym_function_modifiers] = STATE(2472),
    [sym_extern_modifier] = STATE(1526),
    [sym_visibility_modifier] = STATE(645),
    [sym__type] = STATE(1917),
    [sym_bracketed_type] = STATE(2378),
    [sym_lifetime] = STATE(2471),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2379),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1390),
    [sym_scoped_identifier] = STATE(2157),
    [sym_scoped_type_identifier] = STATE(1331),
    [aux_sym_enum_variant_list_repeat1] = STATE(557),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(2309),
    [anon_sym_LPAREN] = ACTIONS(856),
    [anon_sym_LBRACK] = ACTIONS(860),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(862),
    [anon_sym_i8] = ACTIONS(862),
    [anon_sym_u16] = ACTIONS(862),
    [anon_sym_i16] = ACTIONS(862),
    [anon_sym_u32] = ACTIONS(862),
    [anon_sym_i32] = ACTIONS(862),
    [anon_sym_u64] = ACTIONS(862),
    [anon_sym_i64] = ACTIONS(862),
    [anon_sym_u128] = ACTIONS(862),
    [anon_sym_i128] = ACTIONS(862),
    [anon_sym_isize] = ACTIONS(862),
    [anon_sym_usize] = ACTIONS(862),
    [anon_sym_f32] = ACTIONS(862),
    [anon_sym_f64] = ACTIONS(862),
    [anon_sym_bool] = ACTIONS(862),
    [anon_sym_str] = ACTIONS(862),
    [anon_sym_char] = ACTIONS(862),
    [anon_sym_SQUOTE] = ACTIONS(2313),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(664),
    [anon_sym_default] = ACTIONS(864),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_pub] = ACTIONS(2315),
    [anon_sym_union] = ACTIONS(866),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_POUND] = ACTIONS(2317),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(870),
    [anon_sym_AMP] = ACTIONS(872),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(876),
    [sym_super] = ACTIONS(876),
    [sym_crate] = ACTIONS(2321),
    [sym_metavariable] = ACTIONS(878),
    [sym_block_comment] = ACTIONS(3),
  },
  [556] = {
    [sym_parameter] = STATE(1980),
    [sym_bracketed_type] = STATE(2489),
    [sym_generic_type] = STATE(2488),
    [sym_generic_type_with_turbofish] = STATE(2487),
    [sym_macro_invocation] = STATE(1422),
    [sym_scoped_identifier] = STATE(1337),
    [sym_scoped_type_identifier] = STATE(2013),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1798),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [sym_identifier] = ACTIONS(1524),
    [anon_sym_LPAREN] = ACTIONS(1526),
    [anon_sym_LBRACK] = ACTIONS(1530),
    [anon_sym_u8] = ACTIONS(1532),
    [anon_sym_i8] = ACTIONS(1532),
    [anon_sym_u16] = ACTIONS(1532),
    [anon_sym_i16] = ACTIONS(1532),
    [anon_sym_u32] = ACTIONS(1532),
    [anon_sym_i32] = ACTIONS(1532),
    [anon_sym_u64] = ACTIONS(1532),
    [anon_sym_i64] = ACTIONS(1532),
    [anon_sym_u128] = ACTIONS(1532),
    [anon_sym_i128] = ACTIONS(1532),
    [anon_sym_isize] = ACTIONS(1532),
    [anon_sym_usize] = ACTIONS(1532),
    [anon_sym_f32] = ACTIONS(1532),
    [anon_sym_f64] = ACTIONS(1532),
    [anon_sym_bool] = ACTIONS(1532),
    [anon_sym_str] = ACTIONS(1532),
    [anon_sym_char] = ACTIONS(1532),
    [anon_sym_const] = ACTIONS(1534),
    [anon_sym_default] = ACTIONS(1536),
    [anon_sym_union] = ACTIONS(1536),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(1538),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(1540),
    [sym_mutable_specifier] = ACTIONS(2335),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [anon_sym_PIPE] = ACTIONS(2337),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(2339),
    [sym_super] = ACTIONS(1542),
    [sym_crate] = ACTIONS(1542),
    [sym_metavariable] = ACTIONS(1544),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [557] = {
    [sym_attribute_item] = STATE(1153),
    [sym_function_modifiers] = STATE(2472),
    [sym_extern_modifier] = STATE(1526),
    [sym_visibility_modifier] = STATE(691),
    [sym__type] = STATE(1883),
    [sym_bracketed_type] = STATE(2378),
    [sym_lifetime] = STATE(2471),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2379),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1390),
    [sym_scoped_identifier] = STATE(2157),
    [sym_scoped_type_identifier] = STATE(1331),
    [aux_sym_enum_variant_list_repeat1] = STATE(1153),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(2309),
    [anon_sym_LPAREN] = ACTIONS(856),
    [anon_sym_LBRACK] = ACTIONS(860),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(862),
    [anon_sym_i8] = ACTIONS(862),
    [anon_sym_u16] = ACTIONS(862),
    [anon_sym_i16] = ACTIONS(862),
    [anon_sym_u32] = ACTIONS(862),
    [anon_sym_i32] = ACTIONS(862),
    [anon_sym_u64] = ACTIONS(862),
    [anon_sym_i64] = ACTIONS(862),
    [anon_sym_u128] = ACTIONS(862),
    [anon_sym_i128] = ACTIONS(862),
    [anon_sym_isize] = ACTIONS(862),
    [anon_sym_usize] = ACTIONS(862),
    [anon_sym_f32] = ACTIONS(862),
    [anon_sym_f64] = ACTIONS(862),
    [anon_sym_bool] = ACTIONS(862),
    [anon_sym_str] = ACTIONS(862),
    [anon_sym_char] = ACTIONS(862),
    [anon_sym_SQUOTE] = ACTIONS(2313),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(664),
    [anon_sym_default] = ACTIONS(864),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_pub] = ACTIONS(2315),
    [anon_sym_union] = ACTIONS(866),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_POUND] = ACTIONS(2317),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(870),
    [anon_sym_AMP] = ACTIONS(872),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(876),
    [sym_super] = ACTIONS(876),
    [sym_crate] = ACTIONS(2321),
    [sym_metavariable] = ACTIONS(878),
    [sym_block_comment] = ACTIONS(3),
  },
  [558] = {
    [sym_identifier] = ACTIONS(2341),
    [anon_sym_LPAREN] = ACTIONS(2343),
    [anon_sym_RPAREN] = ACTIONS(2343),
    [anon_sym_LBRACE] = ACTIONS(2343),
    [anon_sym_RBRACE] = ACTIONS(2343),
    [anon_sym_LBRACK] = ACTIONS(2343),
    [anon_sym_RBRACK] = ACTIONS(2343),
    [anon_sym_COLON] = ACTIONS(2345),
    [anon_sym_DOLLAR] = ACTIONS(2341),
    [anon_sym_u8] = ACTIONS(2341),
    [anon_sym_i8] = ACTIONS(2341),
    [anon_sym_u16] = ACTIONS(2341),
    [anon_sym_i16] = ACTIONS(2341),
    [anon_sym_u32] = ACTIONS(2341),
    [anon_sym_i32] = ACTIONS(2341),
    [anon_sym_u64] = ACTIONS(2341),
    [anon_sym_i64] = ACTIONS(2341),
    [anon_sym_u128] = ACTIONS(2341),
    [anon_sym_i128] = ACTIONS(2341),
    [anon_sym_isize] = ACTIONS(2341),
    [anon_sym_usize] = ACTIONS(2341),
    [anon_sym_f32] = ACTIONS(2341),
    [anon_sym_f64] = ACTIONS(2341),
    [anon_sym_bool] = ACTIONS(2341),
    [anon_sym_str] = ACTIONS(2341),
    [anon_sym_char] = ACTIONS(2341),
    [aux_sym__non_special_token_token1] = ACTIONS(2341),
    [anon_sym_SQUOTE] = ACTIONS(2341),
    [anon_sym_as] = ACTIONS(2341),
    [anon_sym_async] = ACTIONS(2341),
    [anon_sym_await] = ACTIONS(2341),
    [anon_sym_break] = ACTIONS(2341),
    [anon_sym_const] = ACTIONS(2341),
    [anon_sym_continue] = ACTIONS(2341),
    [anon_sym_default] = ACTIONS(2341),
    [anon_sym_enum] = ACTIONS(2341),
    [anon_sym_fn] = ACTIONS(2341),
    [anon_sym_for] = ACTIONS(2341),
    [anon_sym_if] = ACTIONS(2341),
    [anon_sym_impl] = ACTIONS(2341),
    [anon_sym_let] = ACTIONS(2341),
    [anon_sym_loop] = ACTIONS(2341),
    [anon_sym_match] = ACTIONS(2341),
    [anon_sym_mod] = ACTIONS(2341),
    [anon_sym_pub] = ACTIONS(2341),
    [anon_sym_return] = ACTIONS(2341),
    [anon_sym_static] = ACTIONS(2341),
    [anon_sym_struct] = ACTIONS(2341),
    [anon_sym_trait] = ACTIONS(2341),
    [anon_sym_type] = ACTIONS(2341),
    [anon_sym_union] = ACTIONS(2341),
    [anon_sym_unsafe] = ACTIONS(2341),
    [anon_sym_use] = ACTIONS(2341),
    [anon_sym_where] = ACTIONS(2341),
    [anon_sym_while] = ACTIONS(2341),
    [sym_mutable_specifier] = ACTIONS(2341),
    [sym_integer_literal] = ACTIONS(2343),
    [aux_sym_string_literal_token1] = ACTIONS(2343),
    [sym_char_literal] = ACTIONS(2343),
    [anon_sym_true] = ACTIONS(2341),
    [anon_sym_false] = ACTIONS(2341),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2341),
    [sym_super] = ACTIONS(2341),
    [sym_crate] = ACTIONS(2341),
    [sym_metavariable] = ACTIONS(2343),
    [sym_raw_string_literal] = ACTIONS(2343),
    [sym_float_literal] = ACTIONS(2343),
    [sym_block_comment] = ACTIONS(3),
  },
  [559] = {
    [sym_bracketed_type] = STATE(2489),
    [sym_generic_type] = STATE(2488),
    [sym_generic_type_with_turbofish] = STATE(2487),
    [sym_macro_invocation] = STATE(1422),
    [sym_scoped_identifier] = STATE(1337),
    [sym_scoped_type_identifier] = STATE(2013),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1791),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [sym_identifier] = ACTIONS(1524),
    [anon_sym_LPAREN] = ACTIONS(1526),
    [anon_sym_LBRACK] = ACTIONS(1530),
    [anon_sym_RBRACK] = ACTIONS(780),
    [anon_sym_u8] = ACTIONS(1532),
    [anon_sym_i8] = ACTIONS(1532),
    [anon_sym_u16] = ACTIONS(1532),
    [anon_sym_i16] = ACTIONS(1532),
    [anon_sym_u32] = ACTIONS(1532),
    [anon_sym_i32] = ACTIONS(1532),
    [anon_sym_u64] = ACTIONS(1532),
    [anon_sym_i64] = ACTIONS(1532),
    [anon_sym_u128] = ACTIONS(1532),
    [anon_sym_i128] = ACTIONS(1532),
    [anon_sym_isize] = ACTIONS(1532),
    [anon_sym_usize] = ACTIONS(1532),
    [anon_sym_f32] = ACTIONS(1532),
    [anon_sym_f64] = ACTIONS(1532),
    [anon_sym_bool] = ACTIONS(1532),
    [anon_sym_str] = ACTIONS(1532),
    [anon_sym_char] = ACTIONS(1532),
    [anon_sym_const] = ACTIONS(1534),
    [anon_sym_default] = ACTIONS(1536),
    [anon_sym_union] = ACTIONS(1536),
    [anon_sym_COMMA] = ACTIONS(788),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(1538),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(1540),
    [sym_mutable_specifier] = ACTIONS(796),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1542),
    [sym_super] = ACTIONS(1542),
    [sym_crate] = ACTIONS(1542),
    [sym_metavariable] = ACTIONS(1544),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [560] = {
    [sym_bracketed_type] = STATE(2489),
    [sym_generic_type] = STATE(2488),
    [sym_generic_type_with_turbofish] = STATE(2487),
    [sym_macro_invocation] = STATE(1422),
    [sym_scoped_identifier] = STATE(1337),
    [sym_scoped_type_identifier] = STATE(2013),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1759),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [sym_identifier] = ACTIONS(1524),
    [anon_sym_LPAREN] = ACTIONS(1526),
    [anon_sym_RPAREN] = ACTIONS(2347),
    [anon_sym_LBRACK] = ACTIONS(1530),
    [anon_sym_u8] = ACTIONS(1532),
    [anon_sym_i8] = ACTIONS(1532),
    [anon_sym_u16] = ACTIONS(1532),
    [anon_sym_i16] = ACTIONS(1532),
    [anon_sym_u32] = ACTIONS(1532),
    [anon_sym_i32] = ACTIONS(1532),
    [anon_sym_u64] = ACTIONS(1532),
    [anon_sym_i64] = ACTIONS(1532),
    [anon_sym_u128] = ACTIONS(1532),
    [anon_sym_i128] = ACTIONS(1532),
    [anon_sym_isize] = ACTIONS(1532),
    [anon_sym_usize] = ACTIONS(1532),
    [anon_sym_f32] = ACTIONS(1532),
    [anon_sym_f64] = ACTIONS(1532),
    [anon_sym_bool] = ACTIONS(1532),
    [anon_sym_str] = ACTIONS(1532),
    [anon_sym_char] = ACTIONS(1532),
    [anon_sym_const] = ACTIONS(1534),
    [anon_sym_default] = ACTIONS(1536),
    [anon_sym_union] = ACTIONS(1536),
    [anon_sym_COMMA] = ACTIONS(2349),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(1538),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(1540),
    [sym_mutable_specifier] = ACTIONS(796),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1542),
    [sym_super] = ACTIONS(1542),
    [sym_crate] = ACTIONS(1542),
    [sym_metavariable] = ACTIONS(1544),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [561] = {
    [sym_bracketed_type] = STATE(2489),
    [sym_generic_type] = STATE(2488),
    [sym_generic_type_with_turbofish] = STATE(2487),
    [sym_macro_invocation] = STATE(1422),
    [sym_scoped_identifier] = STATE(1337),
    [sym_scoped_type_identifier] = STATE(2013),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1789),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [sym_identifier] = ACTIONS(1524),
    [anon_sym_LPAREN] = ACTIONS(1526),
    [anon_sym_RPAREN] = ACTIONS(2351),
    [anon_sym_LBRACK] = ACTIONS(1530),
    [anon_sym_u8] = ACTIONS(1532),
    [anon_sym_i8] = ACTIONS(1532),
    [anon_sym_u16] = ACTIONS(1532),
    [anon_sym_i16] = ACTIONS(1532),
    [anon_sym_u32] = ACTIONS(1532),
    [anon_sym_i32] = ACTIONS(1532),
    [anon_sym_u64] = ACTIONS(1532),
    [anon_sym_i64] = ACTIONS(1532),
    [anon_sym_u128] = ACTIONS(1532),
    [anon_sym_i128] = ACTIONS(1532),
    [anon_sym_isize] = ACTIONS(1532),
    [anon_sym_usize] = ACTIONS(1532),
    [anon_sym_f32] = ACTIONS(1532),
    [anon_sym_f64] = ACTIONS(1532),
    [anon_sym_bool] = ACTIONS(1532),
    [anon_sym_str] = ACTIONS(1532),
    [anon_sym_char] = ACTIONS(1532),
    [anon_sym_const] = ACTIONS(1534),
    [anon_sym_default] = ACTIONS(1536),
    [anon_sym_union] = ACTIONS(1536),
    [anon_sym_COMMA] = ACTIONS(806),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(1538),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(1540),
    [sym_mutable_specifier] = ACTIONS(796),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1542),
    [sym_super] = ACTIONS(1542),
    [sym_crate] = ACTIONS(1542),
    [sym_metavariable] = ACTIONS(1544),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [562] = {
    [sym_parameter] = STATE(2138),
    [sym_bracketed_type] = STATE(2489),
    [sym_generic_type] = STATE(2488),
    [sym_generic_type_with_turbofish] = STATE(2487),
    [sym_macro_invocation] = STATE(1422),
    [sym_scoped_identifier] = STATE(1337),
    [sym_scoped_type_identifier] = STATE(2013),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(2088),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [sym_identifier] = ACTIONS(1524),
    [anon_sym_LPAREN] = ACTIONS(1526),
    [anon_sym_LBRACK] = ACTIONS(1530),
    [anon_sym_u8] = ACTIONS(1532),
    [anon_sym_i8] = ACTIONS(1532),
    [anon_sym_u16] = ACTIONS(1532),
    [anon_sym_i16] = ACTIONS(1532),
    [anon_sym_u32] = ACTIONS(1532),
    [anon_sym_i32] = ACTIONS(1532),
    [anon_sym_u64] = ACTIONS(1532),
    [anon_sym_i64] = ACTIONS(1532),
    [anon_sym_u128] = ACTIONS(1532),
    [anon_sym_i128] = ACTIONS(1532),
    [anon_sym_isize] = ACTIONS(1532),
    [anon_sym_usize] = ACTIONS(1532),
    [anon_sym_f32] = ACTIONS(1532),
    [anon_sym_f64] = ACTIONS(1532),
    [anon_sym_bool] = ACTIONS(1532),
    [anon_sym_str] = ACTIONS(1532),
    [anon_sym_char] = ACTIONS(1532),
    [anon_sym_const] = ACTIONS(1534),
    [anon_sym_default] = ACTIONS(1536),
    [anon_sym_union] = ACTIONS(1536),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(1538),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(1540),
    [sym_mutable_specifier] = ACTIONS(2335),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(2339),
    [sym_super] = ACTIONS(1542),
    [sym_crate] = ACTIONS(1542),
    [sym_metavariable] = ACTIONS(1544),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [563] = {
    [sym_identifier] = ACTIONS(2353),
    [anon_sym_LPAREN] = ACTIONS(2355),
    [anon_sym_RPAREN] = ACTIONS(2355),
    [anon_sym_LBRACE] = ACTIONS(2355),
    [anon_sym_RBRACE] = ACTIONS(2355),
    [anon_sym_LBRACK] = ACTIONS(2355),
    [anon_sym_RBRACK] = ACTIONS(2355),
    [anon_sym_DOLLAR] = ACTIONS(2353),
    [anon_sym_u8] = ACTIONS(2353),
    [anon_sym_i8] = ACTIONS(2353),
    [anon_sym_u16] = ACTIONS(2353),
    [anon_sym_i16] = ACTIONS(2353),
    [anon_sym_u32] = ACTIONS(2353),
    [anon_sym_i32] = ACTIONS(2353),
    [anon_sym_u64] = ACTIONS(2353),
    [anon_sym_i64] = ACTIONS(2353),
    [anon_sym_u128] = ACTIONS(2353),
    [anon_sym_i128] = ACTIONS(2353),
    [anon_sym_isize] = ACTIONS(2353),
    [anon_sym_usize] = ACTIONS(2353),
    [anon_sym_f32] = ACTIONS(2353),
    [anon_sym_f64] = ACTIONS(2353),
    [anon_sym_bool] = ACTIONS(2353),
    [anon_sym_str] = ACTIONS(2353),
    [anon_sym_char] = ACTIONS(2353),
    [aux_sym__non_special_token_token1] = ACTIONS(2353),
    [anon_sym_SQUOTE] = ACTIONS(2353),
    [anon_sym_as] = ACTIONS(2353),
    [anon_sym_async] = ACTIONS(2353),
    [anon_sym_await] = ACTIONS(2353),
    [anon_sym_break] = ACTIONS(2353),
    [anon_sym_const] = ACTIONS(2353),
    [anon_sym_continue] = ACTIONS(2353),
    [anon_sym_default] = ACTIONS(2353),
    [anon_sym_enum] = ACTIONS(2353),
    [anon_sym_fn] = ACTIONS(2353),
    [anon_sym_for] = ACTIONS(2353),
    [anon_sym_if] = ACTIONS(2353),
    [anon_sym_impl] = ACTIONS(2353),
    [anon_sym_let] = ACTIONS(2353),
    [anon_sym_loop] = ACTIONS(2353),
    [anon_sym_match] = ACTIONS(2353),
    [anon_sym_mod] = ACTIONS(2353),
    [anon_sym_pub] = ACTIONS(2353),
    [anon_sym_return] = ACTIONS(2353),
    [anon_sym_static] = ACTIONS(2353),
    [anon_sym_struct] = ACTIONS(2353),
    [anon_sym_trait] = ACTIONS(2353),
    [anon_sym_type] = ACTIONS(2353),
    [anon_sym_union] = ACTIONS(2353),
    [anon_sym_unsafe] = ACTIONS(2353),
    [anon_sym_use] = ACTIONS(2353),
    [anon_sym_where] = ACTIONS(2353),
    [anon_sym_while] = ACTIONS(2353),
    [sym_mutable_specifier] = ACTIONS(2353),
    [sym_integer_literal] = ACTIONS(2355),
    [aux_sym_string_literal_token1] = ACTIONS(2355),
    [sym_char_literal] = ACTIONS(2355),
    [anon_sym_true] = ACTIONS(2353),
    [anon_sym_false] = ACTIONS(2353),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2353),
    [sym_super] = ACTIONS(2353),
    [sym_crate] = ACTIONS(2353),
    [sym_metavariable] = ACTIONS(2355),
    [sym_raw_string_literal] = ACTIONS(2355),
    [sym_float_literal] = ACTIONS(2355),
    [sym_block_comment] = ACTIONS(3),
  },
  [564] = {
    [sym_identifier] = ACTIONS(2357),
    [anon_sym_LPAREN] = ACTIONS(2359),
    [anon_sym_RPAREN] = ACTIONS(2359),
    [anon_sym_LBRACE] = ACTIONS(2359),
    [anon_sym_RBRACE] = ACTIONS(2359),
    [anon_sym_LBRACK] = ACTIONS(2359),
    [anon_sym_RBRACK] = ACTIONS(2359),
    [anon_sym_DOLLAR] = ACTIONS(2357),
    [anon_sym_u8] = ACTIONS(2357),
    [anon_sym_i8] = ACTIONS(2357),
    [anon_sym_u16] = ACTIONS(2357),
    [anon_sym_i16] = ACTIONS(2357),
    [anon_sym_u32] = ACTIONS(2357),
    [anon_sym_i32] = ACTIONS(2357),
    [anon_sym_u64] = ACTIONS(2357),
    [anon_sym_i64] = ACTIONS(2357),
    [anon_sym_u128] = ACTIONS(2357),
    [anon_sym_i128] = ACTIONS(2357),
    [anon_sym_isize] = ACTIONS(2357),
    [anon_sym_usize] = ACTIONS(2357),
    [anon_sym_f32] = ACTIONS(2357),
    [anon_sym_f64] = ACTIONS(2357),
    [anon_sym_bool] = ACTIONS(2357),
    [anon_sym_str] = ACTIONS(2357),
    [anon_sym_char] = ACTIONS(2357),
    [aux_sym__non_special_token_token1] = ACTIONS(2357),
    [anon_sym_SQUOTE] = ACTIONS(2357),
    [anon_sym_as] = ACTIONS(2357),
    [anon_sym_async] = ACTIONS(2357),
    [anon_sym_await] = ACTIONS(2357),
    [anon_sym_break] = ACTIONS(2357),
    [anon_sym_const] = ACTIONS(2357),
    [anon_sym_continue] = ACTIONS(2357),
    [anon_sym_default] = ACTIONS(2357),
    [anon_sym_enum] = ACTIONS(2357),
    [anon_sym_fn] = ACTIONS(2357),
    [anon_sym_for] = ACTIONS(2357),
    [anon_sym_if] = ACTIONS(2357),
    [anon_sym_impl] = ACTIONS(2357),
    [anon_sym_let] = ACTIONS(2357),
    [anon_sym_loop] = ACTIONS(2357),
    [anon_sym_match] = ACTIONS(2357),
    [anon_sym_mod] = ACTIONS(2357),
    [anon_sym_pub] = ACTIONS(2357),
    [anon_sym_return] = ACTIONS(2357),
    [anon_sym_static] = ACTIONS(2357),
    [anon_sym_struct] = ACTIONS(2357),
    [anon_sym_trait] = ACTIONS(2357),
    [anon_sym_type] = ACTIONS(2357),
    [anon_sym_union] = ACTIONS(2357),
    [anon_sym_unsafe] = ACTIONS(2357),
    [anon_sym_use] = ACTIONS(2357),
    [anon_sym_where] = ACTIONS(2357),
    [anon_sym_while] = ACTIONS(2357),
    [sym_mutable_specifier] = ACTIONS(2357),
    [sym_integer_literal] = ACTIONS(2359),
    [aux_sym_string_literal_token1] = ACTIONS(2359),
    [sym_char_literal] = ACTIONS(2359),
    [anon_sym_true] = ACTIONS(2357),
    [anon_sym_false] = ACTIONS(2357),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2357),
    [sym_super] = ACTIONS(2357),
    [sym_crate] = ACTIONS(2357),
    [sym_metavariable] = ACTIONS(2359),
    [sym_raw_string_literal] = ACTIONS(2359),
    [sym_float_literal] = ACTIONS(2359),
    [sym_block_comment] = ACTIONS(3),
  },
  [565] = {
    [sym_identifier] = ACTIONS(2361),
    [anon_sym_LPAREN] = ACTIONS(2363),
    [anon_sym_RPAREN] = ACTIONS(2363),
    [anon_sym_LBRACE] = ACTIONS(2363),
    [anon_sym_RBRACE] = ACTIONS(2363),
    [anon_sym_LBRACK] = ACTIONS(2363),
    [anon_sym_RBRACK] = ACTIONS(2363),
    [anon_sym_DOLLAR] = ACTIONS(2361),
    [anon_sym_u8] = ACTIONS(2361),
    [anon_sym_i8] = ACTIONS(2361),
    [anon_sym_u16] = ACTIONS(2361),
    [anon_sym_i16] = ACTIONS(2361),
    [anon_sym_u32] = ACTIONS(2361),
    [anon_sym_i32] = ACTIONS(2361),
    [anon_sym_u64] = ACTIONS(2361),
    [anon_sym_i64] = ACTIONS(2361),
    [anon_sym_u128] = ACTIONS(2361),
    [anon_sym_i128] = ACTIONS(2361),
    [anon_sym_isize] = ACTIONS(2361),
    [anon_sym_usize] = ACTIONS(2361),
    [anon_sym_f32] = ACTIONS(2361),
    [anon_sym_f64] = ACTIONS(2361),
    [anon_sym_bool] = ACTIONS(2361),
    [anon_sym_str] = ACTIONS(2361),
    [anon_sym_char] = ACTIONS(2361),
    [aux_sym__non_special_token_token1] = ACTIONS(2361),
    [anon_sym_SQUOTE] = ACTIONS(2361),
    [anon_sym_as] = ACTIONS(2361),
    [anon_sym_async] = ACTIONS(2361),
    [anon_sym_await] = ACTIONS(2361),
    [anon_sym_break] = ACTIONS(2361),
    [anon_sym_const] = ACTIONS(2361),
    [anon_sym_continue] = ACTIONS(2361),
    [anon_sym_default] = ACTIONS(2361),
    [anon_sym_enum] = ACTIONS(2361),
    [anon_sym_fn] = ACTIONS(2361),
    [anon_sym_for] = ACTIONS(2361),
    [anon_sym_if] = ACTIONS(2361),
    [anon_sym_impl] = ACTIONS(2361),
    [anon_sym_let] = ACTIONS(2361),
    [anon_sym_loop] = ACTIONS(2361),
    [anon_sym_match] = ACTIONS(2361),
    [anon_sym_mod] = ACTIONS(2361),
    [anon_sym_pub] = ACTIONS(2361),
    [anon_sym_return] = ACTIONS(2361),
    [anon_sym_static] = ACTIONS(2361),
    [anon_sym_struct] = ACTIONS(2361),
    [anon_sym_trait] = ACTIONS(2361),
    [anon_sym_type] = ACTIONS(2361),
    [anon_sym_union] = ACTIONS(2361),
    [anon_sym_unsafe] = ACTIONS(2361),
    [anon_sym_use] = ACTIONS(2361),
    [anon_sym_where] = ACTIONS(2361),
    [anon_sym_while] = ACTIONS(2361),
    [sym_mutable_specifier] = ACTIONS(2361),
    [sym_integer_literal] = ACTIONS(2363),
    [aux_sym_string_literal_token1] = ACTIONS(2363),
    [sym_char_literal] = ACTIONS(2363),
    [anon_sym_true] = ACTIONS(2361),
    [anon_sym_false] = ACTIONS(2361),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2361),
    [sym_super] = ACTIONS(2361),
    [sym_crate] = ACTIONS(2361),
    [sym_metavariable] = ACTIONS(2363),
    [sym_raw_string_literal] = ACTIONS(2363),
    [sym_float_literal] = ACTIONS(2363),
    [sym_block_comment] = ACTIONS(3),
  },
  [566] = {
    [sym_bracketed_type] = STATE(2489),
    [sym_generic_type] = STATE(2488),
    [sym_generic_type_with_turbofish] = STATE(2487),
    [sym_macro_invocation] = STATE(1422),
    [sym_scoped_identifier] = STATE(1337),
    [sym_scoped_type_identifier] = STATE(2013),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1804),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [sym_identifier] = ACTIONS(1524),
    [anon_sym_LPAREN] = ACTIONS(1526),
    [anon_sym_RPAREN] = ACTIONS(2365),
    [anon_sym_LBRACK] = ACTIONS(1530),
    [anon_sym_u8] = ACTIONS(1532),
    [anon_sym_i8] = ACTIONS(1532),
    [anon_sym_u16] = ACTIONS(1532),
    [anon_sym_i16] = ACTIONS(1532),
    [anon_sym_u32] = ACTIONS(1532),
    [anon_sym_i32] = ACTIONS(1532),
    [anon_sym_u64] = ACTIONS(1532),
    [anon_sym_i64] = ACTIONS(1532),
    [anon_sym_u128] = ACTIONS(1532),
    [anon_sym_i128] = ACTIONS(1532),
    [anon_sym_isize] = ACTIONS(1532),
    [anon_sym_usize] = ACTIONS(1532),
    [anon_sym_f32] = ACTIONS(1532),
    [anon_sym_f64] = ACTIONS(1532),
    [anon_sym_bool] = ACTIONS(1532),
    [anon_sym_str] = ACTIONS(1532),
    [anon_sym_char] = ACTIONS(1532),
    [anon_sym_const] = ACTIONS(1534),
    [anon_sym_default] = ACTIONS(1536),
    [anon_sym_union] = ACTIONS(1536),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(1538),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(1540),
    [sym_mutable_specifier] = ACTIONS(796),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1542),
    [sym_super] = ACTIONS(1542),
    [sym_crate] = ACTIONS(1542),
    [sym_metavariable] = ACTIONS(1544),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [567] = {
    [sym_identifier] = ACTIONS(2367),
    [anon_sym_LPAREN] = ACTIONS(2369),
    [anon_sym_RPAREN] = ACTIONS(2369),
    [anon_sym_LBRACE] = ACTIONS(2369),
    [anon_sym_RBRACE] = ACTIONS(2369),
    [anon_sym_LBRACK] = ACTIONS(2369),
    [anon_sym_RBRACK] = ACTIONS(2369),
    [anon_sym_DOLLAR] = ACTIONS(2367),
    [anon_sym_u8] = ACTIONS(2367),
    [anon_sym_i8] = ACTIONS(2367),
    [anon_sym_u16] = ACTIONS(2367),
    [anon_sym_i16] = ACTIONS(2367),
    [anon_sym_u32] = ACTIONS(2367),
    [anon_sym_i32] = ACTIONS(2367),
    [anon_sym_u64] = ACTIONS(2367),
    [anon_sym_i64] = ACTIONS(2367),
    [anon_sym_u128] = ACTIONS(2367),
    [anon_sym_i128] = ACTIONS(2367),
    [anon_sym_isize] = ACTIONS(2367),
    [anon_sym_usize] = ACTIONS(2367),
    [anon_sym_f32] = ACTIONS(2367),
    [anon_sym_f64] = ACTIONS(2367),
    [anon_sym_bool] = ACTIONS(2367),
    [anon_sym_str] = ACTIONS(2367),
    [anon_sym_char] = ACTIONS(2367),
    [aux_sym__non_special_token_token1] = ACTIONS(2367),
    [anon_sym_SQUOTE] = ACTIONS(2367),
    [anon_sym_as] = ACTIONS(2367),
    [anon_sym_async] = ACTIONS(2367),
    [anon_sym_await] = ACTIONS(2367),
    [anon_sym_break] = ACTIONS(2367),
    [anon_sym_const] = ACTIONS(2367),
    [anon_sym_continue] = ACTIONS(2367),
    [anon_sym_default] = ACTIONS(2367),
    [anon_sym_enum] = ACTIONS(2367),
    [anon_sym_fn] = ACTIONS(2367),
    [anon_sym_for] = ACTIONS(2367),
    [anon_sym_if] = ACTIONS(2367),
    [anon_sym_impl] = ACTIONS(2367),
    [anon_sym_let] = ACTIONS(2367),
    [anon_sym_loop] = ACTIONS(2367),
    [anon_sym_match] = ACTIONS(2367),
    [anon_sym_mod] = ACTIONS(2367),
    [anon_sym_pub] = ACTIONS(2367),
    [anon_sym_return] = ACTIONS(2367),
    [anon_sym_static] = ACTIONS(2367),
    [anon_sym_struct] = ACTIONS(2367),
    [anon_sym_trait] = ACTIONS(2367),
    [anon_sym_type] = ACTIONS(2367),
    [anon_sym_union] = ACTIONS(2367),
    [anon_sym_unsafe] = ACTIONS(2367),
    [anon_sym_use] = ACTIONS(2367),
    [anon_sym_where] = ACTIONS(2367),
    [anon_sym_while] = ACTIONS(2367),
    [sym_mutable_specifier] = ACTIONS(2367),
    [sym_integer_literal] = ACTIONS(2369),
    [aux_sym_string_literal_token1] = ACTIONS(2369),
    [sym_char_literal] = ACTIONS(2369),
    [anon_sym_true] = ACTIONS(2367),
    [anon_sym_false] = ACTIONS(2367),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2367),
    [sym_super] = ACTIONS(2367),
    [sym_crate] = ACTIONS(2367),
    [sym_metavariable] = ACTIONS(2369),
    [sym_raw_string_literal] = ACTIONS(2369),
    [sym_float_literal] = ACTIONS(2369),
    [sym_block_comment] = ACTIONS(3),
  },
  [568] = {
    [sym_identifier] = ACTIONS(2371),
    [anon_sym_LPAREN] = ACTIONS(2373),
    [anon_sym_RPAREN] = ACTIONS(2373),
    [anon_sym_LBRACE] = ACTIONS(2373),
    [anon_sym_RBRACE] = ACTIONS(2373),
    [anon_sym_LBRACK] = ACTIONS(2373),
    [anon_sym_RBRACK] = ACTIONS(2373),
    [anon_sym_DOLLAR] = ACTIONS(2371),
    [anon_sym_u8] = ACTIONS(2371),
    [anon_sym_i8] = ACTIONS(2371),
    [anon_sym_u16] = ACTIONS(2371),
    [anon_sym_i16] = ACTIONS(2371),
    [anon_sym_u32] = ACTIONS(2371),
    [anon_sym_i32] = ACTIONS(2371),
    [anon_sym_u64] = ACTIONS(2371),
    [anon_sym_i64] = ACTIONS(2371),
    [anon_sym_u128] = ACTIONS(2371),
    [anon_sym_i128] = ACTIONS(2371),
    [anon_sym_isize] = ACTIONS(2371),
    [anon_sym_usize] = ACTIONS(2371),
    [anon_sym_f32] = ACTIONS(2371),
    [anon_sym_f64] = ACTIONS(2371),
    [anon_sym_bool] = ACTIONS(2371),
    [anon_sym_str] = ACTIONS(2371),
    [anon_sym_char] = ACTIONS(2371),
    [aux_sym__non_special_token_token1] = ACTIONS(2371),
    [anon_sym_SQUOTE] = ACTIONS(2371),
    [anon_sym_as] = ACTIONS(2371),
    [anon_sym_async] = ACTIONS(2371),
    [anon_sym_await] = ACTIONS(2371),
    [anon_sym_break] = ACTIONS(2371),
    [anon_sym_const] = ACTIONS(2371),
    [anon_sym_continue] = ACTIONS(2371),
    [anon_sym_default] = ACTIONS(2371),
    [anon_sym_enum] = ACTIONS(2371),
    [anon_sym_fn] = ACTIONS(2371),
    [anon_sym_for] = ACTIONS(2371),
    [anon_sym_if] = ACTIONS(2371),
    [anon_sym_impl] = ACTIONS(2371),
    [anon_sym_let] = ACTIONS(2371),
    [anon_sym_loop] = ACTIONS(2371),
    [anon_sym_match] = ACTIONS(2371),
    [anon_sym_mod] = ACTIONS(2371),
    [anon_sym_pub] = ACTIONS(2371),
    [anon_sym_return] = ACTIONS(2371),
    [anon_sym_static] = ACTIONS(2371),
    [anon_sym_struct] = ACTIONS(2371),
    [anon_sym_trait] = ACTIONS(2371),
    [anon_sym_type] = ACTIONS(2371),
    [anon_sym_union] = ACTIONS(2371),
    [anon_sym_unsafe] = ACTIONS(2371),
    [anon_sym_use] = ACTIONS(2371),
    [anon_sym_where] = ACTIONS(2371),
    [anon_sym_while] = ACTIONS(2371),
    [sym_mutable_specifier] = ACTIONS(2371),
    [sym_integer_literal] = ACTIONS(2373),
    [aux_sym_string_literal_token1] = ACTIONS(2373),
    [sym_char_literal] = ACTIONS(2373),
    [anon_sym_true] = ACTIONS(2371),
    [anon_sym_false] = ACTIONS(2371),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2371),
    [sym_super] = ACTIONS(2371),
    [sym_crate] = ACTIONS(2371),
    [sym_metavariable] = ACTIONS(2373),
    [sym_raw_string_literal] = ACTIONS(2373),
    [sym_float_literal] = ACTIONS(2373),
    [sym_block_comment] = ACTIONS(3),
  },
  [569] = {
    [sym_function_modifiers] = STATE(2472),
    [sym_const_parameter] = STATE(2006),
    [sym_constrained_type_parameter] = STATE(1758),
    [sym_optional_type_parameter] = STATE(2006),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(2103),
    [sym_bracketed_type] = STATE(2378),
    [sym_qualified_type] = STATE(2448),
    [sym_lifetime] = STATE(1654),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2379),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1390),
    [sym_scoped_identifier] = STATE(2157),
    [sym_scoped_type_identifier] = STATE(1331),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(2375),
    [anon_sym_LPAREN] = ACTIONS(856),
    [anon_sym_LBRACK] = ACTIONS(860),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(862),
    [anon_sym_i8] = ACTIONS(862),
    [anon_sym_u16] = ACTIONS(862),
    [anon_sym_i16] = ACTIONS(862),
    [anon_sym_u32] = ACTIONS(862),
    [anon_sym_i32] = ACTIONS(862),
    [anon_sym_u64] = ACTIONS(862),
    [anon_sym_i64] = ACTIONS(862),
    [anon_sym_u128] = ACTIONS(862),
    [anon_sym_i128] = ACTIONS(862),
    [anon_sym_isize] = ACTIONS(862),
    [anon_sym_usize] = ACTIONS(862),
    [anon_sym_f32] = ACTIONS(862),
    [anon_sym_f64] = ACTIONS(862),
    [anon_sym_bool] = ACTIONS(862),
    [anon_sym_str] = ACTIONS(862),
    [anon_sym_char] = ACTIONS(862),
    [anon_sym_SQUOTE] = ACTIONS(2313),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(2377),
    [anon_sym_default] = ACTIONS(864),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(866),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(870),
    [anon_sym_AMP] = ACTIONS(872),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(876),
    [sym_super] = ACTIONS(876),
    [sym_crate] = ACTIONS(876),
    [sym_metavariable] = ACTIONS(2379),
    [sym_block_comment] = ACTIONS(3),
  },
  [570] = {
    [sym_bracketed_type] = STATE(2489),
    [sym_generic_type] = STATE(2488),
    [sym_generic_type_with_turbofish] = STATE(2487),
    [sym_macro_invocation] = STATE(1422),
    [sym_scoped_identifier] = STATE(1337),
    [sym_scoped_type_identifier] = STATE(2013),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1804),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [sym_identifier] = ACTIONS(1524),
    [anon_sym_LPAREN] = ACTIONS(1526),
    [anon_sym_RPAREN] = ACTIONS(2381),
    [anon_sym_LBRACK] = ACTIONS(1530),
    [anon_sym_u8] = ACTIONS(1532),
    [anon_sym_i8] = ACTIONS(1532),
    [anon_sym_u16] = ACTIONS(1532),
    [anon_sym_i16] = ACTIONS(1532),
    [anon_sym_u32] = ACTIONS(1532),
    [anon_sym_i32] = ACTIONS(1532),
    [anon_sym_u64] = ACTIONS(1532),
    [anon_sym_i64] = ACTIONS(1532),
    [anon_sym_u128] = ACTIONS(1532),
    [anon_sym_i128] = ACTIONS(1532),
    [anon_sym_isize] = ACTIONS(1532),
    [anon_sym_usize] = ACTIONS(1532),
    [anon_sym_f32] = ACTIONS(1532),
    [anon_sym_f64] = ACTIONS(1532),
    [anon_sym_bool] = ACTIONS(1532),
    [anon_sym_str] = ACTIONS(1532),
    [anon_sym_char] = ACTIONS(1532),
    [anon_sym_const] = ACTIONS(1534),
    [anon_sym_default] = ACTIONS(1536),
    [anon_sym_union] = ACTIONS(1536),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(1538),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(1540),
    [sym_mutable_specifier] = ACTIONS(796),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1542),
    [sym_super] = ACTIONS(1542),
    [sym_crate] = ACTIONS(1542),
    [sym_metavariable] = ACTIONS(1544),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [571] = {
    [sym_identifier] = ACTIONS(2383),
    [anon_sym_LPAREN] = ACTIONS(2385),
    [anon_sym_RPAREN] = ACTIONS(2385),
    [anon_sym_LBRACE] = ACTIONS(2385),
    [anon_sym_RBRACE] = ACTIONS(2385),
    [anon_sym_LBRACK] = ACTIONS(2385),
    [anon_sym_RBRACK] = ACTIONS(2385),
    [anon_sym_DOLLAR] = ACTIONS(2383),
    [anon_sym_u8] = ACTIONS(2383),
    [anon_sym_i8] = ACTIONS(2383),
    [anon_sym_u16] = ACTIONS(2383),
    [anon_sym_i16] = ACTIONS(2383),
    [anon_sym_u32] = ACTIONS(2383),
    [anon_sym_i32] = ACTIONS(2383),
    [anon_sym_u64] = ACTIONS(2383),
    [anon_sym_i64] = ACTIONS(2383),
    [anon_sym_u128] = ACTIONS(2383),
    [anon_sym_i128] = ACTIONS(2383),
    [anon_sym_isize] = ACTIONS(2383),
    [anon_sym_usize] = ACTIONS(2383),
    [anon_sym_f32] = ACTIONS(2383),
    [anon_sym_f64] = ACTIONS(2383),
    [anon_sym_bool] = ACTIONS(2383),
    [anon_sym_str] = ACTIONS(2383),
    [anon_sym_char] = ACTIONS(2383),
    [aux_sym__non_special_token_token1] = ACTIONS(2383),
    [anon_sym_SQUOTE] = ACTIONS(2383),
    [anon_sym_as] = ACTIONS(2383),
    [anon_sym_async] = ACTIONS(2383),
    [anon_sym_await] = ACTIONS(2383),
    [anon_sym_break] = ACTIONS(2383),
    [anon_sym_const] = ACTIONS(2383),
    [anon_sym_continue] = ACTIONS(2383),
    [anon_sym_default] = ACTIONS(2383),
    [anon_sym_enum] = ACTIONS(2383),
    [anon_sym_fn] = ACTIONS(2383),
    [anon_sym_for] = ACTIONS(2383),
    [anon_sym_if] = ACTIONS(2383),
    [anon_sym_impl] = ACTIONS(2383),
    [anon_sym_let] = ACTIONS(2383),
    [anon_sym_loop] = ACTIONS(2383),
    [anon_sym_match] = ACTIONS(2383),
    [anon_sym_mod] = ACTIONS(2383),
    [anon_sym_pub] = ACTIONS(2383),
    [anon_sym_return] = ACTIONS(2383),
    [anon_sym_static] = ACTIONS(2383),
    [anon_sym_struct] = ACTIONS(2383),
    [anon_sym_trait] = ACTIONS(2383),
    [anon_sym_type] = ACTIONS(2383),
    [anon_sym_union] = ACTIONS(2383),
    [anon_sym_unsafe] = ACTIONS(2383),
    [anon_sym_use] = ACTIONS(2383),
    [anon_sym_where] = ACTIONS(2383),
    [anon_sym_while] = ACTIONS(2383),
    [sym_mutable_specifier] = ACTIONS(2383),
    [sym_integer_literal] = ACTIONS(2385),
    [aux_sym_string_literal_token1] = ACTIONS(2385),
    [sym_char_literal] = ACTIONS(2385),
    [anon_sym_true] = ACTIONS(2383),
    [anon_sym_false] = ACTIONS(2383),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2383),
    [sym_super] = ACTIONS(2383),
    [sym_crate] = ACTIONS(2383),
    [sym_metavariable] = ACTIONS(2385),
    [sym_raw_string_literal] = ACTIONS(2385),
    [sym_float_literal] = ACTIONS(2385),
    [sym_block_comment] = ACTIONS(3),
  },
  [572] = {
    [sym_identifier] = ACTIONS(2387),
    [anon_sym_LPAREN] = ACTIONS(2389),
    [anon_sym_RPAREN] = ACTIONS(2389),
    [anon_sym_LBRACE] = ACTIONS(2389),
    [anon_sym_RBRACE] = ACTIONS(2389),
    [anon_sym_LBRACK] = ACTIONS(2389),
    [anon_sym_RBRACK] = ACTIONS(2389),
    [anon_sym_DOLLAR] = ACTIONS(2387),
    [anon_sym_u8] = ACTIONS(2387),
    [anon_sym_i8] = ACTIONS(2387),
    [anon_sym_u16] = ACTIONS(2387),
    [anon_sym_i16] = ACTIONS(2387),
    [anon_sym_u32] = ACTIONS(2387),
    [anon_sym_i32] = ACTIONS(2387),
    [anon_sym_u64] = ACTIONS(2387),
    [anon_sym_i64] = ACTIONS(2387),
    [anon_sym_u128] = ACTIONS(2387),
    [anon_sym_i128] = ACTIONS(2387),
    [anon_sym_isize] = ACTIONS(2387),
    [anon_sym_usize] = ACTIONS(2387),
    [anon_sym_f32] = ACTIONS(2387),
    [anon_sym_f64] = ACTIONS(2387),
    [anon_sym_bool] = ACTIONS(2387),
    [anon_sym_str] = ACTIONS(2387),
    [anon_sym_char] = ACTIONS(2387),
    [aux_sym__non_special_token_token1] = ACTIONS(2387),
    [anon_sym_SQUOTE] = ACTIONS(2387),
    [anon_sym_as] = ACTIONS(2387),
    [anon_sym_async] = ACTIONS(2387),
    [anon_sym_await] = ACTIONS(2387),
    [anon_sym_break] = ACTIONS(2387),
    [anon_sym_const] = ACTIONS(2387),
    [anon_sym_continue] = ACTIONS(2387),
    [anon_sym_default] = ACTIONS(2387),
    [anon_sym_enum] = ACTIONS(2387),
    [anon_sym_fn] = ACTIONS(2387),
    [anon_sym_for] = ACTIONS(2387),
    [anon_sym_if] = ACTIONS(2387),
    [anon_sym_impl] = ACTIONS(2387),
    [anon_sym_let] = ACTIONS(2387),
    [anon_sym_loop] = ACTIONS(2387),
    [anon_sym_match] = ACTIONS(2387),
    [anon_sym_mod] = ACTIONS(2387),
    [anon_sym_pub] = ACTIONS(2387),
    [anon_sym_return] = ACTIONS(2387),
    [anon_sym_static] = ACTIONS(2387),
    [anon_sym_struct] = ACTIONS(2387),
    [anon_sym_trait] = ACTIONS(2387),
    [anon_sym_type] = ACTIONS(2387),
    [anon_sym_union] = ACTIONS(2387),
    [anon_sym_unsafe] = ACTIONS(2387),
    [anon_sym_use] = ACTIONS(2387),
    [anon_sym_where] = ACTIONS(2387),
    [anon_sym_while] = ACTIONS(2387),
    [sym_mutable_specifier] = ACTIONS(2387),
    [sym_integer_literal] = ACTIONS(2389),
    [aux_sym_string_literal_token1] = ACTIONS(2389),
    [sym_char_literal] = ACTIONS(2389),
    [anon_sym_true] = ACTIONS(2387),
    [anon_sym_false] = ACTIONS(2387),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2387),
    [sym_super] = ACTIONS(2387),
    [sym_crate] = ACTIONS(2387),
    [sym_metavariable] = ACTIONS(2389),
    [sym_raw_string_literal] = ACTIONS(2389),
    [sym_float_literal] = ACTIONS(2389),
    [sym_block_comment] = ACTIONS(3),
  },
  [573] = {
    [sym_bracketed_type] = STATE(2489),
    [sym_generic_type] = STATE(2488),
    [sym_generic_type_with_turbofish] = STATE(2487),
    [sym_macro_invocation] = STATE(1422),
    [sym_scoped_identifier] = STATE(1337),
    [sym_scoped_type_identifier] = STATE(2013),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1804),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [sym_identifier] = ACTIONS(1524),
    [anon_sym_LPAREN] = ACTIONS(1526),
    [anon_sym_LBRACK] = ACTIONS(1530),
    [anon_sym_RBRACK] = ACTIONS(2391),
    [anon_sym_u8] = ACTIONS(1532),
    [anon_sym_i8] = ACTIONS(1532),
    [anon_sym_u16] = ACTIONS(1532),
    [anon_sym_i16] = ACTIONS(1532),
    [anon_sym_u32] = ACTIONS(1532),
    [anon_sym_i32] = ACTIONS(1532),
    [anon_sym_u64] = ACTIONS(1532),
    [anon_sym_i64] = ACTIONS(1532),
    [anon_sym_u128] = ACTIONS(1532),
    [anon_sym_i128] = ACTIONS(1532),
    [anon_sym_isize] = ACTIONS(1532),
    [anon_sym_usize] = ACTIONS(1532),
    [anon_sym_f32] = ACTIONS(1532),
    [anon_sym_f64] = ACTIONS(1532),
    [anon_sym_bool] = ACTIONS(1532),
    [anon_sym_str] = ACTIONS(1532),
    [anon_sym_char] = ACTIONS(1532),
    [anon_sym_const] = ACTIONS(1534),
    [anon_sym_default] = ACTIONS(1536),
    [anon_sym_union] = ACTIONS(1536),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(1538),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(1540),
    [sym_mutable_specifier] = ACTIONS(796),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1542),
    [sym_super] = ACTIONS(1542),
    [sym_crate] = ACTIONS(1542),
    [sym_metavariable] = ACTIONS(1544),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [574] = {
    [sym_identifier] = ACTIONS(2393),
    [anon_sym_LPAREN] = ACTIONS(2395),
    [anon_sym_RPAREN] = ACTIONS(2395),
    [anon_sym_LBRACE] = ACTIONS(2395),
    [anon_sym_RBRACE] = ACTIONS(2395),
    [anon_sym_LBRACK] = ACTIONS(2395),
    [anon_sym_RBRACK] = ACTIONS(2395),
    [anon_sym_DOLLAR] = ACTIONS(2393),
    [anon_sym_u8] = ACTIONS(2393),
    [anon_sym_i8] = ACTIONS(2393),
    [anon_sym_u16] = ACTIONS(2393),
    [anon_sym_i16] = ACTIONS(2393),
    [anon_sym_u32] = ACTIONS(2393),
    [anon_sym_i32] = ACTIONS(2393),
    [anon_sym_u64] = ACTIONS(2393),
    [anon_sym_i64] = ACTIONS(2393),
    [anon_sym_u128] = ACTIONS(2393),
    [anon_sym_i128] = ACTIONS(2393),
    [anon_sym_isize] = ACTIONS(2393),
    [anon_sym_usize] = ACTIONS(2393),
    [anon_sym_f32] = ACTIONS(2393),
    [anon_sym_f64] = ACTIONS(2393),
    [anon_sym_bool] = ACTIONS(2393),
    [anon_sym_str] = ACTIONS(2393),
    [anon_sym_char] = ACTIONS(2393),
    [aux_sym__non_special_token_token1] = ACTIONS(2393),
    [anon_sym_SQUOTE] = ACTIONS(2393),
    [anon_sym_as] = ACTIONS(2393),
    [anon_sym_async] = ACTIONS(2393),
    [anon_sym_await] = ACTIONS(2393),
    [anon_sym_break] = ACTIONS(2393),
    [anon_sym_const] = ACTIONS(2393),
    [anon_sym_continue] = ACTIONS(2393),
    [anon_sym_default] = ACTIONS(2393),
    [anon_sym_enum] = ACTIONS(2393),
    [anon_sym_fn] = ACTIONS(2393),
    [anon_sym_for] = ACTIONS(2393),
    [anon_sym_if] = ACTIONS(2393),
    [anon_sym_impl] = ACTIONS(2393),
    [anon_sym_let] = ACTIONS(2393),
    [anon_sym_loop] = ACTIONS(2393),
    [anon_sym_match] = ACTIONS(2393),
    [anon_sym_mod] = ACTIONS(2393),
    [anon_sym_pub] = ACTIONS(2393),
    [anon_sym_return] = ACTIONS(2393),
    [anon_sym_static] = ACTIONS(2393),
    [anon_sym_struct] = ACTIONS(2393),
    [anon_sym_trait] = ACTIONS(2393),
    [anon_sym_type] = ACTIONS(2393),
    [anon_sym_union] = ACTIONS(2393),
    [anon_sym_unsafe] = ACTIONS(2393),
    [anon_sym_use] = ACTIONS(2393),
    [anon_sym_where] = ACTIONS(2393),
    [anon_sym_while] = ACTIONS(2393),
    [sym_mutable_specifier] = ACTIONS(2393),
    [sym_integer_literal] = ACTIONS(2395),
    [aux_sym_string_literal_token1] = ACTIONS(2395),
    [sym_char_literal] = ACTIONS(2395),
    [anon_sym_true] = ACTIONS(2393),
    [anon_sym_false] = ACTIONS(2393),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2393),
    [sym_super] = ACTIONS(2393),
    [sym_crate] = ACTIONS(2393),
    [sym_metavariable] = ACTIONS(2395),
    [sym_raw_string_literal] = ACTIONS(2395),
    [sym_float_literal] = ACTIONS(2395),
    [sym_block_comment] = ACTIONS(3),
  },
  [575] = {
    [sym_identifier] = ACTIONS(2397),
    [anon_sym_LPAREN] = ACTIONS(2399),
    [anon_sym_RPAREN] = ACTIONS(2399),
    [anon_sym_LBRACE] = ACTIONS(2399),
    [anon_sym_RBRACE] = ACTIONS(2399),
    [anon_sym_LBRACK] = ACTIONS(2399),
    [anon_sym_RBRACK] = ACTIONS(2399),
    [anon_sym_DOLLAR] = ACTIONS(2397),
    [anon_sym_u8] = ACTIONS(2397),
    [anon_sym_i8] = ACTIONS(2397),
    [anon_sym_u16] = ACTIONS(2397),
    [anon_sym_i16] = ACTIONS(2397),
    [anon_sym_u32] = ACTIONS(2397),
    [anon_sym_i32] = ACTIONS(2397),
    [anon_sym_u64] = ACTIONS(2397),
    [anon_sym_i64] = ACTIONS(2397),
    [anon_sym_u128] = ACTIONS(2397),
    [anon_sym_i128] = ACTIONS(2397),
    [anon_sym_isize] = ACTIONS(2397),
    [anon_sym_usize] = ACTIONS(2397),
    [anon_sym_f32] = ACTIONS(2397),
    [anon_sym_f64] = ACTIONS(2397),
    [anon_sym_bool] = ACTIONS(2397),
    [anon_sym_str] = ACTIONS(2397),
    [anon_sym_char] = ACTIONS(2397),
    [aux_sym__non_special_token_token1] = ACTIONS(2397),
    [anon_sym_SQUOTE] = ACTIONS(2397),
    [anon_sym_as] = ACTIONS(2397),
    [anon_sym_async] = ACTIONS(2397),
    [anon_sym_await] = ACTIONS(2397),
    [anon_sym_break] = ACTIONS(2397),
    [anon_sym_const] = ACTIONS(2397),
    [anon_sym_continue] = ACTIONS(2397),
    [anon_sym_default] = ACTIONS(2397),
    [anon_sym_enum] = ACTIONS(2397),
    [anon_sym_fn] = ACTIONS(2397),
    [anon_sym_for] = ACTIONS(2397),
    [anon_sym_if] = ACTIONS(2397),
    [anon_sym_impl] = ACTIONS(2397),
    [anon_sym_let] = ACTIONS(2397),
    [anon_sym_loop] = ACTIONS(2397),
    [anon_sym_match] = ACTIONS(2397),
    [anon_sym_mod] = ACTIONS(2397),
    [anon_sym_pub] = ACTIONS(2397),
    [anon_sym_return] = ACTIONS(2397),
    [anon_sym_static] = ACTIONS(2397),
    [anon_sym_struct] = ACTIONS(2397),
    [anon_sym_trait] = ACTIONS(2397),
    [anon_sym_type] = ACTIONS(2397),
    [anon_sym_union] = ACTIONS(2397),
    [anon_sym_unsafe] = ACTIONS(2397),
    [anon_sym_use] = ACTIONS(2397),
    [anon_sym_where] = ACTIONS(2397),
    [anon_sym_while] = ACTIONS(2397),
    [sym_mutable_specifier] = ACTIONS(2397),
    [sym_integer_literal] = ACTIONS(2399),
    [aux_sym_string_literal_token1] = ACTIONS(2399),
    [sym_char_literal] = ACTIONS(2399),
    [anon_sym_true] = ACTIONS(2397),
    [anon_sym_false] = ACTIONS(2397),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2397),
    [sym_super] = ACTIONS(2397),
    [sym_crate] = ACTIONS(2397),
    [sym_metavariable] = ACTIONS(2399),
    [sym_raw_string_literal] = ACTIONS(2399),
    [sym_float_literal] = ACTIONS(2399),
    [sym_block_comment] = ACTIONS(3),
  },
  [576] = {
    [sym_bracketed_type] = STATE(2489),
    [sym_generic_type] = STATE(2488),
    [sym_generic_type_with_turbofish] = STATE(2487),
    [sym_macro_invocation] = STATE(1422),
    [sym_scoped_identifier] = STATE(1337),
    [sym_scoped_type_identifier] = STATE(2013),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1804),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [sym_identifier] = ACTIONS(1524),
    [anon_sym_LPAREN] = ACTIONS(1526),
    [anon_sym_RPAREN] = ACTIONS(2401),
    [anon_sym_LBRACK] = ACTIONS(1530),
    [anon_sym_u8] = ACTIONS(1532),
    [anon_sym_i8] = ACTIONS(1532),
    [anon_sym_u16] = ACTIONS(1532),
    [anon_sym_i16] = ACTIONS(1532),
    [anon_sym_u32] = ACTIONS(1532),
    [anon_sym_i32] = ACTIONS(1532),
    [anon_sym_u64] = ACTIONS(1532),
    [anon_sym_i64] = ACTIONS(1532),
    [anon_sym_u128] = ACTIONS(1532),
    [anon_sym_i128] = ACTIONS(1532),
    [anon_sym_isize] = ACTIONS(1532),
    [anon_sym_usize] = ACTIONS(1532),
    [anon_sym_f32] = ACTIONS(1532),
    [anon_sym_f64] = ACTIONS(1532),
    [anon_sym_bool] = ACTIONS(1532),
    [anon_sym_str] = ACTIONS(1532),
    [anon_sym_char] = ACTIONS(1532),
    [anon_sym_const] = ACTIONS(1534),
    [anon_sym_default] = ACTIONS(1536),
    [anon_sym_union] = ACTIONS(1536),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(1538),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(1540),
    [sym_mutable_specifier] = ACTIONS(796),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1542),
    [sym_super] = ACTIONS(1542),
    [sym_crate] = ACTIONS(1542),
    [sym_metavariable] = ACTIONS(1544),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [577] = {
    [sym_identifier] = ACTIONS(2403),
    [anon_sym_LPAREN] = ACTIONS(2405),
    [anon_sym_RPAREN] = ACTIONS(2405),
    [anon_sym_LBRACE] = ACTIONS(2405),
    [anon_sym_RBRACE] = ACTIONS(2405),
    [anon_sym_LBRACK] = ACTIONS(2405),
    [anon_sym_RBRACK] = ACTIONS(2405),
    [anon_sym_DOLLAR] = ACTIONS(2403),
    [anon_sym_u8] = ACTIONS(2403),
    [anon_sym_i8] = ACTIONS(2403),
    [anon_sym_u16] = ACTIONS(2403),
    [anon_sym_i16] = ACTIONS(2403),
    [anon_sym_u32] = ACTIONS(2403),
    [anon_sym_i32] = ACTIONS(2403),
    [anon_sym_u64] = ACTIONS(2403),
    [anon_sym_i64] = ACTIONS(2403),
    [anon_sym_u128] = ACTIONS(2403),
    [anon_sym_i128] = ACTIONS(2403),
    [anon_sym_isize] = ACTIONS(2403),
    [anon_sym_usize] = ACTIONS(2403),
    [anon_sym_f32] = ACTIONS(2403),
    [anon_sym_f64] = ACTIONS(2403),
    [anon_sym_bool] = ACTIONS(2403),
    [anon_sym_str] = ACTIONS(2403),
    [anon_sym_char] = ACTIONS(2403),
    [aux_sym__non_special_token_token1] = ACTIONS(2403),
    [anon_sym_SQUOTE] = ACTIONS(2403),
    [anon_sym_as] = ACTIONS(2403),
    [anon_sym_async] = ACTIONS(2403),
    [anon_sym_await] = ACTIONS(2403),
    [anon_sym_break] = ACTIONS(2403),
    [anon_sym_const] = ACTIONS(2403),
    [anon_sym_continue] = ACTIONS(2403),
    [anon_sym_default] = ACTIONS(2403),
    [anon_sym_enum] = ACTIONS(2403),
    [anon_sym_fn] = ACTIONS(2403),
    [anon_sym_for] = ACTIONS(2403),
    [anon_sym_if] = ACTIONS(2403),
    [anon_sym_impl] = ACTIONS(2403),
    [anon_sym_let] = ACTIONS(2403),
    [anon_sym_loop] = ACTIONS(2403),
    [anon_sym_match] = ACTIONS(2403),
    [anon_sym_mod] = ACTIONS(2403),
    [anon_sym_pub] = ACTIONS(2403),
    [anon_sym_return] = ACTIONS(2403),
    [anon_sym_static] = ACTIONS(2403),
    [anon_sym_struct] = ACTIONS(2403),
    [anon_sym_trait] = ACTIONS(2403),
    [anon_sym_type] = ACTIONS(2403),
    [anon_sym_union] = ACTIONS(2403),
    [anon_sym_unsafe] = ACTIONS(2403),
    [anon_sym_use] = ACTIONS(2403),
    [anon_sym_where] = ACTIONS(2403),
    [anon_sym_while] = ACTIONS(2403),
    [sym_mutable_specifier] = ACTIONS(2403),
    [sym_integer_literal] = ACTIONS(2405),
    [aux_sym_string_literal_token1] = ACTIONS(2405),
    [sym_char_literal] = ACTIONS(2405),
    [anon_sym_true] = ACTIONS(2403),
    [anon_sym_false] = ACTIONS(2403),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2403),
    [sym_super] = ACTIONS(2403),
    [sym_crate] = ACTIONS(2403),
    [sym_metavariable] = ACTIONS(2405),
    [sym_raw_string_literal] = ACTIONS(2405),
    [sym_float_literal] = ACTIONS(2405),
    [sym_block_comment] = ACTIONS(3),
  },
  [578] = {
    [sym_identifier] = ACTIONS(2407),
    [anon_sym_LPAREN] = ACTIONS(2409),
    [anon_sym_RPAREN] = ACTIONS(2409),
    [anon_sym_LBRACE] = ACTIONS(2409),
    [anon_sym_RBRACE] = ACTIONS(2409),
    [anon_sym_LBRACK] = ACTIONS(2409),
    [anon_sym_RBRACK] = ACTIONS(2409),
    [anon_sym_DOLLAR] = ACTIONS(2407),
    [anon_sym_u8] = ACTIONS(2407),
    [anon_sym_i8] = ACTIONS(2407),
    [anon_sym_u16] = ACTIONS(2407),
    [anon_sym_i16] = ACTIONS(2407),
    [anon_sym_u32] = ACTIONS(2407),
    [anon_sym_i32] = ACTIONS(2407),
    [anon_sym_u64] = ACTIONS(2407),
    [anon_sym_i64] = ACTIONS(2407),
    [anon_sym_u128] = ACTIONS(2407),
    [anon_sym_i128] = ACTIONS(2407),
    [anon_sym_isize] = ACTIONS(2407),
    [anon_sym_usize] = ACTIONS(2407),
    [anon_sym_f32] = ACTIONS(2407),
    [anon_sym_f64] = ACTIONS(2407),
    [anon_sym_bool] = ACTIONS(2407),
    [anon_sym_str] = ACTIONS(2407),
    [anon_sym_char] = ACTIONS(2407),
    [aux_sym__non_special_token_token1] = ACTIONS(2407),
    [anon_sym_SQUOTE] = ACTIONS(2407),
    [anon_sym_as] = ACTIONS(2407),
    [anon_sym_async] = ACTIONS(2407),
    [anon_sym_await] = ACTIONS(2407),
    [anon_sym_break] = ACTIONS(2407),
    [anon_sym_const] = ACTIONS(2407),
    [anon_sym_continue] = ACTIONS(2407),
    [anon_sym_default] = ACTIONS(2407),
    [anon_sym_enum] = ACTIONS(2407),
    [anon_sym_fn] = ACTIONS(2407),
    [anon_sym_for] = ACTIONS(2407),
    [anon_sym_if] = ACTIONS(2407),
    [anon_sym_impl] = ACTIONS(2407),
    [anon_sym_let] = ACTIONS(2407),
    [anon_sym_loop] = ACTIONS(2407),
    [anon_sym_match] = ACTIONS(2407),
    [anon_sym_mod] = ACTIONS(2407),
    [anon_sym_pub] = ACTIONS(2407),
    [anon_sym_return] = ACTIONS(2407),
    [anon_sym_static] = ACTIONS(2407),
    [anon_sym_struct] = ACTIONS(2407),
    [anon_sym_trait] = ACTIONS(2407),
    [anon_sym_type] = ACTIONS(2407),
    [anon_sym_union] = ACTIONS(2407),
    [anon_sym_unsafe] = ACTIONS(2407),
    [anon_sym_use] = ACTIONS(2407),
    [anon_sym_where] = ACTIONS(2407),
    [anon_sym_while] = ACTIONS(2407),
    [sym_mutable_specifier] = ACTIONS(2407),
    [sym_integer_literal] = ACTIONS(2409),
    [aux_sym_string_literal_token1] = ACTIONS(2409),
    [sym_char_literal] = ACTIONS(2409),
    [anon_sym_true] = ACTIONS(2407),
    [anon_sym_false] = ACTIONS(2407),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2407),
    [sym_super] = ACTIONS(2407),
    [sym_crate] = ACTIONS(2407),
    [sym_metavariable] = ACTIONS(2409),
    [sym_raw_string_literal] = ACTIONS(2409),
    [sym_float_literal] = ACTIONS(2409),
    [sym_block_comment] = ACTIONS(3),
  },
  [579] = {
    [sym_identifier] = ACTIONS(2411),
    [anon_sym_LPAREN] = ACTIONS(2413),
    [anon_sym_RPAREN] = ACTIONS(2413),
    [anon_sym_LBRACE] = ACTIONS(2413),
    [anon_sym_RBRACE] = ACTIONS(2413),
    [anon_sym_LBRACK] = ACTIONS(2413),
    [anon_sym_RBRACK] = ACTIONS(2413),
    [anon_sym_DOLLAR] = ACTIONS(2411),
    [anon_sym_u8] = ACTIONS(2411),
    [anon_sym_i8] = ACTIONS(2411),
    [anon_sym_u16] = ACTIONS(2411),
    [anon_sym_i16] = ACTIONS(2411),
    [anon_sym_u32] = ACTIONS(2411),
    [anon_sym_i32] = ACTIONS(2411),
    [anon_sym_u64] = ACTIONS(2411),
    [anon_sym_i64] = ACTIONS(2411),
    [anon_sym_u128] = ACTIONS(2411),
    [anon_sym_i128] = ACTIONS(2411),
    [anon_sym_isize] = ACTIONS(2411),
    [anon_sym_usize] = ACTIONS(2411),
    [anon_sym_f32] = ACTIONS(2411),
    [anon_sym_f64] = ACTIONS(2411),
    [anon_sym_bool] = ACTIONS(2411),
    [anon_sym_str] = ACTIONS(2411),
    [anon_sym_char] = ACTIONS(2411),
    [aux_sym__non_special_token_token1] = ACTIONS(2411),
    [anon_sym_SQUOTE] = ACTIONS(2411),
    [anon_sym_as] = ACTIONS(2411),
    [anon_sym_async] = ACTIONS(2411),
    [anon_sym_await] = ACTIONS(2411),
    [anon_sym_break] = ACTIONS(2411),
    [anon_sym_const] = ACTIONS(2411),
    [anon_sym_continue] = ACTIONS(2411),
    [anon_sym_default] = ACTIONS(2411),
    [anon_sym_enum] = ACTIONS(2411),
    [anon_sym_fn] = ACTIONS(2411),
    [anon_sym_for] = ACTIONS(2411),
    [anon_sym_if] = ACTIONS(2411),
    [anon_sym_impl] = ACTIONS(2411),
    [anon_sym_let] = ACTIONS(2411),
    [anon_sym_loop] = ACTIONS(2411),
    [anon_sym_match] = ACTIONS(2411),
    [anon_sym_mod] = ACTIONS(2411),
    [anon_sym_pub] = ACTIONS(2411),
    [anon_sym_return] = ACTIONS(2411),
    [anon_sym_static] = ACTIONS(2411),
    [anon_sym_struct] = ACTIONS(2411),
    [anon_sym_trait] = ACTIONS(2411),
    [anon_sym_type] = ACTIONS(2411),
    [anon_sym_union] = ACTIONS(2411),
    [anon_sym_unsafe] = ACTIONS(2411),
    [anon_sym_use] = ACTIONS(2411),
    [anon_sym_where] = ACTIONS(2411),
    [anon_sym_while] = ACTIONS(2411),
    [sym_mutable_specifier] = ACTIONS(2411),
    [sym_integer_literal] = ACTIONS(2413),
    [aux_sym_string_literal_token1] = ACTIONS(2413),
    [sym_char_literal] = ACTIONS(2413),
    [anon_sym_true] = ACTIONS(2411),
    [anon_sym_false] = ACTIONS(2411),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2411),
    [sym_super] = ACTIONS(2411),
    [sym_crate] = ACTIONS(2411),
    [sym_metavariable] = ACTIONS(2413),
    [sym_raw_string_literal] = ACTIONS(2413),
    [sym_float_literal] = ACTIONS(2413),
    [sym_block_comment] = ACTIONS(3),
  },
  [580] = {
    [sym_bracketed_type] = STATE(2489),
    [sym_generic_type] = STATE(2488),
    [sym_generic_type_with_turbofish] = STATE(2487),
    [sym_macro_invocation] = STATE(1422),
    [sym_scoped_identifier] = STATE(1337),
    [sym_scoped_type_identifier] = STATE(2013),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1804),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [sym_identifier] = ACTIONS(1524),
    [anon_sym_LPAREN] = ACTIONS(1526),
    [anon_sym_RPAREN] = ACTIONS(2415),
    [anon_sym_LBRACK] = ACTIONS(1530),
    [anon_sym_u8] = ACTIONS(1532),
    [anon_sym_i8] = ACTIONS(1532),
    [anon_sym_u16] = ACTIONS(1532),
    [anon_sym_i16] = ACTIONS(1532),
    [anon_sym_u32] = ACTIONS(1532),
    [anon_sym_i32] = ACTIONS(1532),
    [anon_sym_u64] = ACTIONS(1532),
    [anon_sym_i64] = ACTIONS(1532),
    [anon_sym_u128] = ACTIONS(1532),
    [anon_sym_i128] = ACTIONS(1532),
    [anon_sym_isize] = ACTIONS(1532),
    [anon_sym_usize] = ACTIONS(1532),
    [anon_sym_f32] = ACTIONS(1532),
    [anon_sym_f64] = ACTIONS(1532),
    [anon_sym_bool] = ACTIONS(1532),
    [anon_sym_str] = ACTIONS(1532),
    [anon_sym_char] = ACTIONS(1532),
    [anon_sym_const] = ACTIONS(1534),
    [anon_sym_default] = ACTIONS(1536),
    [anon_sym_union] = ACTIONS(1536),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(1538),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(1540),
    [sym_mutable_specifier] = ACTIONS(796),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1542),
    [sym_super] = ACTIONS(1542),
    [sym_crate] = ACTIONS(1542),
    [sym_metavariable] = ACTIONS(1544),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [581] = {
    [sym_bracketed_type] = STATE(2489),
    [sym_generic_type] = STATE(2488),
    [sym_generic_type_with_turbofish] = STATE(2487),
    [sym_macro_invocation] = STATE(1422),
    [sym_scoped_identifier] = STATE(1337),
    [sym_scoped_type_identifier] = STATE(2013),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1804),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [sym_identifier] = ACTIONS(1524),
    [anon_sym_LPAREN] = ACTIONS(1526),
    [anon_sym_LBRACK] = ACTIONS(1530),
    [anon_sym_RBRACK] = ACTIONS(2417),
    [anon_sym_u8] = ACTIONS(1532),
    [anon_sym_i8] = ACTIONS(1532),
    [anon_sym_u16] = ACTIONS(1532),
    [anon_sym_i16] = ACTIONS(1532),
    [anon_sym_u32] = ACTIONS(1532),
    [anon_sym_i32] = ACTIONS(1532),
    [anon_sym_u64] = ACTIONS(1532),
    [anon_sym_i64] = ACTIONS(1532),
    [anon_sym_u128] = ACTIONS(1532),
    [anon_sym_i128] = ACTIONS(1532),
    [anon_sym_isize] = ACTIONS(1532),
    [anon_sym_usize] = ACTIONS(1532),
    [anon_sym_f32] = ACTIONS(1532),
    [anon_sym_f64] = ACTIONS(1532),
    [anon_sym_bool] = ACTIONS(1532),
    [anon_sym_str] = ACTIONS(1532),
    [anon_sym_char] = ACTIONS(1532),
    [anon_sym_const] = ACTIONS(1534),
    [anon_sym_default] = ACTIONS(1536),
    [anon_sym_union] = ACTIONS(1536),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(1538),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(1540),
    [sym_mutable_specifier] = ACTIONS(796),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1542),
    [sym_super] = ACTIONS(1542),
    [sym_crate] = ACTIONS(1542),
    [sym_metavariable] = ACTIONS(1544),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [582] = {
    [sym_identifier] = ACTIONS(2419),
    [anon_sym_LPAREN] = ACTIONS(2421),
    [anon_sym_RPAREN] = ACTIONS(2421),
    [anon_sym_LBRACE] = ACTIONS(2421),
    [anon_sym_RBRACE] = ACTIONS(2421),
    [anon_sym_LBRACK] = ACTIONS(2421),
    [anon_sym_RBRACK] = ACTIONS(2421),
    [anon_sym_DOLLAR] = ACTIONS(2419),
    [anon_sym_u8] = ACTIONS(2419),
    [anon_sym_i8] = ACTIONS(2419),
    [anon_sym_u16] = ACTIONS(2419),
    [anon_sym_i16] = ACTIONS(2419),
    [anon_sym_u32] = ACTIONS(2419),
    [anon_sym_i32] = ACTIONS(2419),
    [anon_sym_u64] = ACTIONS(2419),
    [anon_sym_i64] = ACTIONS(2419),
    [anon_sym_u128] = ACTIONS(2419),
    [anon_sym_i128] = ACTIONS(2419),
    [anon_sym_isize] = ACTIONS(2419),
    [anon_sym_usize] = ACTIONS(2419),
    [anon_sym_f32] = ACTIONS(2419),
    [anon_sym_f64] = ACTIONS(2419),
    [anon_sym_bool] = ACTIONS(2419),
    [anon_sym_str] = ACTIONS(2419),
    [anon_sym_char] = ACTIONS(2419),
    [aux_sym__non_special_token_token1] = ACTIONS(2419),
    [anon_sym_SQUOTE] = ACTIONS(2419),
    [anon_sym_as] = ACTIONS(2419),
    [anon_sym_async] = ACTIONS(2419),
    [anon_sym_await] = ACTIONS(2419),
    [anon_sym_break] = ACTIONS(2419),
    [anon_sym_const] = ACTIONS(2419),
    [anon_sym_continue] = ACTIONS(2419),
    [anon_sym_default] = ACTIONS(2419),
    [anon_sym_enum] = ACTIONS(2419),
    [anon_sym_fn] = ACTIONS(2419),
    [anon_sym_for] = ACTIONS(2419),
    [anon_sym_if] = ACTIONS(2419),
    [anon_sym_impl] = ACTIONS(2419),
    [anon_sym_let] = ACTIONS(2419),
    [anon_sym_loop] = ACTIONS(2419),
    [anon_sym_match] = ACTIONS(2419),
    [anon_sym_mod] = ACTIONS(2419),
    [anon_sym_pub] = ACTIONS(2419),
    [anon_sym_return] = ACTIONS(2419),
    [anon_sym_static] = ACTIONS(2419),
    [anon_sym_struct] = ACTIONS(2419),
    [anon_sym_trait] = ACTIONS(2419),
    [anon_sym_type] = ACTIONS(2419),
    [anon_sym_union] = ACTIONS(2419),
    [anon_sym_unsafe] = ACTIONS(2419),
    [anon_sym_use] = ACTIONS(2419),
    [anon_sym_where] = ACTIONS(2419),
    [anon_sym_while] = ACTIONS(2419),
    [sym_mutable_specifier] = ACTIONS(2419),
    [sym_integer_literal] = ACTIONS(2421),
    [aux_sym_string_literal_token1] = ACTIONS(2421),
    [sym_char_literal] = ACTIONS(2421),
    [anon_sym_true] = ACTIONS(2419),
    [anon_sym_false] = ACTIONS(2419),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2419),
    [sym_super] = ACTIONS(2419),
    [sym_crate] = ACTIONS(2419),
    [sym_metavariable] = ACTIONS(2421),
    [sym_raw_string_literal] = ACTIONS(2421),
    [sym_float_literal] = ACTIONS(2421),
    [sym_block_comment] = ACTIONS(3),
  },
  [583] = {
    [sym_identifier] = ACTIONS(2423),
    [anon_sym_LPAREN] = ACTIONS(2425),
    [anon_sym_RPAREN] = ACTIONS(2425),
    [anon_sym_LBRACE] = ACTIONS(2425),
    [anon_sym_RBRACE] = ACTIONS(2425),
    [anon_sym_LBRACK] = ACTIONS(2425),
    [anon_sym_RBRACK] = ACTIONS(2425),
    [anon_sym_DOLLAR] = ACTIONS(2423),
    [anon_sym_u8] = ACTIONS(2423),
    [anon_sym_i8] = ACTIONS(2423),
    [anon_sym_u16] = ACTIONS(2423),
    [anon_sym_i16] = ACTIONS(2423),
    [anon_sym_u32] = ACTIONS(2423),
    [anon_sym_i32] = ACTIONS(2423),
    [anon_sym_u64] = ACTIONS(2423),
    [anon_sym_i64] = ACTIONS(2423),
    [anon_sym_u128] = ACTIONS(2423),
    [anon_sym_i128] = ACTIONS(2423),
    [anon_sym_isize] = ACTIONS(2423),
    [anon_sym_usize] = ACTIONS(2423),
    [anon_sym_f32] = ACTIONS(2423),
    [anon_sym_f64] = ACTIONS(2423),
    [anon_sym_bool] = ACTIONS(2423),
    [anon_sym_str] = ACTIONS(2423),
    [anon_sym_char] = ACTIONS(2423),
    [aux_sym__non_special_token_token1] = ACTIONS(2423),
    [anon_sym_SQUOTE] = ACTIONS(2423),
    [anon_sym_as] = ACTIONS(2423),
    [anon_sym_async] = ACTIONS(2423),
    [anon_sym_await] = ACTIONS(2423),
    [anon_sym_break] = ACTIONS(2423),
    [anon_sym_const] = ACTIONS(2423),
    [anon_sym_continue] = ACTIONS(2423),
    [anon_sym_default] = ACTIONS(2423),
    [anon_sym_enum] = ACTIONS(2423),
    [anon_sym_fn] = ACTIONS(2423),
    [anon_sym_for] = ACTIONS(2423),
    [anon_sym_if] = ACTIONS(2423),
    [anon_sym_impl] = ACTIONS(2423),
    [anon_sym_let] = ACTIONS(2423),
    [anon_sym_loop] = ACTIONS(2423),
    [anon_sym_match] = ACTIONS(2423),
    [anon_sym_mod] = ACTIONS(2423),
    [anon_sym_pub] = ACTIONS(2423),
    [anon_sym_return] = ACTIONS(2423),
    [anon_sym_static] = ACTIONS(2423),
    [anon_sym_struct] = ACTIONS(2423),
    [anon_sym_trait] = ACTIONS(2423),
    [anon_sym_type] = ACTIONS(2423),
    [anon_sym_union] = ACTIONS(2423),
    [anon_sym_unsafe] = ACTIONS(2423),
    [anon_sym_use] = ACTIONS(2423),
    [anon_sym_where] = ACTIONS(2423),
    [anon_sym_while] = ACTIONS(2423),
    [sym_mutable_specifier] = ACTIONS(2423),
    [sym_integer_literal] = ACTIONS(2425),
    [aux_sym_string_literal_token1] = ACTIONS(2425),
    [sym_char_literal] = ACTIONS(2425),
    [anon_sym_true] = ACTIONS(2423),
    [anon_sym_false] = ACTIONS(2423),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2423),
    [sym_super] = ACTIONS(2423),
    [sym_crate] = ACTIONS(2423),
    [sym_metavariable] = ACTIONS(2425),
    [sym_raw_string_literal] = ACTIONS(2425),
    [sym_float_literal] = ACTIONS(2425),
    [sym_block_comment] = ACTIONS(3),
  },
  [584] = {
    [sym_identifier] = ACTIONS(2427),
    [anon_sym_LPAREN] = ACTIONS(2429),
    [anon_sym_RPAREN] = ACTIONS(2429),
    [anon_sym_LBRACE] = ACTIONS(2429),
    [anon_sym_RBRACE] = ACTIONS(2429),
    [anon_sym_LBRACK] = ACTIONS(2429),
    [anon_sym_RBRACK] = ACTIONS(2429),
    [anon_sym_DOLLAR] = ACTIONS(2427),
    [anon_sym_u8] = ACTIONS(2427),
    [anon_sym_i8] = ACTIONS(2427),
    [anon_sym_u16] = ACTIONS(2427),
    [anon_sym_i16] = ACTIONS(2427),
    [anon_sym_u32] = ACTIONS(2427),
    [anon_sym_i32] = ACTIONS(2427),
    [anon_sym_u64] = ACTIONS(2427),
    [anon_sym_i64] = ACTIONS(2427),
    [anon_sym_u128] = ACTIONS(2427),
    [anon_sym_i128] = ACTIONS(2427),
    [anon_sym_isize] = ACTIONS(2427),
    [anon_sym_usize] = ACTIONS(2427),
    [anon_sym_f32] = ACTIONS(2427),
    [anon_sym_f64] = ACTIONS(2427),
    [anon_sym_bool] = ACTIONS(2427),
    [anon_sym_str] = ACTIONS(2427),
    [anon_sym_char] = ACTIONS(2427),
    [aux_sym__non_special_token_token1] = ACTIONS(2427),
    [anon_sym_SQUOTE] = ACTIONS(2427),
    [anon_sym_as] = ACTIONS(2427),
    [anon_sym_async] = ACTIONS(2427),
    [anon_sym_await] = ACTIONS(2427),
    [anon_sym_break] = ACTIONS(2427),
    [anon_sym_const] = ACTIONS(2427),
    [anon_sym_continue] = ACTIONS(2427),
    [anon_sym_default] = ACTIONS(2427),
    [anon_sym_enum] = ACTIONS(2427),
    [anon_sym_fn] = ACTIONS(2427),
    [anon_sym_for] = ACTIONS(2427),
    [anon_sym_if] = ACTIONS(2427),
    [anon_sym_impl] = ACTIONS(2427),
    [anon_sym_let] = ACTIONS(2427),
    [anon_sym_loop] = ACTIONS(2427),
    [anon_sym_match] = ACTIONS(2427),
    [anon_sym_mod] = ACTIONS(2427),
    [anon_sym_pub] = ACTIONS(2427),
    [anon_sym_return] = ACTIONS(2427),
    [anon_sym_static] = ACTIONS(2427),
    [anon_sym_struct] = ACTIONS(2427),
    [anon_sym_trait] = ACTIONS(2427),
    [anon_sym_type] = ACTIONS(2427),
    [anon_sym_union] = ACTIONS(2427),
    [anon_sym_unsafe] = ACTIONS(2427),
    [anon_sym_use] = ACTIONS(2427),
    [anon_sym_where] = ACTIONS(2427),
    [anon_sym_while] = ACTIONS(2427),
    [sym_mutable_specifier] = ACTIONS(2427),
    [sym_integer_literal] = ACTIONS(2429),
    [aux_sym_string_literal_token1] = ACTIONS(2429),
    [sym_char_literal] = ACTIONS(2429),
    [anon_sym_true] = ACTIONS(2427),
    [anon_sym_false] = ACTIONS(2427),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2427),
    [sym_super] = ACTIONS(2427),
    [sym_crate] = ACTIONS(2427),
    [sym_metavariable] = ACTIONS(2429),
    [sym_raw_string_literal] = ACTIONS(2429),
    [sym_float_literal] = ACTIONS(2429),
    [sym_block_comment] = ACTIONS(3),
  },
  [585] = {
    [sym_identifier] = ACTIONS(2431),
    [anon_sym_LPAREN] = ACTIONS(2433),
    [anon_sym_RPAREN] = ACTIONS(2433),
    [anon_sym_LBRACE] = ACTIONS(2433),
    [anon_sym_RBRACE] = ACTIONS(2433),
    [anon_sym_LBRACK] = ACTIONS(2433),
    [anon_sym_RBRACK] = ACTIONS(2433),
    [anon_sym_DOLLAR] = ACTIONS(2431),
    [anon_sym_u8] = ACTIONS(2431),
    [anon_sym_i8] = ACTIONS(2431),
    [anon_sym_u16] = ACTIONS(2431),
    [anon_sym_i16] = ACTIONS(2431),
    [anon_sym_u32] = ACTIONS(2431),
    [anon_sym_i32] = ACTIONS(2431),
    [anon_sym_u64] = ACTIONS(2431),
    [anon_sym_i64] = ACTIONS(2431),
    [anon_sym_u128] = ACTIONS(2431),
    [anon_sym_i128] = ACTIONS(2431),
    [anon_sym_isize] = ACTIONS(2431),
    [anon_sym_usize] = ACTIONS(2431),
    [anon_sym_f32] = ACTIONS(2431),
    [anon_sym_f64] = ACTIONS(2431),
    [anon_sym_bool] = ACTIONS(2431),
    [anon_sym_str] = ACTIONS(2431),
    [anon_sym_char] = ACTIONS(2431),
    [aux_sym__non_special_token_token1] = ACTIONS(2431),
    [anon_sym_SQUOTE] = ACTIONS(2431),
    [anon_sym_as] = ACTIONS(2431),
    [anon_sym_async] = ACTIONS(2431),
    [anon_sym_await] = ACTIONS(2431),
    [anon_sym_break] = ACTIONS(2431),
    [anon_sym_const] = ACTIONS(2431),
    [anon_sym_continue] = ACTIONS(2431),
    [anon_sym_default] = ACTIONS(2431),
    [anon_sym_enum] = ACTIONS(2431),
    [anon_sym_fn] = ACTIONS(2431),
    [anon_sym_for] = ACTIONS(2431),
    [anon_sym_if] = ACTIONS(2431),
    [anon_sym_impl] = ACTIONS(2431),
    [anon_sym_let] = ACTIONS(2431),
    [anon_sym_loop] = ACTIONS(2431),
    [anon_sym_match] = ACTIONS(2431),
    [anon_sym_mod] = ACTIONS(2431),
    [anon_sym_pub] = ACTIONS(2431),
    [anon_sym_return] = ACTIONS(2431),
    [anon_sym_static] = ACTIONS(2431),
    [anon_sym_struct] = ACTIONS(2431),
    [anon_sym_trait] = ACTIONS(2431),
    [anon_sym_type] = ACTIONS(2431),
    [anon_sym_union] = ACTIONS(2431),
    [anon_sym_unsafe] = ACTIONS(2431),
    [anon_sym_use] = ACTIONS(2431),
    [anon_sym_where] = ACTIONS(2431),
    [anon_sym_while] = ACTIONS(2431),
    [sym_mutable_specifier] = ACTIONS(2431),
    [sym_integer_literal] = ACTIONS(2433),
    [aux_sym_string_literal_token1] = ACTIONS(2433),
    [sym_char_literal] = ACTIONS(2433),
    [anon_sym_true] = ACTIONS(2431),
    [anon_sym_false] = ACTIONS(2431),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2431),
    [sym_super] = ACTIONS(2431),
    [sym_crate] = ACTIONS(2431),
    [sym_metavariable] = ACTIONS(2433),
    [sym_raw_string_literal] = ACTIONS(2433),
    [sym_float_literal] = ACTIONS(2433),
    [sym_block_comment] = ACTIONS(3),
  },
  [586] = {
    [sym_bracketed_type] = STATE(2489),
    [sym_generic_type] = STATE(2488),
    [sym_generic_type_with_turbofish] = STATE(2487),
    [sym_macro_invocation] = STATE(1422),
    [sym_scoped_identifier] = STATE(1337),
    [sym_scoped_type_identifier] = STATE(2013),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1451),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [sym_identifier] = ACTIONS(1524),
    [anon_sym_LPAREN] = ACTIONS(1526),
    [anon_sym_LBRACK] = ACTIONS(1530),
    [anon_sym_u8] = ACTIONS(1532),
    [anon_sym_i8] = ACTIONS(1532),
    [anon_sym_u16] = ACTIONS(1532),
    [anon_sym_i16] = ACTIONS(1532),
    [anon_sym_u32] = ACTIONS(1532),
    [anon_sym_i32] = ACTIONS(1532),
    [anon_sym_u64] = ACTIONS(1532),
    [anon_sym_i64] = ACTIONS(1532),
    [anon_sym_u128] = ACTIONS(1532),
    [anon_sym_i128] = ACTIONS(1532),
    [anon_sym_isize] = ACTIONS(1532),
    [anon_sym_usize] = ACTIONS(1532),
    [anon_sym_f32] = ACTIONS(1532),
    [anon_sym_f64] = ACTIONS(1532),
    [anon_sym_bool] = ACTIONS(1532),
    [anon_sym_str] = ACTIONS(1532),
    [anon_sym_char] = ACTIONS(1532),
    [anon_sym_const] = ACTIONS(1534),
    [anon_sym_default] = ACTIONS(1536),
    [anon_sym_union] = ACTIONS(1536),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(1538),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(1540),
    [sym_mutable_specifier] = ACTIONS(796),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1542),
    [sym_super] = ACTIONS(1542),
    [sym_crate] = ACTIONS(1542),
    [sym_metavariable] = ACTIONS(1544),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [587] = {
    [sym_bracketed_type] = STATE(2489),
    [sym_generic_type] = STATE(2488),
    [sym_generic_type_with_turbofish] = STATE(2487),
    [sym_macro_invocation] = STATE(1422),
    [sym_scoped_identifier] = STATE(1337),
    [sym_scoped_type_identifier] = STATE(2013),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(2285),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [sym_identifier] = ACTIONS(1524),
    [anon_sym_LPAREN] = ACTIONS(1526),
    [anon_sym_LBRACK] = ACTIONS(1530),
    [anon_sym_u8] = ACTIONS(1532),
    [anon_sym_i8] = ACTIONS(1532),
    [anon_sym_u16] = ACTIONS(1532),
    [anon_sym_i16] = ACTIONS(1532),
    [anon_sym_u32] = ACTIONS(1532),
    [anon_sym_i32] = ACTIONS(1532),
    [anon_sym_u64] = ACTIONS(1532),
    [anon_sym_i64] = ACTIONS(1532),
    [anon_sym_u128] = ACTIONS(1532),
    [anon_sym_i128] = ACTIONS(1532),
    [anon_sym_isize] = ACTIONS(1532),
    [anon_sym_usize] = ACTIONS(1532),
    [anon_sym_f32] = ACTIONS(1532),
    [anon_sym_f64] = ACTIONS(1532),
    [anon_sym_bool] = ACTIONS(1532),
    [anon_sym_str] = ACTIONS(1532),
    [anon_sym_char] = ACTIONS(1532),
    [anon_sym_const] = ACTIONS(1534),
    [anon_sym_default] = ACTIONS(1536),
    [anon_sym_union] = ACTIONS(1536),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(1538),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(1540),
    [sym_mutable_specifier] = ACTIONS(796),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1542),
    [sym_super] = ACTIONS(1542),
    [sym_crate] = ACTIONS(1542),
    [sym_metavariable] = ACTIONS(1544),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [588] = {
    [sym_bracketed_type] = STATE(2489),
    [sym_generic_type] = STATE(2488),
    [sym_generic_type_with_turbofish] = STATE(2487),
    [sym_macro_invocation] = STATE(1422),
    [sym_scoped_identifier] = STATE(1337),
    [sym_scoped_type_identifier] = STATE(2013),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1890),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [sym_identifier] = ACTIONS(1524),
    [anon_sym_LPAREN] = ACTIONS(1526),
    [anon_sym_LBRACK] = ACTIONS(1530),
    [anon_sym_u8] = ACTIONS(1532),
    [anon_sym_i8] = ACTIONS(1532),
    [anon_sym_u16] = ACTIONS(1532),
    [anon_sym_i16] = ACTIONS(1532),
    [anon_sym_u32] = ACTIONS(1532),
    [anon_sym_i32] = ACTIONS(1532),
    [anon_sym_u64] = ACTIONS(1532),
    [anon_sym_i64] = ACTIONS(1532),
    [anon_sym_u128] = ACTIONS(1532),
    [anon_sym_i128] = ACTIONS(1532),
    [anon_sym_isize] = ACTIONS(1532),
    [anon_sym_usize] = ACTIONS(1532),
    [anon_sym_f32] = ACTIONS(1532),
    [anon_sym_f64] = ACTIONS(1532),
    [anon_sym_bool] = ACTIONS(1532),
    [anon_sym_str] = ACTIONS(1532),
    [anon_sym_char] = ACTIONS(1532),
    [anon_sym_const] = ACTIONS(1534),
    [anon_sym_default] = ACTIONS(1536),
    [anon_sym_union] = ACTIONS(1536),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(1538),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(1540),
    [sym_mutable_specifier] = ACTIONS(796),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1542),
    [sym_super] = ACTIONS(1542),
    [sym_crate] = ACTIONS(1542),
    [sym_metavariable] = ACTIONS(1544),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [589] = {
    [sym_identifier] = ACTIONS(2427),
    [anon_sym_LPAREN] = ACTIONS(2429),
    [anon_sym_RPAREN] = ACTIONS(2429),
    [anon_sym_LBRACE] = ACTIONS(2429),
    [anon_sym_RBRACE] = ACTIONS(2429),
    [anon_sym_LBRACK] = ACTIONS(2429),
    [anon_sym_RBRACK] = ACTIONS(2429),
    [anon_sym_DOLLAR] = ACTIONS(2429),
    [anon_sym_u8] = ACTIONS(2427),
    [anon_sym_i8] = ACTIONS(2427),
    [anon_sym_u16] = ACTIONS(2427),
    [anon_sym_i16] = ACTIONS(2427),
    [anon_sym_u32] = ACTIONS(2427),
    [anon_sym_i32] = ACTIONS(2427),
    [anon_sym_u64] = ACTIONS(2427),
    [anon_sym_i64] = ACTIONS(2427),
    [anon_sym_u128] = ACTIONS(2427),
    [anon_sym_i128] = ACTIONS(2427),
    [anon_sym_isize] = ACTIONS(2427),
    [anon_sym_usize] = ACTIONS(2427),
    [anon_sym_f32] = ACTIONS(2427),
    [anon_sym_f64] = ACTIONS(2427),
    [anon_sym_bool] = ACTIONS(2427),
    [anon_sym_str] = ACTIONS(2427),
    [anon_sym_char] = ACTIONS(2427),
    [aux_sym__non_special_token_token1] = ACTIONS(2427),
    [anon_sym_SQUOTE] = ACTIONS(2427),
    [anon_sym_as] = ACTIONS(2427),
    [anon_sym_async] = ACTIONS(2427),
    [anon_sym_await] = ACTIONS(2427),
    [anon_sym_break] = ACTIONS(2427),
    [anon_sym_const] = ACTIONS(2427),
    [anon_sym_continue] = ACTIONS(2427),
    [anon_sym_default] = ACTIONS(2427),
    [anon_sym_enum] = ACTIONS(2427),
    [anon_sym_fn] = ACTIONS(2427),
    [anon_sym_for] = ACTIONS(2427),
    [anon_sym_if] = ACTIONS(2427),
    [anon_sym_impl] = ACTIONS(2427),
    [anon_sym_let] = ACTIONS(2427),
    [anon_sym_loop] = ACTIONS(2427),
    [anon_sym_match] = ACTIONS(2427),
    [anon_sym_mod] = ACTIONS(2427),
    [anon_sym_pub] = ACTIONS(2427),
    [anon_sym_return] = ACTIONS(2427),
    [anon_sym_static] = ACTIONS(2427),
    [anon_sym_struct] = ACTIONS(2427),
    [anon_sym_trait] = ACTIONS(2427),
    [anon_sym_type] = ACTIONS(2427),
    [anon_sym_union] = ACTIONS(2427),
    [anon_sym_unsafe] = ACTIONS(2427),
    [anon_sym_use] = ACTIONS(2427),
    [anon_sym_where] = ACTIONS(2427),
    [anon_sym_while] = ACTIONS(2427),
    [sym_mutable_specifier] = ACTIONS(2427),
    [sym_integer_literal] = ACTIONS(2429),
    [aux_sym_string_literal_token1] = ACTIONS(2429),
    [sym_char_literal] = ACTIONS(2429),
    [anon_sym_true] = ACTIONS(2427),
    [anon_sym_false] = ACTIONS(2427),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2427),
    [sym_super] = ACTIONS(2427),
    [sym_crate] = ACTIONS(2427),
    [sym_raw_string_literal] = ACTIONS(2429),
    [sym_float_literal] = ACTIONS(2429),
    [sym_block_comment] = ACTIONS(3),
  },
  [590] = {
    [sym_bracketed_type] = STATE(2489),
    [sym_generic_type] = STATE(2488),
    [sym_generic_type_with_turbofish] = STATE(2487),
    [sym_macro_invocation] = STATE(1422),
    [sym_scoped_identifier] = STATE(1337),
    [sym_scoped_type_identifier] = STATE(2013),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1842),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [sym_identifier] = ACTIONS(1524),
    [anon_sym_LPAREN] = ACTIONS(1526),
    [anon_sym_LBRACK] = ACTIONS(1530),
    [anon_sym_u8] = ACTIONS(1532),
    [anon_sym_i8] = ACTIONS(1532),
    [anon_sym_u16] = ACTIONS(1532),
    [anon_sym_i16] = ACTIONS(1532),
    [anon_sym_u32] = ACTIONS(1532),
    [anon_sym_i32] = ACTIONS(1532),
    [anon_sym_u64] = ACTIONS(1532),
    [anon_sym_i64] = ACTIONS(1532),
    [anon_sym_u128] = ACTIONS(1532),
    [anon_sym_i128] = ACTIONS(1532),
    [anon_sym_isize] = ACTIONS(1532),
    [anon_sym_usize] = ACTIONS(1532),
    [anon_sym_f32] = ACTIONS(1532),
    [anon_sym_f64] = ACTIONS(1532),
    [anon_sym_bool] = ACTIONS(1532),
    [anon_sym_str] = ACTIONS(1532),
    [anon_sym_char] = ACTIONS(1532),
    [anon_sym_const] = ACTIONS(1534),
    [anon_sym_default] = ACTIONS(1536),
    [anon_sym_union] = ACTIONS(1536),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(1538),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(1540),
    [sym_mutable_specifier] = ACTIONS(796),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(2435),
    [sym_super] = ACTIONS(1542),
    [sym_crate] = ACTIONS(1542),
    [sym_metavariable] = ACTIONS(1544),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [591] = {
    [sym_bracketed_type] = STATE(2489),
    [sym_generic_type] = STATE(2488),
    [sym_generic_type_with_turbofish] = STATE(2487),
    [sym_macro_invocation] = STATE(1422),
    [sym_scoped_identifier] = STATE(1337),
    [sym_scoped_type_identifier] = STATE(2013),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(2133),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [sym_identifier] = ACTIONS(1524),
    [anon_sym_LPAREN] = ACTIONS(1526),
    [anon_sym_LBRACK] = ACTIONS(1530),
    [anon_sym_u8] = ACTIONS(1532),
    [anon_sym_i8] = ACTIONS(1532),
    [anon_sym_u16] = ACTIONS(1532),
    [anon_sym_i16] = ACTIONS(1532),
    [anon_sym_u32] = ACTIONS(1532),
    [anon_sym_i32] = ACTIONS(1532),
    [anon_sym_u64] = ACTIONS(1532),
    [anon_sym_i64] = ACTIONS(1532),
    [anon_sym_u128] = ACTIONS(1532),
    [anon_sym_i128] = ACTIONS(1532),
    [anon_sym_isize] = ACTIONS(1532),
    [anon_sym_usize] = ACTIONS(1532),
    [anon_sym_f32] = ACTIONS(1532),
    [anon_sym_f64] = ACTIONS(1532),
    [anon_sym_bool] = ACTIONS(1532),
    [anon_sym_str] = ACTIONS(1532),
    [anon_sym_char] = ACTIONS(1532),
    [anon_sym_const] = ACTIONS(1534),
    [anon_sym_default] = ACTIONS(1536),
    [anon_sym_union] = ACTIONS(1536),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(1538),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(1540),
    [sym_mutable_specifier] = ACTIONS(796),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1542),
    [sym_super] = ACTIONS(1542),
    [sym_crate] = ACTIONS(1542),
    [sym_metavariable] = ACTIONS(1544),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [592] = {
    [sym_identifier] = ACTIONS(2431),
    [anon_sym_LPAREN] = ACTIONS(2433),
    [anon_sym_RPAREN] = ACTIONS(2433),
    [anon_sym_LBRACE] = ACTIONS(2433),
    [anon_sym_RBRACE] = ACTIONS(2433),
    [anon_sym_LBRACK] = ACTIONS(2433),
    [anon_sym_RBRACK] = ACTIONS(2433),
    [anon_sym_DOLLAR] = ACTIONS(2433),
    [anon_sym_u8] = ACTIONS(2431),
    [anon_sym_i8] = ACTIONS(2431),
    [anon_sym_u16] = ACTIONS(2431),
    [anon_sym_i16] = ACTIONS(2431),
    [anon_sym_u32] = ACTIONS(2431),
    [anon_sym_i32] = ACTIONS(2431),
    [anon_sym_u64] = ACTIONS(2431),
    [anon_sym_i64] = ACTIONS(2431),
    [anon_sym_u128] = ACTIONS(2431),
    [anon_sym_i128] = ACTIONS(2431),
    [anon_sym_isize] = ACTIONS(2431),
    [anon_sym_usize] = ACTIONS(2431),
    [anon_sym_f32] = ACTIONS(2431),
    [anon_sym_f64] = ACTIONS(2431),
    [anon_sym_bool] = ACTIONS(2431),
    [anon_sym_str] = ACTIONS(2431),
    [anon_sym_char] = ACTIONS(2431),
    [aux_sym__non_special_token_token1] = ACTIONS(2431),
    [anon_sym_SQUOTE] = ACTIONS(2431),
    [anon_sym_as] = ACTIONS(2431),
    [anon_sym_async] = ACTIONS(2431),
    [anon_sym_await] = ACTIONS(2431),
    [anon_sym_break] = ACTIONS(2431),
    [anon_sym_const] = ACTIONS(2431),
    [anon_sym_continue] = ACTIONS(2431),
    [anon_sym_default] = ACTIONS(2431),
    [anon_sym_enum] = ACTIONS(2431),
    [anon_sym_fn] = ACTIONS(2431),
    [anon_sym_for] = ACTIONS(2431),
    [anon_sym_if] = ACTIONS(2431),
    [anon_sym_impl] = ACTIONS(2431),
    [anon_sym_let] = ACTIONS(2431),
    [anon_sym_loop] = ACTIONS(2431),
    [anon_sym_match] = ACTIONS(2431),
    [anon_sym_mod] = ACTIONS(2431),
    [anon_sym_pub] = ACTIONS(2431),
    [anon_sym_return] = ACTIONS(2431),
    [anon_sym_static] = ACTIONS(2431),
    [anon_sym_struct] = ACTIONS(2431),
    [anon_sym_trait] = ACTIONS(2431),
    [anon_sym_type] = ACTIONS(2431),
    [anon_sym_union] = ACTIONS(2431),
    [anon_sym_unsafe] = ACTIONS(2431),
    [anon_sym_use] = ACTIONS(2431),
    [anon_sym_where] = ACTIONS(2431),
    [anon_sym_while] = ACTIONS(2431),
    [sym_mutable_specifier] = ACTIONS(2431),
    [sym_integer_literal] = ACTIONS(2433),
    [aux_sym_string_literal_token1] = ACTIONS(2433),
    [sym_char_literal] = ACTIONS(2433),
    [anon_sym_true] = ACTIONS(2431),
    [anon_sym_false] = ACTIONS(2431),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2431),
    [sym_super] = ACTIONS(2431),
    [sym_crate] = ACTIONS(2431),
    [sym_raw_string_literal] = ACTIONS(2433),
    [sym_float_literal] = ACTIONS(2433),
    [sym_block_comment] = ACTIONS(3),
  },
  [593] = {
    [sym_bracketed_type] = STATE(2489),
    [sym_generic_type] = STATE(2488),
    [sym_generic_type_with_turbofish] = STATE(2487),
    [sym_macro_invocation] = STATE(1422),
    [sym_scoped_identifier] = STATE(1337),
    [sym_scoped_type_identifier] = STATE(2013),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1939),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [sym_identifier] = ACTIONS(1524),
    [anon_sym_LPAREN] = ACTIONS(1526),
    [anon_sym_LBRACK] = ACTIONS(1530),
    [anon_sym_u8] = ACTIONS(1532),
    [anon_sym_i8] = ACTIONS(1532),
    [anon_sym_u16] = ACTIONS(1532),
    [anon_sym_i16] = ACTIONS(1532),
    [anon_sym_u32] = ACTIONS(1532),
    [anon_sym_i32] = ACTIONS(1532),
    [anon_sym_u64] = ACTIONS(1532),
    [anon_sym_i64] = ACTIONS(1532),
    [anon_sym_u128] = ACTIONS(1532),
    [anon_sym_i128] = ACTIONS(1532),
    [anon_sym_isize] = ACTIONS(1532),
    [anon_sym_usize] = ACTIONS(1532),
    [anon_sym_f32] = ACTIONS(1532),
    [anon_sym_f64] = ACTIONS(1532),
    [anon_sym_bool] = ACTIONS(1532),
    [anon_sym_str] = ACTIONS(1532),
    [anon_sym_char] = ACTIONS(1532),
    [anon_sym_const] = ACTIONS(1534),
    [anon_sym_default] = ACTIONS(1536),
    [anon_sym_union] = ACTIONS(1536),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(1538),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(1540),
    [sym_mutable_specifier] = ACTIONS(796),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1542),
    [sym_super] = ACTIONS(1542),
    [sym_crate] = ACTIONS(1542),
    [sym_metavariable] = ACTIONS(1544),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [594] = {
    [sym_function_modifiers] = STATE(2472),
    [sym_higher_ranked_trait_bound] = STATE(1568),
    [sym_removed_trait_bound] = STATE(1568),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(1570),
    [sym_bracketed_type] = STATE(2378),
    [sym_lifetime] = STATE(1575),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2379),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1390),
    [sym_scoped_identifier] = STATE(2157),
    [sym_scoped_type_identifier] = STATE(1331),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(2309),
    [anon_sym_LPAREN] = ACTIONS(856),
    [anon_sym_LBRACK] = ACTIONS(860),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_QMARK] = ACTIONS(2437),
    [anon_sym_u8] = ACTIONS(862),
    [anon_sym_i8] = ACTIONS(862),
    [anon_sym_u16] = ACTIONS(862),
    [anon_sym_i16] = ACTIONS(862),
    [anon_sym_u32] = ACTIONS(862),
    [anon_sym_i32] = ACTIONS(862),
    [anon_sym_u64] = ACTIONS(862),
    [anon_sym_i64] = ACTIONS(862),
    [anon_sym_u128] = ACTIONS(862),
    [anon_sym_i128] = ACTIONS(862),
    [anon_sym_isize] = ACTIONS(862),
    [anon_sym_usize] = ACTIONS(862),
    [anon_sym_f32] = ACTIONS(862),
    [anon_sym_f64] = ACTIONS(862),
    [anon_sym_bool] = ACTIONS(862),
    [anon_sym_str] = ACTIONS(862),
    [anon_sym_char] = ACTIONS(862),
    [anon_sym_SQUOTE] = ACTIONS(2313),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(664),
    [anon_sym_default] = ACTIONS(864),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(2439),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(866),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(870),
    [anon_sym_AMP] = ACTIONS(872),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(876),
    [sym_super] = ACTIONS(876),
    [sym_crate] = ACTIONS(876),
    [sym_metavariable] = ACTIONS(878),
    [sym_block_comment] = ACTIONS(3),
  },
  [595] = {
    [sym_function_modifiers] = STATE(2472),
    [sym_higher_ranked_trait_bound] = STATE(1568),
    [sym_removed_trait_bound] = STATE(1568),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(1591),
    [sym_bracketed_type] = STATE(2378),
    [sym_lifetime] = STATE(1593),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2379),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1390),
    [sym_scoped_identifier] = STATE(2157),
    [sym_scoped_type_identifier] = STATE(1331),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(2309),
    [anon_sym_LPAREN] = ACTIONS(856),
    [anon_sym_LBRACK] = ACTIONS(860),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_QMARK] = ACTIONS(2437),
    [anon_sym_u8] = ACTIONS(862),
    [anon_sym_i8] = ACTIONS(862),
    [anon_sym_u16] = ACTIONS(862),
    [anon_sym_i16] = ACTIONS(862),
    [anon_sym_u32] = ACTIONS(862),
    [anon_sym_i32] = ACTIONS(862),
    [anon_sym_u64] = ACTIONS(862),
    [anon_sym_i64] = ACTIONS(862),
    [anon_sym_u128] = ACTIONS(862),
    [anon_sym_i128] = ACTIONS(862),
    [anon_sym_isize] = ACTIONS(862),
    [anon_sym_usize] = ACTIONS(862),
    [anon_sym_f32] = ACTIONS(862),
    [anon_sym_f64] = ACTIONS(862),
    [anon_sym_bool] = ACTIONS(862),
    [anon_sym_str] = ACTIONS(862),
    [anon_sym_char] = ACTIONS(862),
    [anon_sym_SQUOTE] = ACTIONS(2313),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(664),
    [anon_sym_default] = ACTIONS(864),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(2439),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(866),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(870),
    [anon_sym_AMP] = ACTIONS(872),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(876),
    [sym_super] = ACTIONS(876),
    [sym_crate] = ACTIONS(876),
    [sym_metavariable] = ACTIONS(878),
    [sym_block_comment] = ACTIONS(3),
  },
  [596] = {
    [sym_function_modifiers] = STATE(2472),
    [sym_higher_ranked_trait_bound] = STATE(1568),
    [sym_removed_trait_bound] = STATE(1568),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(1591),
    [sym_bracketed_type] = STATE(2378),
    [sym_lifetime] = STATE(1575),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2379),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1390),
    [sym_scoped_identifier] = STATE(2157),
    [sym_scoped_type_identifier] = STATE(1331),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(2309),
    [anon_sym_LPAREN] = ACTIONS(856),
    [anon_sym_LBRACK] = ACTIONS(860),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_QMARK] = ACTIONS(2437),
    [anon_sym_u8] = ACTIONS(862),
    [anon_sym_i8] = ACTIONS(862),
    [anon_sym_u16] = ACTIONS(862),
    [anon_sym_i16] = ACTIONS(862),
    [anon_sym_u32] = ACTIONS(862),
    [anon_sym_i32] = ACTIONS(862),
    [anon_sym_u64] = ACTIONS(862),
    [anon_sym_i64] = ACTIONS(862),
    [anon_sym_u128] = ACTIONS(862),
    [anon_sym_i128] = ACTIONS(862),
    [anon_sym_isize] = ACTIONS(862),
    [anon_sym_usize] = ACTIONS(862),
    [anon_sym_f32] = ACTIONS(862),
    [anon_sym_f64] = ACTIONS(862),
    [anon_sym_bool] = ACTIONS(862),
    [anon_sym_str] = ACTIONS(862),
    [anon_sym_char] = ACTIONS(862),
    [anon_sym_SQUOTE] = ACTIONS(2313),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(664),
    [anon_sym_default] = ACTIONS(864),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(2439),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(866),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(870),
    [anon_sym_AMP] = ACTIONS(872),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(876),
    [sym_super] = ACTIONS(876),
    [sym_crate] = ACTIONS(876),
    [sym_metavariable] = ACTIONS(878),
    [sym_block_comment] = ACTIONS(3),
  },
  [597] = {
    [sym_bracketed_type] = STATE(2489),
    [sym_generic_type] = STATE(2488),
    [sym_generic_type_with_turbofish] = STATE(2487),
    [sym_macro_invocation] = STATE(1422),
    [sym_scoped_identifier] = STATE(1337),
    [sym_scoped_type_identifier] = STATE(2013),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(2246),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [sym_identifier] = ACTIONS(1524),
    [anon_sym_LPAREN] = ACTIONS(1526),
    [anon_sym_LBRACK] = ACTIONS(1530),
    [anon_sym_u8] = ACTIONS(1532),
    [anon_sym_i8] = ACTIONS(1532),
    [anon_sym_u16] = ACTIONS(1532),
    [anon_sym_i16] = ACTIONS(1532),
    [anon_sym_u32] = ACTIONS(1532),
    [anon_sym_i32] = ACTIONS(1532),
    [anon_sym_u64] = ACTIONS(1532),
    [anon_sym_i64] = ACTIONS(1532),
    [anon_sym_u128] = ACTIONS(1532),
    [anon_sym_i128] = ACTIONS(1532),
    [anon_sym_isize] = ACTIONS(1532),
    [anon_sym_usize] = ACTIONS(1532),
    [anon_sym_f32] = ACTIONS(1532),
    [anon_sym_f64] = ACTIONS(1532),
    [anon_sym_bool] = ACTIONS(1532),
    [anon_sym_str] = ACTIONS(1532),
    [anon_sym_char] = ACTIONS(1532),
    [anon_sym_const] = ACTIONS(1534),
    [anon_sym_default] = ACTIONS(1536),
    [anon_sym_union] = ACTIONS(1536),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(1538),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(1540),
    [sym_mutable_specifier] = ACTIONS(796),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1542),
    [sym_super] = ACTIONS(1542),
    [sym_crate] = ACTIONS(1542),
    [sym_metavariable] = ACTIONS(1544),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [598] = {
    [sym_bracketed_type] = STATE(2489),
    [sym_generic_type] = STATE(2488),
    [sym_generic_type_with_turbofish] = STATE(2487),
    [sym_macro_invocation] = STATE(1422),
    [sym_scoped_identifier] = STATE(1337),
    [sym_scoped_type_identifier] = STATE(2013),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1736),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [sym_identifier] = ACTIONS(1524),
    [anon_sym_LPAREN] = ACTIONS(1526),
    [anon_sym_LBRACK] = ACTIONS(1530),
    [anon_sym_u8] = ACTIONS(1532),
    [anon_sym_i8] = ACTIONS(1532),
    [anon_sym_u16] = ACTIONS(1532),
    [anon_sym_i16] = ACTIONS(1532),
    [anon_sym_u32] = ACTIONS(1532),
    [anon_sym_i32] = ACTIONS(1532),
    [anon_sym_u64] = ACTIONS(1532),
    [anon_sym_i64] = ACTIONS(1532),
    [anon_sym_u128] = ACTIONS(1532),
    [anon_sym_i128] = ACTIONS(1532),
    [anon_sym_isize] = ACTIONS(1532),
    [anon_sym_usize] = ACTIONS(1532),
    [anon_sym_f32] = ACTIONS(1532),
    [anon_sym_f64] = ACTIONS(1532),
    [anon_sym_bool] = ACTIONS(1532),
    [anon_sym_str] = ACTIONS(1532),
    [anon_sym_char] = ACTIONS(1532),
    [anon_sym_const] = ACTIONS(1534),
    [anon_sym_default] = ACTIONS(1536),
    [anon_sym_union] = ACTIONS(1536),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(1538),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(1540),
    [sym_mutable_specifier] = ACTIONS(2441),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1542),
    [sym_super] = ACTIONS(1542),
    [sym_crate] = ACTIONS(1542),
    [sym_metavariable] = ACTIONS(1544),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [599] = {
    [sym_bracketed_type] = STATE(2489),
    [sym_generic_type] = STATE(2488),
    [sym_generic_type_with_turbofish] = STATE(2487),
    [sym_macro_invocation] = STATE(1422),
    [sym_scoped_identifier] = STATE(1337),
    [sym_scoped_type_identifier] = STATE(2013),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1418),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [sym_identifier] = ACTIONS(1524),
    [anon_sym_LPAREN] = ACTIONS(1526),
    [anon_sym_LBRACK] = ACTIONS(1530),
    [anon_sym_u8] = ACTIONS(1532),
    [anon_sym_i8] = ACTIONS(1532),
    [anon_sym_u16] = ACTIONS(1532),
    [anon_sym_i16] = ACTIONS(1532),
    [anon_sym_u32] = ACTIONS(1532),
    [anon_sym_i32] = ACTIONS(1532),
    [anon_sym_u64] = ACTIONS(1532),
    [anon_sym_i64] = ACTIONS(1532),
    [anon_sym_u128] = ACTIONS(1532),
    [anon_sym_i128] = ACTIONS(1532),
    [anon_sym_isize] = ACTIONS(1532),
    [anon_sym_usize] = ACTIONS(1532),
    [anon_sym_f32] = ACTIONS(1532),
    [anon_sym_f64] = ACTIONS(1532),
    [anon_sym_bool] = ACTIONS(1532),
    [anon_sym_str] = ACTIONS(1532),
    [anon_sym_char] = ACTIONS(1532),
    [anon_sym_const] = ACTIONS(1534),
    [anon_sym_default] = ACTIONS(1536),
    [anon_sym_union] = ACTIONS(1536),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(1538),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(1540),
    [sym_mutable_specifier] = ACTIONS(796),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1542),
    [sym_super] = ACTIONS(1542),
    [sym_crate] = ACTIONS(1542),
    [sym_metavariable] = ACTIONS(1544),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [600] = {
    [sym_bracketed_type] = STATE(2489),
    [sym_generic_type] = STATE(2488),
    [sym_generic_type_with_turbofish] = STATE(2487),
    [sym_macro_invocation] = STATE(1422),
    [sym_scoped_identifier] = STATE(1337),
    [sym_scoped_type_identifier] = STATE(2013),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1746),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [sym_identifier] = ACTIONS(1524),
    [anon_sym_LPAREN] = ACTIONS(1526),
    [anon_sym_LBRACK] = ACTIONS(1530),
    [anon_sym_u8] = ACTIONS(1532),
    [anon_sym_i8] = ACTIONS(1532),
    [anon_sym_u16] = ACTIONS(1532),
    [anon_sym_i16] = ACTIONS(1532),
    [anon_sym_u32] = ACTIONS(1532),
    [anon_sym_i32] = ACTIONS(1532),
    [anon_sym_u64] = ACTIONS(1532),
    [anon_sym_i64] = ACTIONS(1532),
    [anon_sym_u128] = ACTIONS(1532),
    [anon_sym_i128] = ACTIONS(1532),
    [anon_sym_isize] = ACTIONS(1532),
    [anon_sym_usize] = ACTIONS(1532),
    [anon_sym_f32] = ACTIONS(1532),
    [anon_sym_f64] = ACTIONS(1532),
    [anon_sym_bool] = ACTIONS(1532),
    [anon_sym_str] = ACTIONS(1532),
    [anon_sym_char] = ACTIONS(1532),
    [anon_sym_const] = ACTIONS(1534),
    [anon_sym_default] = ACTIONS(1536),
    [anon_sym_union] = ACTIONS(1536),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(1538),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(1540),
    [sym_mutable_specifier] = ACTIONS(796),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1542),
    [sym_super] = ACTIONS(1542),
    [sym_crate] = ACTIONS(1542),
    [sym_metavariable] = ACTIONS(1544),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [601] = {
    [sym_identifier] = ACTIONS(2367),
    [anon_sym_LPAREN] = ACTIONS(2369),
    [anon_sym_RPAREN] = ACTIONS(2369),
    [anon_sym_LBRACE] = ACTIONS(2369),
    [anon_sym_RBRACE] = ACTIONS(2369),
    [anon_sym_LBRACK] = ACTIONS(2369),
    [anon_sym_RBRACK] = ACTIONS(2369),
    [anon_sym_DOLLAR] = ACTIONS(2369),
    [anon_sym_u8] = ACTIONS(2367),
    [anon_sym_i8] = ACTIONS(2367),
    [anon_sym_u16] = ACTIONS(2367),
    [anon_sym_i16] = ACTIONS(2367),
    [anon_sym_u32] = ACTIONS(2367),
    [anon_sym_i32] = ACTIONS(2367),
    [anon_sym_u64] = ACTIONS(2367),
    [anon_sym_i64] = ACTIONS(2367),
    [anon_sym_u128] = ACTIONS(2367),
    [anon_sym_i128] = ACTIONS(2367),
    [anon_sym_isize] = ACTIONS(2367),
    [anon_sym_usize] = ACTIONS(2367),
    [anon_sym_f32] = ACTIONS(2367),
    [anon_sym_f64] = ACTIONS(2367),
    [anon_sym_bool] = ACTIONS(2367),
    [anon_sym_str] = ACTIONS(2367),
    [anon_sym_char] = ACTIONS(2367),
    [aux_sym__non_special_token_token1] = ACTIONS(2367),
    [anon_sym_SQUOTE] = ACTIONS(2367),
    [anon_sym_as] = ACTIONS(2367),
    [anon_sym_async] = ACTIONS(2367),
    [anon_sym_await] = ACTIONS(2367),
    [anon_sym_break] = ACTIONS(2367),
    [anon_sym_const] = ACTIONS(2367),
    [anon_sym_continue] = ACTIONS(2367),
    [anon_sym_default] = ACTIONS(2367),
    [anon_sym_enum] = ACTIONS(2367),
    [anon_sym_fn] = ACTIONS(2367),
    [anon_sym_for] = ACTIONS(2367),
    [anon_sym_if] = ACTIONS(2367),
    [anon_sym_impl] = ACTIONS(2367),
    [anon_sym_let] = ACTIONS(2367),
    [anon_sym_loop] = ACTIONS(2367),
    [anon_sym_match] = ACTIONS(2367),
    [anon_sym_mod] = ACTIONS(2367),
    [anon_sym_pub] = ACTIONS(2367),
    [anon_sym_return] = ACTIONS(2367),
    [anon_sym_static] = ACTIONS(2367),
    [anon_sym_struct] = ACTIONS(2367),
    [anon_sym_trait] = ACTIONS(2367),
    [anon_sym_type] = ACTIONS(2367),
    [anon_sym_union] = ACTIONS(2367),
    [anon_sym_unsafe] = ACTIONS(2367),
    [anon_sym_use] = ACTIONS(2367),
    [anon_sym_where] = ACTIONS(2367),
    [anon_sym_while] = ACTIONS(2367),
    [sym_mutable_specifier] = ACTIONS(2367),
    [sym_integer_literal] = ACTIONS(2369),
    [aux_sym_string_literal_token1] = ACTIONS(2369),
    [sym_char_literal] = ACTIONS(2369),
    [anon_sym_true] = ACTIONS(2367),
    [anon_sym_false] = ACTIONS(2367),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2367),
    [sym_super] = ACTIONS(2367),
    [sym_crate] = ACTIONS(2367),
    [sym_raw_string_literal] = ACTIONS(2369),
    [sym_float_literal] = ACTIONS(2369),
    [sym_block_comment] = ACTIONS(3),
  },
  [602] = {
    [sym_bracketed_type] = STATE(2489),
    [sym_generic_type] = STATE(2488),
    [sym_generic_type_with_turbofish] = STATE(2487),
    [sym_macro_invocation] = STATE(1422),
    [sym_scoped_identifier] = STATE(1337),
    [sym_scoped_type_identifier] = STATE(2013),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1458),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [sym_identifier] = ACTIONS(1524),
    [anon_sym_LPAREN] = ACTIONS(1526),
    [anon_sym_LBRACK] = ACTIONS(1530),
    [anon_sym_u8] = ACTIONS(1532),
    [anon_sym_i8] = ACTIONS(1532),
    [anon_sym_u16] = ACTIONS(1532),
    [anon_sym_i16] = ACTIONS(1532),
    [anon_sym_u32] = ACTIONS(1532),
    [anon_sym_i32] = ACTIONS(1532),
    [anon_sym_u64] = ACTIONS(1532),
    [anon_sym_i64] = ACTIONS(1532),
    [anon_sym_u128] = ACTIONS(1532),
    [anon_sym_i128] = ACTIONS(1532),
    [anon_sym_isize] = ACTIONS(1532),
    [anon_sym_usize] = ACTIONS(1532),
    [anon_sym_f32] = ACTIONS(1532),
    [anon_sym_f64] = ACTIONS(1532),
    [anon_sym_bool] = ACTIONS(1532),
    [anon_sym_str] = ACTIONS(1532),
    [anon_sym_char] = ACTIONS(1532),
    [anon_sym_const] = ACTIONS(1534),
    [anon_sym_default] = ACTIONS(1536),
    [anon_sym_union] = ACTIONS(1536),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(1538),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(1540),
    [sym_mutable_specifier] = ACTIONS(796),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1542),
    [sym_super] = ACTIONS(1542),
    [sym_crate] = ACTIONS(1542),
    [sym_metavariable] = ACTIONS(1544),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [603] = {
    [sym_bracketed_type] = STATE(2489),
    [sym_generic_type] = STATE(2488),
    [sym_generic_type_with_turbofish] = STATE(2487),
    [sym_macro_invocation] = STATE(1422),
    [sym_scoped_identifier] = STATE(1337),
    [sym_scoped_type_identifier] = STATE(2013),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1842),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [sym_identifier] = ACTIONS(1524),
    [anon_sym_LPAREN] = ACTIONS(1526),
    [anon_sym_LBRACK] = ACTIONS(1530),
    [anon_sym_u8] = ACTIONS(1532),
    [anon_sym_i8] = ACTIONS(1532),
    [anon_sym_u16] = ACTIONS(1532),
    [anon_sym_i16] = ACTIONS(1532),
    [anon_sym_u32] = ACTIONS(1532),
    [anon_sym_i32] = ACTIONS(1532),
    [anon_sym_u64] = ACTIONS(1532),
    [anon_sym_i64] = ACTIONS(1532),
    [anon_sym_u128] = ACTIONS(1532),
    [anon_sym_i128] = ACTIONS(1532),
    [anon_sym_isize] = ACTIONS(1532),
    [anon_sym_usize] = ACTIONS(1532),
    [anon_sym_f32] = ACTIONS(1532),
    [anon_sym_f64] = ACTIONS(1532),
    [anon_sym_bool] = ACTIONS(1532),
    [anon_sym_str] = ACTIONS(1532),
    [anon_sym_char] = ACTIONS(1532),
    [anon_sym_const] = ACTIONS(1534),
    [anon_sym_default] = ACTIONS(1536),
    [anon_sym_union] = ACTIONS(1536),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(1538),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(1540),
    [sym_mutable_specifier] = ACTIONS(796),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(2443),
    [sym_super] = ACTIONS(1542),
    [sym_crate] = ACTIONS(1542),
    [sym_metavariable] = ACTIONS(1544),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [604] = {
    [sym_bracketed_type] = STATE(2489),
    [sym_generic_type] = STATE(2488),
    [sym_generic_type_with_turbofish] = STATE(2487),
    [sym_macro_invocation] = STATE(1422),
    [sym_scoped_identifier] = STATE(1337),
    [sym_scoped_type_identifier] = STATE(2013),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1452),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [sym_identifier] = ACTIONS(1524),
    [anon_sym_LPAREN] = ACTIONS(1526),
    [anon_sym_LBRACK] = ACTIONS(1530),
    [anon_sym_u8] = ACTIONS(1532),
    [anon_sym_i8] = ACTIONS(1532),
    [anon_sym_u16] = ACTIONS(1532),
    [anon_sym_i16] = ACTIONS(1532),
    [anon_sym_u32] = ACTIONS(1532),
    [anon_sym_i32] = ACTIONS(1532),
    [anon_sym_u64] = ACTIONS(1532),
    [anon_sym_i64] = ACTIONS(1532),
    [anon_sym_u128] = ACTIONS(1532),
    [anon_sym_i128] = ACTIONS(1532),
    [anon_sym_isize] = ACTIONS(1532),
    [anon_sym_usize] = ACTIONS(1532),
    [anon_sym_f32] = ACTIONS(1532),
    [anon_sym_f64] = ACTIONS(1532),
    [anon_sym_bool] = ACTIONS(1532),
    [anon_sym_str] = ACTIONS(1532),
    [anon_sym_char] = ACTIONS(1532),
    [anon_sym_const] = ACTIONS(1534),
    [anon_sym_default] = ACTIONS(1536),
    [anon_sym_union] = ACTIONS(1536),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(1538),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(1540),
    [sym_mutable_specifier] = ACTIONS(796),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1542),
    [sym_super] = ACTIONS(1542),
    [sym_crate] = ACTIONS(1542),
    [sym_metavariable] = ACTIONS(1544),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [605] = {
    [sym_identifier] = ACTIONS(440),
    [anon_sym_LPAREN] = ACTIONS(438),
    [anon_sym_RPAREN] = ACTIONS(438),
    [anon_sym_LBRACE] = ACTIONS(438),
    [anon_sym_RBRACE] = ACTIONS(438),
    [anon_sym_LBRACK] = ACTIONS(438),
    [anon_sym_RBRACK] = ACTIONS(438),
    [anon_sym_DOLLAR] = ACTIONS(438),
    [anon_sym_u8] = ACTIONS(440),
    [anon_sym_i8] = ACTIONS(440),
    [anon_sym_u16] = ACTIONS(440),
    [anon_sym_i16] = ACTIONS(440),
    [anon_sym_u32] = ACTIONS(440),
    [anon_sym_i32] = ACTIONS(440),
    [anon_sym_u64] = ACTIONS(440),
    [anon_sym_i64] = ACTIONS(440),
    [anon_sym_u128] = ACTIONS(440),
    [anon_sym_i128] = ACTIONS(440),
    [anon_sym_isize] = ACTIONS(440),
    [anon_sym_usize] = ACTIONS(440),
    [anon_sym_f32] = ACTIONS(440),
    [anon_sym_f64] = ACTIONS(440),
    [anon_sym_bool] = ACTIONS(440),
    [anon_sym_str] = ACTIONS(440),
    [anon_sym_char] = ACTIONS(440),
    [aux_sym__non_special_token_token1] = ACTIONS(440),
    [anon_sym_SQUOTE] = ACTIONS(440),
    [anon_sym_as] = ACTIONS(440),
    [anon_sym_async] = ACTIONS(440),
    [anon_sym_await] = ACTIONS(440),
    [anon_sym_break] = ACTIONS(440),
    [anon_sym_const] = ACTIONS(440),
    [anon_sym_continue] = ACTIONS(440),
    [anon_sym_default] = ACTIONS(440),
    [anon_sym_enum] = ACTIONS(440),
    [anon_sym_fn] = ACTIONS(440),
    [anon_sym_for] = ACTIONS(440),
    [anon_sym_if] = ACTIONS(440),
    [anon_sym_impl] = ACTIONS(440),
    [anon_sym_let] = ACTIONS(440),
    [anon_sym_loop] = ACTIONS(440),
    [anon_sym_match] = ACTIONS(440),
    [anon_sym_mod] = ACTIONS(440),
    [anon_sym_pub] = ACTIONS(440),
    [anon_sym_return] = ACTIONS(440),
    [anon_sym_static] = ACTIONS(440),
    [anon_sym_struct] = ACTIONS(440),
    [anon_sym_trait] = ACTIONS(440),
    [anon_sym_type] = ACTIONS(440),
    [anon_sym_union] = ACTIONS(440),
    [anon_sym_unsafe] = ACTIONS(440),
    [anon_sym_use] = ACTIONS(440),
    [anon_sym_where] = ACTIONS(440),
    [anon_sym_while] = ACTIONS(440),
    [sym_mutable_specifier] = ACTIONS(440),
    [sym_integer_literal] = ACTIONS(438),
    [aux_sym_string_literal_token1] = ACTIONS(438),
    [sym_char_literal] = ACTIONS(438),
    [anon_sym_true] = ACTIONS(440),
    [anon_sym_false] = ACTIONS(440),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(440),
    [sym_super] = ACTIONS(440),
    [sym_crate] = ACTIONS(440),
    [sym_raw_string_literal] = ACTIONS(438),
    [sym_float_literal] = ACTIONS(438),
    [sym_block_comment] = ACTIONS(3),
  },
  [606] = {
    [sym_bracketed_type] = STATE(2489),
    [sym_generic_type] = STATE(2488),
    [sym_generic_type_with_turbofish] = STATE(2487),
    [sym_macro_invocation] = STATE(1422),
    [sym_scoped_identifier] = STATE(1337),
    [sym_scoped_type_identifier] = STATE(2013),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1804),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [sym_identifier] = ACTIONS(1524),
    [anon_sym_LPAREN] = ACTIONS(1526),
    [anon_sym_LBRACK] = ACTIONS(1530),
    [anon_sym_u8] = ACTIONS(1532),
    [anon_sym_i8] = ACTIONS(1532),
    [anon_sym_u16] = ACTIONS(1532),
    [anon_sym_i16] = ACTIONS(1532),
    [anon_sym_u32] = ACTIONS(1532),
    [anon_sym_i32] = ACTIONS(1532),
    [anon_sym_u64] = ACTIONS(1532),
    [anon_sym_i64] = ACTIONS(1532),
    [anon_sym_u128] = ACTIONS(1532),
    [anon_sym_i128] = ACTIONS(1532),
    [anon_sym_isize] = ACTIONS(1532),
    [anon_sym_usize] = ACTIONS(1532),
    [anon_sym_f32] = ACTIONS(1532),
    [anon_sym_f64] = ACTIONS(1532),
    [anon_sym_bool] = ACTIONS(1532),
    [anon_sym_str] = ACTIONS(1532),
    [anon_sym_char] = ACTIONS(1532),
    [anon_sym_const] = ACTIONS(1534),
    [anon_sym_default] = ACTIONS(1536),
    [anon_sym_union] = ACTIONS(1536),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(1538),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(1540),
    [sym_mutable_specifier] = ACTIONS(796),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1542),
    [sym_super] = ACTIONS(1542),
    [sym_crate] = ACTIONS(1542),
    [sym_metavariable] = ACTIONS(1544),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [607] = {
    [sym_bracketed_type] = STATE(2489),
    [sym_generic_type] = STATE(2488),
    [sym_generic_type_with_turbofish] = STATE(2487),
    [sym_macro_invocation] = STATE(1422),
    [sym_scoped_identifier] = STATE(1337),
    [sym_scoped_type_identifier] = STATE(2013),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1426),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [sym_identifier] = ACTIONS(1524),
    [anon_sym_LPAREN] = ACTIONS(1526),
    [anon_sym_LBRACK] = ACTIONS(1530),
    [anon_sym_u8] = ACTIONS(1532),
    [anon_sym_i8] = ACTIONS(1532),
    [anon_sym_u16] = ACTIONS(1532),
    [anon_sym_i16] = ACTIONS(1532),
    [anon_sym_u32] = ACTIONS(1532),
    [anon_sym_i32] = ACTIONS(1532),
    [anon_sym_u64] = ACTIONS(1532),
    [anon_sym_i64] = ACTIONS(1532),
    [anon_sym_u128] = ACTIONS(1532),
    [anon_sym_i128] = ACTIONS(1532),
    [anon_sym_isize] = ACTIONS(1532),
    [anon_sym_usize] = ACTIONS(1532),
    [anon_sym_f32] = ACTIONS(1532),
    [anon_sym_f64] = ACTIONS(1532),
    [anon_sym_bool] = ACTIONS(1532),
    [anon_sym_str] = ACTIONS(1532),
    [anon_sym_char] = ACTIONS(1532),
    [anon_sym_const] = ACTIONS(1534),
    [anon_sym_default] = ACTIONS(1536),
    [anon_sym_union] = ACTIONS(1536),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(1538),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(1540),
    [sym_mutable_specifier] = ACTIONS(796),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1542),
    [sym_super] = ACTIONS(1542),
    [sym_crate] = ACTIONS(1542),
    [sym_metavariable] = ACTIONS(1544),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [608] = {
    [sym_bracketed_type] = STATE(2489),
    [sym_generic_type] = STATE(2488),
    [sym_generic_type_with_turbofish] = STATE(2487),
    [sym_macro_invocation] = STATE(1422),
    [sym_scoped_identifier] = STATE(1337),
    [sym_scoped_type_identifier] = STATE(2013),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1453),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [sym_identifier] = ACTIONS(1524),
    [anon_sym_LPAREN] = ACTIONS(1526),
    [anon_sym_LBRACK] = ACTIONS(1530),
    [anon_sym_u8] = ACTIONS(1532),
    [anon_sym_i8] = ACTIONS(1532),
    [anon_sym_u16] = ACTIONS(1532),
    [anon_sym_i16] = ACTIONS(1532),
    [anon_sym_u32] = ACTIONS(1532),
    [anon_sym_i32] = ACTIONS(1532),
    [anon_sym_u64] = ACTIONS(1532),
    [anon_sym_i64] = ACTIONS(1532),
    [anon_sym_u128] = ACTIONS(1532),
    [anon_sym_i128] = ACTIONS(1532),
    [anon_sym_isize] = ACTIONS(1532),
    [anon_sym_usize] = ACTIONS(1532),
    [anon_sym_f32] = ACTIONS(1532),
    [anon_sym_f64] = ACTIONS(1532),
    [anon_sym_bool] = ACTIONS(1532),
    [anon_sym_str] = ACTIONS(1532),
    [anon_sym_char] = ACTIONS(1532),
    [anon_sym_const] = ACTIONS(1534),
    [anon_sym_default] = ACTIONS(1536),
    [anon_sym_union] = ACTIONS(1536),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(1538),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(1540),
    [sym_mutable_specifier] = ACTIONS(2445),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1542),
    [sym_super] = ACTIONS(1542),
    [sym_crate] = ACTIONS(1542),
    [sym_metavariable] = ACTIONS(1544),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [609] = {
    [sym_bracketed_type] = STATE(2489),
    [sym_generic_type] = STATE(2488),
    [sym_generic_type_with_turbofish] = STATE(2487),
    [sym_macro_invocation] = STATE(1422),
    [sym_scoped_identifier] = STATE(1337),
    [sym_scoped_type_identifier] = STATE(2013),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1870),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [sym_identifier] = ACTIONS(1524),
    [anon_sym_LPAREN] = ACTIONS(1526),
    [anon_sym_LBRACK] = ACTIONS(1530),
    [anon_sym_u8] = ACTIONS(1532),
    [anon_sym_i8] = ACTIONS(1532),
    [anon_sym_u16] = ACTIONS(1532),
    [anon_sym_i16] = ACTIONS(1532),
    [anon_sym_u32] = ACTIONS(1532),
    [anon_sym_i32] = ACTIONS(1532),
    [anon_sym_u64] = ACTIONS(1532),
    [anon_sym_i64] = ACTIONS(1532),
    [anon_sym_u128] = ACTIONS(1532),
    [anon_sym_i128] = ACTIONS(1532),
    [anon_sym_isize] = ACTIONS(1532),
    [anon_sym_usize] = ACTIONS(1532),
    [anon_sym_f32] = ACTIONS(1532),
    [anon_sym_f64] = ACTIONS(1532),
    [anon_sym_bool] = ACTIONS(1532),
    [anon_sym_str] = ACTIONS(1532),
    [anon_sym_char] = ACTIONS(1532),
    [anon_sym_const] = ACTIONS(1534),
    [anon_sym_default] = ACTIONS(1536),
    [anon_sym_union] = ACTIONS(1536),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(1538),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(1540),
    [sym_mutable_specifier] = ACTIONS(796),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1542),
    [sym_super] = ACTIONS(1542),
    [sym_crate] = ACTIONS(1542),
    [sym_metavariable] = ACTIONS(1544),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [610] = {
    [sym_function_modifiers] = STATE(2472),
    [sym_higher_ranked_trait_bound] = STATE(1537),
    [sym_removed_trait_bound] = STATE(1537),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(1540),
    [sym_bracketed_type] = STATE(2378),
    [sym_lifetime] = STATE(1541),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2379),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1390),
    [sym_scoped_identifier] = STATE(2157),
    [sym_scoped_type_identifier] = STATE(1331),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(2309),
    [anon_sym_LPAREN] = ACTIONS(856),
    [anon_sym_LBRACK] = ACTIONS(860),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_QMARK] = ACTIONS(2437),
    [anon_sym_u8] = ACTIONS(862),
    [anon_sym_i8] = ACTIONS(862),
    [anon_sym_u16] = ACTIONS(862),
    [anon_sym_i16] = ACTIONS(862),
    [anon_sym_u32] = ACTIONS(862),
    [anon_sym_i32] = ACTIONS(862),
    [anon_sym_u64] = ACTIONS(862),
    [anon_sym_i64] = ACTIONS(862),
    [anon_sym_u128] = ACTIONS(862),
    [anon_sym_i128] = ACTIONS(862),
    [anon_sym_isize] = ACTIONS(862),
    [anon_sym_usize] = ACTIONS(862),
    [anon_sym_f32] = ACTIONS(862),
    [anon_sym_f64] = ACTIONS(862),
    [anon_sym_bool] = ACTIONS(862),
    [anon_sym_str] = ACTIONS(862),
    [anon_sym_char] = ACTIONS(862),
    [anon_sym_SQUOTE] = ACTIONS(2313),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(664),
    [anon_sym_default] = ACTIONS(864),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(2439),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(866),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(870),
    [anon_sym_AMP] = ACTIONS(872),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(876),
    [sym_super] = ACTIONS(876),
    [sym_crate] = ACTIONS(876),
    [sym_metavariable] = ACTIONS(878),
    [sym_block_comment] = ACTIONS(3),
  },
  [611] = {
    [sym_bracketed_type] = STATE(2489),
    [sym_generic_type] = STATE(2488),
    [sym_generic_type_with_turbofish] = STATE(2487),
    [sym_macro_invocation] = STATE(1422),
    [sym_scoped_identifier] = STATE(1337),
    [sym_scoped_type_identifier] = STATE(2013),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(2187),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [sym_identifier] = ACTIONS(1524),
    [anon_sym_LPAREN] = ACTIONS(1526),
    [anon_sym_LBRACK] = ACTIONS(1530),
    [anon_sym_u8] = ACTIONS(1532),
    [anon_sym_i8] = ACTIONS(1532),
    [anon_sym_u16] = ACTIONS(1532),
    [anon_sym_i16] = ACTIONS(1532),
    [anon_sym_u32] = ACTIONS(1532),
    [anon_sym_i32] = ACTIONS(1532),
    [anon_sym_u64] = ACTIONS(1532),
    [anon_sym_i64] = ACTIONS(1532),
    [anon_sym_u128] = ACTIONS(1532),
    [anon_sym_i128] = ACTIONS(1532),
    [anon_sym_isize] = ACTIONS(1532),
    [anon_sym_usize] = ACTIONS(1532),
    [anon_sym_f32] = ACTIONS(1532),
    [anon_sym_f64] = ACTIONS(1532),
    [anon_sym_bool] = ACTIONS(1532),
    [anon_sym_str] = ACTIONS(1532),
    [anon_sym_char] = ACTIONS(1532),
    [anon_sym_const] = ACTIONS(1534),
    [anon_sym_default] = ACTIONS(1536),
    [anon_sym_union] = ACTIONS(1536),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(1538),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(1540),
    [sym_mutable_specifier] = ACTIONS(796),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1542),
    [sym_super] = ACTIONS(1542),
    [sym_crate] = ACTIONS(1542),
    [sym_metavariable] = ACTIONS(1544),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [612] = {
    [sym_bracketed_type] = STATE(2489),
    [sym_generic_type] = STATE(2488),
    [sym_generic_type_with_turbofish] = STATE(2487),
    [sym_macro_invocation] = STATE(1422),
    [sym_scoped_identifier] = STATE(1337),
    [sym_scoped_type_identifier] = STATE(2013),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(2277),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [sym_identifier] = ACTIONS(1524),
    [anon_sym_LPAREN] = ACTIONS(1526),
    [anon_sym_LBRACK] = ACTIONS(1530),
    [anon_sym_u8] = ACTIONS(1532),
    [anon_sym_i8] = ACTIONS(1532),
    [anon_sym_u16] = ACTIONS(1532),
    [anon_sym_i16] = ACTIONS(1532),
    [anon_sym_u32] = ACTIONS(1532),
    [anon_sym_i32] = ACTIONS(1532),
    [anon_sym_u64] = ACTIONS(1532),
    [anon_sym_i64] = ACTIONS(1532),
    [anon_sym_u128] = ACTIONS(1532),
    [anon_sym_i128] = ACTIONS(1532),
    [anon_sym_isize] = ACTIONS(1532),
    [anon_sym_usize] = ACTIONS(1532),
    [anon_sym_f32] = ACTIONS(1532),
    [anon_sym_f64] = ACTIONS(1532),
    [anon_sym_bool] = ACTIONS(1532),
    [anon_sym_str] = ACTIONS(1532),
    [anon_sym_char] = ACTIONS(1532),
    [anon_sym_const] = ACTIONS(1534),
    [anon_sym_default] = ACTIONS(1536),
    [anon_sym_union] = ACTIONS(1536),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(1538),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(1540),
    [sym_mutable_specifier] = ACTIONS(796),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1542),
    [sym_super] = ACTIONS(1542),
    [sym_crate] = ACTIONS(1542),
    [sym_metavariable] = ACTIONS(1544),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [613] = {
    [sym_bracketed_type] = STATE(2489),
    [sym_generic_type] = STATE(2488),
    [sym_generic_type_with_turbofish] = STATE(2487),
    [sym_macro_invocation] = STATE(1422),
    [sym_scoped_identifier] = STATE(1337),
    [sym_scoped_type_identifier] = STATE(2013),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1735),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [sym_identifier] = ACTIONS(1524),
    [anon_sym_LPAREN] = ACTIONS(1526),
    [anon_sym_LBRACK] = ACTIONS(1530),
    [anon_sym_u8] = ACTIONS(1532),
    [anon_sym_i8] = ACTIONS(1532),
    [anon_sym_u16] = ACTIONS(1532),
    [anon_sym_i16] = ACTIONS(1532),
    [anon_sym_u32] = ACTIONS(1532),
    [anon_sym_i32] = ACTIONS(1532),
    [anon_sym_u64] = ACTIONS(1532),
    [anon_sym_i64] = ACTIONS(1532),
    [anon_sym_u128] = ACTIONS(1532),
    [anon_sym_i128] = ACTIONS(1532),
    [anon_sym_isize] = ACTIONS(1532),
    [anon_sym_usize] = ACTIONS(1532),
    [anon_sym_f32] = ACTIONS(1532),
    [anon_sym_f64] = ACTIONS(1532),
    [anon_sym_bool] = ACTIONS(1532),
    [anon_sym_str] = ACTIONS(1532),
    [anon_sym_char] = ACTIONS(1532),
    [anon_sym_const] = ACTIONS(1534),
    [anon_sym_default] = ACTIONS(1536),
    [anon_sym_union] = ACTIONS(1536),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(1538),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(1540),
    [sym_mutable_specifier] = ACTIONS(796),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1542),
    [sym_super] = ACTIONS(1542),
    [sym_crate] = ACTIONS(1542),
    [sym_metavariable] = ACTIONS(1544),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [614] = {
    [sym_identifier] = ACTIONS(2397),
    [anon_sym_LPAREN] = ACTIONS(2399),
    [anon_sym_RPAREN] = ACTIONS(2399),
    [anon_sym_LBRACE] = ACTIONS(2399),
    [anon_sym_RBRACE] = ACTIONS(2399),
    [anon_sym_LBRACK] = ACTIONS(2399),
    [anon_sym_RBRACK] = ACTIONS(2399),
    [anon_sym_DOLLAR] = ACTIONS(2399),
    [anon_sym_u8] = ACTIONS(2397),
    [anon_sym_i8] = ACTIONS(2397),
    [anon_sym_u16] = ACTIONS(2397),
    [anon_sym_i16] = ACTIONS(2397),
    [anon_sym_u32] = ACTIONS(2397),
    [anon_sym_i32] = ACTIONS(2397),
    [anon_sym_u64] = ACTIONS(2397),
    [anon_sym_i64] = ACTIONS(2397),
    [anon_sym_u128] = ACTIONS(2397),
    [anon_sym_i128] = ACTIONS(2397),
    [anon_sym_isize] = ACTIONS(2397),
    [anon_sym_usize] = ACTIONS(2397),
    [anon_sym_f32] = ACTIONS(2397),
    [anon_sym_f64] = ACTIONS(2397),
    [anon_sym_bool] = ACTIONS(2397),
    [anon_sym_str] = ACTIONS(2397),
    [anon_sym_char] = ACTIONS(2397),
    [aux_sym__non_special_token_token1] = ACTIONS(2397),
    [anon_sym_SQUOTE] = ACTIONS(2397),
    [anon_sym_as] = ACTIONS(2397),
    [anon_sym_async] = ACTIONS(2397),
    [anon_sym_await] = ACTIONS(2397),
    [anon_sym_break] = ACTIONS(2397),
    [anon_sym_const] = ACTIONS(2397),
    [anon_sym_continue] = ACTIONS(2397),
    [anon_sym_default] = ACTIONS(2397),
    [anon_sym_enum] = ACTIONS(2397),
    [anon_sym_fn] = ACTIONS(2397),
    [anon_sym_for] = ACTIONS(2397),
    [anon_sym_if] = ACTIONS(2397),
    [anon_sym_impl] = ACTIONS(2397),
    [anon_sym_let] = ACTIONS(2397),
    [anon_sym_loop] = ACTIONS(2397),
    [anon_sym_match] = ACTIONS(2397),
    [anon_sym_mod] = ACTIONS(2397),
    [anon_sym_pub] = ACTIONS(2397),
    [anon_sym_return] = ACTIONS(2397),
    [anon_sym_static] = ACTIONS(2397),
    [anon_sym_struct] = ACTIONS(2397),
    [anon_sym_trait] = ACTIONS(2397),
    [anon_sym_type] = ACTIONS(2397),
    [anon_sym_union] = ACTIONS(2397),
    [anon_sym_unsafe] = ACTIONS(2397),
    [anon_sym_use] = ACTIONS(2397),
    [anon_sym_where] = ACTIONS(2397),
    [anon_sym_while] = ACTIONS(2397),
    [sym_mutable_specifier] = ACTIONS(2397),
    [sym_integer_literal] = ACTIONS(2399),
    [aux_sym_string_literal_token1] = ACTIONS(2399),
    [sym_char_literal] = ACTIONS(2399),
    [anon_sym_true] = ACTIONS(2397),
    [anon_sym_false] = ACTIONS(2397),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(2397),
    [sym_super] = ACTIONS(2397),
    [sym_crate] = ACTIONS(2397),
    [sym_raw_string_literal] = ACTIONS(2399),
    [sym_float_literal] = ACTIONS(2399),
    [sym_block_comment] = ACTIONS(3),
  },
  [615] = {
    [sym_identifier] = ACTIONS(616),
    [anon_sym_LPAREN] = ACTIONS(614),
    [anon_sym_RPAREN] = ACTIONS(614),
    [anon_sym_LBRACE] = ACTIONS(614),
    [anon_sym_RBRACE] = ACTIONS(614),
    [anon_sym_LBRACK] = ACTIONS(614),
    [anon_sym_RBRACK] = ACTIONS(614),
    [anon_sym_DOLLAR] = ACTIONS(614),
    [anon_sym_u8] = ACTIONS(616),
    [anon_sym_i8] = ACTIONS(616),
    [anon_sym_u16] = ACTIONS(616),
    [anon_sym_i16] = ACTIONS(616),
    [anon_sym_u32] = ACTIONS(616),
    [anon_sym_i32] = ACTIONS(616),
    [anon_sym_u64] = ACTIONS(616),
    [anon_sym_i64] = ACTIONS(616),
    [anon_sym_u128] = ACTIONS(616),
    [anon_sym_i128] = ACTIONS(616),
    [anon_sym_isize] = ACTIONS(616),
    [anon_sym_usize] = ACTIONS(616),
    [anon_sym_f32] = ACTIONS(616),
    [anon_sym_f64] = ACTIONS(616),
    [anon_sym_bool] = ACTIONS(616),
    [anon_sym_str] = ACTIONS(616),
    [anon_sym_char] = ACTIONS(616),
    [aux_sym__non_special_token_token1] = ACTIONS(616),
    [anon_sym_SQUOTE] = ACTIONS(616),
    [anon_sym_as] = ACTIONS(616),
    [anon_sym_async] = ACTIONS(616),
    [anon_sym_await] = ACTIONS(616),
    [anon_sym_break] = ACTIONS(616),
    [anon_sym_const] = ACTIONS(616),
    [anon_sym_continue] = ACTIONS(616),
    [anon_sym_default] = ACTIONS(616),
    [anon_sym_enum] = ACTIONS(616),
    [anon_sym_fn] = ACTIONS(616),
    [anon_sym_for] = ACTIONS(616),
    [anon_sym_if] = ACTIONS(616),
    [anon_sym_impl] = ACTIONS(616),
    [anon_sym_let] = ACTIONS(616),
    [anon_sym_loop] = ACTIONS(616),
    [anon_sym_match] = ACTIONS(616),
    [anon_sym_mod] = ACTIONS(616),
    [anon_sym_pub] = ACTIONS(616),
    [anon_sym_return] = ACTIONS(616),
    [anon_sym_static] = ACTIONS(616),
    [anon_sym_struct] = ACTIONS(616),
    [anon_sym_trait] = ACTIONS(616),
    [anon_sym_type] = ACTIONS(616),
    [anon_sym_union] = ACTIONS(616),
    [anon_sym_unsafe] = ACTIONS(616),
    [anon_sym_use] = ACTIONS(616),
    [anon_sym_where] = ACTIONS(616),
    [anon_sym_while] = ACTIONS(616),
    [sym_mutable_specifier] = ACTIONS(616),
    [sym_integer_literal] = ACTIONS(614),
    [aux_sym_string_literal_token1] = ACTIONS(614),
    [sym_char_literal] = ACTIONS(614),
    [anon_sym_true] = ACTIONS(616),
    [anon_sym_false] = ACTIONS(616),
    [sym_line_comment] = ACTIONS(912),
    [sym_self] = ACTIONS(616),
    [sym_super] = ACTIONS(616),
    [sym_crate] = ACTIONS(616),
    [sym_raw_string_literal] = ACTIONS(614),
    [sym_float_literal] = ACTIONS(614),
    [sym_block_comment] = ACTIONS(3),
  },
  [616] = {
    [sym_bracketed_type] = STATE(2489),
    [sym_generic_type] = STATE(2488),
    [sym_generic_type_with_turbofish] = STATE(2487),
    [sym_macro_invocation] = STATE(1422),
    [sym_scoped_identifier] = STATE(1337),
    [sym_scoped_type_identifier] = STATE(2013),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(1751),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [sym_identifier] = ACTIONS(1524),
    [anon_sym_LPAREN] = ACTIONS(1526),
    [anon_sym_LBRACK] = ACTIONS(1530),
    [anon_sym_u8] = ACTIONS(1532),
    [anon_sym_i8] = ACTIONS(1532),
    [anon_sym_u16] = ACTIONS(1532),
    [anon_sym_i16] = ACTIONS(1532),
    [anon_sym_u32] = ACTIONS(1532),
    [anon_sym_i32] = ACTIONS(1532),
    [anon_sym_u64] = ACTIONS(1532),
    [anon_sym_i64] = ACTIONS(1532),
    [anon_sym_u128] = ACTIONS(1532),
    [anon_sym_i128] = ACTIONS(1532),
    [anon_sym_isize] = ACTIONS(1532),
    [anon_sym_usize] = ACTIONS(1532),
    [anon_sym_f32] = ACTIONS(1532),
    [anon_sym_f64] = ACTIONS(1532),
    [anon_sym_bool] = ACTIONS(1532),
    [anon_sym_str] = ACTIONS(1532),
    [anon_sym_char] = ACTIONS(1532),
    [anon_sym_const] = ACTIONS(1534),
    [anon_sym_default] = ACTIONS(1536),
    [anon_sym_union] = ACTIONS(1536),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(1538),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(1540),
    [sym_mutable_specifier] = ACTIONS(2447),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1542),
    [sym_super] = ACTIONS(1542),
    [sym_crate] = ACTIONS(1542),
    [sym_metavariable] = ACTIONS(1544),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [617] = {
    [sym_bracketed_type] = STATE(2489),
    [sym_generic_type] = STATE(2488),
    [sym_generic_type_with_turbofish] = STATE(2487),
    [sym_macro_invocation] = STATE(1422),
    [sym_scoped_identifier] = STATE(1337),
    [sym_scoped_type_identifier] = STATE(2013),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(2197),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [sym_identifier] = ACTIONS(1524),
    [anon_sym_LPAREN] = ACTIONS(1526),
    [anon_sym_LBRACK] = ACTIONS(1530),
    [anon_sym_u8] = ACTIONS(1532),
    [anon_sym_i8] = ACTIONS(1532),
    [anon_sym_u16] = ACTIONS(1532),
    [anon_sym_i16] = ACTIONS(1532),
    [anon_sym_u32] = ACTIONS(1532),
    [anon_sym_i32] = ACTIONS(1532),
    [anon_sym_u64] = ACTIONS(1532),
    [anon_sym_i64] = ACTIONS(1532),
    [anon_sym_u128] = ACTIONS(1532),
    [anon_sym_i128] = ACTIONS(1532),
    [anon_sym_isize] = ACTIONS(1532),
    [anon_sym_usize] = ACTIONS(1532),
    [anon_sym_f32] = ACTIONS(1532),
    [anon_sym_f64] = ACTIONS(1532),
    [anon_sym_bool] = ACTIONS(1532),
    [anon_sym_str] = ACTIONS(1532),
    [anon_sym_char] = ACTIONS(1532),
    [anon_sym_const] = ACTIONS(1534),
    [anon_sym_default] = ACTIONS(1536),
    [anon_sym_union] = ACTIONS(1536),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(1538),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(1540),
    [sym_mutable_specifier] = ACTIONS(796),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1542),
    [sym_super] = ACTIONS(1542),
    [sym_crate] = ACTIONS(1542),
    [sym_metavariable] = ACTIONS(1544),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [618] = {
    [sym_bracketed_type] = STATE(2489),
    [sym_generic_type] = STATE(2488),
    [sym_generic_type_with_turbofish] = STATE(2487),
    [sym_macro_invocation] = STATE(1422),
    [sym_scoped_identifier] = STATE(1337),
    [sym_scoped_type_identifier] = STATE(2013),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(2273),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [sym_identifier] = ACTIONS(1524),
    [anon_sym_LPAREN] = ACTIONS(1526),
    [anon_sym_LBRACK] = ACTIONS(1530),
    [anon_sym_u8] = ACTIONS(1532),
    [anon_sym_i8] = ACTIONS(1532),
    [anon_sym_u16] = ACTIONS(1532),
    [anon_sym_i16] = ACTIONS(1532),
    [anon_sym_u32] = ACTIONS(1532),
    [anon_sym_i32] = ACTIONS(1532),
    [anon_sym_u64] = ACTIONS(1532),
    [anon_sym_i64] = ACTIONS(1532),
    [anon_sym_u128] = ACTIONS(1532),
    [anon_sym_i128] = ACTIONS(1532),
    [anon_sym_isize] = ACTIONS(1532),
    [anon_sym_usize] = ACTIONS(1532),
    [anon_sym_f32] = ACTIONS(1532),
    [anon_sym_f64] = ACTIONS(1532),
    [anon_sym_bool] = ACTIONS(1532),
    [anon_sym_str] = ACTIONS(1532),
    [anon_sym_char] = ACTIONS(1532),
    [anon_sym_const] = ACTIONS(1534),
    [anon_sym_default] = ACTIONS(1536),
    [anon_sym_union] = ACTIONS(1536),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(1538),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(1540),
    [sym_mutable_specifier] = ACTIONS(796),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1542),
    [sym_super] = ACTIONS(1542),
    [sym_crate] = ACTIONS(1542),
    [sym_metavariable] = ACTIONS(1544),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [619] = {
    [sym_bracketed_type] = STATE(2489),
    [sym_generic_type] = STATE(2488),
    [sym_generic_type_with_turbofish] = STATE(2487),
    [sym_macro_invocation] = STATE(1422),
    [sym_scoped_identifier] = STATE(1337),
    [sym_scoped_type_identifier] = STATE(2013),
    [sym_const_block] = STATE(1422),
    [sym__pattern] = STATE(2198),
    [sym_tuple_pattern] = STATE(1422),
    [sym_slice_pattern] = STATE(1422),
    [sym_tuple_struct_pattern] = STATE(1422),
    [sym_struct_pattern] = STATE(1422),
    [sym_remaining_field_pattern] = STATE(1422),
    [sym_mut_pattern] = STATE(1422),
    [sym_range_pattern] = STATE(1422),
    [sym_ref_pattern] = STATE(1422),
    [sym_captured_pattern] = STATE(1422),
    [sym_reference_pattern] = STATE(1422),
    [sym_or_pattern] = STATE(1422),
    [sym__literal_pattern] = STATE(1396),
    [sym_negative_literal] = STATE(1397),
    [sym_string_literal] = STATE(1397),
    [sym_boolean_literal] = STATE(1397),
    [sym_identifier] = ACTIONS(1524),
    [anon_sym_LPAREN] = ACTIONS(1526),
    [anon_sym_LBRACK] = ACTIONS(1530),
    [anon_sym_u8] = ACTIONS(1532),
    [anon_sym_i8] = ACTIONS(1532),
    [anon_sym_u16] = ACTIONS(1532),
    [anon_sym_i16] = ACTIONS(1532),
    [anon_sym_u32] = ACTIONS(1532),
    [anon_sym_i32] = ACTIONS(1532),
    [anon_sym_u64] = ACTIONS(1532),
    [anon_sym_i64] = ACTIONS(1532),
    [anon_sym_u128] = ACTIONS(1532),
    [anon_sym_i128] = ACTIONS(1532),
    [anon_sym_isize] = ACTIONS(1532),
    [anon_sym_usize] = ACTIONS(1532),
    [anon_sym_f32] = ACTIONS(1532),
    [anon_sym_f64] = ACTIONS(1532),
    [anon_sym_bool] = ACTIONS(1532),
    [anon_sym_str] = ACTIONS(1532),
    [anon_sym_char] = ACTIONS(1532),
    [anon_sym_const] = ACTIONS(1534),
    [anon_sym_default] = ACTIONS(1536),
    [anon_sym_union] = ACTIONS(1536),
    [anon_sym_ref] = ACTIONS(686),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(1538),
    [anon_sym__] = ACTIONS(792),
    [anon_sym_AMP] = ACTIONS(1540),
    [sym_mutable_specifier] = ACTIONS(796),
    [anon_sym_DOT_DOT] = ACTIONS(798),
    [anon_sym_DASH] = ACTIONS(702),
    [sym_integer_literal] = ACTIONS(704),
    [aux_sym_string_literal_token1] = ACTIONS(706),
    [sym_char_literal] = ACTIONS(704),
    [anon_sym_true] = ACTIONS(708),
    [anon_sym_false] = ACTIONS(708),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(1542),
    [sym_super] = ACTIONS(1542),
    [sym_crate] = ACTIONS(1542),
    [sym_metavariable] = ACTIONS(1544),
    [sym_raw_string_literal] = ACTIONS(704),
    [sym_float_literal] = ACTIONS(704),
    [sym_block_comment] = ACTIONS(3),
  },
  [620] = {
    [sym_function_modifiers] = STATE(2472),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(1398),
    [sym_bracketed_type] = STATE(2378),
    [sym_lifetime] = STATE(2471),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2379),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1390),
    [sym_scoped_identifier] = STATE(2157),
    [sym_scoped_type_identifier] = STATE(1331),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(2309),
    [anon_sym_LPAREN] = ACTIONS(856),
    [anon_sym_LBRACK] = ACTIONS(860),
    [anon_sym_PLUS] = ACTIONS(2449),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(862),
    [anon_sym_i8] = ACTIONS(862),
    [anon_sym_u16] = ACTIONS(862),
    [anon_sym_i16] = ACTIONS(862),
    [anon_sym_u32] = ACTIONS(862),
    [anon_sym_i32] = ACTIONS(862),
    [anon_sym_u64] = ACTIONS(862),
    [anon_sym_i64] = ACTIONS(862),
    [anon_sym_u128] = ACTIONS(862),
    [anon_sym_i128] = ACTIONS(862),
    [anon_sym_isize] = ACTIONS(862),
    [anon_sym_usize] = ACTIONS(862),
    [anon_sym_f32] = ACTIONS(862),
    [anon_sym_f64] = ACTIONS(862),
    [anon_sym_bool] = ACTIONS(862),
    [anon_sym_str] = ACTIONS(862),
    [anon_sym_char] = ACTIONS(862),
    [anon_sym_SQUOTE] = ACTIONS(2313),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(664),
    [anon_sym_default] = ACTIONS(864),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(866),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(870),
    [anon_sym_AMP] = ACTIONS(872),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_mutable_specifier] = ACTIONS(2451),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(2453),
    [sym_super] = ACTIONS(876),
    [sym_crate] = ACTIONS(876),
    [sym_metavariable] = ACTIONS(878),
    [sym_block_comment] = ACTIONS(3),
  },
  [621] = {
    [sym_attribute_item] = STATE(621),
    [aux_sym_enum_variant_list_repeat1] = STATE(621),
    [sym_identifier] = ACTIONS(2455),
    [anon_sym_LPAREN] = ACTIONS(2457),
    [anon_sym_LBRACE] = ACTIONS(2457),
    [anon_sym_LBRACK] = ACTIONS(2457),
    [anon_sym_RBRACK] = ACTIONS(2457),
    [anon_sym_STAR] = ACTIONS(2457),
    [anon_sym_u8] = ACTIONS(2455),
    [anon_sym_i8] = ACTIONS(2455),
    [anon_sym_u16] = ACTIONS(2455),
    [anon_sym_i16] = ACTIONS(2455),
    [anon_sym_u32] = ACTIONS(2455),
    [anon_sym_i32] = ACTIONS(2455),
    [anon_sym_u64] = ACTIONS(2455),
    [anon_sym_i64] = ACTIONS(2455),
    [anon_sym_u128] = ACTIONS(2455),
    [anon_sym_i128] = ACTIONS(2455),
    [anon_sym_isize] = ACTIONS(2455),
    [anon_sym_usize] = ACTIONS(2455),
    [anon_sym_f32] = ACTIONS(2455),
    [anon_sym_f64] = ACTIONS(2455),
    [anon_sym_bool] = ACTIONS(2455),
    [anon_sym_str] = ACTIONS(2455),
    [anon_sym_char] = ACTIONS(2455),
    [anon_sym_SQUOTE] = ACTIONS(2455),
    [anon_sym_async] = ACTIONS(2455),
    [anon_sym_break] = ACTIONS(2455),
    [anon_sym_const] = ACTIONS(2455),
    [anon_sym_continue] = ACTIONS(2455),
    [anon_sym_default] = ACTIONS(2455),
    [anon_sym_for] = ACTIONS(2455),
    [anon_sym_if] = ACTIONS(2455),
    [anon_sym_loop] = ACTIONS(2455),
    [anon_sym_match] = ACTIONS(2455),
    [anon_sym_return] = ACTIONS(2455),
    [anon_sym_union] = ACTIONS(2455),
    [anon_sym_unsafe] = ACTIONS(2455),
    [anon_sym_while] = ACTIONS(2455),
    [anon_sym_POUND] = ACTIONS(2459),
    [anon_sym_BANG] = ACTIONS(2457),
    [anon_sym_COMMA] = ACTIONS(2457),
    [anon_sym_ref] = ACTIONS(2455),
    [anon_sym_LT] = ACTIONS(2457),
    [anon_sym_COLON_COLON] = ACTIONS(2457),
    [anon_sym__] = ACTIONS(2455),
    [anon_sym_AMP] = ACTIONS(2457),
    [sym_mutable_specifier] = ACTIONS(2455),
    [anon_sym_DOT_DOT] = ACTIONS(2457),
    [anon_sym_DASH] = ACTIONS(2457),
    [anon_sym_PIPE] = ACTIONS(2457),
    [anon_sym_yield] = ACTIONS(2455),
    [anon_sym_move] = ACTIONS(2455),
    [sym_integer_literal] = ACTIONS(2457),
    [aux_sym_string_literal_token1] = ACTIONS(2457),
    [sym_char_literal] = ACTIONS(2457),
    [anon_sym_true] = ACTIONS(2455),
    [anon_sym_false] = ACTIONS(2455),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(2455),
    [sym_super] = ACTIONS(2455),
    [sym_crate] = ACTIONS(2455),
    [sym_metavariable] = ACTIONS(2457),
    [sym_raw_string_literal] = ACTIONS(2457),
    [sym_float_literal] = ACTIONS(2457),
    [sym_block_comment] = ACTIONS(3),
  },
  [622] = {
    [sym_function_modifiers] = STATE(2472),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(2289),
    [sym_bracketed_type] = STATE(2378),
    [sym_lifetime] = STATE(2471),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2379),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1390),
    [sym_scoped_identifier] = STATE(2157),
    [sym_scoped_type_identifier] = STATE(1331),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(2309),
    [anon_sym_LPAREN] = ACTIONS(856),
    [anon_sym_LBRACK] = ACTIONS(860),
    [anon_sym_PLUS] = ACTIONS(2449),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(862),
    [anon_sym_i8] = ACTIONS(862),
    [anon_sym_u16] = ACTIONS(862),
    [anon_sym_i16] = ACTIONS(862),
    [anon_sym_u32] = ACTIONS(862),
    [anon_sym_i32] = ACTIONS(862),
    [anon_sym_u64] = ACTIONS(862),
    [anon_sym_i64] = ACTIONS(862),
    [anon_sym_u128] = ACTIONS(862),
    [anon_sym_i128] = ACTIONS(862),
    [anon_sym_isize] = ACTIONS(862),
    [anon_sym_usize] = ACTIONS(862),
    [anon_sym_f32] = ACTIONS(862),
    [anon_sym_f64] = ACTIONS(862),
    [anon_sym_bool] = ACTIONS(862),
    [anon_sym_str] = ACTIONS(862),
    [anon_sym_char] = ACTIONS(862),
    [anon_sym_SQUOTE] = ACTIONS(2313),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(664),
    [anon_sym_default] = ACTIONS(864),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(866),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(870),
    [anon_sym_AMP] = ACTIONS(872),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_mutable_specifier] = ACTIONS(2462),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(876),
    [sym_super] = ACTIONS(876),
    [sym_crate] = ACTIONS(876),
    [sym_metavariable] = ACTIONS(878),
    [sym_block_comment] = ACTIONS(3),
  },
  [623] = {
    [sym_function_modifiers] = STATE(2386),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(941),
    [sym_bracketed_type] = STATE(2491),
    [sym_lifetime] = STATE(2513),
    [sym_array_type] = STATE(1040),
    [sym_for_lifetimes] = STATE(1195),
    [sym_function_type] = STATE(1040),
    [sym_tuple_type] = STATE(1040),
    [sym_unit_type] = STATE(1040),
    [sym_generic_type] = STATE(790),
    [sym_generic_type_with_turbofish] = STATE(2492),
    [sym_bounded_type] = STATE(1040),
    [sym_reference_type] = STATE(1040),
    [sym_pointer_type] = STATE(1040),
    [sym_empty_type] = STATE(1040),
    [sym_abstract_type] = STATE(1040),
    [sym_dynamic_type] = STATE(1040),
    [sym_macro_invocation] = STATE(1040),
    [sym_scoped_identifier] = STATE(2216),
    [sym_scoped_type_identifier] = STATE(760),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(2464),
    [anon_sym_LPAREN] = ACTIONS(2466),
    [anon_sym_LBRACK] = ACTIONS(2468),
    [anon_sym_PLUS] = ACTIONS(2470),
    [anon_sym_STAR] = ACTIONS(2472),
    [anon_sym_u8] = ACTIONS(2474),
    [anon_sym_i8] = ACTIONS(2474),
    [anon_sym_u16] = ACTIONS(2474),
    [anon_sym_i16] = ACTIONS(2474),
    [anon_sym_u32] = ACTIONS(2474),
    [anon_sym_i32] = ACTIONS(2474),
    [anon_sym_u64] = ACTIONS(2474),
    [anon_sym_i64] = ACTIONS(2474),
    [anon_sym_u128] = ACTIONS(2474),
    [anon_sym_i128] = ACTIONS(2474),
    [anon_sym_isize] = ACTIONS(2474),
    [anon_sym_usize] = ACTIONS(2474),
    [anon_sym_f32] = ACTIONS(2474),
    [anon_sym_f64] = ACTIONS(2474),
    [anon_sym_bool] = ACTIONS(2474),
    [anon_sym_str] = ACTIONS(2474),
    [anon_sym_char] = ACTIONS(2474),
    [anon_sym_SQUOTE] = ACTIONS(2313),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(664),
    [anon_sym_default] = ACTIONS(2476),
    [anon_sym_fn] = ACTIONS(2478),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(2480),
    [anon_sym_union] = ACTIONS(2482),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_BANG] = ACTIONS(2484),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(2486),
    [anon_sym_AMP] = ACTIONS(2488),
    [anon_sym_dyn] = ACTIONS(2490),
    [sym_mutable_specifier] = ACTIONS(2492),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(2494),
    [sym_super] = ACTIONS(2494),
    [sym_crate] = ACTIONS(2494),
    [sym_metavariable] = ACTIONS(2496),
    [sym_block_comment] = ACTIONS(3),
  },
  [624] = {
    [sym_function_modifiers] = STATE(2472),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(1398),
    [sym_bracketed_type] = STATE(2378),
    [sym_lifetime] = STATE(2471),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2379),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1390),
    [sym_scoped_identifier] = STATE(2157),
    [sym_scoped_type_identifier] = STATE(1331),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(2309),
    [anon_sym_LPAREN] = ACTIONS(856),
    [anon_sym_LBRACK] = ACTIONS(860),
    [anon_sym_PLUS] = ACTIONS(2449),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(862),
    [anon_sym_i8] = ACTIONS(862),
    [anon_sym_u16] = ACTIONS(862),
    [anon_sym_i16] = ACTIONS(862),
    [anon_sym_u32] = ACTIONS(862),
    [anon_sym_i32] = ACTIONS(862),
    [anon_sym_u64] = ACTIONS(862),
    [anon_sym_i64] = ACTIONS(862),
    [anon_sym_u128] = ACTIONS(862),
    [anon_sym_i128] = ACTIONS(862),
    [anon_sym_isize] = ACTIONS(862),
    [anon_sym_usize] = ACTIONS(862),
    [anon_sym_f32] = ACTIONS(862),
    [anon_sym_f64] = ACTIONS(862),
    [anon_sym_bool] = ACTIONS(862),
    [anon_sym_str] = ACTIONS(862),
    [anon_sym_char] = ACTIONS(862),
    [anon_sym_SQUOTE] = ACTIONS(2313),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(664),
    [anon_sym_default] = ACTIONS(864),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(866),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(870),
    [anon_sym_AMP] = ACTIONS(872),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_mutable_specifier] = ACTIONS(2498),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(876),
    [sym_super] = ACTIONS(876),
    [sym_crate] = ACTIONS(876),
    [sym_metavariable] = ACTIONS(878),
    [sym_block_comment] = ACTIONS(3),
  },
  [625] = {
    [sym_function_modifiers] = STATE(2472),
    [sym_type_parameters] = STATE(704),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(1668),
    [sym_bracketed_type] = STATE(2378),
    [sym_lifetime] = STATE(2471),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1639),
    [sym_generic_type_with_turbofish] = STATE(2379),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1390),
    [sym_scoped_identifier] = STATE(2157),
    [sym_scoped_type_identifier] = STATE(1499),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(2500),
    [anon_sym_LPAREN] = ACTIONS(856),
    [anon_sym_LBRACK] = ACTIONS(860),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(862),
    [anon_sym_i8] = ACTIONS(862),
    [anon_sym_u16] = ACTIONS(862),
    [anon_sym_i16] = ACTIONS(862),
    [anon_sym_u32] = ACTIONS(862),
    [anon_sym_i32] = ACTIONS(862),
    [anon_sym_u64] = ACTIONS(862),
    [anon_sym_i64] = ACTIONS(862),
    [anon_sym_u128] = ACTIONS(862),
    [anon_sym_i128] = ACTIONS(862),
    [anon_sym_isize] = ACTIONS(862),
    [anon_sym_usize] = ACTIONS(862),
    [anon_sym_f32] = ACTIONS(862),
    [anon_sym_f64] = ACTIONS(862),
    [anon_sym_bool] = ACTIONS(862),
    [anon_sym_str] = ACTIONS(862),
    [anon_sym_char] = ACTIONS(862),
    [anon_sym_SQUOTE] = ACTIONS(2313),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(664),
    [anon_sym_default] = ACTIONS(864),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(866),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_LT] = ACTIONS(2502),
    [anon_sym_COLON_COLON] = ACTIONS(870),
    [anon_sym_AMP] = ACTIONS(872),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(876),
    [sym_super] = ACTIONS(876),
    [sym_crate] = ACTIONS(876),
    [sym_metavariable] = ACTIONS(878),
    [sym_block_comment] = ACTIONS(3),
  },
  [626] = {
    [sym_function_modifiers] = STATE(2472),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(2305),
    [sym_bracketed_type] = STATE(2378),
    [sym_lifetime] = STATE(622),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2379),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1390),
    [sym_scoped_identifier] = STATE(2157),
    [sym_scoped_type_identifier] = STATE(1331),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(2309),
    [anon_sym_LPAREN] = ACTIONS(856),
    [anon_sym_LBRACK] = ACTIONS(860),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(862),
    [anon_sym_i8] = ACTIONS(862),
    [anon_sym_u16] = ACTIONS(862),
    [anon_sym_i16] = ACTIONS(862),
    [anon_sym_u32] = ACTIONS(862),
    [anon_sym_i32] = ACTIONS(862),
    [anon_sym_u64] = ACTIONS(862),
    [anon_sym_i64] = ACTIONS(862),
    [anon_sym_u128] = ACTIONS(862),
    [anon_sym_i128] = ACTIONS(862),
    [anon_sym_isize] = ACTIONS(862),
    [anon_sym_usize] = ACTIONS(862),
    [anon_sym_f32] = ACTIONS(862),
    [anon_sym_f64] = ACTIONS(862),
    [anon_sym_bool] = ACTIONS(862),
    [anon_sym_str] = ACTIONS(862),
    [anon_sym_char] = ACTIONS(862),
    [anon_sym_SQUOTE] = ACTIONS(2504),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(664),
    [anon_sym_default] = ACTIONS(864),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(866),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(870),
    [anon_sym_AMP] = ACTIONS(872),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_mutable_specifier] = ACTIONS(2506),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(876),
    [sym_super] = ACTIONS(876),
    [sym_crate] = ACTIONS(876),
    [sym_metavariable] = ACTIONS(878),
    [sym_block_comment] = ACTIONS(3),
  },
  [627] = {
    [sym_function_modifiers] = STATE(2472),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(1399),
    [sym_bracketed_type] = STATE(2378),
    [sym_lifetime] = STATE(624),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2379),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1390),
    [sym_scoped_identifier] = STATE(2157),
    [sym_scoped_type_identifier] = STATE(1331),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(2309),
    [anon_sym_LPAREN] = ACTIONS(856),
    [anon_sym_LBRACK] = ACTIONS(860),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(862),
    [anon_sym_i8] = ACTIONS(862),
    [anon_sym_u16] = ACTIONS(862),
    [anon_sym_i16] = ACTIONS(862),
    [anon_sym_u32] = ACTIONS(862),
    [anon_sym_i32] = ACTIONS(862),
    [anon_sym_u64] = ACTIONS(862),
    [anon_sym_i64] = ACTIONS(862),
    [anon_sym_u128] = ACTIONS(862),
    [anon_sym_i128] = ACTIONS(862),
    [anon_sym_isize] = ACTIONS(862),
    [anon_sym_usize] = ACTIONS(862),
    [anon_sym_f32] = ACTIONS(862),
    [anon_sym_f64] = ACTIONS(862),
    [anon_sym_bool] = ACTIONS(862),
    [anon_sym_str] = ACTIONS(862),
    [anon_sym_char] = ACTIONS(862),
    [anon_sym_SQUOTE] = ACTIONS(2504),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(664),
    [anon_sym_default] = ACTIONS(864),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(866),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(870),
    [anon_sym_AMP] = ACTIONS(872),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_mutable_specifier] = ACTIONS(2508),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(876),
    [sym_super] = ACTIONS(876),
    [sym_crate] = ACTIONS(876),
    [sym_metavariable] = ACTIONS(878),
    [sym_block_comment] = ACTIONS(3),
  },
  [628] = {
    [sym_function_modifiers] = STATE(2472),
    [sym_type_parameters] = STATE(712),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(1656),
    [sym_bracketed_type] = STATE(2378),
    [sym_lifetime] = STATE(2471),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1651),
    [sym_generic_type_with_turbofish] = STATE(2379),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1390),
    [sym_scoped_identifier] = STATE(2157),
    [sym_scoped_type_identifier] = STATE(1493),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(2510),
    [anon_sym_LPAREN] = ACTIONS(856),
    [anon_sym_LBRACK] = ACTIONS(860),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(862),
    [anon_sym_i8] = ACTIONS(862),
    [anon_sym_u16] = ACTIONS(862),
    [anon_sym_i16] = ACTIONS(862),
    [anon_sym_u32] = ACTIONS(862),
    [anon_sym_i32] = ACTIONS(862),
    [anon_sym_u64] = ACTIONS(862),
    [anon_sym_i64] = ACTIONS(862),
    [anon_sym_u128] = ACTIONS(862),
    [anon_sym_i128] = ACTIONS(862),
    [anon_sym_isize] = ACTIONS(862),
    [anon_sym_usize] = ACTIONS(862),
    [anon_sym_f32] = ACTIONS(862),
    [anon_sym_f64] = ACTIONS(862),
    [anon_sym_bool] = ACTIONS(862),
    [anon_sym_str] = ACTIONS(862),
    [anon_sym_char] = ACTIONS(862),
    [anon_sym_SQUOTE] = ACTIONS(2313),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(664),
    [anon_sym_default] = ACTIONS(864),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(866),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_LT] = ACTIONS(2502),
    [anon_sym_COLON_COLON] = ACTIONS(870),
    [anon_sym_AMP] = ACTIONS(872),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(876),
    [sym_super] = ACTIONS(876),
    [sym_crate] = ACTIONS(876),
    [sym_metavariable] = ACTIONS(878),
    [sym_block_comment] = ACTIONS(3),
  },
  [629] = {
    [sym_function_modifiers] = STATE(2472),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(2051),
    [sym_bracketed_type] = STATE(2378),
    [sym_lifetime] = STATE(2471),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2379),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1390),
    [sym_scoped_identifier] = STATE(2157),
    [sym_scoped_type_identifier] = STATE(1331),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(2309),
    [anon_sym_LPAREN] = ACTIONS(856),
    [anon_sym_RPAREN] = ACTIONS(2512),
    [anon_sym_LBRACK] = ACTIONS(860),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(862),
    [anon_sym_i8] = ACTIONS(862),
    [anon_sym_u16] = ACTIONS(862),
    [anon_sym_i16] = ACTIONS(862),
    [anon_sym_u32] = ACTIONS(862),
    [anon_sym_i32] = ACTIONS(862),
    [anon_sym_u64] = ACTIONS(862),
    [anon_sym_i64] = ACTIONS(862),
    [anon_sym_u128] = ACTIONS(862),
    [anon_sym_i128] = ACTIONS(862),
    [anon_sym_isize] = ACTIONS(862),
    [anon_sym_usize] = ACTIONS(862),
    [anon_sym_f32] = ACTIONS(862),
    [anon_sym_f64] = ACTIONS(862),
    [anon_sym_bool] = ACTIONS(862),
    [anon_sym_str] = ACTIONS(862),
    [anon_sym_char] = ACTIONS(862),
    [anon_sym_SQUOTE] = ACTIONS(2313),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(664),
    [anon_sym_default] = ACTIONS(864),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(866),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(870),
    [anon_sym_AMP] = ACTIONS(872),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(876),
    [sym_super] = ACTIONS(876),
    [sym_crate] = ACTIONS(876),
    [sym_metavariable] = ACTIONS(878),
    [sym_block_comment] = ACTIONS(3),
  },
  [630] = {
    [sym_function_modifiers] = STATE(2472),
    [sym_type_parameters] = STATE(659),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(1623),
    [sym_bracketed_type] = STATE(2378),
    [sym_lifetime] = STATE(2471),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1659),
    [sym_generic_type_with_turbofish] = STATE(2379),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1390),
    [sym_scoped_identifier] = STATE(2157),
    [sym_scoped_type_identifier] = STATE(1482),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(2514),
    [anon_sym_LPAREN] = ACTIONS(856),
    [anon_sym_LBRACK] = ACTIONS(860),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(862),
    [anon_sym_i8] = ACTIONS(862),
    [anon_sym_u16] = ACTIONS(862),
    [anon_sym_i16] = ACTIONS(862),
    [anon_sym_u32] = ACTIONS(862),
    [anon_sym_i32] = ACTIONS(862),
    [anon_sym_u64] = ACTIONS(862),
    [anon_sym_i64] = ACTIONS(862),
    [anon_sym_u128] = ACTIONS(862),
    [anon_sym_i128] = ACTIONS(862),
    [anon_sym_isize] = ACTIONS(862),
    [anon_sym_usize] = ACTIONS(862),
    [anon_sym_f32] = ACTIONS(862),
    [anon_sym_f64] = ACTIONS(862),
    [anon_sym_bool] = ACTIONS(862),
    [anon_sym_str] = ACTIONS(862),
    [anon_sym_char] = ACTIONS(862),
    [anon_sym_SQUOTE] = ACTIONS(2313),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(664),
    [anon_sym_default] = ACTIONS(864),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(866),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_LT] = ACTIONS(2502),
    [anon_sym_COLON_COLON] = ACTIONS(870),
    [anon_sym_AMP] = ACTIONS(872),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(876),
    [sym_super] = ACTIONS(876),
    [sym_crate] = ACTIONS(876),
    [sym_metavariable] = ACTIONS(878),
    [sym_block_comment] = ACTIONS(3),
  },
  [631] = {
    [sym_function_modifiers] = STATE(2472),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(1827),
    [sym_bracketed_type] = STATE(2378),
    [sym_lifetime] = STATE(2471),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2379),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1390),
    [sym_scoped_identifier] = STATE(2157),
    [sym_scoped_type_identifier] = STATE(1331),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(2309),
    [anon_sym_LPAREN] = ACTIONS(856),
    [anon_sym_RPAREN] = ACTIONS(2516),
    [anon_sym_LBRACK] = ACTIONS(860),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(862),
    [anon_sym_i8] = ACTIONS(862),
    [anon_sym_u16] = ACTIONS(862),
    [anon_sym_i16] = ACTIONS(862),
    [anon_sym_u32] = ACTIONS(862),
    [anon_sym_i32] = ACTIONS(862),
    [anon_sym_u64] = ACTIONS(862),
    [anon_sym_i64] = ACTIONS(862),
    [anon_sym_u128] = ACTIONS(862),
    [anon_sym_i128] = ACTIONS(862),
    [anon_sym_isize] = ACTIONS(862),
    [anon_sym_usize] = ACTIONS(862),
    [anon_sym_f32] = ACTIONS(862),
    [anon_sym_f64] = ACTIONS(862),
    [anon_sym_bool] = ACTIONS(862),
    [anon_sym_str] = ACTIONS(862),
    [anon_sym_char] = ACTIONS(862),
    [anon_sym_SQUOTE] = ACTIONS(2313),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(664),
    [anon_sym_default] = ACTIONS(864),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(866),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(870),
    [anon_sym_AMP] = ACTIONS(872),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(876),
    [sym_super] = ACTIONS(876),
    [sym_crate] = ACTIONS(876),
    [sym_metavariable] = ACTIONS(878),
    [sym_block_comment] = ACTIONS(3),
  },
  [632] = {
    [sym_function_modifiers] = STATE(2472),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(2051),
    [sym_bracketed_type] = STATE(2378),
    [sym_lifetime] = STATE(2471),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2379),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1390),
    [sym_scoped_identifier] = STATE(2157),
    [sym_scoped_type_identifier] = STATE(1331),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(2309),
    [anon_sym_LPAREN] = ACTIONS(856),
    [anon_sym_RPAREN] = ACTIONS(2518),
    [anon_sym_LBRACK] = ACTIONS(860),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(862),
    [anon_sym_i8] = ACTIONS(862),
    [anon_sym_u16] = ACTIONS(862),
    [anon_sym_i16] = ACTIONS(862),
    [anon_sym_u32] = ACTIONS(862),
    [anon_sym_i32] = ACTIONS(862),
    [anon_sym_u64] = ACTIONS(862),
    [anon_sym_i64] = ACTIONS(862),
    [anon_sym_u128] = ACTIONS(862),
    [anon_sym_i128] = ACTIONS(862),
    [anon_sym_isize] = ACTIONS(862),
    [anon_sym_usize] = ACTIONS(862),
    [anon_sym_f32] = ACTIONS(862),
    [anon_sym_f64] = ACTIONS(862),
    [anon_sym_bool] = ACTIONS(862),
    [anon_sym_str] = ACTIONS(862),
    [anon_sym_char] = ACTIONS(862),
    [anon_sym_SQUOTE] = ACTIONS(2313),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(664),
    [anon_sym_default] = ACTIONS(864),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(866),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(870),
    [anon_sym_AMP] = ACTIONS(872),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(876),
    [sym_super] = ACTIONS(876),
    [sym_crate] = ACTIONS(876),
    [sym_metavariable] = ACTIONS(878),
    [sym_block_comment] = ACTIONS(3),
  },
  [633] = {
    [sym_function_modifiers] = STATE(2472),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(2051),
    [sym_bracketed_type] = STATE(2378),
    [sym_lifetime] = STATE(2471),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2379),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1390),
    [sym_scoped_identifier] = STATE(2157),
    [sym_scoped_type_identifier] = STATE(1331),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(2309),
    [anon_sym_LPAREN] = ACTIONS(856),
    [anon_sym_RPAREN] = ACTIONS(2520),
    [anon_sym_LBRACK] = ACTIONS(860),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(862),
    [anon_sym_i8] = ACTIONS(862),
    [anon_sym_u16] = ACTIONS(862),
    [anon_sym_i16] = ACTIONS(862),
    [anon_sym_u32] = ACTIONS(862),
    [anon_sym_i32] = ACTIONS(862),
    [anon_sym_u64] = ACTIONS(862),
    [anon_sym_i64] = ACTIONS(862),
    [anon_sym_u128] = ACTIONS(862),
    [anon_sym_i128] = ACTIONS(862),
    [anon_sym_isize] = ACTIONS(862),
    [anon_sym_usize] = ACTIONS(862),
    [anon_sym_f32] = ACTIONS(862),
    [anon_sym_f64] = ACTIONS(862),
    [anon_sym_bool] = ACTIONS(862),
    [anon_sym_str] = ACTIONS(862),
    [anon_sym_char] = ACTIONS(862),
    [anon_sym_SQUOTE] = ACTIONS(2313),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(664),
    [anon_sym_default] = ACTIONS(864),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(866),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(870),
    [anon_sym_AMP] = ACTIONS(872),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(876),
    [sym_super] = ACTIONS(876),
    [sym_crate] = ACTIONS(876),
    [sym_metavariable] = ACTIONS(878),
    [sym_block_comment] = ACTIONS(3),
  },
  [634] = {
    [sym_function_modifiers] = STATE(2472),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(2103),
    [sym_bracketed_type] = STATE(2378),
    [sym_qualified_type] = STATE(2448),
    [sym_lifetime] = STATE(2471),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2379),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1390),
    [sym_scoped_identifier] = STATE(2157),
    [sym_scoped_type_identifier] = STATE(1331),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(2309),
    [anon_sym_LPAREN] = ACTIONS(856),
    [anon_sym_LBRACK] = ACTIONS(860),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(862),
    [anon_sym_i8] = ACTIONS(862),
    [anon_sym_u16] = ACTIONS(862),
    [anon_sym_i16] = ACTIONS(862),
    [anon_sym_u32] = ACTIONS(862),
    [anon_sym_i32] = ACTIONS(862),
    [anon_sym_u64] = ACTIONS(862),
    [anon_sym_i64] = ACTIONS(862),
    [anon_sym_u128] = ACTIONS(862),
    [anon_sym_i128] = ACTIONS(862),
    [anon_sym_isize] = ACTIONS(862),
    [anon_sym_usize] = ACTIONS(862),
    [anon_sym_f32] = ACTIONS(862),
    [anon_sym_f64] = ACTIONS(862),
    [anon_sym_bool] = ACTIONS(862),
    [anon_sym_str] = ACTIONS(862),
    [anon_sym_char] = ACTIONS(862),
    [anon_sym_SQUOTE] = ACTIONS(2313),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(664),
    [anon_sym_default] = ACTIONS(864),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(866),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(870),
    [anon_sym_AMP] = ACTIONS(872),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(876),
    [sym_super] = ACTIONS(876),
    [sym_crate] = ACTIONS(876),
    [sym_metavariable] = ACTIONS(878),
    [sym_block_comment] = ACTIONS(3),
  },
  [635] = {
    [sym_function_modifiers] = STATE(2472),
    [sym_type_parameters] = STATE(723),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(1652),
    [sym_bracketed_type] = STATE(2378),
    [sym_lifetime] = STATE(2471),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1650),
    [sym_generic_type_with_turbofish] = STATE(2379),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1390),
    [sym_scoped_identifier] = STATE(2157),
    [sym_scoped_type_identifier] = STATE(1506),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(2522),
    [anon_sym_LPAREN] = ACTIONS(856),
    [anon_sym_LBRACK] = ACTIONS(860),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(862),
    [anon_sym_i8] = ACTIONS(862),
    [anon_sym_u16] = ACTIONS(862),
    [anon_sym_i16] = ACTIONS(862),
    [anon_sym_u32] = ACTIONS(862),
    [anon_sym_i32] = ACTIONS(862),
    [anon_sym_u64] = ACTIONS(862),
    [anon_sym_i64] = ACTIONS(862),
    [anon_sym_u128] = ACTIONS(862),
    [anon_sym_i128] = ACTIONS(862),
    [anon_sym_isize] = ACTIONS(862),
    [anon_sym_usize] = ACTIONS(862),
    [anon_sym_f32] = ACTIONS(862),
    [anon_sym_f64] = ACTIONS(862),
    [anon_sym_bool] = ACTIONS(862),
    [anon_sym_str] = ACTIONS(862),
    [anon_sym_char] = ACTIONS(862),
    [anon_sym_SQUOTE] = ACTIONS(2313),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(664),
    [anon_sym_default] = ACTIONS(864),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(866),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_LT] = ACTIONS(2502),
    [anon_sym_COLON_COLON] = ACTIONS(870),
    [anon_sym_AMP] = ACTIONS(872),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(876),
    [sym_super] = ACTIONS(876),
    [sym_crate] = ACTIONS(876),
    [sym_metavariable] = ACTIONS(878),
    [sym_block_comment] = ACTIONS(3),
  },
  [636] = {
    [sym_function_modifiers] = STATE(2386),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(1005),
    [sym_bracketed_type] = STATE(2491),
    [sym_lifetime] = STATE(623),
    [sym_array_type] = STATE(1040),
    [sym_for_lifetimes] = STATE(1195),
    [sym_function_type] = STATE(1040),
    [sym_tuple_type] = STATE(1040),
    [sym_unit_type] = STATE(1040),
    [sym_generic_type] = STATE(790),
    [sym_generic_type_with_turbofish] = STATE(2492),
    [sym_bounded_type] = STATE(1040),
    [sym_reference_type] = STATE(1040),
    [sym_pointer_type] = STATE(1040),
    [sym_empty_type] = STATE(1040),
    [sym_abstract_type] = STATE(1040),
    [sym_dynamic_type] = STATE(1040),
    [sym_macro_invocation] = STATE(1040),
    [sym_scoped_identifier] = STATE(2216),
    [sym_scoped_type_identifier] = STATE(760),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(2464),
    [anon_sym_LPAREN] = ACTIONS(2466),
    [anon_sym_LBRACK] = ACTIONS(2468),
    [anon_sym_STAR] = ACTIONS(2472),
    [anon_sym_u8] = ACTIONS(2474),
    [anon_sym_i8] = ACTIONS(2474),
    [anon_sym_u16] = ACTIONS(2474),
    [anon_sym_i16] = ACTIONS(2474),
    [anon_sym_u32] = ACTIONS(2474),
    [anon_sym_i32] = ACTIONS(2474),
    [anon_sym_u64] = ACTIONS(2474),
    [anon_sym_i64] = ACTIONS(2474),
    [anon_sym_u128] = ACTIONS(2474),
    [anon_sym_i128] = ACTIONS(2474),
    [anon_sym_isize] = ACTIONS(2474),
    [anon_sym_usize] = ACTIONS(2474),
    [anon_sym_f32] = ACTIONS(2474),
    [anon_sym_f64] = ACTIONS(2474),
    [anon_sym_bool] = ACTIONS(2474),
    [anon_sym_str] = ACTIONS(2474),
    [anon_sym_char] = ACTIONS(2474),
    [anon_sym_SQUOTE] = ACTIONS(2504),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(664),
    [anon_sym_default] = ACTIONS(2476),
    [anon_sym_fn] = ACTIONS(2478),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(2480),
    [anon_sym_union] = ACTIONS(2482),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_BANG] = ACTIONS(2484),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(2486),
    [anon_sym_AMP] = ACTIONS(2488),
    [anon_sym_dyn] = ACTIONS(2490),
    [sym_mutable_specifier] = ACTIONS(2524),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(2494),
    [sym_super] = ACTIONS(2494),
    [sym_crate] = ACTIONS(2494),
    [sym_metavariable] = ACTIONS(2496),
    [sym_block_comment] = ACTIONS(3),
  },
  [637] = {
    [sym_function_modifiers] = STATE(2472),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(2051),
    [sym_bracketed_type] = STATE(2378),
    [sym_lifetime] = STATE(2471),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2379),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1390),
    [sym_scoped_identifier] = STATE(2157),
    [sym_scoped_type_identifier] = STATE(1331),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(2309),
    [anon_sym_LPAREN] = ACTIONS(856),
    [anon_sym_RPAREN] = ACTIONS(2526),
    [anon_sym_LBRACK] = ACTIONS(860),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(862),
    [anon_sym_i8] = ACTIONS(862),
    [anon_sym_u16] = ACTIONS(862),
    [anon_sym_i16] = ACTIONS(862),
    [anon_sym_u32] = ACTIONS(862),
    [anon_sym_i32] = ACTIONS(862),
    [anon_sym_u64] = ACTIONS(862),
    [anon_sym_i64] = ACTIONS(862),
    [anon_sym_u128] = ACTIONS(862),
    [anon_sym_i128] = ACTIONS(862),
    [anon_sym_isize] = ACTIONS(862),
    [anon_sym_usize] = ACTIONS(862),
    [anon_sym_f32] = ACTIONS(862),
    [anon_sym_f64] = ACTIONS(862),
    [anon_sym_bool] = ACTIONS(862),
    [anon_sym_str] = ACTIONS(862),
    [anon_sym_char] = ACTIONS(862),
    [anon_sym_SQUOTE] = ACTIONS(2313),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(664),
    [anon_sym_default] = ACTIONS(864),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(866),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(870),
    [anon_sym_AMP] = ACTIONS(872),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(876),
    [sym_super] = ACTIONS(876),
    [sym_crate] = ACTIONS(876),
    [sym_metavariable] = ACTIONS(878),
    [sym_block_comment] = ACTIONS(3),
  },
  [638] = {
    [sym_function_modifiers] = STATE(2472),
    [sym_extern_modifier] = STATE(1526),
    [sym__type] = STATE(1831),
    [sym_bracketed_type] = STATE(2378),
    [sym_lifetime] = STATE(2471),
    [sym_array_type] = STATE(1390),
    [sym_for_lifetimes] = STATE(1196),
    [sym_function_type] = STATE(1390),
    [sym_tuple_type] = STATE(1390),
    [sym_unit_type] = STATE(1390),
    [sym_generic_type] = STATE(1359),
    [sym_generic_type_with_turbofish] = STATE(2379),
    [sym_bounded_type] = STATE(1390),
    [sym_reference_type] = STATE(1390),
    [sym_pointer_type] = STATE(1390),
    [sym_empty_type] = STATE(1390),
    [sym_abstract_type] = STATE(1390),
    [sym_dynamic_type] = STATE(1390),
    [sym_macro_invocation] = STATE(1390),
    [sym_scoped_identifier] = STATE(2157),
    [sym_scoped_type_identifier] = STATE(1331),
    [aux_sym_function_modifiers_repeat1] = STATE(1526),
    [sym_identifier] = ACTIONS(2309),
    [anon_sym_LPAREN] = ACTIONS(856),
    [anon_sym_RPAREN] = ACTIONS(2528),
    [anon_sym_LBRACK] = ACTIONS(860),
    [anon_sym_STAR] = ACTIONS(658),
    [anon_sym_u8] = ACTIONS(862),
    [anon_sym_i8] = ACTIONS(862),
    [anon_sym_u16] = ACTIONS(862),
    [anon_sym_i16] = ACTIONS(862),
    [anon_sym_u32] = ACTIONS(862),
    [anon_sym_i32] = ACTIONS(862),
    [anon_sym_u64] = ACTIONS(862),
    [anon_sym_i64] = ACTIONS(862),
    [anon_sym_u128] = ACTIONS(862),
    [anon_sym_i128] = ACTIONS(862),
    [anon_sym_isize] = ACTIONS(862),
    [anon_sym_usize] = ACTIONS(862),
    [anon_sym_f32] = ACTIONS(862),
    [anon_sym_f64] = ACTIONS(862),
    [anon_sym_bool] = ACTIONS(862),
    [anon_sym_str] = ACTIONS(862),
    [anon_sym_char] = ACTIONS(862),
    [anon_sym_SQUOTE] = ACTIONS(2313),
    [anon_sym_async] = ACTIONS(664),
    [anon_sym_const] = ACTIONS(664),
    [anon_sym_default] = ACTIONS(864),
    [anon_sym_fn] = ACTIONS(670),
    [anon_sym_for] = ACTIONS(672),
    [anon_sym_impl] = ACTIONS(674),
    [anon_sym_union] = ACTIONS(866),
    [anon_sym_unsafe] = ACTIONS(664),
    [anon_sym_BANG] = ACTIONS(680),
    [anon_sym_extern] = ACTIONS(684),
    [anon_sym_LT] = ACTIONS(77),
    [anon_sym_COLON_COLON] = ACTIONS(870),
    [anon_sym_AMP] = ACTIONS(872),
    [anon_sym_dyn] = ACTIONS(696),
    [sym_line_comment] = ACTIONS(3),
    [sym_self] = ACTIONS(876),
    [sym_super] = ACTIONS(876),
    [sym_crate] = ACTIONS(876),
    [sym_metavariable] = ACTIONS(878),
    [sym_block_comment] = ACTIONS(3),
  },
};

static const uint16_t ts_small_parse_table[] = {
  [0] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1400), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [129] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2291), 1,
      sym__type,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [258] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(2038), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [387] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1663), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [516] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1613), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [645] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1383), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [774] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1883), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [903] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1387), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [1032] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1879), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [1161] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1614), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [1290] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1385), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [1419] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(2068), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [1548] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1811), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [1677] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1633), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [1806] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1398), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [1935] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1624), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [2064] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(2104), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [2193] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(2045), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [2322] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(2117), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [2451] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1875), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [2580] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    ACTIONS(2530), 1,
      sym_identifier,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1491), 1,
      sym_scoped_type_identifier,
    STATE(1657), 1,
      sym__type,
    STATE(1679), 1,
      sym_generic_type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [2709] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1869), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [2838] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1405), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [2967] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1625), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [3096] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1817), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [3225] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1379), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [3354] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1631), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [3483] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1394), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [3612] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1647), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [3741] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1874), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [3870] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1381), 1,
      sym__type,
    STATE(1382), 1,
      sym_lifetime,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [3999] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    ACTIONS(2464), 1,
      sym_identifier,
    ACTIONS(2466), 1,
      anon_sym_LPAREN,
    ACTIONS(2468), 1,
      anon_sym_LBRACK,
    ACTIONS(2472), 1,
      anon_sym_STAR,
    ACTIONS(2476), 1,
      anon_sym_default,
    ACTIONS(2478), 1,
      anon_sym_fn,
    ACTIONS(2480), 1,
      anon_sym_impl,
    ACTIONS(2482), 1,
      anon_sym_union,
    ACTIONS(2484), 1,
      anon_sym_BANG,
    ACTIONS(2486), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2488), 1,
      anon_sym_AMP,
    ACTIONS(2490), 1,
      anon_sym_dyn,
    ACTIONS(2496), 1,
      sym_metavariable,
    STATE(760), 1,
      sym_scoped_type_identifier,
    STATE(790), 1,
      sym_generic_type,
    STATE(935), 1,
      sym__type,
    STATE(1195), 1,
      sym_for_lifetimes,
    STATE(2216), 1,
      sym_scoped_identifier,
    STATE(2386), 1,
      sym_function_modifiers,
    STATE(2491), 1,
      sym_bracketed_type,
    STATE(2492), 1,
      sym_generic_type_with_turbofish,
    STATE(2513), 1,
      sym_lifetime,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(2494), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1040), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(2474), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [4128] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    ACTIONS(2464), 1,
      sym_identifier,
    ACTIONS(2466), 1,
      anon_sym_LPAREN,
    ACTIONS(2468), 1,
      anon_sym_LBRACK,
    ACTIONS(2472), 1,
      anon_sym_STAR,
    ACTIONS(2476), 1,
      anon_sym_default,
    ACTIONS(2478), 1,
      anon_sym_fn,
    ACTIONS(2480), 1,
      anon_sym_impl,
    ACTIONS(2482), 1,
      anon_sym_union,
    ACTIONS(2484), 1,
      anon_sym_BANG,
    ACTIONS(2486), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2488), 1,
      anon_sym_AMP,
    ACTIONS(2490), 1,
      anon_sym_dyn,
    ACTIONS(2496), 1,
      sym_metavariable,
    STATE(760), 1,
      sym_scoped_type_identifier,
    STATE(790), 1,
      sym_generic_type,
    STATE(942), 1,
      sym__type,
    STATE(1195), 1,
      sym_for_lifetimes,
    STATE(2216), 1,
      sym_scoped_identifier,
    STATE(2386), 1,
      sym_function_modifiers,
    STATE(2491), 1,
      sym_bracketed_type,
    STATE(2492), 1,
      sym_generic_type_with_turbofish,
    STATE(2513), 1,
      sym_lifetime,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(2494), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1040), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(2474), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [4257] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    ACTIONS(2464), 1,
      sym_identifier,
    ACTIONS(2466), 1,
      anon_sym_LPAREN,
    ACTIONS(2468), 1,
      anon_sym_LBRACK,
    ACTIONS(2472), 1,
      anon_sym_STAR,
    ACTIONS(2476), 1,
      anon_sym_default,
    ACTIONS(2478), 1,
      anon_sym_fn,
    ACTIONS(2480), 1,
      anon_sym_impl,
    ACTIONS(2482), 1,
      anon_sym_union,
    ACTIONS(2484), 1,
      anon_sym_BANG,
    ACTIONS(2486), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2488), 1,
      anon_sym_AMP,
    ACTIONS(2490), 1,
      anon_sym_dyn,
    ACTIONS(2496), 1,
      sym_metavariable,
    STATE(760), 1,
      sym_scoped_type_identifier,
    STATE(790), 1,
      sym_generic_type,
    STATE(951), 1,
      sym__type,
    STATE(1195), 1,
      sym_for_lifetimes,
    STATE(2216), 1,
      sym_scoped_identifier,
    STATE(2386), 1,
      sym_function_modifiers,
    STATE(2491), 1,
      sym_bracketed_type,
    STATE(2492), 1,
      sym_generic_type_with_turbofish,
    STATE(2513), 1,
      sym_lifetime,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(2494), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1040), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(2474), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [4386] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(2051), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [4515] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2282), 1,
      sym__type,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [4644] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    ACTIONS(2464), 1,
      sym_identifier,
    ACTIONS(2466), 1,
      anon_sym_LPAREN,
    ACTIONS(2468), 1,
      anon_sym_LBRACK,
    ACTIONS(2472), 1,
      anon_sym_STAR,
    ACTIONS(2476), 1,
      anon_sym_default,
    ACTIONS(2478), 1,
      anon_sym_fn,
    ACTIONS(2480), 1,
      anon_sym_impl,
    ACTIONS(2482), 1,
      anon_sym_union,
    ACTIONS(2484), 1,
      anon_sym_BANG,
    ACTIONS(2486), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2488), 1,
      anon_sym_AMP,
    ACTIONS(2490), 1,
      anon_sym_dyn,
    ACTIONS(2496), 1,
      sym_metavariable,
    STATE(760), 1,
      sym_scoped_type_identifier,
    STATE(790), 1,
      sym_generic_type,
    STATE(953), 1,
      sym__type,
    STATE(1195), 1,
      sym_for_lifetimes,
    STATE(2216), 1,
      sym_scoped_identifier,
    STATE(2386), 1,
      sym_function_modifiers,
    STATE(2491), 1,
      sym_bracketed_type,
    STATE(2492), 1,
      sym_generic_type_with_turbofish,
    STATE(2513), 1,
      sym_lifetime,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(2494), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1040), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(2474), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [4773] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1619), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [4902] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1936), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [5031] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1381), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [5160] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1670), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [5289] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1694), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [5418] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(2025), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [5547] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1782), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [5676] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(2031), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [5805] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1889), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [5934] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1636), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [6063] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1823), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [6192] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1655), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [6321] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1684), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [6450] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2177), 1,
      sym__type,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [6579] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2252), 1,
      sym__type,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [6708] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1867), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [6837] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1840), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [6966] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1831), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [7095] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1671), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [7224] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1669), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [7353] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(2017), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [7482] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1928), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [7611] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2192), 1,
      sym__type,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [7740] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1678), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [7869] = 33,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    ACTIONS(2532), 1,
      sym_self,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1383), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(876), 2,
      sym_super,
      sym_crate,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [8000] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1985), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [8129] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1572), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [8258] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1701), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [8387] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    ACTIONS(2534), 1,
      sym_identifier,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1497), 1,
      sym_scoped_type_identifier,
    STATE(1637), 1,
      sym_generic_type,
    STATE(1640), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [8516] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1703), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [8645] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1999), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [8774] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2222), 1,
      sym__type,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [8903] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1692), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [9032] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1689), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [9161] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2230), 1,
      sym__type,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [9290] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(2021), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [9419] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    ACTIONS(2536), 1,
      sym_identifier,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1507), 1,
      sym_scoped_type_identifier,
    STATE(1642), 1,
      sym__type,
    STATE(1643), 1,
      sym_generic_type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [9548] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1950), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [9677] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(2009), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [9806] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1660), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [9935] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2203), 1,
      sym__type,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [10064] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1622), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [10193] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    ACTIONS(2464), 1,
      sym_identifier,
    ACTIONS(2466), 1,
      anon_sym_LPAREN,
    ACTIONS(2468), 1,
      anon_sym_LBRACK,
    ACTIONS(2472), 1,
      anon_sym_STAR,
    ACTIONS(2476), 1,
      anon_sym_default,
    ACTIONS(2478), 1,
      anon_sym_fn,
    ACTIONS(2480), 1,
      anon_sym_impl,
    ACTIONS(2482), 1,
      anon_sym_union,
    ACTIONS(2484), 1,
      anon_sym_BANG,
    ACTIONS(2486), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2488), 1,
      anon_sym_AMP,
    ACTIONS(2490), 1,
      anon_sym_dyn,
    ACTIONS(2496), 1,
      sym_metavariable,
    STATE(760), 1,
      sym_scoped_type_identifier,
    STATE(790), 1,
      sym_generic_type,
    STATE(856), 1,
      sym__type,
    STATE(1195), 1,
      sym_for_lifetimes,
    STATE(2216), 1,
      sym_scoped_identifier,
    STATE(2386), 1,
      sym_function_modifiers,
    STATE(2491), 1,
      sym_bracketed_type,
    STATE(2492), 1,
      sym_generic_type_with_turbofish,
    STATE(2513), 1,
      sym_lifetime,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(2494), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1040), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(2474), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [10322] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1392), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [10451] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1402), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [10580] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1704), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [10709] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2180), 1,
      sym__type,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [10838] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    ACTIONS(2538), 1,
      sym_identifier,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1483), 1,
      sym_scoped_type_identifier,
    STATE(1685), 1,
      sym__type,
    STATE(1686), 1,
      sym_generic_type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [10967] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    ACTIONS(2464), 1,
      sym_identifier,
    ACTIONS(2466), 1,
      anon_sym_LPAREN,
    ACTIONS(2468), 1,
      anon_sym_LBRACK,
    ACTIONS(2472), 1,
      anon_sym_STAR,
    ACTIONS(2476), 1,
      anon_sym_default,
    ACTIONS(2478), 1,
      anon_sym_fn,
    ACTIONS(2480), 1,
      anon_sym_impl,
    ACTIONS(2482), 1,
      anon_sym_union,
    ACTIONS(2484), 1,
      anon_sym_BANG,
    ACTIONS(2486), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2488), 1,
      anon_sym_AMP,
    ACTIONS(2490), 1,
      anon_sym_dyn,
    ACTIONS(2496), 1,
      sym_metavariable,
    STATE(760), 1,
      sym_scoped_type_identifier,
    STATE(790), 1,
      sym_generic_type,
    STATE(862), 1,
      sym__type,
    STATE(1195), 1,
      sym_for_lifetimes,
    STATE(2216), 1,
      sym_scoped_identifier,
    STATE(2386), 1,
      sym_function_modifiers,
    STATE(2491), 1,
      sym_bracketed_type,
    STATE(2492), 1,
      sym_generic_type_with_turbofish,
    STATE(2513), 1,
      sym_lifetime,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(2494), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1040), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(2474), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [11096] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    ACTIONS(2464), 1,
      sym_identifier,
    ACTIONS(2466), 1,
      anon_sym_LPAREN,
    ACTIONS(2468), 1,
      anon_sym_LBRACK,
    ACTIONS(2472), 1,
      anon_sym_STAR,
    ACTIONS(2476), 1,
      anon_sym_default,
    ACTIONS(2478), 1,
      anon_sym_fn,
    ACTIONS(2480), 1,
      anon_sym_impl,
    ACTIONS(2482), 1,
      anon_sym_union,
    ACTIONS(2484), 1,
      anon_sym_BANG,
    ACTIONS(2486), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2488), 1,
      anon_sym_AMP,
    ACTIONS(2490), 1,
      anon_sym_dyn,
    ACTIONS(2496), 1,
      sym_metavariable,
    STATE(760), 1,
      sym_scoped_type_identifier,
    STATE(790), 1,
      sym_generic_type,
    STATE(865), 1,
      sym__type,
    STATE(1195), 1,
      sym_for_lifetimes,
    STATE(2216), 1,
      sym_scoped_identifier,
    STATE(2386), 1,
      sym_function_modifiers,
    STATE(2491), 1,
      sym_bracketed_type,
    STATE(2492), 1,
      sym_generic_type_with_turbofish,
    STATE(2513), 1,
      sym_lifetime,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(2494), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1040), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(2474), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [11225] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    ACTIONS(2464), 1,
      sym_identifier,
    ACTIONS(2466), 1,
      anon_sym_LPAREN,
    ACTIONS(2468), 1,
      anon_sym_LBRACK,
    ACTIONS(2472), 1,
      anon_sym_STAR,
    ACTIONS(2476), 1,
      anon_sym_default,
    ACTIONS(2478), 1,
      anon_sym_fn,
    ACTIONS(2480), 1,
      anon_sym_impl,
    ACTIONS(2482), 1,
      anon_sym_union,
    ACTIONS(2484), 1,
      anon_sym_BANG,
    ACTIONS(2486), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2488), 1,
      anon_sym_AMP,
    ACTIONS(2490), 1,
      anon_sym_dyn,
    ACTIONS(2496), 1,
      sym_metavariable,
    STATE(760), 1,
      sym_scoped_type_identifier,
    STATE(790), 1,
      sym_generic_type,
    STATE(867), 1,
      sym__type,
    STATE(1195), 1,
      sym_for_lifetimes,
    STATE(2216), 1,
      sym_scoped_identifier,
    STATE(2386), 1,
      sym_function_modifiers,
    STATE(2491), 1,
      sym_bracketed_type,
    STATE(2492), 1,
      sym_generic_type_with_turbofish,
    STATE(2513), 1,
      sym_lifetime,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(2494), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1040), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(2474), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [11354] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2289), 1,
      sym__type,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [11483] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1992), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [11612] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1688), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [11741] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    ACTIONS(2464), 1,
      sym_identifier,
    ACTIONS(2466), 1,
      anon_sym_LPAREN,
    ACTIONS(2468), 1,
      anon_sym_LBRACK,
    ACTIONS(2472), 1,
      anon_sym_STAR,
    ACTIONS(2476), 1,
      anon_sym_default,
    ACTIONS(2478), 1,
      anon_sym_fn,
    ACTIONS(2480), 1,
      anon_sym_impl,
    ACTIONS(2482), 1,
      anon_sym_union,
    ACTIONS(2484), 1,
      anon_sym_BANG,
    ACTIONS(2486), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2488), 1,
      anon_sym_AMP,
    ACTIONS(2490), 1,
      anon_sym_dyn,
    ACTIONS(2496), 1,
      sym_metavariable,
    STATE(760), 1,
      sym_scoped_type_identifier,
    STATE(790), 1,
      sym_generic_type,
    STATE(1083), 1,
      sym__type,
    STATE(1195), 1,
      sym_for_lifetimes,
    STATE(2216), 1,
      sym_scoped_identifier,
    STATE(2386), 1,
      sym_function_modifiers,
    STATE(2491), 1,
      sym_bracketed_type,
    STATE(2492), 1,
      sym_generic_type_with_turbofish,
    STATE(2513), 1,
      sym_lifetime,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(2494), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1040), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(2474), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [11870] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1919), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [11999] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1344), 20,
      sym_raw_string_literal,
      sym_float_literal,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_STAR,
      anon_sym_POUND,
      anon_sym_BANG,
      anon_sym_COMMA,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      sym_integer_literal,
      aux_sym_string_literal_token1,
      sym_char_literal,
      sym_metavariable,
    ACTIONS(1346), 42,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_SQUOTE,
      anon_sym_async,
      anon_sym_break,
      anon_sym_const,
      anon_sym_continue,
      anon_sym_default,
      anon_sym_for,
      anon_sym_if,
      anon_sym_loop,
      anon_sym_match,
      anon_sym_return,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_while,
      anon_sym_ref,
      anon_sym__,
      sym_mutable_specifier,
      anon_sym_yield,
      anon_sym_move,
      anon_sym_true,
      anon_sym_false,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [12070] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1837), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [12199] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2279), 1,
      sym__type,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [12328] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1915), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [12457] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1617), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [12586] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1621), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [12715] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1846), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [12844] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1683), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [12973] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    ACTIONS(2464), 1,
      sym_identifier,
    ACTIONS(2466), 1,
      anon_sym_LPAREN,
    ACTIONS(2468), 1,
      anon_sym_LBRACK,
    ACTIONS(2472), 1,
      anon_sym_STAR,
    ACTIONS(2476), 1,
      anon_sym_default,
    ACTIONS(2478), 1,
      anon_sym_fn,
    ACTIONS(2480), 1,
      anon_sym_impl,
    ACTIONS(2482), 1,
      anon_sym_union,
    ACTIONS(2484), 1,
      anon_sym_BANG,
    ACTIONS(2486), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2488), 1,
      anon_sym_AMP,
    ACTIONS(2490), 1,
      anon_sym_dyn,
    ACTIONS(2496), 1,
      sym_metavariable,
    STATE(760), 1,
      sym_scoped_type_identifier,
    STATE(790), 1,
      sym_generic_type,
    STATE(946), 1,
      sym__type,
    STATE(1195), 1,
      sym_for_lifetimes,
    STATE(2216), 1,
      sym_scoped_identifier,
    STATE(2386), 1,
      sym_function_modifiers,
    STATE(2491), 1,
      sym_bracketed_type,
    STATE(2492), 1,
      sym_generic_type_with_turbofish,
    STATE(2513), 1,
      sym_lifetime,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(2494), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1040), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(2474), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [13102] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(2093), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [13231] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1766), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [13360] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(1595), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [13489] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    ACTIONS(2464), 1,
      sym_identifier,
    ACTIONS(2466), 1,
      anon_sym_LPAREN,
    ACTIONS(2468), 1,
      anon_sym_LBRACK,
    ACTIONS(2472), 1,
      anon_sym_STAR,
    ACTIONS(2476), 1,
      anon_sym_default,
    ACTIONS(2478), 1,
      anon_sym_fn,
    ACTIONS(2480), 1,
      anon_sym_impl,
    ACTIONS(2482), 1,
      anon_sym_union,
    ACTIONS(2484), 1,
      anon_sym_BANG,
    ACTIONS(2486), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2488), 1,
      anon_sym_AMP,
    ACTIONS(2490), 1,
      anon_sym_dyn,
    ACTIONS(2496), 1,
      sym_metavariable,
    STATE(760), 1,
      sym_scoped_type_identifier,
    STATE(790), 1,
      sym_generic_type,
    STATE(932), 1,
      sym__type,
    STATE(1195), 1,
      sym_for_lifetimes,
    STATE(2216), 1,
      sym_scoped_identifier,
    STATE(2386), 1,
      sym_function_modifiers,
    STATE(2491), 1,
      sym_bracketed_type,
    STATE(2492), 1,
      sym_generic_type_with_turbofish,
    STATE(2513), 1,
      sym_lifetime,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(2494), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1040), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(2474), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [13618] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    ACTIONS(2464), 1,
      sym_identifier,
    ACTIONS(2466), 1,
      anon_sym_LPAREN,
    ACTIONS(2468), 1,
      anon_sym_LBRACK,
    ACTIONS(2472), 1,
      anon_sym_STAR,
    ACTIONS(2476), 1,
      anon_sym_default,
    ACTIONS(2478), 1,
      anon_sym_fn,
    ACTIONS(2480), 1,
      anon_sym_impl,
    ACTIONS(2482), 1,
      anon_sym_union,
    ACTIONS(2484), 1,
      anon_sym_BANG,
    ACTIONS(2486), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2488), 1,
      anon_sym_AMP,
    ACTIONS(2490), 1,
      anon_sym_dyn,
    ACTIONS(2496), 1,
      sym_metavariable,
    STATE(760), 1,
      sym_scoped_type_identifier,
    STATE(790), 1,
      sym_generic_type,
    STATE(941), 1,
      sym__type,
    STATE(1195), 1,
      sym_for_lifetimes,
    STATE(2216), 1,
      sym_scoped_identifier,
    STATE(2386), 1,
      sym_function_modifiers,
    STATE(2491), 1,
      sym_bracketed_type,
    STATE(2492), 1,
      sym_generic_type_with_turbofish,
    STATE(2513), 1,
      sym_lifetime,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(2494), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1040), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(2474), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [13747] = 32,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(658), 1,
      anon_sym_STAR,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(674), 1,
      anon_sym_impl,
    ACTIONS(680), 1,
      anon_sym_BANG,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(696), 1,
      anon_sym_dyn,
    ACTIONS(856), 1,
      anon_sym_LPAREN,
    ACTIONS(860), 1,
      anon_sym_LBRACK,
    ACTIONS(864), 1,
      anon_sym_default,
    ACTIONS(866), 1,
      anon_sym_union,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(872), 1,
      anon_sym_AMP,
    ACTIONS(878), 1,
      sym_metavariable,
    ACTIONS(2309), 1,
      sym_identifier,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1331), 1,
      sym_scoped_type_identifier,
    STATE(1359), 1,
      sym_generic_type,
    STATE(2139), 1,
      sym__type,
    STATE(2157), 1,
      sym_scoped_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2471), 1,
      sym_lifetime,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1390), 11,
      sym_array_type,
      sym_function_type,
      sym_tuple_type,
      sym_unit_type,
      sym_bounded_type,
      sym_reference_type,
      sym_pointer_type,
      sym_empty_type,
      sym_abstract_type,
      sym_dynamic_type,
      sym_macro_invocation,
    ACTIONS(862), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [13876] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2542), 17,
      sym_raw_string_literal,
      sym_float_literal,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_DASH_GT,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_PIPE,
      sym_integer_literal,
      aux_sym_string_literal_token1,
      sym_char_literal,
      sym_metavariable,
    ACTIONS(2540), 40,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_SQUOTE,
      anon_sym_async,
      anon_sym_break,
      anon_sym_const,
      anon_sym_continue,
      anon_sym_default,
      anon_sym_for,
      anon_sym_if,
      anon_sym_loop,
      anon_sym_match,
      anon_sym_return,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_while,
      anon_sym_DASH,
      anon_sym_yield,
      anon_sym_move,
      anon_sym_true,
      anon_sym_false,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [13942] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2546), 17,
      sym_raw_string_literal,
      sym_float_literal,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_DASH_GT,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_PIPE,
      sym_integer_literal,
      aux_sym_string_literal_token1,
      sym_char_literal,
      sym_metavariable,
    ACTIONS(2544), 40,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_SQUOTE,
      anon_sym_async,
      anon_sym_break,
      anon_sym_const,
      anon_sym_continue,
      anon_sym_default,
      anon_sym_for,
      anon_sym_if,
      anon_sym_loop,
      anon_sym_match,
      anon_sym_return,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_while,
      anon_sym_DASH,
      anon_sym_yield,
      anon_sym_move,
      anon_sym_true,
      anon_sym_false,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [14008] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(502), 18,
      sym_raw_string_literal,
      sym_float_literal,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      sym_integer_literal,
      aux_sym_string_literal_token1,
      sym_char_literal,
      sym_metavariable,
    ACTIONS(2548), 39,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_SQUOTE,
      anon_sym_async,
      anon_sym_break,
      anon_sym_const,
      anon_sym_continue,
      anon_sym_default,
      anon_sym_for,
      anon_sym_if,
      anon_sym_loop,
      anon_sym_match,
      anon_sym_return,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_while,
      anon_sym_yield,
      anon_sym_move,
      anon_sym_true,
      anon_sym_false,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [14074] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(842), 17,
      sym_raw_string_literal,
      sym_float_literal,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_DASH_GT,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_PIPE,
      sym_integer_literal,
      aux_sym_string_literal_token1,
      sym_char_literal,
      sym_metavariable,
    ACTIONS(840), 40,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_SQUOTE,
      anon_sym_async,
      anon_sym_break,
      anon_sym_const,
      anon_sym_continue,
      anon_sym_default,
      anon_sym_for,
      anon_sym_if,
      anon_sym_loop,
      anon_sym_match,
      anon_sym_return,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_while,
      anon_sym_DASH,
      anon_sym_yield,
      anon_sym_move,
      anon_sym_true,
      anon_sym_false,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [14140] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1344), 15,
      sym_raw_string_literal,
      sym_float_literal,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      anon_sym_AMP,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DASH,
      sym_integer_literal,
      aux_sym_string_literal_token1,
      sym_char_literal,
      sym_metavariable,
    ACTIONS(1346), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_SQUOTE,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_for,
      anon_sym_impl,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_extern,
      anon_sym_ref,
      anon_sym__,
      anon_sym_dyn,
      sym_mutable_specifier,
      anon_sym_DOT_DOT,
      anon_sym_true,
      anon_sym_false,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [14202] = 9,
    ACTIONS(2552), 1,
      anon_sym_LPAREN,
    ACTIONS(2556), 1,
      anon_sym_BANG,
    ACTIONS(2558), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2560), 1,
      anon_sym_LT2,
    STATE(797), 1,
      sym_type_arguments,
    STATE(807), 1,
      sym_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2554), 17,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT_EQ,
      anon_sym_DOT,
    ACTIONS(2550), 27,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_GT_GT_EQ,
  [14273] = 8,
    ACTIONS(2552), 1,
      anon_sym_LPAREN,
    ACTIONS(2558), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2560), 1,
      anon_sym_LT2,
    STATE(797), 1,
      sym_type_arguments,
    STATE(807), 1,
      sym_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2564), 17,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT_EQ,
      anon_sym_DOT,
    ACTIONS(2562), 27,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_GT_GT_EQ,
  [14341] = 8,
    ACTIONS(2552), 1,
      anon_sym_LPAREN,
    ACTIONS(2558), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2560), 1,
      anon_sym_LT2,
    STATE(797), 1,
      sym_type_arguments,
    STATE(807), 1,
      sym_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2568), 17,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT_EQ,
      anon_sym_DOT,
    ACTIONS(2566), 27,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_GT_GT_EQ,
  [14409] = 5,
    ACTIONS(2574), 1,
      anon_sym_BANG,
    ACTIONS(2576), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2572), 17,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT_EQ,
      anon_sym_DOT,
    ACTIONS(2570), 29,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_LT2,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_GT_GT_EQ,
  [14470] = 5,
    ACTIONS(2582), 1,
      anon_sym_BANG,
    ACTIONS(2584), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2580), 17,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT_EQ,
      anon_sym_DOT,
    ACTIONS(2578), 29,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_LT2,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_GT_GT_EQ,
  [14531] = 7,
    ACTIONS(2552), 1,
      anon_sym_LPAREN,
    ACTIONS(2560), 1,
      anon_sym_LT2,
    STATE(799), 1,
      sym_type_arguments,
    STATE(803), 1,
      sym_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2588), 17,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT_EQ,
      anon_sym_DOT,
    ACTIONS(2586), 27,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_GT_GT_EQ,
  [14596] = 7,
    ACTIONS(2556), 1,
      anon_sym_BANG,
    ACTIONS(2590), 1,
      anon_sym_LBRACE,
    ACTIONS(2592), 1,
      anon_sym_COLON_COLON,
    STATE(1121), 1,
      sym_field_initializer_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(460), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(458), 29,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [14661] = 7,
    ACTIONS(2552), 1,
      anon_sym_LPAREN,
    ACTIONS(2560), 1,
      anon_sym_LT2,
    STATE(799), 1,
      sym_type_arguments,
    STATE(803), 1,
      sym_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2596), 17,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT_EQ,
      anon_sym_DOT,
    ACTIONS(2594), 27,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_GT_GT_EQ,
  [14726] = 7,
    ACTIONS(2552), 1,
      anon_sym_LPAREN,
    ACTIONS(2560), 1,
      anon_sym_LT2,
    STATE(799), 1,
      sym_type_arguments,
    STATE(803), 1,
      sym_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2600), 17,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT_EQ,
      anon_sym_DOT,
    ACTIONS(2598), 27,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_GT_GT_EQ,
  [14791] = 5,
    ACTIONS(2606), 1,
      anon_sym_BANG,
    ACTIONS(2608), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2604), 17,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT_EQ,
      anon_sym_DOT,
    ACTIONS(2602), 29,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_LT2,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_GT_GT_EQ,
  [14852] = 5,
    ACTIONS(2614), 1,
      anon_sym_BANG,
    ACTIONS(2616), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2612), 17,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT_EQ,
      anon_sym_DOT,
    ACTIONS(2610), 29,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_LT2,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_GT_GT_EQ,
  [14913] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1924), 9,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1926), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [14969] = 4,
    ACTIONS(2618), 1,
      anon_sym_LBRACE,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2582), 16,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2584), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_COLON_COLON,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [15027] = 6,
    ACTIONS(2626), 1,
      anon_sym_BANG,
    ACTIONS(2628), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2624), 6,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_unsafe,
      anon_sym_extern,
    ACTIONS(2622), 16,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_as,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2620), 23,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_QMARK,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [15089] = 4,
    ACTIONS(2630), 1,
      anon_sym_LBRACE,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2614), 16,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2616), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_COLON_COLON,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [15147] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1936), 9,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1938), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [15203] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1352), 9,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1354), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [15259] = 4,
    ACTIONS(2632), 1,
      anon_sym_LBRACE,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2606), 16,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2608), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_COLON_COLON,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [15317] = 5,
    ACTIONS(2638), 1,
      anon_sym_SQUOTE,
    STATE(1073), 1,
      sym_loop_label,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2636), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2634), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [15377] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2582), 16,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2584), 31,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_COLON_COLON,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [15433] = 4,
    ACTIONS(2640), 1,
      anon_sym_LBRACE,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2574), 16,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2576), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_COLON_COLON,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [15491] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1096), 9,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1098), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [15547] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1376), 9,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1378), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [15603] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2644), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2642), 31,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [15658] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2648), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2646), 31,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [15713] = 5,
    ACTIONS(2652), 1,
      anon_sym_LPAREN,
    STATE(1081), 1,
      sym_arguments,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2654), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2650), 29,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [15772] = 4,
    ACTIONS(2656), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(460), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(458), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [15829] = 4,
    ACTIONS(2662), 1,
      anon_sym_DASH_GT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2660), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2658), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [15886] = 5,
    ACTIONS(2628), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2664), 1,
      anon_sym_BANG,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2622), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2620), 29,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [15945] = 4,
    ACTIONS(2670), 1,
      anon_sym_DASH_GT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2668), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2666), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [16002] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2674), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2672), 31,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_COLON_COLON,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [16057] = 4,
    ACTIONS(2680), 1,
      anon_sym_DASH_GT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2678), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2676), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [16114] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(814), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(816), 31,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [16169] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2684), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2682), 31,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_COLON_COLON,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [16224] = 4,
    ACTIONS(2686), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2588), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2586), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [16281] = 4,
    ACTIONS(2688), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2600), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2598), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [16338] = 4,
    ACTIONS(2690), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2600), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2598), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [16395] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2580), 17,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_EQ,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_LT_LT_EQ,
      anon_sym_DOT,
    ACTIONS(2578), 29,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_LT2,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_GT_GT_EQ,
  [16450] = 4,
    ACTIONS(2686), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2600), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2598), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [16507] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2694), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2692), 31,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [16562] = 5,
    ACTIONS(2556), 1,
      anon_sym_BANG,
    ACTIONS(2696), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(460), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(458), 29,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [16621] = 4,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2700), 2,
      anon_sym_LBRACE,
      anon_sym_COLON_COLON,
    ACTIONS(2702), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2698), 29,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [16678] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2706), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2704), 31,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_COLON_COLON,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [16733] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2710), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2708), 31,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [16788] = 4,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2712), 2,
      anon_sym_LBRACE,
      anon_sym_COLON_COLON,
    ACTIONS(2702), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2698), 29,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [16845] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2716), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2714), 31,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_COLON_COLON,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [16900] = 4,
    ACTIONS(2722), 1,
      anon_sym_DASH_GT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2720), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2718), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [16957] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2726), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2724), 31,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_COLON_COLON,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [17012] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2730), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2728), 31,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_COLON_COLON,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [17067] = 5,
    ACTIONS(2732), 1,
      anon_sym_else,
    STATE(952), 1,
      sym_else_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(402), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(400), 29,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [17126] = 4,
    ACTIONS(2738), 1,
      anon_sym_DASH_GT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2736), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2734), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [17183] = 4,
    ACTIONS(2744), 1,
      anon_sym_DASH_GT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2742), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2740), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [17240] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2748), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2746), 31,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_COLON_COLON,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [17295] = 4,
    ACTIONS(2686), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2596), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2594), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [17352] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2752), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2750), 31,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [17407] = 4,
    ACTIONS(2758), 1,
      anon_sym_DASH_GT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2756), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2754), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [17464] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1232), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1234), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [17518] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1296), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1298), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [17572] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1558), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1560), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [17626] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1618), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1620), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [17680] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1630), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1632), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [17734] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2762), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2760), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [17788] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1648), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1650), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [17842] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2431), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2433), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [17896] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2766), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2764), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [17950] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2770), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2768), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [18004] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(432), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(430), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [18058] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1684), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1686), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [18112] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1688), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1690), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [18166] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1692), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1694), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [18220] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1696), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1698), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [18274] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1700), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1702), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [18328] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1704), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1706), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [18382] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1708), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1710), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [18436] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1712), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1714), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [18490] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1724), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1726), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [18544] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1732), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1734), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [18598] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1736), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1738), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [18652] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1740), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1742), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [18706] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1756), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1758), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [18760] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1772), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1774), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [18814] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1780), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1782), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [18868] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1784), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1786), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [18922] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1792), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1794), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [18976] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1804), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1806), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [19030] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1812), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1814), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [19084] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2774), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2772), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [19138] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1816), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1818), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [19192] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1820), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1822), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [19246] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1824), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1826), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [19300] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2778), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2776), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [19354] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2782), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2780), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [19408] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1840), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1842), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [19462] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1056), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1058), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [19516] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1852), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1854), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [19570] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1856), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1858), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [19624] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1880), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1882), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [19678] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1888), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1890), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [19732] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1904), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1906), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [19786] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1952), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1954), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [19840] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(494), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(492), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [19894] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1982), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1984), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [19948] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1986), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1988), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [20002] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1994), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1996), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [20056] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2786), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2784), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [20110] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1998), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(2000), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [20164] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1728), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1730), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [20218] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1978), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1980), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [20272] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1962), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1964), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [20326] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1958), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1960), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [20380] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2790), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2788), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [20434] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1940), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1942), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [20488] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1932), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1934), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [20542] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2794), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2792), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [20596] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2798), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2796), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [20650] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2802), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2800), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [20704] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2806), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2804), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [20758] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(448), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(446), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [20812] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(464), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(462), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [20866] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2810), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2808), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [20920] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2814), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2812), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [20974] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1892), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1894), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [21028] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1884), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1886), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [21082] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1868), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1870), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [21136] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1864), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1866), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [21190] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1760), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1762), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [21244] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1744), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1746), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [21298] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(616), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(614), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [21352] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1672), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1674), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [21406] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1582), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1584), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [21460] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1574), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1576), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [21514] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(486), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(484), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [21568] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1492), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1494), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [21622] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1416), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1418), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [21676] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1408), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1410), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [21730] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1380), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1382), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [21784] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1372), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1374), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [21838] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1368), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1370), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [21892] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1364), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1366), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [21946] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2702), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2698), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [22000] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1356), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1358), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [22054] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1348), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1350), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [22108] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1634), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1636), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [22162] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1614), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1616), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [22216] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1340), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1342), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [22270] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1328), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1330), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [22324] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2818), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2816), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [22378] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1316), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1318), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [22432] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1288), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1290), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [22486] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1280), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1282), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [22540] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1276), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1278), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [22594] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1260), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1262), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [22648] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1244), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1246), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [22702] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2822), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2820), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [22756] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1748), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1750), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [22810] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1752), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1754), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [22864] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1164), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1166), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [22918] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1160), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1162), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [22972] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1156), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1158), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [23026] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1152), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1154), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [23080] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1100), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1102), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [23134] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1512), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1514), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [23188] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1768), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1770), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [23242] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(478), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(476), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [23296] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1088), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1090), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [23350] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1504), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1506), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [23404] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(600), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(602), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [23458] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1776), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1778), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [23512] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1808), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1810), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [23566] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1432), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1434), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [23620] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2826), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2824), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [23674] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1460), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1462), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [23728] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2830), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2828), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [23782] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1828), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1830), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [23836] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1080), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1082), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [23890] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1638), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1640), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [23944] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1076), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1078), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [23998] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1456), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1458), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [24052] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(596), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(598), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [24106] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2834), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2832), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [24160] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2834), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2832), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [24214] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1072), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1074), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [24268] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1068), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1070), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [24322] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2838), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2836), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [24376] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1452), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1454), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [24430] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1832), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1834), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [24484] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1836), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1838), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [24538] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1448), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1450), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [24592] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1896), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1898), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [24646] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2842), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2840), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [24700] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2846), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2844), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [24754] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1908), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1910), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [24808] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1912), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1914), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [24862] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1444), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1446), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [24916] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2850), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2848), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [24970] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1440), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1442), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [25024] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2854), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2852), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [25078] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2858), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2856), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [25132] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2862), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2860), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [25186] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2866), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2864), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [25240] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(482), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(480), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [25294] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2870), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2868), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [25348] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1436), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1438), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [25402] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1060), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1062), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [25456] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1392), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1394), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [25510] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1488), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1490), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [25564] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1428), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1430), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [25618] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1420), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1422), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [25672] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1412), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1414), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [25726] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(610), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(608), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [25780] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1388), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1390), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [25834] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1656), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1658), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [25888] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1660), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1662), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [25942] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1788), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1790), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [25996] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(440), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(438), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [26050] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(412), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(410), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [26104] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1872), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1874), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [26158] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1900), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1902), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [26212] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1948), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1950), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [26266] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1944), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1946), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [26320] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1860), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1862), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [26374] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(598), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(596), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [26428] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(474), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(472), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [26482] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1716), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1718), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [26536] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1642), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1644), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [26590] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1626), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1628), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [26644] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1610), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1612), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [26698] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2874), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2872), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [26752] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1590), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1592), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [26806] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1308), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1310), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [26860] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1578), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1580), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [26914] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1554), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1556), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [26968] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1550), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1552), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [27022] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(424), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(422), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [27076] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1304), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1306), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [27130] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2878), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2876), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [27184] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1508), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1510), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [27238] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2367), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2369), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [27292] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1300), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1302), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [27346] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1920), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1922), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [27400] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1520), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1522), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [27454] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1292), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1294), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [27508] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1284), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1286), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [27562] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1272), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1274), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [27616] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1264), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1266), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [27670] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1472), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1474), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [27724] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1464), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1466), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [27778] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1256), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1258), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [27832] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1248), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1250), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [27886] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1240), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1242), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [27940] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2588), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2586), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [27994] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1928), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1930), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [28048] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1236), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1238), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [28102] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2882), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2880), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [28156] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1228), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1230), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [28210] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1224), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1226), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [28264] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1220), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1222), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [28318] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2886), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2884), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [28372] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1216), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1218), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [28426] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2890), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2888), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [28480] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2596), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2594), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [28534] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1212), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1214), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [28588] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1208), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1210), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [28642] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2894), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2892), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [28696] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1200), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1202), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [28750] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2898), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2896), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [28804] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1196), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1198), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [28858] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1188), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1190), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [28912] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1176), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1178), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [28966] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1172), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1174), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [29020] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1180), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1182), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [29074] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(602), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(600), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [29128] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(606), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(604), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [29182] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(408), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(406), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [29236] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1990), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1992), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [29290] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1974), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1976), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [29344] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2427), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2429), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [29398] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2902), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2900), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [29452] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1148), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1150), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [29506] = 4,
    ACTIONS(2904), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(460), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(458), 29,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [29562] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1192), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1194), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [29616] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1204), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1206), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [29670] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1144), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1146), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [29724] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1136), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1138), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [29778] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1128), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1130), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [29832] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1108), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1110), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [29886] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(428), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(426), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [29940] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(436), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(434), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [29994] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2600), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2598), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [30048] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1092), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1094), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [30102] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2908), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2906), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [30156] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1064), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1066), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [30210] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1970), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1972), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [30264] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(452), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(450), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [30318] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2912), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2910), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [30372] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(490), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(488), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [30426] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1680), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1682), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [30480] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(416), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(414), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [30534] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1848), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1850), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [30588] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2916), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2914), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [30642] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2920), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2918), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [30696] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1966), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1968), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [30750] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2654), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2650), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [30804] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1084), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1086), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [30858] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1116), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1118), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [30912] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1120), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1122), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [30966] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1606), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1608), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [31020] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1124), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1126), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [31074] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2924), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2922), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [31128] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2928), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2926), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [31182] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1132), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1134), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [31236] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1562), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1564), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [31290] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1500), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1502), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [31344] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1916), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1918), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [31398] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1140), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1142), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [31452] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2932), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2930), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [31506] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1586), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1588), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [31560] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2936), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2934), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [31614] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2940), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2938), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [31668] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2944), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2942), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [31722] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1496), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1498), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [31776] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2948), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2946), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [31830] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1360), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1362), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [31884] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1168), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1170), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [31938] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1184), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1186), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [31992] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1252), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1254), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [32046] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1268), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1270), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [32100] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2952), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2950), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [32154] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1344), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1346), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [32208] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2956), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2954), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [32262] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2960), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2958), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [32316] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2964), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2962), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [32370] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1312), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1314), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [32424] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1876), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1878), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [32478] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1320), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1322), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [32532] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2968), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2966), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [32586] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1324), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1326), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [32640] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1332), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1334), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [32694] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1844), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1846), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [32748] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1336), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1338), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [32802] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1384), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1386), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [32856] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1396), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1398), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [32910] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1400), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1402), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [32964] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1404), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1406), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [33018] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1424), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1426), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [33072] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1468), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1470), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [33126] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1476), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1478), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [33180] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(410), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(412), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [33234] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1480), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1482), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [33288] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1484), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1486), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [33342] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(406), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(408), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [33396] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1516), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1518), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [33450] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1546), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1548), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [33504] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1800), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1802), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [33558] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1566), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1568), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [33612] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(414), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(416), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [33666] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1570), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1572), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [33720] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1796), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1798), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [33774] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1104), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1106), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [33828] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(312), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(310), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [33882] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1594), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1596), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [33936] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1598), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1600), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [33990] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(614), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(616), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [34044] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1602), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1604), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [34098] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1622), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1624), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [34152] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1764), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1766), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [34206] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2397), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2399), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [34260] = 4,
    ACTIONS(2628), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2622), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2620), 29,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [34316] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1652), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1654), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [34370] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2972), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2970), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [34424] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1664), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1666), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [34478] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1668), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1670), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [34532] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(438), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(440), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [34586] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1676), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1678), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [34640] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1720), 7,
      anon_sym_SEMI,
      anon_sym_macro_rules_BANG,
      anon_sym_RBRACE,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(1722), 38,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_impl,
      anon_sym_let,
      anon_sym_mod,
      anon_sym_pub,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [34694] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(444), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(442), 30,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [34748] = 14,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2990), 1,
      anon_sym_DOT_DOT,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2988), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2984), 3,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2974), 25,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [34823] = 11,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2990), 1,
      anon_sym_DOT_DOT,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2988), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2984), 6,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
    ACTIONS(2974), 25,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [34892] = 7,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2990), 1,
      anon_sym_DOT_DOT,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2988), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3002), 13,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3000), 26,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [34953] = 21,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    ACTIONS(3004), 1,
      sym_identifier,
    ACTIONS(3008), 1,
      anon_sym_LPAREN,
    ACTIONS(3010), 1,
      anon_sym_STAR,
    ACTIONS(3016), 1,
      anon_sym_for,
    ACTIONS(3018), 1,
      anon_sym_AMP,
    ACTIONS(3020), 1,
      sym_metavariable,
    STATE(1788), 1,
      sym_scoped_type_identifier,
    STATE(1887), 1,
      sym_generic_type,
    STATE(1953), 1,
      sym_where_predicate,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2404), 1,
      sym_scoped_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3006), 2,
      anon_sym_SEMI,
      anon_sym_LBRACE,
    ACTIONS(3014), 2,
      anon_sym_default,
      anon_sym_union,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(2296), 5,
      sym_higher_ranked_trait_bound,
      sym_lifetime,
      sym_tuple_type,
      sym_reference_type,
      sym_pointer_type,
    ACTIONS(3012), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [35042] = 18,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2990), 1,
      anon_sym_DOT_DOT,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3024), 1,
      anon_sym_EQ,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2988), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3022), 19,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [35125] = 21,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    ACTIONS(3004), 1,
      sym_identifier,
    ACTIONS(3008), 1,
      anon_sym_LPAREN,
    ACTIONS(3010), 1,
      anon_sym_STAR,
    ACTIONS(3016), 1,
      anon_sym_for,
    ACTIONS(3018), 1,
      anon_sym_AMP,
    ACTIONS(3020), 1,
      sym_metavariable,
    STATE(1788), 1,
      sym_scoped_type_identifier,
    STATE(1887), 1,
      sym_generic_type,
    STATE(1953), 1,
      sym_where_predicate,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2404), 1,
      sym_scoped_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3014), 2,
      anon_sym_default,
      anon_sym_union,
    ACTIONS(3034), 2,
      anon_sym_SEMI,
      anon_sym_LBRACE,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(2296), 5,
      sym_higher_ranked_trait_bound,
      sym_lifetime,
      sym_tuple_type,
      sym_reference_type,
      sym_pointer_type,
    ACTIONS(3012), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [35214] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3038), 12,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      anon_sym_AMP,
      sym_metavariable,
    ACTIONS(3036), 32,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_for,
      anon_sym_impl,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_where,
      anon_sym_extern,
      anon_sym_dyn,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [35267] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3042), 12,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      anon_sym_AMP,
      sym_metavariable,
    ACTIONS(3040), 32,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_for,
      anon_sym_impl,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_where,
      anon_sym_extern,
      anon_sym_dyn,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [35320] = 18,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2990), 1,
      anon_sym_DOT_DOT,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3046), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2988), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3044), 19,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [35403] = 4,
    ACTIONS(3052), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3050), 14,
      sym_raw_string_literal,
      sym_float_literal,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      sym_integer_literal,
      aux_sym_string_literal_token1,
      sym_char_literal,
      sym_metavariable,
    ACTIONS(3048), 29,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_const,
      anon_sym_default,
      anon_sym_union,
      anon_sym_ref,
      anon_sym__,
      sym_mutable_specifier,
      anon_sym_true,
      anon_sym_false,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [35458] = 18,
    ACTIONS(288), 1,
      anon_sym_EQ,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2990), 1,
      anon_sym_DOT_DOT,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2988), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(282), 19,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [35541] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3056), 12,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      anon_sym_AMP,
      sym_metavariable,
    ACTIONS(3054), 32,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_for,
      anon_sym_impl,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_where,
      anon_sym_extern,
      anon_sym_dyn,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [35594] = 10,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2990), 1,
      anon_sym_DOT_DOT,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2988), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2984), 8,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2974), 25,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [35661] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3060), 12,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      anon_sym_AMP,
      sym_metavariable,
    ACTIONS(3058), 32,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_for,
      anon_sym_impl,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_where,
      anon_sym_extern,
      anon_sym_dyn,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [35714] = 7,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2990), 1,
      anon_sym_DOT_DOT,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2988), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3064), 13,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3062), 26,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [35775] = 12,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2990), 1,
      anon_sym_DOT_DOT,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2988), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2984), 5,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_CARET,
    ACTIONS(2974), 25,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [35846] = 13,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2990), 1,
      anon_sym_DOT_DOT,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2988), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2984), 4,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
    ACTIONS(2974), 25,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [35919] = 20,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2990), 1,
      anon_sym_DOT_DOT,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2988), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3066), 8,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_COMMA,
      anon_sym_else,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [36006] = 17,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2984), 1,
      anon_sym_EQ,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2990), 1,
      anon_sym_DOT_DOT,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2988), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(2974), 20,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [36087] = 16,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2984), 1,
      anon_sym_EQ,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2990), 1,
      anon_sym_DOT_DOT,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2988), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(2974), 21,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [36166] = 18,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2990), 1,
      anon_sym_DOT_DOT,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3076), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2988), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3074), 19,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [36249] = 4,
    ACTIONS(3082), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3080), 14,
      sym_raw_string_literal,
      sym_float_literal,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      sym_integer_literal,
      aux_sym_string_literal_token1,
      sym_char_literal,
      sym_metavariable,
    ACTIONS(3078), 29,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_const,
      anon_sym_default,
      anon_sym_union,
      anon_sym_ref,
      anon_sym__,
      sym_mutable_specifier,
      anon_sym_true,
      anon_sym_false,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [36304] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3086), 12,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      anon_sym_AMP,
      sym_metavariable,
    ACTIONS(3084), 32,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_for,
      anon_sym_impl,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_where,
      anon_sym_extern,
      anon_sym_dyn,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [36357] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3090), 12,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      anon_sym_AMP,
      sym_metavariable,
    ACTIONS(3088), 32,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_for,
      anon_sym_impl,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_where,
      anon_sym_extern,
      anon_sym_dyn,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [36410] = 18,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2990), 1,
      anon_sym_DOT_DOT,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3094), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2988), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3092), 19,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [36493] = 5,
    ACTIONS(3096), 1,
      anon_sym_POUND,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1153), 2,
      sym_attribute_item,
      aux_sym_enum_variant_list_repeat1,
    ACTIONS(2457), 9,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      anon_sym_AMP,
      sym_metavariable,
    ACTIONS(2455), 32,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_for,
      anon_sym_impl,
      anon_sym_pub,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_extern,
      anon_sym_dyn,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [36550] = 18,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2990), 1,
      anon_sym_DOT_DOT,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3101), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2988), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3099), 19,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [36633] = 7,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2990), 1,
      anon_sym_DOT_DOT,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2988), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3105), 13,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3103), 26,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [36694] = 20,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2990), 1,
      anon_sym_DOT_DOT,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2988), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3107), 8,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_COMMA,
      anon_sym_else,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [36781] = 8,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2990), 1,
      anon_sym_DOT_DOT,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2988), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2984), 13,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2974), 25,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [36844] = 9,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2990), 1,
      anon_sym_DOT_DOT,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2988), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2984), 10,
      anon_sym_PLUS,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2974), 25,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_QMARK,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [36909] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3050), 14,
      sym_raw_string_literal,
      sym_float_literal,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      sym_integer_literal,
      aux_sym_string_literal_token1,
      sym_char_literal,
      sym_metavariable,
    ACTIONS(3048), 29,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_const,
      anon_sym_default,
      anon_sym_union,
      anon_sym_ref,
      anon_sym__,
      sym_mutable_specifier,
      anon_sym_true,
      anon_sym_false,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [36961] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3080), 14,
      sym_raw_string_literal,
      sym_float_literal,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_POUND,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      sym_integer_literal,
      aux_sym_string_literal_token1,
      sym_char_literal,
      sym_metavariable,
    ACTIONS(3078), 29,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_const,
      anon_sym_default,
      anon_sym_union,
      anon_sym_ref,
      anon_sym__,
      sym_mutable_specifier,
      anon_sym_true,
      anon_sym_false,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [37013] = 7,
    ACTIONS(2590), 1,
      anon_sym_LBRACE,
    ACTIONS(2592), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3109), 1,
      anon_sym_BANG,
    STATE(1121), 1,
      sym_field_initializer_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(460), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(458), 24,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [37073] = 20,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    ACTIONS(3004), 1,
      sym_identifier,
    ACTIONS(3008), 1,
      anon_sym_LPAREN,
    ACTIONS(3010), 1,
      anon_sym_STAR,
    ACTIONS(3016), 1,
      anon_sym_for,
    ACTIONS(3018), 1,
      anon_sym_AMP,
    ACTIONS(3020), 1,
      sym_metavariable,
    STATE(1788), 1,
      sym_scoped_type_identifier,
    STATE(1790), 1,
      sym_where_predicate,
    STATE(1887), 1,
      sym_generic_type,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2404), 1,
      sym_scoped_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3014), 2,
      anon_sym_default,
      anon_sym_union,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(2296), 5,
      sym_higher_ranked_trait_bound,
      sym_lifetime,
      sym_tuple_type,
      sym_reference_type,
      sym_pointer_type,
    ACTIONS(3012), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [37158] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(1344), 10,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_SQUOTE,
      anon_sym_POUND,
      anon_sym_BANG,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      anon_sym_AMP,
      sym_metavariable,
    ACTIONS(1346), 32,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_for,
      anon_sym_impl,
      anon_sym_pub,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_extern,
      anon_sym_dyn,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [37209] = 21,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(2478), 1,
      anon_sym_fn,
    ACTIONS(2486), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3111), 1,
      sym_identifier,
    ACTIONS(3115), 1,
      anon_sym_default,
    ACTIONS(3117), 1,
      sym_metavariable,
    STATE(757), 1,
      sym_scoped_type_identifier,
    STATE(786), 1,
      sym_generic_type,
    STATE(1002), 1,
      sym_function_type,
    STATE(1195), 1,
      sym_for_lifetimes,
    STATE(2386), 1,
      sym_function_modifiers,
    STATE(2411), 1,
      sym_scoped_identifier,
    STATE(2491), 1,
      sym_bracketed_type,
    STATE(2492), 1,
      sym_generic_type_with_turbofish,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(2494), 3,
      sym_self,
      sym_super,
      sym_crate,
    ACTIONS(3113), 18,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_union,
  [37296] = 21,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(2478), 1,
      anon_sym_fn,
    ACTIONS(2486), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3115), 1,
      anon_sym_default,
    ACTIONS(3117), 1,
      sym_metavariable,
    ACTIONS(3119), 1,
      sym_identifier,
    STATE(759), 1,
      sym_scoped_type_identifier,
    STATE(805), 1,
      sym_generic_type,
    STATE(1012), 1,
      sym_function_type,
    STATE(1195), 1,
      sym_for_lifetimes,
    STATE(2386), 1,
      sym_function_modifiers,
    STATE(2411), 1,
      sym_scoped_identifier,
    STATE(2491), 1,
      sym_bracketed_type,
    STATE(2492), 1,
      sym_generic_type_with_turbofish,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(2494), 3,
      sym_self,
      sym_super,
      sym_crate,
    ACTIONS(3113), 18,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_union,
  [37383] = 6,
    ACTIONS(2626), 1,
      anon_sym_BANG,
    ACTIONS(2628), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3121), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2622), 16,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_as,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2620), 23,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_QMARK,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [37440] = 21,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3020), 1,
      sym_metavariable,
    ACTIONS(3123), 1,
      sym_identifier,
    ACTIONS(3125), 1,
      anon_sym_default,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1327), 1,
      sym_scoped_type_identifier,
    STATE(1355), 1,
      sym_generic_type,
    STATE(1384), 1,
      sym_function_type,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2404), 1,
      sym_scoped_identifier,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    ACTIONS(3014), 18,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_union,
  [37527] = 20,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    ACTIONS(3004), 1,
      sym_identifier,
    ACTIONS(3008), 1,
      anon_sym_LPAREN,
    ACTIONS(3010), 1,
      anon_sym_STAR,
    ACTIONS(3016), 1,
      anon_sym_for,
    ACTIONS(3018), 1,
      anon_sym_AMP,
    ACTIONS(3020), 1,
      sym_metavariable,
    STATE(1788), 1,
      sym_scoped_type_identifier,
    STATE(1887), 1,
      sym_generic_type,
    STATE(1953), 1,
      sym_where_predicate,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2404), 1,
      sym_scoped_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3014), 2,
      anon_sym_default,
      anon_sym_union,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(2296), 5,
      sym_higher_ranked_trait_bound,
      sym_lifetime,
      sym_tuple_type,
      sym_reference_type,
      sym_pointer_type,
    ACTIONS(3012), 17,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
  [37612] = 21,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(670), 1,
      anon_sym_fn,
    ACTIONS(672), 1,
      anon_sym_for,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3020), 1,
      sym_metavariable,
    ACTIONS(3125), 1,
      anon_sym_default,
    ACTIONS(3127), 1,
      sym_identifier,
    STATE(1196), 1,
      sym_for_lifetimes,
    STATE(1329), 1,
      sym_scoped_type_identifier,
    STATE(1362), 1,
      sym_generic_type,
    STATE(1401), 1,
      sym_function_type,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2404), 1,
      sym_scoped_identifier,
    STATE(2472), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    ACTIONS(3014), 18,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_union,
  [37699] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2892), 10,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      anon_sym_AMP,
      sym_metavariable,
    ACTIONS(2894), 32,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_for,
      anon_sym_impl,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_extern,
      anon_sym_dyn,
      sym_mutable_specifier,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [37750] = 6,
    ACTIONS(2556), 1,
      anon_sym_BANG,
    ACTIONS(3129), 1,
      anon_sym_COLON_COLON,
    STATE(1121), 1,
      sym_field_initializer_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(460), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(458), 23,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [37806] = 5,
    ACTIONS(2696), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3109), 1,
      anon_sym_BANG,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(460), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(458), 24,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [37860] = 16,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(702), 1,
      anon_sym_DASH,
    ACTIONS(706), 1,
      aux_sym_string_literal_token1,
    ACTIONS(1042), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3131), 1,
      sym_identifier,
    ACTIONS(3137), 1,
      sym_metavariable,
    STATE(1410), 1,
      sym_scoped_identifier,
    STATE(1425), 1,
      sym__literal_pattern,
    STATE(2369), 1,
      sym_bracketed_type,
    STATE(2457), 1,
      sym_generic_type_with_turbofish,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(708), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(3135), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1397), 3,
      sym_negative_literal,
      sym_string_literal,
      sym_boolean_literal,
    ACTIONS(704), 4,
      sym_raw_string_literal,
      sym_float_literal,
      sym_integer_literal,
      sym_char_literal,
    ACTIONS(3133), 19,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_default,
      anon_sym_union,
  [37936] = 16,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(702), 1,
      anon_sym_DASH,
    ACTIONS(706), 1,
      aux_sym_string_literal_token1,
    ACTIONS(1042), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3139), 1,
      sym_identifier,
    ACTIONS(3145), 1,
      sym_metavariable,
    STATE(1413), 1,
      sym_scoped_identifier,
    STATE(1447), 1,
      sym__literal_pattern,
    STATE(2369), 1,
      sym_bracketed_type,
    STATE(2457), 1,
      sym_generic_type_with_turbofish,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(708), 2,
      anon_sym_true,
      anon_sym_false,
    ACTIONS(3143), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1397), 3,
      sym_negative_literal,
      sym_string_literal,
      sym_boolean_literal,
    ACTIONS(704), 4,
      sym_raw_string_literal,
      sym_float_literal,
      sym_integer_literal,
      sym_char_literal,
    ACTIONS(3141), 19,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_default,
      anon_sym_union,
  [38012] = 23,
    ACTIONS(624), 1,
      anon_sym_RBRACK,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3147), 1,
      anon_sym_SEMI,
    ACTIONS(3149), 1,
      anon_sym_COMMA,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    STATE(1868), 1,
      aux_sym_array_expression_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [38101] = 5,
    ACTIONS(2556), 1,
      anon_sym_BANG,
    ACTIONS(3155), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(460), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(458), 23,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [38154] = 5,
    ACTIONS(2664), 1,
      anon_sym_BANG,
    ACTIONS(3157), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2622), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2620), 23,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [38207] = 4,
    ACTIONS(3163), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3161), 8,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_LT,
      anon_sym_AMP,
      sym_metavariable,
    ACTIONS(3159), 31,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_for,
      anon_sym_impl,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_extern,
      anon_sym_dyn,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [38258] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3060), 9,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      anon_sym_AMP,
      sym_metavariable,
    ACTIONS(3058), 31,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_for,
      anon_sym_impl,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_extern,
      anon_sym_dyn,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [38307] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3167), 9,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      anon_sym_AMP,
      sym_metavariable,
    ACTIONS(3165), 31,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_for,
      anon_sym_impl,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_extern,
      anon_sym_dyn,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [38356] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2606), 16,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2608), 24,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [38405] = 23,
    ACTIONS(374), 1,
      anon_sym_RBRACK,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3169), 1,
      anon_sym_SEMI,
    ACTIONS(3171), 1,
      anon_sym_COMMA,
    STATE(1974), 1,
      aux_sym_array_expression_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [38494] = 4,
    ACTIONS(3173), 1,
      anon_sym_LPAREN,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3161), 8,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      anon_sym_AMP,
      sym_metavariable,
    ACTIONS(3159), 31,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_for,
      anon_sym_impl,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_extern,
      anon_sym_dyn,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [38545] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3177), 9,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      anon_sym_AMP,
      sym_metavariable,
    ACTIONS(3175), 31,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_for,
      anon_sym_impl,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_extern,
      anon_sym_dyn,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [38594] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3181), 9,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      anon_sym_AMP,
      sym_metavariable,
    ACTIONS(3179), 31,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_for,
      anon_sym_impl,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_extern,
      anon_sym_dyn,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [38643] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2614), 16,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2616), 24,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [38692] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2582), 16,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2584), 24,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [38741] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3042), 9,
      anon_sym_LPAREN,
      anon_sym_LBRACK,
      anon_sym_STAR,
      anon_sym_SQUOTE,
      anon_sym_BANG,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      anon_sym_AMP,
      sym_metavariable,
    ACTIONS(3040), 31,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_for,
      anon_sym_impl,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_extern,
      anon_sym_dyn,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [38790] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2574), 16,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_BANG,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2576), 24,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_COLON_COLON,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [38839] = 4,
    ACTIONS(2700), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2702), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2698), 23,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [38889] = 4,
    ACTIONS(2712), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2702), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2698), 23,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [38939] = 4,
    ACTIONS(3183), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(460), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(458), 23,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [38989] = 22,
    ACTIONS(396), 1,
      anon_sym_RPAREN,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3185), 1,
      anon_sym_COMMA,
    STATE(1976), 1,
      aux_sym_arguments_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [39075] = 4,
    ACTIONS(3157), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2622), 15,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DOT_DOT,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
      anon_sym_DOT,
    ACTIONS(2620), 23,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_LBRACK,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [39125] = 18,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3020), 1,
      sym_metavariable,
    ACTIONS(3125), 1,
      anon_sym_default,
    ACTIONS(3187), 1,
      sym_identifier,
    ACTIONS(3189), 1,
      anon_sym_fn,
    STATE(1818), 1,
      sym_scoped_type_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2404), 1,
      sym_scoped_identifier,
    STATE(2413), 1,
      sym_function_modifiers,
    STATE(2488), 1,
      sym_generic_type,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    ACTIONS(3014), 18,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_union,
  [39203] = 18,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(684), 1,
      anon_sym_extern,
    ACTIONS(870), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3020), 1,
      sym_metavariable,
    ACTIONS(3125), 1,
      anon_sym_default,
    ACTIONS(3191), 1,
      sym_identifier,
    ACTIONS(3193), 1,
      anon_sym_fn,
    STATE(1792), 1,
      sym_scoped_type_identifier,
    STATE(2378), 1,
      sym_bracketed_type,
    STATE(2379), 1,
      sym_generic_type_with_turbofish,
    STATE(2394), 1,
      sym_function_modifiers,
    STATE(2404), 1,
      sym_scoped_identifier,
    STATE(2488), 1,
      sym_generic_type,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1526), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(664), 3,
      anon_sym_async,
      anon_sym_const,
      anon_sym_unsafe,
    ACTIONS(876), 3,
      sym_self,
      sym_super,
      sym_crate,
    ACTIONS(3014), 18,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_union,
  [39281] = 22,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3195), 1,
      anon_sym_RPAREN,
    ACTIONS(3197), 1,
      anon_sym_COMMA,
    STATE(2096), 1,
      aux_sym_arguments_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [39367] = 18,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3046), 1,
      anon_sym_EQ,
    ACTIONS(3205), 1,
      anon_sym_AMP,
    ACTIONS(3209), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3211), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3213), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3215), 1,
      anon_sym_PIPE,
    ACTIONS(3217), 1,
      anon_sym_CARET,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3199), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3203), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3207), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3221), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3201), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3219), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3044), 13,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_QMARK,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [39444] = 21,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3223), 1,
      anon_sym_SEMI,
    ACTIONS(3225), 1,
      anon_sym_else,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [39527] = 7,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3209), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3207), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3105), 13,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3103), 20,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [39582] = 21,
    ACTIONS(284), 1,
      anon_sym_LBRACE,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3205), 1,
      anon_sym_AMP,
    ACTIONS(3211), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3213), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3215), 1,
      anon_sym_PIPE,
    ACTIONS(3217), 1,
      anon_sym_CARET,
    ACTIONS(3227), 1,
      anon_sym_EQ,
    ACTIONS(3231), 1,
      anon_sym_DOT_DOT,
    STATE(1127), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3199), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3203), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3221), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3229), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3201), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3219), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3233), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [39665] = 21,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3235), 1,
      anon_sym_SEMI,
    ACTIONS(3237), 1,
      anon_sym_else,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [39748] = 8,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3209), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3207), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2984), 13,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2974), 19,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_QMARK,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [39805] = 21,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3239), 1,
      anon_sym_SEMI,
    ACTIONS(3241), 1,
      anon_sym_else,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [39888] = 18,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3101), 1,
      anon_sym_EQ,
    ACTIONS(3205), 1,
      anon_sym_AMP,
    ACTIONS(3209), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3211), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3213), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3215), 1,
      anon_sym_PIPE,
    ACTIONS(3217), 1,
      anon_sym_CARET,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3199), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3203), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3207), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3221), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3201), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3219), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3099), 13,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_QMARK,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [39965] = 21,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3243), 1,
      anon_sym_SEMI,
    ACTIONS(3245), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [40048] = 15,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(3247), 1,
      sym_identifier,
    ACTIONS(3249), 1,
      anon_sym_LBRACE,
    ACTIONS(3251), 1,
      anon_sym_RBRACE,
    ACTIONS(3253), 1,
      anon_sym_STAR,
    ACTIONS(3257), 1,
      anon_sym_COMMA,
    ACTIONS(3259), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3263), 1,
      sym_metavariable,
    STATE(1749), 1,
      sym_scoped_identifier,
    STATE(2457), 1,
      sym_generic_type_with_turbofish,
    STATE(2527), 1,
      sym_bracketed_type,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3261), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(1951), 5,
      sym__use_clause,
      sym_scoped_use_list,
      sym_use_list,
      sym_use_as_clause,
      sym_use_wildcard,
    ACTIONS(3255), 19,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_default,
      anon_sym_union,
  [40119] = 7,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3209), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3207), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3002), 13,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3000), 20,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [40174] = 21,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3265), 1,
      anon_sym_SEMI,
    ACTIONS(3267), 1,
      anon_sym_else,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [40257] = 21,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3205), 1,
      anon_sym_AMP,
    ACTIONS(3211), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3213), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3215), 1,
      anon_sym_PIPE,
    ACTIONS(3217), 1,
      anon_sym_CARET,
    ACTIONS(3227), 1,
      anon_sym_EQ,
    ACTIONS(3231), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3269), 1,
      anon_sym_LBRACE,
    STATE(60), 1,
      sym_match_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3199), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3203), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3221), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3229), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3201), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3219), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3233), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [40340] = 21,
    ACTIONS(15), 1,
      anon_sym_LBRACE,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3205), 1,
      anon_sym_AMP,
    ACTIONS(3211), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3213), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3215), 1,
      anon_sym_PIPE,
    ACTIONS(3217), 1,
      anon_sym_CARET,
    ACTIONS(3227), 1,
      anon_sym_EQ,
    ACTIONS(3231), 1,
      anon_sym_DOT_DOT,
    STATE(69), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3199), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3203), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3221), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3229), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3201), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3219), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3233), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [40423] = 20,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3271), 2,
      anon_sym_RBRACK,
      anon_sym_COMMA,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [40504] = 21,
    ACTIONS(270), 1,
      anon_sym_RBRACE,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3243), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [40587] = 21,
    ACTIONS(632), 1,
      anon_sym_LBRACE,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3205), 1,
      anon_sym_AMP,
    ACTIONS(3211), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3213), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3215), 1,
      anon_sym_PIPE,
    ACTIONS(3217), 1,
      anon_sym_CARET,
    ACTIONS(3227), 1,
      anon_sym_EQ,
    ACTIONS(3231), 1,
      anon_sym_DOT_DOT,
    STATE(229), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3199), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3203), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3221), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3229), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3201), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3219), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3233), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [40670] = 21,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3273), 1,
      anon_sym_RBRACE,
    ACTIONS(3275), 1,
      anon_sym_COMMA,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [40753] = 21,
    ACTIONS(272), 1,
      anon_sym_RBRACE,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3243), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [40836] = 21,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3205), 1,
      anon_sym_AMP,
    ACTIONS(3211), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3213), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3215), 1,
      anon_sym_PIPE,
    ACTIONS(3217), 1,
      anon_sym_CARET,
    ACTIONS(3227), 1,
      anon_sym_EQ,
    ACTIONS(3231), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3277), 1,
      anon_sym_LBRACE,
    STATE(985), 1,
      sym_match_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3199), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3203), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3221), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3229), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3201), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3219), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3233), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [40919] = 20,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3279), 2,
      anon_sym_RBRACE,
      anon_sym_COMMA,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [41000] = 21,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3243), 1,
      anon_sym_SEMI,
    ACTIONS(3281), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [41083] = 21,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3243), 1,
      anon_sym_SEMI,
    ACTIONS(3283), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [41166] = 18,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3076), 1,
      anon_sym_EQ,
    ACTIONS(3205), 1,
      anon_sym_AMP,
    ACTIONS(3209), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3211), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3213), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3215), 1,
      anon_sym_PIPE,
    ACTIONS(3217), 1,
      anon_sym_CARET,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3199), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3203), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3207), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3221), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3201), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3219), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3074), 13,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_QMARK,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [41243] = 21,
    ACTIONS(632), 1,
      anon_sym_LBRACE,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3205), 1,
      anon_sym_AMP,
    ACTIONS(3211), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3213), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3215), 1,
      anon_sym_PIPE,
    ACTIONS(3217), 1,
      anon_sym_CARET,
    ACTIONS(3227), 1,
      anon_sym_EQ,
    ACTIONS(3231), 1,
      anon_sym_DOT_DOT,
    STATE(224), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3199), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3203), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3221), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3229), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3201), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3219), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3233), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [41326] = 20,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3285), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [41407] = 10,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3209), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3199), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3207), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3201), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2984), 8,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2974), 19,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_QMARK,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [41468] = 20,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3287), 2,
      anon_sym_RBRACE,
      anon_sym_COMMA,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [41549] = 21,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3289), 1,
      anon_sym_RPAREN,
    ACTIONS(3291), 1,
      anon_sym_COMMA,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [41632] = 20,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3293), 2,
      anon_sym_RBRACE,
      anon_sym_COMMA,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [41713] = 21,
    ACTIONS(107), 1,
      anon_sym_RBRACE,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3243), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [41796] = 21,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3295), 1,
      anon_sym_RBRACE,
    ACTIONS(3297), 1,
      anon_sym_COMMA,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [41879] = 21,
    ACTIONS(274), 1,
      anon_sym_RBRACE,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3243), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [41962] = 21,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3299), 1,
      anon_sym_SEMI,
    ACTIONS(3301), 1,
      anon_sym_else,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [42045] = 12,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3205), 1,
      anon_sym_AMP,
    ACTIONS(3209), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3199), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3207), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3221), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3201), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2984), 5,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
      anon_sym_CARET,
    ACTIONS(2974), 19,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_QMARK,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [42110] = 21,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3303), 1,
      anon_sym_SEMI,
    ACTIONS(3305), 1,
      anon_sym_else,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [42193] = 21,
    ACTIONS(594), 1,
      anon_sym_RPAREN,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3307), 1,
      anon_sym_COMMA,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [42276] = 13,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3205), 1,
      anon_sym_AMP,
    ACTIONS(3209), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3217), 1,
      anon_sym_CARET,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3199), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3207), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3221), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3201), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2984), 4,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_PIPE,
    ACTIONS(2974), 19,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_QMARK,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [42343] = 17,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2984), 1,
      anon_sym_EQ,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3205), 1,
      anon_sym_AMP,
    ACTIONS(3209), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3211), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3215), 1,
      anon_sym_PIPE,
    ACTIONS(3217), 1,
      anon_sym_CARET,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3199), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3203), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3207), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3221), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3201), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3219), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(2974), 14,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_QMARK,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [42418] = 21,
    ACTIONS(284), 1,
      anon_sym_LBRACE,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3205), 1,
      anon_sym_AMP,
    ACTIONS(3211), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3213), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3215), 1,
      anon_sym_PIPE,
    ACTIONS(3217), 1,
      anon_sym_CARET,
    ACTIONS(3227), 1,
      anon_sym_EQ,
    ACTIONS(3231), 1,
      anon_sym_DOT_DOT,
    STATE(870), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3199), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3203), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3221), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3229), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3201), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3219), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3233), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [42501] = 18,
    ACTIONS(288), 1,
      anon_sym_EQ,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3205), 1,
      anon_sym_AMP,
    ACTIONS(3209), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3211), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3213), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3215), 1,
      anon_sym_PIPE,
    ACTIONS(3217), 1,
      anon_sym_CARET,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3199), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3203), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3207), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3221), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3201), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3219), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(282), 13,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_QMARK,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [42578] = 21,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3243), 1,
      anon_sym_SEMI,
    ACTIONS(3309), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [42661] = 16,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2984), 1,
      anon_sym_EQ,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3205), 1,
      anon_sym_AMP,
    ACTIONS(3209), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3215), 1,
      anon_sym_PIPE,
    ACTIONS(3217), 1,
      anon_sym_CARET,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3199), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3203), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3207), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3221), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3201), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3219), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(2974), 15,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_QMARK,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [42734] = 11,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3209), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3199), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3207), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3221), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3201), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2984), 6,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_PIPE,
      anon_sym_CARET,
    ACTIONS(2974), 19,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_QMARK,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [42797] = 21,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3243), 1,
      anon_sym_SEMI,
    ACTIONS(3311), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [42880] = 21,
    ACTIONS(612), 1,
      anon_sym_RPAREN,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3307), 1,
      anon_sym_COMMA,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [42963] = 14,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3205), 1,
      anon_sym_AMP,
    ACTIONS(3209), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3215), 1,
      anon_sym_PIPE,
    ACTIONS(3217), 1,
      anon_sym_CARET,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3199), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3207), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3221), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2984), 3,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3201), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2974), 19,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_QMARK,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [43032] = 21,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3307), 1,
      anon_sym_COMMA,
    ACTIONS(3313), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [43115] = 21,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3315), 1,
      anon_sym_SEMI,
    ACTIONS(3317), 1,
      anon_sym_else,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [43198] = 18,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3094), 1,
      anon_sym_EQ,
    ACTIONS(3205), 1,
      anon_sym_AMP,
    ACTIONS(3209), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3211), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3213), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3215), 1,
      anon_sym_PIPE,
    ACTIONS(3217), 1,
      anon_sym_CARET,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3199), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3203), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3207), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3221), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3201), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3219), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3092), 13,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_QMARK,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [43275] = 21,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3205), 1,
      anon_sym_AMP,
    ACTIONS(3211), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3213), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3215), 1,
      anon_sym_PIPE,
    ACTIONS(3217), 1,
      anon_sym_CARET,
    ACTIONS(3227), 1,
      anon_sym_EQ,
    ACTIONS(3231), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3319), 1,
      anon_sym_LBRACE,
    STATE(215), 1,
      sym_match_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3199), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3203), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3221), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3229), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3201), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3219), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3233), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [43358] = 9,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3209), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3207), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3201), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(2984), 10,
      anon_sym_PLUS,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(2974), 19,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_QMARK,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [43417] = 20,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3321), 2,
      anon_sym_RBRACE,
      anon_sym_COMMA,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [43498] = 21,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3323), 1,
      anon_sym_SEMI,
    ACTIONS(3325), 1,
      anon_sym_else,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [43581] = 20,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3327), 2,
      anon_sym_RBRACE,
      anon_sym_COMMA,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [43662] = 7,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3209), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3207), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3064), 13,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_EQ,
      anon_sym_LT,
      anon_sym_GT,
      anon_sym_AMP,
      anon_sym_DASH,
      anon_sym_PIPE,
      anon_sym_CARET,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3062), 20,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_QMARK,
      anon_sym_as,
      anon_sym_AMP_AMP,
      anon_sym_PIPE_PIPE,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [43717] = 20,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3329), 2,
      anon_sym_RBRACE,
      anon_sym_COMMA,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [43798] = 20,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3331), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [43879] = 20,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3205), 1,
      anon_sym_AMP,
    ACTIONS(3209), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3211), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3213), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3215), 1,
      anon_sym_PIPE,
    ACTIONS(3217), 1,
      anon_sym_CARET,
    ACTIONS(3227), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3066), 2,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
    ACTIONS(3199), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3203), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3207), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3221), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3201), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3219), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3233), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [43960] = 21,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3243), 1,
      anon_sym_SEMI,
    ACTIONS(3333), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [44043] = 20,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3335), 2,
      anon_sym_RBRACE,
      anon_sym_COMMA,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [44124] = 21,
    ACTIONS(15), 1,
      anon_sym_LBRACE,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3205), 1,
      anon_sym_AMP,
    ACTIONS(3211), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3213), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3215), 1,
      anon_sym_PIPE,
    ACTIONS(3217), 1,
      anon_sym_CARET,
    ACTIONS(3227), 1,
      anon_sym_EQ,
    ACTIONS(3231), 1,
      anon_sym_DOT_DOT,
    STATE(65), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3199), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3203), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3221), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3229), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3201), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3219), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3233), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [44207] = 21,
    ACTIONS(109), 1,
      anon_sym_RBRACE,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3243), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [44290] = 21,
    ACTIONS(276), 1,
      anon_sym_RBRACE,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3243), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [44373] = 20,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3205), 1,
      anon_sym_AMP,
    ACTIONS(3209), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3211), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3213), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3215), 1,
      anon_sym_PIPE,
    ACTIONS(3217), 1,
      anon_sym_CARET,
    ACTIONS(3227), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3107), 2,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
    ACTIONS(3199), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3203), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3207), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3221), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3201), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3219), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3233), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [44454] = 18,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3024), 1,
      anon_sym_EQ,
    ACTIONS(3205), 1,
      anon_sym_AMP,
    ACTIONS(3209), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3211), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3213), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3215), 1,
      anon_sym_PIPE,
    ACTIONS(3217), 1,
      anon_sym_CARET,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3199), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3203), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3207), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3221), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3201), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3219), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3022), 13,
      anon_sym_LPAREN,
      anon_sym_LBRACE,
      anon_sym_QMARK,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [44531] = 20,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3337), 1,
      anon_sym_RBRACK,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [44611] = 20,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3339), 1,
      anon_sym_RBRACK,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [44691] = 20,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3243), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [44771] = 20,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3341), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [44851] = 19,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3205), 1,
      anon_sym_AMP,
    ACTIONS(3213), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3215), 1,
      anon_sym_PIPE,
    ACTIONS(3217), 1,
      anon_sym_CARET,
    ACTIONS(3227), 1,
      anon_sym_EQ,
    ACTIONS(3231), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3199), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3203), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3221), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3229), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3343), 2,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
    ACTIONS(3201), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3219), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3233), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [44929] = 20,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3345), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [45009] = 19,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3205), 1,
      anon_sym_AMP,
    ACTIONS(3213), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3215), 1,
      anon_sym_PIPE,
    ACTIONS(3217), 1,
      anon_sym_CARET,
    ACTIONS(3227), 1,
      anon_sym_EQ,
    ACTIONS(3231), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3199), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3203), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3221), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3229), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3347), 2,
      anon_sym_LBRACE,
      anon_sym_AMP_AMP,
    ACTIONS(3201), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3219), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3233), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [45087] = 20,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3349), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [45167] = 20,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3205), 1,
      anon_sym_AMP,
    ACTIONS(3213), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3215), 1,
      anon_sym_PIPE,
    ACTIONS(3217), 1,
      anon_sym_CARET,
    ACTIONS(3227), 1,
      anon_sym_EQ,
    ACTIONS(3231), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3351), 1,
      anon_sym_LBRACE,
    ACTIONS(3353), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3199), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(3203), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3221), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3229), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3201), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3219), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3233), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [45247] = 20,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3355), 1,
      anon_sym_COMMA,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [45327] = 20,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3357), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [45407] = 20,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3359), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [45487] = 20,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3361), 1,
      anon_sym_RBRACK,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [45567] = 20,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3363), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [45647] = 20,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3365), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [45727] = 20,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3367), 1,
      anon_sym_RBRACK,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [45807] = 19,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3347), 2,
      anon_sym_EQ_GT,
      anon_sym_AMP_AMP,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [45885] = 20,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3369), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [45965] = 20,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3371), 1,
      anon_sym_RBRACK,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [46045] = 19,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3343), 2,
      anon_sym_EQ_GT,
      anon_sym_AMP_AMP,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [46123] = 20,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3373), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [46203] = 20,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3375), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [46283] = 20,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3377), 1,
      anon_sym_COMMA,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [46363] = 20,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3351), 1,
      anon_sym_EQ_GT,
    ACTIONS(3379), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [46443] = 20,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3381), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [46523] = 14,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(3247), 1,
      sym_identifier,
    ACTIONS(3249), 1,
      anon_sym_LBRACE,
    ACTIONS(3253), 1,
      anon_sym_STAR,
    ACTIONS(3259), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3263), 1,
      sym_metavariable,
    ACTIONS(3383), 1,
      anon_sym_RBRACE,
    STATE(1749), 1,
      sym_scoped_identifier,
    STATE(2457), 1,
      sym_generic_type_with_turbofish,
    STATE(2527), 1,
      sym_bracketed_type,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3261), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(2207), 5,
      sym__use_clause,
      sym_scoped_use_list,
      sym_use_list,
      sym_use_as_clause,
      sym_use_wildcard,
    ACTIONS(3255), 19,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_default,
      anon_sym_union,
  [46591] = 20,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3385), 1,
      anon_sym_RBRACK,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [46671] = 20,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3307), 1,
      anon_sym_COMMA,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [46751] = 20,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3387), 1,
      anon_sym_COMMA,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [46831] = 20,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3389), 1,
      anon_sym_RBRACK,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [46911] = 20,
    ACTIONS(2976), 1,
      anon_sym_LBRACK,
    ACTIONS(2982), 1,
      anon_sym_as,
    ACTIONS(2986), 1,
      anon_sym_AMP,
    ACTIONS(2992), 1,
      anon_sym_PIPE,
    ACTIONS(2994), 1,
      anon_sym_CARET,
    ACTIONS(2998), 1,
      anon_sym_DOT,
    ACTIONS(3028), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3030), 1,
      anon_sym_PIPE_PIPE,
    ACTIONS(3068), 1,
      anon_sym_QMARK,
    ACTIONS(3070), 1,
      anon_sym_EQ,
    ACTIONS(3153), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3391), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2978), 2,
      anon_sym_PLUS,
      anon_sym_DASH,
    ACTIONS(2996), 2,
      anon_sym_LT_LT,
      anon_sym_GT_GT,
    ACTIONS(3026), 2,
      anon_sym_LT,
      anon_sym_GT,
    ACTIONS(3151), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2980), 3,
      anon_sym_STAR,
      anon_sym_SLASH,
      anon_sym_PERCENT,
    ACTIONS(3032), 4,
      anon_sym_EQ_EQ,
      anon_sym_BANG_EQ,
      anon_sym_LT_EQ,
      anon_sym_GT_EQ,
    ACTIONS(3072), 10,
      anon_sym_PLUS_EQ,
      anon_sym_DASH_EQ,
      anon_sym_STAR_EQ,
      anon_sym_SLASH_EQ,
      anon_sym_PERCENT_EQ,
      anon_sym_AMP_EQ,
      anon_sym_PIPE_EQ,
      anon_sym_CARET_EQ,
      anon_sym_LT_LT_EQ,
      anon_sym_GT_GT_EQ,
  [46991] = 14,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(3247), 1,
      sym_identifier,
    ACTIONS(3249), 1,
      anon_sym_LBRACE,
    ACTIONS(3253), 1,
      anon_sym_STAR,
    ACTIONS(3259), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3263), 1,
      sym_metavariable,
    ACTIONS(3393), 1,
      anon_sym_RBRACE,
    STATE(1749), 1,
      sym_scoped_identifier,
    STATE(2457), 1,
      sym_generic_type_with_turbofish,
    STATE(2527), 1,
      sym_bracketed_type,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3261), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(2207), 5,
      sym__use_clause,
      sym_scoped_use_list,
      sym_use_list,
      sym_use_as_clause,
      sym_use_wildcard,
    ACTIONS(3255), 19,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_default,
      anon_sym_union,
  [47059] = 13,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(3247), 1,
      sym_identifier,
    ACTIONS(3249), 1,
      anon_sym_LBRACE,
    ACTIONS(3253), 1,
      anon_sym_STAR,
    ACTIONS(3259), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3263), 1,
      sym_metavariable,
    STATE(1749), 1,
      sym_scoped_identifier,
    STATE(2457), 1,
      sym_generic_type_with_turbofish,
    STATE(2527), 1,
      sym_bracketed_type,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3261), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(2207), 5,
      sym__use_clause,
      sym_scoped_use_list,
      sym_use_list,
      sym_use_as_clause,
      sym_use_wildcard,
    ACTIONS(3255), 19,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_default,
      anon_sym_union,
  [47124] = 13,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(3247), 1,
      sym_identifier,
    ACTIONS(3249), 1,
      anon_sym_LBRACE,
    ACTIONS(3253), 1,
      anon_sym_STAR,
    ACTIONS(3259), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3263), 1,
      sym_metavariable,
    STATE(1749), 1,
      sym_scoped_identifier,
    STATE(2457), 1,
      sym_generic_type_with_turbofish,
    STATE(2527), 1,
      sym_bracketed_type,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3261), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(2330), 5,
      sym__use_clause,
      sym_scoped_use_list,
      sym_use_list,
      sym_use_as_clause,
      sym_use_wildcard,
    ACTIONS(3255), 19,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_default,
      anon_sym_union,
  [47189] = 13,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(3247), 1,
      sym_identifier,
    ACTIONS(3249), 1,
      anon_sym_LBRACE,
    ACTIONS(3253), 1,
      anon_sym_STAR,
    ACTIONS(3259), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3263), 1,
      sym_metavariable,
    STATE(1749), 1,
      sym_scoped_identifier,
    STATE(2457), 1,
      sym_generic_type_with_turbofish,
    STATE(2527), 1,
      sym_bracketed_type,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3261), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(2475), 5,
      sym__use_clause,
      sym_scoped_use_list,
      sym_use_list,
      sym_use_as_clause,
      sym_use_wildcard,
    ACTIONS(3255), 19,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_default,
      anon_sym_union,
  [47254] = 13,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(3247), 1,
      sym_identifier,
    ACTIONS(3249), 1,
      anon_sym_LBRACE,
    ACTIONS(3253), 1,
      anon_sym_STAR,
    ACTIONS(3259), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3263), 1,
      sym_metavariable,
    STATE(1749), 1,
      sym_scoped_identifier,
    STATE(2457), 1,
      sym_generic_type_with_turbofish,
    STATE(2527), 1,
      sym_bracketed_type,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3261), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(2506), 5,
      sym__use_clause,
      sym_scoped_use_list,
      sym_use_list,
      sym_use_as_clause,
      sym_use_wildcard,
    ACTIONS(3255), 19,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_default,
      anon_sym_union,
  [47319] = 13,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(3247), 1,
      sym_identifier,
    ACTIONS(3249), 1,
      anon_sym_LBRACE,
    ACTIONS(3253), 1,
      anon_sym_STAR,
    ACTIONS(3259), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3263), 1,
      sym_metavariable,
    STATE(1749), 1,
      sym_scoped_identifier,
    STATE(2457), 1,
      sym_generic_type_with_turbofish,
    STATE(2527), 1,
      sym_bracketed_type,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3261), 3,
      sym_self,
      sym_super,
      sym_crate,
    STATE(2458), 5,
      sym__use_clause,
      sym_scoped_use_list,
      sym_use_list,
      sym_use_as_clause,
      sym_use_wildcard,
    ACTIONS(3255), 19,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_default,
      anon_sym_union,
  [47384] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3397), 3,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(3395), 28,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [47424] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3401), 3,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(3399), 28,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [47464] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3405), 3,
      anon_sym_LT,
      anon_sym_COLON_COLON,
      sym_metavariable,
    ACTIONS(3403), 28,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_extern,
      sym_identifier,
      sym_self,
      sym_super,
      sym_crate,
  [47504] = 11,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(1042), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3407), 1,
      sym_identifier,
    ACTIONS(3413), 1,
      sym_metavariable,
    STATE(1599), 1,
      sym_scoped_identifier,
    STATE(2444), 1,
      sym_attribute,
    STATE(2457), 1,
      sym_generic_type_with_turbofish,
    STATE(2527), 1,
      sym_bracketed_type,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3411), 3,
      sym_self,
      sym_super,
      sym_crate,
    ACTIONS(3409), 19,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_default,
      anon_sym_union,
  [47559] = 11,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(1042), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3407), 1,
      sym_identifier,
    ACTIONS(3413), 1,
      sym_metavariable,
    STATE(1599), 1,
      sym_scoped_identifier,
    STATE(2457), 1,
      sym_generic_type_with_turbofish,
    STATE(2477), 1,
      sym_attribute,
    STATE(2527), 1,
      sym_bracketed_type,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3411), 3,
      sym_self,
      sym_super,
      sym_crate,
    ACTIONS(3409), 19,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_default,
      anon_sym_union,
  [47614] = 11,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(1042), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3407), 1,
      sym_identifier,
    ACTIONS(3413), 1,
      sym_metavariable,
    STATE(1599), 1,
      sym_scoped_identifier,
    STATE(2342), 1,
      sym_attribute,
    STATE(2457), 1,
      sym_generic_type_with_turbofish,
    STATE(2527), 1,
      sym_bracketed_type,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3411), 3,
      sym_self,
      sym_super,
      sym_crate,
    ACTIONS(3409), 19,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_default,
      anon_sym_union,
  [47669] = 11,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(1042), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3407), 1,
      sym_identifier,
    ACTIONS(3413), 1,
      sym_metavariable,
    STATE(1599), 1,
      sym_scoped_identifier,
    STATE(2433), 1,
      sym_attribute,
    STATE(2457), 1,
      sym_generic_type_with_turbofish,
    STATE(2527), 1,
      sym_bracketed_type,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3411), 3,
      sym_self,
      sym_super,
      sym_crate,
    ACTIONS(3409), 19,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_default,
      anon_sym_union,
  [47724] = 11,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(1042), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3407), 1,
      sym_identifier,
    ACTIONS(3413), 1,
      sym_metavariable,
    STATE(1599), 1,
      sym_scoped_identifier,
    STATE(2363), 1,
      sym_attribute,
    STATE(2457), 1,
      sym_generic_type_with_turbofish,
    STATE(2527), 1,
      sym_bracketed_type,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3411), 3,
      sym_self,
      sym_super,
      sym_crate,
    ACTIONS(3409), 19,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_default,
      anon_sym_union,
  [47779] = 11,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(1042), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3407), 1,
      sym_identifier,
    ACTIONS(3413), 1,
      sym_metavariable,
    STATE(1599), 1,
      sym_scoped_identifier,
    STATE(2322), 1,
      sym_attribute,
    STATE(2457), 1,
      sym_generic_type_with_turbofish,
    STATE(2527), 1,
      sym_bracketed_type,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3411), 3,
      sym_self,
      sym_super,
      sym_crate,
    ACTIONS(3409), 19,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_default,
      anon_sym_union,
  [47834] = 11,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(1042), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3407), 1,
      sym_identifier,
    ACTIONS(3413), 1,
      sym_metavariable,
    STATE(1599), 1,
      sym_scoped_identifier,
    STATE(2339), 1,
      sym_attribute,
    STATE(2457), 1,
      sym_generic_type_with_turbofish,
    STATE(2527), 1,
      sym_bracketed_type,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3411), 3,
      sym_self,
      sym_super,
      sym_crate,
    ACTIONS(3409), 19,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_default,
      anon_sym_union,
  [47889] = 10,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(1042), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3415), 1,
      sym_identifier,
    ACTIONS(3421), 1,
      sym_metavariable,
    STATE(2212), 1,
      sym_scoped_identifier,
    STATE(2457), 1,
      sym_generic_type_with_turbofish,
    STATE(2527), 1,
      sym_bracketed_type,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3419), 3,
      sym_self,
      sym_super,
      sym_crate,
    ACTIONS(3417), 19,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_default,
      anon_sym_union,
  [47941] = 10,
    ACTIONS(77), 1,
      anon_sym_LT,
    ACTIONS(1042), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3423), 1,
      sym_identifier,
    ACTIONS(3429), 1,
      sym_metavariable,
    STATE(2111), 1,
      sym_scoped_identifier,
    STATE(2457), 1,
      sym_generic_type_with_turbofish,
    STATE(2527), 1,
      sym_bracketed_type,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3427), 3,
      sym_self,
      sym_super,
      sym_crate,
    ACTIONS(3425), 19,
      anon_sym_u8,
      anon_sym_i8,
      anon_sym_u16,
      anon_sym_i16,
      anon_sym_u32,
      anon_sym_i32,
      anon_sym_u64,
      anon_sym_i64,
      anon_sym_u128,
      anon_sym_i128,
      anon_sym_isize,
      anon_sym_usize,
      anon_sym_f32,
      anon_sym_f64,
      anon_sym_bool,
      anon_sym_str,
      anon_sym_char,
      anon_sym_default,
      anon_sym_union,
  [47993] = 3,
    ACTIONS(2367), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2369), 21,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_if,
      anon_sym_unsafe,
      anon_sym_COMMA,
      anon_sym_extern,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_in,
      anon_sym_DOT_DOT_EQ,
      anon_sym_PIPE,
  [48024] = 3,
    ACTIONS(2397), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2399), 21,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_if,
      anon_sym_unsafe,
      anon_sym_COMMA,
      anon_sym_extern,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_in,
      anon_sym_DOT_DOT_EQ,
      anon_sym_PIPE,
  [48055] = 11,
    ACTIONS(3433), 1,
      anon_sym_LPAREN,
    ACTIONS(3435), 1,
      anon_sym_LBRACE,
    ACTIONS(3439), 1,
      anon_sym_BANG,
    ACTIONS(3441), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(3447), 1,
      anon_sym_AT,
    STATE(1342), 1,
      sym_type_arguments,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3437), 2,
      anon_sym_COLON,
      anon_sym_EQ,
    ACTIONS(3443), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3431), 10,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [48101] = 9,
    ACTIONS(2554), 1,
      anon_sym_COLON,
    ACTIONS(3439), 1,
      anon_sym_BANG,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(3449), 1,
      anon_sym_LPAREN,
    ACTIONS(3451), 1,
      anon_sym_COLON_COLON,
    STATE(1342), 1,
      sym_type_arguments,
    STATE(1361), 1,
      sym_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2550), 13,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_PIPE,
  [48142] = 4,
    ACTIONS(2612), 1,
      anon_sym_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2616), 2,
      anon_sym_BANG,
      anon_sym_COLON_COLON,
    ACTIONS(2610), 16,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_for,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_LT2,
      anon_sym_PIPE,
  [48172] = 4,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2570), 2,
      anon_sym_LBRACE,
      anon_sym_LT2,
    ACTIONS(2574), 2,
      anon_sym_COLON,
      anon_sym_EQ,
    ACTIONS(2576), 15,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_if,
      anon_sym_BANG,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_COLON_COLON,
      anon_sym_DOT_DOT_DOT,
      anon_sym_in,
      anon_sym_DOT_DOT_EQ,
      anon_sym_PIPE,
  [48202] = 8,
    ACTIONS(2564), 1,
      anon_sym_COLON,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(3449), 1,
      anon_sym_LPAREN,
    ACTIONS(3451), 1,
      anon_sym_COLON_COLON,
    STATE(1342), 1,
      sym_type_arguments,
    STATE(1361), 1,
      sym_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2562), 13,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_PIPE,
  [48240] = 4,
    ACTIONS(2580), 1,
      anon_sym_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2584), 2,
      anon_sym_BANG,
      anon_sym_COLON_COLON,
    ACTIONS(2578), 16,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_for,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_LT2,
      anon_sym_PIPE,
  [48270] = 4,
    ACTIONS(2572), 1,
      anon_sym_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2576), 2,
      anon_sym_BANG,
      anon_sym_COLON_COLON,
    ACTIONS(2570), 16,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_for,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_LT2,
      anon_sym_PIPE,
  [48300] = 4,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2610), 2,
      anon_sym_LBRACE,
      anon_sym_LT2,
    ACTIONS(2614), 2,
      anon_sym_COLON,
      anon_sym_EQ,
    ACTIONS(2616), 15,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_if,
      anon_sym_BANG,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_COLON_COLON,
      anon_sym_DOT_DOT_DOT,
      anon_sym_in,
      anon_sym_DOT_DOT_EQ,
      anon_sym_PIPE,
  [48330] = 4,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2578), 2,
      anon_sym_LBRACE,
      anon_sym_LT2,
    ACTIONS(2582), 2,
      anon_sym_COLON,
      anon_sym_EQ,
    ACTIONS(2584), 15,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_if,
      anon_sym_BANG,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_COLON_COLON,
      anon_sym_DOT_DOT_DOT,
      anon_sym_in,
      anon_sym_DOT_DOT_EQ,
      anon_sym_PIPE,
  [48360] = 4,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2602), 2,
      anon_sym_LBRACE,
      anon_sym_LT2,
    ACTIONS(2606), 2,
      anon_sym_COLON,
      anon_sym_EQ,
    ACTIONS(2608), 15,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_if,
      anon_sym_BANG,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_COLON_COLON,
      anon_sym_DOT_DOT_DOT,
      anon_sym_in,
      anon_sym_DOT_DOT_EQ,
      anon_sym_PIPE,
  [48390] = 4,
    ACTIONS(2604), 1,
      anon_sym_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2608), 2,
      anon_sym_BANG,
      anon_sym_COLON_COLON,
    ACTIONS(2602), 16,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_for,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_LT2,
      anon_sym_PIPE,
  [48420] = 8,
    ACTIONS(2568), 1,
      anon_sym_COLON,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(3449), 1,
      anon_sym_LPAREN,
    ACTIONS(3451), 1,
      anon_sym_COLON_COLON,
    STATE(1342), 1,
      sym_type_arguments,
    STATE(1361), 1,
      sym_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2566), 13,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_PIPE,
  [48458] = 6,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(3449), 1,
      anon_sym_LPAREN,
    STATE(1347), 1,
      sym_type_arguments,
    STATE(1349), 1,
      sym_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2594), 14,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_PIPE,
  [48491] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2614), 2,
      anon_sym_COLON,
      anon_sym_EQ,
    ACTIONS(2616), 16,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_if,
      anon_sym_BANG,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_COLON_COLON,
      anon_sym_in,
      anon_sym_PIPE,
  [48518] = 6,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(3449), 1,
      anon_sym_LPAREN,
    STATE(1347), 1,
      sym_type_arguments,
    STATE(1349), 1,
      sym_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2586), 14,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_PIPE,
  [48551] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2574), 2,
      anon_sym_COLON,
      anon_sym_EQ,
    ACTIONS(2576), 16,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_if,
      anon_sym_BANG,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_COLON_COLON,
      anon_sym_in,
      anon_sym_PIPE,
  [48578] = 6,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(3449), 1,
      anon_sym_LPAREN,
    STATE(1347), 1,
      sym_type_arguments,
    STATE(1349), 1,
      sym_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2598), 14,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_PIPE,
  [48611] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2606), 2,
      anon_sym_COLON,
      anon_sym_EQ,
    ACTIONS(2608), 16,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_LBRACK,
      anon_sym_RBRACK,
      anon_sym_as,
      anon_sym_if,
      anon_sym_BANG,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_COLON_COLON,
      anon_sym_in,
      anon_sym_PIPE,
  [48638] = 3,
    ACTIONS(598), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(596), 16,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_if,
      anon_sym_where,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [48664] = 17,
    ACTIONS(3455), 1,
      anon_sym_const,
    ACTIONS(3457), 1,
      anon_sym_enum,
    ACTIONS(3459), 1,
      anon_sym_fn,
    ACTIONS(3461), 1,
      anon_sym_mod,
    ACTIONS(3463), 1,
      anon_sym_static,
    ACTIONS(3465), 1,
      anon_sym_struct,
    ACTIONS(3467), 1,
      anon_sym_trait,
    ACTIONS(3469), 1,
      anon_sym_type,
    ACTIONS(3471), 1,
      anon_sym_union,
    ACTIONS(3473), 1,
      anon_sym_unsafe,
    ACTIONS(3475), 1,
      anon_sym_use,
    ACTIONS(3477), 1,
      anon_sym_extern,
    STATE(1511), 1,
      sym_extern_modifier,
    STATE(1526), 1,
      aux_sym_function_modifiers_repeat1,
    STATE(2558), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3453), 2,
      anon_sym_async,
      anon_sym_default,
  [48718] = 3,
    ACTIONS(616), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(614), 16,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_if,
      anon_sym_where,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [48744] = 3,
    ACTIONS(602), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(600), 16,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_if,
      anon_sym_where,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [48770] = 7,
    ACTIONS(3433), 1,
      anon_sym_LPAREN,
    ACTIONS(3439), 1,
      anon_sym_BANG,
    ACTIONS(3479), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3437), 2,
      anon_sym_COLON,
      anon_sym_EQ,
    ACTIONS(3443), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3431), 10,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [48804] = 3,
    ACTIONS(440), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(438), 16,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_if,
      anon_sym_where,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [48830] = 17,
    ACTIONS(3481), 1,
      anon_sym_const,
    ACTIONS(3483), 1,
      anon_sym_enum,
    ACTIONS(3485), 1,
      anon_sym_fn,
    ACTIONS(3487), 1,
      anon_sym_mod,
    ACTIONS(3489), 1,
      anon_sym_static,
    ACTIONS(3491), 1,
      anon_sym_struct,
    ACTIONS(3493), 1,
      anon_sym_trait,
    ACTIONS(3495), 1,
      anon_sym_type,
    ACTIONS(3497), 1,
      anon_sym_union,
    ACTIONS(3499), 1,
      anon_sym_unsafe,
    ACTIONS(3501), 1,
      anon_sym_use,
    ACTIONS(3503), 1,
      anon_sym_extern,
    STATE(1517), 1,
      sym_extern_modifier,
    STATE(1526), 1,
      aux_sym_function_modifiers_repeat1,
    STATE(2436), 1,
      sym_function_modifiers,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3453), 2,
      anon_sym_async,
      anon_sym_default,
  [48884] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2578), 17,
      anon_sym_SEMI,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_for,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_LT2,
      anon_sym_PIPE,
  [48908] = 3,
    ACTIONS(3505), 1,
      anon_sym_LPAREN,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3159), 15,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_mod,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
  [48933] = 3,
    ACTIONS(2716), 1,
      anon_sym_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2714), 15,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_for,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_COLON_COLON,
      anon_sym_PIPE,
  [48958] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2892), 16,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      sym_mutable_specifier,
      anon_sym_PIPE,
      sym_self,
  [48981] = 3,
    ACTIONS(2748), 1,
      anon_sym_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2746), 15,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_for,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_COLON_COLON,
      anon_sym_PIPE,
  [49006] = 3,
    ACTIONS(2674), 1,
      anon_sym_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2672), 15,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_for,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_COLON_COLON,
      anon_sym_PIPE,
  [49031] = 3,
    ACTIONS(2684), 1,
      anon_sym_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2682), 15,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_for,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_COLON_COLON,
      anon_sym_PIPE,
  [49056] = 3,
    ACTIONS(2726), 1,
      anon_sym_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2724), 15,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_for,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_COLON_COLON,
      anon_sym_PIPE,
  [49081] = 6,
    ACTIONS(3511), 1,
      anon_sym_BANG,
    ACTIONS(3513), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3509), 2,
      anon_sym_COLON,
      anon_sym_EQ,
    ACTIONS(3515), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3507), 10,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [49112] = 3,
    ACTIONS(3517), 1,
      anon_sym_DASH_GT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2740), 14,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_PIPE,
  [49136] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2692), 15,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_PIPE,
  [49158] = 3,
    ACTIONS(2904), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3161), 14,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_mod,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
  [49182] = 3,
    ACTIONS(3519), 1,
      anon_sym_DASH_GT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2658), 14,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_PIPE,
  [49206] = 3,
    ACTIONS(3521), 1,
      anon_sym_DASH_GT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2666), 14,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_PIPE,
  [49230] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2646), 15,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_PIPE,
  [49252] = 4,
    ACTIONS(2596), 1,
      anon_sym_COLON,
    ACTIONS(3523), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2594), 13,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_PIPE,
  [49278] = 3,
    ACTIONS(3525), 1,
      anon_sym_DASH_GT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2676), 14,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_PIPE,
  [49302] = 3,
    ACTIONS(3527), 1,
      anon_sym_DASH_GT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2734), 14,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_PIPE,
  [49326] = 3,
    ACTIONS(2427), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2429), 14,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_in,
      anon_sym_DOT_DOT_EQ,
      anon_sym_PIPE,
  [49350] = 4,
    ACTIONS(2600), 1,
      anon_sym_COLON,
    ACTIONS(3523), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2598), 13,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_PIPE,
  [49376] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3165), 15,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_mod,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
  [49398] = 3,
    ACTIONS(3529), 1,
      anon_sym_DASH_GT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2754), 14,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_PIPE,
  [49422] = 4,
    ACTIONS(2588), 1,
      anon_sym_COLON,
    ACTIONS(3523), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2586), 13,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_PIPE,
  [49448] = 3,
    ACTIONS(3531), 1,
      anon_sym_DASH_GT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2718), 14,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_PIPE,
  [49472] = 5,
    ACTIONS(3513), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3509), 2,
      anon_sym_COLON,
      anon_sym_EQ,
    ACTIONS(3515), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3507), 10,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [49500] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2708), 15,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_PIPE,
  [49522] = 3,
    ACTIONS(3533), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3161), 14,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_mod,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
  [49546] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2750), 15,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_PIPE,
  [49568] = 4,
    ACTIONS(2600), 1,
      anon_sym_COLON,
    ACTIONS(3163), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2598), 13,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_PIPE,
  [49594] = 13,
    ACTIONS(3435), 1,
      anon_sym_LBRACE,
    ACTIONS(3439), 1,
      anon_sym_BANG,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(3447), 1,
      anon_sym_AT,
    ACTIONS(3535), 1,
      anon_sym_LPAREN,
    ACTIONS(3537), 1,
      anon_sym_RBRACK,
    ACTIONS(3540), 1,
      anon_sym_COLON_COLON,
    STATE(1342), 1,
      sym_type_arguments,
    STATE(1361), 1,
      sym_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2550), 2,
      anon_sym_SEMI,
      anon_sym_PLUS,
    ACTIONS(3431), 2,
      anon_sym_COMMA,
      anon_sym_PIPE,
    ACTIONS(3443), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
  [49638] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2642), 15,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_PIPE,
  [49660] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3175), 15,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_mod,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
  [49682] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3179), 15,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_enum,
      anon_sym_fn,
      anon_sym_mod,
      anon_sym_static,
      anon_sym_struct,
      anon_sym_trait,
      anon_sym_type,
      anon_sym_union,
      anon_sym_unsafe,
      anon_sym_use,
      anon_sym_extern,
      sym_identifier,
  [49704] = 14,
    ACTIONS(2550), 1,
      anon_sym_PLUS,
    ACTIONS(3431), 1,
      anon_sym_PIPE,
    ACTIONS(3435), 1,
      anon_sym_LBRACE,
    ACTIONS(3437), 1,
      anon_sym_COLON,
    ACTIONS(3439), 1,
      anon_sym_BANG,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(3447), 1,
      anon_sym_AT,
    ACTIONS(3542), 1,
      anon_sym_LPAREN,
    ACTIONS(3544), 1,
      anon_sym_COLON_COLON,
    STATE(1342), 1,
      sym_type_arguments,
    STATE(1361), 1,
      sym_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3443), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3537), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [49750] = 4,
    ACTIONS(2600), 1,
      anon_sym_COLON,
    ACTIONS(3546), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2598), 13,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_PIPE,
  [49776] = 13,
    ACTIONS(3431), 1,
      anon_sym_PIPE,
    ACTIONS(3435), 1,
      anon_sym_LBRACE,
    ACTIONS(3437), 1,
      anon_sym_COLON,
    ACTIONS(3439), 1,
      anon_sym_BANG,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(3447), 1,
      anon_sym_AT,
    ACTIONS(3548), 1,
      anon_sym_LPAREN,
    ACTIONS(3550), 1,
      anon_sym_COLON_COLON,
    STATE(1342), 1,
      sym_type_arguments,
    STATE(1361), 1,
      sym_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3443), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2550), 3,
      anon_sym_RPAREN,
      anon_sym_PLUS,
      anon_sym_COMMA,
  [49820] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2860), 14,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_PIPE,
  [49841] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2906), 14,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_PIPE,
  [49862] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2804), 14,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_PIPE,
  [49883] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2800), 14,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_PIPE,
  [49904] = 14,
    ACTIONS(3439), 1,
      anon_sym_BANG,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(3449), 1,
      anon_sym_LPAREN,
    ACTIONS(3451), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3552), 1,
      anon_sym_COLON,
    ACTIONS(3554), 1,
      anon_sym_EQ,
    ACTIONS(3556), 1,
      anon_sym_COMMA,
    ACTIONS(3558), 1,
      anon_sym_GT,
    STATE(1342), 1,
      sym_type_arguments,
    STATE(1361), 1,
      sym_parameters,
    STATE(1968), 1,
      sym_trait_bounds,
    STATE(1969), 1,
      aux_sym_type_parameters_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2550), 2,
      anon_sym_PLUS,
      anon_sym_as,
  [49949] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2832), 14,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_PIPE,
  [49970] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2832), 14,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_PIPE,
  [49991] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2792), 14,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_PIPE,
  [50012] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2594), 14,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_PIPE,
  [50033] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2848), 14,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_PIPE,
  [50054] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2856), 14,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_PIPE,
  [50075] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2788), 14,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_PIPE,
  [50096] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2896), 14,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_PIPE,
  [50117] = 4,
    ACTIONS(3562), 1,
      anon_sym_pat,
    STATE(579), 1,
      sym_fragment_specifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3560), 12,
      anon_sym_block,
      anon_sym_expr,
      anon_sym_ident,
      anon_sym_item,
      anon_sym_lifetime,
      anon_sym_literal,
      anon_sym_meta,
      anon_sym_path,
      anon_sym_stmt,
      anon_sym_tt,
      anon_sym_ty,
      anon_sym_vis,
  [50142] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2598), 14,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_PIPE,
  [50163] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2582), 2,
      anon_sym_COLON,
      anon_sym_EQ,
    ACTIONS(2584), 12,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_if,
      anon_sym_BANG,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_COLON_COLON,
      anon_sym_in,
      anon_sym_PIPE,
  [50186] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2836), 14,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_PIPE,
  [50207] = 3,
    ACTIONS(3566), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3564), 13,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_in,
      anon_sym_DOT_DOT_EQ,
      anon_sym_PIPE,
  [50230] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2962), 14,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_PIPE,
  [50251] = 7,
    ACTIONS(3509), 1,
      anon_sym_COLON,
    ACTIONS(3511), 1,
      anon_sym_BANG,
    ACTIONS(3568), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3515), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3507), 3,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
    ACTIONS(2624), 6,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_unsafe,
      anon_sym_extern,
  [50282] = 4,
    ACTIONS(3437), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3443), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3431), 11,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [50307] = 3,
    ACTIONS(3572), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3570), 13,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_DOT_DOT_DOT,
      anon_sym_in,
      anon_sym_DOT_DOT_EQ,
      anon_sym_PIPE,
  [50330] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2840), 14,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_PIPE,
  [50351] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2880), 14,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_PIPE,
  [50372] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2864), 14,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_PIPE,
  [50393] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2586), 14,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_PIPE,
  [50414] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2868), 14,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_PIPE,
  [50435] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2824), 14,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_PIPE,
  [50456] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2820), 14,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_PIPE,
  [50477] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2784), 14,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_PLUS,
      anon_sym_as,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_PIPE,
  [50498] = 4,
    ACTIONS(3578), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3576), 2,
      anon_sym_COLON,
      anon_sym_EQ,
    ACTIONS(3574), 10,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [50522] = 3,
    ACTIONS(416), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(414), 12,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [50544] = 6,
    ACTIONS(3511), 1,
      anon_sym_BANG,
    ACTIONS(3580), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3515), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3507), 3,
      anon_sym_RBRACK,
      anon_sym_COMMA,
      anon_sym_PIPE,
    ACTIONS(2624), 6,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_unsafe,
      anon_sym_extern,
  [50572] = 4,
    ACTIONS(3586), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3584), 2,
      anon_sym_COLON,
      anon_sym_EQ,
    ACTIONS(3582), 10,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [50596] = 4,
    ACTIONS(3592), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3590), 2,
      anon_sym_COLON,
      anon_sym_EQ,
    ACTIONS(3588), 10,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [50620] = 4,
    ACTIONS(3578), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3596), 2,
      anon_sym_COLON,
      anon_sym_EQ,
    ACTIONS(3594), 10,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [50644] = 4,
    ACTIONS(3533), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3584), 2,
      anon_sym_COLON,
      anon_sym_EQ,
    ACTIONS(3582), 10,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [50668] = 4,
    ACTIONS(3592), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3584), 2,
      anon_sym_COLON,
      anon_sym_EQ,
    ACTIONS(3582), 10,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [50692] = 3,
    ACTIONS(412), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(410), 12,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [50714] = 3,
    ACTIONS(408), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(406), 12,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [50736] = 4,
    ACTIONS(3586), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3590), 2,
      anon_sym_COLON,
      anon_sym_EQ,
    ACTIONS(3588), 10,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [50760] = 4,
    ACTIONS(3533), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3590), 2,
      anon_sym_COLON,
      anon_sym_EQ,
    ACTIONS(3588), 10,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [50784] = 3,
    ACTIONS(3600), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3598), 11,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [50805] = 4,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3602), 2,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
    ACTIONS(2610), 4,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_PLUS,
      anon_sym_LT2,
    ACTIONS(2616), 6,
      anon_sym_BANG,
      anon_sym_COMMA,
      anon_sym_COLON_COLON,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_PIPE,
  [50828] = 3,
    ACTIONS(3607), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3605), 11,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [50849] = 3,
    ACTIONS(3611), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3609), 11,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [50870] = 3,
    ACTIONS(3437), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3431), 11,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [50891] = 5,
    ACTIONS(2582), 1,
      anon_sym_COLON,
    ACTIONS(3613), 1,
      anon_sym_LPAREN,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2578), 5,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_LT2,
    ACTIONS(2584), 5,
      anon_sym_BANG,
      anon_sym_COLON_COLON,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_PIPE,
  [50916] = 4,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3616), 2,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
    ACTIONS(2602), 4,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_PLUS,
      anon_sym_LT2,
    ACTIONS(2608), 6,
      anon_sym_BANG,
      anon_sym_COMMA,
      anon_sym_COLON_COLON,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_PIPE,
  [50939] = 3,
    ACTIONS(3590), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3588), 11,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [50960] = 3,
    ACTIONS(3621), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3619), 11,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [50981] = 3,
    ACTIONS(3625), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3623), 11,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [51002] = 3,
    ACTIONS(3629), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3627), 11,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [51023] = 3,
    ACTIONS(3633), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3631), 11,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [51044] = 4,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3613), 2,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
    ACTIONS(2578), 4,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_PLUS,
      anon_sym_LT2,
    ACTIONS(2584), 6,
      anon_sym_BANG,
      anon_sym_COMMA,
      anon_sym_COLON_COLON,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_PIPE,
  [51067] = 5,
    ACTIONS(2614), 1,
      anon_sym_COLON,
    ACTIONS(3602), 1,
      anon_sym_LPAREN,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2610), 5,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_LT2,
    ACTIONS(2616), 5,
      anon_sym_BANG,
      anon_sym_COLON_COLON,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_PIPE,
  [51092] = 3,
    ACTIONS(3637), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3635), 11,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [51113] = 5,
    ACTIONS(2574), 1,
      anon_sym_COLON,
    ACTIONS(3639), 1,
      anon_sym_LPAREN,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2570), 5,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_LT2,
    ACTIONS(2576), 5,
      anon_sym_BANG,
      anon_sym_COLON_COLON,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_PIPE,
  [51138] = 3,
    ACTIONS(3644), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3642), 11,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [51159] = 4,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3639), 2,
      anon_sym_LPAREN,
      anon_sym_RBRACK,
    ACTIONS(2570), 4,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_PLUS,
      anon_sym_LT2,
    ACTIONS(2576), 6,
      anon_sym_BANG,
      anon_sym_COMMA,
      anon_sym_COLON_COLON,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_PIPE,
  [51182] = 3,
    ACTIONS(3648), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3646), 11,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [51203] = 3,
    ACTIONS(3652), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3650), 11,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [51224] = 3,
    ACTIONS(3656), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3654), 11,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [51245] = 7,
    ACTIONS(3507), 1,
      anon_sym_PIPE,
    ACTIONS(3509), 1,
      anon_sym_COLON,
    ACTIONS(3511), 1,
      anon_sym_BANG,
    ACTIONS(3658), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3515), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2624), 6,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_unsafe,
      anon_sym_extern,
  [51274] = 3,
    ACTIONS(3662), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3660), 11,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [51295] = 3,
    ACTIONS(3666), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3664), 11,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [51316] = 3,
    ACTIONS(3670), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3668), 11,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [51337] = 3,
    ACTIONS(3674), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3672), 11,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [51358] = 3,
    ACTIONS(3678), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3676), 11,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [51379] = 5,
    ACTIONS(2606), 1,
      anon_sym_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2602), 3,
      anon_sym_LBRACE,
      anon_sym_PLUS,
      anon_sym_LT2,
    ACTIONS(3616), 3,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
    ACTIONS(2608), 5,
      anon_sym_BANG,
      anon_sym_COLON_COLON,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_PIPE,
  [51404] = 3,
    ACTIONS(3682), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3680), 11,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [51425] = 3,
    ACTIONS(3584), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3582), 11,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [51446] = 3,
    ACTIONS(3686), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3684), 11,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [51467] = 5,
    ACTIONS(2606), 1,
      anon_sym_COLON,
    ACTIONS(3616), 1,
      anon_sym_LPAREN,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2602), 5,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_LT2,
    ACTIONS(2608), 5,
      anon_sym_BANG,
      anon_sym_COLON_COLON,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_PIPE,
  [51492] = 3,
    ACTIONS(3690), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3688), 11,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [51513] = 3,
    ACTIONS(3694), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3692), 11,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [51534] = 3,
    ACTIONS(3698), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3696), 11,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [51555] = 3,
    ACTIONS(3702), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3700), 11,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [51576] = 3,
    ACTIONS(3706), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3704), 11,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [51597] = 3,
    ACTIONS(452), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(450), 11,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [51618] = 3,
    ACTIONS(3710), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3708), 11,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [51639] = 5,
    ACTIONS(2574), 1,
      anon_sym_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2570), 3,
      anon_sym_LBRACE,
      anon_sym_PLUS,
      anon_sym_LT2,
    ACTIONS(3639), 3,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
    ACTIONS(2576), 5,
      anon_sym_BANG,
      anon_sym_COLON_COLON,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_PIPE,
  [51664] = 3,
    ACTIONS(3714), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3712), 11,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [51685] = 5,
    ACTIONS(2582), 1,
      anon_sym_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2578), 3,
      anon_sym_LBRACE,
      anon_sym_PLUS,
      anon_sym_LT2,
    ACTIONS(3613), 3,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
    ACTIONS(2584), 5,
      anon_sym_BANG,
      anon_sym_COLON_COLON,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_PIPE,
  [51710] = 3,
    ACTIONS(3718), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3716), 11,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
      anon_sym_EQ_GT,
      anon_sym_RBRACK,
      anon_sym_COLON,
      anon_sym_if,
      anon_sym_COMMA,
      anon_sym_else,
      anon_sym_in,
      anon_sym_PIPE,
  [51731] = 5,
    ACTIONS(2614), 1,
      anon_sym_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2610), 3,
      anon_sym_LBRACE,
      anon_sym_PLUS,
      anon_sym_LT2,
    ACTIONS(3602), 3,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_COMMA,
    ACTIONS(2616), 5,
      anon_sym_BANG,
      anon_sym_COLON_COLON,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
      anon_sym_PIPE,
  [51756] = 9,
    ACTIONS(3439), 1,
      anon_sym_BANG,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(3449), 1,
      anon_sym_LPAREN,
    ACTIONS(3451), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3720), 1,
      anon_sym_for,
    STATE(1342), 1,
      sym_type_arguments,
    STATE(1361), 1,
      sym_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2550), 4,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_PLUS,
      anon_sym_where,
  [51788] = 9,
    ACTIONS(3439), 1,
      anon_sym_BANG,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(3449), 1,
      anon_sym_LPAREN,
    ACTIONS(3451), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3722), 1,
      anon_sym_for,
    STATE(1342), 1,
      sym_type_arguments,
    STATE(1361), 1,
      sym_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2550), 4,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_PLUS,
      anon_sym_where,
  [51820] = 5,
    ACTIONS(706), 1,
      aux_sym_string_literal_token1,
    ACTIONS(3726), 1,
      sym_crate,
    STATE(1530), 1,
      sym_string_literal,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3724), 8,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_unsafe,
      anon_sym_extern,
  [51844] = 9,
    ACTIONS(3439), 1,
      anon_sym_BANG,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(3449), 1,
      anon_sym_LPAREN,
    ACTIONS(3451), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3728), 1,
      anon_sym_for,
    STATE(1342), 1,
      sym_type_arguments,
    STATE(1361), 1,
      sym_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2550), 4,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_PLUS,
      anon_sym_where,
  [51876] = 9,
    ACTIONS(3439), 1,
      anon_sym_BANG,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(3449), 1,
      anon_sym_LPAREN,
    ACTIONS(3451), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3730), 1,
      anon_sym_for,
    STATE(1342), 1,
      sym_type_arguments,
    STATE(1361), 1,
      sym_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2550), 4,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_PLUS,
      anon_sym_where,
  [51908] = 5,
    ACTIONS(706), 1,
      aux_sym_string_literal_token1,
    ACTIONS(3732), 1,
      sym_crate,
    STATE(1530), 1,
      sym_string_literal,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3724), 8,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_unsafe,
      anon_sym_extern,
  [51932] = 9,
    ACTIONS(3439), 1,
      anon_sym_BANG,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(3449), 1,
      anon_sym_LPAREN,
    ACTIONS(3451), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3734), 1,
      anon_sym_for,
    STATE(1342), 1,
      sym_type_arguments,
    STATE(1361), 1,
      sym_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2550), 4,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_PLUS,
      anon_sym_where,
  [51964] = 9,
    ACTIONS(3439), 1,
      anon_sym_BANG,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(3449), 1,
      anon_sym_LPAREN,
    ACTIONS(3451), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3736), 1,
      anon_sym_for,
    STATE(1342), 1,
      sym_type_arguments,
    STATE(1361), 1,
      sym_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2550), 4,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_PLUS,
      anon_sym_where,
  [51996] = 9,
    ACTIONS(3439), 1,
      anon_sym_BANG,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(3449), 1,
      anon_sym_LPAREN,
    ACTIONS(3451), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3738), 1,
      anon_sym_for,
    STATE(1342), 1,
      sym_type_arguments,
    STATE(1361), 1,
      sym_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2550), 4,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_PLUS,
      anon_sym_where,
  [52028] = 5,
    ACTIONS(706), 1,
      aux_sym_string_literal_token1,
    ACTIONS(3740), 1,
      sym_crate,
    STATE(1530), 1,
      sym_string_literal,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3724), 8,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_unsafe,
      anon_sym_extern,
  [52052] = 5,
    ACTIONS(706), 1,
      aux_sym_string_literal_token1,
    ACTIONS(3742), 1,
      sym_crate,
    STATE(1530), 1,
      sym_string_literal,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3724), 8,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_unsafe,
      anon_sym_extern,
  [52076] = 9,
    ACTIONS(3439), 1,
      anon_sym_BANG,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(3449), 1,
      anon_sym_LPAREN,
    ACTIONS(3451), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3744), 1,
      anon_sym_for,
    STATE(1342), 1,
      sym_type_arguments,
    STATE(1361), 1,
      sym_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2550), 4,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_PLUS,
      anon_sym_where,
  [52108] = 7,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(3449), 1,
      anon_sym_LPAREN,
    ACTIONS(3746), 1,
      anon_sym_LBRACE,
    STATE(1347), 1,
      sym_type_arguments,
    STATE(1349), 1,
      sym_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2598), 5,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_PLUS,
      anon_sym_COMMA,
  [52135] = 8,
    ACTIONS(2317), 1,
      anon_sym_POUND,
    ACTIONS(3748), 1,
      sym_identifier,
    ACTIONS(3750), 1,
      anon_sym_RBRACE,
    ACTIONS(3752), 1,
      anon_sym_COMMA,
    ACTIONS(3754), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1838), 2,
      sym_attribute_item,
      aux_sym_enum_variant_list_repeat1,
    STATE(1897), 3,
      sym_shorthand_field_initializer,
      sym_field_initializer,
      sym_base_field_initializer,
  [52164] = 6,
    ACTIONS(15), 1,
      anon_sym_LBRACE,
    ACTIONS(3493), 1,
      anon_sym_trait,
    ACTIONS(3756), 1,
      anon_sym_impl,
    STATE(61), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2624), 6,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_unsafe,
      anon_sym_extern,
  [52189] = 9,
    ACTIONS(3439), 1,
      anon_sym_BANG,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(3449), 1,
      anon_sym_LPAREN,
    ACTIONS(3451), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3758), 1,
      anon_sym_EQ,
    STATE(1361), 1,
      sym_parameters,
    STATE(1747), 1,
      sym_type_arguments,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2550), 3,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_GT,
  [52220] = 10,
    ACTIONS(53), 1,
      anon_sym_pub,
    ACTIONS(2317), 1,
      anon_sym_POUND,
    ACTIONS(3760), 1,
      sym_identifier,
    ACTIONS(3762), 1,
      anon_sym_RBRACE,
    ACTIONS(3764), 1,
      anon_sym_COMMA,
    ACTIONS(3766), 1,
      sym_crate,
    STATE(2106), 1,
      sym_field_declaration,
    STATE(2410), 1,
      sym_visibility_modifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1562), 2,
      sym_attribute_item,
      aux_sym_enum_variant_list_repeat1,
  [52253] = 10,
    ACTIONS(53), 1,
      anon_sym_pub,
    ACTIONS(2317), 1,
      anon_sym_POUND,
    ACTIONS(3760), 1,
      sym_identifier,
    ACTIONS(3766), 1,
      sym_crate,
    ACTIONS(3768), 1,
      anon_sym_RBRACE,
    ACTIONS(3770), 1,
      anon_sym_COMMA,
    STATE(2008), 1,
      sym_field_declaration,
    STATE(2410), 1,
      sym_visibility_modifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1527), 2,
      sym_attribute_item,
      aux_sym_enum_variant_list_repeat1,
  [52286] = 10,
    ACTIONS(53), 1,
      anon_sym_pub,
    ACTIONS(2317), 1,
      anon_sym_POUND,
    ACTIONS(3766), 1,
      sym_crate,
    ACTIONS(3772), 1,
      sym_identifier,
    ACTIONS(3774), 1,
      anon_sym_RBRACE,
    ACTIONS(3776), 1,
      anon_sym_COMMA,
    STATE(1880), 1,
      sym_enum_variant,
    STATE(2335), 1,
      sym_visibility_modifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1560), 2,
      sym_attribute_item,
      aux_sym_enum_variant_list_repeat1,
  [52319] = 10,
    ACTIONS(53), 1,
      anon_sym_pub,
    ACTIONS(2317), 1,
      anon_sym_POUND,
    ACTIONS(3766), 1,
      sym_crate,
    ACTIONS(3772), 1,
      sym_identifier,
    ACTIONS(3778), 1,
      anon_sym_RBRACE,
    ACTIONS(3780), 1,
      anon_sym_COMMA,
    STATE(2073), 1,
      sym_enum_variant,
    STATE(2335), 1,
      sym_visibility_modifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1559), 2,
      sym_attribute_item,
      aux_sym_enum_variant_list_repeat1,
  [52352] = 7,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(3449), 1,
      anon_sym_LPAREN,
    ACTIONS(3782), 1,
      anon_sym_for,
    STATE(1347), 1,
      sym_type_arguments,
    STATE(1349), 1,
      sym_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2598), 4,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_PLUS,
      anon_sym_where,
  [52378] = 7,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(3449), 1,
      anon_sym_LPAREN,
    ACTIONS(3784), 1,
      anon_sym_for,
    STATE(1347), 1,
      sym_type_arguments,
    STATE(1349), 1,
      sym_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2598), 4,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_PLUS,
      anon_sym_where,
  [52404] = 5,
    ACTIONS(3786), 1,
      anon_sym_SEMI,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    STATE(280), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2624), 6,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_unsafe,
      anon_sym_extern,
  [52426] = 9,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    ACTIONS(3790), 1,
      sym_identifier,
    ACTIONS(3792), 1,
      anon_sym_const,
    ACTIONS(3794), 1,
      anon_sym_GT,
    ACTIONS(3796), 1,
      sym_metavariable,
    STATE(1773), 1,
      sym_lifetime,
    STATE(2041), 1,
      sym_constrained_type_parameter,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(2164), 2,
      sym_const_parameter,
      sym_optional_type_parameter,
  [52456] = 10,
    ACTIONS(3798), 1,
      anon_sym_SEMI,
    ACTIONS(3800), 1,
      anon_sym_LPAREN,
    ACTIONS(3802), 1,
      anon_sym_LBRACE,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(3806), 1,
      anon_sym_LT,
    STATE(886), 1,
      sym_field_declaration_list,
    STATE(1571), 1,
      sym_type_parameters,
    STATE(2098), 1,
      sym_ordered_field_declaration_list,
    STATE(2130), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [52488] = 9,
    ACTIONS(53), 1,
      anon_sym_pub,
    ACTIONS(2317), 1,
      anon_sym_POUND,
    ACTIONS(3760), 1,
      sym_identifier,
    ACTIONS(3766), 1,
      sym_crate,
    ACTIONS(3808), 1,
      anon_sym_RBRACE,
    STATE(2274), 1,
      sym_field_declaration,
    STATE(2410), 1,
      sym_visibility_modifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1524), 2,
      sym_attribute_item,
      aux_sym_enum_variant_list_repeat1,
  [52518] = 9,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    ACTIONS(3790), 1,
      sym_identifier,
    ACTIONS(3792), 1,
      anon_sym_const,
    ACTIONS(3796), 1,
      sym_metavariable,
    ACTIONS(3810), 1,
      anon_sym_GT,
    STATE(1832), 1,
      sym_lifetime,
    STATE(2041), 1,
      sym_constrained_type_parameter,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(2164), 2,
      sym_const_parameter,
      sym_optional_type_parameter,
  [52548] = 9,
    ACTIONS(53), 1,
      anon_sym_pub,
    ACTIONS(2317), 1,
      anon_sym_POUND,
    ACTIONS(3760), 1,
      sym_identifier,
    ACTIONS(3766), 1,
      sym_crate,
    ACTIONS(3812), 1,
      anon_sym_RBRACE,
    STATE(2274), 1,
      sym_field_declaration,
    STATE(2410), 1,
      sym_visibility_modifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1524), 2,
      sym_attribute_item,
      aux_sym_enum_variant_list_repeat1,
  [52578] = 9,
    ACTIONS(53), 1,
      anon_sym_pub,
    ACTIONS(2317), 1,
      anon_sym_POUND,
    ACTIONS(3766), 1,
      sym_crate,
    ACTIONS(3772), 1,
      sym_identifier,
    ACTIONS(3814), 1,
      anon_sym_RBRACE,
    STATE(2248), 1,
      sym_enum_variant,
    STATE(2335), 1,
      sym_visibility_modifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1558), 2,
      sym_attribute_item,
      aux_sym_enum_variant_list_repeat1,
  [52608] = 7,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(3449), 1,
      anon_sym_LPAREN,
    ACTIONS(3816), 1,
      anon_sym_for,
    STATE(1347), 1,
      sym_type_arguments,
    STATE(1349), 1,
      sym_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2598), 4,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_PLUS,
      anon_sym_where,
  [52634] = 10,
    ACTIONS(3800), 1,
      anon_sym_LPAREN,
    ACTIONS(3802), 1,
      anon_sym_LBRACE,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(3806), 1,
      anon_sym_LT,
    ACTIONS(3818), 1,
      anon_sym_SEMI,
    STATE(980), 1,
      sym_field_declaration_list,
    STATE(1611), 1,
      sym_type_parameters,
    STATE(2036), 1,
      sym_ordered_field_declaration_list,
    STATE(2168), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [52666] = 7,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(3449), 1,
      anon_sym_LPAREN,
    ACTIONS(3820), 1,
      anon_sym_for,
    STATE(1347), 1,
      sym_type_arguments,
    STATE(1349), 1,
      sym_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2598), 4,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_PLUS,
      anon_sym_where,
  [52692] = 9,
    ACTIONS(53), 1,
      anon_sym_pub,
    ACTIONS(2317), 1,
      anon_sym_POUND,
    ACTIONS(3766), 1,
      sym_crate,
    ACTIONS(3772), 1,
      sym_identifier,
    ACTIONS(3822), 1,
      anon_sym_RBRACE,
    STATE(2248), 1,
      sym_enum_variant,
    STATE(2335), 1,
      sym_visibility_modifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1558), 2,
      sym_attribute_item,
      aux_sym_enum_variant_list_repeat1,
  [52722] = 9,
    ACTIONS(53), 1,
      anon_sym_pub,
    ACTIONS(2317), 1,
      anon_sym_POUND,
    ACTIONS(3760), 1,
      sym_identifier,
    ACTIONS(3766), 1,
      sym_crate,
    ACTIONS(3824), 1,
      anon_sym_RBRACE,
    STATE(2274), 1,
      sym_field_declaration,
    STATE(2410), 1,
      sym_visibility_modifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1524), 2,
      sym_attribute_item,
      aux_sym_enum_variant_list_repeat1,
  [52752] = 10,
    ACTIONS(3800), 1,
      anon_sym_LPAREN,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(3806), 1,
      anon_sym_LT,
    ACTIONS(3826), 1,
      anon_sym_SEMI,
    ACTIONS(3828), 1,
      anon_sym_LBRACE,
    STATE(335), 1,
      sym_field_declaration_list,
    STATE(1609), 1,
      sym_type_parameters,
    STATE(2101), 1,
      sym_ordered_field_declaration_list,
    STATE(2116), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [52784] = 7,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(3449), 1,
      anon_sym_LPAREN,
    ACTIONS(3830), 1,
      anon_sym_for,
    STATE(1347), 1,
      sym_type_arguments,
    STATE(1349), 1,
      sym_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2598), 4,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_PLUS,
      anon_sym_where,
  [52810] = 9,
    ACTIONS(53), 1,
      anon_sym_pub,
    ACTIONS(2317), 1,
      anon_sym_POUND,
    ACTIONS(3760), 1,
      sym_identifier,
    ACTIONS(3766), 1,
      sym_crate,
    ACTIONS(3832), 1,
      anon_sym_RBRACE,
    STATE(2274), 1,
      sym_field_declaration,
    STATE(2410), 1,
      sym_visibility_modifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1524), 2,
      sym_attribute_item,
      aux_sym_enum_variant_list_repeat1,
  [52840] = 7,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(3449), 1,
      anon_sym_LPAREN,
    ACTIONS(3834), 1,
      anon_sym_for,
    STATE(1347), 1,
      sym_type_arguments,
    STATE(1349), 1,
      sym_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2598), 4,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_PLUS,
      anon_sym_where,
  [52866] = 9,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    ACTIONS(3790), 1,
      sym_identifier,
    ACTIONS(3792), 1,
      anon_sym_const,
    ACTIONS(3796), 1,
      sym_metavariable,
    ACTIONS(3836), 1,
      anon_sym_GT,
    STATE(1773), 1,
      sym_lifetime,
    STATE(2041), 1,
      sym_constrained_type_parameter,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(2164), 2,
      sym_const_parameter,
      sym_optional_type_parameter,
  [52896] = 5,
    ACTIONS(3838), 1,
      anon_sym_SEMI,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    STATE(1032), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2624), 6,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_unsafe,
      anon_sym_extern,
  [52918] = 9,
    ACTIONS(53), 1,
      anon_sym_pub,
    ACTIONS(2317), 1,
      anon_sym_POUND,
    ACTIONS(3766), 1,
      sym_crate,
    ACTIONS(3772), 1,
      sym_identifier,
    ACTIONS(3842), 1,
      anon_sym_RBRACE,
    STATE(2248), 1,
      sym_enum_variant,
    STATE(2335), 1,
      sym_visibility_modifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1558), 2,
      sym_attribute_item,
      aux_sym_enum_variant_list_repeat1,
  [52948] = 9,
    ACTIONS(53), 1,
      anon_sym_pub,
    ACTIONS(2317), 1,
      anon_sym_POUND,
    ACTIONS(3766), 1,
      sym_crate,
    ACTIONS(3772), 1,
      sym_identifier,
    ACTIONS(3844), 1,
      anon_sym_RBRACE,
    STATE(2248), 1,
      sym_enum_variant,
    STATE(2335), 1,
      sym_visibility_modifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1558), 2,
      sym_attribute_item,
      aux_sym_enum_variant_list_repeat1,
  [52978] = 7,
    ACTIONS(2317), 1,
      anon_sym_POUND,
    ACTIONS(3748), 1,
      sym_identifier,
    ACTIONS(3754), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3846), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1838), 2,
      sym_attribute_item,
      aux_sym_enum_variant_list_repeat1,
    STATE(2209), 3,
      sym_shorthand_field_initializer,
      sym_field_initializer,
      sym_base_field_initializer,
  [53004] = 7,
    ACTIONS(3433), 1,
      anon_sym_LPAREN,
    ACTIONS(3437), 1,
      anon_sym_COLON,
    ACTIONS(3439), 1,
      anon_sym_BANG,
    ACTIONS(3848), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3443), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3431), 3,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
  [53030] = 7,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(3449), 1,
      anon_sym_LPAREN,
    ACTIONS(3850), 1,
      anon_sym_for,
    STATE(1347), 1,
      sym_type_arguments,
    STATE(1349), 1,
      sym_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2598), 4,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_PLUS,
      anon_sym_where,
  [53056] = 7,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(3449), 1,
      anon_sym_LPAREN,
    ACTIONS(3852), 1,
      anon_sym_for,
    STATE(1347), 1,
      sym_type_arguments,
    STATE(1349), 1,
      sym_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2598), 4,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_PLUS,
      anon_sym_where,
  [53082] = 9,
    ACTIONS(53), 1,
      anon_sym_pub,
    ACTIONS(2317), 1,
      anon_sym_POUND,
    ACTIONS(3760), 1,
      sym_identifier,
    ACTIONS(3766), 1,
      sym_crate,
    ACTIONS(3854), 1,
      anon_sym_RBRACE,
    STATE(2274), 1,
      sym_field_declaration,
    STATE(2410), 1,
      sym_visibility_modifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1524), 2,
      sym_attribute_item,
      aux_sym_enum_variant_list_repeat1,
  [53112] = 9,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    ACTIONS(3790), 1,
      sym_identifier,
    ACTIONS(3792), 1,
      anon_sym_const,
    ACTIONS(3796), 1,
      sym_metavariable,
    ACTIONS(3856), 1,
      anon_sym_GT,
    STATE(1773), 1,
      sym_lifetime,
    STATE(2041), 1,
      sym_constrained_type_parameter,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(2164), 2,
      sym_const_parameter,
      sym_optional_type_parameter,
  [53142] = 5,
    ACTIONS(15), 1,
      anon_sym_LBRACE,
    ACTIONS(3858), 1,
      sym_identifier,
    STATE(67), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3860), 6,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_unsafe,
      anon_sym_extern,
  [53164] = 5,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    ACTIONS(3862), 1,
      anon_sym_SEMI,
    STATE(970), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2624), 6,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_unsafe,
      anon_sym_extern,
  [53186] = 9,
    ACTIONS(53), 1,
      anon_sym_pub,
    ACTIONS(2317), 1,
      anon_sym_POUND,
    ACTIONS(3766), 1,
      sym_crate,
    ACTIONS(3772), 1,
      sym_identifier,
    ACTIONS(3864), 1,
      anon_sym_RBRACE,
    STATE(2248), 1,
      sym_enum_variant,
    STATE(2335), 1,
      sym_visibility_modifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1558), 2,
      sym_attribute_item,
      aux_sym_enum_variant_list_repeat1,
  [53216] = 9,
    ACTIONS(53), 1,
      anon_sym_pub,
    ACTIONS(2317), 1,
      anon_sym_POUND,
    ACTIONS(3766), 1,
      sym_crate,
    ACTIONS(3772), 1,
      sym_identifier,
    ACTIONS(3866), 1,
      anon_sym_RBRACE,
    STATE(2248), 1,
      sym_enum_variant,
    STATE(2335), 1,
      sym_visibility_modifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1558), 2,
      sym_attribute_item,
      aux_sym_enum_variant_list_repeat1,
  [53246] = 9,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    ACTIONS(3790), 1,
      sym_identifier,
    ACTIONS(3792), 1,
      anon_sym_const,
    ACTIONS(3796), 1,
      sym_metavariable,
    ACTIONS(3868), 1,
      anon_sym_GT,
    STATE(1773), 1,
      sym_lifetime,
    STATE(2041), 1,
      sym_constrained_type_parameter,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(2164), 2,
      sym_const_parameter,
      sym_optional_type_parameter,
  [53276] = 9,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    ACTIONS(3790), 1,
      sym_identifier,
    ACTIONS(3792), 1,
      anon_sym_const,
    ACTIONS(3796), 1,
      sym_metavariable,
    ACTIONS(3870), 1,
      anon_sym_GT,
    STATE(1773), 1,
      sym_lifetime,
    STATE(2041), 1,
      sym_constrained_type_parameter,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(2164), 2,
      sym_const_parameter,
      sym_optional_type_parameter,
  [53306] = 9,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    ACTIONS(3790), 1,
      sym_identifier,
    ACTIONS(3792), 1,
      anon_sym_const,
    ACTIONS(3796), 1,
      sym_metavariable,
    ACTIONS(3872), 1,
      anon_sym_GT,
    STATE(1773), 1,
      sym_lifetime,
    STATE(2041), 1,
      sym_constrained_type_parameter,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(2164), 2,
      sym_const_parameter,
      sym_optional_type_parameter,
  [53336] = 5,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    ACTIONS(3874), 1,
      anon_sym_SEMI,
    STATE(467), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2624), 6,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_unsafe,
      anon_sym_extern,
  [53358] = 9,
    ACTIONS(53), 1,
      anon_sym_pub,
    ACTIONS(2317), 1,
      anon_sym_POUND,
    ACTIONS(3760), 1,
      sym_identifier,
    ACTIONS(3766), 1,
      sym_crate,
    ACTIONS(3876), 1,
      anon_sym_RBRACE,
    STATE(2274), 1,
      sym_field_declaration,
    STATE(2410), 1,
      sym_visibility_modifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1524), 2,
      sym_attribute_item,
      aux_sym_enum_variant_list_repeat1,
  [53388] = 9,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    ACTIONS(3790), 1,
      sym_identifier,
    ACTIONS(3792), 1,
      anon_sym_const,
    ACTIONS(3796), 1,
      sym_metavariable,
    ACTIONS(3878), 1,
      anon_sym_GT,
    STATE(1773), 1,
      sym_lifetime,
    STATE(2041), 1,
      sym_constrained_type_parameter,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(2164), 2,
      sym_const_parameter,
      sym_optional_type_parameter,
  [53418] = 10,
    ACTIONS(3800), 1,
      anon_sym_LPAREN,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(3806), 1,
      anon_sym_LT,
    ACTIONS(3828), 1,
      anon_sym_LBRACE,
    ACTIONS(3880), 1,
      anon_sym_SEMI,
    STATE(377), 1,
      sym_field_declaration_list,
    STATE(1588), 1,
      sym_type_parameters,
    STATE(1987), 1,
      sym_ordered_field_declaration_list,
    STATE(2228), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [53450] = 9,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    ACTIONS(3790), 1,
      sym_identifier,
    ACTIONS(3792), 1,
      anon_sym_const,
    ACTIONS(3796), 1,
      sym_metavariable,
    ACTIONS(3882), 1,
      anon_sym_GT,
    STATE(1773), 1,
      sym_lifetime,
    STATE(2041), 1,
      sym_constrained_type_parameter,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(2164), 2,
      sym_const_parameter,
      sym_optional_type_parameter,
  [53480] = 7,
    ACTIONS(2317), 1,
      anon_sym_POUND,
    ACTIONS(3748), 1,
      sym_identifier,
    ACTIONS(3754), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3884), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1838), 2,
      sym_attribute_item,
      aux_sym_enum_variant_list_repeat1,
    STATE(2209), 3,
      sym_shorthand_field_initializer,
      sym_field_initializer,
      sym_base_field_initializer,
  [53506] = 5,
    ACTIONS(15), 1,
      anon_sym_LBRACE,
    ACTIONS(3886), 1,
      anon_sym_move,
    STATE(76), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2624), 6,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_unsafe,
      anon_sym_extern,
  [53528] = 8,
    ACTIONS(53), 1,
      anon_sym_pub,
    ACTIONS(2317), 1,
      anon_sym_POUND,
    ACTIONS(3760), 1,
      sym_identifier,
    ACTIONS(3766), 1,
      sym_crate,
    STATE(2304), 1,
      sym_field_declaration,
    STATE(2410), 1,
      sym_visibility_modifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1153), 2,
      sym_attribute_item,
      aux_sym_enum_variant_list_repeat1,
  [53555] = 5,
    ACTIONS(3891), 1,
      anon_sym_fn,
    ACTIONS(3893), 1,
      anon_sym_extern,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1525), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(3888), 4,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_unsafe,
  [53576] = 5,
    ACTIONS(3898), 1,
      anon_sym_fn,
    ACTIONS(3900), 1,
      anon_sym_extern,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1525), 2,
      sym_extern_modifier,
      aux_sym_function_modifiers_repeat1,
    ACTIONS(3896), 4,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_unsafe,
  [53597] = 8,
    ACTIONS(53), 1,
      anon_sym_pub,
    ACTIONS(2317), 1,
      anon_sym_POUND,
    ACTIONS(3760), 1,
      sym_identifier,
    ACTIONS(3766), 1,
      sym_crate,
    STATE(2012), 1,
      sym_field_declaration,
    STATE(2410), 1,
      sym_visibility_modifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1153), 2,
      sym_attribute_item,
      aux_sym_enum_variant_list_repeat1,
  [53624] = 6,
    ACTIONS(3433), 1,
      anon_sym_LPAREN,
    ACTIONS(3439), 1,
      anon_sym_BANG,
    ACTIONS(3902), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3443), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3431), 3,
      anon_sym_RBRACK,
      anon_sym_COMMA,
      anon_sym_PIPE,
  [53647] = 8,
    ACTIONS(798), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3904), 1,
      sym_identifier,
    ACTIONS(3906), 1,
      anon_sym_RBRACE,
    ACTIONS(3908), 1,
      anon_sym_COMMA,
    ACTIONS(3910), 1,
      anon_sym_ref,
    ACTIONS(3912), 1,
      sym_mutable_specifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1931), 2,
      sym_field_pattern,
      sym_remaining_field_pattern,
  [53674] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3914), 8,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_unsafe,
      anon_sym_extern,
  [53689] = 9,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(3916), 1,
      anon_sym_COLON,
    ACTIONS(3918), 1,
      anon_sym_LT,
    STATE(382), 1,
      sym_declaration_list,
    STATE(1697), 1,
      sym_type_parameters,
    STATE(1830), 1,
      sym_trait_bounds,
    STATE(2233), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [53718] = 9,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    ACTIONS(3916), 1,
      anon_sym_COLON,
    ACTIONS(3918), 1,
      anon_sym_LT,
    STATE(896), 1,
      sym_declaration_list,
    STATE(1661), 1,
      sym_type_parameters,
    STATE(1765), 1,
      sym_trait_bounds,
    STATE(2122), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [53747] = 9,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    ACTIONS(3916), 1,
      anon_sym_COLON,
    ACTIONS(3918), 1,
      anon_sym_LT,
    STATE(978), 1,
      sym_declaration_list,
    STATE(1618), 1,
      sym_type_parameters,
    STATE(1814), 1,
      sym_trait_bounds,
    STATE(2166), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [53776] = 9,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(3916), 1,
      anon_sym_COLON,
    ACTIONS(3918), 1,
      anon_sym_LT,
    STATE(402), 1,
      sym_declaration_list,
    STATE(1620), 1,
      sym_type_parameters,
    STATE(1815), 1,
      sym_trait_bounds,
    STATE(2225), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [53805] = 8,
    ACTIONS(53), 1,
      anon_sym_pub,
    ACTIONS(2317), 1,
      anon_sym_POUND,
    ACTIONS(3760), 1,
      sym_identifier,
    ACTIONS(3766), 1,
      sym_crate,
    STATE(2274), 1,
      sym_field_declaration,
    STATE(2410), 1,
      sym_visibility_modifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1524), 2,
      sym_attribute_item,
      aux_sym_enum_variant_list_repeat1,
  [53832] = 8,
    ACTIONS(798), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3904), 1,
      sym_identifier,
    ACTIONS(3910), 1,
      anon_sym_ref,
    ACTIONS(3912), 1,
      sym_mutable_specifier,
    ACTIONS(3920), 1,
      anon_sym_RBRACE,
    ACTIONS(3922), 1,
      anon_sym_COMMA,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1954), 2,
      sym_field_pattern,
      sym_remaining_field_pattern,
  [53859] = 4,
    ACTIONS(3926), 1,
      anon_sym_PLUS,
    STATE(1550), 1,
      aux_sym_trait_bounds_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3924), 6,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
  [53878] = 4,
    ACTIONS(2688), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3928), 1,
      anon_sym_BANG,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2624), 6,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_unsafe,
      anon_sym_extern,
  [53897] = 4,
    ACTIONS(3932), 1,
      anon_sym_PLUS,
    STATE(1539), 1,
      aux_sym_trait_bounds_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3930), 6,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
  [53916] = 4,
    ACTIONS(3935), 1,
      anon_sym_PLUS,
    STATE(1550), 1,
      aux_sym_trait_bounds_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3924), 6,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
  [53935] = 4,
    ACTIONS(3937), 1,
      anon_sym_PLUS,
    STATE(1550), 1,
      aux_sym_trait_bounds_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3924), 6,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
  [53954] = 8,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    ACTIONS(3792), 1,
      anon_sym_const,
    ACTIONS(3939), 1,
      sym_identifier,
    ACTIONS(3941), 1,
      sym_metavariable,
    STATE(1711), 1,
      sym_lifetime,
    STATE(1758), 1,
      sym_constrained_type_parameter,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(2006), 2,
      sym_const_parameter,
      sym_optional_type_parameter,
  [53981] = 8,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    ACTIONS(3792), 1,
      anon_sym_const,
    ACTIONS(3943), 1,
      sym_identifier,
    ACTIONS(3945), 1,
      sym_metavariable,
    STATE(1715), 1,
      sym_lifetime,
    STATE(1822), 1,
      sym_constrained_type_parameter,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(2023), 2,
      sym_const_parameter,
      sym_optional_type_parameter,
  [54008] = 8,
    ACTIONS(3947), 1,
      anon_sym_LPAREN,
    ACTIONS(3952), 1,
      anon_sym_LBRACE,
    ACTIONS(3955), 1,
      anon_sym_LBRACK,
    STATE(1544), 1,
      aux_sym_macro_definition_repeat1,
    STATE(2315), 1,
      sym_token_tree_pattern,
    STATE(2522), 1,
      sym_macro_rule,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3950), 2,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
  [54035] = 6,
    ACTIONS(3580), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3958), 1,
      anon_sym_RBRACK,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2598), 2,
      anon_sym_SEMI,
      anon_sym_PLUS,
    ACTIONS(3507), 2,
      anon_sym_COMMA,
      anon_sym_PIPE,
    ACTIONS(3515), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
  [54058] = 7,
    ACTIONS(2598), 1,
      anon_sym_PLUS,
    ACTIONS(3507), 1,
      anon_sym_PIPE,
    ACTIONS(3509), 1,
      anon_sym_COLON,
    ACTIONS(3568), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3515), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3958), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [54083] = 4,
    ACTIONS(3578), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3961), 1,
      anon_sym_BANG,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2624), 6,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_unsafe,
      anon_sym_extern,
  [54102] = 6,
    ACTIONS(2317), 1,
      anon_sym_POUND,
    ACTIONS(3748), 1,
      sym_identifier,
    ACTIONS(3754), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1838), 2,
      sym_attribute_item,
      aux_sym_enum_variant_list_repeat1,
    STATE(2209), 3,
      sym_shorthand_field_initializer,
      sym_field_initializer,
      sym_base_field_initializer,
  [54125] = 4,
    ACTIONS(858), 1,
      anon_sym_LBRACE,
    STATE(1455), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2624), 6,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_unsafe,
      anon_sym_extern,
  [54144] = 4,
    ACTIONS(3926), 1,
      anon_sym_PLUS,
    STATE(1539), 1,
      aux_sym_trait_bounds_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3963), 6,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
  [54163] = 8,
    ACTIONS(53), 1,
      anon_sym_pub,
    ACTIONS(2317), 1,
      anon_sym_POUND,
    ACTIONS(3766), 1,
      sym_crate,
    ACTIONS(3772), 1,
      sym_identifier,
    STATE(2248), 1,
      sym_enum_variant,
    STATE(2335), 1,
      sym_visibility_modifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1558), 2,
      sym_attribute_item,
      aux_sym_enum_variant_list_repeat1,
  [54190] = 6,
    ACTIONS(3507), 1,
      anon_sym_PIPE,
    ACTIONS(3509), 1,
      anon_sym_COLON,
    ACTIONS(3658), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3515), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2598), 3,
      anon_sym_RPAREN,
      anon_sym_PLUS,
      anon_sym_COMMA,
  [54213] = 4,
    ACTIONS(3467), 1,
      anon_sym_trait,
    ACTIONS(3965), 1,
      anon_sym_impl,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2624), 6,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_unsafe,
      anon_sym_extern,
  [54232] = 8,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    ACTIONS(3792), 1,
      anon_sym_const,
    ACTIONS(3939), 1,
      sym_identifier,
    ACTIONS(3941), 1,
      sym_metavariable,
    STATE(1690), 1,
      sym_lifetime,
    STATE(1758), 1,
      sym_constrained_type_parameter,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(2006), 2,
      sym_const_parameter,
      sym_optional_type_parameter,
  [54259] = 9,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    ACTIONS(3916), 1,
      anon_sym_COLON,
    ACTIONS(3918), 1,
      anon_sym_LT,
    STATE(820), 1,
      sym_declaration_list,
    STATE(1644), 1,
      sym_type_parameters,
    STATE(1807), 1,
      sym_trait_bounds,
    STATE(2194), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [54288] = 8,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    ACTIONS(3790), 1,
      sym_identifier,
    ACTIONS(3792), 1,
      anon_sym_const,
    ACTIONS(3796), 1,
      sym_metavariable,
    STATE(1773), 1,
      sym_lifetime,
    STATE(2041), 1,
      sym_constrained_type_parameter,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(2164), 2,
      sym_const_parameter,
      sym_optional_type_parameter,
  [54315] = 6,
    ACTIONS(3509), 1,
      anon_sym_COLON,
    ACTIONS(3511), 1,
      anon_sym_BANG,
    ACTIONS(3568), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3515), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3507), 3,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
  [54338] = 8,
    ACTIONS(53), 1,
      anon_sym_pub,
    ACTIONS(2317), 1,
      anon_sym_POUND,
    ACTIONS(3766), 1,
      sym_crate,
    ACTIONS(3772), 1,
      sym_identifier,
    STATE(2286), 1,
      sym_enum_variant,
    STATE(2335), 1,
      sym_visibility_modifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1153), 2,
      sym_attribute_item,
      aux_sym_enum_variant_list_repeat1,
  [54365] = 8,
    ACTIONS(53), 1,
      anon_sym_pub,
    ACTIONS(2317), 1,
      anon_sym_POUND,
    ACTIONS(3766), 1,
      sym_crate,
    ACTIONS(3772), 1,
      sym_identifier,
    STATE(2082), 1,
      sym_enum_variant,
    STATE(2335), 1,
      sym_visibility_modifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1153), 2,
      sym_attribute_item,
      aux_sym_enum_variant_list_repeat1,
  [54392] = 8,
    ACTIONS(53), 1,
      anon_sym_pub,
    ACTIONS(2317), 1,
      anon_sym_POUND,
    ACTIONS(3766), 1,
      sym_crate,
    ACTIONS(3772), 1,
      sym_identifier,
    STATE(2075), 1,
      sym_enum_variant,
    STATE(2335), 1,
      sym_visibility_modifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1153), 2,
      sym_attribute_item,
      aux_sym_enum_variant_list_repeat1,
  [54419] = 4,
    ACTIONS(706), 1,
      aux_sym_string_literal_token1,
    STATE(1530), 1,
      sym_string_literal,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3724), 6,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_unsafe,
      anon_sym_extern,
  [54438] = 8,
    ACTIONS(53), 1,
      anon_sym_pub,
    ACTIONS(2317), 1,
      anon_sym_POUND,
    ACTIONS(3760), 1,
      sym_identifier,
    ACTIONS(3766), 1,
      sym_crate,
    STATE(2028), 1,
      sym_field_declaration,
    STATE(2410), 1,
      sym_visibility_modifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1153), 2,
      sym_attribute_item,
      aux_sym_enum_variant_list_repeat1,
  [54465] = 9,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(3916), 1,
      anon_sym_COLON,
    ACTIONS(3918), 1,
      anon_sym_LT,
    STATE(317), 1,
      sym_declaration_list,
    STATE(1680), 1,
      sym_type_parameters,
    STATE(1778), 1,
      sym_trait_bounds,
    STATE(2167), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [54494] = 4,
    ACTIONS(3511), 1,
      anon_sym_BANG,
    ACTIONS(3546), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2624), 6,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_unsafe,
      anon_sym_extern,
  [54513] = 7,
    ACTIONS(3431), 1,
      anon_sym_PIPE,
    ACTIONS(3433), 1,
      anon_sym_LPAREN,
    ACTIONS(3437), 1,
      anon_sym_COLON,
    ACTIONS(3439), 1,
      anon_sym_BANG,
    ACTIONS(3967), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3443), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
  [54537] = 6,
    ACTIONS(3800), 1,
      anon_sym_LPAREN,
    ACTIONS(3802), 1,
      anon_sym_LBRACE,
    ACTIONS(3971), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3969), 2,
      anon_sym_RBRACE,
      anon_sym_COMMA,
    STATE(2079), 2,
      sym_field_declaration_list,
      sym_ordered_field_declaration_list,
  [54559] = 3,
    ACTIONS(3973), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3860), 6,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_unsafe,
      anon_sym_extern,
  [54575] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3930), 7,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_PLUS,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
  [54589] = 3,
    ACTIONS(3975), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3860), 6,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_unsafe,
      anon_sym_extern,
  [54605] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3930), 7,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_PLUS,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
  [54619] = 8,
    ACTIONS(3800), 1,
      anon_sym_LPAREN,
    ACTIONS(3802), 1,
      anon_sym_LBRACE,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(3977), 1,
      anon_sym_SEMI,
    STATE(823), 1,
      sym_field_declaration_list,
    STATE(2015), 1,
      sym_ordered_field_declaration_list,
    STATE(2191), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [54645] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3979), 7,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_PLUS,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
  [54659] = 3,
    ACTIONS(3981), 1,
      anon_sym_trait,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2624), 6,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_unsafe,
      anon_sym_extern,
  [54675] = 7,
    ACTIONS(798), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3904), 1,
      sym_identifier,
    ACTIONS(3910), 1,
      anon_sym_ref,
    ACTIONS(3912), 1,
      sym_mutable_specifier,
    ACTIONS(3983), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(2263), 2,
      sym_field_pattern,
      sym_remaining_field_pattern,
  [54699] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3930), 7,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_PLUS,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
  [54713] = 8,
    ACTIONS(3985), 1,
      anon_sym_LPAREN,
    ACTIONS(3987), 1,
      anon_sym_RPAREN,
    ACTIONS(3989), 1,
      anon_sym_LBRACE,
    ACTIONS(3991), 1,
      anon_sym_LBRACK,
    STATE(1544), 1,
      aux_sym_macro_definition_repeat1,
    STATE(2121), 1,
      sym_macro_rule,
    STATE(2315), 1,
      sym_token_tree_pattern,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [54739] = 3,
    ACTIONS(3993), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3860), 6,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_unsafe,
      anon_sym_extern,
  [54755] = 8,
    ACTIONS(3578), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3995), 1,
      anon_sym_LPAREN,
    ACTIONS(3997), 1,
      anon_sym_LBRACE,
    ACTIONS(3999), 1,
      anon_sym_LBRACK,
    ACTIONS(4001), 1,
      anon_sym_RBRACK,
    ACTIONS(4003), 1,
      anon_sym_EQ,
    STATE(2431), 1,
      sym_delim_token_tree,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [54781] = 8,
    ACTIONS(3985), 1,
      anon_sym_LPAREN,
    ACTIONS(3989), 1,
      anon_sym_LBRACE,
    ACTIONS(3991), 1,
      anon_sym_LBRACK,
    ACTIONS(4005), 1,
      anon_sym_RBRACE,
    STATE(1544), 1,
      aux_sym_macro_definition_repeat1,
    STATE(2143), 1,
      sym_macro_rule,
    STATE(2315), 1,
      sym_token_tree_pattern,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [54807] = 8,
    ACTIONS(3985), 1,
      anon_sym_LPAREN,
    ACTIONS(3989), 1,
      anon_sym_LBRACE,
    ACTIONS(3991), 1,
      anon_sym_LBRACK,
    ACTIONS(4007), 1,
      anon_sym_RBRACE,
    STATE(1544), 1,
      aux_sym_macro_definition_repeat1,
    STATE(2141), 1,
      sym_macro_rule,
    STATE(2315), 1,
      sym_token_tree_pattern,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [54833] = 8,
    ACTIONS(3985), 1,
      anon_sym_LPAREN,
    ACTIONS(3989), 1,
      anon_sym_LBRACE,
    ACTIONS(3991), 1,
      anon_sym_LBRACK,
    ACTIONS(4009), 1,
      anon_sym_RBRACE,
    STATE(1544), 1,
      aux_sym_macro_definition_repeat1,
    STATE(2148), 1,
      sym_macro_rule,
    STATE(2315), 1,
      sym_token_tree_pattern,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [54859] = 8,
    ACTIONS(3985), 1,
      anon_sym_LPAREN,
    ACTIONS(3989), 1,
      anon_sym_LBRACE,
    ACTIONS(3991), 1,
      anon_sym_LBRACK,
    ACTIONS(4011), 1,
      anon_sym_RPAREN,
    STATE(1544), 1,
      aux_sym_macro_definition_repeat1,
    STATE(2147), 1,
      sym_macro_rule,
    STATE(2315), 1,
      sym_token_tree_pattern,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [54885] = 3,
    ACTIONS(4013), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3860), 6,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_unsafe,
      anon_sym_extern,
  [54901] = 8,
    ACTIONS(3985), 1,
      anon_sym_LPAREN,
    ACTIONS(3989), 1,
      anon_sym_LBRACE,
    ACTIONS(3991), 1,
      anon_sym_LBRACK,
    ACTIONS(4015), 1,
      anon_sym_RPAREN,
    STATE(1544), 1,
      aux_sym_macro_definition_repeat1,
    STATE(2120), 1,
      sym_macro_rule,
    STATE(2315), 1,
      sym_token_tree_pattern,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [54927] = 3,
    ACTIONS(4017), 1,
      anon_sym_trait,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2624), 6,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_unsafe,
      anon_sym_extern,
  [54943] = 3,
    ACTIONS(3546), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2624), 6,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_unsafe,
      anon_sym_extern,
  [54959] = 8,
    ACTIONS(3985), 1,
      anon_sym_LPAREN,
    ACTIONS(3989), 1,
      anon_sym_LBRACE,
    ACTIONS(3991), 1,
      anon_sym_LBRACK,
    ACTIONS(4019), 1,
      anon_sym_RPAREN,
    STATE(1576), 1,
      aux_sym_macro_definition_repeat1,
    STATE(2154), 1,
      sym_macro_rule,
    STATE(2315), 1,
      sym_token_tree_pattern,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [54985] = 8,
    ACTIONS(3800), 1,
      anon_sym_LPAREN,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(3828), 1,
      anon_sym_LBRACE,
    ACTIONS(4021), 1,
      anon_sym_SEMI,
    STATE(272), 1,
      sym_field_declaration_list,
    STATE(2010), 1,
      sym_ordered_field_declaration_list,
    STATE(2196), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [55011] = 7,
    ACTIONS(798), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3904), 1,
      sym_identifier,
    ACTIONS(3910), 1,
      anon_sym_ref,
    ACTIONS(3912), 1,
      sym_mutable_specifier,
    ACTIONS(4023), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(2263), 2,
      sym_field_pattern,
      sym_remaining_field_pattern,
  [55035] = 8,
    ACTIONS(3985), 1,
      anon_sym_LPAREN,
    ACTIONS(3989), 1,
      anon_sym_LBRACE,
    ACTIONS(3991), 1,
      anon_sym_LBRACK,
    ACTIONS(4025), 1,
      anon_sym_RBRACE,
    STATE(1579), 1,
      aux_sym_macro_definition_repeat1,
    STATE(2151), 1,
      sym_macro_rule,
    STATE(2315), 1,
      sym_token_tree_pattern,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [55061] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3930), 7,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_PLUS,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
  [55075] = 8,
    ACTIONS(3985), 1,
      anon_sym_LPAREN,
    ACTIONS(3989), 1,
      anon_sym_LBRACE,
    ACTIONS(3991), 1,
      anon_sym_LBRACK,
    ACTIONS(4027), 1,
      anon_sym_RBRACE,
    STATE(1544), 1,
      aux_sym_macro_definition_repeat1,
    STATE(2145), 1,
      sym_macro_rule,
    STATE(2315), 1,
      sym_token_tree_pattern,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [55101] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3930), 7,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_PLUS,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
  [55115] = 8,
    ACTIONS(3985), 1,
      anon_sym_LPAREN,
    ACTIONS(3989), 1,
      anon_sym_LBRACE,
    ACTIONS(3991), 1,
      anon_sym_LBRACK,
    ACTIONS(4029), 1,
      anon_sym_RBRACE,
    STATE(1580), 1,
      aux_sym_macro_definition_repeat1,
    STATE(2156), 1,
      sym_macro_rule,
    STATE(2315), 1,
      sym_token_tree_pattern,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [55141] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4031), 7,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_PLUS,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
  [55155] = 5,
    ACTIONS(3511), 1,
      anon_sym_BANG,
    ACTIONS(3580), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3515), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(3507), 3,
      anon_sym_RBRACK,
      anon_sym_COMMA,
      anon_sym_PIPE,
  [55175] = 8,
    ACTIONS(3985), 1,
      anon_sym_LPAREN,
    ACTIONS(3989), 1,
      anon_sym_LBRACE,
    ACTIONS(3991), 1,
      anon_sym_LBRACK,
    ACTIONS(4033), 1,
      anon_sym_RPAREN,
    STATE(1601), 1,
      aux_sym_macro_definition_repeat1,
    STATE(2311), 1,
      sym_macro_rule,
    STATE(2315), 1,
      sym_token_tree_pattern,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [55201] = 8,
    ACTIONS(3985), 1,
      anon_sym_LPAREN,
    ACTIONS(3989), 1,
      anon_sym_LBRACE,
    ACTIONS(3991), 1,
      anon_sym_LBRACK,
    ACTIONS(4035), 1,
      anon_sym_RBRACE,
    STATE(1592), 1,
      aux_sym_macro_definition_repeat1,
    STATE(2310), 1,
      sym_macro_rule,
    STATE(2315), 1,
      sym_token_tree_pattern,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [55227] = 8,
    ACTIONS(3995), 1,
      anon_sym_LPAREN,
    ACTIONS(3997), 1,
      anon_sym_LBRACE,
    ACTIONS(3999), 1,
      anon_sym_LBRACK,
    ACTIONS(4037), 1,
      anon_sym_RBRACK,
    ACTIONS(4039), 1,
      anon_sym_EQ,
    ACTIONS(4041), 1,
      anon_sym_COLON_COLON,
    STATE(2432), 1,
      sym_delim_token_tree,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [55253] = 3,
    ACTIONS(2688), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2624), 6,
      anon_sym_async,
      anon_sym_const,
      anon_sym_default,
      anon_sym_fn,
      anon_sym_unsafe,
      anon_sym_extern,
  [55269] = 8,
    ACTIONS(3985), 1,
      anon_sym_LPAREN,
    ACTIONS(3989), 1,
      anon_sym_LBRACE,
    ACTIONS(3991), 1,
      anon_sym_LBRACK,
    ACTIONS(4043), 1,
      anon_sym_RPAREN,
    STATE(1544), 1,
      aux_sym_macro_definition_repeat1,
    STATE(2144), 1,
      sym_macro_rule,
    STATE(2315), 1,
      sym_token_tree_pattern,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [55295] = 8,
    ACTIONS(3985), 1,
      anon_sym_LPAREN,
    ACTIONS(3989), 1,
      anon_sym_LBRACE,
    ACTIONS(3991), 1,
      anon_sym_LBRACK,
    ACTIONS(4045), 1,
      anon_sym_RPAREN,
    STATE(1584), 1,
      aux_sym_macro_definition_repeat1,
    STATE(2169), 1,
      sym_macro_rule,
    STATE(2315), 1,
      sym_token_tree_pattern,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [55321] = 8,
    ACTIONS(3995), 1,
      anon_sym_LPAREN,
    ACTIONS(3997), 1,
      anon_sym_LBRACE,
    ACTIONS(3999), 1,
      anon_sym_LBRACK,
    ACTIONS(4037), 1,
      anon_sym_RBRACK,
    ACTIONS(4039), 1,
      anon_sym_EQ,
    ACTIONS(4047), 1,
      anon_sym_COLON_COLON,
    STATE(2432), 1,
      sym_delim_token_tree,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [55347] = 7,
    ACTIONS(798), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3904), 1,
      sym_identifier,
    ACTIONS(3910), 1,
      anon_sym_ref,
    ACTIONS(3912), 1,
      sym_mutable_specifier,
    ACTIONS(4049), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(2263), 2,
      sym_field_pattern,
      sym_remaining_field_pattern,
  [55371] = 8,
    ACTIONS(3985), 1,
      anon_sym_LPAREN,
    ACTIONS(3989), 1,
      anon_sym_LBRACE,
    ACTIONS(3991), 1,
      anon_sym_LBRACK,
    ACTIONS(4051), 1,
      anon_sym_RPAREN,
    STATE(1582), 1,
      aux_sym_macro_definition_repeat1,
    STATE(2309), 1,
      sym_macro_rule,
    STATE(2315), 1,
      sym_token_tree_pattern,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [55397] = 6,
    ACTIONS(3800), 1,
      anon_sym_LPAREN,
    ACTIONS(3802), 1,
      anon_sym_LBRACE,
    ACTIONS(4055), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4053), 2,
      anon_sym_RBRACE,
      anon_sym_COMMA,
    STATE(1959), 2,
      sym_field_declaration_list,
      sym_ordered_field_declaration_list,
  [55419] = 7,
    ACTIONS(798), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3904), 1,
      sym_identifier,
    ACTIONS(3910), 1,
      anon_sym_ref,
    ACTIONS(3912), 1,
      sym_mutable_specifier,
    ACTIONS(4057), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(2263), 2,
      sym_field_pattern,
      sym_remaining_field_pattern,
  [55443] = 8,
    ACTIONS(3995), 1,
      anon_sym_LPAREN,
    ACTIONS(3997), 1,
      anon_sym_LBRACE,
    ACTIONS(3999), 1,
      anon_sym_LBRACK,
    ACTIONS(4037), 1,
      anon_sym_RBRACK,
    ACTIONS(4039), 1,
      anon_sym_EQ,
    ACTIONS(4059), 1,
      anon_sym_COLON_COLON,
    STATE(2432), 1,
      sym_delim_token_tree,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [55469] = 8,
    ACTIONS(3800), 1,
      anon_sym_LPAREN,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(3828), 1,
      anon_sym_LBRACE,
    ACTIONS(4061), 1,
      anon_sym_SEMI,
    STATE(405), 1,
      sym_field_declaration_list,
    STATE(1925), 1,
      sym_ordered_field_declaration_list,
    STATE(2221), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [55495] = 8,
    ACTIONS(3985), 1,
      anon_sym_LPAREN,
    ACTIONS(3989), 1,
      anon_sym_LBRACE,
    ACTIONS(3991), 1,
      anon_sym_LBRACK,
    ACTIONS(4063), 1,
      anon_sym_RBRACE,
    STATE(1581), 1,
      aux_sym_macro_definition_repeat1,
    STATE(2308), 1,
      sym_macro_rule,
    STATE(2315), 1,
      sym_token_tree_pattern,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [55521] = 8,
    ACTIONS(3800), 1,
      anon_sym_LPAREN,
    ACTIONS(3802), 1,
      anon_sym_LBRACE,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(4065), 1,
      anon_sym_SEMI,
    STATE(909), 1,
      sym_field_declaration_list,
    STATE(2105), 1,
      sym_ordered_field_declaration_list,
    STATE(2113), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [55547] = 7,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(4067), 1,
      anon_sym_SEMI,
    ACTIONS(4069), 1,
      anon_sym_LBRACE,
    ACTIONS(4071), 1,
      anon_sym_DASH_GT,
    STATE(1041), 1,
      sym_block,
    STATE(1973), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [55570] = 7,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    ACTIONS(4073), 1,
      anon_sym_SEMI,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    STATE(863), 1,
      sym_declaration_list,
    STATE(2043), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [55593] = 7,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4077), 1,
      anon_sym_SEMI,
    STATE(346), 1,
      sym_declaration_list,
    STATE(1927), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [55616] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3623), 2,
      anon_sym_COLON,
      anon_sym_PIPE,
    ACTIONS(2646), 4,
      anon_sym_RPAREN,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
  [55631] = 7,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(4079), 1,
      anon_sym_SEMI,
    ACTIONS(4081), 1,
      anon_sym_LBRACE,
    ACTIONS(4083), 1,
      anon_sym_DASH_GT,
    STATE(397), 1,
      sym_block,
    STATE(1873), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [55654] = 7,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4085), 1,
      anon_sym_SEMI,
    STATE(334), 1,
      sym_declaration_list,
    STATE(1878), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [55677] = 7,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    ACTIONS(3916), 1,
      anon_sym_COLON,
    STATE(904), 1,
      sym_declaration_list,
    STATE(1771), 1,
      sym_trait_bounds,
    STATE(2115), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [55700] = 7,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(4069), 1,
      anon_sym_LBRACE,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4087), 1,
      anon_sym_SEMI,
    STATE(917), 1,
      sym_block,
    STATE(1991), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [55723] = 7,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(3916), 1,
      anon_sym_COLON,
    STATE(264), 1,
      sym_declaration_list,
    STATE(1772), 1,
      sym_trait_bounds,
    STATE(2283), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [55746] = 7,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4089), 1,
      anon_sym_SEMI,
    STATE(351), 1,
      sym_declaration_list,
    STATE(1877), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [55769] = 7,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(4069), 1,
      anon_sym_LBRACE,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4091), 1,
      anon_sym_SEMI,
    STATE(894), 1,
      sym_block,
    STATE(1924), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [55792] = 7,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4093), 1,
      anon_sym_SEMI,
    STATE(897), 1,
      sym_declaration_list,
    STATE(2100), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [55815] = 7,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4081), 1,
      anon_sym_LBRACE,
    ACTIONS(4095), 1,
      anon_sym_SEMI,
    STATE(460), 1,
      sym_block,
    STATE(1858), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [55838] = 7,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4097), 1,
      anon_sym_SEMI,
    STATE(465), 1,
      sym_declaration_list,
    STATE(2035), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [55861] = 7,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(3806), 1,
      anon_sym_LT,
    ACTIONS(3828), 1,
      anon_sym_LBRACE,
    STATE(390), 1,
      sym_field_declaration_list,
    STATE(1816), 1,
      sym_type_parameters,
    STATE(2242), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [55884] = 7,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(4081), 1,
      anon_sym_LBRACE,
    ACTIONS(4099), 1,
      anon_sym_SEMI,
    ACTIONS(4101), 1,
      anon_sym_DASH_GT,
    STATE(255), 1,
      sym_block,
    STATE(1900), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [55907] = 4,
    ACTIONS(4103), 1,
      anon_sym_RBRACK,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3704), 2,
      anon_sym_COMMA,
      anon_sym_PIPE,
    ACTIONS(2750), 3,
      anon_sym_SEMI,
      anon_sym_PLUS,
      anon_sym_DASH_GT,
  [55924] = 7,
    ACTIONS(3249), 1,
      anon_sym_LBRACE,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(4106), 1,
      sym_identifier,
    ACTIONS(4108), 1,
      anon_sym_STAR,
    STATE(2060), 1,
      sym_use_list,
    STATE(2372), 1,
      sym_type_arguments,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [55947] = 7,
    ACTIONS(3249), 1,
      anon_sym_LBRACE,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(4106), 1,
      sym_identifier,
    ACTIONS(4108), 1,
      anon_sym_STAR,
    STATE(2060), 1,
      sym_use_list,
    STATE(2374), 1,
      sym_type_arguments,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [55970] = 7,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4081), 1,
      anon_sym_LBRACE,
    ACTIONS(4110), 1,
      anon_sym_SEMI,
    STATE(263), 1,
      sym_block,
    STATE(1891), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [55993] = 7,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(3806), 1,
      anon_sym_LT,
    ACTIONS(4112), 1,
      anon_sym_LBRACE,
    STATE(890), 1,
      sym_enum_variant_list,
    STATE(1768), 1,
      sym_type_parameters,
    STATE(2127), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [56016] = 7,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4114), 1,
      anon_sym_SEMI,
    STATE(344), 1,
      sym_declaration_list,
    STATE(1926), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [56039] = 6,
    ACTIONS(3163), 1,
      anon_sym_COLON_COLON,
    ACTIONS(4116), 1,
      anon_sym_COMMA,
    ACTIONS(4118), 1,
      anon_sym_GT,
    STATE(1855), 1,
      aux_sym_type_parameters_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2598), 2,
      anon_sym_PLUS,
      anon_sym_as,
  [56060] = 7,
    ACTIONS(3916), 1,
      anon_sym_COLON,
    ACTIONS(3918), 1,
      anon_sym_LT,
    ACTIONS(4120), 1,
      anon_sym_SEMI,
    ACTIONS(4122), 1,
      anon_sym_EQ,
    STATE(1850), 1,
      sym_type_parameters,
    STATE(2318), 1,
      sym_trait_bounds,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [56083] = 7,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(4069), 1,
      anon_sym_LBRACE,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4124), 1,
      anon_sym_SEMI,
    STATE(991), 1,
      sym_block,
    STATE(1944), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [56106] = 4,
    ACTIONS(3523), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3830), 1,
      anon_sym_for,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2598), 4,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_PLUS,
      anon_sym_where,
  [56123] = 7,
    ACTIONS(3802), 1,
      anon_sym_LBRACE,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(3806), 1,
      anon_sym_LT,
    STATE(885), 1,
      sym_field_declaration_list,
    STATE(1761), 1,
      sym_type_parameters,
    STATE(2132), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [56146] = 4,
    ACTIONS(3523), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3834), 1,
      anon_sym_for,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2598), 4,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_PLUS,
      anon_sym_where,
  [56163] = 7,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4126), 1,
      anon_sym_SEMI,
    STATE(933), 1,
      sym_declaration_list,
    STATE(2097), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [56186] = 7,
    ACTIONS(3916), 1,
      anon_sym_COLON,
    ACTIONS(3918), 1,
      anon_sym_LT,
    ACTIONS(4128), 1,
      anon_sym_SEMI,
    ACTIONS(4130), 1,
      anon_sym_EQ,
    STATE(1813), 1,
      sym_type_parameters,
    STATE(2479), 1,
      sym_trait_bounds,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [56209] = 7,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4132), 1,
      anon_sym_SEMI,
    STATE(250), 1,
      sym_declaration_list,
    STATE(2020), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [56232] = 4,
    ACTIONS(3523), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3852), 1,
      anon_sym_for,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2598), 4,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_PLUS,
      anon_sym_where,
  [56249] = 7,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    ACTIONS(3916), 1,
      anon_sym_COLON,
    STATE(1036), 1,
      sym_declaration_list,
    STATE(1841), 1,
      sym_trait_bounds,
    STATE(2232), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [56272] = 7,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(4069), 1,
      anon_sym_LBRACE,
    ACTIONS(4134), 1,
      anon_sym_SEMI,
    ACTIONS(4136), 1,
      anon_sym_DASH_GT,
    STATE(1008), 1,
      sym_block,
    STATE(1978), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [56295] = 7,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(4069), 1,
      anon_sym_LBRACE,
    ACTIONS(4138), 1,
      anon_sym_SEMI,
    ACTIONS(4140), 1,
      anon_sym_DASH_GT,
    STATE(829), 1,
      sym_block,
    STATE(2019), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [56318] = 7,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4081), 1,
      anon_sym_LBRACE,
    ACTIONS(4142), 1,
      anon_sym_SEMI,
    STATE(422), 1,
      sym_block,
    STATE(1865), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [56341] = 7,
    ACTIONS(3802), 1,
      anon_sym_LBRACE,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(3806), 1,
      anon_sym_LT,
    STATE(976), 1,
      sym_field_declaration_list,
    STATE(1812), 1,
      sym_type_parameters,
    STATE(2163), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [56364] = 7,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(4081), 1,
      anon_sym_LBRACE,
    ACTIONS(4144), 1,
      anon_sym_SEMI,
    ACTIONS(4146), 1,
      anon_sym_DASH_GT,
    STATE(414), 1,
      sym_block,
    STATE(1993), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [56387] = 4,
    ACTIONS(3523), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3850), 1,
      anon_sym_for,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2598), 4,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_PLUS,
      anon_sym_where,
  [56404] = 4,
    ACTIONS(3523), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3820), 1,
      anon_sym_for,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2598), 4,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_PLUS,
      anon_sym_where,
  [56421] = 7,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4148), 1,
      anon_sym_SEMI,
    STATE(314), 1,
      sym_declaration_list,
    STATE(2034), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [56444] = 6,
    ACTIONS(3507), 1,
      anon_sym_PIPE,
    ACTIONS(3509), 1,
      anon_sym_COLON,
    ACTIONS(3511), 1,
      anon_sym_BANG,
    ACTIONS(3658), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3515), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
  [56465] = 7,
    ACTIONS(2449), 1,
      anon_sym_PLUS,
    ACTIONS(3916), 1,
      anon_sym_COLON,
    ACTIONS(4116), 1,
      anon_sym_COMMA,
    ACTIONS(4118), 1,
      anon_sym_GT,
    STATE(1855), 1,
      aux_sym_type_parameters_repeat1,
    STATE(1970), 1,
      sym_trait_bounds,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [56488] = 7,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4081), 1,
      anon_sym_LBRACE,
    ACTIONS(4150), 1,
      anon_sym_SEMI,
    STATE(360), 1,
      sym_block,
    STATE(1941), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [56511] = 7,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4152), 1,
      anon_sym_SEMI,
    STATE(352), 1,
      sym_declaration_list,
    STATE(1997), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [56534] = 7,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4154), 1,
      anon_sym_SEMI,
    STATE(837), 1,
      sym_declaration_list,
    STATE(2026), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [56557] = 5,
    ACTIONS(4158), 1,
      anon_sym_COLON,
    ACTIONS(4160), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3443), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(4156), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [56576] = 4,
    ACTIONS(3523), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3782), 1,
      anon_sym_for,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2598), 4,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_PLUS,
      anon_sym_where,
  [56593] = 7,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(4069), 1,
      anon_sym_LBRACE,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4162), 1,
      anon_sym_SEMI,
    STATE(914), 1,
      sym_block,
    STATE(1938), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [56616] = 7,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    ACTIONS(3916), 1,
      anon_sym_COLON,
    STATE(832), 1,
      sym_declaration_list,
    STATE(1800), 1,
      sym_trait_bounds,
    STATE(2182), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [56639] = 7,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(4081), 1,
      anon_sym_LBRACE,
    ACTIONS(4164), 1,
      anon_sym_SEMI,
    ACTIONS(4166), 1,
      anon_sym_DASH_GT,
    STATE(287), 1,
      sym_block,
    STATE(1907), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [56662] = 7,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4168), 1,
      anon_sym_SEMI,
    STATE(857), 1,
      sym_declaration_list,
    STATE(2037), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [56685] = 7,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(4069), 1,
      anon_sym_LBRACE,
    ACTIONS(4170), 1,
      anon_sym_SEMI,
    ACTIONS(4172), 1,
      anon_sym_DASH_GT,
    STATE(1123), 1,
      sym_block,
    STATE(1949), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [56708] = 5,
    ACTIONS(4176), 1,
      anon_sym_COLON,
    ACTIONS(4178), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3443), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(4174), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [56727] = 4,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2646), 2,
      anon_sym_PLUS,
      anon_sym_DASH_GT,
    ACTIONS(3623), 2,
      anon_sym_COLON,
      anon_sym_PIPE,
    ACTIONS(4180), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [56744] = 7,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(3806), 1,
      anon_sym_LT,
    ACTIONS(4112), 1,
      anon_sym_LBRACE,
    STATE(1022), 1,
      sym_enum_variant_list,
    STATE(1828), 1,
      sym_type_parameters,
    STATE(2200), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [56767] = 7,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4183), 1,
      anon_sym_SEMI,
    STATE(997), 1,
      sym_declaration_list,
    STATE(2029), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [56790] = 7,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4185), 1,
      anon_sym_SEMI,
    STATE(999), 1,
      sym_declaration_list,
    STATE(1979), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [56813] = 7,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4187), 1,
      anon_sym_SEMI,
    STATE(939), 1,
      sym_declaration_list,
    STATE(1990), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [56836] = 7,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4189), 1,
      anon_sym_SEMI,
    STATE(993), 1,
      sym_declaration_list,
    STATE(1983), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [56859] = 7,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(4081), 1,
      anon_sym_LBRACE,
    ACTIONS(4191), 1,
      anon_sym_SEMI,
    ACTIONS(4193), 1,
      anon_sym_DASH_GT,
    STATE(447), 1,
      sym_block,
    STATE(2062), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [56882] = 4,
    ACTIONS(4160), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3443), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2598), 3,
      anon_sym_RPAREN,
      anon_sym_PLUS,
      anon_sym_COMMA,
  [56899] = 7,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(4069), 1,
      anon_sym_LBRACE,
    ACTIONS(4195), 1,
      anon_sym_SEMI,
    ACTIONS(4197), 1,
      anon_sym_DASH_GT,
    STATE(956), 1,
      sym_block,
    STATE(2083), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [56922] = 7,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(3806), 1,
      anon_sym_LT,
    ACTIONS(3828), 1,
      anon_sym_LBRACE,
    STATE(338), 1,
      sym_field_declaration_list,
    STATE(1776), 1,
      sym_type_parameters,
    STATE(2128), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [56945] = 7,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(3806), 1,
      anon_sym_LT,
    ACTIONS(4199), 1,
      anon_sym_LBRACE,
    STATE(323), 1,
      sym_enum_variant_list,
    STATE(1764), 1,
      sym_type_parameters,
    STATE(2161), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [56968] = 7,
    ACTIONS(3554), 1,
      anon_sym_EQ,
    ACTIONS(3916), 1,
      anon_sym_COLON,
    ACTIONS(4201), 1,
      anon_sym_COMMA,
    ACTIONS(4203), 1,
      anon_sym_GT,
    STATE(1968), 1,
      sym_trait_bounds,
    STATE(2092), 1,
      aux_sym_type_parameters_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [56991] = 7,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(4069), 1,
      anon_sym_LBRACE,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4205), 1,
      anon_sym_SEMI,
    STATE(1059), 1,
      sym_block,
    STATE(1967), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [57014] = 4,
    ACTIONS(3523), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3816), 1,
      anon_sym_for,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2598), 4,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_PLUS,
      anon_sym_where,
  [57031] = 7,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(3916), 1,
      anon_sym_COLON,
    STATE(423), 1,
      sym_declaration_list,
    STATE(1844), 1,
      sym_trait_bounds,
    STATE(2206), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [57054] = 4,
    ACTIONS(4180), 1,
      anon_sym_RBRACK,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3623), 2,
      anon_sym_COMMA,
      anon_sym_PIPE,
    ACTIONS(2646), 3,
      anon_sym_SEMI,
      anon_sym_PLUS,
      anon_sym_DASH_GT,
  [57071] = 4,
    ACTIONS(4207), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3443), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2598), 3,
      anon_sym_SEMI,
      anon_sym_RBRACK,
      anon_sym_PLUS,
  [57088] = 7,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4081), 1,
      anon_sym_LBRACE,
    ACTIONS(4209), 1,
      anon_sym_SEMI,
    STATE(372), 1,
      sym_block,
    STATE(1876), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [57111] = 7,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4211), 1,
      anon_sym_SEMI,
    STATE(480), 1,
      sym_declaration_list,
    STATE(2033), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [57134] = 7,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4213), 1,
      anon_sym_SEMI,
    STATE(433), 1,
      sym_declaration_list,
    STATE(2007), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [57157] = 4,
    ACTIONS(3523), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3784), 1,
      anon_sym_for,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2598), 4,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_PLUS,
      anon_sym_where,
  [57174] = 7,
    ACTIONS(3554), 1,
      anon_sym_EQ,
    ACTIONS(3556), 1,
      anon_sym_COMMA,
    ACTIONS(3558), 1,
      anon_sym_GT,
    ACTIONS(3916), 1,
      anon_sym_COLON,
    STATE(1968), 1,
      sym_trait_bounds,
    STATE(1969), 1,
      aux_sym_type_parameters_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [57197] = 7,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4215), 1,
      anon_sym_SEMI,
    STATE(296), 1,
      sym_declaration_list,
    STATE(1910), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [57220] = 7,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(4069), 1,
      anon_sym_LBRACE,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4217), 1,
      anon_sym_SEMI,
    STATE(1108), 1,
      sym_block,
    STATE(1952), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [57243] = 7,
    ACTIONS(3916), 1,
      anon_sym_COLON,
    ACTIONS(4219), 1,
      anon_sym_COMMA,
    ACTIONS(4221), 1,
      anon_sym_GT,
    STATE(1855), 1,
      aux_sym_type_parameters_repeat1,
    STATE(1970), 1,
      sym_trait_bounds,
    STATE(2046), 1,
      aux_sym_for_lifetimes_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [57266] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3704), 2,
      anon_sym_COLON,
      anon_sym_PIPE,
    ACTIONS(2750), 4,
      anon_sym_RPAREN,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_DASH_GT,
  [57281] = 7,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4081), 1,
      anon_sym_LBRACE,
    ACTIONS(4223), 1,
      anon_sym_SEMI,
    STATE(388), 1,
      sym_block,
    STATE(1861), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [57304] = 7,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(3806), 1,
      anon_sym_LT,
    ACTIONS(4199), 1,
      anon_sym_LBRACE,
    STATE(277), 1,
      sym_enum_variant_list,
    STATE(1783), 1,
      sym_type_parameters,
    STATE(2170), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [57327] = 7,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4225), 1,
      anon_sym_SEMI,
    STATE(947), 1,
      sym_declaration_list,
    STATE(1989), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [57350] = 6,
    ACTIONS(798), 1,
      anon_sym_DOT_DOT,
    ACTIONS(3904), 1,
      sym_identifier,
    ACTIONS(3910), 1,
      anon_sym_ref,
    ACTIONS(3912), 1,
      sym_mutable_specifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(2263), 2,
      sym_field_pattern,
      sym_remaining_field_pattern,
  [57371] = 4,
    ACTIONS(4227), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3443), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(2598), 3,
      anon_sym_RPAREN,
      anon_sym_PLUS,
      anon_sym_COMMA,
  [57388] = 7,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(3916), 1,
      anon_sym_COLON,
    STATE(293), 1,
      sym_declaration_list,
    STATE(1797), 1,
      sym_trait_bounds,
    STATE(2124), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [57411] = 7,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(4069), 1,
      anon_sym_LBRACE,
    ACTIONS(4229), 1,
      anon_sym_SEMI,
    ACTIONS(4231), 1,
      anon_sym_DASH_GT,
    STATE(875), 1,
      sym_block,
    STATE(2063), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [57434] = 5,
    ACTIONS(4158), 1,
      anon_sym_COLON,
    ACTIONS(4227), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3443), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(4156), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [57453] = 4,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2750), 2,
      anon_sym_PLUS,
      anon_sym_DASH_GT,
    ACTIONS(3704), 2,
      anon_sym_COLON,
      anon_sym_PIPE,
    ACTIONS(4103), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [57470] = 7,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4233), 1,
      anon_sym_SEMI,
    STATE(1095), 1,
      sym_declaration_list,
    STATE(1957), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [57493] = 7,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(4081), 1,
      anon_sym_LBRACE,
    ACTIONS(4235), 1,
      anon_sym_SEMI,
    ACTIONS(4237), 1,
      anon_sym_DASH_GT,
    STATE(331), 1,
      sym_block,
    STATE(1903), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [57516] = 7,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4239), 1,
      anon_sym_SEMI,
    STATE(1097), 1,
      sym_declaration_list,
    STATE(1956), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [57539] = 7,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4241), 1,
      anon_sym_SEMI,
    STATE(305), 1,
      sym_declaration_list,
    STATE(1911), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [57562] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4243), 5,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
  [57574] = 4,
    ACTIONS(4247), 1,
      anon_sym_as,
    ACTIONS(4249), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4245), 3,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COMMA,
  [57590] = 5,
    ACTIONS(4251), 1,
      anon_sym_RPAREN,
    ACTIONS(4253), 1,
      anon_sym_COMMA,
    STATE(2066), 1,
      aux_sym_parameters_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3431), 2,
      anon_sym_COLON,
      anon_sym_PIPE,
  [57608] = 5,
    ACTIONS(4255), 1,
      anon_sym_RPAREN,
    ACTIONS(4257), 1,
      anon_sym_COMMA,
    STATE(2070), 1,
      aux_sym_parameters_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3431), 2,
      anon_sym_COLON,
      anon_sym_PIPE,
  [57626] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3056), 5,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_COLON,
      anon_sym_where,
      anon_sym_EQ,
  [57638] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3431), 2,
      anon_sym_COLON,
      anon_sym_PIPE,
    ACTIONS(2598), 3,
      anon_sym_RPAREN,
      anon_sym_PLUS,
      anon_sym_COMMA,
  [57652] = 6,
    ACTIONS(3916), 1,
      anon_sym_COLON,
    ACTIONS(4116), 1,
      anon_sym_COMMA,
    ACTIONS(4118), 1,
      anon_sym_GT,
    STATE(1855), 1,
      aux_sym_type_parameters_repeat1,
    STATE(1970), 1,
      sym_trait_bounds,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [57672] = 6,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(3451), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3552), 1,
      anon_sym_COLON,
    STATE(1342), 1,
      sym_type_arguments,
    STATE(2071), 1,
      sym_trait_bounds,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [57692] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4259), 5,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
  [57704] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4261), 5,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
  [57716] = 6,
    ACTIONS(3916), 1,
      anon_sym_COLON,
    ACTIONS(4263), 1,
      anon_sym_COMMA,
    ACTIONS(4265), 1,
      anon_sym_GT,
    STATE(1970), 1,
      sym_trait_bounds,
    STATE(2094), 1,
      aux_sym_type_parameters_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [57736] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3660), 2,
      anon_sym_COLON,
      anon_sym_PIPE,
    ACTIONS(2896), 3,
      anon_sym_RPAREN,
      anon_sym_PLUS,
      anon_sym_COMMA,
  [57750] = 5,
    ACTIONS(4267), 1,
      anon_sym_RPAREN,
    ACTIONS(4270), 1,
      anon_sym_COMMA,
    STATE(2070), 1,
      aux_sym_parameters_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3431), 2,
      anon_sym_COLON,
      anon_sym_PIPE,
  [57768] = 4,
    ACTIONS(4160), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3443), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(4273), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [57784] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4275), 5,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
  [57796] = 4,
    ACTIONS(4227), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3443), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(4174), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [57812] = 4,
    ACTIONS(4247), 1,
      anon_sym_as,
    ACTIONS(4277), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4245), 3,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COMMA,
  [57828] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4279), 5,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
  [57840] = 5,
    ACTIONS(766), 1,
      anon_sym_RPAREN,
    ACTIONS(4281), 1,
      anon_sym_COMMA,
    STATE(1943), 1,
      aux_sym_parameters_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3431), 2,
      anon_sym_COLON,
      anon_sym_PIPE,
  [57858] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4283), 5,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
  [57870] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3042), 5,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_COLON,
      anon_sym_where,
      anon_sym_EQ,
  [57882] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3038), 5,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_COLON,
      anon_sym_where,
      anon_sym_EQ,
  [57894] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4285), 5,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
  [57906] = 4,
    ACTIONS(2598), 1,
      anon_sym_PLUS,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3431), 2,
      anon_sym_COLON,
      anon_sym_PIPE,
    ACTIONS(4287), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [57922] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4290), 5,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
  [57934] = 4,
    ACTIONS(4287), 1,
      anon_sym_RBRACK,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2598), 2,
      anon_sym_SEMI,
      anon_sym_PLUS,
    ACTIONS(3431), 2,
      anon_sym_COMMA,
      anon_sym_PIPE,
  [57950] = 5,
    ACTIONS(762), 1,
      anon_sym_RPAREN,
    ACTIONS(4292), 1,
      anon_sym_COMMA,
    STATE(1995), 1,
      aux_sym_parameters_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3431), 2,
      anon_sym_COLON,
      anon_sym_PIPE,
  [57968] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4294), 5,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
  [57980] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4296), 5,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
  [57992] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4298), 5,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
  [58004] = 6,
    ACTIONS(3598), 1,
      anon_sym_PIPE,
    ACTIONS(4300), 1,
      anon_sym_SEMI,
    ACTIONS(4302), 1,
      anon_sym_COLON,
    ACTIONS(4304), 1,
      anon_sym_EQ,
    ACTIONS(4306), 1,
      anon_sym_else,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [58024] = 6,
    ACTIONS(4308), 1,
      anon_sym_SEMI,
    ACTIONS(4310), 1,
      anon_sym_COLON,
    ACTIONS(4312), 1,
      anon_sym_EQ,
    ACTIONS(4314), 1,
      anon_sym_else,
    ACTIONS(4316), 1,
      anon_sym_PIPE,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [58044] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4318), 5,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
  [58056] = 4,
    ACTIONS(2896), 1,
      anon_sym_PLUS,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3660), 2,
      anon_sym_COLON,
      anon_sym_PIPE,
    ACTIONS(4320), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [58072] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4323), 5,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
  [58084] = 6,
    ACTIONS(2552), 1,
      anon_sym_LPAREN,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(3451), 1,
      anon_sym_COLON_COLON,
    STATE(781), 1,
      sym_parameters,
    STATE(1342), 1,
      sym_type_arguments,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [58104] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3086), 5,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_COLON,
      anon_sym_where,
      anon_sym_EQ,
  [58116] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3060), 5,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_COLON,
      anon_sym_where,
      anon_sym_EQ,
  [58128] = 4,
    ACTIONS(4227), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3443), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(4273), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [58144] = 6,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(3449), 1,
      anon_sym_LPAREN,
    ACTIONS(3451), 1,
      anon_sym_COLON_COLON,
    STATE(1342), 1,
      sym_type_arguments,
    STATE(1353), 1,
      sym_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [58164] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4325), 5,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_where,
      anon_sym_EQ,
      anon_sym_COMMA,
  [58176] = 6,
    ACTIONS(3598), 1,
      anon_sym_PIPE,
    ACTIONS(4327), 1,
      anon_sym_SEMI,
    ACTIONS(4329), 1,
      anon_sym_COLON,
    ACTIONS(4331), 1,
      anon_sym_EQ,
    ACTIONS(4333), 1,
      anon_sym_else,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [58196] = 3,
    ACTIONS(4335), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2714), 4,
      anon_sym_PLUS,
      anon_sym_COMMA,
      anon_sym_GT,
      anon_sym_COLON_COLON,
  [58210] = 6,
    ACTIONS(4316), 1,
      anon_sym_PIPE,
    ACTIONS(4337), 1,
      anon_sym_RPAREN,
    ACTIONS(4339), 1,
      anon_sym_COLON,
    ACTIONS(4341), 1,
      anon_sym_COMMA,
    STATE(2061), 1,
      aux_sym_tuple_pattern_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [58230] = 4,
    ACTIONS(4247), 1,
      anon_sym_as,
    ACTIONS(4343), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4245), 3,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COMMA,
  [58246] = 4,
    ACTIONS(4160), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3443), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
    ACTIONS(4174), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [58262] = 6,
    ACTIONS(4316), 1,
      anon_sym_PIPE,
    ACTIONS(4345), 1,
      anon_sym_SEMI,
    ACTIONS(4347), 1,
      anon_sym_COLON,
    ACTIONS(4349), 1,
      anon_sym_EQ,
    ACTIONS(4351), 1,
      anon_sym_else,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [58282] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3950), 5,
      anon_sym_LPAREN,
      anon_sym_RPAREN,
      anon_sym_LBRACE,
      anon_sym_RBRACE,
      anon_sym_LBRACK,
  [58294] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3090), 5,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_COLON,
      anon_sym_where,
      anon_sym_EQ,
  [58306] = 4,
    ACTIONS(4320), 1,
      anon_sym_RBRACK,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2896), 2,
      anon_sym_SEMI,
      anon_sym_PLUS,
    ACTIONS(3660), 2,
      anon_sym_COMMA,
      anon_sym_PIPE,
  [58322] = 5,
    ACTIONS(3554), 1,
      anon_sym_EQ,
    ACTIONS(3916), 1,
      anon_sym_COLON,
    STATE(1968), 1,
      sym_trait_bounds,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4353), 2,
      anon_sym_COMMA,
      anon_sym_GT,
  [58340] = 4,
    ACTIONS(4357), 1,
      anon_sym_as,
    ACTIONS(4359), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4355), 3,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COMMA,
  [58356] = 4,
    ACTIONS(4158), 1,
      anon_sym_COLON,
    ACTIONS(4178), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3443), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
  [58371] = 5,
    ACTIONS(4116), 1,
      anon_sym_COMMA,
    ACTIONS(4118), 1,
      anon_sym_GT,
    ACTIONS(4361), 1,
      anon_sym_EQ,
    STATE(1855), 1,
      aux_sym_type_parameters_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [58388] = 5,
    ACTIONS(4316), 1,
      anon_sym_PIPE,
    ACTIONS(4337), 1,
      anon_sym_RPAREN,
    ACTIONS(4341), 1,
      anon_sym_COMMA,
    STATE(2061), 1,
      aux_sym_tuple_pattern_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [58405] = 5,
    ACTIONS(4363), 1,
      anon_sym_LPAREN,
    ACTIONS(4365), 1,
      anon_sym_LBRACE,
    ACTIONS(4367), 1,
      anon_sym_LBRACK,
    STATE(1336), 1,
      sym_delim_token_tree,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [58422] = 5,
    ACTIONS(3802), 1,
      anon_sym_LBRACE,
    ACTIONS(3804), 1,
      anon_sym_where,
    STATE(821), 1,
      sym_field_declaration_list,
    STATE(2193), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [58439] = 4,
    ACTIONS(4369), 1,
      anon_sym_DQUOTE,
    STATE(1774), 1,
      aux_sym_string_literal_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4371), 2,
      sym__string_content,
      sym_escape_sequence,
  [58454] = 4,
    ACTIONS(3), 1,
      sym_block_comment,
    ACTIONS(912), 1,
      sym_line_comment,
    ACTIONS(4373), 1,
      aux_sym_token_repetition_pattern_token1,
    ACTIONS(4375), 3,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_QMARK,
  [58469] = 5,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(4199), 1,
      anon_sym_LBRACE,
    STATE(411), 1,
      sym_enum_variant_list,
    STATE(2213), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [58486] = 5,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    STATE(833), 1,
      sym_declaration_list,
    STATE(2181), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [58503] = 3,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4377), 3,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
  [58516] = 5,
    ACTIONS(762), 1,
      anon_sym_RPAREN,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4292), 1,
      anon_sym_COMMA,
    STATE(1995), 1,
      aux_sym_parameters_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [58533] = 5,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(4112), 1,
      anon_sym_LBRACE,
    STATE(827), 1,
      sym_enum_variant_list,
    STATE(2188), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [58550] = 5,
    ACTIONS(4379), 1,
      anon_sym_LPAREN,
    ACTIONS(4381), 1,
      anon_sym_LBRACE,
    ACTIONS(4383), 1,
      anon_sym_LBRACK,
    STATE(81), 1,
      sym_delim_token_tree,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [58567] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3431), 2,
      anon_sym_COLON,
      anon_sym_PIPE,
    ACTIONS(4385), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [58580] = 5,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    STATE(845), 1,
      sym_declaration_list,
    STATE(2176), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [58597] = 5,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    ACTIONS(3804), 1,
      anon_sym_where,
    STATE(385), 1,
      sym_declaration_list,
    STATE(2306), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [58614] = 4,
    ACTIONS(3916), 1,
      anon_sym_COLON,
    STATE(1970), 1,
      sym_trait_bounds,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4387), 2,
      anon_sym_COMMA,
      anon_sym_GT,
  [58629] = 4,
    ACTIONS(4389), 1,
      anon_sym_DQUOTE,
    STATE(1839), 1,
      aux_sym_string_literal_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4391), 2,
      sym__string_content,
      sym_escape_sequence,
  [58644] = 5,
    ACTIONS(3449), 1,
      anon_sym_LPAREN,
    ACTIONS(3806), 1,
      anon_sym_LT,
    STATE(1646), 1,
      sym_parameters,
    STATE(2112), 1,
      sym_type_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [58661] = 5,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(3828), 1,
      anon_sym_LBRACE,
    STATE(403), 1,
      sym_field_declaration_list,
    STATE(2223), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [58678] = 5,
    ACTIONS(3449), 1,
      anon_sym_LPAREN,
    ACTIONS(3806), 1,
      anon_sym_LT,
    STATE(1627), 1,
      sym_parameters,
    STATE(2234), 1,
      sym_type_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [58695] = 5,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    ACTIONS(3804), 1,
      anon_sym_where,
    STATE(425), 1,
      sym_declaration_list,
    STATE(2205), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [58712] = 5,
    ACTIONS(4363), 1,
      anon_sym_LPAREN,
    ACTIONS(4365), 1,
      anon_sym_LBRACE,
    ACTIONS(4367), 1,
      anon_sym_LBRACK,
    STATE(1333), 1,
      sym_delim_token_tree,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [58729] = 4,
    ACTIONS(284), 1,
      anon_sym_LBRACE,
    ACTIONS(4393), 1,
      anon_sym_if,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(869), 2,
      sym_if_expression,
      sym_block,
  [58744] = 5,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4255), 1,
      anon_sym_RPAREN,
    ACTIONS(4257), 1,
      anon_sym_COMMA,
    STATE(2070), 1,
      aux_sym_parameters_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [58761] = 5,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4395), 1,
      anon_sym_SEMI,
    ACTIONS(4397), 1,
      anon_sym_EQ,
    ACTIONS(4399), 1,
      anon_sym_else,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [58778] = 5,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(4199), 1,
      anon_sym_LBRACE,
    STATE(394), 1,
      sym_enum_variant_list,
    STATE(2293), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [58795] = 5,
    ACTIONS(4401), 1,
      anon_sym_LPAREN,
    ACTIONS(4403), 1,
      anon_sym_LBRACE,
    ACTIONS(4405), 1,
      anon_sym_LBRACK,
    STATE(1971), 1,
      sym_token_tree,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [58812] = 5,
    ACTIONS(2449), 1,
      anon_sym_PLUS,
    ACTIONS(4407), 1,
      anon_sym_COMMA,
    ACTIONS(4409), 1,
      anon_sym_GT,
    STATE(2058), 1,
      aux_sym_type_arguments_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [58829] = 5,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4407), 1,
      anon_sym_COMMA,
    ACTIONS(4409), 1,
      anon_sym_GT,
    STATE(2058), 1,
      aux_sym_type_arguments_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [58846] = 5,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4251), 1,
      anon_sym_RPAREN,
    ACTIONS(4253), 1,
      anon_sym_COMMA,
    STATE(2066), 1,
      aux_sym_parameters_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [58863] = 5,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(3916), 1,
      anon_sym_COLON,
    STATE(1347), 1,
      sym_type_arguments,
    STATE(2074), 1,
      sym_trait_bounds,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [58880] = 5,
    ACTIONS(4316), 1,
      anon_sym_PIPE,
    ACTIONS(4411), 1,
      anon_sym_RPAREN,
    ACTIONS(4413), 1,
      anon_sym_COMMA,
    STATE(1904), 1,
      aux_sym_tuple_pattern_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [58897] = 4,
    ACTIONS(4417), 1,
      anon_sym_COMMA,
    STATE(1794), 1,
      aux_sym_where_clause_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4415), 2,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [58912] = 5,
    ACTIONS(4316), 1,
      anon_sym_PIPE,
    ACTIONS(4419), 1,
      anon_sym_RBRACK,
    ACTIONS(4421), 1,
      anon_sym_COMMA,
    STATE(1905), 1,
      aux_sym_tuple_pattern_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [58929] = 5,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(3449), 1,
      anon_sym_LPAREN,
    STATE(1347), 1,
      sym_type_arguments,
    STATE(1356), 1,
      sym_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [58946] = 4,
    ACTIONS(4423), 1,
      anon_sym_DQUOTE,
    STATE(1839), 1,
      aux_sym_string_literal_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4391), 2,
      sym__string_content,
      sym_escape_sequence,
  [58961] = 4,
    ACTIONS(4425), 1,
      anon_sym_COMMA,
    STATE(1825), 1,
      aux_sym_where_clause_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3006), 2,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [58976] = 5,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4427), 1,
      anon_sym_COMMA,
    ACTIONS(4429), 1,
      anon_sym_GT,
    STATE(2047), 1,
      aux_sym_type_arguments_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [58993] = 4,
    ACTIONS(632), 1,
      anon_sym_LBRACE,
    ACTIONS(4431), 1,
      anon_sym_if,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(233), 2,
      sym_if_expression,
      sym_block,
  [59008] = 5,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    ACTIONS(3804), 1,
      anon_sym_where,
    STATE(246), 1,
      sym_declaration_list,
    STATE(2199), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [59025] = 5,
    ACTIONS(4339), 1,
      anon_sym_COLON,
    ACTIONS(4433), 1,
      anon_sym_COMMA,
    ACTIONS(4435), 1,
      anon_sym_PIPE,
    STATE(1920), 1,
      aux_sym_closure_parameters_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [59042] = 5,
    ACTIONS(2449), 1,
      anon_sym_PLUS,
    ACTIONS(4427), 1,
      anon_sym_COMMA,
    ACTIONS(4429), 1,
      anon_sym_GT,
    STATE(2047), 1,
      aux_sym_type_arguments_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [59059] = 5,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    STATE(1004), 1,
      sym_declaration_list,
    STATE(2226), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [59076] = 4,
    ACTIONS(4437), 1,
      anon_sym_DQUOTE,
    STATE(1805), 1,
      aux_sym_string_literal_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4439), 2,
      sym__string_content,
      sym_escape_sequence,
  [59091] = 5,
    ACTIONS(766), 1,
      anon_sym_RPAREN,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4281), 1,
      anon_sym_COMMA,
    STATE(1943), 1,
      aux_sym_parameters_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [59108] = 5,
    ACTIONS(4441), 1,
      anon_sym_LPAREN,
    ACTIONS(4443), 1,
      anon_sym_LBRACE,
    ACTIONS(4445), 1,
      anon_sym_LBRACK,
    STATE(918), 1,
      sym_delim_token_tree,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [59125] = 3,
    ACTIONS(4316), 1,
      anon_sym_PIPE,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4447), 3,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
      anon_sym_COMMA,
  [59138] = 4,
    ACTIONS(4449), 1,
      anon_sym_DQUOTE,
    STATE(1839), 1,
      aux_sym_string_literal_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4391), 2,
      sym__string_content,
      sym_escape_sequence,
  [59153] = 4,
    ACTIONS(4451), 1,
      anon_sym_COMMA,
    STATE(1806), 1,
      aux_sym_tuple_pattern_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4447), 2,
      anon_sym_RPAREN,
      anon_sym_RBRACK,
  [59168] = 5,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    STATE(1035), 1,
      sym_declaration_list,
    STATE(2231), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [59185] = 5,
    ACTIONS(4441), 1,
      anon_sym_LPAREN,
    ACTIONS(4443), 1,
      anon_sym_LBRACE,
    ACTIONS(4445), 1,
      anon_sym_LBRACK,
    STATE(930), 1,
      sym_delim_token_tree,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [59202] = 3,
    ACTIONS(4454), 1,
      anon_sym_in,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4456), 3,
      sym_self,
      sym_super,
      sym_crate,
  [59215] = 5,
    ACTIONS(3449), 1,
      anon_sym_LPAREN,
    ACTIONS(3806), 1,
      anon_sym_LT,
    STATE(1612), 1,
      sym_parameters,
    STATE(2153), 1,
      sym_type_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [59232] = 5,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4458), 1,
      anon_sym_SEMI,
    ACTIONS(4460), 1,
      anon_sym_EQ,
    ACTIONS(4462), 1,
      anon_sym_else,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [59249] = 5,
    ACTIONS(3802), 1,
      anon_sym_LBRACE,
    ACTIONS(3804), 1,
      anon_sym_where,
    STATE(900), 1,
      sym_field_declaration_list,
    STATE(2119), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [59266] = 5,
    ACTIONS(3916), 1,
      anon_sym_COLON,
    ACTIONS(4464), 1,
      anon_sym_SEMI,
    ACTIONS(4466), 1,
      anon_sym_EQ,
    STATE(2437), 1,
      sym_trait_bounds,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [59283] = 5,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    STATE(808), 1,
      sym_declaration_list,
    STATE(2114), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [59300] = 5,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    ACTIONS(3804), 1,
      anon_sym_where,
    STATE(266), 1,
      sym_declaration_list,
    STATE(2280), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [59317] = 5,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(3828), 1,
      anon_sym_LBRACE,
    STATE(304), 1,
      sym_field_declaration_list,
    STATE(2172), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [59334] = 5,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4468), 1,
      anon_sym_SEMI,
    ACTIONS(4470), 1,
      anon_sym_EQ,
    ACTIONS(4472), 1,
      anon_sym_else,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [59351] = 5,
    ACTIONS(2552), 1,
      anon_sym_LPAREN,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    STATE(783), 1,
      sym_parameters,
    STATE(1347), 1,
      sym_type_arguments,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [59368] = 3,
    ACTIONS(4474), 1,
      anon_sym_in,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4476), 3,
      sym_self,
      sym_super,
      sym_crate,
  [59381] = 3,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3431), 2,
      anon_sym_COLON,
      anon_sym_PIPE,
    ACTIONS(4478), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [59394] = 5,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4480), 1,
      anon_sym_RPAREN,
    ACTIONS(4482), 1,
      anon_sym_COMMA,
    STATE(2024), 1,
      aux_sym_ordered_field_declaration_list_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [59411] = 5,
    ACTIONS(4263), 1,
      anon_sym_COMMA,
    ACTIONS(4265), 1,
      anon_sym_GT,
    ACTIONS(4361), 1,
      anon_sym_EQ,
    STATE(2094), 1,
      aux_sym_type_parameters_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [59428] = 5,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4484), 1,
      anon_sym_RPAREN,
    ACTIONS(4486), 1,
      anon_sym_COMMA,
    STATE(1918), 1,
      aux_sym_ordered_field_declaration_list_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [59445] = 4,
    ACTIONS(3), 1,
      sym_block_comment,
    ACTIONS(912), 1,
      sym_line_comment,
    ACTIONS(4488), 1,
      aux_sym_token_repetition_pattern_token1,
    ACTIONS(4490), 3,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_QMARK,
  [59460] = 4,
    ACTIONS(4494), 1,
      anon_sym_COMMA,
    STATE(1825), 1,
      aux_sym_where_clause_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4492), 2,
      anon_sym_SEMI,
      anon_sym_LBRACE,
  [59475] = 4,
    ACTIONS(15), 1,
      anon_sym_LBRACE,
    ACTIONS(4497), 1,
      anon_sym_if,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(66), 2,
      sym_if_expression,
      sym_block,
  [59490] = 5,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4499), 1,
      anon_sym_RPAREN,
    ACTIONS(4501), 1,
      anon_sym_COMMA,
    STATE(2090), 1,
      aux_sym_tuple_type_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [59507] = 5,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(4112), 1,
      anon_sym_LBRACE,
    STATE(963), 1,
      sym_enum_variant_list,
    STATE(2149), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [59524] = 4,
    ACTIONS(4503), 1,
      anon_sym_DQUOTE,
    STATE(1849), 1,
      aux_sym_string_literal_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4505), 2,
      sym__string_content,
      sym_escape_sequence,
  [59539] = 5,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    ACTIONS(3804), 1,
      anon_sym_where,
    STATE(290), 1,
      sym_declaration_list,
    STATE(2185), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [59556] = 5,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4507), 1,
      anon_sym_RPAREN,
    ACTIONS(4509), 1,
      anon_sym_COMMA,
    STATE(1963), 1,
      aux_sym_tuple_type_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [59573] = 4,
    ACTIONS(3916), 1,
      anon_sym_COLON,
    STATE(1970), 1,
      sym_trait_bounds,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4511), 2,
      anon_sym_COMMA,
      anon_sym_GT,
  [59588] = 5,
    ACTIONS(3449), 1,
      anon_sym_LPAREN,
    ACTIONS(3806), 1,
      anon_sym_LT,
    STATE(1702), 1,
      sym_parameters,
    STATE(2175), 1,
      sym_type_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [59605] = 5,
    ACTIONS(4379), 1,
      anon_sym_LPAREN,
    ACTIONS(4381), 1,
      anon_sym_LBRACE,
    ACTIONS(4383), 1,
      anon_sym_LBRACK,
    STATE(83), 1,
      sym_delim_token_tree,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [59622] = 5,
    ACTIONS(3449), 1,
      anon_sym_LPAREN,
    ACTIONS(3806), 1,
      anon_sym_LT,
    STATE(1649), 1,
      sym_parameters,
    STATE(2126), 1,
      sym_type_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [59639] = 4,
    ACTIONS(3), 1,
      sym_block_comment,
    ACTIONS(912), 1,
      sym_line_comment,
    ACTIONS(4514), 1,
      aux_sym_token_repetition_pattern_token1,
    ACTIONS(4516), 3,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_QMARK,
  [59654] = 5,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4518), 1,
      anon_sym_RPAREN,
    ACTIONS(4520), 1,
      anon_sym_COMMA,
    STATE(1882), 1,
      aux_sym_ordered_field_declaration_list_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [59671] = 4,
    ACTIONS(2317), 1,
      anon_sym_POUND,
    ACTIONS(4522), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    STATE(1153), 2,
      sym_attribute_item,
      aux_sym_enum_variant_list_repeat1,
  [59686] = 4,
    ACTIONS(4524), 1,
      anon_sym_DQUOTE,
    STATE(1839), 1,
      aux_sym_string_literal_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4526), 2,
      sym__string_content,
      sym_escape_sequence,
  [59701] = 5,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4529), 1,
      anon_sym_SEMI,
    ACTIONS(4531), 1,
      anon_sym_EQ,
    ACTIONS(4533), 1,
      anon_sym_else,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [59718] = 5,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    STATE(1116), 1,
      sym_declaration_list,
    STATE(2255), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [59735] = 3,
    ACTIONS(4535), 1,
      anon_sym_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3598), 3,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
  [59748] = 4,
    ACTIONS(4176), 1,
      anon_sym_COLON,
    ACTIONS(4178), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3443), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
  [59763] = 5,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    ACTIONS(3804), 1,
      anon_sym_where,
    STATE(291), 1,
      sym_declaration_list,
    STATE(2276), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [59780] = 5,
    ACTIONS(3995), 1,
      anon_sym_LPAREN,
    ACTIONS(3997), 1,
      anon_sym_LBRACE,
    ACTIONS(3999), 1,
      anon_sym_LBRACK,
    STATE(1023), 1,
      sym_delim_token_tree,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [59797] = 3,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4537), 3,
      anon_sym_RPAREN,
      anon_sym_COMMA,
      anon_sym_PIPE,
  [59810] = 4,
    ACTIONS(3), 1,
      sym_block_comment,
    ACTIONS(912), 1,
      sym_line_comment,
    ACTIONS(4539), 1,
      aux_sym_token_repetition_pattern_token1,
    ACTIONS(4541), 3,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_QMARK,
  [59825] = 5,
    ACTIONS(3995), 1,
      anon_sym_LPAREN,
    ACTIONS(3997), 1,
      anon_sym_LBRACE,
    ACTIONS(3999), 1,
      anon_sym_LBRACK,
    STATE(973), 1,
      sym_delim_token_tree,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [59842] = 4,
    ACTIONS(4543), 1,
      anon_sym_DQUOTE,
    STATE(1839), 1,
      aux_sym_string_literal_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4391), 2,
      sym__string_content,
      sym_escape_sequence,
  [59857] = 5,
    ACTIONS(3916), 1,
      anon_sym_COLON,
    ACTIONS(4545), 1,
      anon_sym_SEMI,
    ACTIONS(4547), 1,
      anon_sym_EQ,
    STATE(2422), 1,
      sym_trait_bounds,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [59874] = 5,
    ACTIONS(3249), 1,
      anon_sym_LBRACE,
    ACTIONS(4106), 1,
      sym_identifier,
    ACTIONS(4108), 1,
      anon_sym_STAR,
    STATE(2060), 1,
      sym_use_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [59891] = 5,
    ACTIONS(3449), 1,
      anon_sym_LPAREN,
    ACTIONS(3806), 1,
      anon_sym_LT,
    STATE(1674), 1,
      sym_parameters,
    STATE(2202), 1,
      sym_type_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [59908] = 4,
    ACTIONS(4549), 1,
      anon_sym_DQUOTE,
    STATE(1793), 1,
      aux_sym_string_literal_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4551), 2,
      sym__string_content,
      sym_escape_sequence,
  [59923] = 5,
    ACTIONS(3249), 1,
      anon_sym_LBRACE,
    ACTIONS(4553), 1,
      sym_identifier,
    ACTIONS(4555), 1,
      anon_sym_STAR,
    STATE(2054), 1,
      sym_use_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [59940] = 4,
    ACTIONS(3836), 1,
      anon_sym_GT,
    ACTIONS(4557), 1,
      anon_sym_COMMA,
    STATE(2040), 1,
      aux_sym_type_parameters_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [59954] = 4,
    ACTIONS(284), 1,
      anon_sym_LBRACE,
    ACTIONS(4559), 1,
      anon_sym_move,
    STATE(1047), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [59968] = 4,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(4561), 1,
      sym_identifier,
    STATE(2374), 1,
      sym_type_arguments,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [59982] = 4,
    ACTIONS(4081), 1,
      anon_sym_LBRACE,
    ACTIONS(4563), 1,
      anon_sym_SEMI,
    STATE(400), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [59996] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4565), 3,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_QMARK,
  [60006] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4567), 3,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_QMARK,
  [60016] = 4,
    ACTIONS(4081), 1,
      anon_sym_LBRACE,
    ACTIONS(4569), 1,
      anon_sym_SEMI,
    STATE(322), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60030] = 4,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(4571), 1,
      sym_identifier,
    STATE(2374), 1,
      sym_type_arguments,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60044] = 4,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(4571), 1,
      sym_identifier,
    STATE(2372), 1,
      sym_type_arguments,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60058] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4573), 3,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_QMARK,
  [60068] = 4,
    ACTIONS(4081), 1,
      anon_sym_LBRACE,
    ACTIONS(4575), 1,
      anon_sym_SEMI,
    STATE(381), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60082] = 3,
    ACTIONS(3163), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4577), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [60094] = 3,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4579), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [60106] = 4,
    ACTIONS(626), 1,
      anon_sym_RBRACK,
    ACTIONS(4581), 1,
      anon_sym_COMMA,
    STATE(1972), 1,
      aux_sym_array_expression_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60120] = 4,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4583), 1,
      anon_sym_SEMI,
    ACTIONS(4585), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60134] = 3,
    ACTIONS(4316), 1,
      anon_sym_PIPE,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4587), 2,
      anon_sym_RBRACE,
      anon_sym_COMMA,
  [60146] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2395), 3,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
  [60156] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4589), 3,
      anon_sym_PLUS,
      anon_sym_STAR,
      anon_sym_QMARK,
  [60166] = 4,
    ACTIONS(4081), 1,
      anon_sym_LBRACE,
    ACTIONS(4591), 1,
      anon_sym_SEMI,
    STATE(432), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60180] = 4,
    ACTIONS(284), 1,
      anon_sym_LBRACE,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    STATE(817), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60194] = 4,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4593), 1,
      anon_sym_SEMI,
    ACTIONS(4595), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60208] = 4,
    ACTIONS(4081), 1,
      anon_sym_LBRACE,
    ACTIONS(4597), 1,
      anon_sym_SEMI,
    STATE(457), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60222] = 4,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    ACTIONS(4599), 1,
      anon_sym_SEMI,
    STATE(478), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60236] = 4,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    ACTIONS(4601), 1,
      anon_sym_SEMI,
    STATE(473), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60250] = 3,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4603), 2,
      anon_sym_RBRACE,
      anon_sym_COMMA,
  [60262] = 4,
    ACTIONS(4605), 1,
      anon_sym_RBRACE,
    ACTIONS(4607), 1,
      anon_sym_COMMA,
    STATE(2077), 1,
      aux_sym_enum_variant_list_repeat2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60276] = 4,
    ACTIONS(3806), 1,
      anon_sym_LT,
    ACTIONS(4609), 1,
      anon_sym_EQ,
    STATE(2456), 1,
      sym_type_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60290] = 4,
    ACTIONS(4611), 1,
      anon_sym_RPAREN,
    ACTIONS(4613), 1,
      anon_sym_COMMA,
    STATE(1916), 1,
      aux_sym_ordered_field_declaration_list_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60304] = 3,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4615), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [60316] = 4,
    ACTIONS(4617), 1,
      sym_identifier,
    ACTIONS(4619), 1,
      anon_sym_ref,
    ACTIONS(4621), 1,
      sym_mutable_specifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60330] = 4,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(4623), 1,
      sym_identifier,
    STATE(2372), 1,
      sym_type_arguments,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60344] = 4,
    ACTIONS(3546), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3552), 1,
      anon_sym_COLON,
    STATE(2074), 1,
      sym_trait_bounds,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60358] = 4,
    ACTIONS(3523), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3552), 1,
      anon_sym_COLON,
    STATE(2074), 1,
      sym_trait_bounds,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60372] = 4,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(4623), 1,
      sym_identifier,
    STATE(2374), 1,
      sym_type_arguments,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60386] = 3,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4625), 2,
      anon_sym_COMMA,
      anon_sym_GT,
  [60398] = 3,
    ACTIONS(4316), 1,
      anon_sym_PIPE,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4627), 2,
      anon_sym_RBRACE,
      anon_sym_COMMA,
  [60410] = 4,
    ACTIONS(4081), 1,
      anon_sym_LBRACE,
    ACTIONS(4629), 1,
      anon_sym_SEMI,
    STATE(430), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60424] = 4,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    ACTIONS(4631), 1,
      anon_sym_SEMI,
    STATE(325), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60438] = 3,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4385), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [60450] = 4,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(4633), 1,
      sym_identifier,
    STATE(2374), 1,
      sym_type_arguments,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60464] = 4,
    ACTIONS(3578), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3961), 1,
      anon_sym_BANG,
    ACTIONS(4635), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60478] = 4,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(4633), 1,
      sym_identifier,
    STATE(2372), 1,
      sym_type_arguments,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60492] = 4,
    ACTIONS(4637), 1,
      anon_sym_RBRACE,
    ACTIONS(4639), 1,
      anon_sym_COMMA,
    STATE(2095), 1,
      aux_sym_field_initializer_list_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60506] = 4,
    ACTIONS(4641), 1,
      sym_identifier,
    ACTIONS(4643), 1,
      anon_sym_ref,
    ACTIONS(4645), 1,
      sym_mutable_specifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60520] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2385), 3,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
  [60530] = 4,
    ACTIONS(4081), 1,
      anon_sym_LBRACE,
    ACTIONS(4647), 1,
      anon_sym_SEMI,
    STATE(410), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60544] = 4,
    ACTIONS(4255), 1,
      anon_sym_RPAREN,
    ACTIONS(4257), 1,
      anon_sym_COMMA,
    STATE(2070), 1,
      aux_sym_parameters_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60558] = 4,
    ACTIONS(2449), 1,
      anon_sym_PLUS,
    ACTIONS(4649), 1,
      sym_mutable_specifier,
    ACTIONS(4651), 1,
      sym_self,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60572] = 4,
    ACTIONS(4081), 1,
      anon_sym_LBRACE,
    ACTIONS(4653), 1,
      anon_sym_SEMI,
    STATE(453), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60586] = 4,
    ACTIONS(2381), 1,
      anon_sym_RPAREN,
    ACTIONS(4655), 1,
      anon_sym_COMMA,
    STATE(1806), 1,
      aux_sym_tuple_pattern_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60600] = 4,
    ACTIONS(2391), 1,
      anon_sym_RBRACK,
    ACTIONS(4657), 1,
      anon_sym_COMMA,
    STATE(1806), 1,
      aux_sym_tuple_pattern_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60614] = 4,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(4659), 1,
      sym_identifier,
    STATE(2372), 1,
      sym_type_arguments,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60628] = 4,
    ACTIONS(4081), 1,
      anon_sym_LBRACE,
    ACTIONS(4661), 1,
      anon_sym_SEMI,
    STATE(366), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60642] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4663), 3,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COMMA,
  [60652] = 4,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(4659), 1,
      sym_identifier,
    STATE(2374), 1,
      sym_type_arguments,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60666] = 4,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    ACTIONS(4665), 1,
      anon_sym_SEMI,
    STATE(354), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60680] = 4,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    ACTIONS(4667), 1,
      anon_sym_SEMI,
    STATE(332), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60694] = 4,
    ACTIONS(3854), 1,
      anon_sym_RBRACE,
    ACTIONS(4669), 1,
      anon_sym_COMMA,
    STATE(1913), 1,
      aux_sym_field_declaration_list_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60708] = 4,
    ACTIONS(4671), 1,
      anon_sym_RBRACE,
    ACTIONS(4673), 1,
      anon_sym_COMMA,
    STATE(1913), 1,
      aux_sym_field_declaration_list_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60722] = 3,
    ACTIONS(4678), 1,
      anon_sym_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4676), 2,
      anon_sym_RBRACE,
      anon_sym_COMMA,
  [60734] = 3,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4680), 2,
      anon_sym_RBRACE,
      anon_sym_COMMA,
  [60746] = 4,
    ACTIONS(4682), 1,
      anon_sym_RPAREN,
    ACTIONS(4684), 1,
      anon_sym_COMMA,
    STATE(1916), 1,
      aux_sym_ordered_field_declaration_list_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60760] = 3,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4687), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [60772] = 4,
    ACTIONS(4689), 1,
      anon_sym_RPAREN,
    ACTIONS(4691), 1,
      anon_sym_COMMA,
    STATE(1916), 1,
      aux_sym_ordered_field_declaration_list_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60786] = 4,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4693), 1,
      anon_sym_SEMI,
    ACTIONS(4695), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60800] = 4,
    ACTIONS(4433), 1,
      anon_sym_COMMA,
    ACTIONS(4697), 1,
      anon_sym_PIPE,
    STATE(2091), 1,
      aux_sym_closure_parameters_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60814] = 4,
    ACTIONS(4699), 1,
      sym_identifier,
    ACTIONS(4701), 1,
      anon_sym_ref,
    ACTIONS(4703), 1,
      sym_mutable_specifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60828] = 3,
    ACTIONS(4707), 1,
      anon_sym_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4705), 2,
      anon_sym_RBRACE,
      anon_sym_COMMA,
  [60840] = 4,
    ACTIONS(3806), 1,
      anon_sym_LT,
    ACTIONS(4709), 1,
      anon_sym_EQ,
    STATE(2543), 1,
      sym_type_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60854] = 4,
    ACTIONS(4069), 1,
      anon_sym_LBRACE,
    ACTIONS(4711), 1,
      anon_sym_SEMI,
    STATE(1074), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60868] = 4,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(4713), 1,
      anon_sym_SEMI,
    STATE(2549), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60882] = 4,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    ACTIONS(4715), 1,
      anon_sym_SEMI,
    STATE(299), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60896] = 4,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    ACTIONS(4717), 1,
      anon_sym_SEMI,
    STATE(278), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60910] = 3,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4719), 2,
      anon_sym_COMMA,
      anon_sym_GT,
  [60922] = 4,
    ACTIONS(4721), 1,
      anon_sym_COMMA,
    ACTIONS(4724), 1,
      anon_sym_GT,
    STATE(1929), 1,
      aux_sym_for_lifetimes_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60936] = 4,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    ACTIONS(4726), 1,
      anon_sym_GT,
    STATE(2265), 1,
      sym_lifetime,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60950] = 4,
    ACTIONS(4728), 1,
      anon_sym_RBRACE,
    ACTIONS(4730), 1,
      anon_sym_COMMA,
    STATE(2055), 1,
      aux_sym_struct_pattern_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60964] = 4,
    ACTIONS(4732), 1,
      anon_sym_COMMA,
    ACTIONS(4735), 1,
      anon_sym_GT,
    STATE(1932), 1,
      aux_sym_type_arguments_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [60978] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4737), 3,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COMMA,
  [60988] = 3,
    ACTIONS(2449), 1,
      anon_sym_PLUS,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4735), 2,
      anon_sym_COMMA,
      anon_sym_GT,
  [61000] = 3,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4735), 2,
      anon_sym_COMMA,
      anon_sym_GT,
  [61012] = 3,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4739), 2,
      anon_sym_COMMA,
      anon_sym_GT,
  [61024] = 4,
    ACTIONS(4741), 1,
      anon_sym_RBRACE,
    ACTIONS(4743), 1,
      anon_sym_COMMA,
    STATE(1937), 1,
      aux_sym_struct_pattern_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61038] = 4,
    ACTIONS(4069), 1,
      anon_sym_LBRACE,
    ACTIONS(4746), 1,
      anon_sym_SEMI,
    STATE(1058), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61052] = 3,
    ACTIONS(4316), 1,
      anon_sym_PIPE,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4748), 2,
      anon_sym_RBRACE,
      anon_sym_COMMA,
  [61064] = 3,
    ACTIONS(4752), 1,
      anon_sym_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4750), 2,
      anon_sym_RBRACE,
      anon_sym_COMMA,
  [61076] = 4,
    ACTIONS(4081), 1,
      anon_sym_LBRACE,
    ACTIONS(4754), 1,
      anon_sym_SEMI,
    STATE(267), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61090] = 4,
    ACTIONS(2590), 1,
      anon_sym_LBRACE,
    ACTIONS(4756), 1,
      anon_sym_COLON_COLON,
    STATE(1082), 1,
      sym_field_initializer_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61104] = 4,
    ACTIONS(760), 1,
      anon_sym_RPAREN,
    ACTIONS(4758), 1,
      anon_sym_COMMA,
    STATE(1946), 1,
      aux_sym_parameters_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61118] = 4,
    ACTIONS(4069), 1,
      anon_sym_LBRACE,
    ACTIONS(4760), 1,
      anon_sym_SEMI,
    STATE(1048), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61132] = 3,
    ACTIONS(3163), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4273), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [61144] = 4,
    ACTIONS(4478), 1,
      anon_sym_RPAREN,
    ACTIONS(4762), 1,
      anon_sym_COMMA,
    STATE(1946), 1,
      aux_sym_parameters_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61158] = 3,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4478), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [61170] = 4,
    ACTIONS(4427), 1,
      anon_sym_COMMA,
    ACTIONS(4429), 1,
      anon_sym_GT,
    STATE(2047), 1,
      aux_sym_type_arguments_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61184] = 4,
    ACTIONS(4069), 1,
      anon_sym_LBRACE,
    ACTIONS(4765), 1,
      anon_sym_SEMI,
    STATE(920), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61198] = 4,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4767), 1,
      anon_sym_SEMI,
    ACTIONS(4769), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61212] = 4,
    ACTIONS(4771), 1,
      anon_sym_RBRACE,
    ACTIONS(4773), 1,
      anon_sym_COMMA,
    STATE(2042), 1,
      aux_sym_use_list_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61226] = 4,
    ACTIONS(4069), 1,
      anon_sym_LBRACE,
    ACTIONS(4775), 1,
      anon_sym_SEMI,
    STATE(943), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61240] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4492), 3,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_COMMA,
  [61250] = 4,
    ACTIONS(4777), 1,
      anon_sym_RBRACE,
    ACTIONS(4779), 1,
      anon_sym_COMMA,
    STATE(2052), 1,
      aux_sym_struct_pattern_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61264] = 4,
    ACTIONS(632), 1,
      anon_sym_LBRACE,
    ACTIONS(4781), 1,
      anon_sym_move,
    STATE(223), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61278] = 4,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    ACTIONS(4783), 1,
      anon_sym_SEMI,
    STATE(1026), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61292] = 4,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    ACTIONS(4785), 1,
      anon_sym_SEMI,
    STATE(1044), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61306] = 4,
    ACTIONS(3844), 1,
      anon_sym_RBRACE,
    ACTIONS(4787), 1,
      anon_sym_COMMA,
    STATE(1960), 1,
      aux_sym_enum_variant_list_repeat2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61320] = 3,
    ACTIONS(4791), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4789), 2,
      anon_sym_RBRACE,
      anon_sym_COMMA,
  [61332] = 4,
    ACTIONS(4793), 1,
      anon_sym_RBRACE,
    ACTIONS(4795), 1,
      anon_sym_COMMA,
    STATE(1960), 1,
      aux_sym_enum_variant_list_repeat2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61346] = 4,
    ACTIONS(2590), 1,
      anon_sym_LBRACE,
    ACTIONS(4798), 1,
      anon_sym_COLON_COLON,
    STATE(1082), 1,
      sym_field_initializer_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61360] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3347), 3,
      anon_sym_LBRACE,
      anon_sym_EQ_GT,
      anon_sym_AMP_AMP,
  [61370] = 4,
    ACTIONS(2512), 1,
      anon_sym_RPAREN,
    ACTIONS(4800), 1,
      anon_sym_COMMA,
    STATE(2050), 1,
      aux_sym_tuple_type_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61384] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4802), 3,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COMMA,
  [61394] = 3,
    ACTIONS(4227), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3443), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
  [61406] = 4,
    ACTIONS(4804), 1,
      anon_sym_COMMA,
    ACTIONS(4806), 1,
      anon_sym_GT,
    STATE(2046), 1,
      aux_sym_for_lifetimes_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61420] = 4,
    ACTIONS(4069), 1,
      anon_sym_LBRACE,
    ACTIONS(4808), 1,
      anon_sym_SEMI,
    STATE(1105), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61434] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4810), 3,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
  [61444] = 4,
    ACTIONS(3882), 1,
      anon_sym_GT,
    ACTIONS(4812), 1,
      anon_sym_COMMA,
    STATE(2040), 1,
      aux_sym_type_parameters_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61458] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4814), 3,
      anon_sym_EQ,
      anon_sym_COMMA,
      anon_sym_GT,
  [61468] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4816), 3,
      anon_sym_SEMI,
      anon_sym_RPAREN,
      anon_sym_RBRACE,
  [61478] = 4,
    ACTIONS(3271), 1,
      anon_sym_RBRACK,
    ACTIONS(4818), 1,
      anon_sym_COMMA,
    STATE(1972), 1,
      aux_sym_array_expression_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61492] = 4,
    ACTIONS(4069), 1,
      anon_sym_LBRACE,
    ACTIONS(4821), 1,
      anon_sym_SEMI,
    STATE(1126), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61506] = 4,
    ACTIONS(624), 1,
      anon_sym_RBRACK,
    ACTIONS(3149), 1,
      anon_sym_COMMA,
    STATE(1972), 1,
      aux_sym_array_expression_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61520] = 3,
    ACTIONS(4823), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4825), 2,
      anon_sym_default,
      anon_sym_union,
  [61532] = 4,
    ACTIONS(398), 1,
      anon_sym_RPAREN,
    ACTIONS(4827), 1,
      anon_sym_COMMA,
    STATE(1977), 1,
      aux_sym_arguments_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61546] = 4,
    ACTIONS(3285), 1,
      anon_sym_RPAREN,
    ACTIONS(4829), 1,
      anon_sym_COMMA,
    STATE(1977), 1,
      aux_sym_arguments_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61560] = 4,
    ACTIONS(4069), 1,
      anon_sym_LBRACE,
    ACTIONS(4832), 1,
      anon_sym_SEMI,
    STATE(1104), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61574] = 4,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    ACTIONS(4834), 1,
      anon_sym_SEMI,
    STATE(1100), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61588] = 4,
    ACTIONS(4433), 1,
      anon_sym_COMMA,
    ACTIONS(4836), 1,
      anon_sym_PIPE,
    STATE(1920), 1,
      aux_sym_closure_parameters_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61602] = 4,
    ACTIONS(4316), 1,
      anon_sym_PIPE,
    ACTIONS(4838), 1,
      anon_sym_EQ_GT,
    ACTIONS(4840), 1,
      anon_sym_if,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61616] = 4,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    ACTIONS(4842), 1,
      anon_sym_SEMI,
    STATE(983), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61630] = 4,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    ACTIONS(4844), 1,
      anon_sym_SEMI,
    STATE(1093), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61644] = 4,
    ACTIONS(3876), 1,
      anon_sym_RBRACE,
    ACTIONS(4846), 1,
      anon_sym_COMMA,
    STATE(1913), 1,
      aux_sym_field_declaration_list_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61658] = 4,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4848), 1,
      anon_sym_SEMI,
    ACTIONS(4850), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61672] = 4,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    ACTIONS(4852), 1,
      anon_sym_GT,
    STATE(2265), 1,
      sym_lifetime,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61686] = 4,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(4854), 1,
      anon_sym_SEMI,
    STATE(2412), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61700] = 3,
    ACTIONS(4207), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3443), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
  [61712] = 4,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    ACTIONS(4856), 1,
      anon_sym_SEMI,
    STATE(1078), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61726] = 4,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    ACTIONS(4858), 1,
      anon_sym_SEMI,
    STATE(1076), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61740] = 4,
    ACTIONS(4069), 1,
      anon_sym_LBRACE,
    ACTIONS(4860), 1,
      anon_sym_SEMI,
    STATE(1066), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61754] = 4,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4862), 1,
      anon_sym_SEMI,
    ACTIONS(4864), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61768] = 4,
    ACTIONS(4081), 1,
      anon_sym_LBRACE,
    ACTIONS(4866), 1,
      anon_sym_SEMI,
    STATE(285), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61782] = 4,
    ACTIONS(2560), 1,
      anon_sym_LT2,
    ACTIONS(4868), 1,
      sym_identifier,
    STATE(1191), 1,
      sym_type_arguments,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61796] = 4,
    ACTIONS(764), 1,
      anon_sym_RPAREN,
    ACTIONS(4870), 1,
      anon_sym_COMMA,
    STATE(1946), 1,
      aux_sym_parameters_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61810] = 4,
    ACTIONS(4872), 1,
      anon_sym_RBRACE,
    ACTIONS(4874), 1,
      anon_sym_COMMA,
    STATE(1996), 1,
      aux_sym_field_initializer_list_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61824] = 4,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    ACTIONS(4877), 1,
      anon_sym_SEMI,
    STATE(252), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61838] = 4,
    ACTIONS(4879), 1,
      anon_sym_for,
    ACTIONS(4881), 1,
      anon_sym_loop,
    ACTIONS(4883), 1,
      anon_sym_while,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61852] = 4,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4885), 1,
      anon_sym_SEMI,
    ACTIONS(4887), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61866] = 4,
    ACTIONS(2560), 1,
      anon_sym_LT2,
    ACTIONS(4868), 1,
      sym_identifier,
    STATE(1190), 1,
      sym_type_arguments,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61880] = 3,
    ACTIONS(4178), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3443), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
  [61892] = 4,
    ACTIONS(4889), 1,
      anon_sym_RBRACE,
    ACTIONS(4891), 1,
      anon_sym_COMMA,
    STATE(2002), 1,
      aux_sym_use_list_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61906] = 4,
    ACTIONS(4894), 1,
      anon_sym_for,
    ACTIONS(4896), 1,
      anon_sym_loop,
    ACTIONS(4898), 1,
      anon_sym_while,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61920] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4900), 3,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COMMA,
  [61930] = 4,
    ACTIONS(3822), 1,
      anon_sym_RBRACE,
    ACTIONS(4902), 1,
      anon_sym_COMMA,
    STATE(1960), 1,
      aux_sym_enum_variant_list_repeat2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61944] = 4,
    ACTIONS(4116), 1,
      anon_sym_COMMA,
    ACTIONS(4118), 1,
      anon_sym_GT,
    STATE(1855), 1,
      aux_sym_type_parameters_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61958] = 4,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    ACTIONS(4904), 1,
      anon_sym_SEMI,
    STATE(300), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61972] = 4,
    ACTIONS(4906), 1,
      anon_sym_RBRACE,
    ACTIONS(4908), 1,
      anon_sym_COMMA,
    STATE(2022), 1,
      aux_sym_field_declaration_list_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [61986] = 4,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4910), 1,
      anon_sym_SEMI,
    ACTIONS(4912), 1,
      anon_sym_RBRACK,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62000] = 4,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(4914), 1,
      anon_sym_SEMI,
    STATE(2493), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62014] = 3,
    ACTIONS(4160), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(3443), 2,
      anon_sym_DOT_DOT_DOT,
      anon_sym_DOT_DOT_EQ,
  [62026] = 4,
    ACTIONS(3824), 1,
      anon_sym_RBRACE,
    ACTIONS(4916), 1,
      anon_sym_COMMA,
    STATE(1912), 1,
      aux_sym_field_declaration_list_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62040] = 4,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(3746), 1,
      anon_sym_LBRACE,
    STATE(1347), 1,
      sym_type_arguments,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62054] = 4,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(4918), 1,
      sym_identifier,
    STATE(2372), 1,
      sym_type_arguments,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62068] = 4,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(4920), 1,
      anon_sym_SEMI,
    STATE(2384), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62082] = 4,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(4922), 1,
      sym_identifier,
    STATE(2374), 1,
      sym_type_arguments,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62096] = 4,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4924), 1,
      anon_sym_SEMI,
    ACTIONS(4926), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62110] = 4,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(4918), 1,
      sym_identifier,
    STATE(2374), 1,
      sym_type_arguments,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62124] = 4,
    ACTIONS(4069), 1,
      anon_sym_LBRACE,
    ACTIONS(4928), 1,
      anon_sym_SEMI,
    STATE(1013), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62138] = 4,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    ACTIONS(4930), 1,
      anon_sym_SEMI,
    STATE(471), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62152] = 4,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4932), 1,
      anon_sym_SEMI,
    ACTIONS(4934), 1,
      anon_sym_RBRACK,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62166] = 4,
    ACTIONS(3824), 1,
      anon_sym_RBRACE,
    ACTIONS(4916), 1,
      anon_sym_COMMA,
    STATE(1913), 1,
      aux_sym_field_declaration_list_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62180] = 4,
    ACTIONS(4263), 1,
      anon_sym_COMMA,
    ACTIONS(4265), 1,
      anon_sym_GT,
    STATE(2094), 1,
      aux_sym_type_parameters_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62194] = 4,
    ACTIONS(4936), 1,
      anon_sym_RPAREN,
    ACTIONS(4938), 1,
      anon_sym_COMMA,
    STATE(1916), 1,
      aux_sym_ordered_field_declaration_list_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62208] = 4,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4940), 1,
      anon_sym_SEMI,
    ACTIONS(4942), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62222] = 4,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    ACTIONS(4944), 1,
      anon_sym_SEMI,
    STATE(995), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62236] = 4,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(4922), 1,
      sym_identifier,
    STATE(2372), 1,
      sym_type_arguments,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62250] = 4,
    ACTIONS(3832), 1,
      anon_sym_RBRACE,
    ACTIONS(4946), 1,
      anon_sym_COMMA,
    STATE(1984), 1,
      aux_sym_field_declaration_list_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62264] = 4,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    ACTIONS(4948), 1,
      anon_sym_SEMI,
    STATE(926), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62278] = 4,
    ACTIONS(3832), 1,
      anon_sym_RBRACE,
    ACTIONS(4946), 1,
      anon_sym_COMMA,
    STATE(1913), 1,
      aux_sym_field_declaration_list_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62292] = 4,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(4950), 1,
      anon_sym_SEMI,
    ACTIONS(4952), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62306] = 4,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(4561), 1,
      sym_identifier,
    STATE(2372), 1,
      sym_type_arguments,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62320] = 4,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    ACTIONS(4954), 1,
      anon_sym_SEMI,
    STATE(342), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62334] = 4,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    ACTIONS(4956), 1,
      anon_sym_SEMI,
    STATE(428), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62348] = 4,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    ACTIONS(4958), 1,
      anon_sym_SEMI,
    STATE(348), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62362] = 4,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(4960), 1,
      anon_sym_SEMI,
    STATE(2450), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62376] = 4,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    ACTIONS(4962), 1,
      anon_sym_SEMI,
    STATE(921), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62390] = 3,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4964), 2,
      anon_sym_COMMA,
      anon_sym_GT,
  [62402] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4966), 3,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COMMA,
  [62412] = 4,
    ACTIONS(4387), 1,
      anon_sym_GT,
    ACTIONS(4968), 1,
      anon_sym_COMMA,
    STATE(2040), 1,
      aux_sym_type_parameters_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62426] = 3,
    ACTIONS(4361), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4387), 2,
      anon_sym_COMMA,
      anon_sym_GT,
  [62438] = 4,
    ACTIONS(3383), 1,
      anon_sym_RBRACE,
    ACTIONS(4971), 1,
      anon_sym_COMMA,
    STATE(2002), 1,
      aux_sym_use_list_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62452] = 4,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    ACTIONS(4973), 1,
      anon_sym_SEMI,
    STATE(929), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62466] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4975), 3,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COMMA,
  [62476] = 3,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4977), 2,
      anon_sym_COMMA,
      anon_sym_GT,
  [62488] = 4,
    ACTIONS(4852), 1,
      anon_sym_GT,
    ACTIONS(4979), 1,
      anon_sym_COMMA,
    STATE(1929), 1,
      aux_sym_for_lifetimes_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62502] = 4,
    ACTIONS(868), 1,
      anon_sym_GT,
    ACTIONS(4981), 1,
      anon_sym_COMMA,
    STATE(1932), 1,
      aux_sym_type_arguments_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62516] = 3,
    ACTIONS(4983), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4985), 2,
      anon_sym_default,
      anon_sym_union,
  [62528] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4987), 3,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COMMA,
  [62538] = 4,
    ACTIONS(4989), 1,
      anon_sym_RPAREN,
    ACTIONS(4991), 1,
      anon_sym_COMMA,
    STATE(2050), 1,
      aux_sym_tuple_type_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62552] = 3,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4989), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [62564] = 4,
    ACTIONS(4057), 1,
      anon_sym_RBRACE,
    ACTIONS(4994), 1,
      anon_sym_COMMA,
    STATE(1937), 1,
      aux_sym_struct_pattern_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62578] = 4,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    ACTIONS(4996), 1,
      anon_sym_SEMI,
    STATE(888), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62592] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4998), 3,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COMMA,
  [62602] = 4,
    ACTIONS(4049), 1,
      anon_sym_RBRACE,
    ACTIONS(5000), 1,
      anon_sym_COMMA,
    STATE(1937), 1,
      aux_sym_struct_pattern_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62616] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(5002), 3,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COMMA,
  [62626] = 3,
    ACTIONS(5006), 1,
      anon_sym_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(5004), 2,
      anon_sym_RBRACE,
      anon_sym_COMMA,
  [62638] = 4,
    ACTIONS(880), 1,
      anon_sym_GT,
    ACTIONS(5008), 1,
      anon_sym_COMMA,
    STATE(1932), 1,
      aux_sym_type_arguments_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62652] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(5010), 3,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COMMA,
  [62662] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(5012), 3,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COMMA,
  [62672] = 4,
    ACTIONS(2401), 1,
      anon_sym_RPAREN,
    ACTIONS(5014), 1,
      anon_sym_COMMA,
    STATE(1806), 1,
      aux_sym_tuple_pattern_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62686] = 4,
    ACTIONS(4081), 1,
      anon_sym_LBRACE,
    ACTIONS(5016), 1,
      anon_sym_SEMI,
    STATE(364), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62700] = 4,
    ACTIONS(4069), 1,
      anon_sym_LBRACE,
    ACTIONS(5018), 1,
      anon_sym_SEMI,
    STATE(992), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62714] = 4,
    ACTIONS(762), 1,
      anon_sym_RPAREN,
    ACTIONS(4292), 1,
      anon_sym_COMMA,
    STATE(1995), 1,
      aux_sym_parameters_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62728] = 4,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(4106), 1,
      sym_identifier,
    STATE(2372), 1,
      sym_type_arguments,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62742] = 4,
    ACTIONS(762), 1,
      anon_sym_RPAREN,
    ACTIONS(4292), 1,
      anon_sym_COMMA,
    STATE(1946), 1,
      aux_sym_parameters_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62756] = 4,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    ACTIONS(5020), 1,
      anon_sym_SEMI,
    STATE(368), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62770] = 4,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(5022), 1,
      anon_sym_SEMI,
    ACTIONS(5024), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62784] = 4,
    ACTIONS(766), 1,
      anon_sym_RPAREN,
    ACTIONS(4281), 1,
      anon_sym_COMMA,
    STATE(1943), 1,
      aux_sym_parameters_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62798] = 4,
    ACTIONS(766), 1,
      anon_sym_RPAREN,
    ACTIONS(4281), 1,
      anon_sym_COMMA,
    STATE(1946), 1,
      aux_sym_parameters_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62812] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(5026), 3,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_COMMA,
  [62822] = 4,
    ACTIONS(5028), 1,
      anon_sym_for,
    ACTIONS(5030), 1,
      anon_sym_loop,
    ACTIONS(5032), 1,
      anon_sym_while,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62836] = 4,
    ACTIONS(5034), 1,
      anon_sym_RBRACE,
    ACTIONS(5036), 1,
      anon_sym_COMMA,
    STATE(2084), 1,
      aux_sym_enum_variant_list_repeat2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62850] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(5038), 3,
      anon_sym_SEMI,
      anon_sym_LBRACE,
      anon_sym_COMMA,
  [62860] = 4,
    ACTIONS(3842), 1,
      anon_sym_RBRACE,
    ACTIONS(5040), 1,
      anon_sym_COMMA,
    STATE(1958), 1,
      aux_sym_enum_variant_list_repeat2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62874] = 4,
    ACTIONS(3445), 1,
      anon_sym_LT2,
    ACTIONS(4106), 1,
      sym_identifier,
    STATE(2374), 1,
      sym_type_arguments,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62888] = 4,
    ACTIONS(3842), 1,
      anon_sym_RBRACE,
    ACTIONS(5040), 1,
      anon_sym_COMMA,
    STATE(1960), 1,
      aux_sym_enum_variant_list_repeat2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62902] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(5042), 3,
      anon_sym_SEMI,
      anon_sym_RBRACE,
      anon_sym_COMMA,
  [62912] = 3,
    ACTIONS(5046), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(5044), 2,
      anon_sym_RBRACE,
      anon_sym_COMMA,
  [62924] = 4,
    ACTIONS(4251), 1,
      anon_sym_RPAREN,
    ACTIONS(4253), 1,
      anon_sym_COMMA,
    STATE(2066), 1,
      aux_sym_parameters_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62938] = 4,
    ACTIONS(2560), 1,
      anon_sym_LT2,
    ACTIONS(5048), 1,
      sym_identifier,
    STATE(793), 1,
      sym_type_arguments,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62952] = 4,
    ACTIONS(3814), 1,
      anon_sym_RBRACE,
    ACTIONS(5050), 1,
      anon_sym_COMMA,
    STATE(2005), 1,
      aux_sym_enum_variant_list_repeat2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62966] = 4,
    ACTIONS(4069), 1,
      anon_sym_LBRACE,
    ACTIONS(5052), 1,
      anon_sym_SEMI,
    STATE(873), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62980] = 4,
    ACTIONS(3814), 1,
      anon_sym_RBRACE,
    ACTIONS(5050), 1,
      anon_sym_COMMA,
    STATE(1960), 1,
      aux_sym_enum_variant_list_repeat2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [62994] = 4,
    ACTIONS(5054), 1,
      sym_identifier,
    ACTIONS(5056), 1,
      anon_sym_await,
    ACTIONS(5058), 1,
      sym_integer_literal,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63008] = 4,
    ACTIONS(4407), 1,
      anon_sym_COMMA,
    ACTIONS(4409), 1,
      anon_sym_GT,
    STATE(2058), 1,
      aux_sym_type_arguments_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63022] = 4,
    ACTIONS(5060), 1,
      sym_identifier,
    ACTIONS(5062), 1,
      anon_sym_ref,
    ACTIONS(5064), 1,
      sym_mutable_specifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63036] = 3,
    ACTIONS(4339), 1,
      anon_sym_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(5066), 2,
      anon_sym_COMMA,
      anon_sym_PIPE,
  [63048] = 4,
    ACTIONS(3249), 1,
      anon_sym_LBRACE,
    ACTIONS(5068), 1,
      sym_identifier,
    STATE(1933), 1,
      sym_use_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63062] = 4,
    ACTIONS(2518), 1,
      anon_sym_RPAREN,
    ACTIONS(5070), 1,
      anon_sym_COMMA,
    STATE(2050), 1,
      aux_sym_tuple_type_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63076] = 4,
    ACTIONS(5066), 1,
      anon_sym_PIPE,
    ACTIONS(5072), 1,
      anon_sym_COMMA,
    STATE(2091), 1,
      aux_sym_closure_parameters_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63090] = 4,
    ACTIONS(3872), 1,
      anon_sym_GT,
    ACTIONS(5075), 1,
      anon_sym_COMMA,
    STATE(2040), 1,
      aux_sym_type_parameters_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63104] = 4,
    ACTIONS(284), 1,
      anon_sym_LBRACE,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    STATE(1009), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63118] = 4,
    ACTIONS(3870), 1,
      anon_sym_GT,
    ACTIONS(5077), 1,
      anon_sym_COMMA,
    STATE(2040), 1,
      aux_sym_type_parameters_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63132] = 4,
    ACTIONS(3846), 1,
      anon_sym_RBRACE,
    ACTIONS(5079), 1,
      anon_sym_COMMA,
    STATE(1996), 1,
      aux_sym_field_initializer_list_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63146] = 4,
    ACTIONS(396), 1,
      anon_sym_RPAREN,
    ACTIONS(3185), 1,
      anon_sym_COMMA,
    STATE(1977), 1,
      aux_sym_arguments_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63160] = 4,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    ACTIONS(5081), 1,
      anon_sym_SEMI,
    STATE(860), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63174] = 4,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(5083), 1,
      anon_sym_SEMI,
    STATE(2417), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63188] = 3,
    ACTIONS(5087), 1,
      anon_sym_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(5085), 2,
      anon_sym_RBRACE,
      anon_sym_COMMA,
  [63200] = 4,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    ACTIONS(5089), 1,
      anon_sym_SEMI,
    STATE(835), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63214] = 4,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(5091), 1,
      anon_sym_SEMI,
    STATE(2507), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63228] = 4,
    ACTIONS(2560), 1,
      anon_sym_LT2,
    ACTIONS(5048), 1,
      sym_identifier,
    STATE(796), 1,
      sym_type_arguments,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63242] = 4,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(5093), 1,
      anon_sym_as,
    ACTIONS(5095), 1,
      anon_sym_GT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63256] = 4,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(5097), 1,
      anon_sym_SEMI,
    ACTIONS(5099), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63270] = 4,
    ACTIONS(3804), 1,
      anon_sym_where,
    ACTIONS(5101), 1,
      anon_sym_SEMI,
    STATE(2423), 1,
      sym_where_clause,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63284] = 4,
    ACTIONS(5103), 1,
      anon_sym_RBRACE,
    ACTIONS(5105), 1,
      anon_sym_COMMA,
    STATE(2030), 1,
      aux_sym_field_declaration_list_repeat1,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63298] = 3,
    ACTIONS(5107), 1,
      anon_sym_EQ_GT,
    ACTIONS(5109), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63309] = 3,
    ACTIONS(3578), 1,
      anon_sym_COLON_COLON,
    ACTIONS(5111), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63320] = 3,
    ACTIONS(4059), 1,
      anon_sym_COLON_COLON,
    ACTIONS(5113), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63331] = 3,
    ACTIONS(4047), 1,
      anon_sym_COLON_COLON,
    ACTIONS(5113), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63342] = 3,
    ACTIONS(4041), 1,
      anon_sym_COLON_COLON,
    ACTIONS(5113), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63353] = 3,
    ACTIONS(3449), 1,
      anon_sym_LPAREN,
    STATE(1645), 1,
      sym_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63364] = 3,
    ACTIONS(3802), 1,
      anon_sym_LBRACE,
    STATE(848), 1,
      sym_field_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63375] = 3,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    STATE(847), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63386] = 3,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    STATE(846), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63397] = 3,
    ACTIONS(3828), 1,
      anon_sym_LBRACE,
    STATE(407), 1,
      sym_field_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63408] = 3,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(5115), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63419] = 3,
    ACTIONS(5107), 1,
      anon_sym_LBRACE,
    ACTIONS(5117), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63430] = 3,
    ACTIONS(3802), 1,
      anon_sym_LBRACE,
    STATE(840), 1,
      sym_field_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63441] = 3,
    ACTIONS(5119), 1,
      anon_sym_SEMI,
    ACTIONS(5121), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63452] = 3,
    ACTIONS(5119), 1,
      anon_sym_SEMI,
    ACTIONS(5123), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63463] = 3,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    STATE(834), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63474] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(5125), 2,
      sym_identifier,
      sym_metavariable,
  [63483] = 3,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    STATE(443), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63494] = 3,
    ACTIONS(284), 1,
      anon_sym_LBRACE,
    STATE(2429), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63505] = 3,
    ACTIONS(3449), 1,
      anon_sym_LPAREN,
    STATE(1662), 1,
      sym_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63516] = 3,
    ACTIONS(4112), 1,
      anon_sym_LBRACE,
    STATE(828), 1,
      sym_enum_variant_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63527] = 3,
    ACTIONS(3828), 1,
      anon_sym_LBRACE,
    STATE(404), 1,
      sym_field_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63538] = 3,
    ACTIONS(5127), 1,
      anon_sym_SEMI,
    ACTIONS(5129), 1,
      anon_sym_as,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63549] = 3,
    ACTIONS(3802), 1,
      anon_sym_LBRACE,
    STATE(825), 1,
      sym_field_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63560] = 3,
    ACTIONS(2552), 1,
      anon_sym_LPAREN,
    STATE(798), 1,
      sym_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63571] = 3,
    ACTIONS(3802), 1,
      anon_sym_LBRACE,
    STATE(822), 1,
      sym_field_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63582] = 3,
    ACTIONS(4316), 1,
      anon_sym_PIPE,
    ACTIONS(5131), 1,
      anon_sym_in,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63593] = 3,
    ACTIONS(15), 1,
      anon_sym_LBRACE,
    STATE(62), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63604] = 3,
    ACTIONS(5133), 1,
      anon_sym_SEMI,
    ACTIONS(5135), 1,
      anon_sym_as,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63615] = 3,
    ACTIONS(284), 1,
      anon_sym_LBRACE,
    STATE(818), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63626] = 3,
    ACTIONS(15), 1,
      anon_sym_LBRACE,
    STATE(72), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63637] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(5066), 2,
      anon_sym_COMMA,
      anon_sym_PIPE,
  [63646] = 3,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(5137), 1,
      anon_sym_GT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63657] = 3,
    ACTIONS(284), 1,
      anon_sym_LBRACE,
    STATE(2392), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63668] = 3,
    ACTIONS(5119), 1,
      anon_sym_SEMI,
    ACTIONS(5139), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63679] = 3,
    ACTIONS(3511), 1,
      anon_sym_BANG,
    ACTIONS(3546), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63690] = 3,
    ACTIONS(5119), 1,
      anon_sym_SEMI,
    ACTIONS(5141), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63701] = 3,
    ACTIONS(5119), 1,
      anon_sym_SEMI,
    ACTIONS(5143), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63712] = 3,
    ACTIONS(5119), 1,
      anon_sym_SEMI,
    ACTIONS(5145), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63723] = 3,
    ACTIONS(284), 1,
      anon_sym_LBRACE,
    STATE(2389), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63734] = 3,
    ACTIONS(5119), 1,
      anon_sym_SEMI,
    ACTIONS(5147), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63745] = 3,
    ACTIONS(5119), 1,
      anon_sym_SEMI,
    ACTIONS(5149), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63756] = 3,
    ACTIONS(4112), 1,
      anon_sym_LBRACE,
    STATE(877), 1,
      sym_enum_variant_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63767] = 3,
    ACTIONS(5060), 1,
      sym_identifier,
    ACTIONS(5064), 1,
      sym_mutable_specifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63778] = 3,
    ACTIONS(4005), 1,
      anon_sym_RBRACE,
    ACTIONS(5119), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63789] = 3,
    ACTIONS(2560), 1,
      anon_sym_LT2,
    STATE(891), 1,
      sym_type_arguments,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63800] = 3,
    ACTIONS(3449), 1,
      anon_sym_LPAREN,
    STATE(1664), 1,
      sym_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63811] = 3,
    ACTIONS(3987), 1,
      anon_sym_RPAREN,
    ACTIONS(5119), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63822] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4174), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [63831] = 3,
    ACTIONS(4007), 1,
      anon_sym_RBRACE,
    ACTIONS(5119), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63842] = 3,
    ACTIONS(3439), 1,
      anon_sym_BANG,
    ACTIONS(5151), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63853] = 3,
    ACTIONS(284), 1,
      anon_sym_LBRACE,
    STATE(2380), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63864] = 3,
    ACTIONS(284), 1,
      anon_sym_LBRACE,
    STATE(883), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63875] = 3,
    ACTIONS(284), 1,
      anon_sym_LBRACE,
    STATE(2325), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63886] = 3,
    ACTIONS(4199), 1,
      anon_sym_LBRACE,
    STATE(413), 1,
      sym_enum_variant_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63897] = 3,
    ACTIONS(5153), 1,
      anon_sym_SEMI,
    ACTIONS(5155), 1,
      anon_sym_as,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63908] = 3,
    ACTIONS(3802), 1,
      anon_sym_LBRACE,
    STATE(901), 1,
      sym_field_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63919] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4387), 2,
      anon_sym_COMMA,
      anon_sym_GT,
  [63928] = 3,
    ACTIONS(3449), 1,
      anon_sym_LPAREN,
    STATE(1357), 1,
      sym_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63939] = 3,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    STATE(908), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63950] = 3,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    STATE(426), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63961] = 3,
    ACTIONS(3802), 1,
      anon_sym_LBRACE,
    STATE(911), 1,
      sym_field_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63972] = 3,
    ACTIONS(4015), 1,
      anon_sym_RPAREN,
    ACTIONS(5119), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63983] = 3,
    ACTIONS(4199), 1,
      anon_sym_LBRACE,
    STATE(395), 1,
      sym_enum_variant_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [63994] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(5157), 2,
      anon_sym_const,
      sym_mutable_specifier,
  [64003] = 3,
    ACTIONS(3828), 1,
      anon_sym_LBRACE,
    STATE(435), 1,
      sym_field_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64014] = 3,
    ACTIONS(284), 1,
      anon_sym_LBRACE,
    STATE(2538), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64025] = 3,
    ACTIONS(284), 1,
      anon_sym_LBRACE,
    STATE(2542), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64036] = 3,
    ACTIONS(3449), 1,
      anon_sym_LPAREN,
    STATE(1672), 1,
      sym_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64047] = 3,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    STATE(986), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64058] = 3,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(5159), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64069] = 3,
    ACTIONS(284), 1,
      anon_sym_LBRACE,
    STATE(2373), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64080] = 3,
    ACTIONS(2552), 1,
      anon_sym_LPAREN,
    STATE(779), 1,
      sym_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64091] = 3,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(5161), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64102] = 3,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    STATE(1000), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64113] = 3,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    STATE(1001), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64124] = 3,
    ACTIONS(15), 1,
      anon_sym_LBRACE,
    STATE(48), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64135] = 3,
    ACTIONS(284), 1,
      anon_sym_LBRACE,
    STATE(2455), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64146] = 3,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    STATE(444), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64157] = 3,
    ACTIONS(284), 1,
      anon_sym_LBRACE,
    STATE(1038), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64168] = 3,
    ACTIONS(4316), 1,
      anon_sym_PIPE,
    ACTIONS(5163), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64179] = 3,
    ACTIONS(4112), 1,
      anon_sym_LBRACE,
    STATE(1016), 1,
      sym_enum_variant_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64190] = 3,
    ACTIONS(284), 1,
      anon_sym_LBRACE,
    STATE(2370), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64201] = 3,
    ACTIONS(3351), 1,
      anon_sym_LBRACE,
    ACTIONS(5117), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64212] = 3,
    ACTIONS(3802), 1,
      anon_sym_LBRACE,
    STATE(1020), 1,
      sym_field_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64223] = 3,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(5165), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64234] = 3,
    ACTIONS(3802), 1,
      anon_sym_LBRACE,
    STATE(1030), 1,
      sym_field_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64245] = 3,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    STATE(1034), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64256] = 3,
    ACTIONS(5167), 1,
      anon_sym_LT,
    STATE(743), 1,
      sym_type_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64267] = 3,
    ACTIONS(3828), 1,
      anon_sym_LBRACE,
    STATE(450), 1,
      sym_field_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64278] = 3,
    ACTIONS(4316), 1,
      anon_sym_PIPE,
    ACTIONS(5169), 1,
      anon_sym_in,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64289] = 3,
    ACTIONS(4316), 1,
      anon_sym_PIPE,
    ACTIONS(5171), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64300] = 3,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    STATE(308), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64311] = 3,
    ACTIONS(4112), 1,
      anon_sym_LBRACE,
    STATE(964), 1,
      sym_enum_variant_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64322] = 3,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(1966), 1,
      sym_lifetime,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64333] = 3,
    ACTIONS(3449), 1,
      anon_sym_LPAREN,
    STATE(1698), 1,
      sym_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64344] = 3,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(5173), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64355] = 3,
    ACTIONS(3449), 1,
      anon_sym_LPAREN,
    STATE(1352), 1,
      sym_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64366] = 3,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    STATE(294), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64377] = 3,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    STATE(292), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64388] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4889), 2,
      anon_sym_RBRACE,
      anon_sym_COMMA,
  [64397] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(5175), 2,
      sym_float_literal,
      sym_integer_literal,
  [64406] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4872), 2,
      anon_sym_RBRACE,
      anon_sym_COMMA,
  [64415] = 3,
    ACTIONS(284), 1,
      anon_sym_LBRACE,
    STATE(2383), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64426] = 3,
    ACTIONS(15), 1,
      anon_sym_LBRACE,
    STATE(84), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64437] = 3,
    ACTIONS(4041), 1,
      anon_sym_COLON_COLON,
    ACTIONS(5177), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64448] = 3,
    ACTIONS(4199), 1,
      anon_sym_LBRACE,
    STATE(282), 1,
      sym_enum_variant_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64459] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(5179), 2,
      sym_identifier,
      sym_metavariable,
  [64468] = 3,
    ACTIONS(4047), 1,
      anon_sym_COLON_COLON,
    ACTIONS(5177), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64479] = 3,
    ACTIONS(5181), 1,
      anon_sym_BANG,
    ACTIONS(5183), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64490] = 3,
    ACTIONS(284), 1,
      anon_sym_LBRACE,
    STATE(974), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64501] = 3,
    ACTIONS(284), 1,
      anon_sym_LBRACE,
    STATE(1045), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64512] = 3,
    ACTIONS(4059), 1,
      anon_sym_COLON_COLON,
    ACTIONS(5177), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64523] = 3,
    ACTIONS(3578), 1,
      anon_sym_COLON_COLON,
    ACTIONS(5185), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64534] = 3,
    ACTIONS(3828), 1,
      anon_sym_LBRACE,
    STATE(276), 1,
      sym_field_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64545] = 3,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(5187), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64556] = 3,
    ACTIONS(3828), 1,
      anon_sym_LBRACE,
    STATE(269), 1,
      sym_field_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64567] = 3,
    ACTIONS(2688), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3928), 1,
      anon_sym_BANG,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64578] = 3,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    STATE(268), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64589] = 3,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    STATE(1101), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64600] = 3,
    ACTIONS(5189), 1,
      anon_sym_LPAREN,
    ACTIONS(5191), 1,
      anon_sym_LBRACE,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64611] = 3,
    ACTIONS(3828), 1,
      anon_sym_LBRACE,
    STATE(270), 1,
      sym_field_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64622] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(5193), 2,
      anon_sym_const,
      sym_mutable_specifier,
  [64631] = 3,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(5195), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64642] = 3,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    STATE(1113), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64653] = 3,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    STATE(1115), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64664] = 3,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    STATE(273), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64675] = 3,
    ACTIONS(3449), 1,
      anon_sym_LPAREN,
    STATE(1616), 1,
      sym_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64686] = 3,
    ACTIONS(284), 1,
      anon_sym_LBRACE,
    STATE(2314), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64697] = 3,
    ACTIONS(2552), 1,
      anon_sym_LPAREN,
    STATE(802), 1,
      sym_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64708] = 3,
    ACTIONS(3449), 1,
      anon_sym_LPAREN,
    STATE(1363), 1,
      sym_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64719] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(5197), 2,
      sym_identifier,
      sym_metavariable,
  [64728] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(5199), 2,
      anon_sym_const,
      sym_mutable_specifier,
  [64737] = 3,
    ACTIONS(284), 1,
      anon_sym_LBRACE,
    STATE(801), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64748] = 3,
    ACTIONS(15), 1,
      anon_sym_LBRACE,
    STATE(75), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64759] = 3,
    ACTIONS(3828), 1,
      anon_sym_LBRACE,
    STATE(302), 1,
      sym_field_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64770] = 3,
    ACTIONS(3351), 1,
      anon_sym_EQ_GT,
    ACTIONS(5109), 1,
      anon_sym_AMP_AMP,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64781] = 3,
    ACTIONS(5201), 1,
      anon_sym_LBRACK,
    ACTIONS(5203), 1,
      anon_sym_BANG,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64792] = 3,
    ACTIONS(5205), 1,
      anon_sym_LPAREN,
    ACTIONS(5207), 1,
      anon_sym_LBRACE,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64803] = 3,
    ACTIONS(4316), 1,
      anon_sym_PIPE,
    ACTIONS(5209), 1,
      anon_sym_in,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64814] = 3,
    ACTIONS(2590), 1,
      anon_sym_LBRACE,
    STATE(1082), 1,
      sym_field_initializer_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64825] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4793), 2,
      anon_sym_RBRACE,
      anon_sym_COMMA,
  [64834] = 3,
    ACTIONS(284), 1,
      anon_sym_LBRACE,
    STATE(1024), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64845] = 3,
    ACTIONS(632), 1,
      anon_sym_LBRACE,
    STATE(228), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64856] = 3,
    ACTIONS(858), 1,
      anon_sym_LBRACE,
    STATE(1455), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64867] = 3,
    ACTIONS(4031), 1,
      anon_sym_COLON,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64878] = 3,
    ACTIONS(85), 1,
      anon_sym_PIPE,
    STATE(98), 1,
      sym_closure_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64889] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(5211), 2,
      sym_identifier,
      sym_metavariable,
  [64898] = 3,
    ACTIONS(3840), 1,
      anon_sym_LBRACE,
    STATE(937), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64909] = 3,
    ACTIONS(284), 1,
      anon_sym_LBRACE,
    STATE(1039), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64920] = 3,
    ACTIONS(5213), 1,
      anon_sym_LPAREN,
    ACTIONS(5215), 1,
      anon_sym_LBRACE,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64931] = 3,
    ACTIONS(5217), 1,
      anon_sym_LPAREN,
    ACTIONS(5219), 1,
      anon_sym_LBRACE,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64942] = 3,
    ACTIONS(85), 1,
      anon_sym_PIPE,
    STATE(90), 1,
      sym_closure_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64953] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4478), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [64962] = 3,
    ACTIONS(5221), 1,
      anon_sym_LBRACK,
    ACTIONS(5223), 1,
      anon_sym_BANG,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [64973] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4273), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [64982] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4741), 2,
      anon_sym_RBRACE,
      anon_sym_COMMA,
  [64991] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4735), 2,
      anon_sym_COMMA,
      anon_sym_GT,
  [65000] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4724), 2,
      anon_sym_COMMA,
      anon_sym_GT,
  [65009] = 3,
    ACTIONS(284), 1,
      anon_sym_LBRACE,
    STATE(2556), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65020] = 3,
    ACTIONS(632), 1,
      anon_sym_LBRACE,
    STATE(226), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65031] = 3,
    ACTIONS(5225), 1,
      anon_sym_SEMI,
    ACTIONS(5227), 1,
      anon_sym_as,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65042] = 3,
    ACTIONS(632), 1,
      anon_sym_LBRACE,
    STATE(227), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65053] = 3,
    ACTIONS(5229), 1,
      sym_identifier,
    ACTIONS(5231), 1,
      sym_mutable_specifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65064] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(5233), 2,
      sym_identifier,
      sym_metavariable,
  [65073] = 3,
    ACTIONS(632), 1,
      anon_sym_LBRACE,
    STATE(234), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65084] = 3,
    ACTIONS(4316), 1,
      anon_sym_PIPE,
    ACTIONS(5235), 1,
      anon_sym_in,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65095] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4671), 2,
      anon_sym_RBRACE,
      anon_sym_COMMA,
  [65104] = 3,
    ACTIONS(632), 1,
      anon_sym_LBRACE,
    STATE(210), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65115] = 3,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    STATE(355), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65126] = 3,
    ACTIONS(4316), 1,
      anon_sym_PIPE,
    ACTIONS(5237), 1,
      anon_sym_in,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65137] = 3,
    ACTIONS(632), 1,
      anon_sym_LBRACE,
    STATE(217), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65148] = 3,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(5239), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65159] = 3,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    STATE(379), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65170] = 3,
    ACTIONS(4316), 1,
      anon_sym_PIPE,
    ACTIONS(4339), 1,
      anon_sym_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65181] = 3,
    ACTIONS(2792), 1,
      anon_sym_COLON,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65192] = 3,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    STATE(380), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65203] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(5241), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [65212] = 3,
    ACTIONS(4316), 1,
      anon_sym_PIPE,
    ACTIONS(5243), 1,
      anon_sym_in,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65223] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(5245), 2,
      anon_sym_RBRACE,
      anon_sym_COMMA,
  [65232] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(4385), 2,
      anon_sym_RPAREN,
      anon_sym_COMMA,
  [65241] = 3,
    ACTIONS(3586), 1,
      anon_sym_COLON_COLON,
    ACTIONS(5247), 1,
      anon_sym_BANG,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65252] = 3,
    ACTIONS(2840), 1,
      anon_sym_COLON,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65263] = 3,
    ACTIONS(3592), 1,
      anon_sym_COLON_COLON,
    ACTIONS(5247), 1,
      anon_sym_BANG,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65274] = 3,
    ACTIONS(2848), 1,
      anon_sym_COLON,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65285] = 3,
    ACTIONS(632), 1,
      anon_sym_LBRACE,
    STATE(219), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65296] = 3,
    ACTIONS(4199), 1,
      anon_sym_LBRACE,
    STATE(420), 1,
      sym_enum_variant_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65307] = 3,
    ACTIONS(284), 1,
      anon_sym_LBRACE,
    STATE(2555), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65318] = 3,
    ACTIONS(2313), 1,
      anon_sym_SQUOTE,
    STATE(2265), 1,
      sym_lifetime,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65329] = 3,
    ACTIONS(3916), 1,
      anon_sym_COLON,
    STATE(2074), 1,
      sym_trait_bounds,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65340] = 3,
    ACTIONS(3806), 1,
      anon_sym_LT,
    STATE(690), 1,
      sym_type_parameters,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65351] = 3,
    ACTIONS(4617), 1,
      sym_identifier,
    ACTIONS(4621), 1,
      sym_mutable_specifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65362] = 3,
    ACTIONS(15), 1,
      anon_sym_LBRACE,
    STATE(63), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65373] = 3,
    ACTIONS(632), 1,
      anon_sym_LBRACE,
    STATE(220), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65384] = 3,
    ACTIONS(5249), 1,
      sym_identifier,
    ACTIONS(5251), 1,
      sym_mutable_specifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65395] = 3,
    ACTIONS(284), 1,
      anon_sym_LBRACE,
    STATE(2561), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65406] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(5253), 2,
      sym_identifier,
      sym_metavariable,
  [65415] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(5255), 2,
      anon_sym_RBRACE,
      anon_sym_COMMA,
  [65424] = 3,
    ACTIONS(2880), 1,
      anon_sym_COLON,
    ACTIONS(4075), 1,
      anon_sym_PLUS,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65435] = 3,
    ACTIONS(3788), 1,
      anon_sym_LBRACE,
    STATE(438), 1,
      sym_declaration_list,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65446] = 3,
    ACTIONS(284), 1,
      anon_sym_LBRACE,
    STATE(2503), 1,
      sym_block,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65457] = 3,
    ACTIONS(4009), 1,
      anon_sym_RBRACE,
    ACTIONS(5119), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65468] = 3,
    ACTIONS(4011), 1,
      anon_sym_RPAREN,
    ACTIONS(5119), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65479] = 3,
    ACTIONS(4027), 1,
      anon_sym_RBRACE,
    ACTIONS(5119), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65490] = 3,
    ACTIONS(4043), 1,
      anon_sym_RPAREN,
    ACTIONS(5119), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65501] = 3,
    ACTIONS(5257), 1,
      sym_identifier,
    ACTIONS(5259), 1,
      sym_mutable_specifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65512] = 2,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
    ACTIONS(2433), 2,
      anon_sym_COMMA,
      anon_sym_GT,
  [65521] = 2,
    ACTIONS(5261), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65529] = 2,
    ACTIONS(5263), 1,
      anon_sym_EQ_GT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65537] = 2,
    ACTIONS(4605), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65545] = 2,
    ACTIONS(5265), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65553] = 2,
    ACTIONS(5267), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65561] = 2,
    ACTIONS(4571), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65569] = 2,
    ACTIONS(5269), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65577] = 2,
    ACTIONS(4009), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65585] = 2,
    ACTIONS(5271), 1,
      anon_sym_RBRACK,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65593] = 2,
    ACTIONS(4027), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65601] = 2,
    ACTIONS(5273), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65609] = 2,
    ACTIONS(5275), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65617] = 2,
    ACTIONS(3195), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65625] = 2,
    ACTIONS(3161), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65633] = 2,
    ACTIONS(5277), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65641] = 2,
    ACTIONS(5279), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65649] = 2,
    ACTIONS(5281), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65657] = 2,
    ACTIONS(5283), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65665] = 2,
    ACTIONS(5285), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65673] = 2,
    ACTIONS(5287), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65681] = 2,
    ACTIONS(5289), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65689] = 2,
    ACTIONS(5291), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65697] = 2,
    ACTIONS(4623), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65705] = 2,
    ACTIONS(5293), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65713] = 2,
    ACTIONS(5295), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65721] = 2,
    ACTIONS(5297), 1,
      anon_sym_RBRACK,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65729] = 2,
    ACTIONS(5299), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65737] = 2,
    ACTIONS(5301), 1,
      anon_sym_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65745] = 2,
    ACTIONS(5303), 1,
      anon_sym_RBRACK,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65753] = 2,
    ACTIONS(4255), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65761] = 2,
    ACTIONS(5305), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65769] = 2,
    ACTIONS(3578), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65777] = 2,
    ACTIONS(5307), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65785] = 2,
    ACTIONS(4633), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65793] = 2,
    ACTIONS(5309), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65801] = 2,
    ACTIONS(5311), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65809] = 2,
    ACTIONS(4659), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65817] = 2,
    ACTIONS(5313), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65825] = 2,
    ACTIONS(5315), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65833] = 2,
    ACTIONS(5317), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65841] = 2,
    ACTIONS(5319), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65849] = 2,
    ACTIONS(4637), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65857] = 2,
    ACTIONS(4337), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65865] = 2,
    ACTIONS(5321), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65873] = 2,
    ACTIONS(4728), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65881] = 2,
    ACTIONS(5323), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65889] = 2,
    ACTIONS(5325), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65897] = 2,
    ACTIONS(5327), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65905] = 2,
    ACTIONS(5329), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65913] = 2,
    ACTIONS(5331), 1,
      anon_sym_RBRACK,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65921] = 2,
    ACTIONS(5068), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65929] = 2,
    ACTIONS(5333), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65937] = 2,
    ACTIONS(5335), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65945] = 2,
    ACTIONS(3183), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65953] = 2,
    ACTIONS(5337), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65961] = 2,
    ACTIONS(3533), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65969] = 2,
    ACTIONS(5339), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65977] = 2,
    ACTIONS(5341), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65985] = 2,
    ACTIONS(2712), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [65993] = 2,
    ACTIONS(5343), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66001] = 2,
    ACTIONS(2700), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66009] = 2,
    ACTIONS(4771), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66017] = 2,
    ACTIONS(4777), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66025] = 2,
    ACTIONS(5345), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66033] = 2,
    ACTIONS(3163), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66041] = 2,
    ACTIONS(5347), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66049] = 2,
    ACTIONS(5349), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66057] = 2,
    ACTIONS(5351), 1,
      anon_sym_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66065] = 2,
    ACTIONS(5353), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66073] = 2,
    ACTIONS(5355), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66081] = 2,
    ACTIONS(5357), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66089] = 2,
    ACTIONS(5359), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66097] = 2,
    ACTIONS(5361), 1,
      anon_sym_fn,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66105] = 2,
    ACTIONS(5363), 1,
      anon_sym_EQ_GT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66113] = 2,
    ACTIONS(5365), 1,
      anon_sym_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66121] = 2,
    ACTIONS(5367), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66129] = 2,
    ACTIONS(5369), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66137] = 2,
    ACTIONS(5371), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66145] = 2,
    ACTIONS(5373), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66153] = 2,
    ACTIONS(5375), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66161] = 2,
    ACTIONS(5377), 1,
      anon_sym_fn,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66169] = 2,
    ACTIONS(5379), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66177] = 2,
    ACTIONS(5381), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66185] = 2,
    ACTIONS(5383), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66193] = 2,
    ACTIONS(5385), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66201] = 2,
    ACTIONS(5387), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66209] = 2,
    ACTIONS(5389), 1,
      anon_sym_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66217] = 2,
    ACTIONS(5391), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66225] = 2,
    ACTIONS(5393), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66233] = 2,
    ACTIONS(4906), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66241] = 2,
    ACTIONS(5151), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66249] = 2,
    ACTIONS(3546), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66257] = 2,
    ACTIONS(5395), 1,
      anon_sym_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66265] = 2,
    ACTIONS(5397), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66273] = 2,
    ACTIONS(5399), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66281] = 2,
    ACTIONS(2688), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66289] = 2,
    ACTIONS(5401), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66297] = 2,
    ACTIONS(5183), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66305] = 2,
    ACTIONS(5403), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66313] = 2,
    ACTIONS(5405), 1,
      anon_sym_fn,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66321] = 2,
    ACTIONS(5407), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66329] = 2,
    ACTIONS(5409), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66337] = 2,
    ACTIONS(4922), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66345] = 2,
    ACTIONS(5411), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66353] = 2,
    ACTIONS(5413), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66361] = 2,
    ACTIONS(5415), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66369] = 2,
    ACTIONS(4419), 1,
      anon_sym_RBRACK,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66377] = 2,
    ACTIONS(4411), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66385] = 2,
    ACTIONS(5417), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66393] = 2,
    ACTIONS(5419), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66401] = 2,
    ACTIONS(5421), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66409] = 2,
    ACTIONS(5423), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66417] = 2,
    ACTIONS(624), 1,
      anon_sym_RBRACK,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66425] = 2,
    ACTIONS(5425), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66433] = 2,
    ACTIONS(5427), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66441] = 2,
    ACTIONS(5429), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66449] = 2,
    ACTIONS(4553), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66457] = 2,
    ACTIONS(5431), 1,
      anon_sym_RBRACK,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66465] = 2,
    ACTIONS(5433), 1,
      anon_sym_RBRACK,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66473] = 2,
    ACTIONS(5435), 1,
      anon_sym_RBRACK,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66481] = 2,
    ACTIONS(5437), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66489] = 2,
    ACTIONS(4106), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66497] = 2,
    ACTIONS(5439), 1,
      anon_sym_fn,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66505] = 2,
    ACTIONS(5441), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66513] = 2,
    ACTIONS(5443), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66521] = 2,
    ACTIONS(5445), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66529] = 2,
    ACTIONS(5141), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66537] = 2,
    ACTIONS(5447), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66545] = 2,
    ACTIONS(5139), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66553] = 2,
    ACTIONS(5449), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66561] = 2,
    ACTIONS(5451), 1,
      anon_sym_RBRACK,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66569] = 2,
    ACTIONS(5048), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66577] = 2,
    ACTIONS(5453), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66585] = 2,
    ACTIONS(5455), 1,
      anon_sym_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66593] = 2,
    ACTIONS(5095), 1,
      anon_sym_GT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66601] = 2,
    ACTIONS(5457), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66609] = 2,
    ACTIONS(5459), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66617] = 2,
    ACTIONS(5103), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66625] = 2,
    ACTIONS(5461), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66633] = 2,
    ACTIONS(5463), 1,
      anon_sym_LBRACK,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66641] = 2,
    ACTIONS(5465), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66649] = 2,
    ACTIONS(5467), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66657] = 2,
    ACTIONS(5469), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66665] = 2,
    ACTIONS(5471), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66673] = 2,
    ACTIONS(5473), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66681] = 2,
    ACTIONS(2359), 1,
      anon_sym_EQ_GT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66689] = 2,
    ACTIONS(5475), 1,
      anon_sym_LPAREN,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66697] = 2,
    ACTIONS(5145), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66705] = 2,
    ACTIONS(5149), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66713] = 2,
    ACTIONS(5477), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66721] = 2,
    ACTIONS(5479), 1,
      anon_sym_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66729] = 2,
    ACTIONS(5481), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66737] = 2,
    ACTIONS(4251), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66745] = 2,
    ACTIONS(5034), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66753] = 2,
    ACTIONS(5483), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66761] = 2,
    ACTIONS(5485), 1,
      sym_self,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66769] = 2,
    ACTIONS(4005), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66777] = 2,
    ACTIONS(2449), 1,
      anon_sym_PLUS,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66785] = 2,
    ACTIONS(5487), 1,
      anon_sym_fn,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66793] = 2,
    ACTIONS(4007), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66801] = 2,
    ACTIONS(5489), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66809] = 2,
    ACTIONS(5491), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66817] = 2,
    ACTIONS(5493), 1,
      anon_sym_LT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66825] = 2,
    ACTIONS(5495), 1,
      anon_sym_RBRACK,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66833] = 2,
    ACTIONS(5497), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66841] = 2,
    ACTIONS(5499), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66849] = 2,
    ACTIONS(5501), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66857] = 2,
    ACTIONS(5503), 1,
      anon_sym_EQ_GT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66865] = 2,
    ACTIONS(5505), 1,
      anon_sym_RBRACE,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66873] = 2,
    ACTIONS(5507), 1,
      anon_sym_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66881] = 2,
    ACTIONS(5509), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66889] = 2,
    ACTIONS(5511), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66897] = 2,
    ACTIONS(4918), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66905] = 2,
    ACTIONS(5513), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66913] = 2,
    ACTIONS(3523), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66921] = 2,
    ACTIONS(4178), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66929] = 2,
    ACTIONS(5515), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66937] = 2,
    ACTIONS(2690), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66945] = 2,
    ACTIONS(5517), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66953] = 2,
    ACTIONS(5519), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66961] = 2,
    ACTIONS(5521), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66969] = 2,
    ACTIONS(5523), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66977] = 2,
    ACTIONS(5525), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66985] = 2,
    ACTIONS(5527), 1,
      anon_sym_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [66993] = 2,
    ACTIONS(4561), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67001] = 2,
    ACTIONS(5529), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67009] = 2,
    ACTIONS(374), 1,
      anon_sym_RBRACK,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67017] = 2,
    ACTIONS(5531), 1,
      anon_sym_LBRACK,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67025] = 2,
    ACTIONS(4868), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67033] = 2,
    ACTIONS(5533), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67041] = 2,
    ACTIONS(4227), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67049] = 2,
    ACTIONS(5535), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67057] = 2,
    ACTIONS(5537), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67065] = 2,
    ACTIONS(5539), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67073] = 2,
    ACTIONS(4160), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67081] = 2,
    ACTIONS(5541), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67089] = 2,
    ACTIONS(5543), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67097] = 2,
    ACTIONS(4207), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67105] = 2,
    ACTIONS(5545), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67113] = 2,
    ACTIONS(2470), 1,
      anon_sym_PLUS,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67121] = 2,
    ACTIONS(2373), 1,
      anon_sym_EQ_GT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67129] = 2,
    ACTIONS(5547), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67137] = 2,
    ACTIONS(5549), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67145] = 2,
    ACTIONS(5551), 1,
      anon_sym_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67153] = 2,
    ACTIONS(5553), 1,
      anon_sym_LBRACK,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67161] = 2,
    ACTIONS(5555), 1,
      anon_sym_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67169] = 2,
    ACTIONS(5557), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67177] = 2,
    ACTIONS(5559), 1,
      anon_sym_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67185] = 2,
    ACTIONS(5119), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67193] = 2,
    ACTIONS(5561), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67201] = 2,
    ACTIONS(5563), 1,
      anon_sym_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67209] = 2,
    ACTIONS(5565), 1,
      anon_sym_LBRACK,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67217] = 2,
    ACTIONS(5567), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67225] = 2,
    ACTIONS(4047), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67233] = 2,
    ACTIONS(3485), 1,
      anon_sym_fn,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67241] = 2,
    ACTIONS(5569), 1,
      ts_builtin_sym_end,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67249] = 2,
    ACTIONS(2904), 1,
      anon_sym_COLON_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67257] = 2,
    ACTIONS(5571), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67265] = 2,
    ACTIONS(5573), 1,
      anon_sym_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67273] = 2,
    ACTIONS(5575), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67281] = 2,
    ACTIONS(5577), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67289] = 2,
    ACTIONS(5579), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67297] = 2,
    ACTIONS(5581), 1,
      anon_sym_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67305] = 2,
    ACTIONS(5583), 1,
      anon_sym_RPAREN,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67313] = 2,
    ACTIONS(5585), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67321] = 2,
    ACTIONS(5587), 1,
      anon_sym_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67329] = 2,
    ACTIONS(5589), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67337] = 2,
    ACTIONS(5591), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67345] = 2,
    ACTIONS(5593), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67353] = 2,
    ACTIONS(5595), 1,
      anon_sym_EQ,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67361] = 2,
    ACTIONS(5597), 1,
      anon_sym_EQ_GT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67369] = 2,
    ACTIONS(5599), 1,
      anon_sym_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67377] = 2,
    ACTIONS(5601), 1,
      anon_sym_COLON,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67385] = 2,
    ACTIONS(5603), 1,
      anon_sym_LBRACK,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67393] = 2,
    ACTIONS(5605), 1,
      anon_sym_EQ_GT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67401] = 2,
    ACTIONS(5607), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67409] = 2,
    ACTIONS(5609), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67417] = 2,
    ACTIONS(5611), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67425] = 2,
    ACTIONS(3459), 1,
      anon_sym_fn,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67433] = 2,
    ACTIONS(5613), 1,
      anon_sym_LPAREN,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67441] = 2,
    ACTIONS(5615), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67449] = 2,
    ACTIONS(5617), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67457] = 2,
    ACTIONS(5619), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67465] = 2,
    ACTIONS(5621), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67473] = 2,
    ACTIONS(5623), 1,
      anon_sym_fn,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67481] = 2,
    ACTIONS(5625), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67489] = 2,
    ACTIONS(5627), 1,
      anon_sym_EQ_GT,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67497] = 2,
    ACTIONS(5629), 1,
      anon_sym_SEMI,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67505] = 2,
    ACTIONS(5631), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67513] = 2,
    ACTIONS(5633), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
  [67521] = 2,
    ACTIONS(5635), 1,
      sym_identifier,
    ACTIONS(3), 2,
      sym_block_comment,
      sym_line_comment,
};

static const uint32_t ts_small_parse_table_map[] = {
  [SMALL_STATE(639)] = 0,
  [SMALL_STATE(640)] = 129,
  [SMALL_STATE(641)] = 258,
  [SMALL_STATE(642)] = 387,
  [SMALL_STATE(643)] = 516,
  [SMALL_STATE(644)] = 645,
  [SMALL_STATE(645)] = 774,
  [SMALL_STATE(646)] = 903,
  [SMALL_STATE(647)] = 1032,
  [SMALL_STATE(648)] = 1161,
  [SMALL_STATE(649)] = 1290,
  [SMALL_STATE(650)] = 1419,
  [SMALL_STATE(651)] = 1548,
  [SMALL_STATE(652)] = 1677,
  [SMALL_STATE(653)] = 1806,
  [SMALL_STATE(654)] = 1935,
  [SMALL_STATE(655)] = 2064,
  [SMALL_STATE(656)] = 2193,
  [SMALL_STATE(657)] = 2322,
  [SMALL_STATE(658)] = 2451,
  [SMALL_STATE(659)] = 2580,
  [SMALL_STATE(660)] = 2709,
  [SMALL_STATE(661)] = 2838,
  [SMALL_STATE(662)] = 2967,
  [SMALL_STATE(663)] = 3096,
  [SMALL_STATE(664)] = 3225,
  [SMALL_STATE(665)] = 3354,
  [SMALL_STATE(666)] = 3483,
  [SMALL_STATE(667)] = 3612,
  [SMALL_STATE(668)] = 3741,
  [SMALL_STATE(669)] = 3870,
  [SMALL_STATE(670)] = 3999,
  [SMALL_STATE(671)] = 4128,
  [SMALL_STATE(672)] = 4257,
  [SMALL_STATE(673)] = 4386,
  [SMALL_STATE(674)] = 4515,
  [SMALL_STATE(675)] = 4644,
  [SMALL_STATE(676)] = 4773,
  [SMALL_STATE(677)] = 4902,
  [SMALL_STATE(678)] = 5031,
  [SMALL_STATE(679)] = 5160,
  [SMALL_STATE(680)] = 5289,
  [SMALL_STATE(681)] = 5418,
  [SMALL_STATE(682)] = 5547,
  [SMALL_STATE(683)] = 5676,
  [SMALL_STATE(684)] = 5805,
  [SMALL_STATE(685)] = 5934,
  [SMALL_STATE(686)] = 6063,
  [SMALL_STATE(687)] = 6192,
  [SMALL_STATE(688)] = 6321,
  [SMALL_STATE(689)] = 6450,
  [SMALL_STATE(690)] = 6579,
  [SMALL_STATE(691)] = 6708,
  [SMALL_STATE(692)] = 6837,
  [SMALL_STATE(693)] = 6966,
  [SMALL_STATE(694)] = 7095,
  [SMALL_STATE(695)] = 7224,
  [SMALL_STATE(696)] = 7353,
  [SMALL_STATE(697)] = 7482,
  [SMALL_STATE(698)] = 7611,
  [SMALL_STATE(699)] = 7740,
  [SMALL_STATE(700)] = 7869,
  [SMALL_STATE(701)] = 8000,
  [SMALL_STATE(702)] = 8129,
  [SMALL_STATE(703)] = 8258,
  [SMALL_STATE(704)] = 8387,
  [SMALL_STATE(705)] = 8516,
  [SMALL_STATE(706)] = 8645,
  [SMALL_STATE(707)] = 8774,
  [SMALL_STATE(708)] = 8903,
  [SMALL_STATE(709)] = 9032,
  [SMALL_STATE(710)] = 9161,
  [SMALL_STATE(711)] = 9290,
  [SMALL_STATE(712)] = 9419,
  [SMALL_STATE(713)] = 9548,
  [SMALL_STATE(714)] = 9677,
  [SMALL_STATE(715)] = 9806,
  [SMALL_STATE(716)] = 9935,
  [SMALL_STATE(717)] = 10064,
  [SMALL_STATE(718)] = 10193,
  [SMALL_STATE(719)] = 10322,
  [SMALL_STATE(720)] = 10451,
  [SMALL_STATE(721)] = 10580,
  [SMALL_STATE(722)] = 10709,
  [SMALL_STATE(723)] = 10838,
  [SMALL_STATE(724)] = 10967,
  [SMALL_STATE(725)] = 11096,
  [SMALL_STATE(726)] = 11225,
  [SMALL_STATE(727)] = 11354,
  [SMALL_STATE(728)] = 11483,
  [SMALL_STATE(729)] = 11612,
  [SMALL_STATE(730)] = 11741,
  [SMALL_STATE(731)] = 11870,
  [SMALL_STATE(732)] = 11999,
  [SMALL_STATE(733)] = 12070,
  [SMALL_STATE(734)] = 12199,
  [SMALL_STATE(735)] = 12328,
  [SMALL_STATE(736)] = 12457,
  [SMALL_STATE(737)] = 12586,
  [SMALL_STATE(738)] = 12715,
  [SMALL_STATE(739)] = 12844,
  [SMALL_STATE(740)] = 12973,
  [SMALL_STATE(741)] = 13102,
  [SMALL_STATE(742)] = 13231,
  [SMALL_STATE(743)] = 13360,
  [SMALL_STATE(744)] = 13489,
  [SMALL_STATE(745)] = 13618,
  [SMALL_STATE(746)] = 13747,
  [SMALL_STATE(747)] = 13876,
  [SMALL_STATE(748)] = 13942,
  [SMALL_STATE(749)] = 14008,
  [SMALL_STATE(750)] = 14074,
  [SMALL_STATE(751)] = 14140,
  [SMALL_STATE(752)] = 14202,
  [SMALL_STATE(753)] = 14273,
  [SMALL_STATE(754)] = 14341,
  [SMALL_STATE(755)] = 14409,
  [SMALL_STATE(756)] = 14470,
  [SMALL_STATE(757)] = 14531,
  [SMALL_STATE(758)] = 14596,
  [SMALL_STATE(759)] = 14661,
  [SMALL_STATE(760)] = 14726,
  [SMALL_STATE(761)] = 14791,
  [SMALL_STATE(762)] = 14852,
  [SMALL_STATE(763)] = 14913,
  [SMALL_STATE(764)] = 14969,
  [SMALL_STATE(765)] = 15027,
  [SMALL_STATE(766)] = 15089,
  [SMALL_STATE(767)] = 15147,
  [SMALL_STATE(768)] = 15203,
  [SMALL_STATE(769)] = 15259,
  [SMALL_STATE(770)] = 15317,
  [SMALL_STATE(771)] = 15377,
  [SMALL_STATE(772)] = 15433,
  [SMALL_STATE(773)] = 15491,
  [SMALL_STATE(774)] = 15547,
  [SMALL_STATE(775)] = 15603,
  [SMALL_STATE(776)] = 15658,
  [SMALL_STATE(777)] = 15713,
  [SMALL_STATE(778)] = 15772,
  [SMALL_STATE(779)] = 15829,
  [SMALL_STATE(780)] = 15886,
  [SMALL_STATE(781)] = 15945,
  [SMALL_STATE(782)] = 16002,
  [SMALL_STATE(783)] = 16057,
  [SMALL_STATE(784)] = 16114,
  [SMALL_STATE(785)] = 16169,
  [SMALL_STATE(786)] = 16224,
  [SMALL_STATE(787)] = 16281,
  [SMALL_STATE(788)] = 16338,
  [SMALL_STATE(789)] = 16395,
  [SMALL_STATE(790)] = 16450,
  [SMALL_STATE(791)] = 16507,
  [SMALL_STATE(792)] = 16562,
  [SMALL_STATE(793)] = 16621,
  [SMALL_STATE(794)] = 16678,
  [SMALL_STATE(795)] = 16733,
  [SMALL_STATE(796)] = 16788,
  [SMALL_STATE(797)] = 16845,
  [SMALL_STATE(798)] = 16900,
  [SMALL_STATE(799)] = 16957,
  [SMALL_STATE(800)] = 17012,
  [SMALL_STATE(801)] = 17067,
  [SMALL_STATE(802)] = 17126,
  [SMALL_STATE(803)] = 17183,
  [SMALL_STATE(804)] = 17240,
  [SMALL_STATE(805)] = 17295,
  [SMALL_STATE(806)] = 17352,
  [SMALL_STATE(807)] = 17407,
  [SMALL_STATE(808)] = 17464,
  [SMALL_STATE(809)] = 17518,
  [SMALL_STATE(810)] = 17572,
  [SMALL_STATE(811)] = 17626,
  [SMALL_STATE(812)] = 17680,
  [SMALL_STATE(813)] = 17734,
  [SMALL_STATE(814)] = 17788,
  [SMALL_STATE(815)] = 17842,
  [SMALL_STATE(816)] = 17896,
  [SMALL_STATE(817)] = 17950,
  [SMALL_STATE(818)] = 18004,
  [SMALL_STATE(819)] = 18058,
  [SMALL_STATE(820)] = 18112,
  [SMALL_STATE(821)] = 18166,
  [SMALL_STATE(822)] = 18220,
  [SMALL_STATE(823)] = 18274,
  [SMALL_STATE(824)] = 18328,
  [SMALL_STATE(825)] = 18382,
  [SMALL_STATE(826)] = 18436,
  [SMALL_STATE(827)] = 18490,
  [SMALL_STATE(828)] = 18544,
  [SMALL_STATE(829)] = 18598,
  [SMALL_STATE(830)] = 18652,
  [SMALL_STATE(831)] = 18706,
  [SMALL_STATE(832)] = 18760,
  [SMALL_STATE(833)] = 18814,
  [SMALL_STATE(834)] = 18868,
  [SMALL_STATE(835)] = 18922,
  [SMALL_STATE(836)] = 18976,
  [SMALL_STATE(837)] = 19030,
  [SMALL_STATE(838)] = 19084,
  [SMALL_STATE(839)] = 19138,
  [SMALL_STATE(840)] = 19192,
  [SMALL_STATE(841)] = 19246,
  [SMALL_STATE(842)] = 19300,
  [SMALL_STATE(843)] = 19354,
  [SMALL_STATE(844)] = 19408,
  [SMALL_STATE(845)] = 19462,
  [SMALL_STATE(846)] = 19516,
  [SMALL_STATE(847)] = 19570,
  [SMALL_STATE(848)] = 19624,
  [SMALL_STATE(849)] = 19678,
  [SMALL_STATE(850)] = 19732,
  [SMALL_STATE(851)] = 19786,
  [SMALL_STATE(852)] = 19840,
  [SMALL_STATE(853)] = 19894,
  [SMALL_STATE(854)] = 19948,
  [SMALL_STATE(855)] = 20002,
  [SMALL_STATE(856)] = 20056,
  [SMALL_STATE(857)] = 20110,
  [SMALL_STATE(858)] = 20164,
  [SMALL_STATE(859)] = 20218,
  [SMALL_STATE(860)] = 20272,
  [SMALL_STATE(861)] = 20326,
  [SMALL_STATE(862)] = 20380,
  [SMALL_STATE(863)] = 20434,
  [SMALL_STATE(864)] = 20488,
  [SMALL_STATE(865)] = 20542,
  [SMALL_STATE(866)] = 20596,
  [SMALL_STATE(867)] = 20650,
  [SMALL_STATE(868)] = 20704,
  [SMALL_STATE(869)] = 20758,
  [SMALL_STATE(870)] = 20812,
  [SMALL_STATE(871)] = 20866,
  [SMALL_STATE(872)] = 20920,
  [SMALL_STATE(873)] = 20974,
  [SMALL_STATE(874)] = 21028,
  [SMALL_STATE(875)] = 21082,
  [SMALL_STATE(876)] = 21136,
  [SMALL_STATE(877)] = 21190,
  [SMALL_STATE(878)] = 21244,
  [SMALL_STATE(879)] = 21298,
  [SMALL_STATE(880)] = 21352,
  [SMALL_STATE(881)] = 21406,
  [SMALL_STATE(882)] = 21460,
  [SMALL_STATE(883)] = 21514,
  [SMALL_STATE(884)] = 21568,
  [SMALL_STATE(885)] = 21622,
  [SMALL_STATE(886)] = 21676,
  [SMALL_STATE(887)] = 21730,
  [SMALL_STATE(888)] = 21784,
  [SMALL_STATE(889)] = 21838,
  [SMALL_STATE(890)] = 21892,
  [SMALL_STATE(891)] = 21946,
  [SMALL_STATE(892)] = 22000,
  [SMALL_STATE(893)] = 22054,
  [SMALL_STATE(894)] = 22108,
  [SMALL_STATE(895)] = 22162,
  [SMALL_STATE(896)] = 22216,
  [SMALL_STATE(897)] = 22270,
  [SMALL_STATE(898)] = 22324,
  [SMALL_STATE(899)] = 22378,
  [SMALL_STATE(900)] = 22432,
  [SMALL_STATE(901)] = 22486,
  [SMALL_STATE(902)] = 22540,
  [SMALL_STATE(903)] = 22594,
  [SMALL_STATE(904)] = 22648,
  [SMALL_STATE(905)] = 22702,
  [SMALL_STATE(906)] = 22756,
  [SMALL_STATE(907)] = 22810,
  [SMALL_STATE(908)] = 22864,
  [SMALL_STATE(909)] = 22918,
  [SMALL_STATE(910)] = 22972,
  [SMALL_STATE(911)] = 23026,
  [SMALL_STATE(912)] = 23080,
  [SMALL_STATE(913)] = 23134,
  [SMALL_STATE(914)] = 23188,
  [SMALL_STATE(915)] = 23242,
  [SMALL_STATE(916)] = 23296,
  [SMALL_STATE(917)] = 23350,
  [SMALL_STATE(918)] = 23404,
  [SMALL_STATE(919)] = 23458,
  [SMALL_STATE(920)] = 23512,
  [SMALL_STATE(921)] = 23566,
  [SMALL_STATE(922)] = 23620,
  [SMALL_STATE(923)] = 23674,
  [SMALL_STATE(924)] = 23728,
  [SMALL_STATE(925)] = 23782,
  [SMALL_STATE(926)] = 23836,
  [SMALL_STATE(927)] = 23890,
  [SMALL_STATE(928)] = 23944,
  [SMALL_STATE(929)] = 23998,
  [SMALL_STATE(930)] = 24052,
  [SMALL_STATE(931)] = 24106,
  [SMALL_STATE(932)] = 24160,
  [SMALL_STATE(933)] = 24214,
  [SMALL_STATE(934)] = 24268,
  [SMALL_STATE(935)] = 24322,
  [SMALL_STATE(936)] = 24376,
  [SMALL_STATE(937)] = 24430,
  [SMALL_STATE(938)] = 24484,
  [SMALL_STATE(939)] = 24538,
  [SMALL_STATE(940)] = 24592,
  [SMALL_STATE(941)] = 24646,
  [SMALL_STATE(942)] = 24700,
  [SMALL_STATE(943)] = 24754,
  [SMALL_STATE(944)] = 24808,
  [SMALL_STATE(945)] = 24862,
  [SMALL_STATE(946)] = 24916,
  [SMALL_STATE(947)] = 24970,
  [SMALL_STATE(948)] = 25024,
  [SMALL_STATE(949)] = 25078,
  [SMALL_STATE(950)] = 25132,
  [SMALL_STATE(951)] = 25186,
  [SMALL_STATE(952)] = 25240,
  [SMALL_STATE(953)] = 25294,
  [SMALL_STATE(954)] = 25348,
  [SMALL_STATE(955)] = 25402,
  [SMALL_STATE(956)] = 25456,
  [SMALL_STATE(957)] = 25510,
  [SMALL_STATE(958)] = 25564,
  [SMALL_STATE(959)] = 25618,
  [SMALL_STATE(960)] = 25672,
  [SMALL_STATE(961)] = 25726,
  [SMALL_STATE(962)] = 25780,
  [SMALL_STATE(963)] = 25834,
  [SMALL_STATE(964)] = 25888,
  [SMALL_STATE(965)] = 25942,
  [SMALL_STATE(966)] = 25996,
  [SMALL_STATE(967)] = 26050,
  [SMALL_STATE(968)] = 26104,
  [SMALL_STATE(969)] = 26158,
  [SMALL_STATE(970)] = 26212,
  [SMALL_STATE(971)] = 26266,
  [SMALL_STATE(972)] = 26320,
  [SMALL_STATE(973)] = 26374,
  [SMALL_STATE(974)] = 26428,
  [SMALL_STATE(975)] = 26482,
  [SMALL_STATE(976)] = 26536,
  [SMALL_STATE(977)] = 26590,
  [SMALL_STATE(978)] = 26644,
  [SMALL_STATE(979)] = 26698,
  [SMALL_STATE(980)] = 26752,
  [SMALL_STATE(981)] = 26806,
  [SMALL_STATE(982)] = 26860,
  [SMALL_STATE(983)] = 26914,
  [SMALL_STATE(984)] = 26968,
  [SMALL_STATE(985)] = 27022,
  [SMALL_STATE(986)] = 27076,
  [SMALL_STATE(987)] = 27130,
  [SMALL_STATE(988)] = 27184,
  [SMALL_STATE(989)] = 27238,
  [SMALL_STATE(990)] = 27292,
  [SMALL_STATE(991)] = 27346,
  [SMALL_STATE(992)] = 27400,
  [SMALL_STATE(993)] = 27454,
  [SMALL_STATE(994)] = 27508,
  [SMALL_STATE(995)] = 27562,
  [SMALL_STATE(996)] = 27616,
  [SMALL_STATE(997)] = 27670,
  [SMALL_STATE(998)] = 27724,
  [SMALL_STATE(999)] = 27778,
  [SMALL_STATE(1000)] = 27832,
  [SMALL_STATE(1001)] = 27886,
  [SMALL_STATE(1002)] = 27940,
  [SMALL_STATE(1003)] = 27994,
  [SMALL_STATE(1004)] = 28048,
  [SMALL_STATE(1005)] = 28102,
  [SMALL_STATE(1006)] = 28156,
  [SMALL_STATE(1007)] = 28210,
  [SMALL_STATE(1008)] = 28264,
  [SMALL_STATE(1009)] = 28318,
  [SMALL_STATE(1010)] = 28372,
  [SMALL_STATE(1011)] = 28426,
  [SMALL_STATE(1012)] = 28480,
  [SMALL_STATE(1013)] = 28534,
  [SMALL_STATE(1014)] = 28588,
  [SMALL_STATE(1015)] = 28642,
  [SMALL_STATE(1016)] = 28696,
  [SMALL_STATE(1017)] = 28750,
  [SMALL_STATE(1018)] = 28804,
  [SMALL_STATE(1019)] = 28858,
  [SMALL_STATE(1020)] = 28912,
  [SMALL_STATE(1021)] = 28966,
  [SMALL_STATE(1022)] = 29020,
  [SMALL_STATE(1023)] = 29074,
  [SMALL_STATE(1024)] = 29128,
  [SMALL_STATE(1025)] = 29182,
  [SMALL_STATE(1026)] = 29236,
  [SMALL_STATE(1027)] = 29290,
  [SMALL_STATE(1028)] = 29344,
  [SMALL_STATE(1029)] = 29398,
  [SMALL_STATE(1030)] = 29452,
  [SMALL_STATE(1031)] = 29506,
  [SMALL_STATE(1032)] = 29562,
  [SMALL_STATE(1033)] = 29616,
  [SMALL_STATE(1034)] = 29670,
  [SMALL_STATE(1035)] = 29724,
  [SMALL_STATE(1036)] = 29778,
  [SMALL_STATE(1037)] = 29832,
  [SMALL_STATE(1038)] = 29886,
  [SMALL_STATE(1039)] = 29940,
  [SMALL_STATE(1040)] = 29994,
  [SMALL_STATE(1041)] = 30048,
  [SMALL_STATE(1042)] = 30102,
  [SMALL_STATE(1043)] = 30156,
  [SMALL_STATE(1044)] = 30210,
  [SMALL_STATE(1045)] = 30264,
  [SMALL_STATE(1046)] = 30318,
  [SMALL_STATE(1047)] = 30372,
  [SMALL_STATE(1048)] = 30426,
  [SMALL_STATE(1049)] = 30480,
  [SMALL_STATE(1050)] = 30534,
  [SMALL_STATE(1051)] = 30588,
  [SMALL_STATE(1052)] = 30642,
  [SMALL_STATE(1053)] = 30696,
  [SMALL_STATE(1054)] = 30750,
  [SMALL_STATE(1055)] = 30804,
  [SMALL_STATE(1056)] = 30858,
  [SMALL_STATE(1057)] = 30912,
  [SMALL_STATE(1058)] = 30966,
  [SMALL_STATE(1059)] = 31020,
  [SMALL_STATE(1060)] = 31074,
  [SMALL_STATE(1061)] = 31128,
  [SMALL_STATE(1062)] = 31182,
  [SMALL_STATE(1063)] = 31236,
  [SMALL_STATE(1064)] = 31290,
  [SMALL_STATE(1065)] = 31344,
  [SMALL_STATE(1066)] = 31398,
  [SMALL_STATE(1067)] = 31452,
  [SMALL_STATE(1068)] = 31506,
  [SMALL_STATE(1069)] = 31560,
  [SMALL_STATE(1070)] = 31614,
  [SMALL_STATE(1071)] = 31668,
  [SMALL_STATE(1072)] = 31722,
  [SMALL_STATE(1073)] = 31776,
  [SMALL_STATE(1074)] = 31830,
  [SMALL_STATE(1075)] = 31884,
  [SMALL_STATE(1076)] = 31938,
  [SMALL_STATE(1077)] = 31992,
  [SMALL_STATE(1078)] = 32046,
  [SMALL_STATE(1079)] = 32100,
  [SMALL_STATE(1080)] = 32154,
  [SMALL_STATE(1081)] = 32208,
  [SMALL_STATE(1082)] = 32262,
  [SMALL_STATE(1083)] = 32316,
  [SMALL_STATE(1084)] = 32370,
  [SMALL_STATE(1085)] = 32424,
  [SMALL_STATE(1086)] = 32478,
  [SMALL_STATE(1087)] = 32532,
  [SMALL_STATE(1088)] = 32586,
  [SMALL_STATE(1089)] = 32640,
  [SMALL_STATE(1090)] = 32694,
  [SMALL_STATE(1091)] = 32748,
  [SMALL_STATE(1092)] = 32802,
  [SMALL_STATE(1093)] = 32856,
  [SMALL_STATE(1094)] = 32910,
  [SMALL_STATE(1095)] = 32964,
  [SMALL_STATE(1096)] = 33018,
  [SMALL_STATE(1097)] = 33072,
  [SMALL_STATE(1098)] = 33126,
  [SMALL_STATE(1099)] = 33180,
  [SMALL_STATE(1100)] = 33234,
  [SMALL_STATE(1101)] = 33288,
  [SMALL_STATE(1102)] = 33342,
  [SMALL_STATE(1103)] = 33396,
  [SMALL_STATE(1104)] = 33450,
  [SMALL_STATE(1105)] = 33504,
  [SMALL_STATE(1106)] = 33558,
  [SMALL_STATE(1107)] = 33612,
  [SMALL_STATE(1108)] = 33666,
  [SMALL_STATE(1109)] = 33720,
  [SMALL_STATE(1110)] = 33774,
  [SMALL_STATE(1111)] = 33828,
  [SMALL_STATE(1112)] = 33882,
  [SMALL_STATE(1113)] = 33936,
  [SMALL_STATE(1114)] = 33990,
  [SMALL_STATE(1115)] = 34044,
  [SMALL_STATE(1116)] = 34098,
  [SMALL_STATE(1117)] = 34152,
  [SMALL_STATE(1118)] = 34206,
  [SMALL_STATE(1119)] = 34260,
  [SMALL_STATE(1120)] = 34316,
  [SMALL_STATE(1121)] = 34370,
  [SMALL_STATE(1122)] = 34424,
  [SMALL_STATE(1123)] = 34478,
  [SMALL_STATE(1124)] = 34532,
  [SMALL_STATE(1125)] = 34586,
  [SMALL_STATE(1126)] = 34640,
  [SMALL_STATE(1127)] = 34694,
  [SMALL_STATE(1128)] = 34748,
  [SMALL_STATE(1129)] = 34823,
  [SMALL_STATE(1130)] = 34892,
  [SMALL_STATE(1131)] = 34953,
  [SMALL_STATE(1132)] = 35042,
  [SMALL_STATE(1133)] = 35125,
  [SMALL_STATE(1134)] = 35214,
  [SMALL_STATE(1135)] = 35267,
  [SMALL_STATE(1136)] = 35320,
  [SMALL_STATE(1137)] = 35403,
  [SMALL_STATE(1138)] = 35458,
  [SMALL_STATE(1139)] = 35541,
  [SMALL_STATE(1140)] = 35594,
  [SMALL_STATE(1141)] = 35661,
  [SMALL_STATE(1142)] = 35714,
  [SMALL_STATE(1143)] = 35775,
  [SMALL_STATE(1144)] = 35846,
  [SMALL_STATE(1145)] = 35919,
  [SMALL_STATE(1146)] = 36006,
  [SMALL_STATE(1147)] = 36087,
  [SMALL_STATE(1148)] = 36166,
  [SMALL_STATE(1149)] = 36249,
  [SMALL_STATE(1150)] = 36304,
  [SMALL_STATE(1151)] = 36357,
  [SMALL_STATE(1152)] = 36410,
  [SMALL_STATE(1153)] = 36493,
  [SMALL_STATE(1154)] = 36550,
  [SMALL_STATE(1155)] = 36633,
  [SMALL_STATE(1156)] = 36694,
  [SMALL_STATE(1157)] = 36781,
  [SMALL_STATE(1158)] = 36844,
  [SMALL_STATE(1159)] = 36909,
  [SMALL_STATE(1160)] = 36961,
  [SMALL_STATE(1161)] = 37013,
  [SMALL_STATE(1162)] = 37073,
  [SMALL_STATE(1163)] = 37158,
  [SMALL_STATE(1164)] = 37209,
  [SMALL_STATE(1165)] = 37296,
  [SMALL_STATE(1166)] = 37383,
  [SMALL_STATE(1167)] = 37440,
  [SMALL_STATE(1168)] = 37527,
  [SMALL_STATE(1169)] = 37612,
  [SMALL_STATE(1170)] = 37699,
  [SMALL_STATE(1171)] = 37750,
  [SMALL_STATE(1172)] = 37806,
  [SMALL_STATE(1173)] = 37860,
  [SMALL_STATE(1174)] = 37936,
  [SMALL_STATE(1175)] = 38012,
  [SMALL_STATE(1176)] = 38101,
  [SMALL_STATE(1177)] = 38154,
  [SMALL_STATE(1178)] = 38207,
  [SMALL_STATE(1179)] = 38258,
  [SMALL_STATE(1180)] = 38307,
  [SMALL_STATE(1181)] = 38356,
  [SMALL_STATE(1182)] = 38405,
  [SMALL_STATE(1183)] = 38494,
  [SMALL_STATE(1184)] = 38545,
  [SMALL_STATE(1185)] = 38594,
  [SMALL_STATE(1186)] = 38643,
  [SMALL_STATE(1187)] = 38692,
  [SMALL_STATE(1188)] = 38741,
  [SMALL_STATE(1189)] = 38790,
  [SMALL_STATE(1190)] = 38839,
  [SMALL_STATE(1191)] = 38889,
  [SMALL_STATE(1192)] = 38939,
  [SMALL_STATE(1193)] = 38989,
  [SMALL_STATE(1194)] = 39075,
  [SMALL_STATE(1195)] = 39125,
  [SMALL_STATE(1196)] = 39203,
  [SMALL_STATE(1197)] = 39281,
  [SMALL_STATE(1198)] = 39367,
  [SMALL_STATE(1199)] = 39444,
  [SMALL_STATE(1200)] = 39527,
  [SMALL_STATE(1201)] = 39582,
  [SMALL_STATE(1202)] = 39665,
  [SMALL_STATE(1203)] = 39748,
  [SMALL_STATE(1204)] = 39805,
  [SMALL_STATE(1205)] = 39888,
  [SMALL_STATE(1206)] = 39965,
  [SMALL_STATE(1207)] = 40048,
  [SMALL_STATE(1208)] = 40119,
  [SMALL_STATE(1209)] = 40174,
  [SMALL_STATE(1210)] = 40257,
  [SMALL_STATE(1211)] = 40340,
  [SMALL_STATE(1212)] = 40423,
  [SMALL_STATE(1213)] = 40504,
  [SMALL_STATE(1214)] = 40587,
  [SMALL_STATE(1215)] = 40670,
  [SMALL_STATE(1216)] = 40753,
  [SMALL_STATE(1217)] = 40836,
  [SMALL_STATE(1218)] = 40919,
  [SMALL_STATE(1219)] = 41000,
  [SMALL_STATE(1220)] = 41083,
  [SMALL_STATE(1221)] = 41166,
  [SMALL_STATE(1222)] = 41243,
  [SMALL_STATE(1223)] = 41326,
  [SMALL_STATE(1224)] = 41407,
  [SMALL_STATE(1225)] = 41468,
  [SMALL_STATE(1226)] = 41549,
  [SMALL_STATE(1227)] = 41632,
  [SMALL_STATE(1228)] = 41713,
  [SMALL_STATE(1229)] = 41796,
  [SMALL_STATE(1230)] = 41879,
  [SMALL_STATE(1231)] = 41962,
  [SMALL_STATE(1232)] = 42045,
  [SMALL_STATE(1233)] = 42110,
  [SMALL_STATE(1234)] = 42193,
  [SMALL_STATE(1235)] = 42276,
  [SMALL_STATE(1236)] = 42343,
  [SMALL_STATE(1237)] = 42418,
  [SMALL_STATE(1238)] = 42501,
  [SMALL_STATE(1239)] = 42578,
  [SMALL_STATE(1240)] = 42661,
  [SMALL_STATE(1241)] = 42734,
  [SMALL_STATE(1242)] = 42797,
  [SMALL_STATE(1243)] = 42880,
  [SMALL_STATE(1244)] = 42963,
  [SMALL_STATE(1245)] = 43032,
  [SMALL_STATE(1246)] = 43115,
  [SMALL_STATE(1247)] = 43198,
  [SMALL_STATE(1248)] = 43275,
  [SMALL_STATE(1249)] = 43358,
  [SMALL_STATE(1250)] = 43417,
  [SMALL_STATE(1251)] = 43498,
  [SMALL_STATE(1252)] = 43581,
  [SMALL_STATE(1253)] = 43662,
  [SMALL_STATE(1254)] = 43717,
  [SMALL_STATE(1255)] = 43798,
  [SMALL_STATE(1256)] = 43879,
  [SMALL_STATE(1257)] = 43960,
  [SMALL_STATE(1258)] = 44043,
  [SMALL_STATE(1259)] = 44124,
  [SMALL_STATE(1260)] = 44207,
  [SMALL_STATE(1261)] = 44290,
  [SMALL_STATE(1262)] = 44373,
  [SMALL_STATE(1263)] = 44454,
  [SMALL_STATE(1264)] = 44531,
  [SMALL_STATE(1265)] = 44611,
  [SMALL_STATE(1266)] = 44691,
  [SMALL_STATE(1267)] = 44771,
  [SMALL_STATE(1268)] = 44851,
  [SMALL_STATE(1269)] = 44929,
  [SMALL_STATE(1270)] = 45009,
  [SMALL_STATE(1271)] = 45087,
  [SMALL_STATE(1272)] = 45167,
  [SMALL_STATE(1273)] = 45247,
  [SMALL_STATE(1274)] = 45327,
  [SMALL_STATE(1275)] = 45407,
  [SMALL_STATE(1276)] = 45487,
  [SMALL_STATE(1277)] = 45567,
  [SMALL_STATE(1278)] = 45647,
  [SMALL_STATE(1279)] = 45727,
  [SMALL_STATE(1280)] = 45807,
  [SMALL_STATE(1281)] = 45885,
  [SMALL_STATE(1282)] = 45965,
  [SMALL_STATE(1283)] = 46045,
  [SMALL_STATE(1284)] = 46123,
  [SMALL_STATE(1285)] = 46203,
  [SMALL_STATE(1286)] = 46283,
  [SMALL_STATE(1287)] = 46363,
  [SMALL_STATE(1288)] = 46443,
  [SMALL_STATE(1289)] = 46523,
  [SMALL_STATE(1290)] = 46591,
  [SMALL_STATE(1291)] = 46671,
  [SMALL_STATE(1292)] = 46751,
  [SMALL_STATE(1293)] = 46831,
  [SMALL_STATE(1294)] = 46911,
  [SMALL_STATE(1295)] = 46991,
  [SMALL_STATE(1296)] = 47059,
  [SMALL_STATE(1297)] = 47124,
  [SMALL_STATE(1298)] = 47189,
  [SMALL_STATE(1299)] = 47254,
  [SMALL_STATE(1300)] = 47319,
  [SMALL_STATE(1301)] = 47384,
  [SMALL_STATE(1302)] = 47424,
  [SMALL_STATE(1303)] = 47464,
  [SMALL_STATE(1304)] = 47504,
  [SMALL_STATE(1305)] = 47559,
  [SMALL_STATE(1306)] = 47614,
  [SMALL_STATE(1307)] = 47669,
  [SMALL_STATE(1308)] = 47724,
  [SMALL_STATE(1309)] = 47779,
  [SMALL_STATE(1310)] = 47834,
  [SMALL_STATE(1311)] = 47889,
  [SMALL_STATE(1312)] = 47941,
  [SMALL_STATE(1313)] = 47993,
  [SMALL_STATE(1314)] = 48024,
  [SMALL_STATE(1315)] = 48055,
  [SMALL_STATE(1316)] = 48101,
  [SMALL_STATE(1317)] = 48142,
  [SMALL_STATE(1318)] = 48172,
  [SMALL_STATE(1319)] = 48202,
  [SMALL_STATE(1320)] = 48240,
  [SMALL_STATE(1321)] = 48270,
  [SMALL_STATE(1322)] = 48300,
  [SMALL_STATE(1323)] = 48330,
  [SMALL_STATE(1324)] = 48360,
  [SMALL_STATE(1325)] = 48390,
  [SMALL_STATE(1326)] = 48420,
  [SMALL_STATE(1327)] = 48458,
  [SMALL_STATE(1328)] = 48491,
  [SMALL_STATE(1329)] = 48518,
  [SMALL_STATE(1330)] = 48551,
  [SMALL_STATE(1331)] = 48578,
  [SMALL_STATE(1332)] = 48611,
  [SMALL_STATE(1333)] = 48638,
  [SMALL_STATE(1334)] = 48664,
  [SMALL_STATE(1335)] = 48718,
  [SMALL_STATE(1336)] = 48744,
  [SMALL_STATE(1337)] = 48770,
  [SMALL_STATE(1338)] = 48804,
  [SMALL_STATE(1339)] = 48830,
  [SMALL_STATE(1340)] = 48884,
  [SMALL_STATE(1341)] = 48908,
  [SMALL_STATE(1342)] = 48933,
  [SMALL_STATE(1343)] = 48958,
  [SMALL_STATE(1344)] = 48981,
  [SMALL_STATE(1345)] = 49006,
  [SMALL_STATE(1346)] = 49031,
  [SMALL_STATE(1347)] = 49056,
  [SMALL_STATE(1348)] = 49081,
  [SMALL_STATE(1349)] = 49112,
  [SMALL_STATE(1350)] = 49136,
  [SMALL_STATE(1351)] = 49158,
  [SMALL_STATE(1352)] = 49182,
  [SMALL_STATE(1353)] = 49206,
  [SMALL_STATE(1354)] = 49230,
  [SMALL_STATE(1355)] = 49252,
  [SMALL_STATE(1356)] = 49278,
  [SMALL_STATE(1357)] = 49302,
  [SMALL_STATE(1358)] = 49326,
  [SMALL_STATE(1359)] = 49350,
  [SMALL_STATE(1360)] = 49376,
  [SMALL_STATE(1361)] = 49398,
  [SMALL_STATE(1362)] = 49422,
  [SMALL_STATE(1363)] = 49448,
  [SMALL_STATE(1364)] = 49472,
  [SMALL_STATE(1365)] = 49500,
  [SMALL_STATE(1366)] = 49522,
  [SMALL_STATE(1367)] = 49546,
  [SMALL_STATE(1368)] = 49568,
  [SMALL_STATE(1369)] = 49594,
  [SMALL_STATE(1370)] = 49638,
  [SMALL_STATE(1371)] = 49660,
  [SMALL_STATE(1372)] = 49682,
  [SMALL_STATE(1373)] = 49704,
  [SMALL_STATE(1374)] = 49750,
  [SMALL_STATE(1375)] = 49776,
  [SMALL_STATE(1376)] = 49820,
  [SMALL_STATE(1377)] = 49841,
  [SMALL_STATE(1378)] = 49862,
  [SMALL_STATE(1379)] = 49883,
  [SMALL_STATE(1380)] = 49904,
  [SMALL_STATE(1381)] = 49949,
  [SMALL_STATE(1382)] = 49970,
  [SMALL_STATE(1383)] = 49991,
  [SMALL_STATE(1384)] = 50012,
  [SMALL_STATE(1385)] = 50033,
  [SMALL_STATE(1386)] = 50054,
  [SMALL_STATE(1387)] = 50075,
  [SMALL_STATE(1388)] = 50096,
  [SMALL_STATE(1389)] = 50117,
  [SMALL_STATE(1390)] = 50142,
  [SMALL_STATE(1391)] = 50163,
  [SMALL_STATE(1392)] = 50186,
  [SMALL_STATE(1393)] = 50207,
  [SMALL_STATE(1394)] = 50230,
  [SMALL_STATE(1395)] = 50251,
  [SMALL_STATE(1396)] = 50282,
  [SMALL_STATE(1397)] = 50307,
  [SMALL_STATE(1398)] = 50330,
  [SMALL_STATE(1399)] = 50351,
  [SMALL_STATE(1400)] = 50372,
  [SMALL_STATE(1401)] = 50393,
  [SMALL_STATE(1402)] = 50414,
  [SMALL_STATE(1403)] = 50435,
  [SMALL_STATE(1404)] = 50456,
  [SMALL_STATE(1405)] = 50477,
  [SMALL_STATE(1406)] = 50498,
  [SMALL_STATE(1407)] = 50522,
  [SMALL_STATE(1408)] = 50544,
  [SMALL_STATE(1409)] = 50572,
  [SMALL_STATE(1410)] = 50596,
  [SMALL_STATE(1411)] = 50620,
  [SMALL_STATE(1412)] = 50644,
  [SMALL_STATE(1413)] = 50668,
  [SMALL_STATE(1414)] = 50692,
  [SMALL_STATE(1415)] = 50714,
  [SMALL_STATE(1416)] = 50736,
  [SMALL_STATE(1417)] = 50760,
  [SMALL_STATE(1418)] = 50784,
  [SMALL_STATE(1419)] = 50805,
  [SMALL_STATE(1420)] = 50828,
  [SMALL_STATE(1421)] = 50849,
  [SMALL_STATE(1422)] = 50870,
  [SMALL_STATE(1423)] = 50891,
  [SMALL_STATE(1424)] = 50916,
  [SMALL_STATE(1425)] = 50939,
  [SMALL_STATE(1426)] = 50960,
  [SMALL_STATE(1427)] = 50981,
  [SMALL_STATE(1428)] = 51002,
  [SMALL_STATE(1429)] = 51023,
  [SMALL_STATE(1430)] = 51044,
  [SMALL_STATE(1431)] = 51067,
  [SMALL_STATE(1432)] = 51092,
  [SMALL_STATE(1433)] = 51113,
  [SMALL_STATE(1434)] = 51138,
  [SMALL_STATE(1435)] = 51159,
  [SMALL_STATE(1436)] = 51182,
  [SMALL_STATE(1437)] = 51203,
  [SMALL_STATE(1438)] = 51224,
  [SMALL_STATE(1439)] = 51245,
  [SMALL_STATE(1440)] = 51274,
  [SMALL_STATE(1441)] = 51295,
  [SMALL_STATE(1442)] = 51316,
  [SMALL_STATE(1443)] = 51337,
  [SMALL_STATE(1444)] = 51358,
  [SMALL_STATE(1445)] = 51379,
  [SMALL_STATE(1446)] = 51404,
  [SMALL_STATE(1447)] = 51425,
  [SMALL_STATE(1448)] = 51446,
  [SMALL_STATE(1449)] = 51467,
  [SMALL_STATE(1450)] = 51492,
  [SMALL_STATE(1451)] = 51513,
  [SMALL_STATE(1452)] = 51534,
  [SMALL_STATE(1453)] = 51555,
  [SMALL_STATE(1454)] = 51576,
  [SMALL_STATE(1455)] = 51597,
  [SMALL_STATE(1456)] = 51618,
  [SMALL_STATE(1457)] = 51639,
  [SMALL_STATE(1458)] = 51664,
  [SMALL_STATE(1459)] = 51685,
  [SMALL_STATE(1460)] = 51710,
  [SMALL_STATE(1461)] = 51731,
  [SMALL_STATE(1462)] = 51756,
  [SMALL_STATE(1463)] = 51788,
  [SMALL_STATE(1464)] = 51820,
  [SMALL_STATE(1465)] = 51844,
  [SMALL_STATE(1466)] = 51876,
  [SMALL_STATE(1467)] = 51908,
  [SMALL_STATE(1468)] = 51932,
  [SMALL_STATE(1469)] = 51964,
  [SMALL_STATE(1470)] = 51996,
  [SMALL_STATE(1471)] = 52028,
  [SMALL_STATE(1472)] = 52052,
  [SMALL_STATE(1473)] = 52076,
  [SMALL_STATE(1474)] = 52108,
  [SMALL_STATE(1475)] = 52135,
  [SMALL_STATE(1476)] = 52164,
  [SMALL_STATE(1477)] = 52189,
  [SMALL_STATE(1478)] = 52220,
  [SMALL_STATE(1479)] = 52253,
  [SMALL_STATE(1480)] = 52286,
  [SMALL_STATE(1481)] = 52319,
  [SMALL_STATE(1482)] = 52352,
  [SMALL_STATE(1483)] = 52378,
  [SMALL_STATE(1484)] = 52404,
  [SMALL_STATE(1485)] = 52426,
  [SMALL_STATE(1486)] = 52456,
  [SMALL_STATE(1487)] = 52488,
  [SMALL_STATE(1488)] = 52518,
  [SMALL_STATE(1489)] = 52548,
  [SMALL_STATE(1490)] = 52578,
  [SMALL_STATE(1491)] = 52608,
  [SMALL_STATE(1492)] = 52634,
  [SMALL_STATE(1493)] = 52666,
  [SMALL_STATE(1494)] = 52692,
  [SMALL_STATE(1495)] = 52722,
  [SMALL_STATE(1496)] = 52752,
  [SMALL_STATE(1497)] = 52784,
  [SMALL_STATE(1498)] = 52810,
  [SMALL_STATE(1499)] = 52840,
  [SMALL_STATE(1500)] = 52866,
  [SMALL_STATE(1501)] = 52896,
  [SMALL_STATE(1502)] = 52918,
  [SMALL_STATE(1503)] = 52948,
  [SMALL_STATE(1504)] = 52978,
  [SMALL_STATE(1505)] = 53004,
  [SMALL_STATE(1506)] = 53030,
  [SMALL_STATE(1507)] = 53056,
  [SMALL_STATE(1508)] = 53082,
  [SMALL_STATE(1509)] = 53112,
  [SMALL_STATE(1510)] = 53142,
  [SMALL_STATE(1511)] = 53164,
  [SMALL_STATE(1512)] = 53186,
  [SMALL_STATE(1513)] = 53216,
  [SMALL_STATE(1514)] = 53246,
  [SMALL_STATE(1515)] = 53276,
  [SMALL_STATE(1516)] = 53306,
  [SMALL_STATE(1517)] = 53336,
  [SMALL_STATE(1518)] = 53358,
  [SMALL_STATE(1519)] = 53388,
  [SMALL_STATE(1520)] = 53418,
  [SMALL_STATE(1521)] = 53450,
  [SMALL_STATE(1522)] = 53480,
  [SMALL_STATE(1523)] = 53506,
  [SMALL_STATE(1524)] = 53528,
  [SMALL_STATE(1525)] = 53555,
  [SMALL_STATE(1526)] = 53576,
  [SMALL_STATE(1527)] = 53597,
  [SMALL_STATE(1528)] = 53624,
  [SMALL_STATE(1529)] = 53647,
  [SMALL_STATE(1530)] = 53674,
  [SMALL_STATE(1531)] = 53689,
  [SMALL_STATE(1532)] = 53718,
  [SMALL_STATE(1533)] = 53747,
  [SMALL_STATE(1534)] = 53776,
  [SMALL_STATE(1535)] = 53805,
  [SMALL_STATE(1536)] = 53832,
  [SMALL_STATE(1537)] = 53859,
  [SMALL_STATE(1538)] = 53878,
  [SMALL_STATE(1539)] = 53897,
  [SMALL_STATE(1540)] = 53916,
  [SMALL_STATE(1541)] = 53935,
  [SMALL_STATE(1542)] = 53954,
  [SMALL_STATE(1543)] = 53981,
  [SMALL_STATE(1544)] = 54008,
  [SMALL_STATE(1545)] = 54035,
  [SMALL_STATE(1546)] = 54058,
  [SMALL_STATE(1547)] = 54083,
  [SMALL_STATE(1548)] = 54102,
  [SMALL_STATE(1549)] = 54125,
  [SMALL_STATE(1550)] = 54144,
  [SMALL_STATE(1551)] = 54163,
  [SMALL_STATE(1552)] = 54190,
  [SMALL_STATE(1553)] = 54213,
  [SMALL_STATE(1554)] = 54232,
  [SMALL_STATE(1555)] = 54259,
  [SMALL_STATE(1556)] = 54288,
  [SMALL_STATE(1557)] = 54315,
  [SMALL_STATE(1558)] = 54338,
  [SMALL_STATE(1559)] = 54365,
  [SMALL_STATE(1560)] = 54392,
  [SMALL_STATE(1561)] = 54419,
  [SMALL_STATE(1562)] = 54438,
  [SMALL_STATE(1563)] = 54465,
  [SMALL_STATE(1564)] = 54494,
  [SMALL_STATE(1565)] = 54513,
  [SMALL_STATE(1566)] = 54537,
  [SMALL_STATE(1567)] = 54559,
  [SMALL_STATE(1568)] = 54575,
  [SMALL_STATE(1569)] = 54589,
  [SMALL_STATE(1570)] = 54605,
  [SMALL_STATE(1571)] = 54619,
  [SMALL_STATE(1572)] = 54645,
  [SMALL_STATE(1573)] = 54659,
  [SMALL_STATE(1574)] = 54675,
  [SMALL_STATE(1575)] = 54699,
  [SMALL_STATE(1576)] = 54713,
  [SMALL_STATE(1577)] = 54739,
  [SMALL_STATE(1578)] = 54755,
  [SMALL_STATE(1579)] = 54781,
  [SMALL_STATE(1580)] = 54807,
  [SMALL_STATE(1581)] = 54833,
  [SMALL_STATE(1582)] = 54859,
  [SMALL_STATE(1583)] = 54885,
  [SMALL_STATE(1584)] = 54901,
  [SMALL_STATE(1585)] = 54927,
  [SMALL_STATE(1586)] = 54943,
  [SMALL_STATE(1587)] = 54959,
  [SMALL_STATE(1588)] = 54985,
  [SMALL_STATE(1589)] = 55011,
  [SMALL_STATE(1590)] = 55035,
  [SMALL_STATE(1591)] = 55061,
  [SMALL_STATE(1592)] = 55075,
  [SMALL_STATE(1593)] = 55101,
  [SMALL_STATE(1594)] = 55115,
  [SMALL_STATE(1595)] = 55141,
  [SMALL_STATE(1596)] = 55155,
  [SMALL_STATE(1597)] = 55175,
  [SMALL_STATE(1598)] = 55201,
  [SMALL_STATE(1599)] = 55227,
  [SMALL_STATE(1600)] = 55253,
  [SMALL_STATE(1601)] = 55269,
  [SMALL_STATE(1602)] = 55295,
  [SMALL_STATE(1603)] = 55321,
  [SMALL_STATE(1604)] = 55347,
  [SMALL_STATE(1605)] = 55371,
  [SMALL_STATE(1606)] = 55397,
  [SMALL_STATE(1607)] = 55419,
  [SMALL_STATE(1608)] = 55443,
  [SMALL_STATE(1609)] = 55469,
  [SMALL_STATE(1610)] = 55495,
  [SMALL_STATE(1611)] = 55521,
  [SMALL_STATE(1612)] = 55547,
  [SMALL_STATE(1613)] = 55570,
  [SMALL_STATE(1614)] = 55593,
  [SMALL_STATE(1615)] = 55616,
  [SMALL_STATE(1616)] = 55631,
  [SMALL_STATE(1617)] = 55654,
  [SMALL_STATE(1618)] = 55677,
  [SMALL_STATE(1619)] = 55700,
  [SMALL_STATE(1620)] = 55723,
  [SMALL_STATE(1621)] = 55746,
  [SMALL_STATE(1622)] = 55769,
  [SMALL_STATE(1623)] = 55792,
  [SMALL_STATE(1624)] = 55815,
  [SMALL_STATE(1625)] = 55838,
  [SMALL_STATE(1626)] = 55861,
  [SMALL_STATE(1627)] = 55884,
  [SMALL_STATE(1628)] = 55907,
  [SMALL_STATE(1629)] = 55924,
  [SMALL_STATE(1630)] = 55947,
  [SMALL_STATE(1631)] = 55970,
  [SMALL_STATE(1632)] = 55993,
  [SMALL_STATE(1633)] = 56016,
  [SMALL_STATE(1634)] = 56039,
  [SMALL_STATE(1635)] = 56060,
  [SMALL_STATE(1636)] = 56083,
  [SMALL_STATE(1637)] = 56106,
  [SMALL_STATE(1638)] = 56123,
  [SMALL_STATE(1639)] = 56146,
  [SMALL_STATE(1640)] = 56163,
  [SMALL_STATE(1641)] = 56186,
  [SMALL_STATE(1642)] = 56209,
  [SMALL_STATE(1643)] = 56232,
  [SMALL_STATE(1644)] = 56249,
  [SMALL_STATE(1645)] = 56272,
  [SMALL_STATE(1646)] = 56295,
  [SMALL_STATE(1647)] = 56318,
  [SMALL_STATE(1648)] = 56341,
  [SMALL_STATE(1649)] = 56364,
  [SMALL_STATE(1650)] = 56387,
  [SMALL_STATE(1651)] = 56404,
  [SMALL_STATE(1652)] = 56421,
  [SMALL_STATE(1653)] = 56444,
  [SMALL_STATE(1654)] = 56465,
  [SMALL_STATE(1655)] = 56488,
  [SMALL_STATE(1656)] = 56511,
  [SMALL_STATE(1657)] = 56534,
  [SMALL_STATE(1658)] = 56557,
  [SMALL_STATE(1659)] = 56576,
  [SMALL_STATE(1660)] = 56593,
  [SMALL_STATE(1661)] = 56616,
  [SMALL_STATE(1662)] = 56639,
  [SMALL_STATE(1663)] = 56662,
  [SMALL_STATE(1664)] = 56685,
  [SMALL_STATE(1665)] = 56708,
  [SMALL_STATE(1666)] = 56727,
  [SMALL_STATE(1667)] = 56744,
  [SMALL_STATE(1668)] = 56767,
  [SMALL_STATE(1669)] = 56790,
  [SMALL_STATE(1670)] = 56813,
  [SMALL_STATE(1671)] = 56836,
  [SMALL_STATE(1672)] = 56859,
  [SMALL_STATE(1673)] = 56882,
  [SMALL_STATE(1674)] = 56899,
  [SMALL_STATE(1675)] = 56922,
  [SMALL_STATE(1676)] = 56945,
  [SMALL_STATE(1677)] = 56968,
  [SMALL_STATE(1678)] = 56991,
  [SMALL_STATE(1679)] = 57014,
  [SMALL_STATE(1680)] = 57031,
  [SMALL_STATE(1681)] = 57054,
  [SMALL_STATE(1682)] = 57071,
  [SMALL_STATE(1683)] = 57088,
  [SMALL_STATE(1684)] = 57111,
  [SMALL_STATE(1685)] = 57134,
  [SMALL_STATE(1686)] = 57157,
  [SMALL_STATE(1687)] = 57174,
  [SMALL_STATE(1688)] = 57197,
  [SMALL_STATE(1689)] = 57220,
  [SMALL_STATE(1690)] = 57243,
  [SMALL_STATE(1691)] = 57266,
  [SMALL_STATE(1692)] = 57281,
  [SMALL_STATE(1693)] = 57304,
  [SMALL_STATE(1694)] = 57327,
  [SMALL_STATE(1695)] = 57350,
  [SMALL_STATE(1696)] = 57371,
  [SMALL_STATE(1697)] = 57388,
  [SMALL_STATE(1698)] = 57411,
  [SMALL_STATE(1699)] = 57434,
  [SMALL_STATE(1700)] = 57453,
  [SMALL_STATE(1701)] = 57470,
  [SMALL_STATE(1702)] = 57493,
  [SMALL_STATE(1703)] = 57516,
  [SMALL_STATE(1704)] = 57539,
  [SMALL_STATE(1705)] = 57562,
  [SMALL_STATE(1706)] = 57574,
  [SMALL_STATE(1707)] = 57590,
  [SMALL_STATE(1708)] = 57608,
  [SMALL_STATE(1709)] = 57626,
  [SMALL_STATE(1710)] = 57638,
  [SMALL_STATE(1711)] = 57652,
  [SMALL_STATE(1712)] = 57672,
  [SMALL_STATE(1713)] = 57692,
  [SMALL_STATE(1714)] = 57704,
  [SMALL_STATE(1715)] = 57716,
  [SMALL_STATE(1716)] = 57736,
  [SMALL_STATE(1717)] = 57750,
  [SMALL_STATE(1718)] = 57768,
  [SMALL_STATE(1719)] = 57784,
  [SMALL_STATE(1720)] = 57796,
  [SMALL_STATE(1721)] = 57812,
  [SMALL_STATE(1722)] = 57828,
  [SMALL_STATE(1723)] = 57840,
  [SMALL_STATE(1724)] = 57858,
  [SMALL_STATE(1725)] = 57870,
  [SMALL_STATE(1726)] = 57882,
  [SMALL_STATE(1727)] = 57894,
  [SMALL_STATE(1728)] = 57906,
  [SMALL_STATE(1729)] = 57922,
  [SMALL_STATE(1730)] = 57934,
  [SMALL_STATE(1731)] = 57950,
  [SMALL_STATE(1732)] = 57968,
  [SMALL_STATE(1733)] = 57980,
  [SMALL_STATE(1734)] = 57992,
  [SMALL_STATE(1735)] = 58004,
  [SMALL_STATE(1736)] = 58024,
  [SMALL_STATE(1737)] = 58044,
  [SMALL_STATE(1738)] = 58056,
  [SMALL_STATE(1739)] = 58072,
  [SMALL_STATE(1740)] = 58084,
  [SMALL_STATE(1741)] = 58104,
  [SMALL_STATE(1742)] = 58116,
  [SMALL_STATE(1743)] = 58128,
  [SMALL_STATE(1744)] = 58144,
  [SMALL_STATE(1745)] = 58164,
  [SMALL_STATE(1746)] = 58176,
  [SMALL_STATE(1747)] = 58196,
  [SMALL_STATE(1748)] = 58210,
  [SMALL_STATE(1749)] = 58230,
  [SMALL_STATE(1750)] = 58246,
  [SMALL_STATE(1751)] = 58262,
  [SMALL_STATE(1752)] = 58282,
  [SMALL_STATE(1753)] = 58294,
  [SMALL_STATE(1754)] = 58306,
  [SMALL_STATE(1755)] = 58322,
  [SMALL_STATE(1756)] = 58340,
  [SMALL_STATE(1757)] = 58356,
  [SMALL_STATE(1758)] = 58371,
  [SMALL_STATE(1759)] = 58388,
  [SMALL_STATE(1760)] = 58405,
  [SMALL_STATE(1761)] = 58422,
  [SMALL_STATE(1762)] = 58439,
  [SMALL_STATE(1763)] = 58454,
  [SMALL_STATE(1764)] = 58469,
  [SMALL_STATE(1765)] = 58486,
  [SMALL_STATE(1766)] = 58503,
  [SMALL_STATE(1767)] = 58516,
  [SMALL_STATE(1768)] = 58533,
  [SMALL_STATE(1769)] = 58550,
  [SMALL_STATE(1770)] = 58567,
  [SMALL_STATE(1771)] = 58580,
  [SMALL_STATE(1772)] = 58597,
  [SMALL_STATE(1773)] = 58614,
  [SMALL_STATE(1774)] = 58629,
  [SMALL_STATE(1775)] = 58644,
  [SMALL_STATE(1776)] = 58661,
  [SMALL_STATE(1777)] = 58678,
  [SMALL_STATE(1778)] = 58695,
  [SMALL_STATE(1779)] = 58712,
  [SMALL_STATE(1780)] = 58729,
  [SMALL_STATE(1781)] = 58744,
  [SMALL_STATE(1782)] = 58761,
  [SMALL_STATE(1783)] = 58778,
  [SMALL_STATE(1784)] = 58795,
  [SMALL_STATE(1785)] = 58812,
  [SMALL_STATE(1786)] = 58829,
  [SMALL_STATE(1787)] = 58846,
  [SMALL_STATE(1788)] = 58863,
  [SMALL_STATE(1789)] = 58880,
  [SMALL_STATE(1790)] = 58897,
  [SMALL_STATE(1791)] = 58912,
  [SMALL_STATE(1792)] = 58929,
  [SMALL_STATE(1793)] = 58946,
  [SMALL_STATE(1794)] = 58961,
  [SMALL_STATE(1795)] = 58976,
  [SMALL_STATE(1796)] = 58993,
  [SMALL_STATE(1797)] = 59008,
  [SMALL_STATE(1798)] = 59025,
  [SMALL_STATE(1799)] = 59042,
  [SMALL_STATE(1800)] = 59059,
  [SMALL_STATE(1801)] = 59076,
  [SMALL_STATE(1802)] = 59091,
  [SMALL_STATE(1803)] = 59108,
  [SMALL_STATE(1804)] = 59125,
  [SMALL_STATE(1805)] = 59138,
  [SMALL_STATE(1806)] = 59153,
  [SMALL_STATE(1807)] = 59168,
  [SMALL_STATE(1808)] = 59185,
  [SMALL_STATE(1809)] = 59202,
  [SMALL_STATE(1810)] = 59215,
  [SMALL_STATE(1811)] = 59232,
  [SMALL_STATE(1812)] = 59249,
  [SMALL_STATE(1813)] = 59266,
  [SMALL_STATE(1814)] = 59283,
  [SMALL_STATE(1815)] = 59300,
  [SMALL_STATE(1816)] = 59317,
  [SMALL_STATE(1817)] = 59334,
  [SMALL_STATE(1818)] = 59351,
  [SMALL_STATE(1819)] = 59368,
  [SMALL_STATE(1820)] = 59381,
  [SMALL_STATE(1821)] = 59394,
  [SMALL_STATE(1822)] = 59411,
  [SMALL_STATE(1823)] = 59428,
  [SMALL_STATE(1824)] = 59445,
  [SMALL_STATE(1825)] = 59460,
  [SMALL_STATE(1826)] = 59475,
  [SMALL_STATE(1827)] = 59490,
  [SMALL_STATE(1828)] = 59507,
  [SMALL_STATE(1829)] = 59524,
  [SMALL_STATE(1830)] = 59539,
  [SMALL_STATE(1831)] = 59556,
  [SMALL_STATE(1832)] = 59573,
  [SMALL_STATE(1833)] = 59588,
  [SMALL_STATE(1834)] = 59605,
  [SMALL_STATE(1835)] = 59622,
  [SMALL_STATE(1836)] = 59639,
  [SMALL_STATE(1837)] = 59654,
  [SMALL_STATE(1838)] = 59671,
  [SMALL_STATE(1839)] = 59686,
  [SMALL_STATE(1840)] = 59701,
  [SMALL_STATE(1841)] = 59718,
  [SMALL_STATE(1842)] = 59735,
  [SMALL_STATE(1843)] = 59748,
  [SMALL_STATE(1844)] = 59763,
  [SMALL_STATE(1845)] = 59780,
  [SMALL_STATE(1846)] = 59797,
  [SMALL_STATE(1847)] = 59810,
  [SMALL_STATE(1848)] = 59825,
  [SMALL_STATE(1849)] = 59842,
  [SMALL_STATE(1850)] = 59857,
  [SMALL_STATE(1851)] = 59874,
  [SMALL_STATE(1852)] = 59891,
  [SMALL_STATE(1853)] = 59908,
  [SMALL_STATE(1854)] = 59923,
  [SMALL_STATE(1855)] = 59940,
  [SMALL_STATE(1856)] = 59954,
  [SMALL_STATE(1857)] = 59968,
  [SMALL_STATE(1858)] = 59982,
  [SMALL_STATE(1859)] = 59996,
  [SMALL_STATE(1860)] = 60006,
  [SMALL_STATE(1861)] = 60016,
  [SMALL_STATE(1862)] = 60030,
  [SMALL_STATE(1863)] = 60044,
  [SMALL_STATE(1864)] = 60058,
  [SMALL_STATE(1865)] = 60068,
  [SMALL_STATE(1866)] = 60082,
  [SMALL_STATE(1867)] = 60094,
  [SMALL_STATE(1868)] = 60106,
  [SMALL_STATE(1869)] = 60120,
  [SMALL_STATE(1870)] = 60134,
  [SMALL_STATE(1871)] = 60146,
  [SMALL_STATE(1872)] = 60156,
  [SMALL_STATE(1873)] = 60166,
  [SMALL_STATE(1874)] = 60180,
  [SMALL_STATE(1875)] = 60194,
  [SMALL_STATE(1876)] = 60208,
  [SMALL_STATE(1877)] = 60222,
  [SMALL_STATE(1878)] = 60236,
  [SMALL_STATE(1879)] = 60250,
  [SMALL_STATE(1880)] = 60262,
  [SMALL_STATE(1881)] = 60276,
  [SMALL_STATE(1882)] = 60290,
  [SMALL_STATE(1883)] = 60304,
  [SMALL_STATE(1884)] = 60316,
  [SMALL_STATE(1885)] = 60330,
  [SMALL_STATE(1886)] = 60344,
  [SMALL_STATE(1887)] = 60358,
  [SMALL_STATE(1888)] = 60372,
  [SMALL_STATE(1889)] = 60386,
  [SMALL_STATE(1890)] = 60398,
  [SMALL_STATE(1891)] = 60410,
  [SMALL_STATE(1892)] = 60424,
  [SMALL_STATE(1893)] = 60438,
  [SMALL_STATE(1894)] = 60450,
  [SMALL_STATE(1895)] = 60464,
  [SMALL_STATE(1896)] = 60478,
  [SMALL_STATE(1897)] = 60492,
  [SMALL_STATE(1898)] = 60506,
  [SMALL_STATE(1899)] = 60520,
  [SMALL_STATE(1900)] = 60530,
  [SMALL_STATE(1901)] = 60544,
  [SMALL_STATE(1902)] = 60558,
  [SMALL_STATE(1903)] = 60572,
  [SMALL_STATE(1904)] = 60586,
  [SMALL_STATE(1905)] = 60600,
  [SMALL_STATE(1906)] = 60614,
  [SMALL_STATE(1907)] = 60628,
  [SMALL_STATE(1908)] = 60642,
  [SMALL_STATE(1909)] = 60652,
  [SMALL_STATE(1910)] = 60666,
  [SMALL_STATE(1911)] = 60680,
  [SMALL_STATE(1912)] = 60694,
  [SMALL_STATE(1913)] = 60708,
  [SMALL_STATE(1914)] = 60722,
  [SMALL_STATE(1915)] = 60734,
  [SMALL_STATE(1916)] = 60746,
  [SMALL_STATE(1917)] = 60760,
  [SMALL_STATE(1918)] = 60772,
  [SMALL_STATE(1919)] = 60786,
  [SMALL_STATE(1920)] = 60800,
  [SMALL_STATE(1921)] = 60814,
  [SMALL_STATE(1922)] = 60828,
  [SMALL_STATE(1923)] = 60840,
  [SMALL_STATE(1924)] = 60854,
  [SMALL_STATE(1925)] = 60868,
  [SMALL_STATE(1926)] = 60882,
  [SMALL_STATE(1927)] = 60896,
  [SMALL_STATE(1928)] = 60910,
  [SMALL_STATE(1929)] = 60922,
  [SMALL_STATE(1930)] = 60936,
  [SMALL_STATE(1931)] = 60950,
  [SMALL_STATE(1932)] = 60964,
  [SMALL_STATE(1933)] = 60978,
  [SMALL_STATE(1934)] = 60988,
  [SMALL_STATE(1935)] = 61000,
  [SMALL_STATE(1936)] = 61012,
  [SMALL_STATE(1937)] = 61024,
  [SMALL_STATE(1938)] = 61038,
  [SMALL_STATE(1939)] = 61052,
  [SMALL_STATE(1940)] = 61064,
  [SMALL_STATE(1941)] = 61076,
  [SMALL_STATE(1942)] = 61090,
  [SMALL_STATE(1943)] = 61104,
  [SMALL_STATE(1944)] = 61118,
  [SMALL_STATE(1945)] = 61132,
  [SMALL_STATE(1946)] = 61144,
  [SMALL_STATE(1947)] = 61158,
  [SMALL_STATE(1948)] = 61170,
  [SMALL_STATE(1949)] = 61184,
  [SMALL_STATE(1950)] = 61198,
  [SMALL_STATE(1951)] = 61212,
  [SMALL_STATE(1952)] = 61226,
  [SMALL_STATE(1953)] = 61240,
  [SMALL_STATE(1954)] = 61250,
  [SMALL_STATE(1955)] = 61264,
  [SMALL_STATE(1956)] = 61278,
  [SMALL_STATE(1957)] = 61292,
  [SMALL_STATE(1958)] = 61306,
  [SMALL_STATE(1959)] = 61320,
  [SMALL_STATE(1960)] = 61332,
  [SMALL_STATE(1961)] = 61346,
  [SMALL_STATE(1962)] = 61360,
  [SMALL_STATE(1963)] = 61370,
  [SMALL_STATE(1964)] = 61384,
  [SMALL_STATE(1965)] = 61394,
  [SMALL_STATE(1966)] = 61406,
  [SMALL_STATE(1967)] = 61420,
  [SMALL_STATE(1968)] = 61434,
  [SMALL_STATE(1969)] = 61444,
  [SMALL_STATE(1970)] = 61458,
  [SMALL_STATE(1971)] = 61468,
  [SMALL_STATE(1972)] = 61478,
  [SMALL_STATE(1973)] = 61492,
  [SMALL_STATE(1974)] = 61506,
  [SMALL_STATE(1975)] = 61520,
  [SMALL_STATE(1976)] = 61532,
  [SMALL_STATE(1977)] = 61546,
  [SMALL_STATE(1978)] = 61560,
  [SMALL_STATE(1979)] = 61574,
  [SMALL_STATE(1980)] = 61588,
  [SMALL_STATE(1981)] = 61602,
  [SMALL_STATE(1982)] = 61616,
  [SMALL_STATE(1983)] = 61630,
  [SMALL_STATE(1984)] = 61644,
  [SMALL_STATE(1985)] = 61658,
  [SMALL_STATE(1986)] = 61672,
  [SMALL_STATE(1987)] = 61686,
  [SMALL_STATE(1988)] = 61700,
  [SMALL_STATE(1989)] = 61712,
  [SMALL_STATE(1990)] = 61726,
  [SMALL_STATE(1991)] = 61740,
  [SMALL_STATE(1992)] = 61754,
  [SMALL_STATE(1993)] = 61768,
  [SMALL_STATE(1994)] = 61782,
  [SMALL_STATE(1995)] = 61796,
  [SMALL_STATE(1996)] = 61810,
  [SMALL_STATE(1997)] = 61824,
  [SMALL_STATE(1998)] = 61838,
  [SMALL_STATE(1999)] = 61852,
  [SMALL_STATE(2000)] = 61866,
  [SMALL_STATE(2001)] = 61880,
  [SMALL_STATE(2002)] = 61892,
  [SMALL_STATE(2003)] = 61906,
  [SMALL_STATE(2004)] = 61920,
  [SMALL_STATE(2005)] = 61930,
  [SMALL_STATE(2006)] = 61944,
  [SMALL_STATE(2007)] = 61958,
  [SMALL_STATE(2008)] = 61972,
  [SMALL_STATE(2009)] = 61986,
  [SMALL_STATE(2010)] = 62000,
  [SMALL_STATE(2011)] = 62014,
  [SMALL_STATE(2012)] = 62026,
  [SMALL_STATE(2013)] = 62040,
  [SMALL_STATE(2014)] = 62054,
  [SMALL_STATE(2015)] = 62068,
  [SMALL_STATE(2016)] = 62082,
  [SMALL_STATE(2017)] = 62096,
  [SMALL_STATE(2018)] = 62110,
  [SMALL_STATE(2019)] = 62124,
  [SMALL_STATE(2020)] = 62138,
  [SMALL_STATE(2021)] = 62152,
  [SMALL_STATE(2022)] = 62166,
  [SMALL_STATE(2023)] = 62180,
  [SMALL_STATE(2024)] = 62194,
  [SMALL_STATE(2025)] = 62208,
  [SMALL_STATE(2026)] = 62222,
  [SMALL_STATE(2027)] = 62236,
  [SMALL_STATE(2028)] = 62250,
  [SMALL_STATE(2029)] = 62264,
  [SMALL_STATE(2030)] = 62278,
  [SMALL_STATE(2031)] = 62292,
  [SMALL_STATE(2032)] = 62306,
  [SMALL_STATE(2033)] = 62320,
  [SMALL_STATE(2034)] = 62334,
  [SMALL_STATE(2035)] = 62348,
  [SMALL_STATE(2036)] = 62362,
  [SMALL_STATE(2037)] = 62376,
  [SMALL_STATE(2038)] = 62390,
  [SMALL_STATE(2039)] = 62402,
  [SMALL_STATE(2040)] = 62412,
  [SMALL_STATE(2041)] = 62426,
  [SMALL_STATE(2042)] = 62438,
  [SMALL_STATE(2043)] = 62452,
  [SMALL_STATE(2044)] = 62466,
  [SMALL_STATE(2045)] = 62476,
  [SMALL_STATE(2046)] = 62488,
  [SMALL_STATE(2047)] = 62502,
  [SMALL_STATE(2048)] = 62516,
  [SMALL_STATE(2049)] = 62528,
  [SMALL_STATE(2050)] = 62538,
  [SMALL_STATE(2051)] = 62552,
  [SMALL_STATE(2052)] = 62564,
  [SMALL_STATE(2053)] = 62578,
  [SMALL_STATE(2054)] = 62592,
  [SMALL_STATE(2055)] = 62602,
  [SMALL_STATE(2056)] = 62616,
  [SMALL_STATE(2057)] = 62626,
  [SMALL_STATE(2058)] = 62638,
  [SMALL_STATE(2059)] = 62652,
  [SMALL_STATE(2060)] = 62662,
  [SMALL_STATE(2061)] = 62672,
  [SMALL_STATE(2062)] = 62686,
  [SMALL_STATE(2063)] = 62700,
  [SMALL_STATE(2064)] = 62714,
  [SMALL_STATE(2065)] = 62728,
  [SMALL_STATE(2066)] = 62742,
  [SMALL_STATE(2067)] = 62756,
  [SMALL_STATE(2068)] = 62770,
  [SMALL_STATE(2069)] = 62784,
  [SMALL_STATE(2070)] = 62798,
  [SMALL_STATE(2071)] = 62812,
  [SMALL_STATE(2072)] = 62822,
  [SMALL_STATE(2073)] = 62836,
  [SMALL_STATE(2074)] = 62850,
  [SMALL_STATE(2075)] = 62860,
  [SMALL_STATE(2076)] = 62874,
  [SMALL_STATE(2077)] = 62888,
  [SMALL_STATE(2078)] = 62902,
  [SMALL_STATE(2079)] = 62912,
  [SMALL_STATE(2080)] = 62924,
  [SMALL_STATE(2081)] = 62938,
  [SMALL_STATE(2082)] = 62952,
  [SMALL_STATE(2083)] = 62966,
  [SMALL_STATE(2084)] = 62980,
  [SMALL_STATE(2085)] = 62994,
  [SMALL_STATE(2086)] = 63008,
  [SMALL_STATE(2087)] = 63022,
  [SMALL_STATE(2088)] = 63036,
  [SMALL_STATE(2089)] = 63048,
  [SMALL_STATE(2090)] = 63062,
  [SMALL_STATE(2091)] = 63076,
  [SMALL_STATE(2092)] = 63090,
  [SMALL_STATE(2093)] = 63104,
  [SMALL_STATE(2094)] = 63118,
  [SMALL_STATE(2095)] = 63132,
  [SMALL_STATE(2096)] = 63146,
  [SMALL_STATE(2097)] = 63160,
  [SMALL_STATE(2098)] = 63174,
  [SMALL_STATE(2099)] = 63188,
  [SMALL_STATE(2100)] = 63200,
  [SMALL_STATE(2101)] = 63214,
  [SMALL_STATE(2102)] = 63228,
  [SMALL_STATE(2103)] = 63242,
  [SMALL_STATE(2104)] = 63256,
  [SMALL_STATE(2105)] = 63270,
  [SMALL_STATE(2106)] = 63284,
  [SMALL_STATE(2107)] = 63298,
  [SMALL_STATE(2108)] = 63309,
  [SMALL_STATE(2109)] = 63320,
  [SMALL_STATE(2110)] = 63331,
  [SMALL_STATE(2111)] = 63342,
  [SMALL_STATE(2112)] = 63353,
  [SMALL_STATE(2113)] = 63364,
  [SMALL_STATE(2114)] = 63375,
  [SMALL_STATE(2115)] = 63386,
  [SMALL_STATE(2116)] = 63397,
  [SMALL_STATE(2117)] = 63408,
  [SMALL_STATE(2118)] = 63419,
  [SMALL_STATE(2119)] = 63430,
  [SMALL_STATE(2120)] = 63441,
  [SMALL_STATE(2121)] = 63452,
  [SMALL_STATE(2122)] = 63463,
  [SMALL_STATE(2123)] = 63474,
  [SMALL_STATE(2124)] = 63483,
  [SMALL_STATE(2125)] = 63494,
  [SMALL_STATE(2126)] = 63505,
  [SMALL_STATE(2127)] = 63516,
  [SMALL_STATE(2128)] = 63527,
  [SMALL_STATE(2129)] = 63538,
  [SMALL_STATE(2130)] = 63549,
  [SMALL_STATE(2131)] = 63560,
  [SMALL_STATE(2132)] = 63571,
  [SMALL_STATE(2133)] = 63582,
  [SMALL_STATE(2134)] = 63593,
  [SMALL_STATE(2135)] = 63604,
  [SMALL_STATE(2136)] = 63615,
  [SMALL_STATE(2137)] = 63626,
  [SMALL_STATE(2138)] = 63637,
  [SMALL_STATE(2139)] = 63646,
  [SMALL_STATE(2140)] = 63657,
  [SMALL_STATE(2141)] = 63668,
  [SMALL_STATE(2142)] = 63679,
  [SMALL_STATE(2143)] = 63690,
  [SMALL_STATE(2144)] = 63701,
  [SMALL_STATE(2145)] = 63712,
  [SMALL_STATE(2146)] = 63723,
  [SMALL_STATE(2147)] = 63734,
  [SMALL_STATE(2148)] = 63745,
  [SMALL_STATE(2149)] = 63756,
  [SMALL_STATE(2150)] = 63767,
  [SMALL_STATE(2151)] = 63778,
  [SMALL_STATE(2152)] = 63789,
  [SMALL_STATE(2153)] = 63800,
  [SMALL_STATE(2154)] = 63811,
  [SMALL_STATE(2155)] = 63822,
  [SMALL_STATE(2156)] = 63831,
  [SMALL_STATE(2157)] = 63842,
  [SMALL_STATE(2158)] = 63853,
  [SMALL_STATE(2159)] = 63864,
  [SMALL_STATE(2160)] = 63875,
  [SMALL_STATE(2161)] = 63886,
  [SMALL_STATE(2162)] = 63897,
  [SMALL_STATE(2163)] = 63908,
  [SMALL_STATE(2164)] = 63919,
  [SMALL_STATE(2165)] = 63928,
  [SMALL_STATE(2166)] = 63939,
  [SMALL_STATE(2167)] = 63950,
  [SMALL_STATE(2168)] = 63961,
  [SMALL_STATE(2169)] = 63972,
  [SMALL_STATE(2170)] = 63983,
  [SMALL_STATE(2171)] = 63994,
  [SMALL_STATE(2172)] = 64003,
  [SMALL_STATE(2173)] = 64014,
  [SMALL_STATE(2174)] = 64025,
  [SMALL_STATE(2175)] = 64036,
  [SMALL_STATE(2176)] = 64047,
  [SMALL_STATE(2177)] = 64058,
  [SMALL_STATE(2178)] = 64069,
  [SMALL_STATE(2179)] = 64080,
  [SMALL_STATE(2180)] = 64091,
  [SMALL_STATE(2181)] = 64102,
  [SMALL_STATE(2182)] = 64113,
  [SMALL_STATE(2183)] = 64124,
  [SMALL_STATE(2184)] = 64135,
  [SMALL_STATE(2185)] = 64146,
  [SMALL_STATE(2186)] = 64157,
  [SMALL_STATE(2187)] = 64168,
  [SMALL_STATE(2188)] = 64179,
  [SMALL_STATE(2189)] = 64190,
  [SMALL_STATE(2190)] = 64201,
  [SMALL_STATE(2191)] = 64212,
  [SMALL_STATE(2192)] = 64223,
  [SMALL_STATE(2193)] = 64234,
  [SMALL_STATE(2194)] = 64245,
  [SMALL_STATE(2195)] = 64256,
  [SMALL_STATE(2196)] = 64267,
  [SMALL_STATE(2197)] = 64278,
  [SMALL_STATE(2198)] = 64289,
  [SMALL_STATE(2199)] = 64300,
  [SMALL_STATE(2200)] = 64311,
  [SMALL_STATE(2201)] = 64322,
  [SMALL_STATE(2202)] = 64333,
  [SMALL_STATE(2203)] = 64344,
  [SMALL_STATE(2204)] = 64355,
  [SMALL_STATE(2205)] = 64366,
  [SMALL_STATE(2206)] = 64377,
  [SMALL_STATE(2207)] = 64388,
  [SMALL_STATE(2208)] = 64397,
  [SMALL_STATE(2209)] = 64406,
  [SMALL_STATE(2210)] = 64415,
  [SMALL_STATE(2211)] = 64426,
  [SMALL_STATE(2212)] = 64437,
  [SMALL_STATE(2213)] = 64448,
  [SMALL_STATE(2214)] = 64459,
  [SMALL_STATE(2215)] = 64468,
  [SMALL_STATE(2216)] = 64479,
  [SMALL_STATE(2217)] = 64490,
  [SMALL_STATE(2218)] = 64501,
  [SMALL_STATE(2219)] = 64512,
  [SMALL_STATE(2220)] = 64523,
  [SMALL_STATE(2221)] = 64534,
  [SMALL_STATE(2222)] = 64545,
  [SMALL_STATE(2223)] = 64556,
  [SMALL_STATE(2224)] = 64567,
  [SMALL_STATE(2225)] = 64578,
  [SMALL_STATE(2226)] = 64589,
  [SMALL_STATE(2227)] = 64600,
  [SMALL_STATE(2228)] = 64611,
  [SMALL_STATE(2229)] = 64622,
  [SMALL_STATE(2230)] = 64631,
  [SMALL_STATE(2231)] = 64642,
  [SMALL_STATE(2232)] = 64653,
  [SMALL_STATE(2233)] = 64664,
  [SMALL_STATE(2234)] = 64675,
  [SMALL_STATE(2235)] = 64686,
  [SMALL_STATE(2236)] = 64697,
  [SMALL_STATE(2237)] = 64708,
  [SMALL_STATE(2238)] = 64719,
  [SMALL_STATE(2239)] = 64728,
  [SMALL_STATE(2240)] = 64737,
  [SMALL_STATE(2241)] = 64748,
  [SMALL_STATE(2242)] = 64759,
  [SMALL_STATE(2243)] = 64770,
  [SMALL_STATE(2244)] = 64781,
  [SMALL_STATE(2245)] = 64792,
  [SMALL_STATE(2246)] = 64803,
  [SMALL_STATE(2247)] = 64814,
  [SMALL_STATE(2248)] = 64825,
  [SMALL_STATE(2249)] = 64834,
  [SMALL_STATE(2250)] = 64845,
  [SMALL_STATE(2251)] = 64856,
  [SMALL_STATE(2252)] = 64867,
  [SMALL_STATE(2253)] = 64878,
  [SMALL_STATE(2254)] = 64889,
  [SMALL_STATE(2255)] = 64898,
  [SMALL_STATE(2256)] = 64909,
  [SMALL_STATE(2257)] = 64920,
  [SMALL_STATE(2258)] = 64931,
  [SMALL_STATE(2259)] = 64942,
  [SMALL_STATE(2260)] = 64953,
  [SMALL_STATE(2261)] = 64962,
  [SMALL_STATE(2262)] = 64973,
  [SMALL_STATE(2263)] = 64982,
  [SMALL_STATE(2264)] = 64991,
  [SMALL_STATE(2265)] = 65000,
  [SMALL_STATE(2266)] = 65009,
  [SMALL_STATE(2267)] = 65020,
  [SMALL_STATE(2268)] = 65031,
  [SMALL_STATE(2269)] = 65042,
  [SMALL_STATE(2270)] = 65053,
  [SMALL_STATE(2271)] = 65064,
  [SMALL_STATE(2272)] = 65073,
  [SMALL_STATE(2273)] = 65084,
  [SMALL_STATE(2274)] = 65095,
  [SMALL_STATE(2275)] = 65104,
  [SMALL_STATE(2276)] = 65115,
  [SMALL_STATE(2277)] = 65126,
  [SMALL_STATE(2278)] = 65137,
  [SMALL_STATE(2279)] = 65148,
  [SMALL_STATE(2280)] = 65159,
  [SMALL_STATE(2281)] = 65170,
  [SMALL_STATE(2282)] = 65181,
  [SMALL_STATE(2283)] = 65192,
  [SMALL_STATE(2284)] = 65203,
  [SMALL_STATE(2285)] = 65212,
  [SMALL_STATE(2286)] = 65223,
  [SMALL_STATE(2287)] = 65232,
  [SMALL_STATE(2288)] = 65241,
  [SMALL_STATE(2289)] = 65252,
  [SMALL_STATE(2290)] = 65263,
  [SMALL_STATE(2291)] = 65274,
  [SMALL_STATE(2292)] = 65285,
  [SMALL_STATE(2293)] = 65296,
  [SMALL_STATE(2294)] = 65307,
  [SMALL_STATE(2295)] = 65318,
  [SMALL_STATE(2296)] = 65329,
  [SMALL_STATE(2297)] = 65340,
  [SMALL_STATE(2298)] = 65351,
  [SMALL_STATE(2299)] = 65362,
  [SMALL_STATE(2300)] = 65373,
  [SMALL_STATE(2301)] = 65384,
  [SMALL_STATE(2302)] = 65395,
  [SMALL_STATE(2303)] = 65406,
  [SMALL_STATE(2304)] = 65415,
  [SMALL_STATE(2305)] = 65424,
  [SMALL_STATE(2306)] = 65435,
  [SMALL_STATE(2307)] = 65446,
  [SMALL_STATE(2308)] = 65457,
  [SMALL_STATE(2309)] = 65468,
  [SMALL_STATE(2310)] = 65479,
  [SMALL_STATE(2311)] = 65490,
  [SMALL_STATE(2312)] = 65501,
  [SMALL_STATE(2313)] = 65512,
  [SMALL_STATE(2314)] = 65521,
  [SMALL_STATE(2315)] = 65529,
  [SMALL_STATE(2316)] = 65537,
  [SMALL_STATE(2317)] = 65545,
  [SMALL_STATE(2318)] = 65553,
  [SMALL_STATE(2319)] = 65561,
  [SMALL_STATE(2320)] = 65569,
  [SMALL_STATE(2321)] = 65577,
  [SMALL_STATE(2322)] = 65585,
  [SMALL_STATE(2323)] = 65593,
  [SMALL_STATE(2324)] = 65601,
  [SMALL_STATE(2325)] = 65609,
  [SMALL_STATE(2326)] = 65617,
  [SMALL_STATE(2327)] = 65625,
  [SMALL_STATE(2328)] = 65633,
  [SMALL_STATE(2329)] = 65641,
  [SMALL_STATE(2330)] = 65649,
  [SMALL_STATE(2331)] = 65657,
  [SMALL_STATE(2332)] = 65665,
  [SMALL_STATE(2333)] = 65673,
  [SMALL_STATE(2334)] = 65681,
  [SMALL_STATE(2335)] = 65689,
  [SMALL_STATE(2336)] = 65697,
  [SMALL_STATE(2337)] = 65705,
  [SMALL_STATE(2338)] = 65713,
  [SMALL_STATE(2339)] = 65721,
  [SMALL_STATE(2340)] = 65729,
  [SMALL_STATE(2341)] = 65737,
  [SMALL_STATE(2342)] = 65745,
  [SMALL_STATE(2343)] = 65753,
  [SMALL_STATE(2344)] = 65761,
  [SMALL_STATE(2345)] = 65769,
  [SMALL_STATE(2346)] = 65777,
  [SMALL_STATE(2347)] = 65785,
  [SMALL_STATE(2348)] = 65793,
  [SMALL_STATE(2349)] = 65801,
  [SMALL_STATE(2350)] = 65809,
  [SMALL_STATE(2351)] = 65817,
  [SMALL_STATE(2352)] = 65825,
  [SMALL_STATE(2353)] = 65833,
  [SMALL_STATE(2354)] = 65841,
  [SMALL_STATE(2355)] = 65849,
  [SMALL_STATE(2356)] = 65857,
  [SMALL_STATE(2357)] = 65865,
  [SMALL_STATE(2358)] = 65873,
  [SMALL_STATE(2359)] = 65881,
  [SMALL_STATE(2360)] = 65889,
  [SMALL_STATE(2361)] = 65897,
  [SMALL_STATE(2362)] = 65905,
  [SMALL_STATE(2363)] = 65913,
  [SMALL_STATE(2364)] = 65921,
  [SMALL_STATE(2365)] = 65929,
  [SMALL_STATE(2366)] = 65937,
  [SMALL_STATE(2367)] = 65945,
  [SMALL_STATE(2368)] = 65953,
  [SMALL_STATE(2369)] = 65961,
  [SMALL_STATE(2370)] = 65969,
  [SMALL_STATE(2371)] = 65977,
  [SMALL_STATE(2372)] = 65985,
  [SMALL_STATE(2373)] = 65993,
  [SMALL_STATE(2374)] = 66001,
  [SMALL_STATE(2375)] = 66009,
  [SMALL_STATE(2376)] = 66017,
  [SMALL_STATE(2377)] = 66025,
  [SMALL_STATE(2378)] = 66033,
  [SMALL_STATE(2379)] = 66041,
  [SMALL_STATE(2380)] = 66049,
  [SMALL_STATE(2381)] = 66057,
  [SMALL_STATE(2382)] = 66065,
  [SMALL_STATE(2383)] = 66073,
  [SMALL_STATE(2384)] = 66081,
  [SMALL_STATE(2385)] = 66089,
  [SMALL_STATE(2386)] = 66097,
  [SMALL_STATE(2387)] = 66105,
  [SMALL_STATE(2388)] = 66113,
  [SMALL_STATE(2389)] = 66121,
  [SMALL_STATE(2390)] = 66129,
  [SMALL_STATE(2391)] = 66137,
  [SMALL_STATE(2392)] = 66145,
  [SMALL_STATE(2393)] = 66153,
  [SMALL_STATE(2394)] = 66161,
  [SMALL_STATE(2395)] = 66169,
  [SMALL_STATE(2396)] = 66177,
  [SMALL_STATE(2397)] = 66185,
  [SMALL_STATE(2398)] = 66193,
  [SMALL_STATE(2399)] = 66201,
  [SMALL_STATE(2400)] = 66209,
  [SMALL_STATE(2401)] = 66217,
  [SMALL_STATE(2402)] = 66225,
  [SMALL_STATE(2403)] = 66233,
  [SMALL_STATE(2404)] = 66241,
  [SMALL_STATE(2405)] = 66249,
  [SMALL_STATE(2406)] = 66257,
  [SMALL_STATE(2407)] = 66265,
  [SMALL_STATE(2408)] = 66273,
  [SMALL_STATE(2409)] = 66281,
  [SMALL_STATE(2410)] = 66289,
  [SMALL_STATE(2411)] = 66297,
  [SMALL_STATE(2412)] = 66305,
  [SMALL_STATE(2413)] = 66313,
  [SMALL_STATE(2414)] = 66321,
  [SMALL_STATE(2415)] = 66329,
  [SMALL_STATE(2416)] = 66337,
  [SMALL_STATE(2417)] = 66345,
  [SMALL_STATE(2418)] = 66353,
  [SMALL_STATE(2419)] = 66361,
  [SMALL_STATE(2420)] = 66369,
  [SMALL_STATE(2421)] = 66377,
  [SMALL_STATE(2422)] = 66385,
  [SMALL_STATE(2423)] = 66393,
  [SMALL_STATE(2424)] = 66401,
  [SMALL_STATE(2425)] = 66409,
  [SMALL_STATE(2426)] = 66417,
  [SMALL_STATE(2427)] = 66425,
  [SMALL_STATE(2428)] = 66433,
  [SMALL_STATE(2429)] = 66441,
  [SMALL_STATE(2430)] = 66449,
  [SMALL_STATE(2431)] = 66457,
  [SMALL_STATE(2432)] = 66465,
  [SMALL_STATE(2433)] = 66473,
  [SMALL_STATE(2434)] = 66481,
  [SMALL_STATE(2435)] = 66489,
  [SMALL_STATE(2436)] = 66497,
  [SMALL_STATE(2437)] = 66505,
  [SMALL_STATE(2438)] = 66513,
  [SMALL_STATE(2439)] = 66521,
  [SMALL_STATE(2440)] = 66529,
  [SMALL_STATE(2441)] = 66537,
  [SMALL_STATE(2442)] = 66545,
  [SMALL_STATE(2443)] = 66553,
  [SMALL_STATE(2444)] = 66561,
  [SMALL_STATE(2445)] = 66569,
  [SMALL_STATE(2446)] = 66577,
  [SMALL_STATE(2447)] = 66585,
  [SMALL_STATE(2448)] = 66593,
  [SMALL_STATE(2449)] = 66601,
  [SMALL_STATE(2450)] = 66609,
  [SMALL_STATE(2451)] = 66617,
  [SMALL_STATE(2452)] = 66625,
  [SMALL_STATE(2453)] = 66633,
  [SMALL_STATE(2454)] = 66641,
  [SMALL_STATE(2455)] = 66649,
  [SMALL_STATE(2456)] = 66657,
  [SMALL_STATE(2457)] = 66665,
  [SMALL_STATE(2458)] = 66673,
  [SMALL_STATE(2459)] = 66681,
  [SMALL_STATE(2460)] = 66689,
  [SMALL_STATE(2461)] = 66697,
  [SMALL_STATE(2462)] = 66705,
  [SMALL_STATE(2463)] = 66713,
  [SMALL_STATE(2464)] = 66721,
  [SMALL_STATE(2465)] = 66729,
  [SMALL_STATE(2466)] = 66737,
  [SMALL_STATE(2467)] = 66745,
  [SMALL_STATE(2468)] = 66753,
  [SMALL_STATE(2469)] = 66761,
  [SMALL_STATE(2470)] = 66769,
  [SMALL_STATE(2471)] = 66777,
  [SMALL_STATE(2472)] = 66785,
  [SMALL_STATE(2473)] = 66793,
  [SMALL_STATE(2474)] = 66801,
  [SMALL_STATE(2475)] = 66809,
  [SMALL_STATE(2476)] = 66817,
  [SMALL_STATE(2477)] = 66825,
  [SMALL_STATE(2478)] = 66833,
  [SMALL_STATE(2479)] = 66841,
  [SMALL_STATE(2480)] = 66849,
  [SMALL_STATE(2481)] = 66857,
  [SMALL_STATE(2482)] = 66865,
  [SMALL_STATE(2483)] = 66873,
  [SMALL_STATE(2484)] = 66881,
  [SMALL_STATE(2485)] = 66889,
  [SMALL_STATE(2486)] = 66897,
  [SMALL_STATE(2487)] = 66905,
  [SMALL_STATE(2488)] = 66913,
  [SMALL_STATE(2489)] = 66921,
  [SMALL_STATE(2490)] = 66929,
  [SMALL_STATE(2491)] = 66937,
  [SMALL_STATE(2492)] = 66945,
  [SMALL_STATE(2493)] = 66953,
  [SMALL_STATE(2494)] = 66961,
  [SMALL_STATE(2495)] = 66969,
  [SMALL_STATE(2496)] = 66977,
  [SMALL_STATE(2497)] = 66985,
  [SMALL_STATE(2498)] = 66993,
  [SMALL_STATE(2499)] = 67001,
  [SMALL_STATE(2500)] = 67009,
  [SMALL_STATE(2501)] = 67017,
  [SMALL_STATE(2502)] = 67025,
  [SMALL_STATE(2503)] = 67033,
  [SMALL_STATE(2504)] = 67041,
  [SMALL_STATE(2505)] = 67049,
  [SMALL_STATE(2506)] = 67057,
  [SMALL_STATE(2507)] = 67065,
  [SMALL_STATE(2508)] = 67073,
  [SMALL_STATE(2509)] = 67081,
  [SMALL_STATE(2510)] = 67089,
  [SMALL_STATE(2511)] = 67097,
  [SMALL_STATE(2512)] = 67105,
  [SMALL_STATE(2513)] = 67113,
  [SMALL_STATE(2514)] = 67121,
  [SMALL_STATE(2515)] = 67129,
  [SMALL_STATE(2516)] = 67137,
  [SMALL_STATE(2517)] = 67145,
  [SMALL_STATE(2518)] = 67153,
  [SMALL_STATE(2519)] = 67161,
  [SMALL_STATE(2520)] = 67169,
  [SMALL_STATE(2521)] = 67177,
  [SMALL_STATE(2522)] = 67185,
  [SMALL_STATE(2523)] = 67193,
  [SMALL_STATE(2524)] = 67201,
  [SMALL_STATE(2525)] = 67209,
  [SMALL_STATE(2526)] = 67217,
  [SMALL_STATE(2527)] = 67225,
  [SMALL_STATE(2528)] = 67233,
  [SMALL_STATE(2529)] = 67241,
  [SMALL_STATE(2530)] = 67249,
  [SMALL_STATE(2531)] = 67257,
  [SMALL_STATE(2532)] = 67265,
  [SMALL_STATE(2533)] = 67273,
  [SMALL_STATE(2534)] = 67281,
  [SMALL_STATE(2535)] = 67289,
  [SMALL_STATE(2536)] = 67297,
  [SMALL_STATE(2537)] = 67305,
  [SMALL_STATE(2538)] = 67313,
  [SMALL_STATE(2539)] = 67321,
  [SMALL_STATE(2540)] = 67329,
  [SMALL_STATE(2541)] = 67337,
  [SMALL_STATE(2542)] = 67345,
  [SMALL_STATE(2543)] = 67353,
  [SMALL_STATE(2544)] = 67361,
  [SMALL_STATE(2545)] = 67369,
  [SMALL_STATE(2546)] = 67377,
  [SMALL_STATE(2547)] = 67385,
  [SMALL_STATE(2548)] = 67393,
  [SMALL_STATE(2549)] = 67401,
  [SMALL_STATE(2550)] = 67409,
  [SMALL_STATE(2551)] = 67417,
  [SMALL_STATE(2552)] = 67425,
  [SMALL_STATE(2553)] = 67433,
  [SMALL_STATE(2554)] = 67441,
  [SMALL_STATE(2555)] = 67449,
  [SMALL_STATE(2556)] = 67457,
  [SMALL_STATE(2557)] = 67465,
  [SMALL_STATE(2558)] = 67473,
  [SMALL_STATE(2559)] = 67481,
  [SMALL_STATE(2560)] = 67489,
  [SMALL_STATE(2561)] = 67497,
  [SMALL_STATE(2562)] = 67505,
  [SMALL_STATE(2563)] = 67513,
  [SMALL_STATE(2564)] = 67521,
};

static const TSParseActionEntry ts_parse_actions[] = {
  [0] = {.entry = {.count = 0, .reusable = false}},
  [1] = {.entry = {.count = 1, .reusable = false}}, RECOVER(),
  [3] = {.entry = {.count = 1, .reusable = true}}, SHIFT_EXTRA(),
  [5] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_source_file, 0),
  [7] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1161),
  [9] = {.entry = {.count = 1, .reusable = true}}, SHIFT(253),
  [11] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2048),
  [13] = {.entry = {.count = 1, .reusable = true}}, SHIFT(42),
  [15] = {.entry = {.count = 1, .reusable = true}}, SHIFT(6),
  [17] = {.entry = {.count = 1, .reusable = true}}, SHIFT(35),
  [19] = {.entry = {.count = 1, .reusable = true}}, SHIFT(148),
  [21] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1119),
  [23] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2564),
  [25] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1523),
  [27] = {.entry = {.count = 1, .reusable = false}}, SHIFT(22),
  [29] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1510),
  [31] = {.entry = {.count = 1, .reusable = false}}, SHIFT(770),
  [33] = {.entry = {.count = 1, .reusable = false}}, SHIFT(765),
  [35] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2563),
  [37] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2214),
  [39] = {.entry = {.count = 1, .reusable = false}}, SHIFT(617),
  [41] = {.entry = {.count = 1, .reusable = false}}, SHIFT(43),
  [43] = {.entry = {.count = 1, .reusable = false}}, SHIFT(628),
  [45] = {.entry = {.count = 1, .reusable = false}}, SHIFT(598),
  [47] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2299),
  [49] = {.entry = {.count = 1, .reusable = false}}, SHIFT(155),
  [51] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2550),
  [53] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1341),
  [55] = {.entry = {.count = 1, .reusable = false}}, SHIFT(24),
  [57] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1921),
  [59] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2541),
  [61] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2540),
  [63] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2535),
  [65] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1166),
  [67] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1476),
  [69] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1300),
  [71] = {.entry = {.count = 1, .reusable = false}}, SHIFT(50),
  [73] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2261),
  [75] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1467),
  [77] = {.entry = {.count = 1, .reusable = true}}, SHIFT(634),
  [79] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2533),
  [81] = {.entry = {.count = 1, .reusable = true}}, SHIFT(91),
  [83] = {.entry = {.count = 1, .reusable = true}}, SHIFT(20),
  [85] = {.entry = {.count = 1, .reusable = true}}, SHIFT(556),
  [87] = {.entry = {.count = 1, .reusable = false}}, SHIFT(23),
  [89] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2259),
  [91] = {.entry = {.count = 1, .reusable = true}}, SHIFT(815),
  [93] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1853),
  [95] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1028),
  [97] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1031),
  [99] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2530),
  [101] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1351),
  [103] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1031),
  [105] = {.entry = {.count = 1, .reusable = true}}, SHIFT(212),
  [107] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1025),
  [109] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1415),
  [111] = {.entry = {.count = 1, .reusable = true}}, SHIFT(245),
  [113] = {.entry = {.count = 1, .reusable = true}}, SHIFT(56),
  [115] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1407),
  [117] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2),
  [119] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(1161),
  [122] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(253),
  [125] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(2048),
  [128] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(42),
  [131] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(6),
  [134] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(35),
  [137] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(148),
  [140] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(1119),
  [143] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(2564),
  [146] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(1523),
  [149] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(22),
  [152] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(1510),
  [155] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(770),
  [158] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(765),
  [161] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(2563),
  [164] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(2214),
  [167] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(617),
  [170] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(43),
  [173] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(628),
  [176] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(598),
  [179] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(2299),
  [182] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(155),
  [185] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(2550),
  [188] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(1341),
  [191] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(24),
  [194] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(1921),
  [197] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(2541),
  [200] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(2540),
  [203] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(2535),
  [206] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(1166),
  [209] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(1476),
  [212] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(1300),
  [215] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(50),
  [218] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(2261),
  [221] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(1467),
  [224] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(634),
  [227] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(2533),
  [230] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(91),
  [233] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(20),
  [236] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(556),
  [239] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(23),
  [242] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(2259),
  [245] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(815),
  [248] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(1853),
  [251] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(1028),
  [254] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(1031),
  [257] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(2530),
  [260] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(1351),
  [263] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_source_file_repeat1, 2), SHIFT_REPEAT(1031),
  [266] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_source_file, 1),
  [268] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1049),
  [270] = {.entry = {.count = 1, .reusable = true}}, SHIFT(330),
  [272] = {.entry = {.count = 1, .reusable = true}}, SHIFT(211),
  [274] = {.entry = {.count = 1, .reusable = true}}, SHIFT(53),
  [276] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1102),
  [278] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1107),
  [280] = {.entry = {.count = 1, .reusable = false}}, SHIFT(758),
  [282] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_break_expression, 2),
  [284] = {.entry = {.count = 1, .reusable = true}}, SHIFT(10),
  [286] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2003),
  [288] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_break_expression, 2),
  [290] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1856),
  [292] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2218),
  [294] = {.entry = {.count = 1, .reusable = false}}, SHIFT(780),
  [296] = {.entry = {.count = 1, .reusable = false}}, SHIFT(597),
  [298] = {.entry = {.count = 1, .reusable = false}}, SHIFT(39),
  [300] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2256),
  [302] = {.entry = {.count = 1, .reusable = false}}, SHIFT(142),
  [304] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2186),
  [306] = {.entry = {.count = 1, .reusable = false}}, SHIFT(41),
  [308] = {.entry = {.count = 1, .reusable = false}}, SHIFT(148),
  [310] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_range_expression, 2),
  [312] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_range_expression, 2),
  [314] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_range_expression, 1),
  [316] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_range_expression, 1),
  [318] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_break_expression, 1),
  [320] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_break_expression, 1),
  [322] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2534),
  [324] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_yield_expression, 1),
  [326] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_yield_expression, 1),
  [328] = {.entry = {.count = 1, .reusable = false}}, SHIFT(634),
  [330] = {.entry = {.count = 1, .reusable = false}}, SHIFT(91),
  [332] = {.entry = {.count = 1, .reusable = false}}, SHIFT(19),
  [334] = {.entry = {.count = 1, .reusable = false}}, SHIFT(556),
  [336] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_return_expression, 1),
  [338] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_return_expression, 1),
  [340] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1171),
  [342] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1194),
  [344] = {.entry = {.count = 1, .reusable = false}}, SHIFT(30),
  [346] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1177),
  [348] = {.entry = {.count = 1, .reusable = false}}, SHIFT(29),
  [350] = {.entry = {.count = 1, .reusable = false}}, SHIFT(161),
  [352] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2531),
  [354] = {.entry = {.count = 1, .reusable = false}}, SHIFT(27),
  [356] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2253),
  [358] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1192),
  [360] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2367),
  [362] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1192),
  [364] = {.entry = {.count = 1, .reusable = false}}, SHIFT(95),
  [366] = {.entry = {.count = 1, .reusable = false}}, SHIFT(32),
  [368] = {.entry = {.count = 1, .reusable = true}}, SHIFT(948),
  [370] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2501),
  [372] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2326),
  [374] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1029),
  [376] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2426),
  [378] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1079),
  [380] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2500),
  [382] = {.entry = {.count = 1, .reusable = false}}, SHIFT(611),
  [384] = {.entry = {.count = 1, .reusable = true}}, SHIFT(161),
  [386] = {.entry = {.count = 1, .reusable = false}}, SHIFT(619),
  [388] = {.entry = {.count = 1, .reusable = true}}, SHIFT(95),
  [390] = {.entry = {.count = 1, .reusable = true}}, SHIFT(28),
  [392] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1087),
  [394] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1060),
  [396] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1046),
  [398] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1051),
  [400] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_expression, 3, .production_id = 18),
  [402] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_expression, 3, .production_id = 18),
  [404] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1826),
  [406] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block, 3),
  [408] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_block, 3),
  [410] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block, 4),
  [412] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_block, 4),
  [414] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_block, 2),
  [416] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_block, 2),
  [418] = {.entry = {.count = 1, .reusable = true}}, SHIFT(32),
  [420] = {.entry = {.count = 1, .reusable = true}}, SHIFT(19),
  [422] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_expression, 3, .production_id = 27),
  [424] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_match_expression, 3, .production_id = 27),
  [426] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_unsafe_block, 2),
  [428] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_unsafe_block, 2),
  [430] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_while_expression, 5, .production_id = 133),
  [432] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_while_expression, 5, .production_id = 133),
  [434] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_loop_expression, 2, .production_id = 2),
  [436] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_loop_expression, 2, .production_id = 2),
  [438] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_delim_token_tree, 2),
  [440] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_delim_token_tree, 2),
  [442] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_expression, 7, .production_id = 218),
  [444] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_expression, 7, .production_id = 218),
  [446] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_else_clause, 2),
  [448] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_else_clause, 2),
  [450] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_const_block, 2, .production_id = 2),
  [452] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_const_block, 2, .production_id = 2),
  [454] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__statement, 1),
  [456] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__statement, 1),
  [458] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expression_except_range, 1),
  [460] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__expression_except_range, 1),
  [462] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_expression, 5, .production_id = 98),
  [464] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_expression, 5, .production_id = 98),
  [466] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expression_statement, 1),
  [468] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_expression_statement, 1),
  [470] = {.entry = {.count = 1, .reusable = true}}, SHIFT(872),
  [472] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_while_expression, 3, .production_id = 31),
  [474] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_while_expression, 3, .production_id = 31),
  [476] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_block, 2),
  [478] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_match_block, 2),
  [480] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_if_expression, 4, .production_id = 59),
  [482] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_if_expression, 4, .production_id = 59),
  [484] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_loop_expression, 4, .production_id = 90),
  [486] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_loop_expression, 4, .production_id = 90),
  [488] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_async_block, 2),
  [490] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_async_block, 2),
  [492] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_block, 3),
  [494] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_match_block, 3),
  [496] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_tuple_expression_repeat1, 2), SHIFT_REPEAT(758),
  [499] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_tuple_expression_repeat1, 2), SHIFT_REPEAT(42),
  [502] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_tuple_expression_repeat1, 2),
  [504] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_tuple_expression_repeat1, 2), SHIFT_REPEAT(10),
  [507] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_tuple_expression_repeat1, 2), SHIFT_REPEAT(35),
  [510] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_tuple_expression_repeat1, 2), SHIFT_REPEAT(148),
  [513] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_tuple_expression_repeat1, 2), SHIFT_REPEAT(1119),
  [516] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_tuple_expression_repeat1, 2), SHIFT_REPEAT(2564),
  [519] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_tuple_expression_repeat1, 2), SHIFT_REPEAT(1856),
  [522] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_tuple_expression_repeat1, 2), SHIFT_REPEAT(22),
  [525] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_tuple_expression_repeat1, 2), SHIFT_REPEAT(2218),
  [528] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_tuple_expression_repeat1, 2), SHIFT_REPEAT(770),
  [531] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_tuple_expression_repeat1, 2), SHIFT_REPEAT(780),
  [534] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_tuple_expression_repeat1, 2), SHIFT_REPEAT(597),
  [537] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_tuple_expression_repeat1, 2), SHIFT_REPEAT(39),
  [540] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_tuple_expression_repeat1, 2), SHIFT_REPEAT(2256),
  [543] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_tuple_expression_repeat1, 2), SHIFT_REPEAT(142),
  [546] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_tuple_expression_repeat1, 2), SHIFT_REPEAT(24),
  [549] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_tuple_expression_repeat1, 2), SHIFT_REPEAT(2186),
  [552] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_tuple_expression_repeat1, 2), SHIFT_REPEAT(41),
  [555] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_tuple_expression_repeat1, 2), SHIFT_REPEAT(634),
  [558] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_tuple_expression_repeat1, 2), SHIFT_REPEAT(2533),
  [561] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_tuple_expression_repeat1, 2), SHIFT_REPEAT(91),
  [564] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_tuple_expression_repeat1, 2), SHIFT_REPEAT(20),
  [567] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_tuple_expression_repeat1, 2), SHIFT_REPEAT(556),
  [570] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_tuple_expression_repeat1, 2), SHIFT_REPEAT(23),
  [573] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_tuple_expression_repeat1, 2), SHIFT_REPEAT(2259),
  [576] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_tuple_expression_repeat1, 2), SHIFT_REPEAT(815),
  [579] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_tuple_expression_repeat1, 2), SHIFT_REPEAT(1853),
  [582] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_tuple_expression_repeat1, 2), SHIFT_REPEAT(1028),
  [585] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_tuple_expression_repeat1, 2), SHIFT_REPEAT(1031),
  [588] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_tuple_expression_repeat1, 2), SHIFT_REPEAT(2530),
  [591] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_tuple_expression_repeat1, 2), SHIFT_REPEAT(1031),
  [594] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1067),
  [596] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_macro_invocation, 3, .production_id = 34),
  [598] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_macro_invocation, 3, .production_id = 34),
  [600] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_macro_invocation, 3, .production_id = 14),
  [602] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_macro_invocation, 3, .production_id = 14),
  [604] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_async_block, 3),
  [606] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_async_block, 3),
  [608] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_block, 4),
  [610] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_match_block, 4),
  [612] = {.entry = {.count = 1, .reusable = true}}, SHIFT(838),
  [614] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_delim_token_tree, 3),
  [616] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_delim_token_tree, 3),
  [618] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1071),
  [620] = {.entry = {.count = 1, .reusable = true}}, SHIFT(741),
  [622] = {.entry = {.count = 1, .reusable = false}}, SHIFT(149),
  [624] = {.entry = {.count = 1, .reusable = true}}, SHIFT(866),
  [626] = {.entry = {.count = 1, .reusable = true}}, SHIFT(843),
  [628] = {.entry = {.count = 1, .reusable = true}}, SHIFT(668),
  [630] = {.entry = {.count = 1, .reusable = false}}, SHIFT(150),
  [632] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2),
  [634] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1955),
  [636] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2269),
  [638] = {.entry = {.count = 1, .reusable = false}}, SHIFT(612),
  [640] = {.entry = {.count = 1, .reusable = false}}, SHIFT(49),
  [642] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2267),
  [644] = {.entry = {.count = 1, .reusable = false}}, SHIFT(120),
  [646] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2250),
  [648] = {.entry = {.count = 1, .reusable = false}}, SHIFT(37),
  [650] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1373),
  [652] = {.entry = {.count = 1, .reusable = true}}, SHIFT(195),
  [654] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1615),
  [656] = {.entry = {.count = 1, .reusable = true}}, SHIFT(194),
  [658] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2171),
  [660] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1546),
  [662] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2478),
  [664] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1526),
  [666] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1549),
  [668] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1395),
  [670] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2165),
  [672] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2476),
  [674] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1167),
  [676] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1557),
  [678] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2547),
  [680] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1377),
  [682] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2537),
  [684] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1561),
  [686] = {.entry = {.count = 1, .reusable = false}}, SHIFT(586),
  [688] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2329),
  [690] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1717),
  [692] = {.entry = {.count = 1, .reusable = true}}, SHIFT(208),
  [694] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2284),
  [696] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1169),
  [698] = {.entry = {.count = 1, .reusable = false}}, SHIFT(603),
  [700] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1421),
  [702] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2208),
  [704] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1397),
  [706] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1829),
  [708] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1358),
  [710] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1658),
  [712] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2011),
  [714] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1673),
  [716] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1666),
  [718] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2401),
  [720] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1681),
  [722] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2332),
  [724] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1375),
  [726] = {.entry = {.count = 1, .reusable = true}}, SHIFT(197),
  [728] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1354),
  [730] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1552),
  [732] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1439),
  [734] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1653),
  [736] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2343),
  [738] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2324),
  [740] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1708),
  [742] = {.entry = {.count = 1, .reusable = true}}, SHIFT(204),
  [744] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1699),
  [746] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1965),
  [748] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1696),
  [750] = {.entry = {.count = 1, .reusable = true}}, SHIFT(776),
  [752] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2466),
  [754] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1707),
  [756] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1365),
  [758] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1820),
  [760] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1370),
  [762] = {.entry = {.count = 1, .reusable = true}}, SHIFT(791),
  [764] = {.entry = {.count = 1, .reusable = true}}, SHIFT(775),
  [766] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1350),
  [768] = {.entry = {.count = 1, .reusable = true}}, SHIFT(795),
  [770] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1770),
  [772] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1723),
  [774] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1731),
  [776] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1369),
  [778] = {.entry = {.count = 1, .reusable = true}}, SHIFT(196),
  [780] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1443),
  [782] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1545),
  [784] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1408),
  [786] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1596),
  [788] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2420),
  [790] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2346),
  [792] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1422),
  [794] = {.entry = {.count = 1, .reusable = true}}, SHIFT(205),
  [796] = {.entry = {.count = 1, .reusable = false}}, SHIFT(599),
  [798] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1421),
  [800] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1988),
  [802] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1682),
  [804] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1738),
  [806] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2421),
  [808] = {.entry = {.count = 1, .reusable = true}}, SHIFT(202),
  [810] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1754),
  [812] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1716),
  [814] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_loop_label, 2),
  [816] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_loop_label, 2),
  [818] = {.entry = {.count = 1, .reusable = true}}, SHIFT(203),
  [820] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1743),
  [822] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2515),
  [824] = {.entry = {.count = 1, .reusable = false}}, SHIFT(207),
  [826] = {.entry = {.count = 1, .reusable = false}}, SHIFT(200),
  [828] = {.entry = {.count = 1, .reusable = false}}, SHIFT(199),
  [830] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1720),
  [832] = {.entry = {.count = 1, .reusable = false}}, SHIFT(201),
  [834] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1718),
  [836] = {.entry = {.count = 1, .reusable = false}}, SHIFT(206),
  [838] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1750),
  [840] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_closure_parameters, 3),
  [842] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_closure_parameters, 3),
  [844] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1796),
  [846] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_match_arm, 3, .production_id = 155),
  [848] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_arm, 3, .production_id = 155),
  [850] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_match_arm, 4, .production_id = 99),
  [852] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_arm, 4, .production_id = 99),
  [854] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1477),
  [856] = {.entry = {.count = 1, .reusable = true}}, SHIFT(638),
  [858] = {.entry = {.count = 1, .reusable = true}}, SHIFT(7),
  [860] = {.entry = {.count = 1, .reusable = true}}, SHIFT(714),
  [862] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1374),
  [864] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1564),
  [866] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2142),
  [868] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1346),
  [870] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2520),
  [872] = {.entry = {.count = 1, .reusable = true}}, SHIFT(627),
  [874] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2313),
  [876] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2378),
  [878] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1368),
  [880] = {.entry = {.count = 1, .reusable = true}}, SHIFT(785),
  [882] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1345),
  [884] = {.entry = {.count = 1, .reusable = true}}, SHIFT(782),
  [886] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_token_tree_pattern_repeat1, 2), SHIFT_REPEAT(239),
  [889] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_token_tree_pattern_repeat1, 2), SHIFT_REPEAT(492),
  [892] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_token_tree_pattern_repeat1, 2),
  [894] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_token_tree_pattern_repeat1, 2), SHIFT_REPEAT(481),
  [897] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_token_tree_pattern_repeat1, 2), SHIFT_REPEAT(494),
  [900] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_token_tree_pattern_repeat1, 2), SHIFT_REPEAT(2460),
  [903] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_token_tree_pattern_repeat1, 2), SHIFT_REPEAT(585),
  [906] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_token_tree_pattern_repeat1, 2), SHIFT_REPEAT(1801),
  [909] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_token_tree_pattern_repeat1, 2), SHIFT_REPEAT(584),
  [912] = {.entry = {.count = 1, .reusable = false}}, SHIFT_EXTRA(),
  [914] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_token_tree_pattern_repeat1, 2), SHIFT_REPEAT(558),
  [917] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2), SHIFT_REPEAT(2288),
  [920] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2), SHIFT_REPEAT(1055),
  [923] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2), SHIFT_REPEAT(1975),
  [926] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2),
  [928] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2), SHIFT_REPEAT(2345),
  [931] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2), SHIFT_REPEAT(1526),
  [934] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2), SHIFT_REPEAT(1577),
  [937] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2), SHIFT_REPEAT(1547),
  [940] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2), SHIFT_REPEAT(2354),
  [943] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2), SHIFT_REPEAT(2238),
  [946] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2), SHIFT_REPEAT(625),
  [949] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2), SHIFT_REPEAT(616),
  [952] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2), SHIFT_REPEAT(2359),
  [955] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2), SHIFT_REPEAT(1341),
  [958] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2), SHIFT_REPEAT(1898),
  [961] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2), SHIFT_REPEAT(2360),
  [964] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2), SHIFT_REPEAT(2361),
  [967] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2), SHIFT_REPEAT(2362),
  [970] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2), SHIFT_REPEAT(1895),
  [973] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2), SHIFT_REPEAT(1553),
  [976] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2), SHIFT_REPEAT(1299),
  [979] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2), SHIFT_REPEAT(2244),
  [982] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2), SHIFT_REPEAT(1464),
  [985] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2), SHIFT_REPEAT(634),
  [988] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2), SHIFT_REPEAT(2364),
  [991] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2), SHIFT_REPEAT(2369),
  [994] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_declaration_list_repeat1, 2), SHIFT_REPEAT(1366),
  [997] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_declaration_list_repeat1, 2), SHIFT_REPEAT(2369),
  [1000] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2288),
  [1002] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1055),
  [1004] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1975),
  [1006] = {.entry = {.count = 1, .reusable = true}}, SHIFT(892),
  [1008] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2345),
  [1010] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1577),
  [1012] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1547),
  [1014] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2354),
  [1016] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2238),
  [1018] = {.entry = {.count = 1, .reusable = false}}, SHIFT(625),
  [1020] = {.entry = {.count = 1, .reusable = false}}, SHIFT(616),
  [1022] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2359),
  [1024] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1898),
  [1026] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2360),
  [1028] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2361),
  [1030] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2362),
  [1032] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1895),
  [1034] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1553),
  [1036] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1299),
  [1038] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2244),
  [1040] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1464),
  [1042] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2364),
  [1044] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2369),
  [1046] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1366),
  [1048] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2369),
  [1050] = {.entry = {.count = 1, .reusable = true}}, SHIFT(972),
  [1052] = {.entry = {.count = 1, .reusable = true}}, SHIFT(445),
  [1054] = {.entry = {.count = 1, .reusable = true}}, SHIFT(321),
  [1056] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_trait_item, 5, .production_id = 116),
  [1058] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_trait_item, 5, .production_id = 116),
  [1060] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_macro_definition, 7, .production_id = 5),
  [1062] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_macro_definition, 7, .production_id = 5),
  [1064] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_macro_definition, 7, .production_id = 48),
  [1066] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_macro_definition, 7, .production_id = 48),
  [1068] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_impl_item, 4, .production_id = 66),
  [1070] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_impl_item, 4, .production_id = 66),
  [1072] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_impl_item, 4, .production_id = 67),
  [1074] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_impl_item, 4, .production_id = 67),
  [1076] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_impl_item, 4, .production_id = 22),
  [1078] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_impl_item, 4, .production_id = 22),
  [1080] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_impl_item, 4, .production_id = 68),
  [1082] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_impl_item, 4, .production_id = 68),
  [1084] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_empty_statement, 1),
  [1086] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_empty_statement, 1),
  [1088] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_declaration, 4, .production_id = 71),
  [1090] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_declaration, 4, .production_id = 71),
  [1092] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_item, 6, .production_id = 181),
  [1094] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_item, 6, .production_id = 181),
  [1096] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration_list, 2),
  [1098] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_declaration_list, 2),
  [1100] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_item, 4, .production_id = 15),
  [1102] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_item, 4, .production_id = 15),
  [1104] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_item, 7, .production_id = 124),
  [1106] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_item, 7, .production_id = 124),
  [1108] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_signature_item, 6, .production_id = 180),
  [1110] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_signature_item, 6, .production_id = 180),
  [1112] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_expression_statement, 2),
  [1114] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_expression_statement, 2),
  [1116] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_variant_list, 5),
  [1118] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_variant_list, 5),
  [1120] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_signature_item, 7, .production_id = 186),
  [1122] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_signature_item, 7, .production_id = 186),
  [1124] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_item, 7, .production_id = 187),
  [1126] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_item, 7, .production_id = 187),
  [1128] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_trait_item, 6, .production_id = 179),
  [1130] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_trait_item, 6, .production_id = 179),
  [1132] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_signature_item, 7, .production_id = 137),
  [1134] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_signature_item, 7, .production_id = 137),
  [1136] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_trait_item, 6, .production_id = 178),
  [1138] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_trait_item, 6, .production_id = 178),
  [1140] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_item, 7, .production_id = 188),
  [1142] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_item, 7, .production_id = 188),
  [1144] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_trait_item, 6, .production_id = 177),
  [1146] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_trait_item, 6, .production_id = 177),
  [1148] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_union_item, 6, .production_id = 169),
  [1150] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_union_item, 6, .production_id = 169),
  [1152] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_item, 4, .production_id = 50),
  [1154] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_item, 4, .production_id = 50),
  [1156] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_item, 4, .production_id = 72),
  [1158] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_item, 4, .production_id = 72),
  [1160] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_item, 4, .production_id = 51),
  [1162] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_item, 4, .production_id = 51),
  [1164] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_trait_item, 4, .production_id = 50),
  [1166] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_trait_item, 4, .production_id = 50),
  [1168] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_impl_item, 7, .production_id = 145),
  [1170] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_impl_item, 7, .production_id = 145),
  [1172] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_item, 6, .production_id = 176),
  [1174] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_item, 6, .production_id = 176),
  [1176] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_item, 6, .production_id = 169),
  [1178] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_item, 6, .production_id = 169),
  [1180] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_item, 3, .production_id = 15),
  [1182] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_item, 3, .production_id = 15),
  [1184] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_impl_item, 7, .production_id = 191),
  [1186] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_impl_item, 7, .production_id = 191),
  [1188] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_item, 6, .production_id = 124),
  [1190] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_item, 6, .production_id = 124),
  [1192] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_foreign_mod_item, 2, .production_id = 2),
  [1194] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_foreign_mod_item, 2, .production_id = 2),
  [1196] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_item, 6, .production_id = 76),
  [1198] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_item, 6, .production_id = 76),
  [1200] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_item, 6, .production_id = 169),
  [1202] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_item, 6, .production_id = 169),
  [1204] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_foreign_mod_item, 2),
  [1206] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_foreign_mod_item, 2),
  [1208] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_const_item, 6, .production_id = 157),
  [1210] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_const_item, 6, .production_id = 157),
  [1212] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_item, 6, .production_id = 175),
  [1214] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_item, 6, .production_id = 175),
  [1216] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_signature_item, 6, .production_id = 128),
  [1218] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_signature_item, 6, .production_id = 128),
  [1220] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_item, 6, .production_id = 174),
  [1222] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_item, 6, .production_id = 174),
  [1224] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_signature_item, 6, .production_id = 173),
  [1226] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_signature_item, 6, .production_id = 173),
  [1228] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_extern_crate_declaration, 6, .production_id = 171),
  [1230] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_extern_crate_declaration, 6, .production_id = 171),
  [1232] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_trait_item, 4, .production_id = 73),
  [1234] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_trait_item, 4, .production_id = 73),
  [1236] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_trait_item, 6, .production_id = 170),
  [1238] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_trait_item, 6, .production_id = 170),
  [1240] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_trait_item, 6, .production_id = 169),
  [1242] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_trait_item, 6, .production_id = 169),
  [1244] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_trait_item, 4, .production_id = 51),
  [1246] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_trait_item, 4, .production_id = 51),
  [1248] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_trait_item, 6, .production_id = 168),
  [1250] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_trait_item, 6, .production_id = 168),
  [1252] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_impl_item, 7, .production_id = 147),
  [1254] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_impl_item, 7, .production_id = 147),
  [1256] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_impl_item, 6, .production_id = 167),
  [1258] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_impl_item, 6, .production_id = 167),
  [1260] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_associated_type, 4, .production_id = 74),
  [1262] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_associated_type, 4, .production_id = 74),
  [1264] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_impl_item, 6, .production_id = 166),
  [1266] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_impl_item, 6, .production_id = 166),
  [1268] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_impl_item, 7, .production_id = 192),
  [1270] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_impl_item, 7, .production_id = 192),
  [1272] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_impl_item, 6, .production_id = 165),
  [1274] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_impl_item, 6, .production_id = 165),
  [1276] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_associated_type, 4, .production_id = 72),
  [1278] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_associated_type, 4, .production_id = 72),
  [1280] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_union_item, 4, .production_id = 50),
  [1282] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_union_item, 4, .production_id = 50),
  [1284] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_impl_item, 6, .production_id = 119),
  [1286] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_impl_item, 6, .production_id = 119),
  [1288] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_union_item, 4, .production_id = 51),
  [1290] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_union_item, 4, .production_id = 51),
  [1292] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_impl_item, 6, .production_id = 164),
  [1294] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_impl_item, 6, .production_id = 164),
  [1296] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_impl_item, 6, .production_id = 163),
  [1298] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_impl_item, 6, .production_id = 163),
  [1300] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_item, 6, .production_id = 162),
  [1302] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_item, 6, .production_id = 162),
  [1304] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_trait_item, 6, .production_id = 161),
  [1306] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_trait_item, 6, .production_id = 161),
  [1308] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_item, 6, .production_id = 51),
  [1310] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_item, 6, .production_id = 51),
  [1312] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_declaration, 7, .production_id = 194),
  [1314] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_declaration, 7, .production_id = 194),
  [1316] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_impl_item, 4, .production_id = 61),
  [1318] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_impl_item, 4, .production_id = 61),
  [1320] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_declaration, 7, .production_id = 195),
  [1322] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_declaration, 7, .production_id = 195),
  [1324] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_declaration, 7, .production_id = 196),
  [1326] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_declaration, 7, .production_id = 196),
  [1328] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_impl_item, 4, .production_id = 75),
  [1330] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_impl_item, 4, .production_id = 75),
  [1332] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_static_item, 7, .production_id = 197),
  [1334] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_static_item, 7, .production_id = 197),
  [1336] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_static_item, 7, .production_id = 183),
  [1338] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_static_item, 7, .production_id = 183),
  [1340] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_trait_item, 4, .production_id = 76),
  [1342] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_trait_item, 4, .production_id = 76),
  [1344] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute_item, 4),
  [1346] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_attribute_item, 4),
  [1348] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_extern_crate_declaration, 4, .production_id = 83),
  [1350] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_extern_crate_declaration, 4, .production_id = 83),
  [1352] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration_list, 4),
  [1354] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_declaration_list, 4),
  [1356] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration_list, 3),
  [1358] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration_list, 3),
  [1360] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_item, 10, .production_id = 245),
  [1362] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_item, 10, .production_id = 245),
  [1364] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_item, 4, .production_id = 76),
  [1366] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_item, 4, .production_id = 76),
  [1368] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_mod_item, 4, .production_id = 83),
  [1370] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_mod_item, 4, .production_id = 83),
  [1372] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_mod_item, 4, .production_id = 86),
  [1374] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_mod_item, 4, .production_id = 86),
  [1376] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration_list, 5),
  [1378] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_declaration_list, 5),
  [1380] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_item, 4, .production_id = 87),
  [1382] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_item, 4, .production_id = 87),
  [1384] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_impl_item, 7, .production_id = 163),
  [1386] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_impl_item, 7, .production_id = 163),
  [1388] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_static_item, 6, .production_id = 157),
  [1390] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_static_item, 6, .production_id = 157),
  [1392] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_item, 4, .production_id = 53),
  [1394] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_item, 4, .production_id = 53),
  [1396] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_impl_item, 7, .production_id = 200),
  [1398] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_impl_item, 7, .production_id = 200),
  [1400] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_impl_item, 7, .production_id = 201),
  [1402] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_impl_item, 7, .production_id = 201),
  [1404] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_impl_item, 7, .production_id = 202),
  [1406] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_impl_item, 7, .production_id = 202),
  [1408] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_item, 4, .production_id = 76),
  [1410] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_item, 4, .production_id = 76),
  [1412] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_declaration, 6, .production_id = 154),
  [1414] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_declaration, 6, .production_id = 154),
  [1416] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_union_item, 4, .production_id = 76),
  [1418] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_union_item, 4, .production_id = 76),
  [1420] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_declaration, 6, .production_id = 153),
  [1422] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_declaration, 6, .production_id = 153),
  [1424] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_impl_item, 7, .production_id = 203),
  [1426] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_impl_item, 7, .production_id = 203),
  [1428] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_declaration, 6, .production_id = 152),
  [1430] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_declaration, 6, .production_id = 152),
  [1432] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_impl_item, 6, .production_id = 151),
  [1434] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_impl_item, 6, .production_id = 151),
  [1436] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_impl_item, 6, .production_id = 109),
  [1438] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_impl_item, 6, .production_id = 109),
  [1440] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_impl_item, 6, .production_id = 148),
  [1442] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_impl_item, 6, .production_id = 148),
  [1444] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_impl_item, 6, .production_id = 147),
  [1446] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_impl_item, 6, .production_id = 147),
  [1448] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_impl_item, 6, .production_id = 146),
  [1450] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_impl_item, 6, .production_id = 146),
  [1452] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_impl_item, 6, .production_id = 145),
  [1454] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_impl_item, 6, .production_id = 145),
  [1456] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_impl_item, 6, .production_id = 143),
  [1458] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_impl_item, 6, .production_id = 143),
  [1460] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_impl_item, 6, .production_id = 104),
  [1462] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_impl_item, 6, .production_id = 104),
  [1464] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_impl_item, 3, .production_id = 22),
  [1466] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_impl_item, 3, .production_id = 22),
  [1468] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_impl_item, 7, .production_id = 204),
  [1470] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_impl_item, 7, .production_id = 204),
  [1472] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_impl_item, 3, .production_id = 24),
  [1474] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_impl_item, 3, .production_id = 24),
  [1476] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_impl_item, 7, .production_id = 166),
  [1478] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_impl_item, 7, .production_id = 166),
  [1480] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_impl_item, 7, .production_id = 205),
  [1482] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_impl_item, 7, .production_id = 205),
  [1484] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_trait_item, 7, .production_id = 206),
  [1486] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_trait_item, 7, .production_id = 206),
  [1488] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_signature_item, 4, .production_id = 52),
  [1490] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_signature_item, 4, .production_id = 52),
  [1492] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_use_declaration, 4, .production_id = 88),
  [1494] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_use_declaration, 4, .production_id = 88),
  [1496] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_signature_item, 10, .production_id = 240),
  [1498] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_signature_item, 10, .production_id = 240),
  [1500] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_static_item, 10, .production_id = 244),
  [1502] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_static_item, 10, .production_id = 244),
  [1504] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_item, 6, .production_id = 138),
  [1506] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_item, 6, .production_id = 138),
  [1508] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_declaration, 3, .production_id = 26),
  [1510] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_declaration, 3, .production_id = 26),
  [1512] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_signature_item, 6, .production_id = 137),
  [1514] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_signature_item, 6, .production_id = 137),
  [1516] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_signature_item, 7, .production_id = 173),
  [1518] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_signature_item, 7, .production_id = 173),
  [1520] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_item, 6, .production_id = 136),
  [1522] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_item, 6, .production_id = 136),
  [1524] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1315),
  [1526] = {.entry = {.count = 1, .reusable = true}}, SHIFT(561),
  [1528] = {.entry = {.count = 1, .reusable = true}}, SHIFT(73),
  [1530] = {.entry = {.count = 1, .reusable = true}}, SHIFT(559),
  [1532] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1364),
  [1534] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2251),
  [1536] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1348),
  [1538] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2496),
  [1540] = {.entry = {.count = 1, .reusable = true}}, SHIFT(608),
  [1542] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2001),
  [1544] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2001),
  [1546] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_item, 7, .production_id = 207),
  [1548] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_item, 7, .production_id = 207),
  [1550] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_mod_item, 3, .production_id = 5),
  [1552] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_mod_item, 3, .production_id = 5),
  [1554] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_mod_item, 3, .production_id = 28),
  [1556] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_mod_item, 3, .production_id = 28),
  [1558] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_signature_item, 6, .production_id = 94),
  [1560] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_signature_item, 6, .production_id = 94),
  [1562] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_declaration, 10, .production_id = 243),
  [1564] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_declaration, 10, .production_id = 243),
  [1566] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_signature_item, 7, .production_id = 208),
  [1568] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_signature_item, 7, .production_id = 208),
  [1570] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_item, 7, .production_id = 209),
  [1572] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_item, 7, .production_id = 209),
  [1574] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_macro_definition, 5, .production_id = 48),
  [1576] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_macro_definition, 5, .production_id = 48),
  [1578] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_item, 3, .production_id = 6),
  [1580] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_item, 3, .production_id = 6),
  [1582] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_macro_definition, 5, .production_id = 5),
  [1584] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_macro_definition, 5, .production_id = 5),
  [1586] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_declaration, 8, .production_id = 224),
  [1588] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_declaration, 8, .production_id = 224),
  [1590] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_item, 3, .production_id = 15),
  [1592] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_item, 3, .production_id = 15),
  [1594] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_item, 7, .production_id = 210),
  [1596] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_item, 7, .production_id = 210),
  [1598] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_trait_item, 7, .production_id = 211),
  [1600] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_trait_item, 7, .production_id = 211),
  [1602] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_trait_item, 7, .production_id = 212),
  [1604] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_trait_item, 7, .production_id = 212),
  [1606] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_item, 9, .production_id = 242),
  [1608] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_item, 9, .production_id = 242),
  [1610] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_trait_item, 3, .production_id = 15),
  [1612] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_trait_item, 3, .production_id = 15),
  [1614] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_signature_item, 9, .production_id = 235),
  [1616] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_signature_item, 9, .production_id = 235),
  [1618] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_variant_list, 4),
  [1620] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_variant_list, 4),
  [1622] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_trait_item, 7, .production_id = 213),
  [1624] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_trait_item, 7, .production_id = 213),
  [1626] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_associated_type, 3, .production_id = 6),
  [1628] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_associated_type, 3, .production_id = 6),
  [1630] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_macro_definition, 6, .production_id = 5),
  [1632] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_macro_definition, 6, .production_id = 5),
  [1634] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_item, 9, .production_id = 241),
  [1636] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_item, 9, .production_id = 241),
  [1638] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_signature_item, 9, .production_id = 240),
  [1640] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_signature_item, 9, .production_id = 240),
  [1642] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_union_item, 3, .production_id = 15),
  [1644] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_union_item, 3, .production_id = 15),
  [1646] = {.entry = {.count = 1, .reusable = true}}, SHIFT(225),
  [1648] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_macro_definition, 6, .production_id = 48),
  [1650] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_macro_definition, 6, .production_id = 48),
  [1652] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_extern_crate_declaration, 7, .production_id = 214),
  [1654] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_extern_crate_declaration, 7, .production_id = 214),
  [1656] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_item, 4, .production_id = 51),
  [1658] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_item, 4, .production_id = 51),
  [1660] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_item, 4, .production_id = 50),
  [1662] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_item, 4, .production_id = 50),
  [1664] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_signature_item, 7, .production_id = 215),
  [1666] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_signature_item, 7, .production_id = 215),
  [1668] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_item, 7, .production_id = 216),
  [1670] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_item, 7, .production_id = 216),
  [1672] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_const_item, 5, .production_id = 92),
  [1674] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_const_item, 5, .production_id = 92),
  [1676] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_signature_item, 7, .production_id = 180),
  [1678] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_signature_item, 7, .production_id = 180),
  [1680] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_item, 9, .production_id = 239),
  [1682] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_item, 9, .production_id = 239),
  [1684] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_extern_crate_declaration, 5, .production_id = 132),
  [1686] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_extern_crate_declaration, 5, .production_id = 132),
  [1688] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_trait_item, 5, .production_id = 131),
  [1690] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_trait_item, 5, .production_id = 131),
  [1692] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_union_item, 5, .production_id = 124),
  [1694] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_union_item, 5, .production_id = 124),
  [1696] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_union_item, 5, .production_id = 122),
  [1698] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_union_item, 5, .production_id = 122),
  [1700] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_item, 5, .production_id = 124),
  [1702] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_item, 5, .production_id = 124),
  [1704] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_item, 5, .production_id = 130),
  [1706] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_item, 5, .production_id = 130),
  [1708] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_item, 5, .production_id = 122),
  [1710] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_item, 5, .production_id = 122),
  [1712] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_item, 5, .production_id = 76),
  [1714] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_item, 5, .production_id = 76),
  [1716] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_use_declaration, 3, .production_id = 30),
  [1718] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_use_declaration, 3, .production_id = 30),
  [1720] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_item, 7, .production_id = 217),
  [1722] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_item, 7, .production_id = 217),
  [1724] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_item, 5, .production_id = 124),
  [1726] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_item, 5, .production_id = 124),
  [1728] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_signature_item, 9, .production_id = 229),
  [1730] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_signature_item, 9, .production_id = 229),
  [1732] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_item, 5, .production_id = 122),
  [1734] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_item, 5, .production_id = 122),
  [1736] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_item, 5, .production_id = 129),
  [1738] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_item, 5, .production_id = 129),
  [1740] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_signature_item, 5, .production_id = 128),
  [1742] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_signature_item, 5, .production_id = 128),
  [1744] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_variant_list, 3),
  [1746] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_variant_list, 3),
  [1748] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_static_item, 9, .production_id = 238),
  [1750] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_static_item, 9, .production_id = 238),
  [1752] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_declaration, 9, .production_id = 237),
  [1754] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_declaration, 9, .production_id = 237),
  [1756] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_inner_attribute_item, 5),
  [1758] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_inner_attribute_item, 5),
  [1760] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_item, 5, .production_id = 93),
  [1762] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_item, 5, .production_id = 93),
  [1764] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_variant_list, 6),
  [1766] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_variant_list, 6),
  [1768] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_item, 8, .production_id = 236),
  [1770] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_item, 8, .production_id = 236),
  [1772] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_trait_item, 5, .production_id = 124),
  [1774] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_trait_item, 5, .production_id = 124),
  [1776] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_signature_item, 8, .production_id = 235),
  [1778] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_signature_item, 8, .production_id = 235),
  [1780] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_trait_item, 5, .production_id = 123),
  [1782] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_trait_item, 5, .production_id = 123),
  [1784] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_trait_item, 5, .production_id = 122),
  [1786] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_trait_item, 5, .production_id = 122),
  [1788] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_variant_list, 2),
  [1790] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_enum_variant_list, 2),
  [1792] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_impl_item, 5, .production_id = 121),
  [1794] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_impl_item, 5, .production_id = 121),
  [1796] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_signature_item, 8, .production_id = 186),
  [1798] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_signature_item, 8, .production_id = 186),
  [1800] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_item, 8, .production_id = 220),
  [1802] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_item, 8, .production_id = 220),
  [1804] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_impl_item, 5, .production_id = 61),
  [1806] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_impl_item, 5, .production_id = 61),
  [1808] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_item, 8, .production_id = 234),
  [1810] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_item, 8, .production_id = 234),
  [1812] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_impl_item, 5, .production_id = 120),
  [1814] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_impl_item, 5, .production_id = 120),
  [1816] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_impl_item, 5, .production_id = 119),
  [1818] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_impl_item, 5, .production_id = 119),
  [1820] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_union_item, 5, .production_id = 93),
  [1822] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_union_item, 5, .production_id = 93),
  [1824] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_associated_type, 5, .production_id = 118),
  [1826] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_associated_type, 5, .production_id = 118),
  [1828] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_signature_item, 8, .production_id = 215),
  [1830] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_signature_item, 8, .production_id = 215),
  [1832] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_trait_item, 8, .production_id = 233),
  [1834] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_trait_item, 8, .production_id = 233),
  [1836] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_static_item, 8, .production_id = 232),
  [1838] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_static_item, 8, .production_id = 232),
  [1840] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_item, 5, .production_id = 117),
  [1842] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_item, 5, .production_id = 117),
  [1844] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_declaration, 8, .production_id = 222),
  [1846] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_declaration, 8, .production_id = 222),
  [1848] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_const_item, 7, .production_id = 183),
  [1850] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_const_item, 7, .production_id = 183),
  [1852] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_trait_item, 5, .production_id = 93),
  [1854] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_trait_item, 5, .production_id = 93),
  [1856] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_trait_item, 5, .production_id = 115),
  [1858] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_trait_item, 5, .production_id = 115),
  [1860] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_declaration_list, 2),
  [1862] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_declaration_list, 2),
  [1864] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_signature_item, 5, .production_id = 94),
  [1866] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_signature_item, 5, .production_id = 94),
  [1868] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_item, 5, .production_id = 95),
  [1870] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_item, 5, .production_id = 95),
  [1872] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_macro_definition, 4, .production_id = 5),
  [1874] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_macro_definition, 4, .production_id = 5),
  [1876] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_declaration, 8, .production_id = 223),
  [1878] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_declaration, 8, .production_id = 223),
  [1880] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_item, 5, .production_id = 93),
  [1882] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_item, 5, .production_id = 93),
  [1884] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_signature_item, 5, .production_id = 52),
  [1886] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_signature_item, 5, .production_id = 52),
  [1888] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_item, 5, .production_id = 51),
  [1890] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_item, 5, .production_id = 51),
  [1892] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_item, 5, .production_id = 96),
  [1894] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_item, 5, .production_id = 96),
  [1896] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_const_item, 8, .production_id = 225),
  [1898] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_const_item, 8, .production_id = 225),
  [1900] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_macro_definition, 4, .production_id = 48),
  [1902] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_macro_definition, 4, .production_id = 48),
  [1904] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_item, 5, .production_id = 15),
  [1906] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_item, 5, .production_id = 15),
  [1908] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_item, 8, .production_id = 231),
  [1910] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_item, 8, .production_id = 231),
  [1912] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_signature_item, 8, .production_id = 208),
  [1914] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_signature_item, 8, .production_id = 208),
  [1916] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_static_item, 8, .production_id = 225),
  [1918] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_static_item, 8, .production_id = 225),
  [1920] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_item, 8, .production_id = 230),
  [1922] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_item, 8, .production_id = 230),
  [1924] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration_list, 3),
  [1926] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_declaration_list, 3),
  [1928] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_signature_item, 8, .production_id = 229),
  [1930] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_signature_item, 8, .production_id = 229),
  [1932] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_impl_item, 5, .production_id = 104),
  [1934] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_impl_item, 5, .production_id = 104),
  [1936] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration_list, 6),
  [1938] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_declaration_list, 6),
  [1940] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_impl_item, 5, .production_id = 105),
  [1942] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_impl_item, 5, .production_id = 105),
  [1944] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_foreign_mod_item, 3),
  [1946] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_foreign_mod_item, 3),
  [1948] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_foreign_mod_item, 3, .production_id = 39),
  [1950] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_foreign_mod_item, 3, .production_id = 39),
  [1952] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_static_item, 5, .production_id = 92),
  [1954] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_static_item, 5, .production_id = 92),
  [1956] = {.entry = {.count = 1, .reusable = true}}, SHIFT(915),
  [1958] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_impl_item, 5, .production_id = 66),
  [1960] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_impl_item, 5, .production_id = 66),
  [1962] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_impl_item, 5, .production_id = 107),
  [1964] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_impl_item, 5, .production_id = 107),
  [1966] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_impl_item, 8, .production_id = 201),
  [1968] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_impl_item, 8, .production_id = 201),
  [1970] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_impl_item, 8, .production_id = 227),
  [1972] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_impl_item, 8, .production_id = 227),
  [1974] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_impl_item, 8, .production_id = 203),
  [1976] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_impl_item, 8, .production_id = 203),
  [1978] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_impl_item, 5, .production_id = 109),
  [1980] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_impl_item, 5, .production_id = 109),
  [1982] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_declaration, 5, .production_id = 113),
  [1984] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_declaration, 5, .production_id = 113),
  [1986] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_declaration, 5, .production_id = 99),
  [1988] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_declaration, 5, .production_id = 99),
  [1990] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_impl_item, 8, .production_id = 228),
  [1992] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_impl_item, 8, .production_id = 228),
  [1994] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_declaration, 5, .production_id = 112),
  [1996] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_let_declaration, 5, .production_id = 112),
  [1998] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_impl_item, 5, .production_id = 110),
  [2000] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_impl_item, 5, .production_id = 110),
  [2002] = {.entry = {.count = 1, .reusable = false}}, SHIFT(497),
  [2004] = {.entry = {.count = 1, .reusable = true}}, SHIFT(492),
  [2006] = {.entry = {.count = 1, .reusable = true}}, SHIFT(481),
  [2008] = {.entry = {.count = 1, .reusable = true}}, SHIFT(564),
  [2010] = {.entry = {.count = 1, .reusable = true}}, SHIFT(494),
  [2012] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2460),
  [2014] = {.entry = {.count = 1, .reusable = true}}, SHIFT(585),
  [2016] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1801),
  [2018] = {.entry = {.count = 1, .reusable = false}}, SHIFT(584),
  [2020] = {.entry = {.count = 1, .reusable = true}}, SHIFT(558),
  [2022] = {.entry = {.count = 1, .reusable = false}}, SHIFT(491),
  [2024] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1763),
  [2026] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_delim_token_tree_repeat1, 2), SHIFT_REPEAT(483),
  [2029] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_delim_token_tree_repeat1, 2), SHIFT_REPEAT(522),
  [2032] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_delim_token_tree_repeat1, 2),
  [2034] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_delim_token_tree_repeat1, 2), SHIFT_REPEAT(530),
  [2037] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_delim_token_tree_repeat1, 2), SHIFT_REPEAT(534),
  [2040] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_delim_token_tree_repeat1, 2), SHIFT_REPEAT(483),
  [2043] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_delim_token_tree_repeat1, 2), SHIFT_REPEAT(592),
  [2046] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_delim_token_tree_repeat1, 2), SHIFT_REPEAT(1762),
  [2049] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_delim_token_tree_repeat1, 2), SHIFT_REPEAT(589),
  [2052] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_token_tree_repeat1, 2), SHIFT_REPEAT(485),
  [2055] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_token_tree_repeat1, 2), SHIFT_REPEAT(535),
  [2058] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_token_tree_repeat1, 2),
  [2060] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_token_tree_repeat1, 2), SHIFT_REPEAT(544),
  [2063] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_token_tree_repeat1, 2), SHIFT_REPEAT(533),
  [2066] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_token_tree_repeat1, 2), SHIFT_REPEAT(2553),
  [2069] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_token_tree_repeat1, 2), SHIFT_REPEAT(585),
  [2072] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_token_tree_repeat1, 2), SHIFT_REPEAT(1801),
  [2075] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_token_tree_repeat1, 2), SHIFT_REPEAT(584),
  [2078] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_token_tree_repeat1, 2), SHIFT_REPEAT(485),
  [2081] = {.entry = {.count = 1, .reusable = false}}, SHIFT(239),
  [2083] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2514),
  [2085] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1824),
  [2087] = {.entry = {.count = 1, .reusable = false}}, SHIFT(493),
  [2089] = {.entry = {.count = 1, .reusable = true}}, SHIFT(568),
  [2091] = {.entry = {.count = 1, .reusable = false}}, SHIFT(495),
  [2093] = {.entry = {.count = 1, .reusable = false}}, SHIFT(489),
  [2095] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2459),
  [2097] = {.entry = {.count = 1, .reusable = false}}, SHIFT(488),
  [2099] = {.entry = {.count = 1, .reusable = false}}, SHIFT(487),
  [2101] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_match_block_repeat1, 2), SHIFT_REPEAT(1315),
  [2104] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_match_block_repeat1, 2), SHIFT_REPEAT(561),
  [2107] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_match_block_repeat1, 2), SHIFT_REPEAT(559),
  [2110] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_match_block_repeat1, 2), SHIFT_REPEAT(1364),
  [2113] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_match_block_repeat1, 2), SHIFT_REPEAT(2251),
  [2116] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_match_block_repeat1, 2), SHIFT_REPEAT(1348),
  [2119] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_match_block_repeat1, 2), SHIFT_REPEAT(2501),
  [2122] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_match_block_repeat1, 2), SHIFT_REPEAT(586),
  [2125] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_match_block_repeat1, 2), SHIFT_REPEAT(634),
  [2128] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_match_block_repeat1, 2), SHIFT_REPEAT(2496),
  [2131] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_match_block_repeat1, 2), SHIFT_REPEAT(1422),
  [2134] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_match_block_repeat1, 2), SHIFT_REPEAT(608),
  [2137] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_match_block_repeat1, 2), SHIFT_REPEAT(599),
  [2140] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_match_block_repeat1, 2), SHIFT_REPEAT(1421),
  [2143] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_match_block_repeat1, 2), SHIFT_REPEAT(2208),
  [2146] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_match_block_repeat1, 2), SHIFT_REPEAT(1397),
  [2149] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_match_block_repeat1, 2), SHIFT_REPEAT(1829),
  [2152] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_match_block_repeat1, 2), SHIFT_REPEAT(1358),
  [2155] = {.entry = {.count = 2, .reusable = false}}, REDUCE(aux_sym_match_block_repeat1, 2), SHIFT_REPEAT(2001),
  [2158] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_match_block_repeat1, 2), SHIFT_REPEAT(2001),
  [2161] = {.entry = {.count = 1, .reusable = false}}, SHIFT(483),
  [2163] = {.entry = {.count = 1, .reusable = true}}, SHIFT(522),
  [2165] = {.entry = {.count = 1, .reusable = true}}, SHIFT(530),
  [2167] = {.entry = {.count = 1, .reusable = true}}, SHIFT(879),
  [2169] = {.entry = {.count = 1, .reusable = true}}, SHIFT(534),
  [2171] = {.entry = {.count = 1, .reusable = true}}, SHIFT(483),
  [2173] = {.entry = {.count = 1, .reusable = true}}, SHIFT(592),
  [2175] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1762),
  [2177] = {.entry = {.count = 1, .reusable = false}}, SHIFT(589),
  [2179] = {.entry = {.count = 1, .reusable = true}}, SHIFT(615),
  [2181] = {.entry = {.count = 1, .reusable = false}}, SHIFT(485),
  [2183] = {.entry = {.count = 1, .reusable = true}}, SHIFT(535),
  [2185] = {.entry = {.count = 1, .reusable = true}}, SHIFT(544),
  [2187] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1871),
  [2189] = {.entry = {.count = 1, .reusable = true}}, SHIFT(533),
  [2191] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2553),
  [2193] = {.entry = {.count = 1, .reusable = true}}, SHIFT(485),
  [2195] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1335),
  [2197] = {.entry = {.count = 1, .reusable = false}}, SHIFT(513),
  [2199] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1124),
  [2201] = {.entry = {.count = 1, .reusable = true}}, SHIFT(513),
  [2203] = {.entry = {.count = 1, .reusable = false}}, SHIFT(519),
  [2205] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1338),
  [2207] = {.entry = {.count = 1, .reusable = true}}, SHIFT(519),
  [2209] = {.entry = {.count = 1, .reusable = false}}, SHIFT(542),
  [2211] = {.entry = {.count = 1, .reusable = true}}, SHIFT(542),
  [2213] = {.entry = {.count = 1, .reusable = false}}, SHIFT(504),
  [2215] = {.entry = {.count = 1, .reusable = true}}, SHIFT(504),
  [2217] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1114),
  [2219] = {.entry = {.count = 1, .reusable = false}}, SHIFT(529),
  [2221] = {.entry = {.count = 1, .reusable = true}}, SHIFT(64),
  [2223] = {.entry = {.count = 1, .reusable = true}}, SHIFT(529),
  [2225] = {.entry = {.count = 1, .reusable = false}}, SHIFT(525),
  [2227] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1899),
  [2229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(525),
  [2231] = {.entry = {.count = 1, .reusable = false}}, SHIFT(503),
  [2233] = {.entry = {.count = 1, .reusable = true}}, SHIFT(503),
  [2235] = {.entry = {.count = 1, .reusable = false}}, SHIFT(506),
  [2237] = {.entry = {.count = 1, .reusable = true}}, SHIFT(506),
  [2239] = {.entry = {.count = 1, .reusable = false}}, SHIFT(528),
  [2241] = {.entry = {.count = 1, .reusable = true}}, SHIFT(528),
  [2243] = {.entry = {.count = 1, .reusable = true}}, SHIFT(574),
  [2245] = {.entry = {.count = 1, .reusable = false}}, SHIFT(512),
  [2247] = {.entry = {.count = 1, .reusable = true}}, SHIFT(512),
  [2249] = {.entry = {.count = 1, .reusable = false}}, SHIFT(536),
  [2251] = {.entry = {.count = 1, .reusable = true}}, SHIFT(966),
  [2253] = {.entry = {.count = 1, .reusable = true}}, SHIFT(536),
  [2255] = {.entry = {.count = 1, .reusable = false}}, SHIFT(540),
  [2257] = {.entry = {.count = 1, .reusable = true}}, SHIFT(605),
  [2259] = {.entry = {.count = 1, .reusable = true}}, SHIFT(540),
  [2261] = {.entry = {.count = 1, .reusable = false}}, SHIFT(510),
  [2263] = {.entry = {.count = 1, .reusable = true}}, SHIFT(510),
  [2265] = {.entry = {.count = 1, .reusable = false}}, SHIFT(501),
  [2267] = {.entry = {.count = 1, .reusable = true}}, SHIFT(501),
  [2269] = {.entry = {.count = 1, .reusable = true}}, SHIFT(88),
  [2271] = {.entry = {.count = 1, .reusable = false}}, SHIFT(543),
  [2273] = {.entry = {.count = 1, .reusable = true}}, SHIFT(543),
  [2275] = {.entry = {.count = 1, .reusable = false}}, SHIFT(524),
  [2277] = {.entry = {.count = 1, .reusable = true}}, SHIFT(524),
  [2279] = {.entry = {.count = 1, .reusable = false}}, SHIFT(527),
  [2281] = {.entry = {.count = 1, .reusable = true}}, SHIFT(527),
  [2283] = {.entry = {.count = 1, .reusable = false}}, SHIFT(538),
  [2285] = {.entry = {.count = 1, .reusable = true}}, SHIFT(571),
  [2287] = {.entry = {.count = 1, .reusable = true}}, SHIFT(538),
  [2289] = {.entry = {.count = 1, .reusable = false}}, SHIFT(502),
  [2291] = {.entry = {.count = 1, .reusable = true}}, SHIFT(502),
  [2293] = {.entry = {.count = 1, .reusable = false}}, SHIFT(518),
  [2295] = {.entry = {.count = 1, .reusable = true}}, SHIFT(518),
  [2297] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1847),
  [2299] = {.entry = {.count = 1, .reusable = false}}, SHIFT(537),
  [2301] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1836),
  [2303] = {.entry = {.count = 1, .reusable = true}}, SHIFT(537),
  [2305] = {.entry = {.count = 1, .reusable = false}}, SHIFT(541),
  [2307] = {.entry = {.count = 1, .reusable = true}}, SHIFT(541),
  [2309] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1316),
  [2311] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1727),
  [2313] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2478),
  [2315] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1183),
  [2317] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2518),
  [2319] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2402),
  [2321] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1178),
  [2323] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1745),
  [2325] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1724),
  [2327] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1722),
  [2329] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1713),
  [2331] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1732),
  [2333] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1714),
  [2335] = {.entry = {.count = 1, .reusable = false}}, SHIFT(590),
  [2337] = {.entry = {.count = 1, .reusable = true}}, SHIFT(748),
  [2339] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1757),
  [2341] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__token_pattern, 1),
  [2343] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__token_pattern, 1),
  [2345] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1389),
  [2347] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1427),
  [2349] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2356),
  [2351] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1440),
  [2353] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_token_repetition, 4),
  [2355] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_token_repetition, 4),
  [2357] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_token_tree_pattern, 2),
  [2359] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_token_tree_pattern, 2),
  [2361] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_token_repetition_pattern, 5),
  [2363] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_token_repetition_pattern, 5),
  [2365] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1437),
  [2367] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string_literal, 3),
  [2369] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_string_literal, 3),
  [2371] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_token_tree_pattern, 3),
  [2373] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_token_tree_pattern, 3),
  [2375] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1380),
  [2377] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1567),
  [2379] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1634),
  [2381] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1428),
  [2383] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_token_tree, 2),
  [2385] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_token_tree, 2),
  [2387] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_token_repetition_pattern, 4),
  [2389] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_token_repetition_pattern, 4),
  [2391] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1432),
  [2393] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_token_tree, 3),
  [2395] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_token_tree, 3),
  [2397] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_string_literal, 2),
  [2399] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_string_literal, 2),
  [2401] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1446),
  [2403] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_token_repetition, 5),
  [2405] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_token_repetition, 5),
  [2407] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_fragment_specifier, 1),
  [2409] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_fragment_specifier, 1),
  [2411] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_token_binding_pattern, 3, .production_id = 182),
  [2413] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_token_binding_pattern, 3, .production_id = 182),
  [2415] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1460),
  [2417] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1436),
  [2419] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_token_repetition, 6),
  [2421] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_token_repetition, 6),
  [2423] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_token_repetition_pattern, 6),
  [2425] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_token_repetition_pattern, 6),
  [2427] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_boolean_literal, 1),
  [2429] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_boolean_literal, 1),
  [2431] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__literal, 1),
  [2433] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__literal, 1),
  [2435] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1843),
  [2437] = {.entry = {.count = 1, .reusable = true}}, SHIFT(702),
  [2439] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2195),
  [2441] = {.entry = {.count = 1, .reusable = false}}, SHIFT(613),
  [2443] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1665),
  [2445] = {.entry = {.count = 1, .reusable = false}}, SHIFT(607),
  [2447] = {.entry = {.count = 1, .reusable = false}}, SHIFT(600),
  [2449] = {.entry = {.count = 1, .reusable = true}}, SHIFT(678),
  [2451] = {.entry = {.count = 1, .reusable = false}}, SHIFT(700),
  [2453] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1945),
  [2455] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_enum_variant_list_repeat1, 2),
  [2457] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_enum_variant_list_repeat1, 2),
  [2459] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_enum_variant_list_repeat1, 2), SHIFT_REPEAT(2501),
  [2462] = {.entry = {.count = 1, .reusable = false}}, SHIFT(674),
  [2464] = {.entry = {.count = 1, .reusable = false}}, SHIFT(752),
  [2466] = {.entry = {.count = 1, .reusable = true}}, SHIFT(631),
  [2468] = {.entry = {.count = 1, .reusable = true}}, SHIFT(711),
  [2470] = {.entry = {.count = 1, .reusable = true}}, SHIFT(744),
  [2472] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2239),
  [2474] = {.entry = {.count = 1, .reusable = false}}, SHIFT(787),
  [2476] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1538),
  [2478] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2236),
  [2480] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1165),
  [2482] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2224),
  [2484] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1042),
  [2486] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2353),
  [2488] = {.entry = {.count = 1, .reusable = true}}, SHIFT(636),
  [2490] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1164),
  [2492] = {.entry = {.count = 1, .reusable = false}}, SHIFT(725),
  [2494] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2491),
  [2496] = {.entry = {.count = 1, .reusable = true}}, SHIFT(788),
  [2498] = {.entry = {.count = 1, .reusable = false}}, SHIFT(644),
  [2500] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1462),
  [2502] = {.entry = {.count = 1, .reusable = true}}, SHIFT(569),
  [2504] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2515),
  [2506] = {.entry = {.count = 1, .reusable = false}}, SHIFT(727),
  [2508] = {.entry = {.count = 1, .reusable = false}}, SHIFT(653),
  [2510] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1463),
  [2512] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1378),
  [2514] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1466),
  [2516] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1017),
  [2518] = {.entry = {.count = 1, .reusable = true}}, SHIFT(868),
  [2520] = {.entry = {.count = 1, .reusable = true}}, SHIFT(905),
  [2522] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1465),
  [2524] = {.entry = {.count = 1, .reusable = false}}, SHIFT(745),
  [2526] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1404),
  [2528] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1388),
  [2530] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1469),
  [2532] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1866),
  [2534] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1468),
  [2536] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1473),
  [2538] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1470),
  [2540] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_closure_parameters, 4),
  [2542] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_closure_parameters, 4),
  [2544] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_closure_parameters, 2),
  [2546] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_closure_parameters, 2),
  [2548] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_tuple_expression_repeat1, 2),
  [2550] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type, 1, .production_id = 4),
  [2552] = {.entry = {.count = 1, .reusable = true}}, SHIFT(183),
  [2554] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__type, 1, .production_id = 4),
  [2556] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1848),
  [2558] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1896),
  [2560] = {.entry = {.count = 1, .reusable = true}}, SHIFT(237),
  [2562] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_dynamic_type, 2, .production_id = 20),
  [2564] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_dynamic_type, 2, .production_id = 20),
  [2566] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_abstract_type, 2, .production_id = 20),
  [2568] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_abstract_type, 2, .production_id = 20),
  [2570] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_scoped_type_identifier, 3, .production_id = 41),
  [2572] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_scoped_type_identifier, 3, .production_id = 41),
  [2574] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_scoped_identifier, 3, .production_id = 40),
  [2576] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_scoped_identifier, 3, .production_id = 40),
  [2578] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_scoped_type_identifier, 3, .production_id = 36),
  [2580] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_scoped_type_identifier, 3, .production_id = 36),
  [2582] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_scoped_identifier, 3, .production_id = 35),
  [2584] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_scoped_identifier, 3, .production_id = 35),
  [2586] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_dynamic_type, 2, .production_id = 21),
  [2588] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_dynamic_type, 2, .production_id = 21),
  [2590] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1475),
  [2592] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2102),
  [2594] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_abstract_type, 2, .production_id = 21),
  [2596] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_abstract_type, 2, .production_id = 21),
  [2598] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__type, 1),
  [2600] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__type, 1),
  [2602] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_scoped_type_identifier, 2, .production_id = 6),
  [2604] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_scoped_type_identifier, 2, .production_id = 6),
  [2606] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_scoped_identifier, 2, .production_id = 5),
  [2608] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_scoped_identifier, 2, .production_id = 5),
  [2610] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_scoped_type_identifier, 3, .production_id = 13),
  [2612] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_scoped_type_identifier, 3, .production_id = 13),
  [2614] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_scoped_identifier, 3, .production_id = 12),
  [2616] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_scoped_identifier, 3, .production_id = 12),
  [2618] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_scoped_type_identifier_in_expression_position, 3, .production_id = 36),
  [2620] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expression_except_range, 1, .production_id = 1),
  [2622] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__expression_except_range, 1, .production_id = 1),
  [2624] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_function_modifiers_repeat1, 1),
  [2626] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1834),
  [2628] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2499),
  [2630] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_scoped_type_identifier_in_expression_position, 3, .production_id = 13),
  [2632] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_scoped_type_identifier_in_expression_position, 2, .production_id = 6),
  [2634] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_continue_expression, 1),
  [2636] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_continue_expression, 1),
  [2638] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2564),
  [2640] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_scoped_type_identifier_in_expression_position, 3, .production_id = 41),
  [2642] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameters, 5),
  [2644] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameters, 5),
  [2646] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameters, 2),
  [2648] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameters, 2),
  [2650] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__expression, 1),
  [2652] = {.entry = {.count = 1, .reusable = true}}, SHIFT(33),
  [2654] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__expression, 1),
  [2656] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2152),
  [2658] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_type, 3, .production_id = 65),
  [2660] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_type, 3, .production_id = 65),
  [2662] = {.entry = {.count = 1, .reusable = true}}, SHIFT(670),
  [2664] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1845),
  [2666] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_type, 3, .production_id = 69),
  [2668] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_type, 3, .production_id = 69),
  [2670] = {.entry = {.count = 1, .reusable = true}}, SHIFT(672),
  [2672] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_arguments, 5),
  [2674] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_arguments, 5),
  [2676] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_type, 3, .production_id = 70),
  [2678] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_type, 3, .production_id = 70),
  [2680] = {.entry = {.count = 1, .reusable = true}}, SHIFT(675),
  [2682] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_arguments, 4),
  [2684] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_arguments, 4),
  [2686] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2485),
  [2688] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2349),
  [2690] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2347),
  [2692] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameters, 4),
  [2694] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameters, 4),
  [2696] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2081),
  [2698] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generic_function, 3, .production_id = 37),
  [2700] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generic_type_with_turbofish, 3, .production_id = 47),
  [2702] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_generic_function, 3, .production_id = 37),
  [2704] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_expression, 3, .production_id = 46),
  [2706] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_expression, 3, .production_id = 46),
  [2708] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameters, 6),
  [2710] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameters, 6),
  [2712] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generic_type_with_turbofish, 3, .production_id = 38),
  [2714] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generic_type, 2, .production_id = 16),
  [2716] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_generic_type, 2, .production_id = 16),
  [2718] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_type, 4, .production_id = 108),
  [2720] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_type, 4, .production_id = 108),
  [2722] = {.entry = {.count = 1, .reusable = true}}, SHIFT(730),
  [2724] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_generic_type, 2, .production_id = 17),
  [2726] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_generic_type, 2, .production_id = 17),
  [2728] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_expression, 3, .production_id = 45),
  [2730] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_expression, 3, .production_id = 45),
  [2732] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1780),
  [2734] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_type, 2, .production_id = 19),
  [2736] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_type, 2, .production_id = 19),
  [2738] = {.entry = {.count = 1, .reusable = true}}, SHIFT(726),
  [2740] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_type, 2, .production_id = 25),
  [2742] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_type, 2, .production_id = 25),
  [2744] = {.entry = {.count = 1, .reusable = true}}, SHIFT(718),
  [2746] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_arguments, 3),
  [2748] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_arguments, 3),
  [2750] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameters, 3),
  [2752] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parameters, 3),
  [2754] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_type, 2, .production_id = 23),
  [2756] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_type, 2, .production_id = 23),
  [2758] = {.entry = {.count = 1, .reusable = true}}, SHIFT(724),
  [2760] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_arguments, 3),
  [2762] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_arguments, 3),
  [2764] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_index_expression, 4),
  [2766] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_index_expression, 4),
  [2768] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_closure_expression, 4, .production_id = 89),
  [2770] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_closure_expression, 4, .production_id = 89),
  [2772] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tuple_expression, 5),
  [2774] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_tuple_expression, 5),
  [2776] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_expression, 5, .production_id = 91),
  [2778] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_expression, 5, .production_id = 91),
  [2780] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_expression, 5),
  [2782] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_expression, 5),
  [2784] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_type, 4, .production_id = 111),
  [2786] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_type, 4, .production_id = 111),
  [2788] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_type, 4, .production_id = 106),
  [2790] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_type, 4, .production_id = 106),
  [2792] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_reference_type, 4, .production_id = 103),
  [2794] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_reference_type, 4, .production_id = 103),
  [2796] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_expression, 4),
  [2798] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_expression, 4),
  [2800] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_type, 4, .production_id = 100),
  [2802] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_type, 4, .production_id = 100),
  [2804] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tuple_type, 4),
  [2806] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_tuple_type, 4),
  [2808] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_range_expression, 3),
  [2810] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_range_expression, 3),
  [2812] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tuple_expression, 4),
  [2814] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_tuple_expression, 4),
  [2816] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_await_expression, 3),
  [2818] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_await_expression, 3),
  [2820] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tuple_type, 5),
  [2822] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_tuple_type, 5),
  [2824] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_type, 5, .production_id = 142),
  [2826] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_type, 5, .production_id = 142),
  [2828] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_initializer_list, 5),
  [2830] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_initializer_list, 5),
  [2832] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_bounded_type, 3),
  [2834] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_bounded_type, 3),
  [2836] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_type, 5, .production_id = 144),
  [2838] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_type, 5, .production_id = 144),
  [2840] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_reference_type, 3, .production_id = 61),
  [2842] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_reference_type, 3, .production_id = 61),
  [2844] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_cast_expression, 3, .production_id = 43),
  [2846] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_cast_expression, 3, .production_id = 43),
  [2848] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_pointer_type, 3, .production_id = 61),
  [2850] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_pointer_type, 3, .production_id = 61),
  [2852] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_arguments, 2),
  [2854] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_arguments, 2),
  [2856] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_type, 3, .production_id = 60),
  [2858] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_type, 3, .production_id = 60),
  [2860] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tuple_type, 3),
  [2862] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_tuple_type, 3),
  [2864] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_type, 5, .production_id = 149),
  [2866] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_type, 5, .production_id = 149),
  [2868] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_type, 5, .production_id = 150),
  [2870] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_type, 5, .production_id = 150),
  [2872] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_initializer_list, 3),
  [2874] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_initializer_list, 3),
  [2876] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_initializer_list, 2),
  [2878] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_initializer_list, 2),
  [2880] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_reference_type, 2, .production_id = 22),
  [2882] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_reference_type, 2, .production_id = 22),
  [2884] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_closure_expression, 5, .production_id = 126),
  [2886] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_closure_expression, 5, .production_id = 126),
  [2888] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_initializer_list, 4),
  [2890] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_field_initializer_list, 4),
  [2892] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_lifetime, 2),
  [2894] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_lifetime, 2),
  [2896] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_unit_type, 2),
  [2898] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_unit_type, 2),
  [2900] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_expression, 3),
  [2902] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_expression, 3),
  [2904] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2445),
  [2906] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_empty_type, 1),
  [2908] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_empty_type, 1),
  [2910] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_arguments, 4),
  [2912] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_arguments, 4),
  [2914] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_arguments, 5),
  [2916] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_arguments, 5),
  [2918] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parenthesized_expression, 3),
  [2920] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_parenthesized_expression, 3),
  [2922] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_unit_expression, 2),
  [2924] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_unit_expression, 2),
  [2926] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tuple_expression, 7),
  [2928] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_tuple_expression, 7),
  [2930] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tuple_expression, 6),
  [2932] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_tuple_expression, 6),
  [2934] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_try_expression, 2),
  [2936] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_try_expression, 2),
  [2938] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_expression, 6, .production_id = 134),
  [2940] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_expression, 6, .production_id = 134),
  [2942] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_expression, 6),
  [2944] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_expression, 6),
  [2946] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_continue_expression, 2),
  [2948] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_continue_expression, 2),
  [2950] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_array_expression, 2),
  [2952] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_array_expression, 2),
  [2954] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_call_expression, 2, .production_id = 10),
  [2956] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_call_expression, 2, .production_id = 10),
  [2958] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_expression, 2, .production_id = 9),
  [2960] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_expression, 2, .production_id = 9),
  [2962] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_type, 6, .production_id = 193),
  [2964] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_function_type, 6, .production_id = 193),
  [2966] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_arguments, 6),
  [2968] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_arguments, 6),
  [2970] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_expression, 2, .production_id = 8),
  [2972] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_expression, 2, .production_id = 8),
  [2974] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_binary_expression, 3, .production_id = 42),
  [2976] = {.entry = {.count = 1, .reusable = true}}, SHIFT(154),
  [2978] = {.entry = {.count = 1, .reusable = false}}, SHIFT(177),
  [2980] = {.entry = {.count = 1, .reusable = false}}, SHIFT(175),
  [2982] = {.entry = {.count = 1, .reusable = true}}, SHIFT(671),
  [2984] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_binary_expression, 3, .production_id = 42),
  [2986] = {.entry = {.count = 1, .reusable = false}}, SHIFT(167),
  [2988] = {.entry = {.count = 1, .reusable = true}}, SHIFT(163),
  [2990] = {.entry = {.count = 1, .reusable = false}}, SHIFT(21),
  [2992] = {.entry = {.count = 1, .reusable = false}}, SHIFT(141),
  [2994] = {.entry = {.count = 1, .reusable = false}}, SHIFT(139),
  [2996] = {.entry = {.count = 1, .reusable = false}}, SHIFT(132),
  [2998] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2085),
  [3000] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_reference_expression, 2, .production_id = 7),
  [3002] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_reference_expression, 2, .production_id = 7),
  [3004] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1712),
  [3006] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_where_clause, 3),
  [3008] = {.entry = {.count = 1, .reusable = true}}, SHIFT(693),
  [3010] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2229),
  [3012] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1886),
  [3014] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2405),
  [3016] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2297),
  [3018] = {.entry = {.count = 1, .reusable = true}}, SHIFT(626),
  [3020] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2378),
  [3022] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_return_expression, 2),
  [3024] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_return_expression, 2),
  [3026] = {.entry = {.count = 1, .reusable = false}}, SHIFT(171),
  [3028] = {.entry = {.count = 1, .reusable = true}}, SHIFT(146),
  [3030] = {.entry = {.count = 1, .reusable = true}}, SHIFT(143),
  [3032] = {.entry = {.count = 1, .reusable = true}}, SHIFT(171),
  [3034] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_where_clause, 4),
  [3036] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_parameters, 4, .production_id = 62),
  [3038] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_parameters, 4, .production_id = 62),
  [3040] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_parameters, 4),
  [3042] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_parameters, 4),
  [3044] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_compound_assignment_expr, 3, .production_id = 42),
  [3046] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_compound_assignment_expr, 3, .production_id = 42),
  [3048] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_match_arm, 5, .production_id = 99),
  [3050] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_arm, 5, .production_id = 99),
  [3052] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_last_match_arm, 5, .production_id = 99),
  [3054] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_parameters, 5, .production_id = 62),
  [3056] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_parameters, 5, .production_id = 62),
  [3058] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_parameters, 3),
  [3060] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_parameters, 3),
  [3062] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_unary_expression, 2),
  [3064] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_unary_expression, 2),
  [3066] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_closure_expression, 2, .production_id = 11),
  [3068] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1069),
  [3070] = {.entry = {.count = 1, .reusable = false}}, SHIFT(172),
  [3072] = {.entry = {.count = 1, .reusable = true}}, SHIFT(129),
  [3074] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_yield_expression, 2),
  [3076] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_yield_expression, 2),
  [3078] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_match_arm, 4, .production_id = 155),
  [3080] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_arm, 4, .production_id = 155),
  [3082] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_last_match_arm, 4, .production_id = 155),
  [3084] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_parameters, 3, .production_id = 62),
  [3086] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_parameters, 3, .production_id = 62),
  [3088] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_type_parameters, 5),
  [3090] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_parameters, 5),
  [3092] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_assignment_expression, 3, .production_id = 44),
  [3094] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_assignment_expression, 3, .production_id = 44),
  [3096] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_enum_variant_list_repeat1, 2), SHIFT_REPEAT(2518),
  [3099] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_break_expression, 3),
  [3101] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_break_expression, 3),
  [3103] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_reference_expression, 3, .production_id = 32),
  [3105] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_reference_expression, 3, .production_id = 32),
  [3107] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_closure_expression, 3, .production_id = 33),
  [3109] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1769),
  [3111] = {.entry = {.count = 1, .reusable = false}}, SHIFT(753),
  [3113] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2409),
  [3115] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1600),
  [3117] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2491),
  [3119] = {.entry = {.count = 1, .reusable = false}}, SHIFT(754),
  [3121] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1626),
  [3123] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1326),
  [3125] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1586),
  [3127] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1319),
  [3129] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1994),
  [3131] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1416),
  [3133] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1411),
  [3135] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1417),
  [3137] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1417),
  [3139] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1409),
  [3141] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1406),
  [3143] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1412),
  [3145] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1412),
  [3147] = {.entry = {.count = 1, .reusable = true}}, SHIFT(114),
  [3149] = {.entry = {.count = 1, .reusable = true}}, SHIFT(93),
  [3151] = {.entry = {.count = 1, .reusable = true}}, SHIFT(158),
  [3153] = {.entry = {.count = 1, .reusable = false}}, SHIFT(18),
  [3155] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2000),
  [3157] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2526),
  [3159] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_visibility_modifier, 1),
  [3161] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_visibility_modifier, 1),
  [3163] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2486),
  [3165] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_visibility_modifier, 4),
  [3167] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_visibility_modifier, 4),
  [3169] = {.entry = {.count = 1, .reusable = true}}, SHIFT(102),
  [3171] = {.entry = {.count = 1, .reusable = true}}, SHIFT(92),
  [3173] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1809),
  [3175] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_visibility_modifier, 5, .production_id = 114),
  [3177] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_visibility_modifier, 5, .production_id = 114),
  [3179] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_visibility_modifier, 5),
  [3181] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_visibility_modifier, 5),
  [3183] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2502),
  [3185] = {.entry = {.count = 1, .reusable = true}}, SHIFT(47),
  [3187] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1740),
  [3189] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2179),
  [3191] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1744),
  [3193] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2204),
  [3195] = {.entry = {.count = 1, .reusable = true}}, SHIFT(813),
  [3197] = {.entry = {.count = 1, .reusable = true}}, SHIFT(45),
  [3199] = {.entry = {.count = 1, .reusable = false}}, SHIFT(152),
  [3201] = {.entry = {.count = 1, .reusable = false}}, SHIFT(153),
  [3203] = {.entry = {.count = 1, .reusable = false}}, SHIFT(169),
  [3205] = {.entry = {.count = 1, .reusable = false}}, SHIFT(165),
  [3207] = {.entry = {.count = 1, .reusable = true}}, SHIFT(162),
  [3209] = {.entry = {.count = 1, .reusable = false}}, SHIFT(26),
  [3211] = {.entry = {.count = 1, .reusable = true}}, SHIFT(136),
  [3213] = {.entry = {.count = 1, .reusable = true}}, SHIFT(128),
  [3215] = {.entry = {.count = 1, .reusable = false}}, SHIFT(110),
  [3217] = {.entry = {.count = 1, .reusable = false}}, SHIFT(109),
  [3219] = {.entry = {.count = 1, .reusable = true}}, SHIFT(169),
  [3221] = {.entry = {.count = 1, .reusable = false}}, SHIFT(106),
  [3223] = {.entry = {.count = 1, .reusable = true}}, SHIFT(441),
  [3225] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2307),
  [3227] = {.entry = {.count = 1, .reusable = false}}, SHIFT(166),
  [3229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(100),
  [3231] = {.entry = {.count = 1, .reusable = false}}, SHIFT(31),
  [3233] = {.entry = {.count = 1, .reusable = true}}, SHIFT(105),
  [3235] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1090),
  [3237] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2189),
  [3239] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1084),
  [3241] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2178),
  [3243] = {.entry = {.count = 1, .reusable = true}}, SHIFT(260),
  [3245] = {.entry = {.count = 1, .reusable = true}}, SHIFT(337),
  [3247] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1721),
  [3249] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1207),
  [3251] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1964),
  [3253] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2078),
  [3255] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1756),
  [3257] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2375),
  [3259] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2089),
  [3261] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1706),
  [3263] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1706),
  [3265] = {.entry = {.count = 1, .reusable = true}}, SHIFT(339),
  [3267] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2266),
  [3269] = {.entry = {.count = 1, .reusable = true}}, SHIFT(365),
  [3271] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_array_expression_repeat1, 2),
  [3273] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_last_match_arm, 4, .production_id = 99),
  [3275] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1137),
  [3277] = {.entry = {.count = 1, .reusable = true}}, SHIFT(469),
  [3279] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_variant, 5, .production_id = 219),
  [3281] = {.entry = {.count = 1, .reusable = true}}, SHIFT(54),
  [3283] = {.entry = {.count = 1, .reusable = true}}, SHIFT(213),
  [3285] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_arguments_repeat1, 2),
  [3287] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_initializer, 3, .production_id = 127),
  [3289] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1052),
  [3291] = {.entry = {.count = 1, .reusable = true}}, SHIFT(71),
  [3293] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_initializer, 4, .production_id = 172),
  [3295] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_last_match_arm, 3, .production_id = 155),
  [3297] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1149),
  [3299] = {.entry = {.count = 1, .reusable = true}}, SHIFT(959),
  [3301] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2160),
  [3303] = {.entry = {.count = 1, .reusable = true}}, SHIFT(310),
  [3305] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2302),
  [3307] = {.entry = {.count = 1, .reusable = true}}, SHIFT(749),
  [3309] = {.entry = {.count = 1, .reusable = true}}, SHIFT(967),
  [3311] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1414),
  [3313] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1061),
  [3315] = {.entry = {.count = 1, .reusable = true}}, SHIFT(854),
  [3317] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2146),
  [3319] = {.entry = {.count = 1, .reusable = true}}, SHIFT(391),
  [3321] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_variant, 4, .production_id = 185),
  [3323] = {.entry = {.count = 1, .reusable = true}}, SHIFT(477),
  [3325] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2174),
  [3327] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_variant, 3, .production_id = 135),
  [3329] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_variant, 4, .production_id = 184),
  [3331] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_arguments_repeat1, 3),
  [3333] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1099),
  [3335] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_base_field_initializer, 2),
  [3337] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute, 3, .production_id = 125),
  [3339] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1403),
  [3341] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1064),
  [3343] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_let_condition, 4, .production_id = 99),
  [3345] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1091),
  [3347] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__let_chain, 3),
  [3349] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1065),
  [3351] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__condition, 1),
  [3353] = {.entry = {.count = 1, .reusable = true}}, SHIFT(58),
  [3355] = {.entry = {.count = 1, .reusable = true}}, SHIFT(86),
  [3357] = {.entry = {.count = 1, .reusable = true}}, SHIFT(454),
  [3359] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1050),
  [3361] = {.entry = {.count = 1, .reusable = true}}, SHIFT(816),
  [3363] = {.entry = {.count = 1, .reusable = true}}, SHIFT(906),
  [3365] = {.entry = {.count = 1, .reusable = true}}, SHIFT(316),
  [3367] = {.entry = {.count = 1, .reusable = true}}, SHIFT(922),
  [3369] = {.entry = {.count = 1, .reusable = true}}, SHIFT(442),
  [3371] = {.entry = {.count = 1, .reusable = true}}, SHIFT(842),
  [3373] = {.entry = {.count = 1, .reusable = true}}, SHIFT(459),
  [3375] = {.entry = {.count = 1, .reusable = true}}, SHIFT(940),
  [3377] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1159),
  [3379] = {.entry = {.count = 1, .reusable = true}}, SHIFT(59),
  [3381] = {.entry = {.count = 1, .reusable = true}}, SHIFT(417),
  [3383] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2004),
  [3385] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute, 3, .production_id = 32),
  [3387] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1160),
  [3389] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1070),
  [3391] = {.entry = {.count = 1, .reusable = true}}, SHIFT(359),
  [3393] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1908),
  [3395] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_lifetimes, 6),
  [3397] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_lifetimes, 6),
  [3399] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_lifetimes, 4),
  [3401] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_lifetimes, 4),
  [3403] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_for_lifetimes, 5),
  [3405] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_for_lifetimes, 5),
  [3407] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1608),
  [3409] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1578),
  [3411] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1603),
  [3413] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1603),
  [3415] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2219),
  [3417] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2220),
  [3419] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2215),
  [3421] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2215),
  [3423] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2109),
  [3425] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2108),
  [3427] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2110),
  [3429] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2110),
  [3431] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__pattern, 1),
  [3433] = {.entry = {.count = 1, .reusable = true}}, SHIFT(560),
  [3435] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1529),
  [3437] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__pattern, 1),
  [3439] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1779),
  [3441] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2027),
  [3443] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1174),
  [3445] = {.entry = {.count = 1, .reusable = true}}, SHIFT(238),
  [3447] = {.entry = {.count = 1, .reusable = true}}, SHIFT(604),
  [3449] = {.entry = {.count = 1, .reusable = true}}, SHIFT(182),
  [3451] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2014),
  [3453] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1526),
  [3455] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1583),
  [3457] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2396),
  [3459] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2254),
  [3461] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2397),
  [3463] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1884),
  [3465] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2398),
  [3467] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2393),
  [3469] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2557),
  [3471] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2399),
  [3473] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1573),
  [3475] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1298),
  [3477] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1472),
  [3479] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2016),
  [3481] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1569),
  [3483] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2443),
  [3485] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2123),
  [3487] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2441),
  [3489] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2087),
  [3491] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2439),
  [3493] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2463),
  [3495] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2438),
  [3497] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2428),
  [3499] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1585),
  [3501] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1297),
  [3503] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1471),
  [3505] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1819),
  [3507] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__pattern, 1, .production_id = 1),
  [3509] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__pattern, 1, .production_id = 1),
  [3511] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1760),
  [3513] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2419),
  [3515] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1173),
  [3517] = {.entry = {.count = 1, .reusable = true}}, SHIFT(661),
  [3519] = {.entry = {.count = 1, .reusable = true}}, SHIFT(719),
  [3521] = {.entry = {.count = 1, .reusable = true}}, SHIFT(639),
  [3523] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2415),
  [3525] = {.entry = {.count = 1, .reusable = true}}, SHIFT(720),
  [3527] = {.entry = {.count = 1, .reusable = true}}, SHIFT(664),
  [3529] = {.entry = {.count = 1, .reusable = true}}, SHIFT(646),
  [3531] = {.entry = {.count = 1, .reusable = true}}, SHIFT(666),
  [3533] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2498),
  [3535] = {.entry = {.count = 1, .reusable = true}}, SHIFT(181),
  [3537] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__type, 1, .production_id = 4), REDUCE(sym__pattern, 1),
  [3540] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1906),
  [3542] = {.entry = {.count = 1, .reusable = true}}, SHIFT(180),
  [3544] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1885),
  [3546] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2490),
  [3548] = {.entry = {.count = 1, .reusable = true}}, SHIFT(179),
  [3550] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1863),
  [3552] = {.entry = {.count = 1, .reusable = false}}, SHIFT(610),
  [3554] = {.entry = {.count = 1, .reusable = true}}, SHIFT(656),
  [3556] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1521),
  [3558] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1150),
  [3560] = {.entry = {.count = 1, .reusable = true}}, SHIFT(578),
  [3562] = {.entry = {.count = 1, .reusable = false}}, SHIFT(578),
  [3564] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_negative_literal, 2),
  [3566] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_negative_literal, 2),
  [3568] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2334),
  [3570] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__literal_pattern, 1),
  [3572] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym__literal_pattern, 1),
  [3574] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_range_pattern, 3, .production_id = 58),
  [3576] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_range_pattern, 3, .production_id = 58),
  [3578] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2430),
  [3580] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2348),
  [3582] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_range_pattern, 3),
  [3584] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_range_pattern, 3),
  [3586] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2032),
  [3588] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_range_pattern, 3, .production_id = 1),
  [3590] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_range_pattern, 3, .production_id = 1),
  [3592] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1857),
  [3594] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_range_pattern, 3, .production_id = 54),
  [3596] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_range_pattern, 3, .production_id = 54),
  [3598] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_mut_pattern, 2),
  [3600] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_mut_pattern, 2),
  [3602] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_scoped_identifier, 3, .production_id = 12), REDUCE(sym_scoped_type_identifier, 3, .production_id = 13),
  [3605] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_pattern, 4, .production_id = 55),
  [3607] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_pattern, 4, .production_id = 55),
  [3609] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_remaining_field_pattern, 1),
  [3611] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_remaining_field_pattern, 1),
  [3613] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_scoped_identifier, 3, .production_id = 35), REDUCE(sym_scoped_type_identifier, 3, .production_id = 36),
  [3616] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_scoped_identifier, 2, .production_id = 5), REDUCE(sym_scoped_type_identifier, 2, .production_id = 6),
  [3619] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_reference_pattern, 3),
  [3621] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_reference_pattern, 3),
  [3623] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tuple_struct_pattern, 3, .production_id = 55),
  [3625] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_tuple_struct_pattern, 3, .production_id = 55),
  [3627] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tuple_pattern, 4),
  [3629] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_tuple_pattern, 4),
  [3631] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_pattern, 3, .production_id = 56),
  [3633] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_pattern, 3, .production_id = 56),
  [3635] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_slice_pattern, 4),
  [3637] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_slice_pattern, 4),
  [3639] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_scoped_identifier, 3, .production_id = 40), REDUCE(sym_scoped_type_identifier, 3, .production_id = 41),
  [3642] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tuple_pattern, 3),
  [3644] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_tuple_pattern, 3),
  [3646] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_slice_pattern, 5),
  [3648] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_slice_pattern, 5),
  [3650] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tuple_struct_pattern, 6, .production_id = 55),
  [3652] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_tuple_struct_pattern, 6, .production_id = 55),
  [3654] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_pattern, 6, .production_id = 56),
  [3656] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_pattern, 6, .production_id = 56),
  [3658] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2317),
  [3660] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tuple_pattern, 2),
  [3662] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_tuple_pattern, 2),
  [3664] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_pattern, 5, .production_id = 55),
  [3666] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_pattern, 5, .production_id = 55),
  [3668] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_pattern, 6, .production_id = 55),
  [3670] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_pattern, 6, .production_id = 55),
  [3672] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_slice_pattern, 2),
  [3674] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_slice_pattern, 2),
  [3676] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_pattern, 4, .production_id = 56),
  [3678] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_pattern, 4, .production_id = 56),
  [3680] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tuple_struct_pattern, 5, .production_id = 55),
  [3682] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_tuple_struct_pattern, 5, .production_id = 55),
  [3684] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_pattern, 5, .production_id = 56),
  [3686] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_pattern, 5, .production_id = 56),
  [3688] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_slice_pattern, 3),
  [3690] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_slice_pattern, 3),
  [3692] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_ref_pattern, 2),
  [3694] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_ref_pattern, 2),
  [3696] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_captured_pattern, 3),
  [3698] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_captured_pattern, 3),
  [3700] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_reference_pattern, 2),
  [3702] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_reference_pattern, 2),
  [3704] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tuple_struct_pattern, 4, .production_id = 55),
  [3706] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_tuple_struct_pattern, 4, .production_id = 55),
  [3708] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_struct_pattern, 3, .production_id = 55),
  [3710] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_struct_pattern, 3, .production_id = 55),
  [3712] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_or_pattern, 3),
  [3714] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_or_pattern, 3),
  [3716] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_tuple_pattern, 5),
  [3718] = {.entry = {.count = 1, .reusable = false}}, REDUCE(sym_tuple_pattern, 5),
  [3720] = {.entry = {.count = 1, .reusable = true}}, SHIFT(643),
  [3722] = {.entry = {.count = 1, .reusable = true}}, SHIFT(662),
  [3724] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_extern_modifier, 1),
  [3726] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2395),
  [3728] = {.entry = {.count = 1, .reusable = true}}, SHIFT(721),
  [3730] = {.entry = {.count = 1, .reusable = true}}, SHIFT(694),
  [3732] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2452),
  [3734] = {.entry = {.count = 1, .reusable = true}}, SHIFT(679),
  [3736] = {.entry = {.count = 1, .reusable = true}}, SHIFT(703),
  [3738] = {.entry = {.count = 1, .reusable = true}}, SHIFT(736),
  [3740] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2328),
  [3742] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2425),
  [3744] = {.entry = {.count = 1, .reusable = true}}, SHIFT(648),
  [3746] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1536),
  [3748] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1914),
  [3750] = {.entry = {.count = 1, .reusable = true}}, SHIFT(987),
  [3752] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2355),
  [3754] = {.entry = {.count = 1, .reusable = true}}, SHIFT(137),
  [3756] = {.entry = {.count = 1, .reusable = true}}, SHIFT(635),
  [3758] = {.entry = {.count = 1, .reusable = true}}, SHIFT(677),
  [3760] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2406),
  [3762] = {.entry = {.count = 1, .reusable = true}}, SHIFT(773),
  [3764] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2451),
  [3766] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2327),
  [3768] = {.entry = {.count = 1, .reusable = true}}, SHIFT(256),
  [3770] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2403),
  [3772] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1566),
  [3774] = {.entry = {.count = 1, .reusable = true}}, SHIFT(427),
  [3776] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2316),
  [3778] = {.entry = {.count = 1, .reusable = true}}, SHIFT(965),
  [3780] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2467),
  [3782] = {.entry = {.count = 1, .reusable = true}}, SHIFT(695),
  [3784] = {.entry = {.count = 1, .reusable = true}}, SHIFT(737),
  [3786] = {.entry = {.count = 1, .reusable = true}}, SHIFT(283),
  [3788] = {.entry = {.count = 1, .reusable = true}}, SHIFT(243),
  [3790] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1755),
  [3792] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2340),
  [3794] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1139),
  [3796] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2164),
  [3798] = {.entry = {.count = 1, .reusable = true}}, SHIFT(887),
  [3800] = {.entry = {.count = 1, .reusable = true}}, SHIFT(545),
  [3802] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1478),
  [3804] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1162),
  [3806] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1542),
  [3808] = {.entry = {.count = 1, .reusable = true}}, SHIFT(767),
  [3810] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1188),
  [3812] = {.entry = {.count = 1, .reusable = true}}, SHIFT(464),
  [3814] = {.entry = {.count = 1, .reusable = true}}, SHIFT(811),
  [3816] = {.entry = {.count = 1, .reusable = true}}, SHIFT(705),
  [3818] = {.entry = {.count = 1, .reusable = true}}, SHIFT(982),
  [3820] = {.entry = {.count = 1, .reusable = true}}, SHIFT(688),
  [3822] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1056),
  [3824] = {.entry = {.count = 1, .reusable = true}}, SHIFT(320),
  [3826] = {.entry = {.count = 1, .reusable = true}}, SHIFT(327),
  [3828] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1479),
  [3830] = {.entry = {.count = 1, .reusable = true}}, SHIFT(680),
  [3832] = {.entry = {.count = 1, .reusable = true}}, SHIFT(768),
  [3834] = {.entry = {.count = 1, .reusable = true}}, SHIFT(642),
  [3836] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1135),
  [3838] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1033),
  [3840] = {.entry = {.count = 1, .reusable = true}}, SHIFT(242),
  [3842] = {.entry = {.count = 1, .reusable = true}}, SHIFT(384),
  [3844] = {.entry = {.count = 1, .reusable = true}}, SHIFT(261),
  [3846] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1011),
  [3848] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1888),
  [3850] = {.entry = {.count = 1, .reusable = true}}, SHIFT(729),
  [3852] = {.entry = {.count = 1, .reusable = true}}, SHIFT(652),
  [3854] = {.entry = {.count = 1, .reusable = true}}, SHIFT(326),
  [3856] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1151),
  [3858] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2497),
  [3860] = {.entry = {.count = 1, .reusable = false}}, REDUCE(aux_sym_function_modifiers_repeat1, 1),
  [3862] = {.entry = {.count = 1, .reusable = true}}, SHIFT(971),
  [3864] = {.entry = {.count = 1, .reusable = true}}, SHIFT(421),
  [3866] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1117),
  [3868] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1709),
  [3870] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1725),
  [3872] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1726),
  [3874] = {.entry = {.count = 1, .reusable = true}}, SHIFT(466),
  [3876] = {.entry = {.count = 1, .reusable = true}}, SHIFT(774),
  [3878] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1753),
  [3880] = {.entry = {.count = 1, .reusable = true}}, SHIFT(374),
  [3882] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1134),
  [3884] = {.entry = {.count = 1, .reusable = true}}, SHIFT(924),
  [3886] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2211),
  [3888] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_function_modifiers_repeat1, 2), SHIFT_REPEAT(1525),
  [3891] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_function_modifiers_repeat1, 2),
  [3893] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_function_modifiers_repeat1, 2), SHIFT_REPEAT(1561),
  [3896] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1525),
  [3898] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_function_modifiers, 1),
  [3900] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1561),
  [3902] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1909),
  [3904] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1922),
  [3906] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1429),
  [3908] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2358),
  [3910] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2270),
  [3912] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2368),
  [3914] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_extern_modifier, 2),
  [3916] = {.entry = {.count = 1, .reusable = true}}, SHIFT(610),
  [3918] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1543),
  [3920] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1456),
  [3922] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2376),
  [3924] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_trait_bounds, 2),
  [3926] = {.entry = {.count = 1, .reusable = true}}, SHIFT(594),
  [3928] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1845),
  [3930] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_trait_bounds_repeat1, 2),
  [3932] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_trait_bounds_repeat1, 2), SHIFT_REPEAT(594),
  [3935] = {.entry = {.count = 1, .reusable = true}}, SHIFT(595),
  [3937] = {.entry = {.count = 1, .reusable = true}}, SHIFT(596),
  [3939] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1687),
  [3941] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2006),
  [3943] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1677),
  [3945] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2023),
  [3947] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_macro_definition_repeat1, 2), SHIFT_REPEAT(496),
  [3950] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_macro_definition_repeat1, 2),
  [3952] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_macro_definition_repeat1, 2), SHIFT_REPEAT(498),
  [3955] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_macro_definition_repeat1, 2), SHIFT_REPEAT(499),
  [3958] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__type, 1), REDUCE(sym__pattern, 1, .production_id = 1),
  [3961] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1803),
  [3963] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_trait_bounds, 3),
  [3965] = {.entry = {.count = 1, .reusable = true}}, SHIFT(630),
  [3967] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1862),
  [3969] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_variant, 1, .production_id = 49),
  [3971] = {.entry = {.count = 1, .reusable = true}}, SHIFT(99),
  [3973] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2381),
  [3975] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2341),
  [3977] = {.entry = {.count = 1, .reusable = true}}, SHIFT(824),
  [3979] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_removed_trait_bound, 2),
  [3981] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2424),
  [3983] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1438),
  [3985] = {.entry = {.count = 1, .reusable = true}}, SHIFT(496),
  [3987] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2440),
  [3989] = {.entry = {.count = 1, .reusable = true}}, SHIFT(498),
  [3991] = {.entry = {.count = 1, .reusable = true}}, SHIFT(499),
  [3993] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2519),
  [3995] = {.entry = {.count = 1, .reusable = true}}, SHIFT(531),
  [3997] = {.entry = {.count = 1, .reusable = true}}, SHIFT(526),
  [3999] = {.entry = {.count = 1, .reusable = true}}, SHIFT(521),
  [4001] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute, 1, .production_id = 1),
  [4003] = {.entry = {.count = 1, .reusable = true}}, SHIFT(118),
  [4005] = {.entry = {.count = 1, .reusable = true}}, SHIFT(881),
  [4007] = {.entry = {.count = 1, .reusable = true}}, SHIFT(882),
  [4009] = {.entry = {.count = 1, .reusable = true}}, SHIFT(375),
  [4011] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2462),
  [4013] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2536),
  [4015] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2442),
  [4017] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2331),
  [4019] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2470),
  [4021] = {.entry = {.count = 1, .reusable = true}}, SHIFT(271),
  [4023] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1442),
  [4025] = {.entry = {.count = 1, .reusable = true}}, SHIFT(968),
  [4027] = {.entry = {.count = 1, .reusable = true}}, SHIFT(373),
  [4029] = {.entry = {.count = 1, .reusable = true}}, SHIFT(969),
  [4031] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_higher_ranked_trait_bound, 3, .production_id = 66),
  [4033] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2323),
  [4035] = {.entry = {.count = 1, .reusable = true}}, SHIFT(455),
  [4037] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute, 1),
  [4039] = {.entry = {.count = 1, .reusable = true}}, SHIFT(138),
  [4041] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2076),
  [4043] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2461),
  [4045] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2473),
  [4047] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2435),
  [4049] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1448),
  [4051] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2321),
  [4053] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_variant, 2, .production_id = 5),
  [4055] = {.entry = {.count = 1, .reusable = true}}, SHIFT(160),
  [4057] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1441),
  [4059] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2065),
  [4061] = {.entry = {.count = 1, .reusable = true}}, SHIFT(406),
  [4063] = {.entry = {.count = 1, .reusable = true}}, SHIFT(448),
  [4065] = {.entry = {.count = 1, .reusable = true}}, SHIFT(910),
  [4067] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1037),
  [4069] = {.entry = {.count = 1, .reusable = true}}, SHIFT(16),
  [4071] = {.entry = {.count = 1, .reusable = true}}, SHIFT(715),
  [4073] = {.entry = {.count = 1, .reusable = true}}, SHIFT(864),
  [4075] = {.entry = {.count = 1, .reusable = true}}, SHIFT(669),
  [4077] = {.entry = {.count = 1, .reusable = true}}, SHIFT(347),
  [4079] = {.entry = {.count = 1, .reusable = true}}, SHIFT(396),
  [4081] = {.entry = {.count = 1, .reusable = true}}, SHIFT(5),
  [4083] = {.entry = {.count = 1, .reusable = true}}, SHIFT(708),
  [4085] = {.entry = {.count = 1, .reusable = true}}, SHIFT(333),
  [4087] = {.entry = {.count = 1, .reusable = true}}, SHIFT(913),
  [4089] = {.entry = {.count = 1, .reusable = true}}, SHIFT(340),
  [4091] = {.entry = {.count = 1, .reusable = true}}, SHIFT(927),
  [4093] = {.entry = {.count = 1, .reusable = true}}, SHIFT(899),
  [4095] = {.entry = {.count = 1, .reusable = true}}, SHIFT(462),
  [4097] = {.entry = {.count = 1, .reusable = true}}, SHIFT(463),
  [4099] = {.entry = {.count = 1, .reusable = true}}, SHIFT(259),
  [4101] = {.entry = {.count = 1, .reusable = true}}, SHIFT(667),
  [4103] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_parameters, 3), REDUCE(sym_tuple_struct_pattern, 4, .production_id = 55),
  [4106] = {.entry = {.count = 1, .reusable = true}}, SHIFT(771),
  [4108] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2059),
  [4110] = {.entry = {.count = 1, .reusable = true}}, SHIFT(262),
  [4112] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1481),
  [4114] = {.entry = {.count = 1, .reusable = true}}, SHIFT(345),
  [4116] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1500),
  [4118] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1141),
  [4120] = {.entry = {.count = 1, .reusable = true}}, SHIFT(386),
  [4122] = {.entry = {.count = 1, .reusable = true}}, SHIFT(722),
  [4124] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1003),
  [4126] = {.entry = {.count = 1, .reusable = true}}, SHIFT(934),
  [4128] = {.entry = {.count = 1, .reusable = true}}, SHIFT(977),
  [4130] = {.entry = {.count = 1, .reusable = true}}, SHIFT(657),
  [4132] = {.entry = {.count = 1, .reusable = true}}, SHIFT(249),
  [4134] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1007),
  [4136] = {.entry = {.count = 1, .reusable = true}}, SHIFT(685),
  [4138] = {.entry = {.count = 1, .reusable = true}}, SHIFT(830),
  [4140] = {.entry = {.count = 1, .reusable = true}}, SHIFT(709),
  [4142] = {.entry = {.count = 1, .reusable = true}}, SHIFT(424),
  [4144] = {.entry = {.count = 1, .reusable = true}}, SHIFT(415),
  [4146] = {.entry = {.count = 1, .reusable = true}}, SHIFT(739),
  [4148] = {.entry = {.count = 1, .reusable = true}}, SHIFT(311),
  [4150] = {.entry = {.count = 1, .reusable = true}}, SHIFT(362),
  [4152] = {.entry = {.count = 1, .reusable = true}}, SHIFT(350),
  [4154] = {.entry = {.count = 1, .reusable = true}}, SHIFT(839),
  [4156] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_self_parameter, 1),
  [4158] = {.entry = {.count = 1, .reusable = false}}, SHIFT(742),
  [4160] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2336),
  [4162] = {.entry = {.count = 1, .reusable = true}}, SHIFT(919),
  [4164] = {.entry = {.count = 1, .reusable = true}}, SHIFT(288),
  [4166] = {.entry = {.count = 1, .reusable = true}}, SHIFT(654),
  [4168] = {.entry = {.count = 1, .reusable = true}}, SHIFT(859),
  [4170] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1122),
  [4172] = {.entry = {.count = 1, .reusable = true}}, SHIFT(717),
  [4174] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_self_parameter, 2),
  [4176] = {.entry = {.count = 1, .reusable = false}}, SHIFT(738),
  [4178] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2416),
  [4180] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_parameters, 2), REDUCE(sym_tuple_struct_pattern, 3, .production_id = 55),
  [4183] = {.entry = {.count = 1, .reusable = true}}, SHIFT(998),
  [4185] = {.entry = {.count = 1, .reusable = true}}, SHIFT(996),
  [4187] = {.entry = {.count = 1, .reusable = true}}, SHIFT(936),
  [4189] = {.entry = {.count = 1, .reusable = true}}, SHIFT(809),
  [4191] = {.entry = {.count = 1, .reusable = true}}, SHIFT(446),
  [4193] = {.entry = {.count = 1, .reusable = true}}, SHIFT(665),
  [4195] = {.entry = {.count = 1, .reusable = true}}, SHIFT(957),
  [4197] = {.entry = {.count = 1, .reusable = true}}, SHIFT(676),
  [4199] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1480),
  [4201] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1516),
  [4203] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1741),
  [4205] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1057),
  [4207] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2350),
  [4209] = {.entry = {.count = 1, .reusable = true}}, SHIFT(371),
  [4211] = {.entry = {.count = 1, .reusable = true}}, SHIFT(475),
  [4213] = {.entry = {.count = 1, .reusable = true}}, SHIFT(434),
  [4215] = {.entry = {.count = 1, .reusable = true}}, SHIFT(298),
  [4217] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1106),
  [4219] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1488),
  [4221] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1179),
  [4223] = {.entry = {.count = 1, .reusable = true}}, SHIFT(389),
  [4225] = {.entry = {.count = 1, .reusable = true}}, SHIFT(945),
  [4227] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2319),
  [4229] = {.entry = {.count = 1, .reusable = true}}, SHIFT(876),
  [4231] = {.entry = {.count = 1, .reusable = true}}, SHIFT(699),
  [4233] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1094),
  [4235] = {.entry = {.count = 1, .reusable = true}}, SHIFT(356),
  [4237] = {.entry = {.count = 1, .reusable = true}}, SHIFT(687),
  [4239] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1096),
  [4241] = {.entry = {.count = 1, .reusable = true}}, SHIFT(306),
  [4243] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_ordered_field_declaration_list, 5, .production_id = 198),
  [4245] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__use_clause, 1),
  [4247] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2366),
  [4249] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1851),
  [4251] = {.entry = {.count = 1, .reusable = true}}, SHIFT(806),
  [4253] = {.entry = {.count = 1, .reusable = true}}, SHIFT(186),
  [4255] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1367),
  [4257] = {.entry = {.count = 1, .reusable = true}}, SHIFT(188),
  [4259] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_ordered_field_declaration_list, 6, .production_id = 103),
  [4261] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_ordered_field_declaration_list, 6, .production_id = 198),
  [4263] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1515),
  [4265] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1742),
  [4267] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__pattern, 1), SHIFT(1367),
  [4270] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__pattern, 1), SHIFT(188),
  [4273] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_self_parameter, 3),
  [4275] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_ordered_field_declaration_list, 4, .production_id = 158),
  [4277] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1629),
  [4279] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_ordered_field_declaration_list, 5, .production_id = 61),
  [4281] = {.entry = {.count = 1, .reusable = true}}, SHIFT(185),
  [4283] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_ordered_field_declaration_list, 4, .production_id = 22),
  [4285] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_ordered_field_declaration_list, 2),
  [4287] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym__type, 1), REDUCE(sym__pattern, 1),
  [4290] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_ordered_field_declaration_list, 6, .production_id = 226),
  [4292] = {.entry = {.count = 1, .reusable = true}}, SHIFT(187),
  [4294] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_ordered_field_declaration_list, 5, .production_id = 158),
  [4296] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_ordered_field_declaration_list, 4, .production_id = 61),
  [4298] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_ordered_field_declaration_list, 3, .production_id = 22),
  [4300] = {.entry = {.count = 1, .reusable = true}}, SHIFT(254),
  [4302] = {.entry = {.count = 1, .reusable = true}}, SHIFT(663),
  [4304] = {.entry = {.count = 1, .reusable = true}}, SHIFT(119),
  [4306] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2235),
  [4308] = {.entry = {.count = 1, .reusable = true}}, SHIFT(361),
  [4310] = {.entry = {.count = 1, .reusable = true}}, SHIFT(692),
  [4312] = {.entry = {.count = 1, .reusable = true}}, SHIFT(170),
  [4314] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2210),
  [4316] = {.entry = {.count = 1, .reusable = true}}, SHIFT(602),
  [4318] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_ordered_field_declaration_list, 3),
  [4320] = {.entry = {.count = 2, .reusable = true}}, REDUCE(sym_unit_type, 2), REDUCE(sym_tuple_pattern, 2),
  [4323] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_ordered_field_declaration_list, 5, .production_id = 103),
  [4325] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_ordered_field_declaration_list, 7, .production_id = 226),
  [4327] = {.entry = {.count = 1, .reusable = true}}, SHIFT(916),
  [4329] = {.entry = {.count = 1, .reusable = true}}, SHIFT(682),
  [4331] = {.entry = {.count = 1, .reusable = true}}, SHIFT(112),
  [4333] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2125),
  [4335] = {.entry = {.count = 1, .reusable = true}}, SHIFT(684),
  [4337] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1454),
  [4339] = {.entry = {.count = 1, .reusable = true}}, SHIFT(742),
  [4341] = {.entry = {.count = 1, .reusable = true}}, SHIFT(576),
  [4343] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1630),
  [4345] = {.entry = {.count = 1, .reusable = true}}, SHIFT(988),
  [4347] = {.entry = {.count = 1, .reusable = true}}, SHIFT(651),
  [4349] = {.entry = {.count = 1, .reusable = true}}, SHIFT(115),
  [4351] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2184),
  [4353] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_type_parameters_repeat1, 2, .production_id = 62),
  [4355] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__use_clause, 1, .production_id = 1),
  [4357] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2371),
  [4359] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1854),
  [4361] = {.entry = {.count = 1, .reusable = true}}, SHIFT(641),
  [4363] = {.entry = {.count = 1, .reusable = true}}, SHIFT(509),
  [4365] = {.entry = {.count = 1, .reusable = true}}, SHIFT(508),
  [4367] = {.entry = {.count = 1, .reusable = true}}, SHIFT(507),
  [4369] = {.entry = {.count = 1, .reusable = true}}, SHIFT(614),
  [4371] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1774),
  [4373] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1872),
  [4375] = {.entry = {.count = 1, .reusable = false}}, SHIFT(572),
  [4377] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter, 3, .production_id = 85),
  [4379] = {.entry = {.count = 1, .reusable = true}}, SHIFT(532),
  [4381] = {.entry = {.count = 1, .reusable = true}}, SHIFT(517),
  [4383] = {.entry = {.count = 1, .reusable = true}}, SHIFT(511),
  [4385] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_parameters_repeat1, 3),
  [4387] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_type_parameters_repeat1, 2),
  [4389] = {.entry = {.count = 1, .reusable = true}}, SHIFT(601),
  [4391] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1839),
  [4393] = {.entry = {.count = 1, .reusable = true}}, SHIFT(39),
  [4395] = {.entry = {.count = 1, .reusable = true}}, SHIFT(958),
  [4397] = {.entry = {.count = 1, .reusable = true}}, SHIFT(135),
  [4399] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2158),
  [4401] = {.entry = {.count = 1, .reusable = true}}, SHIFT(516),
  [4403] = {.entry = {.count = 1, .reusable = true}}, SHIFT(515),
  [4405] = {.entry = {.count = 1, .reusable = true}}, SHIFT(514),
  [4407] = {.entry = {.count = 1, .reusable = true}}, SHIFT(231),
  [4409] = {.entry = {.count = 1, .reusable = true}}, SHIFT(804),
  [4411] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1434),
  [4413] = {.entry = {.count = 1, .reusable = true}}, SHIFT(570),
  [4415] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_where_clause, 2),
  [4417] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1131),
  [4419] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1450),
  [4421] = {.entry = {.count = 1, .reusable = true}}, SHIFT(573),
  [4423] = {.entry = {.count = 1, .reusable = true}}, SHIFT(989),
  [4425] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1133),
  [4427] = {.entry = {.count = 1, .reusable = true}}, SHIFT(230),
  [4429] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1344),
  [4431] = {.entry = {.count = 1, .reusable = true}}, SHIFT(49),
  [4433] = {.entry = {.count = 1, .reusable = true}}, SHIFT(562),
  [4435] = {.entry = {.count = 1, .reusable = true}}, SHIFT(209),
  [4437] = {.entry = {.count = 1, .reusable = true}}, SHIFT(575),
  [4439] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1805),
  [4441] = {.entry = {.count = 1, .reusable = true}}, SHIFT(523),
  [4443] = {.entry = {.count = 1, .reusable = true}}, SHIFT(520),
  [4445] = {.entry = {.count = 1, .reusable = true}}, SHIFT(505),
  [4447] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_tuple_pattern_repeat1, 2),
  [4449] = {.entry = {.count = 1, .reusable = true}}, SHIFT(567),
  [4451] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_tuple_pattern_repeat1, 2), SHIFT_REPEAT(606),
  [4454] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1312),
  [4456] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2480),
  [4458] = {.entry = {.count = 1, .reusable = true}}, SHIFT(855),
  [4460] = {.entry = {.count = 1, .reusable = true}}, SHIFT(168),
  [4462] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2140),
  [4464] = {.entry = {.count = 1, .reusable = true}}, SHIFT(902),
  [4466] = {.entry = {.count = 1, .reusable = true}}, SHIFT(689),
  [4468] = {.entry = {.count = 1, .reusable = true}}, SHIFT(341),
  [4470] = {.entry = {.count = 1, .reusable = true}}, SHIFT(176),
  [4472] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2294),
  [4474] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1311),
  [4476] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2391),
  [4478] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_parameters_repeat1, 2),
  [4480] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1734),
  [4482] = {.entry = {.count = 1, .reusable = true}}, SHIFT(549),
  [4484] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1733),
  [4486] = {.entry = {.count = 1, .reusable = true}}, SHIFT(550),
  [4488] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1864),
  [4490] = {.entry = {.count = 1, .reusable = false}}, SHIFT(565),
  [4492] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_where_clause_repeat1, 2),
  [4494] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_where_clause_repeat1, 2), SHIFT_REPEAT(1168),
  [4497] = {.entry = {.count = 1, .reusable = true}}, SHIFT(43),
  [4499] = {.entry = {.count = 1, .reusable = true}}, SHIFT(950),
  [4501] = {.entry = {.count = 1, .reusable = true}}, SHIFT(632),
  [4503] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1314),
  [4505] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1849),
  [4507] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1376),
  [4509] = {.entry = {.count = 1, .reusable = true}}, SHIFT(629),
  [4511] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_type_parameters_repeat1, 2), REDUCE(aux_sym_for_lifetimes_repeat1, 2),
  [4514] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1860),
  [4516] = {.entry = {.count = 1, .reusable = false}}, SHIFT(563),
  [4518] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1739),
  [4520] = {.entry = {.count = 1, .reusable = true}}, SHIFT(551),
  [4522] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2099),
  [4524] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_string_literal_repeat1, 2),
  [4526] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_string_literal_repeat1, 2), SHIFT_REPEAT(1839),
  [4529] = {.entry = {.count = 1, .reusable = true}}, SHIFT(479),
  [4531] = {.entry = {.count = 1, .reusable = true}}, SHIFT(121),
  [4533] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2173),
  [4535] = {.entry = {.count = 1, .reusable = true}}, SHIFT(738),
  [4537] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_parameter, 4, .production_id = 112),
  [4539] = {.entry = {.count = 1, .reusable = false}}, SHIFT(1859),
  [4541] = {.entry = {.count = 1, .reusable = false}}, SHIFT(577),
  [4543] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1313),
  [4545] = {.entry = {.count = 1, .reusable = true}}, SHIFT(301),
  [4547] = {.entry = {.count = 1, .reusable = true}}, SHIFT(716),
  [4549] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1118),
  [4551] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1793),
  [4553] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1328),
  [4555] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2049),
  [4557] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1509),
  [4559] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2249),
  [4561] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1391),
  [4563] = {.entry = {.count = 1, .reusable = true}}, SHIFT(412),
  [4565] = {.entry = {.count = 1, .reusable = true}}, SHIFT(582),
  [4567] = {.entry = {.count = 1, .reusable = true}}, SHIFT(577),
  [4569] = {.entry = {.count = 1, .reusable = true}}, SHIFT(358),
  [4571] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1423),
  [4573] = {.entry = {.count = 1, .reusable = true}}, SHIFT(583),
  [4575] = {.entry = {.count = 1, .reusable = true}}, SHIFT(383),
  [4577] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_self_parameter, 4),
  [4579] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_ordered_field_declaration_list_repeat1, 4, .production_id = 103),
  [4581] = {.entry = {.count = 1, .reusable = true}}, SHIFT(89),
  [4583] = {.entry = {.count = 1, .reusable = true}}, SHIFT(398),
  [4585] = {.entry = {.count = 1, .reusable = true}}, SHIFT(107),
  [4587] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_pattern, 5, .production_id = 221),
  [4589] = {.entry = {.count = 1, .reusable = true}}, SHIFT(565),
  [4591] = {.entry = {.count = 1, .reusable = true}}, SHIFT(437),
  [4593] = {.entry = {.count = 1, .reusable = true}}, SHIFT(439),
  [4595] = {.entry = {.count = 1, .reusable = true}}, SHIFT(117),
  [4597] = {.entry = {.count = 1, .reusable = true}}, SHIFT(458),
  [4599] = {.entry = {.count = 1, .reusable = true}}, SHIFT(474),
  [4601] = {.entry = {.count = 1, .reusable = true}}, SHIFT(472),
  [4603] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration, 4, .production_id = 199),
  [4605] = {.entry = {.count = 1, .reusable = true}}, SHIFT(416),
  [4607] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1502),
  [4609] = {.entry = {.count = 1, .reusable = true}}, SHIFT(707),
  [4611] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1729),
  [4613] = {.entry = {.count = 1, .reusable = true}}, SHIFT(548),
  [4615] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_ordered_field_declaration_list_repeat1, 3, .production_id = 61),
  [4617] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2532),
  [4619] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2312),
  [4621] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2559),
  [4623] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1459),
  [4625] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 4, .production_id = 190),
  [4627] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_pattern, 4, .production_id = 189),
  [4629] = {.entry = {.count = 1, .reusable = true}}, SHIFT(429),
  [4631] = {.entry = {.count = 1, .reusable = true}}, SHIFT(324),
  [4633] = {.entry = {.count = 1, .reusable = true}}, SHIFT(756),
  [4635] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1648),
  [4637] = {.entry = {.count = 1, .reusable = true}}, SHIFT(979),
  [4639] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1504),
  [4641] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2524),
  [4643] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2298),
  [4645] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2554),
  [4647] = {.entry = {.count = 1, .reusable = true}}, SHIFT(399),
  [4649] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2469),
  [4651] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2155),
  [4653] = {.entry = {.count = 1, .reusable = true}}, SHIFT(451),
  [4655] = {.entry = {.count = 1, .reusable = true}}, SHIFT(580),
  [4657] = {.entry = {.count = 1, .reusable = true}}, SHIFT(581),
  [4659] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1430),
  [4661] = {.entry = {.count = 1, .reusable = true}}, SHIFT(363),
  [4663] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_use_list, 5),
  [4665] = {.entry = {.count = 1, .reusable = true}}, SHIFT(353),
  [4667] = {.entry = {.count = 1, .reusable = true}}, SHIFT(328),
  [4669] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1489),
  [4671] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_field_declaration_list_repeat1, 2),
  [4673] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_field_declaration_list_repeat1, 2), SHIFT_REPEAT(1535),
  [4676] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_shorthand_field_initializer, 1),
  [4678] = {.entry = {.count = 1, .reusable = true}}, SHIFT(147),
  [4680] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_declaration, 3, .production_id = 160),
  [4682] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_ordered_field_declaration_list_repeat1, 2, .production_id = 159),
  [4684] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_ordered_field_declaration_list_repeat1, 2, .production_id = 159), SHIFT_REPEAT(555),
  [4687] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_ordered_field_declaration_list_repeat1, 2, .production_id = 22),
  [4689] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1705),
  [4691] = {.entry = {.count = 1, .reusable = true}}, SHIFT(553),
  [4693] = {.entry = {.count = 1, .reusable = true}}, SHIFT(315),
  [4695] = {.entry = {.count = 1, .reusable = true}}, SHIFT(140),
  [4697] = {.entry = {.count = 1, .reusable = true}}, SHIFT(747),
  [4699] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2464),
  [4701] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2150),
  [4703] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2465),
  [4705] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_pattern, 1, .production_id = 57),
  [4707] = {.entry = {.count = 1, .reusable = true}}, SHIFT(593),
  [4709] = {.entry = {.count = 1, .reusable = true}}, SHIFT(698),
  [4711] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1072),
  [4713] = {.entry = {.count = 1, .reusable = true}}, SHIFT(279),
  [4715] = {.entry = {.count = 1, .reusable = true}}, SHIFT(295),
  [4717] = {.entry = {.count = 1, .reusable = true}}, SHIFT(274),
  [4719] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_const_parameter, 4, .production_id = 92),
  [4721] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_for_lifetimes_repeat1, 2), SHIFT_REPEAT(2295),
  [4724] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_for_lifetimes_repeat1, 2),
  [4726] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1301),
  [4728] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1444),
  [4730] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1604),
  [4732] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_type_arguments_repeat1, 2), SHIFT_REPEAT(236),
  [4735] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_type_arguments_repeat1, 2),
  [4737] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_scoped_use_list, 2, .production_id = 29),
  [4739] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_type_binding, 3, .production_id = 141),
  [4741] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_struct_pattern_repeat1, 2),
  [4743] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_struct_pattern_repeat1, 2), SHIFT_REPEAT(1695),
  [4746] = {.entry = {.count = 1, .reusable = true}}, SHIFT(895),
  [4748] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_pattern, 3, .production_id = 140),
  [4750] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_pattern, 3, .production_id = 139),
  [4752] = {.entry = {.count = 1, .reusable = true}}, SHIFT(609),
  [4754] = {.entry = {.count = 1, .reusable = true}}, SHIFT(265),
  [4756] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2495),
  [4758] = {.entry = {.count = 1, .reusable = true}}, SHIFT(184),
  [4760] = {.entry = {.count = 1, .reusable = true}}, SHIFT(858),
  [4762] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_parameters_repeat1, 2), SHIFT_REPEAT(190),
  [4765] = {.entry = {.count = 1, .reusable = true}}, SHIFT(925),
  [4767] = {.entry = {.count = 1, .reusable = true}}, SHIFT(938),
  [4769] = {.entry = {.count = 1, .reusable = true}}, SHIFT(122),
  [4771] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2039),
  [4773] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1289),
  [4775] = {.entry = {.count = 1, .reusable = true}}, SHIFT(944),
  [4777] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1420),
  [4779] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1607),
  [4781] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2272),
  [4783] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1027),
  [4785] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1053),
  [4787] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1512),
  [4789] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_variant, 3, .production_id = 28),
  [4791] = {.entry = {.count = 1, .reusable = true}}, SHIFT(101),
  [4793] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_enum_variant_list_repeat2, 2),
  [4795] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_enum_variant_list_repeat2, 2), SHIFT_REPEAT(1551),
  [4798] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2434),
  [4800] = {.entry = {.count = 1, .reusable = true}}, SHIFT(637),
  [4802] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_use_list, 2),
  [4804] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1986),
  [4806] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1302),
  [4808] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1109),
  [4810] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constrained_type_parameter, 2, .production_id = 63),
  [4812] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1485),
  [4814] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_constrained_type_parameter, 2, .production_id = 64),
  [4816] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_macro_rule, 3, .production_id = 44),
  [4818] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_array_expression_repeat1, 2), SHIFT_REPEAT(104),
  [4821] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1125),
  [4823] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2257),
  [4825] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2258),
  [4827] = {.entry = {.count = 1, .reusable = true}}, SHIFT(38),
  [4829] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_arguments_repeat1, 2), SHIFT_REPEAT(52),
  [4832] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1103),
  [4834] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1098),
  [4836] = {.entry = {.count = 1, .reusable = true}}, SHIFT(750),
  [4838] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_pattern, 1),
  [4840] = {.entry = {.count = 1, .reusable = true}}, SHIFT(36),
  [4842] = {.entry = {.count = 1, .reusable = true}}, SHIFT(984),
  [4844] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1092),
  [4846] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1487),
  [4848] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1089),
  [4850] = {.entry = {.count = 1, .reusable = true}}, SHIFT(124),
  [4852] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1303),
  [4854] = {.entry = {.count = 1, .reusable = true}}, SHIFT(257),
  [4856] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1077),
  [4858] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1075),
  [4860] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1062),
  [4862] = {.entry = {.count = 1, .reusable = true}}, SHIFT(284),
  [4864] = {.entry = {.count = 1, .reusable = true}}, SHIFT(113),
  [4866] = {.entry = {.count = 1, .reusable = true}}, SHIFT(286),
  [4868] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1187),
  [4870] = {.entry = {.count = 1, .reusable = true}}, SHIFT(189),
  [4872] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_field_initializer_list_repeat1, 2),
  [4874] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_field_initializer_list_repeat1, 2), SHIFT_REPEAT(1548),
  [4877] = {.entry = {.count = 1, .reusable = true}}, SHIFT(251),
  [4879] = {.entry = {.count = 1, .reusable = true}}, SHIFT(587),
  [4881] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2292),
  [4883] = {.entry = {.count = 1, .reusable = true}}, SHIFT(40),
  [4885] = {.entry = {.count = 1, .reusable = true}}, SHIFT(468),
  [4887] = {.entry = {.count = 1, .reusable = true}}, SHIFT(126),
  [4889] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_use_list_repeat1, 2),
  [4891] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_use_list_repeat1, 2), SHIFT_REPEAT(1296),
  [4894] = {.entry = {.count = 1, .reusable = true}}, SHIFT(618),
  [4896] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2159),
  [4898] = {.entry = {.count = 1, .reusable = true}}, SHIFT(46),
  [4900] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_use_list, 4),
  [4902] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1513),
  [4904] = {.entry = {.count = 1, .reusable = true}}, SHIFT(303),
  [4906] = {.entry = {.count = 1, .reusable = true}}, SHIFT(461),
  [4908] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1495),
  [4910] = {.entry = {.count = 1, .reusable = true}}, SHIFT(157),
  [4912] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1386),
  [4914] = {.entry = {.count = 1, .reusable = true}}, SHIFT(452),
  [4916] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1508),
  [4918] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1320),
  [4920] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1019),
  [4922] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1323),
  [4924] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1014),
  [4926] = {.entry = {.count = 1, .reusable = true}}, SHIFT(130),
  [4928] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1010),
  [4930] = {.entry = {.count = 1, .reusable = true}}, SHIFT(470),
  [4932] = {.entry = {.count = 1, .reusable = true}}, SHIFT(108),
  [4934] = {.entry = {.count = 1, .reusable = true}}, SHIFT(949),
  [4936] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1719),
  [4938] = {.entry = {.count = 1, .reusable = true}}, SHIFT(552),
  [4940] = {.entry = {.count = 1, .reusable = true}}, SHIFT(329),
  [4942] = {.entry = {.count = 1, .reusable = true}}, SHIFT(164),
  [4944] = {.entry = {.count = 1, .reusable = true}}, SHIFT(994),
  [4946] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1518),
  [4948] = {.entry = {.count = 1, .reusable = true}}, SHIFT(928),
  [4950] = {.entry = {.count = 1, .reusable = true}}, SHIFT(962),
  [4952] = {.entry = {.count = 1, .reusable = true}}, SHIFT(133),
  [4954] = {.entry = {.count = 1, .reusable = true}}, SHIFT(343),
  [4956] = {.entry = {.count = 1, .reusable = true}}, SHIFT(431),
  [4958] = {.entry = {.count = 1, .reusable = true}}, SHIFT(349),
  [4960] = {.entry = {.count = 1, .reusable = true}}, SHIFT(912),
  [4962] = {.entry = {.count = 1, .reusable = true}}, SHIFT(954),
  [4964] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_optional_type_parameter, 3, .production_id = 102),
  [4966] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_use_list, 3),
  [4968] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_type_parameters_repeat1, 2), SHIFT_REPEAT(1556),
  [4971] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1295),
  [4973] = {.entry = {.count = 1, .reusable = true}}, SHIFT(923),
  [4975] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_use_as_clause, 3, .production_id = 77),
  [4977] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_optional_type_parameter, 3, .production_id = 101),
  [4979] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1930),
  [4981] = {.entry = {.count = 1, .reusable = true}}, SHIFT(232),
  [4983] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2227),
  [4985] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2245),
  [4987] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_use_wildcard, 3, .production_id = 1),
  [4989] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_tuple_type_repeat1, 2),
  [4991] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_tuple_type_repeat1, 2), SHIFT_REPEAT(673),
  [4994] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1589),
  [4996] = {.entry = {.count = 1, .reusable = true}}, SHIFT(889),
  [4998] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_scoped_use_list, 3, .production_id = 78),
  [5000] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1574),
  [5002] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_use_as_clause, 3, .production_id = 79),
  [5004] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_field_pattern, 2, .production_id = 97),
  [5006] = {.entry = {.count = 1, .reusable = true}}, SHIFT(588),
  [5008] = {.entry = {.count = 1, .reusable = true}}, SHIFT(235),
  [5010] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_use_wildcard, 3),
  [5012] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_scoped_use_list, 3, .production_id = 80),
  [5014] = {.entry = {.count = 1, .reusable = true}}, SHIFT(566),
  [5016] = {.entry = {.count = 1, .reusable = true}}, SHIFT(369),
  [5018] = {.entry = {.count = 1, .reusable = true}}, SHIFT(810),
  [5020] = {.entry = {.count = 1, .reusable = true}}, SHIFT(367),
  [5022] = {.entry = {.count = 1, .reusable = true}}, SHIFT(880),
  [5024] = {.entry = {.count = 1, .reusable = true}}, SHIFT(174),
  [5026] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_where_predicate, 2, .production_id = 63),
  [5028] = {.entry = {.count = 1, .reusable = true}}, SHIFT(591),
  [5030] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2241),
  [5032] = {.entry = {.count = 1, .reusable = true}}, SHIFT(44),
  [5034] = {.entry = {.count = 1, .reusable = true}}, SHIFT(878),
  [5036] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1490),
  [5038] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_where_predicate, 2, .production_id = 64),
  [5040] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1503),
  [5042] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_use_wildcard, 1),
  [5044] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_enum_variant, 2, .production_id = 9),
  [5046] = {.entry = {.count = 1, .reusable = true}}, SHIFT(156),
  [5048] = {.entry = {.count = 1, .reusable = true}}, SHIFT(764),
  [5050] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1494),
  [5052] = {.entry = {.count = 1, .reusable = true}}, SHIFT(874),
  [5054] = {.entry = {.count = 1, .reusable = false}}, SHIFT(794),
  [5056] = {.entry = {.count = 1, .reusable = false}}, SHIFT(898),
  [5058] = {.entry = {.count = 1, .reusable = true}}, SHIFT(800),
  [5060] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2388),
  [5062] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2301),
  [5064] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2390),
  [5066] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_closure_parameters_repeat1, 2),
  [5068] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1332),
  [5070] = {.entry = {.count = 1, .reusable = true}}, SHIFT(633),
  [5072] = {.entry = {.count = 2, .reusable = true}}, REDUCE(aux_sym_closure_parameters_repeat1, 2), SHIFT_REPEAT(562),
  [5075] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1514),
  [5077] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1519),
  [5079] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1522),
  [5081] = {.entry = {.count = 1, .reusable = true}}, SHIFT(861),
  [5083] = {.entry = {.count = 1, .reusable = true}}, SHIFT(826),
  [5085] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_shorthand_field_initializer, 2),
  [5087] = {.entry = {.count = 1, .reusable = true}}, SHIFT(145),
  [5089] = {.entry = {.count = 1, .reusable = true}}, SHIFT(836),
  [5091] = {.entry = {.count = 1, .reusable = true}}, SHIFT(408),
  [5093] = {.entry = {.count = 1, .reusable = true}}, SHIFT(746),
  [5095] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2357),
  [5097] = {.entry = {.count = 1, .reusable = true}}, SHIFT(851),
  [5099] = {.entry = {.count = 1, .reusable = true}}, SHIFT(159),
  [5101] = {.entry = {.count = 1, .reusable = true}}, SHIFT(849),
  [5103] = {.entry = {.count = 1, .reusable = true}}, SHIFT(763),
  [5105] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1498),
  [5107] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym__condition, 1, .production_id = 3),
  [5109] = {.entry = {.count = 1, .reusable = true}}, SHIFT(82),
  [5111] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1184),
  [5113] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1185),
  [5115] = {.entry = {.count = 1, .reusable = true}}, SHIFT(844),
  [5117] = {.entry = {.count = 1, .reusable = true}}, SHIFT(80),
  [5119] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1752),
  [5121] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2408),
  [5123] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2407),
  [5125] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1835),
  [5127] = {.entry = {.count = 1, .reusable = true}}, SHIFT(401),
  [5129] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2510),
  [5131] = {.entry = {.count = 1, .reusable = true}}, SHIFT(151),
  [5133] = {.entry = {.count = 1, .reusable = true}}, SHIFT(819),
  [5135] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2468),
  [5137] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_qualified_type, 3, .production_id = 84),
  [5139] = {.entry = {.count = 1, .reusable = true}}, SHIFT(814),
  [5141] = {.entry = {.count = 1, .reusable = true}}, SHIFT(812),
  [5143] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2516),
  [5145] = {.entry = {.count = 1, .reusable = true}}, SHIFT(392),
  [5147] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2523),
  [5149] = {.entry = {.count = 1, .reusable = true}}, SHIFT(387),
  [5151] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2018),
  [5153] = {.entry = {.count = 1, .reusable = true}}, SHIFT(893),
  [5155] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2449),
  [5157] = {.entry = {.count = 1, .reusable = true}}, SHIFT(649),
  [5159] = {.entry = {.count = 1, .reusable = true}}, SHIFT(990),
  [5161] = {.entry = {.count = 1, .reusable = true}}, SHIFT(440),
  [5163] = {.entry = {.count = 1, .reusable = true}}, SHIFT(173),
  [5165] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1021),
  [5167] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1554),
  [5169] = {.entry = {.count = 1, .reusable = true}}, SHIFT(116),
  [5171] = {.entry = {.count = 1, .reusable = true}}, SHIFT(127),
  [5173] = {.entry = {.count = 1, .reusable = true}}, SHIFT(307),
  [5175] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1393),
  [5177] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1372),
  [5179] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1833),
  [5181] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1848),
  [5183] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1894),
  [5185] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1371),
  [5187] = {.entry = {.count = 1, .reusable = true}}, SHIFT(275),
  [5189] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1605),
  [5191] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1610),
  [5193] = {.entry = {.count = 1, .reusable = true}}, SHIFT(640),
  [5195] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1112),
  [5197] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1852),
  [5199] = {.entry = {.count = 1, .reusable = true}}, SHIFT(740),
  [5201] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1309),
  [5203] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2525),
  [5205] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1597),
  [5207] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1598),
  [5209] = {.entry = {.count = 1, .reusable = true}}, SHIFT(125),
  [5211] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1775),
  [5213] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1587),
  [5215] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1590),
  [5217] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1602),
  [5219] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1594),
  [5221] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1308),
  [5223] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2453),
  [5225] = {.entry = {.count = 1, .reusable = true}}, SHIFT(319),
  [5227] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2377),
  [5229] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2057),
  [5231] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2474),
  [5233] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1810),
  [5235] = {.entry = {.count = 1, .reusable = true}}, SHIFT(144),
  [5237] = {.entry = {.count = 1, .reusable = true}}, SHIFT(111),
  [5239] = {.entry = {.count = 1, .reusable = true}}, SHIFT(378),
  [5241] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_variadic_parameter, 1),
  [5243] = {.entry = {.count = 1, .reusable = true}}, SHIFT(103),
  [5245] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_enum_variant_list_repeat2, 3),
  [5247] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1808),
  [5249] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2400),
  [5251] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2446),
  [5253] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1777),
  [5255] = {.entry = {.count = 1, .reusable = true}}, REDUCE(aux_sym_field_declaration_list_repeat1, 3),
  [5257] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2539),
  [5259] = {.entry = {.count = 1, .reusable = false}}, SHIFT(2562),
  [5261] = {.entry = {.count = 1, .reusable = true}}, SHIFT(336),
  [5263] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1784),
  [5265] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1431),
  [5267] = {.entry = {.count = 1, .reusable = true}}, SHIFT(297),
  [5269] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1433),
  [5271] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1080),
  [5273] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1449),
  [5275] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1068),
  [5277] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2129),
  [5279] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1445),
  [5281] = {.entry = {.count = 1, .reusable = true}}, SHIFT(357),
  [5283] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1534),
  [5285] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1628),
  [5287] = {.entry = {.count = 1, .reusable = true}}, SHIFT(216),
  [5289] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1461),
  [5291] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1606),
  [5293] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1457),
  [5295] = {.entry = {.count = 1, .reusable = true}}, SHIFT(222),
  [5297] = {.entry = {.count = 1, .reusable = true}}, SHIFT(751),
  [5299] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2381),
  [5301] = {.entry = {.count = 1, .reusable = true}}, SHIFT(728),
  [5303] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1163),
  [5305] = {.entry = {.count = 1, .reusable = true}}, SHIFT(755),
  [5307] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1424),
  [5309] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1419),
  [5311] = {.entry = {.count = 1, .reusable = true}}, SHIFT(762),
  [5313] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1435),
  [5315] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1015),
  [5317] = {.entry = {.count = 1, .reusable = true}}, SHIFT(761),
  [5319] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1667),
  [5321] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_bracketed_type, 3),
  [5323] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1982),
  [5325] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1492),
  [5327] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1533),
  [5329] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1641),
  [5331] = {.entry = {.count = 1, .reusable = true}}, SHIFT(318),
  [5333] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1330),
  [5335] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2056),
  [5337] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2057),
  [5339] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1063),
  [5341] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2044),
  [5343] = {.entry = {.count = 1, .reusable = true}}, SHIFT(907),
  [5345] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2494),
  [5347] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2484),
  [5349] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1085),
  [5351] = {.entry = {.count = 1, .reusable = true}}, SHIFT(697),
  [5353] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1120),
  [5355] = {.entry = {.count = 1, .reusable = true}}, SHIFT(476),
  [5357] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1110),
  [5359] = {.entry = {.count = 1, .reusable = true}}, SHIFT(77),
  [5361] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2179),
  [5363] = {.entry = {.count = 1, .reusable = true}}, SHIFT(123),
  [5365] = {.entry = {.count = 1, .reusable = true}}, SHIFT(681),
  [5367] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1088),
  [5369] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2400),
  [5371] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1360),
  [5373] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1086),
  [5375] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1532),
  [5377] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2237),
  [5379] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2162),
  [5381] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1632),
  [5383] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2053),
  [5385] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1486),
  [5387] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1638),
  [5389] = {.entry = {.count = 1, .reusable = true}}, SHIFT(731),
  [5391] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1700),
  [5393] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1737),
  [5395] = {.entry = {.count = 1, .reusable = true}}, SHIFT(735),
  [5397] = {.entry = {.count = 1, .reusable = true}}, SHIFT(955),
  [5399] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1043),
  [5401] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2483),
  [5403] = {.entry = {.count = 1, .reusable = true}}, SHIFT(456),
  [5405] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2131),
  [5407] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1318),
  [5409] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1340),
  [5411] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1018),
  [5413] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1006),
  [5415] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1322),
  [5417] = {.entry = {.count = 1, .reusable = true}}, SHIFT(436),
  [5419] = {.entry = {.count = 1, .reusable = true}}, SHIFT(981),
  [5421] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1555),
  [5423] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2135),
  [5425] = {.entry = {.count = 1, .reusable = true}}, SHIFT(961),
  [5427] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1675),
  [5429] = {.entry = {.count = 1, .reusable = true}}, SHIFT(960),
  [5431] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute, 2, .production_id = 81),
  [5433] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_attribute, 2, .production_id = 82),
  [5435] = {.entry = {.count = 1, .reusable = true}}, SHIFT(419),
  [5437] = {.entry = {.count = 1, .reusable = true}}, SHIFT(772),
  [5439] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2303),
  [5441] = {.entry = {.count = 1, .reusable = true}}, SHIFT(841),
  [5443] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1881),
  [5445] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1496),
  [5447] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1892),
  [5449] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1676),
  [5451] = {.entry = {.count = 1, .reusable = true}}, SHIFT(831),
  [5453] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2447),
  [5455] = {.entry = {.count = 1, .reusable = true}}, SHIFT(658),
  [5457] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2418),
  [5459] = {.entry = {.count = 1, .reusable = true}}, SHIFT(850),
  [5461] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2268),
  [5463] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1307),
  [5465] = {.entry = {.count = 1, .reusable = true}}, SHIFT(852),
  [5467] = {.entry = {.count = 1, .reusable = true}}, SHIFT(853),
  [5469] = {.entry = {.count = 1, .reusable = true}}, SHIFT(734),
  [5471] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2365),
  [5473] = {.entry = {.count = 1, .reusable = true}}, SHIFT(409),
  [5475] = {.entry = {.count = 1, .reusable = true}}, SHIFT(482),
  [5477] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1563),
  [5479] = {.entry = {.count = 1, .reusable = true}}, SHIFT(706),
  [5481] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2388),
  [5483] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2382),
  [5485] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2262),
  [5487] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2204),
  [5489] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1940),
  [5491] = {.entry = {.count = 1, .reusable = true}}, SHIFT(884),
  [5493] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2201),
  [5495] = {.entry = {.count = 1, .reusable = true}}, SHIFT(732),
  [5497] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1343),
  [5499] = {.entry = {.count = 1, .reusable = true}}, SHIFT(903),
  [5501] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1180),
  [5503] = {.entry = {.count = 1, .reusable = true}}, SHIFT(178),
  [5505] = {.entry = {.count = 1, .reusable = true}}, SHIFT(85),
  [5507] = {.entry = {.count = 1, .reusable = true}}, SHIFT(647),
  [5509] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1321),
  [5511] = {.entry = {.count = 1, .reusable = true}}, SHIFT(789),
  [5513] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2414),
  [5515] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1317),
  [5517] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2344),
  [5519] = {.entry = {.count = 1, .reusable = true}}, SHIFT(309),
  [5521] = {.entry = {.count = 1, .reusable = true}}, SHIFT(289),
  [5523] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1189),
  [5525] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1324),
  [5527] = {.entry = {.count = 1, .reusable = true}}, SHIFT(660),
  [5529] = {.entry = {.count = 1, .reusable = true}}, SHIFT(766),
  [5531] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1305),
  [5533] = {.entry = {.count = 1, .reusable = true}}, SHIFT(370),
  [5535] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2320),
  [5537] = {.entry = {.count = 1, .reusable = true}}, SHIFT(975),
  [5539] = {.entry = {.count = 1, .reusable = true}}, SHIFT(281),
  [5541] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2337),
  [5543] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2551),
  [5545] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2351),
  [5547] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1170),
  [5549] = {.entry = {.count = 1, .reusable = true}}, SHIFT(248),
  [5551] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2003),
  [5553] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1306),
  [5555] = {.entry = {.count = 1, .reusable = true}}, SHIFT(650),
  [5557] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1325),
  [5559] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2072),
  [5561] = {.entry = {.count = 1, .reusable = true}}, SHIFT(247),
  [5563] = {.entry = {.count = 1, .reusable = true}}, SHIFT(655),
  [5565] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1304),
  [5567] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1186),
  [5569] = {.entry = {.count = 1, .reusable = true}},  ACCEPT_INPUT(),
  [5571] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1181),
  [5573] = {.entry = {.count = 1, .reusable = true}}, SHIFT(683),
  [5575] = {.entry = {.count = 1, .reusable = true}}, SHIFT(769),
  [5577] = {.entry = {.count = 1, .reusable = true}}, SHIFT(198),
  [5579] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1635),
  [5581] = {.entry = {.count = 1, .reusable = true}}, SHIFT(696),
  [5583] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1691),
  [5585] = {.entry = {.count = 1, .reusable = true}}, SHIFT(312),
  [5587] = {.entry = {.count = 1, .reusable = true}}, SHIFT(701),
  [5589] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1531),
  [5591] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1520),
  [5593] = {.entry = {.count = 1, .reusable = true}}, SHIFT(313),
  [5595] = {.entry = {.count = 1, .reusable = true}}, SHIFT(710),
  [5597] = {.entry = {.count = 1, .reusable = true}}, REDUCE(sym_match_pattern, 3, .production_id = 156),
  [5599] = {.entry = {.count = 1, .reusable = true}}, SHIFT(713),
  [5601] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1998),
  [5603] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1310),
  [5605] = {.entry = {.count = 1, .reusable = true}}, SHIFT(134),
  [5607] = {.entry = {.count = 1, .reusable = true}}, SHIFT(258),
  [5609] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2067),
  [5611] = {.entry = {.count = 1, .reusable = true}}, SHIFT(393),
  [5613] = {.entry = {.count = 1, .reusable = true}}, SHIFT(539),
  [5615] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2532),
  [5617] = {.entry = {.count = 1, .reusable = true}}, SHIFT(449),
  [5619] = {.entry = {.count = 1, .reusable = true}}, SHIFT(376),
  [5621] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1923),
  [5623] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2271),
  [5625] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2539),
  [5627] = {.entry = {.count = 1, .reusable = true}}, SHIFT(131),
  [5629] = {.entry = {.count = 1, .reusable = true}}, SHIFT(418),
  [5631] = {.entry = {.count = 1, .reusable = true}}, SHIFT(2545),
  [5633] = {.entry = {.count = 1, .reusable = true}}, SHIFT(1693),
  [5635] = {.entry = {.count = 1, .reusable = true}}, SHIFT(784),
};

#ifdef __cplusplus
extern "C" {
#endif
void *tree_sitter_rust_external_scanner_create(void);
void tree_sitter_rust_external_scanner_destroy(void *);
bool tree_sitter_rust_external_scanner_scan(void *, TSLexer *, const bool *);
unsigned tree_sitter_rust_external_scanner_serialize(void *, char *);
void tree_sitter_rust_external_scanner_deserialize(void *, const char *, unsigned);

#ifdef _WIN32
#define extern __declspec(dllexport)
#endif

extern const TSLanguage *tree_sitter_rust(void) {
  static const TSLanguage language = {
    .version = LANGUAGE_VERSION,
    .symbol_count = SYMBOL_COUNT,
    .alias_count = ALIAS_COUNT,
    .token_count = TOKEN_COUNT,
    .external_token_count = EXTERNAL_TOKEN_COUNT,
    .state_count = STATE_COUNT,
    .large_state_count = LARGE_STATE_COUNT,
    .production_id_count = PRODUCTION_ID_COUNT,
    .field_count = FIELD_COUNT,
    .max_alias_sequence_length = MAX_ALIAS_SEQUENCE_LENGTH,
    .parse_table = &ts_parse_table[0][0],
    .small_parse_table = ts_small_parse_table,
    .small_parse_table_map = ts_small_parse_table_map,
    .parse_actions = ts_parse_actions,
    .symbol_names = ts_symbol_names,
    .field_names = ts_field_names,
    .field_map_slices = ts_field_map_slices,
    .field_map_entries = ts_field_map_entries,
    .symbol_metadata = ts_symbol_metadata,
    .public_symbol_map = ts_symbol_map,
    .alias_map = ts_non_terminal_alias_map,
    .alias_sequences = &ts_alias_sequences[0][0],
    .lex_modes = ts_lex_modes,
    .lex_fn = ts_lex,
    .keyword_lex_fn = ts_lex_keywords,
    .keyword_capture_token = sym_identifier,
    .external_scanner = {
      &ts_external_scanner_states[0][0],
      ts_external_scanner_symbol_map,
      tree_sitter_rust_external_scanner_create,
      tree_sitter_rust_external_scanner_destroy,
      tree_sitter_rust_external_scanner_scan,
      tree_sitter_rust_external_scanner_serialize,
      tree_sitter_rust_external_scanner_deserialize,
    },
    .primary_state_ids = ts_primary_state_ids,
  };
  return &language;
}
#ifdef __cplusplus
}
#endif
